#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "grid.h"
#ifdef USE_MPI
#include <mpi.h>
#endif

/** \example test_read_icon.c
 * This contains examples for read_icon_grid.
 */

/**
 * reads in an icon grid netcdf file and generates a struct %grid from it
 * @param[in] filename name of the icon grid netcdf file
 * @returns %grid structure that contains the icon grid
 */
struct grid * read_icon_grid(char * filename);

/**
 * reads in an icon grid netcdf file and return the grid information in
 * a format that is supported by the YAC user interface.
 * @param[in]  filename              name of the icon grid netcdf file
 * @param[out] nbr_vertices          number of vertices in the grid
 * @param[out] nbr_cells             number of cells in the grid
 * @param[out] num_vertices_per_cell number of vertices per cell
 * @param[out] cell_to_vertex        vertex indices for each cell
 * @param[out] x_vertices            longitudes of vertices
 * @param[out] y_vertices            latitudes of vertices
 * @param[out] x_cells               longitudes of cell center
 * @param[out] y_cells               latitudes of cell center
 * @param[out] cell_mask             mask for cells
 */
void read_icon_grid_information(const char * filename, int * nbr_vertices,
                                int * nbr_cells, int ** num_vertices_per_cell,
                                int ** cell_to_vertex, double ** x_vertices,
                                double ** y_vertices, double ** x_cells,
                                double ** y_cells, int ** cell_mask);

/**
 * reads in an icon grid netcdf file and return the grid information in
 * a format that is supported by the YAC user interface. The reading is done
 * in parallel and a basic domain decomposition is applied.
 * @param[in]  filename              name of the icon grid netcdf file
 * @param[in]  comm                  MPI communicator containing all proceses
 *                                   that will get a part of the grid
 * @param[out] nbr_vertices          number of vertices in the local part of the
 *                                   grid
 * @param[out] nbr_cells             number of cells in the local part of the
 *                                   grid
 * @param[out] num_vertices_per_cell number of vertices per cell
 * @param[out] cell_to_vertex        vertex indices for each cell
 * @param[out] global_cell_ids       global ids of local cells (core and halo
 *                                   cells)
 * @param[out] cell_owner            owner of each cell (locally owned cells
 *                                   are marked with -1)
 * @param[out] global_vertex_ids     global ids of local vertices (core and halo
 *                                   vertices)
 * @param[out] vertex_owner          owner of each vertex (locally owned
 *                                   vertices are marked with -1)
 * @param[out] x_vertices            longitudes of vertices
 * @param[out] y_vertices            latitudes of vertices
 * @param[out] x_cells               longitudes of cell center
 * @param[out] y_cells               latitudes of cell center
 * @param[out] cell_mask             mask for cells
 * @param[in]  io_proc_step          every io_proc_step's process will read a
 *                                   part of the file
 * @remark in case (io_proc_step >= comm_size) => io_proc_step = sqrt(comm_size)
 */
#ifdef USE_MPI
void read_icon_grid_information_parallel(
  const char * filename, MPI_Comm comm, int * nbr_vertices, int * nbr_cells,
  int ** num_vertices_per_cell, int ** cell_to_vertex, int ** global_cell_ids,
  int ** cell_owner, int ** global_vertex_ids, int ** vertex_owner,
  double ** x_vertices, double ** y_vertices, double ** x_cells,
  double ** y_cells, int ** cell_mask, unsigned io_proc_step);
#endif