#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#ifdef HAVE_C_NETCDF
#include <netcdf.h>
#include <stdint.h>

#include "grid.h"
#include "geometry.h"
#include "points.h"
#include "grid_unstruct.h"
#include "dep_list.h"

static void open_mpiom_file    ( int *ncid, const char *filename);

static void close_mpiom_file   ( int ncid );

static void get_mpiom_vertices ( int ncid,
                                 int *vertex_type,  // data type for vertex coordinates
                                 void **vertex_lon, // longitude of vertex coordinates
                                 void **vertex_lat, // latitude of vertex coordinates
                                 size_t *nbr_vertices );

static void get_mpiom_cell_center ( int ncid,
                                    int *cell_type,  // data type for cell coordinates
                                    void **cell_lon, // longitude of cell coordinates
                                    void **cell_lat, // latitude of cell coordinates
                                    size_t *nbr_cells );

static void get_mpiom_cell_mask ( int ncid,
                                  int *cell_type,  // data type for mask
                                  int **cell_mask, // data for mask
                                  size_t *nbr_cells );

void remove_duplicated_vertices(void * temp_vertex_lon, void * temp_vertex_lat,
                                size_t * temp_nbr_vertices, int * old_to_new_id,
                                int vertex_type);

void remove_duplicated_cells ( void * temp_cell_lon, void * temp_cell_lat,
                               int * temp_cell_mask,
                               size_t * temp_nbr_cells, int * old_to_new_id,
                               int cell_type);

static void handle_error(int status);

void read_mpiom_grid_information(const char * filename, int * nbr_vertices,
                                 int * nbr_cells, int ** num_vertices_per_cell,
                                 int ** cell_to_vertex, double ** x_vertices,
                                 double ** y_vertices, double ** x_cells,
                                 double ** y_cells, int ** cell_mask) {

   int ncid;

   /* Open file */

   open_mpiom_file ( &ncid, filename );

   /* Get vertex longitudes and latitudes of cells and
   relations between vertices and cells*/

   {
      void * temp_vertex_lon;
      void * temp_vertex_lat;
      size_t temp_nbr_vertices;
      int vertex_type;

      get_mpiom_vertices (ncid, &vertex_type, &temp_vertex_lon,
                          &temp_vertex_lat, &temp_nbr_vertices);

      *nbr_cells = temp_nbr_vertices / 4;

      int * old_to_new_id = malloc(temp_nbr_vertices * sizeof(*old_to_new_id));

      remove_duplicated_vertices(temp_vertex_lon, temp_vertex_lat,
                                 &temp_nbr_vertices, old_to_new_id,
                                 vertex_type);

      *nbr_vertices = temp_nbr_vertices;

      if (vertex_type == NC_FLOAT) {

         *x_vertices = malloc(temp_nbr_vertices * sizeof(**x_vertices));
         *y_vertices = malloc(temp_nbr_vertices * sizeof(**y_vertices));

         for (int i = 0; i < temp_nbr_vertices; ++i) {

            (*x_vertices)[i] = ((float*)temp_vertex_lon)[i];
            (*y_vertices)[i] = ((float*)temp_vertex_lat)[i];
         }
         free(temp_vertex_lon);
         free(temp_vertex_lat);
      } else {
         *x_vertices = realloc(temp_vertex_lon,
               temp_nbr_vertices * sizeof(double));
         *y_vertices = realloc(temp_vertex_lat,
               temp_nbr_vertices * sizeof(double));
      }

      *num_vertices_per_cell = malloc(*nbr_cells * sizeof(**num_vertices_per_cell));
      for (int i = 0; i < *nbr_cells; ++i)
         (*num_vertices_per_cell)[i] = 4;

      *cell_to_vertex = malloc(*nbr_cells * 4 * sizeof(**cell_to_vertex));

      // Unfortunately the data is only available in Fortran order
      for (int i = 0; i < *nbr_cells; ++i) {

         (*cell_to_vertex)[4*i+0] = old_to_new_id[4*i+0];
         (*cell_to_vertex)[4*i+1] = old_to_new_id[4*i+1];
         (*cell_to_vertex)[4*i+2] = old_to_new_id[4*i+2];
         (*cell_to_vertex)[4*i+3] = old_to_new_id[4*i+3];
      }

      free(old_to_new_id);

   }

   {
      /* Get longitudes and latitudes of cell center points */

      void * temp_cell_lon;
      void * temp_cell_lat;
      int *  temp_cell_mask;

      size_t temp_nbr_cells;
      size_t temp_nbr_masks;

      int cell_type;
      int mask_type;

      get_mpiom_cell_center (ncid, &cell_type, &temp_cell_lon,
                             &temp_cell_lat, &temp_nbr_cells);

      get_mpiom_cell_mask (ncid, &mask_type, &temp_cell_mask,
                           &temp_nbr_masks);

      if ( temp_nbr_cells != temp_nbr_masks ) exit(EXIT_FAILURE);

      int * old_to_new_id = malloc(temp_nbr_cells * sizeof(*old_to_new_id));

      remove_duplicated_cells(temp_cell_lon, temp_cell_lat, temp_cell_mask,
                              &temp_nbr_cells, old_to_new_id,
                              cell_type);

      *nbr_cells = temp_nbr_cells;

      if (cell_type == NC_FLOAT) {

         *x_cells = malloc(temp_nbr_cells * sizeof(**x_cells));
         *y_cells = malloc(temp_nbr_cells * sizeof(**y_cells));

         for (int i = 0; i < temp_nbr_cells; ++i) {

            (*x_cells)[i] = ((float*)temp_cell_lon)[i];
            (*y_cells)[i] = ((float*)temp_cell_lat)[i];
         }
         free(temp_cell_lon);
         free(temp_cell_lat);
      } else {
         *x_cells = realloc(temp_cell_lon,
               temp_nbr_cells * sizeof(double));
         *y_cells = realloc(temp_cell_lat,
               temp_nbr_cells * sizeof(double));
      }

      *cell_mask = realloc(temp_cell_mask,temp_nbr_cells * sizeof(int));

   }
   close_mpiom_file ( ncid );
}

struct grid * read_mpiom_grid(char * filename) {

   int nbr_vertices;
   int nbr_cells;
   int * num_vertices_per_cell;
   int * cell_to_vertex;

   int * cell_mask;

   double * x_vertices;
   double * y_vertices;
   double * x_cells;
   double * y_cells;

   read_mpiom_grid_information(filename, &nbr_vertices, &nbr_cells,
                              &num_vertices_per_cell, &cell_to_vertex, 
                              &x_vertices, &y_vertices,
                              &x_cells, &y_cells, &cell_mask);

   struct dep_list cell_to_vertex_dep_list;

   yac_set_dependencies(&cell_to_vertex_dep_list, nbr_cells,
                        (unsigned*)num_vertices_per_cell, (unsigned*)cell_to_vertex);

   return yac_unstruct_grid_new(x_vertices, y_vertices, nbr_vertices,
                                cell_to_vertex_dep_list);
}

static void open_mpiom_file (int *ncid, const char *filename) {

  int id, status;

  /* open netCDF dataset */

  status = nc_open(filename, NC_NOWRITE, &id);
  if (status != NC_NOERR) handle_error(status);

  *ncid = id;

}

/* ---------------------------------------------------------------- */

static void close_mpiom_file (int ncid) {

   int status;

  /* close netCDF dataset */

  status = nc_close(ncid);
  if (status != NC_NOERR) handle_error(status);
}

/* ---------------------------------------------------------------- */

static void get_mpiom_vertices ( int ncid,
                                 int  *vertex_type,
                                 void **vertex_lon,
                                 void **vertex_lat,
                                 size_t *nbr_vertices ) {

  int status;

  int glat_id;                        // Various NetCDF IDs
  int glon_id;

  int glon_ndims;                     // number of dimensions in file
  int glat_ndims;

  int glon_dimids[NC_MAX_VAR_DIMS];   // dimension NetCDF IDs
  int glat_dimids[NC_MAX_VAR_DIMS];

  nc_type glon_type;                  // variable type
  nc_type glat_type;

  int nbr_atts;                       // number of attributes

  status = nc_inq_varid (ncid, "lon_bounds", &glon_id);
  if (status != NC_NOERR) {
    status = nc_inq_varid (ncid, "lon_bnds", &glon_id);
    if (status != NC_NOERR) handle_error(status);
  }

  status = nc_inq_varid (ncid, "lat_bounds", &glat_id);
  if (status != NC_NOERR) {
    status = nc_inq_varid (ncid, "lat_bnds", &glat_id);
    if (status != NC_NOERR) handle_error(status);
  }

  status = nc_inq_var (ncid, glon_id, 0, &glon_type, &glon_ndims, glon_dimids, &nbr_atts);
  if (status != NC_NOERR) handle_error(status);

  status = nc_inq_var (ncid, glat_id, 0, &glat_type, &glat_ndims, glat_dimids, &nbr_atts);
  if (status != NC_NOERR) handle_error(status);

  /* Allocate memory to read in coordinates */

  if ( glon_type != glat_type ) {
    exit(EXIT_FAILURE);
  }

  if (glon_ndims != 3 || glat_ndims != 3) {
    exit(EXIT_FAILURE);
  }

  /* get size of arrays */

  size_t dimlen_lat[3]; // dimension size for latitude
  size_t dimlen_lon[3]; // dimension size for longitude

  for (int i = 0; i < 3; ++i) {
    status = nc_inq_dimlen(ncid, glon_dimids[i], dimlen_lon + i);
    if (status != NC_NOERR) handle_error(status);
    status = nc_inq_dimlen(ncid, glat_dimids[i], dimlen_lat + i);
    if (status != NC_NOERR) handle_error(status);
    if (dimlen_lon[i] != dimlen_lat[i])
      exit(EXIT_FAILURE);
  }

  if (dimlen_lon[2] != 4)
    exit(EXIT_FAILURE);

  *nbr_vertices = dimlen_lon[0] * dimlen_lon[1] * dimlen_lon[2];

  /* read coordinates and convert radians into degrees */

  size_t start[3] = {0,0,0};

  switch(glon_type) {

  case NC_FLOAT:

    *vertex_lon = (float * ) malloc ( *nbr_vertices * sizeof ( float ) );
    *vertex_lat = (float * ) malloc ( *nbr_vertices * sizeof ( float ) );

    status = nc_get_vara (ncid, glon_id, start, dimlen_lon, (*vertex_lon));
    if (status != NC_NOERR) handle_error(status);

    status = nc_get_vara (ncid, glat_id, start, dimlen_lon, (*vertex_lat));
    if (status != NC_NOERR) handle_error(status);

    break;

  case NC_DOUBLE:

    *vertex_lon = (double * ) malloc ( *nbr_vertices * sizeof ( double ) );
    *vertex_lat = (double * ) malloc ( *nbr_vertices * sizeof ( double ) );

    status = nc_get_vara_double (ncid, glon_id, start, dimlen_lon, (*vertex_lon));
    if (status != NC_NOERR) handle_error(status);

    status = nc_get_vara_double (ncid, glat_id, start, dimlen_lon, (*vertex_lat));
    if (status != NC_NOERR) handle_error(status);

    break;

  default:

    exit(EXIT_FAILURE);

  }

  *vertex_type = glon_type;
}

/* ---------------------------------------------------------------- */

static void get_mpiom_cell_center ( int ncid,
                                    int  *cell_type,
                                    void **cell_lon,
                                    void **cell_lat,
                                    size_t *nbr_cells ) {

  int status;

  int glat_id;                        // Various NetCDF IDs
  int glon_id;

  int glon_ndims;                     // number of dimensions in file
  int glat_ndims;

  int glon_dimids[NC_MAX_VAR_DIMS];   // dimension NetCDF IDs
  int glat_dimids[NC_MAX_VAR_DIMS];

  nc_type glon_type;                  // variable type
  nc_type glat_type;

  int nbr_atts;                       // number of attributes

  status = nc_inq_varid (ncid, "lon", &glon_id);
  if (status != NC_NOERR) handle_error(status);

  status = nc_inq_varid (ncid, "lat", &glat_id);
  if (status != NC_NOERR) handle_error(status);

  status = nc_inq_var (ncid, glon_id, 0, &glon_type, &glon_ndims, glon_dimids, &nbr_atts);
  if (status != NC_NOERR) handle_error(status);

  status = nc_inq_var (ncid, glat_id, 0, &glat_type, &glat_ndims, glat_dimids, &nbr_atts);
  if (status != NC_NOERR) handle_error(status);

  /* Allocate memory to read in coordinates */

  if ( glon_type != glat_type ) {
    exit(EXIT_FAILURE);
  }

  if (glon_ndims != 2 || glat_ndims != 2) {
    exit(EXIT_FAILURE);
  }

  /* get size of arrays */

  size_t dimlen_lat[2]; // dimension size for latitude
  size_t dimlen_lon[2]; // dimension size for longitude

  for (int i = 0; i < 2; ++i) {
    status = nc_inq_dimlen(ncid, glon_dimids[i], dimlen_lon + i);
    if (status != NC_NOERR) handle_error(status);
    status = nc_inq_dimlen(ncid, glat_dimids[i], dimlen_lat + i);
    if (status != NC_NOERR) handle_error(status);
    if (dimlen_lon[i] != dimlen_lat[i])
      exit(EXIT_FAILURE);
  }

  *nbr_cells = dimlen_lon[0] * dimlen_lon[1];

  /* read coordinates and convert radians into degrees */

  size_t start[2] = {0,0};

  switch(glon_type) {

  case NC_FLOAT:

    *cell_lon = (float * ) malloc ( *nbr_cells * sizeof ( float ) );
    *cell_lat = (float * ) malloc ( *nbr_cells * sizeof ( float ) );

    status = nc_get_vara (ncid, glon_id, start, dimlen_lon, (*cell_lon));
    if (status != NC_NOERR) handle_error(status);

    status = nc_get_vara (ncid, glat_id, start, dimlen_lon, (*cell_lat));
    if (status != NC_NOERR) handle_error(status);

    break;

  case NC_DOUBLE:

    *cell_lon = (double * ) malloc ( *nbr_cells * sizeof ( double ) );
    *cell_lat = (double * ) malloc ( *nbr_cells * sizeof ( double ) );

    status = nc_get_vara_double (ncid, glon_id, start, dimlen_lon, (*cell_lon));
    if (status != NC_NOERR) handle_error(status);

    status = nc_get_vara_double (ncid, glat_id, start, dimlen_lon, (*cell_lat));
    if (status != NC_NOERR) handle_error(status);

    break;

  default:

    exit(EXIT_FAILURE);

  }

  *cell_type = glon_type;
}

/* ---------------------------------------------------------------- */

static void get_mpiom_cell_mask ( int ncid,
                                  int  *mask_type,
                                  int **cell_mask,
                                  size_t *nbr_cells ) {

  int status;

  int mask_id;                        // Various NetCDF IDs
  int mask_ndims;                     // number of dimensions in file
  int mask_dimids[NC_MAX_VAR_DIMS];   // dimension NetCDF IDs

  nc_type this_mask_type;             // variable type

  int nbr_atts;                       // number of attributes

  status = nc_inq_varid (ncid, "var1", &mask_id);
  if (status != NC_NOERR) {
      status = nc_inq_varid (ncid, "zo", &mask_id);
      if (status != NC_NOERR) handle_error(status);
  }

  status = nc_inq_var (ncid, mask_id, 0, &this_mask_type, &mask_ndims, mask_dimids, &nbr_atts);
  if (status != NC_NOERR) handle_error(status);

  /* Allocate memory to read in coordinates */

  if (mask_ndims < 3 || mask_ndims > 4) {
     printf ("ERROR in reading MPIOM mask dimension.\n");
     exit(EXIT_FAILURE);
  }

  /*  /po	et size of arrays */

  size_t * start;
  size_t * dimlen_mask;

  dimlen_mask = malloc ( mask_ndims * sizeof(*dimlen_mask)) ; // dimension size for mask

  start = (size_t * ) malloc ( mask_ndims * sizeof(size_t) ) ;
  for ( unsigned i = 0; i < mask_ndims; ++ i ) start[i] = 0;


  switch(mask_ndims) {

  case 3:

    for (int i = 0; i < mask_ndims; ++i) {
      status = nc_inq_dimlen(ncid, mask_dimids[i], dimlen_mask + i);
      if (status != NC_NOERR) handle_error(status);
    }

    *nbr_cells = dimlen_mask[1] * dimlen_mask[2];

    break;

  case 4:

    for (int i = 0; i < mask_ndims; ++i) {
      status = nc_inq_dimlen(ncid, mask_dimids[i], dimlen_mask + i);
      if (status != NC_NOERR) handle_error(status);
    }

    *nbr_cells = dimlen_mask[2] * dimlen_mask[3];

    break;
  }

  /* read mask */

  switch(this_mask_type) {

  float * temp_cell_mask;

  case NC_FLOAT:

    temp_cell_mask = malloc ( *nbr_cells * sizeof ( float ) );

    status = nc_get_vara (ncid, mask_id, start, dimlen_mask, temp_cell_mask);
    if (status != NC_NOERR) handle_error(status);

    if ( mask_ndims == 4 ) {
      for ( int i = 0; i < *nbr_cells; ++i ) {
         if (temp_cell_mask[i] == 1) temp_cell_mask[i] = 0;
         else temp_cell_mask[i] = 1;
      }
    }

    *cell_mask = (int * ) malloc ( *nbr_cells * sizeof ( int ) );

    *cell_mask = (int*) (temp_cell_mask);

    break;

  default:

    printf ( " ERROR reading mask data type! \n" );
    exit(EXIT_FAILURE);

  }

  free(start);
  free(dimlen_mask);

  *mask_type = this_mask_type;
}

/* ---------------------------------------------------------------- */

struct point_with_index {

  int32_t lon, lat;
  int64_t i;
};

static inline int compare_point_with_index(const void * a,const void * b) {

   return ( (*(int64_t*)a < *(int64_t*)b) - (*(int64_t*)a > *(int64_t*)b) );
}

void remove_duplicated_vertices(void * temp_vertex_lon, void * temp_vertex_lat,
                                size_t * temp_nbr_vertices, int * old_to_new_id,
                                int vertex_type) {

   struct point_with_index * sort_array =
      malloc(*temp_nbr_vertices * sizeof(*sort_array));

   if (vertex_type == NC_FLOAT) {

      float const scale = (float)(2 << 20);
      int const periode = 360 * scale;

      for (int i = 0; i < *temp_nbr_vertices; ++i) {

         int32_t curr_lon, curr_lat;

         curr_lon = ((float*)temp_vertex_lon)[i] * scale;
         curr_lat = ((float*)temp_vertex_lat)[i] * scale;

         if (curr_lon <= 0)
            curr_lon = curr_lon - (((curr_lon - periode) / periode) * periode);
         else if (curr_lon > periode)
            curr_lon = curr_lon - ((curr_lon / periode) * periode);

         sort_array[i].lon = curr_lon;
         sort_array[i].lat = curr_lat;

         sort_array[i].i = i;
      }
   } else {

      double const scale = (double)(2 << 20);
      int32_t const periode = 360 * scale;

      for (int i = 0; i < *temp_nbr_vertices; ++i) {

         int curr_lon, curr_lat;

         curr_lon = ((double*)temp_vertex_lon)[i] * scale;
         curr_lat = ((double*)temp_vertex_lat)[i] * scale;

         if (curr_lon <= 0)
            curr_lon = curr_lon - (((curr_lon - periode) / periode) * periode);
         else if (curr_lon > periode)
            curr_lon = curr_lon - ((curr_lon / periode) * periode);

         sort_array[i].lon = curr_lon;
         sort_array[i].lat = curr_lat;

         sort_array[i].i = i;
      }
   }

   yac_mergesort(sort_array, *temp_nbr_vertices, sizeof(*sort_array),
                 compare_point_with_index);

   old_to_new_id[sort_array[0].i] = 1;

   int last_unique_idx = sort_array[0].i;

   for (int i = 1; i < *temp_nbr_vertices; ++i) {

      if (compare_point_with_index(sort_array + i - 1, sort_array + i)) {

         old_to_new_id[sort_array[i].i] = 1;
         last_unique_idx = (int)(sort_array[i].i);

      } else {

         old_to_new_id[sort_array[i].i] = -last_unique_idx;
      }
   }

   free(sort_array);

   size_t new_nbr_vertices = 0;

   for (int i = 0; i < *temp_nbr_vertices; ++i) {

      if (old_to_new_id[i] == 1) {

         if (vertex_type == NC_FLOAT) {

            ((float*)temp_vertex_lon)[new_nbr_vertices] = ((float*)temp_vertex_lon)[i];
            ((float*)temp_vertex_lat)[new_nbr_vertices] = ((float*)temp_vertex_lat)[i];
         } else {

            ((double*)temp_vertex_lon)[new_nbr_vertices] = ((double*)temp_vertex_lon)[i];
            ((double*)temp_vertex_lat)[new_nbr_vertices] = ((double*)temp_vertex_lat)[i];
         }

         old_to_new_id[i] = new_nbr_vertices;

         new_nbr_vertices++;
      }
   }

   for (int i = 0; i < *temp_nbr_vertices; ++i)
      if (old_to_new_id[i] <= 0)
         old_to_new_id[i] = old_to_new_id[-old_to_new_id[i]];

   *temp_nbr_vertices = new_nbr_vertices;
}

/* ---------------------------------------------------------------- */

void remove_duplicated_cells(void * temp_cell_lon, void * temp_cell_lat,
                             int * temp_cell_mask,
                             size_t * temp_nbr_cells, int * old_to_new_id,
                             int cell_type) {

   struct point_with_index * sort_array =
      malloc(*temp_nbr_cells * sizeof(*sort_array));

   if (cell_type == NC_FLOAT) {

      float const scale = (float)(2 << 20);
      int const periode = 360 * scale;

      for (int i = 0; i < *temp_nbr_cells; ++i) {

         int32_t curr_lon, curr_lat;

         curr_lon = ((float*)temp_cell_lon)[i] * scale;
         curr_lat = ((float*)temp_cell_lat)[i] * scale;

         if (curr_lon <= 0)
            curr_lon = curr_lon - (((curr_lon - periode) / periode) * periode);
         else if (curr_lon > periode)
            curr_lon = curr_lon - ((curr_lon / periode) * periode);

         sort_array[i].lon = curr_lon;
         sort_array[i].lat = curr_lat;

         sort_array[i].i = i;
      }
   } else {

      double const scale = (double)(2 << 20);
      int32_t const periode = 360 * scale;

      for (int i = 0; i < *temp_nbr_cells; ++i) {

         int curr_lon, curr_lat;

         curr_lon = ((double*)temp_cell_lon)[i] * scale;
         curr_lat = ((double*)temp_cell_lat)[i] * scale;

         if (curr_lon <= 0)
            curr_lon = curr_lon - (((curr_lon - periode) / periode) * periode);
         else if (curr_lon > periode)
            curr_lon = curr_lon - ((curr_lon / periode) * periode);

         sort_array[i].lon = curr_lon;
         sort_array[i].lat = curr_lat;

         sort_array[i].i = i;
      }
   }

   qsort(sort_array, *temp_nbr_cells, sizeof(*sort_array),
         compare_point_with_index);

   old_to_new_id[sort_array[0].i] = 1;

   int last_unique_idx = (int)(sort_array[0].i);

   for (int i = 1; i < *temp_nbr_cells; ++i) {

      if (compare_point_with_index(sort_array + i - 1, sort_array + i)) {

         old_to_new_id[sort_array[i].i] = 1;
         last_unique_idx = (int)(sort_array[i].i);

      } else {

         old_to_new_id[sort_array[i].i] = -last_unique_idx;
      }
   }

   free(sort_array);

   size_t new_nbr_cells = 0;

   for (int i = 0; i < *temp_nbr_cells; ++i) {

      if (old_to_new_id[i] == 1) {

         if (cell_type == NC_FLOAT) {

            ((float*)temp_cell_lon)[new_nbr_cells] = ((float*)temp_cell_lon)[i];
            ((float*)temp_cell_lat)[new_nbr_cells] = ((float*)temp_cell_lat)[i];
         } else {

            ((double*)temp_cell_lon)[new_nbr_cells] = ((double*)temp_cell_lon)[i];
            ((double*)temp_cell_lat)[new_nbr_cells] = ((double*)temp_cell_lat)[i];
         }

         old_to_new_id[i] = new_nbr_cells;

         new_nbr_cells++;
      }
   }

   for (int i = 0; i < *temp_nbr_cells; ++i)
      if (old_to_new_id[i] <= 0)
         old_to_new_id[i] = old_to_new_id[-old_to_new_id[i]];

   *temp_nbr_cells = new_nbr_cells;
}

/* ---------------------------------------------------------------- */

static void handle_error(int status)
{
  const char *err_string;
  err_string = nc_strerror(status);
  printf ( "%s \n", err_string );
  if (status != NC_NOERR) {
    exit(EXIT_FAILURE);
  }
}
#endif
