/**
 * @file weights2vtk.c
 *
 * @copyright Copyright  (C)  2015 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

#ifdef HAVE_C_NETCDF
#include <netcdf.h>
#endif

#include "read_icon_grid.h"
#include "read_mpiom_grid.h"
#include "utils.h"
#include "geometry.h"
#include "vtk_output.h"
#include "generate_cubed_sphere.h"
#include "interpolation_method_file.h"

#define YAC_WEIGHT_FILE_VERSION_1_0_STRING "yac weight file 1.0"

struct link_data {
  int src_idx, tgt_idx, points_idx;
  double weight;
};

enum grid_type {
  NONE,
  CUBE,
  CURVE,
  UNSTRUCT,
};

struct grid_config {
  union {
    struct {
      unsigned n;
    } cube;
    struct {
      char const * filename;
    } curve, unstruct;
  } config;
  int address_offset;
  enum grid_type type;
};

int file_exists(char const * filename);
void read_link_data(char const * weight_filename, int src_address_offset,
                    int tgt_address_offset, struct link_data ** links,
                    unsigned * num_links, enum yac_location ** src_locations,
                    enum yac_location * tgt_location, unsigned * max_src_idx,
                    unsigned * max_tgt_idx);
void write_data_to_file(char const * filename, struct grid * src_grid,
                        struct grid * tgt_grid, struct link_data * links,
                        enum yac_location * src_locations,
                        enum yac_location tgt_location,
                        unsigned num_links);
void parse_arguments(int argc, char ** argv,
                     struct grid_config * src_grid_config,
                     struct grid_config * tgt_grid_config,
                     char const ** weight_filename,
                     char const ** vtk_filename);
struct grid * create_grid(struct grid_config grid_config);

#ifdef HAVE_C_NETCDF
int main(int argc, char ** argv) {

  struct grid_config src_grid_config, tgt_grid_config;
  char const * weight_filename, * vtk_filename;

  parse_arguments(argc, argv, &src_grid_config, &tgt_grid_config,
                  &weight_filename, &vtk_filename);

  //-------------------------------------
  // read/generate source and target grid
  //-------------------------------------
  struct grid * src_grid = create_grid(src_grid_config);
  struct grid * tgt_grid = create_grid(tgt_grid_config);

  //-----------------
  // read weight file
  //-----------------
  if (!file_exists(weight_filename)) {
    fprintf(stderr, "File %s does not exist.\n", weight_filename);
    return EXIT_FAILURE;
  }
  struct link_data * links;
  unsigned num_links, max_src_idx, max_tgt_idx;
  enum yac_location * src_locations;
  enum yac_location tgt_location;
  read_link_data(weight_filename, src_grid_config.address_offset,
                 tgt_grid_config.address_offset, &links, &num_links,
                 &src_locations, &tgt_location, &max_src_idx, &max_tgt_idx);

  if ((max_src_idx >= yac_get_num_grid_cells(src_grid)) ||
      (max_tgt_idx >= yac_get_num_grid_cells(tgt_grid))) {
    fputs("weight file does not match with source and target grid\n", stderr);
    exit(EXIT_FAILURE);
  }

  //---------------
  // write vtk file
  //---------------
  write_data_to_file(vtk_filename, src_grid, tgt_grid, links, src_locations,
                     tgt_location, num_links);

  //---------
  // clean up
  //---------

  free(src_locations);
  free(links);
  yac_delete_grid(tgt_grid);
  yac_delete_grid(src_grid);

  return EXIT_SUCCESS;
}

int file_exists(const char * filename) {

   FILE * file = fopen(filename, "r");

   if (file != NULL) {
      fclose(file);
      return 1;
   }
   return 0;
}

#ifdef HAVE_C_NETCDF
static void handle_error(int status) {
  const char *err_string;
  err_string = nc_strerror(status);
  printf ("%s \n", err_string);
  if (status != NC_NOERR) {
    exit(EXIT_FAILURE);
  }
}
#endif

void read_link_data(char const * weight_filename, int src_address_offset,
                    int tgt_address_offset, struct link_data ** links,
                    unsigned * num_links, enum yac_location ** src_locations,
                    enum yac_location * tgt_location, unsigned * max_src_idx,
                    unsigned * max_tgt_idx) {

#ifdef HAVE_C_NETCDF

  int ncid, status, dimid, var_id;
  status = nc_open(weight_filename, NC_NOWRITE, &ncid);
  if (status != NC_NOERR) handle_error(status);

  // global attributes
  size_t version_len;
  char * version;
  status = nc_inq_attlen(ncid, NC_GLOBAL, "version", &version_len);
  if (status != NC_NOERR) handle_error(status);
  version = malloc(version_len + 1);
  version[version_len] = '\0';
  status = nc_get_att_text(ncid, NC_GLOBAL, "version", version);
  if (status != NC_NOERR) handle_error(status);
  if ((strlen(YAC_WEIGHT_FILE_VERSION_1_0_STRING) != version_len) ||
      strncmp(version, YAC_WEIGHT_FILE_VERSION_1_0_STRING, version_len)) {
    yac_internal_abort_message(
      "ERROR(read_link_data): version string from weight file is not \""
      YAC_WEIGHT_FILE_VERSION_1_0_STRING "\"\n", __FILE__, __LINE__);
  }
  free(version);

  size_t str_link_len;
  char * str_link;
  var_id = NC_GLOBAL;
  status = nc_inq_attlen(ncid, var_id, "contains_links", &str_link_len);
  if (status != NC_NOERR) handle_error(status);
  str_link = malloc(str_link_len + 1);
  str_link[str_link_len] = '\0';
  status = nc_get_att_text(ncid, var_id, "contains_links", str_link);
  if (status != NC_NOERR) handle_error(status);

  unsigned contains_links = 0;
  if ((strlen("TRUE") == str_link_len) &&
      !strncmp("TRUE", str_link, str_link_len)) {

    contains_links = 1;

  } else if ((strlen("FALSE") != str_link_len) ||
             strncmp("FALSE", str_link, str_link_len))
    yac_internal_abort_message("ERROR(read_link_data): invalid global "
                               "attribute contains_links", __FILE__,
                               __LINE__);
  free(str_link);

  if (!contains_links) {
    *links = NULL;
    *num_links = 0;
    *src_locations = NULL;
    *max_src_idx = 0;
    *max_tgt_idx = 0;
    return;
  }

  size_t num_wgts = 0;
  status = nc_inq_dimid(ncid, "num_wgts", &dimid);
  if (status != NC_NOERR) handle_error(status);
  status = nc_inq_dimlen(ncid, dimid, &num_wgts);
  if (status != NC_NOERR) handle_error(status);
  if (num_wgts != 1)
      yac_internal_abort_message("ERROR(read_link_data): YAC only supports "
                                 "num_wgts == 1\n", __FILE__, __LINE__);

  // get number of links from file
  size_t num_links_in_file = 0;
  status = nc_inq_dimid(ncid, "num_links", &dimid);
  if (status != NC_NOERR) handle_error(status);
  status = nc_inq_dimlen(ncid, dimid, &num_links_in_file);
  if (status != NC_NOERR) handle_error(status);
  if (num_links_in_file == 0)
      yac_internal_abort_message("ERROR(read_link_data): no links defined",
                                 __FILE__, __LINE__);
  *num_links = num_links_in_file;
  *links = malloc(num_links_in_file * sizeof(**links));

  // get number of pointsets
  size_t num_pointsets = 0;
  status = nc_inq_dimid(ncid, "num_src_fields", &dimid);
  if (status != NC_NOERR) handle_error(status);
  status = nc_inq_dimlen(ncid, dimid, &num_pointsets);
  if (status != NC_NOERR) handle_error(status);
  if (num_pointsets == 0)
      yac_internal_abort_message("ERROR(read_link_data): no point sets in file\n",
                                 __FILE__, __LINE__);

  // get max location string length from file
  size_t max_loc_str_len;
  status = nc_inq_dimid(ncid, "max_loc_str_len", &dimid);
  if (status != NC_NOERR) handle_error(status);
  status = nc_inq_dimlen(ncid, dimid, &max_loc_str_len);
  if (status != NC_NOERR) handle_error(status);
  if (max_loc_str_len != YAC_MAX_LOC_STR_LEN)
      yac_internal_abort_message("ERROR(read_link_data): wrong max location "
                                 "string length in weight file", __FILE__,
                                 __LINE__);

  // get source locations
  *src_locations = malloc(num_pointsets * sizeof(**src_locations));
  status = nc_inq_varid(ncid, "src_locations", &var_id);
  if (status != NC_NOERR) handle_error(status);

  for (unsigned i = 0; i < num_pointsets; ++i) {

    char loc_str[YAC_MAX_LOC_STR_LEN];

    size_t str_start[2] = {i, 0};
    size_t str_count[2] = {1, YAC_MAX_LOC_STR_LEN};

    status = nc_get_vara_text(ncid, var_id, str_start, str_count, loc_str);
    if (status != NC_NOERR) handle_error(status);

    (*src_locations)[i] = yac_str2loc(loc_str);
  }

  // get target location
  status = nc_inq_varid(ncid, "dst_location", &var_id);
  {
    char loc_str[YAC_MAX_LOC_STR_LEN];
    status = nc_get_var_text(ncid, var_id, loc_str);
    if (status != NC_NOERR) handle_error(status);
    *tgt_location = yac_str2loc(loc_str);
  }

  // get number of links per pointset
  int * num_links_per_pointset =
    malloc(num_pointsets * sizeof(*num_links_per_pointset));
  status = nc_inq_varid(ncid, "num_links_per_src_field", &var_id);
  if (status != NC_NOERR) handle_error(status);
  status = nc_get_var_int(ncid, var_id, num_links_per_pointset);
  if (status != NC_NOERR) handle_error(status);

  for (unsigned points_idx = 0, link_idx = 0; points_idx < num_pointsets;
       ++points_idx)
    for (unsigned i = 0; i < num_links_per_pointset[points_idx];
         ++i, ++link_idx)
      (*links)[link_idx].points_idx = points_idx;
  free(num_links_per_pointset);

  // get links
  int * address = malloc(num_links_in_file * sizeof(*address));
  status = nc_inq_varid(ncid, "src_address", &var_id);
  if (status != NC_NOERR) handle_error(status);
  status = nc_get_var_int(ncid, var_id, address);
  if (status != NC_NOERR) handle_error(status);
  *max_src_idx = 0;
  for (unsigned i = 0; i < num_links_in_file; ++i) {
    int idx = address[i] + src_address_offset;
    if (idx < 0) {
      fprintf(stderr, "invalid src index (%d)\n", idx);
      exit(EXIT_FAILURE);
    } else if (idx > *max_src_idx) *max_src_idx = idx;
    (*links)[i].src_idx = idx;
  }

  status = nc_inq_varid(ncid, "dst_address", &var_id);
  if (status != NC_NOERR) handle_error(status);
  status = nc_get_var_int(ncid, var_id, address);
  if (status != NC_NOERR) handle_error(status);
  *max_tgt_idx = 0;
  for (unsigned i = 0; i < num_links_in_file; ++i) {
    int idx = address[i] + tgt_address_offset;
    if (idx < 0) {
      fprintf(stderr, "invalid tgt index (%d)\n", idx);
      exit(EXIT_FAILURE);
    } else if (idx > *max_tgt_idx) *max_tgt_idx = idx;
    (*links)[i].tgt_idx = idx;
  }
  free(address);

  double * weights = malloc(num_links_in_file * sizeof(*weights));
  status = nc_inq_varid(ncid, "remap_matrix", &var_id);
  if (status != NC_NOERR) handle_error(status);
  status = nc_get_var_double(ncid, var_id, weights);
  if (status != NC_NOERR) handle_error(status);
  for (size_t i = 0; i < num_links_in_file; ++i)
    (*links)[i].weight = weights[i];
  free(weights);

  status = nc_close(ncid);
  if (status != NC_NOERR) handle_error(status);

#else // HAVE_C_NETCDF
  yac_internal_abort_message("ERROR(read_link_data): HAVE_C_NETCDF not defined",
                             __FILE__, __LINE__);
#endif
}

void get_grid_xyz_coordinates(struct grid * grid, double * points) {

   unsigned i, num_corners;
   double lon, lat;

   num_corners = yac_get_num_grid_corners(grid);

   for (i = 0; i < num_corners; ++i) {
      lon = yac_get_corner_x_coord(grid, i);
      lat = yac_get_corner_y_coord(grid, i);
      LLtoXYZ(lon, lat, points + 3 * i);
   }
}

void get_cell_middle_point(struct grid * grid, unsigned cell_index,
                           double * point) {

  unsigned num_cell_corners = yac_get_num_cell_corners(grid, cell_index);
  unsigned const * cell_corner_indices =
    yac_get_cell_corner_indices(grid, cell_index);
  double const * x_coords = yac_get_x_coords(grid);
  double const * y_coords = yac_get_y_coords(grid);

  point[0] = 0;
  point[1] = 0;
  point[2] = 0;

  for (unsigned i = 0; i < num_cell_corners; ++i) {

    double p[3];
    LLtoXYZ(x_coords[cell_corner_indices[i]], y_coords[cell_corner_indices[i]],
            p);
    point[0] += p[0];
    point[1] += p[1];
    point[2] += p[2];
  }

  double scale = 1.0 / sqrt(point[0] * point[0] + point[1] * point[1] +
                            point[2] * point[2]);

  point[0] *= scale;
  point[1] *= scale;
  point[2] *= scale;
}

void get_link_point_xyz_coordinates(int idx, enum yac_location location,
                                    struct grid * grid, double * point) {

  switch (location) {
    case(CELL):
      get_cell_middle_point(grid, idx, point);
      break;
    case(CORNER):
      LLtoXYZ(yac_get_corner_x_coord(grid, idx),
              yac_get_corner_y_coord(grid, idx), point);
      break;
    case(EDGE):
      yac_internal_abort_message("ERROR(get_link_point_xyz_coordinates): "
                                 "location EDGE is currently not supported",
                                 __FILE__, __LINE__);
      break;
    default:
      yac_internal_abort_message("ERROR(get_link_point_xyz_coordinates): "
                                 "invalid location", __FILE__, __LINE__);
  }
}

void get_link_xyz_coordinates(struct link_data * links, unsigned num_links,
                              struct grid * src_grid, struct grid * tgt_grid,
                              enum yac_location * src_locations,
                              enum yac_location tgt_location,
                              double * points) {

  for (unsigned i = 0; i < num_links; ++i) {

    get_cell_middle_point(src_grid, links[i].src_idx, points + 3 * 2 * i + 0);
    get_cell_middle_point(tgt_grid, links[i].tgt_idx, points + 3 * 2 * i + 3);
  }
}

void get_grid_cell_data(struct grid * grid, unsigned * cell_data,
                        unsigned offset) {

   unsigned i, j, k;
   unsigned num_grid_cells;

   num_grid_cells = yac_get_num_grid_cells(grid);

   for (i = 0, k = 0; i < num_grid_cells; ++i) {

      unsigned const * curr_cell_corners;
      unsigned num_cell_corners;

      curr_cell_corners = yac_get_cell_corner_indices(grid, i);
      num_cell_corners = yac_get_num_cell_corners(grid, i);

      for (j = 0; j < num_cell_corners; ++j)
         cell_data[k++] = curr_cell_corners[j] + offset;
   }
}

void get_link_address_data(struct link_data * links, unsigned num_links,
                           unsigned * polygon_data, unsigned offset) {

  for (unsigned i = 0; i < 2 * num_links; ++i)
    polygon_data[i] = i + offset;
}

void write_data_to_file(char const * filename, struct grid * src_grid,
                        struct grid * tgt_grid, struct link_data * links,
                        enum yac_location * src_locations,
                        enum yac_location tgt_location,
                        unsigned num_links) {

  //------------------------------------
  // generate cell data for the vtk file
  //------------------------------------

  unsigned num_grid_corners[2] = {yac_get_num_grid_corners(src_grid),
                                  yac_get_num_grid_corners(tgt_grid)};
  unsigned total_num_points = num_grid_corners[0] + num_grid_corners[1] +
                              2 * num_links;
  unsigned num_polygons[3] = {yac_get_num_grid_cells(src_grid),
                              yac_get_num_grid_cells(tgt_grid), num_links};
  unsigned total_num_polygons = num_polygons[0] + num_polygons[1] +
                                num_polygons[2];

  double * points = malloc(3 * total_num_points * sizeof(*points));

  // get the point data of the grids
  get_grid_xyz_coordinates(src_grid, points);
  get_grid_xyz_coordinates(tgt_grid, points + num_grid_corners[0] * 3);
  get_link_xyz_coordinates(links, num_links, src_grid, tgt_grid,
                           src_locations, tgt_location, points +
                           (num_grid_corners[0] + num_grid_corners[1]) * 3);

  unsigned * num_points_per_polygon =
    malloc(total_num_polygons * sizeof(*num_points_per_polygon));
  unsigned num_points_per_polygon_sum[3] = {0, 0, 0};
  for (unsigned i = 0; i < num_polygons[0]; ++i) {
    num_points_per_polygon[i] = yac_get_num_cell_corners(src_grid, i);
    num_points_per_polygon_sum[0] += num_points_per_polygon[i];
  }
  for (unsigned i = 0; i < num_polygons[1]; ++i) {
   num_points_per_polygon[num_polygons[0] + i] =
     yac_get_num_cell_corners(tgt_grid, i);
   num_points_per_polygon_sum[1] +=
     num_points_per_polygon[num_polygons[0] + i];
  }
  for (unsigned i = 0; i < num_polygons[2]; ++i)
    num_points_per_polygon[num_polygons[0] + num_polygons[1] + i] = 2;
  num_points_per_polygon_sum[2] = 2 * num_polygons[2];

  unsigned * polygon_data =
    malloc((num_points_per_polygon_sum[0] + num_points_per_polygon_sum[1] +
            num_points_per_polygon_sum[2]) * sizeof(*polygon_data));
  get_grid_cell_data(src_grid, polygon_data, 0);
  get_grid_cell_data(tgt_grid, polygon_data + num_points_per_polygon_sum[0],
                     num_grid_corners[0]);
  get_link_address_data(links, num_links, polygon_data +
                                          num_points_per_polygon_sum[0] +
                                          num_points_per_polygon_sum[1],
                       num_grid_corners[0] + num_grid_corners[1]);

  //------------------------------------
  // generate scalar data
  //------------------------------------

  unsigned * polygon_type = malloc(total_num_polygons * sizeof(*polygon_type));
  for (unsigned i = 0; i < num_polygons[0]; ++i) polygon_type[i] = 0;
  for (unsigned i = 0; i < num_polygons[1]; ++i)
    polygon_type[num_polygons[0] + i] = 1;
  for (unsigned i = 0; i < num_polygons[2]; ++i)
    polygon_type[num_polygons[0] + num_polygons[1] + i] = 2;

  double * weights = malloc(total_num_polygons * sizeof(*weights));
  for (unsigned i = 0; i < num_polygons[0] + num_polygons[1]; ++i)
    weights[i] = -1;
  for (unsigned i = 0; i < num_polygons[2]; ++i)
   weights[i + num_polygons[0] + num_polygons[1]] = links[i].weight;

  //------------------------------------
  // generate the actual vtk file
  //------------------------------------

  VTK_FILE * file = vtk_open(filename, "grid data");
  vtk_write_point_data(file, points, total_num_points);
  vtk_write_cell_data(file, polygon_data, num_points_per_polygon, total_num_polygons);
  vtk_write_cell_scalars_uint(file, polygon_type, total_num_polygons, "polygon_type");
  vtk_write_cell_scalars_double(file, weights, total_num_polygons, "weights");
  vtk_close(file);

  //------------------------------------
  // some final cleanup
  //------------------------------------

  free(weights);
  free(polygon_type);
  free(polygon_data);
  free(num_points_per_polygon);
  free(points);
}

int interpret_grid_arg(struct grid_config * grid_config, char * arg, char * str) {

  if (arg == NULL) {
    fprintf(stderr, "-%c argument is missing\n", str[0]);
    return 1;
  }

  switch (grid_config->type) {
    case CUBE:
      grid_config->config.cube.n = atoi(arg);
      if (grid_config->config.cube.n == 0) {
        fprintf(stderr, "invalid N for cubed sphere %s grid\n", str);
        return 1;
      }
      break;
    case CURVE:
      grid_config->config.curve.filename = arg;
      break;
    case UNSTRUCT:
      grid_config->config.unstruct.filename = arg;
      break;
    case NONE:
    default:
      fprintf(stderr, "invalid %s grid type\n", str);
      return 1;
  }

  return 0;
}

void print_help(char * command) {
  fprintf(stderr, "Usage: %s -S src_grid_type -T tgt_grid_type "
                  "-s src_filename -t tgt_filename -w weight_filename "
                  "-o vtk_filename\n\n"
                  "   grid_type can have the following values:\n"
                  "      'c', 'C': cubed sphere grid (instead of a "
                  "filename provide N, where\n"
                  "                N = sqrt(n/6), with n being the "
                  "total number of cells\n"
                  "      'm', 'M': curvilinear grid\n"
                  "      'i', 'I': unstructured grid\n"
                  "      - a lower case grid type indicates, that the global ids\n"
                  "        of the respective grid were \"0\"-based in the model\n"
                  "        component that generated the weight file\n"
                  "      - an upper case grid type indicates \"1\"-based global\n"
                  "        ids\n"
                  "\n"
                  "Example:\n"
                  "  Configuration:\n"
                  "    source grid type: unstructured\n"
                  "    source grid file: src_grid.nc\n"
                  "    target grid type: cubed sphere\n"
                  "    target grid file: 100 (N instead of filename)\n"
                  "    weight file name: weight_file.nc\n"
                  "    vtk file name:    weights.vtk\n"
                  "  Command:\n"
                  "     %s -S i -T c -s src_grid.nc -t 100 -w weight_file.nc "
                  "-o weights.vtk\n", command, command);
}

void parse_arguments(int argc, char ** argv,
                     struct grid_config * src_grid_config,
                     struct grid_config * tgt_grid_config,
                     char const ** weight_filename,
                     char const ** vtk_filename) {

  src_grid_config->type = NONE;
  tgt_grid_config->type = NONE;
  *weight_filename = NULL;
  *vtk_filename = NULL;

  int error = 0;

  char * src_arg = NULL;
  char * tgt_arg = NULL;

  int opt;
  int digit_optind = 0;
  int aopt = 0, bopt = 0;
  char *copt = 0, *dopt = 0;
  while ((opt = getopt(argc, argv, "?S:T:s:t:w:o:")) != -1) {
    switch (opt) {
    case 'S':
    case 'T':
    {
      struct grid_config * grid_config =
        (opt == 'S')?src_grid_config:tgt_grid_config;

      if (strlen(optarg) > 1) {
        fprintf(stderr, "invalid grid type for argument %c\n", (char)opt);
        exit(EXIT_FAILURE);
      }

      switch (optarg[0]) {
        case 'c':
        case 'C':
          grid_config->type = CUBE;
          break;
        case 'm':
        case 'M':
          grid_config->type = CURVE;
          break;
        case 'i':
        case 'I':
          grid_config->type = UNSTRUCT;
          break;
      };
      grid_config->address_offset = (optarg[0] < 'a')?-2:-1;
      break;
    }
    case 's':
      src_arg = optarg;
      break;
    case 't':
      tgt_arg = optarg;
      break;
    case 'w':
      *weight_filename = optarg;
      break;
    case 'o':
      *vtk_filename = optarg;
      break;
    case '?':
    default:
      error = 1;
    }
  }
  if (optind < argc) {
    printf("non-option ARGV-elements: ");
    while (optind < argc)
        printf ("%s ", argv[optind++]);
    printf("\n");
    error = 1;
  }

  if (argc == 1) error = 1;

  if ((!error) && (*weight_filename == NULL)) {
    fputs("weight_filename argument is missing\n", stderr);
    error = 1;
  }

  if ((!error) && (*vtk_filename == NULL)) {
    fputs("vtk_filename argument is missing\n", stderr);
    error = 1;
  }

  if (!error)
    error |= interpret_grid_arg(src_grid_config, src_arg, "source");
  if (!error)
    error |= interpret_grid_arg(tgt_grid_config, tgt_arg, "target");

  if (error) {
    print_help(argv[0]);
    exit(EXIT_FAILURE);
  }
}

struct grid * create_grid(struct grid_config grid_config) {

  switch(grid_config.type) {
    case CUBE:
      return generate_cubed_sphere_grid(grid_config.config.cube.n);
    case CURVE:
      if (!file_exists(grid_config.config.curve.filename)) {
        fprintf(stderr, "File %s does not exist.\n",
                grid_config.config.curve.filename);
        exit(EXIT_FAILURE);
      }
      return read_mpiom_grid((char*)(grid_config.config.curve.filename));
    case UNSTRUCT:
      if (!file_exists(grid_config.config.unstruct.filename)) {
        fprintf(stderr, "File %s does not exist.\n",
                grid_config.config.unstruct.filename);
        exit(EXIT_FAILURE);
      }
      return read_icon_grid((char*)(grid_config.config.unstruct.filename));
    default:
      return NULL;
  }
}

#else
int main(void) {
    return EXIT_FAILURE;
}
#endif // HAVE_C_NETCDF
