/**
 * @file ICONvtk.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#ifndef ANSI_C
#include <getopt.h>    /* for getopt_long; standard getopt is in unistd.h */
#endif
#include "geometry.h"
#include "grid.h"
#include "read_icon_grid.h"
#include "read_icon_data.h"
#include "read_fortran_data.h"
#include "vtk_output.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

/* ---------------------------------------------------------------- */

struct userInput {
  char gridFileName[128];
  char dataFileName[128];
  char **fieldName;
  int  numFields;
  int  level;
  int  timestepIncrement;
  int  startIdx;
  int  endIdx;
};

int verbose = 0;

#ifdef HAVE_C_NETCDF
static void init_vtk_output_file       ( struct grid * grid, VTK_FILE * fileId );

static void write_vtk_restart          ( struct grid * grid, struct userInput userInput );

static void write_vtk_field            ( struct grid * grid, struct userInput userInput );

static void get_grid_cell_data         ( struct grid * grid, unsigned * cell_data ) ;

static void get_grid_xyz_coordinates   ( struct grid * grid, double * vertices );

static void print_help();
#endif

/* ---------------------------------------------------------------- */

struct grid * icon_grid;
int err = -1;

int main( int argc, char **argv ) {

#ifdef HAVE_C_NETCDF

  struct userInput userInput;
  char *fieldname;

#ifdef ANSI_C

  char fieldList[128];

  printf ("Enter ICON grid file: ");
  if ( scanf("%s",userInput.gridFileName) != 1 ) printf("failed to read name of grid file.\n");

  icon_grid = read_icon_grid( userInput.gridFileName );

  printf ("Enter data file base name <basename>_????.nc or restart_cpl : ");
  if ( scanf("%s",userInput.dataFileName) != 1 ) printf("failed to read basename.\n");
  
  if ( strcmp (userInput.dataFileName,"restart_cpl") == 0 ) {
    write_vtk_restart (icon_grid, userInput);
  }
  else {

    printf ("Enter start index: ");
    if ( scanf ("%d", &userInput.startIdx) != 1 ) printf("failed to read start index.\n");

    printf ("Enter end index: ");
    if ( scanf ("%d", &userInput.endIdx) != 1 ) printf("failed to read end index.\n");

    printf ("Enter field names separated by ',' : ");
    if ( scanf ("%s",  fieldList) != 1 ) printf("failed to read field name.\n");

    printf ("Enter level: ");
    if ( scanf ("%d", &userInput.level) != 1 ) printf("failed to read level. \n");

    printf ("Enter time step increment: ");
    if ( scanf ("%d", &userInput.timestepIncrement) != 1 ) printf("failed to read time increment. \n");

    fieldname = strtok (fieldList,",");
  
    userInput.fieldName = (char**) malloc( 32 * sizeof(char*));

    userInput.numFields = 0;

    while (fieldname != NULL) {
      userInput.fieldName[userInput.numFields] = (char*) malloc( 32 * sizeof(char));
      strcpy(userInput.fieldName[userInput.numFields],fieldname);
      fieldname = strtok (NULL, " ,");
      userInput.numFields++; 
    }
    write_vtk_field (icon_grid, userInput);
  }

#else

  int c;

  /* values set by input argument */

  char * fieldList = NULL;
  int restart  = 0;

  /* based on http://en.wikipedia.org/wiki/Getopt */

  static struct option long_options[] = {
    {"gridfile", 1, 0, 'g'},  // required argument
    {"basename", 1, 0, 'b'},  // optional argument
    {"restart",  1, 0, 'r'},  // optional argument
    {"start",    1, 0, 's'},  // optional argument
    {"end",      1, 0, 'e'},  // optional argument
    {"field",    1, 0, 'f'},  // optional argument
    {"level",    1, 0, 'l'},  // optional argument
    {"timeincr", 1, 0, 't'},  // optional argument
    {"verbose",  0, 0, 'v'},  // no argument
    {"help",     0, 0, 'h'},  // no argument
    {NULL, 0, NULL, 0}
  };
  int option_index = 0;

  if ( argc == 1 ) {
    print_help();
    exit (0);
  };

  userInput.level    = 1;
  userInput.timestepIncrement = 1;
  userInput.startIdx = 0;
  userInput.endIdx   = 0;

  while ((c = getopt_long(argc, argv, "g:b:r:s:e:f:l:t:vh",
         long_options, &option_index)) != -1) {

    switch (c) {
    case 'g':
      strcpy(userInput.gridFileName,optarg);
      printf("input grid file:      %s\n",userInput.gridFileName);
      break;
    case 'b':
      strcpy(userInput.dataFileName,optarg);
      printf("input data file base: %s\n",userInput.dataFileName);
      break;
    case 'r':
      strcpy(userInput.dataFileName,optarg);
      printf("input data file:      %s\n",userInput.dataFileName);
      restart = 1;
      break;
    case 's':
      sscanf(optarg,"%d",&userInput.startIdx);
      printf("start index:          %i\n",userInput.startIdx);
      break;
    case 'e':
      sscanf(optarg,"%d",&userInput.endIdx);
      printf("end index:            %i\n",userInput.endIdx);
      break;
    case 'f':
      fieldList = optarg;
      printf("input field list:     %s\n",fieldList);
      break;
    case 'l':
      sscanf(optarg,"%d",&userInput.level);
      printf("selected level:       %i\n",userInput.level);
      break;
    case 't':
      sscanf(optarg,"%d",&userInput.timestepIncrement);
      printf("time step increment:  %i\n",userInput.timestepIncrement);
      break;
    case 'h':
      print_help();
      break;
    case 'v':
      verbose = 1;
      break;
    case '?':
      break;
    default:
      print_help();
      printf ("?? getopt returned character code 0%o ??\n", c);
    }
  }
  if (optind < argc) {
    printf ("non-option ARGV-elements: ");
    while (optind < argc)
      printf ("%s ", argv[optind++]);
    printf ("\n");
  }

  if ( userInput.endIdx == 0 ) userInput.endIdx = userInput.startIdx;

  icon_grid = read_icon_grid( userInput.gridFileName );

  if ( restart == 1 ) {
    write_vtk_restart (icon_grid, userInput);
  }
  else {

    fieldname = strtok (fieldList,",");
  
    userInput.fieldName = (char**) malloc( 32 * sizeof(char*));

    userInput.numFields = 0;

    while (fieldname != NULL) {
      userInput.fieldName[userInput.numFields] = (char*) malloc( 32 * sizeof(char));
      strcpy(userInput.fieldName[userInput.numFields],fieldname);
      fieldname = strtok (NULL, " ,");
      userInput.numFields++; 
    }
    write_vtk_field (icon_grid, userInput);
  }

#endif

  exit (0);
}

/* ---------------------------------------------------------------- */

static void init_vtk_output_file(struct grid * grid, VTK_FILE * vtk_file) {

   //--------------------------------------
   // generate vertex data for the vtk file
   //--------------------------------------

   unsigned num_corners;
   num_corners = yac_get_num_grid_corners(grid);

   double * vertices;

   vertices = malloc( 3 * num_corners * sizeof(*vertices));

   // get corners converted into cartesian coordinates

   get_grid_xyz_coordinates(grid, vertices);

   // write cartesian coordinates to vtk file

   vtk_write_point_data( vtk_file, vertices, num_corners );

   free (vertices);

   //-----------------------------------------------------
   // generate vertices and connectivity for the vtk file
   //-----------------------------------------------------

   unsigned * num_vertices_per_cell;
   unsigned * cell_data;

   unsigned num_grid_cells;

   num_grid_cells = yac_get_num_grid_cells(grid);

   num_vertices_per_cell = malloc(num_grid_cells * sizeof(*num_vertices_per_cell));

   unsigned i, total_num_cell_vertices;

   // get the number of vertices per grid cell

   for (i = 0; i < num_grid_cells; ++i)
      num_vertices_per_cell[i] = yac_get_num_cell_corners(grid, i);

   // count the total number of cell vertices

   total_num_cell_vertices = 0;
   for (i = 0; i < num_grid_cells; ++i)
      total_num_cell_vertices += num_vertices_per_cell[i];

   cell_data = malloc(total_num_cell_vertices * sizeof(*cell_data));

   // get the vertex coordinates of the grid converted into cartesian coordinates

   get_grid_cell_data(grid, cell_data);

   vtk_write_cell_data(vtk_file, cell_data, num_vertices_per_cell, num_grid_cells );

   free (num_vertices_per_cell);
   free (cell_data);
}


/* Read in, convert and write out self-made ICON
   coupling restart fields from Fortran binary files
   ------------------------------------------------- */

static void write_vtk_restart ( struct grid * grid, struct userInput userInput ) {

  VTK_FILE *vtkFile;
  FILE *inputFile;
  unsigned i, j;

  char fieldName_template[132];
  char fieldName[132];

  char * filename;
  unsigned num_cells;

  struct fieldMetadata fieldInfo;

  double * cell_data;

  num_cells = yac_get_num_grid_cells(grid);

  filename          = calloc((strlen(userInput.dataFileName) + 4), sizeof(char));

  strcpy(filename, userInput.dataFileName);
  strcat(filename, ".vtk");

  strcpy(fieldName_template, "restart_%d");

  if ((inputFile=fopen(userInput.dataFileName, "r")) == NULL) {
    printf("Cannot open file.\n");
    exit(1);
  }

  vtkFile = vtk_open ( filename, "ICON data" );
  init_vtk_output_file ( grid, vtkFile );

  unsigned nbrCouplingFields = 3;

  for ( i = 0; i < nbrCouplingFields; ++i ) {

    read_fortran_dimensions ( inputFile, &fieldInfo, num_cells );

    cell_data = malloc ( fieldInfo.nbrHorPoints * fieldInfo.nbrBundles * sizeof(*cell_data ) );

    read_fortran_data ( cell_data, fieldInfo, inputFile );

    /* ... normalize accumulated fields */

    if ( fieldInfo.accumulatedSteps > 1 ) {
      for ( j = 0; (int)j < fieldInfo.nbrHorPoints * fieldInfo.nbrBundles; j++ ) {
         cell_data[j] = cell_data[j] / (double) fieldInfo.accumulatedSteps;
      }
    }

    /* ... convert fields to vtk format */

    for ( j = 0; (int)j < fieldInfo.nbrBundles; j++ ) {
      if ( fieldInfo.nbrBundles > 1 ) printf ("bundle nr %i \n", j+1);
      sprintf ( fieldName, fieldName_template, i+j );
      vtk_write_cell_scalars_double ( vtkFile, cell_data+j*fieldInfo.nbrHorPoints, 
                                      fieldInfo.nbrHorPoints, fieldName );
    }
    free(cell_data);
  }

  /* ... close files */

  vtk_close ( vtkFile );
  fclose(inputFile);

  /* free memory */

  free(filename);
}


/* Read in, convert and write out fields from ICON NetCDf output files
   ------------------------------------------------------------------- */

static void write_vtk_field ( struct grid * grid, struct userInput userInput) {

  VTK_FILE * vtkFile;
  int iconFile;

  int fieldNumber;
  int fileNumber;
  int timeStep;

  struct fieldMetadata * fieldInfo;
  struct fieldMetadata fieldMeta;

  char input_basename[132];
  char * outputFile_template, * outputFile;
  char * inputFile, * inputFile_template;

  void * field_data;

  /* Create a vtk output file out of the field names to be converted */

  /* TODO:

     calculate the total length of the string including underscore */

  outputFile_template = calloc( 132 , sizeof(char));

  for ( fieldNumber = 0; fieldNumber < userInput.numFields; fieldNumber++ ) {
    strcat(outputFile_template,userInput.fieldName[fieldNumber]);
    strcat(outputFile_template,"_");
  }

  /* .vtk plus integer number with 6 digits (gives space for 10^6 timesteps) */

  outputFile = calloc((strlen(outputFile_template) + 10), sizeof(char));

  strcat(outputFile_template, "%d.vtk");

  /* Pre-construct input file names */

  strcpy(input_basename, userInput.dataFileName);

  /* TODO:

     calculate the total length of the string including underscore based
     on start and end Idx or later based on the date string or import a file list */

  inputFile_template = calloc((strlen(input_basename) + 8), sizeof(char));
  inputFile          = calloc((strlen(input_basename) + 8), sizeof(char));

  strcpy(inputFile_template, input_basename);
  strcat(inputFile_template, "_%04d.nc");

  int processedSteps = 0;
  int timeStart      = 0;

  /* Loop over ICON input files */
 
  for ( fileNumber = userInput.startIdx ; fileNumber < userInput.endIdx+1; fileNumber++ ) {

    sprintf(inputFile, inputFile_template, fileNumber);

    printf("Open input file %s \n", inputFile);

    iconFile = open_icon_output ( inputFile );

    /* ... see what is in the NetCDF file */

    fieldInfo = malloc ( userInput.numFields * sizeof(*fieldInfo) );

    for ( fieldNumber = 0; fieldNumber < userInput.numFields; fieldNumber++ ) {

      read_icon_dimensions ( iconFile, userInput.fieldName[fieldNumber], &fieldInfo[fieldNumber] );

      if ( fieldInfo[fieldNumber].nbrHorPoints != (int) yac_get_num_grid_cells(grid) ) {
        printf ("Dimensions do not match!\n");
        printf ("We got %i horizontal points to process \n",fieldInfo[fieldNumber].nbrHorPoints);
        printf ("but the provided grid consists of %i cell.\n", (int) yac_get_num_grid_cells(grid));
        exit(EXIT_FAILURE);
      }
    }

    /* ... memory of physical fields, assumes that dimensions are identical for all fields. */

    fieldMeta = fieldInfo[0];

    get_memory ( &field_data, fieldMeta );

    for ( timeStep = timeStart; timeStep < fieldInfo[0].nbrTimeSteps;
           timeStep +=userInput.timestepIncrement ) {

      ++processedSteps;

      sprintf(outputFile, outputFile_template, processedSteps);

      if ( verbose ) printf ("Write relative time step %i to file %s \n", timeStep,
                             outputFile );

      vtkFile = vtk_open ( outputFile, "ICON data" );
      init_vtk_output_file ( grid, vtkFile );

      for ( fieldNumber = 0; fieldNumber < userInput.numFields; fieldNumber++ ) {

        /* ... read data from NetCDF file */

        read_icon_timestep_level ( iconFile, field_data, fieldInfo[fieldNumber],
                                   timeStep, userInput.level );

        /* ... and convert fields to vtk format */

        switch( fieldInfo[fieldNumber].varType ) {
          case FLOAT:
            vtk_write_cell_scalars_float ( vtkFile, (float*) field_data,
                                           fieldInfo[fieldNumber].nbrHorPoints,
                                           userInput.fieldName[fieldNumber]);
            break;
          case DOUBLE:
            vtk_write_cell_scalars_double( vtkFile, (double*) field_data,
                                           fieldInfo[fieldNumber].nbrHorPoints,
                                           userInput.fieldName[fieldNumber]);
            break;
          default:
            fprintf(stdout, "Datatype not supported \n");
            exit(EXIT_FAILURE);
        }
      }

      if ( verbose) printf ("Closing %s \n", outputFile );

      vtk_close(vtkFile);

    } // timestep loop

    /* add offset according to requested increment */

    timeStart = ( ( timeStep - fieldInfo[0].nbrTimeSteps ) % userInput.timestepIncrement );

    close_icon_output ( iconFile );

    free_memory ( field_data );
  }

  free(inputFile_template);
  free(outputFile);
  free(outputFile_template);
}

/* Helper routines
   --------------- */

static void get_grid_xyz_coordinates(struct grid * grid, double * vertices) {

   unsigned i, num_corners;
   double lon, lat;

   num_corners = yac_get_num_grid_corners(grid);

   for (i = 0; i < num_corners; ++i) {
      lon = yac_get_corner_x_coord(grid, i);
      lat = yac_get_corner_y_coord(grid, i);
      LLtoXYZ(lon, lat, vertices + 3 * i);
   }
}

static void get_grid_cell_data(struct grid * grid, unsigned * cell_data) {

   unsigned i, j, k;
   unsigned num_grid_cells;

   num_grid_cells = yac_get_num_grid_cells(grid);

   for (i = 0, k = 0; i < num_grid_cells; ++i) {

      unsigned const * curr_cell_corners;
      unsigned num_cell_corners;

      curr_cell_corners = yac_get_cell_corner_indices(grid, i);
      num_cell_corners = yac_get_num_cell_corners(grid, i);

      for (j = 0; j < num_cell_corners; ++j)
         cell_data[k++] = curr_cell_corners[j];
   }
}

/* http://ascii-table.com/ansi-escape-sequences.php */

static void boldOn() {
  char ESC=27;
  printf("%c[1m",ESC);  /* turn on bold */
}

static void lineOn() {
  char ESC=27;
  printf("%c[4m",ESC);  /* turn on underline */
}

static void Off() {
  char ESC=27;
  printf("%c[0m",ESC);  /* reset all */
}


void print_help () {

  lineOn();
  boldOn(); printf ("\n\n\t\t List of available options \n\n"); Off();

  boldOn(); printf ("Required arguments: \n \n");
  printf (" -g, --gridfile\n"); Off();
  printf (" \t input file containing the ICON grid \n \n");

  boldOn();
  printf ("Optional arguments:\n \n");
  printf (" -h, --help \n"); Off();

  printf (" \t print this list \n");
  boldOn(); printf (" -v, --verbose \n");  Off();
  printf (" \t run program in verbose mode \n \n");
  boldOn(); printf (" -l, --level \n"); Off();
  printf (" \t level/layer to be extracted for conversion, default 1\n");
  boldOn();printf (" -t, --timeincr \n"); Off();
  printf (" \t increment of timestep to be converted, default 1\n \n");

  boldOn(); printf ("Semi-Optional arguments, one of the two below is required:\n \n");
  printf (" -b, --basename \n");  Off();
  printf (" \t optional, basename of filename format <basename>_????.nc \n");
  boldOn(); printf (" -r, --restart \n");  Off();
  printf (" \t optional, name of the ICON coupling restart file         \n \n");

  boldOn(); printf ("Required arguments if basename is set: \n \n");
  printf (" -s, --start\n");  Off();
  printf (" \t start index of list of 'basename' files \n");
  boldOn(); printf (" -e, --end \n");  Off();
  printf (" \t end   index of list of 'basename' files, default value is the start index. \n");
  boldOn(); printf (" -f, --field \n");  Off();
  printf (" \t comma separated list of fields contained in the NetCDF file to be converted \n \n");
#else
  printf ("NetCDF is required to run this!\n");
  return EXIT_FAILURE;
#endif
}

