/**
 * @file test_grid_coupling.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * This example is based on a example presented in the following paper:
 * "Development and performance of a new version of the OASIS coupler, OASIS3-MCT_3.0"
 * (http://www.geosci-model-dev-discuss.net/gmd-2017-64/)
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#define EXACT

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef USE_MPI
#include <mpi.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "yac_interface.h"
#include "vtk_output.h"
#include "read_icon_grid.h"
#include "event.h"

#define MIN(a,b) ((a) < (b) ? (a) : (b))

struct {

  unsigned component_idx[3];
  unsigned num_components;

  unsigned grid_idx[3];

} config_per_global_rank[38];

char const * component_names[] = {"comp1-grid1", "comp2-grid2", "comp3",
                                  "comp3-grid3", "comp3-grid4", "comp3-grid5",
                                  "comp4"};
char const * grid_names[] = {"grid1", "grid2", "grid3", "grid4", "grid5"};

struct {

  char const * field_names[8];
  unsigned num_fields;
  char const * file_name;

} grid_config[] = {{.field_names = {"A_1", "D_1", "E_1", "J_1",
                                    "A_2", "D_2", "E_2", "J_2"},
                    .num_fields = 8,
                    .file_name = "iconR2B05-grid.nc"},
                   {.field_names = {"A_1", "B_1", "F_1", "I_1",
                                    "A_2", "B_2", "F_2", "I_2"},
                    .num_fields = 8,
                    .file_name = "iconR2B05-grid.nc"},
                   {.field_names = {"B_1", "C_1", "D_1", "G_1",
                                    "B_2", "C_2", "D_2", "G_2"},
                    .num_fields = 8,
                    .file_name = "iconR2B06-grid.nc"},
                   {.field_names = {"C_1", "E_1", "F_1", "H_1",
                                    "C_2", "E_2", "F_2", "H_2"},
                    .num_fields = 8,
                    .file_name = "iconR2B05-grid.nc"},
                   {.field_names = {"G_1", "H_1", "I_1", "J_1",
                                    "G_2", "H_2", "I_2", "J_2"},
                    .num_fields = 8,
                    .file_name = "iconR2B07-grid.nc"}};

/* -------------------------------------------------------------------- */

static void setup_process_config();

void read_part_icon_grid_information(
  const char * filename, int * nbr_vertices, int * nbr_cells,
  int ** num_vertices_per_cell, int ** cell_to_vertex, double ** x_vertices,
  double ** y_vertices, double ** x_cells, double ** y_cells,
  int ** global_cell_id, int ** cell_mask, int ** global_cell_id_rank,
  int ** global_corner_id, int ** global_corner_id_rank, int rank, int size);

/* -------------------------------------------------------------------- */

int main (void) {

  setup_process_config();

#if defined (USE_MPI) && defined (USE_XML2)

  yac_cinit ("coupling_OASIS3_MCT.xml", "coupling.xsd");
  int global_rank, global_size;
  MPI_Comm_rank(MPI_COMM_WORLD, &global_rank);
  MPI_Comm_size(MPI_COMM_WORLD, &global_size);

  if (global_size != 38) {
    fprintf(stderr, "Wrong number of processes (should be 38)\n");
    exit(EXIT_FAILURE);
  }

  // register component

  char const * local_component_names[3];
  int component_ids[3];
  unsigned num_components = config_per_global_rank[global_rank].num_components;
  for (unsigned i = 0; i < num_components; ++i)
    local_component_names[i] =
      component_names[config_per_global_rank[global_rank].component_idx[i]];
  yac_cdef_comps(local_component_names, num_components, component_ids);

  int component_ranks[3], component_sizes[3];
  for (unsigned i = 0; i < num_components; ++i) {
    MPI_Comm component_comm;
    yac_cget_localcomm(&component_comm, component_ids[i]);
    MPI_Comm_rank(component_comm, &component_ranks[i]);
    MPI_Comm_size(component_comm, &component_sizes[i]);
  }

  // register grid and fields

  int field_ids[3*8];
  unsigned num_fields_per_grid[3];
  unsigned num_fields = 0;

  double * field_data[3][8];

  for (unsigned i = 0; i < num_components; ++i) {

    if (config_per_global_rank[global_rank].grid_idx[i] == -1) {
      num_fields_per_grid[i] = 0;
      continue;
    }

    int nbr_vertices;
    int nbr_cells;
    int * num_vertices_per_cell;
    int * cell_to_vertex;
    double * x_vertices;
    double * y_vertices;
    double * x_cells;
    double * y_cells;

    int * cell_mask;
    int * global_cell_id;
    int * global_cell_id_rank;
    int * global_corner_id;
    int * global_corner_id_rank;

    read_part_icon_grid_information(
      grid_config[config_per_global_rank[global_rank].grid_idx[i]].file_name,
      &nbr_vertices, &nbr_cells, &num_vertices_per_cell, &cell_to_vertex,
      &x_vertices, &y_vertices, &x_cells, &y_cells, &global_cell_id, &cell_mask,
      &global_cell_id_rank, &global_corner_id, &global_corner_id_rank,
      component_ranks[i], component_sizes[i]);

    for (int i = 0; i < nbr_vertices; ++i) {
      x_vertices[i] *= deg;
      y_vertices[i] *= deg;
    }

    for (int i = 0; i < nbr_cells; ++i) {
      x_cells[i] *= deg;
      y_cells[i] *= deg;
    }

    int subdomain_id;
    yac_cdef_subdomain(
      component_ids[i],
      grid_names[config_per_global_rank[global_rank].grid_idx[i]],
      &subdomain_id);

    yac_cdef_elements_unstruct(
      subdomain_id, nbr_vertices, nbr_cells, num_vertices_per_cell, x_vertices,
      y_vertices, cell_to_vertex);

    int corner_point_id, cell_point_id;
    yac_cdef_points_unstruct(
      subdomain_id, nbr_vertices, YAC_LOCATION_CORNER, x_vertices, y_vertices,
      &corner_point_id);
    yac_cdef_points_unstruct(
      subdomain_id, nbr_cells, YAC_LOCATION_CELL, x_cells, y_cells,
      &cell_point_id);

    yac_cdef_index_location(
      subdomain_id, nbr_vertices, YAC_LOCATION_CORNER, global_corner_id,
      global_corner_id_rank);
    yac_cdef_index_location(
      subdomain_id, nbr_cells, YAC_LOCATION_CELL, global_cell_id,
      global_cell_id_rank);

    int domain_id;
    yac_cconnect_subdomains(component_ids[i], 1, &subdomain_id, &domain_id);

    int mask_id;
    yac_cdef_mask(cell_mask, cell_point_id, &mask_id);

    num_fields_per_grid[i] =
      grid_config[config_per_global_rank[global_rank].grid_idx[i]].num_fields;

    for (unsigned j = 0; j < num_fields_per_grid[i]; ++j) {

      yac_cdef_field(
        grid_config[
          config_per_global_rank[global_rank].grid_idx[i]].field_names[j],
        component_ids[i], domain_id, &cell_point_id, &mask_id, 1,
        &field_ids[num_fields]);

      field_data[i][j] = malloc(nbr_cells * sizeof(*(field_data[i][j])));
      for (unsigned k = 0; k < nbr_cells; ++k)
        field_data[i][j][k] =
          sin(x_cells[k] * 2 * rad) * cos(y_cells[k] * 3 * rad);

      num_fields++;
    }

    free(x_vertices);
    free(y_vertices);
    free(x_cells);
    free(y_cells);
    free(cell_to_vertex);
    free(num_vertices_per_cell);
    free(global_cell_id);
    free(global_cell_id_rank);
    free(cell_mask);
    free(global_corner_id);
    free(global_corner_id_rank);
  }

  int ierror;
  yac_csearch(num_components, component_ids, num_fields, field_ids, &ierror);

  // do some ping-pongs

  for (unsigned t = 0; t < 100; ++t) {

    // call put for all local fields
    {
      double *point_set_data[1];
      double **domain_data[1] = {point_set_data};
      double ***collection_data[1] = {domain_data};

      for (unsigned i = 0, k = 0; i < num_components; ++i) {
        for (unsigned j = 0; j < num_fields_per_grid[i]; ++j, ++k) {

          int err, info;
          point_set_data[0] = field_data[i][j];
          yac_cput(field_ids[k], 1, collection_data, &info, &err);
        }
      }
    }

    // call get for all local fields
    {
      double *domain_data[1];
      double **collection_data[1] = {domain_data};

      for (unsigned i = 0, k = 0; i < num_components; ++i) {
        for (unsigned j = 0; j < num_fields_per_grid[i]; ++j, ++k) {

          int err, info;
          domain_data[0] = field_data[i][j];
          yac_cget(field_ids[k], 1, collection_data, &info, &err);
        }
      }
    }
  }

  // clean-up

  for (unsigned i = 0; i < num_components; ++i)
    for (unsigned j = 0; j < num_fields_per_grid[i]; ++j)
      free(field_data[i][j]);

  yac_cfinalize();

#else
  fprintf(stderr, "YAC configured without MPI and/or XML\n");
  exit(EXIT_FAILURE);
#endif
}

/* -------------------------------------------------------------------- */

static void setup_process_config() {

  unsigned rank = 0;

  for (; rank < 6; ++rank) {
    config_per_global_rank[rank].component_idx[0] = 0;
    config_per_global_rank[rank].num_components = 1;
    config_per_global_rank[rank].grid_idx[0] = 0;
  }
  for (; rank < 12; ++rank) {
    config_per_global_rank[rank].component_idx[0] = 1;
    config_per_global_rank[rank].num_components = 1;
    config_per_global_rank[rank].grid_idx[0] = 1;
  }
  for (; rank < 22; ++rank) {
    config_per_global_rank[rank].component_idx[0] = 2;
    config_per_global_rank[rank].component_idx[1] = 3;
    config_per_global_rank[rank].component_idx[2] = 5;
    config_per_global_rank[rank].num_components = 3;
    config_per_global_rank[rank].grid_idx[0] = -1;
    config_per_global_rank[rank].grid_idx[1] = 2;
    config_per_global_rank[rank].grid_idx[2] = 4;
  }
  for (; rank < 27; ++rank) {
    config_per_global_rank[rank].component_idx[0] = 2;
    config_per_global_rank[rank].component_idx[1] = 4;
    config_per_global_rank[rank].component_idx[2] = 5;
    config_per_global_rank[rank].num_components = 3;
    config_per_global_rank[rank].grid_idx[0] = -1;
    config_per_global_rank[rank].grid_idx[1] = 3;
    config_per_global_rank[rank].grid_idx[2] = 4;
  }
  for (; rank < 31; ++rank) {
    config_per_global_rank[rank].component_idx[0] = 2;
    config_per_global_rank[rank].component_idx[1] = 4;
    config_per_global_rank[rank].num_components = 2;
    config_per_global_rank[rank].grid_idx[0] = -1;
    config_per_global_rank[rank].grid_idx[1] = 3;
  }
  for (; rank < 34; ++rank) {
    config_per_global_rank[rank].component_idx[0] = 2;
    config_per_global_rank[rank].num_components = 1;
    config_per_global_rank[rank].grid_idx[0] = -1;
  }
  for (; rank < 38; ++rank) {
    config_per_global_rank[rank].component_idx[0] = 6;
    config_per_global_rank[rank].num_components = 1;
    config_per_global_rank[rank].grid_idx[0] = -1;
  }
}

void read_part_icon_grid_information(
  const char * filename, int * nbr_vertices, int * nbr_cells,
  int ** num_vertices_per_cell, int ** cell_to_vertex, double ** x_vertices,
  double ** y_vertices, double ** x_cells, double ** y_cells,
  int ** global_cell_id, int ** cell_mask, int ** global_cell_id_rank,
  int ** global_corner_id, int ** global_corner_id_rank, int rank, int size) {

#ifdef HAVE_C_NETCDF
  // read the global grid
  read_icon_grid_information(filename, nbr_vertices, nbr_cells,
                             num_vertices_per_cell, cell_to_vertex,
                             x_vertices, y_vertices, x_cells, y_cells, cell_mask);

  // determine local rang
  int num_cells_per_process = (*nbr_cells + size - 1)/size;
  int local_start = rank * num_cells_per_process;
  int num_local_cells = MIN(num_cells_per_process,
                            *nbr_cells - local_start);

  // mask for required vertices and cells
  int * required_vertices = calloc(*nbr_vertices, sizeof(*required_vertices));
  int * required_cells = calloc(*nbr_cells, sizeof(*required_cells));

  int offset = 0;
  for (int i = 0; i < local_start; ++i)
    offset += (*num_vertices_per_cell)[i];

  // mark all local cells and their vertices as required
  for (int i = local_start; i < local_start + num_local_cells; ++i) {

    required_cells[i] = 1;
    for (int j = 0; j < (*num_vertices_per_cell)[i]; ++j)
      required_vertices[(*cell_to_vertex)[offset+j]] = 1;

    offset += (*num_vertices_per_cell)[i];
  }

  // mask for halo cells
  int * cell_is_on_rank = malloc(*nbr_cells * sizeof(*cell_is_on_rank));

  // mark all halo cells as required and generate cell_is_on_rank
  offset = 0;
  for (int i = 0; i < *nbr_cells; ++i) {

    if (!required_cells[i]) {

      for (int j = 0; j < (*num_vertices_per_cell)[i]; ++j) {

        if (required_vertices[(*cell_to_vertex)[offset+j]]) {

          required_cells[i] = 1;
          cell_is_on_rank[i] = i / num_cells_per_process;
          break;
        }
      }

    } else {
      cell_is_on_rank[i] = -1;
    }

    offset += (*num_vertices_per_cell)[i];
  }

  // mask for halo vertices
  int * vertex_is_on_rank = malloc(*nbr_vertices * sizeof(*vertex_is_on_rank));

  // mark all halo vertices as required and generate vertex_is_on_rank
  offset = 0;
  for (int i = 0; i < *nbr_vertices; ++i)
    if (required_vertices[i])
      vertex_is_on_rank[i] = -1;
  for (int i = 0; i < *nbr_cells; ++i) {

    if (required_cells[i] && cell_is_on_rank[i] != -1) {

      for (int j = 0; j < (*num_vertices_per_cell)[i]; ++j) {

        if (!required_vertices[(*cell_to_vertex)[offset+j]]) {

          required_vertices[(*cell_to_vertex)[offset+j]] = 1;
          vertex_is_on_rank[(*cell_to_vertex)[offset+j]] = cell_is_on_rank[i];
        }
      }
    }

    offset += (*num_vertices_per_cell)[i];
  }

  // count the number of cells and vertices
  int part_num_vertices = 0;
  int part_num_cells = 0;
  for (int i = 0; i < *nbr_vertices; ++i)
    if (required_vertices[i])
      part_num_vertices++;
  for (int i = 0; i < *nbr_cells; ++i)
    if(required_cells[i])
      part_num_cells++;

  *global_cell_id = malloc(part_num_cells * sizeof(**global_cell_id));
  *global_cell_id_rank = malloc(part_num_cells * sizeof(**global_cell_id_rank));
  *global_corner_id = malloc(part_num_vertices * sizeof(**global_corner_id));
  *global_corner_id_rank = malloc(part_num_vertices * sizeof(**global_corner_id_rank));

  // generate final vertex data
  part_num_vertices = 0;
  int * global_to_local_vertex = malloc(*nbr_vertices * sizeof(*global_to_local_vertex));
  for (int i = 0; i < *nbr_vertices; ++i) {

    if (required_vertices[i]) {

      (*global_corner_id)[part_num_vertices] = i;
      (*global_corner_id_rank)[part_num_vertices] = vertex_is_on_rank[i];
      (*x_vertices)[part_num_vertices] = (*x_vertices)[i];
      (*y_vertices)[part_num_vertices] = (*y_vertices)[i];
      global_to_local_vertex[i] = part_num_vertices;
      part_num_vertices++;
    }
  }

  *x_vertices = realloc(*x_vertices, part_num_vertices * sizeof(**x_vertices));
  *y_vertices = realloc(*y_vertices, part_num_vertices * sizeof(**y_vertices));
  free(vertex_is_on_rank);
  *nbr_vertices = part_num_vertices;
  free(required_vertices);

  // generate final cell data
  int num_cell_vertex_dependencies = 0;
  part_num_cells = 0;
  offset = 0;
  for (int i = 0; i < *nbr_cells; ++i) {

    if (required_cells[i]) {

      (*global_cell_id)[part_num_cells] = i;
      (*global_cell_id_rank)[part_num_cells] = cell_is_on_rank[i];
      (*x_cells)[part_num_cells] = (*x_cells)[i];
      (*y_cells)[part_num_cells] = (*y_cells)[i];
      (*cell_mask)[part_num_cells] = (*cell_mask)[i];

      for (int j = 0; j < (*num_vertices_per_cell)[i]; ++j)
        (*cell_to_vertex)[num_cell_vertex_dependencies++] =
          global_to_local_vertex[(*cell_to_vertex)[offset+j]];

      (*num_vertices_per_cell)[part_num_cells] = (*num_vertices_per_cell)[i];

      part_num_cells++;
    }

    offset += (*num_vertices_per_cell)[i];
  }

  *x_cells   = realloc(*x_cells, part_num_cells * sizeof(**x_cells));
  *y_cells   = realloc(*y_cells, part_num_cells * sizeof(**y_cells));
  *cell_mask = realloc(*cell_mask, part_num_cells * sizeof(**cell_mask));

  *num_vertices_per_cell = realloc(*num_vertices_per_cell, part_num_cells *
                                   sizeof(**num_vertices_per_cell));
  *cell_to_vertex = realloc(*cell_to_vertex, num_cell_vertex_dependencies *
                            sizeof(**cell_to_vertex));
  free(cell_is_on_rank);
  *nbr_cells = part_num_cells;
  free(required_cells);
  free(global_to_local_vertex);

#else
  fprintf(stderr, "YAC configured without C_NETCDF\n");
  exit(EXIT_FAILURE);
#endif // HAVE_C_NETCDF
}
