#ifdef USE_MPI
#include <mpi.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "yac_interface.h"
#include "points.h"

#define NBR_CELLS 2
#define NBR_VERTICES 4
#define NBR_SUBDOMAIN_IDS 1

int main () {

  const char * fieldName[] = { "TAUX",    // bundled field containing two components
                               "TAUY",    // bundled field containing two components
                               "SFWFLX",  // bundled field containing three components
                               "SFTEMP",
                               "THFLX",   // bundled field containing four components
                               "ICEATM",  // bundled field containing four components
                               "SST",
                               "OCEANU",
                               "OCEANV",
                               "ICEOCE" }; // bundled field containing four components

  char * xml_filename = "coupling_aquaplanet.xml";
  char * xsd_filename = "coupling.xsd";

  char * comp_name = "dummy_atmosphere";
  char * grid_name = "dummy_atmosphere_grid";

  const int no_of_fields = 10;

  const int nbr_subdomain_ids = NBR_SUBDOMAIN_IDS;

  const int nbr_cells = NBR_CELLS;
  const int nbr_vertices = NBR_VERTICES;
  int nbr_vertices_per_cell[NBR_CELLS];

  int i, info, ierror;
  int comp_id;
  int comp_ids[1];

  int cell_point_ids[1];
  int cell_mask_ids[1];

  int domain_id;
  int subdomain_id;
  int subdomain_ids[NBR_SUBDOMAIN_IDS];

  int nbr_fields;
  int global_index[NBR_CELLS];
  int owner_local[NBR_CELLS];

  double * buffer;
  double * buffer_lon;
  double * buffer_lat;
  int * cell_to_vertex;

  int * cell_mask;
  int * field_id;

#ifdef USE_MPI
  MPI_Comm local_comm;
#else
  int local_comm;
#endif

  int size, rank;

#ifdef USE_MPI
  MPI_Init (0, NULL);
#endif

  // Initialise the coupler
  yac_cinit ( xml_filename, xsd_filename );

  // Inform the coupler about what we are
  yac_cdef_comp ( comp_name, &comp_id );
  comp_ids[0] = comp_id;

  yac_cget_localcomm ( &local_comm, comp_id );

#ifdef USE_MPI
  MPI_Comm_rank ( local_comm, &rank );
  MPI_Comm_size ( local_comm, &size );
#else
  rank = 0;
  size = 1;
#endif

  printf (" %s rank %d : local size %d \n", comp_name, rank, size );

  // Announce one subdomain (patch) to the coupler
  yac_cdef_subdomain ( comp_id, grid_name, &subdomain_id );

  subdomain_ids[0] = subdomain_id;

  buffer_lon = malloc ( nbr_vertices * sizeof(*buffer_lon) );
  buffer_lat = malloc ( nbr_vertices * sizeof(*buffer_lat) );
  cell_to_vertex = malloc ( 3 * nbr_cells * sizeof(*cell_to_vertex) );

  /* Define vertices

               0
              / \
             / o \
            /     \
           1-------2   Eq.
            \     /
             \ o /
              \ /
               3

  */

  buffer_lon[0] =  0.0; buffer_lat[0] =  1.1;
  buffer_lon[1] = -1.0; buffer_lat[1] =  0.0;
  buffer_lon[2] =  1.0; buffer_lat[2] =  0.0;
  buffer_lon[3] =  0.0; buffer_lat[3] = -1.0;

  // Connectivity

  cell_to_vertex[0] = 0; cell_to_vertex[1] = 1; cell_to_vertex[2] = 2; // cell 1
  cell_to_vertex[3] = 1; cell_to_vertex[4] = 3; cell_to_vertex[5] = 2; // cell 2

  for ( i = 0; i < nbr_cells; ++i ) nbr_vertices_per_cell[i] = 3;

  // Description of elements, here as unstructured grid

  yac_cdef_elements_unstruct ( subdomain_id,
                               nbr_vertices,
                               nbr_cells,
                               nbr_vertices_per_cell,
                               buffer_lon,
                               buffer_lat,
                               cell_to_vertex );

  // Center points in cells (needed e.g. for patch recovery and nearest neighbour)

  buffer_lon[0] = 0.0; buffer_lat[0] =  0.5;
  buffer_lon[1] = 0.0; buffer_lat[1] = -0.5;

  yac_cdef_points_unstruct ( subdomain_id,
                             nbr_cells,
                             CELL,
                             buffer_lon,
                             buffer_lat,
                             &cell_point_ids[0] );

  free ( buffer_lon );
  free ( buffer_lat );

  // Decomposition information

  for ( i = 0; i < nbr_cells; ++i ) {
    global_index[i] = i;
    owner_local[i] = -1;
  }

  yac_cdef_index_location ( subdomain_id,
                            nbr_cells,
                            CELL,
                            global_index,
                            owner_local );

  // Connect subdomains
  yac_cconnect_subdomains ( comp_id,
                            nbr_subdomain_ids,
                            subdomain_ids,
                            &domain_id );

  // Mask generation

  cell_mask = malloc (nbr_cells * sizeof(*cell_mask));

  for ( i = 0; i < nbr_cells; ++i )
    cell_mask[i] = 1;

  yac_cdef_mask ( cell_mask,
                  cell_point_ids[0],
                  &cell_mask_ids[0] );

  free (cell_mask);

  field_id = malloc ( no_of_fields * sizeof(*field_id));

  for ( i = 0; i < no_of_fields; ++i )
    yac_cdef_field ( fieldName[i],
                     comp_id,
                     domain_id,
                     cell_point_ids,
                     cell_mask_ids,
                     1,
                     &field_id[i] );

  yac_csearch ( 1, comp_ids, no_of_fields, field_id, &ierror );

  yac_cget_pair_rootcomm ( &local_comm, comp_name, "dummy_ocean" );
  MPI_Comm_size ( local_comm, &size );
  printf (" %s rank %d : size of pair is %d \n", comp_name, rank, size );

  // these queries can only be fulfilled once the search has been completed.

  char * timestep_string;
  timestep_string = (char *) calloc ( 132, sizeof(timestep_string));

  for ( i = 0; i < no_of_fields; ++i ) {
    yac_cget_model_timestep_id ( field_id[i], timestep_string );
    printf ( "Field ID %d: model step %s \n", field_id[i], timestep_string);
  }

  for ( i = 0; i < no_of_fields; ++i ) {
    yac_cget_coupling_timestep_name ( fieldName[i], timestep_string );
    printf ( "Field Name %s: coupling step: %s \n", fieldName[i], timestep_string);
  }
  free (timestep_string);

  free ( field_id );

  // Data exchange

  buffer = malloc (5 * nbr_cells * sizeof (*buffer));

  /* field_id[0] represents "TAUX"   wind stress component
     field_id[1] represents "TAUY"   wind stress component
     field_id[2] represents "SFWFLX" surface fresh water flux
     field_id[3] represents "SFTEMP" surface temperature
     field_id[4] represents "THFLX"  total heat flux
     field_id[5] represents "ICEATM" ice temperatures and melt potential

     field_id[6] represents "SST"    sea surface temperature
     field_id[7] represents "OCEANU" u component of ocean surface current
     field_id[8] represents "OCEANV" v component of ocean surface current
     field_id[9]represents "ICEOCE" ice thickness, concentration and temperatures
  */

  yac_cget_nbr_fields ( &nbr_fields );
  field_id = malloc ( nbr_fields * sizeof(*field_id));

  yac_cget_field_ids ( nbr_fields, field_id );

  // Send fields to ocean
  // --------------------

  {
    double *point_set_data[5];
    double **domain_data[5];
    double ***collection_data[5];

    for (int i = 0; i < 5; ++i) {
      point_set_data[i] = buffer + i * nbr_cells;
      domain_data[i] = &(point_set_data[i]);
      collection_data[i] = &(domain_data[i]);
    }

    // meridional wind stress
    for ( i = 0; i < nbr_cells; ++i ) buffer[0*nbr_cells+i] = 10.1;
    for ( i = 0; i < nbr_cells; ++i ) buffer[1*nbr_cells+i] = 10.2;

    yac_cput ( field_id[0], 2, collection_data, &info, &ierror );
    if ( info > 0 ) printf ( "atmosphere CPL TAUX 1 %f \n", buffer[0*nbr_cells+0]);
    if ( info > 0 ) printf ( "atmosphere CPL TAUX 2 %f \n", buffer[1*nbr_cells+0]);

    // zonal  wind stress
    for ( i = 0; i < nbr_cells; ++i ) buffer[0*nbr_cells+i] = 20.1;
    for ( i = 0; i < nbr_cells; ++i ) buffer[1*nbr_cells+i] = 20.2;

    yac_cput ( field_id[1], 2, collection_data, &info, &ierror );
    if ( info > 0 ) printf ( "atmosphere CPL TAUY 1 %f \n", buffer[0*nbr_cells+0]);
    if ( info > 0 ) printf ( "atmosphere CPL TAUY 2 %f \n", buffer[1*nbr_cells+0]);

    // surface fresh water flux
    for ( i = 0; i < nbr_cells; ++i ) buffer[0*nbr_cells+i] = 30.1;
    for ( i = 0; i < nbr_cells; ++i ) buffer[1*nbr_cells+i] = 30.2;
    for ( i = 0; i < nbr_cells; ++i ) buffer[2*nbr_cells+i] = 30.3;

    yac_cput ( field_id[2], 3, collection_data, &info, &ierror );

    // surface temperature

    for ( i = 0; i < nbr_cells; ++i ) buffer[0*nbr_cells+i] = 40.1;

    yac_cput ( field_id[3], 1, collection_data, &info, &ierror );

    // total heat flux

    for ( i = 0; i < nbr_cells; ++i ) buffer[0*nbr_cells+i] = 50.1;
    for ( i = 0; i < nbr_cells; ++i ) buffer[1*nbr_cells+i] = 50.2;
    for ( i = 0; i < nbr_cells; ++i ) buffer[2*nbr_cells+i] = 50.3;
    for ( i = 0; i < nbr_cells; ++i ) buffer[3*nbr_cells+i] = 50.4;


    yac_cput ( field_id[4], 4, collection_data, &info, &ierror );

    // ice temperatures and melt potential

    for ( i = 0; i < nbr_cells; ++i ) buffer[0*nbr_cells+i] = 60.1;
    for ( i = 0; i < nbr_cells; ++i ) buffer[1*nbr_cells+i] = 60.2;
    for ( i = 0; i < nbr_cells; ++i ) buffer[2*nbr_cells+i] = 60.3;
    for ( i = 0; i < nbr_cells; ++i ) buffer[3*nbr_cells+i] = 60.4;

    yac_cput ( field_id[5], 4, collection_data, &info, &ierror );
  }

  //
  // Receive fields from ocean
  // -------------------------

  {
    double *domain_data[5];
    double **collection_data[5];

    for (int i = 0; i < 5; ++i) {
      domain_data[i] = buffer + i * nbr_cells;
      collection_data[i] = &(domain_data[i]);
    }

    // SST

    yac_cget ( field_id[6], 1, collection_data, &info, &ierror );
    if ( info > 0 ) printf ( "atmosphere CPL SST %f \n", buffer[0] );

    // zonal velocity

    yac_cget ( field_id[7], 1, collection_data, &info, &ierror );
    if ( info > 0 ) printf ( "atmosphere CPL OCEANU %f \n", buffer[0] );

    // meridional velocity

    yac_cget ( field_id[8], 1, collection_data, &info, &ierror );
    if ( info > 0 ) printf ( "atmosphere CPL OCEANV %f \n", buffer[0] );

    // Ice thickness, concentration, T1 and T2

    yac_cget ( field_id[9], 5, collection_data, &info, &ierror );
    if ( info > 0 ) {
      printf ( "atmosphere CPL ice 1 %f \n", buffer[0*nbr_cells+0] );
      printf ( "atmosphere CPL ice 2 %f \n", buffer[1*nbr_cells+0] );
      printf ( "atmosphere CPL ice 3 %f \n", buffer[2*nbr_cells+0] );
      printf ( "atmosphere CPL ice 4 %f \n", buffer[3*nbr_cells+0] );
      printf ( "atmosphere CPL ice 5 %f \n", buffer[4*nbr_cells+0] );
    }
  }

  free (buffer);
  free (field_id);

  yac_cfinalize ();

#ifdef USE_MPI
  MPI_Finalize ();
#endif

  return 0;
}
