/**
 * @file explore_patch_recovery.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#define ANSI_COLOR_RED     "\x1b[31m"
#define ANSI_COLOR_GREEN   "\x1b[32m"
#define ANSI_COLOR_BLUE    "\x1b[34m"
#define ANSI_COLOR_RESET   "\x1b[0m"

#define GNOMONIC

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "grid.h"
#include "utils.h"
#include "geometry.h"
#include "patch_recovery.h"

int main () {

  unsigned nbr_points;

  int i, n, m;
  unsigned gauss_order;
  unsigned polynom_order;

  int count;

  int gauss_order_start   = 1;  // min allowed 1
  int gauss_order_end     = 7;  // max allowed 7
  int polynom_order_start = 0;  // min allowed 0
  int polynom_order_end   = 3;  // max allowed 3

  double *triangle_location;    // location in real triangle

  double *abscissa;             // location in reference element
  double reference[3];
  double target_location[3];

  double rms_error[8][4];
  double error[8][4];
  double rms;

  /* trigger of the number of target points investigated */

  int ndim = 10;
  int mdim = 10;

  double dx = 2.0/ndim;
  double dy = 2.0/ndim;

  unsigned gauss_order_min = 0;
  unsigned gauss_order_max = 0;

  unsigned poly_order_min  = 0;
  unsigned poly_order_max  = 0;

  double error_min =  9999999.9;
  double error_max = -9999999.9;

  double error_rel = 0.0;

  /* patch recovery test
     -------------------

     Here we explore the whole range of Gauss integration points and polynomial.

     There is a significant reduction in the error when moving away from the node
     to the Gauss points, i.e. from Gauss order 0 to order 1. This improves significantly
     if we take the average of all surrounding cells for a particular node.
     In this case we can exactly reproduce the linear function.

     With constant 0 order polynomial the error remains
     constant and high for all sets of Gauss points.

     When moving to Gauss points, the linear fits give best results.
     There is no benefit in using a higher order of Gauss points.

     For this example higher order polynomial give worse results. This
     is expected as our test function has a linear dependency. 2nd and
     3rd order polynomial produce errors on the same order.

     Here we consider a whole set of target points which all are located
     inside triangle 7. For individual points a much smaller error can be
     obtained with a combination which differs from the one finally selected
     (marked in green) based on the results for the whole set.

     Define a patch of triangles

     |     Element numbers are given in the middle of the triangles
     |     Node number are given at each node
     | 
     6                     10________11
     |                    /\        /\
     |                   /  \  11  /  \
     |                  /    \    /    \
     |                 /  10  \  /  12  \
     4                7________\8________9
     |               /\        /\        /\
     |              /  \  6   /  \  8   /  \
     |             /    \    /    \    /    \
     |            /  5   \  /  7   \  /  9   \
     2           3________\4________\5________6
     |           \        /\        /\        /
     |            \  0   /  \  2   /  \  4   /
     |             \    /    \    /    \    /
     |              \  /  1   \  /  3   \  /
     0               \0________\1________\2
     |
     |_________________________________________________  
                0         2         4          6

     Element number 7 is the center element around which
     the test patch is constructed.

     Elements 1, 0, 5         with value 2.0
     Elements 2, 3, 6,  7, 10 with value 3.0
     Elements 4, 8, 9, 11, 12 with value 4.0

     For the test function we assume a constant linear increase
     in the direction normal to edge 0-3 from 1.5 along edge 0-3
     to 2.5 along edge 1-4-7 to 3.5 along edge 2-5-8-10 and
     4.5 along 6-9-11 .

     Alternative: constant value of 2.0 for each element

*/

#define NUM_CELLS 13

  unsigned nbr_sampling_points;

  unsigned nbr_total_gauss_points[NUM_CELLS];

  unsigned static num_cells = NUM_CELLS;

  double node_coordinates[2*(NUM_CELLS-1)] = { 1.0, 0.0, 3.0, 0.0, 5.0, 0.0,
                                               0.0, 2.0, 2.0, 2.0, 4.0, 2.0, 6.0, 2.0,
                                               1.0, 4.0, 3.0, 4.0, 5.0, 4.0,
                                               2.0, 6.0, 4.0, 6.0 };

  unsigned connectivity[3*NUM_CELLS] = { 0,  4,  3,
                                         0,  1,  4,
                                         1,  5,  4,
                                         1,  2,  5,
                                         2,  6,  5,
                                         3,  4,  7,
                                         4,  8,  7,
                                         4,  5,  8,
                                         5,  9,  8,
                                         5,  6,  9,
                                         7,  8, 10,
                                         8, 11, 10,
                                         8,  9, 11 };

  /* Define values at the center of the triangles */

  double cell_value[NUM_CELLS] = { 2.0, 2.0, 3.0, 3.0, 4.0,
                                   2.0, 3.0, 3.0, 4.0, 4.0,
                                   3.0, 4.0, 4.0 };

  /* Construct cells */

  struct xyz_triangle cell[NUM_CELLS];

  for ( i = 0; i < num_cells; ++i ) {
     for ( n = 0; n < 3; ++n ) {

        int ci = connectivity[i*3+n];

        if ( n == 0 )
           LLtoXYZ ( node_coordinates[ci*2+0]*rad, node_coordinates[ci*2+1]*rad, cell[i].a );
        if ( n == 1 )
           LLtoXYZ ( node_coordinates[ci*2+0]*rad, node_coordinates[ci*2+1]*rad, cell[i].b );
        if ( n == 2 )
           LLtoXYZ ( node_coordinates[ci*2+0]*rad, node_coordinates[ci*2+1]*rad, cell[i].c );

     }
  }

  /* Set the reference coordinates for the patch */

  double reference_tmp[3];

  reference_tmp[0] = node_coordinates[4*2];
  reference_tmp[1] = node_coordinates[4*2+1];
  reference_tmp[2] = 0.0;

  yac_reference_to_xyz_triangle ( 1, cell[7], reference_tmp, reference );

  /* Initialise error */

  for ( gauss_order = gauss_order_start; gauss_order <= gauss_order_end; ++gauss_order )
    for ( polynom_order = polynom_order_start; polynom_order <= polynom_order_end; ++polynom_order )
       error[gauss_order][polynom_order] = 0.0;

  /* Start main loops */

  for ( gauss_order = gauss_order_start; gauss_order <= gauss_order_end; ++gauss_order ) {
     for ( polynom_order = polynom_order_start; polynom_order <= polynom_order_end; ++polynom_order ) {

        nbr_points = yac_get_number_of_quadrature_points (gauss_order);

        for ( i = 0; i < num_cells; ++i )
           nbr_total_gauss_points[i] = nbr_points;

        abscissa = malloc ( 2 * nbr_points * sizeof(*abscissa) );

        nbr_sampling_points = num_cells * nbr_points;

        yac_triangle_gauss_legendre ( gauss_order, abscissa );

        /* Get the gauss integration points (triangle_location)
           for each cell in the patch.

           For simplicity we first assume that we have a P0 element,
           i.e. the value is constant of the whole element. */

        triangle_location = malloc ( 3 * nbr_sampling_points * sizeof(*triangle_location) );

        double * abscissa_3d = malloc ( 3 * nbr_points * sizeof(*abscissa) );

        for ( i = 0; i < nbr_points; ++ i ) {
           abscissa_3d[i*3+0] = abscissa[i*2+0];
           abscissa_3d[i*3+1] = abscissa[i*2+1];
           abscissa_3d[i*3+2] = 0.0;
        }

        for ( i = 0; i < num_cells; ++ i )
           yac_reference_to_xyz_triangle ( nbr_points, cell[i], abscissa_3d, &triangle_location[i*3*nbr_points] );

        free (abscissa_3d);

        /* Here we now loop over a whole set of target points all of which are located inside source element 7 */

        count = 0;

        unsigned nbr_coeff = yac_get_num_polynomial_coefficients ( polynom_order );

        rms = 0.0;

        /* a sufficient number of sampling points is required to get reasonable results */

        if ( nbr_sampling_points >= 2 * nbr_coeff ) {

           /* Determine Polynomial terms P for each integration point */

           double *PT = malloc ( nbr_sampling_points*nbr_coeff * sizeof(*PT) );

           yac_set_polynomial_terms ( polynom_order, nbr_sampling_points, triangle_location, reference, PT );

           double *target_pterm = malloc ( nbr_coeff * sizeof(*target_pterm) );

           double *weights = malloc ( num_cells * sizeof(*weights) );

           for ( n = 0; n <= ndim; ++n ) {
              for ( m = 0; m <= mdim; ++m ) {

                 if ( abs(n-ndim/2)*dx + 0.5*m*dy <= 1.0 ) {

                    /* the reference value to compare with has the following characteristics:

                         1.5  2.0  2.5  3.0
                       1.0  1.5  2.0  2.5  3.0
                         1.0  1.5  2.0  2.5  3.0
                           1.0  1.5  2.0  2.5  3.0
                             1.0  1.5  2.0  2.5  3.0
                               1.0  1.5  2.0  2.5  3.0
                                 1.0  1.5  2.0  2.5
                     */

                    target_location[0] = 3.0 + (n-ndim/2)*dx;
                    target_location[1] = 2.0 + m*dy;

                    double ref_value = ( 1.0 + 0.5 * target_location[0] + 0.25 * target_location[1] );

                    /* Set location of target in 3D Cartesian */

                    LLtoXYZ( target_location[0]*rad, target_location[1]*rad, target_location );

                    /* 2nd) Get the P for the target point */

                    yac_set_polynomial_terms ( polynom_order, 1, target_location, reference, target_pterm );

                    /* 3rd) Get the new value at the target point */

                    yac_patch_P0weights ( nbr_sampling_points, nbr_coeff, num_cells, nbr_total_gauss_points, target_pterm, PT, weights );

                    double value = yac_patch_P0value ( num_cells, weights, cell_value );

                    error_rel = fabs(value-ref_value) / ref_value * 100.0;

                    error[gauss_order][polynom_order] += error_rel;
                    rms += (value-ref_value) * (value-ref_value);
                    count++;
                 }
              }
           }

           free(target_pterm);
           free(PT);
           free(weights);

           error[gauss_order][polynom_order] /= (double) count;
           rms /= (double) count;
           rms_error[gauss_order][polynom_order] = sqrt(rms);
        }
        else
           printf ("skipped gauss order %d polynomial order %d\n", gauss_order, polynom_order );

        free(triangle_location);

        if ( gauss_order > 0 ) free(abscissa);
     }
  }

  /* Analyse error results */

  for ( gauss_order = gauss_order_start; gauss_order <= gauss_order_end; ++gauss_order ) {
     for ( polynom_order = polynom_order_start; polynom_order <= polynom_order_end; ++polynom_order ) {

        nbr_points = yac_get_number_of_quadrature_points (gauss_order);
        nbr_sampling_points = num_cells * nbr_points;
        if ( gauss_order == 0 ) nbr_sampling_points = num_cells - 1;

        unsigned nbr_coeff = yac_get_num_polynomial_coefficients ( polynom_order );

        if ( nbr_sampling_points >= 2 * nbr_coeff ) {

           double test;

           test = MIN(error_min,error[gauss_order][polynom_order]) ;

           if ( test < error_min ) {
              error_min = test;
              gauss_order_min = gauss_order;
              poly_order_min  = polynom_order;
           }

           test = MAX(error_max,error[gauss_order][polynom_order]);

           if ( test > error_max ) {
              error_max = test;
              gauss_order_max = gauss_order;
              poly_order_max  = polynom_order;
           }
        }
     }
  }

  printf ("statistics are based on %d points\n", count);

  for ( polynom_order = polynom_order_start; polynom_order <= polynom_order_end; ++polynom_order ) {
     for ( gauss_order = gauss_order_start; gauss_order <= gauss_order_end; ++gauss_order ) {

        nbr_points = yac_get_number_of_quadrature_points (gauss_order);
        nbr_sampling_points = num_cells * nbr_points;
        if ( gauss_order == 0 ) nbr_sampling_points = num_cells - 1;

        unsigned nbr_coeff = yac_get_num_polynomial_coefficients ( polynom_order );

        if ( nbr_sampling_points >= 2 * nbr_coeff ) {

           if ( gauss_order == gauss_order_min && polynom_order == poly_order_min )
              printf (ANSI_COLOR_GREEN
                    "mean error for gauss order %i with polynomial fit of degree %i is %f per cent, rms error %f" ANSI_COLOR_RESET "\n",
                    gauss_order, polynom_order, error[gauss_order][polynom_order], rms_error[gauss_order][polynom_order] );
           else if ( gauss_order == gauss_order_max && polynom_order == poly_order_max )
              printf (ANSI_COLOR_RED
                    "mean error for gauss order %i with polynomial fit of degree %i is %f per cent, rms error %f"ANSI_COLOR_RESET "\n",
                    gauss_order, polynom_order, error[gauss_order][polynom_order], rms_error[gauss_order][polynom_order]);
           else
              printf ("mean error for gauss order %i with polynomial fit of degree %i is %f per cent, rms error %f\n",
                    gauss_order, polynom_order, error[gauss_order][polynom_order], rms_error[gauss_order][polynom_order]);
        }
     }
  }
  return EXIT_SUCCESS;
}
