// #define VERBOSE

/**
 * @file perf_toy_cube.c
 *
 * @copyright Copyright  (C)  2015 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if defined USE_MPI && defined HAVE_C_NETCDF
#include <mpi.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "config_xml.h"
#include "tests.h"
#include "utils.h"
#include "component.h"
#include "fields.h"
#include "yac_interface.h"
#include "geometry.h"
#include "vtk_output.h"
#include "generate_cubed_sphere.h"

/* ------------------------------------------------- */

const char * fieldName[] = {"icon_out", "cube_out"};

void generate_part_cube_grid_information(
  unsigned n, unsigned * nbr_vertices, unsigned * nbr_cells,
  unsigned ** num_vertices_per_cell, unsigned ** cell_to_vertex,
  double ** x_vertices, double ** y_vertices, double ** x_cells,
  double ** y_cells, int ** global_cell_id, int ** global_cell_id_rank,
  int ** global_corner_id, int ** global_corner_id_rank, int rank, int size);

int main () {

#ifdef USE_XML2

  // Initialisation of MPI

  MPI_Init (0, NULL);

  /* The initialisation phase includes the reading of the XML file */
#ifdef VERBOSE
  printf (". main: calling yac_cinit\n");
#endif

  double tic, toc, time;

  MPI_Barrier(MPI_COMM_WORLD);

  tic=MPI_Wtime();

  yac_cinit ("perf_toy.xml", "coupling.xsd");

  /* The usual component definition, here for two sequential components on the
     same process */

#ifdef VERBOSE
  printf (". main: calling yac_cdef_comp\n");
#endif

  int comp_id;
  char * comp_name = "CUBE";

  yac_cdef_comp ( comp_name, &comp_id );
#ifdef VERBOSE
  printf(". main: defined %s with local comp ID %i \n", comp_name, comp_id);
#endif

  MPI_Comm local_comm;

  yac_cget_localcomm(&local_comm, comp_id);

  int rank, size;

  MPI_Comm_rank(local_comm,&rank);
  MPI_Comm_size(local_comm,&size);

  int ierror = 0;

  int nbr_fields = 2;
  int cell_point_id;

  int field_ids[2];
  int domain_id;
  int subdomain_id;

  unsigned n = 405;
  // unsigned n = 202;

  unsigned nbr_vertices;
  unsigned nbr_cells;
  unsigned * num_vertices_per_cell;
  unsigned * cell_to_vertex;
  double * x_vertices;
  double * y_vertices;
  double * x_cells;
  double * y_cells;

  int * global_cell_id;
  int * global_cell_id_rank;
  int * global_corner_id;
  int * global_corner_id_rank;

  generate_part_cube_grid_information(n, &nbr_vertices, &nbr_cells,
                                      &num_vertices_per_cell, &cell_to_vertex,
                                      &x_vertices, &y_vertices, &x_cells,
                                      &y_cells, &global_cell_id,
                                      &global_cell_id_rank, &global_corner_id,
                                      &global_corner_id_rank, rank, size);

  double * x_points, * y_points;

  x_points = x_vertices;
  y_points = y_vertices;

  for (unsigned i = 0; i < nbr_vertices; ++i) {
    x_points[i] *= deg;
    y_points[i] *= deg;
  }

  double * x_center, * y_center;

  x_center = x_cells;
  y_center = y_cells;

  for (unsigned i = 0; i < nbr_cells; ++i) {
    x_center[i] *= deg;
    y_center[i] *= deg;
  }

  yac_cdef_subdomain(comp_id, "cube_grid", &subdomain_id);

  yac_cdef_elements_unstruct(subdomain_id, nbr_vertices, nbr_cells,
                             (int*)num_vertices_per_cell, x_vertices,
                             y_vertices, (int*)cell_to_vertex);

  yac_cdef_points_unstruct(subdomain_id, nbr_cells, CELL, x_center, y_center,
                           &cell_point_id);

  yac_cdef_index_location(subdomain_id, nbr_cells, CELL, global_cell_id,
                          global_cell_id_rank);

  yac_cconnect_subdomains(comp_id, 1, &subdomain_id, &domain_id);

  /* Field definition for the component */

#ifdef VERBOSE
  struct component * component = yac_unique_id_to_pointer(comp_id);
  int global_comp_id = yac_component_get_global_id(component);
  printf (". main: number of fields in component %i is %i\n", global_comp_id, nbr_fields);
#endif

  yac_cdef_field(fieldName[0], comp_id, domain_id, &cell_point_id, NULL, 1,
                 &field_ids[0]);
  yac_cdef_field(fieldName[1], comp_id, domain_id, &cell_point_id, NULL, 1,
                 &field_ids[1]);
  
  toc=MPI_Wtime();
  time = toc-tic;
  printf ("CUBE: Time for initialisation %f\n", time ); 

  /* Search. */

#ifdef VERBOSE
  printf (". main: calling yac_csearch\n");
#endif

  MPI_Barrier(MPI_COMM_WORLD);

  tic=MPI_Wtime();

  yac_csearch(1, &comp_id, nbr_fields, field_ids, &ierror);

  toc=MPI_Wtime();
  time = toc-tic;
  printf("CUBE: Time for search %f\n", time ); 

  double * conserv_in = malloc(nbr_cells * sizeof(*conserv_in));

  for (unsigned i = 0; i < nbr_cells; ++i)
    conserv_in[i] = -10;

  double * cell_out = malloc(nbr_cells * sizeof(*cell_out));

  int err;
  int info;

  for (int i = 0; i < nbr_cells; ++i)
    cell_out[i] = sin(x_center[i] * 2 * rad) * cos(y_center[i] * 3 * rad);

  MPI_Barrier(MPI_COMM_WORLD);

  tic=MPI_Wtime();

  {
    double *point_set_data[1];
    double **domain_data[1] = {point_set_data};
    double ***collection_data[1] = {domain_data};

    point_set_data[0] = cell_out;
    yac_cput(field_ids[1], 1, collection_data, &info, &err);
  }

  {
    double *domain_data[1];
    double **collection_data[1] = {domain_data};

    domain_data[0] = conserv_in;
    yac_cget(field_ids[0], 1, collection_data, &info, &err);
  }

  toc=MPI_Wtime();
  time = toc-tic;
  printf ("CUBE: Time for ping-pong %f\n", time );

#ifdef VTK_OUTPUT
  //----------------------------------------------------------
  // write field to vtk output file
  //----------------------------------------------------------

  char vtk_filename[32];

  sprintf(vtk_filename, "perf_toy_cube_%d.vtk", rank);

  double point_data[nbr_vertices][3];
  for (int i = 0; i < nbr_vertices; ++i) {
   LLtoXYZ_deg(x_vertices[i], y_vertices[i], point_data[i]);
  }

  VTK_FILE *vtk_file = vtk_open(vtk_filename, "cube_out");
  vtk_write_point_data(vtk_file, (double *)point_data, nbr_vertices);
  vtk_write_cell_data(vtk_file, (unsigned *)cell_to_vertex,
                      (unsigned*)num_vertices_per_cell, nbr_cells);
  vtk_write_point_scalars_int(vtk_file, global_corner_id_rank, nbr_vertices,
                              "vertex_is_on_rank");
  vtk_write_point_scalars_int(vtk_file, global_corner_id, nbr_vertices,
                              "global_corner_id");
  vtk_write_cell_scalars_int(vtk_file, global_cell_id_rank, nbr_cells,
                             "cell_is_on_rank");
  vtk_write_cell_scalars_int(vtk_file, global_cell_id, nbr_cells,
                             "global_cell_id");

  vtk_write_cell_scalars_double(vtk_file, conserv_in, nbr_cells, "conserv_in");
  vtk_write_cell_scalars_double(vtk_file, cell_out, nbr_cells, "cell_out");

  vtk_close(vtk_file);
#endif // VTK_OUTPUT

  free(global_corner_id_rank);
  free(global_corner_id);
  free(global_cell_id_rank);
  free(global_cell_id);
  free(x_cells);
  free(y_cells);
  free(x_vertices);
  free(y_vertices);
  free(num_vertices_per_cell);
  free(cell_to_vertex);

  yac_cfinalize();

  MPI_Finalize();

  free(conserv_in);
  free(cell_out);

  return EXIT_SUCCESS;

#else

  fprintf(stderr, "ERROR: perf_toy_cube requires XML2\n");

  return EXIT_FAILURE;

#endif // USE_XML2
}

void decompose_domain_simple(unsigned n, int size, int * cell_owner) {

  unsigned nbr_cells = n * n * 6;

  for (unsigned i = 0; i < nbr_cells; ++i)
    cell_owner[i] = (i * size + size - 1) / nbr_cells;
}

void compute_reg2d_decomp(int size, int * num_procs) {

  int best_num_procs[2] = {size, 1};
  float curr_proc_ratio = (float)size;
  float best_proc_ratio = (float)size;

  for (int i = size - 1; i > 0; --i) {

    if (size % i == 0) {

      curr_proc_ratio = (float)(i) / (float)(size/i);

      if (fabs(1.0 - curr_proc_ratio) < fabs(1.0 - best_proc_ratio)) {
        best_num_procs[0] = i;
        best_num_procs[1] = size/i;
        best_proc_ratio = curr_proc_ratio;
      }
    }
  }

  num_procs[0] = best_num_procs[0];
  num_procs[1] = best_num_procs[1];
}

void decompose_domain_2d(unsigned n, int size, int * cell_owner) {

  // distribute processes among all six sides of the cubed sphere
  int proc_start[6 + 1];
  for (int i = 0; i < 6 + 1; ++i)
    proc_start[i] = (size * i) / 6;

  // for each side of the cube
  for (unsigned i = 0, offset = 0; i < 6; ++i) {

    int num_procs[2];
    int curr_num_procs = proc_start[i+1] - proc_start[i];

    compute_reg2d_decomp(curr_num_procs, num_procs);

    int local_start_x[num_procs[0] + 1];
    int local_start_y[num_procs[1] + 1];

    for (int j = 0; j < num_procs[0] + 1; ++j)
      local_start_x[j] = (n * j) / num_procs[0];
    for (int j = 0; j < num_procs[1] + 1; ++j)
      local_start_y[j] = (n * j) / num_procs[1];

    for (int p_y = 0; p_y < num_procs[1]; ++p_y) {
      for (int row = local_start_y[p_y]; row < local_start_y[p_y+1]; ++row) {
        for (int p_x = 0; p_x < num_procs[0]; ++p_x) {
          for (int column = local_start_x[p_x]; column < local_start_x[p_x+1];
               ++column) {

            cell_owner[offset++] = proc_start[i] + p_x + num_procs[0] * p_y;
          }
        }
      }
    }
  }
}

void decompose_domain(unsigned n, int size, int * cell_owner) {

  if (size <= 6) {
    decompose_domain_simple(n, size, cell_owner);
  } else {
    decompose_domain_2d(n, size, cell_owner);
  }
}

void generate_part_cube_grid_information(
  unsigned n, unsigned * nbr_vertices, unsigned * nbr_cells,
  unsigned ** num_vertices_per_cell, unsigned ** cell_to_vertex,
  double ** x_vertices, double ** y_vertices, double ** x_cells,
  double ** y_cells, int ** global_cell_id, int ** global_cell_id_rank,
  int ** global_corner_id, int ** global_corner_id_rank, int rank, int size) {

  double * x_vertices_3d, * y_vertices_3d, * z_vertices_3d;
  unsigned * dummy;

  // generate global grid
  generate_cubed_sphere_grid_information(n, nbr_cells, nbr_vertices,
                                         &x_vertices_3d, &y_vertices_3d,
                                         &z_vertices_3d, cell_to_vertex,
                                         &dummy);
  *num_vertices_per_cell = malloc(*nbr_cells * sizeof(**num_vertices_per_cell));
  for (unsigned i = 0; i < *nbr_cells; ++i) (*num_vertices_per_cell)[i] = 4;
  free(dummy);

  double * x_cell_3d = malloc(*nbr_cells * sizeof(*x_cell_3d));
  double * y_cell_3d = malloc(*nbr_cells * sizeof(*y_cell_3d));
  double * z_cell_3d = malloc(*nbr_cells * sizeof(*z_cell_3d));

  for (unsigned i = 0; i < *nbr_cells; ++i) {

    x_cell_3d[i] = y_cell_3d[i] = z_cell_3d[i] = 0;

    for (unsigned j = 0; j < 4; ++j) {

      x_cell_3d[i] += x_vertices_3d[(*cell_to_vertex)[4 * i + j]];
      y_cell_3d[i] += y_vertices_3d[(*cell_to_vertex)[4 * i + j]];
      z_cell_3d[i] += z_vertices_3d[(*cell_to_vertex)[4 * i + j]];
    }

    double scale = 1.0 / sqrt(x_cell_3d[i] * x_cell_3d[i] +
                              y_cell_3d[i] * y_cell_3d[i] +
                              z_cell_3d[i] * z_cell_3d[i]);

    x_cell_3d[i] *= scale;
    y_cell_3d[i] *= scale;
    z_cell_3d[i] *= scale;
  }

  int * cell_is_on_rank = malloc(*nbr_cells * sizeof(*cell_is_on_rank));

  decompose_domain(n, size, cell_is_on_rank);

  // mask for required vertices and cells
  int * required_vertices = calloc(*nbr_vertices, sizeof(*required_vertices));
  int * required_cells = calloc(*nbr_cells, sizeof(*required_cells));

  // mark all local cells and their vertices as required
  for (int i = 0, offset = 0; i < *nbr_cells;
       offset += (*num_vertices_per_cell)[i++]) {

    if (cell_is_on_rank[i] != rank) continue;

    cell_is_on_rank[i] = -1;

    required_cells[i] = 1;
    for (int j = 0; j < (*num_vertices_per_cell)[i]; ++j)
      required_vertices[(*cell_to_vertex)[offset+j]] = 1;
  }

  // mark all halo cells as required and generate cell_is_on_rank
  for (int i = 0, offset = 0; i < *nbr_cells;
       offset += (*num_vertices_per_cell)[i++]) {

    if (!required_cells[i]) {

      for (int j = 0; j < (*num_vertices_per_cell)[i]; ++j) {

        if (required_vertices[(*cell_to_vertex)[offset+j]]) {

          required_cells[i] = 1;
          break;
        }
      }

    }
  }

  // mask for halo vertices
  int * vertex_is_on_rank = malloc(*nbr_vertices * sizeof(*vertex_is_on_rank));

  // mark all halo vertices as required and generate vertex_is_on_rank
  for (int i = 0; i < *nbr_vertices; ++i)
    if (required_vertices[i])
      vertex_is_on_rank[i] = -1;
  for (int i = 0, offset = 0; i < *nbr_cells;
       offset += (*num_vertices_per_cell)[i++]) {

    if (required_cells[i] && cell_is_on_rank[i] != -1) {

      for (int j = 0; j < (*num_vertices_per_cell)[i]; ++j) {

        if (!required_vertices[(*cell_to_vertex)[offset+j]]) {

          required_vertices[(*cell_to_vertex)[offset+j]] = 1;
          vertex_is_on_rank[(*cell_to_vertex)[offset+j]] = cell_is_on_rank[i];
        }
      }
    }
  }

  // count the number of cells and vertices
  int part_num_vertices = 0;
  int part_num_cells = 0;
  for (int i = 0; i < *nbr_vertices; ++i)
    if (required_vertices[i])
      part_num_vertices++;
  for (int i = 0; i < *nbr_cells; ++i)
    if(required_cells[i])
      part_num_cells++;

  *global_cell_id = malloc(part_num_cells * sizeof(**global_cell_id));
  *global_cell_id_rank = malloc(part_num_cells * sizeof(**global_cell_id_rank));
  *global_corner_id = malloc(part_num_vertices * sizeof(**global_corner_id));
  *global_corner_id_rank = malloc(part_num_vertices * sizeof(**global_corner_id_rank));

  *x_cells = malloc(part_num_cells * sizeof(**x_cells));
  *y_cells = malloc(part_num_cells * sizeof(**y_cells));
  *x_vertices = malloc(part_num_vertices * sizeof(**x_vertices));
  *y_vertices = malloc(part_num_vertices * sizeof(**y_vertices));

  // generate final vertex data
  part_num_vertices = 0;
  int * global_to_local_vertex = malloc(*nbr_vertices * sizeof(*global_to_local_vertex));
  for (int i = 0; i < *nbr_vertices; ++i) {

    if (required_vertices[i]) {

      (*global_corner_id)[part_num_vertices] = i;
      (*global_corner_id_rank)[part_num_vertices] = vertex_is_on_rank[i];
      double p[3] = {x_vertices_3d[i], y_vertices_3d[i], z_vertices_3d[i]};
      XYZtoLL(p, (*x_vertices) + part_num_vertices, (*y_vertices) + part_num_vertices);
      global_to_local_vertex[i] = part_num_vertices;
      part_num_vertices++;
    }
  }

  free(vertex_is_on_rank);
  *nbr_vertices = part_num_vertices;
  free(required_vertices);

  // generate final cell data
  int num_cell_vertex_dependencies = 0;
  part_num_cells = 0;
  for (int i = 0, offset = 0; i < *nbr_cells;
       offset += (*num_vertices_per_cell)[i++]) {

    if (required_cells[i]) {

      (*global_cell_id)[part_num_cells] = i;
      (*global_cell_id_rank)[part_num_cells] = cell_is_on_rank[i];
      double middle_point[3] = {x_cell_3d[i],y_cell_3d[i],z_cell_3d[i]};

      

      for (int j = 0; j < (*num_vertices_per_cell)[i]; ++j) {
        (*cell_to_vertex)[num_cell_vertex_dependencies+j] =
          global_to_local_vertex[(*cell_to_vertex)[offset+j]];
      }

      XYZtoLL(middle_point, (*x_cells)+part_num_cells, (*y_cells)+part_num_cells);

      num_cell_vertex_dependencies += (*num_vertices_per_cell)[i];

      (*num_vertices_per_cell)[part_num_cells] = (*num_vertices_per_cell)[i];

      part_num_cells++;
    }
  }

  free(x_cell_3d);
  free(y_cell_3d);
  free(z_cell_3d);
  free(x_vertices_3d);
  free(y_vertices_3d);
  free(z_vertices_3d);

  *num_vertices_per_cell = realloc(*num_vertices_per_cell, part_num_cells *
                                   sizeof(**num_vertices_per_cell));
  *cell_to_vertex = realloc(*cell_to_vertex, num_cell_vertex_dependencies *
                            sizeof(**cell_to_vertex));
  free(cell_is_on_rank);
  *nbr_cells = part_num_cells;
  free(required_cells);
  free(global_to_local_vertex);
}

#else
#include <stdlib.h>
#include <stdio.h>
int main () {
  printf ("Examples requires compiling with MPI and NetCDF.\n");
  return EXIT_FAILURE;
}
#endif
