/**
 * @file test_grid_coupling.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#define EXACT

#ifdef USE_MPI
#include <mpi.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "yac_interface.h"
#include "points.h"
#include "geometry.h"
#include "vtk_output.h"

const char * fieldName[] = { "AtoB",
                             "BtoA"};

char * xml_filename = "coupling_grid.xml";
char * xsd_filename = "coupling.xsd";

const int no_of_fields = 2;

int i, j, info, ierror;
int comp_id;
int comp_ids[1];

int cell_point_id;

const int nbr_subdomain_ids = 1;
int subdomain_ids[1];
int domain_id;
int subdomain_id;

int nbr_fields;

double * recv_buffer;
double * send_buffer;

double * buffer = NULL;

int * global_index;
int * owner_local;

int * field_id;

void dummy_compA ();
void dummy_compB ();

/* -------------------------------------------------------------------- */

int main (void) {

  int size, rank;

#ifdef USE_MPI
  MPI_Init (0, NULL);

  MPI_Comm_rank ( MPI_COMM_WORLD, &rank );
  MPI_Comm_size ( MPI_COMM_WORLD, &size );

  switch ( rank ) {

  case 0:
    dummy_compA ( );
    break;

  case 1:
    dummy_compB ( );
    break;

  default:
    yac_internal_abort_message ("Too many processes have been launched", __FILE__, __LINE__ );

  }

  MPI_Finalize ();
#endif

}

/* -------------------------------------------------------------------- */

#ifdef USE_MPI

struct point_with_index {

  struct point p;
  int i;
};

int compare_point_with_index(const void * a,const void * b) {

   const struct point_with_index * a_ = (struct point_with_index*)a;
   const struct point_with_index * b_ = (struct point_with_index*)b;

  int lon_diff = fabs(a_->p.lon - b_->p.lon) > 1e-9;
  int lat_diff = fabs(a_->p.lat - b_->p.lat) > 1e-9;

  if (lon_diff) {

    if (a_->p.lon > b_->p.lon) return -1;
    else return 1;

  } else if (lat_diff) {

    if (a_->p.lat > b_->p.lat) return -1;
    else return 1;
  } else
    return 0;
}

void remove_duplicated_vertices(double * temp_vertex_lon,
                                double * temp_vertex_lat,
                                int * temp_nbr_vertices,
                                int * old_to_new_id) {

   struct point_with_index * sort_array =
      malloc(*temp_nbr_vertices * sizeof(*sort_array));

   for (int i = 0; i < *temp_nbr_vertices; ++i) {

      double curr_lon, curr_lat;

      curr_lon = temp_vertex_lon[i];
      curr_lat = temp_vertex_lat[i];

      while (curr_lon < 0.0) curr_lon += 2.0 * M_PI;
      while (curr_lon >= 2.0 * M_PI) curr_lon -= 2.0 * M_PI;

      sort_array[i].p.lon = curr_lon;
      sort_array[i].p.lat = curr_lat;

      sort_array[i].i = i;
   }

   yac_mergesort(sort_array, *temp_nbr_vertices, sizeof(*sort_array),
                 compare_point_with_index);

   old_to_new_id[sort_array[0].i] = 1;

   int last_unique_idx = sort_array[0].i;

   for (int i = 1; i < *temp_nbr_vertices; ++i) {

      if (compare_point_with_index(sort_array + i - 1, sort_array + i)) {

         old_to_new_id[sort_array[i].i] = 1;
         last_unique_idx = sort_array[i].i;

      } else {

         old_to_new_id[sort_array[i].i] = -last_unique_idx;
      }
   }

   free(sort_array);

   size_t new_nbr_vertices = 0;

   for (int i = 0; i < *temp_nbr_vertices; ++i) {

      if (old_to_new_id[i] == 1) {

         temp_vertex_lon[new_nbr_vertices] = temp_vertex_lon[i];
         temp_vertex_lat[new_nbr_vertices] = temp_vertex_lat[i];

         old_to_new_id[i] = new_nbr_vertices;

         new_nbr_vertices++;
      }
   }

   for (int i = 0; i < *temp_nbr_vertices; ++i)
      if (old_to_new_id[i] <= 0)
         old_to_new_id[i] = old_to_new_id[-old_to_new_id[i]];

   *temp_nbr_vertices = new_nbr_vertices;
}

void read_compA_grid_data(int * nbr_vertices, int * nbr_cells,
                          int ** num_vertices_per_cell, int ** cell_to_vertex,
                          double ** buffer_lonv, double ** buffer_latv,
                          double ** buffer_lon, double ** buffer_lat) {

  *nbr_cells = 20;
#ifdef EXACT
  int nbr_vertices_per_cell = 12;
#else
  int nbr_vertices_per_cell = 7;
#endif

  // Read in grid data

  FILE *inData;

#ifdef EXACT
  char * inputFile = "mesh_compA_exact.txt";
#else
  char * inputFile = "mesh_compA_convex.txt";
#endif

  char* inFormat0 = "%s %*s %*s";
  char* inFormat1 = "%*10c %lf";
#ifdef EXACT
  char* inFormat2 = "%*10c %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf";
#else
  char* inFormat2 = "%*10c %lf %lf %lf %lf %lf %lf %lf";
#endif

  if ((inData=fopen(inputFile, "r")) == NULL) {
      printf("Cannot open file.\n");
      exit(1);
  }

  char dummy_string[32];

  //skip first three lines

  for ( i = 0; i < 3; ++i )
    fscanf (inData, inFormat0, dummy_string);

  *nbr_vertices = nbr_vertices_per_cell * (*nbr_cells);

  *buffer_lon = malloc ( *nbr_cells * sizeof(*buffer_lon) );
  *buffer_lat = malloc ( *nbr_cells * sizeof(*buffer_lat) );
  *buffer_lonv = malloc ( (*nbr_vertices) * sizeof(*buffer_lonv) );
  *buffer_latv = malloc ( (*nbr_vertices) * sizeof(*buffer_latv) );

  for ( i = 0; i < *nbr_cells; ++i ) {
    fscanf ( inData, inFormat1, &((*buffer_lon)[i]) );
    printf ("point buffer_lat %i : %lf \n", i, (*buffer_lon)[i]);
  }

  for ( i = 0; i < *nbr_cells; ++i ) {
    fscanf ( inData, inFormat2,
             (*buffer_lonv) + i*nbr_vertices_per_cell +  0,
             (*buffer_lonv) + i*nbr_vertices_per_cell +  1,
             (*buffer_lonv) + i*nbr_vertices_per_cell +  2,
             (*buffer_lonv) + i*nbr_vertices_per_cell +  3,
             (*buffer_lonv) + i*nbr_vertices_per_cell +  4,
             (*buffer_lonv) + i*nbr_vertices_per_cell +  5,
             (*buffer_lonv) + i*nbr_vertices_per_cell +  6,
             (*buffer_lonv) + i*nbr_vertices_per_cell +  7,
             (*buffer_lonv) + i*nbr_vertices_per_cell +  8,
             (*buffer_lonv) + i*nbr_vertices_per_cell +  9,
             (*buffer_lonv) + i*nbr_vertices_per_cell + 10,
             (*buffer_lonv) + i*nbr_vertices_per_cell + 11);
  }

  for ( i = 0; i < *nbr_cells; ++i ) {
    fscanf ( inData, inFormat1, &((*buffer_lat)[i]) );
    printf ("point buffer_lat %i : %lf \n", i, (*buffer_lat)[i]);
  }

  for ( i = 0; i < *nbr_cells; ++i ) {
    fscanf ( inData, inFormat2,
             (*buffer_latv) + i*nbr_vertices_per_cell +  0,
             (*buffer_latv) + i*nbr_vertices_per_cell +  1,
             (*buffer_latv) + i*nbr_vertices_per_cell +  2,
             (*buffer_latv) + i*nbr_vertices_per_cell +  3,
             (*buffer_latv) + i*nbr_vertices_per_cell +  4,
             (*buffer_latv) + i*nbr_vertices_per_cell +  5,
             (*buffer_latv) + i*nbr_vertices_per_cell +  6,
             (*buffer_latv) + i*nbr_vertices_per_cell +  7,
             (*buffer_latv) + i*nbr_vertices_per_cell +  8,
             (*buffer_latv) + i*nbr_vertices_per_cell +  9,
             (*buffer_latv) + i*nbr_vertices_per_cell + 10,
             (*buffer_latv) + i*nbr_vertices_per_cell + 11);
  }

  for ( i = 0; i < *nbr_vertices; ++i )
    printf ("%i vertex buffer_lonv : %lf   buffer_latv : %lf \n", i,
            (*buffer_lonv)[i], (*buffer_latv)[i]);

  // Connectivity

  (*cell_to_vertex) = malloc ( (*nbr_cells) * nbr_vertices_per_cell *
                                sizeof(**cell_to_vertex) );
  (*num_vertices_per_cell) = malloc ( *nbr_cells * sizeof(**num_vertices_per_cell) );

  int * old_to_new_id = malloc((*nbr_vertices) * sizeof(*old_to_new_id));
  remove_duplicated_vertices(*buffer_lonv, *buffer_latv, nbr_vertices,
                             old_to_new_id);

  int count = 0;

  for (int i = 0; i < *nbr_cells; ++i) {

    (*cell_to_vertex)[count++] = old_to_new_id[i * nbr_vertices_per_cell];
    (*num_vertices_per_cell)[i] = 1;

    for (int j = 1; j < nbr_vertices_per_cell; ++j) {

      if (old_to_new_id[i * nbr_vertices_per_cell + j - 1] !=
          old_to_new_id[i * nbr_vertices_per_cell + j]) {

        (*cell_to_vertex)[count++] = old_to_new_id[i * nbr_vertices_per_cell + j];
        (*num_vertices_per_cell)[i]++;
      }
    }
  }

  fclose (inData);

  free(old_to_new_id);
}

void dummy_compA () {

  int size, rank;

  MPI_Comm local_comm;

  char * comp_name = "dummy_compA";
  char * grid_name = "dummy_compA_grid";

  // Initialise the coupler
  yac_cinit ( xml_filename, xsd_filename );

  // Inform the coupler about what we are
  yac_cdef_comp ( comp_name, &comp_id );
  comp_ids[0] = comp_id;

  yac_cget_localcomm ( &local_comm, comp_id );

  MPI_Comm_rank ( local_comm, &rank );
  MPI_Comm_size ( local_comm, &size );

  int lenstr = (int) strlen(comp_name);
  yac_redirstdout ( comp_name, lenstr, 1, rank, size, &ierror );

  // Announce one subdomain (patch) to the coupler
  yac_cdef_subdomain ( comp_id, grid_name, &subdomain_id );

  subdomain_ids[0] = subdomain_id;

  int nbr_vertices, nbr_cells, * num_vertices_per_cell, * cell_to_vertex;
  double * buffer_lonv, * buffer_latv, * buffer_lon, * buffer_lat;


  read_compA_grid_data(&nbr_vertices, &nbr_cells, &num_vertices_per_cell,
                       &cell_to_vertex, &buffer_lonv, &buffer_latv,
                       &buffer_lon, &buffer_lat);

  // Description of elements

  yac_cdef_elements_unstruct ( subdomain_id,
                               nbr_vertices,
                               nbr_cells,
                               num_vertices_per_cell,
                               buffer_lonv,
                               buffer_latv,
                               cell_to_vertex );

  // Center points in cells

  yac_cdef_points_unstruct ( subdomain_id,
                             nbr_cells,
                             CELL,
                             buffer_lon,
                             buffer_lat,
                             &cell_point_id );


  // Decomposition information

  global_index = malloc ( nbr_cells * sizeof(*global_index) ) ;
  owner_local  = malloc ( nbr_cells * sizeof(*owner_local) );

  for ( i = 0; i < nbr_cells; ++i ) {
    global_index[i] = i;
    owner_local[i] = -1;
  }

  yac_cdef_index_location ( subdomain_id,
                            nbr_cells,
                            CELL,
                            global_index,
                            owner_local );

  // Connect subdomains

  yac_cconnect_subdomains ( comp_id,
                            nbr_subdomain_ids,
                            subdomain_ids,
                            &domain_id );

  // Define fields

  field_id = malloc ( no_of_fields * sizeof(*field_id));

  for ( i = 0; i < no_of_fields; ++i )
    yac_cdef_field ( fieldName[i],
                     comp_id,
                     domain_id,
                     &cell_point_id,
                     NULL,
                     1,
                     &field_id[i] );

  // Invoke the search

  yac_csearch ( 1, comp_ids, no_of_fields, field_id, &ierror );

  // Data exchange

  recv_buffer = malloc (nbr_cells * sizeof (*recv_buffer) );
  send_buffer = malloc (nbr_cells * sizeof (*send_buffer) );

  for ( i = 0; i < nbr_cells; ++i ) recv_buffer[i] = 20.0;
  for ( i = 0; i < nbr_cells; ++i ) send_buffer[i] = (double) i;

  // Send fields to compB
  // --------------------

  {
    double *point_set_data[1];
    double **domain_data[1];
    double ***collection_data[1];

    point_set_data[0] = send_buffer;
    domain_data[0] = &(point_set_data[0]);
    collection_data[0] = &(domain_data[0]);

    // AtoB

    yac_cput ( field_id[0], 1, collection_data, &info, &ierror );

  }

/*
  //
  // Receive fields from compB
  // -------------------------

  {
    double *domain_data[1];
    double **collection_data[1];

    domain_data[0] = recv_buffer;
    collection_data[0] = &(domain_data[0]);

    // BtoA

    yac_cget ( field_id[1], 1, collection_data, &info, &ierror );

  }
*/

  //----------------------------------------------------------
  // write field to vtk output file
  //----------------------------------------------------------

  char vtk_filename[32];

  sprintf(vtk_filename, "compA_out_%d.vtk", rank);

  double point_data[nbr_vertices][3];
  for (int i = 0; i < nbr_vertices; ++i) {
   LLtoXYZ_deg(buffer_lonv[i], buffer_latv[i], point_data[i]);
  }

  VTK_FILE *vtk_file = vtk_open(vtk_filename, "compA_out");

  vtk_write_point_data(vtk_file, (double *)point_data, nbr_vertices);

  vtk_write_cell_data(vtk_file, (unsigned *)cell_to_vertex,
                      (unsigned*)num_vertices_per_cell, nbr_cells);

  vtk_write_cell_scalars_int(vtk_file, global_index, nbr_cells,
                             "global_cell_id");

  vtk_write_cell_scalars_double(vtk_file, recv_buffer, nbr_cells, "cell_in");

  vtk_close(vtk_file);

  /* Free memory and finalise */

  free (cell_to_vertex);
  free (num_vertices_per_cell);

  free ( buffer );

  free ( buffer_lon );
  free ( buffer_lat );

  free ( buffer_lonv );
  free ( buffer_latv );

  free( global_index );
  free( owner_local );

  free (recv_buffer);
  free (send_buffer);

  free (field_id);

  yac_cfinalize ();

}

/* -------------------------------------------------------------------- */

void dummy_compB () {

  int size, rank;

  MPI_Comm local_comm;

  char * comp_name = "dummy_compB";
  char * grid_name = "dummy_compB_grid";

  // Initialise the coupler
  yac_cinit ( xml_filename, xsd_filename );

  // Inform the coupler about what we are
  yac_cdef_comp ( comp_name, &comp_id );
  comp_ids[0] = comp_id;

  yac_cget_localcomm ( &local_comm, comp_id );

  MPI_Comm_rank ( local_comm, &rank );
  MPI_Comm_size ( local_comm, &size );

  int lenstr = (int) strlen(comp_name);
  yac_redirstdout ( comp_name, lenstr, 1, rank, size, &ierror );

  // Announce one subdomain (patch) to the coupler
  yac_cdef_subdomain ( comp_id, grid_name, &subdomain_id );

  subdomain_ids[0] = subdomain_id;

  /* Define compB grid */

  int nbr_vertices[2];
  int nbr_cells[2] = {200, 320};
  int cyclic[2] = {0,0};

  double dx = 0.01;

  nbr_vertices[0] = nbr_cells[0] + 1;
  nbr_vertices[1] = nbr_cells[1] + 1;

  double * buffer_lonv = malloc ( nbr_vertices[0] * sizeof(*buffer_lonv) );
  double * buffer_latv = malloc ( nbr_vertices[1] * sizeof(*buffer_latv) );

  // Description of reg2d elements

  buffer_lonv[0] = 265.6 - 0.5 * dx;
  buffer_latv[0] =  58.5 - 0.5 * dx;

  for ( i = 1; i < nbr_vertices[0]; ++i )
    buffer_lonv[i] = buffer_lonv[i-1] + dx;

  for ( i = 1; i < nbr_vertices[1]; ++i )
    buffer_latv[i] = buffer_latv[i-1] + dx;

  yac_cdef_elements_reg2d ( subdomain_id,
                            nbr_vertices,
                            cyclic,
                            buffer_lonv,
                            buffer_latv );

  // Center points in cells

  double * buffer_lon = malloc ( nbr_cells[0] * sizeof(*buffer_lon) );
  double * buffer_lat = malloc ( nbr_cells[1] * sizeof(*buffer_lat) );

  buffer_lon[0] = 265.6;
  buffer_lat[0] =  58.5;

  for ( i = 1; i < nbr_cells[0]; ++i )
    buffer_lon[i] = buffer_lonv[i-1] + dx;

  for ( i = 1; i < nbr_cells[1]; ++i )
    buffer_lat[i] = buffer_latv[i-1] + dx;

  yac_cdef_points_reg2d ( subdomain_id,
                          nbr_cells,
                          CELL,
                          buffer_lon,
                          buffer_lat,
                          &cell_point_id );

  // Decomposition information

  int total_nbr_cells =  nbr_cells[0] * nbr_cells[1];

  global_index = malloc ( total_nbr_cells * sizeof(*global_index) ) ;
  owner_local  = malloc ( total_nbr_cells * sizeof(*owner_local) );

  for ( j = 0; j < nbr_cells[1]; ++j ) {
     for ( i = 0; i < nbr_cells[0]; ++i ) {
       global_index[j*nbr_cells[0]+i] = j*nbr_cells[0]+i;
       owner_local[j*nbr_cells[0]+i] = -1;
     }
  }

  yac_cdef_index_location ( subdomain_id,
                            total_nbr_cells,
                            CELL,
                            global_index,
                            owner_local );

  // Connect subdomains

  yac_cconnect_subdomains ( comp_id,
                            nbr_subdomain_ids,
                            subdomain_ids,
                            &domain_id );

  // Define fields

  field_id = malloc ( no_of_fields * sizeof(*field_id));

  for ( i = 0; i < 2; ++i )
    yac_cdef_field ( fieldName[i],
                     comp_id,
                     domain_id,
                     &cell_point_id,
                     NULL,
                     1,
                     &field_id[i] );

  // Invoke the search

  yac_csearch ( 1, comp_ids, no_of_fields, field_id, &ierror );

 // Data exchange

  recv_buffer = malloc ( total_nbr_cells * sizeof (*recv_buffer) );
  send_buffer = malloc ( total_nbr_cells * sizeof (*send_buffer) );

  for ( j = 0; j < nbr_cells[1]; ++j ) {
     for ( i = 0; i < nbr_cells[0]; ++i ) {
    	 send_buffer[j*nbr_cells[0]+i] = (float) j;
    }
  }

  for ( i = 0; i < total_nbr_cells; ++i ) recv_buffer[i] = 20;

/*
  // Send fields from compB to compA
  // ------------------------------------

  {
    double *point_set_data[1];
    double **domain_data[1];
    double ***collection_data[1];

    point_set_data[0] = send_buffer;
    domain_data[0] = &(point_set_data[0]);
    collection_data[0] = &(domain_data[0]);

    // BtoA

    yac_cput ( field_id[1], 1, collection_data, &info, &ierror );

  }
*/

  // Receive fields from compA
  // ------------------------------

  {
    double *domain_data[1];
    double **collection_data[1];

    domain_data[0] = recv_buffer;
    collection_data[0] = &(domain_data[0]);

    // AtoB

    yac_cget ( field_id[0], 1, collection_data, &info, &ierror );

  }

  //----------------------------------------------------------
  // write field to vtk output file
  //----------------------------------------------------------

  double * point_data = malloc(nbr_vertices[0] * nbr_vertices[1] * 3 * sizeof(*point_data));
  for (int i = 0; i < nbr_vertices[1]; ++i) {
    for (int j = 0; j < nbr_vertices[0]; ++j) {
      LLtoXYZ_deg(buffer_lonv[j], buffer_latv[i], &point_data[3*(i * nbr_vertices[0] + j)]);
    }
  }

  unsigned * cell_corners = malloc(total_nbr_cells * 4 * sizeof(*cell_corners));
  unsigned * num_points_per_cell = malloc(total_nbr_cells * sizeof(*num_points_per_cell));

  for (int i = 0; i < total_nbr_cells; ++i) {

    unsigned x_index, y_index;

    y_index = i / nbr_cells[0];
    x_index = i - y_index * nbr_cells[0];

    cell_corners[i*4+0] =  y_index      * (nbr_cells[0] + 1) + x_index;
    cell_corners[i*4+1] =  y_index      * (nbr_cells[0] + 1) + x_index + 1;
    cell_corners[i*4+2] = (y_index + 1) * (nbr_cells[0] + 1) + x_index + 1;
    cell_corners[i*4+3] = (y_index + 1) * (nbr_cells[0] + 1) + x_index;

    num_points_per_cell[i] = 4;
  }

  char vtk_filename[32];

  sprintf(vtk_filename, "compB_out_%d.vtk", rank);

  VTK_FILE *vtk_file = vtk_open(vtk_filename, "compB_out");

  vtk_write_point_data(vtk_file, point_data, nbr_vertices[0]*nbr_vertices[1]);

  vtk_write_cell_data(vtk_file, cell_corners, num_points_per_cell, total_nbr_cells);

  vtk_write_cell_scalars_int(vtk_file, global_index, total_nbr_cells,
                             "global_cell_id");

  vtk_write_cell_scalars_double(vtk_file, recv_buffer, total_nbr_cells, "cell_in");

  vtk_close(vtk_file);

  /* Free memory */

  free( global_index );
  free( owner_local );

  free(num_points_per_cell);
  free(cell_corners);
  free(point_data);

  free ( buffer );

  free ( buffer_lon );
  free ( buffer_lat );

  free ( buffer_lonv );
  free ( buffer_latv );

  free (send_buffer);
  free (recv_buffer);
  free (field_id);

  yac_cfinalize ();

}

/* -------------------------------------------------------------------- */

#endif


