//#define VERBOSE
/**
 * @file toy_icon_ocn.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */
#undef TOY_SETTING

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if defined USE_MPI && defined HAVE_C_NETCDF
#include <mpi.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "config_xml.h"
#include "tests.h"
#include "utils.h"
#include "component.h"
#include "fields.h"
#include "yac_interface.h"
#include "geometry.h"
#include "vtk_output.h"
#include "read_icon_grid.h"

/* ------------------------------------------------- */

/* For simplicity we define the same 4 fields that are in the xml file */

#include "toy_common.h"

void read_part_icon_grid_information(const char * filename, int * num_vertices,
                                     int * num_cells, int ** num_vertices_per_cell,
                                     int ** cell_to_vertex, double ** x_vertices,
                                     double ** y_vertices, double ** x_cells,
                                     double ** y_cells, int ** global_cell_id,
                                     int ** cell_mask,
                                     int ** global_cell_id_rank, int ** global_corner_id,
                                     int ** global_corner_id_rank, int rank, int size);

int main (int argc, char *argv[]) {

#ifdef USE_XML2

  // Initialisation of MPI

  MPI_Init (0, NULL);

  /* The initialisation phase includes the reading of the XML file */
#ifdef VERBOSE
  printf (". main: calling yac_cinit\n");
#endif

  double tic, toc, time, time_min, time_max, time_ave;

  MPI_Barrier(MPI_COMM_WORLD);

  tic=MPI_Wtime();

  const char * configFilename = xml_filename;
  if ( argc > 2 ) configFilename = argv[2];
#ifdef VERBOSE
  printf ( "coupling configuration file: %s\n",configFilename);
#endif
  yac_cinit (configFilename, xsd_filename);

  /* The usual component definition, here for two sequential components on the same process */

#ifdef VERBOSE
  printf (". main: calling yac_cdef_comp\n");
#endif

  int comp_id;
  char * comp_name = "OCEAN";

  yac_cdef_comp ( comp_name, &comp_id );
#ifdef VERBOSE
  printf ( ". main: defined %s with local comp ID %i \n", "toy-icon-ocean", comp_id );
#endif

  MPI_Comm local_comm;

  yac_cget_localcomm(&local_comm, comp_id);

  int rank, size;

  MPI_Comm_rank(local_comm,&rank);
  MPI_Comm_size(local_comm,&size);

  int ierror = 0;

  int global_comp_id;
  int point_id;
  int mask_id;

  int field_ids[8];
  int domain_id;
  int subdomain_id;

  struct component * component;

  /* Grid definition for the first component (toy-icon-ocean) */

  // char * filename = "iconR2B03-grid.nc";
  char * filename = "iconR2B04-ocean_etopo40_planet_newgeom.nc";
  //char * filename = "iconR2B04-ocean_etopo40_planet_newgeom_noinland.nc";
  if ( argc > 1 ) filename = argv[1];
  printf ( "ocean input grid file: %s\n",filename );

  int num_vertices;
  int num_cells;
  int * num_vertices_per_cell;
  int * cell_to_vertex;
  double * x_vertices;
  double * y_vertices;
  double * x_cells;
  double * y_cells;

  int * cell_mask;
  int * cell_mask_own;
  int * global_cell_id;
  int * global_cell_id_rank;
  int * global_corner_id;
  int * global_corner_id_rank;

  read_part_icon_grid_information(filename, &num_vertices, &num_cells,
                                  &num_vertices_per_cell, &cell_to_vertex,
                                  &x_vertices, &y_vertices, &x_cells,
                                  &y_cells, &global_cell_id,
                                  &cell_mask,
                                  &global_cell_id_rank, &global_corner_id,
                                  &global_corner_id_rank, rank, size);

  double * x_points, * y_points;

  x_points = x_vertices;
  y_points = y_vertices;

  for (int i = 0; i < num_vertices; ++i) {
    x_points[i] *= deg;
    y_points[i] *= deg;
  }

  double * x_center, * y_center;

  x_center = x_cells;
  y_center = y_cells;

  for (int i = 0; i < num_cells; ++i) {
    x_center[i] *= deg;
    y_center[i] *= deg;
  }

  yac_cdef_subdomain ( comp_id, "grid1", &subdomain_id );

  yac_cdef_elements_unstruct ( subdomain_id, num_vertices, num_cells, num_vertices_per_cell,
                               x_vertices, y_vertices, cell_to_vertex );

  yac_cdef_points_unstruct ( subdomain_id, num_cells, CELL, x_center, y_center, &point_id );

  yac_cdef_index_location ( subdomain_id, num_cells, CELL, global_cell_id, global_cell_id_rank );

  yac_cconnect_subdomains ( comp_id, 1, &subdomain_id, &domain_id );

  cell_mask_own = malloc (num_cells * sizeof(*cell_mask_own));

  // negative value represent sea, positive values represent land
  // -+2 inner, -+1 boundary
  //
  // Here we mask out land points:
  //
  // ICON mask values of 0 are left untouched!
  // In case the whole mask is 0 we set all values to true!

  int checksum_mask = 0;

  for ( int i = 0; i < num_cells; ++i )
   checksum_mask += abs(cell_mask[i]);

  if ( checksum_mask > 0 ) {
    for ( int i = 0; i < num_cells; ++i ) {
      if ( cell_mask[i] < 0 ) cell_mask_own[i] = 0;
      else cell_mask_own[i] = 1;
    }
  }
  else {
     for ( int i = 0; i < num_cells; ++i )
        cell_mask_own[i] = 0;
  }

  // overwrite !!!!!
  // for ( int i = 0; i < num_cells; ++i ) cell_mask_own[i] = 0;

  yac_cdef_mask ( cell_mask_own, point_id, &mask_id );

  /* Field definition for the first component (toy-icon-ocean) */

  component = yac_unique_id_to_pointer(comp_id);
  global_comp_id = yac_component_get_global_id(component);
 
  int num_fields = yac_xml_get_number_of_transient_grid_refs_by_component_id(global_comp_id);
  
#ifdef VERBOSE
  printf (". main: number of fields in component %i is %i\n", global_comp_id, num_fields);
#endif

  yac_cdef_field ( fieldName[0], comp_id, domain_id, &point_id, &mask_id, 1, &field_ids[0] );
  yac_cdef_field ( fieldName[2], comp_id, domain_id, &point_id, &mask_id, 1, &field_ids[2] );

  if ( checksum_mask > 0 ) {
    for ( int i = 0; i < num_cells; ++i ) {
      if ( cell_mask[i] < 0 ) cell_mask_own[i] = 0;
      else cell_mask_own[i] = 1;
    }
  }
  else {
     for ( int i = 0; i < num_cells; ++i )
        cell_mask_own[i] = 0;
  }

  yac_cdef_mask ( cell_mask_own, point_id, &mask_id );

  yac_cdef_field ( fieldName[1], comp_id, domain_id, &point_id, &mask_id, 1, &field_ids[1] );
  yac_cdef_field ( fieldName[3], comp_id, domain_id, &point_id, &mask_id, 1, &field_ids[3] );

  free (cell_mask_own);

  toc=MPI_Wtime();
  time = toc-tic;

  MPI_Reduce ( &time, &time_min, 1, MPI_DOUBLE, MPI_MIN, 0, local_comm);
  MPI_Reduce ( &time, &time_max, 1, MPI_DOUBLE, MPI_MAX, 0, local_comm);
  MPI_Reduce ( &time, &time_ave, 1, MPI_DOUBLE, MPI_SUM, 0, local_comm);

  time_ave /= size;

  if ( rank == 0 )
    printf ("toy-icon-ocean:      Time for initialisation %f %f %f \n", time_min, time_max, time_ave );

  /* Search. */

#ifdef VERBOSE
  printf (". main: calling yac_csearch\n");
#endif

  MPI_Barrier(MPI_COMM_WORLD);

  tic=MPI_Wtime();

  yac_csearch ( 1, &comp_id, num_fields, field_ids, &ierror );

  toc=MPI_Wtime();
  time = toc-tic;

  MPI_Reduce ( &time, &time_min, 1, MPI_DOUBLE, MPI_MIN, 0, local_comm);
  MPI_Reduce ( &time, &time_max, 1, MPI_DOUBLE, MPI_MAX, 0, local_comm);
  MPI_Reduce ( &time, &time_ave, 1, MPI_DOUBLE, MPI_SUM, 0, local_comm);

  time_ave /= size;

  if ( rank == 0 )
    printf ("toy-icon-ocean:      Time for search         %f %f %f \n", time_min, time_max, time_ave );

#ifdef TOY_SETTING

  double * cell_out_conserv_data = malloc(num_cells * sizeof(*cell_out_conserv_data));
  double * cell_out_patch_data = malloc(num_cells * sizeof(*cell_out_patch_data));
  double * cell_in_conserv_data = malloc(num_cells * sizeof(*cell_in_conserv_data));
  double * cell_in_patch_data = malloc(num_cells * sizeof(*cell_in_patch_data));

  int err;
  int info;

  int cell_to_vertex_offset = 0;

  for (int i = 0; i < num_cells; ++i) {

    double middle_point[3] = {0, 0, 0};

    for (int j = 0; j < num_vertices_per_cell[i]; ++j) {

      double curr_point[3];

      LLtoXYZ_deg(x_vertices[cell_to_vertex[cell_to_vertex_offset + j]],
                  y_vertices[cell_to_vertex[cell_to_vertex_offset + j]],
                  curr_point);

      middle_point[0] += curr_point[0];
      middle_point[1] += curr_point[1];
      middle_point[2] += curr_point[2];
    }

    double scale = 1.0 / sqrt(middle_point[0] * middle_point[0] + 
                              middle_point[1] * middle_point[1] + 
                              middle_point[2] * middle_point[2]);

    middle_point[0] *= scale;
    middle_point[1] *= scale;
    middle_point[2] *= scale;

    struct point point;

    XYZtoLL(middle_point, &point.lon, &point.lat);

    cell_to_vertex_offset += num_vertices_per_cell[i];

    cell_in_conserv_data[i] = -10;
    cell_in_patch_data[i] = -10;

    cell_out_conserv_data[i] = sin(point.lon * 2) *
                               cos(point.lat * 3);
    cell_out_patch_data[i]   = sin(point.lon * 2) *
                               cos(point.lat * 3);
  }

  for ( int i = 0; i < num_cells; ++i ) {
    if ( cell_mask[i] > 0 ) {
      cell_out_conserv_data[i] = -99999;
      cell_out_patch_data[i]   = -99999;
      cell_in_conserv_data[i]  = -999;
      cell_in_patch_data[i]    = -999;
    }
  }

  double point_data[num_vertices][3];
  for (int i = 0; i < num_vertices; ++i) {
    LLtoXYZ_deg(x_vertices[i], y_vertices[i], point_data[i]);
  }

  for (int step = 0; step < num_steps; ++step) {

    MPI_Barrier(MPI_COMM_WORLD);

    tic=MPI_Wtime();

    {
      double *point_set_data[1];
      double **domain_data[1] = {point_set_data};
      double ***collection_data[1] = {domain_data};

      point_set_data[0] = cell_out_conserv_data;
      yac_cput(field_ids[2], 1, collection_data, &info, &err);
      if ( info == 0 ) yac_internal_abort_message ("check coupling period", __FILE__, __LINE__ );
      point_set_data[0] = cell_out_patch_data;
      yac_cput(field_ids[3], 1, collection_data, &info, &err);
      if ( info == 0 ) yac_internal_abort_message ("check coupling period", __FILE__, __LINE__ );
    }

    {
      double *domain_data[1];
      double **collection_data[1] = {domain_data};

      domain_data[0] = cell_in_conserv_data;
      yac_cget(field_ids[0], 1, collection_data, &info, &err);
      if ( info == 0 ) yac_internal_abort_message ("check coupling period", __FILE__, __LINE__ );
      domain_data[0] = cell_in_patch_data;
      yac_cget(field_ids[1], 1, collection_data, &info, &err);
      if ( info == 0 ) yac_internal_abort_message ("check coupling period", __FILE__, __LINE__ );
    }

    toc=MPI_Wtime();
    time = toc-tic;

    MPI_Reduce ( &time, &time_min, 1, MPI_DOUBLE, MPI_MIN, 0, local_comm);
    MPI_Reduce ( &time, &time_max, 1, MPI_DOUBLE, MPI_MAX, 0, local_comm);
    MPI_Reduce ( &time, &time_ave, 1, MPI_DOUBLE, MPI_SUM, 0, local_comm);

    time_ave /= size;

    if ( rank == 0 )
      printf ("toy-icon-ocean:      Time for ping-pong      %f %f %f \n", time_min, time_max, time_ave );

    //----------------------------------------------------------
    // write field to vtk output file
    //----------------------------------------------------------

    char vtk_filename[32];

    sprintf(vtk_filename, "toy-icon-ocean_out_%d_%d.vtk", rank, step);

    VTK_FILE *vtk_file = vtk_open(vtk_filename, "toy-icon-ocean_out");

    // mask out land points

    vtk_write_point_data(vtk_file, (double *)point_data, num_vertices);
    vtk_write_cell_data(vtk_file, (unsigned *)cell_to_vertex,
                        (unsigned*)num_vertices_per_cell, num_cells);
    vtk_write_point_scalars_int(vtk_file, global_corner_id_rank, num_vertices,
                                "vertex_is_on_rank");
    vtk_write_point_scalars_int(vtk_file, global_corner_id, num_vertices,
                                "global_corner_id");
    vtk_write_cell_scalars_int(vtk_file, global_cell_id_rank, num_cells,
                               "cell_is_on_rank");
    vtk_write_cell_scalars_int(vtk_file, global_cell_id, num_cells,
                               "global_cell_id");

    vtk_write_cell_scalars_int(vtk_file, cell_mask, num_cells, "mask");
    vtk_write_cell_scalars_double(vtk_file, cell_in_conserv_data, num_cells, "cell_in_conserv");
    vtk_write_cell_scalars_double(vtk_file, cell_in_patch_data, num_cells, "cell_in_patch");
    vtk_write_cell_scalars_double(vtk_file, cell_out_conserv_data, num_cells, "cell_out_conserv");
    vtk_write_cell_scalars_double(vtk_file, cell_out_patch_data, num_cells, "cell_out_patch");
    vtk_close(vtk_file);

    for ( int i = 0; i < num_cells; ++i ) {
      if ( cell_mask[i] > 0 ) {
        cell_out_conserv_data[i] = -99999;
        cell_out_patch_data[i]   = -99999;
        cell_in_conserv_data[i]  = -999;
        cell_in_patch_data[i]    = -999;
      } else {
        cell_out_conserv_data[i] = cell_in_conserv_data[i];
        cell_out_patch_data[i]   = cell_in_patch_data[i];
        cell_in_conserv_data[i] = -10;
        cell_in_patch_data[i] = -10;
      }
    }
  }

  free(cell_in_conserv_data);
  free(cell_in_patch_data);
  free(cell_out_conserv_data);
  free(cell_out_patch_data);

#endif

  yac_cfinalize();

  MPI_Finalize();

  return EXIT_SUCCESS;

#else

  fprintf(stderr, "ERROR: toy-icon-ocean requires XML2\n");

  return EXIT_FAILURE;

#endif // USE_XML2
}

void read_part_icon_grid_information(const char * filename, int * num_vertices,
                                     int * num_cells, int ** num_vertices_per_cell,
                                     int ** cell_to_vertex, double ** x_vertices,
                                     double ** y_vertices, double ** x_cells,
                                     double ** y_cells, int ** global_cell_id,
                                     int ** cell_mask,
                                     int ** global_cell_id_rank, int ** global_corner_id,
                                     int ** global_corner_id_rank, int rank, int size) {

  // read the global grid
  read_icon_grid_information(filename, num_vertices, num_cells,
                             num_vertices_per_cell, cell_to_vertex,
                             x_vertices, y_vertices, x_cells, y_cells, cell_mask);

  // determine local rank
  int num_cells_per_process = (*num_cells + size - 1)/size;
  int local_start = rank * num_cells_per_process;
  int num_local_cells = MIN(num_cells_per_process,
                            *num_cells - local_start);

  // mask for required vertices and cells
  int * required_vertices = calloc(*num_vertices, sizeof(*required_vertices));
  int * required_cells = calloc(*num_cells, sizeof(*required_cells));

  int offset = 0;
  for (int i = 0; i < local_start; ++i)
    offset += (*num_vertices_per_cell)[i];

  // mark all local cells and their vertices as required
  for (int i = local_start; i < local_start + num_local_cells; ++i) {

    required_cells[i] = 1;
    for (int j = 0; j < (*num_vertices_per_cell)[i]; ++j)
      required_vertices[(*cell_to_vertex)[offset+j]] = 1;

    offset += (*num_vertices_per_cell)[i];
  }

  // mask for halo cells
  int * cell_is_on_rank = malloc(*num_cells * sizeof(*cell_is_on_rank));

  // mark all halo cells as required and generate cell_is_on_rank
  offset = 0;
  for (int i = 0; i < *num_cells; ++i) {

    if (!required_cells[i]) {

      for (int j = 0; j < (*num_vertices_per_cell)[i]; ++j) {

        if (required_vertices[(*cell_to_vertex)[offset+j]]) {

          required_cells[i] = 1;
          cell_is_on_rank[i] = i / num_cells_per_process;
          break;
        }
      }

    } else {
      cell_is_on_rank[i] = -1;
    }

    offset += (*num_vertices_per_cell)[i];
  }

  // mask for halo vertices
  int * vertex_is_on_rank = malloc(*num_vertices * sizeof(*vertex_is_on_rank));

  // mark all halo vertices as required and generate vertex_is_on_rank
  offset = 0;
  for (int i = 0; i < *num_vertices; ++i)
    if (required_vertices[i])
      vertex_is_on_rank[i] = -1;
  for (int i = 0; i < *num_cells; ++i) {

    if (required_cells[i] && cell_is_on_rank[i] != -1) {

      for (int j = 0; j < (*num_vertices_per_cell)[i]; ++j) {

        if (!required_vertices[(*cell_to_vertex)[offset+j]]) {

          required_vertices[(*cell_to_vertex)[offset+j]] = 1;
          vertex_is_on_rank[(*cell_to_vertex)[offset+j]] = cell_is_on_rank[i];
        }
      }
    }

    offset += (*num_vertices_per_cell)[i];
  }

  // count the number of cells and vertices
  int part_num_vertices = 0;
  int part_num_cells = 0;
  for (int i = 0; i < *num_vertices; ++i)
    if (required_vertices[i])
      part_num_vertices++;
  for (int i = 0; i < *num_cells; ++i)
    if(required_cells[i])
      part_num_cells++;

  *global_cell_id = malloc(part_num_cells * sizeof(**global_cell_id));
  *global_cell_id_rank = malloc(part_num_cells * sizeof(**global_cell_id_rank));
  *global_corner_id = malloc(part_num_vertices * sizeof(**global_corner_id));
  *global_corner_id_rank = malloc(part_num_vertices * sizeof(**global_corner_id_rank));

  // generate final vertex data
  part_num_vertices = 0;
  int * global_to_local_vertex = malloc(*num_vertices * sizeof(*global_to_local_vertex));
  for (int i = 0; i < *num_vertices; ++i) {

    if (required_vertices[i]) {

      (*global_corner_id)[part_num_vertices] = i;
      (*global_corner_id_rank)[part_num_vertices] = vertex_is_on_rank[i];
      (*x_vertices)[part_num_vertices] = (*x_vertices)[i];
      (*y_vertices)[part_num_vertices] = (*y_vertices)[i];
      global_to_local_vertex[i] = part_num_vertices;
      part_num_vertices++;
    }
  }

  *x_vertices = realloc(*x_vertices, part_num_vertices * sizeof(**x_vertices));
  *y_vertices = realloc(*y_vertices, part_num_vertices * sizeof(**y_vertices));
  free(vertex_is_on_rank);
  *num_vertices = part_num_vertices;
  free(required_vertices);

  // generate final cell data
  int num_cell_vertex_dependencies = 0;
  part_num_cells = 0;
  offset = 0;
  for (int i = 0; i < *num_cells; ++i) {

    if (required_cells[i]) {

      (*global_cell_id)[part_num_cells] = i;
      (*global_cell_id_rank)[part_num_cells] = cell_is_on_rank[i];
      (*x_cells)[part_num_cells] = (*x_cells)[i];
      (*y_cells)[part_num_cells] = (*y_cells)[i];
      (*cell_mask)[part_num_cells] = (*cell_mask)[i];

      for (int j = 0; j < (*num_vertices_per_cell)[i]; ++j)
        (*cell_to_vertex)[num_cell_vertex_dependencies++] =
          global_to_local_vertex[(*cell_to_vertex)[offset+j]];

      (*num_vertices_per_cell)[part_num_cells] = (*num_vertices_per_cell)[i];

      part_num_cells++;
    }

    offset += (*num_vertices_per_cell)[i];
  }

  *x_cells   = realloc(*x_cells, part_num_cells * sizeof(**x_cells));
  *y_cells   = realloc(*y_cells, part_num_cells * sizeof(**y_cells));
  *cell_mask = realloc(*cell_mask, part_num_cells * sizeof(**cell_mask));

  *num_vertices_per_cell = realloc(*num_vertices_per_cell, part_num_cells *
                                   sizeof(**num_vertices_per_cell));
  *cell_to_vertex = realloc(*cell_to_vertex, num_cell_vertex_dependencies *
                            sizeof(**cell_to_vertex));
  free(cell_is_on_rank);
  *num_cells = part_num_cells;
  free(required_cells);
  free(global_to_local_vertex);
}

#else
#include <stdlib.h>
#include <stdio.h>
int main () {
  printf ("Examples requires compiling with MPI and NetCDF.\n");
  return EXIT_FAILURE;
}
#endif
