/**
 * 
 */
package de.mpg.mpimet.yacgui.controller;

import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;

import de.mpg.mpimet.yacgui.model.Connection;
import de.mpg.mpimet.yacgui.model.Model;
import de.mpg.mpimet.yacgui.model.Model.Align;
import de.mpg.mpimet.yacgui.ui.ComponentPanel;
import de.mpg.mpimet.yacgui.ui.ConnectionsGlassPane;
import de.mpg.mpimet.yacgui.ui.CouplingGui;
import de.mpg.mpimet.yacgui.ui.SettingsPanel;
import de.mpg.mpimet.yacgui.ui.TransientPanel;
import de.mpg.mpimet.yacgui.ui.XmlFileChooser;
import de.mpg.mpimet.yacgui.ui.dialog.CouplingDialog;
import de.mpg.mpimet.yacgui.ui.dialog.SelectComponentsDialog;
import de.mpg.mpimet.yacgui.ui.dialog.SelectComponentsDialog.SelectComponentsDialogActionsListener;
import de.mpg.mpimet.yacgui.utils.Logger;
import de.mpg.mpimet.yacgui.utils.Utils;
import de.mpg.mpimet.yacgui.utils.XmlDomHandler;
import de.mpg.mpimet.yacgui.utils.XmlDomHandler.XmlType;
import de.mpg.mpimet.yacgui.utils.XmlDomHandlerException;
import de.mpg.mpimet.yacgui.xml.Component;
import de.mpg.mpimet.yacgui.xml.Couple;
import de.mpg.mpimet.yacgui.xml.Coupling;
import de.mpg.mpimet.yacgui.xml.TimestepUnit;
import de.mpg.mpimet.yacgui.xml.TransientCouple;

/**
 * Main controller class, includes most of the logic to build a couple from components
 * 
 * @author Maxim Yastremsky, Teresa Holfeld
 */
public class GuiController implements SelectComponentsDialogActionsListener {

	/** Singleton instance of GuiController */
	private static GuiController instance;
	protected Model model;
	protected GuiState state;
	protected CouplingGui gui;
	protected ConnectionsGlassPane glassPane;
	protected File saveDestination;
	protected FileStatus fileStatus;
	
	/**
	 * List of selected components. Those which are selected by mouse clicks and pressed ctrl.
	 */
	private List<Connection> selectedConnections;

	private SelectComponentsDialog selectComponentDialog;

	/**
	 * Represents the state the GUI is in.<br>
	 * The GUI will only have its full functionality if it is fully loaded (min 2 components).
	 *
	 */
	protected enum GuiState {
		EMPTY,
		LEFT_LOADED,
		RIGHT_LOADED,
		FULLY_LOADED;
	}

	protected GuiController() {
		model = new Model();
		state = GuiState.EMPTY;
		selectedConnections = new ArrayList<Connection>();
	}

	/**
	 * @return singleton instance of GuiController
	 */
	public static GuiController getInstance() {
		if (instance == null) {
			instance = new GuiController();
		}
		return instance;
	}

	/**
	 * Registers CouplingGui if not already present.
	 * 
	 * @param gui
	 *            CouplingGui to register
	 */
	public void registerCouplingGui(CouplingGui gui) {
		if (this.gui == null) {
			this.gui = gui;
			glassPane = gui.getGlassPane();
		}
	}

	/**
	 * Loads a coupling XML file into the GUI.
	 * 
	 * <p>
	 * If couplingFile is null, a JFileChooser dialog will guide the user to
	 * select a coupling XML file.
	 * 
	 * @param couplingFile
	 *            coupling XML file, can be null
	 */
	public void openCoupling(File couplingFile) {

		if (state == GuiState.FULLY_LOADED) {
			if (!fileStatus.isSaved()) {
				int option = gui.showSaveCoplingBeforeChangeDialog();
				if (option == JOptionPane.CANCEL_OPTION || option == JOptionPane.CLOSED_OPTION) {
					return;
				} else if (option == JOptionPane.YES_OPTION) {
					saveCoupling(false);
				}
			} else {
				int option = gui.showConfirmLoadingNewCouplingDialog();
				if (option == JOptionPane.NO_OPTION || option == JOptionPane.CLOSED_OPTION) {
					return;
				}
			}
			reset();
		}

		if (couplingFile == null) {
			couplingFile = XmlFileChooser.chooseFileOpen(gui.getContentPanel());
		}
		// if file choosing dialog has been canceled by the user
		if (couplingFile == null) {
			return;
		}

		Coupling coupling = null;
		XmlDomHandler domHandler = null;
		try {
			domHandler = new XmlDomHandler(couplingFile.getPath(), XmlType.COUPLING);
			coupling = domHandler.buildCoupling();
		} catch (XmlDomHandlerException e) { //bad xml: not expected root element
			gui.showSelectProperXmlFileDialog(e.getMessage());
			return;
		} catch (Exception e) {
			gui.showCouldntParseXmlFileDialog(domHandler.getErrorHandlerMessages());
			return;
		}

		// parsing worked, now load Coupling into GUI
		model.loadCoupling(coupling);
		
		updateGuiWithCoupling();
		updateGlassPaneWithConnections();
		
		saveFileInfo(couplingFile);
		Logger.logInfo("Couple has been loaded into GUI.");

		goToStateFullyLoaded();
		
		updateSelectComponentDialogIfExist();
		
		//in case of some xsd validation messages available - showing them
		String xsdValidationDetails = domHandler.getErrorHandlerMessages();
		if (xsdValidationDetails != null && !xsdValidationDetails.isEmpty()) {
			gui.showXsdValidationDetailsDialog(xsdValidationDetails);
		}
	}

//	public boolean isfileStatusSaved(){
//		return fileStatus.isSaved();
//	}
	
	private void updateSelectComponentDialogIfExist() {
		if (selectComponentDialog != null) {
			selectComponentDialog.updateInfoFromModel();
		}
	}

	/**
	 * Lets the user choose a component XML file and loads it into the GUI at
	 * the side given by align.
	 * 
	 * <p>
	 * If componentFile is null, JFileChooser dialog will guide user to select a
	 * file.
	 * 
	 * @param align
	 *            the side that should be loaded
	 * @param componentFile
	 *            component XML file, can be null
	 */
	public void openComponent(Align align, File componentFile) {
		// we load new component with specific align, check first overwriting
		if (align != null) {
			if (state == GuiState.FULLY_LOADED && model.isComponentCoupled(align)) {
				int option = gui.showOverwriteCoupledComponentWarningDialog(false);
				if (option == JOptionPane.NO_OPTION || option == JOptionPane.CLOSED_OPTION) {
					return;
				}
			}
		}
		
		//get the file(s) to load
		File[] files = null;
		if (componentFile != null) {
			files = new File[] {componentFile};
		} else {
			//if we load components via menu Couple -> Add components, then align = null, so we allow
			//to make multiple components loading
			boolean multipleSelection = align == null;
			
			files = XmlFileChooser.chooseFileOpen(gui.getContentPanel(), multipleSelection);
			
			if (files == null) { //user cancel the dialog
				return;
			}
		}
		
		for (File file : files) {
			if (file == null) {
				continue;
			}
			
			Component component = null;
			XmlDomHandler domHandler = null;
			try {
				domHandler = new XmlDomHandler(file.getPath(), XmlType.COMPONENT);
				component = domHandler.buildComponent();
			} catch (XmlDomHandlerException e) {
				gui.showSelectProperXmlFileDialog(e.getMessage());
				continue;
			} catch (Exception e) {
				gui.showCouldntParseXmlFileDialog(domHandler.getErrorHandlerMessages());
				continue;
			}
			
			if (checkIdenticalComponents(component)) {
				continue;
			}
			
			//no specific align: fill it from left to right
			if (align == null) { 
				if (state == GuiState.EMPTY || state == GuiState.RIGHT_LOADED) {
					align = Align.LEFT;
				} else if (state == GuiState.LEFT_LOADED) {
					align = Align.RIGHT;
				}
			}
			
			model.addComponent(component, align);
			
			if (align != null) {
				updateGuiWithComponent(align);
	
				switch (state) {
				case EMPTY:
					if (align.isLeft()) { // EMPTY -> LEFT_LOADED
						goToStateLeftLoaded();
					} else {	// EMPTY -> RIGHT_LOADED
						goToStateRightLoaded();
					}
					break;
				case LEFT_LOADED:
					if (align.isRight()) { // LEFT_LOADED -> FULLY_LOADED
						goToStateFullyLoaded();
						fileStatus.edit();
					}
					break;
				case RIGHT_LOADED:
					if (align.isLeft()) {	// RIGHT_LOADED -> FULLY_LOADED
						goToStateFullyLoaded();
						fileStatus.edit();
					}
					break;
				case FULLY_LOADED:
					// FULLY_LOADED -> FULLY_LOADED: it means we updated one of the components with another one
					
					goToStateFullyLoaded(); //re-enable fully_loaded state
					fileStatus.edit();
					
					model.getConnections().clear();
					gui.resetGlassPane();
					
					uncheckUnconnectedTransientCheckBoxes(align.getOpposite());
					
					resetLastSelectedTransientPanelIndices();
					enableAllCheckBoxes();
					break;
				default:
					break;
				}
			}
			
			//reset align in case of multiple component loading
			align = null;
		}
		
		updateSelectComponentDialogIfExist();
	}
	
	/**
	 * Checks if the selected component's name is already available in already loaded set
	 * 
	 * @param component to be loaded
	 * 
	 * @return true, if identical component found. False otherwise
	 */
	private boolean checkIdenticalComponents(Component component) {
		if (model.getCoupling().getComponentsMap().size() > 0) {
			Map<Integer, Component> componentsMap = model.getCoupling().getComponentsMap();
			
			for (Map.Entry<Integer, Component> entry : componentsMap.entrySet()) {
				
				if (entry.getValue().getName().equals(component.getName())) {
					gui.showIdenticalComponentWarningDialog(component.getName());
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * Removes a component in specific aligh from model
	 * 
	 * @param align where component should be removed
	 */
	public void removeComponent(Align align) {
		
		if (state == GuiState.FULLY_LOADED && model.isComponentCoupled(align)) {
			int option = gui.showOverwriteCoupledComponentWarningDialog(true);
			if (option == JOptionPane.NO_OPTION || option == JOptionPane.CLOSED_OPTION) {
				return;
			}
		}
		
		model.addComponent(null, align);
		updateGuiWithComponent(align);

		switch (state) {
		case LEFT_LOADED:
			if (align.isLeft()) { // LEFT_LOADED -> EMPTY
				goToStateEmpty();
			}
			break;
		case RIGHT_LOADED:
			if (align.isRight()) {	// RIGHT_LOADED -> EMPTY
				goToStateEmpty();
			}
			break;
		case FULLY_LOADED:
		
			Map<Integer, Component> components = model.getCoupling().getComponentsMap(); 
			
			//if we still have >=2 components to show 
			if (components.size() >= 2) {
				int existingComponentId = model.getComponentId(align.getOpposite());
				
				int secondComponentId = 0;
				
				//first try to find the couple with otherComponentId
				for (String strId : model.getCoupling().getCouples().keySet()) {
					if (strId.contains(String.valueOf(existingComponentId))) {
						secondComponentId = Utils.parseSecondComponentId(strId, existingComponentId);
						break;
					}
				}
				
				//if no couples found, take any other component
				if (secondComponentId == 0) {
					
					for (Integer intId : components.keySet()) {
						if (intId != existingComponentId) {
							secondComponentId = intId;
							break;
						}
					}
				}
				
				//if found any corresponding second component, load them to UI
				if (secondComponentId != 0) {
					if (align.isLeft()) {
						onComponentsSelected(secondComponentId, existingComponentId);
					} else {
						onComponentsSelected(existingComponentId, secondComponentId);
					}
				}
			} else { //only 1 component is left
				if (align.isLeft()) { // FULLY -> RIGHT_LOADED
					goToStateRightLoaded();
				} else {	// FULLY -> LEFT_LOADED
					goToStateLeftLoaded();
				}
				
				model.getConnections().clear();
				gui.resetGlassPane();
				
				uncheckUnconnectedTransientCheckBoxes(align.getOpposite());
				
				resetLastSelectedTransientPanelIndices();
			}
			break;
		default:
			break;
		}
		updateSelectComponentDialogIfExist();
	}
	

	/**
	 * Saves coupling to known file location.<br>
	 * If no file location is set, "save as" dialog will be opened.
	 */
	public void saveCoupling(boolean saveAs) {
		if (saveAs || saveDestination == null ) {
			File destination = XmlFileChooser.chooseFileSave(gui.getContentPanel());
			if (destination == null) {
				return;
			}
			saveFileInfo(destination);
			fileStatus = new FileStatus(gui);
		}
		
		model.getCoupling().setDates(gui.getSettingsPanel().getDates());
		model.getCoupling().setStdoutRedirect(gui.getSettingsPanel().isStdoutRedirect());
		model.getCoupling().setRootRedirect(gui.getSettingsPanel().isRootRedirect());

		XmlDomHandler.writeCouplingToFile(saveDestination, model.getCoupling());
		
		fileStatus.save();
	}
	
	public void setTimestepUnit(TimestepUnit unit) {
		model.getCoupling().setTimestepUnit(unit);
	}

	public boolean isStateFullyLoaded() {
		return (state == GuiState.FULLY_LOADED);
	}

	/**
	 * Callback to be called when the user checks any of checkboxes of transient panels.
	 * Here is implemented the logic of setting up a new connection and keeping track 
	 * already marked checkboxes
	 * 
	 * @param align - left of right
	 * @param index - the index of transient panel
	 */
	public void onTransientCheckBoxCheck(Align align, int index) {
		model.setLastSelectedTransientPanelIndex(align, index);

		if (!model.hasLastSelectedTransient(align.getOpposite())) { // if source (first from two) checkbox

			//disable all other at the same side
			disableAllCheckBoxesBut(align, index);

			// disable all but the ones with the same transient standard name and the same col. size
			String transientStandardName = model.getTransientStandardName(align, index);
			int collectionSize = model.getCollectionSize(align, index);
			disableAllCheckBoxesBut(align.getOpposite(), transientStandardName, collectionSize);
		
		} else { // if target (the second) checkbox for a couple
			fileStatus.edit();
			
			Align sourceAlign = align.getOpposite(); // from there to here
			int sourceIndex = model.getLastSelectedTransientPanelIndex(sourceAlign);
			Connection connection = model.addConnection(sourceAlign, sourceIndex, index);
			
			glassPane.addConnectionArrow(connection);
			pairTransientPanels(connection);
			
			resetLastSelectedTransientPanelIndices();
			enableAllCheckBoxes();
			
			updateSelectComponentDialogIfExist();
		}
	}

	/**
	 * Pairs a transient panel with another one if they were coupled.
	 * 
	 * @param c - a connection to specify two transient panels
	 */
	private void pairTransientPanels(Connection c) {
		TransientPanel panel1 = gui.getComponentPanel(c.getSourceAlign()).getTransientPanel(c.getSourceIndex()); 
		TransientPanel panel2 = gui.getComponentPanel(c.getTargetAlign()).getTransientPanel(c.getTargetIndex());
		
		panel1.setPairedWith(panel2);
		panel2.setPairedWith(panel1);
	}
	
	/**
	 * Unpairs a pair of transient panels, {@link GuiController#pairTransientPanels(Connection)}
	 * 
	 * @param align
	 * @param index
	 */
	private void unpairTransientPanels(Align align, int index) {
		TransientPanel panel1 = gui.getComponentPanel(align).getTransientPanel(index); 
		TransientPanel panel2 = panel1.getPairedPanel();
		
		panel1.setPairedWith(null);
		if (panel2 != null) {
			panel2.setPairedWith(null);
		}
	}

	/**
	 * Callback to be called when the user unchecks any of checkboxes of transient panels.
	 * Here is implemented the logic of removing the connection if it was there and keeping
	 * track already marked checkboxes
	 * 
	 * @param align - left or right
	 * @param index - the index of transient panel
	 */
	public void onTransientCheckBoxUncheck(Align align, int index) {
		if (model.isConnectedAt(align, index)) {
			fileStatus.edit();

			Connection connection = model.getConnectionAt(align, index);
			
			if (selectedConnections.contains(connection)) { 
				selectedConnections.remove(connection);
			}
			
			model.removeConnection(align, index);
			
			glassPane.removeArrowForTransient(align, index);
			unpairTransientPanels(align, index);

			uncheckUnconnectedTransientCheckBoxes(align);
			uncheckUnconnectedTransientCheckBoxes(align.getOpposite());
			
			updateSelectComponentDialogIfExist();
		}
		
		resetLastSelectedTransientPanelIndices();
		enableAllCheckBoxes();
	}

	/**
	 * Callback to be called when the user made a double click on transient panel.
	 * If clicked transient panel is connected shows the coupling dialog for selected transient panels
	 * 
	 * @param transPanel
	 */
	public void onTransientPanelDoubleClick(TransientPanel transPanel) {
		if (state != GuiState.FULLY_LOADED) {
			return;
		}
		
		Align align = transPanel.getAlign();
		int index = transPanel.getIndex();
		
		if (model.isConnectedAt(align, index)) {
			Connection connection = model.getConnectionAt(align, index);
			
			//if the user double-clicked any connection not selected already, 
			// then clear the list and add only this new connection.
			if (!selectedConnections.contains(connection)) { 
				
				clearSelectedTransients();

				selectedConnections.add(connection);
				setTransientsSelectedState(true, align, index);
			}

			showCouplingDialog(connection);
		}
	}
	
	private void showCouplingDialog(Connection clickedConnection) {
		
		StringBuilder dialogCaption = new StringBuilder();
		StringBuilder dialogDescription = new StringBuilder();
		
		for (Connection con : selectedConnections) {
			if (dialogCaption.length() != 0) {
				dialogCaption.append(", ");
				dialogDescription.append("\n");
			}
			
			String transientName = model.getTransientStandardName(con.getSourceAlign(), con.getSourceIndex()); 
			dialogCaption.append(transientName);

			dialogDescription.append(transientName).append(" (");
			dialogDescription.append(model.getGridName(con.getSourceAlign(), con.getSourceIndex()));
			dialogDescription.append(" -> ");
			dialogDescription.append(model.getGridName(con.getTargetAlign(), con.getTargetIndex()));
			dialogDescription.append(")");
		}
		
		dialogCaption.insert(0, "Coupling for ");
		dialogDescription.insert(0, "Coupling parameters for:\n");
		
		TransientCouple transientCouple = new TransientCouple();
		transientCouple.loadTransientCoupleData(model.getTransientCouple(clickedConnection));
		
		CouplingDialog dialog = new CouplingDialog(dialogCaption.toString(), dialogDescription.toString(),
				transientCouple, model.getTimestepUnit());
		dialog.setLocationRelativeTo(gui);
		dialog.setVisible(true);
	}

	private void clearSelectedTransients() {
		
		if (selectedConnections.isEmpty()) {
			return;
		}
		
		for (Connection c : selectedConnections) {
			setTransientsSelectedState(false, c.getSourceAlign(), c.getSourceIndex());
		}
		selectedConnections.clear();
	}

	/**
	 * Makes the selected transient panel selected
	 * 
	 * @param transPanel - transient panel that was clicked
	 * @param controlDown - whether the control button was pressed
	 */
	public void onTransientPanelSingleClick(TransientPanel transPanel, boolean controlDown) {
		
		if (state != GuiState.FULLY_LOADED) {
			return;
		}
		
		Align align = transPanel.getAlign();
		int index = transPanel.getIndex();
		
		if (!controlDown) {
			clearSelectedTransients();
		}
		
		if (model.isConnectedAt(align, index)) {
			Connection connection = model.getConnectionAt(align, index);
			
			selectedConnections.add(connection);
			setTransientsSelectedState(true, align, index);
		}
	}
	
	/**
	 * Updates a state of TransientPanel
	 * 
	 * @param selected
	 * @param align
	 * @param index
	 */
	private void setTransientsSelectedState(boolean selected, Align align, int index) {
		TransientPanel panel1 = gui.getComponentPanel(align).getTransientPanel(index); 
		TransientPanel panel2 = panel1.getPairedPanel();
		
		panel1.setSelectedState(selected);
		if (panel2 != null) {
			panel2.setSelectedState(selected);
		}
	}

	/**
	 * Method is called by CouplingDialog inner logic to save the Transient Couple info specified
	 * by user in the dialog.
	 * 
	 * @param transientCouple
	 */
	public void setTransientCoupleData(TransientCouple transientCouple) {
		fileStatus.edit();
		
		for (Connection con : selectedConnections) {
			con.setHavingInterpolations(transientCouple.getInterpolations() != null && !transientCouple.getInterpolations().isEmpty());
			model.updateTransientCoupleParameters(con, transientCouple);
			glassPane.updateConnectionArrow(con);
		}
	}
	
	/**
	 * Resets the model and gui to empty state
	 */
	private void reset() {
		model.resetData();
		gui.resetComponentPanels();
		gui.resetGlassPane();
		
		updateSelectComponentDialogIfExist();
		
		saveFileInfo(null);
		goToStateEmpty();
	}

	private void goToStateEmpty() {
		fileStatus = null;
		gui.setMenuFileSaveAsEnabled(false);
		state = GuiState.EMPTY;
	}

	private void goToStateLeftLoaded() {
		fileStatus = null;
		gui.setMenuFileSaveAsEnabled(false);
		gui.getComponentPanel(Align.LEFT).setAllCheckBoxesEnabled(false);
		state = GuiState.LEFT_LOADED;
	}

	private void goToStateRightLoaded() {
		fileStatus = null;
		gui.setMenuFileSaveAsEnabled(false);
		gui.getComponentPanel(Align.RIGHT).setAllCheckBoxesEnabled(false);
		state = GuiState.RIGHT_LOADED;
	}

	private void goToStateFullyLoaded() {
		fileStatus = new FileStatus(gui);
		gui.setMenuFileSaveAsEnabled(true);
		gui.getComponentPanel(Align.LEFT).setAllCheckBoxesEnabled(true);
		gui.getComponentPanel(Align.RIGHT).setAllCheckBoxesEnabled(true);
		state = GuiState.FULLY_LOADED;
	}
	
	private void saveFileInfo(File file) {
		saveDestination = file;
		
		if (file != null) {
			gui.setHint(file.getName());
		} else {
			gui.resetHint();
		}
	}

	private void updateGuiWithComponent(Align align) {
		if (align == null) {
			return;
		}
		
		ComponentPanel componentPanel = gui.getComponentPanel(align);
		componentPanel.setComponentPanelTitle(model.getComponentName(align));
		componentPanel.loadTransientPanelsData(model.getTransientPanelsData(align));
		componentPanel.repaint();
	}

	private void updateGuiWithCoupling() {
		
		updateGuiWithComponent(Align.LEFT);
		updateGuiWithComponent(Align.RIGHT);
		
		SettingsPanel settingsPanel = gui.getSettingsPanel();
		settingsPanel.setSettingsValue(model.getCoupling().getDates(), model.getCoupling().getTimestepUnit(),
				model.getCoupling().isStdoutRedirect(), model.getCoupling().isRootRedirect());
	}

	private void updateGlassPaneWithConnections() {
		
		// This has to be done in an extra thread, because the process of 
		// drawing the arrow on the GlassPane has to wait for the panels to 
		// set up the layout and draw themselves. Only after that the 
		// information about the location of the CheckBoxes, that we 
		// need to draw the arrows, will be available.
		SwingUtilities.invokeLater(new Runnable() {			
			@Override
			public void run() {
				// make connections for transient couples
				for (Connection c : model.getConnections()) {
					checkTransientCheckBox(c.getSourceAlign(), c.getSourceIndex());
					checkTransientCheckBox(c.getTargetAlign(), c.getTargetIndex());
					glassPane.addConnectionArrow(c);
					pairTransientPanels(c);
				}
			}
		});		
	}

	/**
	 * Resets last selected transient panel. Forgets all memorized previous
	 * selection of transient panels, of both left and right side.
	 */
	private void resetLastSelectedTransientPanelIndices() {
		model.resetLastSelectedTransientPanelIndex(Align.LEFT);
		model.resetLastSelectedTransientPanelIndex(Align.RIGHT);
	}

	/**
	 * Set checkbox at given align and index <i>enabled</i>.
	 * 
	 * @param align
	 *            LEFT or RIGHT
	 * @param index
	 *            of transient panel
	 */
	private void checkTransientCheckBox(Align align, int index) {
		gui.getComponentPanel(align).setCheckBoxChecked(index, true);
	}

	/**
	 * Set checkbox at given align and index <i>disabled</i>.
	 * 
	 * @param align
	 *            LEFT or RIGHT
	 * @param index
	 *            of transient panel
	 */
	private void uncheckTransientCheckBox(Align align, int index) {
		gui.getComponentPanel(align).setCheckBoxChecked(index, false);
	}

	/**
	 * Uncheck all checkboxes that do not have connections.
	 * 
	 * @param align
	 *            LEFT or RIGHT
	 */
	private void uncheckUnconnectedTransientCheckBoxes(Align align) {
		for (int i = 0; i < model.getSize(align); i++) {
			if (!model.isConnectedAt(align, i)) {
				uncheckTransientCheckBox(align, i);
			}
		}
	}

	/**
	 * Disables all CheckBoxes on side align but the ones that have the given
	 * transient standard name and collectionSize
	 * 
	 * @param align
	 *            LEFT or RIGHT
	 * @param transientStandardName
	 */
	private void disableAllCheckBoxesBut(Align align, String transientStandardName,
			int collectionSize) {
		for (int i = 0; i < model.getSize(align); i++) {
			if (model.getTransientStandardName(align, i).equals(transientStandardName)
					&& model.getCollectionSize(align, i) == collectionSize
					&& !model.isConnectedAt(align, i)) {
				gui.getComponentPanel(align).setCheckBoxEnabled(i, true);
			} else {
				gui.getComponentPanel(align).setCheckBoxEnabled(i, false);
			}
		}
	}

	/**
	 * Disables all CheckBoxes on side align but the one at index.
	 * 
	 * @param align
	 * @param index
	 */
	private void disableAllCheckBoxesBut(Align align, int index) {
		for (int i = 0; i < model.getSize(align); i++) {
			if (i == index) {
				continue;
			}
			gui.getComponentPanel(align).setCheckBoxEnabled(i, false);
		}
	}

	private void enableAllCheckBoxes() {
		gui.getComponentPanel(Align.LEFT).setAllCheckBoxesEnabled(true);
		gui.getComponentPanel(Align.RIGHT).setAllCheckBoxesEnabled(true);
	}

	public void repaintGlassPane() {
		glassPane.repaint();
	}

	public void showSelectComponentsDialog() {
		if (selectComponentDialog == null) {
			selectComponentDialog = new SelectComponentsDialog(this, model);
			selectComponentDialog.setLocationRelativeTo(gui);
			selectComponentDialog.setVisible(true);
		} else {
			//prevent building a new dialog, instead bring the previous dialog to front
			java.awt.EventQueue.invokeLater(new Runnable() {
				@Override
				public void run() {
					if (selectComponentDialog != null) {
						selectComponentDialog.toFront();
						selectComponentDialog.repaint();
					}
				}
			});
		}
	}

	public boolean hasDefaultCouplingParams() {
		return model.getDefaultCouplingParameter() != null;
	}
	
	public void setDefaultCouplingParams(TransientCouple donor) {
		model.setDefaultCouplingParameter(donor);
	}
	
	public TransientCouple getDefaultCouplingParams() {
		return model.getDefaultCouplingParameter();
	}

	/**
	 * Callback method when an user selected two components in SelectComponentsDialog
	 * 
	 * @param leftCompId
	 * @param rightCompId
	 */
	@Override
	public void onComponentsSelected(int leftCompId, int rightCompId) {
		model.loadComponents(leftCompId, rightCompId);
		
		updateGuiWithComponent(Align.LEFT);
		updateGuiWithComponent(Align.RIGHT);
		
		gui.resetGlassPane();
		updateGlassPaneWithConnections();

		resetLastSelectedTransientPanelIndices();
		enableAllCheckBoxes();
		
		state = GuiState.FULLY_LOADED;
		
		selectedConnections.clear();
	}

	@Override
	public void onSelectComponentsDialogClosed() {
		selectComponentDialog = null;
	}
	
	/**
	 * Callback to be called when user changed the timestep unit type
	 * 
	 * @return true, if the user allowed to changed the type, false - otherwise
	 */
	public boolean onNotCompatibleTimestepUnitTypeSelected() {
		Map<String, Couple> couples = model.getCoupling().getCouples();
		
		//having couples means we have at least one transient couple
		if (couples != null && !couples.isEmpty()) {

			int option = gui.showOppositeTimestepUnitTypeWarningDialog();
			if (option == JOptionPane.NO_OPTION || option == JOptionPane.CLOSED_OPTION) {
				return false;
			} else if (option == JOptionPane.YES_OPTION) {
				model.resetTimestepValuesForAllTransientCouples();
				return true;
			}
		}
		//no couples found, allow user to change the type with no constrains
		return true;
	}
}
