package de.mpg.mpimet.yacgui.junit;

import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.junit.Before;
import org.junit.Ignore;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;

import de.mpg.mpimet.yacgui.model.Model;
import de.mpg.mpimet.yacgui.utils.Utils;
import de.mpg.mpimet.yacgui.utils.XmlDomHandler;
import de.mpg.mpimet.yacgui.utils.XmlDomHandler.XmlType;
import de.mpg.mpimet.yacgui.utils.XmlDomHandlerException;
import de.mpg.mpimet.yacgui.xml.Component;
import de.mpg.mpimet.yacgui.xml.Coupling;
import de.mpg.mpimet.yacgui.xml.Grid;
import de.mpg.mpimet.yacgui.xml.Transient;
import de.mpg.mpimet.yacgui.xml.TransientGridRef;

public class ModelTest extends Model {

	private final static String COMPONENT1 = "res/xml/component1.xml";
	private final static String COMPONENT2 = "res/xml/component2.xml";
	private final static String COUPLING = "res/xml/coupling.xml";
	private final static Align LEFT = Align.LEFT;
	private final static Align RIGHT = Align.RIGHT;

	// values to be expected
	List<TransientGridRef> transientGridRefLeft = Arrays.asList(
			new TransientGridRef(1, 1, 1, 2), new TransientGridRef(2, 1, 2, 2),
			new TransientGridRef(3, 3, 3, 2), new TransientGridRef(4, 4, 4, 2),
			new TransientGridRef(5, 5, 5, 2));
	List<TransientGridRef> transientGridRefRight = Arrays.asList(
			new TransientGridRef(1, 1, 2, 2), new TransientGridRef(2, 2, 2, 2),
			new TransientGridRef(3, 3, 2, 2), new TransientGridRef(4, 4, 6, 2),
			new TransientGridRef(5, 5, 6, 2));

	/** Component validating to component.xsd */
	private Component dummyComponent;

	@Rule
	public ExpectedException thrown = ExpectedException.none();

	@Before
	public void setUp() throws Exception {
		if (dummyComponent == null) {
			dummyComponent = new Component(0, "", "", "", new HashMap<Integer, TransientGridRef>());
			dummyComponent.setTransients(new HashMap<Integer, Transient>());
			dummyComponent.setGrids(new HashMap<Integer, Grid>());
		}
	}

	@Test
	public void testModel() {
		Model model = new Model();
		assertTrue("Model is null", model != null);
	}

	@Test
	public void testLoadComponent_leftFirst() {
		try {
			Component component = new XmlDomHandler(COMPONENT1, XmlType.COMPONENT)
					.buildComponent();

			addComponent(component, LEFT);
			assertTrue(leftSide != null);

			List<TransientGridRef> transientGridRefs = new ArrayList<TransientGridRef>(
					component.getTransientGridRefs().values());
			transientGridRefs = Utils
					.sortByTransientGridRefId(transientGridRefs);
			assertTrue(leftSide.sortedTransientGridRefs.equals(transientGridRefs));

			int id = component.getId();
			assertTrue(coupling.getComponentsMap().containsKey(id));
			Component couplingComponent = coupling.getComponentsMap().get(id);
			assertTrue(couplingComponent.getTransientGridRefs().equals(
					component.getTransientGridRefs()));
			assertTrue(coupling.getTransients().equals(
					component.getTransients()));
			assertTrue(coupling.getGrids().equals(component.getGrids()));
		} catch (XmlDomHandlerException e) {
			e.printStackTrace();
		}
	}

	@Test
	public void testLoadComponent_rightFirst() {
		try {
			Component component = new XmlDomHandler(COMPONENT1, XmlType.COMPONENT).buildComponent();

			addComponent(component, RIGHT);
			assertTrue(rightSide != null);

			List<TransientGridRef> transientGridRefs = new ArrayList<TransientGridRef>(
					component.getTransientGridRefs().values());
			transientGridRefs = Utils.sortByTransientGridRefId(transientGridRefs);
			assertTrue(rightSide.sortedTransientGridRefs.equals(transientGridRefs));

			int id = component.getId();
			assertTrue(coupling.getComponentsMap().containsKey(id));
			Component couplingComponent = coupling.getComponentsMap().get(id);
			assertTrue(couplingComponent.getTransientGridRefs().equals(component.getTransientGridRefs()));
			assertTrue(coupling.getTransients().equals(component.getTransients()));
			assertTrue(coupling.getGrids().equals(component.getGrids()));
		} catch (XmlDomHandlerException e) {
			e.printStackTrace();
		}
	}

	@Test
	public void testLoadComponent_leftOverwrite() {
		try {
			// load a component
			Component component1 = new XmlDomHandler(COMPONENT1, XmlType.COMPONENT)
					.buildComponent();

			addComponent(component1, LEFT);

			// overwrite with different component
			Component component2 = new XmlDomHandler(COMPONENT2, XmlType.COMPONENT)
					.buildComponent();
			addComponent(component2, LEFT);

			assertTrue(leftSide != null);
			assertTrue(leftSide.component.equals(component2));

			List<TransientGridRef> transientGridRefs = new ArrayList<TransientGridRef>(
					component2.getTransientGridRefs().values());
			transientGridRefs = Utils
					.sortByTransientGridRefId(transientGridRefs);
			assertTrue(leftSide.sortedTransientGridRefs.equals(transientGridRefs));

			int id = component2.getId();
			assertTrue(coupling.getComponentsMap().containsKey(id));
			Component couplingComponent = coupling.getComponentsMap().get(id);
			assertTrue(couplingComponent.getTransientGridRefs().equals(
					component2.getTransientGridRefs()));
			assertTrue(coupling.getTransients().equals(
					component2.getTransients()));
			assertTrue(coupling.getGrids().equals(component2.getGrids()));
		} catch (XmlDomHandlerException e) {
			e.printStackTrace();
		}
	}

	@Test
	public void testLoadComponent_leftSecond() {
		try {
			// load right component first
			Component component2 = new XmlDomHandler(COMPONENT2, XmlType.COMPONENT)
					.buildComponent();
			addComponent(component2, RIGHT);

			// then load left component
			Component component1 = new XmlDomHandler(COMPONENT1, XmlType.COMPONENT)
					.buildComponent();
			addComponent(component1, LEFT);

			assertTrue(leftSide != null);
			assertTrue(rightSide != null);
			assertTrue(rightSide.component.equals(component2));
			assertFalse(leftSide.component.equals(component1));

			List<TransientGridRef> transientGridRefs2 = new ArrayList<TransientGridRef>(
					component2.getTransientGridRefs().values());
			transientGridRefs2 = Utils
					.sortByTransientGridRefId(transientGridRefs2);
			assertTrue(rightSide.sortedTransientGridRefs.equals(transientGridRefs2));
			List<TransientGridRef> transientGridRefs1 = new ArrayList<TransientGridRef>(
					component1.getTransientGridRefs().values());
			transientGridRefs1 = Utils
					.sortByTransientGridRefId(transientGridRefs1);
			assertTrue(leftSide.sortedTransientGridRefs.equals(transientGridRefs1));

			int idRight = component2.getId();
			int idLeft = component1.getId();
			assertTrue(coupling.getComponentsMap().containsKey(idRight));
			assertTrue(coupling.getComponentsMap().containsKey(idLeft));

			Component couplingComponentRight = coupling.getComponentsMap().get(
					idRight);
			Component couplingComponentLeft = coupling.getComponentsMap().get(
					idLeft);
			assertTrue(couplingComponentRight.getTransientGridRefs().equals(
					component2.getTransientGridRefs()));
			assertTrue(couplingComponentLeft.getTransientGridRefs().equals(
					component1.getTransientGridRefs()));

			List<String> transientsRight = Utils
					.getTransientStandardNames(component2.getTransients());
			List<String> transientsLeft = Utils
					.getTransientStandardNames(component1.getTransients());
			assertTrue(Utils.transientsContainAll(
					coupling.getTransients(), transientsRight));
			assertTrue(Utils.transientsContainAll(
					coupling.getTransients(), transientsLeft));

			List<String> gridsRight = Utils.getAliasNames(component2
					.getGrids());
			List<String> gridsLeft = Utils.getAliasNames(component1
					.getGrids());
			assertTrue(Utils.gridsContainAll(coupling.getGrids(),
					gridsRight));
			assertTrue(Utils.gridsContainAll(coupling.getGrids(),
					gridsLeft));
		} catch (XmlDomHandlerException e) {
			e.printStackTrace();
		}
	}

	@Test
	public void testLoadComponent_rightSecond() {
		try {
			// load left component first
			Component component1 = new XmlDomHandler(COMPONENT1, XmlType.COMPONENT)
					.buildComponent();
			addComponent(component1, LEFT);

			// then load right component
			Component component2 = new XmlDomHandler(COMPONENT2, XmlType.COMPONENT)
					.buildComponent();
			addComponent(component2, RIGHT);

			assertTrue(leftSide != null);
			assertTrue(rightSide != null);
			assertTrue(rightSide.component.equals(component2));
			assertTrue(leftSide.component.equals(component1));

			List<TransientGridRef> transientGridRefs2 = new ArrayList<TransientGridRef>(
					component2.getTransientGridRefs().values());
			transientGridRefs2 = Utils
					.sortByTransientGridRefId(transientGridRefs2);
			assertTrue(rightSide.sortedTransientGridRefs.equals(transientGridRefs2));
			List<TransientGridRef> transientGridRefs1 = new ArrayList<TransientGridRef>(
					component1.getTransientGridRefs().values());
			transientGridRefs1 = Utils
					.sortByTransientGridRefId(transientGridRefs1);
			assertTrue(leftSide.sortedTransientGridRefs.equals(transientGridRefs1));

			int idRight = component2.getId();
			int idLeft = component1.getId();
			assertTrue(coupling.getComponentsMap().containsKey(idRight));
			assertTrue(coupling.getComponentsMap().containsKey(idLeft));

			Component couplingComponentRight = coupling.getComponentsMap().get(
					idRight);
			Component couplingComponentLeft = coupling.getComponentsMap().get(
					idLeft);
			assertTrue(couplingComponentRight.getTransientGridRefs().equals(
					component2.getTransientGridRefs()));
			assertTrue(couplingComponentLeft.getTransientGridRefs().equals(
					component1.getTransientGridRefs()));

			List<String> transientsRight = Utils
					.getTransientStandardNames(component2.getTransients());
			List<String> transientsLeft = Utils
					.getTransientStandardNames(component1.getTransients());
			assertTrue(Utils.transientsContainAll(
					coupling.getTransients(), transientsRight));
			assertTrue(Utils.transientsContainAll(
					coupling.getTransients(), transientsLeft));

			List<String> gridsRight = Utils.getAliasNames(component2
					.getGrids());
			List<String> gridsLeft = Utils.getAliasNames(component1
					.getGrids());
			assertTrue(Utils.gridsContainAll(coupling.getGrids(),
					gridsRight));
			assertTrue(Utils.gridsContainAll(coupling.getGrids(),
					gridsLeft));
		} catch (XmlDomHandlerException e) {
			e.printStackTrace();
		}
	}

	@Test
	public void testLoadCoupling() {
		try {
			// precondition: components are loaded
			Component component1 = new XmlDomHandler(COMPONENT1, XmlType.COMPONENT)
					.buildComponent();
			Component component2 = new XmlDomHandler(COMPONENT2, XmlType.COMPONENT)
					.buildComponent();
			addComponent(component1, LEFT);
			addComponent(component2, RIGHT);

			// load coupling
			createNewCouple();

			// tests
			String coupleId = String.valueOf(component1.getId()) + "-"
					+ String.valueOf(component2.getId());
			assertTrue(coupling.getCouples().containsKey(coupleId));
			assertTrue(coupling.getCouples().get(coupleId)
					.getTransientCouples().isEmpty());
			assertTrue(activeCoupleId.equals(coupleId));
		} catch (XmlDomHandlerException e) {
			e.printStackTrace();
		}
	}

	@Test
	public void testLoadCouplingException_notLoaded() throws Error {
		// specify expected exception
		thrown.expect(IllegalStateException.class);
		thrown.expectMessage("Not both components are loaded");

		createNewCouple();
	}

	@Test
	public void testLoadCouplingException_onlyLeftLoaded()
			throws Error, XmlDomHandlerException {
		// specify expected exception
		thrown.expect(IllegalStateException.class);
		thrown.expectMessage("Not both components are loaded");

		Component component1 = new XmlDomHandler(COMPONENT1, XmlType.COMPONENT).buildComponent();
		addComponent(component1, LEFT);
		createNewCouple();
	}

	@Test
	public void testLoadCouplingException_onlyRightLoaded()
			throws Error, XmlDomHandlerException {
		// specify expected exception
		thrown.expect(IllegalStateException.class);
		thrown.expectMessage("Not both components are loaded");

		Component component1 = new XmlDomHandler(COMPONENT1, XmlType.COMPONENT).buildComponent();
		addComponent(component1, RIGHT);
		createNewCouple();
	}

	@Test
	public void testLoadCouplingCoupling_EmptyCoupling() {
		try {
			// load empty coupling
			// typical usage: components, transients, grids, transientGridRefs
			// are
			// given
			Component component1 = new XmlDomHandler(COMPONENT1, XmlType.COMPONENT)
					.buildComponent();
			Component component2 = new XmlDomHandler(COMPONENT2, XmlType.COMPONENT)
					.buildComponent();
			Coupling coupling = new Coupling();
			coupling.addComponent(component1);
			coupling.addComponent(component2);
			loadCoupling(coupling);

			// test
			assertTrue(this.coupling.equals(coupling));
			assertTrue(leftSide != null);
			assertTrue(rightSide != null);
			assertTrue(leftSide.component != null);
			assertTrue(rightSide.component != null);
		} catch (XmlDomHandlerException e) {
			e.printStackTrace();
		}
	}

	@Test
	public void testLoadCouplingCoupling_fromFile() {
		try {
			// load coupling from file
			Coupling coupling = new XmlDomHandler(COUPLING, XmlType.COUPLING).buildCoupling();
			loadCoupling(coupling);

			// test
			assertTrue(this.coupling.equals(coupling));
			assertTrue(leftSide != null);
			assertTrue(rightSide != null);
			assertTrue(leftSide.component != null);
			assertTrue(rightSide.component != null);
			assertTrue(activeCoupleId != null);
		} catch (XmlDomHandlerException e) {
			e.printStackTrace();
		}
	}

	@Test
	public void testLoadCouplingCouplingException_noComponents() {
		// specify expected exception
		thrown.expect(IllegalStateException.class);
		thrown.expectMessage("Coupling is incomplete");

		// incomplete coupling: no components loaded
		Coupling coupling = new Coupling();
		loadCoupling(coupling);
	}

	@Test
	public void testLoadCouple_detailed() {
		try {
			// load couple
			Coupling coupling = new XmlDomHandler(COUPLING, XmlType.COUPLING).buildCoupling();
			loadCoupling(coupling); // calls loadCouple();

			// tests: assume values from xml file
			assertTrue(leftSide.component.getId() == 1);
			assertTrue(rightSide.component.getId() == 2);
			assertTrue(connections.size() == 2);
			assertTrue(connections.hasConnectionAt(LEFT, 0));
			assertTrue(connections.hasConnectionAt(RIGHT, 2));
		} catch (XmlDomHandlerException e) {
			e.printStackTrace();
		}
	}

	@Test
	public void testResetData() {
		Model model = new Model();
		Model newModel = model;
		model.addComponent(dummyComponent, LEFT);
		model.resetData();
		assertTrue("model data is expected to be the same like new Model",
				model.equals(newModel));
	}

	@Test
	public void testGetSize() {
		loadCouplingXml();
		assertTrue(getSize(LEFT) == 5);
		assertTrue(getSize(RIGHT) == 5);
	}

	@Test
	public void testGetComponentId() {
		loadCouplingXml();
		assertTrue(getComponentId(LEFT) == 1);
		assertTrue(getComponentId(RIGHT) == 2);
	}

	@Test
	public void testGetComponentName() {
		loadCouplingXml();
		assertTrue(getComponentName(LEFT).equals("ICON-ocean"));
		assertTrue(getComponentName(RIGHT).equals("ICON-atmosphere"));
	}

	@Test
	public void testGetComponentModel() {
		loadCouplingXml();
		assertTrue(getComponentModel(LEFT).equals("some_model..."));
		assertTrue(getComponentModel(RIGHT).equals("..."));
	}

	@Test
	public void testGetComponentSimulated() {
		loadCouplingXml();
		assertTrue(getComponentSimulated(LEFT).equals("ocean"));
		assertTrue(getComponentSimulated(RIGHT).equals("atmosphere"));
	}

	@Test
	public void testGetTransientStandardName() {
		loadCouplingXml();
		assertTrue(getTransientStandardName(LEFT, 0).equals(
				"sea_surface_temperature"));
		assertTrue(getTransientStandardName(LEFT, 1).equals(
				"sea_surface_temperature"));
		assertTrue(getTransientStandardName(LEFT, 2).equals(
				"water_flux_into_sea_water"));
		assertTrue(getTransientStandardName(LEFT, 3).equals(
				"grid_eastward_wind"));
		assertTrue(getTransientStandardName(LEFT, 4).equals(
				"grid_northward_wind"));
		assertTrue(getTransientStandardName(RIGHT, 0).equals(
				"sea_surface_temperature"));
		assertTrue(getTransientStandardName(RIGHT, 1).equals("wind_speed"));
		assertTrue(getTransientStandardName(RIGHT, 2).equals(
				"water_flux_into_sea_water"));
		assertTrue(getTransientStandardName(RIGHT, 3).equals(
				"grid_eastward_wind"));
		assertTrue(getTransientStandardName(RIGHT, 4).equals(
				"grid_northward_wind"));
	}

	@Test
	public void testGetTransientStandardNameException() {
		thrown.expect(IndexOutOfBoundsException.class);
		thrown.expectMessage("transientGridRefs of LEFT only has 5 elements");
		loadCouplingXml();
		getTransientStandardName(LEFT, 5);
		thrown.expectMessage("transientGridRefs of RIGHT only has 5 elements");
		getTransientStandardName(RIGHT, 5);
	}

	@Test
	public void testGetGridName() {
		loadCouplingXml();
		assertTrue(getGridName(LEFT, 0).equals("grid1"));
		assertTrue(getGridName(LEFT, 1).equals("grid2"));
		assertTrue(getGridName(LEFT, 2).equals("grid3"));
		assertTrue(getGridName(LEFT, 3).equals("grid4"));
		assertTrue(getGridName(LEFT, 4).equals("grid5"));
		assertTrue(getGridName(RIGHT, 0).equals("grid2"));
		assertTrue(getGridName(RIGHT, 1).equals("grid2"));
		assertTrue(getGridName(RIGHT, 2).equals("grid2"));
		assertTrue(getGridName(RIGHT, 3).equals("grid6"));
		assertTrue(getGridName(RIGHT, 4).equals("grid6"));
	}

	@Test
	public void testGetGridNameException() {
		thrown.expect(IndexOutOfBoundsException.class);
		thrown.expectMessage("transientGridRefs of LEFT only has 5 elements");
		loadCouplingXml();
		getGridName(LEFT, 5);
		thrown.expectMessage("transientGridRefs of RIGHT only has 5 elements");
		getGridName(RIGHT, 5);
	}

	@Test
	public void testGetTransients() {
		loadCouplingXml();
		Map<Integer, Transient> transients = new HashMap<Integer, Transient>();
		transients.put(1, new Transient(1, "sea_surface_temperature"));
		transients.put(2, new Transient(2, "wind_speed"));
		transients.put(3, new Transient(3, "water_flux_into_sea_water"));
		transients.put(4, new Transient(4, "grid_eastward_wind"));
		transients.put(5, new Transient(5, "grid_northward_wind"));
		transients.put(6, new Transient(6, "std_name_1"));
		transients.put(7, new Transient(7, "std_name_2"));
		transients.put(8, new Transient(8, "std_name_3"));
		transients.put(9, new Transient(9, "std_name_4"));
		assertTrue(getTransients().equals(transients));
	}

	@Test
	public void testGetGrids() {
		loadCouplingXml();
		Map<Integer, Grid> grids = new HashMap<Integer, Grid>();
		grids.put(1, new Grid(1, "grid1"));
		grids.put(2, new Grid(2, "grid2"));
		grids.put(3, new Grid(3, "grid3"));
		grids.put(4, new Grid(4, "grid4"));
		grids.put(5, new Grid(5, "grid5"));
		grids.put(6, new Grid(6, "grid6"));
		assertTrue(getGrids().equals(grids));
	}

	@Test
	public void testHasLastSelectedTransient() {
		loadCouplingXml();
		assertFalse(hasLastSelectedTransient(LEFT));
		assertFalse(hasLastSelectedTransient(RIGHT));
		leftSide.lastSelectedTransientPanelIndex = 1;
		assertTrue(hasLastSelectedTransient(LEFT));
		rightSide.lastSelectedTransientPanelIndex = 1;
		assertTrue(hasLastSelectedTransient(RIGHT));
	}

	@Test
	public void testGetLastSelectedTransientPanelIndex() {
		loadCouplingXml();
		assertTrue(getLastSelectedTransientPanelIndex(LEFT) == NONE);
		assertTrue(getLastSelectedTransientPanelIndex(RIGHT) == NONE);
		leftSide.lastSelectedTransientPanelIndex = 1;
		assertTrue(getLastSelectedTransientPanelIndex(LEFT) == 1);
		rightSide.lastSelectedTransientPanelIndex = 1;
		assertTrue(getLastSelectedTransientPanelIndex(RIGHT) == 1);
	}

	@Test
	public void testSetLastSelectedTransientPanelIndex() {
		loadCouplingXml();
		setLastSelectedTransientPanelIndex(LEFT, 2);
		assertTrue(leftSide.lastSelectedTransientPanelIndex == 2);
		setLastSelectedTransientPanelIndex(RIGHT, 2);
		assertTrue(rightSide.lastSelectedTransientPanelIndex == 2);
	}

	@Test
	public void testResetLastSelectedTransientPanelIndex() {
		loadCouplingXml();
		setLastSelectedTransientPanelIndex(LEFT, 3);
		resetLastSelectedTransientPanelIndex(LEFT);
		assertTrue(leftSide.lastSelectedTransientPanelIndex == NONE);
		setLastSelectedTransientPanelIndex(RIGHT, 4);
		resetLastSelectedTransientPanelIndex(RIGHT);
		assertTrue(rightSide.lastSelectedTransientPanelIndex == NONE);
	}

	@Test
	public void testIsConnected() {
		loadCouplingXml();
		assertTrue(isConnectedAt(LEFT, 0));
		assertFalse(isConnectedAt(LEFT, 1));
		assertTrue(isConnectedAt(LEFT, 2));
		assertFalse(isConnectedAt(LEFT, 3));
		assertFalse(isConnectedAt(LEFT, 4));
		assertTrue(isConnectedAt(RIGHT, 0));
		assertFalse(isConnectedAt(RIGHT, 1));
		assertTrue(isConnectedAt(RIGHT, 2));
		assertFalse(isConnectedAt(RIGHT, 3));
		assertFalse(isConnectedAt(RIGHT, 4));
	}

	@Test
	public void testIsSource() {
		loadCouplingXml();
		assertTrue(isSource(LEFT, 0));
		assertFalse(isSource(LEFT, 1));
		assertFalse(isSource(LEFT, 2));
		assertFalse(isSource(LEFT, 3));
		assertFalse(isSource(LEFT, 4));
		assertFalse(isSource(RIGHT, 0));
		assertFalse(isSource(RIGHT, 1));
		assertTrue(isSource(RIGHT, 2));
		assertFalse(isSource(RIGHT, 3));
		assertFalse(isSource(RIGHT, 4));
	}
	
	// TODO test connections

	@Ignore("TODO")
	public void testGetConnection() {
		loadCouplingXml();

//		// source align = LEFT, source index = 0
//		Connection connection = getConnection(LEFT, 0);
//		assertTrue(connection.getSourceAlign() == LEFT);
//		assertTrue(connection.getTargetAlign() == RIGHT);
//		assertTrue(connection.getSourceIndex() == 0);
//		assertTrue(connection.getTargetIndices().equals(
//				Arrays.asList(new Integer(0))));
//
//		// target align = RIGHT, target index = 0
//		Connection connection1 = getConnection(RIGHT, 0);
//		assertTrue(connection1.equals(connection));
//
//		// target align = LEFT, target index = 2
//		Connection connection2 = getConnection(LEFT, 2);
//		assertTrue(connection2.getTargetAlign() == LEFT);
//		assertTrue(connection2.getSourceAlign() == RIGHT);
//		assertTrue(connection2.getTargetIndices().equals(
//				Arrays.asList(new Integer(2))));
//		assertTrue(connection2.getSourceIndex() == 2);
//
//		// source align = RIGHT, source index = 2
//		Connection connection3 = getConnection(RIGHT, 2);
//		assertTrue(connection3.equals(connection2));
//
//		// connection does not exist
//		Connection connection4 = getConnection(LEFT, 1);
//		assertTrue(connection4 == null);
	}

//	@Test
//	public void testSetConnection() {
//		loadCouple();
//
//		// source: LEFT, 4; target: RIGHT, 4
//		left.lastSelectedTransientPanelIndex = 4;
//		setConnection(RIGHT, 4);
//		Connection connection = getConnection(LEFT, 4);
//		assertTrue(connection.getSourceAlign() == LEFT);
//		assertTrue(connection.getTargetAlign() == RIGHT);
//		assertTrue(connection.getSourceIndex() == 4);
//		assertTrue(connection.getTargetIndices().equals(
//				Arrays.asList(new Integer(4))));
//	}
//
//	@Test
//	public void testSetConnectionException() {
//		thrown.expect(Error.class);
//		thrown.expectMessage("Cannot create connection.");
//		loadCouple();
//		setConnection(LEFT, 1);
//		setConnection(RIGHT, 3);
//
//	}
//
//	@Test
//	public void testSetConnectionException_IndexOutOfBounds() {
//		thrown.expect(IndexOutOfBoundsException.class);
//		thrown.expectMessage("transientGridRefs of left only has 5 elements");
//		loadCouple();
//		right.lastSelectedTransientPanelIndex = 5;
//		setConnection(LEFT, 5);
//	}
//
//	@Test
//	public void testRemoveConnection() {
//		loadCouple();
//
//		// source: LEFT, 0; target: right, 0
//		assertTrue(getConnection(LEFT, 0) != null);
//		removeConnection(LEFT, 0);
//		assertTrue(getConnection(LEFT, 0) == null);
//		assertTrue(getConnection(RIGHT, 0) == null);
//
//		// source: RIGHT, 2; target: left, 2
//		assertTrue(getConnection(LEFT, 2) != null);
//		removeConnection(LEFT, 2);
//		assertTrue(getConnection(LEFT, 2) == null);
//		assertTrue(getConnection(RIGHT, 2) == null);
//	}

	@Ignore
	public void testSetCouplingDataStringStringStringListOfString() {
		fail("Not yet implemented");
	}

	@Ignore
	public void testSetCouplingDataStringStringStringIntIntIntStringInt() {
		fail("Not yet implemented");
	}

	@Ignore
	public void testSetCouplingDataStringStringStringDoubleStringBooleanBoolean() {
		fail("Not yet implemented");
	}

	/**
	 * Loads coupling from XML file.
	 */
	private void loadCouplingXml() {
		try {
			Coupling coupling = new XmlDomHandler(COUPLING, XmlType.COUPLING).buildCoupling();
			loadCoupling(coupling);
		} catch (XmlDomHandlerException e) {
			e.printStackTrace();
		}
	}
}
