package de.mpg.mpimet.yacgui.model;

import de.mpg.mpimet.yacgui.model.Model.Align;

/**
 * Helper class to encapsulate the coordinates of couplings
 * 
 * @author Maxim Yastremsky
 */
public class Connection {
	
	private Align sourceAlign;
	private Align targetAlign;
	private int sourceIndex;
	private int targetIndex;
	private boolean hasInterpolations;
	
	/**
	 * Creates a new Connection.
	 * 
	 * @param sourceAlign	LEFT or RIGHT
	 * @param sourceIndex	index of source transient panel
	 * @param targetIndex	index of target transient panel
	 */
	public Connection(Align sourceAlign, int sourceIndex, int targetIndex, boolean hasInterpolation) {
		this(sourceAlign, sourceAlign.getOpposite(), sourceIndex, targetIndex, hasInterpolation);
	}
	
	/**
	 * Creates a new Connection.
	 * 
	 * @param sourceAlign
	 *            LEFT or RIGHT
	 * @param targetAlign
	 *            LEFT or RIGHT (other than sourceAlign)
	 * @param sourceIndex
	 *            index of source transient panel
	 * @param targetIndex
	 *            index of target transient panel
	 * @throws IllegalArgumentException
	 *             if targetAlign is not different from sourceAlign
	 */
	public Connection(Align sourceAlign, Align targetAlign, int sourceIndex, int targetIndex, boolean hasInterpolations) {
		this.sourceAlign = sourceAlign;
		if (targetAlign != sourceAlign.getOpposite()) {
			throw new IllegalArgumentException("Parameter Align targetAlign has to be other than sourceAlign.");
		}
		this.targetAlign = targetAlign;
		this.sourceIndex = sourceIndex;
		this.targetIndex = targetIndex;
		this.hasInterpolations = hasInterpolations;
	}
	
	/**
	 * @return the sourceAlign
	 */
	public Align getSourceAlign() {
		return sourceAlign;
	}
	
	/**
	 * @return the targetAlign
	 */
	public Align getTargetAlign() {
		return targetAlign;
	}
	
	/**
	 * @return the sourceIndex
	 */
	public int getSourceIndex() {
		return sourceIndex;
	}
	
	/**
	 * @return the targetIndex
	 */
	public int getTargetIndex() {
		return targetIndex;
	}
	
	public boolean hasInterpolations() {
		return hasInterpolations;
	}
	
	public void setHavingInterpolations(boolean hasInterp) {
		hasInterpolations = hasInterp;
	}
	
	/**
	 * Gets the index at given Align.
	 * 
	 * @param align	LEFT or RIGHT
	 * @return index at given align
	 */
	public int getIndexAt(Align align) {
		if (align == sourceAlign) {
			return sourceIndex;
		}
		return targetIndex;
	}
	
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + (hasInterpolations ? 1231 : 1237);
		result = prime * result
				+ ((sourceAlign == null) ? 0 : sourceAlign.hashCode());
		result = prime * result + sourceIndex;
		result = prime * result
				+ ((targetAlign == null) ? 0 : targetAlign.hashCode());
		result = prime * result + targetIndex;
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		Connection other = (Connection) obj;
		if (hasInterpolations != other.hasInterpolations)
			return false;
		if (sourceAlign != other.sourceAlign)
			return false;
		if (sourceIndex != other.sourceIndex)
			return false;
		if (targetAlign != other.targetAlign)
			return false;
		if (targetIndex != other.targetIndex)
			return false;
		return true;
	}

	@Override
	public String toString() {
		StringBuffer buf = new StringBuffer("[");
		buf.append(sourceAlign).append(":");
		buf.append(sourceIndex).append("]->[");
		buf.append(targetAlign).append(":");
		buf.append(targetIndex).append("]");
		return buf.toString();
	}

}
