package de.mpg.mpimet.yacgui.ui;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;

import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.KeyStroke;
import javax.swing.SwingUtilities;
import javax.swing.border.EmptyBorder;

//import de.mpg.mpimet.yacgui.controller.FileStatus;
import de.mpg.mpimet.yacgui.controller.GuiController;
import de.mpg.mpimet.yacgui.model.Model.Align;
import de.mpg.mpimet.yacgui.utils.VersionInfo;

/**
 * Main application window.
 */
public class CouplingGui extends JFrame implements WindowListener  {

	private static final long serialVersionUID = -3574101729200493910L;
	
	// window size: we use the standard of SVGA (800x600)
	private static final int WIDTH = 800;
	private static final int HEIGHT = 600;
	
	private static final int DIALOG_WIDTH = 450;
	private static final int DIALOG_HEIGHT = 300;
	
	// menu indices
	private static final int MENU_FILE_INDEX = 0;
	private static final int MENU_COUPLE_INDEX = 1;
	private static final int MENU_HELP_INDEX = 2;
	
	private static final int MENU_OPEN_INDEX = 0;
	private static final int MENU_SAVE_INDEX = 1;
	private static final int MENU_SAVE_AS_INDEX = 2;
	private static final int MENU_EXIT_INDEX = 4;
	
	// menu labels
	private static final String MENU_FILE = "File";
	private static final String MENU_OPEN = "Open...";
	private static final String MENU_SAVE = "Save";
	private static final String MENU_SAVE_AS = "Save as...";
	private static final String MENU_EXIT = "Exit";
	private static final String MENU_HELP = "Help";
	private static final String MENU_ABOUT = "About";
	private static final String MENU_OPEN_TOOLTIP = "Open component XML";
	private static final String MENU_SAVE_TOOLTIP = "Save current coupling";
	private static final String MENU_SAVE_AS_TOOLTIP = "Save coupling as...";
	private static final String MENU_EXIT_TOOLTIP = "Exit application";
	
	private static final String MENU_COUPLE = "Couple";
	private static final String MENU_ADD_COMPONENTS = "Add component(s)";
	private static final String MENU_ADD_COMPONENT_TOOLTIP = "Add additional component to current coupling";
	private static final String MENU_SELECT_COMPONENTS = "Select active components";
	private static final String MENU_SELECT_COMPONENTS_TOOLTIP = "Select active components from already loaded one";
	
	// GUI properties
	private static final String TITLE = "Coupling GUI";
	private static final int MARGIN = 10;
	private boolean isFilesaved;
	
	private HintPanel menuPanel;
	private JPanel contentPanel;
	private SettingsPanel settingsPanel;
	private ComponentPanel leftComponentPanel;
	private ComponentPanel rightComponentPanel;
	private ConnectionsGlassPane glassPane;
	private GuiController guiController;
	;
		
	/**
	 * Initializes AWT/Swing GUI, draws content, and shows GUI.
	 */
	public void createAndShowGui() {
		
		// init base panel
		JPanel base = new JPanel();
		base.setLayout(new BorderLayout());
		setContentPane(base);
		
		initMenuBar();

		// init window
		setTitle(TITLE);
		setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
		setSize(WIDTH, HEIGHT);
		setLocationRelativeTo(null);
		
		// init menu panel
		menuPanel = new HintPanel();
		base.add(menuPanel, BorderLayout.PAGE_START);
		
		// init content panel
		contentPanel = new JPanel();
		contentPanel.setLayout(new GridLayout(1, 2, MARGIN, MARGIN));
		contentPanel.setBorder(BorderFactory.createEmptyBorder(MARGIN, MARGIN, MARGIN, MARGIN));
		base.add(contentPanel, BorderLayout.CENTER);
		
		// init component panels
		leftComponentPanel = new ComponentPanel(Align.LEFT);
		rightComponentPanel = new ComponentPanel(Align.RIGHT);
		contentPanel.add(leftComponentPanel);
		contentPanel.add(rightComponentPanel);
		
		// init settings panel
		settingsPanel = new SettingsPanel();
		base.add(settingsPanel, BorderLayout.PAGE_END);
		settingsPanel.registerCouplingGui(this);
		
		// init glass pane
		glassPane = new ConnectionsGlassPane();
		setGlassPane(glassPane);
		glassPane.setVisible(true);
		
		// init model and controller
		guiController = GuiController.getInstance();
		guiController.registerCouplingGui(this);
		
		// show
		setVisible(true);
		
		//Add window listener 
		addWindowListener(this); 
	}
	
	/**
	 * Resets both left and right component panel.
	 */
	public void resetComponentPanels() {
		contentPanel.removeAll();
		leftComponentPanel = new ComponentPanel(Align.LEFT);
		rightComponentPanel = new ComponentPanel(Align.RIGHT);		
		contentPanel.add(leftComponentPanel);
		contentPanel.add(rightComponentPanel);
	}
	
	/**
	 * Resets component panel at side given by align.
	 * 
	 * @param align - the side which to reset
	 */
	public void resetComponentPanel(Align align) {
		if (align.isLeft()) {
			leftComponentPanel = new ComponentPanel(Align.LEFT);
		} else {
			rightComponentPanel = new ComponentPanel(Align.RIGHT);
		}
	}
	
	/**
	 * Resets the GlassPane, removing all drawn items.
	 */
	public void resetGlassPane() {
		glassPane.removeAll();
	}

	public int showOverwriteCoupledComponentWarningDialog(boolean delete) {
		//providing these labels in order to indicate default "NO" button preselected
		Object[] optionsLabels = {"Yes", "No"};
		int n = JOptionPane.showOptionDialog(this, "Selected component is already coupled with at least one of other components."
				+ "\nIn case of deleting it all related couples will be lost."
				+ "\nAre you sure you want to " + (delete ? "delete" : "overwrite") + " this component?", 
				(delete ? "Delete" : "Overwrite") + " selected component?", 
				JOptionPane.YES_NO_OPTION, JOptionPane.WARNING_MESSAGE, null, optionsLabels, optionsLabels[1]);
		return n;
	}
	
	public int showSaveCoplingBeforeChangeDialog() {
		//providing these labels in order to indicate default "Cancel" button preselected
		Object[] optionsLabels = {"Yes", "No", "Cancel"};
		int n = JOptionPane.showOptionDialog(this, "Do you want to save your current coupling first?",
				"Save current coupling?", JOptionPane.YES_NO_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE, 
				null, optionsLabels, optionsLabels[2]);
		return n;
	}
	
	public int showConfirmLoadingNewCouplingDialog() {
		//providing these labels in order to indicate default "NO" button preselected
		Object[] optionsLabels = {"Yes", "No"};
		int n = JOptionPane.showOptionDialog(this, "Do you want to load a different coupling instead of your current setup?",
				"Change to other coupling?", JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE, 
				null, optionsLabels, optionsLabels[1]);
		return n;
	}
	
	public void showIdenticalComponentWarningDialog(String name) {
		JOptionPane.showMessageDialog(this, "The name of chosen component (" + name + ") is identical to one of "
				+ "the components loaded before.\nPlease load new component.\n\n"
				+ "Hint: All loaded components can be displayed using "
				+ "Menu \"Couple\" -> Select active components.", "Warning", JOptionPane.WARNING_MESSAGE);
	}
	
	public int showOppositeTimestepUnitTypeWarningDialog() {
		//providing these labels in order to indicate default "NO" button preselected
		Object[] optionsLabels = {"Yes", "Cancel"};
		int n = JOptionPane.showOptionDialog(this, "Selected Timestep unit has type, which is not compatible with previous selection."
				+ "\nIn case of proceeding the timestep parameters for all existing transient couples will be reseted to default values."
				+ "\nAre you sure you want to proceed?",  "Changing timestep unit format", 
				JOptionPane.YES_NO_OPTION, JOptionPane.WARNING_MESSAGE, null, optionsLabels, optionsLabels[1]);
		return n;
	}
	
	public ConnectionsGlassPane getGlassPane() {
		return glassPane;
	}
	
	/**
	 * @return JPanel that holds all children GUI elements
	 */
	public JPanel getContentPanel() {
		return contentPanel;
	}
	
	public ComponentPanel getComponentPanel(Align align) {
		if (align.isLeft()) {
			return leftComponentPanel;
		}
		return rightComponentPanel;
	}
	
	public SettingsPanel getSettingsPanel() {
		return settingsPanel;
	}
	
	public boolean isMenuFileSaveEnabled() {
		return getJMenuBar().getMenu(MENU_FILE_INDEX).getItem(MENU_SAVE_INDEX).isEnabled();
	}
	
	public void setMenuFileSaveEnabled(boolean enabled) {
		getJMenuBar().getMenu(MENU_FILE_INDEX).getItem(MENU_SAVE_INDEX).setEnabled(enabled);
	}
	
	public boolean isMenuFileSaveAsEnabled() {
		return getJMenuBar().getMenu(MENU_FILE_INDEX).getItem(MENU_SAVE_AS_INDEX).isEnabled();
	}
	
	public void setMenuFileSaveAsEnabled(boolean enabled) {
		getJMenuBar().getMenu(MENU_FILE_INDEX).getItem(MENU_SAVE_AS_INDEX).setEnabled(enabled);
	}
	
	public void setTitleSaved() {
		setTitle(TITLE);
	}
	
	public void setTitleUnsaved() {
		setTitle("*" + TITLE);
	}
	
	public void setHint(String fileName) {
		menuPanel.setHint(fileName);
	}
	
	public void resetHint() {
		menuPanel.reset();
	}
	
	/**
	 * Launches GUI.
	 */
	public static void main(String[] args) {
		SwingUtilities.invokeLater(new Runnable() {
			@Override
			public void run() {
				CouplingGui gui = new CouplingGui();
				gui.createAndShowGui();
			}
		});
	}
	
	/**
	 * Defines menu entries.
	 */
	private void initMenuBar() {
		JMenuBar menuBar = new JMenuBar();
		
		// Main menu entries
		JMenu fileMenu = new JMenu(MENU_FILE);
		fileMenu.setMnemonic(KeyEvent.VK_F);
		
		// File -> Open
		JMenuItem fileOpen = new JMenuItem(MENU_OPEN);
		fileOpen.setToolTipText(MENU_OPEN_TOOLTIP);
		fileOpen.setMnemonic(KeyEvent.VK_O);
		fileOpen.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_O, ActionEvent.CTRL_MASK));
		fileOpen.addActionListener(new ActionListener() {			
			@Override
			public void actionPerformed(ActionEvent arg0) {
				guiController.openCoupling(null);
			}
		});
		
		// File -> Save
		JMenuItem fileSave = new JMenuItem(MENU_SAVE);
		fileSave.setToolTipText(MENU_SAVE_TOOLTIP);
		fileSave.setMnemonic(KeyEvent.VK_S);
		fileSave.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_S, ActionEvent.CTRL_MASK));
		fileSave.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent arg0) {
				guiController.saveCoupling(false);
			}
		});
		fileSave.setEnabled(false);
		
		// File -> Save as
		JMenuItem fileSaveAs = new JMenuItem(MENU_SAVE_AS);
		fileSaveAs.setToolTipText(MENU_SAVE_AS_TOOLTIP);
		fileSaveAs.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent arg0) {
				guiController.saveCoupling(true);
			}
		});
		fileSaveAs.setEnabled(false);
		
		// File -> Exit
		JMenuItem fileExit = new JMenuItem(MENU_EXIT);
		fileExit.setMnemonic(KeyEvent.VK_C);
		fileExit.setToolTipText(MENU_EXIT_TOOLTIP);
		fileExit.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_W, ActionEvent.CTRL_MASK));
		fileExit.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				isFilesaved = isMenuFileSaveEnabled();
				if (isFilesaved) {
					//int option = gui.showSaveCoplingBeforeChangeDialog();
		    		Object[] optionsLabels = {"Save", "Don't Save","Cancel"};
		            int confirm = JOptionPane.showOptionDialog(CouplingGui.this,"Do you want to save the coupling to file?",
		            "Exit Confirmation", JOptionPane.YES_NO_CANCEL_OPTION,
		            JOptionPane.QUESTION_MESSAGE, null, optionsLabels, optionsLabels[2]);
					if (confirm == JOptionPane.NO_OPTION) {
						 System.exit(1);
					} else if (confirm == JOptionPane.YES_OPTION) {
						guiController.saveCoupling(false);
						System.exit(1);
					}
		    	}
		    	else
		    		System.exit(0);	
			}
		});
		
		fileMenu.add(fileOpen, MENU_OPEN_INDEX);
		fileMenu.add(fileSave, MENU_SAVE_INDEX);
		fileMenu.add(fileSaveAs, MENU_SAVE_AS_INDEX);
		fileMenu.addSeparator();
		fileMenu.add(fileExit, MENU_EXIT_INDEX);
		
		// Main menu entries
		JMenu coupleMenu = new JMenu(MENU_COUPLE);
		fileMenu.setMnemonic(KeyEvent.VK_C);
		
		// Couple -> Add component
		JMenuItem addComp = new JMenuItem(MENU_ADD_COMPONENTS);
		addComp.setToolTipText(MENU_ADD_COMPONENT_TOOLTIP);
		addComp.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_M, ActionEvent.CTRL_MASK));
		addComp.addActionListener(new ActionListener() {			
			@Override
			public void actionPerformed(ActionEvent arg0) {
				guiController.openComponent(null, null);
			}
		});
		
		// Couple -> Selecte active comps
		JMenuItem selectActComp = new JMenuItem(MENU_SELECT_COMPONENTS);
		selectActComp.setToolTipText(MENU_SELECT_COMPONENTS_TOOLTIP);
		selectActComp.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_A, ActionEvent.CTRL_MASK));
		selectActComp.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent arg0) {
				guiController.showSelectComponentsDialog();
			}
		});

		coupleMenu.add(addComp);
		coupleMenu.add(selectActComp);
		
		// Main menu entries
		JMenu helpMenu = new JMenu(MENU_HELP);
		fileMenu.setMnemonic(KeyEvent.VK_H);
		
		JMenuItem aboutItem = new JMenuItem(MENU_ABOUT);
		aboutItem.addActionListener(new ActionListener() {			
			@Override
			public void actionPerformed(ActionEvent arg0) {
				JOptionPane.showMessageDialog(CouplingGui.this, "CouplingGui - it is a java swing application "
						+ "for simplifying the building of coupling xml files from two or more components xml files."
						+ "\n\nVersion " + VersionInfo.number,
						"About CouplingGui application", JOptionPane.INFORMATION_MESSAGE);
			}
		});
		
		helpMenu.add(aboutItem);
		

		menuBar.add(fileMenu, MENU_FILE_INDEX);
		menuBar.add(coupleMenu, MENU_COUPLE_INDEX);
		menuBar.add(helpMenu, MENU_HELP_INDEX);
		setJMenuBar(menuBar);
	}

	public void showSelectProperXmlFileDialog(String message) {
		JOptionPane.showMessageDialog(this, "The selected xml file doesn't seem to be appropriate. " 
				+ message, "Warning", JOptionPane.WARNING_MESSAGE);
	}

	public void showCouldntParseXmlFileDialog(String xsdMessages) {
		
		JPanel dialogPanel = new JPanel();
		dialogPanel.setLayout(new BoxLayout(dialogPanel, BoxLayout.PAGE_AXIS));
		dialogPanel.setBorder(new EmptyBorder(5, 5, 5, 5));

		String dialogMessage = "The selected xml file couldn't be parsed.\n";
		if (xsdMessages == null) {
			dialogMessage += "For more details provide with corresponding xsd file in the same directory.";
			dialogPanel.add(new JLabel(dialogMessage));
		} else {
			dialogPanel.setPreferredSize(new Dimension(DIALOG_WIDTH, DIALOG_HEIGHT));
			dialogMessage += "Logs from xsd validation:";
			dialogPanel.add(new JLabel(dialogMessage));
			
			JTextArea message = new JTextArea(xsdMessages);
            // make it look & act like a label
            message.setWrapStyleWord(true);
            message.setLineWrap(true);
            message.setEditable(false);
            message.setFocusable(false);
            message.setOpaque(false);
			
			JScrollPane scrollPane = new JScrollPane(message);
			dialogPanel.add(scrollPane);
		}
		
		JOptionPane.showMessageDialog(this, dialogPanel, "Warning", JOptionPane.WARNING_MESSAGE);
	}
	
	public void showXsdValidationDetailsDialog(String xsdMessages) {
		
		if (xsdMessages == null || xsdMessages.isEmpty()) {
			return;
		}
		
		JPanel dialogPanel = new JPanel();
		dialogPanel.setLayout(new BoxLayout(dialogPanel, BoxLayout.PAGE_AXIS));
		dialogPanel.setBorder(new EmptyBorder(5, 5, 5, 5));

		dialogPanel.add(new JLabel("Results of xsd validation process:\n"));
		dialogPanel.setPreferredSize(new Dimension(DIALOG_WIDTH, DIALOG_HEIGHT));
			
		JTextArea message = new JTextArea(xsdMessages);
        // make it look & act like a label
        message.setWrapStyleWord(true);
        message.setLineWrap(true);
        message.setEditable(false);
        message.setFocusable(false);
        message.setOpaque(false);
		
		JScrollPane scrollPane = new JScrollPane(message);
		dialogPanel.add(scrollPane);
		
		JOptionPane.showMessageDialog(this, dialogPanel, "Warning", JOptionPane.WARNING_MESSAGE);
	}

	@Override
	public void windowActivated(WindowEvent arg0) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void windowClosed(WindowEvent e) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void windowClosing(WindowEvent e) {
		// TODO Auto-generated method stub
		//if (true) {
		isFilesaved = isMenuFileSaveEnabled();
		if (isFilesaved) {
			//int option = gui.showSaveCoplingBeforeChangeDialog();
    		Object[] optionsLabels = {"Save", "Don't Save","Cancel"};
            int confirm = JOptionPane.showOptionDialog(CouplingGui.this,"Do you want to save the coupling to file?",
            "Exit Confirmation", JOptionPane.YES_NO_CANCEL_OPTION,
            JOptionPane.QUESTION_MESSAGE, null, optionsLabels, optionsLabels[2]);
			if (confirm == JOptionPane.NO_OPTION) {
				 System.exit(1);
			} else if (confirm == JOptionPane.YES_OPTION) {
				guiController.saveCoupling(false);
				System.exit(1);
			}
    	}
    	else
    		System.exit(0);	
		
	}

	@Override
	public void windowDeactivated(WindowEvent e) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void windowDeiconified(WindowEvent e) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void windowIconified(WindowEvent e) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void windowOpened(WindowEvent e) {
		// TODO Auto-generated method stub
		
	}
	
}

