package de.mpg.mpimet.yacgui.ui;

import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;

import javax.swing.BorderFactory;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.border.Border;

import com.toedter.calendar.JDateChooser;

import de.mpg.mpimet.yacgui.controller.FileStatus;
import de.mpg.mpimet.yacgui.controller.GuiController;
import de.mpg.mpimet.yacgui.utils.Default;
import de.mpg.mpimet.yacgui.xml.Calendar;
import de.mpg.mpimet.yacgui.xml.Dates;
import de.mpg.mpimet.yacgui.xml.TimestepUnit;

/**
 * Bottom panel of GUI, contains controls for setting the dates, timestep unit and redirect settings
 * 
 * @author Maxim Yastremsky
 */
public class SettingsPanel extends JPanel {

	private static final long serialVersionUID = -8928464335165056558L;
	
	private static final int MARGIN_TOP = 0;
	private static final int MARGIN_BOTTOM = 10;
	private static final int MARGIN_LEFT = 10;
	private static final int MARGIN_RIGHT = 10;
	
	private static final String TITLE = "Basic settings";
	private static final String CALENDAR = "Calendar:";
	private static final String START_DATE = "Start date:";
	private static final String END_DATE = "End date:";
	private static final String TIMESTEP_UNIT = "Timestep unit:";

	private static final String REDIRECT_STDOUT = "Stdout redirect";
	private static final String REDIRECT_OF_ROOT = "Root redirect";
	
	private JLabel startDateLabel;
	private JDateChooser startDateChooser;
	
	private JLabel endDateLabel;
	private JDateChooser endDateChooser;
	
	private JLabel calendarLabel;
	private JComboBox<Calendar> calendarComboBox;
	
	private JLabel timestepUnitLabel;
	private JComboBox<TimestepUnit> timestepUnitComboBox;

	private JCheckBox stdoutRedirectCheckBox;
	private JCheckBox rootRedirectCheckBox;
	
	private CouplingGui gui;
	protected FileStatus fileStatus;

	//protected ConnectionsGlassPane glassPane;

	public SettingsPanel() {
		createLayout();
	}
	
	/**
	 * Registers CouplingGui if not already present.
	 * 
	 * @param gui
	 *            CouplingGui to register
	 */
	public void registerCouplingGui(CouplingGui gui) {
		if (this.gui == null) {
			this.gui = gui;
			fileStatus = new FileStatus(gui);
		}
	}
	
	private void createLayout() {
		setLayout(new GridBagLayout());
		Border margin = BorderFactory.createEmptyBorder(MARGIN_TOP, MARGIN_LEFT, MARGIN_BOTTOM, MARGIN_RIGHT);
		Border title = BorderFactory.createTitledBorder(TITLE);
		setBorder(BorderFactory.createCompoundBorder(margin, title));
		
		startDateLabel = new JLabel(START_DATE);
		startDateChooser = new JDateChooser();
		startDateChooser.setDateFormatString(Default.DATE_FORMAT);
		startDateChooser.setFormattedDateString(Default.START_DATE);
		startDateChooser.getCalendarButton().addActionListener(new ActionListener() {			
			@Override
			public void actionPerformed(ActionEvent arg0) {
				fileStatus.edit();			
				gui.setMenuFileSaveAsEnabled(true);
			}
		});		
		
		endDateLabel = new JLabel(END_DATE);
		endDateChooser = new JDateChooser();
		endDateChooser.setDateFormatString(Default.DATE_FORMAT);
		endDateChooser.setFormattedDateString(Default.END_DATE);
		endDateChooser.getCalendarButton().addActionListener(new ActionListener() {			
			@Override
			public void actionPerformed(ActionEvent arg0) {
				fileStatus.edit();			
				gui.setMenuFileSaveAsEnabled(true);
			}
		});	
		
		calendarLabel = new JLabel(CALENDAR);
		calendarComboBox = new JComboBox<Calendar>(Calendar.values());
		calendarComboBox.addActionListener(new ActionListener() {			
			@Override
			public void actionPerformed(ActionEvent arg0) {
				
				Calendar cal = (Calendar) calendarComboBox.getSelectedItem();
				startDateChooser.setYACCalendar(cal);
				endDateChooser.setYACCalendar(cal);
				fileStatus.edit();			
				gui.setMenuFileSaveAsEnabled(true);
			}
		});		
		
		timestepUnitLabel = new JLabel(TIMESTEP_UNIT);
		timestepUnitComboBox = new JComboBox<TimestepUnit>(TimestepUnit.values());		
		timestepUnitComboBox.addItemListener(new ItemListener() {
			
			private TimestepUnit deselectedTimestepUnit;

			@Override
			public void itemStateChanged(ItemEvent e) {
				
				switch (e.getStateChange()) {
				case ItemEvent.DESELECTED:
					deselectedTimestepUnit = (TimestepUnit) e.getItem();
					break;
				case ItemEvent.SELECTED:
					
					if (deselectedTimestepUnit == null) {
						return;
					}
					
					TimestepUnit newUnit = (TimestepUnit) e.getItem();
					
					if (!deselectedTimestepUnit.isTypeCompatibleWith(newUnit)) {
						boolean shouldProceed = GuiController.getInstance().onNotCompatibleTimestepUnitTypeSelected();
					
						if (!shouldProceed) {
							timestepUnitComboBox.setSelectedItem(deselectedTimestepUnit);
						}
					}
					GuiController.getInstance().setTimestepUnit((TimestepUnit) timestepUnitComboBox.getSelectedItem());
					break;
				}
				fileStatus.edit();
				gui.setMenuFileSaveAsEnabled(true);
			}
		});
		
		stdoutRedirectCheckBox = new JCheckBox(REDIRECT_STDOUT);
		rootRedirectCheckBox = new JCheckBox(REDIRECT_OF_ROOT);
		rootRedirectCheckBox.setEnabled(false);
		
		stdoutRedirectCheckBox.addItemListener(new ItemListener() {
			@Override
			public void itemStateChanged(ItemEvent e) {
				if (stdoutRedirectCheckBox.isSelected()) {
					rootRedirectCheckBox.setEnabled(true);
				} else {
					rootRedirectCheckBox.setEnabled(false);
					rootRedirectCheckBox.setSelected(false);
				}
				fileStatus.edit();
				gui.setMenuFileSaveAsEnabled(true);
			}
		});
		
		GridBagConstraints c = new GridBagConstraints();
		c.insets = new Insets(5,5,0,5);  //top, left and right
		// add items to layout
		c.gridx = 0;
		c.gridy = 0;
		c.weightx = 0.0;
		c.anchor = GridBagConstraints.EAST;
		add(calendarLabel, c);
		
		c.gridx = 1;
		c.gridy = 0;
		c.weightx = 1.0;
		c.anchor = GridBagConstraints.WEST;
		add(calendarComboBox, c);
		
		c.gridx = 0;
		c.gridy = 1;
		c.weightx = 0.0;
		c.anchor = GridBagConstraints.EAST;
		add(startDateLabel, c);
		
		c.gridx = 1;
		c.gridy = 1;
		c.weightx = 0.5;
		c.anchor = GridBagConstraints.WEST;
		add(startDateChooser, c);
		
		c.gridx = 0;
		c.gridy = 2;
		c.weightx = 0.0;
		c.anchor = GridBagConstraints.EAST;
		add(endDateLabel, c);
		
		c.gridx = 1;
		c.gridy = 2;
		c.weightx = 0.5;
		c.anchor = GridBagConstraints.WEST;
		add(endDateChooser, c);
		
		c.gridx = 3;
		c.gridy = 0;
		c.weightx = 0.0;
		c.anchor = GridBagConstraints.WEST;
		add(timestepUnitLabel, c);
		
		c.gridx = 4;
		c.gridy = 0;
		c.weightx = 1.0;
		c.anchor = GridBagConstraints.WEST;
		add(timestepUnitComboBox, c);
		
		c.gridx = 3;
		c.gridy = 1;
		c.weightx = 0.0;
		c.anchor = GridBagConstraints.WEST;
		add(stdoutRedirectCheckBox, c);
		
		c.gridx = 3;
		c.gridy = 2;
		c.weightx = 0.0;
		c.anchor = GridBagConstraints.WEST;
		add(rootRedirectCheckBox, c);
		
		setSize(getPreferredSize());
	}
	
	public void setSettingsValue(Dates dates, TimestepUnit timestepUnit, boolean redirectStdout, boolean redirectOfRoot) {
		calendarComboBox.setSelectedItem(dates.getCalendar());
		startDateChooser.setFormattedDateString(dates.getStartDate());
		startDateChooser.setYACCalendar(dates.getCalendar());
		endDateChooser.setFormattedDateString(dates.getEndDate());
		endDateChooser.setYACCalendar(dates.getCalendar());
		
		timestepUnitComboBox.setSelectedItem(timestepUnit);

		//set first the value of rootRedirectCheckBox, then the value of stdoutRedirectCheckBox will correct previous
		rootRedirectCheckBox.setSelected(redirectOfRoot);
		stdoutRedirectCheckBox.setSelected(redirectStdout);
	}
	
	public Dates getDates() {
		return new Dates(startDateChooser.getFormattedDateString(), endDateChooser.
				getFormattedDateString(), (Calendar) calendarComboBox.getSelectedItem());
	}
	
	public TimestepUnit getTimestepUnit() {
		return (TimestepUnit) timestepUnitComboBox.getSelectedItem();
	}
	
	public boolean isStdoutRedirect() {
		return stdoutRedirectCheckBox.isSelected();
	}
	
	public boolean isRootRedirect() {
		return rootRedirectCheckBox.isSelected();
	}
}
