package de.mpg.mpimet.yacgui.xml;

import java.util.ArrayList;
import java.util.List;

import de.mpg.mpimet.yacgui.xml.interpolation.Interpolation;

/**
 * Java representation of {@code <transient_couple>} xml element.
 * 
 * @author Maxim Yastremsky
 */
public class TransientCouple {
	
	//default values for new couple
	private static boolean defaultUseSourceMask = true;
	private static boolean defaultUseTargetMask = true;
	private static boolean defaultEnforceWriteRestart = false;
	private static boolean defaultEnforceWriteWeightFile = false;
	//end of default value set
	
	/** ID referencing same Transient on both sides */
	private int transientId;
	
	/** 
	 * sourceComponentRef
	 * 
	 * Should take value only from 1 or 2
	 */
	private int sourceComponentRef;
	
	/** ID referencing TransientGridRef on sourceComponentRef side */
	private int sourceTransientGridRefId;
	
	/** ID referencing TransientGridRef on targetComponentRef side */
	private int targetTransientGridRefId;
	
	/** time step */
	private Timestep timestep;
	
	/** use source mask: {true, false} */
	private boolean useSourceMask;
	
	/** use target mask: {true, false} */
	private boolean useTargetMask;
	
	/** Interpolation sequence */
	private List<Interpolation> interpolations;
	
	/** mask value (optional, can be null) */
	private Double maskValue;
	
	/** debug mode */
	private DebugMode debugMode;
	
	/** enforce write restart */
	private boolean enforceWriteRestart;
	
	/** enforce write weight file */
	private boolean enforceWriteWeightFile;
	
	/** weight file name */
	private String weightFilename;
	
	/**
	 * Creates a new TransientCouple object with all member variables.
	 * 
	 * @param transientId
	 * @param sourceComponentRef
	 * @param sourceTransientGridRefId
	 * @param targetTransienGridtRefId
	 * @param timestep
	 * @param useSourceMask
	 * @param useTargetMask
	 * @param interpolations
	 * @param maskValue
	 * @param debugMode
	 * @param enforceWriteRestart
	 */
	public TransientCouple(int transientId, int sourceComponentRef, int sourceTransientGridRefId, 
			int targetTransienGridtRefId, Timestep timestep, boolean useSourceMask,
			boolean useTargetMask, List<Interpolation> interpolations, Double maskValue,
			DebugMode debugMode, boolean enforceWriteRestart, boolean enforceWriteWeightFile,
			String weightFilename) {
		this.transientId = transientId;
		this.sourceComponentRef = sourceComponentRef;
		this.sourceTransientGridRefId = sourceTransientGridRefId;
		this.targetTransientGridRefId = targetTransienGridtRefId;
		this.timestep = timestep;
		this.useSourceMask = useSourceMask;
		this.useTargetMask = useTargetMask;
		this.interpolations = interpolations;
		this.maskValue = maskValue;
		this.debugMode = debugMode;
		this.enforceWriteRestart = enforceWriteRestart;
		this.enforceWriteWeightFile = enforceWriteWeightFile;
		this.weightFilename = weightFilename;
	}
	
	/**
	 * Creates a completely new TransientCouple object with default parameters and ids with value zero.
	 */
	public TransientCouple() {
		this(0, 0, 0, 0);
	}
	
	/**
	 * Creates a new TransientCouple object with default parameters.
	 * 
	 * @param transientId
	 * @param sourceComponentRef
	 * @param sourceTransientGridRefId
	 * @param targetTransienGridtRefId
	 */
	public TransientCouple(int transientId, int sourceComponentRef, int sourceTransientGridRefId, int targetTransienGridtRefId) {
		this(transientId, sourceComponentRef, sourceTransientGridRefId, targetTransienGridtRefId,
				Timestep.getDefault(), defaultUseSourceMask, defaultUseTargetMask, null, null, 
				DebugMode.getDefault(), defaultEnforceWriteRestart, defaultEnforceWriteWeightFile, null);
	}

	/**
	 * Loads (copies) the TransientCouple parameters from other object of TransientCouple
	 * 
	 * @param couple - donor couple
	 */
	public void loadTransientCoupleData(TransientCouple couple) {
		
		if (couple == null || this.equals(couple)) {
			return;
		}
		
		Timestep step = couple.getTimestep();
		
		if (step != null) {
			this.timestep = new Timestep(step.getSource(), step.getTarget(), step.getCouplingPeriod(), 
					new String(step.getCouplingOperation().toString()), step.getSourceTimeLag(), step.getTargetTimeLag());
		} else {
			this.timestep = Timestep.getDefault();
		}

		this.useSourceMask = couple.useSourceMask;
		this.useTargetMask = couple.useTargetMask;
		
		this.interpolations = couple.interpolations == null ? null : new ArrayList<Interpolation>(couple.interpolations);
		
		this.maskValue = couple.maskValue == null ? null : new Double(couple.maskValue.doubleValue());

		if (couple.debugMode != null) {
			this.debugMode = new DebugMode(couple.debugMode.isAtSourceBeforeInterpolation(),
					couple.debugMode.isAtSourceAfterInterpolation(), 
					couple.debugMode.isAtTarget());
		} else {
			this.debugMode = DebugMode.getDefault();
		}
		
		this.enforceWriteRestart = couple.enforceWriteRestart;
		
		this.enforceWriteWeightFile = couple.enforceWriteWeightFile;
		this.weightFilename = couple.weightFilename == null ? null : new String(couple.weightFilename);
	}

	public int getTransientId() {
		return transientId;
	}

	public int getSourceComponentRef() {
		return sourceComponentRef;
	}

	public int getSourceTransientGridRefId() {
		return sourceTransientGridRefId;
	}
	
	public int getTargetTransientGridRefIf() {
		return targetTransientGridRefId;
	}
	
	public Timestep getTimestep() {
		return timestep;
	}
	
	public void setTimestep(Timestep timestep) {
		this.timestep = timestep;
	}

	public boolean getUseSourceMask() {
		return useSourceMask;
	}
	
	public boolean getUseTargetMask() {
		return useTargetMask;
	}
	
	public void setUseSourceMask(boolean useSourceMask) {
		this.useSourceMask = useSourceMask;
	}
	
	public void setUseTargetMask(boolean useTargetMask) {
		this.useTargetMask = useTargetMask;
	}

	public List<Interpolation> getInterpolations() {
		return interpolations;
	}
	
	public void setInterpolations(List<Interpolation> interpolations) {
		this.interpolations = interpolations;
	}

	public Double getMaskValue() {
		return maskValue;
	}
	
	public void setMaskValue(Double maskValue) {
		this.maskValue = maskValue;
	}

	public DebugMode getDebugMode() {
		return debugMode;
	}

	public void setDebugMode(DebugMode debugMode) {
		this.debugMode = debugMode;
	}
	
	public boolean isEnforceWriteRestart() {
		return enforceWriteRestart;
	}
	
	public void setEnforceWriteRestart(boolean enforceWriteRestart) {
		this.enforceWriteRestart = enforceWriteRestart;
	}
	
	public boolean isEnforceWriteWeightFile() {
		return enforceWriteWeightFile;
	}
	
	public void setEnforceWriteWeightFile(boolean enforceWriteWeightFile) {
		this.enforceWriteWeightFile = enforceWriteWeightFile;
	}
	
	public String getWeightFilename() {
		return weightFilename;
	}
	
	public void setWeightFilename(String weightFilename) {
		this.weightFilename = weightFilename;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((debugMode == null) ? 0 : debugMode.hashCode());
		result = prime * result + (enforceWriteRestart ? 1231 : 1237);
		result = prime * result + (enforceWriteWeightFile ? 1231 : 1237);
		result = prime * result
				+ ((interpolations == null) ? 0 : interpolations.hashCode());
		result = prime * result
				+ ((maskValue == null) ? 0 : maskValue.hashCode());
		result = prime * result + sourceComponentRef;
		result = prime * result + sourceTransientGridRefId;
		result = prime * result + targetTransientGridRefId;
		result = prime * result
				+ ((timestep == null) ? 0 : timestep.hashCode());
		result = prime * result + transientId;
		result = prime * result + (useSourceMask ? 1231 : 1237);
		result = prime * result + (useTargetMask ? 1231 : 1237);
		result = prime * result
				+ ((weightFilename == null) ? 0 : weightFilename.hashCode());
		return result;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		TransientCouple other = (TransientCouple) obj;
		if (debugMode == null) {
			if (other.debugMode != null)
				return false;
		} else if (!debugMode.equals(other.debugMode))
			return false;
		if (enforceWriteRestart != other.enforceWriteRestart)
			return false;
		if (enforceWriteWeightFile != other.enforceWriteWeightFile)
			return false;
		if (interpolations == null) {
			if (other.interpolations != null)
				return false;
		} else if (!interpolations.equals(other.interpolations))
			return false;
		if (maskValue == null) {
			if (other.maskValue != null)
				return false;
		} else if (!maskValue.equals(other.maskValue))
			return false;
		if (sourceComponentRef != other.sourceComponentRef)
			return false;
		if (sourceTransientGridRefId != other.sourceTransientGridRefId)
			return false;
		if (targetTransientGridRefId != other.targetTransientGridRefId)
			return false;
		if (timestep == null) {
			if (other.timestep != null)
				return false;
		} else if (!timestep.equals(other.timestep))
			return false;
		if (transientId != other.transientId)
			return false;
		if (useSourceMask != other.useSourceMask)
			return false;
		if (useTargetMask != other.useTargetMask)
			return false;
		if (weightFilename == null) {
			if (other.weightFilename != null)
				return false;
		} else if (!weightFilename.equals(other.weightFilename))
			return false;
		return true;
	}
}
