package de.mpg.mpimet.yacgui.xml.interpolation;

import java.util.List;

import de.mpg.mpimet.yacgui.utils.Logger;
import de.mpg.mpimet.yacgui.xml.Dom;
import de.mpg.mpimet.yacgui.xml.interpolation.ConservativeInterpolation.Normalisation;
import de.mpg.mpimet.yacgui.xml.interpolation.NNearestNeighborInterpolation.Weighted;
import de.mpg.mpimet.yacgui.xml.interpolation.UserFileInterpolation.YacLocation;

/**
 * Definite the class of InterpolationFactory for loading current interpolations  
 * 
 * @author Maxim Yastremsky
 * @update Junpei Xu 30/06/2016
 */

public class InterpolationFactory {
	
	/**
	 * Creates an implementation createInterpolation of {@link Interpolation} according to 
	 * given method, and the attributes respectively
	 * @param method	String representation of {@link InterpolationMethod}
	 * @param attributes  List of parameters with a concrete InterpolationMethod 
	 * @return function return value of a object {@link InterpolationMethod}.
	 */
	public static Interpolation createInterpolation(String method, List<String[]> attributes) {
		if (!InterpolationMethod.isValidName(method)) {
			throw new IllegalArgumentException("'" + method + "' is not a valid method");
		}
		InterpolationMethod interpolationMethod = InterpolationMethod.fromName(method);
		
		switch (interpolationMethod) {
		case AVERAGE:
			return createAverage(attributes);
		case N_NEAREST_NEIGHBOR:
			return createNNearestNeighbor(attributes);
		case CONSERVATIVE:
			return createConservative(attributes);
		case PATCH_RECOVERY:
			return createPatchRecovery(attributes);
		case FIXED_VALUE:
			return createFixedValue(attributes);
		case USER_FILE:
			return createUserFile(attributes);
		case SOURCE2TARGET_MAP:
			return createSource2TargetMap(attributes);
		case SMOOTHED_PATCH_RECOVERY:
			return createSmoothedPatchRecovery(attributes);
		default:
			throw new IllegalArgumentException("'" + method + "' is not a valid method");
		}
	}

	/**
	 * Creates an implementation createInterpolation of {@link AverageInterpolation} 
	 * @param attributes  List of parameters of AverageInterpolation: partial coverage, weighted
	 * the order of element attributes follows Alphabetical order
	 * @return a object of {@link AverageInterpolation}.
	 */
	private static AverageInterpolation createAverage(List<String[]> attributes) {
		try {
			checkAttributesSize(attributes, 2);
			
			String[] attribute = attributes.get(0);
			checkAttributeLength(attribute);
			checkAttributeName(attribute, Dom.Attribute.PARTIAL_COVERAGE);
			
			String[] attribute2 = attributes.get(1);
			checkAttributeLength(attribute2);
			checkAttributeName(attribute2, Dom.Attribute.WEIGHTED);
			
			boolean partialCoverage = Boolean.parseBoolean(attribute[1]);
			if (!Weighted.isValidString(attribute2[1])) {
				throw new IllegalArgumentException("Weighted String invalid");
			}
			Weighted weighted = Weighted.objectFromString(attribute2[1]);
			
			return new AverageInterpolation(partialCoverage, weighted);
		} catch (IllegalArgumentException e) {
			Logger.logError(e.getMessage());
			return null;
		}
	}
	
	/**
	 * Creates an implementation createInterpolation of {@link NNearestNeighborInterpolation} 
	 * @param attributes  List of parameters of NNearestNeighborInterpolation: n, weighted
	 * n is the number of source point values requested for the nearest neighbor interpolation
	 * the order of element attributes follows Alphabetical order
	 * @return a object of {@link NNearestNeighborInterpolation}.
	 */
	private static NNearestNeighborInterpolation createNNearestNeighbor(List<String[]> attributes) {
		try {
			checkAttributesSize(attributes, 2);
			
			String[] attribute1 = attributes.get(0);
			checkAttributeLength(attribute1);
			checkAttributeName(attribute1, Dom.Attribute.N);
			
			String[] attribute2 = attributes.get(1);
			checkAttributeLength(attribute2);
			checkAttributeName(attribute2, Dom.Attribute.WEIGHTED);
			
			int n = Integer.parseInt(attribute1[1]);
			if (n < 1) {
				throw new IllegalArgumentException("n must be bigger than 0");
			}
			if (!Weighted.isValidString(attribute2[1])) {
				throw new IllegalArgumentException("Weighted String invalid");
			}
			Weighted weighted = Weighted.objectFromString(attribute2[1]);
			
			return new NNearestNeighborInterpolation(n, weighted);
		} catch (IllegalArgumentException e) {
			Logger.logError(e.getMessage());
			return null;
		}
	}
	
	/**
	 * Creates an implementation createInterpolation of {@link ConservativeInterpolation} 
	 * @param attributes  List of parameters of ConservativeInterpolation: enforced conservation, partial coverage, normalisation
 	 * the order of element attributes follows Alphabetical order when they are read from xml 
 	 * meaning attributes[0] -> enforced conservation, attributes[1] -> normalisation, attributes[2] -> partial coverage 
	 * @return a object of {@link ConservativeInterpolation}.
	 */
	private static ConservativeInterpolation createConservative(List<String[]> attributes) {
		try {
			checkAttributesSize(attributes, 3);
			
			String[] attribute1 = attributes.get(0);
			checkAttributeLength(attribute1);
			checkAttributeName(attribute1, Dom.Attribute.ENFORCED_CONSERVATION);
			
			String[] attribute2 = attributes.get(2);
			checkAttributeLength(attribute2);
			checkAttributeName(attribute2, Dom.Attribute.PARTIAL_COVERAGE);
			
			String[] attribute3 = attributes.get(1);
			checkAttributeLength(attribute3);
			checkAttributeName(attribute3, Dom.Attribute.NORMALISATION);
			
			boolean enforcedConservation = Boolean.parseBoolean(attribute1[1]);
			boolean partialCoverage = Boolean.parseBoolean(attribute2[1]);
			
			if (!Normalisation.isValidString(attribute3[1])) {
				throw new IllegalArgumentException("Normalisation String invalid");
			}
			Normalisation normalisation = Normalisation.objectFromString(attribute3[1]);
			return new ConservativeInterpolation(enforcedConservation, partialCoverage, normalisation);
		} catch (IllegalArgumentException e) {
			Logger.logError(e.getMessage());
			return null;
		}
	}
	
	/**
	 * Creates an implementation createInterpolation of {@link PatchRecoveryInterpolation} 
	 * @param attributes  List of parameters of PatchRecoveryInterpolation: allow extrapolation, extend source patch, polynomial order, gauss order
 	 * the order of element attributes follows Alphabetical order when they are read from xml
 	 * meaning attributes[0] -> allow extrapolation, attributes[1] -> extend source patch, attributes[2] -> guass order, attributes[3] -> polynomial order 
	 * @return a object of {@link PatchRecoveryInterpolation}.
	 */
	private static PatchRecoveryInterpolation createPatchRecovery(List<String[]> attributes) {
		try {
			checkAttributesSize(attributes, 4);
			
			String[] attribute1 = attributes.get(0);
			checkAttributeLength(attribute1);
			checkAttributeName(attribute1, Dom.Attribute.ALLOW_EXTRAPOLATION);
			
			String[] attribute2 = attributes.get(1);
			checkAttributeLength(attribute2);
			checkAttributeName(attribute2, Dom.Attribute.EXTEND_SOURCE_PATCH);
			checkPositiveInteger(attribute2[1]);
			
			String[] attribute3 = attributes.get(3);
			checkAttributeLength(attribute3);
			checkAttributeName(attribute3, Dom.Attribute.POLYNOMIAL_ORDER);
			checkPositiveInteger(attribute3[1]);
			
			String[] attribute4 = attributes.get(2);
			checkAttributeLength(attribute4);
			checkAttributeName(attribute4, Dom.Attribute.GAUSS_ORDER);
			checkPositiveInteger(attribute4[1]);
			
			boolean allowExtrapolation = Boolean.parseBoolean(attribute1[1]);
			int extendSourcePatch = Integer.parseInt(attribute2[1]);
			int polynomialOrder = Integer.parseInt(attribute3[1]);
			int gaussOrder = Integer.parseInt(attribute4[1]);
			
			return new PatchRecoveryInterpolation(allowExtrapolation, extendSourcePatch, polynomialOrder, gaussOrder);
		} catch (IllegalArgumentException e) {
			Logger.logError(e.getMessage());
			return null;
		}
	}
	
	/**
	 * Creates an implementation createInterpolation of {@link FixedValueInterpolation} 
	 * @param attributes  List of parameters of FixedValueInterpolation: allow extrapolation, extend source patch, polynomial order, gauss order
	 * @return a object of {@link FixedValueInterpolation}.
	 */
	private static FixedValueInterpolation createFixedValue(List<String[]> attributes) {
		try {
			checkAttributesSize(attributes, 1);
			String[] attribute = attributes.get(0);
			checkAttributeLength(attribute);
			checkAttributeName(attribute, Dom.Attribute.USER_VALUE);
			double userValue = Double.parseDouble(attribute[1]);
			return new FixedValueInterpolation(userValue);
		} catch (IllegalArgumentException e) {
			Logger.logError(e.getMessage());
			return null;
		}
	}
	
	/**
	 * Creates an implementation createInterpolation of {@link UserFileInterpolation} 
	 * @param attributes  List of parameters of UserFileInterpolation: allow extrapolation, extend source patch, polynomial order, gauss order
 	 * the order of element attributes follows Alphabetical order when they are read from xml
 	 * meaning attributes[0] -> allow extrapolation, attributes[1] -> extend source patch, attributes[2] -> guass order, attributes[3] -> polynomial order 
	 * @return a object of {@link UserFileInterpolation}.
	 */
	private static UserFileInterpolation createUserFile(List<String[]> attributes) {
		try {
			checkAttributesSize(attributes, 2);
			
			String[] attribute1 = attributes.get(0);
			checkAttributeLength(attribute1);
			checkAttributeName(attribute1, Dom.Attribute.FILENAME);
			
			String[] attribute2 = attributes.get(1);
			checkAttributeLength(attribute2);
			checkAttributeName(attribute2, Dom.Attribute.YAC_LOCATION);
			
			String filename = attribute1[1];

			if (!YacLocation.isValidString(attribute2[1])) {
				throw new IllegalArgumentException("YacLocation String is invalid");
			}
			YacLocation yacLocation = YacLocation.objectFromString(attribute2[1]);
			return new UserFileInterpolation(filename, yacLocation);
		} catch (IllegalArgumentException e) {
			Logger.logError(e.getMessage());
			return null;
		}
	}
	
	/**
	 * Creates an implementation createInterpolation of {@link Source2TargetMapInterpolation} 
	 * @param attributes  List of parameters of Source2TargetMapInterpolation: allow extrapolation, extend source patch, polynomial order, gauss order
 	 * the order of element attributes follows Alphabetical order when they are read from xml
 	 * meaning attributes[0] -> allow extrapolation, attributes[1] -> extend source patch, attributes[2] -> guass order, attributes[3] -> polynomial order 
	 * @return a object of {@link Source2TargetMapInterpolation}.
	 */
	private static Source2TargetMapInterpolation createSource2TargetMap(List<String[]> attributes) {
		try {
			checkAttributesSize(attributes, 0);
			return new Source2TargetMapInterpolation();
		} catch (IllegalArgumentException e) {
			Logger.logError(e.getMessage());
			return null;
		}
	}
	
	/**
	 * Creates an implementation createInterpolation of {@link SmoothedPatchRecoveryInterpolation} 
	 * @param attributes  List of parameters of SmoothedPatchRecoveryInterpolation: allow extrapolation, extend source patch, polynomial order, gauss order, allow corner weight, corner weight
 	 * the order of element attributes follows Alphabetical order when they are read from xml
 	 * meaning attributes[0] -> allow corner weight, attributes[1] -> allow extrapolation, attributes[2] -> corner weight,
 	 * attributes[3] -> extend source patch, attributes[4] -> gauss order, attributes[5]-> polynomial order 
	 * @return a object of {@link SmoothedPatchRecoveryInterpolation}.
	 */
	private static SmoothedPatchRecoveryInterpolation createSmoothedPatchRecovery(List<String[]> attributes) {
		try {
			checkAttributesSize(attributes, 6);
			
			String[] attribute1 = attributes.get(1);
			checkAttributeLength(attribute1);
			checkAttributeName(attribute1, Dom.Attribute.ALLOW_EXTRAPOLATION);
			
			String[] attribute2 = attributes.get(3);
			checkAttributeLength(attribute2);
			checkAttributeName(attribute2, Dom.Attribute.EXTEND_SOURCE_PATCH);
			checkPositiveInteger(attribute2[1]);
			
			String[] attribute3 = attributes.get(5);
			checkAttributeLength(attribute3);
			checkAttributeName(attribute3, Dom.Attribute.POLYNOMIAL_ORDER);
			checkPositiveInteger(attribute3[1]);
			
			String[] attribute4 = attributes.get(4);
			checkAttributeLength(attribute4);
			checkAttributeName(attribute4, Dom.Attribute.GAUSS_ORDER);
			checkPositiveInteger(attribute4[1]);
			
			String[] attribute5 = attributes.get(0);
			checkAttributeLength(attribute5);
			checkAttributeName(attribute5, Dom.Attribute.ALLOW_CORNER_WEIGHT);
			
			String[] attribute6 = attributes.get(2);
			checkAttributeLength(attribute6);
			checkAttributeName(attribute6, Dom.Attribute.CORNER_WEIGHT);
			float upperBound = 1;
			float lowerBound = 0;
			checkFloatBound(Dom.Attribute.CORNER_WEIGHT,attribute6[1],upperBound,lowerBound);
			
			boolean allowSmoothedExtrapolation = Boolean.parseBoolean(attribute1[1]);
			int extendSmoothedSourcePatch = Integer.parseInt(attribute2[1]);
			int polynomialSmoothedOrder = Integer.parseInt(attribute3[1]);
			int gaussSmoothedOrder = Integer.parseInt(attribute4[1]);
			boolean allowCornerWeight = Boolean.parseBoolean(attribute5[1]);
			float cornerWeight = Float.parseFloat(attribute6[1]);
			
			return new SmoothedPatchRecoveryInterpolation(allowSmoothedExtrapolation, extendSmoothedSourcePatch, polynomialSmoothedOrder, gaussSmoothedOrder,allowCornerWeight,cornerWeight);
		} catch (IllegalArgumentException e) {
			Logger.logError(e.getMessage());
			return null;
		}
	}
	
	/**
	 * dealt with the exception of attributes size overflow
	 * @param attributes  List of parameters of a InterpolationMethod
	 * @param expetation  the number of parameters that the InterpolationMethod requires
	 * @return Exception Message.
	 */
	private static void checkAttributesSize(List<String[]> attributes, int expectation) 
			throws IllegalArgumentException {
		if (attributes.size() != expectation) {
			throw new IllegalArgumentException("attributes must be of size " + expectation);
		}
	}
	
	/**
	 * dealt with exception of the element of attributes size overflow, supposed to be 2
	 * @param attribute   one parameter from attributes list including attribute[0] for the name and attribute[1] for the value  
	 * @param expetation  the number of parameters that the InterpolationMethod requires
	 * @return Exception Message "attribute must be of size 2".
	 */
	private static void checkAttributeLength(String[] attribute) throws IllegalArgumentException {
		if (attribute.length != 2) {
			throw new IllegalArgumentException("attribute must be of size 2");
		}
	}
	
	/**
	 * dealt with exception of attributes disorder
	 * @param attribute   one parameter from attributes list, here only check the name attribute[0] equals to expectation 
	 * @param expectation  the name of parameters that the attribute should be
	 * @return Exception Message "'attribute[0] must be' + expectation".
	 */
	private static void checkAttributeName(String[] attribute, String expectation) {
		if (!attribute[0].equals(expectation)) {
			throw new IllegalArgumentException("attribute[0] must be " + expectation);
		}
	}
	
	/**
	 * dealt with exception of negative integer 
	 * @param attribute   the value of one parameter from attributes list, attribute[1]
	 * @return Exception Message "attribute must be positive integer".
	 */
	private static void checkPositiveInteger(String attribute) {
		int posInt = Integer.parseInt(attribute);
		if (posInt < 1) {
			throw new IllegalArgumentException("attribute must be positive integer");
		}
	}
	
	/**
	 * check if a float value locals within a given bound
	 * @param name        the parameter of a float value to be check 
	 * @param attribute   the value of one parameter from attributes list, attribute[1]
	 * @param upperBound  the maximum value 
	 * @param lowerBound  the minimum value
	 * @return Exception Message "name + ' must be float between 0 and 1'".
	 */
	private static void checkFloatBound(String name, String attribute, float upperBound, float lowerBound) {
		float posFloat = Float.parseFloat(attribute);
		if (posFloat < lowerBound || posFloat > upperBound) {
			throw new IllegalArgumentException( name + " must be float between 0 and 1");
		}
	}
}
