/**
 * @file communicator_mpi.c
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *                                 Thomas Jahns <jahns@dkrz.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 *         Thomas Jahns <jahns@dkrz.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 *             Thomas Jahns <jahns@dkrz.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "communicator.h"
#include "communicator_mpi.h"
#include "yacpar.h"
#include "utils.h"
#include "ensure_array_size.h"

#ifdef USE_MPI

char datarep[10] = "external32";

/**
 * communicator_mpi internally uses its own priority to allow for an addition
 * priority for standard bsends.
 */
enum mpi_message_priority {
  MPI_MSG_PRIO_BASIC_SEND = 0,
  MPI_MSG_PRIO_HIGH = MSG_PRIO_HIGH + 1,
  MPI_MSG_PRIO_MIDDLE = MSG_PRIO_MIDDLE + 1,
  MPI_MSG_PRIO_LOW = MSG_PRIO_LOW + 1,
};

struct comm_request_mpi {
  void (*waitall)(unsigned count, struct comm_request ** requests);
  void (*free_request)(struct comm_request ** request);

  MPI_Request request;

  unsigned request_is_free;
};

struct indexed_strings {
  char const * str;
  unsigned idx;
};

struct mpi_communicator {

   struct communicator_vtable * vtable;

   int rank;
   int size;

   MPI_Comm comm;

   unsigned * message_indices;
   unsigned message_indices_size;
   unsigned num_open_messages;

   struct comm_request_mpi ** requests;
   unsigned requests_array_size;

   struct {
      char * identifier_string;
      unsigned tag;
   } * unique_tags;
   unsigned num_unique_tags;
   unsigned unique_tags_array_size;

   struct {
      char * str;
      unsigned free;
   } * registered_tags;
   unsigned max_num_registered_tags;
};

static unsigned get_comm_size_mpi(struct communicator * comm);
static unsigned get_comm_rank_mpi(struct communicator * comm);
static void comm_bsend_mpi(void const * send_buffer, unsigned count,
                           enum yac_comm_data_type data_type, unsigned dest,
                           unsigned tag, struct communicator * comm);
static void comm_isend_mpi(void const * send_buffer, unsigned count,
                           enum yac_comm_data_type data_type, unsigned dest,
                           unsigned tag, struct communicator * comm,
                           struct comm_request ** request);
static void comm_recv_callback_mpi(unsigned count,
                                   enum yac_comm_data_type data_type,
                                   unsigned source, unsigned tag,
                                   struct communicator * comm, void * user_data,
                                   func_recv_callback recv_callback,
                                   func_recv_callback_cancel recv_callback_cancel,
                                   enum yac_message_priority priority);
static void comm_recv_mpi(void * recv_buffer, unsigned * count,
                          enum yac_comm_data_type data_type, unsigned source,
                          unsigned tag, struct communicator * comm);
static void comm_irecv_mpi(void * recv_buffer, unsigned count,
                           enum yac_comm_data_type data_type, unsigned source,
                           unsigned tag, struct communicator * comm,
                           struct comm_request ** request);
static unsigned comm_wait_mpi(struct communicator * comm, unsigned tag);
static void comm_waitall_mpi(unsigned count, struct comm_request ** requests);
static void comm_cancel_recv_mpi(struct communicator * comm, unsigned tag);
static unsigned get_unique_comm_tag_mpi(struct communicator * comm,
                                        char * identifier_string);
static void get_unique_comm_tags_mpi(struct communicator * comm,
                                     char const * unique_generator_string,
                                     char const ** tag_strings,
                                     unsigned num_tag_strings, unsigned * tags);
static void comm_allgather_mpi(void * send_buffer, unsigned send_count,
                               void * recv_buffer, unsigned recv_count,
                               enum yac_comm_data_type data_type,
                               struct communicator * comm);
static void comm_allgatherv_mpi(void * send_buffer, unsigned send_count,
                                void * recv_buffer, unsigned * recv_counts,
                                enum yac_comm_data_type data_type,
                                struct communicator * comm);
static struct communicator * comm_dup_mpi(struct communicator * comm);
static struct communicator * comm_split_mpi(struct communicator * comm,
                                            int color, int key);
static struct communicator * comm_split_intercomm_mpi(
   struct communicator * comm, int group, unsigned * local_group_ranks,
   unsigned * remote_group_ranks);
static unsigned get_comm_remote_size_mpi(struct communicator * comm);
static void comm_abort_mpi(struct communicator * comm);
static void free_communicator_mpi(struct communicator * comm);
static int comm_pack_size_mpi(int count, enum yac_comm_data_type type,
                              struct communicator * comm);
static void comm_pack_mpi(void * in_buffer, int in_count,
                          enum yac_comm_data_type type, void * out_buffer,
                          int out_buffer_size, int * position,
                          struct communicator * comm);
static void comm_unpack_mpi(void * in_buffer, int in_size, int * position,
                            void * out_buffer, int out_count,
                            enum yac_comm_data_type type,
                            struct communicator * comm);
static void comm_translate_ranks_mpi(struct communicator * comm_a,
                                     struct communicator * comm_b,
                                     unsigned * ranks);
static void comm_bcast_callback_mpi(void * buffer, unsigned count,
                                    enum yac_comm_data_type data_type,
                                    unsigned root, unsigned tag,
                                    struct communicator * comm,
                                    void * user_data,
                                    func_recv_callback bcast_callback,
                                    enum yac_message_priority priority);
static void comm_gatherv_callback_mpi(void * send_buffer, unsigned * counts,
                                      enum yac_comm_data_type data_type, unsigned root,
                                      unsigned tag, struct communicator * comm,
                                      void * user_data,
                                      func_recv_callback gatherv_callback,
                                      enum yac_message_priority priority);
static void comm_allgather_callback_mpi(void * send_buffer, unsigned count,
                                        enum yac_comm_data_type data_type,
                                        unsigned tag,
                                        struct communicator * comm,
                                        void * user_data,
                                        func_recv_callback allgather_callback,
                                        enum yac_message_priority priority);
static void comm_allgatherv_callback_mpi(void * send_buffer, unsigned * counts,
                                         enum yac_comm_data_type data_type,
                                         unsigned tag,
                                         struct communicator * comm,
                                         void * user_data,
                                         func_recv_callback allgatherv_callback,
                                         enum yac_message_priority priority);
static void comm_scatter_callback_mpi(void * send_buffer, unsigned count,
                                      enum yac_comm_data_type data_type,
                                      unsigned root, unsigned tag,
                                      struct communicator * comm,
                                      void * user_data,
                                      func_recv_callback scatter_callback,
                                      enum yac_message_priority priority);
static void comm_scatterv_callback_mpi(void * send_buffer, unsigned * counts,
                                       enum yac_comm_data_type data_type,
                                       unsigned root, unsigned tag,
                                       struct communicator * comm,
                                       void * user_data,
                                       func_recv_callback scatterv_callback,
                                       enum yac_message_priority priority);
static void comm_alltoallv_callback_mpi(void * send_buffer, unsigned * counts,
                                        enum yac_comm_data_type data_type,
                                        unsigned tag, struct communicator * comm,
                                        void * user_data,
                                        func_recv_callback alltoallv_callback,
                                        enum yac_message_priority priority);
static void comm_reduce_callback_mpi(void * buffer, unsigned count,
                                     enum yac_comm_data_type data_type,
                                     func_comm_op op, unsigned root,
                                     unsigned tag, struct communicator * comm,
                                     void * user_data,
                                     func_recv_callback allreduce_callback,
                                     enum yac_message_priority priority);
static void comm_allreduce_callback_mpi(void * buffer, unsigned count,
                                        enum yac_comm_data_type data_type,
                                        func_comm_op op, unsigned tag,
                                        struct communicator * comm,
                                        void * user_data,
                                        func_recv_callback allreduce_callback,
                                        enum yac_message_priority priority);
static void comm_barrier_callback_mpi(unsigned tag, struct communicator * comm,
                                      void * user_data,
                                      func_recv_callback barrier_callback,
                                      enum yac_message_priority priority);
static void comm_check_mpi(struct communicator * comm);
static void comm_finalize_mpi();

static struct communicator_vtable
   mpi_communicator_vtable = {
      .free                = free_communicator_mpi,
      .get_size            = get_comm_size_mpi,
      .get_rank            = get_comm_rank_mpi,
      .bsend               = comm_bsend_mpi,
      .isend               = comm_isend_mpi,
      .recv_callback       = comm_recv_callback_mpi,
      .recv                = comm_recv_mpi,
      .irecv               = comm_irecv_mpi,
      .wait                = comm_wait_mpi,
      .cancel_recv         = comm_cancel_recv_mpi,
      .allgather           = comm_allgather_mpi,
      .allgatherv          = comm_allgatherv_mpi,
      .dup                 = comm_dup_mpi,
      .split               = comm_split_mpi,
      .split_intercomm     = comm_split_intercomm_mpi,
      .get_remote_size     = get_comm_remote_size_mpi,
      .abort               = comm_abort_mpi,
      .get_unique_tag      = get_unique_comm_tag_mpi,
      .get_unique_tags     = get_unique_comm_tags_mpi,
      .pack_size           = comm_pack_size_mpi,
      .pack                = comm_pack_mpi,
      .unpack              = comm_unpack_mpi,
      .translate_ranks     = comm_translate_ranks_mpi,
      .bcast_callback      = comm_bcast_callback_mpi,
      .gatherv_callback    = comm_gatherv_callback_mpi,
      .allgather_callback  = comm_allgather_callback_mpi,
      .allgatherv_callback = comm_allgatherv_callback_mpi,
      .scatter_callback    = comm_scatter_callback_mpi,
      .scatterv_callback   = comm_scatterv_callback_mpi,
      .alltoallv_callback  = comm_alltoallv_callback_mpi,
      .reduce_callback     = comm_reduce_callback_mpi,
      .allreduce_callback  = comm_allreduce_callback_mpi,
      .barrier_callback    = comm_barrier_callback_mpi,
      .check               = comm_check_mpi
   };

struct mpi_communicator_message {

   unsigned message_is_free;

   enum {
     SEND,
     RECV,
   } message_type;

   union {
      struct {
         func_recv_callback func;
         func_recv_callback_cancel func_cancel;
      } recv;
      struct {
        func_send_callback func;
      } send;
   } callback;

   void * user_data;
   struct comm_message message;
   unsigned free_buffer;

   struct communicator * comm;

   enum mpi_message_priority priority;
   unsigned request_idx;
};

struct temp_bcast_data {

  union {
    unsigned root;
    unsigned send_count;
  } data;
  struct comm_message msg;
  enum mpi_message_priority priority;
  func_recv_callback callback;
  void * user_data;
};

struct temp_gatherv_data {

  unsigned * counts;
  unsigned root;
  void * buffer, * buffer_p;
  void * user_data;
  func_recv_callback gatherv_callback;
  enum mpi_message_priority priority;
};

struct temp_allgather_data {
  unsigned count;
  enum mpi_message_priority priority;
  enum yac_comm_data_type data_type;
  unsigned step;
  void * buffer, * buffer_p_recv, * buffer_p_send;
  void * user_data;
  func_recv_callback allgather_callback;
};

struct temp_allgatherv_data {
  unsigned * counts;
  enum mpi_message_priority priority;
  enum yac_comm_data_type data_type;
  unsigned step;
  void * buffer, * buffer_p_recv, * buffer_p_send;
  void * user_data;
  func_recv_callback allgatherv_callback;
};

struct temp_scatter_data {
  unsigned count;
  unsigned root;
  void * user_data;
  enum mpi_message_priority priority;
  func_recv_callback scatter_callback;
};

struct temp_scatterv_data {
  unsigned * counts;
  unsigned root;
  void * user_data;
  enum mpi_message_priority priority;
  func_recv_callback scatterv_callback;
};

struct temp_alltoallv_data {

  unsigned num_ranks; // number of logical ranks (always a power of two and >=
                      //                          comm_size)
  unsigned local_rank;
  unsigned dst_rank_mask;

  void * buffer, * send_buffer, *tmp_buffer;
  unsigned max_buffer_size, max_send_buffer_size, max_tmp_buffer_size;
  unsigned * data_size_per_rank, * send_data_size_per_rank,
           * msg_data_size_per_rank, * tmp_data_size_per_rank;

  enum yac_comm_data_type data_type;
  enum mpi_message_priority priority;

  func_recv_callback alltoallv_callback;
  void * user_data;

  unsigned send_count;
};

struct temp_reduce_data {
  func_comm_op op;
  unsigned root;
  enum mpi_message_priority priority;
  void * buffer;
  void * user_data;
  func_recv_callback reduce_callback;
};

struct temp_allreduce_data {
  func_comm_op op;
  unsigned rank_mask;
  unsigned local_rank;
  enum mpi_message_priority priority;
  void * buffer;
  void * user_data;
  func_recv_callback allreduce_callback;
};

struct temp_barrier_data {

  unsigned step;

  enum mpi_message_priority priority;

  func_recv_callback barrier_callback;
  void * user_data;

  unsigned send_count;
};

//! global requests of all priorities
static MPI_Request * global_requests_ = NULL;
static unsigned * global_message_ids_ = NULL;
static unsigned num_global_requests_ = 0;

static struct {

   MPI_Request * requests;
   unsigned * global_message_idx; // -1 == free request
   unsigned num_requests;

} * global_requests = NULL;

static unsigned max_priority = -1;

static unsigned comm_wait_is_active = 0;

static struct mpi_communicator_message * global_messages = NULL;
static unsigned num_global_messages = 0;

static int mpi_tag_ub = 32767; // minimum defined by MPI standard

static int * msg_prio_stack = NULL;
static int msg_prio_stack_size = 0;
static int curr_msg_prio_stack_pos = -1;

static int yac_initialised_MPI = 0;

static void ** bsend_buffers = NULL;
static size_t * bsend_buffer_sizes = NULL;
static unsigned ** buffer_is_free = NULL;
static unsigned curr_num_bsend_buffers = 0;
static unsigned max_num_bsend_buffers = 0;
static size_t total_bsend_buffer_size = 0;
#define NUM_BSEND_BUFFERS_TO_FREE (64)
#define TOTAL_BSEND_BUFFER_SIZE_TO_FREE (8 * 1024 * 1024)

//taken from http://beige.ucs.indiana.edu/I590/node85.html
static void mpi_err_handler(int error_code, MPI_Comm comm) {

   if (error_code != MPI_SUCCESS) {

      int rank;

      MPI_Comm_rank(comm, &rank);

      char error_string[1024];
      int length_of_error_string, error_class;

      MPI_Error_class(error_code, &error_class);
      MPI_Error_string(error_class, error_string, &length_of_error_string);
      fprintf(stderr, "%3d: %s\n", rank, error_string);
      MPI_Error_string(error_code, error_string, &length_of_error_string);
      fprintf(stderr, "%3d: %s\n", rank, error_string);
      MPI_Abort(comm, error_code);
   }
}

static inline enum mpi_message_priority
convert_msg_prio(enum yac_message_priority prio) {

  return prio + 1;
}

static void init_mpi_env() {

   int is_initialized;

   mpi_err_handler(MPI_Initialized(&is_initialized), MPI_COMM_WORLD);

   if (!is_initialized) {

      yac_initialised_MPI = 1;

      mpi_err_handler(MPI_Init(NULL, NULL), MPI_COMM_WORLD);

      void *attr;
      int  flag;

      mpi_err_handler(MPI_Comm_get_attr(MPI_COMM_WORLD, MPI_TAG_UB, &attr,
                                        &flag), MPI_COMM_WORLD);
      if (!flag)
        yac_internal_abort_message("ERROR: Could not retrieve MPI attribute MPI_TAG_UB.",
                                   __FILE__, __LINE__);

      mpi_tag_ub = *(int*)attr;
      if (mpi_tag_ub < 32767)
        yac_internal_abort_message("ERROR: MPI attribute MPI_TAG_UB is invalid.",
                                   __FILE__, __LINE__);
   }

   yac_add_comm_finalize_callback(comm_finalize_mpi);
}

static struct communicator * mpi_comm_new_(MPI_Comm mpi_comm) {

   struct mpi_communicator * comm = malloc(1 * sizeof(*comm));

   comm->vtable = &mpi_communicator_vtable;
   comm->comm = mpi_comm;
   comm->message_indices = NULL;
   comm->message_indices_size = 0;
   comm->num_open_messages = 0;
   if (mpi_comm != MPI_COMM_NULL) {
      mpi_err_handler(MPI_Comm_rank(mpi_comm, &(comm->rank)), mpi_comm);
      mpi_err_handler(MPI_Comm_size(mpi_comm, &(comm->size)), mpi_comm);
   } else {
      comm->rank = -1;
      comm->size = -1;
   }

   comm->requests = NULL;
   comm->requests_array_size = 0;

   comm->unique_tags = NULL;
   comm->num_unique_tags = 0;
   comm->unique_tags_array_size = 0;

   comm->registered_tags = NULL;
   comm->max_num_registered_tags = 0;

   return (struct communicator *)comm;
}

struct communicator * yac_mpi_comm_new() {

   init_mpi_env();

   MPI_Comm comm_world_copy;

   mpi_err_handler(MPI_Comm_dup(MPI_COMM_WORLD, &comm_world_copy),
                   MPI_COMM_WORLD);

   return mpi_comm_new_(comm_world_copy);
}

static void free_mpi_message(unsigned message_idx) {

   struct mpi_communicator * mpi_comm =
      (struct mpi_communicator*)(global_messages[message_idx].comm);

   if (global_messages[message_idx].free_buffer)
      free(global_messages[message_idx].message.buffer);

   for (unsigned i = 0; i < mpi_comm->message_indices_size; ++i) {
      if (mpi_comm->message_indices[i] == message_idx) {
         mpi_comm->message_indices[i] = -1;
         break;
      }
   }
   mpi_comm->num_open_messages--;
   global_messages[message_idx].comm = NULL;
   global_messages[message_idx].message_is_free = 1 == 1;
   global_messages[message_idx].priority = -1;
   global_messages[message_idx].request_idx = -1;
}

static MPI_Datatype get_mpi_datatype(enum yac_comm_data_type type) {

   switch (type) {

      case(COMM_DBLE):
         return MPI_DOUBLE;
      case(COMM_UINT):
         return MPI_UNSIGNED;
      case(COMM_INT):
         return MPI_INT;
      case(COMM_BYTE):
         return MPI_BYTE;
      case(COMM_PACKED):
         return MPI_PACKED;
      default:
         return MPI_DATATYPE_NULL;
   };
}

static void deliver_callback_message(unsigned message_idx,
                                     MPI_Status * status) {

   struct mpi_communicator_message * mpi_message =
      global_messages + message_idx;
   struct mpi_communicator * mpi_comm =
      (struct mpi_communicator *)mpi_message->comm;

   switch (mpi_message->message_type) {

      case (RECV):
         mpi_err_handler(
            MPI_Get_count(
               status, get_mpi_datatype(mpi_message->message.type),
               (int*)&(mpi_message->message.count)), mpi_comm->comm);

         if (mpi_message->message.src_rank == (unsigned)COMM_ANYSOURCE)
            mpi_message->message.src_rank = status->MPI_SOURCE;

         if (mpi_message->callback.recv.func != NULL)
            mpi_message->callback.recv.func(
               mpi_message->comm, mpi_message->message, mpi_message->user_data);
         break;
      case (SEND):
         if (mpi_message->callback.send.func != NULL)
            mpi_message->callback.send.func(
               mpi_message->comm, mpi_message->message, mpi_message->user_data);
         break;
      default:
         yac_internal_abort_message(
            "ERROR(deliver_callback_message): Invalid message type.",
            __FILE__, __LINE__);
   };

   free_mpi_message(message_idx);
}

static inline void push_msg_prio_stack(int prio) {

   if ((curr_msg_prio_stack_pos == -1) ||
       (curr_msg_prio_stack_pos + 1 >= msg_prio_stack_size)) {

      msg_prio_stack_size = curr_msg_prio_stack_pos + 2;

      msg_prio_stack = realloc(msg_prio_stack, msg_prio_stack_size *
                               sizeof(*msg_prio_stack));
   }

   ++curr_msg_prio_stack_pos;
   msg_prio_stack[curr_msg_prio_stack_pos] = prio;
}

static inline int top_msg_prio_stack() {

   if (curr_msg_prio_stack_pos == -1)
      return -1;
   else
      return msg_prio_stack[curr_msg_prio_stack_pos];
}

static inline void pop_msg_prio_stack() {

   --curr_msg_prio_stack_pos;
}

static unsigned check_open_requests(struct mpi_communicator * mpi_comm) {

   if (max_priority == (unsigned)-1) return 0;

   MPI_Status status;
   int flag, index;

   int max_i;

   // if wait is active
   if (comm_wait_is_active) {
      max_i = top_msg_prio_stack();

      // if there isnt an open request being handled
      if (max_i == -1)
         max_i = max_priority;
      else
         // allow only requests with higher priority to intercept the current one
         --max_i;
   } else {
     // check only the requests of the basic bsends
     max_i = MPI_MSG_PRIO_BASIC_SEND;
   }

   unsigned delivered_a_message = 0;

   for (int i = 0; i <= max_i; ++i) {

      if (global_requests[i].num_requests == 0) continue;

      mpi_err_handler(MPI_Testany(global_requests[i].num_requests,
                                  global_requests[i].requests, &index, &flag,
                                  &status), mpi_comm->comm);

      if (flag && index != MPI_UNDEFINED) {

         unsigned message_idx = global_requests[i].global_message_idx[index];
         global_requests[i].global_message_idx[index] = -1;

         push_msg_prio_stack(i);
         deliver_callback_message(message_idx, &status);
         pop_msg_prio_stack();
         delivered_a_message = 1;

         i = -1;
      }
   }

   return delivered_a_message;
}

static int waitany_open_request_(MPI_Comm comm) {

  MPI_Status status;
  int flag = 0, index = MPI_UNDEFINED;

  // first check the send operations
  if ((max_priority != (unsigned)-1) &&
      (max_priority >= MPI_MSG_PRIO_BASIC_SEND) &&
      (global_requests[MPI_MSG_PRIO_BASIC_SEND].num_requests > 0))
    mpi_err_handler(
      MPI_Testany(global_requests[MPI_MSG_PRIO_BASIC_SEND].num_requests,
      global_requests[MPI_MSG_PRIO_BASIC_SEND].requests, &index, &flag,
      &status), comm);

  unsigned message_idx;
  enum mpi_message_priority priority;
  if (flag && index != MPI_UNDEFINED) {

    message_idx =
      global_requests[MPI_MSG_PRIO_BASIC_SEND].global_message_idx[index];
    global_requests[MPI_MSG_PRIO_BASIC_SEND].global_message_idx[index] = -1;
    priority = MPI_MSG_PRIO_BASIC_SEND;

  // if no valid index was returned by MPI_Testany
  } else {
    mpi_err_handler(MPI_Waitany(num_global_requests_, global_requests_,
                                &index, &status), comm);

    if (index == MPI_UNDEFINED) return 0;

    message_idx = global_message_ids_[index];
    global_message_ids_[index] = -1;
    priority = global_messages[message_idx].priority;
  }

  push_msg_prio_stack((int)priority);
  deliver_callback_message(message_idx, &status);
  pop_msg_prio_stack();

  return 1;
}

static void waitany_open_request(struct mpi_communicator * comm) {

   struct mpi_communicator * mpi_comm = (struct mpi_communicator *)comm;

   unsigned curr_num_open_messages = mpi_comm->num_open_messages;

   if (curr_num_open_messages == 0)
      return;

   while (curr_num_open_messages == mpi_comm->num_open_messages) {

      // if there are no open send or receives
      if (!waitany_open_request_(mpi_comm->comm))
        yac_internal_abort_message(
          "ERROR(waitany_open_request): no open request", __FILE__, __LINE__);
   }
}

static unsigned get_comm_size_mpi(struct communicator * comm) {

   return ((struct mpi_communicator*)comm)->size;
}

static unsigned get_comm_rank_mpi(struct communicator * comm) {

   return ((struct mpi_communicator*)comm)->rank;
}

MPI_Comm yac_get_comm_mpi(struct communicator * comm) {

   return ((struct mpi_communicator*)comm)->comm;
}

static unsigned get_data_type_size(enum yac_comm_data_type type) {

   switch (type) {

      case(COMM_DBLE):
         return sizeof(double);
      case(COMM_UINT):
         return sizeof(unsigned);
      case(COMM_INT):
         return sizeof(int);
      case(COMM_BYTE):
      case(COMM_PACKED):
         return 1;
      default:
         return -1;
   };
}

static unsigned get_free_request_handle(unsigned global_message_idx,
                                        enum mpi_message_priority priority) {

   unsigned const inc_size = 128;

   unsigned p = (unsigned)priority;

   if ((p > max_priority) || (max_priority == (unsigned)-1)) {

      global_requests = realloc(global_requests,
                                (p + 1) * sizeof(*global_requests));
      for (unsigned i = max_priority + 1; i <= p; ++i) {
         global_requests[i].requests = NULL;
         global_requests[i].global_message_idx = NULL;
         global_requests[i].num_requests = 0;
      }
      max_priority = p;
   }

   unsigned i;

   for (i = 0; i < global_requests[p].num_requests; ++i)
      if (global_requests[p].global_message_idx[i] == (unsigned)-1) break;

   if (i == global_requests[p].num_requests) {

      num_global_requests_ += inc_size;

      global_requests_ = realloc(global_requests_, num_global_requests_ *
                                 sizeof(*global_requests_));
      global_message_ids_ = realloc(global_message_ids_, num_global_requests_ *
                                    sizeof(*global_message_ids_));

      unsigned offset = 0;
      unsigned copy_offset = 0;
      unsigned copy_size = 0;

      for (unsigned j = 0; j <= max_priority; ++j) {

         if ((global_requests[j].num_requests == 0) && (j != p)) continue;

         global_requests[j].requests = global_requests_ + offset;
         global_requests[j].global_message_idx = global_message_ids_ + offset;

         offset += global_requests[j].num_requests;

         if (j == p) {
            copy_offset = offset;
            offset += inc_size;
         }

         if (j > p)
            copy_size += global_requests[j].num_requests;
      }

      if (copy_size > 0) {

         memmove(global_requests_ + copy_offset + inc_size,
                 global_requests_ + copy_offset,
                 copy_size * sizeof(*global_requests_));
         memmove(global_message_ids_ + copy_offset + inc_size,
                 global_message_ids_ + copy_offset,
                 copy_size * sizeof(*global_message_ids_));
      }

      unsigned new_num_requests = global_requests[p].num_requests + inc_size;

      for (unsigned j = global_requests[p].num_requests;
           j < new_num_requests; ++j) {
         global_requests[p].requests[j] = MPI_REQUEST_NULL;
         global_requests[p].global_message_idx[j] = (unsigned)-1;
      }
      global_requests[p].num_requests = new_num_requests;
   }

   global_requests[p].global_message_idx[i] = global_message_idx;

   return i;
}

static unsigned get_free_message_handle() {

   unsigned i;

   for (i = 0; i < num_global_messages; ++i)
      if (global_messages[i].message_is_free) break;

   if (i == num_global_messages) {

      unsigned new_global_message_size = num_global_messages + 128;

      global_messages = realloc(global_messages,
                                new_global_message_size *
                                sizeof(*global_messages));

      for (; num_global_messages < new_global_message_size;
           ++num_global_messages) {
         global_messages[num_global_messages].message_is_free = 1 == 1;
      }
   }

   global_messages[i].message_is_free = 1 == 0;

   return i;
}

static void register_message_idx(struct mpi_communicator * mpi_comm,
                                 unsigned message_idx) {

   unsigned i;

   for (i = 0; i < mpi_comm->message_indices_size; ++i)
      if (mpi_comm->message_indices[i] == -1)
         break;

   if (i == mpi_comm->message_indices_size) {

      mpi_comm->message_indices_size += 128;

      mpi_comm->message_indices = realloc(mpi_comm->message_indices,
                                          mpi_comm->message_indices_size *
                                          sizeof(*(mpi_comm->message_indices)));

      for (unsigned j = i + 1; j < mpi_comm->message_indices_size; ++j)
         mpi_comm->message_indices[j] = -1;
   }

   mpi_comm->message_indices[i] = message_idx;
   mpi_comm->num_open_messages++;
}

static void send_internal(void * buffer, unsigned count,
                          enum yac_comm_data_type data_type, unsigned dest,
                          unsigned tag, struct communicator * comm,
                          void * user_data, func_send_callback send_callback,
                          enum mpi_message_priority priority,
                          unsigned free_buffer) {

   struct mpi_communicator * mpi_comm = (struct mpi_communicator *)comm;

   // get free request handle
   unsigned message_idx = get_free_message_handle();
   unsigned request_idx = get_free_request_handle(message_idx, priority);

   global_messages[message_idx].message_type = SEND;

   global_messages[message_idx].callback.send.func = send_callback;

   global_messages[message_idx].user_data = user_data;
   global_messages[message_idx].message.buffer = buffer;
   global_messages[message_idx].message.type = data_type;
   global_messages[message_idx].message.count = count;
   global_messages[message_idx].message.tag = tag;
   global_messages[message_idx].message.dst_rank = dest;
   global_messages[message_idx].message.src_rank = mpi_comm->rank;
   global_messages[message_idx].free_buffer = free_buffer;

   global_messages[message_idx].comm = comm;

   global_messages[message_idx].request_idx = request_idx;
   global_messages[message_idx].priority = priority;

   register_message_idx(mpi_comm, message_idx);

   mpi_err_handler(MPI_Isend(buffer, (int)count, get_mpi_datatype(data_type),
                             dest, tag, mpi_comm->comm,
                             global_requests[priority].requests +
                             request_idx), mpi_comm->comm);

   check_open_requests(mpi_comm);
}

static void free_bsend_buffers() {

  unsigned register i = 0;

  while (i < curr_num_bsend_buffers) {

    if (!*(buffer_is_free[i])) {
      i++;
      continue;
    }

    free(bsend_buffers[i]);
    total_bsend_buffer_size -= bsend_buffer_sizes[i];
    curr_num_bsend_buffers--;

    bsend_buffers[i] = bsend_buffers[curr_num_bsend_buffers];
    bsend_buffer_sizes[i] = bsend_buffer_sizes[curr_num_bsend_buffers];
    unsigned * tmp_buffer_is_free = buffer_is_free[i];
    buffer_is_free[i] = buffer_is_free[curr_num_bsend_buffers];
    buffer_is_free[curr_num_bsend_buffers] = tmp_buffer_is_free;
  }
}

static unsigned get_bsend_buffer(size_t buffer_size) {

  unsigned register buffer_idx;

  // try to reuse a free bsend buffer
  for (buffer_idx = 0; buffer_idx < curr_num_bsend_buffers; ++buffer_idx)
    if (*(buffer_is_free[buffer_idx]) &&
        (bsend_buffer_sizes[buffer_idx] == buffer_size)) break;

  // if we have to allocate a new buffer
  if (buffer_idx == curr_num_bsend_buffers) {

    if (curr_num_bsend_buffers == max_num_bsend_buffers) {

      max_num_bsend_buffers += 32;
      bsend_buffers = realloc(bsend_buffers,
                              max_num_bsend_buffers * sizeof(*bsend_buffers));
      bsend_buffer_sizes = realloc(bsend_buffer_sizes,
                                   max_num_bsend_buffers *
                                   sizeof(*bsend_buffer_sizes));
      buffer_is_free = realloc(buffer_is_free,
                               max_num_bsend_buffers * sizeof(*buffer_is_free));

      for (unsigned i = curr_num_bsend_buffers; i < max_num_bsend_buffers; ++i){
        bsend_buffers[i] = NULL;
        bsend_buffer_sizes[i] = 0;
        buffer_is_free[i] = malloc(sizeof(**buffer_is_free));
        *(buffer_is_free[i]) = 1;
      }
    }

    bsend_buffers[curr_num_bsend_buffers] = malloc(buffer_size);
    bsend_buffer_sizes[curr_num_bsend_buffers] = buffer_size;
    curr_num_bsend_buffers++;
    total_bsend_buffer_size += buffer_size;
  }

  *(buffer_is_free[buffer_idx]) = 0;

  return buffer_idx;
}

static void bsend_callback(struct communicator * comm, struct comm_message msg,
                           void * buffer_is_free) {

  *((unsigned*)buffer_is_free) = 1;
}

static void comm_bsend_mpi(void const * send_buffer, unsigned count,
                           enum yac_comm_data_type data_type, unsigned dest,
                           unsigned tag, struct communicator * comm) {

   struct mpi_communicator * mpi_comm = (struct mpi_communicator *)comm;

   size_t buffer_size = (size_t)count * (size_t)get_data_type_size(data_type);

   unsigned buffer_idx = get_bsend_buffer(buffer_size);
   void * buffer = bsend_buffers[buffer_idx];
   memcpy(buffer, send_buffer, buffer_size);

   send_internal(buffer, count, data_type, dest, tag, comm,
                 buffer_is_free[buffer_idx], bsend_callback,
                 MPI_MSG_PRIO_BASIC_SEND, 0);

  // if we should free some buffers
  if ((curr_num_bsend_buffers > NUM_BSEND_BUFFERS_TO_FREE) ||
      (total_bsend_buffer_size > TOTAL_BSEND_BUFFER_SIZE_TO_FREE))
    free_bsend_buffers();
}

static void free_comm_request_mpi(struct comm_request ** request) {

  struct comm_request_mpi * request_mpi = (struct comm_request_mpi *)*request;

  *request = COMM_REQUEST_NULL;

  if (request_mpi->request != MPI_REQUEST_NULL)
    mpi_err_handler(MPI_Request_free(&(request_mpi->request)), MPI_COMM_WORLD);

  request_mpi->request_is_free = 1;
}

struct comm_request_mpi *
get_free_comm_request_mpi(struct mpi_communicator * comm_mpi) {

  unsigned request_idx;
  for (request_idx = 0; request_idx < comm_mpi->requests_array_size;
       ++request_idx)
    if (((struct comm_request*)(comm_mpi->requests[request_idx]) ==
         COMM_REQUEST_NULL) ||
        (comm_mpi->requests[request_idx]->request_is_free))
      break;

  if (request_idx == comm_mpi->requests_array_size) {

    unsigned tmp_requests_array_size = comm_mpi->requests_array_size;
    comm_mpi->requests_array_size += 128;
    comm_mpi->requests = realloc(comm_mpi->requests,
                                 comm_mpi->requests_array_size *
                                 sizeof(*(comm_mpi->requests)));
    for (unsigned i = tmp_requests_array_size;
         i < comm_mpi->requests_array_size; ++i)
      comm_mpi->requests[i] = (struct comm_request_mpi*)COMM_REQUEST_NULL;
  }

  struct comm_request_mpi * request = comm_mpi->requests[request_idx];
  if ((struct comm_request*)(comm_mpi->requests[request_idx]) ==
      COMM_REQUEST_NULL) {
    comm_mpi->requests[request_idx] =
      malloc(1 * sizeof(**(comm_mpi->requests)));
    comm_mpi->requests[request_idx]->waitall = comm_waitall_mpi;
    comm_mpi->requests[request_idx]->free_request = free_comm_request_mpi;
  }

  comm_mpi->requests[request_idx]->request = MPI_REQUEST_NULL;
  comm_mpi->requests[request_idx]->request_is_free = 0;

  return comm_mpi->requests[request_idx];
}

static void comm_isend_mpi(void const * send_buffer, unsigned count,
                           enum yac_comm_data_type data_type, unsigned dest,
                           unsigned tag, struct communicator * comm,
                           struct comm_request ** request) {

  struct mpi_communicator * mpi_comm = (struct mpi_communicator *)comm;
  struct comm_request_mpi * request_mpi = get_free_comm_request_mpi(mpi_comm);

  *request = (struct comm_request*)request_mpi;

  mpi_err_handler(MPI_Isend((void*)send_buffer, count,
                            get_mpi_datatype(data_type), dest, tag,
                            mpi_comm->comm, &(request_mpi->request)),
                  mpi_comm->comm);
}

static void recv_internal(unsigned count, enum yac_comm_data_type data_type,
                          unsigned source, unsigned tag,
                          struct communicator * comm, void * user_data,
                          func_recv_callback recv_callback,
                          func_recv_callback_cancel recv_callback_cancel,
                          enum mpi_message_priority priority,
                          unsigned free_buffer) {

   struct mpi_communicator * mpi_comm = (struct mpi_communicator *)comm;

   unsigned message_idx;
   unsigned request_idx;

   // allocate memory for message
   void * buffer =
      malloc((size_t)count * (size_t)get_data_type_size(data_type));

   // get free request handle
   message_idx = get_free_message_handle();
   request_idx = get_free_request_handle(message_idx, priority);

   global_messages[message_idx].message_type = RECV;
   global_messages[message_idx].callback.recv.func = recv_callback;
   global_messages[message_idx].callback.recv.func_cancel =
      recv_callback_cancel;
   global_messages[message_idx].user_data = user_data;

   global_messages[message_idx].message.buffer = buffer;
   global_messages[message_idx].message.type = data_type;
   global_messages[message_idx].message.count = count;
   global_messages[message_idx].message.tag = tag;
   global_messages[message_idx].message.dst_rank = mpi_comm->rank;
   global_messages[message_idx].message.src_rank = source;
   global_messages[message_idx].free_buffer = free_buffer;

   global_messages[message_idx].comm = comm;

   global_messages[message_idx].request_idx = request_idx;
   global_messages[message_idx].priority = priority;

   register_message_idx(mpi_comm, message_idx);

   if (source == (unsigned)COMM_ANYSOURCE) source = (unsigned)MPI_ANY_SOURCE;


   // set up irecv
   mpi_err_handler(MPI_Irecv(buffer, count, get_mpi_datatype(data_type), source,
                             tag, mpi_comm->comm,
                             global_requests[priority].requests+request_idx),
                             mpi_comm->comm);

   check_open_requests((struct mpi_communicator *)comm);
}

static void comm_recv_callback_mpi(unsigned count, enum yac_comm_data_type data_type,
                                   unsigned source, unsigned tag,
                                   struct communicator * comm, void * user_data,
                                   func_recv_callback recv_callback,
                                   func_recv_callback_cancel recv_callback_cancel,
                                   enum yac_message_priority priority) {

   recv_internal(count, data_type, source, tag, comm, user_data, recv_callback,
                 recv_callback_cancel, convert_msg_prio(priority), 1);
}

static void comm_recv_mpi(void * recv_buffer, unsigned * count,
                          enum yac_comm_data_type data_type, unsigned source,
                          unsigned tag, struct communicator * comm) {

   struct mpi_communicator * mpi_comm = (struct mpi_communicator *)comm;

   MPI_Status status;

   if (source == (unsigned)COMM_ANYSOURCE)
      source = MPI_ANY_SOURCE;

   mpi_err_handler(MPI_Recv(recv_buffer, *count, get_mpi_datatype(data_type),
                            source, tag, mpi_comm->comm, &status),
                   mpi_comm->comm);

   int received_count;

   mpi_err_handler(MPI_Get_count(&status, get_mpi_datatype(data_type),
                   &received_count), mpi_comm->comm);

   if (received_count == MPI_UNDEFINED)
      yac_internal_abort_message ("ERROR: Amount of received data does not match.",
                                  __FILE__, __LINE__);

   *count = (unsigned)received_count;

   check_open_requests((struct mpi_communicator *)comm);
}

static void comm_irecv_mpi(void * recv_buffer, unsigned count,
                           enum yac_comm_data_type data_type, unsigned source,
                           unsigned tag, struct communicator * comm,
                           struct comm_request ** request) {

  struct mpi_communicator * mpi_comm = (struct mpi_communicator *)comm;
  struct comm_request_mpi * request_mpi = get_free_comm_request_mpi(mpi_comm);

  *request = (struct comm_request*)request_mpi;

  mpi_err_handler(MPI_Irecv(recv_buffer, count, get_mpi_datatype(data_type),
                            source, tag, mpi_comm->comm,
                            &(request_mpi->request)), mpi_comm->comm);
}

static unsigned comm_wait_mpi(struct communicator * comm, unsigned tag) {

   struct mpi_communicator * mpi_comm = (struct mpi_communicator *)comm;

   if (mpi_comm->num_open_messages == 0) return 0;

   unsigned comm_wait_is_active_buffer = comm_wait_is_active;

   comm_wait_is_active = 1;

   unsigned tag_flag, msg_flag = 0;

   do {

      tag_flag = 0;

      for (unsigned i = 0; i < mpi_comm->message_indices_size; ++i) {

         unsigned msg_index = mpi_comm->message_indices[i];

         if (msg_index != -1) {

            if (global_messages[msg_index].message_is_free)
               yac_internal_abort_message("ERROR: invalid message index\n",
                                          __FILE__, __LINE__);

            if (global_messages[msg_index].message.tag == tag) {
               waitany_open_request((struct mpi_communicator *)comm);
               tag_flag = 1;
               msg_flag = 1;
               break;
            }
         }
      }

   } while (tag_flag);

   msg_flag |= check_open_requests((struct mpi_communicator *)comm);

   comm_wait_is_active = comm_wait_is_active_buffer;

   return msg_flag;
}

static void comm_waitall_mpi(unsigned count, struct comm_request ** requests) {

  struct comm_request_mpi ** requests_mpi =
    (struct comm_request_mpi **)requests;

  MPI_Request * array_of_requests = malloc(count * sizeof(*array_of_requests));

  unsigned valid_count = 0;
  for (unsigned i = 0; i < count; ++i)
    if (requests[i] != COMM_REQUEST_NULL)
      array_of_requests[valid_count++] = requests_mpi[i]->request;

  mpi_err_handler(MPI_Waitall(valid_count, array_of_requests,
                              MPI_STATUSES_IGNORE), MPI_COMM_WORLD);

  free(array_of_requests);

  for (unsigned i = 0; i < count; ++i) {
    if (requests[i] == COMM_REQUEST_NULL) continue;
    requests_mpi[i]->request = MPI_REQUEST_NULL;
    requests_mpi[i]->request_is_free = 1;
  }
}

static void cancel_recv_callback_message(unsigned message_idx) {

   struct mpi_communicator * mpi_comm =
      (struct mpi_communicator *)(global_messages[message_idx].comm);

   unsigned priority = (unsigned)(global_messages[message_idx].priority);
   unsigned request_idx = global_messages[message_idx].request_idx;

   if (global_messages[message_idx].message_type != RECV)
      yac_internal_abort_message(
         "ERROR(cancel_recv_callback_message): Invalid message type.",
         __FILE__, __LINE__);

   mpi_err_handler(MPI_Cancel(global_requests[priority].requests +
                              request_idx), mpi_comm->comm);
   mpi_err_handler(MPI_Request_free(global_requests[priority].requests +
                                    request_idx), mpi_comm->comm);
   global_requests[priority].global_message_idx[request_idx] = -1;

   if (global_messages[message_idx].callback.recv.func_cancel != NULL)
      global_messages[message_idx].callback.recv.func_cancel(
         global_messages[message_idx].comm,
         global_messages[message_idx].user_data);

   free_mpi_message(message_idx);
}

static void comm_cancel_recv_mpi(struct communicator * comm, unsigned tag) {

   struct mpi_communicator * mpi_comm = (struct mpi_communicator *)comm;

   for (unsigned i = 0; i < mpi_comm->message_indices_size; ++i) {

      unsigned msg_index = mpi_comm->message_indices[i];

      // if this is a valid message and
      // if this is not a free message and
      // if this is a recv message and
      // if this message has the required tag
      if ((msg_index != -1) &&
          (!global_messages[msg_index].message_is_free) &&
          (global_messages[msg_index].message_type == RECV) &&
          (global_messages[msg_index].message.tag == tag))
         cancel_recv_callback_message(msg_index);
   }

   check_open_requests((struct mpi_communicator *)comm);
}

static unsigned get_unique_comm_tag_mpi(struct communicator * comm,
                                        char * identifier_string) {

   struct mpi_communicator * mpi_comm = (struct mpi_communicator *)comm;

   unsigned i, j, tag;

   if (identifier_string != NULL) {

      for (i = 0; i < mpi_comm->num_unique_tags; ++i)
         if (mpi_comm->unique_tags[i].tag != -1 &&
             mpi_comm->unique_tags[i].identifier_string != NULL &&
             !strcmp(identifier_string, mpi_comm->unique_tags[i].identifier_string))
            break;

      if (i != mpi_comm->num_unique_tags)
         return mpi_comm->unique_tags[i].tag;

      tag = yac_hash(identifier_string)%(mpi_tag_ub+1);

      for (i = 0; i < mpi_comm->num_unique_tags; ++i)
         if (mpi_comm->unique_tags[i].tag == tag)
            break;

      if (i != mpi_comm->num_unique_tags) {
         yac_internal_abort_message("ERROR: tag is not unique\n", __FILE__, __LINE__);
         return -1;
      }

      if (tag == -1) {
         yac_internal_abort_message("ERROR: generate tag was invalid\n", __FILE__, __LINE__);
         return -1;
      }

   } else {

      for (i = 0; i < 100; ++i) {

         tag = rand()%(mpi_tag_ub+1);

         for (j = 0; j < mpi_comm->num_unique_tags; ++j)
            if (mpi_comm->unique_tags[j].tag == tag)
               break;

         if (j > 0 && j == mpi_comm->num_unique_tags) break;
      }

      if (i == 100) {
         yac_internal_abort_message("ERROR: could not generate a unique tag", __FILE__, __LINE__);
         return -1;
      }
   }

   ENSURE_ARRAY_SIZE(mpi_comm->unique_tags, mpi_comm->unique_tags_array_size,
                     mpi_comm->num_unique_tags + 1);

   mpi_comm->unique_tags[mpi_comm->num_unique_tags].tag = tag;

   if (identifier_string != NULL) {
      size_t string_length = strlen(identifier_string) + 1;
      mpi_comm->unique_tags[mpi_comm->num_unique_tags].identifier_string =
         malloc(string_length);
      strcpy(mpi_comm->unique_tags[mpi_comm->num_unique_tags].identifier_string,
             identifier_string);
   } else {
      mpi_comm->unique_tags[mpi_comm->num_unique_tags].identifier_string = NULL;
   }

   mpi_comm->num_unique_tags++;

   return tag;
}

int compare_indexed_strings(const void * a, const void * b)
{
  return strcmp(((struct indexed_strings*)a)->str,
                ((struct indexed_strings*)b)->str);
}

unsigned
register_tag_string(struct mpi_communicator * mpi_comm, char const * str) {

  unsigned i;

  for (i = 0; i < mpi_comm->max_num_registered_tags; ++i)
    if (!mpi_comm->registered_tags[i].free &&
        !strcmp(mpi_comm->registered_tags[i].str, str)) break;

  if (i != mpi_comm->max_num_registered_tags) return i;

  for (i = 0; i < mpi_comm->max_num_registered_tags; ++i)
    if (mpi_comm->registered_tags[i].free) break;

  if (i == mpi_comm->max_num_registered_tags) {
    ENSURE_ARRAY_SIZE(mpi_comm->registered_tags,
                      mpi_comm->max_num_registered_tags,
                      mpi_comm->max_num_registered_tags + 1);
    for (unsigned j = i; j < mpi_comm->max_num_registered_tags; ++j) {
      mpi_comm->registered_tags[j].free = 1;
      mpi_comm->registered_tags[j].str = NULL;
    }
  }

  mpi_comm->registered_tags[i].str = malloc(strlen(str) + 1);
  strcpy(mpi_comm->registered_tags[i].str, str);
  mpi_comm->registered_tags[i].free = 0;

  if (i >= mpi_tag_ub)
    yac_internal_abort_message(
       "ERROR(register_tag_string): too many registered tags",
       __FILE__, __LINE__);

  return i;
}

static void
get_unique_comm_tags_mpi(struct communicator * comm,
                         char const * unique_generator_string,
                         char const ** tag_strings, unsigned num_tag_strings,
                         unsigned * tags) {

  struct mpi_communicator * mpi_comm = (struct mpi_communicator *)comm;

  // min reduce unique_generator_string size + num tags
  unsigned unique_generator_string_len = strlen(unique_generator_string);
  unsigned send_data[2] = {num_tag_strings, unique_generator_string_len};
  unsigned recv_data[2];
  mpi_err_handler(MPI_Allreduce(send_data, recv_data, 2, MPI_UNSIGNED, MPI_MIN,
                                mpi_comm->comm), mpi_comm->comm);

  if ((send_data[0] != recv_data[0]) || (send_data[1] != recv_data[1]))
    yac_internal_abort_message(
       "ERROR(get_unique_comm_tag_mpi): strlen(unique_generator_string) or "
       "num_tag_strings do not match", __FILE__, __LINE__);

  // min reduce unique_generator_string
  char * min_unique_generator_string = malloc(unique_generator_string_len + 1);
  min_unique_generator_string[unique_generator_string_len] = '\0';
  mpi_err_handler(
    MPI_Allreduce(unique_generator_string, min_unique_generator_string,
                  unique_generator_string_len, MPI_CHAR, MPI_MIN,
                  mpi_comm->comm), mpi_comm->comm);

  if (strcmp(min_unique_generator_string, unique_generator_string))
    yac_internal_abort_message(
       "ERROR(get_unique_comm_tag_mpi): unique_generator_string does not match",
       __FILE__, __LINE__);

  free(min_unique_generator_string);

  // sort tag strings
  struct indexed_strings * indexed_tag_strings =
    malloc(num_tag_strings * sizeof(*indexed_tag_strings));
  for (unsigned i = 0; i < num_tag_strings; ++i) {
    indexed_tag_strings[i].str = tag_strings[i];
    indexed_tag_strings[i].idx = i;
  }
  qsort(indexed_tag_strings, num_tag_strings, sizeof(*indexed_tag_strings),
        compare_indexed_strings);

  // assigne free tags to strings
  for (unsigned i = 0; i < num_tag_strings; ++i)
    tags[indexed_tag_strings[i].idx] =
      register_tag_string(mpi_comm, indexed_tag_strings[i].str);

  free(indexed_tag_strings);

  check_open_requests((struct mpi_communicator *)comm);
}

static void comm_allgather_mpi(void * send_buffer, unsigned send_count,
                               void * recv_buffer, unsigned recv_count, 
                               enum yac_comm_data_type data_type,
                               struct communicator * comm) {

   struct mpi_communicator * mpi_comm = (struct mpi_communicator *)comm;

   mpi_err_handler(MPI_Allgather(send_buffer, send_count, get_mpi_datatype(data_type),
                                 recv_buffer, recv_count, get_mpi_datatype(data_type),
                                 mpi_comm->comm), mpi_comm->comm);

   check_open_requests((struct mpi_communicator *)comm);
}

static void comm_allgatherv_mpi(void * send_buffer, unsigned send_count,
                                void * recv_buffer, unsigned * recv_counts, 
                                enum yac_comm_data_type data_type,
                                struct communicator * comm){

   struct mpi_communicator * mpi_comm = (struct mpi_communicator *)comm;

   unsigned comm_size;

   comm_size = get_comm_size_mpi(comm);

   // we assume the simplest method of using MPI_Allgatherv, therefore we can generate
   // the displ array automatically

   int * displs = malloc(comm_size * sizeof(*displs));
   int accum = 0;
   for (unsigned i = 0; i < comm_size; ++i) {
      displs[i] = accum;
      accum += recv_counts[i];
   }

   mpi_err_handler(MPI_Allgatherv(send_buffer, send_count, get_mpi_datatype(data_type),
                                  recv_buffer, (int*)recv_counts, displs,
                                  get_mpi_datatype(data_type), mpi_comm->comm),
                   mpi_comm->comm);

   check_open_requests((struct mpi_communicator *)comm);
}

static struct communicator * comm_dup_mpi(struct communicator * comm) {

   struct mpi_communicator * mpi_comm = (struct mpi_communicator *)comm;

   MPI_Comm new_mpi_comm;

   mpi_err_handler(MPI_Comm_dup(mpi_comm->comm, &new_mpi_comm),
                   mpi_comm->comm);

   return mpi_comm_new_(new_mpi_comm);
}

static struct communicator * comm_split_mpi(struct communicator * comm,
                                            int color, int key) {

   struct mpi_communicator * mpi_comm = (struct mpi_communicator *)comm;

   MPI_Comm new_mpi_comm;

   if ( color < 0 ) color = MPI_UNDEFINED;

   mpi_err_handler(MPI_Comm_split(mpi_comm->comm, color, key, &new_mpi_comm),
                   mpi_comm->comm);

   return mpi_comm_new_(new_mpi_comm);
}

static struct communicator * comm_split_intercomm_mpi(
   struct communicator * comm, int group, unsigned * local_group_ranks,
   unsigned * remote_group_ranks) {

   struct mpi_communicator * mpi_comm = (struct mpi_communicator *)comm;

   if (group < 0 || group > 1) {

      fprintf(stderr, "%3d: wrong group argument for comm_split_intercomm_mpi\n",
              get_comm_rank_mpi(comm));
      MPI_Abort(mpi_comm->comm, -1);
   }

   MPI_Comm global_comm;

   // MPI_Intercomm_create uses point-to-point communication in the common
   // communicator, in order to avoid conflicts with other messages in
   // mpi_comm->comm a copy is made
   mpi_err_handler(MPI_Comm_dup(mpi_comm->comm, &global_comm), mpi_comm->comm);

   MPI_Group global_group;

   mpi_err_handler(MPI_Comm_group(global_comm, &global_group), mpi_comm->comm);

   MPI_Comm local_comm;

   mpi_err_handler(MPI_Comm_split(global_comm, group, get_comm_rank_mpi(comm),
                                  &local_comm), mpi_comm->comm);

   MPI_Group local_group;

   mpi_err_handler(MPI_Comm_group(local_comm, &local_group), mpi_comm->comm);

   MPI_Group remote_group;

   mpi_err_handler(MPI_Group_difference(global_group, local_group, &remote_group),
                   mpi_comm->comm);

   int remote_comm_root = 0;
   int remote_root;

   mpi_err_handler(MPI_Group_translate_ranks(remote_group, 1, &remote_comm_root,
                                             global_group, &remote_root),
                   mpi_comm->comm);

   MPI_Comm new_mpi_comm;

   mpi_err_handler(MPI_Intercomm_create(local_comm, 0, global_comm, remote_root,
                                        0, &new_mpi_comm), mpi_comm->comm);

   struct communicator * new_comm = mpi_comm_new_(new_mpi_comm);

   if (local_group_ranks != NULL || remote_group_ranks != NULL) {

      unsigned local_comm_size, remote_comm_size;

      local_comm_size = yac_get_comm_size(new_comm);
      remote_comm_size = yac_get_comm_remote_size(new_comm);

      unsigned max_size;

      max_size = MAX(local_comm_size, remote_comm_size);

      unsigned * local_ranks;

      local_ranks = malloc(max_size * sizeof(*local_ranks));

      unsigned i;

      for (i = 0; i < max_size; ++i)
         local_ranks[i] = i;

      if (local_group_ranks != NULL)
         mpi_err_handler(MPI_Group_translate_ranks(
            local_group, local_comm_size, (int*)local_ranks, global_group,
            (int*)local_group_ranks), mpi_comm->comm);

      if (remote_group_ranks != NULL)
         mpi_err_handler(MPI_Group_translate_ranks(
            remote_group, remote_comm_size, (int*)local_ranks, global_group,
            (int*)remote_group_ranks), mpi_comm->comm);
      free(local_ranks);
   }

   mpi_err_handler(MPI_Comm_free(&local_comm), mpi_comm->comm);
   mpi_err_handler(MPI_Group_free(&global_group), mpi_comm->comm);
   mpi_err_handler(MPI_Group_free(&local_group), mpi_comm->comm);
   mpi_err_handler(MPI_Group_free(&remote_group), mpi_comm->comm);
   mpi_err_handler(MPI_Comm_free(&global_comm), mpi_comm->comm);

   return new_comm;
}

static unsigned get_comm_remote_size_mpi(struct communicator * comm) {

   struct mpi_communicator * mpi_comm = (struct mpi_communicator *)comm;

   int size;

   mpi_err_handler(MPI_Comm_remote_size(mpi_comm->comm, &size), mpi_comm->comm);

   return size;
}

static void comm_abort_mpi(struct communicator * comm) {

   struct mpi_communicator * mpi_comm = (struct mpi_communicator *)comm;

   mpi_err_handler(MPI_Abort(mpi_comm->comm, EXIT_FAILURE), mpi_comm->comm);
}

static void free_communicator_mpi(struct communicator * comm) {

   check_open_requests((struct mpi_communicator *)comm);

   struct mpi_communicator * mpi_comm = (struct mpi_communicator *)comm;

   unsigned flag;

   do {
      flag = 0;

      for (unsigned i = 0; i < mpi_comm->message_indices_size; ++i) {

         unsigned msg_index = mpi_comm->message_indices[i];

         if (msg_index != -1) {

            flag = 1;

            switch (global_messages[msg_index].message_type) {
               case(SEND):
                  // wait until any open send or receive has been processed
                  waitany_open_request(mpi_comm);
                  break;
               case(RECV):
                  // cancel the receive
                  cancel_recv_callback_message(msg_index);
                  break;
               default:
                  yac_internal_abort_message(
                     "ERROR(free_communicator_mpi): Invalid message type.",
                     __FILE__, __LINE__);
            };
         }
      }
   } while (flag);

   comm_waitall_mpi(mpi_comm->requests_array_size,
                    (struct comm_request**)mpi_comm->requests);
   for (unsigned i = 0; i < mpi_comm->requests_array_size; ++i)
      if (((struct comm_request**)mpi_comm->requests)[i] != COMM_REQUEST_NULL)
         free(mpi_comm->requests[i]);
   free(mpi_comm->requests);

   for (unsigned i = 0; i < mpi_comm->num_unique_tags; ++i)
      free(mpi_comm->unique_tags[i].identifier_string);
   free(mpi_comm->unique_tags);

   for (unsigned i = 0; i < mpi_comm->max_num_registered_tags; ++i)
      free(mpi_comm->registered_tags[i].str);
   free(mpi_comm->registered_tags);

   free(mpi_comm->message_indices);
   if (mpi_comm->comm != MPI_COMM_NULL)
      mpi_err_handler(MPI_Comm_free(&(mpi_comm->comm)), MPI_COMM_WORLD);
   free(mpi_comm);
}

static int comm_pack_size_mpi(int count, enum yac_comm_data_type type,
                              struct communicator * comm) {

   struct mpi_communicator * mpi_comm = (struct mpi_communicator *)comm;

#ifndef WITHOUT_PACK_EXTERNAL
   MPI_Aint asize;

   mpi_err_handler(MPI_Pack_external_size(datarep, count, get_mpi_datatype(type),
                                          &asize), mpi_comm->comm);
   int size = (int) asize;

   if ( size != asize)
     yac_internal_abort_message(
        "ERROR: Conversion in comm_pack_size_mpi does not work!",
        __FILE__, __LINE__);
#else
   int size;

   mpi_err_handler(MPI_Pack_size(count, get_mpi_datatype(type), mpi_comm->comm,
                                 &size), mpi_comm->comm);
#endif
   return size;
}

static void comm_pack_mpi(void * in_buffer, int in_count,
                          enum yac_comm_data_type type, void * out_buffer,
                          int out_buffer_size, int * position,
                          struct communicator * comm) {

   struct mpi_communicator * mpi_comm = (struct mpi_communicator *)comm;

#ifndef WITHOUT_PACK_EXTERNAL

   MPI_Aint a_out_buffer_size = (MPI_Aint) out_buffer_size;
   MPI_Aint a_position        = (MPI_Aint)(*position);

   if ( *position != a_position || a_out_buffer_size != out_buffer_size )
     yac_internal_abort_message(
        "ERROR: Conversion in comm_pack_mpi does not work!",
        __FILE__, __LINE__);

   mpi_err_handler(MPI_Pack_external(
      datarep, in_buffer, in_count, get_mpi_datatype(type), out_buffer,
      a_out_buffer_size, &a_position), mpi_comm->comm);
   *position = (int) a_position;

   if ( *position != a_position )
     yac_internal_abort_message(
        "ERROR: Conversion in comm_pack_mpi does not work!",
        __FILE__, __LINE__);
#else
   mpi_err_handler(MPI_Pack(in_buffer, in_count, get_mpi_datatype(type),
                            out_buffer, out_buffer_size, position,
                            mpi_comm->comm), mpi_comm->comm);
#endif
}

static void comm_unpack_mpi(void * in_buffer, int in_size, int * position,
                            void * out_buffer, int out_count,
                            enum yac_comm_data_type type,
                            struct communicator * comm) {

   struct mpi_communicator * mpi_comm = (struct mpi_communicator *)comm;

#ifndef WITHOUT_PACK_EXTERNAL
   MPI_Aint a_in_size = (MPI_Aint) in_size;
   MPI_Aint a_position = (MPI_Aint)(*position);

   if ( *position != a_position || a_in_size != in_size )
     yac_internal_abort_message(
        "ERROR: Conversion in comm_pack_mpi does not work!",
        __FILE__, __LINE__);

   mpi_err_handler(MPI_Unpack_external(
      datarep, in_buffer, a_in_size, &a_position, out_buffer, out_count,
      get_mpi_datatype(type)), mpi_comm->comm);
   *position = (int) a_position;

   if ( *position != a_position )
     yac_internal_abort_message(
        "ERROR: Conversion in comm_pack_mpi does not work!",
        __FILE__, __LINE__);
#else
   mpi_err_handler(MPI_Unpack(in_buffer, in_size, position, out_buffer,
                              out_count, get_mpi_datatype(type),
                              mpi_comm->comm), mpi_comm->comm);
#endif
}

static void comm_translate_ranks_mpi(struct communicator * comm_a,
                                     struct communicator * comm_b,
                                     unsigned * ranks) {

   struct mpi_communicator * mpi_comm_a = (struct mpi_communicator *)comm_a;
   struct mpi_communicator * mpi_comm_b = (struct mpi_communicator *)comm_b;

   MPI_Group group_a, group_b;

   mpi_err_handler(MPI_Comm_group(mpi_comm_a->comm, &group_a), mpi_comm_a->comm);
   mpi_err_handler(MPI_Comm_group(mpi_comm_b->comm, &group_b), mpi_comm_b->comm);

   int * ranks_b = malloc(mpi_comm_b->size * sizeof(*ranks_b));

   for (unsigned i = 0; i < mpi_comm_b->size; ++i)
      ranks_b[i] = i;

   mpi_err_handler(MPI_Group_translate_ranks(
      group_b, mpi_comm_b->size, ranks_b, group_a, (int*)ranks),
      mpi_comm_a->comm);

   free(ranks_b);
}

static void send_bcast_callback(struct communicator * comm,
                                struct comm_message msg, void * user_data) {

  struct temp_bcast_data * bcast_data = (struct temp_bcast_data *)user_data;

  if (--(bcast_data->data.send_count) == 0) {

    func_recv_callback bcast_callback = bcast_data->callback;

    if (bcast_callback)
      bcast_callback(comm, bcast_data->msg, bcast_data->user_data);
    free(bcast_data->msg.buffer);
    free(bcast_data);
  }
}

static void recv_bcast_callback(struct communicator * comm,
                                struct comm_message msg, void * user_data) {

  struct mpi_communicator * mpi_comm = (struct mpi_communicator*)comm;
  struct temp_bcast_data * bcast_data = (struct temp_bcast_data *)user_data;

  unsigned root = bcast_data->data.root;
  bcast_data->msg = msg;
  bcast_data->msg.src_rank = root;
  bcast_data->data.send_count = 1; // we have to ensure that send_bcast_callback
                                   // is not called prematurely, this is done by
                                   // initialising the send_count with 1

  // on root -> temp_rank == 0
  unsigned temp_rank = (mpi_comm->size + mpi_comm->rank - root)%mpi_comm->size;

  unsigned bit = 1, send_rank;

  while(bit <= temp_rank) bit <<= 1;

  while ((send_rank = temp_rank | bit) < mpi_comm->size) {

    bit <<= 1;

    send_rank = (send_rank + root) % mpi_comm->size;

    bcast_data->data.send_count++; // this needs to be done before the send,
                                   // because it may reduce it again and this
                                   // should not become 0
    send_internal(msg.buffer, msg.count, msg.type, send_rank, msg.tag, comm,
                  bcast_data, send_bcast_callback, bcast_data->priority, 0);
  };

  // this will reduce the send_count by one and call bcast_callback if all
  // messages have already been delivered or no messages had to be been sent
  send_bcast_callback(comm, msg, bcast_data);
}

static void comm_bcast_callback_mpi(void * buffer, unsigned count,
                                    enum yac_comm_data_type data_type,
                                    unsigned root, unsigned tag,
                                    struct communicator * comm,
                                    void * bcast_user_data,
                                    func_recv_callback bcast_callback,
                                    enum yac_message_priority priority) {

  struct mpi_communicator * mpi_comm = (struct mpi_communicator*)comm;

  if (root >= mpi_comm->size)
    yac_internal_abort_message("ERROR: Invalid root rank.", __FILE__, __LINE__);

  struct temp_bcast_data * user_data = malloc(1 * sizeof(*user_data));
  user_data->data.root = root;
  user_data->callback = bcast_callback;
  user_data->user_data = bcast_user_data;
  user_data->priority = convert_msg_prio(priority);

  // uses a hypercube algorithm to distribute the data
  if (root == mpi_comm->rank) {

    struct comm_message temp_message;

    unsigned buffer_size = count * get_data_type_size(data_type);
    temp_message.buffer = malloc(buffer_size);
    memcpy(temp_message.buffer, buffer, buffer_size);
    temp_message.type = data_type;
    temp_message.count = count;
    temp_message.tag = tag;
    temp_message.dst_rank = root;
    temp_message.src_rank = root;

    recv_bcast_callback(comm, temp_message, user_data);

  } else {

    // on root -> temp_rank == 0
    unsigned temp_rank =
      (mpi_comm->size + mpi_comm->rank - root) % mpi_comm->size;

    unsigned bit = 1;

    while (bit <= temp_rank) bit <<= 1;
    bit >>= 1;

    unsigned src_rank = ((temp_rank ^ bit) + root) % mpi_comm->size;

    recv_internal(count, data_type, src_rank, tag, comm, user_data,
                  recv_bcast_callback, NULL, convert_msg_prio(priority), 0);
  }
}

static void recv_gatherv_callback(struct communicator * comm,
                                  struct comm_message msg, void * user_data) {

  struct mpi_communicator * mpi_comm = (struct mpi_communicator*)comm;

  struct temp_gatherv_data * temp_data = (struct temp_gatherv_data*)user_data;

  memcpy(temp_data->buffer_p, msg.buffer, msg.count *
         get_data_type_size(msg.type));

  temp_data->buffer_p = (char*)(temp_data->buffer_p) + msg.count *
                        get_data_type_size(msg.type);

  unsigned step =
    (msg.src_rank - msg.dst_rank + mpi_comm->size)%mpi_comm->size;

  step <<= 1;
  if (step == 0) step = 1;

  unsigned rel_rank =
    (mpi_comm->size + mpi_comm->rank - temp_data->root) % mpi_comm->size;

  void * gather_buffer = NULL;
  unsigned gather_count = 0;

  // if the locally gathered data needs to be sent to another process
  if (rel_rank & step) {

    unsigned local_buffer_size = 0;
    unsigned tmp = MIN (step, mpi_comm->size - rel_rank);

    for (unsigned i = 0; i < tmp; ++i)
      local_buffer_size +=
        temp_data->counts[(rel_rank + i + temp_data->root) % mpi_comm->size];

    unsigned send_rank = ((rel_rank ^ step) + temp_data->root) % mpi_comm->size;

    comm_bsend_mpi(temp_data->buffer, local_buffer_size, msg.type, send_rank,
                   msg.tag, comm);

  // if data from another process needs to be received
  } else if (step < mpi_comm->size) {

    unsigned recv_rank =
      (rel_rank + step + temp_data->root) % mpi_comm->size;
    unsigned rel_recv_rank = rel_rank + step;
    unsigned tmp = MIN (step, mpi_comm->size - rel_recv_rank);
    unsigned recv_count = 0;

    for (unsigned i = 0; i < tmp; ++i)
      recv_count += temp_data->counts[(recv_rank + i)%mpi_comm->size];

    // in case the mpi_comm->size is not a multiple of two this can be true
    if (rel_recv_rank >= mpi_comm->size) {

      // we have to skip one step, which is done by faking the message
      msg.src_rank = recv_rank;
      msg.buffer = NULL;
      msg.count = 0;

      recv_gatherv_callback(comm, msg, user_data);

    } else {
      comm_recv_callback_mpi(recv_count, msg.type, recv_rank, msg.tag, comm,
                             user_data, recv_gatherv_callback, NULL,
                             temp_data->priority);
    }

    return;

  // if all data has been gathered -> this is root
  } else {

    for (unsigned i = 0; i < mpi_comm->size; ++i)
      gather_count += temp_data->counts[i];

    gather_buffer = malloc(gather_count * get_data_type_size(msg.type));

    unsigned offset = 0;

    for (unsigned i = 0; i < mpi_comm->rank; ++i)
      offset += temp_data->counts[i];

    memcpy(gather_buffer, (char*)(temp_data->buffer) + (gather_count - offset) *
           get_data_type_size(msg.type), offset * get_data_type_size(msg.type));
    memcpy((char*)(gather_buffer) + offset * get_data_type_size(msg.type),
           temp_data->buffer, (gather_count - offset) *
           get_data_type_size(msg.type));
  }

  msg.src_rank = mpi_comm->rank;
  msg.buffer = gather_buffer;
  msg.count = gather_count;

  if (temp_data->gatherv_callback)
    temp_data->gatherv_callback(comm, msg, temp_data->user_data);

  free(gather_buffer);
  free(temp_data->buffer);
  free(temp_data->counts);
  free(temp_data);
}

static void comm_gatherv_callback_mpi(void * send_buffer, unsigned * counts,
                                      enum yac_comm_data_type data_type,
                                      unsigned root,
                                      unsigned tag, struct communicator * comm,
                                      void * user_data,
                                      func_recv_callback gatherv_callback,
                                      enum yac_message_priority priority) {

  struct mpi_communicator * mpi_comm = (struct mpi_communicator*)comm;

  unsigned rel_rank = (mpi_comm->size + mpi_comm->rank - root) % mpi_comm->size;

  unsigned local_buffer_size = 0;

  for (unsigned i = rel_rank; i < mpi_comm->size; ++i)
    local_buffer_size += counts[(i + root) % mpi_comm->size];

  struct temp_gatherv_data * temp_data = malloc(1 * sizeof(*temp_data));
  temp_data->counts = malloc(mpi_comm->size * sizeof(*(temp_data->counts)));
  memcpy(temp_data->counts, counts,
         mpi_comm->size * sizeof(*(temp_data->counts)));
  temp_data->root = root;
  temp_data->priority = convert_msg_prio(priority);
  temp_data->buffer = malloc(get_data_type_size(data_type) * local_buffer_size);
  temp_data->buffer_p = temp_data->buffer;
  temp_data->user_data = user_data;
  temp_data->gatherv_callback = gatherv_callback;

  struct comm_message temp_msg;
  temp_msg.buffer = send_buffer;
  temp_msg.type = data_type;
  temp_msg.count = counts[mpi_comm->rank];
  temp_msg.tag = tag;
  temp_msg.dst_rank = mpi_comm->rank;
  temp_msg.src_rank = mpi_comm->rank;

  recv_gatherv_callback(comm, temp_msg, temp_data);
}

static void ring_allgather_callback(struct communicator * comm,
                                    struct comm_message msg,
                                    void * user_data) {

  struct mpi_communicator * mpi_comm = (struct mpi_communicator*)comm;

  struct temp_allgather_data * temp_data =
    (struct temp_allgather_data*)user_data;

  unsigned comm_rank = mpi_comm->rank;
  unsigned comm_size = mpi_comm->size;
  unsigned step = temp_data->step;
  unsigned data_type_size = get_data_type_size(msg.type);
  temp_data->step <<= 1;

  // copy data into buffer
  memcpy(temp_data->buffer_p_recv, msg.buffer,
         msg.count * data_type_size);
  temp_data->buffer_p_recv = (char*)(temp_data->buffer_p_recv) +
                             msg.count * data_type_size;

  // if all data has been gathered
  if (step >= comm_size) {

    // copy data into correct order
    memcpy(temp_data->buffer, ((char*)(temp_data->buffer)) + comm_size *
           temp_data->count * data_type_size, mpi_comm->rank *
           temp_data->count * data_type_size);

    msg.buffer = temp_data->buffer;
    msg.src_rank = comm_rank;
    msg.count = temp_data->count * comm_size;

    if (temp_data->allgather_callback)
      temp_data->allgather_callback(comm, msg, temp_data->user_data);

    free(temp_data->buffer);
    free(temp_data);

  } else {

    unsigned exchange_count = MIN(step, comm_size - step) * temp_data->count;

    // send local data
    unsigned dst_rank = (comm_size + comm_rank - step) % comm_size;
    comm_bsend_mpi(temp_data->buffer_p_send, exchange_count, msg.type, dst_rank,
                   msg.tag, comm);

    // receive remote data
    unsigned src_rank = (comm_rank + step) % comm_size;
    comm_recv_callback_mpi(exchange_count, msg.type, src_rank, msg.tag, comm,
                           user_data, ring_allgather_callback, NULL,
                           temp_data->priority);
  }
}

static void comm_allgather_callback_mpi(void * send_buffer, unsigned count,
                                        enum yac_comm_data_type data_type,
                                        unsigned tag,
                                        struct communicator * comm,
                                        void * user_data,
                                        func_recv_callback allgather_callback,
                                        enum yac_message_priority priority) {

  struct mpi_communicator * mpi_comm = (struct mpi_communicator*)comm;

  struct temp_allgather_data * temp_data = malloc(sizeof(*temp_data));
  temp_data->count = count;
  temp_data->step = 1;
  temp_data->data_type = data_type;
  temp_data->priority = convert_msg_prio(priority);
  temp_data->buffer = malloc(get_data_type_size(data_type) * count *
                             (mpi_comm->size + mpi_comm->rank));
  temp_data->buffer_p_send = ((char*)temp_data->buffer) +
                             get_data_type_size(data_type) * count *
                             mpi_comm->rank;
  temp_data->buffer_p_recv = temp_data->buffer_p_send;
  temp_data->user_data = user_data;
  temp_data->allgather_callback = allgather_callback;

  struct comm_message temp_msg;
  temp_msg.buffer = send_buffer;
  temp_msg.type = data_type;
  temp_msg.count = count;
  temp_msg.tag = tag;
  temp_msg.dst_rank = mpi_comm->rank;
  temp_msg.src_rank = mpi_comm->rank;

  ring_allgather_callback(comm, temp_msg, temp_data);
}

static void ring_allgatherv_callback(struct communicator * comm,
                                     struct comm_message msg,
                                     void * user_data) {

  struct mpi_communicator * mpi_comm = (struct mpi_communicator*)comm;

  struct temp_allgatherv_data * temp_data =
    (struct temp_allgatherv_data*)user_data;

  unsigned comm_rank = mpi_comm->rank;
  unsigned comm_size = mpi_comm->size;
  unsigned step = temp_data->step;
  temp_data->step <<= 1;

  // copy data into buffer
  memcpy(temp_data->buffer_p_recv, msg.buffer,
         msg.count * get_data_type_size(msg.type));
  temp_data->buffer_p_recv = (char*)(temp_data->buffer_p_recv) +
                             msg.count * get_data_type_size(msg.type);

  // if all data has been gathered
  if (step >= comm_size) {

    unsigned total_count = 0, extra_count = 0;
    for (unsigned i = 0; i < mpi_comm->size; ++i)
      total_count += temp_data->counts[i];
    for (unsigned i = 0; i < mpi_comm->rank; ++i)
        extra_count += temp_data->counts[i];

    // copy data into correct order
    memcpy(temp_data->buffer, ((char*)(temp_data->buffer)) +
           total_count * get_data_type_size(msg.type),
           extra_count * get_data_type_size(msg.type));

    msg.buffer = temp_data->buffer;
    msg.src_rank = comm_rank;
    msg.count = total_count;

    if (temp_data->allgatherv_callback)
      temp_data->allgatherv_callback(comm, msg, temp_data->user_data);

    free(temp_data->counts);
    free(temp_data->buffer);
    free(temp_data);

  } else {

    unsigned num_single_counts = MIN(step, comm_size - step);

    unsigned send_count = 0;

    for (unsigned i = 0; i < num_single_counts; ++i)
      send_count += temp_data->counts[(comm_rank + i) % comm_size];

    // send local data
    unsigned dst_rank = (comm_size + comm_rank - step) % comm_size;
    comm_bsend_mpi(temp_data->buffer_p_send, send_count, msg.type, dst_rank,
                   msg.tag, comm);

    unsigned recv_count = 0;

    for (unsigned i = 0; i < num_single_counts; ++i)
      recv_count += temp_data->counts[(comm_rank + step + i) % comm_size];

    // receive remote data
    unsigned src_rank = (comm_rank + step) % comm_size;
    comm_recv_callback_mpi(recv_count, msg.type, src_rank, msg.tag, comm,
                           user_data, ring_allgatherv_callback, NULL,
                           temp_data->priority);
  }
}

static void comm_allgatherv_callback_mpi(void * send_buffer, unsigned * counts,
                                         enum yac_comm_data_type data_type,
                                         unsigned tag,
                                         struct communicator * comm,
                                         void * user_data,
                                         func_recv_callback allgatherv_callback,
                                         enum yac_message_priority priority) {

  struct mpi_communicator * mpi_comm = (struct mpi_communicator*)comm;

  unsigned total_count = 0, extra_count = 0;
  for (unsigned i = 0; i < mpi_comm->size; ++i) total_count += counts[i];
  for (unsigned i = 0; i < mpi_comm->rank; ++i) extra_count += counts[i];

  struct temp_allgatherv_data * temp_data = malloc(sizeof(*temp_data));
  temp_data->counts = malloc(mpi_comm->size * sizeof(*(temp_data->counts)));
  memcpy(temp_data->counts, counts,
         mpi_comm->size * sizeof(*(temp_data->counts)));
  temp_data->step = 1;
  temp_data->data_type = data_type;
  temp_data->priority = convert_msg_prio(priority);
  temp_data->buffer = malloc(get_data_type_size(data_type) *
                             (total_count + extra_count));
  temp_data->buffer_p_send = ((char*)temp_data->buffer) +
                             get_data_type_size(data_type) * extra_count;
  temp_data->buffer_p_recv = temp_data->buffer_p_send;
  temp_data->user_data = user_data;
  temp_data->allgatherv_callback = allgatherv_callback;

  struct comm_message temp_msg;
  temp_msg.buffer = send_buffer;
  temp_msg.type = data_type;
  temp_msg.count = counts[mpi_comm->rank];
  temp_msg.tag = tag;
  temp_msg.dst_rank = mpi_comm->rank;
  temp_msg.src_rank = mpi_comm->rank;

  ring_allgatherv_callback(comm, temp_msg, temp_data);
}

static void recv_scatter_callback(struct communicator * comm,
                                  struct comm_message msg, void * user_data) {

  struct mpi_communicator * mpi_comm = (struct mpi_communicator*)comm;

  struct temp_scatter_data * temp_data = (struct temp_scatter_data*)user_data;

  unsigned rel_rank =
    (mpi_comm->size + mpi_comm->rank - temp_data->root) % mpi_comm->size;

  unsigned step;

  if (mpi_comm->rank == temp_data->root) {
    step = 1;
    while (step < mpi_comm->size) step <<= 1;
  } else {
    step = (msg.dst_rank - msg.src_rank + mpi_comm->size)%mpi_comm->size;
  }

  // for all sends
  while (step >>= 1) {

    if (rel_rank + step >= mpi_comm->size) continue;

    unsigned send_offset =
      temp_data->count * step * get_data_type_size(msg.type);
    unsigned count =
      temp_data->count * MIN(step, mpi_comm->size - rel_rank - step);

    comm_bsend_mpi(((char*)msg.buffer) + send_offset, count, msg.type,
                   (mpi_comm->rank + step)%mpi_comm->size, msg.tag, comm);
  }

  // we can reuse the buffer in msg
  msg.count = temp_data->count;
  msg.src_rank = temp_data->root;

  if (temp_data->scatter_callback)
    temp_data->scatter_callback(comm, msg, temp_data->user_data);

  free(temp_data);
}

static void comm_scatter_callback_mpi(void * send_buffer, unsigned count,
                                      enum yac_comm_data_type data_type,
                                      unsigned root, unsigned tag,
                                      struct communicator * comm,
                                      void * user_data,
                                      func_recv_callback scatter_callback,
                                      enum yac_message_priority priority) {

  struct mpi_communicator * mpi_comm = (struct mpi_communicator*)comm;

  struct temp_scatter_data * temp_data = malloc(1 * sizeof(*temp_data));
  temp_data->count = count;
  temp_data->root = root;
  temp_data->user_data = user_data;
  temp_data->priority = convert_msg_prio(priority);
  temp_data->scatter_callback = scatter_callback;

  unsigned rel_rank = (mpi_comm->size + mpi_comm->rank - root) % mpi_comm->size;

  // if this is root
  if (mpi_comm->rank == root) {

    unsigned total_count = count * mpi_comm->size;

    if (root != 0) {

      void * temp_send_buffer =
        malloc(total_count * get_data_type_size(data_type));

      unsigned offset = count * mpi_comm->rank;

      memcpy(temp_send_buffer, ((char*)send_buffer) + offset *
             get_data_type_size(data_type), (total_count - offset) *
             get_data_type_size(data_type));
      memcpy(((char*)temp_send_buffer) + (total_count - offset) *
             get_data_type_size(data_type), send_buffer, offset *
             get_data_type_size(data_type));

      send_buffer = temp_send_buffer;
    }

    struct comm_message temp_message;

    temp_message.buffer = send_buffer;
    temp_message.type = data_type;
    temp_message.count = total_count;
    temp_message.tag = tag;
    temp_message.dst_rank = root;
    temp_message.src_rank = root;

    recv_scatter_callback(comm, temp_message, temp_data);

    if (root != 0) free(send_buffer);

  } else {

    // bit hack for Reset Lowest Set Bit
    unsigned rel_recv_rank = rel_rank & (rel_rank - 1);
    unsigned recv_rank = (rel_recv_rank + temp_data->root) % mpi_comm->size;
    unsigned step = MIN(rel_rank - rel_recv_rank, mpi_comm->size - rel_rank);
    unsigned recv_count = count * step;

    comm_recv_callback_mpi(recv_count, data_type, recv_rank, tag, comm,
                           temp_data, recv_scatter_callback, NULL,
                           convert_msg_prio(priority));
  }
}

static void recv_scatterv_callback(struct communicator * comm,
                                   struct comm_message msg, void * user_data) {

  struct mpi_communicator * mpi_comm = (struct mpi_communicator*)comm;

  struct temp_scatterv_data * temp_data =
    (struct temp_scatterv_data*)user_data;

  unsigned rel_rank =
    (mpi_comm->size + mpi_comm->rank - temp_data->root) % mpi_comm->size;

  unsigned step;

  if (mpi_comm->rank == temp_data->root) {
    step = 1;
    while (step < mpi_comm->size) step <<= 1;
  } else {
    step = (msg.dst_rank - msg.src_rank + mpi_comm->size)%mpi_comm->size;
  }

  // for all sends
  while (step >>= 1) {

    if (rel_rank + step >= mpi_comm->size) continue;

    unsigned count = 0;
    unsigned send_offset = 0;

    // compute send buffer offset
    for (unsigned i = 0; i < step; ++i)
      send_offset += temp_data->counts[(mpi_comm->rank + i)%mpi_comm->size];

    send_offset *= get_data_type_size(msg.type);

    for (unsigned i = 0; (i < step) && (rel_rank + i + step < mpi_comm->size); ++i)
      count += temp_data->counts[(mpi_comm->rank + step + i)%mpi_comm->size];

    comm_bsend_mpi(((char*)msg.buffer) + send_offset, count, msg.type,
                   (mpi_comm->rank + step)%mpi_comm->size, msg.tag, comm);
  }

  // we can reuse the buffer in msg
  msg.count = temp_data->counts[mpi_comm->rank];
  msg.src_rank = temp_data->root;

  if (temp_data->scatterv_callback)
    temp_data->scatterv_callback(comm, msg, temp_data->user_data);

  free(temp_data->counts);
  free(temp_data);
}

static void comm_scatterv_callback_mpi(void * send_buffer, unsigned * counts,
                                       enum yac_comm_data_type data_type,
                                       unsigned root, unsigned tag,
                                       struct communicator * comm,
                                       void * user_data,
                                       func_recv_callback scatterv_callback,
                                       enum yac_message_priority priority) {

  struct mpi_communicator * mpi_comm = (struct mpi_communicator*)comm;

  struct temp_scatterv_data * temp_data = malloc(1 * sizeof(*temp_data));
  temp_data->counts = malloc(mpi_comm->size * sizeof(*(temp_data->counts)));
  memcpy(temp_data->counts, counts,
         mpi_comm->size * sizeof(*(temp_data->counts)));
  temp_data->root = root;
  temp_data->user_data = user_data;
  temp_data->priority = convert_msg_prio(priority);
  temp_data->scatterv_callback = scatterv_callback;

  unsigned rel_rank = (mpi_comm->size + mpi_comm->rank - root) % mpi_comm->size;

  // if this is root
  if (mpi_comm->rank == root) {

    unsigned total_count = 0;

    for (unsigned i = 0; i < mpi_comm->size; ++i) total_count += counts[i];

    if (root != 0) {

      void * temp_send_buffer =
        malloc(total_count * get_data_type_size(data_type));

      unsigned offset = 0;
      for (unsigned i = 0; i < mpi_comm->rank; ++i) offset += counts[i];

      memcpy(temp_send_buffer, ((char*)send_buffer) + offset *
             get_data_type_size(data_type), (total_count - offset) *
             get_data_type_size(data_type));
      memcpy(((char*)temp_send_buffer) + (total_count - offset) *
             get_data_type_size(data_type), send_buffer, offset *
             get_data_type_size(data_type));

      send_buffer = temp_send_buffer;
    }

    struct comm_message temp_message;

    temp_message.buffer = send_buffer;
    temp_message.type = data_type;
    temp_message.count = total_count;
    temp_message.tag = tag;
    temp_message.dst_rank = root;
    temp_message.src_rank = root;

    recv_scatterv_callback(comm, temp_message, temp_data);

    if (root != 0) free(send_buffer);

  } else {

    // bit hack for Reset Lowest Set Bit
    unsigned rel_recv_rank = rel_rank & (rel_rank - 1);
    unsigned recv_rank = (rel_recv_rank + temp_data->root) % mpi_comm->size;
    unsigned step = MIN(rel_rank - rel_recv_rank, mpi_comm->size - rel_rank);
    unsigned recv_count = 0;

    for (unsigned i = 0; i < step; ++i)
      recv_count += counts[(mpi_comm->rank + i)%mpi_comm->size];

    comm_recv_callback_mpi(recv_count, data_type, recv_rank, tag, comm,
                           temp_data, recv_scatterv_callback, NULL,
                           convert_msg_prio(priority));
  }
}

static struct temp_alltoallv_data *
new_temp_alltoallv_data(unsigned num_ranks, unsigned local_rank,
                        unsigned dst_rank_mask,
                        enum yac_comm_data_type data_type,
                        enum mpi_message_priority priority,
                        func_recv_callback alltoallv_callback,
                        void * user_data) {

  struct temp_alltoallv_data * temp_data = malloc(1 * sizeof(*temp_data));

  temp_data->num_ranks = num_ranks;
  temp_data->local_rank = local_rank;
  temp_data->dst_rank_mask = dst_rank_mask;
  temp_data->buffer = NULL;
  temp_data->send_buffer = NULL;
  temp_data->tmp_buffer = NULL;
  temp_data->max_buffer_size = 0;
  temp_data->max_send_buffer_size = 0;
  temp_data->max_tmp_buffer_size = 0;
  temp_data->data_size_per_rank =
    calloc(num_ranks, sizeof(*(temp_data->data_size_per_rank)));
  temp_data->send_data_size_per_rank =
    calloc(num_ranks, sizeof(*(temp_data->send_data_size_per_rank)));
  temp_data->tmp_data_size_per_rank =
    calloc(num_ranks, sizeof(*(temp_data->tmp_data_size_per_rank)));
  temp_data->msg_data_size_per_rank =
    calloc(num_ranks, sizeof(*(temp_data->msg_data_size_per_rank)));
  temp_data->data_type = data_type;
  temp_data->priority = priority;
  temp_data->alltoallv_callback = alltoallv_callback;
  temp_data->user_data = user_data;
  temp_data->send_count = 1;

  return temp_data;
}

void delete_temp_alltoallv_data(struct temp_alltoallv_data * temp_data) {

  free(temp_data->buffer);
  free(temp_data->data_size_per_rank);
  free(temp_data->send_buffer);
  free(temp_data->send_data_size_per_rank);
  free(temp_data->tmp_buffer);
  free(temp_data->tmp_data_size_per_rank);
  free(temp_data->msg_data_size_per_rank);
  free(temp_data);
}

void merge_buffers(void * in_buffer_a, unsigned * in_size_per_package_a,
                   void * in_buffer_b, unsigned * in_size_per_package_b,
                   void ** out_buffer_a, unsigned * max_out_buffer_size_a,
                   unsigned * out_size_per_package_a,
                   void ** out_buffer_b, unsigned * max_out_buffer_size_b,
                   unsigned * out_size_per_package_b,
                   unsigned num_packages, unsigned even_in_out_a,
                   unsigned a_first) {

  char * in_buffer[2];
  unsigned * in_size_per_package[2];

  if (a_first) {
    in_buffer[0] = in_buffer_a;
    in_buffer[1] = in_buffer_b;
    in_size_per_package[0] = in_size_per_package_a;
    in_size_per_package[1] = in_size_per_package_b;
  } else {
    in_buffer[0] = in_buffer_b;
    in_buffer[1] = in_buffer_a;
    in_size_per_package[0] = in_size_per_package_b;
    in_size_per_package[1] = in_size_per_package_a;
  }

  char * out_buffer[2];
  unsigned * out_size_per_package[2];
  unsigned * max_out_buffer_size[2];

  if (even_in_out_a) {
    out_buffer[0] = *out_buffer_a;
    out_buffer[1] = *out_buffer_b;
    out_size_per_package[0] = out_size_per_package_a;
    out_size_per_package[1] = out_size_per_package_b;
    max_out_buffer_size[0] = max_out_buffer_size_a;
    max_out_buffer_size[1] = max_out_buffer_size_b;
  } else {
    out_buffer[0] = *out_buffer_b;
    out_buffer[1] = *out_buffer_a;
    out_size_per_package[0] = out_size_per_package_b;
    out_size_per_package[1] = out_size_per_package_a;
    max_out_buffer_size[0] = max_out_buffer_size_b;
    max_out_buffer_size[1] = max_out_buffer_size_a;
  }

  { // compute out buffer sizes and ensure that the out_buffers are big enough
    unsigned out_buffer_sizes[2] = {0,0};

    for (unsigned i = 0; i < num_packages; ++i) {
      out_buffer_sizes[i&1] += in_size_per_package[0][i];
      out_buffer_sizes[i&1] += in_size_per_package[1][i];
    }

    for (unsigned i = 0; i < 2; ++i) {

      char * tmp_out_buffer = out_buffer[i];
      ENSURE_ARRAY_SIZE(tmp_out_buffer, *max_out_buffer_size[i],
                        out_buffer_sizes[i]);
      out_buffer[i] = tmp_out_buffer;
    }
  }

  if (even_in_out_a) {
    *out_buffer_a = out_buffer[0];
    *out_buffer_b = out_buffer[1];
  } else {
    *out_buffer_b = out_buffer[0];
    *out_buffer_a = out_buffer[1];
  }

  // merge in buffers into out buffers
  for (unsigned i = 0; i < num_packages; ++i) {

    unsigned flag = i & 1;

    out_size_per_package[flag][i/2] = 0;

    for (unsigned j = 0; j < 2; ++j) {

      memcpy(out_buffer[flag], in_buffer[j], in_size_per_package[j][i]);
      out_size_per_package[flag][i/2] += in_size_per_package[j][i];
      out_buffer[flag] += in_size_per_package[j][i];
      in_buffer[j] += in_size_per_package[j][i];
    }
  }
}

static void send_data_alltoallv_callback(struct communicator * comm,
                                         struct comm_message msg,
                                         void * user_data) {

  struct temp_alltoallv_data * temp_data =
    (struct temp_alltoallv_data *)user_data;

  if (--(temp_data->send_count) == 0) {

    // deliver data
    struct comm_message tmp_msg = msg;

    tmp_msg.buffer = temp_data->buffer;
    tmp_msg.type = temp_data->data_type;
    tmp_msg.count = temp_data->data_size_per_rank[0] /
                    get_data_type_size(temp_data->data_type);
    tmp_msg.src_rank = yac_get_comm_rank(comm);
    tmp_msg.dst_rank = tmp_msg.src_rank;

    if (temp_data->alltoallv_callback != NULL)
      temp_data->alltoallv_callback(comm, tmp_msg, temp_data->user_data);

    // clean up
    delete_temp_alltoallv_data(temp_data);
  }
}

static void recv_data_alltoallv_callback(struct communicator * comm,
                                         struct comm_message msg,
                                         void * user_data);

static void recv_data_size_alltoallv_callback(struct communicator * comm,
                                              struct comm_message msg,
                                              void * user_data) {

  struct temp_alltoallv_data * temp_data =
    (struct temp_alltoallv_data *)user_data;

  memcpy(temp_data->msg_data_size_per_rank, msg.buffer,
         msg.count * get_data_type_size(msg.type));

  unsigned data_size = get_data_type_size(temp_data->data_type);
  unsigned total_count = 0;

  for (unsigned i = 0; i < msg.count; ++i)
    total_count += temp_data->msg_data_size_per_rank[i] / data_size;

  yac_comm_recv_callback(total_count, temp_data->data_type, msg.src_rank,
                         msg.tag, comm, temp_data, recv_data_alltoallv_callback,
                         NULL, temp_data->priority);
}

unsigned compute_orig_rank(unsigned logical_rank, unsigned comm_size,
                           unsigned max_mask) {

  unsigned mask = 1;
  while (mask < max_mask){
    if (logical_rank & mask) {
      unsigned rank = logical_rank ^ mask;
      if (rank >= comm_size) rank =
        compute_orig_rank(rank, comm_size, mask);
      if (rank < comm_size) return rank;
    }
    mask <<= 1;
  };

  return -1;
}

static void recv_data_alltoallv_callback(struct communicator * comm,
                                         struct comm_message msg,
                                         void * user_data) {

  struct temp_alltoallv_data * temp_data =
    (struct temp_alltoallv_data *)user_data;

  unsigned comm_size = yac_get_comm_size(comm);

  merge_buffers(temp_data->buffer, temp_data->data_size_per_rank,
                msg.buffer, temp_data->msg_data_size_per_rank,
                &(temp_data->tmp_buffer), &(temp_data->max_tmp_buffer_size),
                temp_data->tmp_data_size_per_rank,
                &(temp_data->send_buffer), &(temp_data->max_send_buffer_size),
                temp_data->send_data_size_per_rank,
                temp_data->num_ranks,
                !(temp_data->local_rank & temp_data->dst_rank_mask),
                msg.src_rank > msg.dst_rank);

  {
    {
      void * tmp = temp_data->buffer;
      temp_data->buffer = temp_data->tmp_buffer;
      temp_data->tmp_buffer = tmp;
    }
    {
      unsigned * tmp = temp_data->data_size_per_rank;
      temp_data->data_size_per_rank = temp_data->tmp_data_size_per_rank;
      temp_data->tmp_data_size_per_rank = tmp;
    }
    {
      unsigned tmp = temp_data->max_buffer_size;
      temp_data->max_buffer_size = temp_data->max_tmp_buffer_size;
      temp_data->max_tmp_buffer_size = tmp;
    }
  }

  temp_data->num_ranks >>= 1;

  // if only data of the local process is left
  if (temp_data->num_ranks == 0) {

    // this will reduce the send_count by one and call alltoallv_callback if all
    // messages have already been delivered or no messages had to be been sent
    send_data_alltoallv_callback(comm, msg, temp_data);

  } else {

    unsigned dst_rank_logical =
      temp_data->local_rank ^ temp_data->dst_rank_mask;
    temp_data->dst_rank_mask <<= 1;

    // a process with rank dst_rank might not exist, in that case compute the
    // process that mimics the respective rank
    unsigned dst_rank;
    if (dst_rank_logical >= comm_size)
      dst_rank =
        compute_orig_rank(dst_rank_logical, comm_size, temp_data->dst_rank_mask);
    else
      dst_rank = dst_rank_logical;

    // if the destination process is the same as the local one
    if (msg.dst_rank == dst_rank) {

      // the "process" does not yet exist, we have to create it
      struct temp_alltoallv_data * new_temp_data =
        new_temp_alltoallv_data(temp_data->num_ranks, dst_rank_logical,
                                temp_data->dst_rank_mask, temp_data->data_type,
                                temp_data->priority, NULL, NULL);

      struct comm_message tmp_msg = msg;

      tmp_msg.count = 0;
      for (unsigned i = 0; i < temp_data->num_ranks; ++i)
        tmp_msg.count += temp_data->send_data_size_per_rank[i];
      tmp_msg.buffer = temp_data->send_buffer;
      tmp_msg.src_rank = tmp_msg.dst_rank;
      {
        unsigned * tmp = temp_data->send_data_size_per_rank;
        temp_data->send_data_size_per_rank =
          new_temp_data->msg_data_size_per_rank;
        new_temp_data->msg_data_size_per_rank = tmp;
      }

      // call recv_data_alltoallv_callback for the "new" process
      recv_data_alltoallv_callback(comm, tmp_msg, new_temp_data);

      tmp_msg.count = 0;
      tmp_msg.buffer = NULL;
      memset(temp_data->msg_data_size_per_rank, 0,
             temp_data->num_ranks *
             sizeof(*(temp_data->msg_data_size_per_rank)));

      // call recv_data_alltoallv_callback for the "original" process
      recv_data_alltoallv_callback(comm, tmp_msg, temp_data);

    } else {

      // send data
      unsigned total_size = 0;
      for (unsigned i = 0; i < temp_data->num_ranks; ++i)
        total_size += temp_data->send_data_size_per_rank[i];
      void * send_buffer_copy = malloc(total_size);
      memcpy(send_buffer_copy, temp_data->send_buffer, total_size);
      total_size /= get_data_type_size(temp_data->data_type);
      yac_comm_bsend(temp_data->send_data_size_per_rank, temp_data->num_ranks,
                     COMM_UINT, dst_rank, msg.tag, comm);
      temp_data->send_count++;
      send_internal(send_buffer_copy, total_size, temp_data->data_type,
                    dst_rank, msg.tag, comm, temp_data,
                    send_data_alltoallv_callback, temp_data->priority, 1);

      // set up receive for 
      yac_comm_recv_callback(temp_data->num_ranks, COMM_UINT, dst_rank, msg.tag,
                             comm, temp_data, recv_data_size_alltoallv_callback,
                             NULL, temp_data->priority);
    }
  }
}

static void comm_alltoallv_callback_mpi(void * send_buffer, unsigned * counts,
                                        enum yac_comm_data_type data_type,
                                        unsigned tag, struct communicator * comm,
                                        void * user_data,
                                        func_recv_callback alltoallv_callback,
                                        enum yac_message_priority priority) {

  unsigned comm_size = yac_get_comm_size(comm);
  unsigned comm_rank = yac_get_comm_rank(comm);

  unsigned num_ranks = 1;
  while (num_ranks < comm_size) num_ranks <<= 1;

  struct temp_alltoallv_data * temp_data =
    new_temp_alltoallv_data(num_ranks, comm_rank, 1, data_type,
                            convert_msg_prio(priority), alltoallv_callback,
                            user_data);

  unsigned data_size = get_data_type_size(data_type);
  memcpy(temp_data->data_size_per_rank, counts, comm_size * sizeof(*counts));
  for (unsigned i = 0; i < comm_size; ++i)
    temp_data->data_size_per_rank[i] *= data_size;

  unsigned buffer_size = 0;
  for (unsigned i = 0; i < comm_size; ++i)
    buffer_size += counts[i];
  buffer_size *= data_size;
  temp_data->buffer = malloc(buffer_size);
  memcpy(temp_data->buffer, send_buffer, buffer_size);

  struct comm_message msg = {.buffer = NULL,
                             .type = data_type,
                             .count = 0,
                             .tag = tag,
                             .dst_rank = comm_rank,
                             .src_rank = comm_rank};

  recv_data_alltoallv_callback(comm, msg, temp_data);
}

static void recv_data_reduce_callback(struct communicator * comm,
                                      struct comm_message msg,
                                      void * user_data) {

  struct mpi_communicator * mpi_comm = (struct mpi_communicator*)comm;

  struct temp_reduce_data * temp_data = (struct temp_reduce_data*)user_data;

  if (msg.src_rank >= mpi_comm->size)
    msg.src_rank = -(msg.src_rank + 1);
  else if ((msg.src_rank != msg.dst_rank) && (msg.src_rank < mpi_comm->size))
    temp_data->op(msg.buffer, temp_data->buffer, msg.count, msg.type);

  unsigned step = (msg.src_rank - msg.dst_rank + mpi_comm->size)%mpi_comm->size;

  step <<= 1;
  if (step == 0) step = 1;

  unsigned rel_rank =
    (mpi_comm->size + mpi_comm->rank - temp_data->root) % mpi_comm->size;

  // if the data needs to be sent to another process
  if (rel_rank & step) {

    unsigned send_rank = ((rel_rank ^ step) + temp_data->root) % mpi_comm->size;

    comm_bsend_mpi(temp_data->buffer, msg.count, msg.type, send_rank,
                   msg.tag, comm);

  // if data from another process needs to be received
  } else if (step < mpi_comm->size) {

    unsigned recv_rank =
      (rel_rank + step + temp_data->root) % mpi_comm->size;
    unsigned rel_recv_rank = rel_rank + step;

    // in case the mpi_comm->size is not a multiple of two this can be true
    if (rel_recv_rank >= mpi_comm->size) {

      // we have to skip one step, which is done by faking the message
      msg.src_rank = -(recv_rank + 1);

      recv_data_reduce_callback(comm, msg, user_data);

    } else {
      comm_recv_callback_mpi(msg.count, msg.type, recv_rank, msg.tag, comm,
                             user_data, recv_data_reduce_callback, NULL,
                             temp_data->priority);
    }

    return;
  }

  msg.src_rank = mpi_comm->rank;
  msg.buffer = (mpi_comm->rank == temp_data->root)?temp_data->buffer:NULL;
  if (mpi_comm->rank != temp_data->root) msg.count = 0;

  if (temp_data->reduce_callback)
    temp_data->reduce_callback(comm, msg, temp_data->user_data);

  free(temp_data->buffer);
  free(temp_data);
}

static void comm_reduce_callback_mpi(void * buffer, unsigned count,
                                     enum yac_comm_data_type data_type,
                                     func_comm_op op, unsigned root,
                                     unsigned tag, struct communicator * comm,
                                     void * user_data,
                                     func_recv_callback reduce_callback,
                                     enum yac_message_priority priority) {

  struct mpi_communicator * mpi_comm = (struct mpi_communicator*)comm;

  struct temp_reduce_data * temp_data = malloc(1 * sizeof(*temp_data));
  temp_data->op = op;
  temp_data->root = root;
  temp_data->priority = convert_msg_prio(priority);
  unsigned buffer_size = count * get_data_type_size(data_type);
  temp_data->buffer = malloc(buffer_size);
  memcpy(temp_data->buffer, buffer, buffer_size);
  temp_data->user_data = user_data;
  temp_data->reduce_callback = reduce_callback;

  struct comm_message msg = {.buffer = temp_data->buffer,
                             .type = data_type,
                             .count = count,
                             .tag = tag,
                             .dst_rank = mpi_comm->rank,
                             .src_rank = mpi_comm->rank};

  recv_data_reduce_callback(comm, msg, temp_data);
}

static void recv_data_allreduce_callback(struct communicator * comm,
                                         struct comm_message msg,
                                         void * user_data) {

  struct temp_allreduce_data * temp_data =
    (struct temp_allreduce_data *)user_data;

  unsigned comm_size = yac_get_comm_size(comm);
  unsigned data_size = get_data_type_size(msg.type);

  if (msg.src_rank != msg.dst_rank)
    // apply reduce operation
    temp_data->op(msg.buffer, temp_data->buffer, msg.count, msg.type);

  // if reduction is finished
  if (temp_data->rank_mask >= comm_size) {

    msg.src_rank = msg.dst_rank;
    msg.buffer = temp_data->buffer;

    if (temp_data->allreduce_callback != NULL)
      temp_data->allreduce_callback(comm, msg, temp_data->user_data);

    free(temp_data->buffer);
    free(temp_data);

  } else {

    unsigned logical_exchange_rank =
      temp_data->local_rank ^ temp_data->rank_mask;
    temp_data->rank_mask <<= 1;

    // a process with rank logical_exchange_rank might not exist, in that case
    // compute the process that mimics the respective rank
    unsigned exchange_rank;
    if (logical_exchange_rank >= comm_size)
      exchange_rank = compute_orig_rank(logical_exchange_rank, comm_size,
                                        temp_data->rank_mask);
    else
      exchange_rank = logical_exchange_rank;

    // if the exchange rank is the same as the local on
    if (exchange_rank == msg.dst_rank) {

      // the "process" does not yet exist, we have to create it
      struct temp_allreduce_data * new_temp_data =
        malloc(1 * sizeof(*new_temp_data));
      new_temp_data->op = temp_data->op;
      new_temp_data->rank_mask = temp_data->rank_mask;
      new_temp_data->local_rank = logical_exchange_rank;
      new_temp_data->priority = temp_data->priority;
      unsigned buffer_size = msg.count * data_size;
      new_temp_data->buffer = malloc(buffer_size);
      memcpy(new_temp_data->buffer, temp_data->buffer, buffer_size);
      new_temp_data->user_data = NULL;
      new_temp_data->allreduce_callback = NULL;

      struct comm_message tmp_msg = msg;

      tmp_msg.src_rank = msg.dst_rank;
      tmp_msg.buffer = new_temp_data->buffer;

      // call recv_data_allreduce_callback for the "new" process
      recv_data_allreduce_callback(comm, tmp_msg, new_temp_data);

      tmp_msg.src_rank = tmp_msg.dst_rank;
      tmp_msg.buffer = temp_data->buffer;

      // call recv_data_allreduce_callback for the "original" process
      recv_data_allreduce_callback(comm, tmp_msg, temp_data);

    } else {

      yac_comm_bsend(temp_data->buffer, msg.count, msg.type, exchange_rank,
                     msg.tag, comm);
      yac_comm_recv_callback(msg.count, msg.type, exchange_rank, msg.tag, comm,
                             temp_data, recv_data_allreduce_callback, NULL,
                             temp_data->priority);
    }
  }
}

static void comm_allreduce_callback_mpi(void * buffer, unsigned count,
                                        enum yac_comm_data_type data_type,
                                        func_comm_op op, unsigned tag,
                                        struct communicator * comm,
                                        void * user_data,
                                        func_recv_callback allreduce_callback,
                                        enum yac_message_priority priority) {

  unsigned rank = yac_get_comm_rank(comm);
  struct temp_allreduce_data * temp_data = malloc(1 * sizeof(*temp_data));
  temp_data->op = op;
  temp_data->rank_mask = 1;
  temp_data->local_rank = rank;
  temp_data->priority = convert_msg_prio(priority);
  unsigned buffer_size = count * get_data_type_size(data_type);
  temp_data->buffer = malloc(buffer_size);
  memcpy(temp_data->buffer, buffer, buffer_size);
  temp_data->user_data = user_data;
  temp_data->allreduce_callback = allreduce_callback;

  struct comm_message msg = {.buffer = temp_data->buffer,
                             .type = data_type,
                             .count = count,
                             .tag = tag,
                             .dst_rank = rank,
                             .src_rank = rank};

  recv_data_allreduce_callback(comm, msg, temp_data);
}

static void send_data_barrier_callback(struct communicator * comm,
                                       struct comm_message msg,
                                       void * user_data) {

  struct temp_barrier_data * temp_data = (struct temp_barrier_data *)user_data;

  if (--(temp_data->send_count) == 0) {

    msg.src_rank = msg.dst_rank;

    if (temp_data->barrier_callback != NULL)
      temp_data->barrier_callback(comm, msg, temp_data->user_data);

    free(temp_data);
  }
}

static void recv_data_barrier_callback(struct communicator * comm,
                                       struct comm_message msg,
                                       void * user_data) {

  struct temp_barrier_data * temp_data = (struct temp_barrier_data *)user_data;

  unsigned comm_size = yac_get_comm_size(comm);

  if (temp_data->step >= comm_size) {

    // this will reduce the send_count by one and call barrier_callback if all
    // send messages have already been delivered
    send_data_barrier_callback(comm, msg, user_data);
    return;
  }

  unsigned dst_rank = (msg.dst_rank + temp_data->step) % comm_size;
  unsigned src_rank = (comm_size + msg.dst_rank - temp_data->step) % comm_size;
  temp_data->step <<= 1;
  temp_data->send_count++;

  send_internal(NULL, 0, COMM_BYTE, dst_rank, msg.tag, comm, user_data,
                send_data_barrier_callback, temp_data->priority, 0);
  yac_comm_recv_callback(0, COMM_BYTE, src_rank, msg.tag, comm, temp_data,
                         recv_data_barrier_callback, NULL, temp_data->priority);
}

static void comm_barrier_callback_mpi(unsigned tag, struct communicator * comm,
                                      void * user_data,
                                      func_recv_callback barrier_callback,
                                      enum yac_message_priority priority) {

  struct temp_barrier_data * temp_data = malloc(1 * sizeof(*temp_data));

  temp_data->user_data = user_data;
  temp_data->barrier_callback = barrier_callback;
  temp_data->priority = convert_msg_prio(priority);
  temp_data->step = 1;
  temp_data->send_count = 1;

  struct comm_message msg = {.buffer = NULL, .type = COMM_BYTE, .count = 0,
                             .tag = tag, .dst_rank = yac_get_comm_rank(comm),
                             .src_rank = yac_get_comm_rank(comm)};

  recv_data_barrier_callback(comm, msg, temp_data);
}

static void comm_check_mpi(struct communicator * comm) {

  struct mpi_communicator * mpi_comm = (struct mpi_communicator*)comm;

  check_open_requests(mpi_comm);
}

static void comm_finalize_mpi() {

   // ensure that all asynchronous receive operations are complete
   while (waitany_open_request_(MPI_COMM_WORLD));
   // free memory allocated statically in this file and reinitialize it
   free(global_requests_);
   free(global_message_ids_);
   free(global_requests);
   free(global_messages);
   free(msg_prio_stack);
   global_requests_ = NULL;
   global_message_ids_ = NULL;
   num_global_requests_ = 0;
   global_requests = NULL;
   max_priority = -1;
   comm_wait_is_active = 0;
   global_messages = NULL;
   num_global_messages = 0;
   msg_prio_stack = NULL;
   msg_prio_stack_size = 0;
   curr_msg_prio_stack_pos = -1;

   for (unsigned i = 0; i < curr_num_bsend_buffers; ++i)
      free(bsend_buffers[i]);
   for (unsigned i = 0; i < max_num_bsend_buffers; ++i)
      free(buffer_is_free[i]);

   free(bsend_buffers);
   free(bsend_buffer_sizes);
   free(buffer_is_free);
   bsend_buffers = NULL;
   bsend_buffer_sizes = NULL;
   buffer_is_free = NULL;
   curr_num_bsend_buffers = 0;
   max_num_bsend_buffers = 0;
   total_bsend_buffer_size = 0;

  /* -------------------------------------------------------------------
     Check MPI Finalization
     ------------------------------------------------------------------- */

   int is_finalised;

   mpi_err_handler(MPI_Finalized(&is_finalised), MPI_COMM_WORLD);

   if (!is_finalised && yac_initialised_MPI) {

   /* -------------------------------------------------------------------
      Only finalise here when initialised by yac
      ------------------------------------------------------------------- */

      int ierr = MPI_Finalize();
         
      if (ierr != MPI_SUCCESS ) {

         int err_len;
         char err_string[MPI_MAX_ERROR_STRING];
         MPI_Error_string(ierr, err_string, &err_len);
         printf("Error in MPI_Finalize %s \n", err_string );
      }
   }
}

#else

struct communicator * yac_mpi_comm_new(struct communicator * comm) {
   yac_internal_abort_message ( "ERROR: not implemented.", __FILE__, __LINE__ );
   return NULL;
}
#endif
