/**
 * @file component.c
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "utils.h"
#include "component.h"
#include "yacpar.h"
#include "xml/config_xml.h"
#include "communicator_local.h"

static int yac_components_new_first_call = 1;

// general component struct
struct component {

  // general component data

  char * name;
  unsigned int global_comp_id;
  struct communicator * communicator;
  unsigned io_proc_step;
};

static unsigned compute_io_proc_step(unsigned comm_size) {

  unsigned num_procs_per_node;

  // estimate the number of processes per node
  if (comm_size % 36 == 0) {
    num_procs_per_node = 36;
  } else if (comm_size % 24 == 0) {
    num_procs_per_node = 24;
  } else if (comm_size % 16 == 0) {
    num_procs_per_node = 16;
  } else if (comm_size % 12 == 0) {
    num_procs_per_node = 12;
  } else if (comm_size % 8 == 0) {
    num_procs_per_node = 8;
  } else if (comm_size % 6 == 0) {
    num_procs_per_node = 6;
  } else if (comm_size % 4 == 0) {
    num_procs_per_node = 4;
  } else if (comm_size % 2 == 0) {
    num_procs_per_node = 2;
  } else {
    num_procs_per_node = 1;
  }

  unsigned num_nodes = comm_size / num_procs_per_node;
  unsigned num_io_procs = MIN((unsigned)sqrt((float)num_nodes), 64);
  return ((num_nodes + num_io_procs - 1) / num_io_procs) * num_procs_per_node;
}

void yac_components_new(char const ** names, unsigned num_names,
                        struct component ** comps) {

   if (yac_components_new_first_call) yac_components_new_first_call = 0;
   else yac_internal_abort_message(
     "ERROR: yac_component_new can only be called once", __FILE__, __LINE__ );

#ifdef USE_XML2
  int total_num_components = yac_xml_get_number_of_components();
#else
  int total_num_components = num_names;
#endif

  unsigned * global_comp_ids = malloc(num_names * sizeof(*global_comp_ids));
  int * idx = malloc(num_names * sizeof(*idx));

  // for all names get component ids from xml
  for (int i = 0; i < num_names; ++i) {
#ifdef USE_XML2
    global_comp_ids[i] = yac_xml_get_component_id_by_component_name(names[i]);
#else
    global_comp_ids[i] = yac_hash(name);
#endif
    if (global_comp_ids[i] == (unsigned)-1) {
      fprintf (stderr, "Check def_comp call and xml file for %s!\n", names[i]);
      yac_internal_abort_message(
        "ERROR(yac_components_new): mismatch in component name",
        __FILE__, __LINE__);
    }
    idx[i] = i;
  }

  yac_quicksort_index((int*)global_comp_ids, num_names, idx);

  // check for duplicated component ids
  for (int i = 1; i < num_names; ++i) {
    if (global_comp_ids[i-1] == global_comp_ids[i])
      yac_internal_abort_message(
        "ERROR(yac_components_new): duplicated component ids",
        __FILE__, __LINE__);
  }

  // for all components in xml
  for (int i = 0; i < total_num_components; ++i) {

#ifdef USE_XML2
    int curr_comp_id = yac_xml_get_component_id(i);
#else
    int curr_comp_id = global_comp_ids[i];
#endif

    int curr_comp_idx;
    for (curr_comp_idx = 0; curr_comp_idx < num_names; ++curr_comp_idx)
      if (global_comp_ids[curr_comp_idx] == curr_comp_id) break;

    // if the local process has the current component
    if (curr_comp_idx < num_names) {

      struct component * comp = malloc(1 * sizeof(*comp));

      char const * curr_comp_name = names[idx[curr_comp_idx]];
      comp->name = malloc(strlen(curr_comp_name) + 1);
      strcpy(comp->name, curr_comp_name);
      
      comp->global_comp_id = curr_comp_id;

      if (yac_is_local_comm(yac_world_comm))
        yac_named_local_comm_new(yac_world_comm, curr_comp_name);
       else
        comp->communicator = yac_comm_split(yac_world_comm, 0, 0);
      comps[idx[curr_comp_idx]] = comp;

    } else {

      if (!yac_is_local_comm(yac_world_comm)) {
        struct communicator * dummy_comm = yac_comm_split(yac_world_comm, 1, 0);
        yac_free_communicator(dummy_comm);
      }
    }
  }

  for (int i = 0; i < num_names; ++i)
    comps[i]->io_proc_step =
      compute_io_proc_step(yac_get_comm_size(comps[i]->communicator));

  free(idx);
  free(global_comp_ids);
}

struct component * yac_component_new(char const * name) {

  struct component * comp;

  yac_components_new(&name, 1, &comp);

  return comp;
}

const char * yac_component_get_name(struct component * comp) {

   if (comp == NULL)
      yac_internal_abort_message("ERROR: invalid component pointer",
                                 __FILE__, __LINE__ );

  return comp->name;
}

unsigned int yac_component_get_global_id(struct component * comp) {

   if (comp == NULL)
      yac_internal_abort_message("ERROR: invalid component pointer",
                                 __FILE__, __LINE__ );

   return comp->global_comp_id;
}

struct communicator * yac_component_get_communicator(struct component * comp) {

   if (comp == NULL)
      yac_internal_abort_message("ERROR: invalid component pointer",
                                 __FILE__, __LINE__ );

   return comp->communicator;
}

void yac_component_set_io_proc_step(struct component * comp,
                                    unsigned io_proc_step) {

   if (comp == NULL)
      yac_internal_abort_message("ERROR: invalid component pointer",
                                 __FILE__, __LINE__ );

   if (io_proc_step == 0)
      yac_internal_abort_message("ERROR: invalid io_proc_step",
                                 __FILE__, __LINE__ );

   comp->io_proc_step = io_proc_step;
}

unsigned yac_component_get_io_proc_step(struct component * comp) {

   if (comp == NULL)
      yac_internal_abort_message("ERROR: invalid component pointer",
                                 __FILE__, __LINE__ );

   return comp->io_proc_step;
}

void yac_component_delete(struct component * comp) {

   if (comp == NULL)
      yac_internal_abort_message("ERROR: invalid component pointer",
                                 __FILE__, __LINE__ );

  free(comp->name);
  yac_free_communicator(comp->communicator);
  free(comp);
}

void yac_components_new_first_call_reset() {
  yac_components_new_first_call = 1;
}
