/**
 * @file couple_config.c
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "couple_config.h"
#include "utils.h"
#include "communicator.h"
#include "domain.h"
#include "config_xml.h"
#include "yac_interface.h"
#include "event.h"
#include "yacpar.h"
#include "fields.h"
#include "component.h"
#include "global_search.h"
#include "xml/interpolation_enum.h"
#include "interpolation_method.h"
#include "interpolation_method_avg.h"
#include "interpolation_method_nnn.h"
#include "interpolation_method_fixed.h"
#include "interpolation_method_conserv.h"
#include "interpolation_method_patch.h"
#include "interpolation_method_patch_smooth.h"
#include "interpolation_method_spmap.h"
#include "interpolation_method_file.h"
#include "interpolation_method_check.h"
#include "field_interpolation_gen.h"
#include "field_interpolation_gen_src.h"
#include "field_interpolation_gen_tgt.h"
#include "field_interpolation_gen_src_tgt.h"
#include "ensure_array_size.h"

#ifdef USE_XML2
#define MAX_FILE_NAME_LENGHT (512)
#define MAX_ROUTINE_NAME_LENGTH (256)

enum couple_status {
  INITIALISED  = 0,
  FINALIZED    = 1,
  FREED        = 2
};

enum transient_type {

  SRC = 1,
  TGT = 2
};

/**
 * contains all information of a transient
 */
struct transient {

  struct coupling_field * field;
  struct global_search * search;
  enum transient_type type;

  struct field_interpolation_gen * interpolation_gen;

  int couple_idx;    //!< couple index for XML access
  int transient_idx; //!< transient index for XML access

  union {
    struct {
      union { // - every interpolation type needs to have the entry type as first
              //   entry in order to be able to compare them using memcmp
              // - no pointers are allowed here in order to be comparable
        struct {
          enum yac_interpolation_list type;
        } general;
        struct {
          enum yac_interpolation_list type;
          unsigned distance_weighted;
          unsigned partial_coverage;
          unsigned points_idx;
        } average;
        struct {
          enum yac_interpolation_list type;
          int nn;
          unsigned distance_weighted;
          unsigned points_idx;
        } n_nearest_neighbor;
        struct {
          enum yac_interpolation_list type;
          int enforced_conserv;
          int partial_cov;
          enum yac_interpolation_method_conserv_normalisation normalisation;
          unsigned points_idx;
        } conservative;
        struct {
          enum yac_interpolation_list type;
          unsigned p_order;
          unsigned g_order;
          unsigned extension;
          unsigned extrapolation;
          unsigned points_idx;
        } patch_recovery;
        struct {
          enum yac_interpolation_list type;
          unsigned p_order;
          unsigned g_order;
          double corner_weight;
          unsigned extension;
          unsigned extrapolation;
          unsigned points_idx;
        } smooth_patch_recovery;
        struct {
          enum yac_interpolation_list type;
          unsigned points_idx;
        } spmap;
        struct {
          enum yac_interpolation_list type;
          char filename[MAX_FILE_NAME_LENGHT];
          unsigned io_proc_step;
        } user_file;
        struct {
          enum yac_interpolation_list type;
          double value;
        } fixed;
        struct {
          enum yac_interpolation_list type;
          char constructor_key[MAX_ROUTINE_NAME_LENGTH];
          char do_search_key[MAX_ROUTINE_NAME_LENGTH];
        } check;
      } * interpolation_methods;
      int num_interpolations;
      struct interpolation_method ** interpolation_method_stack;
      int use_source_mask;
      const char * weight_file_name;
    } src;

    struct {
      enum yac_field_interpolation_tgt_mask_handling mask_handling;
      unsigned points_idx;
    } tgt;
  } data;

  int mapping_on_source;

  int event_id;

  int field_interpolation_config_idx; //!< used internally to identify identical
                                      //!< configurations
};

/**
 * contains all coupling information on a specific domain pair
 */
struct domain_pair_couple {

  struct domain * local_domain;
  char * remote_domain_name;

  struct global_search * search;

  int local_is_source, local_is_target;

  struct transient * transients;
  unsigned num_transients;
};

/**
 * contains all coupling information on a specific component pair
 */
struct component_pair_couple {

  struct component * local_component;
  const char* remote_component_name;

  struct communicator * comm;
  unsigned free_comm;

  struct domain_pair_couple * domain_pairs;
  unsigned num_domain_pairs;

  struct transient * transients;
  unsigned num_transients;
};

struct temp_field_interpolation_config_idx_data {

  struct component_pair_couple * comp_pair;
  unsigned tag;
};

struct global_search_information {

  char const * component_names[2];
  char const * domain_names[2];
  unsigned comp_pair_idx, domain_pair_idx;
};

/**
 * global coupling information
 */
static struct {

  enum couple_status status;

  struct component_pair_couple * component_pairs;
  unsigned num_component_pairs;

} couple_config = {.status = FREED,
                   .component_pairs = NULL,
                   .num_component_pairs = 0};

static void check_status(enum couple_status desired_status) {

  if (couple_config.status != desired_status) {

    char const * msg_template = "ERROR: couple_config status is %d, "
                                "but should be %d\n";
    char msg[sizeof(msg_template)];
    sprintf(msg, msg_template, couple_config.status, desired_status);
    yac_internal_abort_message(msg, __FILE__, __LINE__);
    exit(EXIT_FAILURE);
  }
}

static void check_status_or(enum couple_status desired_status_a,
                            enum couple_status desired_status_b) {

  if (couple_config.status != desired_status_a &&
      couple_config.status != desired_status_b) {

    char const * msg_template = "ERROR: couple_config status is %d, "
                                "but should be %d or %d\n";
    char msg[sizeof(msg_template)];
    sprintf(msg, msg_template, couple_config.status, desired_status_a,
            desired_status_b);
    yac_internal_abort_message(msg, __FILE__, __LINE__);
    exit(EXIT_FAILURE);
  }
}

static struct interpolation_method **
generate_interpolation_method_stack(
  struct global_search * search, char const * method_stack_name,
  struct transient * transient) {

  struct interpolation_method ** interpolation_method_stack =
    malloc((transient->data.src.num_interpolations + 1) *
           sizeof(*interpolation_method_stack));
  interpolation_method_stack[transient->data.src.num_interpolations] = NULL;

  struct points *** points = yac_get_coupling_field_pointsets(transient->field);
  struct mask *** masks = yac_get_coupling_field_masks(transient->field);
  unsigned num_pointsets =
    yac_get_coupling_field_num_pointsets(transient->field);

  for (int i = 0; i < transient->data.src.num_interpolations; ++i) {

    struct interpolation_method * interp_method;

    switch (transient->data.src.interpolation_methods[i].general.type) {
      case (AVERAGE):
      {
        interp_method =
          yac_new_interpolation_method_avg(
            search, yac_get_coupling_field_domain(transient->field),
            method_stack_name, points, masks, num_pointsets,
            transient->data.src.interpolation_methods[i]
                          .average.points_idx,
            transient->data.src.interpolation_methods[i]
                          .average.distance_weighted,
            transient->data.src.interpolation_methods[i]
              .average.partial_coverage);
        break;
      }
      case (N_NEAREST_NEIGHBOR):
      {
        interp_method =
          yac_new_interpolation_method_nnn_mask(
            search, yac_get_coupling_field_domain(transient->field),
            method_stack_name, points, masks, num_pointsets,
            transient->data.src.interpolation_methods[i]
              .n_nearest_neighbor.distance_weighted,
            transient->data.src.interpolation_methods[i]
              .n_nearest_neighbor.nn,
            transient->data.src.interpolation_methods[i]
              .n_nearest_neighbor.points_idx,
            transient->data.src.use_source_mask);
        break;
      }
      case (CONSERVATIVE):
      {
        interp_method =
          yac_new_interpolation_method_conserv(
            search, yac_get_coupling_field_domain(transient->field),
            method_stack_name, points, masks, num_pointsets,
            transient->data.src.interpolation_methods[i]
              .conservative.points_idx,
            transient->data.src.interpolation_methods[i]
              .conservative.enforced_conserv,
            transient->data.src.interpolation_methods[i]
              .conservative.partial_cov,
            transient->data.src.interpolation_methods[i]
              .conservative.normalisation);
        break;
      }
      case (PATCH_RECOVERY):
      {
        interp_method =
          yac_new_interpolation_method_patch_mask(
            search, yac_get_coupling_field_domain(transient->field),
            method_stack_name, points, masks, num_pointsets,
            transient->data.src.interpolation_methods[i]
              .patch_recovery.p_order,
            transient->data.src.interpolation_methods[i]
              .patch_recovery.g_order,
            transient->data.src.interpolation_methods[i]
              .patch_recovery.extension,
            transient->data.src.interpolation_methods[i]
              .patch_recovery.extrapolation,
            transient->data.src.interpolation_methods[i]
              .patch_recovery.points_idx,
            transient->data.src.use_source_mask);
        break;
      }
      case (SMOOTH_PATCH_RECOVERY):
      {
        interp_method =
          yac_new_interpolation_method_patch_smooth_mask(
            search, yac_get_coupling_field_domain(transient->field),
            method_stack_name, points, masks, num_pointsets,
            transient->data.src.interpolation_methods[i]
              .smooth_patch_recovery.p_order,
            transient->data.src.interpolation_methods[i]
              .smooth_patch_recovery.g_order,
            transient->data.src.interpolation_methods[i]
              .smooth_patch_recovery.corner_weight,
            transient->data.src.interpolation_methods[i]
              .smooth_patch_recovery.extension,
            transient->data.src.interpolation_methods[i]
              .smooth_patch_recovery.extrapolation,
            transient->data.src.interpolation_methods[i]
              .smooth_patch_recovery.points_idx,
            transient->data.src.use_source_mask);
        break;
      }
      case (SOURCE_TO_TARGET_MAP):
      {
        interp_method =
          yac_new_interpolation_method_spmap_mask(
              search, yac_get_coupling_field_domain(transient->field),
              method_stack_name, points, masks, num_pointsets,
              transient->data.src.interpolation_methods[i]
                .spmap.points_idx, transient->data.src.use_source_mask);
        break;
      }
      case (USER_FILE):
      {
        enum yac_location * locations =
          malloc(num_pointsets * sizeof(*locations));

        for (unsigned i = 0; i < num_pointsets; ++i)
          locations[i] =
            (points != NULL && points[0] != NULL && points[0][i] != NULL)?
              (points[0][i]->location):(LOC_UNDEFINED);

        interp_method =
          yac_new_interpolation_method_file(
            search, yac_get_coupling_field_domain(transient->field),
            method_stack_name, locations, num_pointsets,
            transient->data.src.interpolation_methods[i].user_file.filename,
            transient->data.src.interpolation_methods[i].user_file.io_proc_step);

        free(locations);
        break;
      }
      case (FIXED_VALUE):
      {
        interp_method = yac_new_interpolation_method_fixed(
          transient->data.src.interpolation_methods[i].fixed.value, search);
        break;
      }
      case (CHECK):
      {
        func_constructor constructor_callback = NULL;
        void * constructor_user_data = NULL;
        func_do_search do_search_callback = NULL;
        void * do_search_user_data = NULL;
        if (transient->data.src.interpolation_methods[i].
              check.constructor_key[0] != '\0') {
          yac_interpolation_method_get_constructor_callback(
            transient->data.src.interpolation_methods[i].check.constructor_key,
            &constructor_callback, &constructor_user_data);
        }
        if (transient->data.src.interpolation_methods[i].
              check.do_search_key[0] != '\0') {
          yac_interpolation_method_get_do_search_callback(
            transient->data.src.interpolation_methods[i].check.do_search_key,
            &do_search_callback, &do_search_user_data);
        }
        interp_method = yac_new_interpolation_method_check(
          constructor_callback, constructor_user_data, do_search_callback,
          do_search_user_data, search);
        break;
      }
      default:
        break;
    };

    interpolation_method_stack[i] = interp_method;
  }

  return interpolation_method_stack;
}

static
void generate_transient_src(struct transient * transient) {

  int num_interpolations =
    yac_xml_get_number_of_interpolations(transient->couple_idx,
                                         transient->transient_idx);

  transient->data.src.interpolation_methods =
    calloc(num_interpolations,
           sizeof(*(transient->data.src.interpolation_methods)));

  for (int i = 0; i < num_interpolations; ++i) {

    transient->data.src.interpolation_methods[i].general.type =
      yac_xml_get_interpolation_method(transient->couple_idx,
                                       transient->transient_idx, i);

    if (transient->data.src.interpolation_methods[i].general.type == UNDEFINED)
      yac_internal_abort_message(
        "ERROR: couple config: unsupported or undefined interpolation in xml "
        "file!", __FILE__, __LINE__ );

    switch (transient->data.src.interpolation_methods[i].general.type) {

      case (AVERAGE): {
        unsigned partial_cov = 0;

        if (yac_xml_interpolation_has_parameter(transient->couple_idx,
                                                transient->transient_idx,
                                                i, "partial_coverage")) {

          const char * partial_cov_str =
            yac_xml_get_interpolation_parameter_value_string(
              transient->couple_idx, transient->transient_idx, i,
              "partial_coverage");

          if (!strcmp(partial_cov_str, "true") ||
              !strcmp(partial_cov_str, "True") ||
              !strcmp(partial_cov_str, "TRUE")) {

            partial_cov = 1;
          } else if (!(!strcmp(partial_cov_str, "false") ||
                       !strcmp(partial_cov_str, "False") ||
                       !strcmp(partial_cov_str, "FALSE"))) {

            yac_internal_abort_message(
              "ERROR: invalid value for average remapping XML parameter "
              "partial_coverage", __FILE__, __LINE__);
          }

          yac_xml_free_xml_string(partial_cov_str);
        }

        unsigned distance_weighted = 0;

        if (yac_xml_interpolation_has_parameter(transient->couple_idx,
                                                transient->transient_idx,
                                                i, "weighted")) {
            const char * weighted_str =
              yac_xml_get_interpolation_parameter_value_string(
                transient->couple_idx, transient->transient_idx, i,
                "weighted");

            if (!strcmp(weighted_str, "distance_weighted") ||
                !strcmp(weighted_str, "DISTANCE_WEIGHTED")) {

                distance_weighted = 1;
            }
            yac_xml_free_xml_string(weighted_str);
        }

        transient->data.src.interpolation_methods[i].average.partial_coverage =
          partial_cov;
        transient->data.src.interpolation_methods[i].
          average.distance_weighted = distance_weighted;
        transient->data.src.interpolation_methods[i].
          average.points_idx = 0; //! \todo read points_idx from XML

        break;
      }
      case (N_NEAREST_NEIGHBOR): {
        int nn;

        if (yac_xml_interpolation_has_parameter(transient->couple_idx,
                                                transient->transient_idx, i,
                                                "n") == 0)
          yac_internal_abort_message(
            "ERROR: start_search:  N_NEAREST_NEIGHBOR requires n!",
            __FILE__, __LINE__ );

        nn =
          yac_xml_get_interpolation_parameter_value_int(
            transient->couple_idx, transient->transient_idx, i, "n" );

        unsigned distance_weighted = 0;

        if (yac_xml_interpolation_has_parameter(transient->couple_idx,
                                                transient->transient_idx,
                                                i, "weighted")) {

            const char * weighted_str =
              yac_xml_get_interpolation_parameter_value_string(
                transient->couple_idx, transient->transient_idx, i,
                "weighted");

            if (!strcmp(weighted_str, "distance_weighted") ||
                !strcmp(weighted_str, "DISTANCE_WEIGHTED")) {

                distance_weighted = 1;
            }

            yac_xml_free_xml_string(weighted_str);
        }

        transient->data.src.interpolation_methods[i].n_nearest_neighbor.nn = nn;
        transient->data.src.interpolation_methods[i].
          n_nearest_neighbor.distance_weighted = distance_weighted;
        transient->data.src.interpolation_methods[i].
          n_nearest_neighbor.points_idx = 0; //! \todo read points_idx from XML
        break;
      }
      case (CONSERVATIVE): {
        int enforced_conserv = 0;
        int partial_cov = 0;
        enum yac_interpolation_method_conserv_normalisation
          normalisation = CONSERV_DESTAREA;

        if (yac_xml_interpolation_has_parameter(transient->couple_idx,
                                                transient->transient_idx,
                                                i, "enforced_conservation")) {
          const char * enforced_conserv_str =
            yac_xml_get_interpolation_parameter_value_string(
              transient->couple_idx, transient->transient_idx, i,
              "enforced_conservation");
          if (!strcmp(enforced_conserv_str, "true") ||
              !strcmp(enforced_conserv_str, "True") ||
              !strcmp(enforced_conserv_str, "TRUE")) {

            enforced_conserv = 1;
          } else if (!(!strcmp(enforced_conserv_str, "false") ||
                       !strcmp(enforced_conserv_str, "False") ||
                       !strcmp(enforced_conserv_str, "FALSE"))) {

            yac_internal_abort_message(
              "ERROR: invalid value for conservative remapping XML parameter "
              "enforced_conservation", __FILE__, __LINE__);
          }

          yac_xml_free_xml_string(enforced_conserv_str);
        }

        if (yac_xml_interpolation_has_parameter(transient->couple_idx,
                                                transient->transient_idx,
                                                i, "partial_coverage")) {

          const char * partial_cov_str =
            yac_xml_get_interpolation_parameter_value_string(
              transient->couple_idx, transient->transient_idx, i,
              "partial_coverage");

          if (!strcmp(partial_cov_str, "true") ||
              !strcmp(partial_cov_str, "True") ||
              !strcmp(partial_cov_str, "TRUE")) {

            partial_cov = 1;
          } else if (!(!strcmp(partial_cov_str, "false") ||
                       !strcmp(partial_cov_str, "False") ||
                       !strcmp(partial_cov_str, "FALSE"))) {

            yac_internal_abort_message(
              "ERROR: invalid value for conservative remapping XML parameter "
              "partial_coverage", __FILE__, __LINE__);
          }

          yac_xml_free_xml_string(partial_cov_str);
        }

        if (yac_xml_interpolation_has_parameter(transient->couple_idx,
                                                transient->transient_idx, i,
                                                "normalisation")) {

          const char * normalisation_str =
            yac_xml_get_interpolation_parameter_value_string(
              transient->couple_idx, transient->transient_idx, i,
              "normalisation");

          if (!strcmp(normalisation_str, "fracarea") ||
              !strcmp(normalisation_str, "FRACAREA")) {

            normalisation = CONSERV_FRACAREA;

          } else if (!(!strcmp(normalisation_str, "destarea") ||
                       !strcmp(normalisation_str, "DESTAREA"))) {

            yac_internal_abort_message(
              "ERROR: invalid value for conservative remapping XML parameter "
              "normalisation", __FILE__, __LINE__);
          }

          yac_xml_free_xml_string(normalisation_str);
        }

        transient->data.src.interpolation_methods[i]
          .conservative.enforced_conserv = enforced_conserv;
        transient->data.src.interpolation_methods[i]
          .conservative.partial_cov = partial_cov;
        transient->data.src.interpolation_methods[i]
          .conservative.normalisation = normalisation;
        transient->data.src.interpolation_methods[i].
          conservative.points_idx = 0; //! \todo read points_idx from XML
        break;
      }
      case (PATCH_RECOVERY):
      {
        unsigned p_order;
        unsigned g_order;
        unsigned extension;
        unsigned extrapolation;

        if (yac_xml_interpolation_has_parameter(transient->couple_idx,
                                                transient->transient_idx, i,
                                                "polynomial_order" ) == 1)
          p_order =
            yac_xml_get_interpolation_parameter_value_int(
              transient->couple_idx, transient->transient_idx, i,
              "polynomial_order" );
        else
          p_order = 1;

        // if gauss_order is zero we work on nodes else we have to get the gauss points
        if (yac_xml_interpolation_has_parameter(transient->couple_idx,
                                                transient->transient_idx, i,
                                                "gauss_order" ) == 1)
          g_order =
            yac_xml_get_interpolation_parameter_value_int(
              transient->couple_idx, transient->transient_idx, i,
              "gauss_order");
        else
          g_order = 1;

        if (yac_xml_interpolation_has_parameter(transient->couple_idx,
                                                transient->transient_idx, i,
                                                "extend_source_patch" ) == 1)
          extension =
            yac_xml_get_interpolation_parameter_value_int(
              transient->couple_idx, transient->transient_idx, i,
              "extend_source_patch" );
        else
          extension = 1;

        extrapolation = 0;

        if (yac_xml_interpolation_has_parameter(transient->couple_idx,
                                                transient->transient_idx, i,
                                                "allow_extrapolation" ) == 1) {
          const char * extrapolation_str =
            yac_xml_get_interpolation_parameter_value_string(
              transient->couple_idx, transient->transient_idx, i,
              "allow_extrapolation" );

          if (!strcmp(extrapolation_str, "true") ||
              !strcmp(extrapolation_str, "True") ||
              !strcmp(extrapolation_str, "TRUE"))
            extrapolation = 1;

          yac_xml_free_xml_string(extrapolation_str);
        }

        transient->data.src.interpolation_methods[i]
          .patch_recovery.p_order = p_order;
        transient->data.src.interpolation_methods[i]
          .patch_recovery.g_order = g_order;
        transient->data.src.interpolation_methods[i]
          .patch_recovery.extension = extension;
        transient->data.src.interpolation_methods[i]
          .patch_recovery.extrapolation = extrapolation;
        transient->data.src.interpolation_methods[i].
          patch_recovery.points_idx = 0; //! \todo read points_idx from XML
        break;
      }
      case (SMOOTH_PATCH_RECOVERY):
      {
        unsigned p_order;
        unsigned g_order;
        unsigned extension;
        unsigned extrapolation;
        double corner_weight;

        if (yac_xml_interpolation_has_parameter(transient->couple_idx,
                                                transient->transient_idx, i,
                                                "polynomial_order" ) == 1)
          p_order =
            yac_xml_get_interpolation_parameter_value_int(
              transient->couple_idx, transient->transient_idx, i,
              "polynomial_order" );
        else
          p_order = 1;

        // if gauss_order is zero we work on nodes else we have to get the gauss points

        if (yac_xml_interpolation_has_parameter(transient->couple_idx,
                                                transient->transient_idx, i,
                                                "gauss_order" ) == 1)
          g_order =
            yac_xml_get_interpolation_parameter_value_int(
              transient->couple_idx, transient->transient_idx, i,
              "gauss_order");
        else
          g_order = 1;

        // get paramter to shift computation from target corner to target center
        if (yac_xml_interpolation_has_parameter(transient->couple_idx,
                                                transient->transient_idx, i,
                                                "corner_weight" ) == 1)
          corner_weight =
            yac_xml_get_interpolation_parameter_value_double(
              transient->couple_idx, transient->transient_idx, i,
              "corner_weight");
        else
          corner_weight = 1.0;

        if (yac_xml_interpolation_has_parameter(transient->couple_idx,
                                                transient->transient_idx, i,
                                                "extend_source_patch" ) == 1)
          extension =
            yac_xml_get_interpolation_parameter_value_int(
              transient->couple_idx, transient->transient_idx, i,
              "extend_source_patch" );
        else
          extension = 1;

        extrapolation = 0;

        if (yac_xml_interpolation_has_parameter(transient->couple_idx,
                                                transient->transient_idx, i,
                                                "allow_extrapolation" ) == 1) {
          const char * extrapolation_str =
            yac_xml_get_interpolation_parameter_value_string(
              transient->couple_idx, transient->transient_idx, i,
              "allow_extrapolation" );

          if (!strcmp(extrapolation_str, "true") ||
              !strcmp(extrapolation_str, "True") ||
              !strcmp(extrapolation_str, "TRUE"))
            extrapolation = 1;

          yac_xml_free_xml_string(extrapolation_str);
        }

        transient->data.src.interpolation_methods[i]
          .smooth_patch_recovery.p_order = p_order;
        transient->data.src.interpolation_methods[i]
          .smooth_patch_recovery.g_order = g_order;
        transient->data.src.interpolation_methods[i]
          .smooth_patch_recovery.corner_weight = corner_weight;
        transient->data.src.interpolation_methods[i]
          .smooth_patch_recovery.extension = extension;
        transient->data.src.interpolation_methods[i]
          .smooth_patch_recovery.extrapolation = extrapolation;
        break;
      }
      case (SOURCE_TO_TARGET_MAP):
      {
        break;
      }
      case (FIXED_VALUE):
      {
        double fixed_value;

        if (yac_xml_interpolation_has_parameter(transient->couple_idx,
                                                transient->transient_idx, i,
                                                "user_value" ) == 0)
          yac_internal_abort_message(
            "ERROR: start_search: FIXED_VALUE requires a value defined by the "
            "user!", __FILE__, __LINE__);

        fixed_value =
          yac_xml_get_interpolation_parameter_value_double(
            transient->couple_idx, transient->transient_idx, i, "user_value" );

        transient->data.src.interpolation_methods[i].fixed.value = fixed_value;
        break;
      }
      case (USER_FILE):
      {

        /* get the filename containing the weights */

        if (yac_xml_interpolation_has_parameter(transient->couple_idx,
                                                 transient->transient_idx, i,
                                                 "filename" ) == 0)
          yac_internal_abort_message(
            "ERROR: start_search: USER_FILE requires a filename containing "
            "weights defined by the user!", __FILE__, __LINE__);

        char const * filename =
          yac_xml_get_interpolation_parameter_value_string(
            transient->couple_idx, transient->transient_idx, i, "filename" );

        if (strlen(filename) + 1 > MAX_FILE_NAME_LENGHT)
          yac_internal_abort_message("ERROR: USER_FILE filename is too long!",
                                     __FILE__, __LINE__);
        strcpy(transient->data.src.interpolation_methods[i].user_file.filename,
               filename);

        yac_xml_free_xml_string(filename);

        transient->data.src.interpolation_methods[i].user_file.io_proc_step =
          yac_component_get_io_proc_step(
            yac_get_coupling_field_component(transient->field));

        break;
      }
      case (CHECK):
      {
        if (yac_xml_interpolation_has_parameter(transient->couple_idx,
                                                transient->transient_idx,
                                                i, "constructor_key") != 0) {

          char const * constructor_key =
            yac_xml_get_interpolation_parameter_value_string(
              transient->couple_idx, transient->transient_idx, i,
              "constructor_key");

          if (strlen(constructor_key) + 1 > MAX_ROUTINE_NAME_LENGTH)
            yac_internal_abort_message(
              "ERROR: CHECK constructor_key is too long!", __FILE__, __LINE__);
          strcpy(transient->data.src.interpolation_methods[i].
                   check.constructor_key, constructor_key);

          yac_xml_free_xml_string(constructor_key);
        }

        if (yac_xml_interpolation_has_parameter(transient->couple_idx,
                                                transient->transient_idx,
                                                i, "do_search_key") != 0) {

          char const * do_search_key =
            yac_xml_get_interpolation_parameter_value_string(
              transient->couple_idx, transient->transient_idx, i,
              "do_search_key");

          if (strlen(do_search_key) + 1 > MAX_ROUTINE_NAME_LENGTH)
            yac_internal_abort_message(
              "ERROR: CHECK do_search_key is too long!", __FILE__, __LINE__);
          strcpy(transient->data.src.interpolation_methods[i].
                   check.do_search_key, do_search_key);

          yac_xml_free_xml_string(do_search_key);
        }

        break;
      }
      default:
        yac_internal_abort_message(
          "ERROR: start_search: unknown interpolation!", __FILE__, __LINE__ );
    };
  }

  transient->data.src.num_interpolations = num_interpolations;

  transient->data.src.interpolation_method_stack = NULL;

  // use_source_mask is currently required. If the xml validates
  // against the xsd this check is not required.
  if (yac_xml_use_source_mask_is_set(transient->couple_idx,
                                     transient->transient_idx))
    transient->data.src.use_source_mask =
      yac_xml_get_use_source_mask(transient->couple_idx,
                                  transient->transient_idx);
  else
    transient->data.src.use_source_mask = 1; // default

  if (yac_xml_enforce_write_weight_file(transient->couple_idx,
                                        transient->transient_idx)) {

    transient->data.src.weight_file_name =
      yac_xml_get_weight_file_name(transient->couple_idx,
                                   transient->transient_idx);

  } else {
    transient->data.src.weight_file_name = NULL;
  }
}

static
void generate_transient_tgt(struct transient * transient) {

  int use_target_mask = 0;

  // use_target_mask is currently required. If the xml validates
  // against the xsd this check is not required.
  if (yac_xml_use_target_mask_is_set(transient->couple_idx,
                                     transient->transient_idx))
    use_target_mask = yac_xml_get_use_target_mask(transient->couple_idx,
                                                  transient->transient_idx);

  transient->data.tgt.mask_handling =
    (use_target_mask)?TGT_NO_INTERP:TGT_IGNORE_MASK;
  transient->data.tgt.points_idx = 0; //! \todo read points_idx from XML
}

static
void generate_transient(struct domain_pair_couple * domain_pair,
                        enum transient_type type, int couple_idx,
                        int transient_idx, int local_global_component_id,
                        unsigned num_fields,
                        struct coupling_field * coupling_fields[num_fields]) {

  struct transient * curr_transient = NULL;

  if (type == SRC)
    domain_pair->local_is_source = 1;
  else if (type == TGT)
    domain_pair->local_is_target = 1;
  else
    yac_internal_abort_message(
      "ERROR(generate_transient): invalid transient_type", __FILE__, __LINE__);

  domain_pair->transients =
    realloc(domain_pair->transients,
            (domain_pair->num_transients + 1) *
            sizeof(*(domain_pair->transients)));
  curr_transient = domain_pair->transients +
                   domain_pair->num_transients;
  domain_pair->num_transients++;

  const char * transient_name =
    yac_xml_get_transient_standard_name_by_transient_id(
      yac_xml_get_transient_couple_transient_id(couple_idx, transient_idx));

  curr_transient->field = NULL;

  for (int i = 0; i < num_fields; ++i) {

    struct coupling_field * field = coupling_fields[i];

    if ((local_global_component_id ==
         yac_get_coupling_field_global_comp_id(field)) &&
         !strcmp(transient_name,
                 yac_get_coupling_field_name(field))) {
      curr_transient->field = field;
      break;
    }
  }

  if (curr_transient->field == NULL) {
    printf ("no corresponding field for %s \n", transient_name );
    yac_internal_abort_message(
      "ERROR: could not find field for current transient\n",
      __FILE__, __LINE__);
  }
  int local_comp_id;
  int transient_grid_ref_id;

  if ( type == SRC ) {
    local_comp_id =
      yac_xml_get_source_component_id(couple_idx, transient_idx);
    transient_grid_ref_id =
      yac_xml_get_source_transient_grid_ref_id(couple_idx, transient_idx);
  }
  else if ( type == TGT ) {
    local_comp_id =
      yac_xml_get_target_component_id(couple_idx, transient_idx);
    transient_grid_ref_id =
      yac_xml_get_target_transient_grid_ref_id(couple_idx, transient_idx);
  }
  else
    yac_internal_abort_message(
      "ERROR in xml configuration file: unsupported transient unit",
      __FILE__, __LINE__);

  unsigned collection_size =
    yac_xml_get_transient_collection_size(local_comp_id, transient_grid_ref_id);

  yac_set_coupling_field_collection_size ( curr_transient->field , collection_size );

  curr_transient->search = NULL;
  curr_transient->couple_idx = couple_idx;
  curr_transient->transient_idx = transient_idx;
  curr_transient->type = type;
  curr_transient->interpolation_gen = NULL;

  enum yac_reduction_type reduction_operation;

  switch (yac_xml_get_coupling_period_operation(couple_idx, transient_idx)) {
    case (T_AVERAGE):
      reduction_operation = TIME_AVERAGE;
      break;
    case (T_ACCUMULATE):
      reduction_operation = ACCUMULATE;
      break;
    default:
      reduction_operation = INSTANT;
      break;
  };

  enum yac_time_unit_type time_unit;

  switch ( yac_xml_get_timestep_unit() ) {
    case (T_MILLISECOND):
      time_unit = C_MILLISECOND;
      break;
    case (T_SECOND):
      time_unit = C_SECOND;
      break;
    case (T_MINUTE):
      time_unit = C_MINUTE;
      break;
    case (T_HOUR):
      time_unit = C_HOUR;
      break;
    case (T_DAY):
      time_unit = C_DAY;
      break;
    case (T_MONTH):
      time_unit = C_MONTH;
      break;
    case (T_YEAR):
      time_unit = C_YEAR;
      break;
    case (T_ISO_FORMAT):
      time_unit = C_ISO_FORMAT;
      break;
    default:
      time_unit = 0;
      yac_internal_abort_message(
        "ERROR in xml configuration file: unsupported time unit",
        __FILE__, __LINE__ );
  };

  // make sure that the calendar is defined and consistent

  calendarType set_calendar = getCalendarType ();
  calendarType my_calendar  = PROLEPTIC_GREGORIAN; //< \todo get calendar type from XML.

  if ( set_calendar == CALENDAR_NOT_SET )
    initCalendar(my_calendar);
  else if ( set_calendar != my_calendar )
    yac_internal_abort_message(
      "ERROR: different calendars in use! Reconfigure model setup.\n",
      __FILE__, __LINE__);

  // where is the mapping to be executed in the put/get-operation
  curr_transient->mapping_on_source =
    yac_xml_mapping_on_source(couple_idx, transient_idx);

  // generate an event

  yac_event_init(&(curr_transient->event_id));

  int timelag;
  void *timestepPtr;
  void *periodPtr;

  if ( time_unit == C_ISO_FORMAT) {

      char *timestep;
      char *period;

      if ( curr_transient->type == SRC ) {
          timestep = yac_xml_get_source_isotimestep(couple_idx, transient_idx);
          timelag  = yac_xml_get_source_timelag(couple_idx, transient_idx);
      } else {
          timestep = yac_xml_get_target_isotimestep(couple_idx, transient_idx);
          timelag  = yac_xml_get_target_timelag(couple_idx, transient_idx);
      }

      period = yac_xml_get_isocoupling_period(couple_idx, transient_idx);

      timestepPtr = timestep;
      periodPtr = period;


      yac_event_add(curr_transient->event_id, timestepPtr, periodPtr, time_unit,
                    timelag, reduction_operation,
                    yac_get_start_datetime(), yac_get_end_datetime(), NULL );

  } else {

      int timestep;
      int period;

      if ( curr_transient->type == SRC ) {
          timestep = yac_xml_get_source_timestep(couple_idx, transient_idx);
          timelag  = yac_xml_get_source_timelag(couple_idx, transient_idx);
      } else {
          timestep = yac_xml_get_target_timestep(couple_idx, transient_idx);
          timelag  = yac_xml_get_target_timelag(couple_idx, transient_idx);
      }

      period = yac_xml_get_coupling_period(couple_idx, transient_idx);

      timestepPtr = &timestep;
      periodPtr = &period;

      yac_event_add(curr_transient->event_id, timestepPtr, periodPtr, time_unit,
                    timelag, reduction_operation,
                    yac_get_start_datetime(), yac_get_end_datetime(), NULL );

  };

  yac_xml_free_xml_string(transient_name);

  if ( type == SRC )
    generate_transient_src(curr_transient);
  else if ( type == TGT )
    generate_transient_tgt(curr_transient);
}

static struct domain *
get_local_domain_(const char * domain_name, int global_component_id,
                  unsigned num_fields,
                  struct coupling_field * coupling_fields[num_fields]) {

  for (unsigned i = 0; i < num_fields; ++i) {

    struct coupling_field * field = coupling_fields[i];

    if (global_component_id != yac_get_coupling_field_global_comp_id(field))
      continue;

    struct domain * domain = yac_get_coupling_field_domain(field);

    if (strcmp(domain_name, yac_get_domain_name(domain)) == 0)
      return domain;
  }

  return NULL;
}

static
int compare_transients_by_config(const void * a, const void * b) {

  struct transient * t_a = (struct transient *)a,
                   * t_b = (struct transient *)b;

  //MoHa: I belief it is better for the performance if the target fields are
  //      initialised first.
  //      This is because they to not produce much work, but have to answer the
  //      source fields.

  if (t_a->type != t_b->type)
    return (t_a->type == TGT)?-1:1;

  // if the fields are different
  if (t_a->field != t_b->field) {

    struct domain * domains[2] = {yac_get_coupling_field_domain(t_a->field),
                                  yac_get_coupling_field_domain(t_b->field)};

    // compare domains
    {
      if (domains[0] != domains[1]) {
        int ret = strcmp(yac_get_domain_name(domains[0]),
                         yac_get_domain_name(domains[1]));
        if (ret) return ret;
        else
          yac_internal_abort_message(
            "ERROR: different domains have same name.",
            __FILE__, __LINE__);
      }
    }

    if ((yac_get_num_subdomains(domains[0]) == 0) ||
        (yac_get_num_subdomains(domains[1]) == 0))
      yac_internal_abort_message(
        "ERROR: a domain used in an interpolation has no subdomain.",
        __FILE__, __LINE__);


    struct points *** points[2] =
      {yac_get_coupling_field_pointsets(t_a->field),
       yac_get_coupling_field_pointsets(t_b->field)};
    struct mask *** masks[2] =
      {yac_get_coupling_field_masks(t_a->field),
      yac_get_coupling_field_masks(t_b->field)};
    unsigned num_pointsets[2] =
      {yac_get_coupling_field_num_pointsets(t_a->field),
       yac_get_coupling_field_num_pointsets(t_b->field)};

    if ((num_pointsets[0] == 0) || (num_pointsets[1] == 0))
      yac_internal_abort_message(
        "ERROR: transient used in an interpolation has no pointsets.",
        __FILE__, __LINE__);

    // if both transients have different numbers of pointsets
    if (num_pointsets[0] != num_pointsets[1])
      return (int)num_pointsets[0] - (int)num_pointsets[1];

    // if points are defined for only one transient
    if ((points[0] == NULL) ^ (points[1] == NULL))
      return (points[0] == NULL) - (points[1] == NULL);

    if ((points[0] != NULL) && (points[0][0] != NULL) &&
        (points[1] != NULL) && (points[1][0] != NULL))
      for (unsigned i = 0; i < num_pointsets[0]; ++i)
        if (points[0][0][i]->unique_id != points[1][0][i]->unique_id)
          return points[0][0][i]->unique_id - points[1][0][i]->unique_id;

    int compare_masks;

    switch(t_a->type) {

      case(SRC):

        if (t_a->data.src.use_source_mask != t_b->data.src.use_source_mask)
          return (int)t_a->data.src.use_source_mask -
                 (int)t_b->data.src.use_source_mask;

        compare_masks = t_a->data.src.use_source_mask;
        break;

      case(TGT):

        if (t_a->data.tgt.points_idx != t_b->data.tgt.points_idx)
          return (int)t_a->data.tgt.points_idx -
                 (int)t_b->data.tgt.points_idx;

        if (t_a->data.tgt.mask_handling != t_b->data.tgt.mask_handling)
          return (int)t_a->data.tgt.mask_handling -
                 (int)t_b->data.tgt.mask_handling;

        compare_masks = (t_a->data.tgt.mask_handling == TGT_NO_INTERP);
        break;

      default:
        yac_internal_abort_message("ERROR: ", __FILE__, __LINE__);
    };

    if (compare_masks) {
      // if mask is defined for only one transient
      if ((masks[0] == NULL) ^ (masks[1] == NULL))
        return (masks[0] == NULL) - (masks[1] == NULL);

      if ((masks[0] != NULL) && (masks[0][0] != NULL) &&
          (masks[1] != NULL) && (masks[1][0] != NULL))
        for (unsigned i = 0; i < num_pointsets[0]; ++i)
          if (masks[0][0][i]->unique_id != masks[1][0][i]->unique_id)
            return masks[0][0][i]->unique_id - masks[1][0][i]->unique_id;
    }
  }

  if (t_a->type == SRC) {

    if (t_a->data.src.num_interpolations != t_b->data.src.num_interpolations)
      return (int)t_a->data.src.num_interpolations -
             (int)t_b->data.src.num_interpolations;

    {
      // we can use memcmp because interpolation_methods is initialised with
      // zero and every entry in interpolation_methods contains the
      // interpolation type
      int cmp = memcmp(t_a->data.src.interpolation_methods,
                       t_b->data.src.interpolation_methods,
                       t_a->data.src.num_interpolations *
                       sizeof(*(t_a->data.src.interpolation_methods)));

      if (cmp) return cmp;
    }
  }

  return 0;
}

static
int compare_global_search_info_by_config(const void * a, const void * b) {

  struct global_search_information * a_ = (struct global_search_information *)a,
                                   * b_ = (struct global_search_information *)b;

  int       ret = strcmp(a_->component_names[0], b_->component_names[0]);
  if (!ret) ret = strcmp(a_->component_names[1], b_->component_names[1]);
  if (!ret) ret = strcmp(a_->domain_names[0], b_->domain_names[0]);
  if (!ret) ret = strcmp(a_->domain_names[1], b_->domain_names[1]);

  return ret;
}

static
int compare_transients_by_index(const void * a, const void * b) {

  struct transient * t_a = (struct transient *)a,
                   * t_b = (struct transient *)b;

  if (t_a->couple_idx != t_b->couple_idx)
    return t_a->couple_idx - t_b->couple_idx;
  if (t_a->transient_idx != t_b->transient_idx)
    return t_a->transient_idx - t_b->transient_idx;
  if (t_a->type != t_b->type)
    return t_a->type - t_b->type;

  return 0;
}

static
void generate_domain_pairs(struct component_pair_couple * comp_pair,
                           int couple_idx, int local_global_component_id,
                           unsigned num_fields,
                           struct coupling_field * coupling_fields[num_fields]) {

  int num_transients = yac_xml_get_number_of_transient_couples(couple_idx);

  for (int transient_idx = 0; transient_idx < num_transients;
       ++transient_idx) {

    int source_comp_id = yac_xml_get_source_component_id(couple_idx, transient_idx);
    int target_comp_id = yac_xml_get_target_component_id(couple_idx, transient_idx);

    if (source_comp_id == target_comp_id)
      yac_internal_abort_message(
        "ERROR: source and target component are the same for a transient\n",
        __FILE__, __LINE__);

    enum transient_type transient_type;

    int remote_global_component_id;

    if (source_comp_id == local_global_component_id) {
      transient_type = SRC;
      remote_global_component_id = target_comp_id;
    } else if (target_comp_id == local_global_component_id) {
      transient_type = TGT;
      remote_global_component_id = source_comp_id;
    } else {
      yac_internal_abort_message(
        "ERROR: local component is neither source nor target\n",
        __FILE__, __LINE__);
      exit(EXIT_FAILURE);
    }

    int source_transient_grid_ref_id =
      yac_xml_get_source_transient_grid_ref_id(couple_idx, transient_idx);
    int source_grid_id =
      yac_xml_get_grid_id_by_transient_grid_ref_id_by_component_id(source_comp_id,
      source_transient_grid_ref_id);
    const char * source_grid_name =
      yac_xml_get_grid_alias_name_by_grid_id(source_grid_id);
    int target_transient_grid_ref_id =
      yac_xml_get_target_transient_grid_ref_id(couple_idx, transient_idx);
    int target_grid_id =
      yac_xml_get_grid_id_by_transient_grid_ref_id_by_component_id(target_comp_id,
      target_transient_grid_ref_id);
    const char * target_grid_name =
      yac_xml_get_grid_alias_name_by_grid_id(target_grid_id);

    if (MAX(strlen(source_grid_name), strlen(target_grid_name)) > 1023)
      yac_internal_abort_message(
        "ERROR: grid name is too long\n", __FILE__, __LINE__);

    char const * local_domain_name;
    char const * remote_domain_name;

    if (transient_type == SRC) {

      local_domain_name = source_grid_name;
      remote_domain_name = target_grid_name;
    } else {
      local_domain_name = target_grid_name;
      remote_domain_name = source_grid_name;
    }

    int found = 0;

    for (int i = 0; i < comp_pair->num_domain_pairs; ++i) {

      char const * curr_local_domain_name =
        yac_get_domain_name(comp_pair->domain_pairs[i].local_domain);
      char const * curr_remote_domain_name =
        comp_pair->domain_pairs[i].remote_domain_name;

      if (strcmp(local_domain_name, curr_local_domain_name) == 0 &&
          strcmp(remote_domain_name, curr_remote_domain_name) == 0) {

        generate_transient(comp_pair->domain_pairs + i, transient_type,
                           couple_idx, transient_idx, local_global_component_id,
                           num_fields, coupling_fields);
        found = 1;
      }
    }

    if (found) {

      yac_xml_free_xml_string(target_grid_name);
      yac_xml_free_xml_string(source_grid_name);
      continue;
    }

    struct domain * local_domain =
      get_local_domain_(local_domain_name, local_global_component_id,
                        num_fields, coupling_fields);

    if (local_domain == NULL) {
      fprintf ( stderr, "Search for domain name %s or %s in the following list failed.\n",
                local_domain_name, remote_domain_name );
      for (unsigned i = 0; i < num_fields; ++i) {
        struct domain * domain = yac_get_coupling_field_domain(coupling_fields[i]);
        fprintf ( stderr, "- Field %d domain name: %s\n", i, yac_get_domain_name(domain) );
      }
      fprintf ( stderr, "The domain name was generated from component and grid name.\n");
      fprintf ( stderr, "Thus, either your domain name or grid name is inconsistent.\n");

      yac_internal_abort_message(
        "ERROR: could not find matching local domain\n", __FILE__, __LINE__);
    }

    comp_pair->domain_pairs =
      realloc(comp_pair->domain_pairs,
              (comp_pair->num_domain_pairs + 1) *
              sizeof(*(comp_pair->domain_pairs)));

    struct domain_pair_couple * curr_domain_pair =
      comp_pair->domain_pairs + comp_pair->num_domain_pairs;

    comp_pair->num_domain_pairs++;

    curr_domain_pair->local_domain = local_domain;
    curr_domain_pair->remote_domain_name = malloc((strlen(remote_domain_name) + 1));
    strcpy(curr_domain_pair->remote_domain_name, remote_domain_name);
    curr_domain_pair->search = NULL;
    curr_domain_pair->transients = NULL;
    curr_domain_pair->num_transients = 0;

    curr_domain_pair->local_is_source = 0;
    curr_domain_pair->local_is_target = 0;

    generate_transient(curr_domain_pair, transient_type,
                       couple_idx, transient_idx, local_global_component_id,
                       num_fields, coupling_fields);

    if (transient_type == SRC)
      transient_type = TGT;
    else
      transient_type = SRC;

    yac_xml_free_xml_string(target_grid_name);
    yac_xml_free_xml_string(source_grid_name);
  }
}

static struct component *
get_local_component(int global_comp_id, unsigned num_components,
                    struct component * components[num_components]) {

  for (unsigned i = 0; i < num_components; ++i)
    if (yac_component_get_global_id(components[i]) == global_comp_id)
      return components[i];

  return NULL;
}

static
int check_comp_pair(struct communicator * comm,
                    struct component * local_component[2]) {

  unsigned flags = (local_component[0] != NULL) |
                   ((local_component[1] != NULL) << 1);
  unsigned comm_size = yac_get_comm_size(comm);
  unsigned * all_flags = malloc(comm_size * sizeof(*all_flags));

  yac_comm_allgather(&flags, 1, all_flags, 1, COMM_UINT, comm);

  for (unsigned i = 0; i < comm_size; ++i)
    flags |= all_flags[i];

  free(all_flags);

  return flags != 3;
}

static
void generate_component_pairs(unsigned num_components,
                              struct component * components[num_components],
                              unsigned num_fields,
                              struct coupling_field * coupling_fields[num_fields]) {

  int num_couples = yac_xml_get_number_of_couples();

  couple_config.component_pairs =
    malloc(2 * num_couples * sizeof(*(couple_config.component_pairs)));
  couple_config.num_component_pairs = 0;

  struct {
    int local, remote;
  } * global_component_ids = malloc(2 * num_couples * sizeof(*global_component_ids));

  for (int couple_idx = 0; couple_idx < num_couples; ++couple_idx) {

    int global_comp_ids[2] = {yac_xml_get_first_component_id(couple_idx),
                              yac_xml_get_second_component_id(couple_idx)};

    if (global_comp_ids[0] == global_comp_ids[1])
      yac_internal_abort_message(
        "ERROR: both global component ids are identical\n",
        __FILE__, __LINE__);

    int found = 0;

    for (int i = 0; i < couple_config.num_component_pairs; ++i) {
      if (global_component_ids[i].local  == global_comp_ids[0] &&
          global_component_ids[i].remote == global_comp_ids[1]) {

        generate_domain_pairs(couple_config.component_pairs + i, couple_idx,
                              global_component_ids[i].local,
                              num_fields, coupling_fields);
        found = 1;
      } else if (global_component_ids[i].remote == global_comp_ids[0] &&
                 global_component_ids[i].local  == global_comp_ids[1]) {

        generate_domain_pairs(couple_config.component_pairs + i, couple_idx,
                              global_component_ids[i].local,
                              num_fields, coupling_fields);
        found = 1;
      }
    }

    if (found) continue;

    // check if the components are available locally

    struct component * local_component[2] =
      {get_local_component(global_comp_ids[0], num_components, components),
       get_local_component(global_comp_ids[1], num_components, components)};

    struct communicator * comm;

    comm = yac_comm_split(yac_world_comm, (local_component[0] != NULL ||
                          local_component[1] != NULL)?1:-1, 0);

    if ((local_component[0] == NULL && local_component[1] == NULL) ||
        check_comp_pair(comm, local_component)) {

      // none of the components is available locally or one of the components
      // is missing
      // -> we do not need a component pair for this
      yac_free_communicator(comm);
      continue;
    }

    unsigned free_comm = 1;

    for (int i = 0; i < 2; ++i) {

      if (local_component[i] == NULL) continue;

      struct component_pair_couple * curr_comp_pair =
        couple_config.component_pairs + couple_config.num_component_pairs;
      global_component_ids[couple_config.num_component_pairs].local =
        global_comp_ids[i];
      global_component_ids[couple_config.num_component_pairs].remote =
        global_comp_ids[i^1];

      couple_config.num_component_pairs++;

      curr_comp_pair->local_component = local_component[i];
      curr_comp_pair->remote_component_name =
        yac_xml_get_component_name_by_component_id(global_comp_ids[i^1]);
      curr_comp_pair->free_comm = free_comm;
      free_comm = 0;
      curr_comp_pair->comm = comm;
      curr_comp_pair->domain_pairs = NULL;
      curr_comp_pair->num_domain_pairs = 0;
      curr_comp_pair->transients = NULL;
      curr_comp_pair->num_transients = 0;

      generate_domain_pairs(curr_comp_pair, couple_idx, global_comp_ids[i],
                            num_fields, coupling_fields);
    }
  }

  free(global_component_ids);

  couple_config.component_pairs =
    realloc(couple_config.component_pairs,
            couple_config.num_component_pairs *
            sizeof(*(couple_config.component_pairs)));
}

static
void generate_global_search_information() {

  unsigned num_global_search_info = 0;

  for (int component_pair_idx = 0;
       component_pair_idx < couple_config.num_component_pairs;
       ++component_pair_idx)
      num_global_search_info +=
        couple_config.component_pairs[component_pair_idx].num_domain_pairs;

  struct global_search_information * global_search_info =
    malloc(num_global_search_info * sizeof(*global_search_info));

  for (int component_pair_idx = 0, global_search_idx = 0;
       component_pair_idx < couple_config.num_component_pairs;
       ++component_pair_idx) {

    struct component_pair_couple * curr_comp_pair =
      couple_config.component_pairs + component_pair_idx;

    int comp_order =
      strcmp(yac_component_get_name(curr_comp_pair->local_component),
             curr_comp_pair->remote_component_name);

    for (int domain_pair_idx = 0;
         domain_pair_idx < curr_comp_pair->num_domain_pairs;
         ++domain_pair_idx, ++global_search_idx) {

      struct domain_pair_couple * curr_domain_pair =
        curr_comp_pair->domain_pairs + domain_pair_idx;

      int order =
        ((comp_order)?
           (comp_order):
           (strcmp(yac_get_domain_name(curr_domain_pair->local_domain),
                   curr_domain_pair->remote_domain_name))) > 0;

      struct global_search_information * curr_global_search_info =
        global_search_info + global_search_idx;

      curr_global_search_info->component_names[order] =
        yac_component_get_name(curr_comp_pair->local_component);
      curr_global_search_info->component_names[order^1] =
        curr_comp_pair->remote_component_name;
      curr_global_search_info->domain_names[order] =
        yac_get_domain_name(curr_domain_pair->local_domain);
      curr_global_search_info->domain_names[order^1] =
        curr_domain_pair->remote_domain_name;
      curr_global_search_info->comp_pair_idx = component_pair_idx;
      curr_global_search_info->domain_pair_idx = domain_pair_idx;
    }
  }

  qsort((void*)global_search_info, num_global_search_info,
        sizeof(*global_search_info), compare_global_search_info_by_config);

  // find unique global search infos and initialise respective global search
  for (unsigned global_search_idx = 0;
       global_search_idx < num_global_search_info; ++global_search_idx) {

    // if this configuration is unique
    if ((global_search_idx + 1 == num_global_search_info) ||
        compare_global_search_info_by_config(
          global_search_info + global_search_idx,
          global_search_info + global_search_idx + 1)) {

      struct component_pair_couple * curr_comp_pair =
        couple_config.component_pairs +
        global_search_info[global_search_idx].comp_pair_idx;
      struct domain_pair_couple * curr_domain_pair =
        curr_comp_pair->domain_pairs +
        global_search_info[global_search_idx].domain_pair_idx;

      curr_domain_pair->search =
        yac_global_search_new(curr_domain_pair->local_domain,
                              curr_domain_pair->remote_domain_name,
                              curr_comp_pair->comm,
                              curr_domain_pair->local_is_source,
                              curr_domain_pair->local_is_target,
                              yac_component_get_name(
                                curr_comp_pair->local_component),
                              curr_comp_pair->remote_component_name);

    // the configuration is not unique -> both domains are available locally
    } else {

      struct component_pair_couple * curr_comp_pairs[2] =
        {couple_config.component_pairs +
         global_search_info[global_search_idx].comp_pair_idx,
         couple_config.component_pairs +
         global_search_info[global_search_idx + 1].comp_pair_idx};
      struct domain_pair_couple * curr_domain_pairs[2] =
        {curr_comp_pairs[0]->domain_pairs +
         global_search_info[global_search_idx].domain_pair_idx,
         curr_comp_pairs[1]->domain_pairs +
         global_search_info[global_search_idx + 1].domain_pair_idx};

      struct domain * domains[2] = {curr_domain_pairs[0]->local_domain,
                                    curr_domain_pairs[1]->local_domain};
      unsigned domain_is_source[2] = {curr_domain_pairs[0]->local_is_source,
                                      curr_domain_pairs[1]->local_is_source};
      char const * component_names[2] =
        {yac_component_get_name(curr_comp_pairs[0]->local_component),
         yac_component_get_name(curr_comp_pairs[1]->local_component)};
      struct global_search * searches[2];

      yac_global_search_all_local_new(domains, curr_comp_pairs[0]->comm,
                                      domain_is_source, component_names,
                                      searches);
      curr_domain_pairs[0]->search = searches[0];
      curr_domain_pairs[1]->search = searches[1];

      // skip the next
      ++global_search_idx;
    }
  }

  free(global_search_info);

  for (int component_pair_idx = 0;
       component_pair_idx < couple_config.num_component_pairs;
       ++component_pair_idx) {

    struct component_pair_couple * curr_comp_pair =
      couple_config.component_pairs + component_pair_idx;

    for (int domain_pair_idx = 0;
         domain_pair_idx < curr_comp_pair->num_domain_pairs;
         ++domain_pair_idx) {

      struct domain_pair_couple * curr_domain_pair =
        curr_comp_pair->domain_pairs + domain_pair_idx;

      yac_enddef_global_search(curr_domain_pair->search);
    }
  }
}

struct local_remote_index {

  int local_field_interpolation_config_idx;
  int remote_field_interpolation_config_idx;
  unsigned idx;
};

static int
compare_local_remote_index(const void * a, const void * b) {

  struct local_remote_index * i_a, * i_b;

  i_a = (struct local_remote_index *)a;
  i_b = (struct local_remote_index *)b;

  if (i_a->local_field_interpolation_config_idx !=
      i_b->local_field_interpolation_config_idx)
    return i_a->local_field_interpolation_config_idx -
           i_b->local_field_interpolation_config_idx;

  return i_a->remote_field_interpolation_config_idx -
         i_b->remote_field_interpolation_config_idx;
}

static void
bcast_field_interpolation_config_idx_callback (struct communicator * comm,
                                               struct comm_message msg,
                                               void * user_data) {

  struct component_pair_couple * comp_pair =
    (struct component_pair_couple*)user_data;
  struct transient * transients = comp_pair->transients;
  unsigned num_transients = comp_pair->num_transients;

  int * recv_buffer = (int *)(msg.buffer);

  if (num_transients > 0) {

    // consistency check
    for (unsigned j = 0; j < num_transients; ++j) {

      if ((recv_buffer[4*j+0] != transients[j].couple_idx) ||
          (recv_buffer[4*j+1] != transients[j].transient_idx) ||
          (recv_buffer[4*j+2] == transients[j].type)) {

        yac_internal_abort_message(
          "ERROR(recv_field_interpolation_config_idx_callback): remote "
          "component does not have the same setup of coupling fields",
          __FILE__, __LINE__);
      }
    }

    { // use local and remote field_interpolation_config_idx to determine
      // global results
      struct local_remote_index * local_remote_index;

      local_remote_index = malloc(num_transients * sizeof(*local_remote_index));

      for (unsigned i = 0; i < num_transients; ++i) {

        local_remote_index[i].local_field_interpolation_config_idx =
          transients[i].field_interpolation_config_idx;
        local_remote_index[i].remote_field_interpolation_config_idx =
          recv_buffer[4*i+3];
        local_remote_index[i].idx = i;
      }

      qsort(local_remote_index, num_transients, sizeof(*local_remote_index),
            compare_local_remote_index);

      {
        int idx = 0;

        transients[local_remote_index[0].idx].field_interpolation_config_idx =
          idx;

        for (unsigned i = 1; i < num_transients; ++i) {

          if (compare_local_remote_index(local_remote_index + i,
                                         local_remote_index + i - 1)) idx++;

          transients[local_remote_index[i].idx].field_interpolation_config_idx =
            idx;
        }
      }
      free(local_remote_index);
    }
  }
}

static void
recv_field_interpolation_config_idx_callback (struct communicator * comm,
                                              struct comm_message msg,
                                              void * user_data) {

  struct temp_field_interpolation_config_idx_data * temp_data =
    (struct temp_field_interpolation_config_idx_data *)user_data;
  struct component_pair_couple * comp_pair = temp_data->comp_pair;
  struct communicator * comp_comm =
    yac_component_get_communicator(comp_pair->local_component);

  // bcast field_interpolation_config_idx's from the other component
  yac_comm_bcast_callback(msg.buffer, msg.count, msg.type,
                          yac_get_comm_rank(comp_comm), temp_data->tag,
                          comp_comm, comp_pair,
                          bcast_field_interpolation_config_idx_callback,
                          MSG_PRIO_HIGH);
  free(temp_data);
}

struct transient_field_interpolation_config_idx {
  unsigned orig_pos;
  unsigned num_field_interpolation_config_idxs;
  int * field_interpolation_config_idxs;
};

static int
compare_field_interpolation_config_idxs(const void * a, const void * b) {

  struct transient_field_interpolation_config_idx
    * t_a = (struct transient_field_interpolation_config_idx *)a,
    * t_b = (struct transient_field_interpolation_config_idx *)b;

  if (t_a->num_field_interpolation_config_idxs !=
      t_a->num_field_interpolation_config_idxs)
    yac_internal_abort_message("ERROR: internal error in "
                               "compare_field_interpolation_config_idxs\n",
                               __FILE__, __LINE__);

  for (unsigned i = 0; i < t_a->num_field_interpolation_config_idxs; ++i)
    if (t_a->field_interpolation_config_idxs[i] !=
        t_b->field_interpolation_config_idxs[i])
      return t_a->field_interpolation_config_idxs[i] -
             t_b->field_interpolation_config_idxs[i];

  return 0;
}

static int
compare_component_pairs(const void * a, const void * b) {

  struct component_pair_couple * a_ = (struct component_pair_couple *)a,
                               * b_ = (struct component_pair_couple *)b;

  char const * comp_names_a[2];
  char const * comp_names_b[2];

  {
    char const * tmp_comp_name[2] = {yac_component_get_name(a_->local_component),
                                     a_->remote_component_name};

    if (strcmp(tmp_comp_name[0], tmp_comp_name[1]) < 0) {
      comp_names_a[0] = tmp_comp_name[0];
      comp_names_a[1] = tmp_comp_name[1];
    } else {
      comp_names_a[0] = tmp_comp_name[1];
      comp_names_a[1] = tmp_comp_name[0];
    }
  }
  {
    char const * tmp_comp_name[2] = {yac_component_get_name(b_->local_component),
                                     b_->remote_component_name};

    if (strcmp(tmp_comp_name[0], tmp_comp_name[1]) < 0) {
      comp_names_b[0] = tmp_comp_name[0];
      comp_names_b[1] = tmp_comp_name[1];
    } else {
      comp_names_b[0] = tmp_comp_name[1];
      comp_names_b[1] = tmp_comp_name[0];
    }
  }

  int ret = strcmp(comp_names_a[0], comp_names_b[0]);
  if (!ret) ret = strcmp(comp_names_a[1], comp_names_b[1]);
  return ret;
}

static void determine_unique_field_interpolation_configs() {

  for (unsigned i = 0; i < couple_config.num_component_pairs; ++i) {

    unsigned num_transients = 0;

    struct component_pair_couple * curr_comp_pair =
      couple_config.component_pairs + i;

    for (unsigned j = 0; j < curr_comp_pair->num_domain_pairs; ++j) {

      struct domain_pair_couple * curr_domain_pair =
        curr_comp_pair->domain_pairs + j;

      num_transients += curr_domain_pair->num_transients;

      for (unsigned k = 0; k < curr_domain_pair->num_transients; ++k)
        curr_domain_pair->transients[k].search = curr_domain_pair->search;
    }

    curr_comp_pair->num_transients = num_transients;
    curr_comp_pair->transients =
      malloc(num_transients * sizeof(*(curr_comp_pair->transients)));

    unsigned offset = 0;

    for (unsigned j = 0; j < curr_comp_pair->num_domain_pairs; ++j) {

      struct domain_pair_couple * curr_domain_pair =
        curr_comp_pair->domain_pairs + j;

      memcpy(curr_comp_pair->transients + offset, curr_domain_pair->transients,
             curr_domain_pair->num_transients *
             sizeof(*(curr_comp_pair->transients)));

      offset += curr_domain_pair->num_transients;

      free(curr_domain_pair->transients);
      curr_domain_pair->transients = NULL;
      curr_domain_pair->num_transients = 0;
    }
  }

  int * recv_buffer = NULL;
  unsigned recv_buffer_array_size = 0;

  int * send_buffer = NULL;
  unsigned send_buffer_array_size = 0;

  unsigned comp_tags[couple_config.num_component_pairs];
  unsigned comp_pair_root_tags[2 * couple_config.num_component_pairs];

  // for all component pairs
  for (unsigned i = 0; i < couple_config.num_component_pairs; ++i) {

    struct component_pair_couple * curr_comp_pair =
      couple_config.component_pairs + i;
    struct communicator * comp_comm =
      yac_component_get_communicator(curr_comp_pair->local_component);
    unsigned comp_size = yac_get_comm_size(comp_comm);

    // sort transients -> transients with identical configuration are next to
    // each other
    qsort(curr_comp_pair->transients, curr_comp_pair->num_transients,
          sizeof(*(curr_comp_pair->transients)), compare_transients_by_config);

    int field_interpolation_config_idx = 0;

    curr_comp_pair->transients[0].field_interpolation_config_idx =
      field_interpolation_config_idx++;

    // assign field_interpolation_config_idx's
    for (unsigned i = 1; i < curr_comp_pair->num_transients; ++i) {

      if (compare_transients_by_config(
            (void*)(curr_comp_pair->transients + i - 1),
            (void*)(curr_comp_pair->transients + i)))
        curr_comp_pair->transients[i].field_interpolation_config_idx =
          field_interpolation_config_idx++;
      else
        curr_comp_pair->transients[i].field_interpolation_config_idx =
          curr_comp_pair->transients[i-1].field_interpolation_config_idx;
    }

    // sort transients (all processes now have the same order)
    qsort(curr_comp_pair->transients, curr_comp_pair->num_transients,
          sizeof(*(curr_comp_pair->transients)), compare_transients_by_index);

    ENSURE_ARRAY_SIZE(send_buffer, send_buffer_array_size,
                      4 * curr_comp_pair->num_transients);
    ENSURE_ARRAY_SIZE(recv_buffer, recv_buffer_array_size,
                      4 * curr_comp_pair->num_transients * comp_size);

    // pack data for exchange
    for (unsigned i = 0; i < curr_comp_pair->num_transients; ++i) {
      send_buffer[4*i+0] = curr_comp_pair->transients[i].couple_idx;
      send_buffer[4*i+1] = curr_comp_pair->transients[i].transient_idx;
      send_buffer[4*i+2] = curr_comp_pair->transients[i].type;
      send_buffer[4*i+3] =
        curr_comp_pair->transients[i].field_interpolation_config_idx;
    }

    // collect all field_interpolation_config_idx from all processes of the same
    // component
    yac_comm_allgather(send_buffer, 4 * curr_comp_pair->num_transients,
                       recv_buffer, 4 * curr_comp_pair->num_transients,
                       COMM_INT, comp_comm);

    // consistency check
    for (unsigned i = 1; i < comp_size; ++i) {

      for (unsigned j = 0; j < curr_comp_pair->num_transients; ++j) {

        if ((recv_buffer[4*j+0] !=
             recv_buffer[4*curr_comp_pair->num_transients*i + 4*j+0]) ||
            (recv_buffer[4*j+1] !=
             recv_buffer[4*curr_comp_pair->num_transients*i + 4*j+1]) ||
            (recv_buffer[4*j+2] !=
             recv_buffer[4*curr_comp_pair->num_transients*i + 4*j+2])) {

          yac_internal_abort_message(
            "ERROR(determine_unique_field_interpolation_configs): not all "
            "processes of local component have same setup of coupling fields",
            __FILE__, __LINE__);
        }
      }
    }

    if (curr_comp_pair->num_transients > 0) {

      struct transient_field_interpolation_config_idx *
        field_interpolation_config_idxs =
          malloc((comp_size + 1) * curr_comp_pair->num_transients *
                 sizeof(*field_interpolation_config_idxs));
      int * field_interpolation_config_idx_buffer =
          malloc(comp_size * curr_comp_pair->num_transients *
                 sizeof(*field_interpolation_config_idx_buffer));

      for (unsigned i = 0; i < curr_comp_pair->num_transients; ++i) {

        field_interpolation_config_idxs[i].orig_pos = i;
        field_interpolation_config_idxs[i].num_field_interpolation_config_idxs =
          comp_size;
        field_interpolation_config_idxs[i].field_interpolation_config_idxs =
          field_interpolation_config_idx_buffer + i * comp_size;

        for (unsigned j = 0; j < comp_size; ++j)
          field_interpolation_config_idxs[i].field_interpolation_config_idxs[j]=
            recv_buffer[4 * i + 3 + j * 4 * curr_comp_pair->num_transients];
      }

      qsort(field_interpolation_config_idxs, curr_comp_pair->num_transients,
            sizeof(*field_interpolation_config_idxs),
            compare_field_interpolation_config_idxs);

      {
        int idx = 0;

        curr_comp_pair->transients[field_interpolation_config_idxs[0].orig_pos]
          .field_interpolation_config_idx = idx;

        for (unsigned i = 1; i < curr_comp_pair->num_transients; ++i) {

          if (compare_field_interpolation_config_idxs(
                field_interpolation_config_idxs + i - 1,
                field_interpolation_config_idxs + i)) ++idx;

          curr_comp_pair->transients[field_interpolation_config_idxs[i].orig_pos]
            .field_interpolation_config_idx = idx;
        }
      }

      free(field_interpolation_config_idxs);
      free(field_interpolation_config_idx_buffer);
    }
  }

  qsort(couple_config.component_pairs, couple_config.num_component_pairs,
        sizeof(*(couple_config.component_pairs)), compare_component_pairs);

  // for all component pairs
  for (unsigned i = 0; i < couple_config.num_component_pairs; ++i) {

    struct component_pair_couple * curr_comp_pair =
      couple_config.component_pairs + i;
    struct communicator * comp_comm =
      yac_component_get_communicator(curr_comp_pair->local_component);

    if (curr_comp_pair->num_transients > 0) {

      struct global_search * search =
        curr_comp_pair->transients[0].search;
      struct communicator * comm = yac_get_communicator(search);
      unsigned root;
      unsigned remote_comp_root_rank =
        yac_get_lowest_remote_domain_rank(search);
      unsigned local_comp_root_rank =
        yac_get_lowest_local_domain_rank(search, &root);

      // pack transient information
      for (unsigned i = 0; i < curr_comp_pair->num_transients; ++i) {
        send_buffer[4*i+0] = curr_comp_pair->transients[i].couple_idx;
        send_buffer[4*i+1] = curr_comp_pair->transients[i].transient_idx;
        send_buffer[4*i+2] = curr_comp_pair->transients[i].type;
        send_buffer[4*i+3] =
          curr_comp_pair->transients[i].field_interpolation_config_idx;
      }

      { // generate some tags
        char bcast_tag_str[1024], config_tag_str[2][1024],
             unique_generator_string[1024];
        char * tag_str[2] = {config_tag_str[0], config_tag_str[1]};

        if (send_buffer[4*0+2] == SRC) {
          sprintf(bcast_tag_str, "bcast couple config %s%s",
                  yac_component_get_name(curr_comp_pair->local_component),
                  curr_comp_pair->remote_component_name);
          sprintf(unique_generator_string, "generator string %s%s",
                  yac_component_get_name(curr_comp_pair->local_component),
                  curr_comp_pair->remote_component_name);
        } else {
          sprintf(bcast_tag_str, "bcast couple config %s%s",
                  curr_comp_pair->remote_component_name,
                  yac_component_get_name(curr_comp_pair->local_component));
          sprintf(unique_generator_string, "generator string %s%s",
                  curr_comp_pair->remote_component_name,
                  yac_component_get_name(curr_comp_pair->local_component));
        }
        sprintf(config_tag_str[0], "couple config %s%s",
                yac_component_get_name(curr_comp_pair->local_component),
                curr_comp_pair->remote_component_name);
        sprintf(config_tag_str[1], "couple config %s%s",
                curr_comp_pair->remote_component_name,
                yac_component_get_name(curr_comp_pair->local_component));

        if ((i == 0) ||
            ((i > 0) && 
             compare_component_pairs(curr_comp_pair,
                                     couple_config.component_pairs + i - 1))) {

          yac_get_unique_comm_tags(
            comm, unique_generator_string, (const char **)tag_str, 2,
            &comp_pair_root_tags[2*i]);

        } else {
          comp_pair_root_tags[2 * i + 0] = comp_pair_root_tags[2 * (i - 1) + 1];
          comp_pair_root_tags[2 * i + 1] = comp_pair_root_tags[2 * (i - 1) + 0];
        }

        tag_str[0] = bcast_tag_str;
        yac_get_unique_comm_tags(
          yac_component_get_communicator(curr_comp_pair->local_component),
          unique_generator_string, (const char **)tag_str, 1,
          &comp_tags[i]);
      }

      if (local_comp_root_rank == yac_get_comm_rank(comm)) {

        struct temp_field_interpolation_config_idx_data * temp_data =
          malloc(1 * sizeof(*temp_data));

        temp_data->comp_pair = curr_comp_pair;
        temp_data->tag = comp_tags[i];

        // send field_interpolation_config_idx's of local component to root of
        // remote component
        // of other component
        yac_comm_bsend(send_buffer, 4 * curr_comp_pair->num_transients, COMM_INT,
                       remote_comp_root_rank, comp_pair_root_tags[2*i+0], comm);

        // set up receive_callback for field_interpolation_config_idx's from
        // the other component (source is root of other component)
        // once the field_interpolation_config_idx's have been received, they
        // are broadcasted to the other ranks of the local domain
        yac_comm_recv_callback(4 * curr_comp_pair->num_transients, COMM_INT,
                           remote_comp_root_rank, comp_pair_root_tags[2*i+1],
                           comm, temp_data,
                           recv_field_interpolation_config_idx_callback, NULL,
                           MSG_PRIO_HIGH);
      } else {

        // set up bcast_callback for field_interpolation_config_idx's from
        // the other component (source is root of the local component)
        yac_comm_bcast_callback(NULL, 4 * curr_comp_pair->num_transients,
                                COMM_INT, root, comp_tags[i], comp_comm,
                                curr_comp_pair,
                                bcast_field_interpolation_config_idx_callback,
                                MSG_PRIO_HIGH);
      }
    }
  }

  free(recv_buffer);
  free(send_buffer);

  // for all component pairs wait until all field_interpolation_config_idx's
  // have been exchanged and final resuls are available
  for (unsigned i = 0; i < couple_config.num_component_pairs; ++i) {

    if (couple_config.component_pairs[i].num_transients > 0) {

      struct component_pair_couple * curr_comp_pair =
        couple_config.component_pairs + i;

      yac_comm_wait(yac_get_communicator(curr_comp_pair->transients[0].search),
                    comp_pair_root_tags[2*i+1]);
      yac_comm_wait(yac_component_get_communicator(
                      curr_comp_pair->local_component), comp_tags[i]);
    }
  }
}

static void
generate_interpolations() {

  determine_unique_field_interpolation_configs();

  //-------------------------------------
  // generate interpolation method stacks
  //-------------------------------------

  // for all component pairs
  for (unsigned i = 0; i < couple_config.num_component_pairs; ++i) {

    struct component_pair_couple * curr_comp_pair =
      couple_config.component_pairs + i;

    int interpolation_method_stack_flag[curr_comp_pair->num_transients];

    for (unsigned i = 0; i < curr_comp_pair->num_transients; ++i)
      interpolation_method_stack_flag[i] = 0;

    // for all transients
    for (unsigned k = 0; k < curr_comp_pair->num_transients; ++k) {

      struct transient * curr_transient =
        curr_comp_pair->transients + k;

      int field_interpolation_config_idx =
        curr_transient->field_interpolation_config_idx;

      if ((!interpolation_method_stack_flag[field_interpolation_config_idx]) &&
          (curr_transient->type == SRC)) {

        // generate interpolation method stack

        curr_transient->data.src.interpolation_method_stack =
          generate_interpolation_method_stack(curr_transient->search,
            yac_get_coupling_field_name(curr_transient->field),
            curr_transient);

        interpolation_method_stack_flag[field_interpolation_config_idx] = 1;
      }
    }
  }

  //-----------------------------
  // generate interpolation_gen's
  //-----------------------------

  // for all component pairs
  for (unsigned i = 0; i < couple_config.num_component_pairs; ++i) {

    struct component_pair_couple * curr_comp_pair =
      couple_config.component_pairs + i;
    struct component_pair_couple * matching_comp_pair = NULL;
    
    for (unsigned j = i + 1; j < couple_config.num_component_pairs; ++j) {
      if (!(strcmp(yac_component_get_name(curr_comp_pair->local_component),
                   couple_config.component_pairs[j].remote_component_name) ||
            strcmp(yac_component_get_name(
                     couple_config.component_pairs[j].local_component),
                   curr_comp_pair->remote_component_name))) {
    
        matching_comp_pair = couple_config.component_pairs + j;
        break;
      }
    }

    struct field_interpolation_gen *
      interpolation_gen_buffer[curr_comp_pair->num_transients];

    for (unsigned i = 0; i < curr_comp_pair->num_transients; ++i)
      interpolation_gen_buffer[i] = NULL;

    // for all transients
    for (unsigned k = 0; k < curr_comp_pair->num_transients; ++k) {

      struct transient * curr_transient =
        curr_comp_pair->transients + k;

      if (curr_transient->interpolation_gen != NULL) continue;

      int field_interpolation_config_idx =
        curr_transient->field_interpolation_config_idx;

      if (interpolation_gen_buffer[field_interpolation_config_idx] != NULL) {

        curr_transient->interpolation_gen =
          interpolation_gen_buffer[field_interpolation_config_idx];

        if (curr_transient->type == SRC)
          yac_field_interpolation_gen_add_weight_file_name(
            interpolation_gen_buffer[field_interpolation_config_idx],
            curr_transient->data.src.weight_file_name);

      } else {

        // find matching transient and generate src and tgt at the same time

        struct transient * matching_transient = NULL;

        // if there is a matching component pair
        if (matching_comp_pair != NULL) {

          for (unsigned j = 0; j < matching_comp_pair->num_transients; ++j) {
            if ((curr_transient->couple_idx ==
                 matching_comp_pair->transients[j].couple_idx) &&
                (curr_transient->transient_idx ==
                 matching_comp_pair->transients[j].transient_idx)) {
       
              matching_transient = matching_comp_pair->transients + j;
              break;
            }
          }

          if (matching_transient == NULL) {

            char err_str[1024];
            snprintf(err_str, 1024, "ERROR: internal error. Components %s and"
                     " %s are both defined on the same process. But matching "
                     "transients for field %s for both directions could not be "
                     "found.\n",
                     yac_component_get_name(curr_comp_pair->local_component),
                     curr_comp_pair->remote_component_name,
                     yac_get_coupling_field_name(curr_transient->field));
            yac_internal_abort_message(err_str, __FILE__, __LINE__);
          }

          if (curr_transient->type == matching_transient->type) {
            char err_str[1024];
            snprintf(err_str, 1024, "ERROR: problem with transient type of "
                     "field %s.\n",
                     yac_get_coupling_field_name(curr_transient->field));
            yac_internal_abort_message(err_str, __FILE__, __LINE__);
          }
        }

        if (matching_transient != NULL) {

          struct transient * src_transient, * tgt_transient;
          struct component_pair_couple * src_comp_pair, * tgt_comp_pair;

          if (curr_transient->type == SRC) {
            src_transient = curr_transient;
            tgt_transient = matching_transient;
            src_comp_pair = curr_comp_pair;
            tgt_comp_pair = matching_comp_pair;
          } else {
            src_transient = matching_transient;
            tgt_transient = curr_transient;
            src_comp_pair = matching_comp_pair;
            tgt_comp_pair = curr_comp_pair;
          }

          yac_new_field_interpolation_gen_src_tgt_mask_2(
            yac_get_coupling_field_pointsets(src_transient->field),
            yac_get_coupling_field_num_pointsets(src_transient->field),
            yac_get_coupling_field_name(src_transient->field),
            src_transient->search,
            src_transient->data.src.interpolation_method_stack,
            src_transient->data.src.weight_file_name,
            yac_component_get_io_proc_step(
                src_comp_pair->local_component),
            yac_get_coupling_field_pointsets(tgt_transient->field),
            yac_get_coupling_field_masks(tgt_transient->field),
            yac_get_coupling_field_num_pointsets(tgt_transient->field),
            yac_get_coupling_field_name(tgt_transient->field),
            tgt_transient->search, tgt_transient->data.tgt.points_idx,
            tgt_transient->data.tgt.mask_handling,
            &(src_transient->interpolation_gen),
            &(tgt_transient->interpolation_gen),
            src_transient->mapping_on_source);

        } else if (curr_transient->type == SRC) {

          // generate field_interpolation_gen
          curr_transient->interpolation_gen =
            yac_new_field_interpolation_gen_src(
              yac_get_coupling_field_pointsets(curr_transient->field),
              yac_get_coupling_field_num_pointsets(curr_transient->field),
              yac_get_coupling_field_name(curr_transient->field),
              curr_transient->search,
              yac_get_coupling_field_name(curr_transient->field),
              curr_transient->data.src.interpolation_method_stack,
              curr_transient->data.src.weight_file_name,
              yac_component_get_io_proc_step(
                curr_comp_pair->local_component),
              curr_transient->mapping_on_source);
        } else {

          // generate field_interpolation_gen
          curr_transient->interpolation_gen =
            yac_new_field_interpolation_gen_tgt_mask(
              yac_get_coupling_field_pointsets(curr_transient->field),
              yac_get_coupling_field_masks(curr_transient->field),
              yac_get_coupling_field_num_pointsets(curr_transient->field),
              yac_get_coupling_field_name(curr_transient->field),
              curr_transient->search,
              yac_get_coupling_field_name(curr_transient->field),
              curr_transient->data.tgt.points_idx,
              curr_transient->data.tgt.mask_handling,
              curr_transient->mapping_on_source);
        }
      }

      interpolation_gen_buffer[field_interpolation_config_idx] =
        curr_transient->interpolation_gen;
    }
  }
}

static void
finalise_interpolations() {

  // for all component pairs
  for (unsigned i = 0; i < couple_config.num_component_pairs; ++i) {

    struct component_pair_couple * curr_comp_pair =
      couple_config.component_pairs + i;

    struct field_interpolation_gen *
      interpolation_gen_buffer[curr_comp_pair->num_transients];

    // initialise buffer
    for (unsigned i = 0; i < curr_comp_pair->num_transients; ++i)
      interpolation_gen_buffer[i] = NULL;

    // get all unique interpolation gens
    for (unsigned k = 0; k < curr_comp_pair->num_transients; ++k)
      interpolation_gen_buffer[
        curr_comp_pair->transients[k].field_interpolation_config_idx] =
          curr_comp_pair->transients[k].interpolation_gen;

    // finalise generation of interpolation gens
    for (unsigned i = 0; i < curr_comp_pair->num_transients; ++i)
      if (interpolation_gen_buffer[i] != NULL)
        yac_enddef_field_interpolation_gen(interpolation_gen_buffer[i]);

    // for all transients
    for (unsigned k = 0; k < curr_comp_pair->num_transients; ++k) {

      struct transient * curr_transient = curr_comp_pair->transients + k;

      if (curr_transient->type == SRC) {

        yac_set_coupling_field_put_op(curr_transient->field,
                                      curr_transient->event_id,
                                      yac_get_field_interpolation(
                                        curr_transient->interpolation_gen,
                                        yac_get_coupling_field_name(
                                          curr_transient->field),
                                        yac_get_coupling_field_name(
                                          curr_transient->field)));

        free(curr_transient->data.src.interpolation_methods);
        curr_transient->data.src.interpolation_methods = NULL;
        curr_transient->data.src.num_interpolations = 0;
        yac_xml_free_xml_string(curr_transient->data.src.weight_file_name);

      } else
        yac_set_coupling_field_get_op(curr_transient->field,
                                      curr_transient->event_id,
                                      yac_get_field_interpolation(
                                        curr_transient->interpolation_gen,
                                        yac_get_coupling_field_name(
                                          curr_transient->field),
                                        yac_get_coupling_field_name(
                                          curr_transient->field)));

      curr_transient->event_id = -1;
    }

    // clean up interpolation method stacks
    for (unsigned k = 0; k < curr_comp_pair->num_transients; ++k) {

      struct transient * curr_transient = curr_comp_pair->transients + k;

      if ((curr_transient->type == SRC) &&
          (curr_transient->data.src.interpolation_method_stack != NULL)) {

        for (int i = 0;
             curr_transient->data.src.interpolation_method_stack[i] != NULL;
             ++i)
          yac_delete_interpolation_method(
            curr_transient->data.src.interpolation_method_stack[i]);

        free(curr_transient->data.src.interpolation_method_stack);
        curr_transient->data.src.interpolation_method_stack = NULL;
      }

      curr_transient->interpolation_gen = NULL;
    }

    // clean up interpolation gens
    for (unsigned k = 0; k < curr_comp_pair->num_transients; ++k)
      if (interpolation_gen_buffer[k] != NULL)
        yac_delete_field_interpolation_gen(interpolation_gen_buffer[k]);
  }
}

void yac_setup_couple_config(unsigned num_components,
                             struct component * components[num_components],
                             unsigned num_fields,
                             struct coupling_field * coupling_fields[num_fields]) {

  check_status(FREED);

  generate_component_pairs(num_components, components, num_fields, coupling_fields);

  generate_global_search_information();

  generate_interpolations();

  couple_config.status = INITIALISED;
}

struct communicator * yac_get_pair_rootcomm_ ( char const * comp_name1,
                                               char const * comp_name2 ) {

  struct communicator * pair_rootcomm;

  for (unsigned i = 0; i < couple_config.num_component_pairs; ++i) {

    struct component_pair_couple * curr_comp_pair =
      couple_config.component_pairs + i;

    if (curr_comp_pair->num_transients > 0) {

      const char * local_component_name = yac_component_get_name(curr_comp_pair->local_component);

      if ( ( strcmp(comp_name1,local_component_name                  ) == 0  &&
             strcmp(comp_name2,curr_comp_pair->remote_component_name ) == 0 ) ||
           ( strcmp(comp_name2,local_component_name                  ) == 0  &&
             strcmp(comp_name1,curr_comp_pair->remote_component_name ) == 0 ) ) {

        struct global_search * search =
          curr_comp_pair->transients[0].search;

        struct communicator * comm = yac_get_communicator(search);

        unsigned remote_comp_root_rank =
          yac_get_lowest_remote_domain_rank(search);
        unsigned local_comp_root_rank  =
          yac_get_lowest_local_domain_rank(search, NULL);

        int key = 0;
        int color = 0;

        if ( local_comp_root_rank  == yac_get_comm_rank(comm) ||
             remote_comp_root_rank == yac_get_comm_rank(comm) ) {
          color = 1;
        }
        pair_rootcomm = yac_comm_split ( curr_comp_pair->comm, color, key );
      }
    }
  }
  return pair_rootcomm;
}

void yac_enddef_couple_config() {

  check_status(INITIALISED);

  finalise_interpolations();

  couple_config.status = FINALIZED;
}

static void
free_transient(struct transient * transient) {

  // note: interpolation_method_stack has already been deleted by
  //       finalise_interpolations
  // note: event_id and interpolation are freed by the coupling field
  //       they were generated for

  return;
}

static void
free_domain_pair(struct domain_pair_couple * domain_pair) {

  free(domain_pair->remote_domain_name);
  yac_global_search_delete(domain_pair->search);
}

static void
free_component_pair(struct component_pair_couple * comp_pair) {

  for (unsigned i = 0; i < comp_pair->num_domain_pairs; ++i)
    free_domain_pair(comp_pair->domain_pairs + i);
  free(comp_pair->domain_pairs);

  for (unsigned i = 0; i < comp_pair->num_transients; ++i)
    if ((i == 0) ||
        (comp_pair->transients[i-1].field_interpolation_config_idx !=
         comp_pair->transients[i].field_interpolation_config_idx))
      free_transient(comp_pair->transients + i);
  free(comp_pair->transients);

  if (comp_pair->free_comm) yac_free_communicator(comp_pair->comm);
}

void yac_free_couple_config() {

  check_status_or(FINALIZED, FREED);

  for (unsigned i = 0; i < couple_config.num_component_pairs; ++i)
    free_component_pair(couple_config.component_pairs + i);

  freeCalendar();

  free(couple_config.component_pairs);

  couple_config.component_pairs = NULL;
  couple_config.num_component_pairs = 0;

  couple_config.status = FREED;
}

#else // USE_XML2

void yac_setup_couple_config(unsigned num_components,
                             struct component * components[num_components],
                             unsigned num_fields,
                             struct coupling_field * coupling_fields[num_fields]) {
  fprintf (stderr, "WARNING: yac was compiled without xml support!\n");
  fprintf (stdout, "WARNING: yac was compiled without xml support!\n");
}
void yac_enddef_couple_config() {return;}
void yac_free_couple_config() {return;}

#endif // USE_XML2
