/**
 * @file couple_config.c
 * @brief Structs and interfaces to defined coupling configuration
 *
 * Stores reads in coupling configuration and sets up data structures
 * to do the coupling.
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#ifndef COUPLE_CONFIG_H
#define COUPLE_CONFIG_H

#include "fields.h"
#include "component.h"

/** \example test_couple_config.c
 * This example show how to set up a couple configuration. It uses three
 * processes.
 */

/** \example test_couple_config_2.c
 * This example show how to set up a couple configuration. It uses nine
 * processes.
 */

/**
 * reads coupling configuration from configuration xml file and initiates
 * the generation of all data structures required for interpolation
 * @param[in] num_components  number of locally defined components
 * @param[in] components      array of locally defined components
 * @param[in] num_fields      number of locally defined fields
 * @param[in] coupling_fields array of locally defined fields
 * \remarks a call to \ref yac_enddef_couple_config is required to finish the
 *          generation of the data structures
 * \remarks this routine can only be called once
 * \see yac_enddef_couple_config
 */
void yac_setup_couple_config(unsigned num_components,
                             struct component * components[num_components],
                             unsigned num_fields,
                             struct coupling_field * coupling_fields[num_fields]);

/**
 * Finishes the generation of the data structures required for coupling
 * \remarks this routine can only be called once
 * \see yac_free_couple_config
 */
void yac_enddef_couple_config();

/**
 * returns a communicator for communicating between pairs of components
 *
 * @param[in]  comp_name1     name of first component
 * @param[in]  comp_name2     name of second component
 */
struct communicator * yac_get_pair_rootcomm_ ( char const * comp_name1,
                                               char const * comp_name2 );

/**
 * frees all memory allocated for the coupling configuration
 * \remarks this routine can only be called once
 * \remarks this routine must be called after \ref yac_enddef_couple_config
 * \remarks the couple configuration generates communicators for all
 *          active component pairs, this routine can only be called once
 *          it is guaranteed that non of these will be used anymore
 * \see yac_enddef_couple_config
 */
void yac_free_couple_config();

#endif // COUPLE_CONFIG_H
