//#define VERBOSE
/**
 * @file event.c
 *
 * @copyright Copyright  (C)  2013 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "utils.h"
#include "event.h"

#define SECONDS_PER_DAY 86400

int yac_number_of_events;

/* ---------------------------------------------------------------------- */

void yac_event_init ( int * event_id ) {

  struct event * event;

  /* allocate memory for one event */

  event = malloc(1 * sizeof (*event));

  /* get a handle to the event */

  *event_id = yac_pointer_to_unique_id(event);

  /* initialise the event struct */

  event->is_set             = 0;
  event->lag                = 0;

  event->time_operation     = INSTANT;

  ++yac_number_of_events;

}

/* ---------------------------------------------------------------------- */

void yac_event_add ( int event_id,
                     void *delta_model_time,
                     void *delta_coupling_time,
                     enum  yac_time_unit_type time_unit,
                     int lag,
                     enum yac_reduction_type time_operation,
                     const char * startdate,
                     const char * stopdate,
                     const char * offset ) {

  /* Get the event from the handle */

  struct event * event;
  event = yac_unique_id_to_pointer(event_id);

  /* check for free event handles is the existing list */

  if ( event->is_set != 0 ) {
    printf ("event_add: Event is not initialised or event_add has\n");
    printf ("           already been called for this event.\n");
  }

  /* set the event triggering parameters */

  char PTstring[MAX_TIMEDELTA_STR_LEN];

  event->is_set             = 1;
  event->lag                = lag;
  event->time_operation     = time_operation;

  /* create dummy_event */

  event->dummy_time_step = newTimeDelta ( getPTStringFromMS ( 0, PTstring ) );

  /* model time step and coupling time step need to be converted */

  int *delta_model_time_intPtr = (int *) delta_model_time;
  int *delta_coupling_time_intPtr = (int *) delta_coupling_time;

  switch (time_unit) {

  case (C_MILLISECOND):
   {
     event->model_time_step = newTimeDelta ( getPTStringFromMS ( (long int) *delta_model_time_intPtr, PTstring ) );
     if ( event->model_time_step == NULL ) yac_internal_abort_message ( "ERROR in event definition", __FILE__, __LINE__ );
     event->coupling_time_step = newTimeDelta ( getPTStringFromMS ( (long int) *delta_coupling_time_intPtr, PTstring ) );
     if ( event->coupling_time_step == NULL ) yac_internal_abort_message ( "ERROR in event definition", __FILE__, __LINE__ );
     break;
   }

  case (C_SECOND):
   {
     event->model_time_step = newTimeDelta ( getPTStringFromSeconds ( (long int) *delta_model_time_intPtr, PTstring ) );
     if ( event->model_time_step == NULL ) yac_internal_abort_message ( "ERROR in event definition", __FILE__, __LINE__ );
     event->coupling_time_step = newTimeDelta ( getPTStringFromSeconds ( (long int) *delta_coupling_time_intPtr, PTstring ) );
     if ( event->coupling_time_step == NULL ) yac_internal_abort_message ( "ERROR in event definition", __FILE__, __LINE__ );
     break;
   }

  case (C_MINUTE):
   {
     event->model_time_step = newTimeDelta ( getPTStringFromMinutes ( (long int) *delta_model_time_intPtr, PTstring ) );
     if ( event->model_time_step == NULL ) yac_internal_abort_message ( "ERROR in event definition", __FILE__, __LINE__ );
     event->coupling_time_step = newTimeDelta ( getPTStringFromMinutes ( (long int) *delta_coupling_time_intPtr, PTstring ) );
     if ( event->coupling_time_step == NULL ) yac_internal_abort_message ( "ERROR in event definition", __FILE__, __LINE__ );
     break;
   }

  case (C_HOUR):
   {
     event->model_time_step = newTimeDelta ( getPTStringFromHours ( (long int) *delta_model_time_intPtr, PTstring ) );
     if ( event->model_time_step == NULL ) yac_internal_abort_message ( "ERROR in event definition", __FILE__, __LINE__ );
     event->coupling_time_step = newTimeDelta ( getPTStringFromHours ( (long int) *delta_coupling_time_intPtr, PTstring ) );
     if ( event->coupling_time_step == NULL ) yac_internal_abort_message ( "ERROR in event definition", __FILE__, __LINE__ );
     break;
   }

  case (C_ISO_FORMAT):
   {
     event->model_time_step = newTimeDelta ( (char*) delta_model_time );
     if ( event->model_time_step == NULL ) yac_internal_abort_message ( "ERROR in event definition", __FILE__, __LINE__ );
     event->coupling_time_step = newTimeDelta ( (char *) delta_coupling_time );
     if ( event->coupling_time_step == NULL ) yac_internal_abort_message ( "ERROR in event definition", __FILE__, __LINE__ );
     break;
   }

  default:
     yac_internal_abort_message ( "ERROR in event definition: unsupported time unit", __FILE__, __LINE__ );

  };

  // if the model time step and the coupling time step are identical, a time
  // operation does not make sense
  if (!compareTimeDelta(event->model_time_step, event->coupling_time_step)) {
    event->time_operation = INSTANT;
  }

  calendarType set_calender = getCalendarType ();

  if ( set_calender == CALENDAR_NOT_SET )
         yac_internal_abort_message("ERROR: Calender is not set.\n",  __FILE__, __LINE__);

  event->startdate = newDateTime(startdate);
  if ( event->startdate == NULL ) yac_internal_abort_message ( "ERROR in event definition", __FILE__, __LINE__ );

  event->currentdate = newDateTime(startdate);
  if ( event->currentdate == NULL ) yac_internal_abort_message ( "ERROR in event definition", __FILE__, __LINE__ );

  event->stopdate = newDateTime(stopdate);
  if ( event->stopdate == NULL ) yac_internal_abort_message ( "ERROR in event definition", __FILE__, __LINE__ );

  char timedelta_str[MAX_TIMEDELTA_STR_LEN];

  event->coupling = newEvent("Coupling", (char *) startdate, (char *) startdate, (char *) stopdate,
                             timedeltaToString(event->coupling_time_step, timedelta_str), NULL );

  if ( event->coupling == NULL ) yac_internal_abort_message ( "ERROR in event definition", __FILE__, __LINE__ );

#ifdef VERBOSE
  char datetime_str[MAX_DATETIME_STR_LEN];
  printf ("event add for event ID %d \n", event_id );
  printf ("- start   date:    %s\n", datetimeToString(event->startdate,datetime_str) );
  printf ("- current date:    %s\n", datetimeToString(event->currentdate,datetime_str) );
  printf ("- stop    date:    %s\n", datetimeToString(event->stopdate,datetime_str) );
  printf ("- model    time step: %s\n", timedeltaToString(event->model_time_step,timedelta_str) );
  printf ("- coupling time step: %s\n", timedeltaToString(event->coupling_time_step,timedelta_str) );
#endif

  if ( event->lag < 0 )
    yac_internal_abort_message ( "ERROR: lag has to be greater than zero", __FILE__, __LINE__ );

  /* add lag times delta(t) to the current date such that we can detect
     a restart event which has to be performed lag times delta(t) before
     the final date is reached. */
  if ( event->lag > 0 ) {
    for ( int lag = 0; lag < event->lag; ++lag )
      event->currentdate = addTimeDeltaToDateTime ( event->currentdate, event->model_time_step, event->currentdate );
#ifdef VERBOSE
    printf ("- model   time lag:  %d\n", event->lag );
    printf ("- new current date:  %s\n", datetimeToString(event->currentdate,datetime_str) );
#endif
  }
}

/* ---------------------------------------------------------------------- */

void yac_event_update ( int event_id ) {

   /* Get the event from the handle */

   struct event * event;
   event = yac_unique_id_to_pointer(event_id);

   if ( event->is_set == 0 ) {
      printf ("event_update: Event is initialised, but not filled with data.\n");
      printf ("event_update: Call event_add prior to event_update.\n");
   }

   event->currentdate = addTimeDeltaToDateTime ( event->currentdate, event->model_time_step, event->currentdate );
}

/* ---------------------------------------------------------------------- */

int yac_event_check ( int event_id ) {

  /* Get the event from the handle */

  struct event * event;
  event = yac_unique_id_to_pointer(event_id);

  /* Return value action:
     --------------------

     action = 0  : no action to be performed
     action = 1  : coupling action required
     action = 2  : restart file needs to be written
     action = 3-6: last put/get for restart
     action = 7  : date is out of bound, beyond restart

     -------------------------------------------------------------------- */

#ifdef VERBOSE
  char datetime_str[MAX_DATETIME_STR_LEN];
  printf ("- current date: %s\n", datetimeToString(event->currentdate,datetime_str) );
  printf ("- stop    date: %s\n", datetimeToString(event->stopdate,datetime_str) );
#endif

  event->action = NONE;

  /* Detect a restart event */

  if ( compareDatetime(event->currentdate,event->stopdate) == 0 ) {
#ifdef VERBOSE
     printf ("event check for event ID %d delivers RESTART\n", event_id );
#endif
     event->action = RESTART;
     return RESTART;

  } else if ( compareDatetime(event->currentdate,event->stopdate) > 0 ) {
#ifdef VERBOSE
     printf ("event check for event ID %d delivers OUT_OF_BOUND\n", event_id );
#endif
     event->action = OUT_OF_BOUND;
     return OUT_OF_BOUND;
  }

  /* Detect a coupling event */

  if ( isCurrentEventActive ( event->coupling,
                              event->currentdate,
                              event->dummy_time_step,
                              event->dummy_time_step ) != 0 ) {

#ifdef VERBOSE
     printf ("event check for event ID %d delivers COUPLING\n", event_id );
#endif
     event->action = COUPLING;
     return COUPLING;
  }

  // event->action = CHECKPOINT;
  // return CHECKPOINT;

  return NONE;
}

/* ---------------------------------------------------------------------- */

void yac_event_free ( int event_id ) {

  /* Get the event from the handle */

  struct event * event;
  event = yac_unique_id_to_pointer(event_id);

  deallocateDateTime(event->startdate);
  deallocateDateTime(event->currentdate);
  deallocateDateTime(event->stopdate);

  deallocateTimeDelta(event->dummy_time_step);
  deallocateTimeDelta(event->model_time_step);
  deallocateTimeDelta(event->coupling_time_step);

  deallocateEvent(event->coupling);

  free(event);

  --yac_number_of_events;
}

/* ---------------------------------------------------------------------- */

int yac_get_event_lag ( int event_id ) {

  struct event * event;

  /* Get the event from the handle */

  event = yac_unique_id_to_pointer(event_id);

  return event->lag;
}

/* ---------------------------------------------------------------------- */

int yac_get_event_time_operation ( int event_id ) {

  struct event * event;

  /* Get the event from the handle */

  event = yac_unique_id_to_pointer(event_id);

  return event->time_operation;
}

/* ---------------------------------------------------------------------- */

char * yac_get_event_coupling_timestep ( int event_id, char * timedelta_str ) {

  memset(timedelta_str,'\0',MAX_TIMEDELTA_STR_LEN);

  struct event * event;

  /* Get the event from the handle */

  event = yac_unique_id_to_pointer(event_id);

  timedeltaToString(event->coupling_time_step,timedelta_str);

  return timedelta_str;
}


/* ---------------------------------------------------------------------- */

char * yac_get_event_model_timestep ( int event_id, char * timedelta_str ) {

  memset(timedelta_str,'\0',MAX_TIMEDELTA_STR_LEN);

  struct event * event;

  /* Get the event from the handle */

  event = yac_unique_id_to_pointer(event_id);

  timedeltaToString(event->model_time_step,timedelta_str);

  return timedelta_str;
}

