/**
 * @file field_interpolation_gen.c
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#include <stdlib.h>
#include <string.h>

#include "field_interpolation_gen.h"
#include "field_interpolation_gen_src.h"
#include "field_interpolation_gen_src_tgt.h"
#include "field_interpolation_gen_tgt.h"
#include "interpolation_method.h"
#include "interpolation_plan.h"
#include "utils.h"
#include "global_search.h"

static void generate_tag_string(char * buf, char const * tag_base_str,
                                char const * source_comp_name,
                                char const * target_comp_name,
                                char const * source_domain_name,
                                char const * target_domain_name,
                                char const * source_field_name,
                                char const * target_field_name,
                                char const * tag_str) {

   strcpy(buf, tag_base_str);
   strcat(buf, source_comp_name);
   strcat(buf, target_comp_name);
   strcat(buf, source_domain_name);
   strcat(buf, target_domain_name);
   strcat(buf, source_field_name);
   strcat(buf, target_field_name);
   strcat(buf, tag_str);
}

struct field_interpolation_tags
yac_field_interpolation_generate_tag(struct communicator * comm,
                                     char const * source_comp_name,
                                     char const * target_comp_name,
                                     char const * source_domain_name,
                                     char const * target_domain_name,
                                     char const * source_field_name,
                                     char const * target_field_name) {

   char const * str_base_str              = "field interpolation gen";
   char const * str_generator_string      = "generator string";
   char const * str_results               = "results";
   char const * str_point_data_header     = "point data header";
   char const * str_point_data            = "point data";
   char const * str_tgt_point_location    = "tgt point location";
   char const * str_interp_data_msg_count = "interp data msg count";
   char const * str_interp_data_header    = "interp data header";
   char const * str_interp_data           = "interp data";

   char const * str_point_put_get           = "point put get";

   unsigned max_tag_strlen = strlen(str_interp_data_msg_count);

   unsigned max_str_buf_size = max_tag_strlen + strlen(str_base_str) +
                               strlen(source_comp_name) +
                               strlen(target_comp_name) +
                               strlen(source_domain_name) +
                               strlen(target_domain_name) +
                               strlen(source_field_name) +
                               strlen(target_field_name) + 1;

   char * unique_generator_string = malloc(max_str_buf_size);
   generate_tag_string(unique_generator_string, str_base_str, target_comp_name,
                       source_comp_name, target_domain_name, source_domain_name,
                       target_field_name, source_field_name,
                       str_generator_string);

   char * tag_strings[8];
   tag_strings[0] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[0], str_base_str, target_comp_name,
                       source_comp_name, target_domain_name, source_domain_name,
                       target_field_name, source_field_name, str_results);

   tag_strings[1] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[1], str_base_str, source_comp_name,
                       target_comp_name, source_domain_name, target_domain_name,
                       source_field_name, target_field_name, str_point_data);

   tag_strings[2] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[2], str_base_str, source_comp_name,
                       target_comp_name, source_domain_name, target_domain_name,
                       source_field_name, target_field_name,
                       str_point_data_header);

   tag_strings[3] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[3], str_base_str, target_comp_name,
                       source_comp_name, target_domain_name, source_domain_name,
                       target_field_name, source_field_name,
                       str_tgt_point_location);

   tag_strings[4] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[4], str_base_str, source_comp_name,
                       target_comp_name, source_domain_name, target_domain_name,
                       source_field_name, target_field_name,
                       str_interp_data_msg_count);

   tag_strings[5] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[5], str_base_str, source_comp_name,
                       target_comp_name, source_domain_name, target_domain_name,
                       source_field_name, target_field_name,
                       str_interp_data_header);

   tag_strings[6] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[6], str_base_str, source_comp_name,
                       target_comp_name, source_domain_name, target_domain_name,
                       source_field_name, target_field_name,
                       str_interp_data);

   tag_strings[7] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[7], str_base_str, target_comp_name,
                       source_comp_name, target_domain_name, source_domain_name,
                       target_field_name, source_field_name, str_point_put_get);

   unsigned search_tags[8];

   yac_get_unique_comm_tags(
      comm, unique_generator_string, (char const **)tag_strings, 8,
      search_tags);

   free(unique_generator_string);

   struct field_interpolation_tags tags;

   tags.results_exchange = search_tags[0];
   tags.point_data_exchange = search_tags[1];
   tags.point_data_header_exchange = search_tags[2];
   tags.tgt_point_location_exchange = search_tags[3];
   tags.interp_data_msg_count = search_tags[4];
   tags.interp_data_header = search_tags[5];
   tags.interp_data = search_tags[6];
   tags.exchange = search_tags[7];

   for (unsigned i = 0; i < 8; ++i) free(tag_strings[i]);

   return tags;
}

/**
 * initialisation routines for target and source processes
 */
struct field_interpolation_gen *
yac_new_field_interpolation_gen_mask(struct points *** points,
                                     struct mask *** masks,
                                     unsigned num_pointsets,
                                     char const * field_config_name,
                                     struct global_search * search,
                                     const char * other_field_config_name,
                                     unsigned is_source, unsigned is_target,
                                     struct interpolation_method ** methods,
                                     unsigned points_idx,
                                     enum yac_field_interpolation_tgt_mask_handling
                                     mask_handling,
                                     char const * weight_file_name,
                                     unsigned io_proc_step,
                                     int mapping_on_source) {


   if (is_source && is_target)
      return yac_new_field_interpolation_gen_src_tgt_mask(
        points, masks, num_pointsets, field_config_name, search,
        other_field_config_name, methods, points_idx, mask_handling,
        weight_file_name, io_proc_step, mapping_on_source);
   else if (is_source)
      return yac_new_field_interpolation_gen_src(
        points, num_pointsets, field_config_name, search,
        other_field_config_name, methods,
        weight_file_name, io_proc_step, mapping_on_source);
   else if (is_target)
      return yac_new_field_interpolation_gen_tgt_mask(
        points, masks, num_pointsets, field_config_name, search,
        other_field_config_name, points_idx, mask_handling, mapping_on_source);
   else
      return NULL;
}

/**
 * initialisation routines for target and source processes
 */
struct field_interpolation_gen *
yac_new_field_interpolation_gen(struct points *** points,
                                unsigned num_pointsets,
                                char const * field_config_name,
                                struct global_search * search,
                                const char * other_field_config_name,
                                unsigned is_source, unsigned is_target,
                                struct interpolation_method ** methods,
                                unsigned points_idx,
                                char const * weight_file_name,
                                unsigned io_proc_step, int mapping_on_source) {


   if (is_source && is_target)
      return yac_new_field_interpolation_gen_src_tgt(points, num_pointsets,
                                                     field_config_name, search,
                                                     other_field_config_name,
                                                     methods, points_idx,
                                                     weight_file_name,
                                                     io_proc_step,
                                                     mapping_on_source);
   else if (is_source)
      return yac_new_field_interpolation_gen_src(points, num_pointsets,
                                                 field_config_name, search,
                                                 other_field_config_name, methods,
                                                 weight_file_name, io_proc_step,
                                                 mapping_on_source);
   else if (is_target)
      return yac_new_field_interpolation_gen_tgt(points, num_pointsets,
                                                 field_config_name, search,
                                                 other_field_config_name,
                                                 points_idx, mapping_on_source);
   else
      return NULL;
}

/**
 * enddef routine for source an target processes
 */
void yac_enddef_field_interpolation_gen
  (struct field_interpolation_gen * interpolation_gen) {

   interpolation_gen->vtable->enddef(interpolation_gen);
}

struct field_interpolation *
yac_get_field_interpolation(struct field_interpolation_gen * interpolation_gen,
                            char const * local_field_name,
                            char const * remote_field_name) {

  return
    interpolation_gen->vtable->get_field_interpolation(interpolation_gen,
                                                       local_field_name,
                                                       remote_field_name);
}

void yac_field_interpolation_gen_add_weight_file_name(
  field_interpolation_gen interpolation_gen, char const * weight_file_name) {

  interpolation_gen->vtable->add_weight_file_name(interpolation_gen,
                                                  weight_file_name);
}

void yac_delete_field_interpolation_gen
  (struct field_interpolation_gen * interpolation_gen) {

   interpolation_gen->vtable->delete(interpolation_gen);
}

void yac_put_field(struct field_interpolation * interpolation, double **** data, int collection_size) {

   interpolation->vtable->put(interpolation, data, collection_size);
}

void yac_get_field(struct field_interpolation * interpolation, double *** data, int collection_size) {

   interpolation->vtable->get(interpolation, data, collection_size);
}

void yac_delete_field_interpolation(struct field_interpolation * interpolation) {

   interpolation->vtable->delete(interpolation);
}
