/**
 * @file field_interpolation_gen_tgt.c
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#include <stdlib.h>
#include <string.h>

#include "field_interpolation_gen.h"
#include "field_interpolation_gen_tgt.h"
#include "utils.h"
#include "ensure_array_size.h"
#include "global_search.h"
#include "points.h"

static void field_interpolation_gen_tgt_enddef (
  struct field_interpolation_gen * interpolation_gen);
static struct field_interpolation *
field_interpolation_gen_tgt_get_field_interpolation
  (struct field_interpolation_gen * interpolation_gen,
   char const * local_field_name,
   char const * remote_field_name);
static void
field_interpolation_gen_tgt_add_weight_file_name
  (field_interpolation_gen interpolation_gen, char const * weight_file_name);
static void field_interpolation_gen_tgt_delete(
  struct field_interpolation_gen * interpolation_gen);

static struct field_interpolation_gen_vtable
  field_interpolation_gen_tgt_vtable = {
    field_interpolation_gen_tgt_enddef,
    field_interpolation_gen_tgt_get_field_interpolation,
    field_interpolation_gen_tgt_add_weight_file_name,
    field_interpolation_gen_tgt_delete};

static void
field_interpolation_tgt_put(struct field_interpolation * interpolation,
                            double **** data, int collection_size);
static void
field_interpolation_tgt_get_mapping_on_src(
  struct field_interpolation * interpolation, double *** data,
  int collection_size);
static void
field_interpolation_tgt_get_mapping_on_tgt(
  struct field_interpolation * interpolation, double *** data,
  int collection_size);
static void
field_interpolation_tgt_delete(struct field_interpolation * interpolation);

static struct field_interpolation_vtable
  field_interpolation_tgt_mapping_on_src_vtable =
    {field_interpolation_tgt_put,
     field_interpolation_tgt_get_mapping_on_src,
     field_interpolation_tgt_delete},
  field_interpolation_tgt_mapping_on_tgt_vtable =
    {field_interpolation_tgt_put,
     field_interpolation_tgt_get_mapping_on_tgt,
     field_interpolation_tgt_delete};

struct field_interpolation_gen_tags {

   unsigned results_recv;
   unsigned point_data_send;
   unsigned point_data_header_send;
   unsigned tgt_point_location_send;
   unsigned interp_data_msg_count;
   unsigned interp_data_header;
   unsigned interp_data;
};

struct field_interpolation_gen_tgt_tags {

   unsigned scatter_interp_data_msg_count;
};

struct field_interpolation_exchange_tags {

   unsigned get;
};

struct recv_msg_src {
   struct dep_list list_from; //!< contains for each source process that interpolates
                              //!< points for the local process the positions in the
                              //!< receive buffer of data designated for each local
                              //!< subdomain (dimension [message idx]; element in
                              //!< list_from = domain)
   struct dep_list list_to;   //!< contains for each source process that interpolates
                              //!< points for the local process the positions where to
                              //!< copy the interpolated points in a get operation
                              //!< (dimension [message idx]; element in
                              //!< list_to = domain)
   unsigned rank;             //!< source process ranks for the messages in recv_list
};

struct src_indices {
  unsigned index;
  unsigned collection_factor;
};

struct mapping_data_src {
  struct recv_msg_src * recv_msgs;
  unsigned num_recv_messages; //!< number of message to receive in a get
                              //!< operation
  unsigned total_recv_buffer_size;
};

struct recv_msg_tgt {
  struct {
    unsigned count, offset, rank;
  } * msgs;
  unsigned num_msgs;
  unsigned total_buffer_size;
};

struct mapping_data_tgt {

  // link data
  struct {
    unsigned * num_points_per_subdomain;
    unsigned * num_links_per_point;
    struct src_indices * src_indices;
    unsigned * tgt_indices;
    double * weights;
    unsigned num_points;
    unsigned num_weights_per_link;
  } link_data;

  // fixed data
  struct {
    unsigned * tgt_indices;
    double * values;
    unsigned * num_values_per_subdomain;
    unsigned * num_points_per_value;
  } fixed;

  // receive data
  struct recv_msg_tgt recv;
};

union mapping_data {

  // mapping on source
  struct mapping_data_src src;
  // mapping on target
  struct mapping_data_tgt tgt;
};

struct field_interpolation_data_tgt {

  union mapping_data mapping_on;

  struct communicator * comm;
  unsigned num_subdomains;

  int mapping_on_source;

  unsigned reference_counter; //!< number of active references if the data in
                              //!< this field interpolation
};

struct field_interpolation_gen_tgt {

   struct field_interpolation_gen_vtable * vtable;

   struct field_interpolation_gen_tags tags;
   struct field_interpolation_gen_tgt_tags tgt_tags;
   struct field_interpolation_exchange_tags put_get_tags;

   struct field_interpolation_data_tgt * interp_data;

   struct global_search * search;

   enum yac_location point_location;
};

struct field_interpolation_tgt {

   struct field_interpolation_vtable * vtable;

   struct field_interpolation_exchange_tags tags;

   struct field_interpolation_data_tgt * interp_data;
};

struct temp_search_data {

   struct domain * local_domain;
   enum yac_location point_location;
   struct field_interpolation_gen_tgt * interp_tgt;
};

struct link {
  int src_index;
  int tgt_index;
  unsigned subdomain_id;
  unsigned pointset_index;
  double * weights;
  unsigned src_rank;
  unsigned index;
};

struct fixed {
  int * tgt_indices;
  unsigned num_points;
  double value;
};

struct temp_interp_data {

  struct field_interpolation_gen_tgt * interp_tgt;

  // link data
  struct link * links;
  unsigned num_links;
  double * weights;
  unsigned num_weights_per_link;

  // fixed data
  struct fixed * fixed;
  unsigned num_fixed;
  int * fixed_tgt_indices;
  unsigned num_fixed_tgt_indices;

  unsigned num_interp_data_msgs_received;
  unsigned num_interp_data_msgs_to_receive;

  struct domain * local_domain;
  enum yac_location point_location;
};

static void generate_tag_string(char * buf, char const * tag_base_str,
                                char const * source_comp_name,
                                char const * target_comp_name,
                                char const * source_domain_name,
                                char const * target_domain_name,
                                char const * source_field_name,
                                char const * target_field_name,
                                char const * tag_str) {

   strcpy(buf, tag_base_str);
   strcat(buf, source_comp_name);
   strcat(buf, target_comp_name);
   strcat(buf, source_domain_name);
   strcat(buf, target_domain_name);
   strcat(buf, source_field_name);
   strcat(buf, target_field_name);
   strcat(buf, tag_str);
}

struct field_interpolation_gen_tgt_tags
yac_field_interpolation_generate_tgt_tag(struct communicator * comm,
                                         char const * source_comp_name,
                                         char const * target_comp_name,
                                         char const * source_domain_name,
                                         char const * target_domain_name,
                                         char const * source_field_name,
                                         char const * target_field_name) {

   char const * str_base_str           = "field interpolation gen tgt";
   char const * str_generator_string   = "generator string";
   char const * str_scatter_interp_data_msg_count = "scatter interp data msg count";

   unsigned max_tag_strlen = strlen(str_scatter_interp_data_msg_count);

   unsigned max_str_buf_size = max_tag_strlen + strlen(str_base_str) +
                               strlen(source_comp_name) +
                               strlen(target_comp_name) +
                               strlen(source_domain_name) +
                               strlen(target_domain_name) +
                               strlen(source_field_name) +
                               strlen(target_field_name) + 1;

   char * unique_generator_string = malloc(max_str_buf_size);
   generate_tag_string(unique_generator_string, str_base_str, target_comp_name,
                       source_comp_name, target_domain_name, source_domain_name,
                       target_field_name, source_field_name,
                       str_generator_string);

   char * tag_strings[1];

   tag_strings[0] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[0], str_base_str, source_comp_name,
                       target_comp_name, source_domain_name, target_domain_name,
                       source_field_name, target_field_name,
                       str_scatter_interp_data_msg_count);

   unsigned search_tags[1];

   yac_get_unique_comm_tags(
      comm, unique_generator_string, (char const **)tag_strings, 1,
      search_tags);

   free(unique_generator_string);

   struct field_interpolation_gen_tgt_tags tags;

   tags.scatter_interp_data_msg_count = search_tags[0];

   for (unsigned i = 0; i < 1; ++i) free(tag_strings[i]);

   return tags;
}

/**
 * receives from source process the global ids of points which are interpolated by it
 *
 * the matching call to comm_recv_callback can be found in \ref send_interpolation_data
 */
static void
receive_interpolation_results_callback(struct communicator * comm,
                                       struct comm_message msg,
                                       void * user_data) {

   struct temp_search_data * search_data = (struct temp_search_data*)user_data;
   struct mapping_data_src * mapping_on_src =
      &(search_data->interp_tgt->interp_data->mapping_on.src);
   unsigned num_subdomains =
      search_data->interp_tgt->interp_data->num_subdomains;
   struct recv_msg_src * curr_recv_msg =
      mapping_on_src->recv_msgs + mapping_on_src->num_recv_messages;

   mapping_on_src->num_recv_messages++;
   mapping_on_src->total_recv_buffer_size += msg.count;
   curr_recv_msg->rank = msg.src_rank;

   unsigned * recv_points = malloc(msg.count * sizeof(*recv_points));
   unsigned * num_recv_points_per_subdomain =
      malloc(num_subdomains * sizeof(*num_recv_points_per_subdomain));
   unsigned * matching_ids_idx = malloc(msg.count * sizeof(*matching_ids_idx));

   unsigned offset = 0;

   // match the received global ids with the global ids of each
   // local subdomain and generate the receive list
   for (unsigned i = 0; i < num_subdomains; ++i) {

      // the global ids ((int*)(msg.buffer)) have to be sorted
      yac_match_global_ids(
        search_data->local_domain, i, search_data->point_location,
        (int*)(msg.buffer), msg.count, NULL, matching_ids_idx + offset,
        num_recv_points_per_subdomain+i, recv_points + offset,0);

      offset += num_recv_points_per_subdomain[i];
   }

   if (offset != msg.count)
      yac_internal_abort_message(
        "ERROR: could not find all received global ids in local data "
        "(receive_interpolation_results_callback)\n", __FILE__, __LINE__);

   yac_set_dependencies(&(curr_recv_msg->list_from), num_subdomains,
                        num_recv_points_per_subdomain, matching_ids_idx);
   unsigned * num_recv_points_per_subdomain_ =
      malloc(num_subdomains * sizeof(*num_recv_points_per_subdomain_));
   memcpy(num_recv_points_per_subdomain_, num_recv_points_per_subdomain,
          num_subdomains * sizeof(*num_recv_points_per_subdomain_));
   yac_set_dependencies(&(curr_recv_msg->list_to), num_subdomains,
                        num_recv_points_per_subdomain_, recv_points);

   free(search_data);
}

static void
pack_points_data_cell(void ** buffer, unsigned * buffer_size, unsigned * has_coordinates,
                      unsigned * num_points_to_send, struct global_search * search,
                      enum yac_location point_location, struct points *** points,
                      struct mask *** masks, unsigned points_idx, unsigned rank,
                      struct communicator * comm) {

   unsigned num_subdomains =
      yac_get_num_subdomains(yac_get_local_domain(search));

   struct dep_list const * search_results;

   search_results = yac_get_matching_local_cell_ids(search, rank);

   unsigned const * matching_local_cell_ids;
   unsigned num_matching_local_cell_ids;

   *num_points_to_send = 0;

   if (search_results != NULL)
      // for all subdomains
      for (unsigned i = 0; i < num_subdomains; ++i)
         *num_points_to_send += search_results->num_deps_per_element[i];

   *has_coordinates = 1;

   for (unsigned i = 0; i < num_subdomains; ++i)
      if (points == NULL || points[i] == NULL || points[i][points_idx] == NULL ||
          points[i][points_idx]->coordinates_x == NULL ||
          points[i][points_idx]->coordinates_x == NULL) {
         *has_coordinates = 0;
         break;
      }

   unsigned max_buffer_size =
      yac_comm_pack_size(*num_points_to_send, COMM_INT, comm) +
      ((*has_coordinates)?
        (yac_comm_pack_size(2 * *num_points_to_send, COMM_DBLE, comm)):(0));
   *buffer = malloc(max_buffer_size);
   *buffer_size = 0;

   if (search_results == NULL) return;

   // for all subdomains
   for (unsigned i = 0; i < num_subdomains; ++i) {

      matching_local_cell_ids =
        yac_get_dependencies_of_element(*search_results, i);
      num_matching_local_cell_ids = search_results->num_deps_per_element[i];

      int const * subdomain_global_ids =
        yac_get_global_ids(yac_get_local_domain(search), i, point_location);

      if (subdomain_global_ids == NULL)
         yac_internal_abort_message(
            "ERROR(field_interpolation_gen_tgt):global cell ids are "
            "undefined\n", __FILE__, __LINE__);

      struct grid * point_grid = NULL;

      if (*has_coordinates)
         point_grid = yac_get_point_grid(points[i][points_idx]);

      double temp_coord_buffer[2];

      for (unsigned j = 0; j < num_matching_local_cell_ids; ++j) {

         unsigned local_id;
         local_id = matching_local_cell_ids[j];

         if (masks != NULL && masks[i] != NULL && masks[i][points_idx] != NULL &&
             masks[i][points_idx]->mask[local_id] == 1) {

            --*num_points_to_send;
            continue;
         }

         yac_comm_pack((void*)(subdomain_global_ids + local_id), 1, COMM_INT,
                       *buffer, max_buffer_size, (int*)buffer_size, comm);
         if (*has_coordinates) {
            temp_coord_buffer[0] = yac_get_corner_x_coord(point_grid, local_id);
            temp_coord_buffer[1] = yac_get_corner_y_coord(point_grid, local_id);
            yac_comm_pack(temp_coord_buffer, 2, COMM_DBLE, *buffer,
                          max_buffer_size, (int*)buffer_size, comm);
         }
      }
   }
}

static int compare_int(const void * a, const void * b) {

  return ((*(int*)a) > (*(int*)b)) - (*((int*)a) < (*(int*)b));
}

/**
 * packs CORNER or EDGE point data, of points that might be able to be interpolated
 * by the  specified source process (defined by rank), of the local target process
 * into the array point_coords and global_ids
 *
 * this routine is slightly more complex than pack_points_data_cell, because the
 * global_search object only contains information about cells the overlap with the
 * source data, from this we have to derive the respective information for corners
 * or edges
 */
static void
pack_points_data_corner_edge(void ** buffer, unsigned * buffer_size, unsigned * has_coordinates,
                             unsigned * num_points_to_send, struct global_search * search,
                             enum yac_location point_location, struct points *** points,
                             struct mask *** masks, unsigned points_idx, unsigned rank,
                             struct communicator * comm) {

   struct domain * domain = yac_get_local_domain(search);
   unsigned num_subdomains = yac_get_num_subdomains(domain);

   unsigned * local_ids = NULL; // local corner or edge ids
   unsigned local_ids_array_size = 0;

   *num_points_to_send = 0;

   struct dep_list const * search_results;

   search_results = yac_get_matching_local_cell_ids(search, rank);

   if (search_results == NULL) return;

   unsigned const * matching_local_cell_ids;
   unsigned num_matching_local_cell_ids;

   *has_coordinates = 1;

   for (unsigned i = 0; i < num_subdomains; ++i)
      if (points == NULL || points[i] == NULL || points[i][points_idx] == NULL ||
          points[i][points_idx]->coordinates_x == NULL ||
          points[i][points_idx]->coordinates_x == NULL) {
         *has_coordinates = 0;
         break;
      }

   *buffer = NULL;
   *buffer_size = 0;
   unsigned max_buffer_size = 0;

   // for all subdomains
   for (unsigned i = 0; i < num_subdomains; ++i) {

      unsigned num_local_ids = 0;

      matching_local_cell_ids = yac_get_dependencies_of_element(*search_results, i);
      num_matching_local_cell_ids = search_results->num_deps_per_element[i];

      struct grid * grid;

      grid = yac_get_subdomain_grid(*domain, i);

      for (unsigned j = 0; j < num_matching_local_cell_ids; ++j) {

         unsigned const * curr_local_ids;
         unsigned curr_num_points;

         if (point_location == EDGE) {
            curr_local_ids = yac_get_cell_edge_indices(grid, matching_local_cell_ids[j]);
            curr_num_points = yac_get_num_cell_edges(grid, matching_local_cell_ids[j]);
         } else {
            curr_local_ids = yac_get_cell_corner_indices(grid, matching_local_cell_ids[j]);
            curr_num_points = yac_get_num_cell_corners(grid, matching_local_cell_ids[j]);
         }

         ENSURE_ARRAY_SIZE(local_ids, local_ids_array_size,
                           num_local_ids + curr_num_points);

         memcpy(local_ids + num_local_ids, curr_local_ids,
                curr_num_points * sizeof(*curr_local_ids));

         num_local_ids += curr_num_points;
      }

      // sort local_ids
      yac_mergesort(local_ids, num_local_ids, sizeof(*local_ids), compare_int);

      if (num_local_ids > 0) {

         // remove duplicated points
         yac_remove_duplicates_uint(local_ids, &num_local_ids);

         unsigned const * global_ranks = yac_get_global_ranks(domain, i,
                                                              point_location);
         unsigned num_halo_points =
            yac_get_num_total_halo_points(domain, i, point_location);

         // remove halo points
         if (num_halo_points > 0 && global_ranks != NULL) {

            unsigned j;

            // find first non-halo point
            for(j = 0; j < num_local_ids; ++j)
               if (global_ranks[local_ids[j]] == (unsigned)-1)
                  break;

            if (j < num_local_ids) {

               unsigned curr_pos = 0;

               if (j > 0) local_ids[0] = local_ids[j];

               for (unsigned i = j + 1; i < num_local_ids; ++i)
                  if ((local_ids[curr_pos] != local_ids[i]) &&
                      (global_ranks[local_ids[i]] == (unsigned)-1))
                     local_ids[++curr_pos] = local_ids[i];

               num_local_ids = curr_pos + 1;

            } else {

               num_local_ids = 0;
            }
         }
      }

      max_buffer_size =
        *buffer_size + yac_comm_pack_size(num_local_ids, COMM_INT, comm) +
        ((*has_coordinates)?(yac_comm_pack_size(2 * num_local_ids, COMM_DBLE,
                                               comm)):(0));

      *buffer = realloc(*buffer, max_buffer_size);

      int const * subdomain_global_ids;

      subdomain_global_ids = yac_get_global_ids(domain, i, point_location);

      if (subdomain_global_ids == NULL)
         yac_internal_abort_message("ERROR(field_interpolation_gen_tgt):global corner or edge ids "
                                    "are undefined\n", __FILE__, __LINE__);

      double temp_coord_buffer[2];

      // pack the corners into the buffers
      for (unsigned j = 0; j < num_local_ids; ++j) {

         if (masks != NULL && masks[i] != NULL && masks[i][points_idx] != NULL &&
             masks[i][points_idx]->mask[local_ids[j]] == 1)
            continue;

         yac_comm_pack((void*)(subdomain_global_ids + local_ids[j]), 1, COMM_INT,
                       *buffer, max_buffer_size, (int*)buffer_size, comm);

         if (*has_coordinates) {
            yac_get_point_coordinates(points[i][points_idx], local_ids[j],
                                      temp_coord_buffer);
            yac_comm_pack((void*)temp_coord_buffer, 2, COMM_DBLE, *buffer,
                          max_buffer_size, (int*)buffer_size, comm);
         }

         ++*num_points_to_send;
      }
   }

   free(local_ids);
}

/**
 * packs point data, of points that might be able to be interpolated by the
 * specified source process (defined by rank), of the local target process
 * into the array point_coords and global_ids
 */
static void
pack_point_data(void ** buffer, unsigned * buffer_size, unsigned * has_coordinates,
                unsigned * num_points_to_send, struct global_search * search,
                enum yac_location point_location, struct points *** points,
                struct mask *** masks, unsigned points_idx, unsigned rank,
                struct communicator * comm) {

   switch(point_location) {

      case (CELL):
         pack_points_data_cell(buffer, buffer_size, has_coordinates,
                               num_points_to_send, search, point_location,
                               points, masks, points_idx, rank, comm);
         break;
      case (CORNER):
      case (EDGE):
         pack_points_data_corner_edge(buffer, buffer_size, has_coordinates,
                                      num_points_to_send, search,
                                      point_location, points, masks, points_idx,
                                      rank, comm);
         break;
      default:
         yac_internal_abort_message(
          "ERROR: invalid point location in pack_point_data\n",
          __FILE__, __LINE__);
         break;
   };
}

/**
 * send the data about points that need to be interpolated to the
 * specified source process and sets up a comm_recv_callback for the results
 */
static void
send_interpolation_data(struct field_interpolation_gen_tgt * interp_tgt, unsigned rank,
                        struct global_search * search, struct points *** points,
                        struct mask *** masks, unsigned points_idx,
                        enum yac_field_interpolation_tgt_mask_handling mask_handling) {

   enum yac_location point_location = points[0][points_idx]->location;
   struct communicator * comm = interp_tgt->interp_data->comm;

   void * buffer;
   unsigned buffer_size;
   unsigned num_points_to_send;

   if (mask_handling != TGT_NO_INTERP)
      masks = NULL;

   unsigned has_coordinates;

   // generate list of cells/points might be interpolated by the other process
   pack_point_data(&buffer, &buffer_size, &has_coordinates,
                   &num_points_to_send, search, point_location, points, masks,
                   points_idx, rank, comm);

   if (interp_tgt->interp_data->mapping_on_source) {

      struct temp_search_data * search_data = malloc(1 * sizeof(*search_data));
      search_data->local_domain = yac_get_local_domain(search);
      search_data->point_location = point_location;
      search_data->interp_tgt = interp_tgt;

      // set up receive for results
      // the matching send can be found in the field_interpolation_gen_src routine
      // return_interpolation_points
      yac_comm_recv_callback(num_points_to_send, COMM_INT, rank,
                             interp_tgt->tags.results_recv, comm, search_data,
                             receive_interpolation_results_callback, NULL,
                             MSG_PRIO_HIGH);
   }

   unsigned header[4];

   header[0] = num_points_to_send;
   header[1] = point_location;
   header[2] = has_coordinates;
   header[3] = buffer_size;

   // send header (buffer size, point_location)
   // the matching receive can be found in the field_interpolation_gen_src constructor
   // (new_field_interpolation_gen_src)
   yac_comm_bsend(header, 4, COMM_UINT, rank,
                  interp_tgt->tags.point_data_header_send, comm);

   // send point data
   if (num_points_to_send > 0)
      // send packed data
      yac_comm_bsend(buffer, buffer_size, COMM_PACKED, rank,
                     interp_tgt->tags.point_data_send, comm);

   free(buffer);
}

static int compare_links_src_rank(const void * a, const void * b) {

  struct link * link_a = (struct link*)a;
  struct link * link_b = (struct link*)b;

  if (link_a->src_rank != link_b->src_rank)
    return ( link_a->src_rank - link_b->src_rank );
  if (link_a->pointset_index != link_b->pointset_index)
    return ( link_a->pointset_index - link_b->pointset_index );
  return ( link_a->src_index - link_b->src_index );
}

static int compare_links_subdomain_id(const void * a, const void * b) {

  struct link * link_a = (struct link*)a;
  struct link * link_b = (struct link*)b;

  if (link_a->subdomain_id != link_b->subdomain_id)
    return ( link_a->subdomain_id - link_b->subdomain_id );
  if (link_a->tgt_index != link_b->tgt_index)
    return ( link_a->tgt_index - link_b->tgt_index );
  if (link_a->pointset_index != link_b->pointset_index)
    return ( link_a->pointset_index - link_b->pointset_index );
  return ( link_a->src_index - link_b->src_index );
}

static void
generate_receive_data(struct mapping_data_tgt * mapping_on_tgt,
                      unsigned * recv_count_per_rank, unsigned comm_size,
                      unsigned * src_index_to_collection_factor) {

  unsigned num_recv_messages = 0;
  for (unsigned i = 0; i < comm_size; ++i)
    if (recv_count_per_rank[i] != 0) ++num_recv_messages;

  mapping_on_tgt->recv.msgs =
    malloc(num_recv_messages * sizeof(*(mapping_on_tgt->recv.msgs)));
  mapping_on_tgt->recv.num_msgs = num_recv_messages;

  unsigned total_buffer_size = 0;

  for (unsigned i = 0, j = 0, l = 0; i < comm_size; ++i) {
    if (recv_count_per_rank[i] != 0) {

      unsigned count = recv_count_per_rank[i];
      mapping_on_tgt->recv.msgs[j].count = count;
      mapping_on_tgt->recv.msgs[j].offset = total_buffer_size;
      mapping_on_tgt->recv.msgs[j].rank = i;
      total_buffer_size += count;
      ++j;
      for (unsigned k = 0; k < count; ++k, ++l)
        src_index_to_collection_factor[l] = total_buffer_size;
    }
  }
  mapping_on_tgt->recv.total_buffer_size = total_buffer_size;
}

static int compare_fixed(const void * a, const void * b) {

  return (((struct fixed *)a)->value > ((struct fixed *)b)->value) -
         (((struct fixed *)a)->value < ((struct fixed *)b)->value);
}

static void
generate_interpolation_data(struct temp_interp_data * temp_interp_data,
                            unsigned * src_index_to_collection_factor) {

  struct mapping_data_tgt * mapping_on_tgt =
    &(temp_interp_data->interp_tgt->interp_data->mapping_on.tgt);
  unsigned num_subdomains =
    temp_interp_data->interp_tgt->interp_data->num_subdomains;
  unsigned num_links = temp_interp_data->num_links;

  if (num_links > 0) {

    unsigned num_linked_points = 1;

    // sort links by subdomain_id, tgt_index, pointset_index, src_index
    qsort(temp_interp_data->links, num_links,
          sizeof(*(temp_interp_data->links)), compare_links_subdomain_id);

    for (unsigned i = 1; i < num_links; ++i)
      if ((temp_interp_data->links[i].tgt_index !=
           temp_interp_data->links[i-1].tgt_index) ||
          (temp_interp_data->links[i].subdomain_id !=
           temp_interp_data->links[i-1].subdomain_id)) ++num_linked_points;

    mapping_on_tgt->link_data.num_links_per_point =
      calloc(num_linked_points,
             sizeof(*(mapping_on_tgt->link_data.num_links_per_point)));
    mapping_on_tgt->link_data.src_indices =
      malloc(num_links * sizeof(*(mapping_on_tgt->link_data.src_indices)));
    mapping_on_tgt->link_data.tgt_indices =
      malloc(num_linked_points *
             sizeof(*(mapping_on_tgt->link_data.tgt_indices)));
    mapping_on_tgt->link_data.weights =
      malloc(num_links * sizeof(*(mapping_on_tgt->link_data.weights)));
    mapping_on_tgt->link_data.num_points = num_linked_points;
    mapping_on_tgt->link_data.num_weights_per_link =
      temp_interp_data->num_weights_per_link;

    unsigned curr_subdomain_id = -1;
    int curr_tgt_index = -1;
    for (unsigned i = 0, points_idx_link = -1; i < num_links; ++i) {

      mapping_on_tgt->link_data.src_indices[i].index =
        temp_interp_data->links[i].index;
      mapping_on_tgt->link_data.src_indices[i].collection_factor =
          src_index_to_collection_factor[temp_interp_data->links[i].index];
      for (unsigned k = 0; k < temp_interp_data->num_weights_per_link; ++k)
        mapping_on_tgt->link_data.weights
          [i * temp_interp_data->num_weights_per_link + k] =
            temp_interp_data->links[i].weights[k];

      if ((temp_interp_data->links[i].subdomain_id != curr_subdomain_id) ||
          (temp_interp_data->links[i].tgt_index != curr_tgt_index)) {

        ++points_idx_link;
        mapping_on_tgt->link_data.tgt_indices[points_idx_link] =
          temp_interp_data->links[i].tgt_index;
        curr_tgt_index = temp_interp_data->links[i].tgt_index;
        curr_subdomain_id = temp_interp_data->links[i].subdomain_id;
        mapping_on_tgt->link_data.num_points_per_subdomain[curr_subdomain_id]++;
      }
      mapping_on_tgt->link_data.num_links_per_point[points_idx_link]++;
    }
  }

  if (temp_interp_data->num_fixed > 0) {

    qsort(temp_interp_data->fixed, temp_interp_data->num_fixed,
          sizeof(*(temp_interp_data->fixed)), compare_fixed);

    unsigned num_fixed_points = 0;

    for (unsigned i = 0; i < temp_interp_data->num_fixed; ++i)
      num_fixed_points += temp_interp_data->fixed[i].num_points;

    unsigned tgt_indices_array_size = 0;
    unsigned num_tgt_indices = 0;
    mapping_on_tgt->fixed.values =
      malloc(num_subdomains * temp_interp_data->num_fixed *
             sizeof(*(mapping_on_tgt->fixed.values)));
    mapping_on_tgt->fixed.num_points_per_value =
      calloc(num_subdomains * temp_interp_data->num_fixed,
             sizeof(*(mapping_on_tgt->fixed.num_points_per_value)));
    unsigned num_values = 0;

    // sort global ids for each fixed value
    for (unsigned i = 0; i < temp_interp_data->num_fixed; ++i)
      yac_quicksort_index(temp_interp_data->fixed[i].tgt_indices,
                          temp_interp_data->fixed[i].num_points, NULL);

    // for all subdomains
    for (unsigned subdomain_idx = 0;
         subdomain_idx < num_subdomains;
         ++subdomain_idx) {

      unsigned temp_num_tgt_indices = num_tgt_indices;
      unsigned temp_num_values = num_values;

      // for all fixed values
      for (unsigned i = 0; i < temp_interp_data->num_fixed; ++i) {

        unsigned num_matching_ids;

        ENSURE_ARRAY_SIZE(
          mapping_on_tgt->fixed.tgt_indices,
          tgt_indices_array_size, num_tgt_indices +
          temp_interp_data->fixed[i].num_points);

        // get the local ids
        yac_match_global_ids(
          temp_interp_data->local_domain, subdomain_idx,
          temp_interp_data->point_location,
          temp_interp_data->fixed[i].tgt_indices,
          temp_interp_data->fixed[i].num_points, NULL, NULL, &num_matching_ids,
          mapping_on_tgt->fixed.tgt_indices + num_tgt_indices, 0);

        if (num_matching_ids > 0) {

          num_tgt_indices += num_matching_ids;
          mapping_on_tgt->fixed.values[num_values] =
            temp_interp_data->fixed[i].value;
          mapping_on_tgt->fixed.num_values_per_subdomain[subdomain_idx]++;
          mapping_on_tgt->fixed.num_points_per_value[num_values] =
            num_matching_ids;
          num_values++;
        }
      }

      // merge entries with identical fixed values
      for (unsigned i = temp_num_values + 1, j = temp_num_values;
           i < num_values; ++i) {

        if (mapping_on_tgt->fixed.values[j] ==
            mapping_on_tgt->fixed.values[i]) {

          mapping_on_tgt->fixed.num_values_per_subdomain[subdomain_idx]--;
          mapping_on_tgt->fixed.num_points_per_value[j] +=
            mapping_on_tgt->fixed.num_points_per_value[i];
          mapping_on_tgt->fixed.num_points_per_value[i] = 0;
        } else {
          ++j;
        }
      }
      for (unsigned i = temp_num_values; i < num_values; ++i) {

        if (mapping_on_tgt->fixed.num_points_per_value[i] > 0) {
            if ((i != num_values - 1) &&
                (mapping_on_tgt->fixed.num_points_per_value[i+1] == 0))
              yac_quicksort_index(
                (int*)(mapping_on_tgt->fixed.tgt_indices) +
                temp_num_tgt_indices,
                mapping_on_tgt->fixed.num_points_per_value[i], NULL);
          temp_num_tgt_indices += mapping_on_tgt->fixed.num_points_per_value[i];

          if (i != temp_num_values) {

            mapping_on_tgt->fixed.values[temp_num_values] =
              mapping_on_tgt->fixed.values[i];
            mapping_on_tgt->fixed.num_points_per_value[temp_num_values] =
              mapping_on_tgt->fixed.num_points_per_value[i];
          }
          ++temp_num_values;
        }
      }
      num_values = temp_num_values;
    }

    mapping_on_tgt->fixed.num_points_per_value =
      realloc(mapping_on_tgt->fixed.num_points_per_value, num_values *
              sizeof(*(mapping_on_tgt->fixed.num_points_per_value)));
    mapping_on_tgt->fixed.tgt_indices =
      realloc(mapping_on_tgt->fixed.tgt_indices, num_tgt_indices *
              sizeof(*(mapping_on_tgt->fixed.tgt_indices)));
  }
}

static void
finalise_interp_data(struct temp_interp_data * temp_interp_data) {

  struct field_interpolation_gen_tgt * interp_tgt =
    temp_interp_data->interp_tgt;

  // cancel receive for links data header message, because we have received all
  // interpolation data from the source processes
  yac_comm_cancel_recv(interp_tgt->interp_data->comm,
                       interp_tgt->tags.interp_data_header);

  unsigned * src_index_to_collection_factor = NULL;

  // if the interpolation data contains links
  if (temp_interp_data->num_links > 0) {

    struct link * links = temp_interp_data->links;
    unsigned num_links = temp_interp_data->num_links;

    for (unsigned i = 0; i < num_links; ++i)
      links[i].weights = temp_interp_data->num_weights_per_link * i +
                         temp_interp_data->weights;

    // sort links by source rank, pointset_index, src_index
    qsort(links, num_links, sizeof(*links), compare_links_src_rank);

    unsigned * recv_count_per_rank =
      calloc(yac_get_comm_size(interp_tgt->interp_data->comm),
             sizeof(*recv_count_per_rank));

    recv_count_per_rank[links[0].src_rank] = 1;
    links[0].index = 0;
    unsigned total_buffer_size = 0;
    for (unsigned i = 1; i < num_links; ++i) {

      if ((links[i-1].src_rank != links[i].src_rank) ||
          (links[i-1].pointset_index != links[i].pointset_index) ||
          (links[i-1].src_index != links[i].src_index)) {
        ++recv_count_per_rank[links[i].src_rank];
        ++total_buffer_size;
      }
      links[i].index = total_buffer_size;
    }
    ++total_buffer_size;

    src_index_to_collection_factor =
      malloc(total_buffer_size * sizeof(*src_index_to_collection_factor));

    // generate receive data
    generate_receive_data(&(interp_tgt->interp_data->mapping_on.tgt),
                          recv_count_per_rank,
                          yac_get_comm_size(interp_tgt->interp_data->comm),
                          src_index_to_collection_factor);
    free(recv_count_per_rank);
  }

  {
    struct fixed * fixed = temp_interp_data->fixed;
    unsigned num_fixed = temp_interp_data->num_fixed;
    int * fixed_tgt_indices = temp_interp_data->fixed_tgt_indices;
    for (unsigned i = 0, offset = 0; i < num_fixed;
         offset += fixed[i++].num_points)
      fixed[i].tgt_indices = fixed_tgt_indices + offset;
  }

  // generate interpolation data
  generate_interpolation_data(temp_interp_data, src_index_to_collection_factor);

  free(src_index_to_collection_factor);
  free(temp_interp_data->links);
  free(temp_interp_data->weights);
  free(temp_interp_data->fixed);
  free(temp_interp_data->fixed_tgt_indices);
  free(temp_interp_data);
}

static void
receive_intep_data_callback(struct communicator * comm,
                            struct comm_message msg,
                            void * user_data);

/**
 * receives size of link message from source processes
 */
static void
receive_interp_data_header_callback(struct communicator * comm,
                                    struct comm_message msg,
                                    void * user_data) {

  struct temp_interp_data * temp_interp_data =
    (struct temp_interp_data *)user_data;

  // set up receive for link data
  yac_comm_recv_callback(*(unsigned*)(msg.buffer), COMM_PACKED, msg.src_rank,
                         temp_interp_data->interp_tgt->tags.interp_data, comm,
                         temp_interp_data, receive_intep_data_callback, NULL,
                         MSG_PRIO_MIDDLE);
}

static void unpack_link_data(
  struct comm_message msg, int * pos, struct communicator * comm,
  struct temp_interp_data * temp_interp_data) {

  unsigned num_weights_per_link;
  yac_comm_unpack(msg.buffer, msg.count, pos, &num_weights_per_link, 1,
                  COMM_UINT, comm);

  if (temp_interp_data->num_weights_per_link == -1)
    temp_interp_data->num_weights_per_link = num_weights_per_link;
  else if (temp_interp_data->num_weights_per_link != num_weights_per_link)
    yac_internal_abort_message(
      "ERROR(receive_intep_data_callback): num_weights_per_link do not match",
      __FILE__, __LINE__);

  unsigned num_packed_links;
  yac_comm_unpack(msg.buffer, msg.count, pos, &num_packed_links, 1, COMM_UINT,
                  comm);

  temp_interp_data->links =
    realloc(temp_interp_data->links,
            (temp_interp_data->num_links + num_packed_links) *
            sizeof(*(temp_interp_data->links)));
  temp_interp_data->weights =
    realloc(temp_interp_data->weights,
            (temp_interp_data->num_links + num_packed_links) *
            num_weights_per_link * sizeof(*(temp_interp_data->weights)));

  int * global_target_ids =
    malloc(num_packed_links * sizeof(*global_target_ids));
  int * orig_pos = malloc(num_packed_links * sizeof(*orig_pos));

  struct link * links = temp_interp_data->links + temp_interp_data->num_links;
  double * weights = temp_interp_data->weights +
                     temp_interp_data->num_links * num_weights_per_link;
  temp_interp_data->num_links += num_packed_links;

  for (unsigned i = 0; i < num_packed_links; ++i) {

    yac_comm_unpack(msg.buffer, msg.count, pos, &(links[i].src_index), 1,
                    COMM_INT, comm);
    yac_comm_unpack(msg.buffer, msg.count, pos, global_target_ids + i, 1,
                    COMM_INT, comm);
    yac_comm_unpack(msg.buffer, msg.count, pos, &(links[i].pointset_index), 1,
                    COMM_UINT, comm);
    yac_comm_unpack(msg.buffer, msg.count, pos, i * num_weights_per_link +
                    weights, num_weights_per_link, COMM_DBLE, comm);
    links[i].src_rank = msg.src_rank;
    links[i].tgt_index = -1;
    links[i].subdomain_id = -1;
    orig_pos[i] = i;
  }

  unsigned * local_ids = malloc(num_packed_links * sizeof(*local_ids));
  unsigned * matching_ids_idx =
    malloc(num_packed_links * sizeof(*matching_ids_idx));

  yac_quicksort_index(global_target_ids, num_packed_links, orig_pos);

  struct domain * local_domain = temp_interp_data->local_domain;
  unsigned num_subdomains =
    temp_interp_data->interp_tgt->interp_data->num_subdomains;

  for (unsigned i = 0; i < num_subdomains; ++i) {

    unsigned num_matching_ids;
    yac_match_global_ids(local_domain, i, temp_interp_data->point_location,
                         global_target_ids, num_packed_links, NULL,
                         matching_ids_idx, &num_matching_ids, local_ids,0);

    for (unsigned j = 0; j < num_matching_ids; ++j) {

      links[orig_pos[matching_ids_idx[j]]].tgt_index = local_ids[j];
      links[orig_pos[matching_ids_idx[j]]].subdomain_id = i;
    }
  }

  free(matching_ids_idx);
  free(local_ids);
  free(orig_pos);
  free(global_target_ids);
}

static void unpack_fixed_data(
  struct comm_message msg, int * pos, struct communicator * comm,
  struct temp_interp_data * temp_interp_data) {

  unsigned curr_num_packed_fixed, curr_num_fixed_tgt_indices;
  yac_comm_unpack(msg.buffer, msg.count, pos, &curr_num_packed_fixed, 1, COMM_UINT,
                  comm);
  yac_comm_unpack(msg.buffer, msg.count, pos, &curr_num_fixed_tgt_indices, 1,
                  COMM_UINT, comm);

  unsigned num_fixed = temp_interp_data->num_fixed;
  struct fixed * fixed = temp_interp_data->fixed;
  int * fixed_tgt_indices = temp_interp_data->fixed_tgt_indices;
  unsigned num_fixed_tgt_indices = temp_interp_data->num_fixed_tgt_indices;

  fixed = realloc(fixed, (num_fixed + curr_num_packed_fixed) * sizeof(*fixed));
  fixed_tgt_indices =
    realloc(fixed_tgt_indices,
            (num_fixed_tgt_indices + curr_num_fixed_tgt_indices) *
            sizeof(*fixed_tgt_indices));

  temp_interp_data->fixed = fixed;
  temp_interp_data->num_fixed += curr_num_packed_fixed;
  temp_interp_data->fixed_tgt_indices = fixed_tgt_indices;
  temp_interp_data->num_fixed_tgt_indices += curr_num_fixed_tgt_indices;

  fixed += num_fixed;

  for (unsigned i = 0, offset = num_fixed_tgt_indices;
       i < curr_num_packed_fixed; ++i) {

    unsigned num_points;
    yac_comm_unpack(msg.buffer, msg.count, pos, &num_points, 1, COMM_UINT,
                    comm);
    fixed[i].num_points = num_points;
    yac_comm_unpack(msg.buffer, msg.count, pos, fixed_tgt_indices + offset,
                    num_points, COMM_INT, comm);
    yac_comm_unpack(msg.buffer, msg.count, pos, &(fixed[i].value), 1,
                    COMM_DBLE, comm);
    offset += num_points;
  }
}

/**
 * receives interpolation data from source processes
 */
static void
receive_intep_data_callback(struct communicator * comm,
                            struct comm_message msg,
                            void * user_data) {

  struct temp_interp_data * temp_interp_data =
    (struct temp_interp_data *)user_data;

  int pos = 0;
  unpack_link_data(msg, &pos, comm, temp_interp_data);
  unpack_fixed_data(msg, &pos, comm, temp_interp_data);

  temp_interp_data->num_interp_data_msgs_received++;

  if ((temp_interp_data->num_interp_data_msgs_to_receive != -1) &&
      (temp_interp_data->num_interp_data_msgs_to_receive ==
       temp_interp_data->num_interp_data_msgs_received))
    finalise_interp_data(temp_interp_data);
  else
    // set up receive for next header of link data from source processes
    yac_comm_recv_callback(
      1, COMM_UINT, COMM_ANYSOURCE,
      temp_interp_data->interp_tgt->tags.interp_data_header, comm,
      temp_interp_data, receive_interp_data_header_callback, NULL,
      MSG_PRIO_MIDDLE);
}

/**
 * receives the number of link messges per processes from the root of the source
 * component
 */
static void
scatter_interp_data_msg_count_callback(struct communicator * comm,
                                       struct comm_message msg,
                                       void * user_data) {

  struct temp_interp_data * interp_data = (struct temp_interp_data *)user_data;
  unsigned num_interp_data_msgs_to_receive = *(unsigned*)(msg.buffer);

  if (interp_data->num_interp_data_msgs_received ==
      num_interp_data_msgs_to_receive)
    finalise_interp_data(interp_data);
  else
    interp_data->num_interp_data_msgs_to_receive =
      num_interp_data_msgs_to_receive;
}

/**
 * receives the number of link messges per processes from the root of the source
 * component
 */
static void
receive_interp_data_msg_count_callback(struct communicator * comm,
                                       struct comm_message msg,
                                       void * user_data) {

  struct temp_interp_data * interp_data = (struct temp_interp_data *)user_data;
  struct field_interpolation_gen_tgt * interp_tgt = interp_data->interp_tgt;
  struct communicator * domain_comm =
    yac_get_domain_comm(interp_data->local_domain);
  unsigned domain_comm_rank = yac_get_comm_rank(domain_comm);
  unsigned domain_comm_size = yac_get_comm_size(domain_comm);

  unsigned * comm_ranks = malloc(domain_comm_size * sizeof(*comm_ranks));
  yac_comm_translate_ranks(comm, domain_comm, comm_ranks);

  unsigned * num_interp_data_msgs_per_target =
    malloc(domain_comm_size * sizeof(*num_interp_data_msgs_per_target));
  for (unsigned i = 0;  i < domain_comm_size; ++i)
    num_interp_data_msgs_per_target[i] =
      ((unsigned*)(msg.buffer))[comm_ranks[i]];

  yac_comm_scatter_callback(num_interp_data_msgs_per_target, 1, COMM_UINT,
                            domain_comm_rank,
                            interp_tgt->tgt_tags.scatter_interp_data_msg_count,
                            domain_comm, interp_data,
                            scatter_interp_data_msg_count_callback,
                            MSG_PRIO_LOW);

  free(num_interp_data_msgs_per_target);
  free(comm_ranks);
}

static void
receive_interp_data(struct field_interpolation_gen_tgt * interp_tgt) {

  struct communicator * comm = interp_tgt->interp_data->comm;
  unsigned comm_rank = yac_get_comm_rank(comm);
  unsigned comm_size = yac_get_comm_size(comm);
  struct domain * local_domain = yac_get_local_domain(interp_tgt->search);
  struct communicator * domain_comm = yac_get_domain_comm(local_domain);
  unsigned domain_comm_rank = yac_get_comm_rank(domain_comm);
  unsigned domain_comm_size = yac_get_comm_size(domain_comm);
  unsigned domain_comm_local_root_rank;
  unsigned local_root_rank =
    yac_get_lowest_local_domain_rank(interp_tgt->search,
                                     &domain_comm_local_root_rank);
  unsigned remote_root_rank =
    yac_get_lowest_remote_domain_rank(interp_tgt->search);

  struct temp_interp_data * interp_data = malloc(1 * sizeof(*interp_data));

  interp_data->interp_tgt = interp_tgt;
  interp_data->links = NULL;
  interp_data->num_links = 0;
  interp_data->weights = NULL;
  interp_data->fixed = NULL;
  interp_data->num_fixed = 0;
  interp_data->fixed_tgt_indices = NULL;
  interp_data->num_fixed_tgt_indices = 0;
  interp_data->num_interp_data_msgs_to_receive = -1;
  interp_data->num_interp_data_msgs_received = 0;
  interp_data->num_weights_per_link = -1;
  interp_data->local_domain = local_domain;
  interp_data->point_location = interp_tgt->point_location;

  // set up receive for header of link data from source processes
  yac_comm_recv_callback(1, COMM_UINT, COMM_ANYSOURCE,
                         interp_tgt->tags.interp_data_header, comm,
                         interp_data, receive_interp_data_header_callback, NULL,
                         MSG_PRIO_MIDDLE);

  // if the local process is the root of target component
  if (local_root_rank == comm_rank) {

    // set up receive for number of link message per rank from source to the
    // target processes
    yac_comm_recv_callback(comm_size, COMM_UINT, remote_root_rank,
                           interp_tgt->tags.interp_data_msg_count, comm,
                           interp_data, receive_interp_data_msg_count_callback,
                           NULL, MSG_PRIO_LOW);

  } else {

    // set up receive for number of link messages to be received by the local
    // process
    yac_comm_scatter_callback(NULL, 1, COMM_UINT, domain_comm_local_root_rank,
                              interp_tgt->tgt_tags.
                                scatter_interp_data_msg_count,
                              domain_comm, interp_data,
                              scatter_interp_data_msg_count_callback,
                              MSG_PRIO_LOW);
  }
}

static void init_mapping_data(union mapping_data * mapping_on,
                              unsigned mapping_on_source,
                              unsigned num_result_requests,
                              unsigned num_subdomains) {

  if (mapping_on_source) {

    mapping_on->src.recv_msgs =
      malloc(num_result_requests * sizeof(*(mapping_on->src.recv_msgs)));
    mapping_on->src.num_recv_messages = 0;
    mapping_on->src.total_recv_buffer_size = 0;

  } else {

    mapping_on->tgt.link_data.num_points_per_subdomain =
      calloc(num_subdomains,
             sizeof(*(mapping_on->tgt.link_data.num_points_per_subdomain)));
    mapping_on->tgt.link_data.num_links_per_point = NULL;
    mapping_on->tgt.link_data.src_indices = NULL;
    mapping_on->tgt.link_data.tgt_indices = NULL;
    mapping_on->tgt.link_data.weights = NULL;
    mapping_on->tgt.link_data.num_points = 0;
    mapping_on->tgt.link_data.num_weights_per_link = 0;

    mapping_on->tgt.fixed.tgt_indices = NULL;
    mapping_on->tgt.fixed.values = NULL;
    mapping_on->tgt.fixed.num_values_per_subdomain =
      calloc(num_subdomains,
             sizeof(*(mapping_on->tgt.fixed.num_values_per_subdomain)));
    mapping_on->tgt.fixed.num_points_per_value = NULL;

    mapping_on->tgt.recv.msgs = NULL;
    mapping_on->tgt.recv.num_msgs = 0;
    mapping_on->tgt.recv.total_buffer_size = 0;
  }
};

struct field_interpolation_gen *
yac_new_field_interpolation_gen_tgt_mask_(
   struct points *** points, struct mask *** masks, unsigned num_pointsets,
   char const * field_config_name, struct global_search * search,
   const char * other_field_config_name, unsigned points_idx,
   enum yac_field_interpolation_tgt_mask_handling mask_handling,
   struct field_interpolation_tags tags, int mapping_on_source) {

   if (num_pointsets == 0 || points == NULL)
      yac_internal_abort_message("ERROR: no point set defined\n", __FILE__, __LINE__);

   struct field_interpolation_gen_tgt * interp_tgt;

   interp_tgt = malloc(1 * sizeof(*interp_tgt));
   interp_tgt->interp_data = malloc(1 * sizeof(*(interp_tgt->interp_data)));

   interp_tgt->vtable = &field_interpolation_gen_tgt_vtable;
   interp_tgt->interp_data->num_subdomains =
     yac_get_num_subdomains(yac_get_local_domain(search));
   struct communicator * comm = yac_get_communicator(search);
   interp_tgt->interp_data->comm = comm;
   interp_tgt->point_location = points[0][points_idx]->location;
   interp_tgt->interp_data->reference_counter = 1;
   interp_tgt->search = search;
   interp_tgt->interp_data->mapping_on_source = mapping_on_source;

   interp_tgt->tags.results_recv = tags.results_exchange;
   interp_tgt->tags.point_data_send = tags.point_data_exchange;
   interp_tgt->tags.point_data_header_send = tags.point_data_header_exchange;
   interp_tgt->tags.tgt_point_location_send = tags.tgt_point_location_exchange;
   interp_tgt->tags.interp_data_msg_count = tags.interp_data_msg_count;
   interp_tgt->tags.interp_data_header = tags.interp_data_header;
   interp_tgt->tags.interp_data = tags.interp_data;
   interp_tgt->put_get_tags.get = tags.exchange;

   interp_tgt->tgt_tags =
      yac_field_interpolation_generate_tgt_tag(
         yac_get_domain_comm(yac_get_local_domain(search)),
         yac_get_remote_component_name(search),
         yac_get_local_component_name(search),
         yac_get_remote_domain_name(search), yac_get_local_domain_name(search),
         other_field_config_name, field_config_name);

   unsigned comm_size = yac_get_comm_size(comm);
   unsigned comm_rank = yac_get_comm_rank(comm);

   // if this is the target process with the lowest rank
   if (comm_rank == yac_get_lowest_local_domain_rank(search, NULL)) {

      int tgt_location = points[0][points_idx]->location;

      // send the target location to a source process, which distributes
      // it among the other source processes
      // some source that have no overlap with any target process have no
      // communication partner that could give them this information
      yac_comm_bsend(&tgt_location, 1, COMM_INT,
                     yac_get_lowest_remote_domain_rank(search),
                     interp_tgt->tags.tgt_point_location_send, comm);
   }

   int * has_overlap = malloc(comm_size * sizeof(*has_overlap));
   unsigned num_result_requests = 0;

   for (unsigned i = 0; i < comm_size; ++i) {
      has_overlap[i] = yac_target_process_has_overlap_with(search, i);
      if (has_overlap[i]) num_result_requests++;
   }

   init_mapping_data(&(interp_tgt->interp_data->mapping_on), mapping_on_source,
                     num_result_requests,
                     yac_get_num_subdomains(yac_get_local_domain(search)));

   // if the mapping is supposed to be done on the target, set up receive for
   // interpolation data
   if (!mapping_on_source) receive_interp_data(interp_tgt);

   for (unsigned i = 0; i < comm_size; ++i)
      if (has_overlap[i])
         // send points to be interpolated to the source processes
         send_interpolation_data(interp_tgt, i, search, points, masks,
                                 points_idx, mask_handling);

   free(has_overlap);

   return (struct field_interpolation_gen*)interp_tgt;
}

struct field_interpolation_gen *
yac_new_field_interpolation_gen_tgt_mask(
   struct points *** points, struct mask *** masks, unsigned num_pointsets,
   char const * field_config_name, struct global_search * search,
   const char * other_field_config_name, unsigned points_idx,
   enum yac_field_interpolation_tgt_mask_handling mask_handling,
   int mapping_on_source) {

   if (num_pointsets == 0 || points == NULL)
      yac_internal_abort_message(
         "ERROR: no point set defined\n", __FILE__, __LINE__);

   struct field_interpolation_tags tags =
      yac_field_interpolation_generate_tag(
         yac_get_communicator(search),
         yac_get_remote_component_name(search),
         yac_get_local_component_name(search),
         yac_get_remote_domain_name(search),
         yac_get_local_domain_name(search),
         other_field_config_name,
         field_config_name);

   return
      (struct field_interpolation_gen*)
         yac_new_field_interpolation_gen_tgt_mask_(
            points, masks, num_pointsets, field_config_name, search,
            other_field_config_name, points_idx, mask_handling, tags,
            mapping_on_source);
}

struct field_interpolation_gen *
yac_new_field_interpolation_gen_tgt_(struct points *** points,
                                     unsigned num_pointsets,
                                     char const * field_config_name,
                                     struct global_search * search,
                                     const char * other_field_config_name,
                                     unsigned points_idx,
                                     struct field_interpolation_tags tags,
                                     int mapping_on_source) {

   return
      yac_new_field_interpolation_gen_tgt_mask_(
         points, NULL, num_pointsets, field_config_name, search,
         other_field_config_name, points_idx, TGT_IGNORE_MASK, tags,
         mapping_on_source);
}

struct field_interpolation_gen *
yac_new_field_interpolation_gen_tgt(struct points *** points,
                                    unsigned num_pointsets,
                                    char const * field_config_name,
                                    struct global_search * search,
                                    const char * other_field_config_name,
                                    unsigned points_idx,
                                    int mapping_on_source) {

   return yac_new_field_interpolation_gen_tgt_mask(points, NULL, num_pointsets,
                                                   field_config_name, search,
                                                   other_field_config_name,
                                                   points_idx, TGT_IGNORE_MASK,
                                                   mapping_on_source);
}

static struct field_interpolation *
field_interpolation_gen_tgt_get_field_interpolation(
  struct field_interpolation_gen * interpolation_gen,
  char const * local_field_name, char const * remote_field_name) {

  struct field_interpolation_gen_tgt * interp_gen_tgt;
  struct field_interpolation_tgt * interp_tgt;

  interp_gen_tgt = (struct field_interpolation_gen_tgt*)interpolation_gen;
  interp_tgt = malloc(1 * sizeof(*interp_tgt));

  if (interp_gen_tgt->interp_data->mapping_on_source)
    interp_tgt->vtable = &field_interpolation_tgt_mapping_on_src_vtable;
  else
    interp_tgt->vtable = &field_interpolation_tgt_mapping_on_tgt_vtable;
  interp_tgt->interp_data = interp_gen_tgt->interp_data;
  interp_gen_tgt->interp_data->reference_counter++;

  interp_tgt->tags = interp_gen_tgt->put_get_tags;

  return (struct field_interpolation *)interp_tgt;
}

static int compare_recv_msgs (const void * a, const void * b) {
  return ( ((struct recv_msg_src*)b)->rank - ((struct recv_msg_src*)a)->rank );
}

static void field_interpolation_gen_tgt_enddef(
   struct field_interpolation_gen * interpolation_gen) {

   struct field_interpolation_gen_tgt * interp_tgt =
      (struct field_interpolation_gen_tgt*)interpolation_gen;

   struct communicator * comm = interp_tgt->interp_data->comm;
   struct communicator * domain_comm =
      yac_get_domain_comm(yac_get_local_domain(interp_tgt->search));

   while (yac_comm_wait(comm, interp_tgt->tags.tgt_point_location_send) ||
          yac_comm_wait(comm, interp_tgt->tags.results_recv) ||
          yac_comm_wait(comm, interp_tgt->tags.interp_data_msg_count) ||
          yac_comm_wait(comm, interp_tgt->tags.interp_data_header) ||
          yac_comm_wait(comm, interp_tgt->tags.interp_data) ||
          yac_comm_wait(domain_comm,
                        interp_tgt->tgt_tags.scatter_interp_data_msg_count));

   unsigned rank = yac_get_comm_rank(comm);
   unsigned size = yac_get_comm_size(comm);

   if (interp_tgt->interp_data->mapping_on_source) {

      struct mapping_data_src * mapping_on_src =
         &(interp_tgt->interp_data->mapping_on.src);
      unsigned num_recv_messages = mapping_on_src->num_recv_messages;
      struct recv_msg_src * recv_msgs =
         realloc(mapping_on_src->recv_msgs,
                 mapping_on_src->num_recv_messages *
                 sizeof(*(mapping_on_src->recv_msgs)));
      mapping_on_src->recv_msgs = recv_msgs;

      // reorder receive messages
      for (unsigned i = 0; i < num_recv_messages; ++i)
        if (recv_msgs[i].rank <= rank) recv_msgs[i].rank += size;

      // the receive order is invers to the send order
      qsort(recv_msgs, num_recv_messages, sizeof(*(recv_msgs)),
            compare_recv_msgs);

      for (unsigned i = 0; i < num_recv_messages; ++i)
        if (recv_msgs[i].rank >= size) recv_msgs[i].rank -= size;
   }
}

static void
field_interpolation_tgt_put(struct field_interpolation * interpolation,
                            double **** data, int collection_size) {

   yac_internal_abort_message(
      "ERROR: put operation is not defined for a target field interpolation\n",
      __FILE__, __LINE__);
}

static void
field_interpolation_tgt_get_mapping_on_src(
  struct field_interpolation * interpolation, double *** data,
  int collection_size) {

   struct field_interpolation_tgt * interp_tgt =
     (struct field_interpolation_tgt*)interpolation;

   struct mapping_data_src * mapping_on_src =
      &(interp_tgt->interp_data->mapping_on.src);
   unsigned num_recv_messages = mapping_on_src->num_recv_messages;
   struct recv_msg_src * recv_msgs = mapping_on_src->recv_msgs;
   unsigned num_requests = mapping_on_src->num_recv_messages;
   unsigned num_subdomains = interp_tgt->interp_data->num_subdomains;

   struct comm_request ** requests = malloc(num_requests * sizeof(*requests));

   double * recv_buffer =
      malloc(collection_size * mapping_on_src->total_recv_buffer_size *
             sizeof(*recv_buffer));

   double * restrict curr_recv_buffer = recv_buffer;
   for (unsigned i = 0; i < num_recv_messages; ++i) {

      //----------------
      // receive results
      //----------------

      unsigned num_points_to_recv =
         collection_size * yac_get_total_num_dependencies(recv_msgs[i].list_to);

      if (num_points_to_recv == 0) {
         requests[i] = COMM_REQUEST_NULL;
         continue;
      }

      yac_comm_irecv(curr_recv_buffer, num_points_to_recv, COMM_DBLE,
                     recv_msgs[i].rank, interp_tgt->tags.get,
                     interp_tgt->interp_data->comm, &(requests[i]));

      curr_recv_buffer += num_points_to_recv;
   }

   yac_comm_waitall(num_requests, requests);

   curr_recv_buffer = recv_buffer;
   for (unsigned i = 0; i < num_recv_messages; ++i) {

      //---------------------------
      // copy results to data array
      //---------------------------

      unsigned num_points_to_recv =
         collection_size * yac_get_total_num_dependencies(recv_msgs[i].list_to);

      if (num_points_to_recv == 0) continue;

      for (unsigned j = 0; j < num_subdomains; ++j) {

         unsigned const * restrict copy_points_to =
            yac_get_dependencies_of_element(recv_msgs[i].list_to, j);
         unsigned const * restrict copy_points_from =
            yac_get_dependencies_of_element(recv_msgs[i].list_from, j);
         unsigned num_points_to_copy =
            recv_msgs[i].list_to.num_deps_per_element[j];

         for (unsigned h = 0; h < collection_size; ++h)
            for (unsigned k = 0; k < num_points_to_copy; ++k)
               data[h][j][copy_points_to[k]] =
                  curr_recv_buffer[h*num_points_to_copy+copy_points_from[k]];
      }

      curr_recv_buffer += num_points_to_recv;
   }

   free(requests);
   free(recv_buffer);
}

static void
field_interpolation_tgt_get_mapping_on_tgt(
  struct field_interpolation * interpolation, double *** data,
  int collection_size) {

  struct field_interpolation_tgt * interp_tgt =
    (struct field_interpolation_tgt*)interpolation;

  struct mapping_data_tgt * mapping_on_tgt =
    &(interp_tgt->interp_data->mapping_on.tgt);
  struct recv_msg_tgt * recv = &(mapping_on_tgt->recv);

  // receive data
  unsigned num_requests = recv->num_msgs;
  struct comm_request ** requests = malloc(num_requests * sizeof(*requests));

  double * recv_buffer =
    malloc(collection_size * recv->total_buffer_size * sizeof(*recv_buffer));

  for (unsigned i = 0; i < recv->num_msgs; ++i) {

    //-------------
    // receive data
    //-------------

    unsigned num_points_to_recv = collection_size * recv->msgs[i].count;

    if (num_points_to_recv == 0) {
       requests[i] = COMM_REQUEST_NULL;
       continue;
    }

    yac_comm_irecv(recv->msgs[i].offset + recv_buffer, num_points_to_recv,
                   COMM_DBLE, recv->msgs[i].rank, interp_tgt->tags.get,
                   interp_tgt->interp_data->comm, &(requests[i]));
  }

  yac_comm_waitall(num_requests, requests);
  free(requests);

  // interpolate data

  if (((mapping_on_tgt->link_data.num_points > 0) &&
       (mapping_on_tgt->link_data.num_weights_per_link == 0)) ||
       (mapping_on_tgt->link_data.num_weights_per_link > 1))
    yac_internal_abort_message(
      "ERROR(field_interpolation_tgt_get_mapping_on_tgt): invalid number of "
      "weights per link", __FILE__, __LINE__);

  // apply links and assigning fixed values
  unsigned num_subdomains = interp_tgt->interp_data->num_subdomains;
  unsigned * restrict num_points_per_subdomain =
    mapping_on_tgt->link_data.num_points_per_subdomain;
  unsigned * restrict num_links_per_point =
    mapping_on_tgt->link_data.num_links_per_point;
  struct src_indices * restrict src_indices =
    mapping_on_tgt->link_data.src_indices;
  unsigned * restrict tgt_indices_link =
    mapping_on_tgt->link_data.tgt_indices;
  double * restrict weights =
    mapping_on_tgt->link_data.weights;

  unsigned * restrict tgt_indices_fixed =
    mapping_on_tgt->fixed.tgt_indices;
  double * restrict values =
    mapping_on_tgt->fixed.values;
  unsigned * restrict num_values_per_subdomain =
    mapping_on_tgt->fixed.num_values_per_subdomain;
  unsigned * restrict num_points_per_value =
    mapping_on_tgt->fixed.num_points_per_value;

  for (unsigned collection_idx = 0; collection_idx < collection_size;
       ++collection_idx) {

    for (unsigned subdomain_idx = 0, points_idx_link = 0, weight_idx = 0,
                  value_idx = 0, points_idx_fixed = 0;
         subdomain_idx < num_subdomains; ++subdomain_idx) {

      for (unsigned i = 0; i < num_points_per_subdomain[subdomain_idx];
           ++i, ++points_idx_link) {

        double accu = 0;

        for (unsigned j = 0; j < num_links_per_point[points_idx_link];
             ++j, ++weight_idx)
          accu += weights[weight_idx] *
                  recv_buffer[src_indices[weight_idx].index +
                              collection_idx *
                              src_indices[weight_idx].collection_factor];

        data[collection_idx][subdomain_idx]
            [tgt_indices_link[points_idx_link]] = accu;
      }

      for (unsigned i = 0; i < num_values_per_subdomain[subdomain_idx];
           ++i, ++value_idx) {

        double value = values[value_idx];
        unsigned num_points = num_points_per_value[value_idx];
        for (unsigned j = 0; j < num_points; ++j, ++points_idx_fixed)
          data[collection_idx][subdomain_idx]
           [tgt_indices_fixed[points_idx_fixed]] = value;
      }
    }
  }
  free(recv_buffer);
}

static void
field_interpolation_gen_tgt_add_weight_file_name
  (field_interpolation_gen interpolation_gen, char const * weight_file_name) {

  yac_internal_abort_message(
    "ERROR: add_weight_file_name is not defined for a target field "
    "interpolation\n", __FILE__, __LINE__);
}

static void free_mapping_data(union mapping_data * mapping_on,
                              unsigned mapping_on_source) {

  if (mapping_on_source) {

    for (unsigned i = 0; i < mapping_on->src.num_recv_messages; ++i) {
      yac_free_dep_list(&(mapping_on->src.recv_msgs[i].list_from));
      yac_free_dep_list(&(mapping_on->src.recv_msgs[i].list_to));
    }
    free(mapping_on->src.recv_msgs);

  } else {

    free(mapping_on->tgt.link_data.num_points_per_subdomain);
    free(mapping_on->tgt.link_data.num_links_per_point);
    free(mapping_on->tgt.link_data.src_indices);
    free(mapping_on->tgt.link_data.tgt_indices);
    free(mapping_on->tgt.link_data.weights);
    free(mapping_on->tgt.fixed.tgt_indices);
    free(mapping_on->tgt.fixed.values);
    free(mapping_on->tgt.fixed.num_values_per_subdomain);
    free(mapping_on->tgt.fixed.num_points_per_value);
    free(mapping_on->tgt.recv.msgs);
  }
}

static void
field_interpolation_data_tgt_delete(
  struct field_interpolation_data_tgt * interp_data) {

  interp_data->reference_counter--;

  if (!interp_data->reference_counter) {

    free_mapping_data(&(interp_data->mapping_on),
                      interp_data->mapping_on_source);
    free(interp_data);
  }
}

static void
field_interpolation_gen_tgt_delete(struct field_interpolation_gen * interpolation_gen) {

   struct field_interpolation_gen_tgt * interp_tgt;

   interp_tgt = (struct field_interpolation_gen_tgt*)interpolation_gen;

   field_interpolation_data_tgt_delete(interp_tgt->interp_data);

   free(interpolation_gen);
}

static void
field_interpolation_tgt_delete(struct field_interpolation * interpolation) {

   struct field_interpolation_tgt * interp_tgt;

   interp_tgt = (struct field_interpolation_tgt*)interpolation;

   field_interpolation_data_tgt_delete(interp_tgt->interp_data);

   free(interpolation);
}
