/**
 * @file fields.c
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#include <stdlib.h>
#include <string.h>
#include "event.h"
#include "utils.h"
#include "component.h"
#include "fields.h"
#include "field_interpolation_gen.h"

struct coupling_field {

  // general field related information

  char * name;

  int local_comp_id;
  int global_comp_id;

  struct domain * domain;

  // multiple points per subdomain are required when the field data
  // is described by a set of functions that for example use
  // sampling points at the corners and edges (currently only one
  // "struct points" per subdomain is normally required)
  struct points *** points; //!< dimensions [subdomain idx][point set idx]
  struct mask *** masks;    //!< dimensions [subdomain idx][point set idx]
  unsigned num_pointsets;   //!< number of point indices
  unsigned collection_size; //!< number of vertical levels or bundles

  // data exchange

  int exchange_type;

  union {
    struct {
      struct {
        int event_id;
        struct field_interpolation * interpolation;
        int time_accumulation_count;
        double **** send_field_acc;
      } * puts;
      unsigned num_puts;
    } put;
    struct {
      int event_id;
      struct field_interpolation * interpolation;
    } get;
  } exchange_data;
};

struct coupling_field *
yac_coupling_field_new(char * field_name, const int local_comp_id,
                       struct domain * domain, struct points *** points,
                       struct mask *** masks, unsigned num_pointsets ) {

   struct coupling_field * cpl_field = malloc(1 * sizeof(*cpl_field));
   struct component * component = yac_unique_id_to_pointer(local_comp_id);

   if ( component != NULL ) 
       cpl_field->global_comp_id  = yac_component_get_global_id(component);
   else
       cpl_field->global_comp_id = -1;

   cpl_field->name            = malloc((strlen(field_name) + 1));
   strcpy(cpl_field->name, field_name);
   cpl_field->local_comp_id   = local_comp_id;
   cpl_field->domain          = domain;
   if (points != NULL) {
      unsigned nbr_subdomains = yac_get_num_subdomains(domain);
      cpl_field->points =
         malloc(nbr_subdomains * sizeof(*cpl_field->points));
      for (unsigned i = 0; i < nbr_subdomains; ++i) {
         cpl_field->points[i] =
            malloc(num_pointsets * sizeof(**cpl_field->points));
         for (unsigned j = 0; j < num_pointsets; ++j)
            cpl_field->points[i][j] = points[i][j];
      }
   } else {
      cpl_field->points = NULL;
   }
   if (masks != NULL) {
      unsigned nbr_subdomains = yac_get_num_subdomains(domain);
      cpl_field->masks =
         malloc(nbr_subdomains * sizeof(*cpl_field->masks));
      for (unsigned i = 0; i < nbr_subdomains; ++i) {
         cpl_field->masks[i] =
            malloc(num_pointsets * sizeof(**cpl_field->masks));
         for (unsigned j = 0; j < num_pointsets; ++j)
            cpl_field->masks[i][j] = masks[i][j];
      }
   } else {
      cpl_field->masks = NULL;
   }
   cpl_field->num_pointsets   = num_pointsets;
   cpl_field->collection_size = 0;
   cpl_field->exchange_type   = NOTHING;

   if ((num_pointsets != 0) && (points == NULL))
      yac_internal_abort_message("ERROR: num_pointsets > 0 but not point sets provided",
                                 __FILE__, __LINE__);

   //-------------------------------
   // test consistency of input data
   //-------------------------------

   if (points != NULL) {

      unsigned num_subdomains, i, j;

      num_subdomains = yac_get_num_subdomains(domain);

      for (i = 1; i < num_subdomains; ++i)
         for (j = 0; j < num_pointsets; ++j)
            if (points[0][j]->location != points[i][j]->location)
               yac_internal_abort_message ( "ERROR: inconsistent points location in init_coupling_field",
                                            __FILE__, __LINE__ );
   }

   return cpl_field;
}

unsigned yac_get_coupling_field_num_pointsets(struct coupling_field * field) {

   return field->num_pointsets;
}

unsigned yac_get_coupling_field_collection_size(struct coupling_field * field) {

   return field->collection_size;
}

void yac_set_coupling_field_collection_size(struct coupling_field * field, unsigned collection_size ) {

   field->collection_size = collection_size;
}

const char * yac_get_coupling_field_name(struct coupling_field * field) {

  return field->name;
}

struct domain * yac_get_coupling_field_domain(struct coupling_field * field) {

  return field->domain;
}

int yac_get_coupling_field_global_comp_id(struct coupling_field * field) {

  return field->global_comp_id;
}

struct component * yac_get_coupling_field_component(struct coupling_field * field) {

  return (struct component *)yac_unique_id_to_pointer(field->local_comp_id);
}

enum yac_location
yac_get_coupling_field_pointset_location(struct coupling_field * field,
                                         unsigned points_idx) {

  if (field->points == NULL)
    yac_internal_abort_message("ERROR: not point set defined", __FILE__, __LINE__);

  if (points_idx >= field->num_pointsets)
    yac_internal_abort_message("ERROR: invalid point set index", __FILE__, __LINE__);

  return field->points[0][points_idx]->location;
}

struct points *** yac_get_coupling_field_pointsets(struct coupling_field * field) {

  return field->points;
}

struct mask *** yac_get_coupling_field_masks(struct coupling_field * field) {

  return field->masks;
}

enum yac_field_exchange_type
yac_get_coupling_field_exchange_type(struct coupling_field * field) {

  return field->exchange_type;
}

double ****
yac_get_coupling_field_put_op_send_field_acc(struct coupling_field * field,
                                             unsigned put_idx) {

  if (field->exchange_type != SOURCE)
    yac_internal_abort_message("ERROR: wrong field exchange type", __FILE__, __LINE__);

  if (put_idx >= field->exchange_data.put.num_puts)
    yac_internal_abort_message("ERROR: put_idx is invalid", __FILE__, __LINE__);

  if (field->exchange_data.put.puts[put_idx].send_field_acc != NULL)
    return field->exchange_data.put.puts[put_idx].send_field_acc;

  int num_subdomains = yac_get_num_subdomains(field->domain);
  int num_pointsets = field->num_pointsets;
  int collection_size = field->collection_size;
  double **** send_field_acc;

  /* Allocate memory for the accumulation */

  send_field_acc = malloc(collection_size * sizeof(*send_field_acc));

  for (int h = 0; h < collection_size; ++h) {

    send_field_acc[h] = malloc(num_subdomains * sizeof(**send_field_acc));

    for (int i = 0; i < num_subdomains; i++) {
      send_field_acc[h][i] = malloc(num_pointsets * sizeof(***send_field_acc));

      for (int j = 0; j < num_pointsets; j++ ) {
        send_field_acc[h][i][j] = malloc(yac_get_data_size(*(field->points[i][j])) *
                                         sizeof(****send_field_acc));
      }
    }
  }

  field->exchange_data.put.puts[put_idx].send_field_acc = send_field_acc;

  /* Initialise memory for the accumulation */

  yac_init_coupling_field_put_op_send_field_acc(field, put_idx, 0.0);
  

  return send_field_acc;
}

double ****
yac_init_coupling_field_put_op_send_field_acc(struct coupling_field * field,
                                              unsigned put_idx,
                                              double init_value) {

  if (field->exchange_type != SOURCE)
    yac_internal_abort_message("ERROR: wrong field exchange type", __FILE__, __LINE__);

  if (put_idx >= field->exchange_data.put.num_puts)
    yac_internal_abort_message("ERROR: put_idx is invalid", __FILE__, __LINE__);

  if (field->exchange_data.put.puts[put_idx].send_field_acc == NULL) {

    return yac_get_coupling_field_put_op_send_field_acc(field, put_idx);
  }

  int num_subdomains = yac_get_num_subdomains(field->domain);
  int num_pointsets = field->num_pointsets;
  int collection_size = field->collection_size;
  double **** send_field_acc =
    field->exchange_data.put.puts[put_idx].send_field_acc;

  for (int h = 0; h < collection_size; ++h) {
    for (int i = 0; i < num_subdomains; i++) {
      for (int j = 0; j < num_pointsets; j++) {
        int num_points = yac_get_data_size(*(field->points[i][j]));
        for (int m = 0; m < num_points; m++ ) {
          send_field_acc[h][i][j][m] = init_value;
        }
      }
    }
  }

  return send_field_acc;
}

int yac_get_coupling_field_put_op_event_id(struct coupling_field * field,
                                           unsigned put_idx) {

  if (field->exchange_type != SOURCE)
    yac_internal_abort_message("ERROR: wrong field exchange type", __FILE__, __LINE__);

  if (put_idx >= field->exchange_data.put.num_puts)
    yac_internal_abort_message("ERROR: put_idx is invalid", __FILE__, __LINE__);

  return field->exchange_data.put.puts[put_idx].event_id;
}

struct field_interpolation *
yac_get_coupling_field_put_op_interpolation(struct coupling_field * field,
                                            unsigned put_idx) {

  if (field->exchange_type != SOURCE)
    yac_internal_abort_message("ERROR: wrong field exchange type", __FILE__, __LINE__);

  if (put_idx >= field->exchange_data.put.num_puts)
    yac_internal_abort_message("ERROR: put_idx is invalid", __FILE__, __LINE__);

  return field->exchange_data.put.puts[put_idx].interpolation;
}

int
yac_get_coupling_field_put_op_time_accumulation_count(struct coupling_field * field,
                                                      unsigned put_idx) {

  if (field->exchange_type != SOURCE)
    yac_internal_abort_message("ERROR: wrong field exchange type", __FILE__, __LINE__);

  if (put_idx >= field->exchange_data.put.num_puts)
    yac_internal_abort_message("ERROR: put_idx is invalid", __FILE__, __LINE__);

  return field->exchange_data.put.puts[put_idx].time_accumulation_count;
}

void
yac_set_coupling_field_put_op_time_accumulation_count(struct coupling_field * field,
                                                      unsigned put_idx, int count) {

  if (field->exchange_type != SOURCE)
    yac_internal_abort_message("ERROR: wrong field exchange type", __FILE__, __LINE__);

  if (put_idx >= field->exchange_data.put.num_puts)
    yac_internal_abort_message("ERROR: put_idx is invalid", __FILE__, __LINE__);

  field->exchange_data.put.puts[put_idx].time_accumulation_count = count;
}

unsigned yac_get_coupling_field_num_puts(struct coupling_field * field) {

  if (field->exchange_type != SOURCE)
    yac_internal_abort_message("ERROR: wrong field exchange type", __FILE__, __LINE__);

  return field->exchange_data.put.num_puts;
}

int yac_get_coupling_field_get_op_event_id(struct coupling_field * field) {

  if (field->exchange_type != TARGET)
    yac_internal_abort_message("ERROR: wrong field exchange type", __FILE__, __LINE__);

  return field->exchange_data.get.event_id;
}

struct field_interpolation *
yac_get_coupling_field_get_op_interpolation(struct coupling_field * field) {

  if (field->exchange_type != TARGET)
    yac_internal_abort_message("ERROR: wrong field exchange type", __FILE__, __LINE__);

  return field->exchange_data.get.interpolation;
}

void yac_set_coupling_field_put_op(struct coupling_field * field, int event_id,
                                   struct field_interpolation * interpolation) {

   if (field->exchange_type == TARGET)
      yac_internal_abort_message("ERROR: a get operation has already been set\n",
                                 __FILE__, __LINE__);

   if (field->exchange_type == NOTHING) {
      field->exchange_data.put.puts = NULL;
      field->exchange_data.put.num_puts = 0;
      field->exchange_type = SOURCE;
   }

   unsigned num_puts = field->exchange_data.put.num_puts;
   field->exchange_data.put.num_puts++;

   field->exchange_data.put.puts =
      realloc(field->exchange_data.put.puts, (num_puts + 1) *
              sizeof(*(field->exchange_data.put.puts)));

   field->exchange_data.put.puts[num_puts].event_id = event_id;
   field->exchange_data.put.puts[num_puts].interpolation = interpolation;
   field->exchange_data.put.puts[num_puts].time_accumulation_count = 0;
   field->exchange_data.put.puts[num_puts].send_field_acc = NULL;
}

void yac_set_coupling_field_get_op(struct coupling_field * field, int event_id,
                                   struct field_interpolation * interpolation) {

   if (field->exchange_type != NOTHING)
      yac_internal_abort_message("ERROR: a put or get operation has already "
                                 "been set\n", __FILE__, __LINE__);

   field->exchange_type = TARGET;
   field->exchange_data.get.event_id = event_id;
   field->exchange_data.get.interpolation = interpolation;
}

void yac_coupling_field_delete(struct coupling_field * cpl_field) {

  unsigned num_subdomains = yac_get_num_subdomains(cpl_field->domain);
  unsigned num_pointsets = cpl_field->num_pointsets;
  int collection_size = cpl_field->collection_size;

  if (cpl_field->exchange_type == SOURCE) {

    for (unsigned put_idx = 0; put_idx < cpl_field->exchange_data.put.num_puts;
         ++put_idx) {

      int event_id = cpl_field->exchange_data.put.puts[put_idx].event_id;
      struct field_interpolation * interpolation =
        cpl_field->exchange_data.put.puts[put_idx].interpolation;
      double ****send_field_acc =
        cpl_field->exchange_data.put.puts[put_idx].send_field_acc;

      if (send_field_acc != NULL) {
        for (int h = 0; h < collection_size; ++h) {
          for (unsigned i = 0; i < num_subdomains; ++i) {
            for (unsigned j = 0; j < num_pointsets; ++j) {
              free(send_field_acc[h][i][j]);
            }
            free(send_field_acc[h][i]);
          }
          free(send_field_acc[h]);
        }
        free(send_field_acc);
      }

      yac_event_free(event_id);
      yac_delete_field_interpolation(interpolation);
    }

    free(cpl_field->exchange_data.put.puts);

  } else if (cpl_field->exchange_type == TARGET) {

    yac_event_free(cpl_field->exchange_data.get.event_id);
    yac_delete_field_interpolation(cpl_field->exchange_data.get.interpolation);
  }

  if (cpl_field->points != NULL)
    for (unsigned i = 0; i < num_subdomains; ++i) free(cpl_field->points[i]);
  if (cpl_field->masks != NULL)
    for (unsigned i = 0; i < num_subdomains; ++i) free(cpl_field->masks[i]);
  free(cpl_field->points);
  free(cpl_field->masks);
  free(cpl_field->name);
  free(cpl_field);
}
