/**
 * @file fields.c
 * @brief Structs and interfaces to defined coupling fields
 *
 * For the coupling fields several types of action have to be stored
 * like the type of interpolation chosen by the user, the time operation
 * and results from the search.
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#ifndef FIELDS_H
#define FIELDS_H


#include "mask.h"
#include "points.h"
#include "domain.h"

// general coupling field struct

enum yac_field_exchange_type {
  NOTHING = 0,
  SOURCE  = 1,
  TARGET  = 2,
};

// forward declaration

struct coupling_field;
struct field_interpolation;

// generation of a field

struct coupling_field *
yac_coupling_field_new(char * field_name, const int local_comp_id,
                       struct domain * domain, struct points *** points,
                       struct mask *** masks, unsigned num_pointsets);

/**
 * gets the number of points sets per subdomain (it is assumes that all subdomains
 * have the same number of points sets, in the same order(in respect to their
 * location type))
 * @param[in] field
 * @return number of points sets per subdomain
 */
unsigned yac_get_coupling_field_num_pointsets(struct coupling_field * field);

/**
 * Sets the number of vertical levels or bundles for a field. This info
 * is comming from the coupling.xml file and is only available if the
 * field is selected for coupling in the xml config fiĺe.
 *
 * @param[in] field
 * @param[in] collection_size of the field as specified in coupling.xml
 */
void yac_set_coupling_field_collection_size(struct coupling_field * field, unsigned collection_size);

/**
 * gets the number of vertical levels or bundles for a field. This info
 * is comming from the coupling.xml file and is only available if the
 * field is selected for coupling in the xml config fiĺe.
 *
 * @param[in] field
 * @return collection size of the field as specified in coupling.xml
 */
unsigned yac_get_coupling_field_collection_size(struct coupling_field * field);

/**
 * gets the name of the coupling field
 * @param[in] field
 * @return name of the coupling field
 */
const char * yac_get_coupling_field_name(struct coupling_field * field);

/**
 * gets the component of the coupling field
 * @param[in] field
 * @return component
 */
struct component * yac_get_coupling_field_component(struct coupling_field * field);

/**
 * gets the global component id of the coupling field
 * @param[in] field
 * @return global component id
 */
int yac_get_coupling_field_global_comp_id(struct coupling_field * field);

/**
 * gets the domain of the coupling field
 * @param[in] field
 * @return domain of the coupling field
 */
struct domain * yac_get_coupling_field_domain(struct coupling_field * field);

/**
 * get point set data location
 * @param[in] field
 * @param[in] points_idx index of point set
 * @return data location of requested point set
 */
enum yac_location
yac_get_coupling_field_pointset_location(struct coupling_field * field,
                                         unsigned points_idx);

/**
 * gets the point sets of the coupling field
 * @param[in] field
 * @return point sets of the field
 * @remarks dimensions: [subdomain idx][point set idx]
 */
struct points *** yac_get_coupling_field_pointsets(struct coupling_field * field);

/**
 * gets the masks of the coupling field
 * @param[in] field
 * @return masks of the field
 * @remarks dimensions: [subdomain idx][point set idx]
 */
struct mask *** yac_get_coupling_field_masks(struct coupling_field * field);

/**
 * gets the exchange type of the coupling field
 * @param[in] field
 * @return exchange type of the coupling field
 */
enum yac_field_exchange_type
yac_get_coupling_field_exchange_type(struct coupling_field * field);

/**
 * gets the number of put operations of the coupling field
 * @param[in] field
 * @return number of put operations of the coupling field
 * @remarks calling this routine for a coupling field that
 *          has a different exchange type than SOURCE yields
 *          undefined results
 */
unsigned yac_get_coupling_field_num_puts(struct coupling_field * field);

/**
 * gets the event id of a specified put operation
 * @param[in] field
 * @param[in] put_idx index of the put operation
 * @return event id of specified put operation
 * @remarks calling this routine for a coupling field that
 *          has a different exchange type than SOURCE yields
 *          undefined results
 * @see get_coupling_field_num_puts
 */
int yac_get_coupling_field_put_op_event_id(struct coupling_field * field,
                                           unsigned put_idx);

/**
 * gets the field interpolation of a specified put operation
 * @param[in] field
 * @param[in] put_idx index of the put operation
 * @return field interpolation of specified put operation
 * @remarks calling this routine for a coupling field that
 *          has a different exchange type than SOURCE yields
 *          undefined results
 * @see get_coupling_field_num_puts
 */
struct field_interpolation *
yac_get_coupling_field_put_op_interpolation(struct coupling_field * field,
                                            unsigned put_idx);

/**
 * gets the send field accumulator of a specified put operation
 * @param[in] field
 * @param[in] put_idx index of the put operation
 * @return send field accumulator accumulator of specified put operation
 * @remarks calling this routine for a coupling field that
 *          has a different exchange type than SOURCE yields
 *          undefined results
 * @remarks when this routine is called the first time for a put operation
 *          it will be allocated and initialised with zeros
 * @see get_coupling_field_num_puts
 */
double ****
yac_get_coupling_field_put_op_send_field_acc(struct coupling_field * field,
                                             unsigned put_idx);

/**
 * initialises the send field accumulator of a specified put operation
 * @param[in] field
 * @param[in] put_idx    index of the put operation
 * @param[in] init_value value that is to be used for the initialisation of the
 *                       send field accumulator
 * @return initialised send field accumulator of specified put operation
 * @remarks calling this routine for a coupling field that
 *          has a different exchange type than SOURCE yields
 *          undefined results
 * @see get_coupling_field_num_puts
 */
double ****
yac_init_coupling_field_put_op_send_field_acc(struct coupling_field * field,
                                              unsigned put_idx,
                                              double init_value);

/**
 * gets the time accumulation count of a specified put operation
 * @param[in] field
 * @param[in] put_idx    index of the put operation
 * @return time accumulation count of specified put operation
 * @remarks calling this routine for a coupling field that
 *          has a different exchange type than SOURCE yields
 *          undefined results
 * @see get_coupling_field_num_puts
 */
int
yac_get_coupling_field_put_op_time_accumulation_count(struct coupling_field * field,
                                                      unsigned put_idx);

/**
 * sets the time accumulation count of a specified put operation
 * @param[in] field
 * @param[in] put_idx index of the put operation
 * @param[in] count   new time accumulation count
 * @remarks calling this routine for a coupling field that
 *          has a different exchange type than SOURCE yields
 *          undefined results
 * @see get_coupling_field_num_puts
 */
void
yac_set_coupling_field_put_op_time_accumulation_count(struct coupling_field * field,
                                                      unsigned put_idx,
                                                      int count);

/**
 * gets the event id of the get operation
 * @param[in] field
 * @return event id of the get operation
 * @remarks calling this routine for a coupling field that
 *          has a different exchange type than TARGET yields
 *          undefined results
 */
int yac_get_coupling_field_get_op_event_id(struct coupling_field * field);

/**
 * gets the field interpolation of the get operation
 * @param[in] field
 * @return field interpolation of the get operation
 * @remarks calling this routine for a coupling field that
 *          has a different exchange type than TARGET yields
 *          undefined results
 */
struct field_interpolation *
yac_get_coupling_field_get_op_interpolation(struct coupling_field * field);

/**
 * sets the put operation for the coupling field
 * @param[in,out] field
 * @param[in]     event_id      event id of the operation
 * @param[in]     interpolation interpolation to be executed
 * @remarks the coupling field takes control of the input data, it will be
 *          free/deleted by the coupling field
 */
void yac_set_coupling_field_put_op(struct coupling_field * field, int event_id,
                                   struct field_interpolation * interpolation);

/**
 * set the get operation for the coupling field
 * @param[in,out] field
 * @param[in]     event_id      event id of the operation
 * @param[in]     interpolation interpolation to be executed
 * @remarks the coupling field takes control of the input data, it will be
 *          free/deleted by the coupling field
 */
void yac_set_coupling_field_get_op(struct coupling_field * field, int event_id,
                                   struct field_interpolation * interpolation);

// destruction of fields

void yac_coupling_field_delete(struct coupling_field * cpl_field);

#endif // FIELDS_H
