/**
 * @file global_search.c
 *
 * \section global_search_communication Communication schema in global search
 *  \msc
 *    width = "800";
 *    Targets,"Target root","Source root",Sources;
 *    |||;
 *    Targets box Sources [label="exchange of global domain ranks",
 *                         URL="\ref global_search_communication_rank_exchange",
 *                         ID="1"];
 *    |||;
 *    "Source root" rbox Sources [label="gather extents of local domain", ID="2.a"];
 *    "Target root"<-"Source root" [label="extents of local source domains", ID="2.a"];
 *    Targets rbox "Target root" [label="broadcast source extents", ID="2.b"];
 *    Targets rbox "Target root" [label="match source extents with local data", ID="3"];
 *    Targets rbox "Target root" [label="gather basic matching information", ID="3"];
 *    "Target root"->"Source root" [label="basic matching information", ID="3"];
 *    "Source root" rbox Sources [label="scatter basic matching information", ID="4"];
 *    Targets->Sources [label="target cells matching source extents", ID="3"];
 *    "Source root" rbox Sources [label="match target cells with local source cells", ID="4"];
 *    Targets<-Sources [label="global cell ids of matching target cells"];
 *    Targets->Sources [label="ranks of source processes that have overlap with same target cell", ID="6"];
 *
 * \endmsc
 *
 * -# The routine \ref yac_global_search_new initiates the exchange of global ranks
 *    between the source and target domains (see
 *    \ref global_search_communication_rank_exchange).
 * -# Once the global ranks are exchanged:
 *    -# the sources processes gather their extents on their root process,
 *       which sends them to the target root
 *    -# the target root receives the extents of all source processes from the source
 *       root and broadcasts them to all target processes
 * -# In \ref global_search_bcast_extents_cont_callback the targets receive the extents of
 *    the source processes and match them with the local domain data. Each target
 *    then send the overlapping cells to the respective source process. Afterwards,
 *    the target root process gathers the basic matching information it to the
 *    source root.
 * -# The source root receives the information on which target and source process
 *    overlap and scatters this among all source processes. Then each source
 *    process receives the overlapping target cells from the respective target
 *    process and returns the indices of all target cells that have an overlap
 *    with a local source cell.
 * -# If the target detects that source cells from different source processes overlap
 *    with the same target cell, the target process informs the respective source
 *    processes about this (\ref return_multi_match_cells).
 *
 * \section global_search_communication_rank_exchange Exchange of global source and target ranks
 *  \msc
 *    width = "800";
 *    Targets,"Target root","Global root","Source root",Sources;
 *    |||;
 *    Targets rbox "Target root" [label="compute global target ranks", ID="1"],
 *    "Source root" rbox Sources [label="compute global source ranks", ID="1"];
 *    "Target root"->"Global root" [label="global target ranks", ID="2"],
 *    "Global root"<-"Source root" [label="global source ranks", ID="2"];
 *    --- [label="if only the source root is also the global root", ID="3"];
 *    "Global root" note "Source root" [label="are identical"];
 *    "Source root" rbox "Source root" [label="receive message originally sent to Global root", ID="3.a"];
 *    "Target root"<-"Source root" [label="global source ranks", ID="3.b"];
 *    --- [label="if only the target root is also the global root", ID="3"];
 *    "Target root" note "Global root" [label="are identical"];
 *    "Target root" rbox "Target root" [label="receive message originally sent to Global root", ID="3.a"];
 *    "Target root"->"Source root" [label="global target ranks", ID="3.b"];
 *    --- [label="after Target/Source root has received the Source/Target ranks", ID="4"];
 *    Targets<-"Target root" [label="broadcast size of source domain", ID="4.a"],
 *    "Source root"->Sources [label="broadcast size of target domain", ID="4.a"];
 *    Targets<-"Target root" [label="broadcast source domain ranks", ID="4.b"],
 *    "Source root"->Sources [label="broadcast target domain ranks", ID="4.b"];
 *
 * \endmsc
 *
 * -# In \ref yac_global_search_new the processes of each domain (source and target)
 *    computes the matching between their ranks in their domain communicator and the
 *    ranks in the global communicator, which contains of the union of both domains.
 * -# The roots of both domains send the global ranks of their domain to the global
 *    root, which is assumed to be part of the respective other domain.
 *    (\ref yac_global_search_new)
 * -# Typically the global root is only part of one domain.
 *    -# By analysing the received global ranks, it can be determined, whether the
 *       root of the other domain is also the global root. In case it is not, the
 *       message sent in 2 is received in order to avoid unreceived messages.
 *       (\ref global_search_receive_remote_domain_ranks_cont)
 *    -# It then sends the global ranks of its domain to the root of the other
 *       domain, whose rank is now known.
 *       (\ref global_search_receive_remote_domain_ranks_cont)
 * -# The size of the remote domain is broadcasted to the broadcasted among the
 *    processes of the local domain by the respective root process.
 *    (send: \ref global_search_receive_remote_domain_ranks_cont,
 *     recv: \ref global_search_bcast_domain_ranks)
 * -# Afterwards, the actual ranks of the remote domain is broadcasted.
 *    (send: \ref global_search_receive_remote_domain_ranks_cont,
 *     receive: \ref global_search_bcast_domain_ranks_cont)
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#include <stdlib.h>
#include <string.h>

#include "grid_search.h"
#include "global_search.h"
#include "domain.h"
#include "communicator.h"
#include "geometry.h"
#include "utils.h"
#include "ensure_array_size.h"

struct global_search_tags {

   unsigned global_search_send_extents_header;
   unsigned global_search_send_extents;
   unsigned global_search_recv_extents_header;
   unsigned global_search_recv_extents;
   unsigned global_search_send_grid_data_header;
   unsigned global_search_send_grid_data;
   unsigned global_search_recv_grid_data_header;
   unsigned global_search_recv_grid_data;
   unsigned global_search_send_domain_ranks_header;
   unsigned global_search_send_domain_ranks;
   unsigned global_search_recv_domain_ranks_header;
   unsigned global_search_recv_domain_ranks;
   unsigned global_search_send_results_header;
   unsigned global_search_send_results;
   unsigned global_search_recv_results_header;
   unsigned global_search_recv_results;
   unsigned global_search_send_multi_match_cells_header;
   unsigned global_search_send_multi_match_cells;
   unsigned global_search_recv_multi_match_cells_header;
   unsigned global_search_recv_multi_match_cells;
   unsigned global_search_send_overlapping_ranks;
   unsigned global_search_recv_overlapping_ranks;
   unsigned global_search_send_num_overlapping_ranks;
   unsigned global_search_recv_num_overlapping_ranks;
};

struct global_search_domain_tags {

   unsigned global_search_gather_extents;
   unsigned global_search_bcast_domain_ranks_header;
   unsigned global_search_bcast_domain_ranks;
   unsigned global_search_bcast_extents_header;
   unsigned global_search_bcast_extents;
   unsigned global_search_gatherv_overlapping_ranks;
   unsigned global_search_scatterv_overlapping_ranks;
   unsigned global_search_allgather_num_overlapping_ranks;
   unsigned global_search_bcast_num_overlapping_ranks;
   unsigned global_search_allgather_num_subdomains;
};

struct tgt_search_result {
  int rank;
  struct dep_list matching_local_cell_ids;
                                  //!< contains for all source processes in comm a
                                  //!< dependency list that contains for each local
                                  //!< subdomain the local ids of cells that have an
                                  //!< overlap with the respective source process
  int * global_tgt_ids;
  unsigned num_global_tgt_ids;
};

struct global_search {

   struct domain * local_domain;
   struct {
      struct domain * remote_domain;
      unsigned rank;
   } * remote_domains;
   unsigned num_remote_domains;
   char * remote_domain_name;

   char * local_component_name;
   char * remote_component_name;

   struct global_search_tags tags;
   struct global_search_domain_tags domain_tags;

   unsigned local_domain_is_source;
   unsigned local_domain_is_target;

   struct communicator * comm;

   unsigned * local_domain_ranks;  //!< ranks of all process of the local domain
                                   //!< within the global search communicator
   unsigned * remote_domain_ranks; //!< ranks of all process of the remote domain
                                   //!< within the global search communicator
   unsigned local_domain_size;  //!< number of processes in the local domain
   unsigned remote_domain_size; //!< number of processes in the remote domain

   // data of target domains
   struct tgt_search_result * tgt_search_results;
   unsigned num_overlapping_src_processes;

   // data of source domains
   struct grid_search ** grid_search;  //!< grid search object for each local %subdomain

  int * overlapping_tgt_ranks; //!< contains ranks of all target processes that have an
                               //!< overlap with this source domain
  unsigned num_overlapping_tgt_processes;
  struct source_search_results * src_search_results; //!< contains search results for all target cells
                                                     //!< is sorted by the the global tgt id
  unsigned num_src_search_results; //!< number of entries in search_results
};

struct intermediate_source_search_results {

  int global_id;
  int tgt_rank;
  struct domain * tgt_domain;
  unsigned local_tgt_cell_id;
  unsigned tgt_subdomain_idx;
  unsigned subdomain_idx;
  unsigned * overlapping_src_cells;
  unsigned num_overlapping_src_cells;
};

struct multimatch_data {

   int global_id;
   unsigned num_other_src_ranks;
   unsigned * other_src_ranks;
};

struct intermediate_global_search_data {

   unsigned num_open_target_requests;
   unsigned num_open_source_requests;
   struct global_search * search;

   struct multimatch_data * temp_multimatch_data;
   unsigned temp_multimatch_data_array_size;
   unsigned num_temp_multimatch_data;

   struct intermediate_source_search_results * temp_src_search_results;
   unsigned temp_src_search_results_array_size;
   unsigned num_temp_src_search_results;

   unsigned * extent_buffer_sizes;

   unsigned local_domain_root_idx;
   unsigned remote_domain_root;
};

struct temp_receive_grid_data {

   struct intermediate_global_search_data * search_data;
   unsigned buffer_dble_size;
   unsigned buffer_int_size;
};

struct temp_overlappping_ranks_data {

   struct intermediate_global_search_data * search_data;
   unsigned * tgt_rank_indices;
};

struct temp_recv_overlapping_ranks_data {

   struct intermediate_global_search_data * search_data;
   unsigned * num_src_per_tgt;
  
};

static void generate_tag_string(char * buf, char const * comp_a,
                                char const * domain_a, char const * comp_b,
                                char const * domain_b, char const * type,
                                char const * tag_str) {

   strcpy(buf, comp_a);
   strcat(buf, domain_a);
   strcat(buf, comp_b);
   strcat(buf, domain_b);
   strcat(buf, type);
   strcat(buf, tag_str);
}

static
void generate_global_search_tags(struct global_search_tags * tags,
                                 char const * comp_a_name,
                                 char const * domain_a_name,
                                 char const * comp_b_name,
                                 char const * domain_b_name,
                                 struct communicator * comm) {

   char const * str_generator_string                       = "generator string";
   char const * str_global_search_extents_header           = "extents header";
   char const * str_global_search_extents                  = "extents";
   char const * str_global_search_grid_data_header         = "grid data header";
   char const * str_global_search_grid_data                = "grid data";
   char const * str_global_search_results_header           = "results header";
   char const * str_global_search_results                  = "results";
   char const * str_global_search_multi_match_cells_header = "multi match cells header";
   char const * str_global_search_multi_match_cells        = "multi match cells";
   char const * str_global_search_overlapping_ranks        = "overlapping ranks";
   char const * str_global_search_num_overlapping_ranks    = "num overlapping ranks";
   char const * str_global_search_ranks_domain_header      = "domain ranks header";
   char const * str_global_search_ranks_domain             = "domain ranks";

   unsigned max_tag_strlen;

   max_tag_strlen = strlen(str_global_search_multi_match_cells_header);

   char const * str_global_search = "global search";

   unsigned max_str_buf_size;

   max_str_buf_size =
    max_tag_strlen + 2 * MAX(strlen(domain_a_name), strlen(domain_b_name)) +
    2 * MAX(strlen(comp_a_name), strlen(comp_b_name)) +
    strlen(str_global_search) + 1;

   char * unique_generator_string = malloc(max_str_buf_size);

   int cmp_comp_names = strcmp(comp_a_name, comp_b_name);
   int cmp_domain_names = strcmp(domain_a_name, domain_b_name);

   if ((cmp_comp_names && (cmp_comp_names > 0)) ||
       (!cmp_comp_names && (cmp_domain_names > 0)))
      generate_tag_string(unique_generator_string,
                          comp_a_name, domain_a_name,
                          comp_b_name, domain_b_name,
                          str_global_search, str_generator_string);
   else
      generate_tag_string(unique_generator_string,
                          comp_b_name, domain_b_name,
                          comp_a_name, domain_a_name,
                          str_global_search, str_generator_string);

   char * tag_strings[24];
   tag_strings[0] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[0], comp_a_name, domain_a_name, comp_b_name,
                       domain_b_name, str_global_search,
                       str_global_search_grid_data_header);

   tag_strings[1] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[1], comp_a_name, domain_a_name, comp_b_name,
                       domain_b_name, str_global_search,
                       str_global_search_grid_data);

   tag_strings[2] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[2], comp_b_name, domain_b_name, comp_a_name,
                       domain_a_name, str_global_search,
                       str_global_search_grid_data_header);

   tag_strings[3] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[3], comp_b_name, domain_b_name, comp_a_name,
                       domain_a_name, str_global_search,
                       str_global_search_grid_data);

   tag_strings[4] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[4], comp_a_name, domain_a_name, comp_b_name,
                       domain_b_name, str_global_search,
                       str_global_search_extents_header);

   tag_strings[5] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[5], comp_a_name, domain_a_name, comp_b_name,
                       domain_b_name, str_global_search,
                       str_global_search_extents);

   tag_strings[6] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[6], comp_b_name, domain_b_name, comp_a_name,
                       domain_a_name, str_global_search,
                       str_global_search_extents_header);

   tag_strings[7] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[7], comp_b_name, domain_b_name, comp_a_name,
                       domain_a_name, str_global_search,
                       str_global_search_extents);

   tag_strings[8] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[8], comp_a_name, domain_a_name, comp_b_name,
                       domain_b_name, str_global_search,
                       str_global_search_results_header);

   tag_strings[9] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[9], comp_a_name, domain_a_name, comp_b_name,
                       domain_b_name, str_global_search,
                       str_global_search_results);

   tag_strings[10] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[10], comp_b_name, domain_b_name, comp_a_name,
                       domain_a_name, str_global_search,
                       str_global_search_results_header);

   tag_strings[11] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[11], comp_b_name, domain_b_name, comp_a_name,
                       domain_a_name, str_global_search,
                       str_global_search_results);

   tag_strings[12] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[12], comp_a_name, domain_a_name, comp_b_name,
                       domain_b_name, str_global_search,
                       str_global_search_multi_match_cells_header);

   tag_strings[13] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[13], comp_a_name, domain_a_name, comp_b_name,
                       domain_b_name, str_global_search,
                       str_global_search_multi_match_cells);

   tag_strings[14] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[14], comp_b_name, domain_b_name, comp_a_name,
                       domain_a_name, str_global_search,
                       str_global_search_multi_match_cells_header);

   tag_strings[15] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[15], comp_b_name, domain_b_name, comp_a_name,
                       domain_a_name, str_global_search,
                       str_global_search_multi_match_cells);

   tag_strings[16] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[16], comp_a_name, domain_a_name, comp_b_name,
                       domain_b_name, str_global_search,
                       str_global_search_overlapping_ranks);

   tag_strings[17] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[17], comp_b_name, domain_b_name, comp_a_name,
                       domain_a_name, str_global_search,
                       str_global_search_overlapping_ranks);

   tag_strings[18] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[18], comp_a_name, domain_a_name, comp_b_name,
                       domain_b_name, str_global_search,
                       str_global_search_num_overlapping_ranks);

   tag_strings[19] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[19], comp_b_name, domain_b_name, comp_a_name,
                       domain_a_name, str_global_search,
                       str_global_search_num_overlapping_ranks);

   tag_strings[20] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[20], comp_a_name, domain_a_name, comp_b_name,
                       domain_b_name, str_global_search,
                       str_global_search_ranks_domain_header);

   tag_strings[21] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[21], comp_a_name, domain_a_name, comp_b_name,
                       domain_b_name, str_global_search,
                       str_global_search_ranks_domain);

   tag_strings[22] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[22], comp_b_name, domain_b_name, comp_a_name,
                       domain_a_name, str_global_search,
                       str_global_search_ranks_domain_header);

   tag_strings[23] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[23], comp_b_name, domain_b_name, comp_a_name,
                       domain_a_name, str_global_search,
                       str_global_search_ranks_domain);

   unsigned global_tags[24];

   yac_get_unique_comm_tags(
      comm, unique_generator_string, (char const **)tag_strings, 24,
      global_tags);

   free(unique_generator_string);

   tags->global_search_send_grid_data_header = global_tags[0];
   tags->global_search_send_grid_data = global_tags[1];
   tags->global_search_recv_grid_data_header = global_tags[2];
   tags->global_search_recv_grid_data = global_tags[3];
   tags->global_search_send_extents_header = global_tags[4];
   tags->global_search_send_extents = global_tags[5];
   tags->global_search_recv_extents_header = global_tags[6];
   tags->global_search_recv_extents = global_tags[7];
   tags->global_search_send_results_header = global_tags[8];
   tags->global_search_send_results = global_tags[9];
   tags->global_search_recv_results_header = global_tags[10];
   tags->global_search_recv_results = global_tags[11];
   tags->global_search_send_multi_match_cells_header = global_tags[12];
   tags->global_search_send_multi_match_cells = global_tags[13];
   tags->global_search_recv_multi_match_cells_header = global_tags[14];
   tags->global_search_recv_multi_match_cells = global_tags[15];
   tags->global_search_send_overlapping_ranks = global_tags[16];
   tags->global_search_recv_overlapping_ranks = global_tags[17];
   tags->global_search_send_num_overlapping_ranks = global_tags[18];
   tags->global_search_recv_num_overlapping_ranks = global_tags[19];
   tags->global_search_recv_domain_ranks_header = global_tags[20];
   tags->global_search_recv_domain_ranks = global_tags[21];
   tags->global_search_send_domain_ranks_header = global_tags[22];
   tags->global_search_send_domain_ranks = global_tags[23];

   for (unsigned i = 0; i < 24; ++i) free(tag_strings[i]);
}

static void generate_global_search_domain_tags(
   struct global_search_domain_tags * domain_tags,
   char const * comp_a_name, char const * domain_a_name,
   char const * comp_b_name, char const * domain_b_name,
   struct communicator * domain_comm) {

   char const * str_generator_string                              = "generator string";
   char const * str_global_search_gather_extents                  = "gather extents";
   char const * str_global_search_ranks_domain_header             = "domain ranks header";
   char const * str_global_search_ranks_domain                    = "domain ranks";
   char const * str_global_search_bcast_extents_header            = "bcast extents header";
   char const * str_global_search_bcast_extents                   = "bcast extents";
   char const * str_global_search_gatherv_overlapping_ranks       = "gatherv overlapping ranks";
   char const * str_global_search_scatterv_overlapping_ranks      = "scatterv overlapping ranks";
   char const * str_global_search_allgather_num_overlapping_ranks = "allgatherv num overlapping ranks";
   char const * str_global_search_bcast_num_overlapping_ranks     = "bcast num overlapping ranks";
   char const * str_global_search_allgather_num_subdomains        = "allgather num subdomains";

   unsigned max_tag_strlen;

   max_tag_strlen = strlen(str_global_search_allgather_num_overlapping_ranks);

   char const * str_global_search = "global search";

   unsigned max_str_buf_size;

   max_str_buf_size =
    max_tag_strlen + 2 * MAX(strlen(domain_a_name), strlen(domain_b_name)) +
    2 * MAX(strlen(comp_a_name), strlen(comp_b_name)) +
    strlen(str_global_search) + 1;

   char * unique_generator_string = malloc(max_str_buf_size);

   int cmp_comp_names = strcmp(comp_a_name, comp_b_name);
   int cmp_domain_names = strcmp(domain_a_name, domain_b_name);

   if ((cmp_comp_names && (cmp_comp_names > 0)) ||
       (!cmp_comp_names && (cmp_domain_names > 0)))
      generate_tag_string(unique_generator_string,
                          comp_a_name, domain_a_name,
                          comp_b_name, domain_b_name,
                          str_global_search, str_generator_string);
   else
      generate_tag_string(unique_generator_string,
                          comp_b_name, domain_b_name,
                          comp_a_name, domain_a_name,
                          str_global_search, str_generator_string);

   char * domain_tag_strings[10];

   domain_tag_strings[0] = malloc(max_str_buf_size);
   generate_tag_string(domain_tag_strings[0], comp_a_name, domain_a_name,
                       comp_b_name, domain_b_name, str_global_search,
                       str_global_search_gather_extents);

   domain_tag_strings[1] = malloc(max_str_buf_size);
   generate_tag_string(domain_tag_strings[1], comp_a_name, domain_a_name,
                       comp_b_name, domain_b_name, str_global_search,
                       str_global_search_ranks_domain_header);

   domain_tag_strings[2] = malloc(max_str_buf_size);
   generate_tag_string(domain_tag_strings[2], comp_a_name, domain_a_name,
                       comp_b_name, domain_b_name, str_global_search,
                       str_global_search_ranks_domain);

   domain_tag_strings[3] = malloc(max_str_buf_size);
   generate_tag_string(domain_tag_strings[3], comp_a_name, domain_a_name,
                       comp_b_name, domain_b_name, str_global_search,
                       str_global_search_bcast_extents_header);

   domain_tag_strings[4] = malloc(max_str_buf_size);
   generate_tag_string(domain_tag_strings[4], comp_a_name, domain_a_name,
                       comp_b_name, domain_b_name, str_global_search,
                       str_global_search_bcast_extents);

   domain_tag_strings[5] = malloc(max_str_buf_size);
   generate_tag_string(domain_tag_strings[5], comp_a_name, domain_a_name,
                       comp_b_name, domain_b_name, str_global_search,
                       str_global_search_gatherv_overlapping_ranks);

   domain_tag_strings[6] = malloc(max_str_buf_size);
   generate_tag_string(domain_tag_strings[6], comp_a_name, domain_a_name,
                       comp_b_name, domain_b_name, str_global_search,
                       str_global_search_scatterv_overlapping_ranks);

   domain_tag_strings[7] = malloc(max_str_buf_size);
   generate_tag_string(domain_tag_strings[7], comp_a_name, domain_a_name,
                       comp_b_name, domain_b_name, str_global_search,
                       str_global_search_allgather_num_overlapping_ranks);

   domain_tag_strings[8] = malloc(max_str_buf_size);
   generate_tag_string(domain_tag_strings[8], comp_a_name, domain_a_name,
                       comp_b_name, domain_b_name, str_global_search,
                       str_global_search_bcast_num_overlapping_ranks);

   domain_tag_strings[9] = malloc(max_str_buf_size);
   generate_tag_string(domain_tag_strings[9], comp_a_name, domain_a_name,
                       comp_b_name, domain_b_name, str_global_search,
                       str_global_search_allgather_num_subdomains);

   unsigned domain_tags_[10];

   yac_get_unique_comm_tags(
      domain_comm, unique_generator_string, (char const **)domain_tag_strings,
      10, domain_tags_);

   free(unique_generator_string);

   domain_tags->global_search_gather_extents = domain_tags_[0];
   domain_tags->global_search_bcast_domain_ranks_header = domain_tags_[1];
   domain_tags->global_search_bcast_domain_ranks = domain_tags_[2];
   domain_tags->global_search_bcast_extents_header = domain_tags_[3];
   domain_tags->global_search_bcast_extents = domain_tags_[4];
   domain_tags->global_search_gatherv_overlapping_ranks = domain_tags_[5];
   domain_tags->global_search_scatterv_overlapping_ranks = domain_tags_[6];
   domain_tags->global_search_allgather_num_overlapping_ranks = domain_tags_[7];
   domain_tags->global_search_bcast_num_overlapping_ranks = domain_tags_[8];
   domain_tags->global_search_allgather_num_subdomains = domain_tags_[9];

   for (unsigned i = 0; i < 10; ++i) free(domain_tag_strings[i]);
}

static
void pack_multi_overlap_cells(struct global_search * search, int ** pack_buffer_per_rank,
                              int * pack_size_per_rank) {

   if (search->num_overlapping_src_processes == 1)
      return;

   int * curr_global_tgt_id_per_rank = calloc(search->num_overlapping_src_processes,
                                              sizeof(*curr_global_tgt_id_per_rank));

   int smallest_idx[2];
   int * other_src_ranks =
      malloc(search->num_overlapping_src_processes * sizeof(*other_src_ranks));
   int * pack_buffer_per_rank_array_size =
      calloc(search->num_overlapping_src_processes,
             sizeof(*pack_buffer_per_rank_array_size));

   while (1) {

      // find rank with the smallest global_tgt_id
      {
         unsigned i;
         for (i = 0; i < search->num_overlapping_src_processes; ++i)
            if (search->tgt_search_results[i].num_global_tgt_ids >
                curr_global_tgt_id_per_rank[i])
               break;

         if (i == search->num_overlapping_src_processes)
            break;

         smallest_idx[0] = i;
      }

      for (unsigned i = 0; i < search->num_overlapping_src_processes; ++i)
         if ((search->tgt_search_results[i].num_global_tgt_ids >
                curr_global_tgt_id_per_rank[i]) &&
             (search->tgt_search_results[smallest_idx[0]].global_tgt_ids[curr_global_tgt_id_per_rank[smallest_idx[0]]] >
              search->tgt_search_results[i].global_tgt_ids[curr_global_tgt_id_per_rank[i]]))
            smallest_idx[0] = i;

      // find rank with the second smallest global_tgt_id
      {
         unsigned i;
         for (i = 0; i < search->num_overlapping_src_processes; ++i)
            if ((i != smallest_idx[0]) &&
                (search->tgt_search_results[i].num_global_tgt_ids >
                 curr_global_tgt_id_per_rank[i]))
               break;

         if (i == search->num_overlapping_src_processes)
            break;

         smallest_idx[1] = i;
      }

      for (unsigned i = 0; i < search->num_overlapping_src_processes; ++i)
         if ((i != smallest_idx[0]) &&
             (search->tgt_search_results[i].num_global_tgt_ids >
              curr_global_tgt_id_per_rank[i]) &&
             (search->tgt_search_results[smallest_idx[1]].global_tgt_ids[curr_global_tgt_id_per_rank[smallest_idx[1]]] >
              search->tgt_search_results[i].global_tgt_ids[curr_global_tgt_id_per_rank[i]]))
            smallest_idx[1] = i;

      // increment global_id counter of smallest rank until it is not smaller than
      // the second smallest
      while ((curr_global_tgt_id_per_rank[smallest_idx[0]] <
              search->tgt_search_results[smallest_idx[0]].num_global_tgt_ids) &&
             (search->tgt_search_results[smallest_idx[0]].global_tgt_ids[curr_global_tgt_id_per_rank[smallest_idx[0]]] <
              search->tgt_search_results[smallest_idx[1]].global_tgt_ids[curr_global_tgt_id_per_rank[smallest_idx[1]]]))
         curr_global_tgt_id_per_rank[smallest_idx[0]]++;

      // if smallest and second smallest global_tgt_id are identical
      if ((curr_global_tgt_id_per_rank[smallest_idx[0]] <
           search->tgt_search_results[smallest_idx[0]].num_global_tgt_ids) &&
          (search->tgt_search_results[smallest_idx[0]].global_tgt_ids[curr_global_tgt_id_per_rank[smallest_idx[0]]] ==
           search->tgt_search_results[smallest_idx[1]].global_tgt_ids[curr_global_tgt_id_per_rank[smallest_idx[1]]])) {

         int curr_global_tgt_id = search->tgt_search_results[smallest_idx[0]].global_tgt_ids[curr_global_tgt_id_per_rank[smallest_idx[0]]];
         int num_other_src_ranks = 0;

         // how many ranks with the same global_tgt_id are there
         for (int i = 0; i < search->num_overlapping_src_processes; ++i)
            if ((curr_global_tgt_id_per_rank[i] <
                 search->tgt_search_results[i].num_global_tgt_ids) &&
                (search->tgt_search_results[i].global_tgt_ids[curr_global_tgt_id_per_rank[i]] ==
                 curr_global_tgt_id))
               other_src_ranks[num_other_src_ranks++] = i;


         // pack the data for the current global_tgt_id
         for (int i = 0; i < num_other_src_ranks; ++i) {

            ENSURE_ARRAY_SIZE(pack_buffer_per_rank[other_src_ranks[i]],
                              pack_buffer_per_rank_array_size[other_src_ranks[i]],
                              pack_size_per_rank[other_src_ranks[i]] + 2 + num_other_src_ranks);

            pack_buffer_per_rank[other_src_ranks[i]][pack_size_per_rank[other_src_ranks[i]]++] =
               curr_global_tgt_id;
            pack_buffer_per_rank[other_src_ranks[i]][pack_size_per_rank[other_src_ranks[i]]++] =
               num_other_src_ranks-1;
            curr_global_tgt_id_per_rank[other_src_ranks[i]]++;

            for (int j = 0; j < num_other_src_ranks; ++j)
               if (i != j)
                  pack_buffer_per_rank[other_src_ranks[i]][pack_size_per_rank[other_src_ranks[i]]++] =
                     search->tgt_search_results[other_src_ranks[j]].rank;
         }
      }
   } // while

   free(other_src_ranks);
   free(curr_global_tgt_id_per_rank);
   free(pack_buffer_per_rank_array_size);
}

static int compare_tgt_search_results(const void * a, const void * b) {

  int rank_a = ((struct tgt_search_result *)a)->rank;
  int rank_b = ((struct tgt_search_result *)b)->rank;

  return ((rank_a) > (rank_b)) - ((rank_a) < (rank_b));
}

static void return_multi_match_cells(struct global_search * search) {

   // sort the search results by source ranks
   qsort(search->tgt_search_results, search->num_overlapping_src_processes,
         sizeof(*(search->tgt_search_results)), compare_tgt_search_results);

   search->tgt_search_results = realloc(search->tgt_search_results,
                                        search->num_overlapping_src_processes *
                                        sizeof(*search->tgt_search_results));

   int * pack_size_per_rank = calloc(search->num_overlapping_src_processes,
                                     sizeof(*pack_size_per_rank));
   int ** pack_buffer_per_rank = calloc(search->num_overlapping_src_processes,
                                        sizeof(*pack_buffer_per_rank));

   pack_multi_overlap_cells(search, pack_buffer_per_rank, pack_size_per_rank);

   for (int i = 0; i < search->num_overlapping_src_processes; ++i) {

      // send to source
      if (pack_size_per_rank[i] != 0)
         yac_comm_bsend(pack_buffer_per_rank[i], pack_size_per_rank[i], COMM_INT,
                        search->tgt_search_results[i].rank,
                        search->tags.global_search_send_multi_match_cells,
                        search->comm);
      yac_comm_bsend(pack_size_per_rank+i, 1, COMM_INT, search->tgt_search_results[i].rank,
                     search->tags.global_search_send_multi_match_cells_header,
                     search->comm);
   }

   for (int i = 0; i < search->num_overlapping_src_processes; ++i)
      free(pack_buffer_per_rank[i]);
   free(pack_size_per_rank);
   free(pack_buffer_per_rank);
}

static void set_source_search_results(
   struct source_search_results * results,
   struct intermediate_source_search_results * temp_results,
   unsigned num_subdomains) {

   results->global_id = temp_results->global_id;
   results->tgt_rank = temp_results->tgt_rank;
   results->tgt_domain = temp_results->tgt_domain;
   results->local_tgt_cell_id = temp_results->local_tgt_cell_id;
   results->tgt_subdomain_idx = temp_results->tgt_subdomain_idx;
   results->overlapping_src_cells =
      calloc(num_subdomains, sizeof(*results->overlapping_src_cells));
   results->overlapping_src_cells[temp_results->subdomain_idx] =
      temp_results->overlapping_src_cells;
   results->num_overlapping_src_cells_per_subdomain =
      calloc(num_subdomains,
             sizeof(*(results->num_overlapping_src_cells_per_subdomain)));
   results->num_overlapping_src_cells_per_subdomain[temp_results->subdomain_idx] =
      temp_results->num_overlapping_src_cells;
   results->other_src_ranks = NULL;
   results->num_other_src_ranks = 0;
}

static void add_source_search_results (
   struct source_search_results * results,
   struct intermediate_source_search_results * temp_results,
   unsigned num_subdomains) {

   unsigned temp =
      results->num_overlapping_src_cells_per_subdomain[
         temp_results->subdomain_idx];

   if (temp != 0) {
      if (temp != temp_results->num_overlapping_src_cells) {

         yac_internal_abort_message("internal error\n", __FILE__, __LINE__);

      } else {

         free(temp_results->overlapping_src_cells);
      }
   } else {

      results->overlapping_src_cells[temp_results->subdomain_idx] =
         temp_results->overlapping_src_cells;
      results->num_overlapping_src_cells_per_subdomain[temp_results->subdomain_idx] =
         temp_results->num_overlapping_src_cells;
   }
}

static int compare_intermediate_source_search_results(const void * a,
                                                      const void * b) {

  int global_id_a = ((struct intermediate_source_search_results *)a)->global_id;
  int global_id_b = ((struct intermediate_source_search_results *)b)->global_id;

  int ret_val = ((global_id_a) > (global_id_b)) -
                ((global_id_a) < (global_id_b));

  if (ret_val == 0) {

    unsigned subdomain_idx_a =
      ((struct intermediate_source_search_results *)a)->subdomain_idx;
    unsigned subdomain_idx_b =
      ((struct intermediate_source_search_results *)b)->subdomain_idx;

   ret_val = ((subdomain_idx_a) > (subdomain_idx_b)) -
             ((subdomain_idx_a) < (subdomain_idx_b));
  }

  if (ret_val == 0) {

    int tgt_rank_a = ((struct intermediate_source_search_results *)a)->tgt_rank;
    int tgt_rank_b = ((struct intermediate_source_search_results *)b)->tgt_rank;

    ret_val = ((tgt_rank_a) > (tgt_rank_b)) - ((tgt_rank_a) < (tgt_rank_b));
  }

  return ret_val;
}

static void
finalise_src_search_results(
   struct intermediate_global_search_data * search_data) {

   struct global_search * search = search_data->search;

   unsigned num_subdomains;

   num_subdomains = yac_get_num_subdomains(search->local_domain);

   // there is a high possibility that most of the data is sorted,
   // in this case the mergesort used here should outperform quicksort
   // (in some measurements qsort had an significant impact on the runtime)
   yac_mergesort(search_data->temp_src_search_results,
                 search_data->num_temp_src_search_results,
                 sizeof(*search_data->temp_src_search_results),
                 compare_intermediate_source_search_results);

   search->src_search_results =
      malloc(search_data->num_temp_src_search_results *
             sizeof(*(search->src_search_results)));

   struct source_search_results *
      curr_search_results = search->src_search_results - 1;

   struct intermediate_source_search_results *
      curr_intermediate_src_search_results =
         search_data->temp_src_search_results;

   int prev_global_id = -1;

   for (unsigned i = 0; i < search_data->num_temp_src_search_results; ++i) {

      if (curr_intermediate_src_search_results->global_id == prev_global_id)
         add_source_search_results(curr_search_results,
                                   curr_intermediate_src_search_results,
                                   num_subdomains);
      else {
         curr_search_results++;
         search->num_src_search_results++;
         prev_global_id = curr_intermediate_src_search_results->global_id;
         set_source_search_results(curr_search_results,
                                   curr_intermediate_src_search_results,
                                   num_subdomains);
      }
      curr_intermediate_src_search_results++;
   }

   search->src_search_results =
      realloc(search->src_search_results,
              search->num_src_search_results *
             sizeof(*(search->src_search_results)));
}

static void
free_temp_src_search_results(
   struct intermediate_global_search_data * search_data) {

   free(search_data->temp_src_search_results);
   search_data->temp_src_search_results = NULL;
   search_data->temp_src_search_results_array_size = 0;
   search_data->num_temp_src_search_results = 0;
}

static int compare_multimatch_data(const void * a, const void * b) {

  int global_id_a = ((struct multimatch_data *)a)->global_id;
  int global_id_b = ((struct multimatch_data *)b)->global_id;

  return ((global_id_a) > (global_id_b)) - ((global_id_a) < (global_id_b));
}

static void
finalise_multimatch_data(
   struct intermediate_global_search_data * search_data) {

   struct global_search * search = search_data->search;

   // unsigned num_subdomains = get_num_subdomains(search->local_domain);

   qsort(search_data->temp_multimatch_data,
         search_data->num_temp_multimatch_data,
         sizeof(*search_data->temp_multimatch_data),
         compare_multimatch_data);

   unsigned j = 0;

   struct multimatch_data * data = search_data->temp_multimatch_data;

   for (unsigned i = 0; i < search_data->num_temp_multimatch_data; ++i) {

      while ((j < search->num_src_search_results) &&
             (search->src_search_results[j].global_id < data->global_id)) ++j;

      if (j >= search->num_src_search_results)
         yac_internal_abort_message("internal error\n", __FILE__, __LINE__);

      if (search->src_search_results[j].num_other_src_ranks != 0) {

         if  (search->src_search_results[j].num_other_src_ranks !=
              data->num_other_src_ranks) {

            yac_internal_abort_message("internal error\n", __FILE__, __LINE__);

         } else {

            free(data->other_src_ranks);
            data++;
            continue;
         }
      }

      search->src_search_results[j].num_other_src_ranks =
         data->num_other_src_ranks;
      search->src_search_results[j].other_src_ranks =
         data->other_src_ranks;

      ++data;
   }
}

static void
free_temp_multimatch_data(
   struct intermediate_global_search_data * search_data) {

   free(search_data->temp_multimatch_data);
   search_data->temp_multimatch_data = NULL;
   search_data->temp_multimatch_data_array_size = 0;
   search_data->num_temp_multimatch_data = 0;
}

static void
decrement_num_open_target_requests(struct intermediate_global_search_data * search_data) {

   // we have to decrement the counter at the end in order to avoid some race
   // conditions due to asynchronous handling of messages, which can occur in
   // return_multi_match_cells
   if (search_data->num_open_target_requests == 1) {

      if (search_data->search->local_domain_is_target)
         return_multi_match_cells(search_data->search);

      if (search_data->num_open_source_requests == 0) {

         finalise_src_search_results(search_data);
         free_temp_src_search_results(search_data);
         finalise_multimatch_data(search_data);
         free_temp_multimatch_data(search_data);
         free(search_data);
         return;
      }
   }

   search_data->num_open_target_requests--;
}

static void
decrement_num_open_source_requests(struct intermediate_global_search_data * search_data) {

   if (search_data->num_open_source_requests == 1 &&
       search_data->num_open_target_requests == 0) {

         finalise_src_search_results(search_data);
         free_temp_src_search_results(search_data);
         finalise_multimatch_data(search_data);
         free_temp_multimatch_data(search_data);
         free(search_data);

   } else
      search_data->num_open_source_requests--;
}

static
void global_search_receive_results_cont_callback(
   struct communicator * comm, struct comm_message msg, void * user_data) {

   struct intermediate_global_search_data * search_data =
      (struct intermediate_global_search_data *)user_data;
   struct global_search * search = search_data->search;

   unsigned num_matching_cells = msg.count;

   int * matching_global_cell_ids =
      malloc(num_matching_cells * sizeof(*matching_global_cell_ids));

   memcpy(matching_global_cell_ids, msg.buffer,
          num_matching_cells * sizeof(*matching_global_cell_ids));

   // the input data to match_global_ids needs to sorted
   yac_quicksort_index(matching_global_cell_ids, num_matching_cells, NULL);

   unsigned num_subdomains, i, offset;

   num_subdomains = yac_get_num_subdomains(search->local_domain);

   unsigned * matching_local_cell_ids;
   unsigned * num_matching_cells_per_subdomain;

   matching_local_cell_ids = malloc(num_matching_cells * sizeof(*matching_local_cell_ids));
   num_matching_cells_per_subdomain = malloc(num_subdomains * sizeof(*num_matching_cells_per_subdomain));

   // convert global to local cell ids
   for (i = 0, offset = 0; i < num_subdomains; ++i) {

      yac_match_global_ids(
         search->local_domain, i, CELL, matching_global_cell_ids,
         num_matching_cells, NULL, NULL, num_matching_cells_per_subdomain+i,
         matching_local_cell_ids + offset, 0);
      offset += num_matching_cells_per_subdomain[i];
   }

   search->tgt_search_results = realloc(search->tgt_search_results,
                                        (search->num_overlapping_src_processes+1) *
                                        sizeof(*(search->tgt_search_results)));
   // store results
   yac_set_dependencies(
      &(search->tgt_search_results[search->num_overlapping_src_processes].matching_local_cell_ids),
      num_subdomains, num_matching_cells_per_subdomain, matching_local_cell_ids);
   search->tgt_search_results[search->num_overlapping_src_processes].rank = msg.src_rank;
   search->tgt_search_results[search->num_overlapping_src_processes].global_tgt_ids = matching_global_cell_ids;
   search->tgt_search_results[search->num_overlapping_src_processes].num_global_tgt_ids = num_matching_cells;
   search->num_overlapping_src_processes++;

   decrement_num_open_target_requests(search_data);
}

static
void global_search_receive_results_callback(struct communicator * comm,
                                            struct comm_message msg, void * user_data) {

   struct intermediate_global_search_data * search_data =
      (struct intermediate_global_search_data *)user_data;
   struct global_search * search = search_data->search;

   unsigned num_matching_cells;

   num_matching_cells = *((unsigned *)(msg.buffer));

   if (num_matching_cells > 0) {

      // receive results
      yac_comm_recv_callback(num_matching_cells, COMM_INT, msg.src_rank,
                             search->tags.global_search_recv_results, comm,
                             user_data,
                             global_search_receive_results_cont_callback, NULL,
                             MSG_PRIO_LOW);
   } else {

      decrement_num_open_target_requests(search_data);
   }
}

static
void global_search_receive_multi_match_tgt_cells_cont_callback(
   struct communicator * comm, struct comm_message msg, void * user_data) {

   struct intermediate_global_search_data * search_data =
      (struct intermediate_global_search_data *)user_data;

   int * recv_buffer = msg.buffer;

   int i = 0;

   while (i < msg.count) {

      ENSURE_ARRAY_SIZE(search_data->temp_multimatch_data,
                        search_data->temp_multimatch_data_array_size,
                        search_data->num_temp_multimatch_data + 1);

      struct multimatch_data * curr_multimatch_data =
         search_data->temp_multimatch_data +
         search_data->num_temp_multimatch_data;

      search_data->num_temp_multimatch_data++;

      curr_multimatch_data->global_id = recv_buffer[i++];
      curr_multimatch_data->num_other_src_ranks = recv_buffer[i++];
      curr_multimatch_data->other_src_ranks =
         malloc(curr_multimatch_data->num_other_src_ranks *
                sizeof(*(curr_multimatch_data->other_src_ranks)));
      memcpy(curr_multimatch_data->other_src_ranks, recv_buffer + i,
             curr_multimatch_data->num_other_src_ranks *
             sizeof(*(curr_multimatch_data->other_src_ranks)));

      i += curr_multimatch_data->num_other_src_ranks;
   }

   decrement_num_open_source_requests(search_data);
}

static
void global_search_receive_multi_match_tgt_cells_callback(struct communicator * comm,
                                                          struct comm_message msg,
                                                          void * user_data) {

   struct intermediate_global_search_data * search_data =
      (struct intermediate_global_search_data *)user_data;
   struct global_search * search = search_data->search;

   unsigned recv_buffer_size = ((unsigned*)(msg.buffer))[0];

   if (recv_buffer_size > 0) {

      yac_comm_recv_callback(recv_buffer_size, COMM_INT, msg.src_rank,
                             search->tags.global_search_recv_multi_match_cells,
                             comm, user_data,
                             global_search_receive_multi_match_tgt_cells_cont_callback,
                             NULL, MSG_PRIO_HIGH);
   } else {
      decrement_num_open_source_requests(search_data);
   }
}

static int compare_int(const void * a, const void * b) {

  return ((*(int*)a) > (*(int*)b)) - (*((int*)a) < (*(int*)b));
}

static struct domain * unpack_domain(char * domain_buffer,
                                     unsigned buffer_dble_size,
                                     unsigned buffer_int_size,
                                     unsigned recv_buffer_size,
                                     struct communicator * comm) {

   double * buffer_dble = malloc(buffer_dble_size * sizeof(*buffer_dble));
   int * buffer_int = malloc(buffer_int_size * sizeof(*buffer_int));

   // receive grid data
   if (recv_buffer_size > 0) {

      int position = 0;

      yac_comm_unpack(domain_buffer, recv_buffer_size, &position,
                      buffer_dble, buffer_dble_size, COMM_DBLE, comm);
      yac_comm_unpack(domain_buffer, recv_buffer_size, &position,
                      buffer_int, buffer_int_size, COMM_INT, comm);
   }

   struct domain * domain = malloc(1 * sizeof(*domain));

   // generate a domain containing the grid data of the other process
   yac_unpack_domain(domain, comm, NULL, buffer_dble, &buffer_dble_size,
                     buffer_int, &buffer_int_size);

   free(buffer_dble);
   free(buffer_int);

   return domain;
}

static
void global_search_receive_grid_cont_callback(struct communicator * comm,
                                              struct comm_message msg,
                                              void * user_data) {

   struct temp_receive_grid_data * temp_data =
      (struct temp_receive_grid_data *)user_data;
   struct intermediate_global_search_data * search_data = temp_data->search_data;
   struct global_search * search = search_data->search;

   unsigned buffer_dble_size = temp_data->buffer_dble_size;
   unsigned buffer_int_size = temp_data->buffer_int_size;

   free(temp_data);

   struct domain * remote_domain =
      unpack_domain(msg.buffer, buffer_dble_size, buffer_int_size, 
                    msg.count, comm);

   // receive remote domain and store it in global search data structure
   search->remote_domains = realloc(search->remote_domains,
                                    (search->num_remote_domains+1) *
                                    sizeof(*(search->remote_domains)));
   search->remote_domains[search->num_remote_domains].remote_domain =
      remote_domain;
   search->remote_domains[search->num_remote_domains].rank = msg.src_rank;
   search->num_remote_domains++;

   unsigned num_subdomains, num_remote_subdomains;

   num_subdomains = yac_get_num_subdomains(search->local_domain);
   num_remote_subdomains = yac_get_num_subdomains(remote_domain);

   int * global_tgt_ids = NULL;
   unsigned global_tgt_ids_array_size = 0;
   unsigned num_global_tgt_ids = 0;

   // do the search and store the intermediate results
   for (unsigned j = 0; j < num_remote_subdomains; ++j) {

      struct grid * remote_subdomain_grid =
         yac_get_subdomain_grid(*remote_domain, j);
      unsigned curr_num_cells = yac_get_num_grid_cells(remote_subdomain_grid);
      int const * curr_global_ids = yac_get_global_ids(remote_domain, j, CELL);

      for (unsigned i = 0; i < num_subdomains; ++i) {

         struct dep_list temp_cell_search_results;

         yac_do_cell_search(search->grid_search[i], remote_subdomain_grid,
                            &temp_cell_search_results);

         for (unsigned k = 0; k < curr_num_cells; ++k) {

            if ((temp_cell_search_results.num_deps_per_element[k] == 0) ||
                yac_is_halo_value(remote_domain, j, CELL, k, NULL)) continue;

            ENSURE_ARRAY_SIZE(global_tgt_ids, global_tgt_ids_array_size,
                              num_global_tgt_ids + 1);

            global_tgt_ids[num_global_tgt_ids++] = curr_global_ids[k];

            unsigned num_src_cells =
               temp_cell_search_results.num_deps_per_element[k];
            const unsigned * src_cells =
               yac_get_dependencies_of_element(temp_cell_search_results, k);

            ENSURE_ARRAY_SIZE(search_data->temp_src_search_results,
                              search_data->temp_src_search_results_array_size,
                              search_data->num_temp_src_search_results + 1);

            struct intermediate_source_search_results * curr_search_results =
               search_data->temp_src_search_results +
               search_data->num_temp_src_search_results;

            search_data->num_temp_src_search_results++;

            curr_search_results->global_id = curr_global_ids[k];
            curr_search_results->tgt_rank = msg.src_rank;
            curr_search_results->tgt_domain = remote_domain;
            curr_search_results->local_tgt_cell_id = k;
            curr_search_results->tgt_subdomain_idx = j;
            curr_search_results->subdomain_idx = i;
            curr_search_results->overlapping_src_cells =
               malloc(num_src_cells *
                      sizeof(*(curr_search_results->overlapping_src_cells)));
            memcpy(curr_search_results->overlapping_src_cells, src_cells,
                   num_src_cells *
                      sizeof(*(curr_search_results->overlapping_src_cells)));
            curr_search_results->num_overlapping_src_cells = num_src_cells;

         }

         yac_free_dep_list(&temp_cell_search_results);

      } // i (num_subdomains)

   } // j (num_remote_subdomains)

   if (num_global_tgt_ids > 0) {

      // sort all global target ids of cells that overlap with local cells
      qsort(global_tgt_ids, num_global_tgt_ids, sizeof(*global_tgt_ids),
            compare_int);

      // remove duplicated entires
      yac_remove_duplicates_int(global_tgt_ids, &num_global_tgt_ids);

      search->overlapping_tgt_ranks =
         realloc(search->overlapping_tgt_ranks,
                 (search->num_overlapping_tgt_processes + 1) *
                 sizeof(*(search->overlapping_tgt_ranks)));

      search->overlapping_tgt_ranks[search->num_overlapping_tgt_processes++] =
         msg.src_rank;
   }

   if (num_global_tgt_ids > 0) {

      // set up receive of target cells that overlap with source cells
      // from multiple processes
      yac_comm_recv_callback(1, COMM_INT, msg.src_rank,
                             search->tags.global_search_recv_multi_match_cells_header,
                             comm, search_data,
                             global_search_receive_multi_match_tgt_cells_callback,
                             NULL, MSG_PRIO_HIGH);

      // send global_tgt_ids to target
      yac_comm_bsend(global_tgt_ids, num_global_tgt_ids, COMM_INT,
                     msg.src_rank, search->tags.global_search_send_results,
                     comm);
   } else
      decrement_num_open_source_requests(search_data);

   free(global_tgt_ids);

   // send header
   yac_comm_bsend(&num_global_tgt_ids, 1, COMM_UINT, msg.src_rank,
                  search->tags.global_search_send_results_header, comm);
}

static
void global_search_receive_grid_callback(struct communicator * comm,
                                         struct comm_message msg, void * user_data) {

   struct intermediate_global_search_data * search_data =
      (struct intermediate_global_search_data *)user_data;
   struct global_search * search = search_data->search;

   unsigned num_matching_cells = ((unsigned *)(msg.buffer))[0];

   // if there is no grid data
   if (num_matching_cells == 0) {

      decrement_num_open_source_requests(search_data);
      return;
   }

   struct temp_receive_grid_data * temp_data = malloc(1 * sizeof(*temp_data));

   temp_data->search_data = search_data;
   temp_data->buffer_dble_size = ((unsigned *)(msg.buffer))[1];
   temp_data->buffer_int_size = ((unsigned *)(msg.buffer))[2];
   unsigned recv_buffer_size = ((unsigned *)(msg.buffer))[3];

   if (recv_buffer_size > 0) {

      yac_comm_recv_callback(recv_buffer_size, COMM_PACKED, msg.src_rank,
                             search->tags.global_search_recv_grid_data, comm,
                             temp_data,
                             global_search_receive_grid_cont_callback, NULL,
                             MSG_PRIO_LOW);
   } else {

     struct comm_message temp_msg = msg;
     temp_msg.buffer = malloc(0);
     temp_msg.type = COMM_PACKED;
     temp_msg.count = 0;
     temp_msg.tag = search->tags.global_search_recv_grid_data;
     global_search_receive_grid_cont_callback(comm, temp_msg, temp_data);
     free(temp_msg.buffer);
   }
}

static
void pack_extents(double * buf, struct domain * domain,
                  unsigned num_subdomains) {

  struct bounding_circle * extents = malloc(num_subdomains * sizeof(*extents));
  yac_get_subdomain_extents(*domain, extents);

  for (unsigned i = 0; i < num_subdomains; ++i) {

    buf[0 + i * 4] = extents[i].base_vector[0];
    buf[1 + i * 4] = extents[i].base_vector[1];
    buf[2 + i * 4] = extents[i].base_vector[2];
    buf[3 + i * 4] = extents[i].inc_angle;
  }
  free(extents);
}

static
void unpack_extents(double * buf, struct bounding_circle * extents,
                    unsigned num_extents) {

   unsigned i;

   for (i = 0; i < num_extents; ++i) {

      extents[i].base_vector[0] = buf[0 + i * 4];
      extents[i].base_vector[1] = buf[1 + i * 4];
      extents[i].base_vector[2] = buf[2 + i * 4];
      extents[i].inc_angle      = buf[3 + i * 4];
   }
}

static
void global_search_scatterv_overlapping_ranks_callback
   (struct communicator * domain_comm, struct comm_message msg, void * user_data) {

  struct intermediate_global_search_data * search_data =
    (struct intermediate_global_search_data*)user_data;
  struct global_search * search = search_data->search;
  struct communicator * comm = search->comm;

  unsigned * tgt_rank_indices =((unsigned*)(msg.buffer));

  unsigned i = 0;

  for (; i < msg.count; ++i) {
    // set up receive for header of cell data matching local extents
    // from the source
    yac_comm_recv_callback(4, COMM_UINT,
                           search->remote_domain_ranks[tgt_rank_indices[i]],
                           search->tags.global_search_recv_grid_data_header,
                           comm, search_data,
                           global_search_receive_grid_callback, NULL,
                           MSG_PRIO_LOW);
  }

  for (; i < search->remote_domain_size; ++i)
    decrement_num_open_source_requests(search_data);
}

static
void global_search_recv_overlapping_ranks_callback
   (struct communicator * comm, struct comm_message msg, void * user_data) {

  struct temp_recv_overlapping_ranks_data * temp_data =
    (struct temp_recv_overlapping_ranks_data *)user_data;
  struct intermediate_global_search_data * search_data = temp_data->search_data;
  struct global_search * search = search_data->search;
  struct communicator * domain_comm = yac_get_domain_comm(search->local_domain);

  unsigned * src_rank_indices_per_tgt = (unsigned*)(msg.buffer);
  unsigned * tgt_rank_indices_per_src =
    malloc(msg.count * sizeof(*tgt_rank_indices_per_src));
  unsigned * num_tgt_per_src =
    calloc(search->local_domain_size, sizeof(*num_tgt_per_src));
  unsigned * offsets = malloc(search->local_domain_size * sizeof(*offsets));

  for (unsigned i = 0; i < msg.count; ++i)
    ++num_tgt_per_src[src_rank_indices_per_tgt[i]];

  unsigned accum = 0;
  for (unsigned i = 0; i < search->local_domain_size; ++i) {
    offsets[i] = accum;
    accum += num_tgt_per_src[i];
  }

  unsigned offset = 0;

  for (unsigned i = 0; i < search->remote_domain_size; ++i) {
    for (unsigned j = 0; j < temp_data->num_src_per_tgt[i]; ++j, ++offset) {
      tgt_rank_indices_per_src[offsets[src_rank_indices_per_tgt[offset]]] = i;
      ++offsets[src_rank_indices_per_tgt[offset]];
    }
  }

  // broadcast number of target ranks overlapping with each source rank
  yac_comm_bcast_callback(num_tgt_per_src, search->local_domain_size, COMM_UINT,
                          yac_get_comm_rank(domain_comm),
                          search->domain_tags.global_search_bcast_num_overlapping_ranks,
                          domain_comm, search_data, NULL, MSG_PRIO_HIGH);

  // scatterv tgt_rank_indices_per_src
  yac_comm_scatterv_callback(tgt_rank_indices_per_src, num_tgt_per_src,
                             COMM_UINT, yac_get_comm_rank(domain_comm),
                             search->domain_tags.global_search_scatterv_overlapping_ranks,
                             domain_comm, search_data,
                             global_search_scatterv_overlapping_ranks_callback,
                             MSG_PRIO_HIGH);

  free(offsets);
  free(num_tgt_per_src);
  free(tgt_rank_indices_per_src);
  free(temp_data->num_src_per_tgt);
  free(temp_data);
}

static
void global_search_receive_num_overlapping_ranks_callback
   (struct communicator * comm, struct comm_message msg, void * user_data) {

  struct intermediate_global_search_data * search_data =
    (struct intermediate_global_search_data *)user_data;
  struct global_search * search = search_data->search;

  unsigned num_total_overlaps = 0;

  for (unsigned i = 0; i < msg.count; ++i)
    num_total_overlaps += ((unsigned*)(msg.buffer))[i];

  struct temp_recv_overlapping_ranks_data * temp_data =
    malloc(1 * sizeof(*temp_data));

  temp_data->num_src_per_tgt =
    malloc(msg.count * sizeof(*(temp_data->num_src_per_tgt)));
  memcpy(temp_data->num_src_per_tgt, msg.buffer,
         msg.count * sizeof(*(temp_data->num_src_per_tgt)));
  temp_data->search_data = search_data;

  // set up recv for the source domain ranks indices that overlap with each
  // target process
  yac_comm_recv_callback(num_total_overlaps, COMM_UINT,
                         search_data->remote_domain_root,
                         search->tags.global_search_recv_overlapping_ranks,
                         comm, temp_data,
                         global_search_recv_overlapping_ranks_callback, NULL,
                         MSG_PRIO_HIGH);
}

static
void global_search_gatherv_overlapping_ranks_callback(
  struct communicator * domain_comm, struct comm_message msg, void * user_data) {

  struct intermediate_global_search_data * search_data =
    (struct intermediate_global_search_data *)user_data;
  struct global_search * search = search_data->search;
  struct communicator * comm = search->comm;

  // send overlapping ranks to remote root
  yac_comm_bsend(msg.buffer, msg.count, COMM_UINT,
                 search_data->remote_domain_root,
                 search->tags.global_search_send_overlapping_ranks, comm);

  // global_search_gatherv_overlapping_ranks_callback is only called on root and
  // if the local domain is target
  // In case the local domain is also the source, we now set up the receive for
  // the message sent above from the other domain. This is done here in order to
  // enforce a certain message order that improves performance
  if (search->local_domain_is_source) {

    // set up receive for number of source processes overlapping with target
    // processes
    yac_comm_recv_callback(
      search->remote_domain_size, COMM_UINT, search_data->remote_domain_root,
      search->tags.global_search_recv_num_overlapping_ranks, comm, search_data,
      global_search_receive_num_overlapping_ranks_callback, NULL,
      MSG_PRIO_HIGH);
  }

  decrement_num_open_target_requests(search_data);
}

static
void global_search_allgather_num_overlapping_ranks_callback(
  struct communicator * domain_comm, struct comm_message msg, void * user_data) {

  struct temp_overlappping_ranks_data * temp_data =
    (struct temp_overlappping_ranks_data*)user_data;
  struct intermediate_global_search_data * search_data = temp_data->search_data;
  struct global_search * search = search_data->search;
  struct communicator * comm = search->comm;

  unsigned is_root = yac_get_comm_rank(domain_comm) ==
                     search_data->local_domain_root_idx;

  // if root
  if (is_root) {

    // send number of overlapping ranks to remote root
    yac_comm_bsend(msg.buffer, search->local_domain_size, COMM_UINT,
                   search_data->remote_domain_root,
                   search->tags.global_search_send_num_overlapping_ranks, comm);
  }

  // gatherv overlapping ranks
  yac_comm_gatherv_callback(temp_data->tgt_rank_indices, msg.buffer, COMM_UINT,
                            search_data->local_domain_root_idx,
                            search->domain_tags.global_search_gatherv_overlapping_ranks,
                            domain_comm, search_data,
                            (is_root)?global_search_gatherv_overlapping_ranks_callback:NULL,
                            MSG_PRIO_HIGH);

  if (!is_root) decrement_num_open_target_requests(search_data);

  free(temp_data->tgt_rank_indices);
  free(temp_data);
}

static
void global_search_bcast_extents_cont_callback(
  struct communicator * domain_comm, struct comm_message msg, void * user_data) {

  struct intermediate_global_search_data * search_data =
    (struct intermediate_global_search_data *)user_data;
  struct global_search * search = search_data->search;
  struct communicator * comm = search->comm;

  unsigned * extent_buffer_sizes =
    malloc(search->remote_domain_size * sizeof(*(extent_buffer_sizes)));

  int position = 0;

  yac_comm_unpack(msg.buffer, msg.count, &position, extent_buffer_sizes,
                  search->remote_domain_size, COMM_UINT, domain_comm);

  unsigned total_extent_buffer_sizes = 0;

  for (unsigned i = 0; i < search->remote_domain_size; ++i)
    total_extent_buffer_sizes += extent_buffer_sizes[i];

  double * extent_buffer =
    malloc(total_extent_buffer_sizes * sizeof(*extent_buffer));

  yac_comm_unpack(msg.buffer, msg.count, &position, extent_buffer,
                  total_extent_buffer_sizes, COMM_DBLE, domain_comm);

  double * extent_buffer_p = extent_buffer;

  unsigned * tgt_rank_indices =
    malloc(search->remote_domain_size * sizeof(*tgt_rank_indices));

  unsigned num_target_requests = 0;

  for (unsigned i = 0; i < search->remote_domain_size; ++i) {

    unsigned num_remote_extents = extent_buffer_sizes[i] / 4;

    struct bounding_circle * extents =
       malloc(num_remote_extents * sizeof(*extents));

    // unpack extents received from source
    unpack_extents(extent_buffer_p, extents, num_remote_extents);

    extent_buffer_p += extent_buffer_sizes[i];

    double * send_buffer_dble;
    int * send_buffer_int;
    unsigned send_buffer_dble_size, send_buffer_int_size;
    unsigned num_matching_cells;

    send_buffer_dble = NULL, send_buffer_int = NULL;
    send_buffer_dble_size = 0, send_buffer_int_size = 0;

    // match local data with received extents
    yac_pack_domain(search->local_domain, extents, num_remote_extents,
                    &send_buffer_dble, &send_buffer_dble_size, &send_buffer_int,
                    &send_buffer_int_size, &num_matching_cells);

    free(extents);

    // send matching cell data to source process
    if (num_matching_cells > 0) {

      unsigned cell_data_header[4];

      cell_data_header[0] = num_matching_cells;
      cell_data_header[1] = send_buffer_dble_size;
      cell_data_header[2] = send_buffer_int_size;
      cell_data_header[3] = 0;

      unsigned remote_rank = search->remote_domain_ranks[i];

      tgt_rank_indices[num_target_requests] = i;

      num_target_requests++;

      unsigned buffer_size =
        yac_comm_pack_size(send_buffer_dble_size, COMM_DBLE, comm) +
        yac_comm_pack_size(send_buffer_int_size, COMM_INT, comm);
      char * buffer = malloc(buffer_size);

      cell_data_header[3] = buffer_size;

      yac_comm_bsend(cell_data_header, 4, COMM_UINT, remote_rank,
                     search->tags.global_search_send_grid_data_header, comm);

      int position = 0;

      yac_comm_pack(send_buffer_dble, send_buffer_dble_size, COMM_DBLE,
                    buffer, buffer_size, &position, comm);
      yac_comm_pack(send_buffer_int, send_buffer_int_size, COMM_INT,
                    buffer, buffer_size, &position, comm);

      yac_comm_bsend(buffer, buffer_size, COMM_PACKED, remote_rank,
                     search->tags.global_search_send_grid_data, comm);

      free(buffer);

      // set up receive for results of the detailed search on the source
      yac_comm_recv_callback(1, COMM_UINT, remote_rank,
                             search->tags.global_search_recv_results_header, comm,
                             search_data, global_search_receive_results_callback,
                             NULL, MSG_PRIO_LOW);

    } else {
      decrement_num_open_target_requests(search_data);
    }

    free(send_buffer_dble);
    free(send_buffer_int);
  }

  {
    struct temp_overlappping_ranks_data * temp_data =
      malloc(1 * sizeof(*temp_data));

    temp_data->search_data = search_data;
    temp_data->tgt_rank_indices = tgt_rank_indices;

    // allgather number of overlapping ranks
    yac_comm_allgather_callback(
      &num_target_requests, 1, COMM_UINT,
      search->domain_tags.global_search_allgather_num_overlapping_ranks, domain_comm,
      temp_data, global_search_allgather_num_overlapping_ranks_callback,
      MSG_PRIO_HIGH);
  }

  free(extent_buffer);
  free(extent_buffer_sizes);
}

static
void global_search_receive_extents_cont_callback(
  struct communicator * comm, struct comm_message msg, void * user_data) {

  struct intermediate_global_search_data * search_data =
     (struct intermediate_global_search_data *)user_data;
  struct global_search * search = search_data->search;
  struct communicator * domain_comm = yac_get_domain_comm(search->local_domain);

  // broadcast extent data to all other processes of the local domain
  yac_comm_bcast_callback(msg.buffer, msg.count, COMM_PACKED,
                          yac_get_comm_rank(domain_comm),
                          search->domain_tags.global_search_bcast_extents,
                          domain_comm, user_data,
                          global_search_bcast_extents_cont_callback,
                          MSG_PRIO_MIDDLE);
}

static
void global_search_bcast_extents_callback(struct communicator * domain_comm,
                                          struct comm_message msg, void * user_data) {

  struct intermediate_global_search_data * search_data =
    (struct intermediate_global_search_data *)user_data;
  struct global_search * search = search_data->search;
  unsigned extent_bcast_buffer_size = ((unsigned *)(msg.buffer))[0];

  yac_comm_bcast_callback(NULL, extent_bcast_buffer_size, COMM_PACKED,
                          msg.src_rank,
                          search->domain_tags.global_search_bcast_extents,
                          domain_comm, user_data,
                          global_search_bcast_extents_cont_callback,
                          MSG_PRIO_MIDDLE);
}

static
void global_search_receive_extents_callback(struct communicator * comm,
                                            struct comm_message msg, void * user_data) {

  struct intermediate_global_search_data * search_data =
    (struct intermediate_global_search_data *)user_data;
  struct global_search * search = search_data->search;
  struct communicator * domain_comm = yac_get_domain_comm(search->local_domain);
  unsigned extents_buffer_size = ((unsigned *)(msg.buffer))[0];

  // broadcast size of extents buffer size to all other processes of the local
  // domain
  yac_comm_bcast_callback(msg.buffer, 1, COMM_UINT,
                          yac_get_comm_rank(domain_comm),
                          search->domain_tags.global_search_bcast_extents_header,
                          domain_comm, user_data, NULL, MSG_PRIO_MIDDLE);

  // set up receive for the actual extent data from the remote domain root
  yac_comm_recv_callback(extents_buffer_size, COMM_PACKED, msg.src_rank,
                         search->tags.global_search_recv_extents, comm,
                         user_data, global_search_receive_extents_cont_callback,
                         NULL, MSG_PRIO_MIDDLE);
}

static
void global_search_gather_extents(struct communicator * domain_comm,
                                  struct comm_message msg, void * user_data) {

  struct intermediate_global_search_data * search_data =
    (struct intermediate_global_search_data *)user_data;
  struct global_search * search = search_data->search;
  struct communicator * comm = search->comm;

  unsigned domain_comm_size = yac_get_comm_size(domain_comm);

  unsigned total_extent_buffer_size = 0;
  for (unsigned i = 0; i < domain_comm_size; ++i)
    total_extent_buffer_size += search_data->extent_buffer_sizes[i];

  unsigned send_buffer_size = yac_comm_pack_size(domain_comm_size, COMM_UINT,
                                                 comm) +
                              yac_comm_pack_size(total_extent_buffer_size,
                                                 COMM_DBLE, comm);

  void * send_buffer = malloc(send_buffer_size);
  int position = 0;

  // pack extent_buffer_sizes of each process of the local domain and the actual
  // extents into the send_buffer
  yac_comm_pack(search_data->extent_buffer_sizes, domain_comm_size, COMM_UINT,
                send_buffer, send_buffer_size, &position, comm);
  yac_comm_pack(msg.buffer, total_extent_buffer_size, COMM_DBLE,
                send_buffer, send_buffer_size, &position, comm);

  free(search_data->extent_buffer_sizes);
  search_data->extent_buffer_sizes = NULL;

  // send size of send_buffer to root of remote domain
  yac_comm_bsend(&send_buffer_size, 1, COMM_UINT,
                 search_data->remote_domain_root,
                 search->tags.global_search_send_extents_header, comm);
  // send actual send buffer to root of remote domain
  yac_comm_bsend(send_buffer, send_buffer_size, COMM_PACKED,
                 search_data->remote_domain_root,
                 search->tags.global_search_send_extents, comm);

   // if the local domain is also target we now set up the receive for the
   // extents from the other domain
   // (global_search_gather_extents is only on the root process of the local
   //  domain)
   if (search->local_domain_is_target)
      // set up receive for the extents header message of the root of the
      // remote domain
      yac_comm_recv_callback(1, COMM_UINT, search_data->remote_domain_root,
                             search->tags.global_search_recv_extents_header,
                             comm, search_data,
                             global_search_receive_extents_callback, NULL,
                             MSG_PRIO_HIGH);

  free(send_buffer);
}

static
void global_search_bcast_num_overlapping_ranks_callback
   (struct communicator * domain_comm, struct comm_message msg, void * user_data) {

  struct intermediate_global_search_data * search_data =
    (struct intermediate_global_search_data*)user_data;
  struct global_search * search = search_data->search;

  // set up scatterv for the remote target domain rank indice that overlap with
  // the local process
  yac_comm_scatterv_callback(NULL, msg.buffer, COMM_UINT,
                             search_data->local_domain_root_idx,
                             search->domain_tags.global_search_scatterv_overlapping_ranks,
                             domain_comm, search_data,
                             global_search_scatterv_overlapping_ranks_callback,
                             MSG_PRIO_HIGH);
}

static
void global_search_allgather_num_subdomains
   (struct communicator * domain_comm, struct comm_message msg, void * user_data) {

  struct intermediate_global_search_data * search_data =
    (struct intermediate_global_search_data *)user_data;
  struct global_search * search = search_data->search;
  struct communicator * comm = search->comm;

  unsigned * extent_buffer_sizes = (unsigned*)(msg.buffer);
  unsigned num_subdomains  = yac_get_num_subdomains(search->local_domain);

  for (unsigned i = 0; i < search->local_domain_size; ++i)
    extent_buffer_sizes[i] *= 4;

  double * extents = malloc(num_subdomains * 4 * sizeof(*extents));
  pack_extents(extents, search->local_domain, num_subdomains);

  unsigned is_root = yac_get_comm_rank(domain_comm) ==
                     search_data->local_domain_root_idx;

  if (is_root) {

    unsigned buffer_size = yac_get_comm_size(domain_comm) *
                           sizeof(*(search_data->extent_buffer_sizes));

    search_data->extent_buffer_sizes = malloc(buffer_size);
    memcpy(search_data->extent_buffer_sizes, extent_buffer_sizes, buffer_size);

  } else {
    search_data->extent_buffer_sizes = NULL;
  }

  // gather all extents on root of the local domain
  yac_comm_gatherv_callback(extents, extent_buffer_sizes, COMM_DBLE,
                            search_data->local_domain_root_idx,
                            search->domain_tags.global_search_gather_extents,
                            domain_comm, user_data,
                            (is_root)?(global_search_gather_extents):(NULL),
                            MSG_PRIO_HIGH);

  free(extents);

  if (is_root) {

    // global_search_allgather_num_subdomains is only called on source domains.
    // In case the local domain also has the role of a target, we delay the
    // following receive in order to enfore a certain message order that
    // improves performance. (the call is then done in
    // global_search_gatherv_overlapping_ranks_callback)
    if (!search->local_domain_is_target) {
      // set up receive for number of source processes overlapping with target
      // processes
      yac_comm_recv_callback(
        search->remote_domain_size, COMM_UINT, search_data->remote_domain_root,
        search->tags.global_search_recv_num_overlapping_ranks, comm,
        search_data, global_search_receive_num_overlapping_ranks_callback, NULL,
        MSG_PRIO_HIGH);
    }
  } else {

    // set up broadcast for number of target ranks overlapping with each source
    // process
    yac_comm_bcast_callback(NULL, search->local_domain_size, COMM_UINT,
                            search_data->local_domain_root_idx,
                            search->domain_tags.global_search_bcast_num_overlapping_ranks,
                            domain_comm, search_data,
                            global_search_bcast_num_overlapping_ranks_callback,
                            MSG_PRIO_HIGH);
  }
}

static
void global_search_receive_domain_ranks_dummy_cont(struct communicator * comm,
                                                   struct comm_message msg,
                                                   void * user_data) {

   return;
}

static
void global_search_receive_domain_ranks_dummy(struct communicator * comm,
                                              struct comm_message msg,
                                              void * user_data) {

   struct intermediate_global_search_data * search_data =
      (struct intermediate_global_search_data *) user_data;
   struct global_search * search = search_data->search;

   yac_comm_recv_callback(search->local_domain_size, COMM_UINT, msg.src_rank,
                          search->tags.global_search_send_domain_ranks, comm,
                          user_data,
                          global_search_receive_domain_ranks_dummy_cont, NULL,
                          MSG_PRIO_HIGH);
}

static
void global_search_bcast_domain_ranks_cont
   (struct communicator * domain_comm, struct comm_message msg, void * user_data) {

   struct intermediate_global_search_data * search_data =
      (struct intermediate_global_search_data *)user_data;
   struct global_search * search = search_data->search;
   struct communicator * comm = search->comm;

   // store the ranks from the remote domain
   search->remote_domain_size = msg.count;
   search->remote_domain_ranks = malloc(search->remote_domain_size *
                                        sizeof(*(search->remote_domain_ranks)));
   memcpy(search->remote_domain_ranks, msg.buffer,
          search->remote_domain_size * sizeof(*(search->remote_domain_ranks)));

   // compute the root process of the remote domain
   unsigned remote_domain_root = search->remote_domain_ranks[0];
   for (unsigned i = 1; i < search->remote_domain_size; ++i)
      if (remote_domain_root > search->remote_domain_ranks[i])
         remote_domain_root = search->remote_domain_ranks[i];

   search_data->remote_domain_root = remote_domain_root;

   search_data->num_open_target_requests =
      (search->local_domain_is_target)?(search->remote_domain_size + 1):0;
   search_data->num_open_source_requests =
      (search->local_domain_is_source)?search->remote_domain_size:0;

   // see comment below on why we need this variable here
   unsigned local_domain_root_idx = search_data->local_domain_root_idx;

   if (search->local_domain_is_source) {

      unsigned num_subdomains  = yac_get_num_subdomains(search->local_domain);

      search->grid_search = malloc(num_subdomains * sizeof(*(search->grid_search)));

      for (unsigned i = 0; i < num_subdomains; ++i)
         search->grid_search[i] =
            yac_get_grid_search(search->local_domain->subdomains[i]->grid);

      // allgather size of extents buffer of all processes in the local domain
      yac_comm_allgather_callback(&num_subdomains, 1, COMM_UINT,
        search->domain_tags.global_search_allgather_num_subdomains, domain_comm,
        search_data, global_search_allgather_num_subdomains, MSG_PRIO_HIGH);
   }

   if (search->local_domain_is_target) {

      if (yac_get_global_ids(search->local_domain, 0, CELL) == NULL)
         yac_internal_abort_message("ERROR: global cell ids are not defined for"
                                    " the local domain", __FILE__, __LINE__);

      // if the local process is the local domain root and a target but not a
      // source process it is possible that search_data is already deallocated
      // at this point, that is why we had to store local_domain_root_idx in a
      // variable

      // if the process is the root of the local domain
      if (yac_get_comm_rank(domain_comm) == local_domain_root_idx) {

        // if the local domain is also a source domain we wait with this receive
        // until root has gather all extents from the local domain
        // the receive is then call from within global_search_gather_extents
        if (!search->local_domain_is_source)
           // set up receive for the extents header message of the root of the
           // remote domain
           yac_comm_recv_callback(1, COMM_UINT, search_data->remote_domain_root,
                                  search->tags.global_search_recv_extents_header,
                                  comm, search_data,
                                  global_search_receive_extents_callback, NULL,
                                  MSG_PRIO_HIGH);

      } else {

        // set up broadcast callback for extents from source domain
        yac_comm_bcast_callback(NULL, 1, COMM_UINT,
                                search_data->local_domain_root_idx,
                                search->domain_tags.global_search_bcast_extents_header,
                                domain_comm, search_data,
                                global_search_bcast_extents_callback,
                                MSG_PRIO_MIDDLE);
      }
   }
}

static
void global_search_receive_remote_domain_ranks_cont
   (struct communicator * comm, struct comm_message msg, void * user_data) {

   struct intermediate_global_search_data * search_data =
      (struct intermediate_global_search_data *)user_data;
   struct global_search * search = search_data->search;

   struct communicator * domain_comm = yac_get_domain_comm(search->local_domain);
   unsigned domain_comm_rank = yac_get_comm_rank(domain_comm);

   unsigned remote_domain_size = msg.count;
   unsigned * remote_domain_ranks = (unsigned*)(msg.buffer);

   unsigned remote_domain_root = remote_domain_ranks[0];

   // compute the root process of the remote domain
   for (unsigned i = 1; i < remote_domain_size; ++i)
      if (remote_domain_root > remote_domain_ranks[i])
         remote_domain_root = remote_domain_ranks[i];

   // if the remote domain root is not the global root
   if (remote_domain_root != 0) {

      // There is no one to receive the ranks of the local domain, which were
      // sent by the the root of the local domain in \ref yac_global_search_new. Therefore,
      // we have to receive this message here and send the ranks of the local domain
      // to the root of the remote domain
      yac_comm_recv_callback(1, COMM_UINT, 0,
                             search->tags.global_search_send_domain_ranks_header, comm,
                             search_data, global_search_receive_domain_ranks_dummy, NULL,
                             MSG_PRIO_HIGH);
      yac_comm_bsend(&(search->local_domain_size), 1, COMM_UINT,
                     remote_domain_root,
                     search->tags.global_search_send_domain_ranks_header, comm);
      yac_comm_bsend(search->local_domain_ranks, search->local_domain_size, COMM_UINT,
                     remote_domain_root,
                     search->tags.global_search_send_domain_ranks, comm);
   }

   // broadcast the size of the remote domain to all other processes in the
   // local domain
   yac_comm_bcast_callback(
     &remote_domain_size, 1, COMM_UINT, domain_comm_rank,
     search->domain_tags.global_search_bcast_domain_ranks_header, domain_comm,
     search_data, NULL, MSG_PRIO_HIGH);

   // broadcast the remote domain ranks
   yac_comm_bcast_callback(
      remote_domain_ranks, remote_domain_size, COMM_UINT, domain_comm_rank,
      search->domain_tags.global_search_bcast_domain_ranks, domain_comm,
      search_data, global_search_bcast_domain_ranks_cont, MSG_PRIO_HIGH);
}

static
void global_search_bcast_domain_ranks
   (struct communicator * domain_comm, struct comm_message msg,
    void * user_data) {

   struct intermediate_global_search_data * search_data =
      (struct intermediate_global_search_data *)user_data;
   struct global_search * search = search_data->search;

   yac_comm_bcast_callback(
      NULL, *((unsigned *)(msg.buffer)), COMM_UINT, msg.src_rank,
      search->domain_tags.global_search_bcast_domain_ranks, domain_comm,
      search_data, global_search_bcast_domain_ranks_cont, MSG_PRIO_HIGH);
}

static
void global_search_receive_remote_domain_ranks
   (struct communicator * comm, struct comm_message msg, void * user_data) {

   struct intermediate_global_search_data * search_data =
      (struct intermediate_global_search_data *)user_data;
   struct global_search * search = search_data->search;

   yac_comm_recv_callback(*(unsigned *)(msg.buffer), COMM_UINT, msg.src_rank,
                          search->tags.global_search_recv_domain_ranks, comm,
                          search_data,
                          global_search_receive_remote_domain_ranks_cont, NULL,
                          MSG_PRIO_HIGH);
}

struct global_search *
yac_global_search_new_(struct domain * local_domain,
                       char const * remote_domain_name,
                       struct global_search_tags tags,
                       struct global_search_domain_tags domain_tags,
                       struct communicator * comm,
                       unsigned local_domain_is_source,
                       unsigned local_domain_is_target,
                       char const * local_component_name,
                       char const * remote_component_name) {

   struct global_search * search = malloc(1 * sizeof(*search));

   struct communicator * domain_comm = yac_get_domain_comm(local_domain);
   unsigned domain_size = yac_get_comm_size(domain_comm);

   unsigned num_subdomains = yac_get_num_subdomains(local_domain);

   search->tags = tags;
   search->domain_tags = domain_tags;
   search->local_domain = local_domain;
   search->remote_domains = NULL;
   search->num_remote_domains = 0;
   search->local_domain_is_source = local_domain_is_source;
   search->local_domain_is_target = local_domain_is_target;
   search->comm = comm;
   search->tgt_search_results = NULL;
   search->num_overlapping_src_processes = 0;
   search->grid_search = NULL;
   search->overlapping_tgt_ranks = NULL;
   search->num_overlapping_tgt_processes = 0;
   search->src_search_results = NULL;
   search->num_src_search_results = 0;

   size_t name_length;

   name_length = strlen(remote_domain_name) + 1;
   search->remote_domain_name = malloc(name_length);
   strcpy(search->remote_domain_name, remote_domain_name);
   name_length = strlen(local_component_name) + 1;
   search->local_component_name = malloc(name_length);
   strcpy(search->local_component_name, local_component_name);
   name_length = strlen(remote_component_name) + 1;
   search->remote_component_name = malloc(name_length);
   strcpy(search->remote_component_name, remote_component_name);

   search->local_domain_size = domain_size;
   search->local_domain_ranks =
      malloc(domain_size * sizeof(*(search->local_domain_ranks)));
   

   // determine ranks of all processes of the local domain within
   // the global search communicator
   yac_comm_translate_ranks(comm, domain_comm, search->local_domain_ranks);

   // find the root of the local domain
   unsigned local_domain_root_idx = 0;
   yac_get_lowest_local_domain_rank(search, &local_domain_root_idx);

   struct intermediate_global_search_data * search_data;

   search_data = malloc(1 * sizeof(*search_data));

   search_data->search = search;
   search_data->temp_multimatch_data = NULL;
   search_data->temp_multimatch_data_array_size = 0;
   search_data->num_temp_multimatch_data = 0;

   search_data->temp_src_search_results = NULL;
   search_data->temp_src_search_results_array_size = 0;
   search_data->num_temp_src_search_results = 0;
   search_data->local_domain_root_idx = local_domain_root_idx;

   // if this is local domain root
   if (local_domain_root_idx == yac_get_comm_rank(domain_comm)) {

      // send ranks of local domain processes to global root
      yac_comm_bsend(&domain_size, 1, COMM_UINT, 0,
                     search->tags.global_search_send_domain_ranks_header,
                     comm);
      yac_comm_bsend(search->local_domain_ranks, domain_size, COMM_UINT, 0,
                     search->tags.global_search_send_domain_ranks, comm);

     // set up receive for remote domain ranks
     yac_comm_recv_callback(1, COMM_UINT, COMM_ANYSOURCE,
                            search->tags.global_search_recv_domain_ranks_header,
                            comm, search_data,
                            global_search_receive_remote_domain_ranks, NULL,
                            MSG_PRIO_HIGH);
   } else {

      // all non-local-domain root processes initialise broadcast for number of
      // remote domain processes
      yac_comm_bcast_callback(
        NULL, 1, COMM_UINT, local_domain_root_idx,
        search->domain_tags.global_search_bcast_domain_ranks_header,
        domain_comm, search_data, global_search_bcast_domain_ranks,
        MSG_PRIO_HIGH);
   }

   return search;
}

struct global_search *
yac_global_search_new(struct domain * local_domain,
                      char const * remote_domain_name,
                      struct communicator * comm,
                      unsigned local_domain_is_source,
                      unsigned local_domain_is_target,
                      char const * local_component_name,
                      char const * remote_component_name) {

   struct global_search_domain_tags domain_tags;
   struct global_search_tags tags;

   // generate tags
   generate_global_search_tags(
      &tags, local_component_name, local_domain->name,
      remote_component_name, remote_domain_name, comm);
   generate_global_search_domain_tags(
      &domain_tags, local_component_name, local_domain->name,
      remote_component_name, remote_domain_name,
      yac_get_domain_comm(local_domain));

   return yac_global_search_new_(
      local_domain, remote_domain_name, tags, domain_tags, comm,
      local_domain_is_source, local_domain_is_target,
      local_component_name, remote_component_name);
}

void yac_global_search_all_local_new(
  struct domain * domains[2], struct communicator * comm,
  unsigned domain_is_source[2], char const * component_names[2],
  struct global_search * searches[2]) {

  struct global_search_domain_tags domain_tags;
  struct global_search_tags tags[2];

  int flag = strcmp(domains[0]->name, domains[1]->name) < 0;

  // generate tags
  generate_global_search_tags(
     &(tags[0]), component_names[flag], domains[flag]->name,
     component_names[flag^1], domains[flag^1]->name, comm);
  generate_global_search_domain_tags(
     &domain_tags, component_names[flag], domains[flag]->name,
     component_names[flag^1], domains[flag^1]->name,
     yac_get_domain_comm(domains[flag]));

  searches[flag] = yac_global_search_new_(
    domains[flag], domains[flag^1]->name, tags[0], domain_tags, comm,
    domain_is_source[flag], domain_is_source[flag^1],
    component_names[flag], component_names[flag^1]);

  // generate tags
  tags[1].global_search_send_grid_data_header =
    tags[0].global_search_recv_grid_data_header;
  tags[1].global_search_send_grid_data =
    tags[0].global_search_recv_grid_data;
  tags[1].global_search_recv_grid_data_header =
    tags[0].global_search_send_grid_data_header;
  tags[1].global_search_recv_grid_data =
    tags[0].global_search_send_grid_data;
  tags[1].global_search_send_extents_header =
    tags[0].global_search_recv_extents_header;
  tags[1].global_search_send_extents =
    tags[0].global_search_recv_extents;
  tags[1].global_search_recv_extents_header =
    tags[0].global_search_send_extents_header;
  tags[1].global_search_recv_extents =
    tags[0].global_search_send_extents;
  tags[1].global_search_send_results_header =
    tags[0].global_search_recv_results_header;
  tags[1].global_search_send_results =
    tags[0].global_search_recv_results;
  tags[1].global_search_recv_results_header =
    tags[0].global_search_send_results_header;
  tags[1].global_search_recv_results =
    tags[0].global_search_send_results;
  tags[1].global_search_send_multi_match_cells_header =
    tags[0].global_search_recv_multi_match_cells_header;
  tags[1].global_search_send_multi_match_cells =
    tags[0].global_search_recv_multi_match_cells;
  tags[1].global_search_recv_multi_match_cells_header =
    tags[0].global_search_send_multi_match_cells_header;
  tags[1].global_search_recv_multi_match_cells =
    tags[0].global_search_send_multi_match_cells;
  tags[1].global_search_send_overlapping_ranks =
    tags[0].global_search_recv_overlapping_ranks;
  tags[1].global_search_recv_overlapping_ranks =
    tags[0].global_search_send_overlapping_ranks;
  tags[1].global_search_send_num_overlapping_ranks =
    tags[0].global_search_recv_num_overlapping_ranks;
  tags[1].global_search_recv_num_overlapping_ranks =
    tags[0].global_search_send_num_overlapping_ranks;
  tags[1].global_search_recv_domain_ranks_header =
    tags[0].global_search_send_domain_ranks_header;
  tags[1].global_search_recv_domain_ranks =
    tags[0].global_search_send_domain_ranks;
  tags[1].global_search_send_domain_ranks_header =
    tags[0].global_search_recv_domain_ranks_header;
  tags[1].global_search_send_domain_ranks =
    tags[0].global_search_recv_domain_ranks;

  generate_global_search_domain_tags(
     &domain_tags, component_names[flag^1], domains[flag^1]->name,
     component_names[flag], domains[flag]->name,
     yac_get_domain_comm(domains[flag^1]));

  searches[flag^1] = yac_global_search_new_(
    domains[flag^1], domains[flag]->name, tags[1], domain_tags, comm,
    domain_is_source[flag^1], domain_is_source[flag],
    component_names[flag^1], component_names[flag]);
}

void yac_enddef_global_search(struct global_search * search) {

   struct communicator * domain_comm = yac_get_domain_comm(search->local_domain);

   // wait until all requests are handled
   while (
      yac_comm_wait(domain_comm,
                    search->domain_tags.global_search_gather_extents) ||
      yac_comm_wait(domain_comm,
                    search->domain_tags.global_search_bcast_domain_ranks_header) ||
      yac_comm_wait(domain_comm,
                    search->domain_tags.global_search_bcast_domain_ranks) ||
      yac_comm_wait(domain_comm,
                    search->domain_tags.global_search_bcast_extents_header) ||
      yac_comm_wait(domain_comm,
                    search->domain_tags.global_search_bcast_extents) ||
      yac_comm_wait(domain_comm,
                    search->domain_tags.global_search_gatherv_overlapping_ranks) ||
      yac_comm_wait(domain_comm,
                    search->domain_tags.global_search_scatterv_overlapping_ranks) ||
      yac_comm_wait(domain_comm,
                    search->domain_tags.global_search_allgather_num_overlapping_ranks) ||
      yac_comm_wait(domain_comm,
                    search->domain_tags.global_search_bcast_num_overlapping_ranks) ||
      yac_comm_wait(domain_comm,
                    search->domain_tags.global_search_allgather_num_subdomains) ||
      yac_comm_wait(search->comm,
                    search->tags.global_search_recv_domain_ranks_header) ||
      yac_comm_wait(search->comm,
                    search->tags.global_search_recv_domain_ranks) ||
      yac_comm_wait(search->comm,
                    search->tags.global_search_recv_extents_header) ||
      yac_comm_wait(search->comm,
                    search->tags.global_search_recv_extents) ||
      yac_comm_wait(search->comm,
                    search->tags.global_search_recv_grid_data_header) ||
      yac_comm_wait(search->comm,
                    search->tags.global_search_recv_grid_data) ||
      yac_comm_wait(search->comm,
                    search->tags.global_search_recv_results_header) ||
      yac_comm_wait(search->comm,
                    search->tags.global_search_recv_results) ||
      yac_comm_wait(search->comm,
                    search->tags.global_search_recv_multi_match_cells_header) ||
      yac_comm_wait(search->comm,
                    search->tags.global_search_recv_multi_match_cells) ||
      yac_comm_wait(search->comm,
                    search->tags.global_search_send_overlapping_ranks) ||
      yac_comm_wait(search->comm,
                    search->tags.global_search_recv_overlapping_ranks) ||
      yac_comm_wait(search->comm,
                    search->tags.global_search_send_num_overlapping_ranks) ||
      yac_comm_wait(search->comm,
                    search->tags.global_search_recv_num_overlapping_ranks) ||
      yac_comm_wait(search->comm,
                    search->tags.global_search_send_results_header) ||
      yac_comm_wait(search->comm,
                    search->tags.global_search_send_results) ||
      yac_comm_wait(search->comm,
                    search->tags.global_search_send_multi_match_cells) ||
      yac_comm_wait(search->comm,
                    search->tags.global_search_send_multi_match_cells_header) ||
      yac_comm_wait(search->comm,
                    search->tags.global_search_send_extents_header) ||
      yac_comm_wait(search->comm,
                    search->tags.global_search_send_extents) ||
      yac_comm_wait(search->comm,
                    search->tags.global_search_send_grid_data_header) ||
      yac_comm_wait(search->comm,
                    search->tags.global_search_send_grid_data) ||
      yac_comm_wait(search->comm,
                    search->tags.global_search_send_domain_ranks_header) ||
      yac_comm_wait(search->comm,
                    search->tags.global_search_send_domain_ranks));

   if (search->local_domain_is_source) {
      // sort ranks of all processes that have an overlap with the local rank
      qsort(search->overlapping_tgt_ranks,
            search->num_overlapping_tgt_processes,
            sizeof(*(search->overlapping_tgt_ranks)), compare_int);
   }
}

struct grid_search * yac_get_subdomain_grid_search(struct global_search * search,
                                                   unsigned subdomain_idx) {

   if (!search->local_domain_is_source)
      yac_internal_abort_message ( "ERROR: local domain is not registered as a souce.", __FILE__, __LINE__ );

   if (subdomain_idx >= search->local_domain->num_subdomains)
      yac_internal_abort_message ( "ERROR: subdomain index out of range.", __FILE__, __LINE__ );

   return search->grid_search[subdomain_idx];
}

int yac_source_process_has_overlap_with(struct global_search * search,
                                        int rank) {

   if (!search->local_domain_is_source)
      yac_internal_abort_message ( "ERROR: local domain is not registered as a souce.", __FILE__, __LINE__ );

   unsigned i;

   for (i = 0; i < search->num_overlapping_tgt_processes; ++i)
      if (search->overlapping_tgt_ranks[i] == rank)
         break;

   return (i != search->num_overlapping_tgt_processes);
}

int yac_target_process_has_overlap_with(struct global_search * search,
                                    int rank) {

   if (!search->local_domain_is_target)
      yac_internal_abort_message ( "ERROR: local domain is not registered as a target.",
                                   __FILE__, __LINE__ );

   unsigned i;

   for (i = 0; i < search->num_overlapping_src_processes; ++i)
      if (search->tgt_search_results[i].rank == rank)
         break;

   return (i != search->num_overlapping_src_processes);
}

struct dep_list const * yac_get_matching_local_cell_ids(struct global_search * search,
                                                        unsigned rank) {

   if (!search->local_domain_is_target)
      yac_internal_abort_message ( "ERROR: local domain is not registered as a target.",
                                   __FILE__, __LINE__ );

   if (rank >= yac_get_comm_size(search->comm))
      yac_internal_abort_message ( "ERROR: local domain is not registered as a target.",
                                   __FILE__, __LINE__ );

   unsigned i;

   for (i = 0; i < search->num_overlapping_src_processes; ++i)
      if (search->tgt_search_results[i].rank == rank)
         break;

   if (i != search->num_overlapping_src_processes)
      return &(search->tgt_search_results[i].matching_local_cell_ids);
   else
      return NULL;
}

struct source_search_results const * yac_get_search_results_for_tgt_cell (
   struct global_search * search, int global_cell_id) {

   int min = 0, max = search->num_src_search_results - 1;

   // use binary search to find the results for the given target cell

   while (max >= min) {

      int middle  = (max + min) >> 1;

      if (search->src_search_results[middle].global_id == global_cell_id)
         return search->src_search_results + middle;
      else if (search->src_search_results[middle].global_id < global_cell_id)
         min = middle + 1;
      else
         max = middle - 1;
   }

   return NULL;
}

void yac_get_search_results_for_tgt_cells (
   struct global_search * search, int * global_cell_ids,
   unsigned num_global_cell_ids,
   struct source_search_results const ** results) {

   if (num_global_cell_ids == 0) return;

   if (search->num_src_search_results == 0) {
      for (unsigned i = 0; i < num_global_cell_ids; ++i)
         results[i] = NULL;
      return;
   }

   unsigned j = 0;

   for (unsigned i = 0; i < num_global_cell_ids; ++i) {

      while ((j < search->num_src_search_results) &&
             (search->src_search_results[j].global_id <
              global_cell_ids[i])) ++j;

      if (j >= search->num_src_search_results) {

         for (; i < num_global_cell_ids; ++i)
            results[i] = NULL;
         break;
      }

      if (search->src_search_results[j].global_id == global_cell_ids[i])
         results[i] = search->src_search_results + j;
      else
         results[i] = NULL;
   }
}

struct communicator * yac_get_communicator(struct global_search * search) {

   return search->comm;
}

unsigned yac_get_lowest_local_domain_rank(struct global_search * search,
                                          unsigned * local_comm_rank) {

   unsigned lowest = search->local_domain_ranks[0];

   if (local_comm_rank == NULL) {

     for (unsigned i = 1; i < search->local_domain_size; ++i)
        if (lowest > search->local_domain_ranks[i])
           lowest = search->local_domain_ranks[i];
   } else {

      *local_comm_rank = 0;
      
      for (unsigned i = 1; i < search->local_domain_size; ++i) {
         if (lowest > search->local_domain_ranks[i]) {
            *local_comm_rank = i;
            lowest = search->local_domain_ranks[i];
         }
      }
   }

   return lowest;
}

unsigned yac_get_lowest_remote_domain_rank(struct global_search * search) {

   unsigned i, lowest;

   lowest = search->remote_domain_ranks[0];
   for (i = 1; i < search->remote_domain_size; ++i)
      if (lowest > search->remote_domain_ranks[i])
         lowest = search->remote_domain_ranks[i];

   return lowest;
}

unsigned yac_get_num_remote_domain_ranks(struct global_search * search) {

   return search->remote_domain_size;
}

unsigned const * yac_get_remote_domain_ranks(struct global_search * search) {

   return search->remote_domain_ranks;
}

char const * yac_get_local_domain_name(struct global_search * search) {

   return search->local_domain->name;
}

struct domain * yac_get_local_domain(struct global_search * search) {

   return search->local_domain;
}

char const * yac_get_remote_domain_name(struct global_search * search) {

   return search->remote_domain_name;
}

struct domain * yac_get_remote_domain(struct global_search * search,
                                      unsigned remote_domain_rank) {

   for (unsigned i = 0; i < search->num_remote_domains; ++i)
      if (search->remote_domains[i].rank == remote_domain_rank)
         return search->remote_domains[i].remote_domain;

   return NULL;
}

char const * yac_get_local_component_name(struct global_search * search) {

   return search->local_component_name;
}

char const * yac_get_remote_component_name(struct global_search * search) {

   return search->remote_component_name;
}

void yac_global_search_delete(struct global_search * search) {

   if (search == NULL)
      return;

   unsigned num_subdomains;

   num_subdomains = yac_get_num_subdomains(search->local_domain);

   unsigned i, j;

   free(search->remote_domain_name);
   free(search->local_component_name);
   free(search->remote_component_name);

   if (search->local_domain_is_target) {

      for (i = 0; i < search->num_overlapping_src_processes; ++i) {
         yac_free_dep_list(&(search->tgt_search_results[i].matching_local_cell_ids));
         free(search->tgt_search_results[i].global_tgt_ids);
      }
      free(search->tgt_search_results);
   }

   if (search->local_domain_is_source) {

      free(search->grid_search);

      for (i = 0; i < search->num_remote_domains; ++i) {
         yac_free_domain(search->remote_domains[i].remote_domain);
         free(search->remote_domains[i].remote_domain);
      }
      free(search->remote_domains);
   }

   free(search->overlapping_tgt_ranks);

   for (i = 0; i < search->num_src_search_results; ++i) {

      for (j = 0; j < num_subdomains; ++j)
         free(search->src_search_results[i].overlapping_src_cells[j]);
      free(search->src_search_results[i].overlapping_src_cells);
      free(search->src_search_results[i].num_overlapping_src_cells_per_subdomain);
      free(search->src_search_results[i].other_src_ranks);
   }
   free(search->src_search_results);

   free(search->local_domain_ranks);
   free(search->remote_domain_ranks);

   free(search);
}
