/**
 * @file interpolation_method.c
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>

#include "interpolation_method.h"
#include "communicator.h"
#include "global_search.h"
#include "ensure_array_size.h"
#include "utils.h"

static struct {

   struct global_search * search_pointer;
   struct interpolation_method * method_pointer;
   int id;
} * id_lookup = NULL;

static unsigned num_ids_in_lookup = 0;

struct global_ids_rank {

  int global_id;
  unsigned interpolated;
  unsigned rank;
};

struct temp_exchange_data {

   // data to continue the actual search
   void (*do_search_continue)(void * user_data, int * global_ids_to_remove,
                              unsigned num_global_ids);
   void * user_data;
   unsigned tag;
};

static int compare_int(const void * a, const void * b) {

  return ((*(int*)a) > (*(int*)b)) - (*((int*)a) < (*(int*)b));
}

int compare_global_ids_rank (const void * a, const void * b) {

  struct global_ids_rank const * a_ = a, * b_ = b;

  if (a_->global_id != b_->global_id) return a_->global_id - b_->global_id;
  if (a_->interpolated != b_->interpolated)
    return - a_->interpolated + b_->interpolated;
  return b_->rank - a_->rank;
}

int compare_global_ids_rank_2 (const void * a, const void * b) {

  struct global_ids_rank const * a_ = a, * b_ = b;

  if (a_->rank != b_->rank) return a_->rank - b_->rank;
  return a_->global_id - b_->global_id;
}

static void recv_global_ids_to_remove_callback(struct communicator * comm,
                                               struct comm_message msg,
                                               void * user_data) {

  struct temp_exchange_data * temp_data =
    (struct temp_exchange_data *)user_data;

  int * global_ids_to_remove = (int*)(msg.buffer);
  unsigned num_global_ids_to_remove = msg.count;

  // sort the list
  yac_mergesort(global_ids_to_remove, num_global_ids_to_remove,
                sizeof(*global_ids_to_remove), compare_int);

  // pass list of points to actual search of interpolation method
  temp_data->do_search_continue(temp_data->user_data, global_ids_to_remove,
                                num_global_ids_to_remove);

  free(temp_data);
}

static void recv_global_ids_callback(struct communicator * comm,
                                     struct comm_message msg,
                                     void * user_data) {

  struct temp_exchange_data * temp_data =
    (struct temp_exchange_data *)user_data;
  unsigned comm_size = yac_get_comm_size(comm);

  unsigned * num_interpolated_values_per_rank =
    malloc(comm_size * sizeof(*num_interpolated_values_per_rank));
  unsigned * num_values_per_rank =
    malloc(comm_size * sizeof(*num_values_per_rank));

  int * global_ids = NULL;
  unsigned global_ids_array_size = 0;
  unsigned num_global_ids = 0;

  if (msg.type != COMM_PACKED)
      yac_internal_abort_message("ERROR(recv_global_ids_callback): woops...",
                                  __FILE__, __LINE__);

  // unpack data
  int position = 0;

  for (unsigned i = 0; i < comm_size; ++i) {

    yac_comm_unpack(msg.buffer, msg.count, &position,
                    num_interpolated_values_per_rank + i, 1, COMM_UINT, comm);
    yac_comm_unpack(msg.buffer, msg.count, &position,
                    num_values_per_rank + i, 1, COMM_UINT, comm);

    ENSURE_ARRAY_SIZE(global_ids, global_ids_array_size,
                      num_global_ids + num_values_per_rank[i]);

    if (num_values_per_rank[i] > 0)
      yac_comm_unpack(msg.buffer, msg.count, &position,
                      global_ids + num_global_ids, num_values_per_rank[i],
                      COMM_INT, comm);

    num_global_ids += num_values_per_rank[i];
  }

  struct global_ids_rank * global_ids_rank =
    malloc(num_global_ids * sizeof(*global_ids_rank));

  for (unsigned i = 0, idx = 0; i < comm_size; ++i) {

    unsigned j;
    for (j = 0; j < num_interpolated_values_per_rank[i]; ++j, ++idx) {
      global_ids_rank[idx].global_id = global_ids[idx];
      global_ids_rank[idx].interpolated = 1;
      global_ids_rank[idx].rank = i;
    }
    for (; j < num_values_per_rank[i]; ++j, ++idx) {
      global_ids_rank[idx].global_id = global_ids[idx];
      global_ids_rank[idx].interpolated = 0;
      global_ids_rank[idx].rank = i;
    }
  }

  free(num_interpolated_values_per_rank);

  // sort data based on global_ids
  qsort(global_ids_rank, num_global_ids, sizeof(*global_ids_rank),
        compare_global_ids_rank);


  // find first value that can be interpolated
  unsigned i;
  for (i = 0; i < num_global_ids; ++i)
    if (global_ids_rank[i].interpolated) break;

  unsigned num_global_ids_to_remove = 0;
  int last_interpolated_global_id;

  if (i < num_global_ids)
    last_interpolated_global_id = global_ids_rank[i].global_id;

  for (++i; i < num_global_ids; ++i) {
    if (last_interpolated_global_id == global_ids_rank[i].global_id) {

      global_ids_rank[num_global_ids_to_remove] = global_ids_rank[i];
      num_global_ids_to_remove++;
    } else if (global_ids_rank[i].interpolated)
      last_interpolated_global_id = global_ids_rank[i].global_id;
  }

  // sort by ranks
  qsort(global_ids_rank, num_global_ids_to_remove, sizeof(*global_ids_rank),
        compare_global_ids_rank_2);

  // pack data for returning global ids to remove
  memset(num_values_per_rank, 0, comm_size * sizeof(*num_values_per_rank));
  for (unsigned i = 0; i < num_global_ids_to_remove; ++i) {
    global_ids[i] = global_ids_rank[i].global_id;
    num_values_per_rank[global_ids_rank[i].rank]++;
  }

  free(global_ids_rank);

  // distribute global ids to be removed
  yac_comm_alltoallv_callback(global_ids, num_values_per_rank, COMM_INT,
                              temp_data->tag, comm, temp_data,
                              recv_global_ids_to_remove_callback,
                              MSG_PRIO_HIGH);

  free(global_ids);
  free(num_values_per_rank);
}

void
yac_exchange_results(struct interpolation_method * method,
                     int * global_ids, unsigned num_interpolated_values,
                     unsigned num_values, unsigned tags[2],
                     struct communicator * comm,
                     void (*do_search_continue)(void * user_data,
                                                int * global_ids_to_remove,
                                                unsigned num_global_ids),
                     void * user_data) {

   unsigned comm_size = yac_get_comm_size(comm);

   if (comm_size == 1) {
      do_search_continue(user_data, NULL, 0);
      return;
   }

   // sort global ids
   int * sorted_global_ids = malloc(num_values * sizeof(*sorted_global_ids));
   {
     memcpy(sorted_global_ids, global_ids, num_values * sizeof(*global_ids));
     yac_mergesort(sorted_global_ids, num_interpolated_values,
                   sizeof(*sorted_global_ids), compare_int);
     yac_mergesort(sorted_global_ids + num_interpolated_values,
                   num_values - num_interpolated_values,
                   sizeof(*sorted_global_ids), compare_int);
   }

   // compute number of global ids per rank responsible rank
   unsigned * num_values_per_rank =
      calloc(2 * comm_size, sizeof(*num_values_per_rank));
   unsigned * num_interpolated_values_per_rank =
      num_values_per_rank + comm_size;
   {
      unsigned i;
      for (i = 0; i < num_interpolated_values; ++i) {
        unsigned rank = (((unsigned*)sorted_global_ids)[i] / 128) % comm_size;
        num_values_per_rank[rank]++;
        num_interpolated_values_per_rank[rank]++;
      }
      for (; i < num_values; ++i) {
        unsigned rank = (((unsigned*)sorted_global_ids)[i] / 128) % comm_size;
        num_values_per_rank[rank]++;
      }
   }

   // pack sorted_global_ids into send buffers
   int send_buffer_size = yac_comm_pack_size(2 * comm_size, COMM_UINT, comm) + 
                          yac_comm_pack_size(num_values, COMM_INT, comm); 
   int * send_buffer_sizes = malloc(comm_size * sizeof(*send_buffer_sizes));
   for (unsigned i = 0; i < comm_size; ++i)
     send_buffer_sizes[i] = yac_comm_pack_size(2, COMM_UINT, comm) + 
                            yac_comm_pack_size(num_values_per_rank[i],
                                               COMM_INT, comm);
   char * send_buffer = malloc(send_buffer_size);
   {
      unsigned * prefix = malloc(comm_size * sizeof(*prefix));
      unsigned accum = 0;
      for (unsigned i = 0; i < comm_size; ++i) {
        prefix[i] = accum;
        accum += send_buffer_sizes[i];
      }
      int * positions = calloc(comm_size, sizeof(*positions));
      for (unsigned i = 0; i < comm_size; ++i) {
        yac_comm_pack(num_interpolated_values_per_rank + i, 1, COMM_UINT,
                      send_buffer + prefix[i], send_buffer_sizes[i],
                      positions + i, comm);
        yac_comm_pack(num_values_per_rank + i, 1, COMM_UINT,
                      send_buffer + prefix[i], send_buffer_sizes[i],
                      positions + i, comm);
      }

      for (unsigned i = 0; i < num_values; ++i) {
        unsigned rank = (((unsigned*)sorted_global_ids)[i] / 128) % comm_size;
        yac_comm_pack(sorted_global_ids + i, 1, COMM_INT,
                      send_buffer + prefix[rank], send_buffer_sizes[rank],
                      positions + rank, comm);
      }
      free(positions);
      free(prefix);
   }

   free(num_values_per_rank);
   free(sorted_global_ids);

   struct temp_exchange_data * temp_data = malloc(1 * sizeof(*temp_data));

   temp_data->user_data = user_data;
   temp_data->do_search_continue = do_search_continue;
   temp_data->tag = tags[1];

   yac_comm_alltoallv_callback(send_buffer, (unsigned*)send_buffer_sizes,
                               COMM_PACKED, tags[0], comm, temp_data,
                               recv_global_ids_callback, MSG_PRIO_HIGH);

  free(send_buffer_sizes);
  free(send_buffer);
}

void yac_reset_interpolation_method(void) {

   free(id_lookup);
   id_lookup = NULL;
   num_ids_in_lookup = 0;
}

void yac_do_search_interpolation_method(struct interpolation_method ** method,
                                        struct interpolation_plan_element ** plan_stack,
                                        int * global_ids,
                                        double * x_coordinates,
                                        double * y_coordinates,
                                        enum yac_location tgt_location,
                                        unsigned num_values,
                                        unsigned * num_interpolated_values,
                                        unsigned num_pointsets) {

   struct interpolation_method_data_request ** data_request;

   unsigned num_interpolation_methods = 0;

   while (method[num_interpolation_methods] != NULL) ++num_interpolation_methods;

   data_request = calloc(num_interpolation_methods, sizeof(*data_request));

   // execute method stack
   method[0]->vtable->do_search( method, plan_stack, global_ids, x_coordinates,
                                 y_coordinates, tgt_location, num_values,
                                 num_interpolated_values, data_request);
}

void yac_enddef_interpolation_method(struct interpolation_method ** method) {

   method[0]->vtable->enddef(method);
}

void yac_get_source_exchange_data_interpolation_method(
   struct interpolation_method ** method,
   struct interpolation_method_data_request ** data_request,
   unsigned num_pointsets) {

   struct interpolation_method_data_request ** method_data_request;

   method[0]->vtable->get_source_exchange_data(method, &method_data_request);

   unsigned i, j;

   for (i = 0; i < num_pointsets; ++i) {
      (*data_request)[i].num_values = 0;
      (*data_request)[i].global_ids = NULL;
      (*data_request)[i].ranks = NULL;
   }

   for (i = 0; method[i] != NULL; ++i) {
      if (method_data_request[i] != NULL) {
         for (j = 0; j < num_pointsets; ++j) {
            if (method_data_request[i][j].num_values){

               (*data_request)[j].num_values += method_data_request[i][j].num_values;
               (*data_request)[j].global_ids = realloc((*data_request)[j].global_ids,
                                                       (*data_request)[j].num_values *
                                                       sizeof(*((*data_request)[j].global_ids)));
               (*data_request)[j].ranks = realloc((*data_request)[j].ranks,
                                                  (*data_request)[j].num_values *
                                                  sizeof(*((*data_request)[j].ranks)));

               memcpy((*data_request)[j].global_ids + (*data_request)[j].num_values -
                      method_data_request[i][j].num_values,
                      method_data_request[i][j].global_ids,
                      method_data_request[i][j].num_values *
                      sizeof(*((*data_request)[j].global_ids)));
               memcpy((*data_request)[j].ranks + (*data_request)[j].num_values -
                      method_data_request[i][j].num_values,
                      method_data_request[i][j].ranks,
                      method_data_request[i][j].num_values *
                      sizeof(*((*data_request)[j].ranks)));
            }
         }
         yac_free_interpolation_method_data_request(method_data_request[i],
                                                num_pointsets);
      }
   }
   free(method_data_request);
}

void yac_set_finish_callback_interpolation_method(struct interpolation_method ** method,
                                                  void (*callback)(void * user_data),
                                                  void * user_data) {

  if ((method == NULL) || (method[0] == NULL))
    yac_internal_abort_message(
      "ERROR(yac_set_finish_callback_interpolation_method): no interpolation "
      "method defined", __FILE__, __LINE__);
  while (method[1] != NULL) ++method;

  method[0]->vtable->set_finish_callback(method[0], callback, user_data);
}

void yac_init_interpolation_method(struct interpolation_method * method,
                               struct global_search * search) {

   int id = 0;
   int i;

   if (method == NULL || search == NULL)
      yac_internal_abort_message("ERROR(init_interpolation_method): invalid input data",
                                  __FILE__, __LINE__);

   for (i = 0; i < num_ids_in_lookup; ++i) {
      if (search == id_lookup[i].search_pointer) ++id;
      if (method == id_lookup[i].method_pointer)
         yac_internal_abort_message("ERROR(init_interpolation_method): method was already initialised",
                                    __FILE__, __LINE__);
   }

   id_lookup = realloc(id_lookup, (num_ids_in_lookup+1) * sizeof(*id_lookup));

   id_lookup[num_ids_in_lookup].search_pointer = search;
   id_lookup[num_ids_in_lookup].method_pointer = method;
   id_lookup[num_ids_in_lookup].id = id;
   num_ids_in_lookup++;
}

int yac_get_unique_id_interpolation_method(struct interpolation_method * method) {

   int i;

   if (method == NULL)
      yac_internal_abort_message("ERROR(get_unique_id_interpolation_method): invalid input data",
                                 __FILE__, __LINE__);

   for (i = 0; i < num_ids_in_lookup; ++i)
      if (id_lookup[i].method_pointer == method)
         return id_lookup[i].id;

   yac_internal_abort_message("ERROR(get_unique_id_interpolation_method): method was not initialised",
                              __FILE__, __LINE__);
   return -1;
}

void
yac_free_interpolation_method_data_request(struct interpolation_method_data_request * request,
                                           unsigned num_pointsets) {

   if (request != NULL) {

      unsigned i;

      for (i = 0; i < num_pointsets; ++i) {
         free(request[i].global_ids);
         free(request[i].ranks);
      }

      free(request);
   }
}

void yac_delete_interpolation_method(struct interpolation_method * method) {

   if (method == NULL)
      yac_internal_abort_message("ERROR(delete_interpolation_method): invalid input data",
                                 __FILE__, __LINE__);

   for (int i = 0; i < num_ids_in_lookup; ++i) {
      if (id_lookup[i].method_pointer == method) {
         id_lookup[i].search_pointer = NULL;
         id_lookup[i].method_pointer = NULL;
      }
   }

   method->vtable->delete(method);
}

void yac_remove_points(int * global_ids, unsigned * num_global_ids,
                       int * global_ids_to_remove,
                       unsigned num_global_ids_to_remove,
                       int ** remaining_points) {

  // sort global ids
  yac_quicksort_index(global_ids_to_remove, num_global_ids_to_remove, NULL);

  unsigned old_num_global_ids = *num_global_ids;

  int * permutation = malloc(old_num_global_ids * sizeof(*permutation));
  for (unsigned i = 0; i < old_num_global_ids; ++i) permutation[i] = i;
  yac_quicksort_index(global_ids, old_num_global_ids, permutation);

  unsigned new_num_global_ids = old_num_global_ids;
  for (unsigned i = 0, j = 0; i < old_num_global_ids; ++i) {
    while ((j < (num_global_ids_to_remove - 1)) &&
           (global_ids_to_remove[j] < global_ids[i])) ++j;

    if (global_ids_to_remove[j] == global_ids[i]) {
      permutation[i] = INT_MAX;
      new_num_global_ids--;
    }
  }

  yac_quicksort_index(permutation, old_num_global_ids, global_ids);

  *remaining_points = permutation;
  *num_global_ids = new_num_global_ids;
}
