/**
 * @file interpolation_method.h
 * @brief methods for using interpolation methods
 *
 * @remarks if interpolation methods are to be used across multiple processes,
 *          the respective methods have to be built in the same order on all
 *          processes.
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#ifndef INTERPOLATION_METHOD_H
#define INTERPOLATION_METHOD_H

/** \example test_interpolation_method.c
 * This shows how to use the interpolation method routines
 */

/** \example test_interpolation_method_parallel.c
 * This contains a parallel example for the usage of interpolation method. The
 * example uses two source and two target processes.
 */

#include "fields.h"
#include "global_search.h"
#include "dep_list.h"
#include "interpolation_plan.h"

struct interpolation_method;

struct interpolation_method_data_request {

   unsigned num_values; //!< number of required values (on points or cells)
   int * global_ids;    //!< global ids of required points
   int * ranks;         //!< rank of process which has the respective points
};

struct interpolation_method_vtable {

   void (*do_search)(struct interpolation_method ** method,
                     struct interpolation_plan_element ** plan_stack,
                     int * global_ids,
                     double * x_coordinates,
                     double * y_coordinates,
                     enum yac_location tgt_location,
                     unsigned num_points,
                     unsigned * num_interpolated_points,
                     struct interpolation_method_data_request ** data_request);

   void (*enddef)(struct interpolation_method ** method);

   void (*get_source_exchange_data)(struct interpolation_method ** method,
                                    struct interpolation_method_data_request *** data_request);

   // recv_buffer points to data requested in recv_list
   // after the interpolation the method will call the next method on the stack
   // and set receive buffer to first entry after its own data


   // cleanup (close open requests and free memory)
   void (*delete)(struct interpolation_method * method);

   struct global_search * (*get_global_search)(struct interpolation_method * method);

   void (*set_finish_callback)(struct interpolation_method * method,
                               void (*callback)(void * user_data),
                               void * user_data);
};

struct interpolation_method {

   struct interpolation_method_vtable *vtable;
};

/**
 * frees memory associated to an interpolation_method_data_request object
 *
 * @param[in] request
 * @param[in] num_pointsets number of struct points available for
 *                                     each subdomain
 */
void
yac_free_interpolation_method_data_request(struct interpolation_method_data_request * request,
                                           unsigned num_pointsets);

/**
 * initialises the generation of an interpolation plan
 *
 * @param[in]     method                   Contains an array of interpolation_method pointers.
 *                                         These methods are used to interpolate the given
 *                                         points. For each point only one method is used. If
 *                                         a method fails for a number of points, the next
 *                                         method in the array is applied.
 * @param[out]    plan_stack               plan stack that does the actual interpolation
 * @param[in,out] global_ids               Global ids of the points to be interpolated. Points
 *                                         that are not interpolated (because they are by
 *                                         another process or because they cannot be interpolated)
 *                                         are moved to the end of this array.
 * @param[in,out] x_coordinates            x coordinates of the points to be interpolated.
 * @param[in,out] y_coordinates            y coordinates of the points to be interpolated.
 * @param[in]     tgt_location             location(CELL, CORNER, EDGE) of the target points
 * @param[in]     num_points               number of points in global_ids, x_coordinates and
 *                                         y_coordinates.
 * @param[out]    num_interpolated_points  Each method returns in this array the number of
 *                                         points that it interpolates.
 * @param[in]     num_pointsets number of struct points available for
 *                                         each subdomain
 *
 * \remarks This routine only initialises the generation of the interpolation plan. It
 *          needs to be finalised by a call to \ref yac_enddef_interpolation_method.
 * \remarks It is valid to pass \c NULL as arguments for x_coordinates and y_coordinates,
 *          some interpolations might not be able to handle this case and just all points
 *          to the next interpolation on the method stack.
 */
void yac_do_search_interpolation_method(struct interpolation_method ** method,
                                        struct interpolation_plan_element ** plan_stack,
                                        int * global_ids, double * x_coordinates,
                                        double * y_coordinates,
                                        enum yac_location tgt_location,
                                        unsigned num_points,
                                        unsigned * num_interpolated_points,
                                        unsigned num_pointsets);

/**
 * waits until the interpolation plan is successfully generated
 *
 * @param[in] method interpolation_method array that was used in a call to
 *                   \ref yac_do_search_interpolation_method.
 */

void yac_enddef_interpolation_method(struct interpolation_method ** method);

/**
 * returns an interpolation_method_data_request object containing data points
 * required from other source processes for the interpolation
 *
 * @param[in]  method                   interpolation_method array that was used
 *                                      in a call to
 *                                      \ref yac_do_search_interpolation_method.
 * @param[out] data_request             contains points required from other
 *                                      source processes, which are required for
 *                                      the interpolation
 * @param[in]  num_pointsets number of struct points available for
 *                                      each subdomain
 *
 * \remarks it needs to be ensured that an interpolation is already generated, this
 *          is done be a call to \ref yac_enddef_interpolation_method, otherwise the behaviour
 *          is undefined
 * \remarks it can only be called once for each method stack, otherwise the behaviour
 *          is undefined
 */
void yac_get_source_exchange_data_interpolation_method(
   struct interpolation_method ** method,
   struct interpolation_method_data_request ** data_request,
   unsigned num_pointsets);

/**
 * registers a callback routine that is called once the lowest element on the method
 * stack is finished with its work
 *
 * @param[in] method    method stack
 * @param[in] callback  callback routine pointer
 * @param[in] user_data pointer that is passed to callback
 */
void yac_set_finish_callback_interpolation_method(struct interpolation_method ** method,
                                                  void (*callback)(void * user_data),
                                                  void * user_data);

/**
 * Frees all memory associated to an interpolation method.
 * @param[in] method interpolation method
 **/
void yac_delete_interpolation_method(struct interpolation_method * method);

/**
 * Every interpolation method constructor needs to call this routine. This
 * ensures that calls to \ref yac_exchange_results from different
 * interpolation methods do not interfere with each other.
 *
 * @param[in] method interpolation method
 * @param[in] search global search object used in the interpolation method
 */
void yac_init_interpolation_method(struct interpolation_method * method,
                                   struct global_search * search);

/**
 * Returns a id that is unique for the global search that is used in the
 * interpolation method and identical among all processes of the global search.
 *
 * @param[in] method interpolation method
 *
 * @remarks the method must have called \ref yac_init_interpolation_method in order
 *          for this routine not to generate an error
 */
 int yac_get_unique_id_interpolation_method(struct interpolation_method * method);
 
/**
 * Exchanges search results of an interpolation method between the source
 * processes and matches the search results of the local process with the
 * remote once. It then calls a callback routine, which allows the
 * interpolation method to adjust its results in order to avoid target points
 * to be handled by multiple source processes.
 *
 * @param[in]     method                  interpolation method
 * @param[in,out] global_ids              array containing the global ids of
 *                                        the target points
 * @param[in,out] num_interpolated_points number of interpolated points
 * @param[in,out] num_points              total number of target points
 * @param[in]     tags                    two different tags that are unique for
 *                                        this operation but identical on all
 *                                         participating processes
 * @param[in]     comm                    communicator that contains all source processes
 * @param[in]     do_search_continue      callback routine that is called once
 *                                        the exchange of results is finished,
 *                                        it passes a list of global ids, that need
 *                                        to be removed from the local global_ids,
 *                                        x_coordinates and y_coordinates arrays,
 *                                        because these respective target points are
 *                                        handled by another source process
 * @param[in]     user_data               pointer that is passed to the callback
 *                                        routine
 * @remark in order to ensure that the operation is finished, the user has to
 *         call \ref yac_comm_wait for the provided tags\n
 *         while(yac_comm_wait(comm, tags[0]) || yac_comm_wait(comm, tags[1]));
 */
void yac_exchange_results(struct interpolation_method * method,
                          int * global_ids,
                          unsigned num_interpolated_points,
                          unsigned num_points,
                          unsigned tags[2],
                          struct communicator * comm,
                          void (*do_search_continue)(
                             void * user_data,
                             int * global_ids_to_remove,
                             unsigned num_global_ids),
                          void * user_data);

/**
 * utility routine for interpolation methods
 * moves entries from global_ids that are in global_ids_to_remove to the end of
 * global_ids while remaining the original order
 *
 * @param[in,out] global_ids               array containing global indices
 * @param[in,out] num_global_ids           number of entries in global_ids, is
 *                                         changed by this routine to the number
 *                                         of remaining global_ids
 * @param[in]     global_ids_to_remove     global ids that need to be removed
 *                                         from global_ids
 * @param[in]     num_global_ids_to_remove number of entries in
 *                                         global_ids_to_remove
 * @param[out]    remaining_points         positions all all points that remain
 *                                         in global_ids
 *
 * @remark the user is responsible for freeing remaining_points
 */
void yac_remove_points(int * global_ids, unsigned * num_global_ids,
                       int * global_ids_to_remove,
                       unsigned num_global_ids_to_remove,
                       int ** remaining_points);

/**
 * This routine is only needed by some tests that need to reset the internal
 * state of interpolation method at the end of each test.
 * In addition it is used by \ref yac_finalize to free the memory stored
 * statically in interpolation_method.c
 */
void yac_reset_interpolation_method(void);

#endif //INTERPOLATION_METHOD_H
