/**
 * @file interpolation_method_avg.c
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "interpolation_method.h"
#include "interpolation_method_avg.h"
#include "interpolation_plan.h"
#include "grid_search.h"
#include "global_search.h"
#include "ensure_array_size.h"
#include "utils.h"

enum masked {
   NOT_MASKED     = 0, // associated source points are not masked
   PARTIAL_MASKED = 1, // associated source points are partially masked
   MASKED         = 2, // all associated source points are masked
};

static void
do_search_interpolation_method_avg(struct interpolation_method ** method,
                                   struct interpolation_plan_element ** plan_stack,
                                   int * global_ids, double * x_coordinates,
                                   double * y_coordinates,
                                   enum yac_location tgt_location,
                                   unsigned num_points,
                                   unsigned * num_interpolated_points,
                                   struct interpolation_method_data_request ** data_request);
static void
enddef_interpolation_method_avg(struct interpolation_method ** method);
static void
get_source_exchange_data_interpolation_method_avg(
   struct interpolation_method ** method,
   struct interpolation_method_data_request *** data_request);
static void
delete_interpolation_method_avg(struct interpolation_method * method);
static struct global_search *
get_global_search(struct interpolation_method * method);
static void
set_finish_callback(struct interpolation_method * method,
                    void (*callback)(void * user_data),
                    void * user_data);

static struct interpolation_method_vtable
   interpolation_method_avg_vtable = {do_search_interpolation_method_avg,
                                      enddef_interpolation_method_avg,
                                      get_source_exchange_data_interpolation_method_avg,
                                      delete_interpolation_method_avg,
                                      get_global_search,
                                      set_finish_callback};

struct interpolation_method_avg_tags {

   unsigned global_id_exchange[2];
};

struct interpolation_method_avg {

   struct interpolation_method_vtable * vtable;
   void (*finish_callback)(void * user_data);   //!< callback routine that is called once the interpolation is finished
   void * finish_user_data;                     //!< user data that is passed to finish_callback
   struct global_search * search;               //!< global search object used a the basis for the interpolation
   struct domain * domain;
   struct points *** points;
   struct mask *** masks;
   unsigned num_pointsets;
   unsigned distance_weighted;                  //!< flag to switch on and off weighting by distance
   unsigned points_idx;                         //!< index of the points set of field that is to be used for the interpolation
   struct interpolation_method_data_request ** data_request;
   unsigned use_source_mask;
   unsigned partial_coverage;
   struct interpolation_method_avg_tags tags;
};

static void generate_tag_string(char * buf, size_t buffer_size,
                                char const * tag_base_str,
                                char const * source_comp_name,
                                char const * target_comp_name,
                                char const * source_domain_name,
                                char const * target_domain_name,
                                char const * stack_name,
                                int points_idx,
                                unsigned distance_weighted,
                                int unique_interpolation_method_id,
                                char const * tag_str) {

   int ret_val;
   ret_val = snprintf(buf, buffer_size, "%s%s%s%s%s%s%d%d%d%s", tag_base_str,
                      source_comp_name, target_comp_name, source_domain_name,
                      target_domain_name, stack_name, points_idx,
                      distance_weighted, unique_interpolation_method_id,
                      tag_str);

   if (ret_val < 0 || ret_val >= buffer_size)
      yac_internal_abort_message("ERROR(generate_tag_string): truncation",
                                 __FILE__, __LINE__);
}

static void
generate_interpolation_method_avg_tag(struct interpolation_method_avg * method,
                                      char const * stack_name) {

   char const * str_base_str             = "interpolation method avg";
   char const * str_generator_string     = "generator string";
   char const * str_global_id_exchange_a = "global id exchange a";
   char const * str_global_id_exchange_b = "global id exchange b";

   struct global_search * search;

   search = get_global_search((struct interpolation_method *)method);

   struct communicator * comm;

   comm = yac_get_domain_comm(yac_get_local_domain(search));

   char const * source_comp_name = yac_get_local_component_name(search);
   char const * target_comp_name = yac_get_remote_component_name(search);

   char const * source_domain_name = yac_get_local_domain_name(search);
   char const * target_domain_name = yac_get_remote_domain_name(search);

   unsigned max_tag_strlen;

   max_tag_strlen = strlen(str_global_id_exchange_a);

   unsigned max_str_buf_size;

   max_str_buf_size = max_tag_strlen + strlen(str_base_str) +
                      strlen(source_comp_name) +
                      strlen(target_comp_name) +
                      strlen(source_domain_name) +
                      strlen(target_domain_name) +
                      strlen(stack_name) + 10 + 10 + 1; // (+ 10) for points_idx
                                                        // (+ 10) for distance_weighted
                                                        // (+ 10) for unique_method_id
                                                        // (+ 1)  for '\0'

   char * unique_generator_string = malloc(max_str_buf_size);
   generate_tag_string(unique_generator_string, max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name, source_domain_name, target_domain_name,
                       stack_name, method->points_idx, method->distance_weighted,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                       str_generator_string);

   char * tag_strings[2];
   tag_strings[0] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[0], max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name, source_domain_name, target_domain_name,
                       stack_name, method->points_idx, method->distance_weighted,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                       str_global_id_exchange_a);

   tag_strings[1] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[1], max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name, source_domain_name, target_domain_name,
                       stack_name, method->points_idx, method->distance_weighted,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                       str_global_id_exchange_b);

   unsigned tags[2];

   yac_get_unique_comm_tags(
      comm, unique_generator_string, (char const **)tag_strings, 2, tags);

   free(unique_generator_string);

   method->tags.global_id_exchange[0] = tags[0];
   method->tags.global_id_exchange[1] = tags[1];

   for (unsigned i = 0; i < 2; ++i) free(tag_strings[i]);
}

struct interpolation_method *
yac_new_interpolation_method_avg_mask(struct global_search * search,
                                      struct domain * domain,
                                      char const * stack_name,
                                      struct points *** points,
                                      struct mask *** masks,
                                      unsigned num_pointsets,
                                      unsigned points_idx,
                                      unsigned distance_weighted,
                                      unsigned use_source_mask,
                                      unsigned partial_coverage) {

   struct interpolation_method_avg * method;

   method = malloc(1 * sizeof(*method));

   method->vtable = &interpolation_method_avg_vtable;
   method->search = search;
   method->finish_callback = NULL;
   method->finish_user_data = NULL;
   method->use_source_mask = use_source_mask;
   method->partial_coverage = partial_coverage;
   method->points_idx = points_idx;
   method->distance_weighted = distance_weighted;
   method->domain = domain;
   method->points = points;
   method->masks = masks;
   method->num_pointsets = num_pointsets;

   yac_init_interpolation_method((struct interpolation_method *)method, search);
   generate_interpolation_method_avg_tag(method, stack_name);

   // ensure that the right point data is available

   if (num_pointsets == 0)
      yac_internal_abort_message ("ERROR(new_interpolation_method_avg_mask): missing "
                                  "data points.", __FILE__, __LINE__);

   if (points_idx >= num_pointsets)
      yac_internal_abort_message ("ERROR(new_interpolation_method_avg_mask): invalid "
                                  "points_idx.", __FILE__, __LINE__);

   if (points == NULL || points[0] == NULL || points[0][points_idx] == NULL)
      yac_internal_abort_message ("ERROR(new_interpolation_method_avg_mask): no point set "
                                  "data available", __FILE__, __LINE__);
   if (points[0][points_idx]->location != CORNER)
      yac_internal_abort_message ("ERROR(new_interpolation_method_avg_mask): data points "
                                  "have wrong location type (needs to be CORNER).",
                                  __FILE__, __LINE__);

   return (struct interpolation_method *)method;
}

struct interpolation_method *
yac_new_interpolation_method_avg(struct global_search * search,
                                 struct domain * domain,
                                 char const * stack_name,
                                 struct points *** points,
                                 struct mask *** masks,
                                 unsigned num_pointsets,
                                 unsigned points_idx,
                                 unsigned distance_weighted,
                                 unsigned partial_coverage) {

   // ignore source mask is the current default,
   // test_field_interpolation_parallel2 relies on this
   unsigned use_source_mask = 0;

   return yac_new_interpolation_method_avg_mask(search, domain, stack_name, points,
                                                masks, num_pointsets, points_idx,
                                                distance_weighted,
                                                use_source_mask, partial_coverage);
}

static void
insertion_sort_int(int element, int * a, unsigned * curr_length) {

   unsigned i, j;

   if (*curr_length > 0) {

      for (i = 0; i < *curr_length; ++i)
         if (a[i] >= element) break;

      // if the element is already in the array
      if ((i != *curr_length) && (a[i] == element)) return;

      // copy new element into array and move bigger elements one position up
      for (j = *curr_length; j > i; --j) {

         a[j] = a[j-1];
      }

      a[i] = element;

      // increase array length indicator
      ++(*curr_length);

   } else {

      a[0] = element;
      *curr_length = 1;
   }
}

/** reorders the arrays global_ids, x_coordinates and y_coordinates such
  * that the points specified in the put_to_front array are at the
  * beginning of the three arrays
  */
static void
reorder_points(int * global_ids, double * x_coordinates, double * y_coordinates,
               int * put_to_front, unsigned num_points, unsigned num_points_to_front) {

   unsigned i, j;

   for (i = 0; i < num_points_to_front; ++i) {

      if (!put_to_front[i]) {

         // find a point to replace the current one
         for (j = i + 1;!put_to_front[j];++j);

         double temp_x, temp_y;
         int temp_id, temp_put_to_front;

         if (x_coordinates != NULL)
            temp_x = x_coordinates[i], x_coordinates[i] = x_coordinates[j], x_coordinates[j] = temp_x;
         if (y_coordinates != NULL)
            temp_y = y_coordinates[i], y_coordinates[i] = y_coordinates[j], y_coordinates[j] = temp_y;
         temp_id = global_ids[i], global_ids[i] = global_ids[j], global_ids[j] = temp_id;
         temp_put_to_front = put_to_front[i], put_to_front[i] = put_to_front[j], put_to_front[j] = temp_put_to_front;
      }
   }
}

/**
 * Data structure that is used to pass data from the routine
 * do_search_interpolation_method_avg to do_search_continue_avg.
 * (do_search_continue_avg is the callback routine passed to
 *  exchange_results_init)
 */
struct search_data_avg {

   struct interpolation_method ** method_stack;
   struct interpolation_plan_element ** plan_stack;

   int * global_ids;                     //!< global ids of target points that are to be
                                         //!< interpolated by the local process
   double * x_coordinates;               //!< x coordinates of the target points
   double * y_coordinates;               //!< y coordinates of the target points
   enum yac_location tgt_location;       //!< location of the target points
   unsigned num_points;                  //!< number of target points
   unsigned * num_interpolated_points;   //!< number of points that are interpolated by
                                         //!< the avg interpolation method (this is
                                         //!< determined in \ref do_search_continue_avg 
                                         //!< ( \ref remove_points ))

   struct dep_list * compact_tgt_to_src; //!< dependency list containing the source to
                                         //!< target point relations
   unsigned * tgt_to_subdomain;          //!< mapping of the target points to local
                                         //!< subdomain, which is able to interpolate
                                         //!< the point
   enum masked * masked;                 //!< are the associated source points for each
                                         //!< target points masked
};

/**
 * Removes points from the list of target points (even if they could
 * be interpolated by the local process). This because they are going
 * to be interpolated by another source process, which might later
 * request some data from the local process to do the interpolation.
 * @param[in] num_points_to_be_removed number of points that are interpolated
 *                                     by another source process
 * @param[in] global_ids_to_be_removed array with global ids of target points
 *                                     that are to be removed from the local list
 * @param[in] data                     basic search data (see \ref search_data_avg)
 */
static void
remove_points (unsigned num_points_to_be_removed, int * global_ids_to_be_removed,
               struct search_data_avg * data) {

   if (num_points_to_be_removed > 0) {

      unsigned new_num_points = data->num_points;
      int * remaining_points;
      yac_remove_points(data->global_ids, &new_num_points,
                        global_ids_to_be_removed, num_points_to_be_removed,
                        &remaining_points);

      unsigned new_num_interpolated_points = 0;
      unsigned * new_compact_tgt_to_src_deps =
        malloc(
          yac_get_total_num_dependencies(*(data->compact_tgt_to_src)) *
                                         sizeof(*new_compact_tgt_to_src_deps));
      unsigned new_num_tgt_to_src_deps = 0;
      unsigned * new_num_src_per_tgt =
        malloc(*(data->num_interpolated_points) * sizeof(*new_num_src_per_tgt));

      for (unsigned i = 0; i < new_num_points; ++i) {

        int idx = remaining_points[i];

        if (i != idx) {

           data->x_coordinates[i] = data->x_coordinates[idx];
           data->y_coordinates[i] = data->y_coordinates[idx];

           // only for points that can be interpolated the target to subdomain
           // information is available
           if (idx < *(data->num_interpolated_points))
              data->tgt_to_subdomain[i] = data->tgt_to_subdomain[idx];
        }

        // copy the target to source point dependencies for the current target
        // point
        if (idx < *(data->num_interpolated_points)) {
           memcpy(new_compact_tgt_to_src_deps + new_num_tgt_to_src_deps,
                  yac_get_dependencies_of_element(*(data->compact_tgt_to_src),
                                                  idx),
                  data->compact_tgt_to_src->num_deps_per_element[idx] *
                  sizeof(*new_compact_tgt_to_src_deps));
           new_num_tgt_to_src_deps +=
              data->compact_tgt_to_src->num_deps_per_element[idx];
           new_num_src_per_tgt[i] =
              data->compact_tgt_to_src->num_deps_per_element[idx];
           new_num_interpolated_points++;
        }
      }

      free(remaining_points);

      // write back the data that was reduced by the points that are interpolated by
      // another source process
      *(data->num_interpolated_points) = new_num_interpolated_points;
      data->num_points = new_num_points;
      data->tgt_to_subdomain = realloc(data->tgt_to_subdomain,
                                       new_num_interpolated_points *
                                       sizeof(*(data->tgt_to_subdomain)));
      yac_free_dep_list(data->compact_tgt_to_src);
      new_num_src_per_tgt = realloc(new_num_src_per_tgt, new_num_interpolated_points *
                                    sizeof(*new_num_src_per_tgt));
      new_compact_tgt_to_src_deps = realloc(new_compact_tgt_to_src_deps,
                                            new_num_tgt_to_src_deps *
                                            sizeof(*new_compact_tgt_to_src_deps));
      yac_set_dependencies(data->compact_tgt_to_src, new_num_interpolated_points,
                           new_num_src_per_tgt, new_compact_tgt_to_src_deps);
   }
}

/**
 * Check for source points the are required for the interpolation, which located on
 * other source processes or on another local subdomain (which the target point is
 * not assigned to).
 * @param[out] remote_points              global ids of source points, required for
 *                                        the interpolation, that are located on
 *                                        remote source domains
 * @param[out] num_remote_points          number of points in remote_points
 * @param[out] other_subdomain_points     global ids of source points, required for the
 *                                        interpolation, that are located on a local
 *                                        subdomain which is not the one the respective
 *                                        target point is assigned to (the global ids
 *                                        are sorted)
 * @param[out] num_other_subdomain_points number of points in other_subdomain_points
 * @param[in]  data                       basic search data (see \ref search_data_avg)
 */
static void
check_for_remote_points(int ** remote_points, unsigned * num_remote_points,
                        int ** other_subdomain_points,
                        unsigned * num_other_subdomain_points,
                        struct search_data_avg * data) {

   struct interpolation_method_avg * method_avg;

   method_avg = (struct interpolation_method_avg*)(data->method_stack[0]);

   unsigned remote_points_size = 0;
   unsigned other_subdomain_points_size = 0;
   int rank, comm_rank;
   unsigned i, j;
   struct domain * domain = method_avg->domain;

   comm_rank = yac_get_comm_rank(yac_get_domain_comm(domain));

   for (i = 0; i < *(data->num_interpolated_points); ++i) {
   //check for remote points that are required for the interpolation

      unsigned const * src_points;
      unsigned num_src_points;
      int const * global_src_ids;

      src_points = yac_get_dependencies_of_element(*(data->compact_tgt_to_src), i);
      num_src_points = data->compact_tgt_to_src->num_deps_per_element[i];
      global_src_ids = yac_get_global_ids(domain, data->tgt_to_subdomain[i],
                                          CORNER);

      // store requests to other source processes if necessary
      for (j = 0; j < num_src_points; ++j) {

         if (yac_is_halo_value(domain, data->tgt_to_subdomain[i], CORNER,
                               src_points[j], (unsigned *)&rank)) {

            if (rank != comm_rank) {
               ENSURE_ARRAY_SIZE(*remote_points, remote_points_size,
                                 *num_remote_points+1);

               insertion_sort_int(global_src_ids[src_points[j]],
                                  *remote_points, num_remote_points);

            } else {

               ENSURE_ARRAY_SIZE(*other_subdomain_points, other_subdomain_points_size,
                                 *num_other_subdomain_points+1);
               insertion_sort_int(global_src_ids[src_points[j]],
                                  *other_subdomain_points, num_other_subdomain_points);
            }
         }
      }
   }
}

/**
 * Gets the subdomain indices and the local indices of source points that are
 * located on local subdomains, which are not the subdomains the target points
 * are assigned to.
 * @param[in]  other_subdomain_points           global ids of source points, required for
 *                                              the interpolation, that are located on a
 *                                              local subdomain which is not the one the
 *                                              respective target point is assigned to
 * @param[in]  num_other_subdomain_points       number of points in other_subdomain_points
 * @param[out] other_subdomain_indices          subdomain indices for the points listed in
 *                                              other_subdomain_points
 * @param[out] other_subdomain_points_local_ids local ids of the points listed in
 *                                              other_subdomain_points
 * @param[in]  domain                           the local domain
 */
static void
get_subdomain_indices(int * other_subdomain_points, unsigned num_other_subdomain_points,
                      unsigned * other_subdomain_indices,
                      unsigned * other_subdomain_points_local_ids,
                      struct domain * domain) {

   unsigned i, j, k;

   if (num_other_subdomain_points != 0) {

      unsigned num_subdomains;
      num_subdomains = yac_get_num_subdomains(domain);

      unsigned * matching_ids_idx;
      unsigned num_matching_ids;
      unsigned * temp_local_ids;
      unsigned num_open_other_subdomain_points;

      num_open_other_subdomain_points = num_other_subdomain_points;

      matching_ids_idx = malloc(num_other_subdomain_points * sizeof(*matching_ids_idx));
      temp_local_ids = malloc(num_other_subdomain_points * sizeof(*temp_local_ids));

      // set subdomain index to not-found
      for (i = 0; i < num_other_subdomain_points; ++i)
         other_subdomain_indices[i] = -1;

      for (i = 0; i < num_subdomains; ++i) {

         // search in the current subdomains for the points in the open list
         // (it is ensured that other_subdomain_points (computed by
         //  check_for_remote_points) is sorted)
         yac_match_global_ids(domain, i, CORNER, other_subdomain_points,
                              num_open_other_subdomain_points, NULL, matching_ids_idx,
                              &num_matching_ids, temp_local_ids, 0);

         // set the subdomain index and the local index for the points that
         // where found in the current domain
         for (j = 0; j < num_matching_ids; ++j) {
            other_subdomain_indices[matching_ids_idx[j]] = i;
            other_subdomain_points_local_ids[matching_ids_idx[j]] = temp_local_ids[j];
         }

         // reduce the number of open points by the number of found points
         num_open_other_subdomain_points -= num_matching_ids;

         if (num_open_other_subdomain_points == 0)
            break;

         // put all not yet found points to the beginning of the lists
         // (the algorithm ensures that the not yet found global ids
         // stay in a sorted order)
         for (j = 0, k = 0; j < num_open_other_subdomain_points; ++j) {

            // if the current point was not yet found
            if (other_subdomain_indices[j] == (unsigned)-1)
               continue;

            if (k < j) k = j + 1;

            // search for the next point that was not yet found
            while (k < num_other_subdomain_points &&
                   other_subdomain_indices[k] != (unsigned)-1) ++k;

            int tmp_int;
            unsigned tmp_uint;

            tmp_int = other_subdomain_points[j];
            other_subdomain_points[j] = other_subdomain_points[k];
            other_subdomain_points[k] = tmp_int;
            tmp_uint = other_subdomain_indices[j];
            other_subdomain_indices[j] = other_subdomain_indices[k];
            other_subdomain_indices[k] = tmp_uint;
            tmp_uint = other_subdomain_points_local_ids[j];
            other_subdomain_points_local_ids[j] =
               other_subdomain_points_local_ids[k];
            other_subdomain_points_local_ids[k] = tmp_uint;
         }
      }
      free(temp_local_ids);
      free(matching_ids_idx);
   }
}

/**
 * generates the interpolation_plan_element based on the collected search data
 * @param[in]  data                             basic search data (see \ref search_data_avg)
 * @param[in]  remote_points                    global ids of source points, required for
 *                                              the interpolation, that are located on
 *                                              remote source domains
 * @param[out] remote_point_ranks               ranks for the points listed in
 *                                              remote_points
 * @param[in]  num_remote_points                number of remote points listed in
 *                                              remote_points
 * @param[in]  other_subdomain_points           global ids of source points, required for
 *                                              the interpolation, that are located on a
 *                                              local subdomain which is not the one the
 *                                              respective target point is assigned to
 * @param[in]  other_subdomain_indices          subdomain indices for the points listed in
 *                                              other_subdomain_points
 * @param[in]  other_subdomain_points_local_ids local ids of the points listed in
 *                                              other_subdomain_points
 * @param[in]  num_other_subdomain_points       number of points in other_subdomain_points
 */
static struct interpolation_plan_element *
generate_interpolation_plan_avg(struct search_data_avg * data,
                                int * remote_points, int * remote_point_ranks,
                                unsigned num_remote_points,
                                int * other_subdomain_points,
                                unsigned * other_subdomain_indices,
                                unsigned * other_subdomain_points_local_ids,
                                unsigned num_other_subdomain_points) {

   if (*(data->num_interpolated_points) == 0) return NULL;

   struct interpolation_method_avg * method_avg =
      (struct interpolation_method_avg*)(data->method_stack[0]);
   struct domain * domain = method_avg->domain;
   unsigned points_idx = method_avg->points_idx;

   unsigned i, j, k;
   unsigned num_tgt_to_src_deps;
   unsigned * src_indices;
   unsigned * num_src_per_tgt;
   double * weights;
   unsigned num_src_indices;

   unsigned distance_weighted = method_avg->distance_weighted;

   num_tgt_to_src_deps = yac_get_total_num_dependencies(*(data->compact_tgt_to_src));

   src_indices = malloc(3 * num_tgt_to_src_deps * sizeof(*src_indices));
   num_src_per_tgt = malloc(*(data->num_interpolated_points) * sizeof(*num_src_per_tgt));
   memcpy(num_src_per_tgt, data->compact_tgt_to_src->num_deps_per_element,
          *(data->num_interpolated_points) * sizeof(*num_src_per_tgt));
   weights = malloc(num_tgt_to_src_deps * sizeof(*weights));

   int rank, comm_rank;
   comm_rank = yac_get_comm_rank(yac_get_domain_comm(domain));

   // for all points that need to be interpolated
   for (i = 0, num_src_indices = 0; i < *(data->num_interpolated_points); ++i) {

      unsigned const * src_points;
      int const * global_src_ids;
      unsigned subdomain_idx = data->tgt_to_subdomain[i];

      src_points = yac_get_dependencies_of_element(*(data->compact_tgt_to_src), i);
      global_src_ids = yac_get_global_ids(domain, subdomain_idx, CORNER);

      double target_xyz[3];
      struct grid * src_grid = NULL;

      if (method_avg->distance_weighted) {

         // determine coordinates of the target point
         LLtoXYZ(data->x_coordinates[i], data->y_coordinates[i], target_xyz);


        if (method_avg->points != NULL &&
            method_avg->points[subdomain_idx] != NULL &&
            method_avg->points[subdomain_idx][points_idx] != NULL)
          src_grid =
            yac_get_point_grid(method_avg->points[subdomain_idx][points_idx]);
        else
          src_grid = yac_get_subdomain_grid(*domain, subdomain_idx);
      }

      // store requests to other source processes if necessary
      for (k = 0; k < num_src_per_tgt[i]; ++k) {

         if (yac_is_halo_value(domain, subdomain_idx, CORNER, src_points[k],
                               (unsigned *)&rank)) {

            if (rank != comm_rank) {

                for (j = 0; j < num_remote_points; ++j)
                  if(remote_points[j] == global_src_ids[src_points[k]])
                     break;

               src_indices[3*num_src_indices+0] = 0;
               src_indices[3*num_src_indices+1] = points_idx;
               src_indices[3*num_src_indices+2] = j;
               remote_point_ranks[j] = rank;

            } else {

               j = 0;
               while(other_subdomain_points[j] != global_src_ids[src_points[k]]) ++j;

               src_indices[3*num_src_indices+0] = other_subdomain_indices[j] + 1;
               src_indices[3*num_src_indices+1] = points_idx;
               src_indices[3*num_src_indices+2] = other_subdomain_points_local_ids[j];
            }
         } else {

            src_indices[3*num_src_indices+0] = subdomain_idx + 1;
            src_indices[3*num_src_indices+1] = points_idx;
            src_indices[3*num_src_indices+2] = src_points[k];
         }

         // if distance_weighted
         if ( src_grid != NULL ) {


             // get the source vertex coordinates
             double source_xyz[3];
             LLtoXYZ(yac_get_corner_x_coord(src_grid, src_points[k]),
                     yac_get_corner_x_coord(src_grid, src_points[k]),
                     source_xyz);

             // get the distance and use it as an equivalent for weight

             weights[num_src_indices] = get_vector_angle(target_xyz, source_xyz );

         } else {

             weights[num_src_indices] = 1.0 / (double) num_src_per_tgt[i];

         }
         num_src_indices++;
      }
   }

   if ( distance_weighted ) {

       for (i = 0, num_src_indices = 0; i < *(data->num_interpolated_points); ++i) {

           double distance_sum = 0.0;

           for (k = 0; k < num_src_per_tgt[i]; ++k) {

               if ( weights[num_src_indices+k] > 0.0 ) {
                   distance_sum += 1.0 / weights[num_src_indices+k];
               } else {
                   // go for special case as source and target are at the same location
                   break;
               }
           }

           if ( k < num_src_per_tgt[i] ) {

               // special case: source and target at same location

               for ( int jj = 0; jj < num_src_per_tgt[i]; ++jj )
                 weights[num_src_indices+jj] = 0.0;

               weights[num_src_indices+k] = 1.0;

               num_src_indices++;

           } else {

               // normal case: all points contribute

               distance_sum = 1.0 / distance_sum;

               for (k = 0; k < num_src_per_tgt[i]; ++k) {
                   weights[num_src_indices] = ( 1.0 / weights[num_src_indices] ) * distance_sum;
                  num_src_indices++;
               }
           }
       }
   }

   unsigned num_pointsets = method_avg->num_pointsets;

   unsigned * num_remote_points_;

   num_remote_points_ = calloc(num_pointsets, sizeof(*num_remote_points_));
   num_remote_points_[points_idx] = num_remote_points;

   // generate a plan to execute the interpolation
   struct interpolation_plan_element * plan;

   plan = malloc(1 * sizeof(*plan));

   yac_init_interpolation_plan_element_wpointsum(plan, *(data->num_interpolated_points),
                                                 src_indices, num_src_per_tgt, weights,
                                                 num_remote_points_, num_pointsets);

   free(num_remote_points_);
   free(num_src_per_tgt);
   free(src_indices);
   free(weights);

   return plan;
}

/**
 * callback routine passed to exchange_results_init by
 * do_search_interpolation_method_avg
 */
static void
do_search_continue_avg(void * user_data,
                       int * global_ids_to_remove,
                       unsigned num_global_ids) {

   struct search_data_avg * data;
   struct interpolation_method_avg * method_avg;

   data = user_data;
   method_avg = (struct interpolation_method_avg*)(data->method_stack[0]);

   //--------------------------------------------------------------
   // remove points that are interpolated by another source process
   //--------------------------------------------------------------
   remove_points(num_global_ids, global_ids_to_remove, data);

   int * remote_points = NULL;
   unsigned num_remote_points = 0;
   int * other_subdomain_points = NULL;
   unsigned num_other_subdomain_points = 0;

   //--------------------------------------------------------------------
   // Check for source points required for the interpolation that are not
   // in the subdomain, which the target points are assigned to. (can be
   // on another local subdomain or on a remote source domain)
   //--------------------------------------------------------------------
   check_for_remote_points(&remote_points, &num_remote_points,
                           &other_subdomain_points, &num_other_subdomain_points,
                           data);

   unsigned * other_subdomain_indices;
   unsigned * other_subdomain_points_local_ids;

   other_subdomain_indices = malloc(num_other_subdomain_points *
                                    sizeof(*other_subdomain_indices));
   other_subdomain_points_local_ids =  malloc(num_other_subdomain_points *
                                              sizeof(*other_subdomain_points_local_ids));

   //------------------------------------------------------------------
   // get the subdomain indices and local indices for all source points
   // in other_subdomain_points
   //------------------------------------------------------------------
   get_subdomain_indices(other_subdomain_points, num_other_subdomain_points,
                         other_subdomain_indices, other_subdomain_points_local_ids,
                         method_avg->domain);

   int * remote_point_ranks;

   if (num_remote_points > 0)
      remote_point_ranks = malloc(num_remote_points * sizeof(*remote_point_ranks));
   else
      remote_point_ranks = NULL;

   for (unsigned i = 0; i < num_other_subdomain_points; ++i)
      if (other_subdomain_indices[i] == -1)
         yac_internal_abort_message("ERROR(do_search_continue_avg): halo "
                                    "information between subdomains does not"
                                    " match", __FILE__, __LINE__);

   //----------------------------------------------------------
   // generate a interpolation plan based on the collected data
   //----------------------------------------------------------
   data->plan_stack[0] = generate_interpolation_plan_avg(data, remote_points,
                                                         remote_point_ranks,
                                                         num_remote_points,
                                                         other_subdomain_points,
                                                         other_subdomain_indices,
                                                         other_subdomain_points_local_ids,
                                                         num_other_subdomain_points);

   // advance interpolation plan stack if an interpolation plan was generated
   if (data->plan_stack[0] != NULL) data->plan_stack++;

   // cleanup
   free(other_subdomain_points);
   free(other_subdomain_indices);
   free(other_subdomain_points_local_ids);

   unsigned num_pointsets;
   unsigned points_idx;

   points_idx = method_avg->points_idx;

   num_pointsets = method_avg->num_pointsets;
   method_avg->data_request[0] = malloc(num_pointsets *
                                        sizeof(*(method_avg->data_request[0])));

   method_avg->data_request[0][points_idx].num_values = num_remote_points;
   method_avg->data_request[0][points_idx].global_ids = remote_points;
   method_avg->data_request[0][points_idx].ranks = remote_point_ranks;

   unsigned i;

   for (i = 0; i < num_pointsets; ++i) {
      if (i != points_idx) {
         method_avg->data_request[0][i].num_values = 0;
         method_avg->data_request[0][i].global_ids = NULL;
         method_avg->data_request[0][i].ranks = NULL;
      }
   }

   data->num_points -= *(data->num_interpolated_points);

   if (data->x_coordinates != NULL)
      data->x_coordinates += *(data->num_interpolated_points);
   if (data->y_coordinates != NULL)
      data->y_coordinates += *(data->num_interpolated_points);

   // call the do search interpolation_method for the next method
   if (data->method_stack[1] != NULL)
      data->method_stack[1]->vtable->do_search(
         data->method_stack+1, data->plan_stack,
         data->global_ids+*(data->num_interpolated_points),
         data->x_coordinates, data->y_coordinates,
         data->tgt_location, data->num_points,
         data->num_interpolated_points+1,
         method_avg->data_request+1);

   free(data->tgt_to_subdomain);
   yac_free_dep_list(data->compact_tgt_to_src);
   free(data->compact_tgt_to_src);
   free(data);

   if (method_avg->finish_callback != NULL)
      method_avg->finish_callback(method_avg->finish_user_data);
}

/**
 * checks for masked source points in the search results
 * @param[out] is_masked  contains for each target point information on
 *                        on masked source points
 * @param[in]  tgt_to_src contains for each target point the associated
 *                        source points
 * @param[in]  mask       source points mask
 */
static void
check_mask(enum masked * is_masked, struct dep_list tgt_to_src,
           struct mask * mask) {

   unsigned i, j;

   unsigned num_tgt_points = tgt_to_src.num_elements;

   for (i = 0; i < num_tgt_points; ++i) {
      if (tgt_to_src.num_deps_per_element[i] > 0) {

         unsigned num_source_points;
         unsigned const * source_points;

         num_source_points = tgt_to_src.num_deps_per_element[i];
         source_points = yac_get_dependencies_of_element(tgt_to_src, i);

         unsigned num_masked_points = 0;

         for (j = 0; j < num_source_points; ++j)
            if (mask->mask[source_points[j]])
               ++num_masked_points;

         if (num_masked_points) {
            if (num_masked_points == num_source_points)
               is_masked[i] = MASKED;
            else
               is_masked[i] = PARTIAL_MASKED;
         }
      }
   }
}

/**
 * compresses the search results into a single dependency list and an array containing
 * a matching between found target     points and an associated local subdomain
 * @param[out] compact_tgt_to_src      dependency list containing for each target point,
 *                                     which can be interpolated by this domain, the
 *                                     associated source points
 * @param[out] tgt_to_subdomain        array containing for each target the local
 *                                     subdomain that contains the source points used
 *                                     to interpolate it
 * @param[out] num_interpolated_points number of target points that can be interpolated
 *                                     by the local domain
 * @param[in]  num_tgt_points          number of target points
 * @param[in]  tgt_to_src              array of dependency lists of size num_subdomains
 *                                     containing for each target point the associated
 *                                     source points
 * @param[in]  num_subdomains          number of subdomains in the local domain
 * @param[in]  is_masked               containing for each target points whether its
 *                                     associated source points are masked
 * @param[in]  use_source_mask         true ( >0 ), if source mask shall be applied
 * @param[in]  partial_coverage        precise how to handle target points whose associated
 *                                     source points are masked
 * @param[in]  mask                    mask that is to be used in case partial_coverage is .true.
 * @param[in]  points_idx              which points set of the domain is to be used for
 *                                     the interpolation
 */
 static void
 compact_search_results(struct dep_list * compact_tgt_to_src,
                        unsigned * tgt_to_subdomain, unsigned * num_interpolated_points,
                        unsigned num_tgt_points, struct dep_list * tgt_to_src,
                        unsigned num_subdomains,
                        enum masked * is_masked,
                        unsigned use_source_mask,
                        unsigned partial_coverage,
                        struct mask *** mask, unsigned points_idx) {

   unsigned i, j, k;

   // contains for each interpolated target point the local source
   // indices (independent of the subdomain index)
   unsigned * compact_tgt_to_src_deps = NULL;
   unsigned compact_tgt_to_src_deps_array_size = 0;
   unsigned num_compact_tgt_to_src_deps = 0;
   unsigned * num_src_per_tgt;

   num_src_per_tgt = malloc(num_tgt_points * sizeof(*num_src_per_tgt));

   *num_interpolated_points = 0;

   // collects the search results for all target points and prepare them to be
   // stored in a more compact data format
   for (i = 0; i < num_tgt_points; ++i) {
      for (j = 0; j < num_subdomains; ++j) {
         if (tgt_to_src[j].num_deps_per_element[i] > 0) {
            if (!use_source_mask || !is_masked[i]) {

               unsigned const * curr_deps;

               curr_deps = yac_get_dependencies_of_element(tgt_to_src[j], i);

               ENSURE_ARRAY_SIZE(compact_tgt_to_src_deps,
                                 compact_tgt_to_src_deps_array_size,
                                 num_compact_tgt_to_src_deps +
                                 tgt_to_src[j].num_deps_per_element[i]);

               memcpy(compact_tgt_to_src_deps+num_compact_tgt_to_src_deps,
                      curr_deps, tgt_to_src[j].num_deps_per_element[i] *
                      sizeof(*compact_tgt_to_src_deps));

               num_compact_tgt_to_src_deps += tgt_to_src[j].num_deps_per_element[i];
               num_src_per_tgt[*num_interpolated_points] = tgt_to_src[j].num_deps_per_element[i];

               tgt_to_subdomain[*num_interpolated_points] = j;

               ++*num_interpolated_points;

               break;
            } else if (is_masked[i] == PARTIAL_MASKED && partial_coverage) {

               unsigned const * curr_deps;

               curr_deps = yac_get_dependencies_of_element(tgt_to_src[j], i);

               ENSURE_ARRAY_SIZE(compact_tgt_to_src_deps,
                                 compact_tgt_to_src_deps_array_size,
                                 num_compact_tgt_to_src_deps +
                                 tgt_to_src[j].num_deps_per_element[i]);

               unsigned num_unmasked_source_points = 0;

               for (k = 0; k < tgt_to_src[j].num_deps_per_element[i]; ++k)
                  if (!mask[j][points_idx]->mask[curr_deps[k]])
                     compact_tgt_to_src_deps[num_compact_tgt_to_src_deps+
                                             num_unmasked_source_points++] = curr_deps[k];

               num_compact_tgt_to_src_deps += num_unmasked_source_points;
               num_src_per_tgt[*num_interpolated_points] = num_unmasked_source_points;

               tgt_to_subdomain[*num_interpolated_points] = j;

               ++*num_interpolated_points;

               break;
            }
         }
      }
   }

   // if not all points can be interpolated
   if (*num_interpolated_points != num_tgt_points)
      num_src_per_tgt = realloc(num_src_per_tgt, *num_interpolated_points * sizeof(*num_src_per_tgt));


   // store search results in a more compact data format
   compact_tgt_to_src_deps = realloc(compact_tgt_to_src_deps,
                                     num_compact_tgt_to_src_deps *
                                     sizeof(*compact_tgt_to_src_deps));
         
   yac_set_dependencies(compact_tgt_to_src, *num_interpolated_points,
                        num_src_per_tgt, compact_tgt_to_src_deps);
}

static void
do_search_interpolation_method_avg(struct interpolation_method ** method,
                                   struct interpolation_plan_element ** plan_stack,
                                   int * global_ids, double * x_coordinates,
                                   double * y_coordinates,
                                   enum yac_location tgt_location,
                                   unsigned num_points,
                                   unsigned * num_interpolated_points,
                                   struct interpolation_method_data_request ** data_request) {

   unsigned i, j;

   struct interpolation_method_avg * method_avg;
   method_avg = (struct interpolation_method_avg*)(method[0]);

   method_avg->data_request = data_request;

   struct search_data_avg * user_data;

   user_data = malloc(1 * sizeof(*user_data));

   if (x_coordinates != NULL && y_coordinates != NULL) {

      unsigned points_idx;

      points_idx = method_avg->points_idx;

      struct dep_list * tgt_to_src;
      unsigned num_subdomains;

      num_subdomains =
         yac_get_num_subdomains(method_avg->domain);

      tgt_to_src = malloc(num_subdomains * sizeof(*tgt_to_src));

      struct points *** points = method_avg->points;

      // search for the received points in the local data
      for (i = 0; i < num_subdomains; ++i)
         yac_do_point_search_p3(yac_get_subdomain_grid_search(method_avg->search, i), x_coordinates,
                                y_coordinates, num_points, tgt_to_src+i,
                                points[i][points_idx]);

      // check for masked source points
      enum masked * is_masked;
      is_masked = calloc(num_points, sizeof(*(is_masked)));
      struct mask *** masks = method_avg->masks;
      if (masks != NULL)
      for (i = 0; i < num_subdomains; ++i)
         if (masks[i] != NULL && masks[i][points_idx] != NULL)
            check_mask(is_masked, tgt_to_src[i], masks[i][points_idx]);

      // contains for each interpolated target point the local source
      // indices (independent of the subdomain index)
      struct dep_list * compact_tgt_to_src;
      // contains for each interpolated target point the respective subdomain index
      unsigned * tgt_to_subdomain;

      compact_tgt_to_src = malloc(1 * sizeof(*compact_tgt_to_src));
      tgt_to_subdomain = malloc(num_points * sizeof(*tgt_to_subdomain));

      //-------------------------------------------------------------
      // compresses the search results returned by do_point_search_p3
      //-------------------------------------------------------------
      compact_search_results(compact_tgt_to_src, tgt_to_subdomain, num_interpolated_points,
                             num_points, tgt_to_src, num_subdomains, is_masked,
                             method_avg->use_source_mask,
                             method_avg->partial_coverage, masks,
                             method_avg->points_idx);

      // if not all points can be interpolated
      if (*num_interpolated_points != num_points) {

         int * point_can_be_interpolated;
         point_can_be_interpolated = calloc(num_points, sizeof(*point_can_be_interpolated));

         for (i = 0; i < num_points && !point_can_be_interpolated[i]; ++i)
            for (j = 0; j < num_subdomains; ++j)
               if (tgt_to_src[j].num_deps_per_element[i] > 0 &&
                    (!method_avg->use_source_mask ||
                    (!is_masked[i]) ||
                    ((is_masked[i] == PARTIAL_MASKED) &&
                     (method_avg->partial_coverage))))
                  { point_can_be_interpolated[i] = 1 == 1; break; }

         // reorder global_ids, x_coordinates and y_coordinates such that the interpolated
         // points are at the beginning of the array, while keeping the order of the
         // points that can be interpolated intact
         reorder_points(global_ids, x_coordinates, y_coordinates,
                        point_can_be_interpolated, num_points, *num_interpolated_points);

         free(point_can_be_interpolated);
      }

      free(is_masked);

      // if not all points can be interpolated
      if (*num_interpolated_points != num_points)
         tgt_to_subdomain = realloc(tgt_to_subdomain, *num_interpolated_points * sizeof(*tgt_to_subdomain));

      // cleanup
      for (i = 0; i < num_subdomains; ++i)
         yac_free_dep_list(tgt_to_src+i);
      free(tgt_to_src);

      // store data that is to be passed to do_search_continue_avg
      user_data->method_stack = method;
      user_data->plan_stack = plan_stack;
      user_data->global_ids = global_ids;
      user_data->x_coordinates = x_coordinates;
      user_data->y_coordinates = y_coordinates;
      user_data->tgt_location = tgt_location;
      user_data->num_points = num_points;
      user_data->num_interpolated_points = num_interpolated_points;
      user_data->compact_tgt_to_src = compact_tgt_to_src;
      user_data->tgt_to_subdomain = tgt_to_subdomain;

   } else { // if no target point coordinates were passed into this routine

      user_data->method_stack = method;
      user_data->plan_stack = plan_stack;
      user_data->global_ids = global_ids;
      user_data->x_coordinates = x_coordinates;
      user_data->y_coordinates = y_coordinates;
      user_data->tgt_location = tgt_location;
      user_data->num_points = num_points;
      *num_interpolated_points = 0;
      user_data->num_interpolated_points = num_interpolated_points;
      user_data->compact_tgt_to_src = NULL;
      user_data->tgt_to_subdomain = NULL;
   }

   // exchange results with other source processes
   // (processing continues asynchronously at do_search_continue_avg
   // once the results are exchanged between the source processes)
   yac_exchange_results(method[0], global_ids, *num_interpolated_points,
                        num_points, method_avg->tags.global_id_exchange,
                        yac_get_domain_comm(method_avg->domain),
                        do_search_continue_avg, user_data);
}

static void 
enddef_interpolation_method_avg(struct interpolation_method ** method) {

   struct interpolation_method_avg * method_avg;

   method_avg = (struct interpolation_method_avg*)(method[0]);

   while(yac_comm_wait(yac_get_domain_comm(method_avg->domain),
                       method_avg->tags.global_id_exchange[0]) ||
         yac_comm_wait(yac_get_domain_comm(method_avg->domain),
                       method_avg->tags.global_id_exchange[1]));

   if (method[1] != NULL)
      method[1]->vtable->enddef(method + 1);
}

static void 
get_source_exchange_data_interpolation_method_avg(
   struct interpolation_method ** method,
   struct interpolation_method_data_request *** data_request) {

   struct interpolation_method_avg * method_avg;

   method_avg = (struct interpolation_method_avg*)(method[0]);

   *data_request = method_avg->data_request;
   method_avg->data_request = NULL;
}

static
void delete_interpolation_method_avg(struct interpolation_method * method) {

   free(method);
}

 static struct global_search *
 get_global_search(struct interpolation_method * method) {
 
   return ((struct interpolation_method_avg*)(method))->search;
 }

 static void
 set_finish_callback(struct interpolation_method * method,
                     void (*callback)(void * user_data),
                     void * user_data) {

   struct interpolation_method_avg * method_avg;

   method_avg = (struct interpolation_method_avg*)method;
   
   method_avg->finish_callback = callback;
   method_avg->finish_user_data = user_data;
}
