/**
 * @file interpolation_method_conserv.c
 *
 * \brief implementation of the 1st order conservative remapping method
 *
 * \section interpolation_method_conserv Communication schema for conservative
 *          remapping
 *
 * Prerequisite for this algorithm is the availability of a list of target
 * cells with its associated overlapping source cell on the local rank
 * plus information about ranks which may contain additional source cell
 * overlapping the target.
 *
 * Intersection cells between source and target cells are determined
 * and the area is calculated for each of these cells.
 *
 * The source processes exchange information, and for each target cell
 * a master is identified, which collects the partial_areas from other source
 * ranks where an overlap with the particular target cell was detected.
 * Information is sent back from the master ranks to the servant ranks,
 * whether target cells can be processed or not. Target cells that can be
 * handled by the master ranks are eliminated on the server ranks, and the
 * next interpolation method is called.
 *
 * The weights are normalised and corrected such that the sum of all weights
 * per target cell equals 1.0.
 *
 *  \msc
 *    width = "800";
 *    "Master Sources","Servant Sources";
 *    |||;
 *    "Master Sources" box "Servant Sources" [label="build target and overlapping source cells",
 *                       URL="\ref do_search_interpolation_method_conserv"];
 *    |||;
 *    "Master Sources" box "Servant Sources" [label="construct source and target cells"];
 *    "Master Sources" box "Servant Sources" [label="exclude source cells that are located in halos"];
 *    "Master Sources" box "Servant Sources" [label="calculate partial areas"];
 *    "Master Sources" box "Servant Sources" [label="identify master and servant ranks"];
 *    "Master Sources" box "Servant Sources" [label="determine send and receive information"];
 *    "Servant Sources" box "Servant Sources" [label="post receive_callback_accepted for answer from master ranks"];
 *    "Master Sources"<-"Servant Sources" [label="partial areas and provenance data"];
 *    "Master Sources" box "Master Sources" [label="post receive_callback_partial_areas for partial areas and provenance data"];
 *    "Servant Sources" box "Servant Sources" [label="post 2nd set of callback receive message"];
 *    |||;
 *    "Master Sources" box "Servant Sources" [label="continue with second part",
 *                       URL="\ref do_search_interpolation_method_conserv_continue_2"];
 *    |||;
 *    "Master Sources" box "Master Sources" [label="check if target cell is fully covered"];
 *    "Master Sources"->"Servant Sources" [label="list of cells that can be processed by the master"];
 *    "Master Sources" box "Servant Sources" [label="reorder weights according to increasing global cell ID"];
 *    "Servant Sources" box "Servant Sources" [label="erase target cells from local list"];
 *    |||;
 *  \endmsc
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "grid.h"
#include "interpolation_method.h"
#include "interpolation_method_conserv.h"
#include "interpolation_plan.h"
#include "clipping.h"
#include "search.h"
#include "global_search.h"
#include "ensure_array_size.h"
#include "utils.h"
#include "area.h"

// -----------------------------------------------------------------------------

static void
do_search_interpolation_method_conserv ( struct interpolation_method ** method,
                                         struct interpolation_plan_element ** plan_stack,
                                         int * global_ids, double * x_coordinates,
                                         double * y_coordinates,
                                         enum yac_location tgt_location,
                                         unsigned num_cells,
                                         unsigned * num_interpolated_cells,
                                         struct interpolation_method_data_request ** data_request );

static void
enddef_interpolation_method_conserv ( struct interpolation_method ** method );

static void
get_source_exchange_data_interpolation_method_conserv ( struct interpolation_method ** method,
                                                        struct interpolation_method_data_request *** data_request );

static void
delete_interpolation_method_conserv ( struct interpolation_method * method );

static struct global_search *
get_global_search ( struct interpolation_method * method );

static void
set_finish_callback ( struct interpolation_method * method,
                      void (*callback)(void * user_data),
                      void * user_data);

// -----------------------------------------------------------------------------

static struct interpolation_method_vtable
interpolation_method_conserv_vtable = {do_search_interpolation_method_conserv,
                                       enddef_interpolation_method_conserv,
                                       get_source_exchange_data_interpolation_method_conserv,
                                       delete_interpolation_method_conserv,
                                       get_global_search,
                                       set_finish_callback};

struct interpolation_method_conserv_tags {
  unsigned alltoallv_uncertain_tgt;
  unsigned alltoallv_master_ranks;
  unsigned header1_exchange;
  unsigned header2_exchange;
  unsigned global_id_exchange;
  unsigned partial_areas_exchange;
  unsigned accepts_exchange;
};

struct interpolation_method_conserv {
  struct interpolation_method_vtable * vtable;
  void (*finish_callback)(void * user_data);
  void * finish_user_data;
  struct global_search * search;
  struct domain * domain;
  struct points *** points;
  struct mask *** masks;
  unsigned num_pointsets;
  unsigned points_idx;
  struct interpolation_method_data_request ** data_request;
  struct interpolation_method_conserv_tags tags;
  int enforced_conserv;
  int partial_coverage;
  enum yac_interpolation_method_conserv_normalisation normalisation;
  unsigned use_source_mask;
};

//! contains information on a single overlap between a source and a target cell
struct overlap {
  int subdomain_id;   //!< global source cell subdomain ID
  int global_cell_id; //!< global source cell ID
  int local_cell_id;  //!< local source cell ID
  int rank;           //!< source cell rank
  double area;        //!< area associated with cell
  double weight;      //!< source cell partial weight
};

//! contains all information the is relevant for the interpolation of a single
//! target cell
struct weight_provenance {
  int target_cell_id;           //!< id of the target cell
  int master_rank;              //!< rank of the source process that is
                                //!< responsible for interpolating this
                                //!< target cell
  int nbr_servant_ranks;        //!< number of source processes (not including
                                //!< the master) that have cells, which have
                                //!< an overlap with the target cell (this is
                                //!< only available on the respective master)
  unsigned * servant_ranks;     //!< ranks of source processes (not including
                                //!< the master) that have cells, which have
                                //!< an overlap with the target cell (this is
                                //!< only available on the respective master)
  int nbr_overlaps;             //!< total number of overlaps
  struct overlap * overlaps;    //!< array containing information on all
                                //!< overlaps
  int cell_can_be_interpolated; //!< flag that indicates whether the target
                                //!< cell can be interpolated
  double target_cell_area;      //!< area of the target cell in km^2
};

struct send_buffer {
  unsigned num_target_cells;    //!< number of target cells to be sent
  unsigned buffer_array_size;   //!< size of buffer
  unsigned buffer_size;         //!< size of buffer
  int position;                 //!< position of packed buffer
  int * ibuffer;                //!< target cell ids and number of partial areas per target cell
  char * cbuffer;               //!< buffer for packing
                                //!< - source cell ids associated to the partial areas
                                //!< - partial areas in km^2
                                //!< - target cell ids
                                //!< - number of partial areas per target cell
};

struct temp_recv_partial_areas_data {

  struct search_data_conserv * search_data;
  unsigned num_target_cells;
};

struct temp_uncertain_tgt {

  unsigned src_rank;       // which rank sent this point
                           // or later the master rank
  unsigned master_rank;    //
  int global_id;           // global id of the target point
  unsigned num_src_points; // how many points did the rank found for this tgt
};

// -----------------------------------------------------------------------------

static void
recv_callback_partial_areas ( struct communicator * comm,
                              struct comm_message message,
                              void * user_data );

static void
recv_callback_accepts ( struct communicator * comm, struct comm_message message,
                        void * user_data );

// -----------------------------------------------------------------------------

static void
generate_tag_string ( char * buf, size_t buffer_size,
                      char const * tag_base_str,
                      char const * source_comp_name,
                      char const * target_comp_name,
                      char const * source_domain_name,
                      char const * target_domain_name,
                      char const * stack_name,
                      int points_idx,
                      int unique_interpolation_method_id,
                      int enforced_conserv,
                      int partial_coverage,
                      enum yac_interpolation_method_conserv_normalisation normalisation,
                      unsigned use_source_mask,
                      char const * tag_str) {

  int ret_val;
  ret_val = snprintf ( buf, buffer_size, "%s%s%s%s%s%s%d%d%d%d%d%d%s", tag_base_str,
                       source_comp_name, target_comp_name, source_domain_name,
                       target_domain_name, stack_name, points_idx,
                       unique_interpolation_method_id, enforced_conserv,
                       partial_coverage, normalisation, use_source_mask, tag_str);

  if ( ret_val < 0 || ret_val >= buffer_size )
    yac_internal_abort_message ( "ERROR(generate_tag_string): truncation", __FILE__, __LINE__ );
}

// -----------------------------------------------------------------------------

static void
generate_interpolation_method_conserv_tag ( struct interpolation_method_conserv * method,
                                            char const * stack_name ) {

  char const * str_base_str                = "interpolation method conserv";
  char const * str_generator_string        = "generator string";
  char const * str_alltoallv_uncertain_tgt = "alltoallv uncertain tgt";
  char const * str_alltoallv_master_ranks  = "alltoallv master ranks";
  char const * str_header1_exchange        = "header1 exchange";
  char const * str_header2_exchange        = "header2 exchange";
  char const * str_global_id_exchange      = "global id exchange";
  char const * str_partial_areas_exchange  = "partial_areas exchange";
  char const * str_accepts_exchange        = "accepts exchange";

  struct global_search * search;

  search = get_global_search((struct interpolation_method *)method);

  struct communicator * comm;

  comm = yac_get_domain_comm(yac_get_local_domain(search));

  char const * source_comp_name = yac_get_local_component_name(search);
  char const * target_comp_name = yac_get_remote_component_name(search);

  char const * source_domain_name = yac_get_local_domain_name(search);
  char const * target_domain_name = yac_get_remote_domain_name(search);

  unsigned max_tag_strlen;

  max_tag_strlen = strlen(str_alltoallv_uncertain_tgt);

  unsigned max_str_buf_size;

  max_str_buf_size = max_tag_strlen +
                     strlen(str_base_str) +
                     strlen(source_comp_name) +
                     strlen(target_comp_name) +
                     strlen(source_domain_name) +
                     strlen(target_domain_name) +
                     strlen(stack_name) + 10 // for points_idx
                                        + 10 // for unique_method_id
                                        + 1  // for enforced_conserv (boolean)
                                        + 1  // for partial_cov (boolean)
                                        + 10 // for normalisation
                                        + 10 // for use_source_mask
                                        + 1; //  for '\0'

   char * unique_generator_string = malloc(max_str_buf_size);
   generate_tag_string ( unique_generator_string, max_str_buf_size, str_base_str,
                         source_comp_name, target_comp_name,
                         source_domain_name, target_domain_name,
                         stack_name, method->points_idx,
                         yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                         method->enforced_conserv,
                         method->partial_coverage,
                         method->normalisation,
                         method->use_source_mask,
                         str_generator_string );

   char * tag_strings[7];
   tag_strings[0] = malloc(max_str_buf_size);
   generate_tag_string ( tag_strings[0], max_str_buf_size, str_base_str,
                         source_comp_name, target_comp_name,
                         source_domain_name, target_domain_name,
                         stack_name, method->points_idx,
                         yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                         method->enforced_conserv,
                         method->partial_coverage,
                         method->normalisation,
                         method->use_source_mask,
                         str_alltoallv_uncertain_tgt );

   tag_strings[1] = malloc(max_str_buf_size);
   generate_tag_string ( tag_strings[1], max_str_buf_size, str_base_str,
                         source_comp_name, target_comp_name,
                         source_domain_name, target_domain_name,
                         stack_name, method->points_idx,
                         yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                         method->enforced_conserv,
                         method->partial_coverage,
                         method->normalisation,
                         method->use_source_mask,
                         str_alltoallv_master_ranks );

   tag_strings[2] = malloc(max_str_buf_size);
   generate_tag_string ( tag_strings[2], max_str_buf_size, str_base_str,
                         source_comp_name, target_comp_name,
                         source_domain_name, target_domain_name,
                         stack_name, method->points_idx,
                         yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                         method->enforced_conserv,
                         method->partial_coverage,
                         method->normalisation,
                         method->use_source_mask,
                         str_header1_exchange );

   tag_strings[3] = malloc(max_str_buf_size);
   generate_tag_string ( tag_strings[3], max_str_buf_size, str_base_str,
                         source_comp_name, target_comp_name,
                         source_domain_name, target_domain_name,
                         stack_name, method->points_idx,
                         yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                         method->enforced_conserv,
                         method->partial_coverage,
                         method->normalisation,
                         method->use_source_mask,
                         str_header2_exchange );

   tag_strings[4] = malloc(max_str_buf_size);
   generate_tag_string ( tag_strings[4], max_str_buf_size, str_base_str,
                         source_comp_name, target_comp_name,
                         source_domain_name, target_domain_name,
                         stack_name, method->points_idx,
                         yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                         method->enforced_conserv,
                         method->partial_coverage,
                         method->normalisation,
                         method->use_source_mask,
                         str_global_id_exchange );

   tag_strings[5] = malloc(max_str_buf_size);
   generate_tag_string ( tag_strings[5], max_str_buf_size, str_base_str,
                         source_comp_name, target_comp_name,
                         source_domain_name, target_domain_name,
                         stack_name, method->points_idx,
                         yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                         method->enforced_conserv,
                         method->partial_coverage,
                         method->normalisation,
                         method->use_source_mask,
                         str_partial_areas_exchange );

   tag_strings[6] = malloc(max_str_buf_size);
   generate_tag_string ( tag_strings[6], max_str_buf_size, str_base_str,
                         source_comp_name, target_comp_name,
                         source_domain_name, target_domain_name,
                         stack_name, method->points_idx,
                         yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                         method->enforced_conserv,
                         method->partial_coverage,
                         method->normalisation,
                         method->use_source_mask,
                         str_accepts_exchange );

   unsigned tags[7];

   yac_get_unique_comm_tags(
      comm, unique_generator_string, (char const **)tag_strings, 7, tags);

   free(unique_generator_string);

   method->tags.alltoallv_uncertain_tgt = tags[0];
   method->tags.alltoallv_master_ranks = tags[1];
   method->tags.header1_exchange = tags[2];
   method->tags.header2_exchange = tags[3];
   method->tags.global_id_exchange = tags[4];
   method->tags.partial_areas_exchange = tags[5];
   method->tags.accepts_exchange = tags[6];

   for (unsigned i = 0; i < 7; ++i) free(tag_strings[i]);
}

// -----------------------------------------------------------------------------

struct interpolation_method *
yac_new_interpolation_method_conserv_mask (struct global_search * search,
                                           struct domain * domain,
                                           char const * stack_name,
                                           struct points *** points,
                                           struct mask *** masks,
                                           unsigned num_pointsets,
                                           unsigned points_idx,
                                           int enforced_conserv,
                                           int partial_coverage,
                                           enum yac_interpolation_method_conserv_normalisation normalisation,
                                           unsigned use_source_mask) {

  // ensure that the right point data is available

  if (num_pointsets == 0)
    yac_internal_abort_message ("ERROR: missing data points.", __FILE__, __LINE__);
  if (points_idx >= num_pointsets)
    yac_internal_abort_message ("ERROR(new_interpolation_method_conserv_mask): invalid "
                                "points_idx.", __FILE__, __LINE__);
  if (points == NULL || points[0] == NULL || points[0][points_idx] == NULL)
    yac_internal_abort_message ("ERROR(new_interpolation_method_conserv_mask): no point set "
                                "data available", __FILE__, __LINE__);
  if (points[0][points_idx]->location != CELL)
    yac_internal_abort_message ("ERROR(new_interpolation_method_conserv_mask): data points "
                                "have wrong location type (needs to be CELL).",
                                __FILE__, __LINE__);

  struct interpolation_method_conserv * method;

  method = malloc(1 * sizeof(*method));

  method->vtable = &interpolation_method_conserv_vtable;
  method->search = search;
  method->finish_callback = NULL;
  method->finish_user_data = NULL;
  method->enforced_conserv = enforced_conserv;
  method->partial_coverage = partial_coverage;
  method->normalisation = normalisation;
  method->points_idx = points_idx;
  method->use_source_mask = use_source_mask;
  method->domain = domain;
  method->points = points;
  method->masks = masks;
  method->num_pointsets = num_pointsets;

  yac_init_interpolation_method((struct interpolation_method *)method, search);

  generate_interpolation_method_conserv_tag(method, stack_name);

  return (struct interpolation_method *)method;
}

struct interpolation_method *
yac_new_interpolation_method_conserv ( struct global_search * search,
                                       struct domain * domain,
                                       char const * stack_name,
                                       struct points *** points,
                                       struct mask *** masks,
                                       unsigned num_pointsets,
                                       unsigned points_idx,
                                       int enforced_conserv,
                                       int partial_coverage,
                                       enum yac_interpolation_method_conserv_normalisation normalisation ) {
  unsigned use_source_mask = 1;

  return yac_new_interpolation_method_conserv_mask ( search, domain, stack_name, points, masks, num_pointsets,
                                                     points_idx, enforced_conserv, partial_coverage,
                                                     normalisation, use_source_mask );
}
// -----------------------------------------------------------------------------

static int
index_location ( unsigned len, int * list, unsigned list_req ) {

  //
  // simple bisection algorithm, list must be ordered.
  // -------------------------------------------------

  int j1, j2, j3;

  // Initialize lower and upper limits

  int location = -1;

  j1 = 0;

  j3 = len;

  while ( (j3-j1) > 0 ) {

    // Compute midpoint

    j2 = (j3+j1)/2;

    // replace lower or upper limit

    if ( (list[len-1] > list[0]) == (list_req > list[j2]) )
      j1 = j2;
    else
      j3 = j2;

    if ( list_req == list[j2] ) {
      location = j2;
      break;
    }
  }
  return location;
}

// -----------------------------------------------------------------------------

static void
insertion_sort_int ( int element, int * a, unsigned * curr_length ) {

  unsigned i, j;

  if (*curr_length > 0) {

    for (i = 0; i < *curr_length; ++i)
      if (a[i] >= element) break;

    // if the element is already in the array
    if ((i != *curr_length) && (a[i] == element)) return;

    // copy new element into array and move bigger elements one position up
    for (j = *curr_length; j > i; --j) {

      a[j] = a[j-1];
    }

    a[i] = element;

    // increase array length indicator
    ++(*curr_length);

  } else {

    a[0] = element;
    *curr_length = 1;

  }
}

// -----------------------------------------------------------------------------

/** reorders the arrays global_ids, x_coordinates and y_coordinates such
  * that the cells specified in the put_to_front array are at the
  * beginning of the three arrays
  */

static void
reorder_cells ( int * global_ids, double * x_coordinates, double * y_coordinates,
                struct weight_provenance * weight_provenance,
                unsigned num_cells, unsigned num_cells_to_front ) {

  unsigned i, j;

  for (i = 0; i < num_cells_to_front; ++i) {

    if ( !weight_provenance[i].cell_can_be_interpolated ) {

      // find a cell to replace the current one
      for (j = i + 1;!weight_provenance[j].cell_can_be_interpolated;++j);

      double temp_x, temp_y;
      int temp_id;
      struct weight_provenance temp_weight_provenance;

      if ( x_coordinates != NULL )
        temp_x = x_coordinates[i], x_coordinates[i] = x_coordinates[j], x_coordinates[j] = temp_x;
      if ( y_coordinates != NULL )
        temp_y = y_coordinates[i], y_coordinates[i] = y_coordinates[j], y_coordinates[j] = temp_y;

      temp_id = global_ids[i], global_ids[i] = global_ids[j], global_ids[j] = temp_id;
      temp_weight_provenance = weight_provenance[i];
      weight_provenance[i] = weight_provenance[j];
      weight_provenance[j] = temp_weight_provenance;
    }
  }
}


// -----------------------------------------------------------------------------

struct search_data_conserv {

  struct interpolation_method ** method_stack;
  struct interpolation_plan_element ** plan_stack;

  int * global_ids;                               //!< global ids of target cells that are to be
                                                  //!< interpolated by the local process
  double * x_coordinates;                         //!< x (centre) coordinates of the target cells
  double * y_coordinates;                         //!< y (centre) coordinates of the target cells
  unsigned num_cells;                             //!< number of target cells
  unsigned * num_interpolated_cells;              //!< number of cells that are interpolated by
                                                  //!< the conservative remapping interpolation method (this is
                                                  //!< determined in \ref do_search_interpolation_method_conserv_continue_2
                                                  //!< ( \ref remove_cells ))
  unsigned num_partial_areas_messages_to_receive; //!< number of open message for source cell exchange
  unsigned num_accepts_messages_to_receive;       //!< number of open message for rejected target cells

  struct weight_provenance * weight_provenance;   //!< partial areas, weights and metadata calculated based on intersections

  int * target_cell_ids_to_remove;                //!< array containing all global target cell ids of cells that have to be removed
  int num_target_cell_ids_to_remove;              //!< number of ids in target_cell_ids_to_remove
};

/**
 * Removes cells from the list of target cells (even if they could
 * be interpolated by the local process). This is because they are going
 * to be interpolated by another source process which might later
 * request some data from the local process to do the interpolation.
 *
 * @param[in] num_cells_to_be_removed  number of cells that are interpolated
 *                                     by another source process
 * @param[in] global_ids_to_be_removed array with global ids of target cells
 *                                     that are to be removed from the local list
 * @param[in] data                     basic search data (see \ref search_data_conserv)
 */
static void
remove_cells ( unsigned num_cells_to_be_removed, int * global_ids_to_be_removed,
               struct search_data_conserv * data ) {

  if (num_cells_to_be_removed == 0) return;

  unsigned new_num_cells = data->num_cells;
  int * remaining_cells;
  yac_remove_points(data->global_ids, &new_num_cells, global_ids_to_be_removed,
                    num_cells_to_be_removed, &remaining_cells);

  int idx = (new_num_cells > 0)?(remaining_cells[0]):(-1);
  for (unsigned i = 0, j = 0; i < data->num_cells; ++i) {

    if (i == idx) {

      if (i != j) {

        data->weight_provenance[j] = data->weight_provenance[idx];
        if (data->x_coordinates != NULL)
          data->x_coordinates[j] = data->x_coordinates[idx];
        if (data->y_coordinates != NULL)
          data->y_coordinates[j] = data->y_coordinates[idx];
      }
      if (j < new_num_cells - 1) idx = remaining_cells[++j];
    } else {
      free (data->weight_provenance[i].overlaps);
      free (data->weight_provenance[i].servant_ranks);
    }
  }

  free(remaining_cells);

  // write back the data that was reduced by the cells that are interpolated by
  // another source process

  data->num_cells = new_num_cells;
}
// -----------------------------------------------------------------------------

/**
 * Check for source cells that are required for the interpolation, which are located on
 * other source processes or on another local subdomain (which the target cell is
 * not assigned to).
 *
 * @param[out] num_remote_cells  number of cells in remote_cells
 * @param[out] remote_cells      global ids of source cells, required for
 *                               the interpolation, that are located on
 *                               remote source domains
 * @param[out] remote_cell_ranks rank for process that has the respective
 *                               remote cell
 * @param[in]  data              basic search data (see \ref search_data_conserv)
 */
static void
check_for_remote_cells ( unsigned * num_remote_cells,
                         int ** remote_cells,
                         int ** remote_cell_ranks,
                         struct search_data_conserv * data ) {

  struct interpolation_method_conserv * method_conserv;
  method_conserv = (struct interpolation_method_conserv*)(data->method_stack[0]);
  struct communicator * comm = yac_get_domain_comm(yac_get_local_domain(method_conserv->search));
  int comm_rank = yac_get_comm_rank(comm);

  unsigned i, j, k;
  int remote_cells_size = 0;
  int remote_cell_ranks_size = 0;

  struct weight_provenance * weight_provenance = data->weight_provenance;

  *num_remote_cells = 0;
  *remote_cells = NULL;
  *remote_cell_ranks = NULL;

  for ( i = 0; i < *(data->num_interpolated_cells); ++i ) {

    for  ( k = 0; k < weight_provenance[i].nbr_overlaps; ++k ) {

      if ( weight_provenance[i].overlaps[k].rank != comm_rank ) {

        // avoid double entries
        for (j = 0; j < *num_remote_cells; ++j)
          if ((*remote_cells)[j] == weight_provenance[i].overlaps[k].global_cell_id)
            break;

        if ((j != *num_remote_cells) && (*num_remote_cells > 0)) continue;

        // remote cells
        *num_remote_cells = *num_remote_cells + 1;

        ENSURE_ARRAY_SIZE(*remote_cells, remote_cells_size,
                          *num_remote_cells);

        ENSURE_ARRAY_SIZE(*remote_cell_ranks, remote_cell_ranks_size,
                          *num_remote_cells);

        (*remote_cells)[*num_remote_cells-1] = weight_provenance[i].overlaps[k].global_cell_id;
        (*remote_cell_ranks)[*num_remote_cells-1] = weight_provenance[i].overlaps[k].rank;

      }
    }
  }

  yac_quicksort_index ( *remote_cells, *num_remote_cells, *remote_cell_ranks );
}

// -----------------------------------------------------------------------------

/**
 * generates the interpolation_plan_element based on the collected search data
 *
 * @param[in]  data                             basic search data (see \ref search_data_conserv)
 * @param[in]  remote_cells                     global ids of source cells, required for
 *                                              the interpolation, that are located on
 *                                              remote source domains
 * @param[in]  num_remote_cells                 number of remote cells listed in
 *                                              remote_cells
 */

static struct interpolation_plan_element *
generate_interpolation_plan_conserv ( struct search_data_conserv * data,
                                      int * remote_cells,
                                      unsigned num_remote_cells ) {

  if (*(data->num_interpolated_cells) == 0) return NULL;

  struct weight_provenance *weight_provenance;
  struct interpolation_method_conserv * method_conserv;
  unsigned points_idx;

  weight_provenance = data->weight_provenance;
  method_conserv = (struct interpolation_method_conserv*)(data->method_stack[0]);
  points_idx = method_conserv->points_idx;

  int max_num_total_weights = 0;
  for (unsigned n = 0; n < *(data->num_interpolated_cells); ++n ) {
    max_num_total_weights += weight_provenance[n].nbr_overlaps;
  }

  unsigned * src_indices = malloc(3 * max_num_total_weights * sizeof(*src_indices));
  unsigned * num_src_per_tgt =
    calloc(*(data->num_interpolated_cells), sizeof(*num_src_per_tgt));
  double * weights = malloc(max_num_total_weights * sizeof(*weights));

  int comm_rank =
    yac_get_comm_rank(
      yac_get_domain_comm(yac_get_local_domain(method_conserv->search)));

  // loop over all target cells that can be interpolated by this process

  for ( unsigned n = 0, num_src_indices = 0; n < *(data->num_interpolated_cells); ++n ) {

    // store requests to other source processes if necessary
    for ( unsigned i = 0; i < weight_provenance[n].nbr_overlaps; ++i ) {

      if (weight_provenance[n].overlaps[i].weight == 0.0) continue;

      if ( weight_provenance[n].overlaps[i].rank != comm_rank ) {

        // source cell weights come from a remote rank

        int index = index_location ( num_remote_cells, remote_cells,
                                     weight_provenance[n].overlaps[i].global_cell_id );

        if (index == -1)
          yac_internal_abort_message("ERROR: could not find source point in remote data\n",
                                     __FILE__, __LINE__);

        src_indices[3*num_src_indices+0] = 0;
        src_indices[3*num_src_indices+1] = points_idx;
        src_indices[3*num_src_indices+2] = index;

      } else {

        // we have the source cell data in our own subdomain
        src_indices[3*num_src_indices+0] = weight_provenance[n].overlaps[i].subdomain_id + 1; // source subdomain ID + 1
        src_indices[3*num_src_indices+1] = points_idx;
        src_indices[3*num_src_indices+2] = weight_provenance[n].overlaps[i].local_cell_id;    // local source cell ID
      }

      weights[num_src_indices] = weight_provenance[n].overlaps[i].weight;
      num_src_per_tgt[n]++;
      num_src_indices++;
    }
  }

  unsigned num_pointsets = method_conserv->num_pointsets;

  unsigned * num_remote_cells_;

  num_remote_cells_ = calloc(num_pointsets, sizeof(*num_remote_cells_));
  num_remote_cells_[points_idx] = num_remote_cells;

  // generate a plan to execute the interpolation
  struct interpolation_plan_element * plan;

  plan = malloc(1 * sizeof(*plan));

  yac_init_interpolation_plan_element_wpointsum ( plan, *(data->num_interpolated_cells),
                                                  src_indices, num_src_per_tgt, weights,
                                                  num_remote_cells_, num_pointsets );

  free(num_remote_cells_);
  free(num_src_per_tgt);
  free(src_indices);
  free(weights);

  return plan;
}

// -----------------------------------------------------------------------------

/**
 * finalising the search for conservative remapping
 *
 * call remove, reorder, generate plan, generate data request,
 * call next method, set_finish_callback
 *
 * in addition it could do the weight correction as well
 *
 */

static void
do_search_final_conserv ( struct search_data_conserv * data ) {

  struct interpolation_method_conserv * method_conserv;
  method_conserv = (struct interpolation_method_conserv*)(data->method_stack[0]);

  // remove all cells that are interpolated by other source processes

  yac_quicksort_index ( data->target_cell_ids_to_remove,
                    data->num_target_cell_ids_to_remove, NULL );

  remove_cells ( data->num_target_cell_ids_to_remove,
                 data->target_cell_ids_to_remove, data );

  free(data->target_cell_ids_to_remove);

  // if not all cells can be interpolated

  unsigned num_cells = data->num_cells;
  unsigned num_interpolated_cells = data->num_interpolated_cells[0];

  if (num_interpolated_cells != num_cells) {

    // The criteria for cell_can_be_interpolated needs to be revised
    // Note that masks of source cells are not yet communicated to the master
    // nor are they used or stored.

    // move cells to the end for further treatment that cannot be interpolated

    reorder_cells ( data->global_ids, data->x_coordinates, data->y_coordinates,
                    data->weight_provenance, num_cells, num_interpolated_cells );
  }

  //--------------------------------------------------------------------
  // Check for source cells required for the interpolation that are not
  // in the subdomain, which the target cells are assigned to. (can be
  // on another local subdomain or on a remote source domain)
  //--------------------------------------------------------------------

  int * remote_cells = NULL;
  int * remote_cell_ranks = NULL;
  unsigned num_remote_cells = 0;

  check_for_remote_cells ( &num_remote_cells, &remote_cells, &remote_cell_ranks, data );

  if (!method_conserv->partial_coverage &&
      (method_conserv->normalisation == CONSERV_DESTAREA)) {

    // normalise the partial areas (weights) of the intersections with the target area
    // in other words: minimise the error and take care that the sum of weights is as
    // close as possible to 1.0

    unsigned nmaxWeights = 0;
    double * weight = NULL;
    unsigned i, n;

    for ( n = 0; n < data->num_cells; ++n ) {

      if (data->weight_provenance[n].cell_can_be_interpolated) {
         ENSURE_ARRAY_SIZE ( weight, nmaxWeights, data->weight_provenance[n].nbr_overlaps );

         for ( i = 0; i < data->weight_provenance[n].nbr_overlaps; ++i )
           weight[i] = data->weight_provenance[n].overlaps[i].weight;

         yac_correct_weights ( data->weight_provenance[n].nbr_overlaps, weight );

         for ( i = 0; i < data->weight_provenance[n].nbr_overlaps; ++i )
            data->weight_provenance[n].overlaps[i].weight = weight[i];
      }
    }
    free(weight);
  }

  //----------------------------------------------------------
  // generate a interpolation plan based on the collected data
  //----------------------------------------------------------

  data->plan_stack[0] = generate_interpolation_plan_conserv ( data,
                                                              remote_cells,
                                                              num_remote_cells );

  // advance interpolation plan stack if an interpolation plan was generated

  if (data->plan_stack[0] != NULL) data->plan_stack++;

  unsigned num_pointsets = method_conserv->num_pointsets;
  unsigned points_idx = method_conserv->points_idx;

  // structure of data_request is given and cannot be modified

  method_conserv->data_request[0] = malloc(num_pointsets *
                                           sizeof(*(method_conserv->data_request[0])));

  method_conserv->data_request[0][points_idx].num_values = num_remote_cells;
  method_conserv->data_request[0][points_idx].global_ids = remote_cells;
  method_conserv->data_request[0][points_idx].ranks = remote_cell_ranks;

  for (unsigned i = 0; i < num_pointsets; ++i) {
    if (i != points_idx) {
      method_conserv->data_request[0][i].num_values = 0;
      method_conserv->data_request[0][i].global_ids = NULL;
      method_conserv->data_request[0][i].ranks = NULL;
    }
  }

  for (unsigned n = 0; n < data->num_cells; ++n ) {
    free(data->weight_provenance[n].servant_ranks);
    free(data->weight_provenance[n].overlaps);
  }

  free(data->weight_provenance);

  data->num_cells -= num_interpolated_cells;

  if (data->x_coordinates != NULL)
    data->x_coordinates += num_interpolated_cells;
  if (data->y_coordinates != NULL)
    data->y_coordinates += num_interpolated_cells;

  // call the do search interpolation_method for the next method

  if (data->method_stack[1] != NULL)
    data->method_stack[1]->vtable->do_search ( data->method_stack+1, data->plan_stack,
                                               data->global_ids+num_interpolated_cells,
                                               data->x_coordinates, data->y_coordinates,
                                               CELL, data->num_cells,
                                               data->num_interpolated_cells+1,
                                               method_conserv->data_request+1);
  free(data);

  if (method_conserv->finish_callback != NULL)
    method_conserv->finish_callback(method_conserv->finish_user_data);
}

// ------------------------------------------------

static int
compare_overlaps(const void * a, const void * b) {

  struct overlap * w_a = (struct overlap *)a;
  struct overlap * w_b = (struct overlap *)b;

  return (w_a->global_cell_id > w_b->global_cell_id) -
         (w_a->global_cell_id < w_b->global_cell_id);
}

static void
do_search_interpolation_method_conserv_continue_2(
  struct search_data_conserv * user_data) {

  struct interpolation_method_conserv * method_conserv;
  method_conserv = (struct interpolation_method_conserv*)(user_data->method_stack[0]);

  // 2nd part
  //
  // once all partial areas have been received by comm_recv_callback_partial_areas
  // the partial area list shall be complete.
  // now we can check whether the source cell information is complete for the
  // target cells. The master ranks send back a list of target cells, which can
  // be processed (are accepted) by the master. Thus, these can be erased on the
  // servant side (see recv_callback_accepts)

  int n, m, i;

  int * sends_rank_list = NULL;
  struct send_buffer * send_msg = NULL;

  unsigned num_sends = 0;
  int sends_rank_list_size = 0;

  user_data->num_interpolated_cells[0] = 0;

  struct weight_provenance * weight_provenance = user_data->weight_provenance;
  struct communicator * comm = yac_get_domain_comm(yac_get_local_domain(method_conserv->search));
  int comm_rank = yac_get_comm_rank(comm);

  for ( n = 0; n < user_data->num_cells; ++n ) {

    if ( weight_provenance[n].master_rank == comm_rank ) {

      for ( m = 0; m < weight_provenance[n].nbr_servant_ranks; ++m ) {

        ENSURE_ARRAY_SIZE ( sends_rank_list, sends_rank_list_size, num_sends+1 );

        insertion_sort_int ( weight_provenance[n].servant_ranks[m],
                             sends_rank_list, &num_sends );
      }
    }
  }

  send_msg = malloc(num_sends * sizeof(*send_msg));

  for ( i = 0; i < num_sends; ++i ) {
    send_msg[i].ibuffer          = NULL;
    send_msg[i].buffer_size      = 0;
    send_msg[i].num_target_cells = 0;
  }

  for ( n = 0; n < user_data->num_cells; ++n ) {

    if ( weight_provenance[n].master_rank == comm_rank ) {

      double target_cell_area_diff = weight_provenance[n].target_cell_area;
      double area_tol = weight_provenance[n].target_cell_area * 1e-8;

      for ( m = 0; m < weight_provenance[n].nbr_overlaps; m++ )
        target_cell_area_diff -= weight_provenance[n].overlaps[m].area;

#ifdef DEBUG_CONSERV
      if ( weight_provenance[n].target_cell_id == -1 ) {
        printf ("cell %d area tol %e area diff %e num overlaps %d target area %e \n",
        weight_provenance[n].target_cell_id,
                  area_tol, target_cell_area_diff,
                  weight_provenance[n].nbr_overlaps,
                  weight_provenance[n].target_cell_area);
        for ( int i = 0; i < weight_provenance[n].nbr_overlaps; ++i )
          printf (" %d : cell id %d has area %.16e \n", i,
                      weight_provenance[n].overlaps[i].global_cell_id,
                      weight_provenance[n].overlaps[i].area );
      }
#endif

      // if the target cell is completely covered by non-masked source cells or
      // partial coverage is allowed
      if ((fabs(target_cell_area_diff) < area_tol) ||
          (method_conserv->partial_coverage &&
           weight_provenance[n].nbr_overlaps > 0)) {

        // reorder overlaps in accordance with increasing source cell id

        unsigned reorder = 0;

        for ( i = 1; i < weight_provenance[n].nbr_overlaps; ++i )
          if ( weight_provenance[n].overlaps[i].global_cell_id <
               weight_provenance[n].overlaps[i-1].global_cell_id ) reorder = 1;

        if ( reorder == 1 )
          qsort(weight_provenance[n].overlaps, weight_provenance[n].nbr_overlaps,
                sizeof(*weight_provenance[n].overlaps), compare_overlaps);

        double norm_factor;

        switch (method_conserv->normalisation) {

          case(CONSERV_DESTAREA):
            norm_factor = 1.0 / weight_provenance[n].target_cell_area;
            break;
          case(CONSERV_FRACAREA):
          {
            double fracarea = 0;
            for ( m = 0; m < weight_provenance[n].nbr_overlaps; m++ )
              fracarea += weight_provenance[n].overlaps[m].area;
            norm_factor = 1.0 / fracarea;
            break;
          }
          default:
            norm_factor = 0.0;
            yac_internal_abort_message("ERROR: invalid normalisation option in conservative"
                                       " remapping\n", __FILE__, __LINE__);
        }

        // compute the weights
        for ( m = 0; m < weight_provenance[n].nbr_overlaps; m++ )
          weight_provenance[n].overlaps[m].weight =
            weight_provenance[n].overlaps[m].area * norm_factor;

        user_data->num_interpolated_cells[0]++;

        for ( m = 0; m < weight_provenance[n].nbr_servant_ranks; ++m ) {

          int index = index_location ( num_sends,
                                       sends_rank_list,
                                       weight_provenance[n].servant_ranks[m] );

          ENSURE_ARRAY_SIZE ( send_msg[index].ibuffer,
                              send_msg[index].buffer_size,
                              send_msg[index].num_target_cells + 1 );

          // store target cell IDs which can be erased on the servant side in send buffer

          send_msg[index].ibuffer[send_msg[index].num_target_cells++] =
            weight_provenance[n].target_cell_id;

        }

        // flag target cell IDs as being interpolated

        weight_provenance[n].cell_can_be_interpolated = 1;

      } else {

//#define DEBUG_CONSERV
#ifdef DEBUG_CONSERV

        {
          int global_id = -1; // provide global cell id for the cell that you
                              // want to investigate

          if (global_id == -1 ||
              weight_provenance[n].target_cell_id == global_id) {

            struct source_search_results const * source_target_relation =
              yac_get_search_results_for_tgt_cell(
                method_conserv->search, weight_provenance[n].target_cell_id);

            if (source_target_relation != NULL) {

              struct grid_cell tgt_cell;
              yac_init_grid_cell(&tgt_cell);
              yac_get_grid_cell(yac_get_subdomain_grid(
                *source_target_relation->tgt_domain,
                source_target_relation->tgt_subdomain_idx),
                source_target_relation->local_tgt_cell_id, &tgt_cell);

              fprintf(stderr, "Target cell %d could not be interpolated\n"
                      "target area = %.16f\n"
                      "target_cell_area_diff = %.16f\n",
                      weight_provenance[n].target_cell_id,
                      weight_provenance[n].target_cell_area,
                      target_cell_area_diff);
              print_grid_cell(stderr, tgt_cell, "target cell");
              yac_free_grid_cell(&tgt_cell);

              struct grid_cell src_cell;
              yac_init_grid_cell(&src_cell);
              for (int i = 0; i < weight_provenance[n].nbr_overlaps; ++i) {
                fprintf(stderr, "%d global src cell id %d overlap area %.16f\n", i,
                        weight_provenance[n].overlaps[i].global_cell_id,
                        weight_provenance[n].overlaps[i].area);

                if (weight_provenance[n].overlaps[i].subdomain_id != -1) {

                  yac_get_grid_cell(yac_get_subdomain_grid(*method_conserv->domain,
                    weight_provenance[n].overlaps[i].subdomain_id),
                    weight_provenance[n].overlaps[i].local_cell_id, &src_cell);
                  print_grid_cell(stderr, src_cell, "source cell");
                } else {
                  fputs("source cell:\n"
                        "coordinates not available on this process\n", stderr);
                }
              }
              yac_free_grid_cell(&src_cell);
            }
          }
        }
#endif

        weight_provenance[n].cell_can_be_interpolated = 0;

        // compute the weights
        for ( m = 0; m < weight_provenance[n].nbr_overlaps; m++ )
          weight_provenance[n].overlaps[m].weight = -1.0;
      }
    }
  }

  // send target IDs that can be treated

  for ( i = 0; i < num_sends; ++i ) {

    // send data before sending header, this does not give a
    // conflict as messages are identified by their tag.

    if ( send_msg[i].num_target_cells > 0 ) {
      yac_comm_bsend ( send_msg[i].ibuffer, send_msg[i].num_target_cells, COMM_INT, sends_rank_list[i],
                       method_conserv->tags.accepts_exchange, comm );

      free(send_msg[i].ibuffer);

    }

    // send header

    yac_comm_bsend ( &send_msg[i].num_target_cells, 1, COMM_UINT, sends_rank_list[i],
                     method_conserv->tags.header2_exchange, comm );
  }

  free(sends_rank_list);
  free(send_msg);
}

static unsigned
translate_rank(unsigned rank, unsigned * ranks, unsigned num_ranks) {

   for (unsigned i = 0; i < num_ranks; ++i)
      if (rank == ranks[i])
         return i;

   yac_internal_abort_message("ERROR: could not translate rank\n", __FILE__, __LINE__);

   return -1;
}

static void sort_target_points(int * global_ids, double * x_coordinates,
                               double * y_coordinates, unsigned num_cells) {

  // check wether we need to sort the points by global ids
  unsigned i;
  for (i = 1; i < num_cells; ++i) if (global_ids[i-1] > global_ids[i]) break;

  if (i == num_cells) return;

  int * orig_order = malloc(num_cells * sizeof(*orig_order));

  for (unsigned i = 0; i < num_cells; ++i) orig_order[i] = i;

  yac_quicksort_index(global_ids, num_cells, orig_order);

  double * tmp_x_coordinates = malloc(num_cells * sizeof(*tmp_x_coordinates));
  double * tmp_y_coordinates = malloc(num_cells * sizeof(*tmp_y_coordinates));

  memcpy(tmp_x_coordinates, x_coordinates, num_cells * sizeof(*x_coordinates));
  memcpy(tmp_y_coordinates, y_coordinates, num_cells * sizeof(*y_coordinates));

  for (unsigned i = 0; i < num_cells; ++i) {

    x_coordinates[i] = tmp_x_coordinates[orig_order[i]];
    y_coordinates[i] = tmp_y_coordinates[orig_order[i]];
  }

  free(tmp_y_coordinates);
  free(tmp_x_coordinates);
  free(orig_order);
}

static void
init_weight_provenance_data(
  struct search_data_conserv * user_data) {

  struct interpolation_method_conserv * method_conserv =
    (struct interpolation_method_conserv*)(user_data->method_stack[0]);

  struct domain * domain = method_conserv->domain;
  unsigned num_subdomains = yac_get_num_subdomains(domain);

  struct communicator * comm =
    yac_get_domain_comm(yac_get_local_domain(method_conserv->search));

  unsigned comm_rank = yac_get_comm_rank(comm);
  unsigned comm_size = yac_get_comm_size(comm);

  unsigned * domain_to_global_ranks =
    malloc(comm_size * sizeof(*domain_to_global_ranks));

  yac_comm_translate_ranks(yac_get_communicator(method_conserv->search), comm,
                           domain_to_global_ranks);

  struct grid_cell * SourceCell = NULL;
  struct grid_cell  TargetCell;  unsigned nSourceCells;
  unsigned nmaxSourceCells = 0;
  unsigned initialisedSourceCells = 0;
  double * areas = NULL;
  unsigned nmaxAreas = 0;

  struct weight_provenance * weight_provenance = user_data->weight_provenance;
  unsigned num_cells = user_data->num_cells;
  double * x_coordinates = user_data->x_coordinates;
  double * y_coordinates = user_data->y_coordinates;
  int * global_ids = user_data->global_ids;

  yac_init_grid_cell ( &TargetCell );

  // Loop over all target cells

  for (unsigned n = 0; n < num_cells; ++n ) {

    if ((n % 1024) == 0) yac_comm_check(comm);

    weight_provenance[n].target_cell_id = global_ids[n];
    weight_provenance[n].cell_can_be_interpolated = 0;

    struct source_search_results const * source_target_relation =
      yac_get_search_results_for_tgt_cell(method_conserv->search, global_ids[n]);

    // if there are no search results for a requested cell, we do not know how the
    // other source processes handle this cell -> we cannot deal with it
    if (source_target_relation == NULL)
      yac_internal_abort_message(
        "ERROR: there are no search results for a requested cell\n",
        __FILE__, __LINE__);

    // Build target cell

    struct grid * grid =
      yac_get_subdomain_grid(*source_target_relation->tgt_domain,
                             source_target_relation->tgt_subdomain_idx );

    yac_get_grid_cell(grid, source_target_relation->local_tgt_cell_id,
                      &TargetCell);

    // initialise the list of source cells

    nSourceCells = 0;

    for (unsigned i = 0; i < num_subdomains; ++i )
      nSourceCells +=
        source_target_relation->num_overlapping_src_cells_per_subdomain[i];

    ENSURE_ARRAY_SIZE ( SourceCell, nmaxSourceCells, nSourceCells );

    while (initialisedSourceCells < nSourceCells)
      yac_init_grid_cell ( &SourceCell[initialisedSourceCells++] );

    // allocate memory for weights and associated provenance data
    // probably we only have to store provenance info if we do not
    // stay local. If we are reaching the halo we allocated more
    // memory than needed. But doing this already here avoid some
    // additional costly checking in loops further down.

    weight_provenance[n].overlaps =
      malloc(nSourceCells * sizeof(*(weight_provenance[n].overlaps)));

    // loop over subdomains and collect geographical information plus provenance data for each cell

    for (unsigned src_subdomain_idx = 0, i = 0;
         src_subdomain_idx < num_subdomains; ++src_subdomain_idx) {

      struct grid *grid = yac_get_subdomain_grid ( *domain, src_subdomain_idx );

      int const * global_src_ids = yac_get_global_ids ( domain,
                                                        src_subdomain_idx, CELL );
      const unsigned num_overlapping_src_cells =
        source_target_relation->num_overlapping_src_cells_per_subdomain[src_subdomain_idx];
      const unsigned * overlapping_src_cells =
        source_target_relation->overlapping_src_cells[src_subdomain_idx];
      struct mask *** masks = method_conserv->masks;
      unsigned points_idx = method_conserv->points_idx;

      int use_mask = method_conserv->use_source_mask &&
                     (masks != NULL) && (masks[src_subdomain_idx] != NULL) &&
                     (masks[src_subdomain_idx][points_idx] != NULL);

      for (unsigned j = 0; j < num_overlapping_src_cells; ++j ) {

        // Perform mask handling for source cells already here and
        // do not consider halo cells for the weights to avoid duplication

        if (yac_is_halo_value(domain, src_subdomain_idx, CELL,
                              overlapping_src_cells[j], NULL) ||
            (use_mask &&
             masks[src_subdomain_idx][points_idx]->mask[overlapping_src_cells[j]])) {
          nSourceCells--;
          continue;

        }

        // add cell for weight calculation

        yac_get_grid_cell ( grid, overlapping_src_cells[j], &SourceCell[i] );

        weight_provenance[n].overlaps[i].local_cell_id  = overlapping_src_cells[j];
        weight_provenance[n].overlaps[i].subdomain_id   = src_subdomain_idx;
        weight_provenance[n].overlaps[i].global_cell_id = global_src_ids[overlapping_src_cells[j]];
        weight_provenance[n].overlaps[i].rank           = comm_rank;
        ++i;
      }
    }

    // set up prerequisites for communication among source processes
    //
    // check whether we are alone or if we have to share information
    // with other sources for this particular target cell.
    // in case we have to share information we have to determine
    // master and servants for each target cells

    weight_provenance[n].master_rank       = comm_rank;
    weight_provenance[n].nbr_servant_ranks = 0;
    weight_provenance[n].servant_ranks     = NULL;

    if ( source_target_relation->num_other_src_ranks > 0 ) {

      for (unsigned i = 0; i < source_target_relation->num_other_src_ranks; ++i )
        weight_provenance[n].master_rank =
          MIN(weight_provenance[n].master_rank,
              translate_rank(source_target_relation->other_src_ranks[i],
                             domain_to_global_ranks, comm_size));

      weight_provenance[n].nbr_servant_ranks =
        source_target_relation->num_other_src_ranks;
      weight_provenance[n].servant_ranks =
        malloc ( source_target_relation->num_other_src_ranks *
                 sizeof(*(weight_provenance[n].servant_ranks)) );
      for (unsigned i = 0; i < source_target_relation->num_other_src_ranks;
           ++i)
        weight_provenance[n].servant_ranks[i] =
          translate_rank(source_target_relation->other_src_ranks[i],
                         domain_to_global_ranks, comm_size);
    }

    // determine the partial areas. the sequence of returned areas corresponds to the sequence
    // of ingoing SourceCells (and the sequence of source cell ids).

    ENSURE_ARRAY_SIZE ( areas, nmaxAreas, nSourceCells );

/*    compute_overlap_areas ( nSourceCells, SourceCell, TargetCell, areas);*/

    yac_compute_concave_overlap_areas ( nSourceCells, SourceCell, TargetCell,
                                        &x_coordinates[n], &y_coordinates[n], areas);

    double target_cell_area = yac_huiliers_area(TargetCell);

    //! \todo how to we handle the case when the area of the target cell is
    //!       around the area tolerance of the computation
    // if (target_cell_area <= area_tol * 100)
      // abort_message("ERROR: target cell is too small\n", __FILE__, __LINE__);

    // scale area tolerance to target cell area
    weight_provenance[n].target_cell_area = target_cell_area;

    // delete cells that have no overlap with the target cell
    unsigned new_nSourceCells = 0;
    for (unsigned i = 0; i < nSourceCells; ++i) {
      if (areas[i] > 0) {

        if (new_nSourceCells != i)
          weight_provenance[n].overlaps[new_nSourceCells] =
            weight_provenance[n].overlaps[i];

        weight_provenance[n].overlaps[new_nSourceCells].area = areas[i];
        weight_provenance[n].overlaps[new_nSourceCells].weight = 0.0;
        new_nSourceCells++;
      }
    }
    if (new_nSourceCells != nSourceCells)
      weight_provenance[n].overlaps =
        realloc(weight_provenance[n].overlaps,
                new_nSourceCells * sizeof(*(weight_provenance[n].overlaps)));

    // store current size of weight_provenance[n].overlaps

    weight_provenance[n].nbr_overlaps = new_nSourceCells;
  }


  for (unsigned i = 0; i < initialisedSourceCells; ++i )
    yac_free_grid_cell ( &SourceCell[i] );
  yac_free_grid_cell ( &TargetCell );

  free(SourceCell);
  free(domain_to_global_ranks);
  free(areas);
}

static void
do_search_interpolation_method_conserv_continue(
  struct search_data_conserv * user_data) {

  struct interpolation_method_conserv * method_conserv =
    (struct interpolation_method_conserv*)(user_data->method_stack[0]);

  struct communicator * comm =
    yac_get_domain_comm(yac_get_local_domain(method_conserv->search));
  unsigned comm_rank = yac_get_comm_rank(comm);

  // from here on we work with the weight_provenance data, which does not
  // contain any halo cells

  int *receives_rank_list = NULL;
  int *sends_rank_list = NULL;

  unsigned receives_rank_list_size = 0;
  unsigned sends_rank_list_size = 0;

  unsigned num_receives = 0;
  unsigned num_sends = 0;

  struct weight_provenance * weight_provenance = user_data->weight_provenance;
  unsigned num_cells = user_data->num_cells;

  // determine the send and receive operations

  for (unsigned n = 0; n < num_cells; ++n ) {

    if ( weight_provenance[n].nbr_servant_ranks != 0 ) {

      // prepare exchange among source cells for this target.
      // note that master_rank can be different for individual target cells.

      if ( weight_provenance[n].master_rank == comm_rank ) {

        // build/fill receive lists

        for (unsigned i = 0; i < weight_provenance[n].nbr_servant_ranks; ++i) {

          ENSURE_ARRAY_SIZE ( receives_rank_list, receives_rank_list_size,
                              num_receives+1 );

          insertion_sort_int(weight_provenance[n].servant_ranks[i],
                             receives_rank_list, &num_receives);
        }
      } else {

        // build/fill send lists

        ENSURE_ARRAY_SIZE(sends_rank_list, sends_rank_list_size, num_sends+1);

        insertion_sort_int ( weight_provenance[n].master_rank,
                             sends_rank_list, &num_sends );
      }
    }
  }

  // store data that is to be passed to do_search_final_conserv (via
  // recv_callback_partial_areas)
  user_data->num_partial_areas_messages_to_receive = num_receives;
  user_data->num_accepts_messages_to_receive = num_sends;

  // if we do not require any weights from other source processes, we have to call
  // do_search_interpolation_method_conserv_continue_2 here in order for it to not
  // interfere with a call to do_search_final_conserv from recv_callback_accepts
  if (num_receives == 0)
    do_search_interpolation_method_conserv_continue_2(user_data);

  // fill send buffer

  struct send_buffer * send_msg = NULL;

  if ( num_sends > 0 ) {

    send_msg = malloc ( num_sends * sizeof(*send_msg) );

    for (unsigned i = 0; i < num_sends; ++i ) {
      send_msg[i].num_target_cells   = 0;
      send_msg[i].buffer_size        = 0;
      send_msg[i].buffer_array_size  = 0;
      send_msg[i].position           = 0;
      send_msg[i].cbuffer            = NULL;
    }

    // determine buffer size

    for (unsigned n = 0; n < num_cells; ++n ) {

      if ( ( weight_provenance[n].master_rank != comm_rank ) &&
           ( weight_provenance[n].nbr_servant_ranks < 0    ) ) {

        int index = index_location ( num_sends, sends_rank_list,
                                     weight_provenance[n].master_rank );

        send_msg[index].num_target_cells++;

        int nbr_overlaps = weight_provenance[n].nbr_overlaps;

        send_msg[index].buffer_size += yac_comm_pack_size(2, COMM_INT, comm);
        send_msg[index].buffer_size += yac_comm_pack_size(nbr_overlaps, COMM_INT, comm);
        send_msg[index].buffer_size += yac_comm_pack_size(nbr_overlaps, COMM_DBLE, comm);
      }
    }

    // allocate sufficient memory

    for (unsigned i = 0; i < num_sends; ++i)
        ENSURE_ARRAY_SIZE(send_msg[i].cbuffer, send_msg[i].buffer_array_size, send_msg[i].buffer_size);

    // store data in send buffer

    for (unsigned n = 0; n < num_cells; ++n) {

      if ( ( weight_provenance[n].master_rank != comm_rank ) &&
           ( weight_provenance[n].nbr_servant_ranks < 0    ) ) {

        int index = index_location ( num_sends, sends_rank_list,
                                     weight_provenance[n].master_rank );

        yac_comm_pack(&(weight_provenance[n].target_cell_id), 1, COMM_INT,
                      send_msg[index].cbuffer, send_msg[index].buffer_size,
                      &(send_msg[index].position), comm);

        yac_comm_pack(&(weight_provenance[n].nbr_overlaps), 1, COMM_INT,
                      send_msg[index].cbuffer, send_msg[index].buffer_size,
                      &(send_msg[index].position), comm);

        for (unsigned i = 0; i < weight_provenance[n].nbr_overlaps; ++i ) {
            yac_comm_pack(&(weight_provenance[n].overlaps[i].global_cell_id), 1, COMM_INT,
                          send_msg[index].cbuffer, send_msg[index].buffer_size,
                          &(send_msg[index].position), comm);
            yac_comm_pack(&(weight_provenance[n].overlaps[i].area), 1, COMM_DBLE,
                          send_msg[index].cbuffer, send_msg[index].buffer_size,
                          &(send_msg[index].position), comm);
        }
      }
    }

    // We expect an answer from the master rank in the second part in any case

    yac_comm_recv_callback ( 1, COMM_UINT, COMM_ANYSOURCE,
                             method_conserv->tags.header2_exchange,
                             comm, user_data, recv_callback_accepts, NULL,
                             MSG_PRIO_HIGH );
  }

  for (unsigned i = 0; i < num_sends; ++i) {

    // send data before sending header, this does not give a
    // conflict as messages are identified by their tag.

    if ( send_msg[i].buffer_size > 0 ) {

      yac_comm_bsend(send_msg[i].cbuffer, send_msg[i].buffer_size, COMM_PACKED,
                     sends_rank_list[i],
                     method_conserv->tags.partial_areas_exchange, comm );

      free(send_msg[i].cbuffer);
    }

    // send header

    {
      unsigned nbr_items[2] = {send_msg[i].num_target_cells,
                               send_msg[i].buffer_size};
      yac_comm_bsend(nbr_items, 2, COMM_UINT, sends_rank_list[i],
                     method_conserv->tags.header1_exchange, comm );
    }
  }

  free(sends_rank_list);
  free(receives_rank_list);
  free(send_msg);

  // receive first header to get the length of data, all
  // following header messages are received inside recv_callback_partial_areas
  // for a well defined source use receives_rank_list[0], note
  // that these are currently freed and not available.

  if ( num_receives > 0 )
    yac_comm_recv_callback ( 2, COMM_UINT, COMM_ANYSOURCE,
                             method_conserv->tags.header1_exchange,
                             comm, user_data, recv_callback_partial_areas, NULL,
                             MSG_PRIO_HIGH );

  // do_search_final_conserv and do_search_interpolation_method_conserv_continue_2
  // are called by the callback function in case
  // messages have to be processed from other source ranks. If no messages
  // are expected we call do_search_final_conserv here.

  if ( num_receives == 0 && num_sends == 0 )
    do_search_final_conserv ( user_data );
}

static void
alltoallv_master_ranks_callback(
  struct communicator * comm, struct comm_message msg, void * user_data ) {

  struct search_data_conserv * search_data =
    (struct search_data_conserv *)user_data;

  struct interpolation_method_conserv * method_conserv =
    (struct interpolation_method_conserv*)(search_data->method_stack[0]);

  unsigned comm_rank = yac_get_comm_rank(comm);

  int * recv_buffer = (int*)(msg.buffer);
  unsigned num_tgt_points = msg.count / 2;

  int * tgt_global_id = malloc(num_tgt_points * sizeof(*tgt_global_id));
  unsigned * master_ranks = malloc(num_tgt_points * sizeof(*master_ranks));

  for (unsigned i = 0; i < num_tgt_points; ++i) {

    tgt_global_id[i] = recv_buffer[2*i+0];
    master_ranks[i] = recv_buffer[2*i+1];
  }

  yac_quicksort_index(tgt_global_id, num_tgt_points, (int*)master_ranks);

  struct weight_provenance * weight_provenance = search_data->weight_provenance;
  unsigned num_cells = search_data->num_cells;

  for (unsigned i = 0, j = 0; i < num_tgt_points; ++i) {

    // find the respective entry in the weight provenance data
    while ((j < num_cells) &&
           (weight_provenance[j].target_cell_id < tgt_global_id[i])) ++j;

    if ((j >= num_cells) ||
        (weight_provenance[j].target_cell_id != tgt_global_id[i]))
      yac_internal_abort_message(
        "ERROR(alltoallv_master_ranks_callback): could not find matching "
        "global target cell id in local weight provenance data",
        __FILE__, __LINE__);

    weight_provenance[j].master_rank = master_ranks[i];
  }

  free(master_ranks);
  free(tgt_global_id);

  for (unsigned n = 0; n < num_cells; ++n ) {
    if ( comm_rank != weight_provenance[n].master_rank ) {
      // servants get negative numbers, used to distinguish later
      weight_provenance[n].nbr_servant_ranks =
        - weight_provenance[n].nbr_servant_ranks;
      free(weight_provenance[n].servant_ranks);
      weight_provenance[n].servant_ranks = NULL;
    }
  }

  // now servants and master are determined
  // continue with the interpolation
  do_search_interpolation_method_conserv_continue(user_data);
}

static int
compare_uncertain_tgt_global_id(const void * a, const void * b) {

  struct temp_uncertain_tgt * t_a = (struct temp_uncertain_tgt *)a;
  struct temp_uncertain_tgt * t_b = (struct temp_uncertain_tgt *)b;

  int ret;
  if (ret = (t_a->global_id > t_b->global_id) -
            (t_a->global_id < t_b->global_id)) return ret;
  if (ret = (t_a->num_src_points < t_b->num_src_points) -
            (t_a->num_src_points > t_b->num_src_points)) return ret;
  return (t_a->src_rank > t_b->src_rank) - (t_a->src_rank < t_b->src_rank);
}

static int
compare_uncertain_tgt_src_rank(const void * a, const void * b) {

  struct temp_uncertain_tgt * t_a = (struct temp_uncertain_tgt *)a;
  struct temp_uncertain_tgt * t_b = (struct temp_uncertain_tgt *)b;

  int ret;
  if (ret = (t_a->src_rank > t_b->src_rank) -
            (t_a->src_rank < t_b->src_rank)) return ret;
  return (t_a->global_id > t_b->global_id) - (t_a->global_id < t_b->global_id);
}

static void
alltoallv_uncertain_tgt_callback(
  struct communicator * comm, struct comm_message msg, void * user_data ) {

  struct search_data_conserv * search_data =
    (struct search_data_conserv *)user_data;

  struct interpolation_method_conserv * method_conserv =
    (struct interpolation_method_conserv*)(search_data->method_stack[0]);

  unsigned comm_rank = yac_get_comm_rank(comm);
  unsigned comm_size = yac_get_comm_size(comm);

  int * recv_buffer = (int*)(msg.buffer);
  unsigned offset = 0;

  struct temp_uncertain_tgt * tgt_points =
    malloc((msg.count / 2) * sizeof(*tgt_points));

  unsigned num_received_tgt = 0;

  while (offset < msg.count) {

    unsigned curr_src_rank = recv_buffer[offset++];
    unsigned curr_num_tgt = recv_buffer[offset++];

    for (unsigned i = 0; i < curr_num_tgt; ++i) {

      tgt_points[num_received_tgt + i].src_rank = curr_src_rank;
      tgt_points[num_received_tgt + i].global_id = recv_buffer[offset++];
      tgt_points[num_received_tgt + i].num_src_points = recv_buffer[offset++];
    }
    num_received_tgt += curr_num_tgt;
  }

  // sort the received target points by their global ids, then their number of
  // source points (descending order) followed by their source rank
  qsort(tgt_points, num_received_tgt, sizeof(*tgt_points),
        compare_uncertain_tgt_global_id);

  struct weight_provenance * weight_provenance = search_data->weight_provenance;
  unsigned num_cells = search_data->num_cells;

  // determine the master rank
  for (unsigned i = 0, j = 0; i < num_received_tgt; ++i) {

    // if we have already determined the master rank for this global id
    if ((i > 0) && (tgt_points[i-1].global_id == tgt_points[i].global_id)) {
      tgt_points[i].master_rank = tgt_points[i-1].master_rank;
      continue;
    }

    // find the respective entry in the weight provenance data
    while ((j < num_cells) &&
           (weight_provenance[j].target_cell_id < tgt_points[i].global_id)) ++j;

    if ((j >= num_cells) ||
        (weight_provenance[j].target_cell_id != tgt_points[i].global_id))
      yac_internal_abort_message(
        "ERROR(alltoallv_uncertain_tgt_callback): could not find matching "
        "global target cell id in local weight provenance data",
        __FILE__, __LINE__);

    // if the local process is not the master for the current target point
    if (tgt_points[i].num_src_points > weight_provenance[j].nbr_overlaps) {
      weight_provenance[j].master_rank = tgt_points[i].src_rank;
      tgt_points[i].master_rank = tgt_points[i].src_rank;
    } else
      tgt_points[i].master_rank = comm_rank;
  }

  // sort the received target points by source rank followd by the global id
  qsort(tgt_points, num_received_tgt, sizeof(*tgt_points),
        compare_uncertain_tgt_src_rank);

  // return results
  int * send_list = malloc(2 * num_received_tgt * sizeof(*send_list));
  unsigned * send_size_per_rank =
    calloc(comm_size, sizeof(*send_size_per_rank));

  for (unsigned i = 0; i < num_received_tgt; ++i) {

    send_list[2*i+0] = tgt_points[i].global_id;
    send_list[2*i+1] = tgt_points[i].master_rank;
    send_size_per_rank[tgt_points[i].src_rank] += 2;
  }

  yac_comm_alltoallv_callback(send_list, send_size_per_rank, COMM_INT,
                              method_conserv->tags.alltoallv_master_ranks,
                              comm, user_data, alltoallv_master_ranks_callback,
                              MSG_PRIO_HIGH);

  free(send_list);
  free(send_size_per_rank);
  free(tgt_points);
}

static void
determine_master_for_tgt_cells(struct search_data_conserv * user_data) {

  struct interpolation_method_conserv * method_conserv =
    (struct interpolation_method_conserv*)(user_data->method_stack[0]);

  struct communicator * comm =
    yac_get_domain_comm(yac_get_local_domain(method_conserv->search));
  unsigned comm_rank = yac_get_comm_rank(comm);
  unsigned comm_size = yac_get_comm_size(comm);

  struct weight_provenance * weight_provenance = user_data->weight_provenance;
  unsigned num_cells = user_data->num_cells;

  unsigned send_list_size = 0;
  unsigned * send_size_per_rank =
    calloc(comm_size, sizeof(send_size_per_rank));
  unsigned * send_list_offsets =
    malloc(comm_size * sizeof(send_list_offsets));

  // count number of target cells that are assigned to more than one source
  // process
  for (unsigned i = 0; i < num_cells; ++i) {
    if (weight_provenance[i].master_rank != comm_rank) {
      send_list_size++;
      send_size_per_rank[weight_provenance[i].master_rank] += 2;
    }
  }
  for (unsigned i = 0; i < comm_size; ++i) {
    if (send_size_per_rank[i] > 0) {
      send_list_size++;
      send_size_per_rank[i] += 2; // for the local rank
    }
  }
  send_list_size *= 2;
  unsigned accum = 0;
  for (unsigned i = 0; i < comm_size; ++i) {
    send_list_offsets[i] = accum;
    accum += send_size_per_rank[i];
  }

  int * send_list = malloc(2 * send_list_size * sizeof(*send_list));
  // set local rank and number of target points in front of each message
  for (unsigned i = 0; i < comm_size; ++i) {
    if (send_size_per_rank[i] > 0) {

      unsigned offset = send_list_offsets[i];

      if (offset == -1) fputs("ups\n", stderr);

      send_list_offsets[i] += 2;
      send_list[offset+0] = comm_rank;
      send_list[offset+1] = send_size_per_rank[i] / 2 - 1;
    }
  }

  for (unsigned i = 0; i < num_cells; ++i) {
    if (weight_provenance[i].master_rank != comm_rank) {

      unsigned offset = send_list_offsets[weight_provenance[i].master_rank];
      send_list_offsets[weight_provenance[i].master_rank] += 2;
      send_list[offset+0] = weight_provenance[i].target_cell_id;
      send_list[offset+1] = weight_provenance[i].nbr_overlaps;
    }
  }

  // send uncertain target points to current masters
  yac_comm_alltoallv_callback(send_list, send_size_per_rank, COMM_INT,
                              method_conserv->tags.alltoallv_uncertain_tgt,
                              comm, user_data, alltoallv_uncertain_tgt_callback,
                              MSG_PRIO_HIGH);

  free(send_list);
  free(send_list_offsets);
  free(send_size_per_rank);
}

// -----------------------------------------------------------------------------

static void
do_search_interpolation_method_conserv ( struct interpolation_method ** method,
                                         struct interpolation_plan_element ** plan_stack,
                                         int * global_ids, double * x_coordinates,
                                         double * y_coordinates,
                                         enum yac_location tgt_location,
                                         unsigned num_cells,
                                         unsigned * num_interpolated_cells,
                                         struct interpolation_method_data_request ** data_request ) {

  if (tgt_location != CELL)
    yac_internal_abort_message("ERROR(do_search_interpolation_method_conserv): "
                               "invalid target location (must be CELL)\n",
                               __FILE__, __LINE__);

  struct interpolation_method_conserv * method_conserv;
  method_conserv = (struct interpolation_method_conserv*)(method[0]);
  method_conserv->data_request = data_request;
  num_interpolated_cells[0] = 0;

  struct search_data_conserv * user_data = malloc(1 * sizeof(*user_data));

  // sort target points by global ids
  sort_target_points(global_ids, x_coordinates, y_coordinates, num_cells);

  // store data that is to be passed to do_search_final_conserv (via
  // recv_callback_partial_areas)
  user_data->method_stack = method;
  user_data->plan_stack = plan_stack;
  user_data->global_ids = global_ids;
  user_data->x_coordinates = x_coordinates;
  user_data->y_coordinates = y_coordinates;
  user_data->num_cells = num_cells;
  user_data->num_interpolated_cells = num_interpolated_cells;
  user_data->num_partial_areas_messages_to_receive = 0;
  user_data->num_accepts_messages_to_receive = 0;
  user_data->weight_provenance =
    malloc(num_cells * sizeof(*(user_data->weight_provenance)));
  user_data->target_cell_ids_to_remove = NULL;
  user_data->num_target_cell_ids_to_remove = 0;

  init_weight_provenance_data(user_data);

  determine_master_for_tgt_cells(user_data);
}

// -----------------------------------------------------------------------------

static void
enddef_interpolation_method_conserv ( struct interpolation_method ** method ) {

  struct interpolation_method_conserv * method_conserv;
  method_conserv = (struct interpolation_method_conserv*)(method[0]);

  struct communicator * comm =
    yac_get_domain_comm(yac_get_local_domain(method_conserv->search));

  while (yac_comm_wait(comm, method_conserv->tags.alltoallv_uncertain_tgt) ||
         yac_comm_wait(comm, method_conserv->tags.alltoallv_master_ranks) ||
         yac_comm_wait(comm, method_conserv->tags.header1_exchange) ||
         yac_comm_wait(comm, method_conserv->tags.partial_areas_exchange) ||
         yac_comm_wait(comm, method_conserv->tags.header2_exchange) ||
         yac_comm_wait(comm, method_conserv->tags.accepts_exchange));

  if (method[1] != NULL)
    method[1]->vtable->enddef(method + 1);
}

// -----------------------------------------------------------------------------

static void
get_source_exchange_data_interpolation_method_conserv (
  struct interpolation_method ** method,
  struct interpolation_method_data_request *** data_request ) {

  struct interpolation_method_conserv * method_conserv;

  method_conserv = (struct interpolation_method_conserv*)(method[0]);

  *data_request = method_conserv->data_request;
  method_conserv->data_request = NULL;
}

// -----------------------------------------------------------------------------

static void
delete_interpolation_method_conserv(struct interpolation_method * method) {
  free(method);
}

// -----------------------------------------------------------------------------

static struct global_search *
get_global_search(struct interpolation_method * method) {

  return ((struct interpolation_method_conserv*)(method))->search;
}

// -----------------------------------------------------------------------------

static void
recv_callback_partial_areas_cont ( struct communicator * comm,
                                   struct comm_message msg,
                                   void * user_data ) {

  struct temp_recv_partial_areas_data * temp_data =
    (struct temp_recv_partial_areas_data *)user_data;

  unsigned num_target_cells   = temp_data->num_target_cells;
  struct search_data_conserv * search_data = temp_data->search_data;
  struct interpolation_method_conserv * method_conserv =
    (struct interpolation_method_conserv*)(search_data->method_stack[0]);

  free(temp_data);

  int position = 0;
  unsigned offset = 0;

  for (unsigned n = 0; n < num_target_cells; ++n ) {

    // stupid n^2 search to find matching m
    // something better is needed, probably a sorted lists

    int target_id;
    int nbr_overlaps;

    yac_comm_unpack(msg.buffer, msg.count, &position,
                    &(target_id), 1, COMM_INT, comm);

    yac_comm_unpack(msg.buffer, msg.count, &position,
                    &(nbr_overlaps), 1, COMM_INT, comm);

    unsigned m;

    for (m = 0; m < search_data->num_cells; ++m)
      if ( target_id == search_data->weight_provenance[m].target_cell_id ) break;

    if ( m == search_data->num_cells ) continue;

    // increase size of search_data->weight_provenance[m].overlaps

    search_data->weight_provenance[m].overlaps =
      realloc ( search_data->weight_provenance[m].overlaps,
              ( search_data->weight_provenance[m].nbr_overlaps + nbr_overlaps ) *
                sizeof(*(search_data->weight_provenance[m].overlaps)) );

    offset = search_data->weight_provenance[m].nbr_overlaps;
    search_data->weight_provenance[m].nbr_overlaps += nbr_overlaps;

    for (unsigned i = 0; i < nbr_overlaps; ++i ) {
      search_data->weight_provenance[m].overlaps[offset+i].subdomain_id = -1;
      search_data->weight_provenance[m].overlaps[offset+i].local_cell_id = -1;
      search_data->weight_provenance[m].overlaps[offset+i].rank = msg.src_rank;

      yac_comm_unpack(msg.buffer, msg.count, &position,
                      &(search_data->weight_provenance[m].overlaps[offset+i].global_cell_id),
                      1, COMM_INT, comm);

      yac_comm_unpack(msg.buffer, msg.count, &position,
                      &(search_data->weight_provenance[m].overlaps[offset+i].area),
                      1, COMM_DBLE, comm);
    }
  }

  // send the next header message

  // Even though the following part looks rather simple, you have to be really
  // careful when changing it!
  // do_search_final_conserv has to be called once all accept messages and all
  // partial_area messages have been received. It is only allowed to be called
  // ONCE by EITHER recv_callback_accepts or recv_callback_partial_areas. It is
  // possible that both routines are on the stack. So you have to ensure, that
  // only one call on the stack sees:
  // "search_data->num_accepts_messages_to_receive == 0 &&
  //  search_data->num_partial_areas_messages_to_receive == 0"
  // On the other hand you also have to ensure that when calling
  // comm_recv_callback the respective message counter
  // (num_accepts_messages_to_receive or num_partial_areas_messages_to_receive)
  // is correct. Because otherwise comm_recv_callback will be called to often,
  // which causes a deadlock in enddef_interpolation_method_conserv.

  if ( search_data->num_partial_areas_messages_to_receive > 1 ) {

    search_data->num_partial_areas_messages_to_receive--;

    yac_comm_recv_callback ( 2, COMM_UINT, COMM_ANYSOURCE,
                             method_conserv->tags.header1_exchange,
                             comm, search_data, recv_callback_partial_areas,
                             NULL, MSG_PRIO_HIGH );

  } else {

    // num_partial_areas_messages_to_receive must be decreased after
    // do_search_interpolation_method_conserv_continue_2 was called, because
    // otherwise it would be possible that recv_callback_accepts might call
    // do_search_final_conserv before do_search_interpolation_method_conserv_continue_2
    // was finished, which is mandatory

    do_search_interpolation_method_conserv_continue_2 ( search_data );

    search_data->num_partial_areas_messages_to_receive--;

    if ( search_data->num_accepts_messages_to_receive == 0 &&
         search_data->num_partial_areas_messages_to_receive == 0 )
      do_search_final_conserv ( search_data );
  }
}

// -----------------------------------------------------------------------------

static void
recv_callback_partial_areas ( struct communicator * comm,
                              struct comm_message msg,
                              void * user_data ) {

  unsigned packed_buffer_size = ((unsigned *)(msg.buffer))[1];

  struct temp_recv_partial_areas_data * temp_data =
    malloc(1 * sizeof(*temp_data));

  temp_data->num_target_cells = ((unsigned *)(msg.buffer))[0];
  temp_data->search_data = user_data;
  struct interpolation_method_conserv * method_conserv =
    (struct interpolation_method_conserv*)(temp_data->search_data->method_stack[0]);

  if ( packed_buffer_size > 0 ) {

    // receive owners (ranks) of source cell IDs, source cell IDs, and number of vertices per source cell
    yac_comm_recv_callback(packed_buffer_size, COMM_PACKED, msg.src_rank,
                           method_conserv->tags.partial_areas_exchange, comm,
                           temp_data, recv_callback_partial_areas_cont, NULL,
                           MSG_PRIO_HIGH);
  } else {

    struct comm_message temp_msg = msg;

    temp_msg.buffer = malloc(0);
    temp_msg.type = COMM_PACKED;
    temp_msg.count = 0;
    temp_msg.tag = method_conserv->tags.partial_areas_exchange;
    recv_callback_partial_areas_cont(comm, temp_msg, temp_data);
    free(temp_msg.buffer);
  }
}

// -----------------------------------------------------------------------------

static void
recv_callback_accepts_cont ( struct communicator * comm,
                        struct comm_message message,
                        void * user_data ) {

  struct search_data_conserv * search_data =
    (struct search_data_conserv *)user_data;
  struct interpolation_method_conserv * method_conserv =
    (struct interpolation_method_conserv*)(search_data->method_stack[0]);

  unsigned num_target_cells = message.count;

  // copy list of target cell IDs that can be processed by the master

  if ( num_target_cells > 0 ) {

    search_data->target_cell_ids_to_remove =
      realloc(search_data->target_cell_ids_to_remove,
              (search_data->num_target_cell_ids_to_remove + num_target_cells) *
              sizeof(*(search_data->target_cell_ids_to_remove)));

    memcpy(search_data->target_cell_ids_to_remove +
           search_data->num_target_cell_ids_to_remove, message.buffer,
           num_target_cells * sizeof(*(search_data->target_cell_ids_to_remove)));

    search_data->num_target_cell_ids_to_remove += num_target_cells;
  }

  // send the next header message

  // Even though the following part looks rather simple, you have to be really
  // careful when changing it!
  // do_search_final_conserv has to be called once all accept messages and all
  // partial_area messages have been received. It is only allowed to be called
  // ONCE by EITHER recv_callback_accepts or recv_callback_partial_areas. It is
  // possible that both routines are on the stack. So you have to ensure, that
  // only one call on the stack sees:
  // "search_data->num_accepts_messages_to_receive == 0 &&
  //  search_data->num_partial_areas_messages_to_receive == 0"
  // On the other hand you also have to ensure that when calling
  // comm_recv_callback the respective message counter
  // (num_accepts_messages_to_receive or num_partial_areas_messages_to_receive)
  // is correct. Because otherwise comm_recv_callback will be called to often,
  // which causes a deadlock in enddef_interpolation_method_conserv.

  search_data->num_accepts_messages_to_receive--;

  if ( search_data->num_accepts_messages_to_receive > 0 ) {

    yac_comm_recv_callback ( 1, COMM_UINT, COMM_ANYSOURCE,
                             method_conserv->tags.header2_exchange,
                             comm, user_data, recv_callback_accepts, NULL,
                             MSG_PRIO_HIGH );

  } else if ( search_data->num_partial_areas_messages_to_receive == 0) {

    do_search_final_conserv ( search_data );
  }
}

// -----------------------------------------------------------------------------

static void
recv_callback_accepts ( struct communicator * comm,
                        struct comm_message message,
                        void * user_data ) {

  struct search_data_conserv * search_data =
    (struct search_data_conserv *)user_data;
  struct interpolation_method_conserv * method_conserv =
    (struct interpolation_method_conserv*)(search_data->method_stack[0]);
  unsigned num_target_cells = *(unsigned *)(message.buffer);

  // receive list of target cell IDs that can be processed by the master

  if ( num_target_cells > 0 ) {

    yac_comm_recv_callback(num_target_cells, COMM_INT, message.src_rank,
                           method_conserv->tags.accepts_exchange, comm,
                           user_data, recv_callback_accepts_cont, NULL,
                           MSG_PRIO_HIGH);
  } else {

    struct comm_message temp_msg = message;

    temp_msg.buffer = malloc(0);
    temp_msg.type = COMM_INT;
    temp_msg.count = 0;
    temp_msg.tag = method_conserv->tags.accepts_exchange;
    recv_callback_accepts_cont(comm, temp_msg, user_data);
    free(temp_msg.buffer);
  }
}

// -----------------------------------------------------------------------------

static void
set_finish_callback ( struct interpolation_method * method,
                      void (*callback)(void * user_data),
                      void * user_data ) {

  struct interpolation_method_conserv * method_conserv;

  method_conserv = (struct interpolation_method_conserv*)method;

  method_conserv->finish_callback = callback;
  method_conserv->finish_user_data = user_data;
}
