/**
 * @file interpolation_method_nnn.c
 * @brief implementation of the n-Nearest Neighbour interpolation method
 *
 * @copyright Copyright  (C)  2017 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "interpolation_method.h"
#include "interpolation_method_nnn.h"
#include "interpolation_plan.h"
#include "grid_search.h"
#include "global_search.h"
#include "ensure_array_size.h"
#include "utils.h"
#include "geometry.h"
#include "grid.h"
#include "sphere_part.h"

static void
do_search_interpolation_method_nnn(struct interpolation_method ** method,
                                   struct interpolation_plan_element ** plan_stack,
                                   int * global_ids, double * coordinates_x,
                                   double * coordinates_y,
                                   enum yac_location tgt_location,
                                   unsigned num_points,
                                   unsigned * num_interpolated_points,
                                   struct interpolation_method_data_request ** data_request);
static void
enddef_interpolation_method_nnn(struct interpolation_method ** method);
static void
get_source_exchange_data_interpolation_method_nnn(
   struct interpolation_method ** method,
   struct interpolation_method_data_request *** data_request);
static void
delete_interpolation_method_nnn(struct interpolation_method * method);
static struct global_search *
get_global_search(struct interpolation_method * method);
static void
set_finish_callback(struct interpolation_method * method,
                    void (*callback)(void * user_data),
                    void * user_data);

static struct interpolation_method_vtable
   interpolation_method_nnn_vtable = {
      .do_search                = do_search_interpolation_method_nnn,
      .enddef                   = enddef_interpolation_method_nnn,
      .get_source_exchange_data = get_source_exchange_data_interpolation_method_nnn,
      .delete                   = delete_interpolation_method_nnn,
      .get_global_search        = get_global_search,
      .set_finish_callback      = set_finish_callback};


struct interpolation_method_nnn_tags {

   unsigned results_or_search_request_header; //!< tag for messages that contain the header for a search request or for an answer to a search request
   unsigned search_request;           //!< tag for messages that contain the data for a search request
   unsigned results;                  //!< tag for messages that contain the results of a search request
   unsigned finished;                 //!< tag for messages that is send to inform other processes that a process has finished all its local work
   unsigned global_id_exchange_1[2];  //!< tags for the first global id exchange in \ref do_search_interpolation_method_nnn
   unsigned global_id_exchange_2[2];  //!< tags for the second global id exchange in \ref do_search_continue_nnn
   unsigned exchange_results[2];      //!< tags for exchanging which target points are actually interpolated
};

struct interpolation_method_nnn {

   struct interpolation_method_vtable * vtable;
   void (*finish_callback)(void * user_data); //!< callback routine that is called once the interpolation is finished
   void * finish_user_data;                   //!< user data that is passed to finish_callback
   struct global_search * search;             //!< global search object used a the basis for the interpolation
   struct domain * domain;
   struct points *** points;
   struct mask *** masks;
   unsigned num_pointsets;
   unsigned distance_weighted;                //!< flag to switch on and off weighting by distance
   unsigned n;                                //!< number of nearest neighbours per target point
   unsigned points_idx;                       //!< index of the points set of field that is to be used for the interpolation
   struct interpolation_method_data_request ** data_request;
   struct interpolation_method_nnn_tags tags;
   unsigned use_source_mask;
   unsigned num_domain_neighbours;            //!< number of neighbour processes
   unsigned * domain_neighbour_ranks;         //!< ranks of all neighbour processes
   struct point_sphere_part_search * non_halo_point_search;
   unsigned * non_halo_local_ids;
   unsigned * num_non_halo_local_ids_per_subdomain_prefix;
   struct point_sphere_part_search ** halo_point_search;
   unsigned finished;
};

enum interp_flag {
  LOCAL = 0,  //!< interpolated on the local process
  NOT = 1,    //!< not assigned to any process yet
  REMOTE = 2, //!< interpolated by a remote process
};

/**
 * contain the information for on source point that was found for a target point
 */
struct tgt_point_results {

   double cos_angle; //!< cosine of angle between the source and target point
   int global_id;    //!< global id of the source point

   int is_local;    //!< true if the point is available in the local domain

   union {

      struct {
         unsigned local_id;
         unsigned subdomain_id;
      } local;

      struct {
         int rank;
      } remote;
   } data;          //!< specific information on the source point (depends on
                    //!< whether the point is locally available or not)
};

/**
 * contains all temporary search specific data for one target point
 */
struct tgt_point_search_data {

   int global_id;                 //!< global id of the target point
   enum interp_flag flag;
   double coordinate_x, coordinate_y;

   double min_cos_angle; //!< minimum cosine angle of the results

   struct tgt_point_results * results; //!< the source points found for the target point
   unsigned num_found_points;          //!< number of found source points
};

struct search_data {

   struct tgt_point_search_data * tgt_point_data;
   unsigned num_tgt_points;           //!< total number of entries in tgt_point_data
   unsigned num_tgt_points_to_interpolate; //!< number of tgt_point_data entries
                                           //!< that are marked as LOCAL

   unsigned num_open_remote_requests; //!< number for forwarded search requests that
                                      //!< that have not yet been answered
};

struct temp_search_request_data {

  struct intermediate_method_data * data;
  unsigned num_tgt_points;
};

#define RESULTS_MSG_TYPE 0
#define SEARCH_REQUEST_MSG_TYPE 1

struct request_record {

  unsigned rank; //!< rank to which (a) request(s) have been sent
  int * global_ids; //!< global ids of target points that have been sent
                    //!< to this rank
  unsigned global_ids_array_size;
  unsigned num_global_ids; //!< number of entries in sent_global_ids
};

//! collection of data that is required through out the nearest neighbourhood search
struct intermediate_method_data {

   struct search_data search_data; //!< search data for the target points that
                                   //!< are assigned to the process

   struct request_record * request_records;
   unsigned num_request_records;

   struct interpolation_method ** method_stack;
   struct interpolation_plan_element ** plan_stack;
   unsigned * num_interpolated_points;
   double * coordinates_x;
   double * coordinates_y;
   int * global_ids;
   enum yac_location tgt_location;
};

static void
generate_tag_string(char * buf,  size_t buffer_size,
                    char const * tag_base_str,
                    char const * source_comp_name,
                    char const * target_comp_name,
                    char const * source_domain_name,
                    char const * target_domain_name,
                    char const * stack_name,
                    unsigned points_idx,
                    unsigned distance_weighted,
                    unsigned n,
                    unsigned use_source_mask,
                    int unique_interpolation_method_id,
                    char const * tag_str) {

   int ret_val;
   ret_val = snprintf(buf, buffer_size, "%s%s%s%s%s%s%d%d%d%d%d%s", tag_base_str,
                      source_comp_name, target_comp_name, source_domain_name,
                      target_domain_name, stack_name, points_idx,
                      distance_weighted, n, use_source_mask,
                      unique_interpolation_method_id, tag_str);

   if (ret_val < 0 || ret_val >= buffer_size)
      yac_internal_abort_message("ERROR(generate_tag_string): truncation", __FILE__, __LINE__);
}

static void
generate_interpolation_method_nnn_tag(struct interpolation_method_nnn * method,
                                      char const * stack_name) {

   char const * str_base_str                         = "interpolation method nnn";
   char const * str_generator_string                 = "generator string";
   char const * str_results_or_search_request_header = "results or search request header";
   char const * str_search_request                   = "search request";
   char const * str_results                          = "results";
   char const * str_global_id_exchange_1_a           = "global id exchange 1 a";
   char const * str_global_id_exchange_1_b           = "global id exchange 1 b";
   char const * str_global_id_exchange_2_a           = "global id exchange 2 a";
   char const * str_global_id_exchange_2_b           = "global id exchange 2 b";
   char const * str_exchange_results_a               = "exchange results a";
   char const * str_exchange_results_b               = "exchange results b";
   char const * str_finished                         = "finished";

   struct global_search * search;

   search = get_global_search((struct interpolation_method *)method);

   struct communicator * comm = yac_get_domain_comm(method->domain);

   char const * source_comp_name = yac_get_local_component_name(search);
   char const * target_comp_name = yac_get_remote_component_name(search);

   char const * source_domain_name = yac_get_local_domain_name(search);
   char const * target_domain_name = yac_get_remote_domain_name(search);

   unsigned max_tag_strlen;

   max_tag_strlen = strlen(str_results_or_search_request_header);

   unsigned max_str_buf_size;

   max_str_buf_size = max_tag_strlen + strlen(str_base_str) +
                      strlen(source_comp_name) + strlen(target_comp_name) +
                      strlen(source_domain_name) + strlen(target_domain_name) +
                      strlen(stack_name) + 10 + 10 + 2 + 10 + 1; // (+ 10) for points_idx
                                                                 // (+ 10) for distance_weighted
                                                                 // (+ 10) for n
                                                                 // (+ 2)  for mask handling
                                                                 // (+ 10) for unique_method_id
                                                                 // (+ 1)  for '\0'

   char * unique_generator_string = malloc(max_str_buf_size);
   generate_tag_string(unique_generator_string, max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name,  source_domain_name, target_domain_name,
                       stack_name, method->points_idx, method->distance_weighted,
                       method->n, method->use_source_mask,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                       str_generator_string);

   char * tag_strings[10];
   tag_strings[0] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[0], max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name,  source_domain_name, target_domain_name,
                       stack_name, method->points_idx, method->distance_weighted,
                       method->n, method->use_source_mask,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                       str_results_or_search_request_header);

   tag_strings[1] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[1], max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name,  source_domain_name, target_domain_name,
                       stack_name, method->points_idx, method->distance_weighted,
                       method->n, method->use_source_mask,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                       str_search_request);

   tag_strings[2] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[2], max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name,  source_domain_name, target_domain_name,
                       stack_name, method->points_idx, method->distance_weighted,
                       method->n, method->use_source_mask,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                       str_results);

   tag_strings[3] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[3], max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name,  source_domain_name, target_domain_name,
                       stack_name, method->points_idx, method->distance_weighted,
                       method->n, method->use_source_mask,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                       str_global_id_exchange_1_a);

   tag_strings[4] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[4], max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name,  source_domain_name, target_domain_name,
                       stack_name, method->points_idx, method->distance_weighted,
                       method->n, method->use_source_mask,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                       str_global_id_exchange_1_b);

   tag_strings[5] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[5], max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name,  source_domain_name, target_domain_name,
                       stack_name, method->points_idx, method->distance_weighted,
                       method->n, method->use_source_mask,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                       str_global_id_exchange_2_a);

   tag_strings[6] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[6], max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name,  source_domain_name, target_domain_name,
                       stack_name, method->points_idx, method->distance_weighted,
                       method->n, method->use_source_mask,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                       str_global_id_exchange_2_b);

   tag_strings[7] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[7], max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name,  source_domain_name, target_domain_name,
                       stack_name, method->points_idx, method->distance_weighted,
                       method->n, method->use_source_mask,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                       str_exchange_results_a);

   tag_strings[8] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[8], max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name,  source_domain_name, target_domain_name,
                       stack_name, method->points_idx, method->distance_weighted,
                       method->n, method->use_source_mask,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                       str_exchange_results_b);

   tag_strings[9] = malloc(max_str_buf_size);
   generate_tag_string(tag_strings[9], max_str_buf_size, str_base_str, source_comp_name,
                       target_comp_name,  source_domain_name, target_domain_name,
                       stack_name, method->points_idx, method->distance_weighted,
                       method->n, method->use_source_mask,
                       yac_get_unique_id_interpolation_method((struct interpolation_method *)method),
                       str_finished);

   unsigned tags[10];

   yac_get_unique_comm_tags(
      comm, unique_generator_string, (char const **)tag_strings, 10, tags);

   free(unique_generator_string);

   method->tags.results_or_search_request_header = tags[0];
   method->tags.search_request = tags[1];
   method->tags.results = tags[2];
   method->tags.global_id_exchange_1[0] = tags[3];
   method->tags.global_id_exchange_1[1] = tags[4];
   method->tags.global_id_exchange_2[0] = tags[5];
   method->tags.global_id_exchange_2[1] = tags[6];
   method->tags.exchange_results[0] = tags[7];
   method->tags.exchange_results[1] = tags[8];
   method->tags.finished = tags[9];

   for (unsigned i = 0; i < 10; ++i) free(tag_strings[i]);
}

static int
compare_int(const void * a, const void * b) {

  return ((*(int*)a) - (*(int*)b));
}

static void
init_tgt_search_data(struct tgt_point_search_data * data,
                     unsigned num_src_per_tgt, double coordinate_x,
                     double coordinate_y, int global_id, double min_cos_angle) {

   data->global_id = global_id;
   data->flag = NOT;
   data->coordinate_x = coordinate_x;
   data->coordinate_y = coordinate_y;

   data->min_cos_angle = min_cos_angle;

   data->results = NULL;
   data->num_found_points = 0;
}

static void
unpack_tgt_search_request(void * buffer, unsigned buffer_size,
                          int * position, unsigned num_tgt_points,
                          struct domain * domain, unsigned num_src_per_tgt,
                          struct search_data * search_data,
                          struct communicator * comm) {

   search_data->num_tgt_points = num_tgt_points;
   search_data->num_tgt_points_to_interpolate = num_tgt_points;
   search_data->num_open_remote_requests = 0;

   search_data->tgt_point_data =
      malloc(num_tgt_points * sizeof(*(search_data->tgt_point_data)));

   for (unsigned i = 0; i < num_tgt_points; ++i) {

      double coordinate_x, coordinate_y, min_cos_angle;
      int global_tgt_id;

      // unpack the information about the target point
      yac_comm_unpack(buffer, buffer_size, position, &global_tgt_id, 1, COMM_INT, comm);
      yac_comm_unpack(buffer, buffer_size, position, &coordinate_x, 1, COMM_DBLE, comm);
      yac_comm_unpack(buffer, buffer_size, position, &coordinate_y, 1, COMM_DBLE, comm);

      // unpack information about the minimum cosine angle of potential results
      yac_comm_unpack(buffer, buffer_size, position, &min_cos_angle, 1, COMM_DBLE, comm);

      // initialise the specific target point information
      init_tgt_search_data(search_data->tgt_point_data+i, num_src_per_tgt,
                           coordinate_x, coordinate_y, global_tgt_id,
                           min_cos_angle);

   }
}

//! returns the result of the search request to the origin of the request
static void
pack_and_send_results(struct request_record * records,
                      struct search_data * search_data,
                      struct interpolation_method_nnn * method,
                      unsigned orig_rank, struct communicator * comm) {

  char * buffer = NULL;
  unsigned buffer_array_size = 0;
  unsigned buffer_size = 0;
  int position = 0;

  // count the number of remote processes that may contain better source points
  unsigned num_remote_ranks = 0;
  for (unsigned i = 0; i < method->num_domain_neighbours; ++i)
    if (records[i].num_global_ids > 0) num_remote_ranks++;

  buffer_size += yac_comm_pack_size(1, COMM_UINT, comm);
  ENSURE_ARRAY_SIZE(buffer, buffer_array_size, buffer_size);

  // pack the number of remote ranks
  yac_comm_pack(&num_remote_ranks, 1, COMM_UINT, buffer, buffer_size,
                &position, comm);

  // pack the global target ids that may have to be sent to the respective
  // remote process
  for (unsigned i = 0; i < method->num_domain_neighbours; ++i) {

    if (records[i].num_global_ids == 0) continue;

    buffer_size += yac_comm_pack_size(2, COMM_UINT, comm);
    buffer_size += yac_comm_pack_size(records[i].num_global_ids, COMM_INT,
                                      comm);
    ENSURE_ARRAY_SIZE(buffer, buffer_array_size, buffer_size);

    yac_comm_pack(&(records[i].rank), 1, COMM_UINT, buffer, buffer_size,
                  &position, comm);
    yac_comm_pack(&(records[i].num_global_ids), 1, COMM_UINT, buffer,
                  buffer_size, &position, comm);
    if (records[i].num_global_ids > 0)
      yac_comm_pack(records[i].global_ids, records[i].num_global_ids, COMM_INT,
                    buffer, buffer_size, &position, comm);
  }

  // pack locally found source points
  for (unsigned i = 0; i < search_data->num_tgt_points; ++i) {

    unsigned num_found_points =
      search_data->tgt_point_data[i].num_found_points;

    if (num_found_points == 0) continue;

    struct tgt_point_search_data * curr_tgt_point_data =
      search_data->tgt_point_data + i;

    buffer_size += yac_comm_pack_size(1, COMM_INT, comm); // global id
    buffer_size += yac_comm_pack_size(1, COMM_UINT, comm); // num_found_points
    buffer_size += yac_comm_pack_size(num_found_points, COMM_INT, comm); // global id
    buffer_size += yac_comm_pack_size(num_found_points, COMM_DBLE, comm); // cos_angle
    ENSURE_ARRAY_SIZE(buffer, buffer_array_size, buffer_size);

    yac_comm_pack(&(curr_tgt_point_data->global_id), 1, COMM_INT, buffer,
                  buffer_size, &position, comm);
    yac_comm_pack(&(curr_tgt_point_data->num_found_points), 1, COMM_UINT,
                  buffer, buffer_size, &position, comm);
    for (unsigned j = 0; j < curr_tgt_point_data->num_found_points; ++j) {
       yac_comm_pack(&(curr_tgt_point_data->results[j].cos_angle), 1, COMM_DBLE,
                     buffer, buffer_size, &position, comm);
       yac_comm_pack(&(curr_tgt_point_data->results[j].global_id), 1, COMM_INT,
                     buffer, buffer_size, &position, comm);
    }
  }

  unsigned header[2] = {RESULTS_MSG_TYPE, buffer_size};

  // send header
  yac_comm_bsend(&header, 2, COMM_UINT, orig_rank,
                 method->tags.results_or_search_request_header, comm);

  // send data
  yac_comm_bsend(buffer, buffer_size, COMM_PACKED, orig_rank,
                 method->tags.results, comm);

  free(buffer);
}

static struct tgt_point_search_data *
tgt_point_search_data_new(unsigned num_points,
                          unsigned num_src_per_tgt,
                          double * coordinates_x,
                          double * coordinates_y,
                          int * global_ids) {

   struct tgt_point_search_data * data;

   data = malloc(num_points * sizeof(*data));

   unsigned i;

   for (i = 0; i < num_points; ++i)
      init_tgt_search_data(data+i, num_src_per_tgt, coordinates_x[i],
                           coordinates_y[i], global_ids[i], -1.0);

   return data;
}

static unsigned
get_num_cell_data_points(struct grid * grid, unsigned cell_id,
                         enum yac_location location) {

   switch (location) {

      case (CELL):
         return 1;
      case (EDGE):
         return yac_get_num_cell_edges(grid, cell_id);
      case (CORNER):
         return yac_get_num_cell_corners(grid, cell_id);
      default:
         return 0;
   };
}

static unsigned const *
get_local_point_ids(struct grid * grid, unsigned * cell_id,
                    enum yac_location location) {

   switch (location) {

      case (CELL):
         return cell_id;
      case (EDGE):
         return yac_get_cell_edge_indices(grid, *cell_id);
      case (CORNER):
         return yac_get_cell_corner_indices(grid, *cell_id);
      default:
         return NULL;
   };
}

static int
compare_results(const void * a, const void * b) {

  int res = (((struct tgt_point_results*)b)->cos_angle >
             ((struct tgt_point_results*)a)->cos_angle) -
            (((struct tgt_point_results*)a)->cos_angle >
             ((struct tgt_point_results*)b)->cos_angle);
  if (res == 0)
    res = ((struct tgt_point_results*)a)->global_id -
          ((struct tgt_point_results*)b)->global_id;
  return res;
}

/**
 * adds a source point to the results for a target point \n
 * the point is only added if the required number of source points per target points
 * is not exceeded \n
 * if there are already enough source points and if there are source points that have a
 * higher distance, the new point will replace the one with the highest distance
 */
static void
add_src_point_to_results(struct tgt_point_results src_point,
                         struct tgt_point_search_data * search_data,
                         unsigned num_src_per_tgt) {

  // if the point is too far
  if (src_point.cos_angle < search_data->min_cos_angle) return;

  // if there are no results for the current target points yet
  if (search_data->num_found_points == 0) {
    search_data->results[0] = src_point;
    search_data->num_found_points = 1;
    if (num_src_per_tgt == 1) search_data->min_cos_angle = src_point.cos_angle;
    return;
  }

  // check whether point is already in the list
  for (unsigned i = 0; i < search_data->num_found_points; ++i)
    if (search_data->results[i].global_id == src_point.global_id) return;

  unsigned i;
  for (i = MIN(num_src_per_tgt-1, search_data->num_found_points); i > 0; --i) {

    if (compare_results(search_data->results+i-1, &src_point) > 0)
      search_data->results[i] = search_data->results[i-1];
    else
      break;
  }

  search_data->results[i] = src_point;
  search_data->num_found_points =
    MIN(search_data->num_found_points + 1, num_src_per_tgt);
  search_data->min_cos_angle =
    search_data->results[search_data->num_found_points-1].cos_angle;
}

//! searches in the local data for source points
static void
nnn_search_local(struct search_data * search_data,
                 struct interpolation_method_nnn * method,
                 struct communicator * comm) {


  unsigned buffer_size = MIN(search_data->num_tgt_points_to_interpolate, 1024);
  double * tgt_coordinates_x = malloc(buffer_size * sizeof(*tgt_coordinates_x));
  double * tgt_coordinates_y = malloc(buffer_size * sizeof(*tgt_coordinates_y));

  unsigned num_subdomains = yac_get_num_subdomains(method->domain);

  double * cos_angles = NULL;
  unsigned cos_angles_array_size = 0;
  unsigned * local_point_ids = NULL;
  unsigned local_point_ids_array_size = 0;
  unsigned * num_local_point_ids =
    malloc(MIN(search_data->num_tgt_points_to_interpolate, 1024) *
           sizeof(*num_local_point_ids));

  union {
    int const * one_sub;
    int const ** n_sub;
  } global_ids;

  enum yac_location location = method->points[0][method->points_idx]->location;
  if (num_subdomains == 1) {
    global_ids.one_sub = yac_get_global_ids(method->domain, 0, location);
  } else {
    global_ids.n_sub = malloc(num_subdomains * sizeof(*(global_ids.n_sub)));
    for (unsigned i = 0; i < num_subdomains; ++i)
      global_ids.n_sub[i] = yac_get_global_ids(method->domain, i, location);
  }

  unsigned num_interations =
    (search_data->num_tgt_points_to_interpolate + 1023) / 1024;
  for (unsigned i = 0; i < num_interations; ++i) {

    unsigned num_tgt_points =
      MIN(1024, search_data->num_tgt_points_to_interpolate - i * 1024);

    struct tgt_point_search_data * curr_tgt_point_data =
      search_data->tgt_point_data + i * 1024;

    yac_comm_check(comm);

    for (unsigned j = 0; j < num_tgt_points; ++j) {
      tgt_coordinates_x[j] = curr_tgt_point_data[j].coordinate_x;
      tgt_coordinates_y[j] = curr_tgt_point_data[j].coordinate_y;
    }

    yac_point_sphere_part_search_NNN(
      method->non_halo_point_search, num_tgt_points, tgt_coordinates_x,
      tgt_coordinates_y, method->n, &cos_angles, &cos_angles_array_size,
      &local_point_ids, &local_point_ids_array_size, num_local_point_ids);

    for (unsigned j = 0, result_offset = 0; j < num_tgt_points; ++j) {

      struct tgt_point_results * results =
        malloc(MAX(num_local_point_ids[j], method->n) * sizeof(*results));

      // if there is only one subdomain
      if (num_subdomains == 1) {

        for (unsigned k = 0; k < num_local_point_ids[j]; ++k, ++result_offset) {
          unsigned local_id =
            method->non_halo_local_ids[local_point_ids[result_offset]];
          results[k].cos_angle = cos_angles[result_offset];
          results[k].global_id = global_ids.one_sub[local_id];
          results[k].is_local = 1;
          results[k].data.local.local_id = local_id;
          results[k].data.local.subdomain_id = 0;
        }
      } else {

        for (unsigned k = 0; k < num_local_point_ids[j]; ++k, ++result_offset) {
          unsigned local_id = local_point_ids[result_offset];
          unsigned subdomain_id;
          for (
            subdomain_id = 0; (subdomain_id < num_subdomains) && (local_id >=
            method->num_non_halo_local_ids_per_subdomain_prefix[subdomain_id]);
            ++subdomain_id);
          local_id = method->non_halo_local_ids[local_id];
          results[k].cos_angle = cos_angles[result_offset];
          results[k].global_id = global_ids.n_sub[subdomain_id][local_id];
          results[k].is_local = 1;
          results[k].data.local.local_id = local_id;
          results[k].data.local.subdomain_id = subdomain_id;
        }
      }

      // sort results first by angle then by global id
      qsort(results, num_local_point_ids[j], sizeof(*results),
            compare_results);

      curr_tgt_point_data[j].results = results;
      curr_tgt_point_data[j].num_found_points =
        MIN(num_local_point_ids[j], method->n);
      if (curr_tgt_point_data[j].num_found_points == method->n)
        curr_tgt_point_data[j].min_cos_angle =
          curr_tgt_point_data[j].results[method->n-1].cos_angle;
    }
  }

  free(local_point_ids);
  free(cos_angles);
  free(num_local_point_ids);
  if (num_subdomains > 1) free(global_ids.n_sub);
  free(tgt_coordinates_y);
  free(tgt_coordinates_x);
}

static void
pack_and_send_search_request(struct intermediate_method_data * data,
                             struct request_record * records,
                             unsigned num_records) {

  struct interpolation_method_nnn * method =
    (struct interpolation_method_nnn*)(data->method_stack[0]);
  struct communicator * comm = yac_get_domain_comm(method->domain);

  char * send_buffer = NULL;
  unsigned send_buffer_array_size = 0;

  // for all entries in records
  for (unsigned i = 0; i < num_records; ++i) {

    if (records[i].num_global_ids == 0) continue;

    unsigned num_points = records[i].num_global_ids;
    unsigned send_buffer_size =
      yac_comm_pack_size(3 * num_points, COMM_DBLE, comm) + // coordinate_x +
                                                            // coordinate_y +
                                                            // min_cos_angle
      yac_comm_pack_size(1 * num_points, COMM_INT, comm);   // global_id
    int position = 0;

    ENSURE_ARRAY_SIZE(send_buffer, send_buffer_array_size, send_buffer_size);

    for (unsigned j = 0, k = 0; j < num_points; ++j) {

      while (data->search_data.tgt_point_data[k].global_id !=
             records[i].global_ids[j]) ++k;

      yac_comm_pack(&(data->search_data.tgt_point_data[k].global_id), 1,
                    COMM_INT, send_buffer, send_buffer_size, &position, comm);
      yac_comm_pack(&(data->search_data.tgt_point_data[k].coordinate_x), 1,
                    COMM_DBLE, send_buffer, send_buffer_size, &position, comm);
      yac_comm_pack(&(data->search_data.tgt_point_data[k].coordinate_y), 1,
                    COMM_DBLE, send_buffer, send_buffer_size, &position, comm);
      yac_comm_pack(&(data->search_data.tgt_point_data[k].min_cos_angle), 1,
                    COMM_DBLE, send_buffer, send_buffer_size, &position, comm);
    }

    data->search_data.num_open_remote_requests++;

    unsigned header[3] = {SEARCH_REQUEST_MSG_TYPE, position, num_points};

    // send header (does not contain data, just informs other process about
    //              pending message)
    yac_comm_bsend(header, 3, COMM_UINT, records[i].rank,
                   method->tags.results_or_search_request_header, comm);

    // send message
    yac_comm_bsend(send_buffer, position, COMM_PACKED, records[i].rank,
                   method->tags.search_request, comm);
  }

  free(send_buffer);
}

static int
compare_request_records(const void * a, const void * b) {

  return (((struct request_record*)a)->rank -
          ((struct request_record*)b)->rank);
}

//! unpacks results received from another process that performed a search
static void
unpack_results(void * buffer, unsigned buffer_size,
               struct interpolation_method_nnn * method,
               struct intermediate_method_data * data,
               struct request_record ** records, unsigned * num_records,
               int remote_rank, struct communicator * comm) {

  int position = 0;

  unsigned num_remote_rank;
  yac_comm_unpack(buffer, buffer_size, &position, &num_remote_rank, 1,
                  COMM_UINT, comm);
  *num_records = num_remote_rank;

  *records = malloc(num_remote_rank * sizeof(**records));

  for (unsigned i = 0; i < num_remote_rank; ++i) {

    struct request_record record;

    // unpack data
    yac_comm_unpack(buffer, buffer_size, &position, &(record.rank), 1,
                    COMM_UINT, comm);
    yac_comm_unpack(buffer, buffer_size, &position, &(record.num_global_ids), 1,
                    COMM_UINT, comm);
    record.global_ids_array_size = record.num_global_ids;
    if (record.num_global_ids > 0) {
      record.global_ids = malloc(record.num_global_ids *
                                 sizeof(*(record.global_ids)));
      yac_comm_unpack(buffer, buffer_size, &position, record.global_ids,
                      record.num_global_ids, COMM_INT, comm);
    } else {
      record.global_ids = NULL;
    }

    // check whether we already sent the current data to the respective rank
    unsigned j;
    for (j = 0; j < data->num_request_records; ++j)
      if (data->request_records[j].rank == record.rank) break;

    if (j == data->num_request_records) {

      data->request_records = realloc(data->request_records,
                                      (data->num_request_records + 1) *
                                      sizeof(*(data->request_records)));

      data->request_records[data->num_request_records] = record;
      data->request_records[data->num_request_records].global_ids =
        malloc(record.num_global_ids * sizeof(*(record.global_ids)));
      memcpy(data->request_records[data->num_request_records].global_ids,
             record.global_ids, record.num_global_ids *
             sizeof(*(record.global_ids)));
      data->num_request_records++;

    } else {

      ENSURE_ARRAY_SIZE(data->request_records[j].global_ids,
                        data->request_records[j].global_ids_array_size,
                        data->request_records[j].num_global_ids +
                        record.num_global_ids);

      unsigned m = 0, sort_flag = 0;
      for (unsigned k = 0, l = 0; k < record.num_global_ids; ++k) {

        while ((l < data->request_records[j].num_global_ids) &&
               (data->request_records[j].global_ids[l] < record.global_ids[k]))
          ++l;

        if ((l == data->request_records[j].num_global_ids) ||
            (data->request_records[j].global_ids[l] != record.global_ids[k])) {
          data->request_records[j].global_ids[l++] = record.global_ids[k];
          data->request_records[j].num_global_ids++;
          sort_flag = 1;
          if (m != k) record.global_ids[m] = record.global_ids[k];
          m++;
        } else if (data->request_records[j].global_ids[l] !=
                   record.global_ids[k]) {
          if (m != k) record.global_ids[m] = record.global_ids[k];
          m++;
        }
      }
      record.num_global_ids = m;

      if (sort_flag) qsort(data->request_records[j].global_ids,
                           data->request_records[j].num_global_ids,
                           sizeof(*(data->request_records[j].global_ids)),
                           compare_int);
    }

    (*records)[i] = record;
  }

  qsort(*records, *num_records, sizeof(**records), compare_request_records);

  struct tgt_point_results src_point;
  src_point.is_local = 0;
  src_point.data.remote.rank = remote_rank;

  unsigned j = 0;

  while (position < buffer_size) {

    int global_id;
    unsigned num_found_src_points;

    yac_comm_unpack(buffer, buffer_size, &position, &global_id, 1, COMM_INT,
                    comm);
    yac_comm_unpack(buffer, buffer_size, &position, &num_found_src_points, 1,
                    COMM_UINT, comm);

    while ((j < data->search_data.num_tgt_points_to_interpolate) &&
           (data->search_data.tgt_point_data[j].global_id != global_id)) ++j;

    if (j >= data->search_data.num_tgt_points_to_interpolate)
      yac_internal_abort_message("ERROR (unpack_results): did not find target "
                                 "point for which results were received\n",
                                 __FILE__, __LINE__);

    for (unsigned k = 0; k < num_found_src_points; ++k) {

       yac_comm_unpack(buffer, buffer_size, &position, &(src_point.cos_angle),
                       1, COMM_DBLE, comm);
       yac_comm_unpack(buffer, buffer_size, &position, &(src_point.global_id),
                       1, COMM_INT, comm);

       add_src_point_to_results(src_point, data->search_data.tgt_point_data+j,
                                method->n);
    }
  }
}

/**
  * reorders search_data such that the points specified in the put_to_front
  * array are at the beginning
  */
static void
reorder_points(struct tgt_point_search_data * search_data, int * put_to_front,
               unsigned num_points, unsigned num_points_to_front) {

   unsigned i, j;

   for (i = 0; i < num_points_to_front; ++i) {

      if (!put_to_front[i]) {

         // find a point to replace the current one
         for (j = i + 1;!put_to_front[j];++j);

         int temp_put_to_front;
         struct tgt_point_search_data temp_search_data;

         temp_search_data = search_data[i];
         search_data[i] = search_data[j];
         search_data[j] = temp_search_data;
         temp_put_to_front = put_to_front[i];
         put_to_front[i] = put_to_front[j];
         put_to_front[j] = temp_put_to_front;
      }
   }
}

/**
 * search for result source points that are marked as remote, but are actually
 * available locally
 */
static void
detect_local_points(struct intermediate_method_data * data) {

   struct search_data * search_data;
   struct interpolation_method_nnn * method;

   search_data = &(data->search_data);
   method = (struct interpolation_method_nnn *)(data->method_stack[0]);

   struct domain * domain = method->domain;

   int comm_rank;

   comm_rank = yac_get_comm_rank(yac_get_domain_comm(domain));

   int * global_ids = NULL;
   unsigned global_ids_array_size = 0;
   unsigned num_global_ids = 0;

   unsigned i, j, k;

   // search for remote points that have the same rank as the local process
   for (i = 0; i < data->num_interpolated_points[0]; ++i) {

      for (j = 0; j < search_data->tgt_point_data[i].num_found_points; ++j) {

         if (search_data->tgt_point_data[i].results[j].is_local == 0 &&
             search_data->tgt_point_data[i].results[j].data.remote.rank == comm_rank) {

            ENSURE_ARRAY_SIZE(global_ids, global_ids_array_size, num_global_ids + 1);
            global_ids[num_global_ids++] = search_data->tgt_point_data[i].results[j].global_id;
         }
      }
   }

   // if some points were found
   if (num_global_ids > 0) {

      enum yac_location location = method->points[0][method->points_idx]->location;

      int * matched_global_ids;
      unsigned * local_ids;
      unsigned * subdomain_ids;
      unsigned num_matched_global_ids = 0;
      unsigned curr_num_matched_global_ids = 0;
      unsigned num_subdomains;

      matched_global_ids = malloc(num_global_ids * sizeof(*matched_global_ids));
      local_ids = malloc(num_global_ids * sizeof(*local_ids));
      subdomain_ids = malloc(num_global_ids * sizeof(*subdomain_ids));

      num_subdomains = yac_get_num_subdomains(domain);

      // sort global_ids (required by match_global_ids)
      qsort(global_ids, num_global_ids, sizeof(*global_ids), compare_int);

      // search for the global ids within the local data
      for (i = 0; i < num_subdomains; ++i) {

         yac_match_global_ids(domain, i, location, global_ids, num_global_ids,
                              matched_global_ids + num_matched_global_ids, NULL,
                              &curr_num_matched_global_ids, local_ids +
                              num_matched_global_ids, 0);


         for (j = 0; j < curr_num_matched_global_ids; ++j)
            subdomain_ids[num_matched_global_ids+j] = i;

         num_matched_global_ids += curr_num_matched_global_ids;
      }

      if (num_matched_global_ids != num_global_ids)
         yac_internal_abort_message("ERROR(detect_local_points): is the definition of the global"
                                    " ids correct?", __FILE__, __LINE__);

      // convert the "remote" points to local ones
      for (i = 0; i < data->num_interpolated_points[0]; ++i) {

         for (j = 0; j < search_data->tgt_point_data[i].num_found_points; ++j) {

            if (search_data->tgt_point_data[i].results[j].is_local == 0 &&
                search_data->tgt_point_data[i].results[j].data.remote.rank == comm_rank) {

               for (k = 0; k < num_global_ids; ++k) {

                  if (search_data->tgt_point_data[i].results[j].global_id ==
                      matched_global_ids[k]) {

                     search_data->tgt_point_data[i].results[j].is_local = 1;
                     search_data->tgt_point_data[i].results[j].data.local.local_id = local_ids[k];
                     search_data->tgt_point_data[i].results[j].data.local.subdomain_id = subdomain_ids[k];
                  }
               }
            }
         }
      }

      free(matched_global_ids);
      free(local_ids);
      free(subdomain_ids);
   }

   free(global_ids);
}

/**
 * analyses the results of the search and generates a data_request object, which
 * is required by the field_interpolation in order to be able to provide the
 * remote points needed for the local interpolation
 */
static void
generate_data_request(struct intermediate_method_data * data) {

   struct search_data * search_data;
   struct interpolation_method_nnn * method;
   struct interpolation_method_data_request * data_request;
   unsigned points_idx;
   unsigned num_pointsets;

   search_data = &(data->search_data);
   method = (struct interpolation_method_nnn *)(data->method_stack[0]);
   points_idx = method->points_idx;
   num_pointsets = method->num_pointsets;
   method->data_request[0] = malloc(num_pointsets *
                                    sizeof(*(method->data_request[0])));
   data_request = method->data_request[0];

   unsigned global_ids_array_size = 0;
   unsigned ranks_array_size = 0;
   unsigned num_requested_points = 0;

   unsigned i, j, k;

   for (i = 0; i < num_pointsets; ++i) {
      data_request[i].num_values = 0;
      data_request[i].global_ids = NULL;
      data_request[i].ranks = NULL;
   }

   data_request += points_idx;

   for (i = 0; i < data->num_interpolated_points[0]; ++i) {

      struct tgt_point_search_data * curr_tgt_point_data;

      curr_tgt_point_data = search_data->tgt_point_data + i;

      for (j = 0; j < curr_tgt_point_data->num_found_points; ++j) {

         if (!curr_tgt_point_data->results[j].is_local) {

            ENSURE_ARRAY_SIZE(data_request->global_ids, global_ids_array_size,
                              num_requested_points + 1);
            ENSURE_ARRAY_SIZE(data_request->ranks, ranks_array_size,
                              num_requested_points + 1);

            for (k = 0; k < num_requested_points; ++k)
               if (curr_tgt_point_data->results[j].global_id ==
                   data_request->global_ids[k]) break;

            if (k == num_requested_points) {

               data_request->global_ids[num_requested_points] =
                  curr_tgt_point_data->results[j].global_id;
               data_request->ranks[num_requested_points] =
                  curr_tgt_point_data->results[j].data.remote.rank;
               num_requested_points++;
            }
         }
      }
   }

   data_request->num_values = num_requested_points;
   data_request->global_ids = realloc(data_request->global_ids, num_requested_points *
                                      sizeof(*(data_request->global_ids)));
   data_request->ranks = realloc(data_request->ranks, num_requested_points *
                                 sizeof(*(data_request->ranks)));
}

static void
free_search_data(struct search_data * search_data) {

   if (search_data->num_open_remote_requests != 0)
      yac_internal_abort_message(
         "ERROR (free_search_data): search_data still has open requests",
         __FILE__, __LINE__);

   unsigned i;

   for (i = 0; i < search_data->num_tgt_points; ++i)
      free(search_data->tgt_point_data[i].results);

   free(search_data->tgt_point_data);
}

static void
request_records_delete(struct request_record * records, unsigned n) {

  for (unsigned i = 0; i < n; ++i) free(records[i].global_ids);
  free(records);
}

static void
free_intermediate_method_data(struct intermediate_method_data * data) {

   request_records_delete(data->request_records, data->num_request_records);
   free_search_data(&(data->search_data));
   free(data);
}

/**
 * generates the interpolation plan based on the results \n
 * uses a point-type plan for the interpolation plan
 */
static struct interpolation_plan_element *
generate_interpolation_plan_1nn(struct intermediate_method_data * data) {

   if (*(data->num_interpolated_points) == 0) return NULL;

   unsigned points_idx;
   unsigned num_interpolated_points;
   struct tgt_point_search_data * tgt_results;
   struct interpolation_method_data_request * data_request;
   unsigned num_pointsets;

   points_idx = ((struct interpolation_method_nnn*)(data->method_stack[0]))->points_idx;
   num_interpolated_points = data->num_interpolated_points[0];
   tgt_results = data->search_data.tgt_point_data;
   data_request = ((struct interpolation_method_nnn*)(data->method_stack[0]))->data_request[0];
   num_pointsets = ((struct interpolation_method_nnn*)(data->method_stack[0]))->num_pointsets;

   unsigned * src_indices;
   src_indices = malloc(3  * num_interpolated_points * sizeof(*src_indices));

   for (unsigned i = 0; i < num_interpolated_points; ++i) {

      if (tgt_results[i].results[0].is_local) {

         src_indices[3*i+0] = tgt_results[i].results[0].data.local.subdomain_id+1;
         src_indices[3*i+1] = points_idx;
         src_indices[3*i+2] = tgt_results[i].results[0].data.local.local_id;

      } else {

         unsigned k;
         for (k = 0; k < data_request->num_values; ++k)
            if (data_request->global_ids[k] == tgt_results[i].results[0].global_id)
               break;

         src_indices[3*i+0] = 0;
         src_indices[3*i+1] = points_idx;
         src_indices[3*i+2] = k;
      }
   }

   unsigned * num_remote_points;

   num_remote_points = calloc(num_pointsets, sizeof(*num_remote_points));
   num_remote_points[points_idx] = data_request->num_values;

   // generate a plan to execute the interpolation
   struct interpolation_plan_element * plan = malloc(1 * sizeof(*plan));

   yac_init_interpolation_plan_element_point(plan, num_interpolated_points,
                                             src_indices, num_remote_points,
                                             num_pointsets);

   free(num_remote_points);
   free(src_indices);

   return plan;
}

/**
 * generates the interpolation plan based on the results \n
 * uses a wpointsum as for the interpolation plan
 */
static struct interpolation_plan_element *
generate_interpolation_plan_nnn(struct intermediate_method_data * data) {

   if (*(data->num_interpolated_points) == 0) return NULL;

   unsigned points_idx;
   unsigned distance_weighted;
   unsigned num_src_per_tgt;
   unsigned num_interpolated_points;
   struct tgt_point_search_data * tgt_results;
   struct interpolation_method_data_request * data_request;
   unsigned num_pointsets;

   points_idx = ((struct interpolation_method_nnn*)(data->method_stack[0]))->points_idx;
   distance_weighted = ((struct interpolation_method_nnn*)(data->method_stack[0]))->distance_weighted;
   num_src_per_tgt = ((struct interpolation_method_nnn*)(data->method_stack[0]))->n;
   num_interpolated_points = data->num_interpolated_points[0];
   tgt_results = data->search_data.tgt_point_data;
   data_request = ((struct interpolation_method_nnn*)(data->method_stack[0]))->data_request[0];
   num_pointsets = ((struct interpolation_method_nnn*)(data->method_stack[0]))->num_pointsets;

   unsigned * src_indices;

   src_indices = malloc(3 * num_src_per_tgt * num_interpolated_points *
                        sizeof(*src_indices));

   unsigned i, j, k;
   unsigned * curr_src_indices;

   curr_src_indices = src_indices;

   for (i = 0; i < num_interpolated_points; ++i) {

      for (j = 0; j < tgt_results[i].num_found_points; ++j) {

         if (tgt_results[i].results[j].is_local) {

            curr_src_indices[0] = tgt_results[i].results[j].data.local.subdomain_id+1;
            curr_src_indices[1] = points_idx;
            curr_src_indices[2] = tgt_results[i].results[j].data.local.local_id;

         } else {

            for (k = 0; k < data_request->num_values; ++k)
               if (data_request->global_ids[k] == tgt_results[i].results[j].global_id)
                  break;

            curr_src_indices[0] = 0;
            curr_src_indices[1] = points_idx;
            curr_src_indices[2] = k;
         }

         curr_src_indices += 3;
      }
   }

   double * weights;
   unsigned * num_src_per_tgt_array;
   unsigned * num_remote_points;
   double * distances;

   weights = malloc(num_interpolated_points * num_src_per_tgt * sizeof(*weights));
   num_src_per_tgt_array = malloc(num_interpolated_points *
                                  sizeof(*num_src_per_tgt_array));
   num_remote_points = calloc(num_pointsets, sizeof(*num_remote_points));
   distances = malloc(num_src_per_tgt * sizeof(*distances));

   if ( distance_weighted ) {

       for (i = 0; i < num_interpolated_points; ++i) {

           if (tgt_results[i].num_found_points > 0) {

               double distance_sum = 0.0;

               for (j = 0; j < num_src_per_tgt; ++j) {
                 distances[j] = acos(tgt_results[i].results[j].cos_angle);
                 distance_sum += 1.0 / distances[j];
               }

               distance_sum = 1.0 / distance_sum;

               for (j = 0; j < num_src_per_tgt; ++j) {
                   if ( distances[j] > 0.0 ) {
                       weights[i*num_src_per_tgt+j] =
                           1.0 / distances[j] * distance_sum;
                   } else {
                       // special case: source and target at same location
                       break;
                   }
               }

               // special case: source and target at same location

               if ( j < num_src_per_tgt ) {

                   for ( int jj = 0; jj < num_src_per_tgt; ++jj )
                     weights[i*num_src_per_tgt+jj] = 0.0;

                   weights[i*num_src_per_tgt+j] = 1.0;

               }
           }

           for (j = num_src_per_tgt; j < num_src_per_tgt; ++j)
             weights[i*num_src_per_tgt+j] = 0.0;
       }

   } else {

       // arithmetic average

       for (i = 0; i < num_interpolated_points * num_src_per_tgt; ++i)
         weights[i] = 1.0 / (double)num_src_per_tgt;
   }

   free(distances);

   for (i = 0; i < num_interpolated_points; ++i)
      num_src_per_tgt_array[i] = num_src_per_tgt;
   num_remote_points[points_idx] = data_request->num_values;

   // generate a plan to execute the interpolation
   struct interpolation_plan_element * plan;

   plan = malloc(1 * sizeof(*plan));

   yac_init_interpolation_plan_element_wpointsum(plan, num_interpolated_points,
                                                 src_indices, num_src_per_tgt_array, weights,
                                                 num_remote_points, num_pointsets);

   free(num_remote_points);
   free(num_src_per_tgt_array);
   free(src_indices);
   free(weights);

   return plan;
}

/**
 * callback routine the is called once a finish message from another
 * source process is received
 */
static void
finish_barrier_callback(struct communicator * comm,
                        struct comm_message msg, void * user_data) {

   struct intermediate_method_data * data;
   struct interpolation_method_nnn * method;

   data = (struct intermediate_method_data *)user_data;
   method = (struct interpolation_method_nnn *)(data->method_stack[0]);

   method->finished = 1;
   yac_comm_cancel_recv(comm, method->tags.results_or_search_request_header);

   free_intermediate_method_data(data);

   if (method->finish_callback != NULL)
      method->finish_callback(method->finish_user_data);
}

static void
retrieve_tgt_point_data(struct tgt_point_search_data * tgt_point_data,
                        unsigned num_tgt_points, double * coordinates_x,
                        double * coordinates_y, int * global_ids) {

  for (unsigned i = 0; i < num_tgt_points; ++i) {
    global_ids[i] = tgt_point_data[i].global_id;
    coordinates_x[i] = tgt_point_data[i].coordinate_x;
    coordinates_y[i] = tgt_point_data[i].coordinate_y;
  }
}

/**
 * Marks points from the list of target points (even if they could
 * be interpolated by the local process). This is done because they are going
 * to be interpolated by another source process, which might later
 * request some data from the local process to do the interpolation.
 * @param[in] num_points_to_be_marked number of points that are interpolated
 *                                    by another source process
 * @param[in] global_ids_to_be_marked array with global ids of target points
 *                                    that are to be marked from the local list
 * @param[in,out] data                basic search data
 *                                    (see \ref intermediate_method_data)
 */
static void
mark_tgt_points(unsigned num_points_to_be_marked, int * global_ids_to_be_marked,
                struct intermediate_method_data * data) {

  if (num_points_to_be_marked == 0) return;

  // first sort all global ids that need to be marked
  yac_quicksort_index(global_ids_to_be_marked, num_points_to_be_marked, NULL);

  // for all target points that need to be marked
  for (unsigned i = 0, j = 0; i < num_points_to_be_marked; ++i) {

    // find the matching entry in the local target points
    while ((j < data->search_data.num_tgt_points) &&
           (data->search_data.tgt_point_data[j].global_id <
            global_ids_to_be_marked[i])) ++j;

    // if no matching entry could be found
    if ((j == data->search_data.num_tgt_points) ||
        (data->search_data.tgt_point_data[j].global_id !=
        global_ids_to_be_marked[i]))
      yac_internal_abort_message(
        "ERROR(mark_tgt_points): invalid global target id\n",
        __FILE__, __LINE__);

    // mark the target point
    data->search_data.tgt_point_data[j].flag = REMOTE;
  }
}

static inline int compare_tgt_point_data(const void * a,const void * b) {

   return ((struct tgt_point_search_data*)a)->global_id -
          ((struct tgt_point_search_data*)b)->global_id;
}

static inline int compare_tgt_point_data_flag(const void * a,const void * b) {

   if (((struct tgt_point_search_data*)a)->flag !=
       ((struct tgt_point_search_data*)b)->flag)
      return ((struct tgt_point_search_data*)a)->flag -
             ((struct tgt_point_search_data*)b)->flag;
   else
      return (((struct tgt_point_search_data*)a)->global_id -
              ((struct tgt_point_search_data*)b)->global_id);
}

/**
 * this routine receive all global ids
 * once the search is finished, this routine generates the data_request object for
 * the field_interpolation and the interpolation_plan \n
 * afterwards it informs all other source processes that the nnn search on this
 * process is finished
 * finally it call the next interpolation method on the interpolation method
 * stack for the remaining target points that could not be interpolated
 */
static void
finish_nnn_interpolation_cont(void * user_data, int * global_ids_to_mark,
                              unsigned num_global_ids) {

   struct intermediate_method_data * data =
      (struct intermediate_method_data *)user_data;
   struct interpolation_method_nnn * method =
      (struct interpolation_method_nnn *)(data->method_stack[0]);
   struct communicator * comm = yac_get_domain_comm(method->domain);

   //--------------------------------------------------------------
   // reset all REMOTE flags to NOT
   //--------------------------------------------------------------
   for (unsigned i = 0; i < data->search_data.num_tgt_points; ++i)
      if (data->search_data.tgt_point_data[i].flag == REMOTE)
         data->search_data.tgt_point_data[i].flag = NOT;

   //--------------------------------------------------------------
   // mark all target points that are interpolated by remote
   // source processes
   //--------------------------------------------------------------
   mark_tgt_points(num_global_ids, global_ids_to_mark, data);

   //--------------------------------------------------------------
   // sort tgt_point_data by the global ids and put the ones
   // flagged as LOCAL in front
   //--------------------------------------------------------------
   qsort(data->search_data.tgt_point_data, data->search_data.num_tgt_points,
         sizeof(*data->search_data.tgt_point_data),
         compare_tgt_point_data_flag);

   // count the number of interpolated points
   unsigned num_interpolated_points = 0;
   unsigned num_not_interpolated_points = 0;
   for (unsigned i = 0; i < data->search_data.num_tgt_points; ++i) {
      if (data->search_data.tgt_point_data[i].flag == LOCAL)
         num_interpolated_points++;
      else if (data->search_data.tgt_point_data[i].flag == NOT)
         num_not_interpolated_points++;
   }
   data->search_data.num_tgt_points_to_interpolate = num_interpolated_points;
   *(data->num_interpolated_points) = num_interpolated_points;

   // detect local result points that are marked as remote ones
   detect_local_points(data);

   // collect all remote points required for the interpolation
   generate_data_request(data);

   // generate final results
   if (method->n == 1)
      data->plan_stack[0] = generate_interpolation_plan_1nn(data);
   else
      data->plan_stack[0] = generate_interpolation_plan_nnn(data);
   if (data->plan_stack[0] != NULL) data->plan_stack++;

   // retrieve coordinates_x, coordinates_y and global_ids from tgt_point_data
   retrieve_tgt_point_data(data->search_data.tgt_point_data,
                           num_interpolated_points +
                           num_not_interpolated_points,
                           data->coordinates_x, data->coordinates_y,
                           data->global_ids);

   struct interpolation_method ** method_stack     = data->method_stack+1;
   struct interpolation_plan_element ** plan_stack = data->plan_stack;
   unsigned * p_num_interpolated_points            = data->num_interpolated_points+1;
   double * coordinates_x                          = data->coordinates_x;
   double * coordinates_y                          = data->coordinates_y;
   int * global_ids                                = data->global_ids+*(data->num_interpolated_points);
   enum yac_location tgt_location                  = data->tgt_location;

   coordinates_x += *(data->num_interpolated_points);
   coordinates_y += *(data->num_interpolated_points);

   yac_comm_barrier_callback(method->tags.finished, comm, data,
                             finish_barrier_callback, MSG_PRIO_LOW);

   // call next interpolation method on stack
   if (*method_stack != NULL)
      (*method_stack)->vtable->do_search(
         method_stack, plan_stack, global_ids, coordinates_x, coordinates_y,
         tgt_location, num_not_interpolated_points, p_num_interpolated_points,
         method->data_request+1);
}

/**
 * once the search is finished, this routine exchanges the points that can be
 * interpolated by the local process with the other source processes. This
 * allows then to remove these points from their target point lists, which are
 * then passed to the next interpolation method in the stack.
 */
static void
finish_nnn_interpolation(struct intermediate_method_data * data){

   struct interpolation_method_nnn * method =
      (struct interpolation_method_nnn *)(data->method_stack[0]);
   struct communicator * comm = yac_get_domain_comm(method->domain);

   // search for target points for which the number of found source
   // points is insufficient
   for (unsigned i = 0; i < data->search_data.num_tgt_points_to_interpolate;
        ++i)
      if (data->search_data.tgt_point_data[i].num_found_points < method->n)
         data->search_data.tgt_point_data[i].flag = NOT;

   // get all points that can be interpolated
   unsigned num_interpolated_points = 0;
   for (unsigned i = 0; i < data->search_data.num_tgt_points_to_interpolate;
        ++i)
      if (data->search_data.tgt_point_data[i].flag == LOCAL)
         data->global_ids[num_interpolated_points++] =
            data->search_data.tgt_point_data[i].global_id;
   // get the remaining points
   for (unsigned i = 0, j = num_interpolated_points;
        i < data->search_data.num_tgt_points; ++i)
      if (data->search_data.tgt_point_data[i].flag != LOCAL)
         data->global_ids[j++] =
            data->search_data.tgt_point_data[i].global_id;

   // sort the target point data by global ids
   qsort(data->search_data.tgt_point_data, data->search_data.num_tgt_points,
         sizeof(*data->search_data.tgt_point_data), compare_tgt_point_data);

   // exchange final results with other source processes
   yac_exchange_results((struct interpolation_method*)method,
                        data->global_ids, num_interpolated_points,
                        data->search_data.num_tgt_points,
                        method->tags.exchange_results,
                        comm, finish_nnn_interpolation_cont, data);
}

static void
recv_results_or_search_request_callback(struct communicator * comm,
                                        struct comm_message msg,
                                        void * user_data);

static void
set_up_recv_results_or_search_request_callback(
   struct interpolation_method_nnn * method,
   struct intermediate_method_data * data) {


   if (!method->finished)
      yac_comm_recv_callback(3, COMM_UINT, COMM_ANYSOURCE,
                             method->tags.results_or_search_request_header,
                             yac_get_domain_comm(method->domain),
                             data, recv_results_or_search_request_callback,
                             NULL, MSG_PRIO_HIGH);
}

static int
nnn_search_remote(struct intermediate_method_data * data,
                  struct search_data * search_data,
                  struct request_record * request_records,
                  unsigned orig_rank);


static void
recv_results_cont_callback(struct communicator * comm, struct comm_message msg,
                           void * user_data) {

   struct intermediate_method_data * data =
      (struct intermediate_method_data *)user_data;
   struct interpolation_method_nnn * method =
      (struct interpolation_method_nnn *)(data->method_stack[0]);

   void * recv_buffer = msg.buffer;
   unsigned recv_buffer_size = msg.count;

   // unpack the results
   struct request_record * records = NULL;
   unsigned num_records = 0;
   unpack_results(recv_buffer, recv_buffer_size, method, data, &records,
                  &num_records, msg.src_rank, comm);

   pack_and_send_search_request(data, records, num_records);
   request_records_delete(records, num_records);

   // if all results have been received
   if (data->search_data.num_open_remote_requests-- == 1)
      finish_nnn_interpolation(data);

   // set up new request
   set_up_recv_results_or_search_request_callback(method, data);
}

/**
 * this is a callback routine, which is called once the results from a forwarded
 * have been received
 */
static void
recv_results_callback(struct communicator * comm, struct comm_message msg,
                      void * user_data) {

   struct intermediate_method_data * data;
   struct interpolation_method_nnn * method;

   data = (struct intermediate_method_data *)user_data;
   method = (struct interpolation_method_nnn *)(data->method_stack[0]);

   unsigned recv_buffer_size = ((unsigned *)(msg.buffer))[0];

   yac_comm_recv_callback(recv_buffer_size, COMM_PACKED, msg.src_rank,
                          method->tags.results, comm, user_data,
                          recv_results_cont_callback, NULL, MSG_PRIO_HIGH);
}

static int compare_uint (const void * a, const void * b) {
  return ( *(unsigned*)a - *(unsigned*)b );
}

static inline void
add_global_id_to_request_record(struct request_record * record, int global_id) {

  ENSURE_ARRAY_SIZE(record->global_ids, record->global_ids_array_size,
                    record->num_global_ids + 1);

  record->global_ids[record->num_global_ids++] = global_id;
}

/**
 * checks whether there may be better source points on neighbouring processes
 */
static int
nnn_search_remote(struct intermediate_method_data * data,
                  struct search_data * search_data,
                  struct request_record * request_records,
                  unsigned orig_rank) {

  struct interpolation_method_nnn * method =
    (struct interpolation_method_nnn*)(data->method_stack[0]);
  struct tgt_point_search_data * tgt_point_data = search_data->tgt_point_data;
  unsigned n = method->n;

  // for each target point
  for (unsigned i = 0; i < search_data->num_tgt_points_to_interpolate; ++i) {

    // if there are not enough results for the current target point 
    if (tgt_point_data[i].num_found_points < n) {

      for (unsigned j = 0; j < method->num_domain_neighbours; ++j)
        if (method->domain_neighbour_ranks[j] != orig_rank)
          add_global_id_to_request_record(
            request_records + j, tgt_point_data[i].global_id);
    } else {

      // get maximum search distance
      struct reduced_bounding_circle circle;
      LLtoXYZ(tgt_point_data[i].coordinate_x, tgt_point_data[i].coordinate_y,
              circle.base_vector);
      circle.cos_inc_angle = tgt_point_data[i].min_cos_angle;

      // determine which neighbours have to receive a search request
      for (unsigned j = 0; j < method->num_domain_neighbours; ++j) {

        if (method->domain_neighbour_ranks[j] == orig_rank) continue;

        // if the current neighbour has halo points in the bounding circle
        if (yac_point_sphere_part_search_bnd_circle_contains_points(
              method->halo_point_search[j], circle))
          add_global_id_to_request_record(
            request_records + j, tgt_point_data[i].global_id);
      }
    }
  }

  for (unsigned i = 0; i < method->num_domain_neighbours; ++i)
    if (request_records[i].num_global_ids > 0) return 1;
  return 0;
}

static struct request_record *
request_records_new(unsigned n, unsigned * ranks) {

  struct request_record * records = malloc(n * sizeof(*records));

  for (unsigned i = 0; i < n; ++i) {
    records[i].rank = ranks[i];
    records[i].global_ids = NULL;
    records[i].global_ids_array_size = 0;
    records[i].num_global_ids = 0;
  }

  return records;
}

static void
receive_search_request_cont_callback(struct communicator * comm,
                                     struct comm_message msg,
                                     void * user_data) {

   struct temp_search_request_data * temp_data =
      (struct temp_search_request_data *)user_data;

   struct intermediate_method_data * data = temp_data->data;
   struct interpolation_method_nnn * method =
      (struct interpolation_method_nnn *)(data->method_stack[0]);
   unsigned num_tgt_points = temp_data->num_tgt_points;

   free(temp_data);

   int position = 0;

   struct search_data search_data;

   unpack_tgt_search_request(msg.buffer, msg.count, &position, num_tgt_points,
                             method->domain, method->n, &search_data, comm);

   // local search
   nnn_search_local(&search_data, method, comm);

   struct request_record * records =
      request_records_new(method->num_domain_neighbours,
                          method->domain_neighbour_ranks);

   // check whether there may be better source points on neighbouring processes
   nnn_search_remote(data, &search_data, records, msg.src_rank);

   // pack and send the results of the local and remote search
   pack_and_send_results(records, &search_data, method, msg.src_rank, comm);
   free_search_data(&search_data);

   request_records_delete(records, method->num_domain_neighbours);

   // set up new request
   set_up_recv_results_or_search_request_callback(method, data);
}

/**
 * callback routine that receives and processes a forwarded search requests
 */
static void
receive_search_request_callback(struct communicator * comm,
                                struct comm_message msg,
                                void * user_data) {

   struct intermediate_method_data * data =
      (struct intermediate_method_data *)user_data;
   struct interpolation_method_nnn * method =
      (struct interpolation_method_nnn *)(data->method_stack[0]);

   struct temp_search_request_data * temp_data = malloc(1 * sizeof(*temp_data));

   unsigned buffer_size = ((unsigned *)(msg.buffer))[0];
   unsigned num_tgt_points = ((unsigned *)(msg.buffer))[1];

   temp_data->data = data;
   temp_data->num_tgt_points = num_tgt_points;

   yac_comm_recv_callback(buffer_size, COMM_PACKED, msg.src_rank,
                          method->tags.search_request, comm, temp_data,
                          receive_search_request_cont_callback, NULL,
                          MSG_PRIO_HIGH);
}

static void
recv_results_or_search_request_callback(struct communicator * comm,
                                        struct comm_message msg,
                                        void * user_data) {

   struct comm_message temp_msg = msg;

   unsigned msg_type = *(unsigned*)(temp_msg.buffer);

   unsigned * temp = (unsigned*)(temp_msg.buffer);
   temp_msg.buffer = ++temp;
   temp_msg.count--;

   switch (msg_type) {

      case (0) :
      {
         recv_results_callback(comm, temp_msg, user_data);
         break;
      }
      case (1) :
      {
         receive_search_request_callback(comm, temp_msg, user_data);
         break;
      }
      default:
      {
         yac_internal_abort_message("ERROR(recv_results_or_search_request_callback): "
                                    "undefined message type\n", __FILE__, __LINE__);
         break;
      }
   }
}

static void
generate_point_sphere_part_searches(struct interpolation_method_nnn * method) {

  unsigned num_subdomains = yac_get_num_subdomains(method->domain);
  unsigned num_neighbours = method->num_domain_neighbours;
  unsigned points_idx = method->points_idx;
  struct mask *** masks = method->masks;
  struct points *** points = method->points;
  enum yac_location location = points[0][points_idx]->location;

  // allocate local_ids, coordinates_x, and coordinates_y
  unsigned * local_ids;
  double * coordinates_x, * coordinates_y;
  unsigned total_num_local_ids = 0;
  {
    unsigned max_num_local_ids = 0;
    for (unsigned subdomain_id = 0; subdomain_id < num_subdomains;
         ++subdomain_id) {
      unsigned data_size =
        yac_get_subdomain_data_size(method->domain, subdomain_id, location);
      if (data_size > max_num_local_ids) max_num_local_ids = data_size;
      total_num_local_ids += data_size;
    }
    local_ids = malloc(max_num_local_ids * sizeof(*local_ids));
    coordinates_x = malloc(total_num_local_ids * sizeof(*coordinates_x));
    coordinates_y = malloc(total_num_local_ids * sizeof(*coordinates_y));
  }

  // generate sphere part point search for all non halo points
  method->non_halo_local_ids =
    malloc(total_num_local_ids * sizeof(*(method->non_halo_local_ids)));
  if (num_subdomains > 1) {
    method->num_non_halo_local_ids_per_subdomain_prefix =
      malloc((num_subdomains + 1) *
             sizeof(*(method->num_non_halo_local_ids_per_subdomain_prefix)));
  } else {
    method->num_non_halo_local_ids_per_subdomain_prefix = NULL;
  }
  unsigned num_non_masked_points = 0;
  for (unsigned subdomain_id = 0; subdomain_id < num_subdomains;
       ++subdomain_id) {

    struct mask * mask = (masks != NULL)?masks[subdomain_id][points_idx]:NULL;

    unsigned num_points =
      yac_get_subdomain_data_size(method->domain, subdomain_id, location) -
      yac_get_num_total_halo_points(method->domain, subdomain_id, location);

    // get the local ids for the non halo points of the current subdomain
    yac_get_non_halo_points(
      method->domain, subdomain_id, location, NULL, local_ids);

    // get the coordinates for the non halo points
    unsigned curr_num_non_masked_points = 0;
    for (unsigned i = 0; i < num_points; ++i) {

      // if the point is masked
      if ((mask != NULL) && (mask->mask[local_ids[i]])) continue;

      double lonlat[2];
      yac_get_point_coordinates(
        points[subdomain_id][points_idx], local_ids[i], lonlat);
      coordinates_x[num_non_masked_points] = lonlat[0];
      coordinates_y[num_non_masked_points] = lonlat[1];
      method->non_halo_local_ids[num_non_masked_points] = local_ids[i];
      num_non_masked_points++;
      curr_num_non_masked_points++;
    }
    if (method->num_non_halo_local_ids_per_subdomain_prefix != NULL) {
      method->num_non_halo_local_ids_per_subdomain_prefix[subdomain_id] =
        curr_num_non_masked_points;
      if (subdomain_id > 0)
        method->num_non_halo_local_ids_per_subdomain_prefix[subdomain_id] +=
          method->num_non_halo_local_ids_per_subdomain_prefix[subdomain_id - 1];
    }
  }
  method->non_halo_point_search =
    yac_point_sphere_part_search_new(
      num_non_masked_points, coordinates_x, coordinates_y);
  if (num_non_masked_points)
    method->non_halo_local_ids =
      realloc(method->non_halo_local_ids, num_non_masked_points *
              sizeof(*(method->non_halo_local_ids)));

  // generate sphere part point search for all halo points
  method->halo_point_search =
    malloc(num_neighbours * sizeof(*(method->halo_point_search)));
  for (unsigned neigh_idx = 0; neigh_idx < num_neighbours; ++neigh_idx) {

    unsigned rank = method->domain_neighbour_ranks[neigh_idx];
    unsigned num_halo_points = 0;

    for (unsigned subdomain_id = 0;
        subdomain_id < num_subdomains; ++subdomain_id) {

      // get the local ids for the halo points of the current subdomain and
      // current domain neighbour
      unsigned num_points;
      yac_get_halo_points(method->domain, subdomain_id, location, NULL,
                          local_ids, &num_points, rank);

      // get the coordinates for the non halo points
      for (unsigned i = 0; i < num_points; ++i, ++num_halo_points) {
        double lonlat[2];
        yac_get_point_coordinates(
          points[subdomain_id][points_idx], local_ids[i], lonlat);
        coordinates_x[num_halo_points] = lonlat[0];
        coordinates_y[num_halo_points] = lonlat[1];
      }
    }
    method->halo_point_search[neigh_idx] =
      yac_point_sphere_part_search_new(
        num_halo_points, coordinates_x, coordinates_y);
  }

  free(coordinates_y);
  free(coordinates_x);
  free(local_ids);
}

//! constructor for a nearest neighbour interpolation method
struct interpolation_method *
yac_new_interpolation_method_nnn_mask(struct global_search * search,
                                      struct domain * domain,
                                      char const * stack_name,
                                      struct points *** points,
                                      struct mask *** masks,
                                      unsigned num_pointsets,
                                      unsigned distance_weighted,
                                      unsigned n, unsigned points_idx,
                                      unsigned use_source_mask) {

   struct interpolation_method_nnn * method;

   method = malloc(1 * sizeof(*method));

   method->vtable = &interpolation_method_nnn_vtable;
   method->search = search;
   method->domain = domain;
   method->points = points;
   method->masks = (use_source_mask)?masks:NULL;
   method->num_pointsets = num_pointsets;
   method->finish_callback = NULL;
   method->finish_user_data = NULL;
   method->distance_weighted = distance_weighted;
   method->n = n;
   method->points_idx = points_idx;
   method->use_source_mask = use_source_mask;
   yac_get_domain_neighbours(*domain, CELL, &(method->domain_neighbour_ranks),
                             &(method->num_domain_neighbours));
   qsort(method->domain_neighbour_ranks, method->num_domain_neighbours,
         sizeof(*(method->domain_neighbour_ranks)), compare_uint);
   generate_point_sphere_part_searches(method);
   method->finished = 0;

   yac_init_interpolation_method((struct interpolation_method *)method, search);

   generate_interpolation_method_nnn_tag(method, stack_name);

   // ensure that the right point data is available

   if (num_pointsets == 0)
      yac_internal_abort_message ("ERROR(new_interpolation_method_nnn): missing "
                                  "data points.", __FILE__, __LINE__);

   if (points_idx >= num_pointsets)
      yac_internal_abort_message ("ERROR(new_interpolation_method_nnn): invalid "
                                  "points_idx.", __FILE__, __LINE__);

   if (points == NULL || points[0] == NULL || points[0][points_idx] == NULL)
      yac_internal_abort_message ("ERROR(new_interpolation_method_nnn): no point set "
                                  "data available", __FILE__, __LINE__);

   return (struct interpolation_method *)method;
}

//! constructor for a nearest neighbour interpolation method
struct interpolation_method *
yac_new_interpolation_method_nnn(struct global_search * search,
                                 struct domain * domain,
                                 char const * stack_name,
                                 struct points *** points,
                                 struct mask *** masks,
                                 unsigned num_pointsets,
                                 unsigned distance_weighted,
                                 unsigned n, unsigned points_idx) {

   unsigned use_source_mask = 1;

   return (struct interpolation_method *)yac_new_interpolation_method_nnn_mask(
      search, domain, stack_name, points, masks, num_pointsets, distance_weighted,
      n, points_idx, use_source_mask);
}

/**
 * callback routine passed to exchange_results_init by
 * do_search_continue_nnn
 */
static void
do_search_continue_nnn_2(void * user_data,
                         int * global_ids_to_mark,
                         unsigned num_global_ids) {

   struct intermediate_method_data * data =
      (struct intermediate_method_data *)user_data;
   struct interpolation_method_nnn * method_nnn =
      (struct interpolation_method_nnn*)(data->method_stack[0]);

   //-------------------------------------------------------------------
   // mark target points that are interpolated by another source process
   //-------------------------------------------------------------------
   mark_tgt_points(num_global_ids, global_ids_to_mark, data);

   //------------------------------------------------------------------------
   // all target points not marked as LOCAL or REMOTE have to be interpolated
   // by the local process
   //------------------------------------------------------------------------
   for (unsigned i = 0; i < data->search_data.num_tgt_points; ++i)
     if (data->search_data.tgt_point_data[i].flag == NOT)
       data->search_data.tgt_point_data[i].flag = LOCAL;

   int request_flag = 0;

   //--------------------------------------------------------------
   // sort tgt_point_data by the global ids and put the ones
   // flagged as LOCAL in front
   //--------------------------------------------------------------
   qsort(data->search_data.tgt_point_data, data->search_data.num_tgt_points,
         sizeof(*data->search_data.tgt_point_data),
         compare_tgt_point_data_flag);

   //-------------------------------------------------------------
   // count the number of target point that are to be interpolated
   // by the local process
   //-------------------------------------------------------------
   unsigned num_tgt_points_to_interpolate = 0;
   for (unsigned i = 0; i < data->search_data.num_tgt_points; ++i)
      if (data->search_data.tgt_point_data[i].flag == LOCAL)
         ++num_tgt_points_to_interpolate;
   data->search_data.num_tgt_points_to_interpolate =
       num_tgt_points_to_interpolate;

   //---------------------------------------------------------
   // ensure that every target point has a initial source cell
   //---------------------------------------------------------
   nnn_search_local(&(data->search_data), method_nnn,
                    yac_get_domain_comm(method_nnn->domain));

   data->request_records =
      request_records_new(method_nnn->num_domain_neighbours,
                          method_nnn->domain_neighbour_ranks);
   data->num_request_records = method_nnn->num_domain_neighbours;

   request_flag = nnn_search_remote(data, &(data->search_data),
                                    data->request_records, -1);

   if (request_flag)
     pack_and_send_search_request(data, data->request_records,
                                  method_nnn->num_domain_neighbours);

   // set up recv_callback for remote search requests
   set_up_recv_results_or_search_request_callback(method_nnn, data);

   // if there are no remote requests
   if (!request_flag) finish_nnn_interpolation(data);
}

/**
 * callback routine passed to yac_exchange_results by
 * do_search_interpolation_method_nnn
 */
static void
do_search_continue_nnn(void * user_data,
                       int * global_ids_to_mark,
                       unsigned num_global_ids) {

   struct intermediate_method_data * data;
   struct interpolation_method_nnn * method_nnn;

   data = (struct intermediate_method_data *)user_data;
   method_nnn = (struct interpolation_method_nnn*)(data->method_stack[0]);

   //-------------------------------------------------------------------
   // mark target points that are interpolated by another source process
   //-------------------------------------------------------------------
   mark_tgt_points(num_global_ids, global_ids_to_mark, data);

   //----------------------------------------------------------------
   // distribute the target points that did not match any source cell
   //----------------------------------------------------------------
   unsigned num_uninterpolated_points = 0;
    for (unsigned i = 0; i < data->search_data.num_tgt_points; ++i)
      if (data->search_data.tgt_point_data[i].flag == NOT)
        data->global_ids[num_uninterpolated_points++] =
          data->search_data.tgt_point_data[i].global_id;

   yac_exchange_results((struct interpolation_method*)method_nnn,
                        data->global_ids, num_uninterpolated_points,
                        num_uninterpolated_points,
                        method_nnn->tags.global_id_exchange_2,
                        yac_get_domain_comm(method_nnn->domain),
                        do_search_continue_nnn_2, data);
}

static unsigned
contains_local_cell(struct domain * domain, unsigned subdomain_id,
                    unsigned const * local_cell_ids, unsigned num_cells) {

  for (unsigned i = 0; i < num_cells; ++i)
    if (!yac_is_halo_value(domain, subdomain_id, CELL, local_cell_ids[i], NULL))
      return 1;
  return 0;
}

static void
do_search_interpolation_method_nnn(struct interpolation_method ** method,
                                   struct interpolation_plan_element ** plan_stack,
                                   int * global_ids, double * coordinates_x,
                                   double * coordinates_y,
                                   enum yac_location tgt_location,
                                   unsigned num_points,
                                   unsigned * num_interpolated_points,
                                   struct interpolation_method_data_request ** data_request) {

  if ( ( coordinates_x == NULL || coordinates_y == NULL ) && num_points > 0 )
    yac_internal_abort_message("ERROR: missing target point coordinates "
                  "(coordinates_x == NULL || coordinates_y == NULL)",
                  __FILE__, __LINE__);

  struct interpolation_method_nnn * method_nnn;
  method_nnn = (struct interpolation_method_nnn*)(method[0]);

  method_nnn->data_request = data_request;

  struct intermediate_method_data * data;
  struct search_data * search_data;

  struct domain * domain = method_nnn->domain;

  struct communicator * domain_comm = yac_get_domain_comm(domain);

  // initialise data that is required for the nearest neighbourhood search

  data = malloc(1 * sizeof(*data));
  data->method_stack = method;
  data->plan_stack = plan_stack;
  data->num_interpolated_points = num_interpolated_points;
  *num_interpolated_points = 0;
  data->coordinates_x = coordinates_x;
  data->coordinates_y = coordinates_y;
  data->global_ids = global_ids;
  data->tgt_location = tgt_location;
  search_data = &(data->search_data);
  search_data->tgt_point_data = tgt_point_search_data_new(
     num_points, method_nnn->n, coordinates_x, coordinates_y, global_ids);
  search_data->num_tgt_points = num_points;
  search_data->num_tgt_points_to_interpolate = num_points;
  search_data->num_open_remote_requests = 0;

  struct dep_list tgt_to_src_cell;
  unsigned num_subdomains;

  num_subdomains = yac_get_num_subdomains(domain);

  // do an initial search that tries to find for each target point a matching
  // local cell
  for (unsigned subdomain_id = 0; subdomain_id < num_subdomains;
       ++subdomain_id) {

    yac_do_point_search_c2(
      yac_get_subdomain_grid_search(method_nnn->search, subdomain_id), coordinates_x,
      coordinates_y, num_points, &tgt_to_src_cell);

    for (unsigned i = 0; i < num_points; ++i)
      if (contains_local_cell(
            domain, subdomain_id,
            yac_get_dependencies_of_element(tgt_to_src_cell, i),
            tgt_to_src_cell.num_deps_per_element[i]))
       search_data->tgt_point_data[i].flag = LOCAL;

    yac_free_dep_list(&tgt_to_src_cell);

    // give the communicator layer the chance to process messages
    yac_comm_check(domain_comm);
  }

  // sort the target point data by global ids
  qsort(data->search_data.tgt_point_data, data->search_data.num_tgt_points,
        sizeof(*data->search_data.tgt_point_data), compare_tgt_point_data);

  // count the number of target points for which at least one matching local
  // source cell was found
  unsigned num_matching_tgt_points = 0;
  for (unsigned i = 0; i < num_points; ++i)
    if (search_data->tgt_point_data[i].flag == LOCAL)
      data->global_ids[num_matching_tgt_points++] =
        search_data->tgt_point_data[i].global_id;
  for (unsigned i = 0, j = num_matching_tgt_points; i < num_points; ++i)
    if (search_data->tgt_point_data[i].flag != LOCAL)
      data->global_ids[j++] =
        search_data->tgt_point_data[i].global_id;

  // exchange the cells results of the initial search with the other source
  // processes and determine points for which multiple source processes have
  // found matching cells once this exchange is finished the algorithms
  // continues in the routine do_search_continue_nnn
  yac_exchange_results((struct interpolation_method*)method_nnn,
                       data->global_ids, num_matching_tgt_points, num_points,
                       method_nnn->tags.global_id_exchange_1,
                       domain_comm, do_search_continue_nnn, data);
}

/**
 * blocking routine that ensures that the search is finished and everything
 * is cleaned up
 */
static void
enddef_interpolation_method_nnn(struct interpolation_method ** method) {

   struct interpolation_method_nnn * method_nnn =
      (struct interpolation_method_nnn*)(method[0]);

   struct communicator * domain_comm = yac_get_domain_comm(method_nnn->domain);

   while(yac_comm_wait(domain_comm, method_nnn->tags.global_id_exchange_1[0]) ||
         yac_comm_wait(domain_comm, method_nnn->tags.global_id_exchange_1[1]) ||
         yac_comm_wait(domain_comm, method_nnn->tags.global_id_exchange_2[0]) ||
         yac_comm_wait(domain_comm, method_nnn->tags.global_id_exchange_2[1]) ||
         yac_comm_wait(domain_comm, method_nnn->tags.exchange_results[0]) ||
         yac_comm_wait(domain_comm, method_nnn->tags.exchange_results[1]) ||
         yac_comm_wait(domain_comm,
                       method_nnn->tags.results_or_search_request_header) ||
         yac_comm_wait(domain_comm, method_nnn->tags.search_request) ||
         yac_comm_wait(domain_comm, method_nnn->tags.results) ||
         yac_comm_wait(domain_comm, method_nnn->tags.finished));

   unsigned num_subdomains = yac_get_num_subdomains(method_nnn->domain);
   unsigned num_neighbours = method_nnn->num_domain_neighbours;

   yac_delete_point_sphere_part_search(method_nnn->non_halo_point_search);
   free(method_nnn->non_halo_local_ids);
   free(method_nnn->num_non_halo_local_ids_per_subdomain_prefix);
   for (unsigned i = 0; i < num_neighbours; ++i)
      yac_delete_point_sphere_part_search(method_nnn->halo_point_search[i]);
   free(method_nnn->halo_point_search);

   free(method_nnn->domain_neighbour_ranks);

   if (method[1] != NULL)
      method[1]->vtable->enddef(method + 1);
}

static void get_source_exchange_data_interpolation_method_nnn (
   struct interpolation_method ** method,
   struct interpolation_method_data_request *** data_request) {

   struct interpolation_method_nnn * method_nnn;

   method_nnn = (struct interpolation_method_nnn*)(method[0]);

   *data_request = method_nnn->data_request;
   method_nnn->data_request = NULL;
}

static void
delete_interpolation_method_nnn(struct interpolation_method * method) {

   free(method);
}

 static struct global_search *
 get_global_search(struct interpolation_method * method) {

   return ((struct interpolation_method_nnn*)(method))->search;
 }

 static void
 set_finish_callback(struct interpolation_method * method,
                     void (*callback)(void * user_data),
                     void * user_data) {

   struct interpolation_method_nnn * method_nnn;

   method_nnn = (struct interpolation_method_nnn*)method;

   method_nnn->finish_callback = callback;
   method_nnn->finish_user_data = user_data;
}
