/**
 * @file interpolation_method_nnn.h
 * @brief n nearest neighbour interpolation method
 *
 * point interpolation method that assigns each target point the average of
 * the n nearest non-masked source points. If the algorithm does not find n
 * source point, no value is assigned.
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#ifndef INTERPOLATION_METHOD_NNN_H
#define INTERPOLATION_METHOD_NNN_H

#include "interpolation_method.h"

 /** \example test_interpolation_method_nnn.c
 * This shows how the n-nearest-neighbour interpolation works
 */

/** \example test_interpolation_method_nnn_parallel.c
 * This contains a parallel example for the n-nearest-neighbour interpolation
 * method. The example uses four source processes and one target process.
 */
 
/**
 * constructor for a interpolation method of type interpolation_method_nnn
 * @param[in] search             global search object that is to be used for the
 *                               interpolation
 * @param[in] domain             %domain of the local process
 * @param[in] stack_name         name of the interpolation stack
 * @param[in] points             points to be used for the weight generation
 * @param[in] masks              mask to be used for the weight generation
 * @param[in] num_pointsets      number of point sets
 * @param[in] distance_weighted  flag to switch on and off weighting by distance
 * @param[in] n                  number of nearest neighbours per target point
 * @param[in] points_idx         which points set of the domain is to be used for the
 *                               interpolation
 *
 * \remarks all subdomains need to have a halo of at least one cell in order to
 *          be able to do interpolation across process boundaries
 */
struct interpolation_method *
yac_new_interpolation_method_nnn(struct global_search * search,
                                 struct domain * domain,
                                 char const * stack_name,
                                 struct points *** points,
                                 struct mask *** masks,
                                 unsigned num_pointsets,
                                 unsigned distance_weighted, unsigned n,
                                 unsigned points_idx);

/**
 * constructor for a interpolation method of type interpolation_method_nnn
 * @param[in] search            global search object that is to be used for the interpolation
 * @param[in] domain            %domain of the local process
 * @param[in] stack_name        name of the interpolation stack
 * @param[in] points            points to be used for the weight generation
 * @param[in] masks             mask to be used for the weight generation
 * @param[in] num_pointsets     number of point sets
 * @param[in] distance_weighted flag to switch on and off weighting by distance
 * @param[in] n                 number of nearest neighbours per target point
 * @param[in] points_idx        which points set of the domain is to be used for the
 *                              interpolation
 * @param[in] use_source_mask   specifies how to handle masked source points
 *
 * \remarks all subdomains need to have a halo of at least one cell in order to
 *          be able to do interpolation across process boundaries
 * \remarks halo points need to have valid mask values
 */
struct interpolation_method *
yac_new_interpolation_method_nnn_mask(struct global_search * search,
                                      struct domain * domain,
                                      char const * stack_name,
                                      struct points *** points,
                                      struct mask *** masks,
                                      unsigned num_pointsets,
                                      unsigned distance_weighted, unsigned n,
                                      unsigned points_idx,
                                      unsigned use_source_mask);

#endif // INTERPOLATION_METHOD_NNN_H
