/**
 * @file interpolation_method_patch.h
 * @brief patch recovery interpolation method
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 */
/*
 * Keywords:
 * Maintainer: Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#ifndef INTERPOLATION_METHOD_PATCH_H
#define INTERPOLATION_METHOD_PATCH_H

/** \example test_interpolation_method_patch.c
 * This shows how the patch recovery interpolation works
 */

/** \example test_interpolation_method_patch_parallel.c
 * This contains a parallel example for the patch recovery interpolation
 * method. The example uses four source and four target processes.
 */

#include "interpolation_method.h"

/**
 * @file interpolation_method_patch.h
 * @brief patch recovery interpolation method
 *
 * cell interpolation method that searches for each target cell the overlapping
 * source cells which make up the patch. A polynomial fit is applied over the
 * which then is used to provide function values at any point within this patch.
 *
 */


/**
 * constructor for a interpolation method of type interpolation_method_patch
 * @param[in] search              global search object that is to be used for the interpolation
 * @param[in] domain              %domain of the local process
 * @param[in] stack_name          name of the interpolation stack
 * @param[in] points              points to be used for the weight generation
 * @param[in] masks               mask to be used for the weight generation
 * @param[in] num_pointsets       number of point sets
 * @param[in] polynom_order       order of the polynomial to be fitted
 * @param[in] gauss_order         gauss_order, is transferred into the number of gauss points per cell
 * @param[in] extent_source_patch allow extension of the overlapping patch by one row of neighbouring cells
 * @param[in] allow_extrapolation allow interpolation for target points that are outside the patch
 * @param[in] points_idx          which points set of the domain is to be used for the interpolation
 *
 */
struct interpolation_method *
yac_new_interpolation_method_patch ( struct global_search * search,
                                     struct domain * domain,
                                     char const * stack_name,
                                     struct points *** points,
                                     struct mask *** masks,
                                     unsigned num_pointsets,
                                     unsigned polynom_order,
                                     unsigned gauss_order,
                                     unsigned extent_source_patch,
                                     unsigned allow_extrapolation,
                                     unsigned points_idx );

/**
 * constructor for a interpolation method of type interpolation_method_patch
 * @param[in] search              global search object that is to be used for the interpolation
 * @param[in] domain              %domain of the local process
 * @param[in] stack_name          name of the interpolation stack
 * @param[in] points              points to be used for the weight generation
 * @param[in] masks               mask to be used for the weight generation
 * @param[in] num_pointsets       number of point sets
 * @param[in] polynom_order       order of the polynomial to be fitted
 * @param[in] gauss_order         gauss_order, is transferred into the number of gauss points per cell
 * @param[in] extent_source_patch allow extension of the overlapping patch by one row of neighbouring cells
 * @param[in] allow_extrapolation allow interpolation for target points that are outside the patch
 * @param[in] points_idx          which points set of the domain is to be used for the interpolation
 * @param[in] use_source_mask     specifies whether masked source cells shall be considered (true) or not (false)
 *
 */
struct interpolation_method *
yac_new_interpolation_method_patch_mask ( struct global_search * search,
                                          struct domain * domain,
                                          char const * stack_name,
                                          struct points *** points,
                                          struct mask *** masks,
                                          unsigned num_pointsets,
                                          unsigned polynom_order,
                                          unsigned gauss_order,
                                          unsigned extent_source_patch,
                                          unsigned allow_extrapolation,
                                          unsigned points_idx,
                                          unsigned use_source_mask );

#endif // INTERPOLATION_METHOD_PATCH_H
