/**
 * @file interpolation_plan.c
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#include <stdlib.h>
#include <string.h>

#include "utils.h"
#include "interpolation_plan.h"

struct fixed_value_sort_data {
  double value;
  unsigned plan_idx;
};

void yac_execute_interpolation_plan_element(struct interpolation_plan_element ** plan_stack,
                                            double *** input_data, double * output_data) {

   if (plan_stack[0] != NULL)
      plan_stack[0]->do_interpolation(plan_stack, input_data, output_data);
}

void yac_free_interpolation_plan_element(struct interpolation_plan_element * plan) {

   if (plan == NULL)
      return;

   plan->free_interpolation_plan_element(plan);
}

//-------------------------------------------------------------------------
// interpolation plan that assigning a fixed value a number of points/cells
//-------------------------------------------------------------------------

struct interpolation_plan_element_fixed {

   unsigned num_points;
   double fixed_value;
};

static void do_interpolation_fixed(struct interpolation_plan_element ** plan_stack,
                                   double *** restrict input_data,
                                   double * restrict output_data) {

   struct interpolation_plan_element_fixed * plan;

   plan = plan_stack[0]->child_data;

   unsigned i;

   for (i = 0; i < plan->num_points; ++i)
      output_data[i] = plan->fixed_value;

   if (plan_stack[1] != NULL)
      plan_stack[1]->do_interpolation(plan_stack+1, input_data, output_data+plan->num_points);
}

static void free_interpolation_plan_element_basic(struct interpolation_plan_element * plan) {

   free(plan->child_data);
}

static void get_link_data_fixed(struct interpolation_plan_element ** plan_stack,
                                int *** global_indices,
                                unsigned * num_targets_per_plan,
                                unsigned * num_links_per_plan,
                                unsigned * num_weights_per_link_per_plan,
                                unsigned ** num_links_per_tgt_per_plan,
                                int ** src_indices_per_plan,
                                unsigned ** pointset_indices_per_plan,
                                double ** weights_per_plan) {

  struct interpolation_plan_element_fixed * plan =
    plan_stack[0]->child_data;

  *num_targets_per_plan = plan->num_points;
  *num_links_per_plan = 0;
  *num_weights_per_link_per_plan = 0;
  *num_links_per_tgt_per_plan = NULL;
  *src_indices_per_plan = NULL;
  *pointset_indices_per_plan = NULL;
  *weights_per_plan = NULL;

  if (plan_stack[1] != NULL)
    plan_stack[1]->get_link_data(plan_stack+1, global_indices,
                                 num_targets_per_plan+1,
                                 num_links_per_plan+1,
                                 num_weights_per_link_per_plan+1,
                                 num_links_per_tgt_per_plan+1,
                                 src_indices_per_plan+1,
                                 pointset_indices_per_plan+1,
                                 weights_per_plan+1);
}

static unsigned get_num_fixed_values_fixed (
  struct interpolation_plan_element ** plan_stack) {

  struct interpolation_plan_element_fixed * plan =
    plan_stack[0]->child_data;

  if (plan_stack[1] != NULL)
    return (plan->num_points > 0) +
           plan_stack[1]->get_num_fixed_values(plan_stack+1);
  else
    return (plan->num_points > 0);
}

static void
get_fixed_data_fixed(struct interpolation_plan_element ** plan_stack,
                     int const * global_ids,
                     int ** global_tgt_indices_per_plan,
                     unsigned * num_tgt_indices_per_plan,
                     double * fixed_value_per_plan) {

  struct interpolation_plan_element_fixed * plan =
    plan_stack[0]->child_data;

  if (plan->num_points > 0) {
    *global_tgt_indices_per_plan =
      malloc(plan->num_points * sizeof(**global_tgt_indices_per_plan));
    *num_tgt_indices_per_plan = plan->num_points;
    *fixed_value_per_plan = plan->fixed_value;

    memcpy(*global_tgt_indices_per_plan, global_ids,
           plan->num_points * sizeof(*global_ids));

    global_tgt_indices_per_plan += 1;
    num_tgt_indices_per_plan += 1;
    fixed_value_per_plan += 1;
  }

  if (plan_stack[1] != NULL)
    plan_stack[1]->get_fixed_data(plan_stack+1, global_ids+plan->num_points,
                                  global_tgt_indices_per_plan,
                                  num_tgt_indices_per_plan,
                                  fixed_value_per_plan);
}

void yac_init_interpolation_plan_element_fixed(struct interpolation_plan_element * plan, unsigned num_points,
                                               double fixed_value) {

   struct interpolation_plan_element_fixed * plan_fixed;

   plan_fixed = malloc(1 * sizeof(*plan_fixed));

   plan_fixed->num_points = num_points;
   plan_fixed->fixed_value = fixed_value;

   plan->child_data = plan_fixed;
   plan->do_interpolation = do_interpolation_fixed;
   plan->get_link_data = get_link_data_fixed;
   plan->get_num_fixed_values = get_num_fixed_values_fixed;
   plan->get_fixed_data = get_fixed_data_fixed;
   plan->free_interpolation_plan_element = free_interpolation_plan_element_basic;
}

//----------------------------------------------------------------------------------
// interpolation plan that generates the output values by directly assiging
// values from the input data to the output
//----------------------------------------------------------------------------------

struct interpolation_plan_element_point {

   unsigned num_interpolated_points;
   unsigned * restrict src_indices;
   unsigned * num_remote_points_per_points_set;
   unsigned num_pointsets;
};

static void
do_interpolation_point(struct interpolation_plan_element ** plan_stack,
                       double *** restrict input_data,
                       double * restrict output_data) {


  struct interpolation_plan_element_point * plan
    = plan_stack[0]->child_data;

  for (unsigned i = 0; i < plan->num_interpolated_points; ++i)
    output_data[i] = input_data[plan->src_indices[i*3+0]]
                               [plan->src_indices[i*3+1]]
                               [plan->src_indices[i*3+2]];

  for (unsigned i = 0; i < plan->num_pointsets; ++i)
    if (plan->num_remote_points_per_points_set[i] > 0)
      input_data[0][i] += plan->num_remote_points_per_points_set[i];
  output_data += plan->num_interpolated_points;

  if (plan_stack[1] != NULL)
    plan_stack[1]->do_interpolation(plan_stack+1, input_data, output_data);

  for (unsigned i = 0; i < plan->num_pointsets; ++i)
    if (plan->num_remote_points_per_points_set[i] > 0)
      input_data[0][i] -= plan->num_remote_points_per_points_set[i];
}

static void
free_interpolation_plan_element_point(
  struct interpolation_plan_element * plan) {

  struct interpolation_plan_element_point * plan_p = plan->child_data;

  free(plan_p->src_indices);
  free(plan_p->num_remote_points_per_points_set);
  free(plan_p);
}

static void
get_link_data_point(struct interpolation_plan_element ** plan_stack,
                    int *** global_indices,
                    unsigned * num_targets_per_plan,
                    unsigned * num_links_per_plan,
                    unsigned * num_weights_per_link_per_plan,
                    unsigned ** num_links_per_tgt_per_plan,
                    int ** src_indices_per_plan,
                    unsigned ** pointset_indices_per_plan,
                    double ** weights_per_plan) {

  struct interpolation_plan_element_point * plan =
    plan_stack[0]->child_data;

  *num_targets_per_plan = plan->num_interpolated_points;
  *num_links_per_plan = plan->num_interpolated_points;
  *num_weights_per_link_per_plan = 1;
  *num_links_per_tgt_per_plan =
    malloc(plan->num_interpolated_points *
           sizeof(**num_links_per_tgt_per_plan));
  for (unsigned i = 0; i < plan->num_interpolated_points; ++i)
    (*num_links_per_tgt_per_plan)[i] = 1;
  *src_indices_per_plan =
    malloc(plan->num_interpolated_points * sizeof(**src_indices_per_plan));
  *pointset_indices_per_plan =
    malloc(plan->num_interpolated_points * sizeof(**pointset_indices_per_plan));
  *weights_per_plan =
    malloc(plan->num_interpolated_points * sizeof(**weights_per_plan));
  for (unsigned i = 0; i < plan->num_interpolated_points; ++i)
    (*weights_per_plan)[i] = 1.0;

  for (unsigned i = 0; i < plan->num_interpolated_points; ++i) {
    (*src_indices_per_plan)[i] =
      global_indices[plan->src_indices[i*3+0]]
                    [plan->src_indices[i*3+1]]
                    [plan->src_indices[i*3+2]];
    (*pointset_indices_per_plan)[i] = plan->src_indices[i*3+1];
  }

  for (unsigned i = 0; i < plan->num_pointsets; ++i)
    if (plan->num_remote_points_per_points_set[i] > 0)
      global_indices[0][i] += plan->num_remote_points_per_points_set[i];

  if (plan_stack[1] != NULL)
    plan_stack[1]->get_link_data(plan_stack+1, global_indices,
                                 num_targets_per_plan+1,
                                 num_links_per_plan+1,
                                 num_weights_per_link_per_plan+1,
                                 num_links_per_tgt_per_plan+1,
                                 src_indices_per_plan+1,
                                 pointset_indices_per_plan+1,
                                 weights_per_plan+1);

  for (unsigned i = 0; i < plan->num_pointsets; ++i)
    if (plan->num_remote_points_per_points_set[i] > 0)
      global_indices[0][i] -= plan->num_remote_points_per_points_set[i];
}

static unsigned get_num_fixed_values_point(
  struct interpolation_plan_element ** plan_stack) {

  if (plan_stack[1] != NULL)
    return plan_stack[1]->get_num_fixed_values(plan_stack+1);
  else return 0;
}

static void
get_fixed_data_point(struct interpolation_plan_element ** plan_stack,
                     int const * global_ids,
                     int ** global_tgt_indices_per_plan,
                     unsigned * num_tgt_indices_per_plan,
                     double * fixed_value_per_plan) {

  struct interpolation_plan_element_point * plan =
    plan_stack[0]->child_data;

  if (plan_stack[1] != NULL)
    plan_stack[1]->get_fixed_data(plan_stack+1,
                                  global_ids+plan->num_interpolated_points,
                                  global_tgt_indices_per_plan,
                                  num_tgt_indices_per_plan,
                                  fixed_value_per_plan);
}

void
yac_init_interpolation_plan_element_point(
  struct interpolation_plan_element * plan, unsigned num_interpolated_points,
  unsigned * src_indices, unsigned * num_remote_points_per_points_set,
  unsigned num_pointsets) {

  struct interpolation_plan_element_point * plan_p
    = malloc(1 * sizeof(*plan_p));

  plan_p->num_interpolated_points = num_interpolated_points;
  plan_p->src_indices =
    malloc(3 * num_interpolated_points * sizeof(*(plan_p->src_indices)));
  plan_p->num_remote_points_per_points_set =
    malloc(num_pointsets * sizeof(*(plan_p->num_remote_points_per_points_set)));
  plan_p->num_pointsets = num_pointsets;
  memcpy(plan_p->num_remote_points_per_points_set,
         num_remote_points_per_points_set,
         num_pointsets * sizeof(*(plan_p->num_remote_points_per_points_set)));
  memcpy(plan_p->src_indices, src_indices,
         3 * num_interpolated_points * sizeof(*(plan_p->src_indices)));

  plan->child_data                      = plan_p;
  plan->do_interpolation                = do_interpolation_point;
  plan->get_link_data                   = get_link_data_point;
  plan->get_num_fixed_values            = get_num_fixed_values_point;
  plan->get_fixed_data                  = get_fixed_data_point;
  plan->free_interpolation_plan_element = free_interpolation_plan_element_point;
}

//----------------------------------------------------------------------------------
// interpolation plan that generates the output values by the sum of some input data
// points
//----------------------------------------------------------------------------------

struct interpolation_plan_element_pointsum {

   unsigned num_interpolated_points;
   unsigned * restrict src_indices;
   unsigned * num_src_per_tgt;
   unsigned * num_remote_points_per_points_set;
   unsigned num_pointsets;
};

static void
do_interpolation_pointsum(struct interpolation_plan_element ** plan_stack,
                          double *** restrict input_data,
                          double * restrict output_data) {


  struct interpolation_plan_element_pointsum * plan
    = plan_stack[0]->child_data;

  for (unsigned i = 0, k = 0; i < plan->num_interpolated_points; ++i) {

    double sum = 0;

    for (unsigned j = 0; j < plan->num_src_per_tgt[i]; ++j, ++k)
      sum +=
        input_data[plan->src_indices[k*3+0]]
                  [plan->src_indices[k*3+1]]
                  [plan->src_indices[k*3+2]];

    output_data[i] = sum;
  }

  for (unsigned i = 0; i < plan->num_pointsets; ++i)
    if (plan->num_remote_points_per_points_set[i] > 0)
      input_data[0][i] += plan->num_remote_points_per_points_set[i];
  output_data += plan->num_interpolated_points;

  if (plan_stack[1] != NULL)
    plan_stack[1]->do_interpolation(plan_stack+1, input_data, output_data);

  for (unsigned i = 0; i < plan->num_pointsets; ++i)
    if (plan->num_remote_points_per_points_set[i] > 0)
      input_data[0][i] -= plan->num_remote_points_per_points_set[i];
}

static void
free_interpolation_plan_element_pointsum(
  struct interpolation_plan_element * plan) {

  struct interpolation_plan_element_pointsum * plan_p = plan->child_data;

  free(plan_p->src_indices);
  free(plan_p->num_src_per_tgt);
  free(plan_p->num_remote_points_per_points_set);

  free(plan_p);
}

static void
get_link_data_pointsum(struct interpolation_plan_element ** plan_stack,
                       int *** global_indices,
                       unsigned * num_targets_per_plan,
                       unsigned * num_links_per_plan,
                       unsigned * num_weights_per_link_per_plan,
                       unsigned ** num_links_per_tgt_per_plan,
                       int ** src_indices_per_plan,
                       unsigned ** pointset_indices_per_plan,
                       double ** weights_per_plan) {

  struct interpolation_plan_element_pointsum * plan =
    plan_stack[0]->child_data;

  unsigned total_num_links = 0;

  for (unsigned i = 0; i < plan->num_interpolated_points; ++i)
    total_num_links += plan->num_src_per_tgt[i];

  *num_targets_per_plan = plan->num_interpolated_points;
  *num_links_per_plan = total_num_links;
  *num_weights_per_link_per_plan = 1;
  *num_links_per_tgt_per_plan =
    malloc(plan->num_interpolated_points *
           sizeof(**num_links_per_tgt_per_plan));
  memcpy(*num_links_per_tgt_per_plan, plan->num_src_per_tgt,
         plan->num_interpolated_points *
         sizeof(**num_links_per_tgt_per_plan));
  *src_indices_per_plan =
    malloc(total_num_links * sizeof(**src_indices_per_plan));
  *pointset_indices_per_plan =
    malloc(total_num_links * sizeof(**pointset_indices_per_plan));
  *weights_per_plan = malloc(total_num_links * sizeof(**weights_per_plan));
  for (unsigned i = 0; i < total_num_links; ++i)
    (*weights_per_plan)[i] = 1.0;

  for (unsigned i = 0, k = 0; i < plan->num_interpolated_points; ++i) {
    for (unsigned j = 0; j < plan->num_src_per_tgt[i]; ++j, ++k) {
      (*src_indices_per_plan)[k] =
        global_indices[plan->src_indices[k*3+0]]
                      [plan->src_indices[k*3+1]]
                      [plan->src_indices[k*3+2]];
      (*pointset_indices_per_plan)[k] = plan->src_indices[k*3+1];
    }
  }

  for (unsigned i = 0; i < plan->num_pointsets; ++i)
    if (plan->num_remote_points_per_points_set[i] > 0)
      global_indices[0][i] += plan->num_remote_points_per_points_set[i];

  if (plan_stack[1] != NULL)
    plan_stack[1]->get_link_data(plan_stack+1, global_indices,
                                 num_targets_per_plan+1,
                                 num_links_per_plan+1,
                                 num_weights_per_link_per_plan+1,
                                 num_links_per_tgt_per_plan+1,
                                 src_indices_per_plan+1,
                                 pointset_indices_per_plan+1,
                                 weights_per_plan+1);

  for (unsigned i = 0; i < plan->num_pointsets; ++i)
    if (plan->num_remote_points_per_points_set[i] > 0)
      global_indices[0][i] -= plan->num_remote_points_per_points_set[i];
}

static unsigned get_num_fixed_values_pointsum(
  struct interpolation_plan_element ** plan_stack) {

  if (plan_stack[1] != NULL)
    return plan_stack[1]->get_num_fixed_values(plan_stack+1);
  else return 0;
}

static void
get_fixed_data_pointsum(struct interpolation_plan_element ** plan_stack,
                        int const * global_ids,
                        int ** global_tgt_indices_per_plan,
                        unsigned * num_tgt_indices_per_plan,
                        double * fixed_value_per_plan) {

  struct interpolation_plan_element_pointsum * plan =
    plan_stack[0]->child_data;

  if (plan_stack[1] != NULL)
    plan_stack[1]->get_fixed_data(plan_stack+1,
                                  global_ids+plan->num_interpolated_points,
                                  global_tgt_indices_per_plan,
                                  num_tgt_indices_per_plan,
                                  fixed_value_per_plan);
}

void
yac_init_interpolation_plan_element_pointsum(
  struct interpolation_plan_element * plan, unsigned num_interpolated_points,
  unsigned * src_indices, unsigned * num_src_per_tgt,
  unsigned * num_remote_points_per_points_set, unsigned num_pointsets) {

  struct interpolation_plan_element_pointsum * plan_p
    = malloc(1 * sizeof(*plan_p));

  unsigned total_num_dependencies = 0;

  for (unsigned i = 0; i < num_interpolated_points; ++i)
    total_num_dependencies += num_src_per_tgt[i];

  plan_p->num_interpolated_points          = num_interpolated_points;
  plan_p->src_indices                      = malloc(3 * total_num_dependencies * sizeof(*(plan_p->src_indices)));
  plan_p->num_src_per_tgt                  = malloc(num_interpolated_points * sizeof(*(plan_p->num_src_per_tgt)));
  plan_p->num_remote_points_per_points_set = malloc(num_pointsets * sizeof(*(plan_p->num_remote_points_per_points_set)));
  plan_p->num_pointsets                    = num_pointsets;
  memcpy(plan_p->num_remote_points_per_points_set,
         num_remote_points_per_points_set,
         num_pointsets * sizeof(*(plan_p->num_remote_points_per_points_set)));
  memcpy(plan_p->num_src_per_tgt,
         num_src_per_tgt,
         num_interpolated_points * sizeof(*(plan_p->num_src_per_tgt)));
  memcpy(plan_p->src_indices, src_indices, 3 * total_num_dependencies * sizeof(*(plan_p->src_indices)));

  plan->child_data                      = plan_p;
  plan->do_interpolation                = do_interpolation_pointsum;
  plan->get_link_data                   = get_link_data_pointsum;
  plan->get_num_fixed_values            = get_num_fixed_values_pointsum;
  plan->get_fixed_data                  = get_fixed_data_pointsum;
  plan->free_interpolation_plan_element = free_interpolation_plan_element_pointsum;
}

//----------------------------------------------------------------------------------
// interpolation plan that generates the output values by the sum of some input data
// points that are multiplied by a predefined value
//----------------------------------------------------------------------------------

struct interpolation_plan_element_wpointsum {

   unsigned num_interpolated_points;
   unsigned * restrict src_indices;
   unsigned * num_src_per_tgt;
   double * restrict weights;
   unsigned * num_remote_points_per_points_set;
   unsigned num_pointsets;
};

static void do_interpolation_wpointsum(struct interpolation_plan_element ** plan_stack,
                                       double *** restrict input_data,
                                       double * restrict output_data) {

   unsigned i, j, k;

   struct interpolation_plan_element_wpointsum * plan;

   plan = plan_stack[0]->child_data;

   for (i = 0, k = 0; i < plan->num_interpolated_points; ++i) {

      double sum = 0;

      for (j = 0; j < plan->num_src_per_tgt[i]; ++j, ++k)
         sum += input_data[plan->src_indices[k*3+0]][plan->src_indices[k*3+1]][plan->src_indices[k*3+2]] *
            plan->weights[k];

      output_data[i] = sum;
   }

   for (i = 0; i < plan->num_pointsets; ++i)
      if (plan->num_remote_points_per_points_set[i] > 0)
         input_data[0][i] += plan->num_remote_points_per_points_set[i];
   output_data += plan->num_interpolated_points;

   if (plan_stack[1] != NULL)
      plan_stack[1]->do_interpolation(plan_stack+1, input_data, output_data);

   for (i = 0; i < plan->num_pointsets; ++i)
      if (plan->num_remote_points_per_points_set[i] > 0)
         input_data[0][i] -= plan->num_remote_points_per_points_set[i];
}

static void free_interpolation_plan_element_wpointsum(struct interpolation_plan_element * plan) {

   struct interpolation_plan_element_wpointsum * plan_wp;

   plan_wp = plan->child_data;

   free(plan_wp->src_indices);
   free(plan_wp->num_src_per_tgt);
   free(plan_wp->weights);
   free(plan_wp->num_remote_points_per_points_set);

   free(plan_wp);
}

static void
get_link_data_wpointsum(struct interpolation_plan_element ** plan_stack,
                        int *** global_indices,
                        unsigned * num_targets_per_plan,
                        unsigned * num_links_per_plan,
                        unsigned * num_weights_per_link_per_plan,
                        unsigned ** num_links_per_tgt_per_plan,
                        int ** src_indices_per_plan,
                        unsigned ** pointset_indices_per_plan,
                        double ** weights_per_plan) {

  struct interpolation_plan_element_wpointsum * plan =
    plan_stack[0]->child_data;

  unsigned total_num_links = 0;

  for (unsigned i = 0; i < plan->num_interpolated_points; ++i)
    total_num_links += plan->num_src_per_tgt[i];

  *num_targets_per_plan = plan->num_interpolated_points;
  *num_links_per_plan = total_num_links;
  *num_weights_per_link_per_plan = 1;
  *num_links_per_tgt_per_plan =
    malloc(plan->num_interpolated_points *
           sizeof(**num_links_per_tgt_per_plan));
  memcpy(*num_links_per_tgt_per_plan, plan->num_src_per_tgt,
         plan->num_interpolated_points *
         sizeof(**num_links_per_tgt_per_plan));
  *src_indices_per_plan =
    malloc(total_num_links * sizeof(**src_indices_per_plan));
  *pointset_indices_per_plan =
    malloc(total_num_links * sizeof(**pointset_indices_per_plan));
  *weights_per_plan = malloc(total_num_links * sizeof(**weights_per_plan));
  memcpy(*weights_per_plan, plan->weights,
         total_num_links * sizeof(**weights_per_plan));

  for (unsigned i = 0, k = 0; i < plan->num_interpolated_points; ++i) {
    for (unsigned j = 0; j < plan->num_src_per_tgt[i]; ++j, ++k) {
      (*src_indices_per_plan)[k] =
        global_indices[plan->src_indices[k*3+0]]
                      [plan->src_indices[k*3+1]]
                      [plan->src_indices[k*3+2]];
      (*pointset_indices_per_plan)[k] = plan->src_indices[k*3+1];
    }
  }

  for (unsigned i = 0; i < plan->num_pointsets; ++i)
    if (plan->num_remote_points_per_points_set[i] > 0)
      global_indices[0][i] += plan->num_remote_points_per_points_set[i];

  if (plan_stack[1] != NULL)
    plan_stack[1]->get_link_data(plan_stack+1, global_indices,
                                 num_targets_per_plan+1,
                                 num_links_per_plan+1,
                                 num_weights_per_link_per_plan+1,
                                 num_links_per_tgt_per_plan+1,
                                 src_indices_per_plan+1,
                                 pointset_indices_per_plan+1,
                                 weights_per_plan+1);

  for (unsigned i = 0; i < plan->num_pointsets; ++i)
    if (plan->num_remote_points_per_points_set[i] > 0)
      global_indices[0][i] -= plan->num_remote_points_per_points_set[i];
}

static unsigned get_num_fixed_values_wpointsum(
  struct interpolation_plan_element ** plan_stack) {

  if (plan_stack[1] != NULL)
    return plan_stack[1]->get_num_fixed_values(plan_stack+1);
  else return 0;
}

static void
get_fixed_data_wpointsum(struct interpolation_plan_element ** plan_stack,
                         int const * global_ids,
                         int ** global_tgt_indices_per_plan,
                         unsigned * num_tgt_indices_per_plan,
                         double * fixed_value_per_plan) {

  struct interpolation_plan_element_wpointsum * plan =
    plan_stack[0]->child_data;

  if (plan_stack[1] != NULL)
    plan_stack[1]->get_fixed_data(plan_stack+1,
                                  global_ids+plan->num_interpolated_points,
                                  global_tgt_indices_per_plan,
                                  num_tgt_indices_per_plan,
                                  fixed_value_per_plan);
}

void yac_init_interpolation_plan_element_wpointsum(struct interpolation_plan_element * plan,
                                                   unsigned num_interpolated_points,
                                                   unsigned * src_indices,
                                                   unsigned * num_src_per_tgt, double * weights,
                                                   unsigned * num_remote_points_per_points_set,
                                                   unsigned num_pointsets) {

   struct interpolation_plan_element_wpointsum * plan_wp;

   plan_wp = malloc(1 * sizeof(*plan_wp));

   unsigned i;
   unsigned total_num_dependencies;

   for (i = 0, total_num_dependencies = 0; i < num_interpolated_points; ++i)
      total_num_dependencies += num_src_per_tgt[i];

   plan_wp->num_interpolated_points          = num_interpolated_points;
   plan_wp->src_indices                      = malloc(3 * total_num_dependencies * sizeof(*(plan_wp->src_indices)));
   plan_wp->num_src_per_tgt                  = malloc(num_interpolated_points * sizeof(*(plan_wp->num_src_per_tgt)));
   plan_wp->weights                          = malloc(total_num_dependencies * sizeof(*(plan_wp->weights)));
   plan_wp->num_remote_points_per_points_set = malloc(num_pointsets * sizeof(*(plan_wp->num_remote_points_per_points_set)));
   plan_wp->num_pointsets                    = num_pointsets;
   memcpy(plan_wp->num_remote_points_per_points_set,
          num_remote_points_per_points_set,
          num_pointsets * sizeof(*(plan_wp->num_remote_points_per_points_set)));
   memcpy(plan_wp->weights,
          weights,
          total_num_dependencies * sizeof(*(plan_wp->weights)));
   memcpy(plan_wp->num_src_per_tgt,
          num_src_per_tgt,
          num_interpolated_points * sizeof(*(plan_wp->num_src_per_tgt)));
   memcpy(plan_wp->src_indices, src_indices, 3 * total_num_dependencies * sizeof(*(plan_wp->src_indices)));

   plan->child_data                      = plan_wp;
   plan->do_interpolation                = do_interpolation_wpointsum;
   plan->get_link_data                   = get_link_data_wpointsum;
   plan->get_num_fixed_values            = get_num_fixed_values_wpointsum;
   plan->get_fixed_data                  = get_fixed_data_wpointsum;
   plan->free_interpolation_plan_element = free_interpolation_plan_element_wpointsum;
}

//------------------------------------------------------------------------------
// interpolation plan that consists of multiple individual interpolation plans
//------------------------------------------------------------------------------

struct interpolation_plan_element_collection {

  unsigned num_plans;
  struct interpolation_plan_element ** plans;
};

static void do_interpolation_collection(
  struct interpolation_plan_element ** plan_stack,
  double *** input_data, double * output_data) {

  struct interpolation_plan_element_collection * plan =
    plan_stack[0]->child_data;

  unsigned num_remaining_plans = 0;

  while (plan_stack[num_remaining_plans+1] != NULL) num_remaining_plans++;

  if (num_remaining_plans > 0) {

    struct interpolation_plan_element ** tmp_plan_stack =
      malloc((plan->num_plans + num_remaining_plans + 1) *
             sizeof(tmp_plan_stack));

    memcpy(tmp_plan_stack, plan->plans,
           plan->num_plans * sizeof(*tmp_plan_stack));
    memcpy(tmp_plan_stack + plan->num_plans, plan_stack + 1,
           num_remaining_plans * sizeof(*tmp_plan_stack));
    tmp_plan_stack[plan->num_plans + num_remaining_plans] = NULL;

    plan_stack = tmp_plan_stack;
  } else {
    plan_stack = plan->plans;
  }

  plan_stack[0]->do_interpolation(plan_stack, input_data, output_data);

  if (num_remaining_plans > 0) free(plan_stack);
}

static void free_interpolation_plan_element_collection(
  struct interpolation_plan_element * plan) {

  struct interpolation_plan_element_collection * plan_coll =
    plan->child_data;

  for (unsigned i = 0; i < plan_coll->num_plans; ++i)
    yac_free_interpolation_plan_element(plan_coll->plans[i]);

  free(plan_coll->plans[0]);
  free(plan_coll->plans);
  free(plan_coll);
}

static void
get_link_data_collection(struct interpolation_plan_element ** plan_stack,
                         int *** global_indices,
                         unsigned * num_targets_per_plan,
                         unsigned * num_links_per_plan,
                         unsigned * num_weights_per_link_per_plan,
                         unsigned ** num_links_per_tgt_per_plan,
                         int ** src_indices_per_plan,
                         unsigned ** pointset_indices_per_plan,
                         double ** weights_per_plan) {

  struct interpolation_plan_element_collection * plan =
    plan_stack[0]->child_data;

  if (plan->num_plans == 0) {
    *num_targets_per_plan = plan->num_plans;
    *num_links_per_plan = 0;
    *num_weights_per_link_per_plan = 0;
    *num_links_per_tgt_per_plan = NULL;
    *src_indices_per_plan = NULL;
    *pointset_indices_per_plan = NULL;
    *weights_per_plan = NULL;
  } else {

    unsigned * tmp_num_targets_per_plan =
      malloc(plan->num_plans * sizeof(*tmp_num_targets_per_plan));
    unsigned * tmp_num_links_per_plan =
      malloc(plan->num_plans * sizeof(*tmp_num_links_per_plan));
    unsigned * tmp_num_weights_per_link_per_plan =
      malloc(plan->num_plans * sizeof(*tmp_num_weights_per_link_per_plan));
    unsigned ** tmp_num_links_per_tgt_per_plan =
      malloc(plan->num_plans * sizeof(*tmp_num_links_per_tgt_per_plan));
    int ** tmp_src_indices_per_plan =
      malloc(plan->num_plans * sizeof(*tmp_src_indices_per_plan));
    unsigned ** tmp_pointset_indices_per_plan =
      malloc(plan->num_plans * sizeof(*tmp_pointset_indices_per_plan));
    double ** tmp_weights_per_plan =
      malloc(plan->num_plans * sizeof(*tmp_weights_per_plan));

    plan->plans[0]->get_link_data(plan->plans, global_indices,
                                  tmp_num_targets_per_plan,
                                  tmp_num_links_per_plan,
                                  tmp_num_weights_per_link_per_plan,
                                  tmp_num_links_per_tgt_per_plan,
                                  tmp_src_indices_per_plan,
                                  tmp_pointset_indices_per_plan,
                                  tmp_weights_per_plan);

    unsigned total_num_tgt = 0;
    unsigned total_num_links = 0;
    unsigned max_num_weights_per_link = 0;

    for (unsigned i = 0; i < plan->num_plans; ++i) {
      total_num_tgt += tmp_num_targets_per_plan[i];
      total_num_links += tmp_num_links_per_plan[i];
      max_num_weights_per_link = MAX(tmp_num_weights_per_link_per_plan[i],
                                     max_num_weights_per_link);
    }

    *num_targets_per_plan = total_num_tgt;
    *num_links_per_plan = total_num_links;
    *num_weights_per_link_per_plan = max_num_weights_per_link;
    *num_links_per_tgt_per_plan =
      malloc(total_num_tgt * sizeof(**num_links_per_tgt_per_plan));
    *src_indices_per_plan =
      malloc(total_num_links * sizeof(**src_indices_per_plan));
    *pointset_indices_per_plan =
      malloc(total_num_links * sizeof(**pointset_indices_per_plan));
    *weights_per_plan =
      malloc(total_num_links * max_num_weights_per_link *
             sizeof(**weights_per_plan));

    for (unsigned i = 0, tgt_offset = 0, link_offset = 0; i < plan->num_plans;
         ++i) {

      if (tmp_num_links_per_plan[i] > 0) {
        memcpy(*num_links_per_tgt_per_plan + tgt_offset,
               tmp_num_links_per_tgt_per_plan[i], tmp_num_targets_per_plan[i] *
               sizeof(**num_links_per_tgt_per_plan));
      } else {
        memset(*num_links_per_tgt_per_plan + tgt_offset, 0,
               tmp_num_targets_per_plan[i] *
               sizeof(**num_links_per_tgt_per_plan));
      }
      free(tmp_num_links_per_tgt_per_plan[i]);
      memcpy(*src_indices_per_plan + link_offset,
             tmp_src_indices_per_plan[i], tmp_num_links_per_plan[i] *
             sizeof(**src_indices_per_plan));
      free(tmp_src_indices_per_plan[i]);
      memcpy(*pointset_indices_per_plan + link_offset,
             tmp_pointset_indices_per_plan[i], tmp_num_links_per_plan[i] *
             sizeof(**pointset_indices_per_plan));
      free(tmp_pointset_indices_per_plan[i]);

      if (tmp_num_weights_per_link_per_plan[i] != max_num_weights_per_link) {
        for (unsigned j = 0, k = 0; j < tmp_num_links_per_plan[i]; ++j) {
          for (k = 0; k < tmp_num_weights_per_link_per_plan[i]; ++k)
            (*weights_per_plan)[(link_offset + j) * max_num_weights_per_link + k] =
              tmp_weights_per_plan[i][j * tmp_num_links_per_plan[i] + k];
          for (; k < max_num_weights_per_link; ++k)
            (*weights_per_plan)[(link_offset + j) * max_num_weights_per_link + k] =
              0.0;
        }
      } else {
        memcpy(*weights_per_plan + link_offset * max_num_weights_per_link,
               tmp_weights_per_plan[i], tmp_num_links_per_plan[i] *
               max_num_weights_per_link * sizeof(**weights_per_plan));
      }
      free(tmp_weights_per_plan[i]);      

      tgt_offset += tmp_num_targets_per_plan[i];
      link_offset += tmp_num_links_per_plan[i];
    }
    free(tmp_num_targets_per_plan);
    free(tmp_num_links_per_plan);
    free(tmp_num_weights_per_link_per_plan);
    free(tmp_num_links_per_tgt_per_plan);
    free(tmp_src_indices_per_plan);
    free(tmp_pointset_indices_per_plan);
    free(tmp_weights_per_plan);
  }

  if (plan_stack[1] != NULL)
    plan_stack[1]->get_link_data(plan_stack+1, global_indices,
                                 num_targets_per_plan+1,
                                 num_links_per_plan+1,
                                 num_weights_per_link_per_plan+1,
                                 num_links_per_tgt_per_plan+1,
                                 src_indices_per_plan+1,
                                 pointset_indices_per_plan+1,
                                 weights_per_plan+1);
}

static unsigned get_num_fixed_values_collection (
  struct interpolation_plan_element ** plan_stack) {

  struct interpolation_plan_element_collection * plan =
    plan_stack[0]->child_data;

  unsigned num_fixed_values =
    (plan->num_plans > 0)?
      plan->plans[0]->get_num_fixed_values(plan->plans):0;

  if (plan_stack[1] != NULL)
    return num_fixed_values + plan_stack[1]->get_num_fixed_values(plan_stack+1);
  else
    return num_fixed_values;
}

static void
get_fixed_data_collection(struct interpolation_plan_element ** plan_stack,
                          int const * global_ids,
                          int ** global_tgt_indices_per_plan,
                          unsigned * num_tgt_indices_per_plan,
                          double * fixed_value_per_plan) {

  struct interpolation_plan_element_collection * plan =
    plan_stack[0]->child_data;

  unsigned num_remaining_plans = 0;

  while (plan_stack[num_remaining_plans+1] != NULL) num_remaining_plans++;

  if (num_remaining_plans > 0) {

    struct interpolation_plan_element ** tmp_plan_stack =
      malloc((plan->num_plans + num_remaining_plans + 1) *
             sizeof(tmp_plan_stack));

    memcpy(tmp_plan_stack, plan->plans,
           plan->num_plans * sizeof(*tmp_plan_stack));
    memcpy(tmp_plan_stack + plan->num_plans, plan_stack + 1,
           num_remaining_plans * sizeof(*tmp_plan_stack));
    tmp_plan_stack[plan->num_plans + num_remaining_plans] = NULL;

    plan_stack = tmp_plan_stack;
  } else {
    plan_stack = plan->plans;
  }

  if (plan_stack[0] != NULL)
    plan_stack[0]->get_fixed_data(plan_stack, global_ids,
                                  global_tgt_indices_per_plan,
                                  num_tgt_indices_per_plan,
                                  fixed_value_per_plan);

  if (num_remaining_plans > 0) free(plan_stack);
}

void yac_init_interpolation_plan_element_collection(
  struct interpolation_plan_element * plan, unsigned num_plans,
  struct interpolation_plan_element ** plans) {

  struct interpolation_plan_element_collection * plan_coll =
    malloc(1 * sizeof(*plan_coll));

  plan_coll->num_plans = num_plans;
  plan_coll->plans = malloc((num_plans + 1) * sizeof(*(plan_coll->plans)));
  plan_coll->plans[num_plans] = NULL;
  if (num_plans > 0) {
    plan_coll->plans[0] = malloc(num_plans * sizeof(**(plan_coll->plans)));
    for (unsigned i = 1; i < num_plans; ++i)
      plan_coll->plans[i] = plan_coll->plans[0] + i;
    for (unsigned i = 0; i < num_plans; ++i)
      *(plan_coll->plans[i]) = *(plans[i]);
  }

  plan->child_data = plan_coll;
  plan->do_interpolation = do_interpolation_collection;
  plan->get_link_data = get_link_data_collection;
  plan->get_num_fixed_values = get_num_fixed_values_collection;
  plan->get_fixed_data = get_fixed_data_collection;
  plan->free_interpolation_plan_element =
    free_interpolation_plan_element_collection;
}

void yac_get_link_data(struct interpolation_plan_element ** plan_stack,
                       int *** global_indices,
                       unsigned num_pointsets,
                       unsigned * num_targets,
                       unsigned ** num_links_per_point_set,
                       unsigned * num_weights_per_link,
                       unsigned ** num_links_per_tgt, int ** src_indices,
                       unsigned ** pointset_indices, double ** weights) {

  *num_links_per_point_set =
    calloc(num_pointsets, sizeof(**num_links_per_point_set));

  unsigned num_plans = 0;

  for (; plan_stack[num_plans] != NULL; ++num_plans);
  
  if (num_plans == 0) {

    *num_targets = 0;
    *num_weights_per_link = 0;
    *num_links_per_tgt = NULL;
    *src_indices = NULL;
    *pointset_indices = NULL;
    *weights = NULL;
    return;
  }

  //----------------------------
  // get results from plan stack
  //----------------------------

  unsigned * num_targets_per_plan = malloc(num_plans * sizeof(*num_targets_per_plan));
  unsigned * num_links_per_plan = malloc(num_plans * sizeof(*num_links_per_plan));
  unsigned * num_weights_per_link_per_plan =
    malloc(num_plans * sizeof(*num_weights_per_link_per_plan));
  unsigned ** num_links_per_tgt_per_plan =
    calloc(num_plans, sizeof(*num_links_per_tgt_per_plan));
  int ** src_indices_per_plan =
    calloc(num_plans, sizeof(*src_indices_per_plan));
  unsigned ** pointset_indices_per_plan =
    calloc(num_plans, sizeof(*pointset_indices_per_plan));
  double ** weights_per_plan = calloc(num_plans, sizeof(*weights_per_plan));

  plan_stack[0]->get_link_data(plan_stack, global_indices, num_targets_per_plan,
                               num_links_per_plan,
                               num_weights_per_link_per_plan,
                               num_links_per_tgt_per_plan,
                               src_indices_per_plan, pointset_indices_per_plan,
                               weights_per_plan);

  unsigned total_num_links = 0;
  unsigned max_num_weights_per_link = 0;
  unsigned total_num_tgt = 0;

  //------------------------------
  // make results plan independant
  //------------------------------

  for (unsigned i = 0; i < num_plans; ++i) {

    total_num_tgt += num_targets_per_plan[i];
    total_num_links += num_links_per_plan[i];

    if (num_weights_per_link_per_plan[i] > max_num_weights_per_link)
      max_num_weights_per_link = num_weights_per_link_per_plan[i];

    for (unsigned j = 0; j < num_links_per_plan[i]; ++j)
      (*num_links_per_point_set)[pointset_indices_per_plan[i][j]]++;
  }
  *num_targets = total_num_tgt;
  *num_weights_per_link = max_num_weights_per_link;

  *num_links_per_tgt = malloc(total_num_tgt * sizeof(**num_links_per_tgt));
  *src_indices = malloc(total_num_links * sizeof(**src_indices));
  *pointset_indices = malloc(total_num_links * sizeof(**pointset_indices));
  *weights = malloc(max_num_weights_per_link * total_num_links *
                    sizeof(**weights));

  for (unsigned i = 0, k = 0, l = 0; i < num_plans; ++i) {

    unsigned curr_num_weights_per_link = num_weights_per_link_per_plan[i];

    if (num_links_per_plan[i] > 0) {
      for (unsigned j = 0; j < num_targets_per_plan[i]; ++j, ++k)
        (*num_links_per_tgt)[k] = num_links_per_tgt_per_plan[i][j];
    } else {
      for (unsigned j = 0; j < num_targets_per_plan[i]; ++j, ++k)
        (*num_links_per_tgt)[k] = 0;
    }

    for (unsigned j = 0; j < num_links_per_plan[i]; ++j, ++l) {

      (*src_indices)[l] = src_indices_per_plan[i][j];
      (*pointset_indices)[l] = pointset_indices_per_plan[i][j];
      for (unsigned m = 0; m < curr_num_weights_per_link; ++m)
        (*weights)[l*max_num_weights_per_link + m] =
          weights_per_plan[i][j*curr_num_weights_per_link + m];
      for (unsigned m  = curr_num_weights_per_link;
           m < max_num_weights_per_link; ++m)
        (*weights)[l*max_num_weights_per_link + m] = 0;
    }
  }

  //---------
  // clean up
  //---------

  for (unsigned i = 0; i < num_plans; ++i) {
    free(num_links_per_tgt_per_plan[i]);
    free(src_indices_per_plan[i]);
    free(pointset_indices_per_plan[i]);
    free(weights_per_plan[i]);
  }

  free(num_targets_per_plan);
  free(num_links_per_plan);
  free(num_weights_per_link_per_plan);
  free(num_links_per_tgt_per_plan);
  free(src_indices_per_plan);
  free(pointset_indices_per_plan);
  free(weights_per_plan);
}

static int compare_fixed_value_sort_data(const void * a, const void * b) {

  return (((struct fixed_value_sort_data*)a)->value >
          ((struct fixed_value_sort_data*)b)->value) -
         (((struct fixed_value_sort_data*)a)->value < 
          ((struct fixed_value_sort_data*)b)->value);
}

static int compare_int(const void * a, const void * b) {

  return ((*(int*)a) > (*(int*)b)) - (*((int*)a) < (*(int*)b));
}

void yac_get_fixed_data(struct interpolation_plan_element ** plan_stack,
                        int const * global_ids, int ** global_tgt_indices,
                        unsigned * num_tgt_indices,
                        unsigned ** num_tgt_per_fixed_value,
                        double ** fixed_values,
                        unsigned * num_fixed_values) {

  unsigned num_plans = 0;

  for (; plan_stack[num_plans] != NULL; ++num_plans);
  
  if (num_plans == 0) {
    *global_tgt_indices = NULL;
    *num_tgt_indices = 0;
    *num_tgt_per_fixed_value = NULL;
    *fixed_values = NULL;
    *num_fixed_values = 0;
    return;
  }

  //----------------------------
  // get results from plan stack
  //----------------------------

  unsigned tmp_num_fixed_values =
    plan_stack[0]->get_num_fixed_values(plan_stack);

  int ** global_tgt_indices_per_plan =
    malloc(tmp_num_fixed_values * sizeof(*global_tgt_indices_per_plan));
  unsigned * num_tgt_indices_per_plan =
    malloc(tmp_num_fixed_values * sizeof(*num_tgt_indices_per_plan));
  double * fixed_value_per_plan =
    malloc(tmp_num_fixed_values * sizeof(*fixed_value_per_plan));

  plan_stack[0]->get_fixed_data(plan_stack, global_ids,
                                global_tgt_indices_per_plan,
                                num_tgt_indices_per_plan,
                                fixed_value_per_plan);

  //------------------------------
  // make results plan independant
  //------------------------------

  unsigned total_num_tgt_indices = 0;
  unsigned total_num_fixed_values = 0;

  for (unsigned i = 0; i < tmp_num_fixed_values; ++i) {
    total_num_tgt_indices += num_tgt_indices_per_plan[i];
    total_num_fixed_values += num_tgt_indices_per_plan[i] > 0;
  }

  // if there are target points that get a fixed value
  if (total_num_tgt_indices == 0) {

    *global_tgt_indices = NULL;
    *num_tgt_indices = 0;
    *num_tgt_per_fixed_value = NULL;
    *fixed_values = NULL;
    *num_fixed_values = 0;

    for (unsigned i = 0; i < tmp_num_fixed_values; ++i)
      free(global_tgt_indices_per_plan[i]);

    free(global_tgt_indices_per_plan);
    free(num_tgt_indices_per_plan);
    free(fixed_value_per_plan);
    return;
  }

  // check for duplicated fixed values
  struct fixed_value_sort_data * tmp_fixed_values =
    malloc(total_num_fixed_values * sizeof(*tmp_fixed_values));

  for (unsigned i = 0, k = 0; i < tmp_num_fixed_values; ++i) {
    if (num_tgt_indices_per_plan[i] == 0) continue;
    tmp_fixed_values[k].value = fixed_value_per_plan[i];
    tmp_fixed_values[k].plan_idx = i;
    ++k;
  }
  free(fixed_value_per_plan);

  // determine number of unique fixed values
  qsort(tmp_fixed_values, total_num_fixed_values, sizeof(*tmp_fixed_values),
        compare_fixed_value_sort_data);

  unsigned num_unique_fixed_values = 1;
  double prev_fixed_value = tmp_fixed_values[0].value;
  for (unsigned i = 1; i < total_num_fixed_values; ++i) {
    if (prev_fixed_value != tmp_fixed_values[i].value) {
      prev_fixed_value = tmp_fixed_values[i].value;
      ++num_unique_fixed_values;
    }
  }

  // copy global target points of all plans into single array
  unsigned * tmp_num_tgt_per_fixed_value =
    calloc(num_unique_fixed_values, sizeof(*tmp_num_tgt_per_fixed_value));
  int * tmp_global_tgt_indices =
    malloc(total_num_tgt_indices * sizeof(*tmp_global_tgt_indices));
  *fixed_values = malloc(num_unique_fixed_values * sizeof(**fixed_values));

  prev_fixed_value = (tmp_fixed_values[0].value == 1.0)?(-1.0):(1.0);

  /* using size_t here because PGI 15.4 miscompiles with unsigned */
  for (size_t i = 0, k = -1, offset = 0; i < total_num_fixed_values; ++i) {

    unsigned curr_plan_idx = tmp_fixed_values[i].plan_idx;
    unsigned curr_num_tgt_indices = num_tgt_indices_per_plan[curr_plan_idx];

    memcpy(tmp_global_tgt_indices + offset,
           global_tgt_indices_per_plan[curr_plan_idx],
           curr_num_tgt_indices * sizeof(*tmp_global_tgt_indices));

    if (prev_fixed_value != tmp_fixed_values[i].value) {
      prev_fixed_value = tmp_fixed_values[i].value;
      ++k;
      (*fixed_values)[k] = tmp_fixed_values[i].value;
    }

    tmp_num_tgt_per_fixed_value[k] += curr_num_tgt_indices;

    offset += curr_num_tgt_indices;

    free(global_tgt_indices_per_plan[curr_plan_idx]);
  }

  free(num_tgt_indices_per_plan);
  free(tmp_fixed_values);
  free(global_tgt_indices_per_plan);

  // sort global target indices that have the same fixed value
  for (unsigned i = 0, offset = 0; i < num_unique_fixed_values; ++i) {
    qsort(tmp_global_tgt_indices + offset, tmp_num_tgt_per_fixed_value[i],
          sizeof(*tmp_global_tgt_indices), compare_int);
    offset += tmp_num_tgt_per_fixed_value[i];
  }

  *global_tgt_indices = tmp_global_tgt_indices;
  *num_tgt_indices = total_num_tgt_indices;
  *num_tgt_per_fixed_value = tmp_num_tgt_per_fixed_value;
  *num_fixed_values = num_unique_fixed_values;
}
