!>
!! @file mo_yac_finterface.f90
!! @brief Implementation of the YAC Fortran user interface description
!!
!! @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
!!                                 Rene Redler <rene.redler@mpimet.mpg.de>
!!                                 Thomas Jahns <jahns@dkrz.de>
!! Fortran user interface for YAC.
!!
!!  Interface descriptions for
!!
!!  - yac_finit
!!  - yac_ffinalize
!!  - yac_fsearch
!!
!!  - yac_fdef_comp
!!  - yac_fdef_comps
!!  - yac_fdef_datetime
!!  - yac_fdef_subdomain
!!  - yac_fconnect_subdomains
!!  - yac_fdef_elements
!!  - yac_fdef_points
!!  - yac_fdef_index_location
!!  - yac_fdef_mask
!!
!!  - yac_fput
!!  - yac_fget
!!
!!  - yac_fget_local_comm
!!  - yac_fget_pair_rootcomm
!!  - yac_fget_nbr_fields
!!  - yac_fget_field_ids
!!  - yac_fget_model_timestep
!!  - yac_fget_coupling_timestep
!!
!!  - yac_abort_message
!!  - yac_redirstdout
!!
!! @version 1.0
!! @author Moritz Hanke <hanke@dkrz.de>
!!         Rene Redler <rene.redler@mpimet.mpg.de>
!!         Thomas Jahns <jahns@dkrz.de>
!!

!
! Keywords:
! Maintainer: Moritz Hanke <hanke@dkrz.de>
!             Rene Redler <rene.redler@mpimet.mpg.de>
!             Thomas Jahns <jahns@dkrz.de>
! URL: https://doc.redmine.dkrz.de/YAC/html/index.html
!
! This file is part of YAC.
!
! YAC is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! YAC is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
!

!> \example test_dummy_coupling.F90
!! This example simulates a whole model setup with three components (ocean,
!! atmosphere, io). It uses one process for each component.

module mo_yac_finterface

  !----------------------------------------------------------------------
  !>
  !!   Constants
  !!
  !----------------------------------------------------------------------

  public

  integer, parameter :: YAC_MAX_CHARLEN = 132

  !----------------------------------------------------------------------
  !>
  !!   Point types (see also points.h)
  !!
  !----------------------------------------------------------------------

  enum, bind(c)
   enumerator :: CELL   = 0
   enumerator :: CORNER = 1
   enumerator :: EDGE   = 2
  end enum

  !----------------------------------------------------------------------
  !>
  !!   Action types (see also event.h)
  !!
  !----------------------------------------------------------------------

  enum, bind(c)
   enumerator :: NONE               = 0
   enumerator :: COUPLING           = 1
   enumerator :: RESTART            = 2
   enumerator :: GET_FOR_RESTART    = 3
   enumerator :: PUT_FOR_RESTART    = 4
   enumerator :: GET_FOR_CHECKPOINT = 5
   enumerator :: PUT_FOR_CHECKPOINT = 6
   enumerator :: OUT_OF_BOUND       = 7
  end enum

  !----------------------------------------------------------------------
  !>
  !!   Reduction types (see also event.h)
  !!
  !----------------------------------------------------------------------

  enum, bind(c)
   enumerator :: INSTANT      = 0
   enumerator :: ACCUMULATE   = 1
   enumerator :: TIME_AVERAGE = 2
  end enum

  !----------------------------------------------------------------------
  !>
  !!   pointer data types
  !!
  !----------------------------------------------------------------------

  type yac_r4_ptr
    real, pointer :: p(:)
  end type yac_r4_ptr

  type yac_r8_ptr
    double precision, pointer :: p(:)
  end type yac_r8_ptr

  !----------------------------------------------------------------------
  !>
  !!   Fortran2C interface for yac collective routines
  !!
  !----------------------------------------------------------------------

  !> \example test_init_final.F90
  !! This contains an example on how to use yac_init and yac_ffinalize.

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for the coupler initialisation
  !!
  !----------------------------------------------------------------------

  interface

     subroutine yac_finit ( xml_filename, schema_filename )

       character(len=*), intent(in) :: xml_filename    !< [IN] xml description
       character(len=*), intent(in) :: schema_filename !< [IN] xml schema

     end subroutine yac_finit

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface
     subroutine yac_ffinalize ()
     end subroutine yac_ffinalize
  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran2C interface for the coupler initialisation
  !!
  !----------------------------------------------------------------------

  interface
     subroutine yac_cinit ( xml_filename, schema_filename ) &
    &      bind ( c, name='yac_cinit' )

       use, intrinsic :: iso_c_binding, only : c_char

       character ( kind=c_char), dimension(*) :: xml_filename
       character ( kind=c_char), dimension(*) :: schema_filename

     end subroutine yac_cinit

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface
     subroutine yac_cfinalize () bind ( c, name='yac_cfinalize' )
     end subroutine yac_cfinalize
  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for the component definition
  !!
  !----------------------------------------------------------------------

  interface

     subroutine yac_fdef_comp ( comp_name, comp_id )

       character(len=*), intent(in) :: comp_name !< [IN]   comp name
       integer, intent(out)         :: comp_id   !< [OUT]  component identifier

     end subroutine yac_fdef_comp

  end interface

  interface

     subroutine yac_fdef_comps ( comp_names, num_comps, comp_ids )

        use, intrinsic :: iso_c_binding, only : c_char
        import :: YAC_MAX_CHARLEN

        integer, intent(in)   :: num_comps
        character(kind=c_char, len=YAC_MAX_CHARLEN), intent(in) :: &
                                 comp_names(num_comps)
        integer, intent(out)  :: comp_ids(num_comps)

     end subroutine yac_fdef_comps

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cdef_comp ( comp_name, comp_id ) &
    &      bind ( c, name='yac_cdef_comp' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       character ( kind=c_char), dimension(*) :: comp_name
       integer   ( kind=c_int )               :: comp_id

     end subroutine yac_cdef_comp

  end interface

  interface

     subroutine yac_cdef_comps ( comp_names, num_comps, comp_ids ) &
    &      bind ( c, name='yac_cdef_comps' )

       use, intrinsic :: iso_c_binding, only : c_int, c_ptr

       type ( c_ptr )                :: comp_names(*)
       integer ( kind=c_int ), value :: num_comps
       integer ( kind=c_int )        :: comp_ids(*)

     end subroutine yac_cdef_comps

  end interface

  !> \example test_def_datetime.F90
  !! This shows how to use yac_fdef_datetime.

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for the definition of start- and end datetime
  !!
  !----------------------------------------------------------------------

  interface

     subroutine yac_fdef_datetime ( start_datetime, end_datetime )

       character(len=*), intent(in), optional :: start_datetime !< [IN] start datetime of job
       character(len=*), intent(in), optional :: end_datetime   !< [IN] end datetime of job

     end subroutine yac_fdef_datetime

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cdef_datetime ( start_datetime, end_datetime ) &
    &      bind ( c, name='yac_cdef_datetime' )

       use, intrinsic :: iso_c_binding, only : c_char

       character ( kind=c_char), dimension(*) :: start_datetime
       character ( kind=c_char), dimension(*) :: end_datetime

     end subroutine yac_cdef_datetime

  end interface

  !> \example test_def_grid.F90
  !! This shows how to use yac_fdef_subdomain.

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for the subdomain definition
  !!
  !----------------------------------------------------------------------

  interface

     subroutine yac_fdef_subdomain ( comp_id,    &
       &                             grid_name,  &
       &                             subdomain_id )

       integer, intent(in)                    :: comp_id      !< [IN]  component identifier
       character(len=*), intent(in)           :: grid_name    !< [IN]  grid name
       integer, intent(out)                   :: subdomain_id !< [OUT] subdomain ID

     end subroutine yac_fdef_subdomain

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cdef_subdomain ( comp_id, grid_name, subdomain_id ) &
    &      bind ( c, name='yac_cdef_subdomain' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       integer   ( kind=c_int ), value         :: comp_id
       character ( kind=c_char ), dimension(*) :: grid_name
       integer   ( kind=c_int )                :: subdomain_id

     end subroutine yac_cdef_subdomain

  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for the connection of subdomains
  !!
  !----------------------------------------------------------------------

  interface

     subroutine yac_fconnect_subdomains ( comp_id,           &
                                          nbr_subdomain_ids, &
                                          subdomain_ids,     &
                                          domain_id )
       integer, intent(in)  :: comp_id                          !< [IN] component ID
       integer, intent(in)  :: nbr_subdomain_ids                !< [IN] number of subdomains
       integer, intent(in)  :: subdomain_ids(nbr_subdomain_ids) !< [IN] list of subdomain IDs
       integer, intent(out) :: domain_id                        !< [OUT] domain ID

     end subroutine yac_fconnect_subdomains

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cconnect_subdomains ( comp_id,           &
                                          nbr_subdomain_ids, &
                                          subdomain_ids,     &
                                          domain_id ) &
    &      bind ( c, name='yac_cconnect_subdomains' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int ), value        :: comp_id
       integer ( kind=c_int ), value        :: nbr_subdomain_ids
       integer ( kind=c_int ), dimension(*) :: subdomain_ids
       integer ( kind=c_int )               :: domain_id


     end subroutine yac_cconnect_subdomains

  end interface

  !> \example test_def_elements.F90
  !! This shows how to use yac_fdef_elements.

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for the definition of subdomain elements
  !!
  !----------------------------------------------------------------------

  interface yac_fdef_elements

     subroutine yac_fdef_elements_nonuniform_r4 ( subdomain_id,          &
                                                  nbr_vertices,          &
                                                  nbr_cells,             &
                                                  nbr_connections,       &
                                                  nbr_vertices_per_cell, &
                                                  x_vertices,            &
                                                  y_vertices,            &
                                                  cell_to_vertex )

       integer, intent(in)  :: subdomain_id                     !< [IN]  subdomain identifier
       integer, intent(in)  :: nbr_vertices                     !< [IN]  number of vertices
       integer, intent(in)  :: nbr_cells                        !< [IN]  number of cells
       integer, intent(in)  :: nbr_connections                  !< [IN]  total size of cell_to_vertex
       integer, intent(in)  :: nbr_vertices_per_cell(nbr_cells) !< [IN]  number of vertices for each cell

       real, intent(in)     :: x_vertices(nbr_vertices)         !< [IN]  longitudes of vertices
       real, intent(in)     :: y_vertices(nbr_vertices)         !< [IN]  latitudes of vertices

       integer, intent(in)  :: cell_to_vertex(nbr_connections)  !< [IN] connectivity between vertices and cells\n
                                                                !!(the vertex indices per cell have to be in clockwise or counterclockwise ordering)

     end subroutine yac_fdef_elements_nonuniform_r4

     subroutine yac_fdef_elements_nonuniform_r8 ( subdomain_id,          &
                                                  nbr_vertices,          &
                                                  nbr_cells,             &
                                                  nbr_connections,       &
                                                  nbr_vertices_per_cell, &
                                                  x_vertices,            &
                                                  y_vertices,            &
                                                  cell_to_vertex )

       integer, intent(in)          :: subdomain_id                     !< [IN]  subdomain identifier
       integer, intent(in)          :: nbr_vertices                     !< [IN]  number of vertices
       integer, intent(in)          :: nbr_cells                        !< [IN]  number of cells
       integer, intent(in)          :: nbr_connections                  !< [IN]  total size of cell_to_vertex
       integer, intent(in)          :: nbr_vertices_per_cell(nbr_cells) !< [IN]  number of vertices for each cell

       double precision, intent(in) :: x_vertices(nbr_vertices)         !< [IN]  longitudes of vertices
       double precision, intent(in) :: y_vertices(nbr_vertices)         !< [IN]  latitudes of vertices

       integer, intent(in)          :: cell_to_vertex(nbr_connections)  !< [IN] connectivity between vertices and cells\n
                                                                        !!(the vertex indices per cell have to be in clockwise or counterclockwise ordering)

     end subroutine yac_fdef_elements_nonuniform_r8

     subroutine yac_fdef_elements_unstruct_r4 ( subdomain_id,          &
                                                nbr_vertices,          &
                                                nbr_cells,             &
                                                nbr_vertices_per_cell, &
                                                x_vertices,            &
                                                y_vertices,            &
                                                cell_to_vertex )

       integer, intent(in)  :: subdomain_id                      !< [IN]  subdomain identifier
       integer, intent(in)  :: nbr_vertices                      !< [IN]  number of vertices
       integer, intent(in)  :: nbr_cells                         !< [IN]  number of cells
       integer, intent(in)  :: nbr_vertices_per_cell             !< [IN]  number of vertices for each cell

       real, intent(in)     :: x_vertices(nbr_vertices)          !< [IN]  longitudes of vertices
       real, intent(in)     :: y_vertices(nbr_vertices)          !< [IN]  latitudes of vertices

       integer, intent(in)  :: cell_to_vertex(nbr_vertices_per_cell,nbr_cells) !< [IN] connectivity between vertices and cells\n
                                                                               !!(the vertex indices per cell have to be in clockwise or counterclockwise ordering)

     end subroutine yac_fdef_elements_unstruct_r4

     subroutine yac_fdef_elements_unstruct_r8 ( subdomain_id,          &
                                                nbr_vertices,          &
                                                nbr_cells,             &
                                                nbr_vertices_per_cell, &
                                                x_vertices,            &
                                                y_vertices,            &
                                                cell_to_vertex )

       integer, intent(in)          :: subdomain_id                     !< [IN]  subdomain identifier
       integer, intent(in)          :: nbr_vertices                     !< [IN]  number of vertices
       integer, intent(in)          :: nbr_cells                        !< [IN]  number of cells
       integer, intent(in)          :: nbr_vertices_per_cell            !< [IN]  number of vertices for each cell

       double precision, intent(in) :: x_vertices(nbr_vertices)         !< [IN]  longitudes of vertices
       double precision, intent(in) :: y_vertices(nbr_vertices)         !< [IN]  latitudes of vertices

       integer, intent(in)          :: cell_to_vertex(nbr_vertices_per_cell, nbr_cells)
                                                                        !< [IN] connectivity between vertices and cells\n
                                                                        !!(the vertex indices per cell have to be in
                                                                        !! clockwise or counterclockwise ordering)

     end subroutine yac_fdef_elements_unstruct_r8

     subroutine yac_fdef_elements_reg2d_r4 ( subdomain_id,          &
                                             nbr_vertices,          &
                                             cyclic,                &
                                             x_vertices,            &
                                             y_vertices )

       integer, intent(in)  :: subdomain_id                            !< [IN]  subdomain identifier
       integer, intent(in)  :: nbr_vertices(2)                         !< [IN]  number of cells in each dimension
       integer, intent(in)  :: cyclic(2)                               !< [IN]  cyclic behavior of cells in each dimension

       real, intent(in)     :: x_vertices(nbr_vertices(1))             !< [IN]  longitudes of vertices
       real, intent(in)     :: y_vertices(nbr_vertices(2))             !< [IN]  latitudes of vertices

     end subroutine yac_fdef_elements_reg2d_r4

     subroutine yac_fdef_elements_reg2d_r8 ( subdomain_id,          &
                                             nbr_vertices,          &
                                             cyclic,                &
                                             x_vertices,            &
                                             y_vertices )

       integer, intent(in)          :: subdomain_id                    !< [IN]  subdomain identifier
       integer, intent(in)          :: nbr_vertices(2)                 !< [IN]  number of cells in each dimension
       integer, intent(in)          :: cyclic(2)                       !< [IN]  cyclic behavior of cells in each dimension

       double precision, intent(in) :: x_vertices(nbr_vertices(1))     !< [IN]  longitudes of vertices
       double precision, intent(in) :: y_vertices(nbr_vertices(2))     !< [IN]  latitudes of vertices

     end subroutine yac_fdef_elements_reg2d_r8

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cdef_elements_reg2d ( subdomain_id, &
                                          nbr_vertices, &
                                          cyclic,       &
                                          x_vertices,   &
                                          y_vertices )  &
    &      bind ( c, name='yac_cdef_elements_reg2d' )

       use, intrinsic :: iso_c_binding, only : c_int, c_double

       integer ( kind=c_int    ), value :: subdomain_id
       integer ( kind=c_int    )        :: nbr_vertices(2)
       integer ( kind=c_int    )        :: cyclic(2)
       real    ( kind=c_double )        :: x_vertices(nbr_vertices(1))
       real    ( kind=c_double )        :: y_vertices(nbr_vertices(2))

     end subroutine yac_cdef_elements_reg2d

  end interface

  interface

     subroutine yac_cdef_elements_unstruct ( subdomain_id,          &
                                             nbr_vertices,          &
                                             nbr_cells,             &
                                             nbr_vertices_per_cell, &
                                             x_vertices,            &
                                             y_vertices,            &
                                             cell_to_vertex )       &
     &     bind ( c, name='yac_cdef_elements_unstruct' )

       use, intrinsic :: iso_c_binding, only : c_int, c_double

       integer ( kind=c_int    ), value :: subdomain_id
       integer ( kind=c_int    ), value :: nbr_vertices
       integer ( kind=c_int    ), value :: nbr_cells
       integer ( kind=c_int    )        :: nbr_vertices_per_cell(nbr_cells)
       real    ( kind=c_double )        :: x_vertices(nbr_vertices)
       real    ( kind=c_double )        :: y_vertices(nbr_vertices)
       integer ( kind=c_int    )        :: cell_to_vertex(nbr_cells,nbr_vertices)

     end subroutine yac_cdef_elements_unstruct

  end interface

  !> \example test_def_points.F90
  !! This shows how to use yac_fdef_points.

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for the definition of points
  !!
  !----------------------------------------------------------------------

  interface yac_fdef_points

     subroutine yac_fdef_points_reg2d_r4 ( subdomain_id,  &
                                           nbr_points,    &
                                           location,      &
                                           x_points_real, &
                                           y_points_real, &
                                           point_id )

       integer, intent(in)  :: subdomain_id                 !< [IN]  subdomain identifier
       integer, intent(in)  :: nbr_points(2)                !< [IN]  number of points
       integer, intent(in)  :: location                     !< [IN]  location, one of center/edge/vertex

       real, intent(in)     :: x_points_real(nbr_points(1)) !< [IN]  longitudes of points
       real, intent(in)     :: y_points_real(nbr_points(2)) !< [IN]  latitudes of points

       integer, intent(out) :: point_id                     !< [OUT] point identifier

     end subroutine yac_fdef_points_reg2d_r4

     subroutine yac_fdef_points_reg2d_r8 ( subdomain_id,  &
                                           nbr_points,    &
                                           location,      &
                                           x_points,      &
                                           y_points,      &
                                           point_id )

       integer, intent(in)  :: subdomain_id                    !< [IN]  subdomain identifier
       integer, intent(in)  :: nbr_points(2)                   !< [IN]  number of points
       integer, intent(in)  :: location                        !< [IN]  location, one of center/edge/vertex

       double precision, intent(in) :: x_points(nbr_points(1)) !< [IN]  longitudes of points
       double precision, intent(in) :: y_points(nbr_points(2)) !< [IN]  latitudes of points

       integer, intent(out) :: point_id                        !< [OUT] point identifier

     end subroutine yac_fdef_points_reg2d_r8

     subroutine yac_fdef_points_unstruct_r4 ( subdomain_id,  &
                                              nbr_points,    &
                                              location,      &
                                              x_points_real, &
                                              y_points_real, &
                                              point_id )

       integer, intent(in)  :: subdomain_id              !< [IN]  subdomain identifier
       integer, intent(in)  :: nbr_points                !< [IN]  number of points
       integer, intent(in)  :: location                  !< [IN]  location, one of center/edge/vertex

       real, intent(in)     :: x_points_real(nbr_points) !< [IN]  longitudes of points
       real, intent(in)     :: y_points_real(nbr_points) !< [IN]  latitudes of points

       integer, intent(out) :: point_id                  !< [OUT] point identifier

     end subroutine yac_fdef_points_unstruct_r4

     subroutine yac_fdef_points_unstruct_r8 ( subdomain_id,  &
                                              nbr_points,    &
                                              location,      &
                                              x_points,      &
                                              y_points,      &
                                              point_id )

       integer, intent(in)  :: subdomain_id                 !< [IN]  subdomain identifier
       integer, intent(in)  :: nbr_points                   !< [IN]  number of points
       integer, intent(in)  :: location                     !< [IN]  location, one of center/edge/vertex

       double precision, intent(in) :: x_points(nbr_points) !< [IN]  longitudes of points
       double precision, intent(in) :: y_points(nbr_points) !< [IN]  latitudes of points

       integer, intent(out) :: point_id                     !< [OUT] point identifier

     end subroutine yac_fdef_points_unstruct_r8

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cdef_points_reg2d ( subdomain_id, &
                                        nbr_points,   &
                                        location,     &
                                        x_points,     &
                                        y_points,     &
                                        point_id )    &
    &      bind ( c, name='yac_cdef_points_reg2d' )

       use, intrinsic :: iso_c_binding, only : c_int, c_double

       integer ( kind=c_int ), value :: subdomain_id
       integer ( kind=c_int )        :: nbr_points(2)
       integer ( kind=c_int ), value :: location

       real    ( kind=c_double )     :: x_points(nbr_points(1))
       real    ( kind=c_double )     :: y_points(nbr_points(2))

       integer ( kind=c_int )        :: point_id

     end subroutine yac_cdef_points_reg2d

  end interface

  interface

     subroutine yac_cdef_points_unstruct ( subdomain_id, &
                                           nbr_points,   &
                                           location,     &
                                           x_points,     &
                                           y_points,     &
                                           point_id )    &
    &      bind ( c, name='yac_cdef_points_unstruct' )

       use, intrinsic :: iso_c_binding, only : c_int, c_double

       integer (kind=c_int), value :: subdomain_id
       integer (kind=c_int), value :: nbr_points
       integer (kind=c_int), value :: location

       real    (kind=c_double)     :: x_points(nbr_points)
       real    (kind=c_double)     :: y_points(nbr_points)

       integer (kind=c_int)        :: point_id

     end subroutine yac_cdef_points_unstruct

  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for the definition of masks
  !!
  !----------------------------------------------------------------------

  interface yac_fdef_mask

     subroutine yac_fdef_imask ( mask_size,  &
                                 imask,      &
                                 points_id,  &
                                 mask_id )

       integer, intent(in)  :: mask_size
       integer, intent(in)  :: imask(mask_size)
       integer, intent(in)  :: points_id
       integer, intent(out) :: mask_id

     end subroutine yac_fdef_imask

     subroutine yac_fdef_lmask ( mask_size,  &
                                 lmask,      &
                                 points_id,  &
                                 mask_id )

       integer, intent(in)  :: mask_size        !< [IN]  size of the mask array
       logical, intent(in)  :: lmask(mask_size) !< [IN]  mask array
       integer, intent(in)  :: points_id        !< [IN]  associated points id
       integer, intent(out) :: mask_id          !< [OUT] mask identifier

     end subroutine yac_fdef_lmask

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cdef_mask ( mask,         &
                                points_id,    &
                                mask_id )     &
    &      bind ( c, name='yac_cdef_mask' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int )        :: mask(*)
       integer ( kind=c_int ), value :: points_id
       integer ( kind=c_int )        :: mask_id

     end subroutine yac_cdef_mask

  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface to specify the process location of indices.
  !!
  !----------------------------------------------------------------------

  interface

     subroutine yac_fdef_index_location ( subdomain_id,        &
                                          nbr_entries,         &
                                          location,            &
                                          global_index,        &
                                          global_index_rank )

       integer, intent (in) :: subdomain_id                   !< [IN] subdomain identifier
       integer, intent (in) :: nbr_entries                    !< [IN] number of indices
       integer, intent (in) :: location                       !< [IN] one of CELL/EDGE/POINT
       integer, intent (in) :: global_index(nbr_entries)      !< [IN] global unique number of data point
       integer, intent (in) :: global_index_rank(nbr_entries) !< [IN] rank of data point

     end subroutine yac_fdef_index_location

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cdef_index_location ( subdomain_id,        &
                                          nbr_entries,         &
                                          location,            &
                                          global_index,        &
                                          global_index_rank )  &
    &      bind ( c, name='yac_cdef_index_location' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int ), value :: subdomain_id
       integer ( kind=c_int ), value :: nbr_entries
       integer ( kind=c_int ), value :: location
       integer ( kind=c_int )        :: global_index(nbr_entries)
       integer ( kind=c_int )        :: global_index_rank(nbr_entries)

     end subroutine yac_cdef_index_location

  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for the definition of coupling fields
  !!
  !----------------------------------------------------------------------

  interface

     subroutine yac_fdef_field ( field_name,               &
                                 component_id,             &
                                 domain_id,                &
                                 point_ids,                &
                                 mask_ids,                 &
                                 num_points_per_subdomain, &
                                 field_id )

       character(len=*), intent (in) :: field_name               !< [IN] short name of the field
       integer, intent (in)          :: component_id             !< [IN] component identifier
       integer, intent (in)          :: domain_id                !< [IN] domain identifier
       integer, intent (in)          :: point_ids(*)             !< [IN] point identifier
       integer, intent (in)          :: mask_ids(*)              !< [IN] mask identifier
       integer, intent (in)          :: num_points_per_subdomain !< [IN] number of pointsets per subdomain
       integer, intent (out)         :: field_id                 !< [OUT] returned field handle

     end subroutine yac_fdef_field

     subroutine yac_fdef_field_nomask ( field_name,               &
                                        component_id,             &
                                        domain_id,                &
                                        point_ids,                &
                                        num_points_per_subdomain, &
                                        field_id )

       character(len=*), intent (in) :: field_name               !< [IN] short name of the field
       integer, intent (in)          :: component_id             !< [IN] component identifier
       integer, intent (in)          :: domain_id                !< [IN] domain identifier
       integer, intent (in)          :: point_ids(*)             !< [IN] point identifier
       integer, intent (in)          :: num_points_per_subdomain !< [IN] number of pointsets per subdomain
       integer, intent (out)         :: field_id                 !< [OUT] returned field handle

     end subroutine yac_fdef_field_nomask

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cdef_field ( field_name,               &
                                 component_id,             &
                                 domain_id,                &
                                 point_ids,                &
                                 mask_ids,                 &
                                 num_points_per_subdomain, &
                                 field_id )                &
    &      bind ( c, name='yac_cdef_field' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       character ( kind=c_char ), dimension(*) :: field_name       !< [IN] short name of the field
       integer ( kind=c_int ), value :: component_id               !< [IN] component identifier
       integer ( kind=c_int ), value :: domain_id                  !< [IN] domain identifier
       integer ( kind=c_int )        :: point_ids(*)               !< [IN] point identifier
       integer ( kind=c_int )        :: mask_ids(*)                !< [IN] mask identifier
       integer ( kind=c_int ), value :: num_points_per_subdomain   !< [IN] number of pointsets per subdomain
       integer ( kind=c_int )        :: field_id                   !< [OUT] returned field handle

     end subroutine yac_cdef_field

  end interface

  interface

     subroutine yac_cdef_field_nomask ( field_name,               &
                                        component_id,             &
                                        domain_id,                &
                                        point_ids,                &
                                        num_points_per_subdomain, &
                                        field_id )                &
    &      bind ( c, name='yac_cdef_field_nomask' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       character ( kind=c_char ), dimension(*) :: field_name       !< [IN] short name of the field
       integer ( kind=c_int ), value :: component_id               !< [IN] component identifier
       integer ( kind=c_int ), value :: domain_id                  !< [IN] domain identifier
       integer ( kind=c_int )        :: point_ids(*)               !< [IN] point identifier
       integer ( kind=c_int ), value :: num_points_per_subdomain   !< [IN] number of pointsets per subdomain
       integer ( kind=c_int )        :: field_id                   !< [OUT] returned field handle

     end subroutine yac_cdef_field_nomask

  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for sending coupling fields
  !!
  !----------------------------------------------------------------------

  interface yac_fput

     subroutine yac_fput_r4 ( field_id,         &
                              nbr_hor_points,   &
                              collection_size,  &
                              nbr_pointsets,    &
                              nbr_subdomains,   &
                              send_field,       &
                              info,             &
                              ierror )

       integer, intent (in)          :: field_id        !< [IN] field identifier
       integer, intent (in)          :: nbr_hor_points  !< [IN] number of horizontal points
       integer, intent (in)          :: collection_size !< [IN] number of vertical level or bundles
       integer, intent (in)          :: nbr_pointsets   !< [IN] number of point sets
       integer, intent (in)          :: nbr_subdomains  !< [IN] number of nbr_subdomain_ids from corresponding
                                                        !! yac_fconnect_subdomains call
       real, intent (in)             :: send_field(nbr_hor_points, &
         &                                         collection_size, &
         &                                         nbr_pointsets, &
         &                                         nbr_subdomains) !< [IN] returned field
       integer, intent (out)         :: info            !< [OUT] returned info argument
       integer, intent (out)         :: ierror          !< [OUT] returned error handler

     end subroutine yac_fput_r4

     subroutine yac_fput_r4_ptr ( field_id,         &
          &                       collection_size,  &
          &                       nbr_pointsets,    &
          &                       nbr_subdomains,   &
          &                       send_field,       &
          &                       info,             &
          &                       ierror )

       import :: yac_r4_ptr

       integer, intent (in)     :: field_id        !< [IN] field identifier
       integer, intent (in)     :: collection_size !< [IN] number of vertical level or bundles
       integer, intent (in)     :: nbr_pointsets   !< [IN] number of point sets
       integer, intent (in)     :: nbr_subdomains  !< [IN] number of nbr_subdomain_ids
                                                   !<      from corresponding
                                                   !<      yac_fconnect_subdomains call
       type(yac_r4_ptr), intent (in) :: send_field(nbr_pointsets, nbr_subdomains, &
         &                                         collection_size)
                                                   !< [IN] returned field
       integer, intent (out)    :: info            !< [OUT] returned info
       integer, intent (out)    :: ierror          !< [OUT] returned error

     end subroutine yac_fput_r4_ptr

     subroutine yac_fput_single_domain_r4 ( field_id,         &
                                            nbr_hor_points,   &
                                            collection_size,  &
                                            nbr_pointsets,    &
                                            nbr_subdomains,   &
                                            send_field,       &
                                            info,             &
                                            ierror )

       integer, intent (in)          :: field_id        !< [IN] field identifier
       integer, intent (in)          :: nbr_hor_points  !< [IN] number of horizontal points
       integer, intent (in)          :: collection_size !< [IN] number of vertical level or bundles
       integer, intent (in)          :: nbr_pointsets   !< [IN] number of point sets
       integer, intent (in)          :: nbr_subdomains  !< [IN] number of nbr_subdomain_ids from corresponding
                                                        !! yac_fconnect_subdomains call
       real, intent (in)             :: send_field(nbr_hor_points, collection_size) !< [IN] returned field
       integer, intent (out)         :: info            !< [OUT] returned info argument
       integer, intent (out)         :: ierror          !< [OUT] returned error handler

     end subroutine yac_fput_single_domain_r4

     subroutine yac_fput_r8 ( field_id,         &
                              nbr_hor_points,   &
                              collection_size,  &
                              nbr_pointsets,    &
                              nbr_subdomains,   &
                              send_field,       &
                              info,             &
                              ierror )

       integer, intent (in)          :: field_id        !< [IN] field identifier
       integer, intent (in)          :: nbr_hor_points  !< [IN] number of horizontal points
       integer, intent (in)          :: collection_size !< [IN] number of vertical level or bundles
       integer, intent (in)          :: nbr_pointsets   !< [IN] number of point sets
       integer, intent (in)          :: nbr_subdomains  !< [IN] number of nbr_subdomain_ids from corresponding
                                                        !! yac_fconnect_subdomains call
       double precision, intent (in) :: send_field(nbr_hor_points, collection_size, nbr_pointsets, nbr_subdomains) !< [IN] returned
                                                                                                               !! field
       integer, intent (out)         :: info            !< [OUT] returned info argument
       integer, intent (out)         :: ierror          !< [OUT] returned error handler

     end subroutine yac_fput_r8

     subroutine yac_fput_r8_ptr ( field_id,         &
          &                       collection_size,  &
          &                       nbr_pointsets,    &
          &                       nbr_subdomains,   &
          &                       send_field,       &
          &                       info,             &
          &                       ierror )

         import :: yac_r8_ptr

         integer, intent (in)     :: field_id        !< [IN] field identifier
         integer, intent (in)     :: collection_size !< [IN] number of vertical level or bundles
         integer, intent (in)     :: nbr_pointsets   !< [IN] number of point sets
         integer, intent (in)     :: nbr_subdomains  !< [IN] number of nbr_subdomain_ids
                                                     !<      from corresponding
                                                     !<      yac_fconnect_subdomains call
         type(yac_r8_ptr), intent (in) :: send_field(nbr_pointsets, nbr_subdomains, &
           &                                         collection_size)
                                                     !< [IN] returned field
         integer, intent (out)    :: info            !< [OUT] returned info
         integer, intent (out)    :: ierror          !< [OUT] returned error

     end subroutine yac_fput_r8_ptr

     subroutine yac_fput_single_domain_r8 ( field_id,         &
                                            nbr_hor_points,   &
                                            collection_size,  &
                                            nbr_pointsets,    &
                                            nbr_subdomains,   &
                                            send_field,       &
                                            info,             &
                                            ierror )

       integer, intent (in)          :: field_id        !< [IN] field identifier
       integer, intent (in)          :: nbr_hor_points  !< [IN] number of horizontal points
       integer, intent (in)          :: collection_size !< [IN] number of vertical level or bundles
       integer, intent (in)          :: nbr_pointsets   !< [IN] number of point sets
       integer, intent (in)          :: nbr_subdomains  !< [IN] number of nbr_subdomain_ids from corresponding
                                                        !! yac_fconnect_subdomains call
       double precision, intent (in) :: send_field(nbr_hor_points, collection_size) !< [IN] returned
                                                                                                               !! field
       integer, intent (out)         :: info            !< [OUT] returned info argument
       integer, intent (out)         :: ierror          !< [OUT] returned error handler

     end subroutine yac_fput_single_domain_r8

  end interface yac_fput

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cput_ ( field_id,         &
                            collection_size,  &
                            send_field,       &
                            info,             &
                            ierror ) bind ( c, name='yac_cput_' )

       use, intrinsic :: iso_c_binding, only : c_int, c_double

       integer ( kind=c_int ), value :: field_id            !< [IN] point identifier
       integer ( kind=c_int ), value :: collection_size     !< [IN] number of vertical level or bundles
       real    ( kind=c_double )     :: send_field(*)       !< [IN] returned field handle
       integer ( kind=c_int )        :: info                !< [OUT] returned info argument
       integer ( kind=c_int )        :: ierror              !< [OUT] returned error handler

     end subroutine yac_cput_

  end interface

  interface

     subroutine yac_cput_ptr_ ( field_id,         &
                                collection_size,  &
                                send_field,       &
                                info,             &
                                ierror ) bind ( c, name='yac_cput_ptr_' )

       use, intrinsic :: iso_c_binding, only : c_int, c_ptr

       integer ( kind=c_int ), value :: field_id            !< [IN] point identifier
       integer ( kind=c_int ), value :: collection_size     !< [IN] number of vertical level or bundles
       type(c_ptr)                   :: send_field(*)       !< [IN] returned field handle
       integer ( kind=c_int )        :: info                !< [OUT] returned info argument
       integer ( kind=c_int )        :: ierror              !< [OUT] returned error handler

     end subroutine yac_cput_ptr_

  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for receiving coupling fields
  !!
  !----------------------------------------------------------------------

  interface yac_fget

     subroutine yac_fget_r4 ( field_id,         &
                              nbr_hor_points,   &
                              collection_size,  &
                              nbr_pointsets,    &
                              nbr_subdomains,   &
                              recv_field,       &
                              info,             &
                              ierror )

       integer, intent (in)  :: field_id        !< [IN] field identifier
       integer, intent (in)  :: nbr_hor_points  !< [IN] number of horizontal points
       integer, intent (in)  :: collection_size !< [IN] number of vertical level or bundles
       integer, intent (in)  :: nbr_pointsets   !< [IN] number of point sets
       integer, intent (in)  :: nbr_subdomains  !< [IN] number of nbr_subdomain_ids from corresponding yac_fconnect_subdomains call
       real, intent (inout)  :: recv_field(nbr_hor_points, collection_size, nbr_pointsets, nbr_subdomains) !< [INOUT] returned field
       integer, intent (out) :: info            !< [OUT] returned info
       integer, intent (out) :: ierror          !< [OUT] returned error handler

     end subroutine yac_fget_r4

     subroutine yac_fget_r4_ptr ( field_id,         &
                                  collection_size,  &
                                  nbr_pointsets,    &
                                  nbr_subdomains,   &
                                  recv_field,       &
                                  info,             &
                                  ierror )

       import :: yac_r4_ptr

       integer, intent (in)  :: field_id        !< [IN] field identifier
       integer, intent (in)  :: collection_size !< [IN] number of vertical level or bundles
       integer, intent (in)  :: nbr_pointsets   !< [IN] number of point sets
       integer, intent (in)  :: nbr_subdomains  !< [IN] number of nbr_subdomain_ids from corresponding yac_fconnect_subdomains call
       type(yac_r4_ptr)      :: recv_field(nbr_pointsets, &
         &                                 nbr_subdomains, &
         &                                 collection_size) !< [OUT] returned field
       integer, intent (out) :: info            !< [OUT] returned info
       integer, intent (out) :: ierror          !< [OUT] returned error handler

     end subroutine yac_fget_r4_ptr

     subroutine yac_fget_single_domain_r4 ( field_id,         &
                                            nbr_hor_points,   &
                                            collection_size,  &
                                            nbr_pointsets,    &
                                            nbr_subdomains,   &
                                            recv_field,       &
                                            info,             &
                                            ierror )

       integer, intent (in)  :: field_id        !< [IN] field identifier
       integer, intent (in)  :: nbr_hor_points  !< [IN] number of horizontal points
       integer, intent (in)  :: collection_size !< [IN] number of vertical level or bundles
       integer, intent (in)  :: nbr_pointsets   !< [IN] number of point sets
       integer, intent (in)  :: nbr_subdomains  !< [IN] number of nbr_subdomain_ids from corresponding yac_fconnect_subdomains call
       real, intent (inout)  :: recv_field(nbr_hor_points, collection_size) !< [INOUT] returned field
       integer, intent (out) :: info            !< [OUT] returned info
       integer, intent (out) :: ierror          !< [OUT] returned error handler

     end subroutine yac_fget_single_domain_r4

     subroutine yac_fget_r8 ( field_id,         &
                              nbr_hor_points,   &
                              collection_size,  &
                              nbr_pointsets,    &
                              nbr_subdomains,   &
                              recv_field,       &
                              info,             &
                              ierror )

       integer, intent (in)  :: field_id        !< [IN] field identifier
       integer, intent (in)  :: nbr_hor_points  !< [IN] number of horizontal points
       integer, intent (in)  :: collection_size !< [IN] number of vertical level or bundles
       integer, intent (in)  :: nbr_pointsets   !< [IN] number of point sets
       integer, intent (in)  :: nbr_subdomains  !< [IN] number of nbr_subdomain_ids from corresponding
                                                !! yac_fconnect_subdomains call
       double precision, intent (inout) :: &
          recv_field(nbr_hor_points, collection_size, nbr_pointsets, nbr_subdomains) !< [INOUT] returned
                                                                                                                      !! field
       integer, intent (out) :: info            !< [OUT] returned info
       integer, intent (out) :: ierror          !< [OUT] returned error handler

     end subroutine yac_fget_r8



     subroutine yac_fget_r8_ptr ( field_id,         &
                                  collection_size,  &
                                  nbr_pointsets,    &
                                  nbr_subdomains,   &
                                  recv_field,       &
                                  info,             &
                                  ierror )

       import :: yac_r8_ptr

       integer, intent (in)  :: field_id        !< [IN] field identifier
       integer, intent (in)  :: collection_size !< [IN] number of vertical level or bundles
       integer, intent (in)  :: nbr_pointsets   !< [IN] number of point sets
       integer, intent (in)  :: nbr_subdomains  !< [IN] number of nbr_subdomain_ids from corresponding yac_fconnect_subdomains call
       type(yac_r8_ptr)      :: recv_field(nbr_pointsets, &
         &                                 nbr_subdomains, &
         &                                 collection_size) !< [OUT] returned field
       integer, intent (out) :: info            !< [OUT] returned info
       integer, intent (out) :: ierror          !< [OUT] returned error handler

     end subroutine yac_fget_r8_ptr

     subroutine yac_fget_single_domain_r8 ( field_id,         &
                                            nbr_hor_points,   &
                                            collection_size,  &
                                            nbr_pointsets,    &
                                            nbr_subdomains,   &
                                            recv_field,       &
                                            info,             &
                                            ierror )

       integer, intent (in)  :: field_id        !< [IN] field identifier
       integer, intent (in)  :: nbr_hor_points  !< [IN] number of horizontal points
       integer, intent (in)  :: collection_size !< [IN] number of vertical level or bundles
       integer, intent (in)  :: nbr_pointsets   !< [IN] number of point sets
       integer, intent (in)  :: nbr_subdomains  !< [IN] number of nbr_subdomain_ids from corresponding yac_fconnect_subdomains call
       double precision, intent (inout) :: recv_field(nbr_hor_points, collection_size) !< [INOUT] returned field
       integer, intent (out) :: info            !< [OUT] returned info
       integer, intent (out) :: ierror          !< [OUT] returned error handler

     end subroutine yac_fget_single_domain_r8

  end interface yac_fget

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cget_ ( field_id,         &
                            collection_size,  &
                            recv_field,       &
                            info,             &
                            ierror ) bind ( c, name='yac_cget_' )

       use, intrinsic :: iso_c_binding, only : c_int, c_double

       integer ( kind=c_int ), value :: field_id            !< [IN] point identifier
       integer ( kind=c_int ), value :: collection_size     !< [IN] collection size
       real    ( kind=c_double )     :: recv_field(*)       !< [OUT] returned field
       integer ( kind=c_int )        :: info                !< [OUT] returned info
       integer ( kind=c_int )        :: ierror              !< [OUT] returned error handler

     end subroutine yac_cget_

  end interface

  interface

     subroutine yac_cget_ptr_ ( field_id,         &
                                collection_size,  &
                                recv_field,       &
                                info,             &
                                ierror ) bind ( c, name='yac_cget_ptr_' )

       use, intrinsic :: iso_c_binding, only : c_int, c_ptr

       integer ( kind=c_int ), value :: field_id            !< [IN] point identifier
       integer ( kind=c_int ), value :: collection_size     !< [IN] collection size
       type(c_ptr)                   :: recv_field(*)       !< [OUT] returned field
       integer ( kind=c_int )        :: info                !< [OUT] returned info
       integer ( kind=c_int )        :: ierror              !< [OUT] returned error handler

     end subroutine yac_cget_ptr_

  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for setting and getting the io_proc_step of a component
  !!
  !----------------------------------------------------------------------

  interface

     subroutine yac_fget_comp_io_proc_step ( comp_id, io_proc_step )

       integer, intent (in)  :: comp_id            !< [IN] component identifier
       integer, intent (out) :: io_proc_step       !< [OUT] stride between ranks of io processes

     end subroutine yac_fget_comp_io_proc_step

     subroutine yac_fset_comp_io_proc_step ( comp_id, io_proc_step )

       integer, intent (in) :: comp_id            !< [IN] component identifier
       integer, intent (in) :: io_proc_step       !< [IN] stride between ranks of io processes

     end subroutine yac_fset_comp_io_proc_step

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cget_comp_io_proc_step ( comp_id, io_proc_step ) &
    &      bind ( c, name='yac_cget_comp_io_proc_step' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int ), value :: comp_id
       integer ( kind=c_int ) :: io_proc_step

     end subroutine yac_cget_comp_io_proc_step

     subroutine yac_cset_comp_io_proc_step ( comp_id, io_proc_step ) &
    &      bind ( c, name='yac_cset_comp_io_proc_step' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int ), value :: comp_id
       integer ( kind=c_int ), value :: io_proc_step

     end subroutine yac_cset_comp_io_proc_step

  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for getting back a local MPI communicator
  !!
  !----------------------------------------------------------------------

  interface

     subroutine yac_fget_localcomm ( local_communicator, comp_id )

       integer, intent (out) :: local_communicator             !< [OUT] MPI communicator
       integer, intent (in)  :: comp_id                        !< [IN]  component ID

     end subroutine yac_fget_localcomm

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_get_localcomm ( local_communicator, comp_id ) &
    &      bind ( c, name='yac_get_localcomm' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int ) :: local_communicator
       integer ( kind=c_int ), value :: comp_id

     end subroutine yac_get_localcomm

  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for getting back a MPI communicator for
  !!   communication between a pair of component root proceses
  !!
  !----------------------------------------------------------------------

  interface

     subroutine yac_fget_pair_rootcomm ( local_communicator, comp_name1, comp_name2 )

       character(len=*), intent(in) :: comp_name1  !< [IN]  first component name
       character(len=*), intent(in) :: comp_name2  !< [IN]  second component name
       integer, intent (out) :: local_communicator !< [OUT] MPI communicator

     end subroutine yac_fget_pair_rootcomm

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_get_pair_rootcomm ( local_communicator, comp_name1, comp_name2 ) &
    &      bind ( c, name='yac_get_pair_rootcomm' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       integer ( kind=c_int ) :: local_communicator
       character ( kind=c_char), dimension(*) :: comp_name1
       character ( kind=c_char), dimension(*) :: comp_name2

     end subroutine yac_get_pair_rootcomm

  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for invoking the end of the definition phase
  !!
  !----------------------------------------------------------------------

  interface

     subroutine yac_fsearch ( nbr_comps, comp_ids, nbr_fields, field_ids, ierror )

       integer, intent (in)  :: nbr_comps             !< [IN]  number of components
       integer, intent (in)  :: comp_ids(nbr_comps)   !< [IN]  list of component IDs
       integer, intent (in)  :: nbr_fields            !< [IN]  number of fields
       integer, intent (in)  :: field_ids(nbr_fields) !< [IN]  list of field IDs
       integer, intent (out) :: ierror                !< [OUT] returned error handler

     end subroutine yac_fsearch

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_csearch ( nbr_comps, comp_ids,  nbr_fields, field_ids, ierror ) &
    &      bind ( c, name='yac_csearch' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int ), value :: nbr_comps            !< [IN]  number of components
       integer ( kind=c_int )        :: comp_ids(nbr_comps)  !< [IN]  list of component IDs
       integer ( kind=c_int ), value :: nbr_fields           !< [IN]  number of fields
       integer ( kind=c_int )        :: field_ids(nbr_comps) !< [IN]  list of field IDs
       integer ( kind=c_int )        :: ierror               !< [OUT] returned error handler

     end subroutine yac_csearch

  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for invoking query functions
  !!
  !----------------------------------------------------------------------

  interface

     subroutine yac_fget_nbr_fields ( nbr_fields )

       integer, intent (out) :: nbr_fields

     end subroutine yac_fget_nbr_fields

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cget_nbr_fields ( nbr_fields ) bind ( c, name='yac_cget_nbr_fields' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int )        :: nbr_fields           !< [OUT]  number of components

     end subroutine yac_cget_nbr_fields

  end interface

  ! ---------------------------------------------------------------------

  interface

     subroutine yac_fget_field_ids ( nbr_fields, field_ids )

       integer, intent (in)  :: nbr_fields
       integer, intent (out) :: field_ids(nbr_fields)

     end subroutine yac_fget_field_ids

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cget_field_ids ( nbr_fields, field_ids ) &
    &      bind ( c, name='yac_cget_field_ids' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int ), value :: nbr_fields            !< [IN]  number of components
       integer ( kind=c_int )        :: field_ids(nbr_fields) !< [OUT] list of field IDS

     end subroutine yac_cget_field_ids

  end interface

  ! ---------------------------------------------------------------------

  interface

     subroutine yac_fget_field_id_from_name ( field_name, field_id )

       character(len=*), intent (in) :: field_name
       integer, intent (out)                       :: field_id

     end subroutine yac_fget_field_id_from_name

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cget_field_id_from_name ( field_name, field_id ) &
    &      bind ( c, name='yac_cget_field_id_from_name' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       character ( kind=c_char), dimension(*) :: field_name  !< [IN]  field name
       integer ( kind=c_int )                 :: field_id    !< [OUT] field ID

     end subroutine yac_cget_field_id_from_name

  end interface

  ! ---------------------------------------------------------------------

  interface

     subroutine yac_fget_field_name_from_id ( field_id, field_name )

       integer, intent (in)                         :: field_id
       character(len=*), intent (out) :: field_name

     end subroutine yac_fget_field_name_from_id

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cget_field_name_from_id ( field_id, field_name ) &
    &      bind ( c, name='yac_cget_field_name_from_id' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       integer ( kind=c_int ), value          :: field_id    !< [IN]  field ID
       character ( kind=c_char), dimension(*) :: field_name  !< [OUT] field name

     end subroutine yac_cget_field_name_from_id

  end interface

  ! ---------------------------------------------------------------------

  interface yac_fget_role_of_field

    subroutine yac_fget_role_of_field_by_id ( field_id, role )

      integer, intent (in)  :: field_id
      integer, intent (out) :: role

    end subroutine yac_fget_role_of_field_by_id

    ! ....................................................................

    subroutine yac_fget_role_of_field_by_name ( field_name, role )

      character (len=*), intent (in)    :: field_name
      integer, intent (out)                           :: role

     end subroutine yac_fget_role_of_field_by_name

  end interface yac_fget_role_of_field

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cget_role_of_field_by_id ( field_id, role ) &
    &      bind ( c, name='yac_cget_role_of_field_by_id' )

       use, intrinsic :: iso_c_binding, only : c_int

       integer ( kind=c_int ), value  :: field_id    !< [IN]  field ID
       integer ( kind=c_int)          :: role        !< [OUT] nothing (0) | put (1) | get (2)

    end subroutine yac_cget_role_of_field_by_id

    ! ...................................................................

    subroutine yac_cget_role_of_field_by_name ( field_name, role ) &
   &      bind ( c, name='yac_cget_role_of_field_by_name' )

      use, intrinsic :: iso_c_binding, only : c_int, c_char

      character ( kind=c_char), dimension(*) :: field_name !< [IN]  field name
      integer ( kind=c_int)                  :: role       !< [OUT] nothing (0) | put (1) | get (2)

    end subroutine yac_cget_role_of_field_by_name

  end interface

! ---------------------------------------------------------------------

  interface yac_fget_model_timestep

     subroutine yac_fget_model_timestep_id ( field_id, string )

       integer, intent (in)                         :: field_id
       character(len=*), intent (out) :: string

     end subroutine yac_fget_model_timestep_id

     ! ..................................................................

     subroutine yac_fget_model_timestep_name ( field_name, string )

       character(len=*), intent (in)  :: field_name
       character(len=*), intent (out) :: string

     end subroutine yac_fget_model_timestep_name

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cget_model_timestep_id ( field_id, string ) &
    &      bind ( c, name='yac_cget_model_timestep_id' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       integer ( kind=c_int ), value          :: field_id    !< [IN] field identifier
       character ( kind=c_char), dimension(*) :: string      !< [OUT] iso time interval

     end subroutine yac_cget_model_timestep_id

     ! ..................................................................

     subroutine yac_cget_model_timestep_name ( field_name, string ) &
    &      bind ( c, name='yac_cget_model_timestep_name' )

       use, intrinsic :: iso_c_binding, only : c_char

       character ( kind=c_char), dimension(*) :: field_name  !< [IN] short name of the field
       character ( kind=c_char), dimension(*) :: string      !< [OUT] iso time interval

     end subroutine yac_cget_model_timestep_name

   end interface

  ! ---------------------------------------------------------------------

  interface yac_fget_coupling_timestep

     subroutine yac_fget_coupling_timestep_id ( field_id, string )

       integer, intent (in)                         :: field_id
       character(len=*), intent (out) :: string

     end subroutine yac_fget_coupling_timestep_id

     ! ..................................................................

     subroutine yac_fget_coupling_timestep_name ( field_name, string )

       character(len=*), intent (in)  :: field_name
       character(len=*), intent (out) :: string

     end subroutine yac_fget_coupling_timestep_name

  end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_cget_coupling_timestep_id ( field_id, string ) &
    &      bind ( c, name='yac_cget_coupling_timestep_id' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       integer ( kind=c_int ), value          :: field_id    !< [IN] field identifier
       character ( kind=c_char), dimension(*) :: string      !< [OUT] iso time interval

     end subroutine yac_cget_coupling_timestep_id

     ! ..................................................................

     subroutine yac_cget_coupling_timestep_name ( field_name, string ) &
    &      bind ( c, name='yac_cget_coupling_timestep_name' )

       use, intrinsic :: iso_c_binding, only : c_char

       character ( kind=c_char), dimension(*) :: field_name  !< [IN] short name of the field
       character ( kind=c_char), dimension(*) :: string      !< [OUT] iso time interval

     end subroutine yac_cget_coupling_timestep_name

   end interface

  ! --- ISO_C interface -------------------------------------------------

  interface

     subroutine yac_abort_message ( text, file, line ) bind ( c, name='yac_abort_message' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       character ( kind=c_char ), dimension(*) :: text       !< [IN] error message
       character ( kind=c_char ), dimension(*) :: file       !< [IN] error message
       integer ( kind=c_int ), value           :: line       !< [IN] line number

     end subroutine yac_abort_message

  end interface

  interface

     subroutine yac_redirstdout ( filestem, lenstr, parallel, my_pe, npes, ierror ) &
    &      bind ( c, name='yac_redirstdout' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       character ( kind=c_char ), dimension(*) :: filestem
       integer ( kind=c_int ), value           :: lenstr
       integer ( kind=c_int ), value           :: parallel
       integer ( kind=c_int ), value           :: my_pe
       integer ( kind=c_int ), value           :: npes
       integer ( kind=c_int )                  :: ierror

     end subroutine yac_redirstdout

  end interface

end module mo_yac_finterface
