/**
 * @file patch_recovery.c
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#if !defined HAVE_LAPACKE && !defined HAVE_MKL_LAPACKE
#  include <assert.h>
#  if defined HAVE_CLAPACK
#    include <clapack.h>
#  else
#    include "cfortran.h"
#  endif
#elif defined HAVE_MKL_LAPACKE
#include <mkl_lapacke.h>
#else
#include <lapacke.h>
#endif
#include "geometry.h"
#include "utils.h"
#include "patch_recovery.h"

/* number of gauss integration points in a unit triangle */

static unsigned const num_gauss_points[8] = {1,1,3,4,6,7,12,13};
static void inverse ( double* A, int n );
static void transposed ( double* A, double* B, int m, int n );

static int mmdotprod ( double* A, int d_a1, int d_a2, double* B, int d_b1, int d_b2,  double* C, int d_c1, int d_c2 );

unsigned yac_get_number_of_quadrature_points ( unsigned order ) {
  if ( order < 1 || order > 7 ) {
    printf ( "Quadrature of order %i not supported.\n", order );
    yac_internal_abort_message ( "ERROR: Quadrature of order not supported", __FILE__, __LINE__ );
  }
  return num_gauss_points[order];
}

/* number of coefficients for a 3D polynomial */

static unsigned const num_polynomial_coefficients[4] = {1,4,10,20};

unsigned yac_get_num_polynomial_coefficients ( unsigned order ) {
  if ( order > 3 ) {
    printf ( "Polynom order %i not supported.\n", order );
    yac_internal_abort_message ( "ERROR: Quadrature of order not supported", __FILE__, __LINE__ );
  }
  return num_polynomial_coefficients[order];
}

/* Abscissa (and weights) for 2D elements */

void yac_linear_gauss_legendre ( unsigned order, double * abscissa ) {

  double const epsilon = 1.0e-14; // relative precision 

  int i, j, m;

  double p1, p2, p3, pp, z, z1;

  /* roots are symmetric in the interval so we only have to find half of them */

  m = (order+1)/2;

  for ( i = 0; i < m; ++i ) {

    z = cos ( M_PI * ( i + 0.75 ) / ( order + 0.5 ) );

    /* Starting with the above approximation to the ith root,
       we enter the main loop of refinement by Newtons's method */

    do {

      p1 = 1.0;
      p2 = 0.0;

      for ( j = 0; j < order; ++j ) {
        p3 = p2;
        p2 = p1;
        p1 = (( 2.0 * j + 1.0  ) * z * p2 - j * p3 ) / ( j + 1 );
      }

      /* p1 is now the desired Legendre polynomial. We next compute pp,
         its derivative, by a standard relation involving also p2, the
         polynomial of one lower order */

      pp = order * ( z * p1 - p2 ) / ( z*z - 1.0 );
      z1 = z;
      z  = z1 - p1/pp;

    } while ( fabs(z-z1) > epsilon );

    /* Abscissa and its symmetric counterpart already scaled to [0,1] */

    abscissa[i]         = 0.5*(1.0-z);
    abscissa[order-1-i] = 0.5*(1.0+z);

    /* weights and their symmetric counterpart

    weight[i]         = 0.5 * (( 1.0 - z*z) * pp*pp );
    weight[order-1-i] = w[i];  */
  }

}

void yac_triangle_gauss_legendre ( unsigned order, double * abscissa ) {

  if ( order == 1 ) {
    abscissa[0]  =  0.33333333333333333;
    abscissa[1]  =  0.33333333333333333;
  }
  else if ( order == 2 ) {
    abscissa[0]  =  0.16666666666666667;
    abscissa[1]  =  0.16666666666666667;
    abscissa[2]  =  0.66666666666666667;
    abscissa[3]  =  0.16666666666666667;
    abscissa[4]  =  0.16666666666666667;
    abscissa[5]  =  0.66666666666666667;
  }
  else if ( order == 3 ) {
    abscissa[0]  =  0.33333333333333333;
    abscissa[1]  =  0.33333333333333333;
    abscissa[2]  =  0.20000000000000000;
    abscissa[3]  =  0.20000000000000000;
    abscissa[4]  =  0.60000000000000000;
    abscissa[5]  =  0.20000000000000000;
    abscissa[6]  =  0.20000000000000000;
    abscissa[7]  =  0.60000000000000000;
  }
  else if ( order == 4 ) {
    abscissa[0]  =  0.091576213509771;
    abscissa[1]  =  0.091576213509771;
    abscissa[2]  =  0.816847572980459;
    abscissa[3]  =  0.091576213509771;
    abscissa[4]  =  0.091576213509771;
    abscissa[5]  =  0.816847572980459;
    abscissa[6]  =  0.445948490915965;
    abscissa[7]  =  0.445948490915965;
    abscissa[8]  =  0.108103018168070;
    abscissa[9]  =  0.445948490915965;
    abscissa[10] =  0.445948490915965;
    abscissa[11] =  0.108103018168070;
  }
  else if ( order == 5 ) {
    abscissa[0]  =  0.333333333333333;
    abscissa[1]  =  0.333333333333333;
    abscissa[2]  =  0.101286507323456;
    abscissa[3]  =  0.101286507323456;
    abscissa[4]  =  0.797426985353087;
    abscissa[5]  =  0.101286507323456;
    abscissa[6]  =  0.101286507323456;
    abscissa[7]  =  0.797426985353087;
    abscissa[8]  =  0.470142064105115;
    abscissa[9]  =  0.470142064105115;
    abscissa[10] =  0.059715871789770;
    abscissa[11] =  0.470142064105115;
    abscissa[12] =  0.470142064105115;
    abscissa[13] =  0.059715871789770;
  }
  else if ( order == 6 ) {
    abscissa[0]  =  0.063089014491502;
    abscissa[1]  =  0.063089014491502;
    abscissa[2]  =  0.873821971016996;
    abscissa[3]  =  0.063089014491502;
    abscissa[4]  =  0.063089014491502;
    abscissa[5]  =  0.873821971016996;
    abscissa[6]  =  0.249286745170910;
    abscissa[7]  =  0.249286745170910;
    abscissa[8]  =  0.501426509658179;
    abscissa[9]  =  0.249286745170910;
    abscissa[10] =  0.249286745170910;
    abscissa[11] =  0.501426509658179;
    abscissa[12] =  0.310352451033785;
    abscissa[13] =  0.053145049844816;
    abscissa[14] =  0.053145049844816;
    abscissa[15] =  0.310352451033785;
    abscissa[16] =  0.636502499121399;
    abscissa[17] =  0.053145049844816;
    abscissa[18] =  0.053145049844816;
    abscissa[19] =  0.636502499121399;
    abscissa[20] =  0.636502499121399;
    abscissa[21] =  0.310352451033785;
    abscissa[22] =  0.310352451033785;
    abscissa[23] =  0.636502499121399;
  }
  else if ( order == 7 ) {
    abscissa[0]  =  0.333333333333333;
    abscissa[1]  =  0.333333333333333;
    abscissa[2]  =  0.260345966079038;
    abscissa[3]  =  0.260345966079038;
    abscissa[4]  =  0.479308067841923;
    abscissa[5]  =  0.260345966079038;
    abscissa[6]  =  0.260345966079038;
    abscissa[7]  =  0.479308067841923;
    abscissa[8]  =  0.065130102902216;
    abscissa[9]  =  0.065130102902216;
    abscissa[10] =  0.869739794195568;
    abscissa[11] =  0.065130102902216;
    abscissa[12] =  0.065130102902216;
    abscissa[13] =  0.869739794195568;
    abscissa[14] =  0.312865496004874;
    abscissa[15] =  0.048690315425316;
    abscissa[16] =  0.048690315425316;
    abscissa[17] =  0.312865496004874;
    abscissa[18] =  0.638444188569809;
    abscissa[19] =  0.048690315425316;
    abscissa[20] =  0.048690315425316;
    abscissa[21] =  0.638444188569809;
    abscissa[22] =  0.638444188569809;
    abscissa[23] =  0.312865496004874;
    abscissa[24] =  0.312865496004874;
    abscissa[25] =  0.638444188569809;
  }
  else {
    printf ( "Quadrature of order %i not supported.\n", order );
    yac_internal_abort_message ( "ERROR: Quadrature of order not supported", __FILE__, __LINE__ );
  }

}

/* ------------- unit quadrilateral --------------- */

void yac_quad_2d_locations ( unsigned order, double * abscissa ) {

  unsigned i, j;
  double absc[order];
  // double *w;
 
  unsigned const dim = 2;

  yac_linear_gauss_legendre ( order, absc );

  for ( i = 0; i < order; ++i ) {
    for ( j = 0; j < order; ++j ) {
      abscissa[(i*order+j)*dim]   = absc[i];
      abscissa[(i*order+j)*dim+1] = absc[j];
      // weights[i*order+j]           = w[i]*w[j];
    }
  }
}

/* ------------ transformation to triangle ------------- */

void yac_reference_to_xyz_triangle ( unsigned num_points, struct xyz_triangle triangle,
                                     double * reference_location,
                                     double * transformed_location ) {
  unsigned i;
  static unsigned const dim3 = 3;

  for ( i = 0; i < num_points; ++i ) {

    transformed_location[i*dim3]   = ( triangle.a[0] - triangle.c[0] ) * reference_location[i*dim3]
                                   + ( triangle.b[0] - triangle.c[0] ) * reference_location[i*dim3+1]
                                   -                   triangle.c[0]   * reference_location[i*dim3+2]
                                   +                   triangle.c[0];

    transformed_location[i*dim3+1] = ( triangle.a[1] - triangle.c[1] ) * reference_location[i*dim3]
                                   + ( triangle.b[1] - triangle.c[1] ) * reference_location[i*dim3+1]
                                   -                   triangle.c[1]   * reference_location[i*dim3+2]
                                   +                   triangle.c[1];

    transformed_location[i*dim3+2] = ( triangle.a[2] - triangle.c[2] ) * reference_location[i*dim3]
                                   + ( triangle.b[2] - triangle.c[2] ) * reference_location[i*dim3+1]
                                   -                   triangle.c[2]   * reference_location[i*dim3+2]
                                   +                   triangle.c[2];
  }
}

void yac_xyz_triangle_to_reference ( unsigned num_points, struct xyz_triangle triangle,
                                     double * reference_location,
                                     double * transformed_location ) {
  unsigned i, j;
  static unsigned const dim = 3;

  double A[3*3] = { triangle.a[0] - triangle.c[0] , triangle.b[0] - triangle.c[0] , -triangle.c[0],
                    triangle.a[1] - triangle.c[1] , triangle.b[1] - triangle.c[1] , -triangle.c[1],
                    triangle.a[2] - triangle.c[2] , triangle.b[2] - triangle.c[2] , -triangle.c[2] };

  inverse ( (double *) A, dim );

  for ( i = 0; i < num_points; ++i )
     for ( j = 0; j < dim; ++j )
        reference_location[i*dim+j] = A[j*dim+0] * ( transformed_location[i*dim+0] - triangle.c[0] )
                                    + A[j*dim+1] * ( transformed_location[i*dim+1] - triangle.c[1] )
                                    + A[j*dim+2] * ( transformed_location[i*dim+2] - triangle.c[2] );
}

/* ------------ transformation to quadrilateral ------------- */

void yac_reference_to_xy_quadrilateral ( unsigned num_points, struct grid_cell cell,
                                         double * reference_location,
                                         double * transformed_location ) {
  unsigned i, n;
  static unsigned const dim = 2;

  double const xsi[4] = {-1.0,  1.0, 1.0, -1.0 };
  double const eta[4] = {-1.0, -1.0, 1.0,  1.0 };

  for ( i = 0; i < num_points; ++i ) {

    transformed_location[i*dim+0] = 0.0;
    transformed_location[i*dim+1] = 0.0;

    for ( n = 0; n < cell.num_corners; ++ n ) {

      transformed_location[i*dim+0] += 0.25
        * ( 1 + xsi[n] * reference_location[i*dim+0] )
        * ( 1 + eta[n] * reference_location[i*dim+1] ) * cell.coordinates_x[n];

      transformed_location[i*dim+1] += 0.25
        * ( 1 + xsi[n] * reference_location[i*dim+0] )
        * ( 1 + eta[n] * reference_location[i*dim+1] ) * cell.coordinates_y[n];
    }
  }  
}

void yac_reference_to_xyz_quadrilateral ( unsigned gauss_order, struct grid_cell cell,
                                          double * reference_location,
                                          double * transformed_location ) {

   /* quadrilaterals (as well as triangles) could also be handled by xyz_polygon
      the resulting positions of integrations points would then be different. */

   static unsigned const dim2 = 2;
   static unsigned const dim3 = 3;

   static unsigned num_lowerhalf = 0;
   static unsigned num_upperhalf = 0;

   struct xyz_triangle xyz_triangle;

   double reference_3Dlocation[3*gauss_order*gauss_order];
   double buffer_3d[3*gauss_order*gauss_order];

   unsigned i, j;

   static double const tol = 1.0e-15;
 
   LLtoXYZ( cell.coordinates_x[0], cell.coordinates_y[0], xyz_triangle.a );
   LLtoXYZ( cell.coordinates_x[1], cell.coordinates_y[1], xyz_triangle.b );
   LLtoXYZ( cell.coordinates_x[2], cell.coordinates_y[2], xyz_triangle.c );

   num_lowerhalf = 0;

   for ( i = 0; i < gauss_order; ++i ) {
      for (j = 0; j < gauss_order; ++j ) {
         if ( reference_location[(i*gauss_order+j)*dim2]+reference_location[(i*gauss_order+j)*dim2+1] <= 1.0+tol ) {
            reference_3Dlocation[num_lowerhalf*dim3+0] = reference_location[(i*gauss_order+j)*dim2];
            reference_3Dlocation[num_lowerhalf*dim3+1] = reference_location[(i*gauss_order+j)*dim2+1];
            reference_3Dlocation[num_lowerhalf*dim3+2] = 0.0;
            ++num_lowerhalf;
         }
      }
   }

  yac_reference_to_xyz_triangle ( num_lowerhalf, xyz_triangle, reference_3Dlocation, buffer_3d );

  for ( i = 0; i < 3*num_lowerhalf; ++i )
     transformed_location[i] = buffer_3d[i];

  LLtoXYZ( cell.coordinates_x[2], cell.coordinates_y[2], xyz_triangle.a );
  LLtoXYZ( cell.coordinates_x[3], cell.coordinates_y[3], xyz_triangle.b );
  LLtoXYZ( cell.coordinates_x[0], cell.coordinates_y[0], xyz_triangle.c );

  num_upperhalf = 0;

  for ( i = 0; i < gauss_order; ++i ) {
     for (j = 0; j < gauss_order; ++j ) {
        if ( reference_location[(i*gauss_order+j)*dim2]+reference_location[(i*gauss_order+j)*dim2+1] < 1.0-tol ) {
           reference_3Dlocation[num_upperhalf*dim3+0] = reference_location[(i*gauss_order+j)*dim2];
           reference_3Dlocation[num_upperhalf*dim3+1] = reference_location[(i*gauss_order+j)*dim2+1];
           reference_3Dlocation[num_upperhalf*dim3+2] = 0.0;
           ++num_upperhalf;
        }
     }
  }

  yac_reference_to_xyz_triangle ( num_upperhalf, xyz_triangle, reference_3Dlocation, buffer_3d );

  for ( i = 0; i < 3*num_upperhalf; ++i )
     transformed_location[3*num_lowerhalf+i] = buffer_3d[i];

  if ( num_lowerhalf+num_upperhalf != gauss_order*gauss_order ) {
     printf ("compared %d %d %d \n",  num_lowerhalf, num_upperhalf, gauss_order*gauss_order );
     yac_internal_abort_message ( "ERROR: Quadrature points do not match", __FILE__, __LINE__ );
  }

}

/* ------------ transformation to polygons ------------- */

static
void get_barycenter ( unsigned num_vertices, struct xyz_polygon polygon, double barycenter[] ) {

   unsigned n;
   unsigned const dim3 = 3;

   for ( n = 0; n < dim3; ++n )
	  barycenter[n] = 0.0;

   for ( n = 0; n < num_vertices; ++n ) {
      barycenter[0] += polygon.x[n];
      barycenter[1] += polygon.y[n];
      barycenter[2] += polygon.z[n];
   }

   for ( n = 0; n < dim3; ++n )
      barycenter[n] /= (double) num_vertices;

   /* project barycenter to surface of unit sphere */

   double length =
         sqrt ( barycenter[0]*barycenter[0] +
                barycenter[1]*barycenter[1] +
                barycenter[2]*barycenter[2] );

   for ( n = 0; n < dim3; ++n )
      barycenter[n] /= length;
}

void yac_reference_to_xyz_polygon ( unsigned num_gauss_points,
                                    struct xyz_polygon polygon,
                                    double * reference_3Dlocation,
                                    double * transformed_location ) {

   /* the total number of transformed locations to be returned
      equals num_gauss_points*num_vertices */

   unsigned const dim3 = 3;
   unsigned n, ngp;

   double barycenter[dim3];
   struct xyz_triangle xyz_triangle;

   double buffer_3d[dim3*num_gauss_points];

   /* the barycenter should always give us a point inside a convex polygon */

   unsigned num_vertices = polygon.num_vertices;

   get_barycenter ( num_vertices, polygon, barycenter );

   for ( n = 0; n < dim3; ++n )
      xyz_triangle.c[n] = barycenter[n];

   for ( n = 0; n < num_vertices; ++n ) {

      unsigned n0 = n%num_vertices;
      unsigned n1 = (n+1)%num_vertices;

      xyz_triangle.a[0] = polygon.x[n0];
      xyz_triangle.a[1] = polygon.y[n0];
      xyz_triangle.a[2] = polygon.z[n0];

      xyz_triangle.b[0] = polygon.x[n1];
      xyz_triangle.b[1] = polygon.y[n1];
      xyz_triangle.b[2] = polygon.z[n1];

      yac_reference_to_xyz_triangle ( num_gauss_points, xyz_triangle, reference_3Dlocation, buffer_3d );

      unsigned offset = n * num_gauss_points * dim3;

      for ( ngp = 0; ngp < dim3*num_gauss_points; ++ngp )
         transformed_location[offset+ngp] = buffer_3d[ngp];
   }
}

#if !defined HAVE_LAPACKE && !defined HAVE_MKL_LAPACKE
#  if defined HAVE_CLAPACK
#    define LAPACK_COL_MAJOR CblasColMajor
static int
LAPACKE_dgels_work(int matrix_order, char trans, int M, int N,
                   int NRHS, double *A, int LDA, double *B, int LDB,
                   double *work, int lwork)
{
  assert(matrix_order == LAPACK_COL_MAJOR);
  return clapack_dgels((enum CBLAS_ORDER)matrix_order,
                       (enum CBLAS_TRANSPOSE)trans,
                       M, N, NRHS, A, LDA, B, LDB);
}

static int
LAPACKE_dgetrf(int matrix_order, int M, int N,
               double *A, int LDA, int *IPIV)
{
  assert(matrix_order == LAPACK_COL_MAJOR);
  return clapack_dgetrf((enum CBLAS_ORDER)matrix_order,
                        M, N, A, LDA, IPIV);
}

static int
LAPACKE_dgetri_work(int matrix_order, int N, double *A,
                    int LDA, const int *IPIV,
                    double *WORK, int LWORK)
{
  assert(matrix_order == LAPACK_COL_MAJOR);
  return clapack_dgetri((enum CBLAS_ORDER)matrix_order,
                        N, A, LDA, IPIV);
}

#  else
#    define LAPACK_COL_MAJOR 102
static int
LAPACKE_dgels_work(int matrix_order, char trans, int M, int N,
                   int NRHS, double *A, int LDA, double *B, int LDB,
                   double *work, int lwork)
{
  assert(matrix_order == LAPACK_COL_MAJOR);
  PROTOCCALLSFSUB11(DGELS,dgels,STRING,INT,INT,INT,DOUBLEV,
                    INT,DOUBLEV,INT,DOUBLEV,INT,
                    PINT);

  int info;
  char trans_farg[2] = { trans, '\0' };

  CCALLSFSUB11(DGELS,dgels,STRING,INT,INT,INT,DOUBLEV,
               INT,DOUBLEV,INT,DOUBLEV,INT,PINT,
               trans_farg,M,N,NRHS,A,LDA,B,LDB,work,lwork,info);

  return info;
}

static int
LAPACKE_dgetrf(int matrix_order, int M, int N,
               double *A, int LDA, int *IPIV)
{
  assert(matrix_order == LAPACK_COL_MAJOR);
  PROTOCCALLSFSUB6(DGETRF,dgetrf,INT,INT,DOUBLEV,INT,INTV,PINT);

  int info;

  CCALLSFSUB6(DGETRF,dgetrf,INT,INT,DOUBLEV,INT,INTV,PINT,
               M,N,A,LDA,IPIV,info);
  return info;
}

static int
LAPACKE_dgetri_work(int matrix_order, int N, double *A,
                    int LDA, const int *IPIV,
                    double *WORK, int LWORK)
{
  assert(matrix_order == LAPACK_COL_MAJOR);
  PROTOCCALLSFSUB7(DGETRI,dgetri,INT,DOUBLEV,INT,INTV,DOUBLEV,INT,PINT);

  int info;

  CCALLSFSUB7(DGETRI,dgetri,INT,DOUBLEV,INT,INTV,DOUBLEV,INT,PINT,
              N,A,LDA,(int *)IPIV,WORK,LWORK,info);
  return info;
}
#  endif
#endif

/* --------------- polynomial treatment ---------------- */
void yac_set_polynomial_terms ( unsigned order, unsigned num_rows,
                                double * coordinates, double * reference, double *PT ) {

  unsigned i;

  if ( order > 3 )
     yac_internal_abort_message ( "ERROR: polynomial order not in the range 0 <= order <= 3 ", __FILE__, __LINE__ );

  for ( i = 0; i < num_rows; ++i )
      PT[0*num_rows+i] = 1.0;

  if ( order == 0 ) return;

  if ( num_rows > 65536 ) yac_internal_abort_message("internal ERROR: too much memory required", __FILE__, __LINE__);

  double x[num_rows], y[num_rows], z[num_rows];

  /* actually this give us the transposed of the matrix if we think of P
     containing only line for each integration points. */

  for ( i = 0; i < num_rows; ++i ) {
    x[i] = coordinates[i*3+0] - reference[0];
    y[i] = coordinates[i*3+1] - reference[1];
    z[i] = coordinates[i*3+2] - reference[2];
  }

  for ( i = 0; i < num_rows; ++i ) {
      PT[1*num_rows+i] = x[i];
      PT[2*num_rows+i] = y[i];
      PT[3*num_rows+i] = z[i];
  }

  if ( order > 1 ) {
    for ( i = 0; i < num_rows; ++i ) {
      PT[4*num_rows+i] = x[i] * x[i];
      PT[5*num_rows+i] = y[i] * y[i];
      PT[6*num_rows+i] = z[i] * z[i];
      PT[7*num_rows+i] = x[i] * y[i];
      PT[8*num_rows+i] = x[i] * z[i];
      PT[9*num_rows+i] = y[i] * z[i];
    }
  }

  if ( order > 2 ) {
    for ( i = 0; i < num_rows; ++i ) {
      PT[10*num_rows+i] = x[i] * x[i] * x[i];
      PT[11*num_rows+i] = y[i] * y[i] * y[i];
      PT[12*num_rows+i] = z[i] * z[i] * z[i];
      PT[13*num_rows+i] = x[i] * x[i] * y[i];
      PT[14*num_rows+i] = x[i] * x[i] * z[i];
      PT[15*num_rows+i] = y[i] * y[i] * x[i];
      PT[16*num_rows+i] = y[i] * y[i] * z[i];
      PT[17*num_rows+i] = z[i] * z[i] * x[i];
      PT[18*num_rows+i] = z[i] * z[i] * y[i];
      PT[19*num_rows+i] = x[i] * y[i] * z[i];
    }
  }
}

void yac_get_polynomial_coefficients ( unsigned num_points, unsigned num_coeff,
                                       double *PT, double *b, double*a ) {

  int lwork = num_points * 64;
  int nrhs  = 1;

  /* int lwork = MIN(num_coeff, num_points) * 64;
     work = malloc ( lwork * sizeof(*work) ); */

  double work[num_points*64];

  /* Fortran ordering */

  int info = LAPACKE_dgels_work(LAPACK_COL_MAJOR, 'N', num_points, num_coeff,
                                nrhs, PT, num_points, b, num_points, work, lwork);
  if ( info != 0 ) yac_internal_abort_message("internal ERROR: dgels", __FILE__, __LINE__);

  memcpy ( a , b, num_coeff * sizeof(*b) );

}

/* ----------- methods to calculate patch values ----------------- */

void yac_patch_P0weights ( unsigned num_points, unsigned num_coeff,
                           unsigned num_cells, unsigned * num_gausspoints,
                           double *Pt, double *PT, double *weight ) {

  unsigned i, n, k;
  unsigned mi, mj;
  unsigned offset[num_cells];

  double Minv[num_coeff*num_coeff];
  double MinvPT[num_coeff*num_points];

  // PT already depends on the individual target point, thus we have to
  // perform all operations for each target point individually.

  // Calculate elements (mi,mj) of M as sum over all n points
  //
  // Mathematically it is the matrix product of PT * P, where
  // PT is the transposed of P. Here we do two steps in one
  // and use PT directly but step through such that it is equivalent
  // to PT * P.

  for ( mi = 0; mi < num_coeff; ++mi ) {
     for ( mj = 0; mj < num_coeff; ++mj ) {
        Minv[mi*num_coeff+mj] = 0;
        for ( n = 0; n < num_points; ++n )
            Minv[mi*num_coeff+mj] += PT[mi*num_points+n]*PT[mj*num_points+n];
     }
  }

  inverse ( Minv, num_coeff );

  for ( n = 0; n < num_points; ++n ) {
     for ( mi = 0; mi < num_coeff; ++mi ) {
        MinvPT[n*num_coeff+mi] = 0;
        for ( mj = 0; mj < num_coeff; ++mj )
           MinvPT[n*num_coeff+mi] += Minv[mi*num_coeff+mj] * PT[mj*num_points+n];
     }
  }

  offset[0] = 0;
  for ( n = 1; n < num_cells; ++n )
	  offset[n] = offset[n-1]+num_gausspoints[n];

  for ( n = 0; n < num_cells; ++n ) {
     weight[n] = 0.0;
     for ( i = 0; i < num_gausspoints[n]; ++i )
        for ( k = 0; k < num_coeff; ++k )
           weight[n] += Pt[k] * MinvPT[(offset[n]+i)*num_coeff+k];
  }

}

void yac_patch_P0weights_opt ( unsigned num_points, unsigned num_coeff,
                               unsigned num_cells, unsigned * num_gausspoints,
                               double *PT, double *weight ) {

  unsigned i, n;
  unsigned mi, mj;

  unsigned offset[num_cells];
  double Minv[num_coeff*num_coeff];
  double MinvPT[num_points];

  // PT already depends on the individual target point, thus we
  // have to perform all operations for each target point individually.

  // Calculate elements (mi,mj) of M as sum over all n points
  //
  // Mathematically it is the matrix product of PT * P, where
  // PT is the transposed of P. Here we do two steps in one
  // and use PT directly but step through such that it is equivalent
  // to PT * P.

  for ( mi = 0; mi < num_coeff; ++mi ) {
     for ( mj = 0; mj < num_coeff; ++mj ) {
        Minv[mi*num_coeff+mj] = 0;
        for ( n = 0; n < num_points; ++n )
            Minv[mi*num_coeff+mj] += PT[mi*num_points+n]*PT[mj*num_points+n];
     }
  }

  inverse ( Minv, num_coeff );

  // By construction the Pt[0] = 1.0 and Pt[n] = 0.0 for all n > 0
  // in the non-optimised version above. Pt, the polynomial evaluated
  // at the target position is different in meaning from PT, the transposed
  // of the polynomial P.
  //
  // Compared to the full version above we can therefore drop some
  // loops. We cannot do much above these lines as we still need
  // the full matrices to calculate the inverse.

  for ( n = 0; n < num_points; ++n ) MinvPT[n] = 0;

  for ( mj = 0; mj < num_coeff; ++mj )
     for ( n = 0; n < num_points; ++n )
        MinvPT[n] += Minv[mj] * PT[mj*num_points+n];

  offset[0] = 0;
  for ( n = 1; n < num_cells; ++n )
	  offset[n] = offset[n-1]+num_gausspoints[n];

  for ( n = 0; n < num_cells; ++n ) {
     weight[n] = 0.0;
#ifdef _CRAYC
#if (_RELEASE_MAJOR == 8) && (_RELEASE_MINOR == 3)
#pragma _CRI novector
#endif
#endif
     for ( i = 0; i < num_gausspoints[n]; ++i )
        weight[n] += MinvPT[offset[n]+i];
  }
}

double yac_patch_P0value ( unsigned num_cells, double *PtMinvPT, double *fi ) {

   unsigned n;
   double value = 0.0;

   for ( n = 0; n < num_cells; ++n ) value += PtMinvPT[n] * fi[n];

   return value;
}

double yac_patch_value_alt1 ( unsigned num_points, unsigned num_coeff,
                              double *Pt, double *PT, double *fi ) {

  unsigned k, n;
  int info;
  double value = 0.0;

  double P[num_points*num_coeff];
  double PTPinv[num_coeff*num_coeff];
  double PTf[num_coeff];
  double PTPinv_PTf[num_coeff];

  transposed ( PT, P, num_coeff, num_points );

  info = mmdotprod ( PT, num_coeff, num_points, P, num_points, num_coeff, PTPinv, num_coeff, num_coeff );
  if ( info < 0 ) yac_internal_abort_message("internal ERROR: wrong dimensions for mmdotprod", __FILE__, __LINE__);

  inverse ( PTPinv, num_coeff );

  for ( k = 0; k < num_coeff; ++k ) {
    PTf[k] = 0.0;
    for ( n = 0; n < num_points; ++n )
       PTf[k] += PT[k*num_points+n]*fi[n];
  }

  for ( k = 0; k < num_coeff; ++k ) {
     PTPinv_PTf[k] = 0.0;
     for ( n = 0; n < num_coeff; ++n )
        PTPinv_PTf[k] += PTPinv[k*num_coeff+n] * PTf[n];
  }

  for ( k = 0; k < num_coeff; ++k )
     value += Pt[k] * PTPinv_PTf[k];

  return value;
}

double yac_patch_value_alt2 ( unsigned num_coeff, double *P, double *a ) {

  unsigned n;
  double value = 0.0;

  for ( n = 0; n < num_coeff; ++n ) value += P[n] * a[n];

  return value;
}

/* ----------- matrix operations ----------------- */

/**  Linear algebra: calculates the inverse of a matrix a of dim n x n.
  *
  *  @param[in] A           transposed of matrix P containing the polynomial terms
  *  @param[in] n           number of rows an columns of A
  *
 **/

static void inverse ( double* A, int n ) {

   int ipiv[n+1];
   int info;
   int lwork = n*n;

   if ( n > 512 ) yac_internal_abort_message("internal ERROR: to much memory required", __FILE__, __LINE__);

   double work[lwork];

   info = LAPACKE_dgetrf(LAPACK_COL_MAJOR, n, n, A, n, ipiv);
   if ( info != 0 )
       yac_internal_abort_message("internal ERROR: dgetrf", __FILE__, __LINE__);

   info = LAPACKE_dgetri_work(LAPACK_COL_MAJOR, n, A, n, ipiv, work, lwork);
   if ( info != 0 )
       yac_internal_abort_message("internal ERROR: dgetri", __FILE__, __LINE__);

}
/**  Linear algebra: calculates the transposed matrix a of dim m x n.
  *
  *  @param[in]  A          matrix A
  *  @param[out] B          transposed of A
  *  @param[in]  M          number of rows (columns) of A (B)
  *  @param[in]  N          number of columns (rows) of A (B)
  *
 **/

static void transposed ( double* A, double* B, int M, int N ) {

  int m, n;

  for ( n = 0; n < N; ++n )
    for ( m = 0; m < M; ++m )
      B[n*M+m] = A[m*N+n];

}

/**  Linear algebra: calculates the transposed matrix a of dim m x n.
  *
  *  @param[in]  A          matrix A
  *  @param[in]  d_a1       number of rows of A
  *  @param[in]  d_a2       number of columns of A, needs to be equal to d_b1
  *  @param[in]  B          matrix A
  *  @param[in]  d_b1       number of rows of B, needs to be equal to d_a2
  *  @param[in]  d_b2       number of columns of B
  *  @param[out] C          matrix C = A B
  *  @param[in]  d_c1       number of rows of C, needs to be equal to d_a1
  *  @param[in]  d_c2       number of columns of C, needs to be equal to d_b2
  *
 **/

static int mmdotprod ( double* A, int d_a1, int d_a2, double* B, int d_b1, int d_b2,  double* C, int d_c1, int d_c2 ) {

  /* could be replaced by

     void BLAS_dgemm( enum blas_order_type order, enum blas_transpose_type transa,
                      enum blas_transpose_type transb, int m, int n, int k,
                      SCALAR_IN alpha, const ARRAY a, int lda, const ARRAY b,
                      int ldb, SCALAR_IN beta, ARRAY c, int ldc );

     http://en.wikipedia.org/wiki/General_Matrix_Multiply

     only needed for alternative 1 to calculate patch value at target point

  */

  int i, j, k;

  if ( d_a2 != d_b1 ) return -1;
  if ( d_c1 != d_a1 ) return -2;
  if ( d_c2 != d_b2 ) return -3;

  for ( i = 0; i < d_a1; ++i ) {
    for ( j = 0; j < d_b2; j++ ) {
       C[i*d_a1+j] = 0.0;
       for ( k = 0; k < d_a2; k++ )
          C[i*d_a1+j] += A[i*d_a2+k]*B[k*d_b2+j];
    }
  }

  return 0;

}
