/**
 * @file config_xml.c
 *
 * @copyright Copyright  (C)  2013 Rene Redler <rene.redler@mpimet.mpg.de>
 *                                 Teresa Holfeld <teresa.holfeld@zmaw.de>
 *
 * @version 1.0
 * @author Rene Redler <rene.redler@mpimet.mpg.de>
 *         Teresa Holfeld <teresa.holfeld@zmaw.de>
 */
/*
 * Keywords:
 * Maintainer: Rene Redler <rene.redler@mpimet.mpg.de>
 *             Teresa Holfeld <teresa.holfeld@zmaw.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "utils.h"
#include "config_xml.h"
#include "interpolation_enum.h"

#ifdef USE_XML2

//* macros and constants representing xml elements and attributes *************/


// first \<transient_couple\> is element of index 3 in \<couple\>
#define FIRSTTRANSIENTCOUPLE 2

// names of xml elements and attributes
static const char UTF8[] = "UTF-8";
static const char REDIRECT[] = "redirect";
static const char REDIRECT_STDOUT[] = "redirect_stdout";
static const char REDIRECT_OF_ROOT[] = "redirect_of_root";
static const char COMPONENTS[] = "components";
static const char COMPONENT[] = "component";
static const char TRANSIENTS[] = "transients";
static const char TRANSIENT[] = "transient";
static const char TRANSIENT_GRID_REFS[] = "transient_grid_refs";
static const char TRANSIENT_GRID_REF[] = "transient_grid_ref";
static const char TRANSIENT_REF[] = "transient_ref";
static const char COLLECTION_SIZE[] = "collection_size";
static const char GRID_REF[] = "grid_ref";
static const char COUPLES[] = "couples";
static const char COUPLE[] = "couple";
static const char DATES[] = "dates";
static const char START_DATE[] = "start_date";
static const char END_DATE[] = "end_date";
static const char CALENDAR[] = "calendar";
static const char TIMESTEP_UNIT[] = "timestep_unit";
static const char TIMESTEP[] = "timestep";
static const char ID[] = "id";
static const char NAME[] = "name";
static const char MODEL[] = "model";
static const char SIMULATED[] = "simulated";
static const char GRIDS[] = "grids";
static const char GRID[] = "grid";
static const char COMPONENT1[] = "component1";
static const char COMPONENT2[] = "component2";
static const char COMPONENT_ID[] = "component_id";
static const char COMPONENT_REF[] = "component_ref";
static const char TRANSIENT_COUPLE[] = "transient_couple";
static const char SOURCE[] = "source";
static const char TARGET[] = "target";
static const char TRANSIENT_ID[] = "transient_id";
static const char TRANSIENT_STANDARD_NAME[] = "transient_standard_name";
static const char ALIAS_NAME[] = "alias_name";
static const char COUPLING_PERIOD[] = "coupling_period";
static const char OPERATION[] = "operation";
static const char SOURCE_TIMELAG[] = "source_timelag";
static const char TARGET_TIMELAG[] = "target_timelag";
static const char INTERPOLATION_REQUIREMENTS[] = "interpolation_requirements";
static const char INTERPOLATION[] = "interpolation";
static const char METHOD[] = "method";
static const char USE_SOURCE_MASK[] = "use_source_mask";
static const char USE_TARGET_MASK[] = "use_target_mask";
static const char MAPPING_ON_SOURCE[] = "mapping_on_source";
static const char DEBUG_MODE[] = "debug_mode";
static const char AT_SOURCE_BEFORE_INTERPOLATION[] =
      "at_source_before_interpolation";
static const char AT_SOURCE_AFTER_INTERPOLATION[] =
      "at_source_after_interpolation";
static const char AT_TARGET[] = "at_target";
static const char ENFORCE_WRITE_RESTART[] = "enforce_write_restart";
static const char ENFORCE_WRITE_WEIGHT_FILE[] = "enforce_write_weight_file";
static const char TRUE_STR[] = "true";
static const char FALSE_STR[] = "false";
static const char ACCUMULATE_STR[] = "accumulate";
static const char AVERAGE_STR[] = "average";
static const char NONE_STR[] = "none";
static const char MILLISECOND_STR[] = "millisecond";
static const char SECOND_STR[] = "second";
static const char MINUTE_STR[] = "minute";
static const char HOUR_STR[] = "hour";
static const char DAY_STR[] = "day";
static const char MONTH_STR[] = "month";
static const char YEAR_STR[] = "year";
static const char ISO_FORMAT_STR[] = "ISO_format";
static const char PROLEPTIC_GREGORIAN_STR[] = "proleptic-gregorian";
static const char REGULAR_360D_STR[] = "360d";
static const char REGULAR_365D_STR[] = "365d";

// strings for error messages
static const char FILE_NOT_FOUND[] = "File %s could not be found.";
static const char FILE_NOT_PARSED[] = "Could not parse file %s";
static const char FILE_EMPTY[] = "File %s is empty.";
static const char XML_NOT_VALIDATING[] = "XML does not validate against schema.";
static const char XML_TREE_NOT_INITIALIZED[] = "XML tree not initialized.";
static const char SCHEMA_MALFORMED[] =
      "Schema cannot be loaded or is not well-formed";
static const char SCHEMA_NO_PARSER[] =
      "Unable to create parser context for schema";
static const char SCHEMA_INVALID[] = "Schema is not valid";
static const char SCHEMA_NO_VALIDATION[] =
      "Unable to create validation context for schema";
static const char STRING_NULL[] = "Passed string is NULL.";
static const char STRING_EMPTY[] = "Passed string is empty.";
static const char NULL_POINTER_COMPONENT[] = "Passed component pointer is NULL.";
static const char NULL_POINTER_TRANSIENT[] = "Passed transient pointer is NULL.";
static const char NULL_POINTER_GRID[] = "Passed grid pointer is NULL.";
static const char PARAMETER_NOT_CONTAINED[] =
      "Interpolation parameter no. %d is not contained in XML tree. \n";
static const char ATTRIBUTE_INDEX_OUT_OF_BOUNDS[] =
      "Invalid parameter: attribute index out of bounds.";
static const char ATTRIBUTE_NOT_FOUND[] = "Attribute called '%s' not found.";
static const char COMPONENT_NOT_CONTAINED[] =
      "Component no. %d is not contained in XML tree. \n";
static const char COMPONENT_INDEX_OUT_OF_BOUNDS[] =
      "Invalid parameter: component index out of bounds.";
static const char TRANSIENT_GRID_REF_NOT_CONTAINED[] =
      "transient_grid_ref no. %d is not contained in component no. %d. \n";
static const char TRANSIENT_GRID_REF_NOT_CONTAINED_IN_COMPONENT_WITH_ID[] =
      "transient_grid_ref no. %d is not contained in component with ID %d. \n";
static const char TRANSIENT_GRID_REF_INDEX_OUT_OF_BOUNDS[] =
      "Invalid parameter: transient_grid_ref index out of bounds.";
static const char TRANSIENT_NOT_CONTAINED[] =
      "Transient no. %d is not contained in XML tree. \n";
static const char TRANSIENT_INDEX_OUT_OF_BOUNDS[] =
      "Invalid parameter: transient index out of bounds.";
static const char GRID_NOT_CONTAINED[] =
      "Grid no. %d is not contained in XML tree. \n";
static const char GRID_INDEX_OUT_OF_BOUNDS[] =
      "Invalid parameter: grid index out of bounds.";
static const char COUPLE_NOT_CONTAINED[] =
      "Couple no. %d is not contained in XML tree. \n";
static const char COUPLE_INDEX_OUT_OF_BOUNDS[] =
      "Invalid parameter: couple index out of bounds.";
static const char TRANSIENT_COUPLE_NOT_CONTAINED[] =
      "Transient couple no. %d is not contained in couple no. %d in XML tree. \n";
static const char TRANSIENT_COUPLE_INDEX_OUT_OF_BOUNDS[] =
      "Invalid parameter: transient couple index out of bounds.";
static const char INTERPOLATION_NOT_CONTAINED[] =
      "Interpolation no. %d is not contained in XML tree. \n";
static const char INTERPOLATION_OUT_OF_BOUNDS[] =
      "Invalid parameter: interpolation index out of bound.";
static const char COMPONENT_ID_NOT_FOUND[] = "No component with ID %d found.";
static const char TRANSIENT_ID_NOT_FOUND[] = "No transient with ID %d found.";
static const char TRANSIENT_GRID_REF_ID_NOT_FOUND[] =
      "No transient_grid_ref element found with ID %d.";
static const char GRID_ID_NOT_FOUND[] = "No grid with ID %d found.";
static const char NUMERIC_VALUE_INVALID[] = "Value '%s' is not a numeric value.";
static const char WRONG_TIME_UNIT[] = "Unsupported time unit '%s'.";

//! holds the document tree
xmlDocPtr document;

//! holds the root element
xmlNodePtr coupling;

static int no_of_components = -1;
static int no_of_transients = -1;
static int no_of_grids = -1;
static int no_of_couples = -1;
static int * no_of_transient_couples = NULL;

//* help function declarations ************************************************/

static int file_exists(const char * filename);

static int is_valid(const xmlDocPtr doc, const char * schema_filename);

static void check_init(const int line);

static void check_string(const char * string, const int line);

static void check_component_pointer(const yac_xml_component * ptr, const int line);

static void check_transient_pointer(const yac_xml_transient * ptr, const int line);

static void check_grid_pointer(const yac_xml_grid * ptr, const int line);

static void check_component_parameter(const int component_index, const int line);

static void check_component_parameters(const int component_index,
      const int transient_grid_ref_index, const int line);

static void check_transient_grid_ref_index(const int component_id,
      const int transient_grid_ref_index, const int line);

static void check_transient_grid_ref_id(xmlNodePtr transient_grid_ref_node,
      const int id, const int line);

static void check_transient_parameter(const int transient_index, const int line);

static void check_grid_parameter(const int grid_index, const int line);

static void check_couple_parameter(const int couple_index, const int line);

static void check_couple_parameters(const int couple_index,
      const int transient_couple_index, const int line);

static void check_component_id(xmlNodePtr component_node, const int id,
      const int line);

static void check_transient_id(const int transient_id, const int line);

static void check_grid_id(const int grid_id, const int line);

static void check_numeric_value(const char * value, const int line);

static void error_message_invalid_component_reference_value(const int line);

static int get_number_of_component_elements(const xmlNodePtr coupling_node);

static int get_number_of_transient_elements(const xmlNodePtr coupling);

static int get_number_of_grid_elements(const xmlNodePtr coupling);

static int get_number_of_couple_elements(const xmlNodePtr coupling_node);

static xmlNodePtr get_child_element_by_id_attribute(
      const xmlNodePtr parent_node, const int id_value);

static xmlNodePtr get_child_element_by_attribute(const xmlNodePtr node,
      const char * attribute_name, const char * attribute_value);

static int get_transient_couple_attribute_value(const int couple_index,
      const int transient_couple_index, const char * attribute_name);

static xmlNodePtr get_component_node(const int component_index);

static xmlNodePtr get_couple_node(const int couple_index);

static xmlNodePtr get_transient_couple_node(const int couple_index,
      const int transient_couple_index);

static xmlNodePtr get_interpolation_node(const int couple_index,
      const int transient_couple_index, const int interpolation_index);

static xmlNodePtr nth_node(xmlNodePtr parent_node, int n);

static xmlAttrPtr nth_attribute(xmlNodePtr parent_node, int n);

static xmlNodePtr get_node_by_name(xmlNodePtr parent_node,
      const char * name);

static int get_number_of_elements_of_name(xmlNodePtr parent_node,
      const char * name);

static enum Yac_xml_time_operation operation_enum_from_string(
      const char * operation_str);

static enum Yac_xml_time_unit time_unit_enum_from_string(
    const char * time_unit_str);

static enum Yac_xml_calendar calendar_enum_from_string(
    const char * calendar_str);

static int is_numeric(const char * s);

//* init and destroy **********************************************************/

void yac_xml_init_config_xml(const char * xml_filename, const char * schema_filename) {
   document = NULL;
   coupling = NULL;

   // error if file does not exist
   if (!file_exists(xml_filename)) {
      char message[100];
      snprintf(message, sizeof(message), FILE_NOT_FOUND, xml_filename);
      yac_internal_abort_message(message, __FILE__, __LINE__);
   }

   // parse the XML file
   document = xmlReadFile(xml_filename, UTF8, 0);
   // error if parsing was not successful
   if (document == NULL) {
      char message[100];
      snprintf(message, sizeof(message), FILE_NOT_PARSED, xml_filename);
      yac_internal_abort_message(message, __FILE__, __LINE__);
   }

   // validate XML against schema
   if (!is_valid(document, schema_filename)) {
      char message[100];
      snprintf(message, sizeof(message), XML_NOT_VALIDATING);
      yac_internal_abort_message(message, __FILE__, __LINE__);
   }
#ifdef VERBOSE_XML
   else
   fprintf(stdout, "XML %s validates against schema.\n", xml_filename);
#endif

   // initialize types
   coupling = xmlDocGetRootElement(document);
   // error if file was empty
   if (coupling == NULL) {
      char message[100];
      snprintf(message, sizeof(message), FILE_EMPTY, xml_filename);
      yac_internal_abort_message(message, __FILE__, __LINE__);
   }

   // init values
   no_of_components = get_number_of_component_elements(coupling);
   no_of_transients = get_number_of_transient_elements(coupling);
   no_of_grids = get_number_of_grid_elements(coupling);
   no_of_couples = get_number_of_couple_elements(coupling);
   no_of_transient_couples = (int *) calloc(no_of_couples, sizeof(int));
#ifdef VERBOSE_XML
   fprintf(stdout, "XML parsed, contains %d components and %d couples. \n",
         no_of_components, no_of_couples);
#endif

   // destroy parser
   xmlCleanupParser();
}

void yac_xml_free_xml_string(const char * string) {
   xmlFree((void *) string);
}

void yac_xml_free_xml_component(yac_xml_component * component) {
   free(component);
}

void yac_xml_free_xml_components(yac_xml_component * components) {
   free(components);
}

void yac_xml_free_xml_transient(yac_xml_transient * transient) {
   xmlFree((void *) transient->name);
   free(transient);
}

void yac_xml_free_xml_transients(yac_xml_transient * transients, const int n) {
   int i;
   for (i = 0; i < n; i++) {
      xmlFree((void *) transients[i].name);
   }
   free(transients);
}

void yac_xml_free_xml_grid(yac_xml_grid * grid) {
   xmlFree((void *) grid->name);
   free(grid);
}

void yac_xml_free_xml_grids(yac_xml_grid * grids, const int n) {
   int i;
   for (i = 0; i < n; i++) {
      xmlFree((void *) grids[i].name);
   }
   free(grids);
}

void yac_xml_destroy_config_xml() {
   if (document != NULL) {
      xmlFreeDoc(document);
   }
   coupling = NULL;
   document = NULL;
   no_of_components = -1;
   no_of_transients = -1;
   no_of_couples = -1;
   free(no_of_transient_couples);
   no_of_transient_couples = NULL;
#ifdef VERBOSE_XML
   fprintf(stdout, "Successfully freed document tree.\n");
#endif
}

//* interface functions for \<components\> ************************************/

int yac_xml_get_number_of_components() {
   check_init(__LINE__);
   return no_of_components;
}

int yac_xml_redirect_stdout() {

  check_init(__LINE__);
  xmlNodePtr redirect_node = get_node_by_name(coupling, REDIRECT);

  int redirect = 0;

  if (redirect_node == NULL) return 0;

  char * value;

  value = (char *) xmlGetProp(redirect_node,
          (const xmlChar *) REDIRECT_STDOUT);

  redirect = !xmlStrcmp((const xmlChar *)value, (const xmlChar *) TRUE_STR);

  if (redirect) {

    xmlFree(value);

    value = (char *) xmlGetProp(redirect_node,
            (const xmlChar *) REDIRECT_OF_ROOT);

    // increment only if redirect of all processes is requested
    redirect += !xmlStrcmp((const xmlChar *)value, (const xmlChar *) FALSE_STR);
  }

  xmlFree(value);

  return redirect;
}

int yac_xml_component_name_exists_in_xml(const char * component_name) {
   check_init(__LINE__);

   // input string well-formed?
   if (component_name == NULL) {
      return 0;
   }
   if (strcmp(component_name, "") == 0) {
      return 0;
   }

   // get xml element with given component name
   xmlNodePtr components_node = get_node_by_name(coupling, COMPONENTS);
   xmlNodePtr children = components_node->children;
   while (children) {
      if (!xmlStrcmp(children->name, (const xmlChar *) COMPONENT)) {
         xmlNodePtr name_node = get_node_by_name(children, NAME);
         if (!xmlStrcmp(name_node->children->content,
               (const xmlChar *) component_name)) {

            // found
            return 1;
         }
      }
      children = children->next;
   }

   // not found
   return 0;
}

int yac_xml_component_id_exists_in_xml(const int component_id) {
   check_init(__LINE__);

   xmlNodePtr components_node = get_node_by_name(coupling, COMPONENTS);
   xmlNodePtr component_node = get_child_element_by_id_attribute(
         components_node, component_id);
   if (component_node != NULL) {
      return 1;
   }
   return 0;
}

int yac_xml_get_component_id(const int component_index) {
   check_component_parameter(component_index, __LINE__);

   xmlNodePtr component_node = get_component_node(component_index);
   char * id_value = (char *) xmlGetProp(component_node, (const xmlChar *) ID);

   check_numeric_value(id_value, __LINE__);
   int ret = atoi(id_value);
   xmlFree(id_value);
   return ret;
}

int yac_xml_get_component_id_by_component_name(const char * component_name) {
   check_string(component_name, __LINE__);

   xmlNodePtr components_node = get_node_by_name(coupling, COMPONENTS);
   xmlNodePtr children = components_node->children;
   char * id_value = NULL;
   while (children) {
      if (!xmlStrcmp(children->name, (const xmlChar *) COMPONENT)) {
         xmlNodePtr name_node = get_node_by_name(children, NAME);
         if (!xmlStrcmp(name_node->children->content,
               (const xmlChar *) component_name)) {
            id_value = (char *) xmlGetProp(children, (const xmlChar *) ID);
            break;
         }
      }
      children = children->next;
   }
   if (id_value == NULL) {
      return -1;
   }

   check_numeric_value(id_value, __LINE__);
   int ret = atoi(id_value);
   xmlFree(id_value);
   return ret;
}

void yac_xml_get_component(const int component_index, yac_xml_component * component) {
   check_component_parameter(component_index, __LINE__);
   check_component_pointer(component, __LINE__);

   xmlNodePtr component_node = get_component_node(component_index);
   xmlNodePtr name_node = get_node_by_name(component_node, NAME);
   char * id_value = (char *) xmlGetProp(component_node, (const xmlChar *) ID);

   check_numeric_value(id_value, __LINE__);
   component->name = (char *) name_node->children->content;
   component->global_id = atoi(id_value);

   xmlFree(id_value);
}

void yac_xml_get_component_by_component_id(const int component_id,
      yac_xml_component * component) {
   check_init(__LINE__);
   check_component_pointer(component, __LINE__);

   xmlNodePtr components_node = get_node_by_name(coupling, COMPONENTS);
   xmlNodePtr component_node = get_child_element_by_id_attribute(
         components_node, component_id);
   check_component_id(component_node, component_id, __LINE__);
   xmlNodePtr name_node = get_node_by_name(component_node, NAME);
   char * id_value = (char *) xmlGetProp(component_node, (const xmlChar *) ID);

   check_numeric_value(id_value, __LINE__);
   component->name = (char *) name_node->children->content;
   component->global_id = atoi(id_value);

   xmlFree(id_value);
}

void yac_xml_get_component_by_component_name(const char * component_name,
      yac_xml_component * component) {
   check_string(component_name, __LINE__);
   check_component_pointer(component, __LINE__);

   int id = yac_xml_get_component_id_by_component_name(component_name);
   yac_xml_get_component_by_component_id(id, component);
}

const char * yac_xml_get_component_name(const int component_index) {
   check_component_parameter(component_index, __LINE__);

   xmlNodePtr component_node = get_component_node(component_index);
   xmlNodePtr name_node = get_node_by_name(component_node, NAME);
   return (const char *) name_node->children->content;
}

const char * yac_xml_get_component_name_by_component_id(
      const int component_id) {
   check_init(__LINE__);

   xmlNodePtr components_node = get_node_by_name(coupling, COMPONENTS);
   xmlNodePtr component_node = get_child_element_by_id_attribute(
         components_node, component_id);
   check_component_id(component_node, component_id, __LINE__);
   xmlNodePtr name_node = get_node_by_name(component_node, NAME);
   return (const char *) name_node->children->content;
}

const char * yac_xml_get_component_model(const int component_index) {
   check_component_parameter(component_index, __LINE__);

   xmlNodePtr component_node = get_component_node(component_index);
   xmlNodePtr model_node = get_node_by_name(component_node, MODEL);
   return (const char *) model_node->children->content;
}

const char * yac_xml_get_component_model_by_component_id(
      const int component_id) {
   check_init(__LINE__);

   xmlNodePtr components_node = get_node_by_name(coupling, COMPONENTS);
   xmlNodePtr component_node = get_child_element_by_id_attribute(
         components_node, component_id);
   check_component_id(component_node, component_id, __LINE__);
   xmlNodePtr name_node = get_node_by_name(component_node, MODEL);
   return (const char *) name_node->children->content;
}

const char * yac_xml_get_component_simulated(const int component_index) {
   check_component_parameter(component_index, __LINE__);

   xmlNodePtr component_node = get_component_node(component_index);
   xmlNodePtr simulated_node = get_node_by_name(component_node, SIMULATED);
   return (const char *) simulated_node->children->content;
}

const char * yac_xml_get_component_simulated_by_component_id(
      const int component_id) {
   check_init(__LINE__);

   xmlNodePtr components_node = get_node_by_name(coupling, COMPONENTS);
   xmlNodePtr component_node = get_child_element_by_id_attribute(
         components_node, component_id);
   check_component_id(component_node, component_id, __LINE__);
   xmlNodePtr simulated_node = get_node_by_name(component_node, SIMULATED);
   return (const char *) simulated_node->children->content;
}

int yac_xml_get_number_of_transient_grid_refs(const int component_index) {
   check_component_parameter(component_index, __LINE__);

   xmlNodePtr component_node = get_component_node(component_index);
   xmlNodePtr transient_grid_refs_node = get_node_by_name(component_node,
         TRANSIENT_GRID_REFS);
   return get_number_of_elements_of_name(transient_grid_refs_node,
         TRANSIENT_GRID_REF);
}

int yac_xml_get_number_of_transient_grid_refs_by_component_id(
      const int component_id) {
   check_init(__LINE__);

   xmlNodePtr components_node = get_node_by_name(coupling, COMPONENTS);
   xmlNodePtr component_node = get_child_element_by_id_attribute(
         components_node, component_id);
   check_component_id(component_node, component_id, __LINE__);

   xmlNodePtr transient_grid_refs_node = get_node_by_name(component_node,
         TRANSIENT_GRID_REFS);
   return get_number_of_elements_of_name(transient_grid_refs_node,
         TRANSIENT_GRID_REF);
}

int yac_xml_get_transient_grid_ref_id(const int component_index,
      const int transient_grid_ref_index) {
   check_component_parameters(component_index, transient_grid_ref_index,
         __LINE__);

   xmlNodePtr component_node = get_component_node(component_index);
   xmlNodePtr transient_grid_refs_node = get_node_by_name(component_node,
         TRANSIENT_GRID_REFS);
   xmlNodePtr transient_grid_ref_node = nth_node(transient_grid_refs_node,
         transient_grid_ref_index);
   char * id_value = (char *) xmlGetProp(transient_grid_ref_node,
         (const xmlChar *) ID);
   check_numeric_value(id_value, __LINE__);
   int id = atoi(id_value);
   xmlFree(id_value);
   return id;
}

int yac_xml_get_transient_grid_ref_id_by_component_id(const int component_id,
      const int transient_grid_ref_index) {
   check_init(__LINE__);

   xmlNodePtr components_node = get_node_by_name(coupling, COMPONENTS);
   xmlNodePtr component_node = get_child_element_by_id_attribute(
         components_node, component_id);
   check_component_id(component_node, component_id, __LINE__);
   check_transient_grid_ref_index(component_id, transient_grid_ref_index,
         __LINE__);

   xmlNodePtr transient_grid_refs_node = get_node_by_name(component_node,
         TRANSIENT_GRID_REFS);
   xmlNodePtr transient_grid_ref_node = nth_node(transient_grid_refs_node,
         transient_grid_ref_index);
   char * id_value = (char *) xmlGetProp(transient_grid_ref_node,
         (const xmlChar *) ID);
   check_numeric_value(id_value, __LINE__);
   int id = atoi(id_value);
   xmlFree(id_value);
   return id;
}

int yac_xml_get_transient_id_by_transient_grid_ref_id(const int component_index,
      const int transient_grid_ref_id) {
   check_component_parameter(component_index, __LINE__);

   xmlNodePtr component_node = get_component_node(component_index);
   xmlNodePtr transient_grid_refs_node = get_node_by_name(component_node,
         TRANSIENT_GRID_REFS);
   xmlNodePtr transient_grid_ref_node =
         get_child_element_by_id_attribute(transient_grid_refs_node,
               transient_grid_ref_id);
   check_transient_grid_ref_id(transient_grid_ref_node, transient_grid_ref_id,
         __LINE__);

   char * id_value = (char *) xmlGetProp(transient_grid_ref_node,
         (const xmlChar *) TRANSIENT_REF);
   check_numeric_value(id_value, __LINE__);
   int id = atoi(id_value);
   xmlFree(id_value);
   return id;
}

int yac_xml_get_transient_id_by_transient_grid_ref_id_by_component_id(
      const int component_id, const int transient_grid_ref_id) {
   check_init(__LINE__);

   xmlNodePtr components_node = get_node_by_name(coupling, COMPONENTS);
   xmlNodePtr component_node = get_child_element_by_id_attribute(
         components_node, component_id);
   check_component_id(component_node, component_id, __LINE__);

   xmlNodePtr transient_grid_refs_node = get_node_by_name(component_node,
         TRANSIENT_GRID_REFS);
   xmlNodePtr transient_grid_ref_node = get_child_element_by_id_attribute(
         transient_grid_refs_node, transient_grid_ref_id);
   check_transient_grid_ref_id(transient_grid_ref_node, transient_grid_ref_id,
         __LINE__);

   char * id_value = (char *) xmlGetProp(transient_grid_ref_node,
         (const xmlChar *) TRANSIENT_REF);
   check_numeric_value(id_value, __LINE__);
   int id = atoi(id_value);
   xmlFree(id_value);
   return id;
}

int yac_xml_get_grid_id_by_transient_grid_ref_id(const int component_index,
      const int transient_grid_ref_id) {
   check_component_parameter(component_index, __LINE__);

   xmlNodePtr component_node = get_component_node(component_index);
   xmlNodePtr transient_grid_refs_node = get_node_by_name(component_node,
         TRANSIENT_GRID_REFS);
   xmlNodePtr transient_grid_ref_node = get_child_element_by_id_attribute(
         transient_grid_refs_node, transient_grid_ref_id);
   check_transient_grid_ref_id(transient_grid_ref_node, transient_grid_ref_id,
         __LINE__);

   char * id_value = (char *) xmlGetProp(transient_grid_ref_node,
         (const xmlChar *) GRID_REF);
   check_numeric_value(id_value, __LINE__);
   int id = atoi(id_value);
   xmlFree(id_value);
   return id;
}

int yac_xml_get_grid_id_by_transient_grid_ref_id_by_component_id(const int component_id,
      const int transient_grid_ref_id) {
   check_init(__LINE__);

   xmlNodePtr components_node = get_node_by_name(coupling, COMPONENTS);
   xmlNodePtr component_node = get_child_element_by_id_attribute(
         components_node, component_id);
   check_component_id(component_node, component_id, __LINE__);

   xmlNodePtr transient_grid_refs_node = get_node_by_name(component_node,
         TRANSIENT_GRID_REFS);
   xmlNodePtr transient_grid_ref_node = get_child_element_by_id_attribute(
         transient_grid_refs_node, transient_grid_ref_id);
   check_transient_grid_ref_id(transient_grid_ref_node, transient_grid_ref_id,
         __LINE__);

   char * id_value = (char *) xmlGetProp(transient_grid_ref_node,
         (const xmlChar *) GRID_REF);
   check_numeric_value(id_value, __LINE__);
   int id = atoi(id_value);
   xmlFree(id_value);
   return id;
}

//* interface functions for \<transients\> ************************************/

int yac_xml_get_number_of_transients() {
   check_init(__LINE__);
   return no_of_transients;
}

int yac_xml_get_transient_id_by_transient_standard_name(
      const char * transient_standard_name) {
   check_string(transient_standard_name, __LINE__);

   xmlNodePtr transients_node = get_node_by_name(coupling, TRANSIENTS);
   xmlNodePtr transient_node = get_child_element_by_attribute(transients_node,
         TRANSIENT_STANDARD_NAME, transient_standard_name);

   if (transient_node == NULL) {
      return -1;
   }

   char * id_value = (char *) xmlGetProp(transient_node, (const xmlChar *) ID);

   check_numeric_value(id_value, __LINE__);
   int ret = atoi(id_value);
   xmlFree(id_value);
   return ret;
}

int yac_xml_transient_id_exists_in_xml(const int transient_id) {
   check_init(__LINE__);

   xmlNodePtr transients_node = get_node_by_name(coupling, TRANSIENTS);
   xmlNodePtr transient_node = get_child_element_by_id_attribute(
         transients_node, transient_id);

   if (transient_node != NULL) {
      return 1; // found
   }
   return 0; // not found
}

const char * yac_xml_get_transient_standard_name_by_transient_id(
      const int transient_id) {
   check_init(__LINE__);
   check_transient_id(transient_id, __LINE__);

   xmlNodePtr transients_node = get_node_by_name(coupling, TRANSIENTS);
   xmlNodePtr transient_node = get_child_element_by_id_attribute(
         transients_node, transient_id);
   return (const char *) xmlGetProp(transient_node,
         (const xmlChar *) TRANSIENT_STANDARD_NAME);
}

int yac_xml_transient_standard_name_exists_in_xml(
      const char * transient_standard_name) {
   check_init(__LINE__);

   xmlNodePtr transients_node = get_node_by_name(coupling, TRANSIENTS);
   xmlNodePtr transient_node = get_child_element_by_attribute(transients_node,
         TRANSIENT_STANDARD_NAME, transient_standard_name);

   if (transient_node != NULL) {
      return 1; // found
   }
   return 0; // not found
}

void yac_xml_get_transient(const int transient_index,
      yac_xml_transient * transient) {
   check_transient_parameter(transient_index, __LINE__);
   check_transient_pointer(transient, __LINE__);

   xmlNodePtr transients_node = get_node_by_name(coupling, TRANSIENTS);
   xmlNodePtr transient_node = nth_node(transients_node, transient_index);
   transient->name = (char *) xmlGetProp(transient_node,
         (const xmlChar *) TRANSIENT_STANDARD_NAME);
   char * id_value = (char *) xmlGetProp(transient_node, (const xmlChar *) ID);

   check_numeric_value(id_value, __LINE__);
   transient->global_id = atoi(id_value);
   xmlFree(id_value);
}

void yac_xml_get_transient_by_id(const int transient_id,
      yac_xml_transient * transient) {
   check_init(__LINE__);
   check_transient_pointer(transient, __LINE__);

   xmlNodePtr transients_node = get_node_by_name(coupling, TRANSIENTS);
   xmlNodePtr transient_node = get_child_element_by_id_attribute(
         transients_node, transient_id);
   if (transient_node != NULL) {
      transient->name = (char *) xmlGetProp(transient_node,
            (const xmlChar *) TRANSIENT_STANDARD_NAME);
      transient->global_id = transient_id;
      return;
   }

   char message[100];
   snprintf(message, sizeof(message), TRANSIENT_ID_NOT_FOUND, transient_id);
   yac_internal_abort_message(message, __FILE__, __LINE__);
}

void yac_xml_get_all_transients(yac_xml_transient transients[]) {
   check_init(__LINE__);
   check_transient_pointer(transients, __LINE__);

   xmlNodePtr transients_node = get_node_by_name(coupling, TRANSIENTS);
   xmlNodePtr child = transients_node->children;

   int i = 0;
   while (child) {
      if (child->type != XML_ELEMENT_NODE
            || xmlStrcmp(child->name, (const xmlChar *) TRANSIENT)) {
         child = child->next;
         continue;
      }
      transients[i].name = (char *) xmlGetProp(child,
            (const xmlChar *) TRANSIENT_STANDARD_NAME);
      char * id_value = (char *) xmlGetProp(child, (const xmlChar *) ID);

      check_numeric_value(id_value, __LINE__);
      transients[i].global_id = atoi(id_value);
      xmlFree(id_value);

      child = child->next;
      i++;
   }
}

//* interface functions for \<grids\> *****************************************/

int yac_xml_get_number_of_grids() {
   check_init(__LINE__);
   return no_of_grids;
}

int yac_xml_grid_id_exists_in_xml(const int grid_id) {
   check_init(__LINE__);

   xmlNodePtr grids_node = get_node_by_name(coupling, GRIDS);
   xmlNodePtr grid_node = get_child_element_by_id_attribute(grids_node,
         grid_id);

   if (grid_node != NULL) {
      return 1; // found
   }
   return 0; // not found
}

const char * yac_xml_get_grid_alias_name_by_grid_id(const int grid_id) {
   check_grid_id(grid_id, __LINE__);

   xmlNodePtr grids_node = get_node_by_name(coupling, GRIDS);
   xmlNodePtr grid_node = get_child_element_by_id_attribute(grids_node,
         grid_id);
   return (const char *) xmlGetProp(grid_node, (const xmlChar *) ALIAS_NAME);
}

void yac_xml_get_grid(const int grid_index, yac_xml_grid * grid) {
   check_grid_parameter(grid_index, __LINE__);
   check_grid_pointer(grid, __LINE__);

   xmlNodePtr grids_node = get_node_by_name(coupling, GRIDS);
   xmlNodePtr grid_node = nth_node(grids_node, grid_index);
   grid->name = (char *) xmlGetProp(grid_node, (const xmlChar *) ALIAS_NAME);
   char * id_value = (char *) xmlGetProp(grid_node, (const xmlChar *) ID);
   check_numeric_value(id_value, __LINE__);
   grid->global_id = atoi(id_value);
   xmlFree(id_value);
}

void yac_xml_get_grid_by_id(const int grid_id, yac_xml_grid * grid) {
   check_grid_id(grid_id, __LINE__);
   check_grid_pointer(grid, __LINE__);

   xmlNodePtr grids_node = get_node_by_name(coupling, GRIDS);
   xmlNodePtr grid_node = get_child_element_by_id_attribute(grids_node,
         grid_id);
   grid->name = (char *) xmlGetProp(grid_node, (const xmlChar *) ALIAS_NAME);
   grid->global_id = grid_id;
}

void yac_xml_get_all_grids(yac_xml_grid grids[]) {
   check_init(__LINE__);
   check_grid_pointer(grids, __LINE__);

   xmlNodePtr grids_node = get_node_by_name(coupling, GRIDS);
   xmlNodePtr child = grids_node->children;
   int i = 0;
   while (child) {
      if (child->type != XML_ELEMENT_NODE
            || xmlStrcmp(child->name, (const xmlChar *) GRID)) {
         child = child->next;
         continue;
      }

      grids[i].name = (char *) xmlGetProp(child, (const xmlChar *) ALIAS_NAME);
      char * id_value = (char *) xmlGetProp(child, (const xmlChar *) ID);
      check_numeric_value(id_value, __LINE__);
      grids[i].global_id = atoi(id_value);
      xmlFree(id_value);

      child = child->next;
      i++;
   }
}

//* interface functions for \<dates\> *****************************************/

const char * yac_xml_get_start_date() {
   check_init(__LINE__);
   xmlNodePtr dates_node = get_node_by_name(coupling, DATES);
   xmlNodePtr start_date_node = get_node_by_name(dates_node, START_DATE);
   char * start_date = (char *) start_date_node->children->content;
   return start_date;
}

const char * yac_xml_get_end_date() {
   check_init(__LINE__);
   xmlNodePtr dates_node = get_node_by_name(coupling, DATES);
   xmlNodePtr end_date_node = get_node_by_name(dates_node, END_DATE);
   char * end_date = (char *) end_date_node->children->content;
   return end_date;
}

yac_xml_calendar yac_xml_get_calendar() {
  check_init(__LINE__);
  xmlNodePtr dates_node = get_node_by_name(coupling, DATES);
  xmlNodePtr calendar_node = get_node_by_name(dates_node, CALENDAR);
  char * calendar_str = (char *) calendar_node->children->content;
  yac_xml_calendar calendar_value = calendar_enum_from_string(calendar_str);
  return calendar_value;
}

//* interface functions for \<timestep_unit\> *********************************/

yac_xml_timestep_unit yac_xml_get_timestep_unit() {
   check_init(__LINE__);
   xmlNodePtr timestep_unit_node = get_node_by_name(coupling, TIMESTEP_UNIT);
   char * timestep_unit_str = (char *) timestep_unit_node->children->content;
   yac_xml_timestep_unit timestep_unit_value = time_unit_enum_from_string(
         timestep_unit_str);
   return timestep_unit_value;
}

//* interface functions for \<couples\> ***************************************/

int yac_xml_get_number_of_couples() {
   check_init(__LINE__);
   return no_of_couples;
}

int yac_xml_get_first_component_id(const int couple_index) {
   check_couple_parameter(couple_index, __LINE__);

   xmlNodePtr couple_node = get_couple_node(couple_index);
   xmlNodePtr component1_node = get_node_by_name(couple_node, COMPONENT1);
   char * attribute_value = (char *) xmlGetProp(component1_node,
         (const xmlChar *) COMPONENT_ID);

   check_numeric_value(attribute_value, __LINE__);
   int ret = atoi(attribute_value);
   xmlFree(attribute_value);
   return ret;
}

int yac_xml_get_second_component_id(const int couple_index) {
   check_couple_parameter(couple_index, __LINE__);

   xmlNodePtr couple_node = get_couple_node(couple_index);
   xmlNodePtr component2_node = get_node_by_name(couple_node, COMPONENT2);
   char * attribute_value = (char *) xmlGetProp(component2_node,
         (const xmlChar *) COMPONENT_ID);

   check_numeric_value(attribute_value, __LINE__);
   int ret = atoi(attribute_value);
   xmlFree(attribute_value);
   return ret;
}

int yac_xml_get_number_of_transient_couples(const int couple_index) {
   check_couple_parameter(couple_index, __LINE__);

   xmlNodePtr couple_node = get_couple_node(couple_index);
   int transient_couples = get_number_of_elements_of_name(couple_node,
         TRANSIENT_COUPLE);
   no_of_transient_couples[couple_index] = transient_couples;
   return transient_couples;
}

int yac_xml_get_transient_couple_transient_id(const int couple_index,
      const int transient_couple_index) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);
   return get_transient_couple_attribute_value(couple_index,
         transient_couple_index, TRANSIENT_ID);
}

void yac_xml_get_transient_couple_transient(const int couple_index,
      const int transient_couple_index, yac_xml_transient * transient) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);

   int transient_id = get_transient_couple_attribute_value(couple_index,
         transient_couple_index, TRANSIENT_ID);
   yac_xml_get_transient_by_id(transient_id, transient);
}

int yac_xml_get_source_component_id(const int couple_index,
      const int transient_couple_index) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);

   xmlNodePtr transient_couple_node = get_transient_couple_node(couple_index,
         transient_couple_index);
   xmlNodePtr source_node = get_node_by_name(transient_couple_node, SOURCE);
   char * attribute_value = (char *) xmlGetProp(source_node,
         (const xmlChar *) COMPONENT_REF);

   check_numeric_value(attribute_value, __LINE__);
   int source_local_id = atoi(attribute_value);
   xmlFree(attribute_value);
   int global_component_id;
   switch (source_local_id) {
   case (1):
      global_component_id = yac_xml_get_first_component_id(couple_index);
      break;
   case (2):
      global_component_id = yac_xml_get_second_component_id(couple_index);
      break;
   default:
      error_message_invalid_component_reference_value(__LINE__);
      global_component_id = -1;
      break;
   }
   return global_component_id;
}

int yac_xml_get_target_component_id(const int couple_index,
      const int transient_couple_index) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);

   xmlNodePtr transient_couple_node = get_transient_couple_node(couple_index,
         transient_couple_index);
   xmlNodePtr source_node = get_node_by_name(transient_couple_node, SOURCE);
   char * attribute_value = (char *) xmlGetProp(source_node,
         (const xmlChar *) COMPONENT_REF);

   check_numeric_value(attribute_value, __LINE__);
   int target_local_id = atoi(attribute_value);
   xmlFree(attribute_value);
   int global_component_id;
   switch (target_local_id) {
   case (1):
      global_component_id = yac_xml_get_second_component_id(couple_index);
      break;
   case (2):
      global_component_id = yac_xml_get_first_component_id(couple_index);
      break;
   default:
      error_message_invalid_component_reference_value(__LINE__);
      global_component_id = -1;
      break;
   }
   return global_component_id;
}

int yac_xml_get_source_transient_grid_ref_id(const int couple_index,
      const int transient_couple_index) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);

   xmlNodePtr transient_couple_node = get_transient_couple_node(couple_index,
         transient_couple_index);
   xmlNodePtr source_node = get_node_by_name(transient_couple_node, SOURCE);
   char * attribute_value = (char *) xmlGetProp(source_node,
         (const xmlChar *) TRANSIENT_GRID_REF);
   check_numeric_value(attribute_value, __LINE__);
   int id = atoi(attribute_value);
   xmlFree(attribute_value);
   return  id;
}

int yac_xml_get_target_transient_grid_ref_id(const int couple_index,
      const int transient_couple_index) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);

   xmlNodePtr transient_couple_node = get_transient_couple_node(couple_index,
         transient_couple_index);
   xmlNodePtr source_node = get_node_by_name(transient_couple_node, TARGET);
   char * attribute_value = (char *) xmlGetProp(source_node,
         (const xmlChar *) TRANSIENT_GRID_REF);
   check_numeric_value(attribute_value, __LINE__);
   int id = atoi(attribute_value);
   xmlFree(attribute_value);
   return id;
}

char * yac_xml_get_source_isotimestep(const int couple_index,
      const int transient_couple_index) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);

   xmlNodePtr transient_couple_node = get_transient_couple_node(couple_index,
         transient_couple_index);
   xmlNodePtr timestep_node = get_node_by_name(transient_couple_node, TIMESTEP);
   xmlNodePtr source_node = get_node_by_name(timestep_node, SOURCE);
   char * value = (char *) source_node->children->content;
   return value;
}

char * yac_xml_get_target_isotimestep(const int couple_index,
      const int transient_couple_index) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);

   xmlNodePtr transient_couple_node = get_transient_couple_node(couple_index,
         transient_couple_index);
   xmlNodePtr timestep_node = get_node_by_name(transient_couple_node, TIMESTEP);
   xmlNodePtr target_node = get_node_by_name(timestep_node, TARGET);
   char * value = (char *) target_node->children->content;
   return value;
}

char * yac_xml_get_isocoupling_period(const int couple_index,
      const int transient_couple_index) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);

   xmlNodePtr transient_couple_node = get_transient_couple_node(couple_index,
         transient_couple_index);
   xmlNodePtr timestep_node = get_node_by_name(transient_couple_node, TIMESTEP);
   xmlNodePtr coupling_period_node = get_node_by_name(timestep_node,
         COUPLING_PERIOD);
   char * value = (char *) coupling_period_node->children->content;
   return value;
}

int yac_xml_get_source_timestep(const int couple_index,
      const int transient_couple_index) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);

   xmlNodePtr transient_couple_node = get_transient_couple_node(couple_index,
         transient_couple_index);
   xmlNodePtr timestep_node = get_node_by_name(transient_couple_node, TIMESTEP);
   xmlNodePtr source_node = get_node_by_name(timestep_node, SOURCE);
   char * value = (char *) source_node->children->content;
   check_numeric_value(value, __LINE__);
   return atoi(value);
}

int yac_xml_get_target_timestep(const int couple_index,
      const int transient_couple_index) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);

   xmlNodePtr transient_couple_node = get_transient_couple_node(couple_index,
         transient_couple_index);
   xmlNodePtr timestep_node = get_node_by_name(transient_couple_node, TIMESTEP);
   xmlNodePtr target_node = get_node_by_name(timestep_node, TARGET);
   char * value = (char *) target_node->children->content;
   check_numeric_value(value, __LINE__);
   return atoi(value);
}

int yac_xml_get_coupling_period(const int couple_index,
      const int transient_couple_index) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);

   xmlNodePtr transient_couple_node = get_transient_couple_node(couple_index,
         transient_couple_index);
   xmlNodePtr timestep_node = get_node_by_name(transient_couple_node, TIMESTEP);
   xmlNodePtr coupling_period_node = get_node_by_name(timestep_node,
         COUPLING_PERIOD);
   char * value = (char *) coupling_period_node->children->content;
   check_numeric_value(value, __LINE__);
   return atoi(value);
}

yac_xml_operation yac_xml_get_coupling_period_operation(const int couple_index,
      const int transient_couple_index) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);

   xmlNodePtr transient_couple_node = get_transient_couple_node(couple_index,
         transient_couple_index);
   xmlNodePtr timestep_node = get_node_by_name(transient_couple_node, TIMESTEP);
   xmlNodePtr coupling_period_node = get_node_by_name(timestep_node,
         COUPLING_PERIOD);
   char * value = (char *) xmlGetProp(coupling_period_node,
         (const xmlChar *) OPERATION);
   yac_xml_operation ret = operation_enum_from_string(value);
   xmlFree(value);
   return ret;
}

int yac_xml_get_source_timelag(const int couple_index,
      const int transient_couple_index) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);

   xmlNodePtr transient_couple_node = get_transient_couple_node(couple_index,
         transient_couple_index);
   xmlNodePtr timestep_node = get_node_by_name(transient_couple_node, TIMESTEP);
   xmlNodePtr lag_node = get_node_by_name(timestep_node, SOURCE_TIMELAG);
   char * value = (char *) lag_node->children->content;

   check_numeric_value(value, __LINE__);
   return atoi(value);
}

int yac_xml_get_target_timelag(const int couple_index,
      const int transient_couple_index) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);

   xmlNodePtr transient_couple_node = get_transient_couple_node(couple_index,
         transient_couple_index);
   xmlNodePtr timestep_node = get_node_by_name(transient_couple_node, TIMESTEP);
   xmlNodePtr lag_node = get_node_by_name(timestep_node, TARGET_TIMELAG);
   char * value = (char *) lag_node->children->content;

   check_numeric_value(value, __LINE__);
   return atoi(value);
}

int yac_xml_get_number_of_interpolations(const int couple_index,
      const int transient_couple_index) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);

   xmlNodePtr transient_couple_node = get_transient_couple_node(couple_index,
         transient_couple_index);
   xmlNodePtr interpolation_requirements_node = get_node_by_name(
         transient_couple_node, INTERPOLATION_REQUIREMENTS);
   int no = get_number_of_elements_of_name(interpolation_requirements_node,
         INTERPOLATION);
   return no;
}

int yac_xml_compare_interpolation_lists(const int couple_index_1,
      const int transient_couple_index_1, const int couple_index_2,
      const int transient_couple_index_2) {
   check_couple_parameters(couple_index_1, transient_couple_index_1, __LINE__);
   check_couple_parameters(couple_index_2, transient_couple_index_2, __LINE__);

   // first list
   xmlNodePtr transient_couple_node_1 = get_transient_couple_node(
         couple_index_1, transient_couple_index_1);
   xmlNodePtr interpolation_requirements_node_1 = get_node_by_name(
         transient_couple_node_1, INTERPOLATION_REQUIREMENTS);

   // second list
   xmlNodePtr transient_couple_node_2 = get_transient_couple_node(
         couple_index_2, transient_couple_index_2);
   xmlNodePtr interpolation_requirements_node_2 = get_node_by_name(
         transient_couple_node_2, INTERPOLATION_REQUIREMENTS);

   int no_1 = get_number_of_elements_of_name(interpolation_requirements_node_1,
         INTERPOLATION);
   int no_2 = get_number_of_elements_of_name(interpolation_requirements_node_2,
         INTERPOLATION);
   if (no_1 != no_2) {
      return 0;
   }

   // compare all interpolations in list
   int i;
   for (i = 0; i < no_1; i++) {
      enum yac_interpolation_list interpolation_method_1 =
            yac_xml_get_interpolation_method(couple_index_1,
                  transient_couple_index_1, i);
      enum yac_interpolation_list interpolation_method_2 =
            yac_xml_get_interpolation_method(couple_index_2,
                  transient_couple_index_2, i);
      if (interpolation_method_1 != interpolation_method_2) {
         return 0;
      }

      int parameters_1 = yac_xml_get_interpolation_number_of_parameters(
            couple_index_1, transient_couple_index_1, i);
      int parameters_2 = yac_xml_get_interpolation_number_of_parameters(
            couple_index_2, transient_couple_index_2, i);
      if (parameters_1 != parameters_2) {
         return 0;
      }

      // compare all parameters in interpolation
      int j;
      for (j = 0; j < parameters_1; j++) {
         // check if parameter names are different
         const char * name_1 = yac_xml_get_interpolation_parameter_name(
               couple_index_1, transient_couple_index_1, i, j);
         const char * name_2 = yac_xml_get_interpolation_parameter_name(
               couple_index_2, transient_couple_index_2, i, j);
         if (strcmp(name_1, name_2)) {
            return 0;
         }

         // check if parameter values are different
         const char * value_1 =
               yac_xml_get_interpolation_parameter_value_string(couple_index_1,
                     transient_couple_index_1, i, name_1);
         const char * value_2 =
               yac_xml_get_interpolation_parameter_value_string(couple_index_2,
                     transient_couple_index_2, i, name_2);
         int result = strcmp(value_1, value_2);
         xmlFree((void*)value_1);
         xmlFree((void*)value_2);
         if (result) {
            return 0;
         }
      }

   }

   return 1;
}

enum yac_interpolation_list yac_xml_get_interpolation_method(const int couple_index,
      const int transient_couple_index, const int interpolation_index) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);

   xmlNodePtr interpolation_node = get_interpolation_node(couple_index,
         transient_couple_index, interpolation_index);
   char * method_value = (char *) xmlGetProp(interpolation_node,
         (const xmlChar *) METHOD);
   char * interpolation_method = yac_verified_interpolation(method_value);
   xmlFree(method_value);
   return yac_interpolation_enum_from_string(interpolation_method);
}

int yac_xml_interpolation_has_parameters(const int couple_index,
      const int transient_couple_index, const int interpolation_index) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);

   xmlNodePtr interpolation_node = get_interpolation_node(couple_index,
         transient_couple_index, interpolation_index);
   xmlAttrPtr attr;
   for (attr = interpolation_node->properties; attr != NULL; attr =
         attr->next) {
      if (xmlStrcmp(attr->name, (const xmlChar *) METHOD)) {
         return 1; // more than just "method" attribute
      }
   }
   return 0; // no other attribute
}

int yac_xml_interpolation_has_parameter(const int couple_index,
      const int transient_couple_index, const int interpolation_index,
      const char * parameter) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);

   xmlNodePtr interpolation_node = get_interpolation_node(couple_index,
         transient_couple_index, interpolation_index);
   xmlChar * attr = xmlGetProp(interpolation_node, (const xmlChar *) parameter);
   if (attr != NULL) {
      xmlFree(attr);
      return 1; // has parameter
   }
   return 0; // does not have parameter
}

int yac_xml_get_interpolation_number_of_parameters(const int couple_index,
      const int transient_couple_index, const int interpolation_index) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);

   xmlNodePtr interpolation_node = get_interpolation_node(couple_index,
         transient_couple_index, interpolation_index);
   xmlAttrPtr attr;
   int number = 0;
   for (attr = interpolation_node->properties; attr != NULL; attr =
         attr->next) {
      // if attribute different from "method"
      if (xmlStrcmp(attr->name, (const xmlChar *) METHOD)) {
         number = number + 1;
      }
   }

   return number;
}

const char * yac_xml_get_interpolation_parameter_name(const int couple_index,
      const int transient_couple_index, const int interpolation_index,
      const int parameter_index) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);

   xmlNodePtr interpolation_node = get_interpolation_node(couple_index,
         transient_couple_index, interpolation_index);
   // parameter_index + 1, because we don't want to count attribute 'method'
   xmlAttrPtr attr = nth_attribute(interpolation_node, parameter_index + 1);
   if (attr == NULL) {
      char message[100];
      snprintf(message, sizeof(message), PARAMETER_NOT_CONTAINED,
            parameter_index);
      strcat(message, ATTRIBUTE_INDEX_OUT_OF_BOUNDS);
      yac_internal_abort_message(message, __FILE__, __LINE__);
   }

   const xmlChar * attr_name = attr->name;
   return (const char *) attr_name;
}

const char * yac_xml_get_interpolation_parameter_value_string(
      const int couple_index, const int transient_couple_index,
      const int interpolation_index, const char * parameter) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);

   xmlNodePtr interpolation_node = get_interpolation_node(couple_index,
         transient_couple_index, interpolation_index);
   xmlChar * attr = xmlGetProp(interpolation_node, (const xmlChar *) parameter);
   if (attr == NULL) {
      char message[100];
      snprintf(message, sizeof(message), ATTRIBUTE_NOT_FOUND, parameter);
      yac_internal_abort_message(message, __FILE__, __LINE__);
   }

   return (const char *) attr;
}

int yac_xml_get_interpolation_parameter_value_int(const int couple_index,
      const int transient_couple_index, const int interpolation_index,
      const char * parameter) {
   const char * value = yac_xml_get_interpolation_parameter_value_string(
         couple_index, transient_couple_index, interpolation_index, parameter);
   check_numeric_value(value, __LINE__);
   int result = atoi(value);
   xmlFree((void*)value);
   return result;
}

double yac_xml_get_interpolation_parameter_value_double(
      const int couple_index, const int transient_couple_index,
      const int interpolation_index, const char * parameter) {
   const char * value = yac_xml_get_interpolation_parameter_value_string(
         couple_index, transient_couple_index, interpolation_index, parameter);
   check_numeric_value(value, __LINE__);
   double result = atof(value);
   xmlFree((void*)value);
   return result;
}

int yac_xml_use_source_mask_is_set(const int couple_index,
                                   const int transient_couple_index) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);

   xmlNodePtr transient_couple_node = get_transient_couple_node(couple_index,
         transient_couple_index);

   xmlNodePtr interpolation_requirements_node = get_node_by_name(
         transient_couple_node, INTERPOLATION_REQUIREMENTS);

   char * value = (char *) xmlGetProp(interpolation_requirements_node,
            (const xmlChar *) USE_SOURCE_MASK);

   int use_source_mask_is_set =
       ( !xmlStrcmp((const xmlChar *)value, (const xmlChar *) TRUE_STR) ) ||
       ( !xmlStrcmp((const xmlChar *)value, (const xmlChar *) FALSE_STR) );

   xmlFree(value);
   return use_source_mask_is_set;
}

int yac_xml_get_use_source_mask(const int couple_index,
                                const int transient_couple_index) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);

   xmlNodePtr transient_couple_node = get_transient_couple_node(couple_index,
         transient_couple_index);

   xmlNodePtr interpolation_requirements_node = get_node_by_name(
         transient_couple_node, INTERPOLATION_REQUIREMENTS);

   char * value = (char *) xmlGetProp(interpolation_requirements_node,
            (const xmlChar *) USE_SOURCE_MASK);

   int retval = !xmlStrcmp((const xmlChar *)value, (const xmlChar *) TRUE_STR);

   xmlFree(value);
   return retval;
}

int yac_xml_use_target_mask_is_set(const int couple_index,
                                   const int transient_couple_index) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);

   xmlNodePtr transient_couple_node = get_transient_couple_node(couple_index,
         transient_couple_index);

   xmlNodePtr interpolation_requirements_node = get_node_by_name(
         transient_couple_node, INTERPOLATION_REQUIREMENTS);

   char * value = (char *) xmlGetProp(interpolation_requirements_node,
            (const xmlChar *) USE_TARGET_MASK);

   int use_target_mask_is_set =
       ( !xmlStrcmp((const xmlChar *)value, (const xmlChar *) TRUE_STR) ) ||
       ( !xmlStrcmp((const xmlChar *)value, (const xmlChar *) FALSE_STR) );

   xmlFree(value);
   return use_target_mask_is_set;
}

int yac_xml_get_use_target_mask(const int couple_index,
                                const int transient_couple_index) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);

   xmlNodePtr transient_couple_node = get_transient_couple_node(couple_index,
         transient_couple_index);

   xmlNodePtr interpolation_requirements_node = get_node_by_name(
         transient_couple_node, INTERPOLATION_REQUIREMENTS);

   char * value = (char *) xmlGetProp(interpolation_requirements_node,
            (const xmlChar *) USE_TARGET_MASK);

   int retval = !xmlStrcmp((const xmlChar *)value, (const xmlChar *) TRUE_STR);

   xmlFree(value);
   return retval;
}

int yac_xml_mapping_on_source(const int couple_index,
      const int transient_couple_index) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);

   xmlNodePtr transient_couple_node = get_transient_couple_node(couple_index,
         transient_couple_index);
   xmlNodePtr mapping_on_source_node = get_node_by_name(
         transient_couple_node, MAPPING_ON_SOURCE);

   if ((mapping_on_source_node == NULL) ||
       !xmlStrcmp(mapping_on_source_node->children->content,
                  (const xmlChar *) TRUE_STR)) {
      return 1; // true
   }
   return 0; // false
}

int yac_xml_debug_mode_at_source_before_interpolation(const int couple_index,
      const int transient_couple_index) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);

   xmlNodePtr transient_couple_node = get_transient_couple_node(couple_index,
         transient_couple_index);
   xmlNodePtr debug_mode_node = get_node_by_name(transient_couple_node,
         DEBUG_MODE);
   xmlChar * value = xmlGetProp(debug_mode_node,
         (const xmlChar *) AT_SOURCE_BEFORE_INTERPOLATION);
   if (!xmlStrcmp(value, (const xmlChar *) TRUE_STR)) {
      xmlFree(value);
      return 1; // true
   }
   xmlFree(value);
   return 0; // false
}

int yac_xml_debug_mode_at_source_after_interpolation(const int couple_index,
      const int transient_couple_index) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);

   xmlNodePtr transient_couple_node = get_transient_couple_node(couple_index,
         transient_couple_index);
   xmlNodePtr debug_mode_node = get_node_by_name(transient_couple_node,
         DEBUG_MODE);
   xmlChar * value = xmlGetProp(debug_mode_node,
         (const xmlChar *) AT_SOURCE_AFTER_INTERPOLATION);
   if (!xmlStrcmp(value, (const xmlChar *) TRUE_STR)) {
      xmlFree(value);
      return 1; // true
   }
   xmlFree(value);
   return 0; // false
}

int yac_xml_debug_mode_at_target(const int couple_index,
      const int transient_couple_index) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);

   xmlNodePtr transient_couple_node = get_transient_couple_node(couple_index,
         transient_couple_index);
   xmlNodePtr debug_mode_node = get_node_by_name(transient_couple_node,
         DEBUG_MODE);
   xmlChar * value = xmlGetProp(debug_mode_node, (const xmlChar *) AT_TARGET);
   if (!xmlStrcmp(value, (const xmlChar *) TRUE_STR)) {
      xmlFree(value);
      return 1; // true
   }
   xmlFree(value);
   return 0; // false
}

int yac_xml_enforce_write_restart(const int couple_index,
      const int transient_couple_index) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);

   xmlNodePtr transient_couple_node = get_transient_couple_node(couple_index,
         transient_couple_index);
   xmlNodePtr enforce_write_restart_node = get_node_by_name(
         transient_couple_node, ENFORCE_WRITE_RESTART);

   if (!xmlStrcmp(enforce_write_restart_node->children->content,
         (const xmlChar *) TRUE_STR)) {
      return 1; // true
   }
   return 0; // false
}

int yac_xml_enforce_write_weight_file(const int couple_index,
      const int transient_couple_index) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);

   xmlNodePtr transient_couple_node = get_transient_couple_node(couple_index,
         transient_couple_index);
   xmlNodePtr enforce_write_weight_file_node = get_node_by_name(
         transient_couple_node, ENFORCE_WRITE_WEIGHT_FILE);

   if (!xmlStrcmp(enforce_write_weight_file_node->children->content,
         (const xmlChar *) TRUE_STR)) {
      return 1; // true
   }
   return 0; // false
}

const char *  yac_xml_get_weight_file_name(const int couple_index,
      const int transient_couple_index) {
   check_couple_parameters(couple_index, transient_couple_index, __LINE__);

   xmlNodePtr transient_couple_node = get_transient_couple_node(couple_index,
         transient_couple_index);
   xmlNodePtr enforce_write_weight_file_node = get_node_by_name(
         transient_couple_node, ENFORCE_WRITE_WEIGHT_FILE);
   xmlChar * attr = xmlGetProp(enforce_write_weight_file_node, (const xmlChar *) "filename");
   if (attr == NULL) {
      char message[100];
      snprintf(message, sizeof(message), ATTRIBUTE_NOT_FOUND, "filename");
      yac_internal_abort_message(message, __FILE__, __LINE__);
   }

   return (const char *) attr;
}

int yac_xml_get_transient_collection_size(const int component_id, const int transient_grid_ref_id) {

  check_init(__LINE__);

  xmlNodePtr components_node = get_node_by_name(coupling, COMPONENTS);
  xmlNodePtr component_node = get_child_element_by_id_attribute(
      components_node, component_id);
  check_component_id(component_node, component_id, __LINE__);

  xmlNodePtr transient_grid_refs_node = get_node_by_name(component_node,
      TRANSIENT_GRID_REFS);
  xmlNodePtr transient_grid_ref_node = get_child_element_by_id_attribute(
      transient_grid_refs_node, transient_grid_ref_id);
  check_transient_grid_ref_id(transient_grid_ref_node, transient_grid_ref_id,
      __LINE__);

  char * size_value = (char *) xmlGetProp(transient_grid_ref_node,
           (const xmlChar *) COLLECTION_SIZE);
  check_numeric_value(size_value, __LINE__);
  int collection_size = atoi(size_value);
  xmlFree(size_value);
  return collection_size;
}


//* help function implementations *********************************************/

static int file_exists(const char * filename) {
   FILE * file;
   if ((file = fopen(filename, "r")) != NULL) {
      fclose(file);
      return 1;
   }
   return 0;
}

static int is_valid(const xmlDocPtr doc, const char * schema_filename) {
   xmlDocPtr schema_doc = xmlReadFile(schema_filename, NULL, XML_PARSE_NONET);
   if (schema_doc == NULL) {
      yac_internal_abort_message((char *) SCHEMA_MALFORMED, __FILE__, __LINE__);
   }
   xmlSchemaParserCtxtPtr parser_ctxt = xmlSchemaNewDocParserCtxt(schema_doc);
   if (parser_ctxt == NULL) {
      xmlFreeDoc(schema_doc);
      yac_internal_abort_message((char *) SCHEMA_NO_PARSER, __FILE__, __LINE__);
   }
   xmlSchemaPtr schema = xmlSchemaParse(parser_ctxt);
   if (schema == NULL) {
      xmlSchemaFreeParserCtxt(parser_ctxt);
      xmlFreeDoc(schema_doc);
      yac_internal_abort_message((char *) SCHEMA_INVALID, __FILE__, __LINE__);
   }
   xmlSchemaValidCtxtPtr valid_ctxt = xmlSchemaNewValidCtxt(schema);
   if (valid_ctxt == NULL) {
      xmlSchemaFree(schema);
      xmlSchemaFreeParserCtxt(parser_ctxt);
      xmlFreeDoc(schema_doc);
      yac_internal_abort_message((char *) SCHEMA_NO_VALIDATION, __FILE__, __LINE__);
   }
   int is_valid = (xmlSchemaValidateDoc(valid_ctxt, doc) == 0);
   xmlSchemaFreeValidCtxt(valid_ctxt);
   xmlSchemaFree(schema);
   xmlSchemaFreeParserCtxt(parser_ctxt);
   xmlFreeDoc(schema_doc);
   return is_valid ? 1 : 0;
}

static void check_init(const int line) {
   if (document == NULL || coupling == NULL
         || no_of_transients == -1|| no_of_components == -1
         || no_of_couples == -1 || no_of_transient_couples == NULL) {
      yac_internal_abort_message ((char *) XML_TREE_NOT_INITIALIZED, __FILE__, line);
   }
}

static void check_string(const char * string, const int line) {
   check_init(line);
   if (string == NULL) {
      yac_internal_abort_message((char *) STRING_NULL, __FILE__, line);
   }
   if (strcmp(string, "") == 0) {
      yac_internal_abort_message((char *) STRING_EMPTY, __FILE__, line);
   }
}

static void check_component_pointer(const yac_xml_component * ptr,
      const int line) {
   if (ptr == NULL) {
      yac_internal_abort_message((char *) NULL_POINTER_COMPONENT, __FILE__, line);
   }
}

static void check_transient_pointer(const yac_xml_transient * ptr,
      const int line) {
   if (ptr == NULL) {
      yac_internal_abort_message((char *) NULL_POINTER_TRANSIENT, __FILE__, line);
   }
}

static void check_grid_pointer(const yac_xml_grid * ptr, const int line) {
   if (ptr == NULL) {
      yac_internal_abort_message((char *) NULL_POINTER_GRID, __FILE__, line);
   }
}

static void check_component_parameter(const int component_index,
      const int line) {
   check_init(line);
   if (component_index >= no_of_components) {
      char message[256];
      snprintf(message, sizeof(message), COMPONENT_NOT_CONTAINED,
            component_index);
      strcat(message, COMPONENT_INDEX_OUT_OF_BOUNDS);
      yac_internal_abort_message(message, __FILE__, line);
   }
}

static void check_component_parameters(const int component_index,
      const int transient_grid_ref_index, const int line) {
   check_component_parameter(component_index, line);
   int no_of_transient_grid_refs = yac_xml_get_number_of_transient_grid_refs(
         component_index);
   if (transient_grid_ref_index >= no_of_transient_grid_refs) {
      char message[256];
      snprintf(message, sizeof(message), TRANSIENT_GRID_REF_NOT_CONTAINED,
            transient_grid_ref_index, component_index);
      strcat(message, TRANSIENT_GRID_REF_INDEX_OUT_OF_BOUNDS);
      yac_internal_abort_message(message, __FILE__, line);
   }
}

static void check_transient_grid_ref_index(const int component_id,
      const int transient_grid_ref_index, const int line) {
   int no_of_transient_grid_refs =
         yac_xml_get_number_of_transient_grid_refs_by_component_id(
               component_id);
   if (transient_grid_ref_index >= no_of_transient_grid_refs) {
      char message[256];
      snprintf(message, sizeof(message),
            TRANSIENT_GRID_REF_NOT_CONTAINED_IN_COMPONENT_WITH_ID,
            transient_grid_ref_index, component_id);
      strcat(message, TRANSIENT_GRID_REF_INDEX_OUT_OF_BOUNDS);
      yac_internal_abort_message(message, __FILE__, line);
   }
}

static void check_transient_parameter(const int transient_index,
      const int line) {
   check_init(line);
   if (transient_index >= no_of_transients) {
      char message[256];
      snprintf(message, sizeof(message), TRANSIENT_NOT_CONTAINED,
            transient_index);
      strcat(message, TRANSIENT_INDEX_OUT_OF_BOUNDS);
      yac_internal_abort_message(message, __FILE__, line);
   }
}

static void check_grid_parameter(const int grid_index, const int line) {
   check_init(line);
   if (grid_index >= no_of_grids) {
      char message[256];
      snprintf(message, sizeof(message), GRID_NOT_CONTAINED, grid_index);
      strcat(message, GRID_INDEX_OUT_OF_BOUNDS);
      yac_internal_abort_message(message, __FILE__, line);
   }
}

static void check_couple_parameter(const int couple_index, const int line) {
   check_init(line);
   if (couple_index >= no_of_couples) {
      char message[256];
      snprintf(message, sizeof(message), COUPLE_NOT_CONTAINED, couple_index);
      strcat(message, COUPLE_INDEX_OUT_OF_BOUNDS);
      yac_internal_abort_message(message, __FILE__, line);
   }
}

static void check_couple_parameters(const int couple_index,
      const int transient_couple_index, const int line) {
   check_couple_parameter(couple_index, line);
   if (no_of_transient_couples[couple_index] == 0) {
      no_of_transient_couples[couple_index] =
            yac_xml_get_number_of_transient_couples(couple_index);
   }
   if (transient_couple_index >= no_of_transient_couples[couple_index]) {
      char message[256];
      snprintf(message, sizeof(message), TRANSIENT_COUPLE_NOT_CONTAINED,
            transient_couple_index, couple_index);
      strcat(message, TRANSIENT_COUPLE_INDEX_OUT_OF_BOUNDS);
      yac_internal_abort_message(message, __FILE__, line);
   }
}

static void check_component_id(xmlNodePtr component_node, const int id,
      const int line) {
   if (component_node == NULL) {
      char message[100];
      snprintf(message, sizeof(message), COMPONENT_ID_NOT_FOUND, id);
      yac_internal_abort_message(message, __FILE__, line);
   }
}

static void check_transient_grid_ref_id(xmlNodePtr transient_grid_ref_node,
      const int id, const int line) {
   if (transient_grid_ref_node == NULL) {
      char message[100];
      snprintf(message, sizeof(message), TRANSIENT_GRID_REF_ID_NOT_FOUND, id);
      yac_internal_abort_message(message, __FILE__, line);
   }
}

static void check_transient_id(const int transient_id, const int line) {
   if (!yac_xml_transient_id_exists_in_xml(transient_id)) {
      char message[100];
      snprintf(message, sizeof(message), TRANSIENT_ID_NOT_FOUND, transient_id);
      yac_internal_abort_message(message, __FILE__, line);
   }
}

static void check_grid_id(const int grid_id, const int line) {
   check_init(line);
   if (!yac_xml_grid_id_exists_in_xml(grid_id)) {
      char message[100];
      snprintf(message, sizeof(message), GRID_ID_NOT_FOUND, grid_id);
      yac_internal_abort_message(message, __FILE__, line);
   }
}

static void check_numeric_value(const char * value, const int line) {
   if (!is_numeric(value)) {
      char message[100];
      snprintf(message, sizeof(message), NUMERIC_VALUE_INVALID, value);
      yac_internal_abort_message(message, __FILE__, line);
//! \todo{Program will terminate here with an error. Is this really wanted?}
//      fprintf(stderr, "Value '%s' is not a numeric value. At line %d", value,
//            line);
   }
}

static void error_message_invalid_component_reference_value(const int line) {
   char message[256];
   snprintf(message, sizeof(message), "Invalid reference to component. ");
   strcat(message, "Reference has to be in [1,2].");
   yac_internal_abort_message(message, __FILE__, __LINE__);
}

static int get_number_of_component_elements(const xmlNodePtr coupling) {
   xmlNodePtr components = get_node_by_name(coupling, COMPONENTS);
   return get_number_of_elements_of_name(components, COMPONENT);
}

static int get_number_of_transient_elements(const xmlNodePtr coupling) {
   xmlNodePtr transients = get_node_by_name(coupling, TRANSIENTS);
   return get_number_of_elements_of_name(transients, TRANSIENT);
}

static int get_number_of_grid_elements(const xmlNodePtr coupling) {
   xmlNodePtr grids = get_node_by_name(coupling, GRIDS);
   return get_number_of_elements_of_name(grids, GRID);
}

static int get_number_of_couple_elements(const xmlNodePtr coupling) {
   xmlNodePtr couples = get_node_by_name(coupling, COUPLES);
   return get_number_of_elements_of_name(couples, COUPLE);
}

static xmlNodePtr get_child_element_by_id_attribute(const xmlNodePtr node,
      const int id_value) {
   xmlNodePtr child = node->children;
   char buffer[50];
   snprintf(buffer, sizeof(buffer), "%d", id_value);
   char * id = buffer;
   int found = 0;
   while (child) {
      xmlChar * id_value = xmlGetProp(child, (const xmlChar *) ID);
      if (!xmlStrcmp(id_value, (const xmlChar *) id)) {
         xmlFree(id_value);
         found = 1;
         break;
      }
      xmlFree(id_value);
      child = child->next;
   }
   if (found) {
      return child;
   } else {
      return NULL;
   }
}

static xmlNodePtr get_child_element_by_attribute(const xmlNodePtr node,
      const char * attribute_name, const char * attribute_value) {
   xmlNodePtr child = node->children;
   int found = 0;
   while (child) {
      xmlChar * attr_value = xmlGetProp(child,
            (const xmlChar *) attribute_name);
      if (!xmlStrcmp(attr_value, (const xmlChar *) attribute_value)) {
         xmlFree(attr_value);
         found = 1;
         break;
      }
      xmlFree(attr_value);
      child = child->next;
   }
   if (found) {
      return child;
   } else {
      return NULL;
   }
}

static int get_transient_couple_attribute_value(const int couple_index,
      const int transient_couple_index, const char * attribute_name) {
   xmlNodePtr transient_couple_node = get_transient_couple_node(couple_index,
         transient_couple_index);
   char * attribute_value = (char *) xmlGetProp(transient_couple_node,
         (const xmlChar *) attribute_name);
   check_numeric_value(attribute_value, __LINE__);
   int ret = atoi(attribute_value);
   xmlFree(attribute_value);
   return ret;
}

static xmlNodePtr get_component_node(const int component_index) {
   xmlNodePtr components_node = get_node_by_name(coupling, COMPONENTS);
   return nth_node(components_node, component_index);
}

static xmlNodePtr get_couple_node(const int couple_index) {
   xmlNodePtr couples_node = get_node_by_name(coupling, COUPLES);
   return nth_node(couples_node, couple_index);
}

static xmlNodePtr get_transient_couple_node(const int couple_index,
      const int transient_couple) {
   xmlNodePtr couple_node = get_couple_node(couple_index);
   return nth_node(couple_node, FIRSTTRANSIENTCOUPLE + transient_couple);
}

static xmlNodePtr get_interpolation_node(const int couple_index,
      const int transient_couple_index, const int interpolation_index) {
   xmlNodePtr transient_couple_node = get_transient_couple_node(couple_index,
         transient_couple_index);
   xmlNodePtr interpolation_requirements_node = get_node_by_name(
         transient_couple_node, INTERPOLATION_REQUIREMENTS);
   int no = get_number_of_elements_of_name(interpolation_requirements_node,
         INTERPOLATION);
   if (interpolation_index >= no) {
      char message[150];
      snprintf(message, sizeof(message), INTERPOLATION_NOT_CONTAINED,
            interpolation_index);
      strcat(message, INTERPOLATION_OUT_OF_BOUNDS);
      yac_internal_abort_message(message, __FILE__, __LINE__);
   }
   return nth_node(interpolation_requirements_node, interpolation_index);
}

/**
 * @returns node at index n (starting with 0)
 */
static xmlNodePtr nth_node(xmlNodePtr parents_node, int n) {
   xmlNodePtr child = parents_node->children;
   while (child) {
      if (child->type != XML_ELEMENT_NODE)
         child = child->next;
      else if (n == 0)
         return child;
      else {
         --n;
         child = child->next;
      }
   }
   return NULL;
}

static xmlAttrPtr nth_attribute(xmlNodePtr parent_node, int n) {
   xmlAttrPtr attr;
   int i = 0;
   for (attr = parent_node->properties; attr != NULL; attr = attr->next) {
      if (i == n) {
         return attr;
      }
      ++i;
   }
   return NULL;
}

static xmlNodePtr get_node_by_name(xmlNodePtr parent_node,
      const char * name) {
   xmlNodePtr node = parent_node->children;
   while (node) {
      if (!xmlStrcmp(node->name, (const xmlChar *) name)) {
         return node;
      }
      node = node->next;
   }
   return NULL;
}

static int get_number_of_elements_of_name(xmlNodePtr parent_node,
      const char * name) {
   xmlNodePtr child = parent_node->children;
   int n = 0;
   while (child) {
      if (!xmlStrcmp(child->name, (const xmlChar *) name)) {
         ++n;
      }
      child = child->next;
   }
   return n;
}

static enum Yac_xml_time_operation operation_enum_from_string(
      const char * operation_str) {
   if (strcmp(operation_str, ACCUMULATE_STR) == 0) {
      return T_ACCUMULATE;
   } else if (strcmp(operation_str, AVERAGE_STR) == 0) {
      return T_AVERAGE;
   } else if (strcmp(operation_str, NONE_STR) == 0) {
      return T_NONE;
   } else {
      return T_NONE;
   }
}

static enum Yac_xml_time_unit time_unit_enum_from_string(
      const char * time_unit_str) {
   if (strcmp(time_unit_str, MILLISECOND_STR) == 0) {
      return T_MILLISECOND;
   } else if (strcmp(time_unit_str, SECOND_STR) == 0) {
      return T_SECOND;
   } else if (strcmp(time_unit_str, MINUTE_STR) == 0) {
      return T_MINUTE;
   } else if (strcmp(time_unit_str, HOUR_STR) == 0) {
      return T_HOUR;
   } else if (strcmp(time_unit_str, DAY_STR) == 0) {
      return T_DAY;
   } else if (strcmp(time_unit_str, MONTH_STR) == 0) {
      return T_MONTH;
   } else if (strcmp(time_unit_str, YEAR_STR) == 0) {
      return T_YEAR;
   } else if (strcmp(time_unit_str, ISO_FORMAT_STR) == 0) {
      return T_ISO_FORMAT;
   } else {
      char message[150];
      snprintf(message, sizeof(message), WRONG_TIME_UNIT, time_unit_str);
      strcat(message, WRONG_TIME_UNIT);
      yac_internal_abort_message(message, __FILE__, __LINE__);
      return T_NO_FORMAT;
   }
}

static enum Yac_xml_calendar calendar_enum_from_string(
    const char * calendar_str) {
  if (strcmp(calendar_str, PROLEPTIC_GREGORIAN_STR) == 0) {
    return C_PROLEPTIC_GREGORIAN;
  } else if (strcmp(calendar_str, REGULAR_360D_STR) == 0) {
    return C_REGULAR_360D;
  } else if (strcmp(calendar_str, REGULAR_365D_STR) == 0) {
    return C_REGULAR_365D;
  } else {
    return C_PROLEPTIC_GREGORIAN;
  }
}

static int is_numeric(const char * s) {
   if (s == NULL || *s == '\0' || isspace(*s))
      return 0;
   char * p;
   strtod(s, &p);
   return (*p == '\0');
}


#else

static const char NO_XML_SUPPORT[] = "YAC configured without XML support";

void yac_xml_init_config_xml(const char * xml_filename, const char * schema_filename) {
#ifdef VERBOSE
  printf ( "WARNING: Compiled without xml support!\n");
#endif
}

const char * yac_xml_get_start_date() {
#ifdef VERBOSE
  printf ( "WARNING: Compiled without xml support!\n");
#endif
  return NO_XML_SUPPORT;
}

const char * yac_xml_get_end_date() {
#ifdef VERBOSE
  printf ( "WARNING: Compiled without xml support!\n");
#endif
  return NO_XML_SUPPORT;
}

int yac_xml_redirect_stdout() {
#ifdef VERBOSE
  printf ( "WARNING: Compiled without xml support!\n");
#endif
  return 0;
}

#endif
