!>
!! @file yac_finterface.f90
!! @brief Implementation of the YAC Fortran user interface
!!
!! @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
!!                                 Rene Redler <rene.redler@mpimet.mpg.de>
!!
!! @version 1.0
!! @author Moritz Hanke <hanke@dkrz.de>
!!         Rene Redler <rene.redler@mpimet.mpg.de>
!!
!! @par Warranty
!!
!! This code has been tested up to a certain level. Defects and weaknesses,
!! which may be included in the code, do not establish any warranties by the
!! authors. The authors do not make any warranty, express or implied, or assume any
!! liability or responsibility for the use, acquisition or application of this
!! software.
!!
!! @par Fortran User API
!!
!! Initialisation
!!
!! - \ref yac_finit
!! - \ref yac_ffinalize
!! - \ref yac_fget_localcomm
!!
!! Definition
!!
!! - \ref yac_fdef_datetime
!! - \ref yac_fdef_comp
!! - \ref yac_fdef_comps
!! - \ref yac_fdef_subdomain
!! - \ref yac_fconnect_subdomains
!! - \ref mo_yac_finterface::yac_fdef_points "yac_fdef_points (*)"
!! - \ref mo_yac_finterface::yac_fdef_elements "yac_fdef_elements (*)"
!! - \ref mo_yac_finterface::yac_fdef_mask "yac_fdef_mask (*)"
!! - \ref yac_fdef_index_location
!!
!! End of definition and search
!!
!! - \ref yac_fsearch
!!
!! Data exchange
!!
!! - \ref mo_yac_finterface::yac_fput "yac_fput (*)"
!! - \ref mo_yac_finterface::yac_fget "yac_fget (*)"
!!
!! Query routines
!!
!! - \ref yac_fget_field_ids
!! - \ref yac_fget_nbr_fields
!! - \ref yac_fget_field_id_from_name
!! - \ref yac_fget_field_name_from_id
!!
!! - \ref mo_yac_finterface::yac_fget_role_of_field "yac_fget_role_of_field (*)"
!! - \ref mo_yac_finterface::yac_fget_model_timestep "yac_fget_model_timestep (*)"
!! - \ref mo_yac_finterface::yac_fget_coupling_timestep "yac_fget_coupling_timestep (*)"
!!
!! Auxillary routines
!!
!! - \ref yac_abort_message
!! - \ref yac_redirstdout
!!
!! For a detailed description of the Fortran interface please see mo_yac_finterface.f90.
!! Functions marked with (*) are overloaded with respect to the data type.
!! See mo_yac_finterface.f90
!! for details.
!!
!! For a detailed description of the c interface please see interface.h
!!

!
! Keywords:
! Maintainer: Moritz Hanke <hanke@dkrz.de>
!             Rene Redler <rene.redler@mpimet.mpg.de>
! URL: https://doc.redmine.dkrz.de/YAC/html/index.html
!
! This file is part of YAC.
!
! YAC is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! YAC is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
!

!
! -------------------------- init -------------------------------------

subroutine yac_finit ( xml_filename, schema_filename )

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_finit

  character(len=*), intent(in) :: xml_filename    !< [IN] name of xml configuration file
  character(len=*), intent(in) :: schema_filename !< [IN] corresponding xsd file

  call yac_check_strlength ( xml_filename )
  call yac_check_strlength ( schema_filename )

  call yac_cinit ( TRIM(xml_filename)    // c_null_char, &
                   TRIM(schema_filename) // c_null_char )

end subroutine yac_finit

! -------------------------- final -------------------------------------

subroutine yac_ffinalize ( )

  use mo_yac_finterface, dummy => yac_ffinalize

  call yac_cfinalize ( )

end subroutine yac_ffinalize

! -------------------------- init -------------------------------------

subroutine yac_fdef_datetime ( start_datetime, end_datetime )

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_fdef_datetime

  character(len=*), intent(in), optional :: start_datetime !< [IN] start datetime of job
  character(len=*), intent(in), optional :: end_datetime   !< [IN] end datetime of job

  integer :: index

  index = 0

  call yac_check_strlength ( start_datetime )
  call yac_check_strlength ( end_datetime )

  if (present(start_datetime)) index = index + 1

  if (present(end_datetime))   index = index + 2

  select case ( index )

    case ( 3 )
      call yac_cdef_datetime ( TRIM(start_datetime) // c_null_char, &
                               TRIM(end_datetime)   // c_null_char )
    case ( 2 )
      call yac_cdef_datetime ( c_null_char, &
                               TRIM(end_datetime)   // c_null_char )
    case ( 1 )
      call yac_cdef_datetime ( TRIM(start_datetime) // c_null_char, &
                               c_null_char )
    end select

end subroutine yac_fdef_datetime

! ------------------------ def_comp ------------------------------------

subroutine yac_fdef_comp ( comp_name, comp_id )

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_fdef_comp

  character(len=*), intent(in) :: comp_name  !< [IN]  component name
  integer, intent(out)         :: comp_id    !< [OUT] returned handle to the component

  call yac_check_strlength ( comp_name )

  call yac_cdef_comp ( TRIM(comp_name) // c_null_char, comp_id )

end subroutine yac_fdef_comp

! ------------------------ def_comps ------------------------------------

subroutine yac_fdef_comps ( comp_names, num_comps, comp_ids )

  use, intrinsic :: iso_c_binding, only : c_null_char, c_ptr, c_loc, c_char
  use mo_yac_finterface, dummy => yac_fdef_comps

  integer, intent(in)   :: num_comps !< [IN]  number of components
  character(kind=c_char, len=YAC_MAX_CHARLEN), intent(in) :: &
                           comp_names(num_comps) !< [IN]  component names
  integer, intent(out)  :: comp_ids(num_comps) !< [OUT] returned handle to the components

  integer :: i, j
  character(kind=c_char), target :: comp_names_cpy(YAC_MAX_CHARLEN+1, num_comps)
  type(c_ptr) :: comp_name_ptrs(num_comps)

  comp_names_cpy = c_null_char

  do i = 1, num_comps
    do j = 1, len_trim(comp_names(i))
      comp_names_cpy(j,i) = comp_names(i)(j:j)
    end do
    comp_name_ptrs(i) = c_loc(comp_names_cpy(1,i));
  end do

  call yac_cdef_comps ( comp_name_ptrs, num_comps, comp_ids )

end subroutine yac_fdef_comps

! ------------------------ def_subdomain ------------------------------------

subroutine yac_fdef_subdomain ( comp_id, grid_name, subdomain_id )

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_fdef_subdomain

  integer, intent(in)          :: comp_id      !< [IN]  component identifier
  character(len=*), intent(in) :: grid_name    !< [IN]  grid name
  integer, intent(out)         :: subdomain_id !< [OUT] returned handle to the subdomain

  call yac_check_strlength ( grid_name )

  call yac_cdef_subdomain ( comp_id,                        &
                            TRIM(grid_name) // c_null_char, &
                            subdomain_id )

end subroutine yac_fdef_subdomain

! ------------------------ connect_subdomain ------------------------------------

subroutine yac_fconnect_subdomains ( comp_id, nbr_subdomain_ids, subdomain_ids, &
   &                                 domain_id )

  use mo_yac_finterface, dummy => yac_fconnect_subdomains

  integer, intent(in)  :: comp_id                          !< [IN] component identifier
  integer, intent(in)  :: nbr_subdomain_ids                !< [IN] number of subdomains
  integer, intent(in)  :: subdomain_ids(nbr_subdomain_ids) !< [IN] list of subdomain identifier
  integer, intent(out) :: domain_id                        !< [OUT] returned handle to the domain

  call yac_cconnect_subdomains ( comp_id,            &
                                 nbr_subdomain_ids,  &
                                 subdomain_ids,      &
                                 domain_id )

end subroutine yac_fconnect_subdomains

! ------------------------- def_points ----------------------------------

subroutine yac_fdef_points_reg2d_r4 ( subdomain_id,  &
                                      nbr_points,    &
                                      location,      &
                                      x_points_real, &
                                      y_points_real, &
                                      point_id )

  use mo_yac_finterface, dummy => yac_fdef_points_reg2d_r4

  integer, intent(in)  :: subdomain_id                 !< [IN]  subdomain identifier
  integer, intent(in)  :: nbr_points(2)                !< [IN]  number of points in x and y
  integer, intent(in)  :: location                     !< [IN]  location, one of center/edge/vertex

  real, intent(in)     :: x_points_real(nbr_points(1)) !< [IN]  longitudes of points
  real, intent(in)     :: y_points_real(nbr_points(2)) !< [IN]  latitudes of points

  integer, intent(out) :: point_id                     !< [OUT] point identifier

  double precision    ::  x_points(nbr_points(1))
  double precision    ::  y_points(nbr_points(2))

  x_points(:) = dble(x_points_real(:))
  y_points(:) = dble(y_points_real(:))

  call yac_cdef_points_reg2d ( subdomain_id, &
                               nbr_points,   &
                               location,     &
                               x_points,     &
                               y_points,     &
                               point_id )

end subroutine yac_fdef_points_reg2d_r4

subroutine yac_fdef_points_reg2d_r8 ( subdomain_id, &
                                      nbr_points,   &
                                      location,     &
                                      x_points,     &
                                      y_points,     &
                                      point_id )

  use mo_yac_finterface, dummy => yac_fdef_points_reg2d_r8

  integer, intent(in)  :: subdomain_id                    !< [IN]  subdomain identifier
  integer, intent(in)  :: nbr_points(2)                   !< [IN]  number of points in x and y
  integer, intent(in)  :: location                        !< [IN]  location, one of center/edge/vertex

  double precision, intent(in) :: x_points(nbr_points(1)) !< [IN]  longitudes of points
  double precision, intent(in) :: y_points(nbr_points(2)) !< [IN]  latitudes of points

  integer, intent(out) :: point_id                        !< [OUT] point identifier

  call yac_cdef_points_reg2d ( subdomain_id, &
                                nbr_points,  &
                                location,    &
                                x_points,    &
                                y_points,    &
                                point_id )

end subroutine yac_fdef_points_reg2d_r8


subroutine yac_fdef_points_unstruct_r4 ( subdomain_id,  &
                                         nbr_points,    &
                                         location,      &
                                         x_points_real, &
                                         y_points_real, &
                                         point_id )

  use mo_yac_finterface, dummy => yac_fdef_points_unstruct_r4

  integer, intent(in)  :: subdomain_id              !< [IN]  subdomain identifier
  integer, intent(in)  :: nbr_points                !< [IN]  number of points
  integer, intent(in)  :: location                  !< [IN]  location, one of center/edge/vertex

  real, intent(in)     :: x_points_real(nbr_points) !< [IN]  longitudes of points
  real, intent(in)     :: y_points_real(nbr_points) !< [IN]  latitudes of points

  integer, intent(out) :: point_id                  !< [OUT] point identifier

  double precision    ::  x_points(nbr_points)
  double precision    ::  y_points(nbr_points)

  x_points(:) = dble(x_points_real(:))
  y_points(:) = dble(y_points_real(:))

  call yac_cdef_points_unstruct ( subdomain_id, &
                                  nbr_points,   &
                                  location,     &
                                  x_points,     &
                                  y_points,     &
                                  point_id )

end subroutine yac_fdef_points_unstruct_r4

subroutine yac_fdef_points_unstruct_r8 ( subdomain_id, &
                                         nbr_points,   &
                                         location,     &
                                         x_points,     &
                                         y_points,     &
                                         point_id )

  use mo_yac_finterface, dummy => yac_fdef_points_unstruct_r8

  integer, intent(in)  :: subdomain_id                 !< [IN]  subdomain identifier
  integer, intent(in)  :: nbr_points                   !< [IN]  number of points
  integer, intent(in)  :: location                     !< [IN]  location, one of center/edge/vertex

  double precision, intent(in) :: x_points(nbr_points) !< [IN]  longitudes of points
  double precision, intent(in) :: y_points(nbr_points) !< [IN]  latitudes of points

  integer, intent(out) :: point_id                     !< [OUT] point identifier

  call yac_cdef_points_unstruct  ( subdomain_id, &
                                   nbr_points,   &
                                   location,     &
                                   x_points,     &
                                   y_points,     &
                                   point_id )

end subroutine yac_fdef_points_unstruct_r8

! ------------------------- def_elements -------------------------------

subroutine yac_fdef_elements_nonuniform_r4 ( subdomain_id,          &
                                             nbr_vertices,          &
                                             nbr_cells,             &
                                             nbr_connections,       &
                                             nbr_vertices_per_cell, &
                                             x_vertices_real,       &
                                             y_vertices_real,       &
                                             cell_to_vertex_in )

  use mo_yac_finterface, dummy => yac_fdef_elements_nonuniform_r4

  integer, intent(in)  :: subdomain_id                     !< [IN]  subdomain identifier
  integer, intent(in)  :: nbr_vertices                     !< [IN]  number of vertices
  integer, intent(in)  :: nbr_cells                        !< [IN]  number of cells
  integer, intent(in)  :: nbr_connections                  !< [IN]  total size of cell_to_vertex
  integer, intent(in)  :: nbr_vertices_per_cell(nbr_cells) !< [IN]  number of vertices for each cell

  real, intent(in)     :: x_vertices_real(nbr_vertices)    !< [IN]  longitudes of vertices
  real, intent(in)     :: y_vertices_real(nbr_vertices)    !< [IN]  latitudes of vertices

  integer, intent(in)  :: cell_to_vertex_in(nbr_connections) !< [IN] connectivity between vertices and cells\n
                                                             !!(the vertex indices per cell have to be in clockwise or counterclockwise ordering)

  double precision       :: x_vertices(nbr_vertices)
  double precision       :: y_vertices(nbr_vertices)

  integer                :: cell_to_vertex(nbr_connections)

  x_vertices(:) = dble(x_vertices_real(:))
  y_vertices(:) = dble(y_vertices_real(:))

  IF (ANY(cell_to_vertex_in < 1)) &
    CALL yac_abort_message("ERROR(yac_fdef_elements_nonuniform_r4): " // &
      &                    "all entries of cell_to_vertex have to be > 0", &
      &                    __FILE__, __LINE__ )

  cell_to_vertex(:) = cell_to_vertex_in(:) - 1

  call yac_cdef_elements_unstruct  ( subdomain_id,          &
                                     nbr_vertices,          &
                                     nbr_cells,             &
                                     nbr_vertices_per_cell, &
                                     x_vertices,            &
                                     y_vertices,            &
                                     cell_to_vertex )

end subroutine yac_fdef_elements_nonuniform_r4

subroutine yac_fdef_elements_nonuniform_r8 ( subdomain_id,          &
                                             nbr_vertices,          &
                                             nbr_cells,             &
                                             nbr_connections,       &
                                             nbr_vertices_per_cell, &
                                             x_vertices,            &
                                             y_vertices,            &
                                             cell_to_vertex_in )

  use mo_yac_finterface, dummy => yac_fdef_elements_nonuniform_r8

  integer, intent(in)          :: subdomain_id                       !< [IN]  subdomain identifier
  integer, intent(in)          :: nbr_vertices                       !< [IN]  number of vertices
  integer, intent(in)          :: nbr_cells                          !< [IN]  number of cells
  integer, intent(in)          :: nbr_connections                    !< [IN]  total size of cell_to_vertex
  integer, intent(in)          :: nbr_vertices_per_cell(nbr_cells)   !< [IN]  number of vertices for each cell

  double precision, intent(in) :: x_vertices(nbr_vertices)           !< [IN]  longitudes of vertices
  double precision, intent(in) :: y_vertices(nbr_vertices)           !< [IN]  latitudes of vertices

  integer, intent(in)          :: cell_to_vertex_in(nbr_connections) !< [IN] connectivity between vertices and cells\n
                                                                     !!(the vertex indices per cell have to be in clockwise or counterclockwise ordering)

  integer                      :: cell_to_vertex(nbr_connections)

  IF (ANY(cell_to_vertex_in < 1)) &
    CALL yac_abort_message("ERROR(yac_fdef_elements_nonuniform_r8): " // &
      &                    "all entries of cell_to_vertex have to be > 0", &
      &                    __FILE__, __LINE__ )

  cell_to_vertex(:) = cell_to_vertex_in(:) - 1

  call yac_cdef_elements_unstruct  ( subdomain_id,          &
                                     nbr_vertices,          &
                                     nbr_cells,             &
                                     nbr_vertices_per_cell, &
                                     x_vertices,            &
                                     y_vertices,            &
                                     cell_to_vertex )

end subroutine yac_fdef_elements_nonuniform_r8

subroutine yac_fdef_elements_unstruct_r4 ( subdomain_id,             &
                                           nbr_vertices,             &
                                           nbr_cells,                &
                                           nbr_vertices_per_cell_in, &
                                           x_vertices_real,          &
                                           y_vertices_real,          &
                                           cell_to_vertex_in )

  use mo_yac_finterface, dummy => yac_fdef_elements_unstruct_r4

  integer, intent(in)  :: subdomain_id                      !< [IN]  subdomain identifier
  integer, intent(in)  :: nbr_vertices                      !< [IN]  number of vertices
  integer, intent(in)  :: nbr_cells                         !< [IN]  number of cells
  integer, intent(in)  :: nbr_vertices_per_cell_in          !< [IN]  number of vertices for each cell

  real, intent(in)     :: x_vertices_real(nbr_vertices)     !< [IN]  longitudes of vertices
  real, intent(in)     :: y_vertices_real(nbr_vertices)     !< [IN]  latitudes of vertices

  integer, intent(in)  :: cell_to_vertex_in(nbr_vertices_per_cell_in,nbr_cells) !< [IN] connectivity between vertices and cells\n
                                                                                !!(the vertex indices per cell have to be in clockwise or counterclockwise ordering)

  double precision       :: x_vertices(nbr_vertices)
  double precision       :: y_vertices(nbr_vertices)

  integer                :: nbr_vertices_per_cell(nbr_cells)
  integer                :: cell_to_vertex(nbr_vertices_per_cell_in,nbr_cells)

  nbr_vertices_per_cell(:) = nbr_vertices_per_cell_in

  x_vertices(:) = dble(x_vertices_real(:))
  y_vertices(:) = dble(y_vertices_real(:))

  IF (ANY(cell_to_vertex_in < 1)) &
    CALL yac_abort_message("ERROR(yac_fdef_elements_unstruct_r4): " // &
      &                    "all entries of cell_to_vertex have to be > 0", &
      &                    __FILE__, __LINE__ )

  cell_to_vertex(:,:) = cell_to_vertex_in(:,:) - 1

  call yac_cdef_elements_unstruct  ( subdomain_id,          &
                                     nbr_vertices,          &
                                     nbr_cells,             &
                                     nbr_vertices_per_cell, &
                                     x_vertices,            &
                                     y_vertices,            &
                                     cell_to_vertex )

end subroutine yac_fdef_elements_unstruct_r4

subroutine yac_fdef_elements_unstruct_r8 ( subdomain_id,             &
                                           nbr_vertices,             &
                                           nbr_cells,                &
                                           nbr_vertices_per_cell_in, &
                                           x_vertices,               &
                                           y_vertices,               &
                                           cell_to_vertex_in )

  use mo_yac_finterface, dummy => yac_fdef_elements_unstruct_r8

  integer, intent(in)          :: subdomain_id                 !< [IN]  subdomain identifier
  integer, intent(in)          :: nbr_vertices                 !< [IN]  number of vertices
  integer, intent(in)          :: nbr_cells                    !< [IN]  number of cells
  integer, intent(in)          :: nbr_vertices_per_cell_in     !< [IN]  number of vertices for each cell

  double precision, intent(in) :: x_vertices(nbr_vertices)     !< [IN]  longitudes of vertices
  double precision, intent(in) :: y_vertices(nbr_vertices)     !< [IN]  latitudes of vertices

  integer, intent(in)  :: cell_to_vertex_in(nbr_vertices_per_cell_in,nbr_cells) !< [IN] connectivity between vertices and cells\n
                                                                                !!(the vertex indices per cell have to be in clockwise or counterclockwise ordering)

  integer                      :: nbr_vertices_per_cell(nbr_cells)
  integer                      :: cell_to_vertex(nbr_vertices_per_cell_in,nbr_cells)

  nbr_vertices_per_cell(:) = nbr_vertices_per_cell_in

  IF (ANY(cell_to_vertex_in < 1)) &
    CALL yac_abort_message("ERROR(yac_fdef_elements_unstruct_r8): " // &
      &                    "all entries of cell_to_vertex have to be > 0", &
      &                    __FILE__, __LINE__ )

  cell_to_vertex(:,:) = cell_to_vertex_in(:,:) - 1

  call yac_cdef_elements_unstruct  ( subdomain_id,          &
                                     nbr_vertices,          &
                                     nbr_cells,             &
                                     nbr_vertices_per_cell, &
                                     x_vertices,            &
                                     y_vertices,            &
                                     cell_to_vertex )

end subroutine yac_fdef_elements_unstruct_r8

subroutine yac_fdef_elements_reg2d_r4 ( subdomain_id,       &
                                        nbr_vertices,       &
                                        cyclic,             &
                                        x_vertices_real,    &
                                        y_vertices_real )

  use mo_yac_finterface, dummy => yac_fdef_elements_reg2d_r4

  integer, intent(in)  :: subdomain_id                     !< [IN]  subdomain identifier
  integer, intent(in)  :: nbr_vertices(2)                  !< [IN]  number of cells in each dimension
  integer, intent(in)  :: cyclic(2)                        !< [IN]  cyclic behavior of cells in each dimension

  real, intent(in)     :: x_vertices_real(nbr_vertices(1)) !< [IN]  longitudes of vertices
  real, intent(in)     :: y_vertices_real(nbr_vertices(2)) !< [IN]  latitudes of vertices

  double precision     ::  x_vertices(nbr_vertices(1))
  double precision     ::  y_vertices(nbr_vertices(2))

  x_vertices(:) = dble(x_vertices_real(:))
  y_vertices(:) = dble(y_vertices_real(:))

  call yac_cdef_elements_reg2d ( subdomain_id, &
                                 nbr_vertices, &
                                 cyclic,       &
                                 x_vertices,   &
                                 y_vertices )

end subroutine yac_fdef_elements_reg2d_r4

subroutine yac_fdef_elements_reg2d_r8 ( subdomain_id, &
                                        nbr_vertices, &
                                        cyclic,       &
                                        x_vertices,   &
                                        y_vertices )

  use mo_yac_finterface, dummy => yac_fdef_elements_reg2d_r8

  integer, intent(in)          :: subdomain_id                !< [IN]  subdomain identifier
  integer, intent(in)          :: nbr_vertices(2)             !< [IN]  number of cells in each dimension
  integer, intent(in)          :: cyclic(2)                   !< [IN]  cyclic behavior of cells in each dimension

  double precision, intent(in) :: x_vertices(nbr_vertices(1)) !< [IN]  longitudes of vertices
  double precision, intent(in) :: y_vertices(nbr_vertices(2)) !< [IN]  latitudes of vertices

  call yac_cdef_elements_reg2d ( subdomain_id, &
                                 nbr_vertices, &
                                 cyclic,       &
                                 x_vertices,   &
                                 y_vertices )

end subroutine yac_fdef_elements_reg2d_r8

! ------------------------- def_mask ------------------------------

subroutine yac_fdef_lmask ( mask_size, &
                            lmask,     &
                            points_id, &
                            mask_id )

  use mo_yac_finterface, dummy => yac_fdef_lmask

  integer, intent(in)  :: mask_size        !< [IN] size of mask
  logical, intent(in)  :: lmask(mask_size) !< [IN] logical mask
                                           !< false, mask is not set
                                           !< true, mask is set
  integer, intent(in)  :: points_id        !< [IN] point set indentifier
  integer, intent(out) :: mask_id          !< [OUT] handle to mask

  integer :: i
  integer :: imask(mask_size)

  do i = 1, mask_size
     if ( lmask(i) ) then
        imask(i) = 1
     else
        imask(i) = 0
     endif
  enddo

  call yac_cdef_mask ( imask,        &
                       points_id,    &
                       mask_id )

end subroutine yac_fdef_lmask

subroutine yac_fdef_imask ( mask_size, &
                            imask,     &
                            points_id, &
                            mask_id )

  use mo_yac_finterface, dummy => yac_fdef_imask

  integer, intent(in)  :: mask_size        !< [IN] size of mask
  integer, intent(in)  :: imask(mask_size) !< [IN] integer mask
                                           !< 0 false, mask is not set
                                           !< 1 true, mask is set
  integer, intent(in)  :: points_id        !< [IN] point set indentifier
  integer, intent(out) :: mask_id          !< [OUT] handle to mask

  call yac_cdef_mask ( imask,     &
                       points_id, &
                       mask_id )

end subroutine yac_fdef_imask

! -------------------------- def_index_location -------------------------

subroutine yac_fdef_index_location ( subdomain_id,        &
                                     nbr_entries,         &
                                     location,            &
                                     global_index,        &
                                     global_index_rank )

  use mo_yac_finterface, dummy => yac_fdef_index_location

  integer, intent (in) :: subdomain_id                   !< [IN] subdomain identifier
  integer, intent (in) :: nbr_entries                    !< [IN] number of indices
  integer, intent (in) :: location                       !< [IN] one of CELL/EDGE/POINT
  integer, intent (in) :: global_index(nbr_entries)      !< [IN] global unique number of data point
  integer, intent (in) :: global_index_rank(nbr_entries) !< [IN] rank of data point

  call yac_cdef_index_location ( subdomain_id,        &
                                 nbr_entries,         &
                                 location,            &
                                 global_index,        &
                                 global_index_rank )

end subroutine yac_fdef_index_location

! ----------------------------- def_field -------------------------------

subroutine yac_fdef_field ( field_name,               &
                            component_id,             &
                            domain_id,                &
                            point_ids,                &
                            mask_ids,                 &
                            num_points_per_subdomain, &
                            field_id )

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_fdef_field

  !
  ! Definition of coupling fields
  !
  character(len=*), intent (in) :: field_name       !< [IN] short name of the field
  integer, intent (in)          :: component_id             !< [IN] component identifier
  integer, intent (in)          :: domain_id                !< [IN] domain identifier
  integer, intent (in)          :: point_ids(*)             !< [IN] point identifier
  integer, intent (in)          :: mask_ids(*)              !< [IN] mask identifier
  integer, intent (in)          :: num_points_per_subdomain !< [IN] number of pointsets per subdomain
  integer, intent (out)         :: field_id                 !< [OUT] returned field handle

  call yac_check_strlength ( field_name )

  call yac_cdef_field ( TRIM(field_name) // c_null_char, &
                        component_id,                    &
                        domain_id,                       &
                        point_ids,                       &
                        mask_ids,                        &
                        num_points_per_subdomain,        &
                        field_id )

end subroutine yac_fdef_field

subroutine yac_fdef_field_nomask ( field_name,               &
                                   component_id,             &
                                   domain_id,                &
                                   point_ids,                &
                                   num_points_per_subdomain, &
                                   field_id )

  use, intrinsic :: iso_c_binding, only : c_null_char
  use mo_yac_finterface, dummy => yac_fdef_field_nomask

  !
  ! Definition of coupling fields
  !
  character(len=*), intent (in) :: field_name       !< [IN] short name of the field
  integer, intent (in)          :: component_id             !< [IN] component identifier
  integer, intent (in)          :: domain_id                !< [IN] domain identifier
  integer, intent (in)          :: point_ids(*)             !< [IN] point identifier
  integer, intent (in)          :: num_points_per_subdomain !< [IN] number of pointsets per subdomain
  integer, intent (out)         :: field_id                 !< [OUT] returned field handle

  call yac_check_strlength ( field_name )

  call yac_cdef_field_nomask ( TRIM(field_name) // c_null_char, &
                               component_id,                    &
                               domain_id,                       &
                               point_ids,                       &
                               num_points_per_subdomain,        &
                               field_id )

end subroutine yac_fdef_field_nomask

! ---------------------------------- put --------------------------------

subroutine yac_fput_r4 ( field_id,         &
     &                   nbr_hor_points,   &
     &                   collection_size,  &
     &                   nbr_pointsets,    &
     &                   nbr_subdomains,   &
     &                   send_field,       &
     &                   info,             &
     &                   ierror )

  use mo_yac_finterface, dummy => yac_fput_r4

  integer, intent (in)  :: field_id        !< [IN] field identifier
  integer, intent (in)  :: nbr_hor_points  !< [IN] number of horizontal points
  integer, intent (in)  :: collection_size !< [IN] number of vertical level or bundles
  integer, intent (in)  :: nbr_pointsets   !< [IN] number of point sets
  integer, intent (in)  :: nbr_subdomains  !< [IN] number of nbr_subdomain_ids
                                           !<      from corresponding
                                           !<      yac_fconnect_subdomains call
  real, intent (in)     :: send_field(nbr_hor_points, collection_size, &
   &                                  nbr_pointsets, nbr_subdomains)
                                           !< [IN] returned field
  integer, intent (out) :: info            !< [OUT] returned info
  integer, intent (out) :: ierror          !< [OUT] returned error

  double precision :: send_buffer(nbr_hor_points, collection_size, &
   &                              nbr_pointsets, nbr_subdomains)

  send_buffer = DBLE(send_field)

  call yac_cput_ ( field_id,         &
     &             collection_size,  &
     &             send_buffer,      &
     &             info,             &
     &             ierror )

end subroutine yac_fput_r4

subroutine yac_fput_r4_ptr ( field_id,         &
     &                       collection_size,  &
     &                       nbr_pointsets,    &
     &                       nbr_subdomains,   &
     &                       send_field,       &
     &                       info,             &
     &                       ierror )

  use mo_yac_finterface, dummy => yac_fput_r4_ptr

  integer, intent (in)     :: field_id        !< [IN] field identifier
  integer, intent (in)     :: collection_size !< [IN] number of vertical level or bundles
  integer, intent (in)     :: nbr_pointsets   !< [IN] number of point sets
  integer, intent (in)     :: nbr_subdomains  !< [IN] number of nbr_subdomain_ids
                                              !<      from corresponding
                                              !<      yac_fconnect_subdomains call
  type(yac_r4_ptr), intent (in) :: send_field(nbr_pointsets, nbr_subdomains, &
    &                                         collection_size)
                                              !< [IN] returned field
  integer, intent (out)    :: info            !< [OUT] returned info
  integer, intent (out)    :: ierror          !< [OUT] returned error

  integer :: i, j, k
  type(yac_r8_ptr) :: send_field_r8(nbr_pointsets, nbr_subdomains, &
    &                               collection_size)

  do i = 1, collection_size
    do j = 1, nbr_subdomains
      do k = 1, nbr_pointsets
        allocate(send_field_r8(k, j, i)%p(SIZE(send_field(k, j, i)%p)))
        send_field_r8(k, j, i)%p = dble(send_field(k, j, i)%p)
      end do
    end do
  end do

  call yac_fput_r8_ptr ( field_id,         &
     &                   collection_size,  &
     &                   nbr_pointsets,    &
     &                   nbr_subdomains,   &
     &                   send_field_r8,    &
     &                   info,             &
     &                   ierror )

  do i = 1, collection_size
    do j = 1, nbr_subdomains
      do k = 1, nbr_pointsets
        deallocate(send_field_r8(k, j, i)%p)
      end do
    end do
  end do

end subroutine yac_fput_r4_ptr

subroutine yac_fput_single_domain_r4 ( field_id,         &
     &                                 nbr_hor_points,   &
     &                                 collection_size,  &
     &                                 nbr_pointsets,    &
     &                                 nbr_subdomains,   &
     &                                 send_field,       &
     &                                 info,             &
     &                                 ierror )

  use mo_yac_finterface, dummy => yac_fput_single_domain_r4

  integer, intent (in)          :: field_id        !< [IN] field identifier
  integer, intent (in)          :: nbr_hor_points  !< [IN] number of horizontal points
  integer, intent (in)          :: collection_size !< [IN] number of vertical level or bundles
  integer, intent (in)          :: nbr_pointsets   !< [IN] number of point sets
  integer, intent (in)          :: nbr_subdomains  !< [IN] number of nbr_subdomain_ids
                                                   !<      from corresponding
                                                   !<      yac_fconnect_subdomains call
  real, intent (in)             :: send_field(nbr_hor_points, collection_size)
                                                   !< [IN] returned field
  integer, intent (out)         :: info            !< [OUT] returned info
  integer, intent (out)         :: ierror          !< [OUT] returned error

  double precision :: send_buffer(nbr_hor_points, collection_size)

  character (len=YAC_MAX_CHARLEN) :: message

  if ( nbr_pointsets /= 1 .OR. nbr_subdomains /= 1 ) then
     message = "nbr_pointsets and nbr_subdomains have to be 1"
     call yac_abort_message ( message , __FILE__, __LINE__ )
  endif

  send_buffer = DBLE(send_field)

  call yac_cput_ ( field_id,         &
     &             collection_size,  &
     &             send_buffer,      &
     &             info,             &
     &             ierror )

end subroutine yac_fput_single_domain_r4

subroutine yac_fput_r8 ( field_id,         &
     &                   nbr_hor_points,   &
     &                   collection_size,  &
     &                   nbr_pointsets,    &
     &                   nbr_subdomains,   &
     &                   send_field,       &
     &                   info,             &
     &                   ierror )

  use mo_yac_finterface, dummy => yac_fput_r8

  integer, intent (in)          :: field_id        !< [IN] field identifier
  integer, intent (in)          :: nbr_hor_points  !< [IN] number of horizontal points
  integer, intent (in)          :: collection_size !< [IN] number of vertical level or bundles
  integer, intent (in)          :: nbr_pointsets   !< [IN] number of point sets
  integer, intent (in)          :: nbr_subdomains  !< [IN] number of nbr_subdomain_ids
                                                   !<      from corresponding
                                                   !<      yac_fconnect_subdomains call
  double precision, intent (in) :: send_field(nbr_hor_points, collection_size, &
    &                                         nbr_pointsets, nbr_subdomains)
                                                   !< [IN] returned field
  integer, intent (out)         :: info            !< [OUT] returned info
  integer, intent (out)         :: ierror          !< [OUT] returned error

  call yac_cput_ ( field_id,         &
     &             collection_size,  &
     &             send_field,       &
     &             info,             &
     &             ierror )

end subroutine yac_fput_r8

subroutine yac_fput_r8_ptr ( field_id,         &
     &                       collection_size,  &
     &                       nbr_pointsets,    &
     &                       nbr_subdomains,   &
     &                       send_field,       &
     &                       info,             &
     &                       ierror )

  use mo_yac_finterface, dummy => yac_fput_r8_ptr
  use iso_c_binding, only: c_ptr, c_loc

  integer, intent (in)     :: field_id        !< [IN] field identifier
  integer, intent (in)     :: collection_size !< [IN] number of vertical level or bundles
  integer, intent (in)     :: nbr_pointsets   !< [IN] number of point sets
  integer, intent (in)     :: nbr_subdomains  !< [IN] number of nbr_subdomain_ids
                                              !<      from corresponding
                                              !<      yac_fconnect_subdomains call
  type(yac_r8_ptr), intent (in) :: send_field(nbr_pointsets, nbr_subdomains, &
    &                                         collection_size)
                                              !< [IN] returned field
  integer, intent (out)    :: info            !< [OUT] returned info
  integer, intent (out)    :: ierror          !< [OUT] returned error

  integer :: i, j, k
  type(c_ptr) :: send_field_(nbr_pointsets, nbr_subdomains, collection_size)

  do i = 1, collection_size
    do j = 1, nbr_subdomains
      do k = 1, nbr_pointsets
        send_field_(k, j, i) = c_loc(send_field(k, j, i)%p(1))
      end do
    end do
  end do

  call yac_cput_ptr_ ( field_id,         &
     &                 collection_size,  &
     &                 send_field_,      &
     &                 info,             &
     &                 ierror )

end subroutine yac_fput_r8_ptr

subroutine yac_fput_single_domain_r8 ( field_id,         &
     &                                 nbr_hor_points,   &
     &                                 collection_size,  &
     &                                 nbr_pointsets,    &
     &                                 nbr_subdomains,   &
     &                                 send_field,       &
     &                                 info,             &
     &                                 ierror )

  use mo_yac_finterface, dummy => yac_fput_single_domain_r8

  integer, intent (in)          :: field_id        !< [IN] field identifier
  integer, intent (in)          :: nbr_hor_points  !< [IN] number of horizontal points
  integer, intent (in)          :: collection_size !< [IN] number of vertical level or bundles
  integer, intent (in)          :: nbr_pointsets   !< [IN] number of point sets
  integer, intent (in)          :: nbr_subdomains  !< [IN] number of nbr_subdomain_ids
                                                   !<      from corresponding
                                                   !<      yac_fconnect_subdomains call
  double precision, intent (in) :: send_field(nbr_hor_points, collection_size)
                                                   !< [IN] returned field
  integer, intent (out)         :: info            !< [OUT] returned info
  integer, intent (out)         :: ierror          !< [OUT] returned error

  character (len=YAC_MAX_CHARLEN) :: message

  if ( nbr_pointsets /= 1 .OR. nbr_subdomains /= 1 ) then
     message = "nbr_pointsets and nbr_subdomains have to be 1"
     call yac_abort_message ( message , __FILE__, __LINE__ )
  endif

  call yac_cput_ ( field_id,         &
     &             collection_size,  &
     &             send_field,       &
     &             info,             &
     &             ierror )

end subroutine yac_fput_single_domain_r8

! ---------------------------------- get -------------------------------

subroutine yac_fget_r4 ( field_id,         &
     &                   nbr_hor_points,   &
     &                   collection_size,  &
     &                   nbr_pointsets,    &
     &                   nbr_subdomains,   &
     &                   recv_field,       &
     &                   info,             &
     &                   ierror )

  use mo_yac_finterface, dummy => yac_fget_r4

  integer, intent (in)          :: field_id        !< [IN] field identifier
  integer, intent (in)          :: nbr_hor_points  !< [IN] number of horizontal points
  integer, intent (in)          :: collection_size !< [IN] number of vertical level or bundles
  integer, intent (in)          :: nbr_pointsets   !< [IN] number of point sets
  integer, intent (in)          :: nbr_subdomains  !< [IN] number of nbr_subdomain_ids
                                                   !<      from corresponding
                                                   !<      yac_fconnect_subdomains call
  real, intent (inout)          :: recv_field(nbr_hor_points, collection_size, &
    &                                         nbr_pointsets, nbr_subdomains)
                                                   !< [INOUT] returned field
  integer, intent (out)         :: info            !< [OUT] returned info
  integer, intent (out)         :: ierror          !< [OUT] returned error

  double precision :: recv_buffer(nbr_hor_points, collection_size, &
    &                             nbr_pointsets, nbr_subdomains)

  call yac_cget_ ( field_id,         &
     &             collection_size,  &
     &             recv_buffer,      &
     &             info,             &
     &             ierror )

  recv_field = REAL(recv_buffer)

end subroutine yac_fget_r4



subroutine yac_fget_r4_ptr ( field_id,         &
  &                          collection_size,  &
  &                          nbr_pointsets,    &
  &                          nbr_subdomains,   &
  &                          recv_field,       &
  &                          info,             &
  &                          ierror )

  use mo_yac_finterface, dummy => yac_fget_r4_ptr

  integer, intent (in)  :: field_id        !< [IN] field identifier
  integer, intent (in)  :: collection_size !< [IN] number of vertical level or bundles
  integer, intent (in)  :: nbr_pointsets   !< [IN] number of point sets
  integer, intent (in)  :: nbr_subdomains  !< [IN] number of nbr_subdomain_ids from corresponding yac_fconnect_subdomains call
  type(yac_r4_ptr)      :: recv_field(nbr_pointsets, &
    &                                 nbr_subdomains, &
    &                                 collection_size) !< [INOUT] returned field
  integer, intent (out) :: info            !< [OUT] returned info
  integer, intent (out) :: ierror          !< [OUT] returned error handler

  integer :: i, j, k
  type(yac_r8_ptr) :: recv_field_r8(nbr_pointsets, nbr_subdomains, &
    &                               collection_size)

  do i = 1, collection_size
    do j = 1, nbr_subdomains
      do k = 1, nbr_pointsets
        allocate(recv_field_r8(k, j, i)%p(SIZE(recv_field(k, j, i)%p)))
        recv_field_r8(k, j, i)%p = dble(recv_field(k, j, i)%p)
      end do
    end do
  end do

  call yac_fget_r8_ptr ( field_id,         &
     &                   collection_size,  &
     &                   nbr_pointsets,    &
     &                   nbr_subdomains,   &
     &                   recv_field_r8,    &
     &                   info,             &
     &                   ierror )

  do i = 1, collection_size
    do j = 1, nbr_subdomains
      do k = 1, nbr_pointsets
        deallocate(recv_field_r8(k, j, i)%p)
      end do
    end do
  end do

end subroutine yac_fget_r4_ptr

subroutine yac_fget_single_domain_r4 ( field_id,         &
     &                                 nbr_hor_points,   &
     &                                 collection_size,  &
     &                                 nbr_pointsets,    &
     &                                 nbr_subdomains,   &
     &                                 recv_field,       &
     &                                 info,             &
     &                                 ierror )

  use mo_yac_finterface, dummy => yac_fget_single_domain_r4

  integer, intent (in)          :: field_id        !< [IN] field identifier
  integer, intent (in)          :: nbr_hor_points  !< [IN] number of horizontal points
  integer, intent (in)          :: collection_size !< [IN] number of vertical level or bundles
  integer, intent (in)          :: nbr_pointsets   !< [IN] number of point sets
  integer, intent (in)          :: nbr_subdomains  !< [IN] number of nbr_subdomain_ids
                                                   !<      from corresponding
                                                   !<      yac_fconnect_subdomains call
  real, intent (inout)          :: recv_field(nbr_hor_points, collection_size)
                                                   !< [INOUT] returned field
  integer, intent (out)         :: info            !< [OUT] returned info
  integer, intent (out)         :: ierror          !< [OUT] returned error

  double precision :: recv_buffer(nbr_hor_points, collection_size)

  character (len=YAC_MAX_CHARLEN) :: message

  if ( nbr_pointsets /= 1 .OR. nbr_subdomains /= 1 ) then
     message = "nbr_pointsets and nbr_subdomains have to be 1"
     call yac_abort_message ( message , __FILE__, __LINE__ )
  endif

  call yac_cget_ ( field_id,         &
     &             collection_size,  &
     &             recv_buffer,      &
     &             info,             &
     &             ierror )

  recv_field = REAL(recv_buffer)

end subroutine yac_fget_single_domain_r4

subroutine yac_fget_r8 ( field_id,         &
     &                   nbr_hor_points,   &
     &                   collection_size,  &
     &                   nbr_pointsets,    &
     &                   nbr_subdomains,   &
     &                   recv_field,       &
     &                   info,             &
     &                   ierror )

  use mo_yac_finterface, dummy => yac_fget_r8

  integer, intent (in)          :: field_id        !< [IN] field identifier
  integer, intent (in)          :: nbr_hor_points  !< [IN] number of horizontal points
  integer, intent (in)          :: collection_size !< [IN] number of vertical level or bundles
  integer, intent (in)          :: nbr_pointsets   !< [IN] number of point sets
  integer, intent (in)          :: nbr_subdomains  !< [IN] number of nbr_subdomain_ids from
                                                   !< corresponding yac_fconnect_subdomains call
  double precision, intent (inout):: recv_field(nbr_hor_points,  &
    &                                           collection_size, &
    &                                           nbr_pointsets,   &
    &                                           nbr_subdomains)
                                                  !< [INOUT] returned field
  integer, intent (out)         :: info           !< [OUT] returned info
  integer, intent (out)         :: ierror         !< [OUT] returned error

  call yac_cget_ ( field_id,         &
     &             collection_size,  &
     &             recv_field,       &
     &             info,             &
     &             ierror )

end subroutine yac_fget_r8

subroutine yac_fget_r8_ptr ( field_id,         &
                             collection_size,  &
                             nbr_pointsets,    &
                             nbr_subdomains,   &
                             recv_field,       &
                             info,             &
                             ierror )

  use mo_yac_finterface, dummy => yac_fget_r8_ptr
  use iso_c_binding, only: c_ptr, c_loc

  integer, intent (in)  :: field_id        !< [IN] field identifier
  integer, intent (in)  :: collection_size !< [IN] number of vertical level or bundles
  integer, intent (in)  :: nbr_pointsets   !< [IN] number of point sets
  integer, intent (in)  :: nbr_subdomains  !< [IN] number of nbr_subdomain_ids from corresponding yac_fconnect_subdomains call
  type(yac_r8_ptr)      :: recv_field(nbr_pointsets, &
    &                                 nbr_subdomains, &
    &                                 collection_size) !< [OUT] returned field
  integer, intent (out) :: info            !< [OUT] returned info
  integer, intent (out) :: ierror          !< [OUT] returned error handler

  integer :: i, j, k
  type(c_ptr) :: recv_field_(nbr_pointsets, nbr_subdomains, collection_size)

  do i = 1, collection_size
    do j = 1, nbr_subdomains
      do k = 1, nbr_pointsets
        recv_field_(k, j, i) = c_loc(recv_field(k, j, i)%p(1))
      end do
    end do
  end do

  call yac_cget_ptr_ ( field_id,         &
     &                 collection_size,  &
     &                 recv_field_,      &
     &                 info,             &
     &                 ierror )

end subroutine yac_fget_r8_ptr

subroutine yac_fget_single_domain_r8 ( field_id,         &
     &                                 nbr_hor_points,   &
     &                                 collection_size,  &
     &                                 nbr_pointsets,    &
     &                                 nbr_subdomains,   &
     &                                 recv_field,       &
     &                                 info,             &
     &                                 ierror )

  use mo_yac_finterface, dummy => yac_fget_single_domain_r8

  integer, intent (in)           :: field_id        !< [IN] field identifier
  integer, intent (in)           :: nbr_hor_points  !< [IN] number of horizontal points
  integer, intent (in)           :: collection_size !< [IN] number of vertical level or bundles
  integer, intent (in)           :: nbr_pointsets   !< [IN] number of point sets
  integer, intent (in)           :: nbr_subdomains  !< [IN] number of nbr_subdomain_ids from
                                                    !<      corresponding yac_fconnect_subdomains call
  double precision, intent (inout) :: recv_field(nbr_hor_points, collection_size)
                                                    !< [INOUT] returned field
  integer, intent (out)          :: info            !< [OUT] returned info
  integer, intent (out)          :: ierror          !< [OUT] returned error

  character (len=YAC_MAX_CHARLEN) :: message

  if ( nbr_pointsets /= 1 .OR. nbr_subdomains /= 1 ) then
     message = "nbr_pointsets and nbr_subdomains have to be 1"
     call yac_abort_message ( message , __FILE__, __LINE__ )
  endif

  call yac_cget_ ( field_id,         &
     &             collection_size,  &
     &             recv_field,       &
     &             info,             &
     &             ierror )

end subroutine yac_fget_single_domain_r8

! ----------------------------------------------------------------------

subroutine yac_fget_localcomm ( local_communicator, comp_id )

   use mo_yac_finterface, dummy => yac_fget_localcomm

   integer, intent (out) :: local_communicator !< [OUT] component MPI communicator
   integer, intent (in)  :: comp_id            !< [IN] component identifier

   call yac_get_localcomm ( local_communicator, comp_id )

end subroutine yac_fget_localcomm

! ----------------------------------------------------------------------

subroutine yac_fget_comp_io_proc_step ( comp_id, io_proc_step )

   use mo_yac_finterface, dummy => yac_fget_comp_io_proc_step

   integer, intent (in)  :: comp_id            !< [IN] component identifier
   integer, intent (out) :: io_proc_step       !< [OUT] stride between ranks of io processes

   call yac_cget_comp_io_proc_step ( comp_id, io_proc_step )

end subroutine yac_fget_comp_io_proc_step

! ----------------------------------------------------------------------

subroutine yac_fset_comp_io_proc_step ( comp_id, io_proc_step )

   use mo_yac_finterface, dummy => yac_fset_comp_io_proc_step

   integer, intent (in) :: comp_id            !< [IN] component identifier
   integer, intent (in) :: io_proc_step       !< [IN] stride between ranks of io processes

   call yac_cset_comp_io_proc_step ( comp_id, io_proc_step )

end subroutine yac_fset_comp_io_proc_step

! ----------------------------------------------------------------------

subroutine yac_fget_pair_rootcomm ( local_communicator, comp_name1, comp_name2 )

   use, intrinsic :: iso_c_binding, only : c_null_char
   use mo_yac_finterface, dummy => yac_fget_pair_rootcomm

   integer, intent (out) :: local_communicator !< [OUT] component MPI communicator
   character(len=*), intent(in) :: comp_name1  !< [IN]  first component name
   character(len=*), intent(in) :: comp_name2  !< [IN]  second component name

   call yac_get_pair_rootcomm ( local_communicator, TRIM(comp_name1) // c_null_char, TRIM(comp_name2)  // c_null_char )

end subroutine yac_fget_pair_rootcomm

! ------------------- search/end of definition -------------------------

subroutine yac_fsearch ( nbr_comps, comp_ids, nbr_fields, field_ids, ierror )

   use mo_yac_finterface, dummy => yac_fsearch

   integer, intent (in)  :: nbr_comps             !< [IN] number of components
   integer, intent (in)  :: comp_ids(nbr_comps)   !< [IN] list of components IDs
   integer, intent (in)  :: nbr_fields            !< [IN] number of coupling fields
   integer, intent (in)  :: field_ids(nbr_fields) !< [IN] list of field IDs
   integer, intent (out) :: ierror                !< [OUT] returned error

   call yac_csearch ( nbr_comps, comp_ids, nbr_fields, field_ids, ierror )

end subroutine yac_fsearch

! ------------------------ query routines -----------------------------

subroutine yac_fget_nbr_fields ( nbr_fields )

   use mo_yac_finterface, dummy => yac_fget_nbr_fields

   integer, intent (out) :: nbr_fields !< [OUT] number of coupling fields

   call yac_cget_nbr_fields ( nbr_fields )

end subroutine yac_fget_nbr_fields

! ---------------------------------------------------------------------

subroutine yac_fget_field_ids ( nbr_fields, field_ids )

   use mo_yac_finterface, dummy => yac_fget_field_ids

   integer, intent (in)  :: nbr_fields            !< [IN] number of coupling fields
   integer, intent (out) :: field_ids(nbr_fields) !< [OUT] list of field IDs

   call yac_cget_field_ids ( nbr_fields, field_ids )

end subroutine yac_fget_field_ids
! ---------------------------------------------------------------------

subroutine yac_fget_role_of_field_by_id ( field_id, role )

   use mo_yac_finterface, dummy => yac_fget_role_of_field_by_id

   integer, intent (in)  :: field_id  !< [IN]  field identifier
   integer, intent (out) :: role      !< [OUT] role for this field
                                      !<       NOTHING (0) | SOURCE (1) | TARGET (2)

   call yac_cget_role_of_field_by_id ( field_id, role )

end subroutine yac_fget_role_of_field_by_id

! ---------------------------------------------------------------------

subroutine yac_fget_role_of_field_by_name ( field_name, role )

   use mo_yac_finterface, dummy => yac_fget_role_of_field_by_name
   use, intrinsic :: iso_c_binding, only: c_null_char

   character (len=*), intent (in)    :: field_name !< [IN] field name
   integer, intent (out)             :: role       !< [OUT] role for this field
                                                   !<       NOTHING (0) | SOURCE (1) | TARGET (2)

   call yac_check_strlength ( field_name )

   call yac_cget_role_of_field_by_name ( TRIM(field_name) // c_null_char, role )

end subroutine yac_fget_role_of_field_by_name

! ---------------------------------------------------------------------

subroutine yac_fget_field_name_from_id ( field_id, field_name )

   use mo_yac_finterface, dummy => yac_fget_field_name_from_id
   use, intrinsic :: iso_c_binding, only: c_null_char

   integer, intent (in)            :: field_id   !< [IN]  field identifier
   character (len=*), intent (out) :: field_name !< [OUT] field name

   integer :: i

   call yac_check_strlength ( field_name )

   call yac_cget_field_name_from_id ( field_id, field_name )

   do i = 1, len(field_name)
      if (field_name(i:i) == c_null_char) exit
   end do

   field_name(i:len(field_name)) = ' '

end subroutine yac_fget_field_name_from_id

! ---------------------------------------------------------------------

subroutine yac_fget_field_id_from_name ( field_name, field_id )

   use mo_yac_finterface, dummy => yac_fget_field_id_from_name
   use, intrinsic :: iso_c_binding, only: c_null_char

   character (len=*), intent (in) :: field_name !< [IN] field name
   integer, intent (out)          :: field_id   !< [OUT]  field identifier

   call yac_check_strlength ( field_name )

   call yac_cget_field_id_from_name &
                     ( TRIM(field_name) // c_null_char, field_id )

end subroutine yac_fget_field_id_from_name

! ---------------------------------------------------------------------

subroutine yac_fget_model_timestep_id ( field_id, string )

   use mo_yac_finterface, dummy => yac_fget_model_timestep_id
   use, intrinsic :: iso_c_binding, only: c_null_char

   integer, intent (in)            :: field_id !< [IN]  field identifier
   character (len=*), intent (out) :: string   !< [OUT] timestep in iso format

   integer :: i

   call yac_cget_model_timestep_id ( field_id, string )

   do i = 1, len(string)
      if (string(i:i) == c_null_char) exit
   end do

   string(i:len(string)) = ' '

end subroutine yac_fget_model_timestep_id

! ---------------------------------------------------------------------

subroutine yac_fget_model_timestep_name ( field_name, string )

   use mo_yac_finterface, dummy => yac_fget_model_timestep_name
   use, intrinsic :: iso_c_binding, only: c_null_char

   character (len=*), intent (in)  :: field_name !< [IN] field name
   character (len=*), intent (out) :: string     !< [OUT] timestep in iso format

   integer :: i

   call yac_check_strlength ( field_name )

   call yac_cget_model_timestep_name &
                     ( TRIM(field_name) // c_null_char, string )

   do i = 1, len(string)
      if (string(i:i) == c_null_char) exit
   end do

   string(i:len(string)) = ' '

end subroutine yac_fget_model_timestep_name

! ---------------------------------------------------------------------

subroutine yac_fget_coupling_timestep_id ( field_id, string )

   use mo_yac_finterface, dummy => yac_fget_coupling_timestep_id
   use, intrinsic :: iso_c_binding, only: c_null_char

   integer, intent (in)            :: field_id !< [IN]  field identifier
   character (len=*), intent (out) :: string   !< [OUT] timestep in iso format

   integer :: i

   call yac_cget_coupling_timestep_id ( field_id, string )

   do i = 1, len(string)
      if (string(i:i) == c_null_char) exit
   end do

   string(i:len(string)) = ' '

end subroutine yac_fget_coupling_timestep_id

! ---------------------------------------------------------------------

subroutine yac_fget_coupling_timestep_name ( field_name, string )

   use mo_yac_finterface, dummy => yac_fget_coupling_timestep_name
   use, intrinsic :: iso_c_binding, only: c_null_char

   character (len=*), intent (in)  :: field_name !< [IN] field name
   character (len=*), intent (out) :: string     !< [OUT] timestep in iso format

   integer :: i

   call yac_check_strlength ( field_name )

   call yac_cget_coupling_timestep_name &
                     ( TRIM(field_name) // c_null_char, string )

   do i = 1, len(string)
      if (string(i:i) == c_null_char) exit
   end do

   string(i:len(string)) = ' '

end subroutine yac_fget_coupling_timestep_name

! ---------------------------------------------------------------------

subroutine yac_check_strlength ( in_string )

  use mo_yac_finterface

  character (len=*), intent (in)  :: in_string
  character (len=YAC_MAX_CHARLEN) :: message

  if ( LEN_TRIM(in_string) > YAC_MAX_CHARLEN ) then
     message = "string too long"
     print *, "string ", TRIM(in_string), "exceeds length of ", YAC_MAX_CHARLEN
     call yac_abort_message ( message , __FILE__, __LINE__ )
  endif

end subroutine
