/**
 * @file yac_interface.c
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *                                 Teresa Holfeld <teresa.holfeld@zmaw.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 *         Teresa Holfeld <teresa.holfeld@zmaw.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 *             Teresa Holfeld <teresa.holfeld@zmaw.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "communicator.h"
#include "communicator_mpi.h"
#include "couple_config.h"
#include "config_xml.h"
#include "component.h"
#include "dep_list.h"
#include "domain.h"
#include "ensure_array_size.h"
#include "event.h"
#include "fields.h"
#include "field_interpolation_gen.h"
#include "grid.h"
#include "grid_reg2d.h"
#include "grid_unstruct.h"
#include "yac_interface.h"
#include "mask.h"
#include "points.h"
#include "search.h"
#include "utils.h"

int const YAC_LOCATION_CELL = 0;
int const YAC_LOCATION_CORNER = 1;
int const YAC_LOCATION_EDGE = 2;

static int nbr_defined_components = 0;
static int current_components_size = 0;
static struct component ** list_of_components = NULL;

static int nbr_defined_subdomains = 0;
static int current_subdomains_size = 0;
static struct {
  struct subdomain * subdomain;
  unsigned initialised;
  char * name;
  struct {
    double * x_vertices;
    double * y_vertices;
    struct dep_list cell_to_vertex;
    struct grid * grid;
  } grid_data;
  struct {
    int * global_ids;
    unsigned * global_ranks;
  } corner, cell, edge;
} * list_of_subdomains = NULL;

static int nbr_defined_domains = 0;
static int current_domains_size = 0;
static struct domain ** list_of_domains = NULL;

static int nbr_defined_fields = 0;
static int current_field_id_size = 0;
static int *list_of_field_ids = NULL;

static int nbr_defined_points = 0;
static int current_defined_points_size = 0;
static struct {
  double * x_points;
  double * y_points;
  struct points * points;
} * list_of_points = NULL;

static int nbr_defined_masks = 0;
static int current_masks_size = 0;
static struct mask ** list_of_masks = NULL;

static char job_start_datetime[MAX_CALENDAR_STR_LEN];
static char job_end_datetime[MAX_CALENDAR_STR_LEN];

/* ---------------------------------------------------------------------- */

void yac_cinit ( const char * xml_filename, const char * schema_filename ) {

  yac_init ( xml_filename, schema_filename );

  /* initialise start- and end dates with setting from xml.
   * This can be overwritten by calls to yac_cdef_datetime.
   */

#ifdef USE_XML2
  strcpy (job_start_datetime, yac_xml_get_start_date());
  strcpy (job_end_datetime, yac_xml_get_end_date());
#endif

  return;
}

/* ---------------------------------------------------------------------- */

void yac_cfinalize () {

  /* free memory */

  yac_free_couple_config();

  int i;

  for (i = 0; i < nbr_defined_components; ++i)
    yac_component_delete(list_of_components[i]);
  free(list_of_components);

  for ( i = 0; i < nbr_defined_fields; ++ i ) {
    struct coupling_field * cpl_field = yac_unique_id_to_pointer(list_of_field_ids[i]);
    yac_coupling_field_delete ( cpl_field );
  }
  free(list_of_field_ids);
  nbr_defined_fields = 0;
  current_field_id_size = 0;
  list_of_field_ids = NULL;

  for (i = 0; i < nbr_defined_domains; ++ i) {
    yac_free_domain(list_of_domains[i]);
    free(list_of_domains[i]);
  }
  free(list_of_domains);
  nbr_defined_domains = 0;
  current_domains_size = 0;
  list_of_domains = NULL;

  for (i = 0; i < nbr_defined_subdomains; ++i) {
    free(list_of_subdomains[i].corner.global_ids);
    free(list_of_subdomains[i].corner.global_ranks);
    free(list_of_subdomains[i].cell.global_ids);
    free(list_of_subdomains[i].cell.global_ranks);
    free(list_of_subdomains[i].edge.global_ids);
    free(list_of_subdomains[i].edge.global_ranks);
    if (list_of_subdomains[i].grid_data.grid != NULL) {
      free(list_of_subdomains[i].grid_data.x_vertices);
      free(list_of_subdomains[i].grid_data.y_vertices);
      yac_free_dep_list(&(list_of_subdomains[i].grid_data.cell_to_vertex));
      yac_delete_grid(list_of_subdomains[i].grid_data.grid);
    }
    free(list_of_subdomains[i].subdomain);
  }
  free(list_of_subdomains);

  for (i = 0; i < nbr_defined_points; ++i) {
    free(list_of_points[i].x_points);
    free(list_of_points[i].y_points);
    yac_free_points(list_of_points[i].points);
    free(list_of_points[i].points);
  }
  free(list_of_points);
  nbr_defined_points = 0;
  current_defined_points_size = 0;
  list_of_points = NULL;

  for (i = 0; i < nbr_defined_masks; ++i) {
    yac_free_mask(list_of_masks[i]);
    free(list_of_masks[i]);
  }
  free(list_of_masks);
  list_of_masks = NULL;
  nbr_defined_masks = 0;
  current_masks_size = 0;

  /* close MPI */

  yac_finalize();

  return;
}

/* ---------------------------------------------------------------------- */

void yac_cdef_datetime ( const char * start_datetime, const char * end_datetime ) {

#ifdef USE_XML2

  if ( start_datetime == NULL || strlen(start_datetime) == 0 )
      strcpy (job_start_datetime, yac_xml_get_start_date());
  else strcpy (job_start_datetime, start_datetime);

  if ( end_datetime == NULL || strlen(end_datetime) == 0 )
       strcpy (job_end_datetime, yac_xml_get_end_date());
  else strcpy (job_end_datetime, end_datetime );

#endif

  return;
}

/* ---------------------------------------------------------------------- */

char * yac_get_start_datetime () {
  return job_start_datetime;
}

/* ---------------------------------------------------------------------- */

char * yac_get_end_datetime () {
  return job_end_datetime;
}

/* ---------------------------------------------------------------------- */

#ifdef USE_MPI

/* internal routine to serve the Fortran interface */

void yac_get_localcomm ( MPI_Fint *comm, int comp_id ) {

  struct component * comp = yac_unique_id_to_pointer(comp_id);
  MPI_Comm tmp_comm;

  /* replace with communicator generated within init_comp */

  MPI_Comm_dup(yac_get_comm_mpi(yac_component_get_communicator(comp)), &tmp_comm);
  *comm = MPI_Comm_c2f(tmp_comm);
}

/* c interface routine */

void yac_cget_localcomm ( MPI_Comm *comm, int comp_id ) {

  struct component * comp = yac_unique_id_to_pointer(comp_id);

  MPI_Comm_dup(yac_get_comm_mpi(yac_component_get_communicator(comp)), comm);
}


#else

/* internal routine to serve the Fortran interface */

void yac_get_localcomm ( int *comm, int comp_id ) {

  comm = 0;

  return;
}

/* c interface routine */

void yac_cget_localcomm ( int * comm, int comp_id ) {

  comm = 0;

  return;
}

#endif

/* ---------------------------------------------------------------------- */

void yac_cset_comp_io_proc_step ( int comp_id, int io_proc_step ) {

  struct component * comp = yac_unique_id_to_pointer(comp_id);

  yac_component_set_io_proc_step(comp, io_proc_step);
}

/* ---------------------------------------------------------------------- */

void yac_cget_comp_io_proc_step ( int comp_id, int * io_proc_step ) {

  struct component * comp = yac_unique_id_to_pointer(comp_id);

  *io_proc_step = yac_component_get_io_proc_step(comp);
}

/* ---------------------------------------------------------------------- */

#ifdef USE_MPI

/* internal routine to serve the Fortran interface */

void yac_get_pair_rootcomm ( MPI_Fint *comm, const char * comp_name1, const char * comp_name2 ) {
#ifdef USE_XML2
  struct communicator * tmp_comm = yac_get_pair_rootcomm_ ( comp_name1, comp_name2 );
  *comm = MPI_Comm_c2f(yac_get_comm_mpi(tmp_comm));
#else
  *comm = MPI_Comm_c2f(MPI_COMM_NULL);
#endif
}

/* c interface routine */

void yac_cget_pair_rootcomm ( MPI_Comm *comm, const char * comp_name1, const char * comp_name2 ) {

#ifdef USE_XML2
  struct communicator * tmp_comm = yac_get_pair_rootcomm_ ( comp_name1, comp_name2 );
  *comm = yac_get_comm_mpi(tmp_comm);
#else
  *comm = MPI_Comm_c2f(MPI_COMM_NULL);
#endif
}

#else

/* internal routine to serve the Fortran interface */

void yac_get_pair_rootcomm ( int *comm, const char * comp_name1, const char * comp_name2 ) {

  comm = 0;

  return;
}

/* c interface routine */

void yac_cget_pair_rootcomm ( int * comm, const char * comp_name1, const char * comp_name2 ) {

  comm = 0;

  return;
}

#endif

/* ---------------------------------------------------------------------- */

void yac_cdef_comp ( char const * comp_name, int * comp_id ) {

  struct component * component = yac_component_new(comp_name);

  *comp_id = yac_pointer_to_unique_id(component);

  int redirect = yac_xml_redirect_stdout();

  if ( redirect ) {
      int ierr;
      int lenstr = (int) strlen(comp_name);
#ifdef USE_MPI
      struct communicator * comm =
        yac_component_get_communicator(yac_unique_id_to_pointer(*comp_id));
      int rank = yac_get_comm_rank (comm);
      int size = yac_get_comm_size (comm);
      if ( redirect == 1 )
        yac_redirstdout ( comp_name, lenstr, 0, rank, size, &ierr ); // all redirect
      else
        yac_redirstdout ( comp_name, lenstr, 1, rank, size, &ierr ); // only root redirect
#else
      yac_redirstdout ( comp_name, lenstr, 1, 0, 1, &ierr );
#endif
  }

  nbr_defined_components++;
  ENSURE_ARRAY_SIZE(list_of_components, current_components_size, nbr_defined_components);
  list_of_components[nbr_defined_components-1] = component;
}

/* ---------------------------------------------------------------------- */

void
yac_cdef_comps ( char const ** comp_names, int num_comps, int * comp_ids ) {

  struct component * components[num_comps];

  yac_components_new(comp_names, num_comps, components);

  for (int i = 0; i < num_comps; ++i)
    comp_ids[i] = yac_pointer_to_unique_id(components[i]);

  int redirect = yac_xml_redirect_stdout();

  if ( redirect ) {
      int ierr;
      int lenstr = (int) strlen(comp_names[0]);
#ifdef USE_MPI
      struct communicator * comm =
        yac_component_get_communicator(yac_unique_id_to_pointer(comp_ids[0]));
      int rank = yac_get_comm_rank (comm);
      int size = yac_get_comm_size (comm);
      if ( redirect == 1 )
        yac_redirstdout ( comp_names[0], lenstr, 0, rank, size, &ierr ); // all redirect
      else
        yac_redirstdout ( comp_names[0], lenstr, 1, rank, size, &ierr ); // only root redirect
#else
      yac_redirstdout ( comp_names[0], lenstr, 1, 0, 1, &ierr );
#endif
  }

  ENSURE_ARRAY_SIZE(list_of_components, current_components_size,
                    nbr_defined_components + num_comps);
  for (int i = 0; i < num_comps; ++i, ++nbr_defined_components)
    list_of_components[nbr_defined_components] = components[i];
}

/* ---------------------------------------------------------------------- */

void yac_cdef_subdomain ( int const comp_id,
                          char const * subdomain_name,
                          int * subdomain_id ) {

  unsigned subdomain_idx = nbr_defined_subdomains;
  nbr_defined_subdomains++;
  ENSURE_ARRAY_SIZE(list_of_subdomains, current_subdomains_size,
                    nbr_defined_subdomains);
  list_of_subdomains[subdomain_idx].subdomain =
    malloc(sizeof(*(list_of_subdomains[subdomain_idx].subdomain)));
  list_of_subdomains[subdomain_idx].initialised = 1 == 0;
  list_of_subdomains[subdomain_idx].name = malloc(strlen(subdomain_name)+1);

  strcpy(list_of_subdomains[subdomain_idx].name, subdomain_name);

  list_of_subdomains[subdomain_idx].grid_data.grid = NULL;

  list_of_subdomains[subdomain_idx].corner.global_ids   = NULL;
  list_of_subdomains[subdomain_idx].corner.global_ranks = NULL;
  list_of_subdomains[subdomain_idx].cell.global_ids     = NULL;
  list_of_subdomains[subdomain_idx].cell.global_ranks   = NULL;
  list_of_subdomains[subdomain_idx].edge.global_ids     = NULL;
  list_of_subdomains[subdomain_idx].edge.global_ranks   = NULL;

  *subdomain_id = subdomain_idx;
}

/* ---------------------------------------------------------------------- */

void yac_cconnect_subdomains ( int const comp_id,
                               int const nbr_subdomain_ids,
                               int const * subdomain_ids,
                               int * domain_id ) {

  if (nbr_subdomain_ids <= 0)
    yac_internal_abort_message ( "ERROR: subdomain index out of range.", __FILE__, __LINE__ );

  struct subdomain ** subdomains =
    malloc(nbr_subdomain_ids * sizeof(*subdomains));

  char * domain_name = list_of_subdomains[subdomain_ids[0]].name;

  struct communicator * component_comm =
    yac_component_get_communicator(yac_unique_id_to_pointer(comp_id));

  for (int i = 0; i < nbr_subdomain_ids; i++ ) {

    unsigned subdomain_idx = subdomain_ids[i];

    subdomains[i] = list_of_subdomains[subdomain_idx].subdomain;

    if (strcmp(domain_name, list_of_subdomains[subdomain_idx].name))
      yac_internal_abort_message("ERROR: domain names do not match.",
                                 __FILE__, __LINE__);

    if (list_of_subdomains[subdomain_idx].initialised)
      yac_internal_abort_message("Subdomain used for more than one domain.",
                                 __FILE__, __LINE__ );

    if (i != 0) free(list_of_subdomains[subdomain_idx].name);
    list_of_subdomains[subdomain_idx].initialised = 1 == 1;

    yac_init_subdomain_unstruct(
      list_of_subdomains[subdomain_idx].subdomain,
      list_of_subdomains[subdomain_idx].grid_data.grid, component_comm,
      list_of_subdomains[subdomain_idx].corner.global_ids,
      list_of_subdomains[subdomain_idx].corner.global_ranks,
      list_of_subdomains[subdomain_idx].cell.global_ids,
      list_of_subdomains[subdomain_idx].cell.global_ranks,
      list_of_subdomains[subdomain_idx].edge.global_ids,
      list_of_subdomains[subdomain_idx].edge.global_ranks);
  }

  // routines to connect the grids

  struct domain * domain = malloc(1 * sizeof(*domain));

  yac_init_domain(domain, component_comm, subdomains, nbr_subdomain_ids,
                  domain_name);

  free(subdomains);

  /**
   *  \todo
   *  - The enddef_domain call should be moved to a dedicated enddef routine
   *    for the initialisation process.
   **/

  yac_enddef_domain(domain);

  free(domain_name);

  *domain_id = yac_pointer_to_unique_id(domain);

  nbr_defined_domains++;
  ENSURE_ARRAY_SIZE(list_of_domains, current_domains_size, nbr_defined_domains);
  list_of_domains[nbr_defined_domains-1] = domain;

  return;
}

/* ---------------------------------------------------------------------- */

void yac_cdef_points_reg2d ( int const subdomain_id,
                             int const *nbr_points,
                             int const located,
                             double const *x_points,
                             double const *y_points,
                             int *point_id ) {

  enum yac_location location;
  unsigned subdomain_idx = subdomain_id;

  unsigned point_array_size[2];

  location = yac_get_location(located);

  double * user_x_points = NULL;
  double * user_y_points = NULL;

  if (list_of_subdomains[subdomain_idx].grid_data.grid == NULL)
    yac_internal_abort_message(
      "ERROR: yac_cdef_points_reg2d: grid not yet defined\n",
      __FILE__, __LINE__);

  yac_get_coordinate_array_sizes(
    list_of_subdomains[subdomain_idx].grid_data.grid, location,
    point_array_size);

  if (x_points != NULL) {
    user_x_points = malloc(point_array_size[0] * sizeof(*user_x_points));
    memcpy(user_x_points, x_points, point_array_size[0] * sizeof(*user_x_points));
    for ( unsigned i = 0; i < point_array_size[0]; ++i ) user_x_points[i] *= rad;
  }

  if (y_points != NULL) {
    user_y_points = malloc(point_array_size[1] * sizeof(*user_y_points));
    memcpy(user_y_points, y_points, point_array_size[1] * sizeof(*user_y_points));
    for ( unsigned i = 0; i < point_array_size[1]; ++i ) user_y_points[i] *= rad;
  }

  struct points * points = malloc(1 * sizeof(points[0]));

  yac_init_points(points, list_of_subdomains[subdomain_idx].grid_data.grid,
                  location, user_x_points, user_y_points);

  *point_id = yac_pointer_to_unique_id(points);

  nbr_defined_points++;
  ENSURE_ARRAY_SIZE(list_of_points, current_defined_points_size, nbr_defined_points);
  list_of_points[nbr_defined_points-1].x_points = user_x_points;
  list_of_points[nbr_defined_points-1].y_points = user_y_points;
  list_of_points[nbr_defined_points-1].points = points;
}

/* ---------------------------------------------------------------------- */

void yac_cdef_points_unstruct ( int const subdomain_id,
                                int const nbr_points,
                                int const located,
                                double const *x_points,
                                double const *y_points,
                                int *point_id ) {

  enum yac_location location;
  unsigned subdomain_idx = subdomain_id;

  unsigned point_array_size[2];

  location = yac_get_location(located);

  double * user_x_points = NULL;
  double * user_y_points = NULL;

  if (list_of_subdomains[subdomain_idx].grid_data.grid == NULL)
    yac_internal_abort_message(
      "ERROR: yac_cdef_points_unstruct: grid not yet defined\n",
      __FILE__, __LINE__);
  yac_get_coordinate_array_sizes(
    list_of_subdomains[subdomain_idx].grid_data.grid, location,
    point_array_size);

  if (x_points != NULL) {
    user_x_points = malloc( nbr_points * sizeof (user_x_points[0]));
    memcpy(user_x_points, x_points, nbr_points * sizeof(user_x_points[0]));
    for ( unsigned i = 0; i < nbr_points; ++i ) user_x_points[i] *= rad;
  }

  if (y_points != NULL) {
    user_y_points = malloc( nbr_points * sizeof (user_y_points[0]));
    memcpy(user_y_points, y_points, nbr_points * sizeof(user_y_points[0]));
    for ( unsigned i = 0; i < nbr_points; ++i ) user_y_points[i] *= rad;
  }

  struct points * points = malloc(1 * sizeof(points[0]));

  yac_init_points(points, list_of_subdomains[subdomain_idx].grid_data.grid,
                  location, user_x_points, user_y_points);

  *point_id = yac_pointer_to_unique_id(points);

  nbr_defined_points++;
  ENSURE_ARRAY_SIZE(list_of_points, current_defined_points_size, nbr_defined_points);
  list_of_points[nbr_defined_points-1].x_points = user_x_points;
  list_of_points[nbr_defined_points-1].y_points = user_y_points;
  list_of_points[nbr_defined_points-1].points = points;
}

/* ---------------------------------------------------------------------- */

void yac_cdef_elements_unstruct ( int const subdomain_id,
                                  int const nbr_vertices,
                                  int const nbr_cells,
                                  int *num_vertices_per_cell_in,
                                  double *x_vertices,
                                  double *y_vertices,
                                  int *cell_to_vertex_in ) {
  int i;
  int len;

  double *user_x_vertices;
  double *user_y_vertices;

  unsigned int *num_vertices_per_cell;
  unsigned int *cell_to_vertex_dep;

  struct dep_list cell_to_vertex;

  unsigned subdomain_idx = subdomain_id;

  if (list_of_subdomains[subdomain_idx].initialised)
    yac_internal_abort_message(
      "Domain was built before grid definition was finished.",
      __FILE__, __LINE__ );

  /* store connectivity data data */

  num_vertices_per_cell = (unsigned *)num_vertices_per_cell_in;

  len = 0;

  for ( i=0; i < nbr_cells; i++ ) {
    len = len + num_vertices_per_cell_in[i];
  }

  num_vertices_per_cell = malloc ( nbr_cells * sizeof (num_vertices_per_cell[0]) );
  cell_to_vertex_dep = malloc ( len * sizeof (cell_to_vertex_dep[0]) );

  memcpy(cell_to_vertex_dep, cell_to_vertex_in,
         len * sizeof (cell_to_vertex_dep[0]));
  memcpy(num_vertices_per_cell, num_vertices_per_cell_in,
         nbr_cells * sizeof (num_vertices_per_cell[0]));

  yac_init_dep_list ( &cell_to_vertex );

  yac_set_dependencies ( &cell_to_vertex,
                         nbr_cells,
                         num_vertices_per_cell,
                         cell_to_vertex_dep );

  /* store vertex coordinates */

  user_x_vertices = malloc( nbr_vertices * sizeof (*user_x_vertices) );
  user_y_vertices = malloc( nbr_vertices * sizeof (*user_y_vertices) );

  memcpy(user_x_vertices, x_vertices,
         nbr_vertices * sizeof (user_x_vertices[0]));
  memcpy(user_y_vertices, y_vertices,
         nbr_vertices * sizeof (user_y_vertices[0]));

  for ( unsigned i = 0; i < nbr_vertices; ++i ) {
     user_x_vertices[i] *= rad;
     user_y_vertices[i] *= rad;
  }

  list_of_subdomains[subdomain_idx].grid_data.grid =
    yac_unstruct_grid_new(user_x_vertices, user_y_vertices, nbr_vertices,
                          cell_to_vertex);
  list_of_subdomains[subdomain_idx].grid_data.x_vertices = user_x_vertices;
  list_of_subdomains[subdomain_idx].grid_data.y_vertices = user_y_vertices;
  list_of_subdomains[subdomain_idx].grid_data.cell_to_vertex = cell_to_vertex;

  return;
}

/* ---------------------------------------------------------------------- */

void yac_cdef_elements_reg2d ( int const subdomain_id,
                               int const nbr_vertices[2],
                               int const cyclic[2],
                               double *x_vertices,
                               double *y_vertices ) {

  unsigned subdomain_idx = subdomain_id;

  if (list_of_subdomains[subdomain_idx].initialised)
    yac_internal_abort_message(
      "Domain was built before grid definition was finished.",
      __FILE__, __LINE__);

  double *user_x_vertices;
  double *user_y_vertices;
  unsigned user_nbr_cells[2];
  unsigned user_cyclic[2];

  user_nbr_cells[0] = (unsigned)nbr_vertices[0] - (cyclic[0]?0:1);
  user_nbr_cells[1] = (unsigned)nbr_vertices[1] - (cyclic[1]?0:1);
  user_cyclic[0] = (unsigned)cyclic[0];
  user_cyclic[1] = (unsigned)cyclic[1];

  /* store vertex coordinates */

  unsigned x_vertices_size, y_vertices_size;

  x_vertices_size = nbr_vertices[0] * sizeof(*user_x_vertices);
  y_vertices_size = nbr_vertices[1] * sizeof(*user_y_vertices);

  user_x_vertices = malloc(x_vertices_size);
  user_y_vertices = malloc(y_vertices_size);

  memcpy(user_x_vertices, x_vertices, x_vertices_size);
  memcpy(user_y_vertices, y_vertices, y_vertices_size);

  for ( unsigned i = 0; i < nbr_vertices[0]; ++i ) user_x_vertices[i] *= rad;
  for ( unsigned i = 0; i < nbr_vertices[1]; ++i ) user_y_vertices[i] *= rad;

  list_of_subdomains[subdomain_idx].grid_data.grid =
    yac_reg2d_grid_new(user_x_vertices, user_y_vertices, user_nbr_cells,
                       user_cyclic);
  list_of_subdomains[subdomain_idx].grid_data.x_vertices = user_x_vertices;
  list_of_subdomains[subdomain_idx].grid_data.y_vertices = user_y_vertices;
  yac_init_dep_list(
    &(list_of_subdomains[subdomain_idx].grid_data.cell_to_vertex));
}

/* ---------------------------------------------------------------------- */

void yac_cdef_mask ( int const * mask_data,
                     int points_id,
                     int * mask_id ) {

  struct mask * mask;
  struct points * points;

  points = (struct points*)yac_unique_id_to_pointer(points_id);

  mask = malloc(1 * sizeof (*mask) );
  yac_init_mask (mask, (unsigned const *)mask_data, points);

  *mask_id = yac_pointer_to_unique_id(mask);

  nbr_defined_masks++;
  ENSURE_ARRAY_SIZE(list_of_masks, current_masks_size, nbr_defined_masks);
  list_of_masks[nbr_defined_masks-1] = mask;
}

/* ---------------------------------------------------------------------- */

void yac_cdef_index_location ( int const subdomain_id,
                               int const nbr_entries,
                               int const location,
                               int const * global_index,
                               int const * global_index_rank ) {

  unsigned subdomain_idx = subdomain_id;

  if (list_of_subdomains[subdomain_idx].initialised)
    yac_internal_abort_message(
      "Domain was built before global id definition was finished.",
      __FILE__, __LINE__);

  switch (location) {

  case (CORNER):

    if (list_of_subdomains[subdomain_idx].corner.global_ids != NULL)
      free(list_of_subdomains[subdomain_idx].corner.global_ids);
    if (list_of_subdomains[subdomain_idx].corner.global_ranks != NULL)
      free(list_of_subdomains[subdomain_idx].corner.global_ranks);
    list_of_subdomains[subdomain_idx].corner.global_ids   =
      malloc(nbr_entries *
             sizeof(list_of_subdomains[subdomain_idx].corner.global_ids[0]));
    list_of_subdomains[subdomain_idx].corner.global_ranks =
      malloc(nbr_entries *
             sizeof(list_of_subdomains[subdomain_idx].corner.global_ranks[0]));

    memcpy(list_of_subdomains[subdomain_idx].corner.global_ids,
           global_index, nbr_entries *
           sizeof(list_of_subdomains[subdomain_idx].corner.global_ids[0]));

    memcpy(list_of_subdomains[subdomain_idx].corner.global_ranks,
           global_index_rank, nbr_entries *
           sizeof(list_of_subdomains[subdomain_idx].corner.global_ranks[0]));
    break;

  case (CELL):

    if (list_of_subdomains[subdomain_idx].cell.global_ids != NULL)
      free(list_of_subdomains[subdomain_idx].cell.global_ids);
    if (list_of_subdomains[subdomain_idx].cell.global_ranks != NULL)
      free(list_of_subdomains[subdomain_idx].cell.global_ranks);
    list_of_subdomains[subdomain_idx].cell.global_ids   =
      malloc(nbr_entries *
             sizeof(list_of_subdomains[subdomain_idx].cell.global_ids[0]));
    list_of_subdomains[subdomain_idx].cell.global_ranks =
      malloc(nbr_entries *
             sizeof(list_of_subdomains[subdomain_idx].cell.global_ranks[0]));

    memcpy(list_of_subdomains[subdomain_idx].cell.global_ids,
           global_index, nbr_entries *
           sizeof(list_of_subdomains[subdomain_idx].cell.global_ids[0]));
          
    memcpy(list_of_subdomains[subdomain_idx].cell.global_ranks,
           global_index_rank, nbr_entries *
           sizeof(list_of_subdomains[subdomain_idx].cell.global_ranks[0]));
    break;

  case (EDGE):

    if (list_of_subdomains[subdomain_idx].edge.global_ids != NULL)
      free(list_of_subdomains[subdomain_idx].edge.global_ids);
    if (list_of_subdomains[subdomain_idx].edge.global_ranks != NULL)
      free(list_of_subdomains[subdomain_idx].edge.global_ranks);
    list_of_subdomains[subdomain_idx].edge.global_ids   =
      malloc(nbr_entries *
             sizeof(list_of_subdomains[subdomain_idx].edge.global_ids[0]));
    list_of_subdomains[subdomain_idx].edge.global_ranks =
      malloc(nbr_entries *
             sizeof(list_of_subdomains[subdomain_idx].edge.global_ranks[0]));

    memcpy(list_of_subdomains[subdomain_idx].edge.global_ids,
           global_index, nbr_entries *
           sizeof(list_of_subdomains[subdomain_idx].edge.global_ids[0]));
          
    memcpy(list_of_subdomains[subdomain_idx].edge.global_ranks,
           global_index_rank, nbr_entries *
           sizeof(list_of_subdomains[subdomain_idx].edge.global_ranks[0]));
    break;
  }

  return;
}

/* ---------------------------------------------------------------------- */

void yac_cdef_field_nomask ( char const * name,
                             int const component_id,
                             int const domain_id,
                             int const * point_ids,
                             int const num_pointsets,
                             int * field_id ) {

  yac_cdef_field(name, component_id, domain_id, point_ids, NULL, num_pointsets,
                 field_id);
}

void yac_cdef_field ( char const * name,
                      int const component_id,
                      int const domain_id,
                      int const * point_ids,
                      int const * mask_ids,
                      int const num_pointsets,
                      int * field_id ) {

  unsigned i;
  int j;

  struct domain * domain;

  domain = yac_unique_id_to_pointer(domain_id);

  char * field_name;
  field_name = malloc((strlen(name) + 1) * sizeof(*field_name));
  strcpy(field_name,name);

  unsigned nbr_subdomains;
  nbr_subdomains = yac_get_num_subdomains(domain);

  struct points *** points;

  if (point_ids != NULL) {
    points = malloc(nbr_subdomains * sizeof(*points));
    for (i = 0; i < nbr_subdomains; ++i) {
      points[i] = malloc(num_pointsets * sizeof(**points));
      for (j = 0; j < num_pointsets; ++j)
        points[i][j] =
          yac_unique_id_to_pointer(point_ids[i * num_pointsets + j]);
    }
  } else {
    points = NULL;
  }

  struct mask *** masks;

  if (mask_ids != NULL) {
    masks = malloc(nbr_subdomains * sizeof(*masks));
    for (i = 0; i < nbr_subdomains; ++i) {
      masks[i] = malloc(num_pointsets * sizeof(**points));
      for (j = 0; j < num_pointsets; ++j)
        masks[i][j] =
          yac_unique_id_to_pointer(mask_ids[i * num_pointsets + j]);
    }
  } else {
    masks = NULL;
  }

  struct coupling_field * cpl_field =
    yac_coupling_field_new(field_name, component_id, domain, points, masks,
                           num_pointsets);

  if (points != NULL)
    for (unsigned i = 0; i < nbr_subdomains; ++i) free(points[i]);
  if (masks != NULL)
    for (unsigned i = 0; i < nbr_subdomains; ++i) free(masks[i]);
  free(points);
  free(masks);
  free(field_name);

  *field_id = yac_pointer_to_unique_id(cpl_field);

  nbr_defined_fields++;
  ENSURE_ARRAY_SIZE(list_of_field_ids, current_field_id_size, nbr_defined_fields);
  list_of_field_ids[nbr_defined_fields-1] = *field_id;
}

/* ---------------------------------------------------------------------- */

void yac_cget_ ( int const field_id,
                 int const collection_size,
                 double *recv_field,   // recv_field[collection_size*nSubdomains]
                                       //           [Points]
                 int    *info,
                 int    *ierr ) {

  /* Needed to transfer from Fortran data structure to C */

  struct coupling_field *cpl_field;

  double ***recv_field_;

  int num_subdomains, num_points;

  cpl_field = yac_unique_id_to_pointer(field_id);

  num_subdomains = yac_get_num_subdomains(yac_get_coupling_field_domain(cpl_field));

  /* we assume that the data coming from Fortran have the same size for each subdomain */

  num_points = yac_get_data_size(***yac_get_coupling_field_pointsets(cpl_field));

  recv_field_ = malloc(collection_size * sizeof(*recv_field_));
  for (int i = 0; i < collection_size; ++i)
    recv_field_[i] = malloc(num_subdomains * sizeof(**recv_field_));

  for (int i = 0; i < collection_size; ++i) {
    for (int j = 0; j < num_subdomains; j++) {
      recv_field_[i][j] = recv_field;
      recv_field += num_points;
    }
  }

  yac_cget ( field_id, collection_size, recv_field_, info, ierr );

  for (int i = 0; i < collection_size; ++i)
    free(recv_field_[i]);
  free(recv_field_);
}

/* ---------------------------------------------------------------------- */

void yac_cget_ptr_ ( int const field_id,
                     int const collection_size,
                     double ** recv_field,  // recv_field[collection_size *
                                            //            nSubdomains][Points]
                     int    *info,
                     int    *ierr ) {

  /* Needed to transfer from Fortran data structure to C */

  struct coupling_field *cpl_field;

  double ***recv_field_;

  int num_subdomains;

  cpl_field = yac_unique_id_to_pointer(field_id);

  num_subdomains = yac_get_num_subdomains(yac_get_coupling_field_domain(cpl_field));

  /* we assume that the data coming from Fortran have the same size for each subdomain */

  recv_field_ = malloc(collection_size * sizeof(*recv_field));

  for (unsigned i = 0; i < collection_size; ++i) {

   recv_field_[i] = malloc(num_subdomains * sizeof(**recv_field_));

    for (unsigned j = 0; j < num_subdomains; ++j) {

      recv_field_[i][j] = *recv_field;
      ++recv_field;
    }
  }


  yac_cget ( field_id, collection_size, recv_field_, info, ierr );

  for (unsigned i = 0; i < collection_size; ++i)
    free(recv_field_[i]);
  free(recv_field_);
}

/* ---------------------------------------------------------------------- */

void yac_cget ( int const field_id,
                int collection_size,
                double *** recv_field, // recv_field[collection_size *
                                       //            nSubdomains][Points]
                int    *info,
                int    *ierr ) {

  struct coupling_field *cpl_field;

  *ierr = -1;
  *info = NONE;

#ifndef USE_XML2
  yac_internal_abort_message("ERROR: Compiled without xml support! "
                             "Coupling is not possible.", __FILE__, __LINE__);
#endif

  cpl_field = yac_unique_id_to_pointer(field_id);

  if (yac_get_coupling_field_exchange_type(cpl_field) != TARGET)
    return;

  /* --------------------------------------------------------------------
     Check for restart and coupling events
     -------------------------------------------------------------------- */

  int event_id = yac_get_coupling_field_get_op_event_id(cpl_field);

  int action = yac_event_check ( event_id );

  /* --------------------------------------------------------------------
     add one model time step (as provided in coupling.xml) to the current event date
     -------------------------------------------------------------------- */

  yac_event_update ( event_id );

  /* ------------------------------------------------------------------
     return in case we are already beyond the end of the run
     ------------------------------------------------------------------ */

  if ( action == OUT_OF_BOUND ) {

      fprintf (stderr, "WARNING: yac get action is beyond end of run date!\n");
      fprintf (stdout, "WARNING: yac get action is beyond end of run date!\n");

      *info = action;
      *ierr = 0;

      return;
  }

  /* --------------------------------------------------------------------
     start actions
     -------------------------------------------------------------------- */

  if ( action == NONE ) {
    *info = action;
    *ierr = 0;

    return;
  }

  if ( collection_size != yac_get_coupling_field_collection_size(cpl_field) )
    yac_internal_abort_message(
      "ERROR: collection size does not match with xml.", __FILE__, __LINE__);

  if ( action == RESTART ) action = GET_FOR_RESTART;

  struct field_interpolation * interpolation =
    yac_get_coupling_field_get_op_interpolation(cpl_field);

  //for (int i = 0; i < collection_size; ++i)
  yac_get_field(interpolation, recv_field, collection_size);

  *info = action;
  *ierr = 0;

  return;
}

/* ---------------------------------------------------------------------- */


void yac_cput_ ( int const field_id,
                 int const collection_size,
                 double *send_field,   // send_field[nSubdomains][nPointSets][Points]
                 int    *info,
                 int    *ierr ) {

  /* Needed to transfer from Fortran data structure to C */

  struct coupling_field *cpl_field;

  double **** send_field_;

  int num_subdomains, num_pointsets, *num_points;

  cpl_field = yac_unique_id_to_pointer(field_id);

  num_subdomains = yac_get_num_subdomains(yac_get_coupling_field_domain(cpl_field));
  num_pointsets  = yac_get_coupling_field_num_pointsets(cpl_field);

  num_points = malloc(num_pointsets * sizeof(*num_points));

  /* we assume that the data coming from Fortran have the same size for each subdomain */

  struct points *** points = yac_get_coupling_field_pointsets(cpl_field);

  for (unsigned j = 0; j < num_pointsets; j++ ) {
    num_points[j] = yac_get_data_size(*(points[0][j]));

    for (unsigned i = 0; i < num_subdomains; ++i)
      if (yac_get_data_size(*(points[i][j])) != num_points[j])
        yac_internal_abort_message("ERROR: yac_cput: point sets of different subdomains "
                      "differ in size", __FILE__, __LINE__);
  }

  send_field_ = malloc(collection_size * sizeof(*send_field_));

  for (unsigned i = 0; i < collection_size; ++i) {

    send_field_[i] = malloc(num_subdomains * sizeof(**send_field_));

    for (unsigned j = 0; j < num_subdomains; ++j) {

      send_field_[i][j] = malloc(num_pointsets * sizeof(***send_field_));

      for (unsigned k = 0; k < num_pointsets; ++k) {

        send_field_[i][j][k] = send_field;
        send_field += num_points[k];
      }
    }
  }

  /* We do not have to allocate a third dimension here. It should be sufficient
     to pass the pointer. Upon return the real data should have been saved by
     the comm_bsend of the MPI if not already communicated to the remote process */

  yac_cput ( field_id, collection_size, send_field_, info, ierr );

  for (unsigned i = 0; i < collection_size; ++i) {
    for (unsigned j = 0; j < num_subdomains; ++j) {
      free(send_field_[i][j]);
    }
    free(send_field_[i]);
  }
  free(send_field_);

  free(num_points);
}

/* ---------------------------------------------------------------------- */


void yac_cput_ptr_ ( int const field_id,
                     int const collection_size,
                     double ** send_field,   // send_field[collection_size *
                                             //            nSubdomains *
                                             //            nPointSets][Points]
                     int    *info,
                     int    *ierr ) {

  /* Needed to transfer from Fortran data structure to C */

  struct coupling_field *cpl_field;

  double **** send_field_;

  int num_subdomains, num_pointsets;

  cpl_field = yac_unique_id_to_pointer(field_id);

  num_subdomains = yac_get_num_subdomains(yac_get_coupling_field_domain(cpl_field));
  num_pointsets  = yac_get_coupling_field_num_pointsets(cpl_field);

  send_field_ = malloc(collection_size * sizeof(*send_field_));

  for (unsigned h = 0; h < collection_size; ++h) {

    send_field_[h] = malloc(num_subdomains * sizeof(**send_field_));

    for (unsigned i = 0; i < num_subdomains; i++) {

      send_field_[h][i] = malloc(num_pointsets * sizeof(***send_field_));

      for (unsigned j = 0; j < num_pointsets; j++) {
        send_field_[h][i][j] = *send_field;
        send_field++;
      }
    }
  }

  /* We do not have to allocate a third dimension here. It should be sufficient
     to pass the pointer. Upon return the real data should have been saved by
     the comm_bsend of the MPI if not already communicated to the remote process */

  yac_cput ( field_id, collection_size, send_field_, info, ierr );

  for (unsigned h = 0; h < collection_size; ++h) {
    for (unsigned i = 0; i < num_subdomains; ++i)
      free(send_field_[h][i]);
    free(send_field_[h]);
  }
  free(send_field_);

}

/* ---------------------------------------------------------------------- */

void yac_cput ( int const field_id,
                int const collection_size,
                double ****send_field,   // send_field[collection_size]
                                         //           [nSubdomains]
                                         //           [nPointSets][Points]
                int    *info,
                int    *ierr ) {

  struct coupling_field *cpl_field;

  *info = NONE;
  *ierr = -1;

#ifndef USE_XML2
  yac_internal_abort_message("ERROR: Compiled without xml support! "
                             "Coupling is not possible.", __FILE__, __LINE__);
#endif

  unsigned i, j, m;

  static int action;

  cpl_field = yac_unique_id_to_pointer(field_id);

  int num_subdomains = yac_get_num_subdomains(yac_get_coupling_field_domain(cpl_field));
  int num_pointsets = yac_get_coupling_field_num_pointsets(cpl_field);
  int * num_points = NULL;

  if (yac_get_coupling_field_exchange_type(cpl_field) != SOURCE)
    return;

  struct points *** points = yac_get_coupling_field_pointsets(cpl_field);

  for (unsigned put_idx = 0; put_idx < yac_get_coupling_field_num_puts(cpl_field);
       ++put_idx) {

    /* ------------------------------------------------------------------
       Check for restart and coupling events
       ------------------------------------------------------------------ */

    int event_id = yac_get_coupling_field_put_op_event_id(cpl_field, put_idx);

    action = yac_event_check ( event_id );

    /* ------------------------------------------------------------------
       add one model time step (as provided in coupling.xml) to the current event date
       ------------------------------------------------------------------ */

    yac_event_update ( event_id );

    /* ------------------------------------------------------------------
       return in case we are already beyond the end of the run
       ------------------------------------------------------------------ */

    if ( action == OUT_OF_BOUND ) {

        fprintf (stderr, "WARNING: put action is beyond end of run date!\n");
        fprintf (stdout, "WARNING: put action is beyond end of run date!\n");

        *info = action;
        *ierr = 0;

        return;
    }

    /* ------------------------------------------------------------------
       start actions
       ------------------------------------------------------------------ */

    int time_operation = yac_get_event_time_operation(event_id);

    if ( action == NONE && time_operation == INSTANT ) continue;

    /* ------------------------------------------------------------------
       If it is time for restart, set appropriate flags and continue
       ------------------------------------------------------------------ */

    if ( action == RESTART ) {
      *info = MAX (*info, PUT_FOR_RESTART);
      *ierr = 0;
    }

    struct field_interpolation * interpolation =
      yac_get_coupling_field_put_op_interpolation(cpl_field, put_idx);

    if ( collection_size != yac_get_coupling_field_collection_size(cpl_field) )
      yac_internal_abort_message(
        "ERROR: collection size does not match with xml.", __FILE__, __LINE__ );

    /* ------------------------------------------------------------------
       First deal with instant sends
       ------------------------------------------------------------------ */

    if ( time_operation == INSTANT ) {

      yac_put_field ( interpolation, send_field, collection_size );

      *info = MAX (*info, COUPLING);
      *ierr = 0;

      continue;
    }

    /* ------------------------------------------------------------------
       Accumulation & Averaging
       ------------------------------------------------------------------ */

    /* prepare for accumulation */

    if (num_points == NULL) {

      num_points = malloc(num_subdomains * num_pointsets * sizeof(*num_points));

      for ( i = 0; i < num_subdomains; i++) {
        for ( j = 0; j < num_pointsets; j++ ) {
          num_points[i*num_pointsets+j] = yac_get_data_size(*(points[i][j]));
        }
      }
    }

    /* accumulate data */

    double **** send_field_acc =
      yac_get_coupling_field_put_op_send_field_acc(cpl_field, put_idx);

    int time_accumulation_count =
      yac_get_coupling_field_put_op_time_accumulation_count(cpl_field, put_idx);

    time_accumulation_count++;

    for (unsigned h = 0; h < collection_size; h++) {
      for ( i = 0; i < num_subdomains; i++) {
        for ( j = 0; j < num_pointsets; j++ ) {
          for ( m = 0; m < num_points[i*num_pointsets+j]; m++ ) {
            send_field_acc[h][i][j][m] += send_field[h][i][j][m];
          }
        }
      }
    }

  /* --------------------------------------------------------------------
     Check whether we have to perform the coupling in this call
     -------------------------------------------------------------------- */

    if ( action == NONE ) {
      yac_set_coupling_field_put_op_time_accumulation_count(
        cpl_field, put_idx, time_accumulation_count);
      continue;
    }

  /* --------------------------------------------------------------------
     Average data if required
     -------------------------------------------------------------------- */

    if ( time_operation == TIME_AVERAGE ) {
      double weight = 1.0 / time_accumulation_count;

      for (unsigned h = 0; h < collection_size; ++h) {
        for ( i = 0; i < num_subdomains; i++) {
          for ( j = 0; j < num_pointsets; j++ ) {
            for ( m = 0; m < num_points[i*num_pointsets+j]; m++ ) {
              send_field_acc[h][i][j][m] *= weight;
            }
          }
        }
      }
    }

  /* --------------------------------------------------------------------
     Apply interpolation and send away
     -------------------------------------------------------------------- */

      yac_put_field ( interpolation, send_field_acc, collection_size );

      *info = MAX (*info, COUPLING);

  /* --------------------------------------------------------------------
     Initialise memory for the next accumulation
     -------------------------------------------------------------------- */

    yac_init_coupling_field_put_op_send_field_acc(cpl_field, put_idx, 0.0);
    yac_set_coupling_field_put_op_time_accumulation_count(cpl_field, put_idx, 0);

  }

  free (num_points);

  *ierr = 0;

  return;
}

/* ---------------------------------------------------------------------- */

void yac_csearch ( int nbr_comps,
                   int * comp_ids,
                   int nbr_fields,
                   int * field_ids,
                   int * ierror ) {

  struct component ** components = malloc(nbr_comps * sizeof(*components));
  for (int i = 0; i < nbr_comps; ++i)
    components[i] = yac_unique_id_to_pointer(comp_ids[i]);
  struct coupling_field ** fields = malloc(nbr_fields * sizeof(*fields));
  for (int i = 0; i < nbr_fields; ++i)
    fields[i] = yac_unique_id_to_pointer(field_ids[i]);
// fputs("MoHa before yac_setup_couple_config\n", stderr);
  yac_setup_couple_config(nbr_comps, components, nbr_fields, fields);
// fputs("MoHa before yac_enddef_couple_config\n", stderr);
  yac_enddef_couple_config();
// fputs("MoHa after yac_enddef_couple_config\n", stderr);

  free(components);
  free(fields);

  *ierror = 0;
}

/* ----------------------------------------------------------------------
                   query functions
   ----------------------------------------------------------------------*/

void yac_cget_nbr_fields ( int *nbr_fields ) {
  *nbr_fields = nbr_defined_fields;
  return;
}

/* ---------------------------------------------------------------------- */

void yac_cget_field_ids ( int nbr_fields, int *field_ids ) {
  if ( nbr_fields != nbr_defined_fields )
    yac_internal_abort_message ( "ERROR: nbr of fields does not match.", __FILE__, __LINE__ );
  for ( int i = 0; i < nbr_defined_fields; ++i )
    field_ids[i] = list_of_field_ids[i];
  return;
}

/* ---------------------------------------------------------------------- */

void yac_cget_field_id_from_name ( const char *field_name,
                                   int *field_id ) {

  int i;

  for ( i = 0; i < nbr_defined_fields; ++i ) {
    struct coupling_field * field = yac_unique_id_to_pointer(list_of_field_ids[i]);
    if ( !strcmp(yac_get_coupling_field_name(field),field_name) ) {
      *field_id = list_of_field_ids[i];
      break;
    }
  }

  if ( i == nbr_defined_fields )
    yac_internal_abort_message ( "ERROR: no matching field name!", __FILE__, __LINE__ );

}

/* ---------------------------------------------------------------------- */

void yac_cget_field_name_from_id ( int field_id,
                                   char *field_name ) {

  struct coupling_field * field;
  field = yac_unique_id_to_pointer(field_id);

  if ( field == NULL )
    yac_internal_abort_message ( "ERROR: field ID not defined!", __FILE__, __LINE__ );

  strcpy(field_name,yac_get_coupling_field_name(field));

  return;
}

/* ---------------------------------------------------------------------- */

void yac_cget_role_of_field_by_id ( int field_id, int *role ) {

  struct coupling_field * field;
  field = yac_unique_id_to_pointer(field_id);

  if ( field == NULL )
    yac_internal_abort_message ( "ERROR: field ID not defined!", __FILE__, __LINE__ );

  *role = yac_get_coupling_field_exchange_type(field);

  return;
}

/* ---------------------------------------------------------------------- */

void yac_cget_role_of_field_by_name ( char *field_name, int *role ) {

  int field_id;
  yac_cget_field_id_from_name ( field_name, &field_id );

  struct coupling_field * field;
  field = yac_unique_id_to_pointer(field_id);
  *role = yac_get_coupling_field_exchange_type(field);

  return;
}

/* ---------------------------------------------------------------------- */

void yac_cget_coupling_timestep_name ( const char * field_name,
                                       char *timestep_string ) {
  int field_id;
  yac_cget_field_id_from_name ( field_name, &field_id );
  yac_cget_coupling_timestep_id ( field_id, timestep_string );

  return;
}

/* ---------------------------------------------------------------------- */

void yac_cget_coupling_timestep_id ( int field_id,
                                     char *timestep_string ) {

  struct coupling_field * field;
  field = yac_unique_id_to_pointer(field_id);
  int event_id;

  if ( field == NULL )
    yac_internal_abort_message ( "ERROR: field ID not defined!", __FILE__, __LINE__ );

  int role = yac_get_coupling_field_exchange_type(field);

  if ( role == SOURCE )
    for (unsigned put_idx = 0; put_idx < yac_get_coupling_field_num_puts(field);
         ++put_idx)
      event_id = yac_get_coupling_field_put_op_event_id(field, put_idx);

  else if ( role == TARGET )
    event_id = yac_get_coupling_field_get_op_event_id(field);

  char my_string[MAX_TIMEDELTA_STR_LEN];

  if ( role == SOURCE || role == TARGET )
    strcpy(timestep_string,yac_get_event_coupling_timestep(event_id,my_string));

  return;
}

/* ---------------------------------------------------------------------- */

void yac_cget_model_timestep_name ( const char * field_name,
                                    char *timestep_string ) {
  int field_id;
  yac_cget_field_id_from_name ( field_name, &field_id );
  yac_cget_model_timestep_id ( field_id, timestep_string );

  return;
}

/* ---------------------------------------------------------------------- */

void yac_cget_model_timestep_id ( int field_id,
                                  char *timestep_string ) {

  struct coupling_field * field;
  field = yac_unique_id_to_pointer(field_id);
  int event_id;

  if ( field == NULL )
    yac_internal_abort_message ( "ERROR: field ID not defined!", __FILE__, __LINE__ );

  int role = yac_get_coupling_field_exchange_type(field);

  if ( role == SOURCE )
    for (unsigned put_idx = 0; put_idx < yac_get_coupling_field_num_puts(field);
         ++put_idx)
      event_id = yac_get_coupling_field_put_op_event_id(field, put_idx);

  else if ( role == TARGET )
    event_id = yac_get_coupling_field_get_op_event_id(field);

  char my_string[MAX_TIMEDELTA_STR_LEN];

  if ( role == SOURCE || role == TARGET )
    strcpy(timestep_string,yac_get_event_model_timestep(event_id,my_string));

  return;
}
