/**
 * @file yac_interface.h
 * @brief C User Interface
 *
 * For a detailed description of the Fortran interface please see
 * \ref mo_yac_finterface.f90 and \ref yac_finterface.f90
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *                                 Teresa Holfeld <teresa.holfeld@zmaw.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 *         Teresa Holfeld <teresa.holfeld@zmaw.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 *             Teresa Holfeld <teresa.holfeld@zmaw.de>
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

#ifndef YAC_INTERFACE_H
#define YAC_INTERFACE_H

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef USE_MPI
#include <mpi.h>
#endif

extern int const YAC_LOCATION_CELL;
extern int const YAC_LOCATION_CORNER;
extern int const YAC_LOCATION_EDGE;

/** \example test_dummy_coupling_c.c
 * This example simulates a whole model setup with three components (ocean,
 * atmosphere, io). It uses one process for each component.
 */

/* -----------------------------------------------------------------------------

    C and Fortran description of the user API

   ----------------------------------------------------------------------------- */

/** Elementary definition of the whole system

     @param[in] xml_filename    - coupling xml configuration
     @param[in] schema_filename - corresponding ml schema file
*/
    void yac_cinit ( const char * xml_filename, const char * schema_filename );

    /** YAC internal elementary definition of the whole system

         @param[in] xml_filename    - coupling xml configuration
         @param[in] schema_filename - corresponding ml schema file
         @return 1 if redirect of stdout is request in coupling xml
    */

    void yac_init ( const char * xml_filename, const char * schema_filename ) ;

/* -------------------------------------------------------------------------------- */

/** Finalisation

*/
     void yac_finalize ();
     void yac_cfinalize ();

/* -------------------------------------------------------------------------------- */

/** Definition of job start and end datetime

     @param[in] start_datetime   - calendar job start datetime
     @param[in] end_datetime     - calendar job end datetime
 */

     void yac_cdef_datetime ( const char * start_datetime,
                              const char * end_datetime );

/* -------------------------------------------------------------------------------- */

/** Elementary definition of the component.

     @param[in] comp_name   - name of the component
     @param[in] comp_id     - component Id
*/

     void yac_cdef_comp ( const char * comp_name,
                          int * comp_id );

/* -------------------------------------------------------------------------------- */

/** Elementary definition of the component.

     @param[in] comp_names   - names of the components
     @param[in] num_comps    - number of components
     @param[in] comp_ids     - component Id's
*/

     void yac_cdef_comps ( const char ** comp_names,
                           int num_comps,
                           int * comp_ids );

/* -------------------------------------------------------------------------------- */

/** Setting io_proc_step for a component.

     @param[in] comp_id      - component Id
     @param[in] io_proc_step - stride between component processes doing io
                               (e.g. writing weight files)
*/

     void yac_cset_comp_io_proc_step ( int comp_id, int io_proc_step );

/* -------------------------------------------------------------------------------- */

/** Getting io_proc_step from a component.

     @param[in] comp_id       - component Id
     @param[out] io_proc_step - stride between component processes doing io
                                (e.g. writing weight files)
*/

     void yac_cget_comp_io_proc_step ( int comp_id, int * io_proc_step );

/* -------------------------------------------------------------------------------- */

/** Providing the local MPI communicator

     @param[in] comp_id            - local component ID (from yac_cdef_comp), handle to the component struct
     @param[in] local_communicator - returned local communicator

*/
#ifdef USE_MPI
     void yac_get_localcomm ( MPI_Fint *local_communicator, int comp_id );
#else
     void yac_get_localcomm ( int *local_communicator, int comp_id );
#endif

#ifdef USE_MPI
     void yac_cget_localcomm ( MPI_Comm *local_communicator, int comp_id );
#else
     void yac_cget_localcomm ( int *local_communicator, int comp_id );
#endif

/* -------------------------------------------------------------------------------- */

/** Providing a MPI communicator for communication between the component root processes

     @param[in] comp_name1  - name of the first component
     @param[in] comp_name2  - name of the second component
     @param[in] local_communicator - returned local communicator

*/
#ifdef USE_MPI
     void yac_get_pair_rootcomm ( MPI_Fint *local_communicator, const char * comp_name1, const char * comp_name2 );
#else
     void yac_get_pair_rootcomm ( int *local_communicator, const char * comp_name1, const char * comp_name2 );
#endif

#ifdef USE_MPI
     void yac_cget_pair_rootcomm ( MPI_Comm *local_communicator, const char * comp_name1, const char * comp_name2 );
#else
     void yac_cget_pair_rootcomm ( int *local_communicator, const char * comp_name1, const char * comp_name2 );
#endif

/* -------------------------------------------------------------------------------- */

/** Elementary definition of the grid.

     @param[in] comp_id        - local component ID (from yac_cdef_comp), handle to the component struct
     @param[in] subdomain_name - name of the complete grid
     @param[in] subdomain_id   - returned handle to the subdomain struct
*/

     void yac_cdef_subdomain ( int const comp_id,
                               const char * subdomain_name,
                               int * subdomain_id );

/* -------------------------------------------------------------------------------- */

/** Connect grid elements

     @param[in] comp_id           - local component ID (from yac_cdef_comp), handle to the component struct
     @param[in] nbr_subdomain_ids - number of subdomains that have to be glued together
     @param[in] subdomain_ids     - array of grid id from def_subdomain that have to be glued together
     @param[in] domain_id         -
*/

     void yac_cconnect_subdomains ( int const comp_id,
                                    int const nbr_subdomain_ids,
                                    int const *subdomain_ids,
                                    int * domain_id );

/* -------------------------------------------------------------------------------- */

/** Definition of a set of points for 2D regular grids.

     @param[in] subdomain_id - handle to the grid struct (returned from yac_cdef_subdomain)
     @param[in] nbr_points   - number of points
     @param[in] location     - location of points [edge|vertex|center]
     @param[in] x_points     - array of point longitudes
     @param[in] y_points     - array of point latitudes
     @param[in] point_id     - returned handle to the points struct
*/

     void yac_cdef_points_reg2d ( int const subdomain_id,
                                  int const *nbr_points,
                                  int const location,
                                  double const *x_points,
                                  double const *y_points,
                                  int *point_id );

/* -------------------------------------------------------------------------------- */

/** Definition of a set of points for 2D unstructured grids.

     @param[in] subdomain_id - handle to the grid struct (returned from yac_cdef_subdomain)
     @param[in] nbr_points   - number of points
     @param[in] location     - location of points [edge|vertex|center]
     @param[in] x_points     - array of point longitudes
     @param[in] y_points     - array of point latitudes
     @param[in] point_id     - returned handle to the points struct
*/

     void yac_cdef_points_unstruct ( int const subdomain_id,
                                     int const nbr_points,
                                     int const location,
                                     double const *x_points,
                                     double const *y_points,
                                     int *point_id );

/* -------------------------------------------------------------------------------- */

/** Definition of a set of elements for an unstructured grid.

     @param[in] subdomain_id             - handle to the grid struct (returned from yac_cdef_subdomain)
     @param[in] nbr_vertices             - number of vertices
     @param[in] nbr_cells                - number of cells
     @param[in] num_vertices_per_cell_in - array containing the number of vertices for each cell
     @param[in] x_vertices               - array of vertex longitudes
     @param[in] y_vertices               - array of vertex latitudes
     @param[in] cell_to_vertex_in        - connectivity of vertices belonging to cells\n
                                         (the vertex indices per cell have to be in clockwise or counterclockwise ordering)

*/

     void yac_cdef_elements_unstruct ( int const subdomain_id,
                                       int const nbr_vertices,
                                       int const nbr_cells,
                                       int *num_vertices_per_cell_in,
                                       double *x_vertices,
                                       double *y_vertices,
                                       int *cell_to_vertex_in );

/* -------------------------------------------------------------------------------- */

/** Definition of a set of elements for a 2d regular grid.

     @param[in] subdomain_id          - handle to the grid struct (returned from yac_cdef_subdomain)
     @param[in] nbr_vertices          - 2d array containing the number of vertices in each dimension
     @param[in] cyclic                - 2d array containing information about cyclic behaviour in each dimension
     @param[in] x_vertices            - array of vertex longitudes
     @param[in] y_vertices            - array of vertex latitudes

*/
     void yac_cdef_elements_reg2d ( int const subdomain_id,
                                    int const nbr_vertices[2],
                                    int const cyclic[2],
                                    double *x_vertices,
                                    double *y_vertices );

/* -------------------------------------------------------------------------------- */

/** Definition of the mask

     @param[in] mask         - mask: 0 for invalid, 1 for valid cells
     @param[in] points_id    - id of points/cells, which the mask is associated to
     @param[in] mask_id      - returned handle to mask struct
*/

     void yac_cdef_mask ( int const * mask,
                          int points_id,
                          int * mask_id );

/* -------------------------------------------------------------------------------- */

/** Definition of the logical process number where the inner point or cell is located

     Local points are marked with their local rank, halo point are marked with the rank
     of the original location.

     @param[in] subdomain_id      - handle to the grid struct (returned from yac_cdef_subdomain)
     @param[in] nbr_entries       - number of values in index_location
     @param[in] location          -
     @param[in] global_index      -
     @param[in] global_index_rank -
*/
     void yac_cdef_index_location ( int const subdomain_id,
                                    int const nbr_entries,
                                    int const location,
                                    int const * global_index,
                                    int const * global_index_rank);

/* -------------------------------------------------------------------------------- */

/** Definition of the coupling field

     @param[in] field_name               - character string providing the name of the coupling field
     @param[in] component_id             - component ID
     @param[in] domain_id                - domain ID
     @param[in] point_ids                - point IDs
     @param[in] mask_ids                 - mask IDs
     @param[in] num_pointsets            - number of pointsets per subdomain
     @param[in] field_id                 - returned field_id which has to be used to identify coupling fields
                                       in yac_cput and yac_cget.
*/

       void yac_cdef_field ( char const * field_name,
                             int const component_id,
                             int const domain_id,
                             int const * point_ids,
                             int const * mask_ids,
                             int const num_pointsets,
                             int * field_id );

/* -------------------------------------------------------------------------------- */

/** Receiving of the coupling fields

     @param[in] field_id           -
     @param[in] collection_size    -
     @param[in] recv_field_        - receive buffer
     @param[in] info               - returned info argument indicating the action performed
     @param[in] ierror             - returned error
*/
     void yac_cget_ ( int const field_id,
                      int const collection_size,
                      double *recv_field_,
                      int    *info,
                      int    *ierror );

     void yac_cget ( int const field_id,
                     int const collection_size,
                     double ***recv_field,
                     int    *info,
                     int    *ierror );

/* -------------------------------------------------------------------------------- */

/** Receiving of the coupling fields

     @param[in] field_id           -
     @param[in] collection_size    -
     @param[in] recv_field_        - receive buffer
     @param[in] info               - returned info argument indicating the action performed
     @param[in] ierror             - returned error
*/

     void yac_cget_ptr_ ( int const field_id,
                          int const collection_size,
                          double **recv_field_,
                          int    *info,
                          int    *ierror );

/* -------------------------------------------------------------------------------- */

/** Sending of the coupling fields

     @param[in] field_id           -
     @param[in] collection_size    -
     @param[in] send_field_        - send buffer
     @param[in] info               - returned info
     @param[in] ierror             - returned error
*/

     void yac_cput_ ( int const field_id,
                      int const collection_size,
                      double *send_field_,
                      int    *info,
                      int    *ierror );

     void yac_cput ( int const field_id,
                     int const collection_size,
                     double **** const send_field,
                     int *info,
                     int *ierror );

/* -------------------------------------------------------------------------------- */

/** Sending of the coupling fields

     @param[in] field_id           -
     @param[in] collection_size    -
     @param[in] send_field_        - send buffer
     @param[in] info               - returned info
     @param[in] ierror             - returned error
*/

void yac_cput_ptr_ ( int const field_id,
                     int const collection_size,
                     double ** send_field_,
                     int    *info,
                     int    *ierror );

/* -------------------------------------------------------------------------------- */

/** End of the definition phase, invocation of the search

     @param[in] nbr_comps          - number of components defined on the calling process
     @param[in] comp_ids           - list of component IDs as they are returned by yac_cdef_comp
     @param[in] nbr_fields         - number of fields defined on the calling process
     @param[in] field_ids          - list of field IDs as they are returned by yac_cdef_field
     @param[in] ierror             - returned error
*/

void yac_csearch ( int nbr_comps,
                   int * comp_ids,
                   int nbr_fields,
                   int * field_ids,
                   int * ierror );

/* --------------------------------------------------------------------------------
           query routines
   -------------------------------------------------------------------------------- */

/** query routine for the start datetime of the job
*/

char * yac_get_start_datetime ();

/** query routine for the end datetime of the job
*/

char * yac_get_end_datetime ();

/** query routine for number of coupling fields defined

     @param[in] nbr_fields         - number of fields defined on the calling process

*/

void yac_cget_nbr_fields ( int *nbr_fields );

/* -------------------------------------------------------------------------------- */

/** query routine to get the list of field IDs for defined fields

     @param[in] nbr_fields         - number of fields defined on the calling process
     @param[in] field_ids          - list of delivered field IDs

*/

void yac_cget_field_ids ( int nbr_fields, int *field_ids );

/* ---------------------------------------------------------------------- */

/** query routine to get the field name for a given ID

     @param[in] field_name         - field name that has been given to yac_cdef_field
     @param[in] field_id           - ID as it has been provided by yac_cdef_field

*/

void yac_cget_field_id_from_name ( const char *field_name, int *field_id );

/* ---------------------------------------------------------------------- */

/** query routine to get the field name for a given ID

     @param[in] field_id           - ID as provided by yac_cdef_field
     @param[in] field_name         - field name that has been given to yac_cdef_field

*/

void yac_cget_field_name_from_id ( int field_id, char *field_name );

/* ---------------------------------------------------------------------- */

/** query routine to get the requested send/recv operation for a given field ID

     @param[in] field_name         - field name as provided by yac_cdef_field
     @param[in] role               - NOTHING (0) || 1 SOURCE (1) || TARGET (2)

*/

void yac_cget_role_of_field_by_name ( char * field_name, int *role );

/* ---------------------------------------------------------------------- */

/** query routine to get the requested send/recv operation for a given field ID

     @param[in] field_id           - ID as provided by yac_cdef_field
     @param[in] role               - NOTHING (0) || 1 SOURCE (1) || TARGET (2)

*/

void yac_cget_role_of_field_by_id ( int field_id, int *role );

/* ---------------------------------------------------------------------- */

/** query routine to get coupling interval for a given field name

     @param[in] field_name         - field name as set in yac_cdef_field
     @param[in] timestep_string    - coupling time step as iso8601 string

*/

void yac_cget_coupling_timestep_name ( const char * field_name,
                                       char *timestep_string );

/* ---------------------------------------------------------------------- */

/** query routine to get coupling interval for a given field id

     @param[in] field_id           - field ID (as representative for a field name)
     @param[in] timestep_string    - coupling time step as iso8601 string

*/

void yac_cget_coupling_timestep_id ( const int field_id,
                                     char *timestep_string );

/* -------------------------------------------------------------------------------- */

/** query routine to get model time step for a given field name

     @param[in] field_name         - field name as set in yac_cdef_field
     @param[in] timestep_string    - model time step as iso8601 string

*/

void yac_cget_model_timestep_name ( const char * field_name,
                                    char *timestep_string );

/* -------------------------------------------------------------------------------- */

/** query routine to get model time step for a given field id

     @param[in] field_id           - field ID (as representative for a field name)
     @param[in] timestep_string    - model time step as iso8601 string

*/

void yac_cget_model_timestep_id ( int field_id,
                                  char *timestep_string );

/* --------------------------------------------------------------------------------
           auxiliary routines
   -------------------------------------------------------------------------------- */

/** auxiliary routine

 * @param  filestem   basename of stdout/stderr files
 * @param  lenstr     significant length (without \0) of "filestem"
 * @param  parallel   flag for direction of output
                             parallel = 1 : Create single output files for each process
                             otherwise    : Create output files only for process 0.
                                           The output of other processes is not redirected.
 * @param  my_pe      local rank
 * @param  npes       total number of processes in set
 * @param  ierror     return code
                             = 0 : no error
                             = 1 : error in malloc
*/

void yac_redirstdout ( const char *filestem, int lenstr,
                       int parallel, int my_pe,
                       int npes, int *ierror );

#endif // YAC_INTERFACE_H
