/**
 * @file yac_redirstdout.c
 *
 * @copyright Copyright  (C)  2014 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Rene Redler <rene.redler@mpimet.mpg.de>
 */

/*
 * Keywords:
 * Maintainer: Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC. If not, see <http://www.gnu.org/licenses/gpl.txt>.
 */

/* ---------------------------------------------------------------------
  Date          Programmer   Description
  ------------  ----------   -----------
  01. Dec 2003  R. Redler    created
  01. Dec 2005  H. Ritzdorf  revised

   Copyright 2006-2010, NEC Europe Ltd., London, UK.
   All rights reserved. Use is subject to OASIS4 license terms.
   --------------------------------------------------------------------- */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "yacpar.h"

/**
 * redirect stderr and stdout into files with names depending on the process ID
 *
 * The original version was part of the OASIS4 coupler and has been adjusted to
 * work for yac.
 *
 * @param[in]  filestem   basename of stdout/stderr files
 * @param[in]  lenstr     significant length (without \0) of "filestem"
 * @param[in]  parallel   flag for direction of output
                             parallel = 1 : Create single output files for each process
                             otherwise    : Create output files only for process 0.
                                           The output of other processes is not redirected.
 * @param[in]  my_pe      local rank
 * @param[in]  npes       total number of processes in set
 * @param[out] ierror     return code
                             = 0 : no error
                             = 1 : error in malloc
 */

void yac_redirstdout ( const char *filestem, int lenstr,
                       int parallel, int my_pe,
                       int npes, int *ierror ) {

        size_t len_alloc = lenstr;

        register int n_digits, num;
        register int redirect = 1;

        char *sname, *ename;
        register char *p_err, *p_std;

        ASSERT (lenstr > 0)
        ASSERT (my_pe  >= 0)

        *ierror = 0;

        /* get number of digits in rank */

        num = (npes > 0) ? npes : 1;
        n_digits = (int) (log10((double)(num) + (double) 0.5)) + 1;
        ASSERT (n_digits > 0)

        /* allocate file names */

#define NUMBER_OF_EXTENSIONS 2
#define LENGTH_OF_ERR 3

        len_alloc += n_digits + NUMBER_OF_EXTENSIONS + LENGTH_OF_ERR + 1;

        sname = (char *) malloc (len_alloc);
        ename = (char *) malloc (len_alloc);

        if (!sname || !ename) {
                fprintf (stderr, "PSMILe_redirstdout: Cannot allocate memory. %d bytes\n", (int) len_alloc);
                *ierror = 1;
                return;
        }

        /* copy base name to standard output and standard error file name */

        memcpy (sname, filestem, (size_t) lenstr);
        memcpy (ename, filestem, (size_t) lenstr);

        p_std = sname + (size_t) lenstr;
        p_err = ename + (size_t) lenstr;

        if (parallel == 1) {

                /* all processes write into their own file */

                sprintf (p_std, ".%0*d",     n_digits, my_pe);
                sprintf (p_err, ".err.%0*d", n_digits, my_pe);
        }

        else {

                /* only local root redirects stdout */

                redirect = (my_pe == 0);

                if( redirect ) {
                        sprintf (p_std, ".log");
                        sprintf (p_err, ".err");
                }
        }

        /* redirect stdout/stderr */

        if (redirect) {
                freopen (sname, "w", stdout);
                freopen (ename, "w", stderr);
        }

        /* free file names */

        free (ename);
        free (sname);
}
