/**
 * @file test_area.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "clipping.h"
#include "area.h"
#include "tests.h"
#include "geometry.h"

//#define VERBOSE

void test_area(struct grid_cell cell, double ref_area,
               double (*area_func)(struct grid_cell));

int main (void) {

  enum yac_edge_type edge_types[5] = {GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE,
                                      GREAT_CIRCLE, GREAT_CIRCLE};

  struct grid_cell Quin = {.coordinates_x = NULL,
                           .coordinates_y = NULL,
                           .coordinates_xyz = (double[]){-1,-1,-1,-1,-1,
                                                         -1,-1,-1,-1,-1,
                                                         -1,-1,-1,-1,-1},
                           .edge_type = edge_types,
                           .num_corners = 5};
  struct grid_cell Quad = {.coordinates_x = NULL,
                           .coordinates_y = NULL,
                           .coordinates_xyz = (double[]){-1,-1,-1,-1,
                                                         -1,-1,-1,-1,
                                                         -1,-1,-1,-1},
                           .edge_type = edge_types,
                           .num_corners = 4};
  struct grid_cell Tri = {.coordinates_x = NULL,
                          .coordinates_y = NULL,
                          .coordinates_xyz = (double[]){-1,-1,-1,
                                                        -1,-1,-1,
                                                        -1,-1,-1},
                          .edge_type = edge_types,
                          .num_corners = 3};

  double area;

#ifdef VERBOSE
  puts ("----- calculating area -----\n");
#endif

  {
    /* set the test data over the Equator

               0.0 (lon)
               0.5 (lat)
               / \
              /   \
    -0.5     /     \  0.5
     0.0     \     /  0.0
              \   /
               \ /
               0.0
              -0.5         
    */

    double coordinates_x[4] = {0.5*rad, 0.0*rad, -0.5*rad, 0.0*rad};
    double coordinates_y[4] = {0.0*rad, 0.5*rad, 0.0*rad, -0.5*rad};

    Quad.coordinates_x = coordinates_x;
    Quad.coordinates_y = coordinates_y;
    for (int i = 0; i < Quad.num_corners; ++i)
      LLtoXYZ(Quad.coordinates_x[i], Quad.coordinates_y[i],
              Quad.coordinates_xyz+i*3);

    area = yac_cell_approx_area ( Quad );
#ifdef VERBOSE
    printf ( "Approx   area of quad   over Equator    is %f sqr Km \n", area * EarthRadius2 );
#endif

    area = yac_girards_area ( Quad );
#ifdef VERBOSE
    printf ( "Girard   area of quad   over Equator    is %f sqr Km \n", area * EarthRadius * EarthRadius );
#endif

    area = yac_cell_area ( Quad );
#ifdef VERBOSE
    printf ( "ICON     area of quad   over Equator    is %f sqr km\n", area * EarthRadius * EarthRadius );
#endif

    area = yac_pole_area ( Quad );
#ifdef VERBOSE
    printf ( "pole     area of quad   over Equator    is %f sqr km\n", area * EarthRadius * EarthRadius );
#endif

    area = yac_huiliers_area ( Quad );
#ifdef VERBOSE
    printf ( "huiliers area of quad   over Equator    is %f sqr km\n", area * EarthRadius * EarthRadius );
#endif
  }

  {
  /* Triangle, half of the above Quad */

    double coordinates_x[3] = {0.5*rad, 0.0*rad, -0.5*rad};
    double coordinates_y[3] = {0.0*rad, 0.5*rad, 0.0*rad};

    Tri.coordinates_x = coordinates_x;
    Tri.coordinates_y = coordinates_y;
    for (int i = 0; i < Tri.num_corners; ++i)
      LLtoXYZ(Tri.coordinates_x[i], Tri.coordinates_y[i],
              Tri.coordinates_xyz+i*3);

    area = yac_cell_approx_area ( Tri );
#ifdef VERBOSE
    printf ( "Approx   area of triangle over Equator  is %f sqr Km \n", area * EarthRadius2 );
#endif

    area = yac_girards_area ( Tri );
#ifdef VERBOSE
    printf ( "Girard   area of triangle over Equator  is %f sqr Km \n", area * EarthRadius * EarthRadius );
#endif

    area = yac_triangle_area ( Tri );
#ifdef VERBOSE
    printf ( "ICON     area of triangle over Equator  is %f sqr km\n", area * EarthRadius * EarthRadius );
#endif

    area = yac_pole_area ( Tri );
#ifdef VERBOSE
    printf ( "pole     area of triangle over Equator  is %f sqr km\n", area * EarthRadius * EarthRadius );
#endif

    area = yac_huiliers_area ( Tri );
#ifdef VERBOSE
    printf ( "huiliers area of triangle over Equator  is %f sqr km\n", area * EarthRadius * EarthRadius );
#endif
  }

  {
    /* set the test data over the Equator

    -0.5               0.5 (lon)
    0.5  -----------  0.5 (lat)
         |         |
         |         |
         |         |
         |         |
         |         |
    -0.5  -----------  0.5
    -0.5              -0.5

    */

    double coordinates_x[4] = {-0.5*rad, 0.5*rad, 0.5*rad, -0.5*rad};
    double coordinates_y[4] = {-0.5*rad, -0.5*rad, 0.5*rad, 0.5*rad};

    Quad.coordinates_x = coordinates_x;
    Quad.coordinates_y = coordinates_y;
    for (int i = 0; i < Quad.num_corners; ++i)
      LLtoXYZ(Quad.coordinates_x[i], Quad.coordinates_y[i],
              Quad.coordinates_xyz+i*3);

    area = yac_cell_approx_area ( Quad );
#ifdef VERBOSE
    printf ( "Approx   area of quad   over Equator    is %f sqr Km \n", area * EarthRadius2 );
#endif

    area = yac_girards_area ( Quad );
#ifdef VERBOSE
    printf ( "Girard   area of quad   over Equator    is %f sqr Km \n", area * EarthRadius * EarthRadius );
#endif

    area = yac_cell_area ( Quad );
#ifdef VERBOSE
    printf ( "ICON     area of quad   over Equator    is %f sqr km\n", area * EarthRadius * EarthRadius );
#endif

    area = yac_pole_area ( Quad );
#ifdef VERBOSE
    printf ( "pole     area of quad   over Equator    is %f sqr km\n", area * EarthRadius * EarthRadius );
#endif

    area = yac_huiliers_area ( Quad );
#ifdef VERBOSE
    printf ( "huiliers area of quad   over Equator    is %f sqr km\n", area * EarthRadius * EarthRadius );
#endif

    double ref_area = yac_pole_area(Quad);

    test_area(Quad, ref_area, yac_pole_area);
    test_area(Quad, ref_area, yac_huiliers_area);
  }

  {
    /* Half of the above Quad */

    double coordinates_x[3] = {-0.5*rad, 0.5*rad, 0.5*rad};
    double coordinates_y[3] = {-0.5*rad, -0.5*rad, 0.5*rad};

    Tri.coordinates_x = coordinates_x;
    Tri.coordinates_y = coordinates_y;
    for (int i = 0; i < Tri.num_corners; ++i)
      LLtoXYZ(Tri.coordinates_x[i], Tri.coordinates_y[i],
              Tri.coordinates_xyz+i*3);

    area = yac_cell_approx_area ( Tri );
#ifdef VERBOSE
    printf ( "Approx   area of triangle over Equator  is %f sqr Km \n", area * EarthRadius2 );
#endif

    area = yac_girards_area ( Tri );
#ifdef VERBOSE
    printf ( "Girard   area of triangle over Equator  is %f sqr Km \n", area * EarthRadius * EarthRadius );
#endif

    area = yac_triangle_area ( Tri );
#ifdef VERBOSE
    printf ( "ICON     area of triangle over Equator  is %f sqr km\n", area * EarthRadius * EarthRadius );
#endif

    area = yac_cell_area ( Tri );
#ifdef VERBOSE
    printf ( "ICON     area of triangle over Equator  is %f sqr km\n", area * EarthRadius * EarthRadius );
#endif

    area = yac_pole_area ( Tri );
#ifdef VERBOSE
    printf ( "pole     area of triangle over Equator  is %f sqr km\n", area * EarthRadius * EarthRadius );
#endif

    area = yac_huiliers_area ( Tri );
#ifdef VERBOSE
    printf ( "huiliers area of triangle over Equator  is %f sqr km\n", area * EarthRadius * EarthRadius );
#endif
  }

  {
    double coordinates_x[3] = {150.0*rad, 180.0*rad, 180.0*rad};
    double coordinates_y[3] = {-60.0*rad, -60.0*rad, -90.0*rad};

    Tri.coordinates_x = coordinates_x;
    Tri.coordinates_y = coordinates_y;
    for (int i = 0; i < Tri.num_corners; ++i)
      LLtoXYZ(Tri.coordinates_x[i], Tri.coordinates_y[i],
              Tri.coordinates_xyz+i*3);

    area = yac_cell_approx_area ( Tri );
#ifdef VERBOSE
    printf ( "Approx   area of triangle over South Pole is %f sqr Km \n", area * EarthRadius2 );
#endif

    area = yac_girards_area ( Tri );
#ifdef VERBOSE
    printf ( "Girard   area of triangle over South Pole is %f sqr Km \n", area * EarthRadius * EarthRadius );
#endif

    area = yac_triangle_area ( Tri );
#ifdef VERBOSE
    printf ( "ICON     area of triangle over South Pole is %f sqr km\n", area * EarthRadius * EarthRadius );
#endif

    area = yac_cell_area ( Tri );
#ifdef VERBOSE
    printf ( "ICON     area of triangle over South Pole is %f sqr km\n", area * EarthRadius * EarthRadius );
#endif

    area = yac_pole_area ( Tri );
#ifdef VERBOSE
    printf ( "pole     area of triangle over South Pole is %f sqr km\n", area * EarthRadius * EarthRadius );
#endif

    area = yac_huiliers_area ( Tri );
#ifdef VERBOSE
    printf ( "huiliers area of triangle over South Pole is %f sqr km\n", area * EarthRadius * EarthRadius );
#endif
  }

  {
    /* set the test data over the North Pole

    90.0                0.0 (lon)
    89.5  -----------  89.5 (lat)
          |         |
          |         |
          |    x    |
          |         |
          |         |
    180.0  ----------- 270.0
    89.5               89.5

    */

    double coordinates_x[4] = {0.0*rad, 90.0*rad, -180.0*rad, -90.0*rad};
    double coordinates_y[4] = {89.5*rad, 89.5*rad, 89.5*rad, 89.5*rad};

    Quad.coordinates_x = coordinates_x;
    Quad.coordinates_y = coordinates_y;
    for (int i = 0; i < Quad.num_corners; ++i)
      LLtoXYZ(Quad.coordinates_x[i], Quad.coordinates_y[i],
              Quad.coordinates_xyz+i*3);

    area = yac_cell_approx_area ( Quad );
#ifdef VERBOSE
    printf ( "Approx area of quad     over North Pole is %f sqr Km \n", area * EarthRadius2 );
#endif

    area = yac_girards_area ( Quad );
#ifdef VERBOSE
    printf ( "Girard area of quad     over North Pole is %f sqr Km \n", area * EarthRadius * EarthRadius );
#endif

    area = yac_cell_area ( Quad );
#ifdef VERBOSE
    printf ( "ICON   area of quad     over North Pole is %f sqr km\n", area * EarthRadius * EarthRadius );
#endif

    area = yac_pole_area ( Quad );
#ifdef VERBOSE
    printf ( "pole   area of quad     over North Pole is %f sqr km\n", area * EarthRadius * EarthRadius );
#endif

    area = yac_huiliers_area ( Quad );
#ifdef VERBOSE
    printf ( "huiliers area of quad   over North Pole is %f sqr km\n", area * EarthRadius * EarthRadius );
#endif
  }

  {

    double coordinates_x[4] = {-180.0*rad, -150.0*rad, -150.0*rad, -180.0*rad};
    double coordinates_y[4] = {60.0*rad, 60.0*rad, 90.0*rad, 90.0*rad};

    Quad.coordinates_x = coordinates_x;
    Quad.coordinates_y = coordinates_y;
    for (int i = 0; i < Quad.num_corners; ++i)
      LLtoXYZ(Quad.coordinates_x[i], Quad.coordinates_y[i],
              Quad.coordinates_xyz+i*3);

    area = yac_cell_approx_area ( Quad );
#ifdef VERBOSE
    printf ( "Approx   area of quad   near North Pole is %f sqr Km \n", area * EarthRadius2 );
#endif

    area = yac_girards_area ( Quad );
#ifdef VERBOSE
    printf ( "Girard   area of quad   near North Pole is %f sqr Km \n", area * EarthRadius * EarthRadius );
#endif

    area = yac_cell_area ( Quad );
#ifdef VERBOSE
    printf ( "ICON     area of quad   near North Pole is %f sqr km\n", area * EarthRadius * EarthRadius );
#endif

    area = yac_pole_area ( Quad );
#ifdef VERBOSE
    printf ( "pole     area of quad   near North Pole is %f sqr km\n", area * EarthRadius * EarthRadius );
#endif

    area = yac_huiliers_area ( Quad );
#ifdef VERBOSE
    printf ( "huiliers area of quad   near North Pole is %f sqr km\n", area * EarthRadius * EarthRadius );
#endif
  }

  {

    double coordinates_x[4] = {-180.0*rad, -150.0*rad, -150.0*rad, -180.0*rad};
    double coordinates_y[4] = {-90.0*rad, -90.0*rad, -60.0*rad, -60.0*rad};

    Quad.coordinates_x = coordinates_x;
    Quad.coordinates_y = coordinates_y;
    for (int i = 0; i < Quad.num_corners; ++i)
      LLtoXYZ(Quad.coordinates_x[i], Quad.coordinates_y[i],
              Quad.coordinates_xyz+i*3);

    area = yac_cell_approx_area ( Quad );
#ifdef VERBOSE
    printf ( "Approx   area of quad   near South Pole is %f sqr Km \n", area * EarthRadius2 );
#endif

    area = yac_girards_area ( Quad );
#ifdef VERBOSE
    printf ( "Girard   area of quad   near South Pole is %f sqr Km \n", area * EarthRadius * EarthRadius );
#endif

    /* would deliver NaN
       area = cell_area ( Quad ); */

#ifdef VERBOSE
    printf ( "ICON     area of quad   near South Pole is %f sqr km\n", area * EarthRadius * EarthRadius );
#endif

    area = yac_pole_area ( Quad );
#ifdef VERBOSE
    printf ( "pole     area of quad   near South Pole is %f sqr Km \n", area * EarthRadius * EarthRadius );
#endif

    area = yac_huiliers_area ( Quad );
#ifdef VERBOSE
    printf ( "huiliers area of quad   near South Pole is %f sqr Km \n", area * EarthRadius * EarthRadius );
#endif
  }

  {

    double coordinates_x[5] = {-180.0*rad, -150.0*rad, 0.0*rad, -150.0*rad, -180.0*rad};
    double coordinates_y[5] = {60.0*rad, 60.0*rad, 90.0*rad, 90.0*rad, 90.0*rad};

    Quin.coordinates_x = coordinates_x;
    Quin.coordinates_y = coordinates_y;
    for (int i = 0; i < Quin.num_corners; ++i)
      LLtoXYZ(Quin.coordinates_x[i], Quin.coordinates_y[i],
              Quin.coordinates_xyz+i*3);

    area = yac_cell_approx_area ( Quin );
#ifdef VERBOSE
    printf ( "Approx   area of quin   near North Pole is %f sqr Km \n", area * EarthRadius2 );
#endif

    area = yac_girards_area ( Quin );
#ifdef VERBOSE
    printf ( "Girard   area of quin   near North Pole is %f sqr Km \n", area * EarthRadius * EarthRadius );
#endif

    area = yac_cell_area ( Quin );
#ifdef VERBOSE
    printf ( "ICON     area of quin   near North Pole is %f sqr km\n", area * EarthRadius * EarthRadius );
#endif

    area = yac_pole_area ( Quin );
#ifdef VERBOSE
    printf ( "pole     area of quin   near North Pole is %f sqr km\n", area * EarthRadius * EarthRadius );
#endif

    area = yac_huiliers_area ( Quin );
#ifdef VERBOSE
    printf ( "huiliers area of quin   near North Pole is %f sqr km\n", area * EarthRadius * EarthRadius );
#endif
  }

  {

    double coordinates_x[2][4] = {{0*rad, 1*rad, 1*rad, 0*rad},
                                  {1*rad, 1*rad, 1*rad}};
    double coordinates_y[2][4] = {{0*rad, 0*rad, 1*rad, 1*rad},
                                  {0*rad, 0*rad, 1*rad}};

    Quad.coordinates_x = coordinates_x[0];
    Quad.coordinates_y = coordinates_y[0];
    Tri.coordinates_x = coordinates_x[1];
    Tri.coordinates_y = coordinates_y[1];

    double base_area = yac_girards_area(Quad);

    /* A dx of 1e-8 deg corresponds to a dx of ~1 mm on the Earth surface */

    double dx[] = {1e-8, 1e-7, 1e-6, 1e-5, 1e-4, 1e-3, 1e-2, 1e-1,
                   0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9};

    for (int i = 0; i < sizeof(dx) / sizeof(dx[0]); ++i) {

      if ((1-dx[i]) == 1) continue;

      coordinates_x[0][1] = (1-dx[i])*rad;
      coordinates_x[1][0] = (1-dx[i])*rad;

      for (int i = 0; i < Tri.num_corners; ++i)
        LLtoXYZ(Tri.coordinates_x[i], Tri.coordinates_y[i],
                Tri.coordinates_xyz+i*3);
      for (int i = 0; i < Quad.num_corners; ++i)
        LLtoXYZ(Quad.coordinates_x[i], Quad.coordinates_y[i],
                Quad.coordinates_xyz+i*3);

      double partial_area = yac_girards_area(Quad) + yac_girards_area(Tri);

      if (fabs(base_area - partial_area) >= yac_area_tol())
        PUT_ERR("girards_area computed wrong area\n");

#ifdef VERBOSE
      printf ("accuracy check for (base_area - partial_area) %8.1e sqr km for dx %.1e\n",
              base_area - partial_area, dx[i]);
#endif
    }
  }

  {

    double coordinates_x[3] = {0.09817477, 0.09817477, 0.098174496};
    double coordinates_y[3] = {0.353325247, 0.353335501, 0.353335609};

    Tri.coordinates_x = coordinates_x;
    Tri.coordinates_y = coordinates_y;
    for (int i = 0; i < Tri.num_corners; ++i)
      LLtoXYZ(Tri.coordinates_x[i], Tri.coordinates_y[i],
              Tri.coordinates_xyz+i*3);

    area = yac_girards_area(Tri);

    if (area < 0)
       PUT_ERR("girards_area computed wrong area\n");

#ifdef VERBOSE
    printf ("Girard accuracy check of very small area %.2e sqr km\n", area * EarthRadius * EarthRadius );
#endif

    area = yac_pole_area(Tri);

    if (area < 0)
       PUT_ERR("pole_area computed wrong area\n");

#ifdef VERBOSE
    printf ("pole   accuracy check of very small area %.2e sqr km\n", area * EarthRadius * EarthRadius );
#endif

    area = yac_huiliers_area(Tri);

    if (area < 0)
       PUT_ERR("huiliers_area computed wrong area\n");

#ifdef VERBOSE
    printf ("huiliers accuracy check of very small area %.2e sqr km\n", area * EarthRadius * EarthRadius );
#endif
  }

  {
    double edge_length_deg = 1.0e-8; // ~1.11 mm

    double coordinates_x[4] = {0, 0, 0, 0};
    double coordinates_y[4] = {0, 0, 0, 0};

    Quad.coordinates_x = coordinates_x;
    Quad.coordinates_y = coordinates_y;

#ifdef VERBOSE
     printf ("\n accuracy check for a unit square\n");
     printf (" edge length [m] |       girard       |        plain       |       simple       |        pole        |      huiliers\n");
     printf (" ----------------+--------------------+--------------------+--------------------+--------------------+--------------------\n");
#endif
     for ( unsigned i = 1; i < 11; ++i ) {

	double edge_length_m  = edge_length_deg * rad * EarthRadius * 1.0e3;

        coordinates_x[1] = edge_length_deg*rad;
        coordinates_x[2] = edge_length_deg*rad;
        coordinates_y[2] = edge_length_deg*rad;
        coordinates_y[3] = edge_length_deg*rad;

        for (int i = 0; i < Quad.num_corners; ++i)
          LLtoXYZ(Quad.coordinates_x[i], Quad.coordinates_y[i],
                  Quad.coordinates_xyz+i*3);

        double t_area = edge_length_m                * edge_length_m;
        double g_area = yac_girards_area(Quad)       * EarthRadius * EarthRadius * 1.0e6;
        double p_area = yac_planar_3dcell_area(Quad) * EarthRadius * EarthRadius * 1.0e6;
        double s_area = yac_pole_area(Quad)          * EarthRadius * EarthRadius * 1.0e6;
        double h_area = yac_huiliers_area(Quad)      * EarthRadius * EarthRadius * 1.0e6;

#ifndef VERBOSE
        if (0)
#endif
          printf (" %.8e m|%.8e sqr m|%.8e sqr m|%.8e sqr m|%.8e sqr m|%.8e sqr m\n",
                  edge_length_m, g_area, p_area, t_area, s_area, h_area);

        edge_length_deg *= 10.0;
     }

  }

  { // cell with two edges (one great circle and one circle of latitude)

    for (double y = -90; y <= 90; y += 5) {

      double coordinates_x[2] = {0.0 * rad, 20.0 * rad};
      double coordinates_y[2] = {y * rad, y * rad};
      enum yac_edge_type edge_types[2] = {GREAT_CIRCLE, LAT_CIRCLE};
      struct grid_cell cell = {.coordinates_x = coordinates_x,
                               .coordinates_y = coordinates_y,
                               .coordinates_xyz = (double[2*3]){-1},
                               .edge_type = edge_types,
                               .num_corners = 2};

      for (int i = 0; i < cell.num_corners; ++i)
        LLtoXYZ(cell.coordinates_x[i], cell.coordinates_y[i],
                cell.coordinates_xyz+i*3);

      // compute reference area
      double ref_area;

      {
        double coordinates_x[3] = {0.0 * rad, 20.0 * rad, 0.0 * rad};
        double coordinates_y[3] = {y * rad, y * rad, 0};
        enum yac_edge_type edge_types[3] = {GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE};
        struct grid_cell cell = {.coordinates_x = coordinates_x,
                                 .coordinates_y = coordinates_y,
                                 .coordinates_xyz = (double[3*3]){-1},
                                 .edge_type = edge_types,
                                 .num_corners = 3};

        if ((fabs(y) == 90.0) || (fabs(y) == 0.0)) {

          ref_area = 0;

        } else {

          if (y < 0) {
            ref_area = (coordinates_x[1] - coordinates_x[0]) *
                        (1 - cos(M_PI_2 + coordinates_y[0]));
            coordinates_y[2] = -90.0 * rad;
          } else {
            ref_area = (coordinates_x[1] - coordinates_x[0]) *
                        (1 - cos(M_PI_2 - coordinates_y[0]));
            coordinates_y[2] = 90.0 * rad;
          }
          for (int i = 0; i < cell.num_corners; ++i)
            LLtoXYZ(cell.coordinates_x[i], cell.coordinates_y[i],
                    cell.coordinates_xyz+i*3);
          // ref_area *= EarthRadius * EarthRadius;
          ref_area -= yac_huiliers_area(cell);
        }
      }

      test_area(cell, ref_area, yac_pole_area);
      test_area(cell, ref_area, yac_huiliers_area);
    }
  }

  { // cell with lon lat edges and one edge on the equator

    double ref_area = sin(0.5 * rad) * (0.5 * rad);
    // ref_area *= EarthRadius * EarthRadius;

    double coordinates_x[2][4] = {{-0.5 * rad, 0.0 * rad, 0.0 * rad, -0.5 * rad},
                                  { 0.0 * rad, 0.5 * rad, 0.5 * rad,  0.0 * rad}};
    double coordinates_y[2][4] = {{-0.5 * rad, -0.5 * rad, 0.0 * rad, 0.0 * rad},
                                  { 0.0 * rad,  0.0 * rad, 0.5 * rad, 0.5 * rad}};
    enum yac_edge_type edge_types[4] = {LAT_CIRCLE, LON_CIRCLE, LAT_CIRCLE, LON_CIRCLE};

    for (int x = 0; x < 2; ++x) {
      for (int y = 0; y < 2; ++y) {
        struct grid_cell cell = {.coordinates_x = coordinates_x[x],
                                 .coordinates_y = coordinates_y[y],
                                 .coordinates_xyz = (double[4*3]){-1},
                                 .edge_type = edge_types,
                                 .num_corners = 4};

        for (int i = 0; i < cell.num_corners; ++i)
          LLtoXYZ(cell.coordinates_x[i], cell.coordinates_y[i],
                  cell.coordinates_xyz+i*3);

        test_area(cell, ref_area, yac_pole_area);
        test_area(cell, ref_area, yac_huiliers_area);
      }
    }
  }

  { // cell with lon lat edges (both lat edges on the different side of the
    // equator)
    double ref_area = (2.0 * sin(0.5 * rad)) * (1.0 * rad);
    // ref_area *= EarthRadius * EarthRadius;

    double coordinates_x[4] = {-0.5 * rad, 0.5 * rad, 0.5 * rad, -0.5 * rad};
    double coordinates_y[4] = {-0.5 * rad, -0.5 * rad, 0.5 * rad, 0.5 * rad};
    enum yac_edge_type edge_types[4] = {LAT_CIRCLE, LON_CIRCLE, LAT_CIRCLE, LON_CIRCLE};
    struct grid_cell cell = {.coordinates_x = coordinates_x,
                             .coordinates_y = coordinates_y,
                             .coordinates_xyz = (double[4*3]){-1},
                             .edge_type = edge_types,
                             .num_corners = 4};

    for (int i = 0; i < cell.num_corners; ++i)
      LLtoXYZ(cell.coordinates_x[i], cell.coordinates_y[i],
              cell.coordinates_xyz+i*3);

    test_area(cell, ref_area, yac_pole_area);
    test_area(cell, ref_area, yac_huiliers_area);
  }

  { // cell with lon lat edges (both lat edges on the same side of the equator)
    double ref_area = (sin(1.0 * rad)-sin(0.5 * rad)) * (1.0 * rad);
    // ref_area *= EarthRadius * EarthRadius;
    
    for (int i = 0; i < 2; ++i) {

      double coordinates_x[4] = {-0.5 * rad, 0.5 * rad, 0.5 * rad, -0.5 * rad};
      double coordinates_y[2][4] = {{ 0.5 * rad, 0.5 * rad, 1.0 * rad,  1.0 * rad},
                                    {-0.5 * rad,-0.5 * rad,-1.0 * rad, -1.0 * rad}};
      enum yac_edge_type edge_types[4] = {LAT_CIRCLE, LON_CIRCLE, LAT_CIRCLE, LON_CIRCLE};
      struct grid_cell cell = {.coordinates_x = coordinates_x,
                               .coordinates_y = coordinates_y[i],
                               .coordinates_xyz = (double[4*3]){-1},
                               .edge_type = edge_types,
                               .num_corners = 4};

      for (int i = 0; i < cell.num_corners; ++i)
        LLtoXYZ(cell.coordinates_x[i], cell.coordinates_y[i],
                cell.coordinates_xyz+i*3);

      test_area(cell, ref_area, yac_pole_area);
      test_area(cell, ref_area, yac_huiliers_area);
    }
  }

  { // concave cell

    struct point intersections[2];

    if (!yac_intersect((struct edge){.points[0] = {.lon = 20.0*rad,
                                               .lat = 59.0*rad},
                                 .points[1] = {.lon =  0.0*rad,
                                               .lat = 80.0*rad},
                                 .edge_type = GREAT_CIRCLE},
                   (struct edge){.points[0] = {.lon =   0.0*rad,
                                               .lat =  60.0*rad},
                                 .points[1] = {.lon =  20.0*rad,
                                               .lat =  60.0*rad},
                                 .edge_type = LAT_CIRCLE},
                     intersections+0))
      return EXIT_FAILURE;
    if (!yac_intersect((struct edge){.points[0] = {.lon =  20.0*rad,
                                               .lat =  59.0*rad},
                                 .points[1] = {.lon = -20.0*rad,
                                               .lat =  59.0*rad},
                                 .edge_type = GREAT_CIRCLE},
                   (struct edge){.points[0] = {.lon =   0.0*rad,
                                               .lat =  60.0*rad},
                                 .points[1] = {.lon =  20.0*rad,
                                               .lat =  60.0*rad},
                                 .edge_type = LAT_CIRCLE},
                     intersections+1))
      return EXIT_FAILURE;

    struct grid_cell partial_ref_cell = {.coordinates_x = (double[]){20.0*rad,
                                                                     intersections[0].lon,
                                                                     intersections[1].lon},
                                         .coordinates_y = (double[]){59.0*rad,
                                                                     60.0*rad,
                                                                     60.0*rad},
                                         .coordinates_xyz = (double[3*3]){-1},
                                         .edge_type = (enum yac_edge_type[]){GREAT_CIRCLE,
                                                                             LAT_CIRCLE,
                                                                             GREAT_CIRCLE},
                                         .num_corners = 3}; // this cell is convex
    for (int i = 0; i < partial_ref_cell.num_corners; ++i)
      LLtoXYZ(partial_ref_cell.coordinates_x[i], partial_ref_cell.coordinates_y[i],
              partial_ref_cell.coordinates_xyz+i*3);
    double partial_area = yac_pole_area(partial_ref_cell);

    if (fabs(partial_area - yac_huiliers_area(partial_ref_cell)) > yac_area_tol())
      PUT_ERR("pole_area != huiliers_area for convex cell");

    double ref_area = 2 * partial_area;

    struct grid_cell cells[2] = {{.coordinates_x = (double[]){20.0*rad,
                                                               intersections[0].lon,
                                                               intersections[1].lon,
                                                              -intersections[1].lon,
                                                              -intersections[0].lon,
                                                              -20.0*rad},
                                  .coordinates_y = (double[]){59.0*rad,
                                                              60.0*rad,
                                                              60.0*rad,
                                                              60.0*rad,
                                                              60.0*rad,
                                                              59.0*rad},
                                  .coordinates_xyz = (double[6*3]){-1},
                                  .edge_type = (enum yac_edge_type[]){GREAT_CIRCLE,
                                                                      LAT_CIRCLE,
                                                                      GREAT_CIRCLE,
                                                                      LAT_CIRCLE,
                                                                      GREAT_CIRCLE,
                                                                      GREAT_CIRCLE},
                                  .num_corners = 6},
                                 {.coordinates_x = (double[]){20.0*rad,
                                                               intersections[0].lon,
                                                              -intersections[0].lon,
                                                              -20.0*rad,
                                                              -intersections[1].lon,
                                                               intersections[1].lon},
                                  .coordinates_y = (double[]){59.0*rad,
                                                              60.0*rad,
                                                              60.0*rad,
                                                              59.0*rad,
                                                              60.0*rad,
                                                              60.0*rad},
                                  .coordinates_xyz = (double[6*3]){-1},
                                  .edge_type = (enum yac_edge_type[]){GREAT_CIRCLE,
                                                                      LAT_CIRCLE,
                                                                      GREAT_CIRCLE,
                                                                      GREAT_CIRCLE,
                                                                      LAT_CIRCLE,
                                                                      GREAT_CIRCLE},
                                  .num_corners = 6}};

    for (int i = 0; i < 2; ++i) {
      for (int j = 0; j < cells[i].num_corners; ++j)
        LLtoXYZ(cells[i].coordinates_x[j], cells[i].coordinates_y[j],
                cells[i].coordinates_xyz+j*3);
      test_area(cells[i], ref_area, yac_pole_area);
      test_area(cells[i], ref_area, yac_huiliers_area);
    }
  }

  return TEST_EXIT_CODE;
}

void test_area(struct grid_cell cell, double ref_area,
               double (*area_func)(struct grid_cell)) {

  int order, start;
  double mem_dummy[2][128];
  double mem_dummy_[128*3];
  enum yac_edge_type edge_dummy[128];
  struct grid_cell test_cell = {.coordinates_x = mem_dummy[0],
                                .coordinates_y = mem_dummy[1],
                                .coordinates_xyz = mem_dummy_,
                                .edge_type = edge_dummy,
                                .num_corners = cell.num_corners};

  for (order = -1; order <= 1; order += 2) {
    for (start = 0; start < cell.num_corners; ++start) {

      for (int i = 0; i < cell.num_corners; ++i) {
        int j = (cell.num_corners+start+i*order)%cell.num_corners;
        test_cell.coordinates_x[i] = cell.coordinates_x[j];
        test_cell.coordinates_y[i] = cell.coordinates_y[j];
        test_cell.coordinates_xyz[0+i*3] = cell.coordinates_xyz[0+j*3];
        test_cell.coordinates_xyz[1+i*3] = cell.coordinates_xyz[1+j*3];
        test_cell.coordinates_xyz[2+i*3] = cell.coordinates_xyz[2+j*3];
        j = (cell.num_corners + j - (order < 0))%cell.num_corners;
        test_cell.edge_type[i] = cell.edge_type[j];
      }

      double area = area_func(test_cell);

      if (fabs(area-ref_area) > yac_area_tol())
        PUT_ERR("ERROR: wrong area\n");
    }
  }
}
