/**
 * @file test_cell_bnd_circle.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "tests.h"
#include "geometry.h"
#include "grid_cell.h"
#include "utils.h"

double const tol = 1.0e-12;

void check_latlon_cell(double * coordinates_x, double * coordinates_y);
void check_gc_triangle(double * coordinates_x, double * coordinates_y);
void check_gc_quad(double * coordinates_x, double * coordinates_y);
// tests whether all corners of cell are within the bounding circle
void test_circle(struct grid_cell cell, struct bounding_circle circle);
void test_cirumscribe_circle(struct grid_cell cell, struct bounding_circle circle);
// tests whether a point is within the bounding circle
unsigned point_in_circle(double point[3], struct bounding_circle circle);

int main (void) {

   { // test regular cell

      check_latlon_cell((double[]){-1*rad,
                                    1*rad,
                                    1*rad,
                                   -1*rad},
                        (double[]){-1*rad,
                                   -1*rad,
                                    1*rad,
                                    1*rad});
   }

   { // test regular cell

      check_latlon_cell((double[]){40*rad,
                                   45*rad,
                                   45*rad,
                                   40*rad},
                        (double[]){20*rad,
                                   20*rad,
                                   25*rad,
                                   25*rad});
   }

   { // test regular cell

      check_latlon_cell((double[]){ 175*rad,
                                   -175*rad,
                                   -175*rad,
                                    175*rad},
                        (double[]){-5*rad,
                                   -5*rad,
                                    5*rad,
                                    5*rad});
   }

   { // test regular cell

      check_latlon_cell((double[]){30*rad,
                                   40*rad,
                                   40*rad,
                                   30*rad},
                        (double[]){80*rad,
                                   80*rad,
                                   85*rad,
                                   85*rad});
   }

   { // test regular cell

      check_latlon_cell((double[]){30*rad,
                                   40*rad,
                                   40*rad,
                                   30*rad},
                        (double[]){80*rad,
                                   80*rad,
                                   90*rad,
                                   90*rad});
   }

   { // test triangle

      check_gc_triangle((double[]){-5*rad,
                                    5*rad,
                                    0*rad},
                        (double[]){-5*rad,
                                    5*rad,
                                   -5*rad});
   }

   { // test triangle

      check_gc_triangle((double[]){   0*rad,
                                    120*rad,
                                   -120*rad},
                        (double[]){85*rad,
                                   85*rad,
                                   85*rad});
   }

   { // test triangle

      check_gc_triangle((double[]){   0*rad,
                                    120*rad,
                                   -120*rad},
                        (double[]){-85*rad,
                                   -85*rad,
                                   -85*rad});
   }

   { // test triangle

      check_gc_triangle((double[]){-5*rad,
                                    5*rad,
                                    1*rad},
                        (double[]){0*rad,
                                   0*rad,
                                   1*rad});
   }

   { // test triangle

      check_gc_triangle((double[]){  0*rad,
                                   170*rad,
                                   260*rad},
                        (double[]){85*rad,
                                   85*rad,
                                   89*rad});
   }

   { // test great circle quad

      check_gc_quad((double[]){-1*rad,
                                1*rad,
                                1*rad,
                               -1*rad},
                    (double[]){-1*rad,
                               -1*rad,
                                1*rad,
                                1*rad});
   }

   { // test great circle quad

      check_gc_quad((double[]){  0*rad,
                                90*rad,
                               180*rad,
                               270*rad},
                    (double[]){85*rad,
                               85*rad,
                               85*rad,
                               85*rad});
   }

   { // test great circle quad

      check_gc_quad((double[]){  0*rad,
                                90*rad,
                               180*rad,
                               270*rad},
                    (double[]){-85*rad,
                               -85*rad,
                               -85*rad,
                               -85*rad});
   }

   { // test great circle quad

      check_gc_quad((double[]){  0*rad,
                                10*rad,
                                 0*rad,
                               -10*rad},
                    (double[]){-10*rad,
                                 0*rad,
                                10*rad,
                                 0*rad});
   }

   return TEST_EXIT_CODE;
}

void check_gc_triangle(double * coordinates_x, double * coordinates_y) {

   enum yac_edge_type edges[] = {GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE};

   for (int order = -1; order <= 1; order += 2) {

      for (int start = 0; start < 3; ++start) {

         double temp_coordinates_x[3];
         double temp_coordinates_y[3];
         double coords[3][3];

         for (int i = 0; i < 3; ++i) {
            temp_coordinates_x[i] = coordinates_x[(3+i*order+start)%3];
            temp_coordinates_y[i] = coordinates_y[(3+i*order+start)%3];
            LLtoXYZ(temp_coordinates_x[i], temp_coordinates_y[i], coords[i]);
         }

         struct grid_cell cell = {.coordinates_x = temp_coordinates_x,
                                  .coordinates_y = temp_coordinates_y,
                                  .coordinates_xyz = (double[3*3]){-1},
                                  .edge_type = edges,
                                  .num_corners = 3};

         for (int i = 0; i < cell.num_corners; ++i)
            LLtoXYZ(cell.coordinates_x[i], cell.coordinates_y[i],
                    cell.coordinates_xyz + i * 3);

         struct bounding_circle bnd_circle[3];

         yac_get_cell_bounding_circle(cell, bnd_circle+0);
         yac_get_cell_bounding_circle_unstruct_triangle(
            coords[0], coords[1], coords[2], bnd_circle+1);
         yac_get_cell_circumscribe_circle_unstruct_triangle(
            coords[0], coords[1], coords[2], bnd_circle+2);

         if (bnd_circle[1].inc_angle - tol > bnd_circle[0].inc_angle)
            PUT_ERR("get_cell_bounding_circle_unstruct_triangle did not "
                    "compute smallest bounding circle\n");

         test_circle(cell, bnd_circle[0]);
         test_circle(cell, bnd_circle[1]);
         test_circle(cell, bnd_circle[2]);
         test_cirumscribe_circle(cell, bnd_circle[2]);
      }
   }
}

void check_gc_quad(double * coordinates_x, double * coordinates_y) {

   enum yac_edge_type edges[] = {GREAT_CIRCLE, GREAT_CIRCLE,
                                 GREAT_CIRCLE, GREAT_CIRCLE};

   for (int order = -1; order <= 1; order += 2) {

      for (int start = 0; start < 4; ++start) {

         double temp_coordinates_x[4];
         double temp_coordinates_y[4];

         for (int i = 0; i < 4; ++i) {
            temp_coordinates_x[i] = coordinates_x[(4+i*order+start)%4];
            temp_coordinates_y[i] = coordinates_y[(4+i*order+start)%4];
         }

         struct grid_cell cell = {.coordinates_x = temp_coordinates_x,
                                  .coordinates_y = temp_coordinates_y,
                                  .coordinates_xyz = (double[4*3]){-1},
                                  .edge_type = edges,
                                  .num_corners = 4};

         for (int i = 0; i < cell.num_corners; ++i)
            LLtoXYZ(cell.coordinates_x[i], cell.coordinates_y[i],
                    cell.coordinates_xyz + i * 3);

         struct bounding_circle bnd_circle;

         yac_get_cell_bounding_circle(cell, &bnd_circle);

         test_circle(cell, bnd_circle);
      }
   }
}

void check_latlon_cell(double * coordinates_x, double * coordinates_y) {

   for (int order = -1; order <= 1; order += 2) {

      for (int start = 0; start < 4; ++start) {

         double temp_coordinates_x[4];
         double temp_coordinates_y[4];
         enum yac_edge_type edges[4];
         double coords[4][3];

         for (int i = 0; i < 4; ++i) {
            temp_coordinates_x[i] = coordinates_x[(4+i*order+start)%4];
            temp_coordinates_y[i] = coordinates_y[(4+i*order+start)%4];
            LLtoXYZ(temp_coordinates_x[i], temp_coordinates_y[i], coords[i]);
         }

         for (int i = 0; i < 4; ++i) {
            if (temp_coordinates_x[i] == temp_coordinates_x[(i+3)%4] &&
                temp_coordinates_y[i] != temp_coordinates_y[(i+3)%4])
               edges[i] = LAT_CIRCLE;
            else if (temp_coordinates_y[i] == temp_coordinates_y[(i+3)%4] &&
                     temp_coordinates_x[i] != temp_coordinates_x[(i+3)%4])
               edges[i] = LON_CIRCLE;
            else
               yac_internal_abort_message("internal error\n", __FILE__, __LINE__);
         }

         struct grid_cell cell = {.coordinates_x = temp_coordinates_x,
                                  .coordinates_y = temp_coordinates_y,
                                  .coordinates_xyz = (double[4*3]){-1},
                                  .edge_type = edges,
                                  .num_corners = 4};

         for (int i = 0; i < cell.num_corners; ++i)
            LLtoXYZ(cell.coordinates_x[i], cell.coordinates_y[i],
                    cell.coordinates_xyz + i * 3);

         struct bounding_circle bnd_circle[3];

         yac_get_cell_bounding_circle(cell, bnd_circle+0);
         yac_get_cell_bounding_circle_reg_quad(coords[0], coords[1],
                                               coords[2], coords[3],
                                               bnd_circle+1);
         yac_get_cell_circumscribe_circle_reg_quad(coords[0], coords[1],
                                                   coords[2], coords[3],
                                                   bnd_circle+2);

         if (bnd_circle[1].inc_angle - tol > bnd_circle[0].inc_angle)
            PUT_ERR("get_cell_bounding_circle_reg_quad did not "
                    "compute smallest bounding circle\n");

         test_circle(cell, bnd_circle[0]);
         test_circle(cell, bnd_circle[1]);
         test_circle(cell, bnd_circle[2]);
         test_cirumscribe_circle(cell, bnd_circle[2]);
      }
   }
}

void test_circle(struct grid_cell cell, struct bounding_circle circle) {

   for (unsigned i = 0; i < cell.num_corners; ++i)
      if (!point_in_circle(cell.coordinates_xyz + i*3, circle))
         PUT_ERR("point is not in bounding circle\n");
}

void test_cirumscribe_circle(struct grid_cell cell,
                             struct bounding_circle circle) {

   for (unsigned i = 0; i < cell.num_corners; ++i) {

      double angle = get_vector_angle(circle.base_vector,
                                      cell.coordinates_xyz + i * 3);

      if (fabs(circle.inc_angle - angle) > tol)
         PUT_ERR("not all points are on the circumscribe circle\n");
   }
}

unsigned point_in_circle(double point[3], struct bounding_circle circle) {

   double const tol = 1.0e-12;
   double const pi = 3.14159265358979323846;

   if (circle.inc_angle + tol >= pi) return 1 == 1;

   return get_vector_angle(circle.base_vector, point) <=
          circle.inc_angle + tol;
}
