/**
 * @file test_communicator_local.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include <stdlib.h>
#include <stdio.h>

#include "tests.h"

#include "communicator.h"
#include "communicator_local.h"

unsigned num_errors = 0;
unsigned num_received_messages = 0;
unsigned num_canceled_messages = 0;

// callback function passed to yac_comm_recv_callback
void recv_callback(struct communicator * comm, struct comm_message message,
                   void * orig_data);
void recv_callback_cancel(struct communicator * comm, void * orig_data);
// callback function passed to yac_comm_bcast_callback
void bcast_callback(struct communicator * comm, struct comm_message message,
                    void * orig_data);
// callback function passed to yac_comm_allgather_callback
void allgather_callback(struct communicator * comm, struct comm_message message,
                        void * orig_data);
// callback function passed to yac_comm_gatherv_callback
void gatherv_callback(struct communicator * comm, struct comm_message message,
                      void * orig_data);
// callback function passed to yac_comm_allgatherv_callback
void allgatherv_callback(struct communicator * comm, struct comm_message message,
                         void * orig_data);
// callback function passed to yac_comm_scatter_callback
void scatter_callback(struct communicator * comm, struct comm_message message,
                      void * user_data);
// callback function passed to yac_comm_scatterv_callback
void scatterv_callback(struct communicator * comm, struct comm_message message,
                       void * orig_data);
// callback function passed to yac_comm_alltoall_callback
void alltoallv_callback(struct communicator * comm, struct comm_message message,
                        void * orig_data);
// callback function passed to yac_comm_reduce_callback
void reduce_callback(struct communicator * comm, struct comm_message message,
                     void * orig_data);
// callback function passed to yac_comm_allreduce_callback
void allreduce_callback(struct communicator * comm, struct comm_message message,
                        void * orig_data);
// callback function passed to yac_comm_barrier_callback
void barrier_callback(struct communicator * comm, struct comm_message message,
                      void * orig_data);

int main (void) {

   { // tests with a single local rank
      struct communicator * comm;
      struct communicator * child_comm;

      comm = yac_local_comm_new();
      child_comm = yac_named_local_comm_new(comm, "child");

      { // tests the size of the communicator
         if (yac_get_comm_size(comm) != 1)
            PUT_ERR("error in get_comm_size\n")
      }

      { // test generation of unique ids
         char const * tag_strings[] = {"corner", "cell", "edge"};
         unsigned ids[3];

         yac_get_unique_comm_tags(
          comm, "generator string", tag_strings, 3, ids);

         if (ids[0] == ids[1] || ids[0] == ids[2] || ids[1] == ids[2])
            PUT_ERR("error in get_unique_comm_tags\n")

         unsigned tag;

         yac_get_unique_comm_tags(
          comm, "generator string", tag_strings, 1, &tag);

         if (tag != ids[0])
            PUT_ERR("error in get_unique_comm_tags\n")
      }

      { // tests sending of data (first send then recv)

         unsigned send_data[4] = {0,1,2,3};

         num_received_messages = 0;

         yac_comm_bsend(send_data, 4, COMM_UINT, 0, 0, comm);

         yac_comm_recv_callback(4, COMM_UINT, 0, 0, comm, send_data,
                                recv_callback, NULL, MSG_PRIO_HIGH);

         yac_comm_wait(comm, 0);

         if (num_received_messages != 1)
            PUT_ERR("error in comm_recv_callback\n")
      }

      { // test sending of data (first recv then send)

         unsigned send_data[4] = {0,1,2,3};

         num_received_messages = 0;

         yac_comm_recv_callback(4, COMM_UINT, 0, 0, comm, send_data,
                                recv_callback, NULL, MSG_PRIO_HIGH);

         yac_comm_bsend(send_data, 4, COMM_UINT, 0, 0, comm);

         yac_comm_wait(comm, 0);

         if (num_received_messages != 1)
            PUT_ERR("error in comm_recv_callback\n")
      }

      { // test sending of two data packages (first send then recv)

         unsigned send_data[2][4] = {{0,1,2,3}, {3,2,1,0}};

         num_received_messages = 0;

         yac_comm_bsend(send_data[0], 4, COMM_UINT, 0, 0, comm);
         yac_comm_bsend(send_data[1], 4, COMM_UINT, 0, 0, comm);

         yac_comm_recv_callback(4, COMM_UINT, 0, 0, comm, send_data[0],
                                recv_callback, NULL, MSG_PRIO_HIGH);
         yac_comm_recv_callback(4, COMM_UINT, 0, 0, comm, send_data[1],
                                recv_callback, NULL, MSG_PRIO_HIGH);

         yac_comm_wait(comm, 0);

         if (num_received_messages != 2)
            PUT_ERR("error in comm_recv_callback\n")
      }

      { // test sending of two data packages (first recv then send)

         unsigned send_data[2][4] = {{0,1,2,3}, {3,2,1,0}};

         num_received_messages = 0;

         yac_comm_recv_callback(4, COMM_UINT, 0, 0, comm, send_data[0],
                                recv_callback, NULL, MSG_PRIO_HIGH);
         yac_comm_recv_callback(4, COMM_UINT, 0, 0, comm, send_data[1],
                                recv_callback, NULL, MSG_PRIO_HIGH);

         yac_comm_bsend(send_data[0], 4, COMM_UINT, 0, 0, comm);
         yac_comm_bsend(send_data[1], 4, COMM_UINT, 0, 0, comm);

         yac_comm_wait(comm, 0);

         if (num_received_messages != 2)
            PUT_ERR("error in comm_recv_callback\n")
      }

      { // test sending of two data packages with different tags

         unsigned send_data[2][4] = {{0,1,2,3}, {3,2,1,0}};

         num_received_messages = 0;

         yac_comm_bsend(send_data[0], 4, COMM_UINT, 0, 0, comm);
         yac_comm_bsend(send_data[1], 4, COMM_UINT, 0, 1, comm);

         yac_comm_recv_callback(4, COMM_UINT, 0, 0, comm, send_data[0],
                                recv_callback, NULL, MSG_PRIO_HIGH);
         yac_comm_recv_callback(4, COMM_UINT, 0, 1, comm, send_data[1],
                                recv_callback, NULL, MSG_PRIO_HIGH);

         yac_comm_wait(comm, 0);
         yac_comm_wait(comm, 1);

         if (num_received_messages != 2)
            PUT_ERR("error in comm_recv_callback\n")
      }

      { // test sending of two data packages with different tags

         unsigned send_data[2][4] = {{0,1,2,3}, {3,2,1,0}};

         num_received_messages = 0;

         yac_comm_bsend(send_data[0], 4, COMM_UINT, 0, 0, comm);
         yac_comm_bsend(send_data[1], 4, COMM_UINT, 0, 1, comm);

         yac_comm_recv_callback(4, COMM_UINT, 0, 1, comm, send_data[1],
                                recv_callback, NULL, MSG_PRIO_HIGH);
         yac_comm_recv_callback(4, COMM_UINT, 0, 0, comm, send_data[0],
                                recv_callback, NULL, MSG_PRIO_HIGH);

         yac_comm_wait(comm, 0);
         yac_comm_wait(comm, 1);

         if (num_received_messages != 2)
            PUT_ERR("error in comm_recv_callback\n")
      }

      { // test sending of two data packages with different tags using blocking recv

         unsigned send_data[2][4] = {{0,1,2,3}, {3,2,1,0}};
         unsigned recv_data[2][4];

         num_received_messages = 0;

         yac_comm_bsend(send_data[0], 4, COMM_UINT, 0, 0, comm);
         yac_comm_bsend(send_data[1], 4, COMM_UINT, 0, 1, comm);

         unsigned recv_count[2] = {4,4};

         yac_comm_recv(recv_data[0], recv_count+0, COMM_UINT, 0, 0, comm);
         yac_comm_recv(recv_data[1], recv_count+1, COMM_UINT, 0, 1, comm);

         if (recv_count[0] != 4 || recv_count[1] != 4)
            PUT_ERR("error in comm_recv count\n")

         unsigned i;

         for (i = 0; i < 4; ++i)
            if (recv_data[0][i] != send_data[0][i] ||
                recv_data[1][i] != send_data[1][i])
               PUT_ERR("error in comm_recv recv_data\n")
      }

      { // test sending of two data packages with different tags using blocking recv

         unsigned send_data[2][4] = {{0,1,2,3}, {3,2,1,0}};
         unsigned recv_data[2][4];

         num_received_messages = 0;

         yac_comm_bsend(send_data[0], 4, COMM_UINT, 0, 0, comm);
         yac_comm_bsend(send_data[1], 4, COMM_UINT, 0, 1, comm);

         unsigned recv_count[2] = {4,4};

         yac_comm_recv(recv_data[1], recv_count+1, COMM_UINT, 0, 1, comm);
         yac_comm_recv(recv_data[0], recv_count+0, COMM_UINT, 0, 0, comm);

         if (recv_count[0] != 4 || recv_count[1] != 4)
            PUT_ERR("error in comm_recv count\n")

         unsigned i;

         for (i = 0; i < 4; ++i)
            if (recv_data[0][i] != send_data[0][i] ||
                recv_data[1][i] != send_data[1][i])
               PUT_ERR("error in comm_recv recv_data\n")
      }

      { // test translation of ranks

         unsigned rank = -1;

         yac_comm_translate_ranks(comm, child_comm, &rank);

         if (rank != 0)
            PUT_ERR("error in comm_translate_ranks\n");
      }

      { // test broadcast with callback

        unsigned data[3] = {0,1,2};
        unsigned tag = 1;
        unsigned root = 0;
        // contents {root, data size, ranks of proc that received the bcast}
        unsigned user_data[3] = {root, sizeof(data)/sizeof(data[0]), -1};

        yac_comm_bcast_callback((void*)data, sizeof(data)/sizeof(data[0]),
                                COMM_UINT, root, tag, comm, user_data,
                                bcast_callback, MSG_PRIO_HIGH);
        yac_comm_wait(comm, tag);

        for (unsigned i = 0; i < yac_get_comm_size(comm); ++i)
          if (user_data[2 + i] != i)
            PUT_ERR("error in yac_comm_bcast_callback\n");
      }

      { // test gatherv with callback

        unsigned data[1] = {0};
        unsigned tag = 1;
        unsigned counts[1] = {1};
        unsigned root = 0;
        unsigned user_data[2] = {root, 0};

        yac_comm_gatherv_callback(data, counts, COMM_UINT, root, tag, comm,
                                  user_data, gatherv_callback,
                                  MSG_PRIO_HIGH);
        yac_comm_wait(comm, tag);

        if (user_data[1] != yac_get_comm_size(comm))
          PUT_ERR("error in yac_comm_gatherv_callback\n");
      }

      { // test allgather with callback

        unsigned data[2] = {0, 1};
        unsigned tag = 1;
        unsigned count = 2;
        unsigned user_data[2] = {count, 0};

        yac_comm_allgather_callback(data, count, COMM_UINT, tag, comm,
                                    user_data, allgather_callback,
                                    MSG_PRIO_HIGH);

        yac_comm_wait(comm, tag);

        if (user_data[1] != yac_get_comm_size(comm))
          PUT_ERR("error in yac_comm_allgather_callback\n");
      }

      { // test allgatherv with callback

        unsigned data[1] = {0};
        unsigned tag = 1;
        unsigned counts[1] = {1};
        unsigned user_data[1] = {0};

        yac_comm_allgatherv_callback(data, counts, COMM_UINT, tag, comm,
                                     user_data, allgatherv_callback,
                                     MSG_PRIO_HIGH);

        yac_comm_wait(comm, tag);

        if (user_data[0] != yac_get_comm_size(comm))
          PUT_ERR("error in yac_comm_allgatherv_callback\n");
      }

      { // test scatter with callback

        unsigned data[1] = {0};
        unsigned tag = 1;
        unsigned count = 1;
        unsigned root = 0;
        unsigned user_data[3] = {root, count, 0};

        yac_comm_scatter_callback(data, count, COMM_UINT, root, tag, comm,
                                  user_data, scatter_callback, MSG_PRIO_HIGH);

        yac_comm_wait(comm, tag);

        if (user_data[2] != yac_get_comm_size(comm))
          PUT_ERR("error in yac_comm_scatter_callback\n");

      }

      { // test scatterv with callback

        unsigned data[2] = {0};
        unsigned tag = 1;
        unsigned counts[1] = {1};
        unsigned root = 0;
        unsigned user_data[2] = {root,0};

        yac_comm_scatterv_callback(data, counts, COMM_UINT, root, tag, comm,
                                   user_data, scatterv_callback, MSG_PRIO_HIGH);

        yac_comm_wait(comm, tag);

        if (user_data[1] != yac_get_comm_size(comm))
          PUT_ERR("error in yac_comm_scatterv_callback\n");

      }

      { // test alltoallv with callback

        unsigned data[1] = {0};
        unsigned tag = 1;
        unsigned counts[1] = {1};
        unsigned user_data[1] = {0};

        yac_comm_alltoallv_callback(data, counts, COMM_UINT, tag, comm,
                                    user_data, alltoallv_callback,
                                    MSG_PRIO_HIGH);

        yac_comm_wait(comm, tag);

        if (user_data[0] != yac_get_comm_size(comm))
          PUT_ERR("error in yac_comm_alltoallv_callback\n");
      }

      { // test reduce with callback

        unsigned data[3] = {0, 1, 2};
        unsigned tag = 1;
        unsigned count = 3;
        unsigned root = 0;
        unsigned user_data[7] = {0, 3, COMM_UINT, root, 0, 1, 2};

        yac_comm_reduce_callback(data, count, COMM_UINT, COMM_OP_SUM, root, tag,
                                 comm, user_data, reduce_callback,
                                 MSG_PRIO_HIGH);

        yac_comm_wait(comm, tag);

        if (user_data[0] != yac_get_comm_size(comm))
          PUT_ERR("error in yac_comm_reduce_callback\n");
      }

      { // test allreduce with callback

        unsigned data[3] = {0, 1, 2};
        unsigned tag = 1;
        unsigned count = 3;
        unsigned user_data[6] = {0, 3, COMM_UINT, 0, 1, 2};

        yac_comm_allreduce_callback(data, count, COMM_UINT, COMM_OP_SUM, tag,
                                    comm, user_data, allreduce_callback,
                                    MSG_PRIO_HIGH);

        yac_comm_wait(comm, tag);

        if (user_data[0] != yac_get_comm_size(comm))
          PUT_ERR("error in yac_comm_allreduce_callback\n");
      }

      { // test barrier with callback

        unsigned tag = 1;
        unsigned user_data = 0;

        yac_comm_barrier_callback(tag, comm, &user_data, barrier_callback,
                                  MSG_PRIO_HIGH);
        yac_comm_wait(comm, tag);

        if (user_data != yac_get_comm_size(comm))
          PUT_ERR("error in yac_comm_barrier_callback\n");
      }

      { // test irecv and waitall

        for (unsigned i = 0; i < 2; ++i) {
          unsigned tag = 1;
          unsigned send_buffer[1] = {0};
          unsigned recv_buffer[1] = {-1};
          struct comm_request * requests[1];

          if (i == 0) yac_comm_bsend(send_buffer, 1, COMM_UINT, 0, tag, comm);
          yac_comm_irecv(recv_buffer, 1, COMM_UINT, 0, tag, comm, requests);
          if (i == 1) yac_comm_bsend(send_buffer, 1, COMM_UINT, 0, tag, comm);
          yac_comm_waitall(1, requests);

          if (recv_buffer[0] != 0)
            PUT_ERR("error in yac_comm_irecv/yac_comm_waitall\n");
        }
        for (unsigned i = 0; i < 2; ++i) {
          unsigned tag[2] = {1, 2};
          unsigned send_buffer[2][2] = {{0,-2},{1,2}};
          unsigned recv_buffer[2][2] = {{-1,-1},{-1,-1}};
          struct comm_request * requests[2];

          if (i == 0) yac_comm_bsend(send_buffer[0], 1, COMM_UINT, 0, tag[0], comm);
          if (i == 0) yac_comm_bsend(send_buffer[1], 2, COMM_UINT, 0, tag[1], comm);
          yac_comm_irecv(recv_buffer[0], 1, COMM_UINT, 0, tag[0], comm, &(requests[0]));
          yac_comm_irecv(recv_buffer[1], 2, COMM_UINT, 0, tag[1], comm, &(requests[1]));
          if (i == 1) yac_comm_bsend(send_buffer[0], 1, COMM_UINT, 0, tag[0], comm);
          if (i == 1) yac_comm_bsend(send_buffer[1], 2, COMM_UINT, 0, tag[1], comm);
          yac_comm_waitall(2, requests);

          if ((recv_buffer[0][0] != 0) || (recv_buffer[0][1] != -1) ||
              (recv_buffer[1][0] != 1) || (recv_buffer[1][1] != 2))
            PUT_ERR("error in yac_comm_irecv/yac_comm_waitall\n");
        }
      }

      { // test isend, irecv and waitall

        for (unsigned i = 0; i < 2; ++i) {
          unsigned tag = 1;
          unsigned send_buffer[1] = {0};
          unsigned recv_buffer[1] = {-1};
          struct comm_request * requests[2];

          if (i == 0) yac_comm_isend(send_buffer, 1, COMM_UINT, 0, tag, comm,
                                     &(requests[1]));
          yac_comm_irecv(recv_buffer, 1, COMM_UINT, 0, tag, comm,
                         &(requests[0]));
          if (i == 1) yac_comm_isend(send_buffer, 1, COMM_UINT, 0, tag, comm,
                                     &(requests[1]));
          yac_comm_waitall(2, requests);

          if (recv_buffer[0] != 0)
            PUT_ERR("error in yac_comm_irecv/yac_comm_waitall\n");
        }
        for (unsigned i = 0; i < 2; ++i) {
          unsigned tag[2] = {1, 2};
          unsigned send_buffer[2][2] = {{0,-2},{1,2}};
          unsigned recv_buffer[2][2] = {{-1,-1},{-1,-1}};
          struct comm_request * requests[4];

          if (i == 0) yac_comm_isend(send_buffer[0], 1, COMM_UINT, 0, tag[0],
                                     comm, &(requests[0]));
          if (i == 0) yac_comm_isend(send_buffer[1], 2, COMM_UINT, 0, tag[1],
                                     comm, &(requests[1]));
          yac_comm_irecv(recv_buffer[0], 1, COMM_UINT, 0, tag[0], comm,
                         &(requests[2]));
          yac_comm_irecv(recv_buffer[1], 2, COMM_UINT, 0, tag[1], comm,
                         &(requests[3]));
          if (i == 1) yac_comm_isend(send_buffer[0], 1, COMM_UINT, 0, tag[0],
                                     comm, &(requests[0]));
          if (i == 1) yac_comm_isend(send_buffer[1], 2, COMM_UINT, 0, tag[1],
                                     comm, &(requests[1]));
          yac_comm_waitall(4, requests);

          if ((recv_buffer[0][0] != 0) || (recv_buffer[0][1] != -1) ||
              (recv_buffer[1][0] != 1) || (recv_buffer[1][1] != 2))
            PUT_ERR("error in yac_comm_irecv/yac_comm_waitall\n");
        }
      }

      yac_free_communicator(child_comm);
      yac_free_communicator(comm);
   }

   { // tests with two local rank (two local communicators)
      struct communicator * comm_a, * comm_b;
      struct communicator * child_comm_a, * child_comm_b;

      comm_a = yac_local_comm_new();
      comm_b = yac_local_comm_new();

      child_comm_a = yac_named_local_comm_new(comm_a, "child");
      child_comm_b = yac_named_local_comm_new(comm_b, "child");

      { // tests size of communicator
         if ((yac_get_comm_size(comm_a) != 2) || (yac_get_comm_size(comm_b) != 2))
            PUT_ERR("error in get_comm_size\n")
      }

      { // tests rank returned by the two communicators (one for each virtual process)
         if ((yac_get_comm_rank(comm_a) != 0) || (yac_get_comm_rank(comm_b) != 1))
            PUT_ERR("error in get_comm_rank\n")
      }

      { // test sending of two data packages with different tags

         unsigned send_data[2][4] = {{0,1,2,3}, {3,2,1,0}};

         num_received_messages = 0;

         yac_comm_bsend(send_data[0], 4, COMM_UINT, 0, 0, comm_a);
         yac_comm_bsend(send_data[1], 4, COMM_UINT, 0, 1, comm_a);

         yac_comm_recv_callback(4, COMM_UINT, 0, 1, comm_a, send_data[1],
                                recv_callback, NULL, MSG_PRIO_HIGH);
         yac_comm_recv_callback(4, COMM_UINT, 0, 0, comm_a, send_data[0],
                                recv_callback, NULL, MSG_PRIO_HIGH);

         yac_comm_wait(comm_a, 0);
         yac_comm_wait(comm_a, 1);

         if (num_received_messages != 2)
            PUT_ERR("error in comm_recv_callback\n")

         num_received_messages = 0;
      }

      { // test sending of two data packages with different tags

         unsigned send_data[2][4] = {{0,1,2,3}, {3,2,1,0}};

         num_received_messages = 0;

         yac_comm_bsend(send_data[0], 4, COMM_UINT, 1, 0, comm_b);
         yac_comm_bsend(send_data[1], 4, COMM_UINT, 1, 1, comm_b);

         yac_comm_recv_callback(4, COMM_UINT, 1, 1, comm_b, send_data[1],
                                recv_callback, NULL, MSG_PRIO_HIGH);
         yac_comm_recv_callback(4, COMM_UINT, 1, 0, comm_b, send_data[0],
                                recv_callback, NULL, MSG_PRIO_HIGH);

         yac_comm_wait(comm_b, 0);
         yac_comm_wait(comm_b, 1);

         if (num_received_messages != 2)
            PUT_ERR("error in comm_recv_callback\n")

         num_received_messages = 0;
      }

      { // test sending of two data packages with different tags

         unsigned send_data[2][4] = {{0,1,2,3}, {3,2,1,0}};

         num_received_messages = 0;

         yac_comm_bsend(send_data[0], 4, COMM_UINT, 1, 0, comm_a);
         yac_comm_bsend(send_data[1], 4, COMM_UINT, 1, 1, comm_a);

         yac_comm_recv_callback(4, COMM_UINT, 0, 1, comm_b, send_data[1],
                                recv_callback, NULL, MSG_PRIO_HIGH);
         yac_comm_recv_callback(4, COMM_UINT, 0, 0, comm_b, send_data[0],
                                recv_callback, NULL, MSG_PRIO_HIGH);

         yac_comm_wait(comm_a, 0);
         yac_comm_wait(comm_a, 1);
         yac_comm_wait(comm_b, 0);
         yac_comm_wait(comm_b, 1);

         if (num_received_messages != 2)
            PUT_ERR("error in comm_recv_callback\n")

         num_received_messages = 0;
      }

      { // test sending of two data packages with different tags

         unsigned send_data[2][4] = {{0,1,2,3}, {3,2,1,0}};

         num_received_messages = 0;

         yac_comm_bsend(send_data[0], 4, COMM_UINT, 0, 0, comm_b);
         yac_comm_bsend(send_data[1], 4, COMM_UINT, 0, 1, comm_b);

         yac_comm_recv_callback(4, COMM_UINT, 1, 1, comm_a, send_data[1],
                                recv_callback, NULL, MSG_PRIO_HIGH);
         yac_comm_recv_callback(4, COMM_UINT, 1, 0, comm_a, send_data[0],
                                recv_callback, NULL, MSG_PRIO_HIGH);

         yac_comm_wait(comm_a, 0);
         yac_comm_wait(comm_a, 1);
         yac_comm_wait(comm_b, 0);
         yac_comm_wait(comm_b, 1);

         if (num_received_messages != 2)
            PUT_ERR("error in comm_recv_callback\n")

         num_received_messages = 0;
      }

      { // test exchange of two data packages with different tags between processes

         unsigned send_data[2][4] = {{0,1,2,3}, {3,2,1,0}};

         num_received_messages = 0;

         yac_comm_recv_callback(4, COMM_UINT, 1, 1, comm_a, send_data[0],
                                recv_callback, NULL, MSG_PRIO_HIGH);
         yac_comm_recv_callback(4, COMM_UINT, 0, 2, comm_b, send_data[0],
                                recv_callback, NULL, MSG_PRIO_HIGH);
         yac_comm_recv_callback(4, COMM_UINT, 1, 3, comm_a, send_data[1],
                                recv_callback, NULL, MSG_PRIO_HIGH);
         yac_comm_recv_callback(4, COMM_UINT, 0, 4, comm_b, send_data[1],
                                recv_callback, NULL, MSG_PRIO_HIGH);

         yac_comm_bsend(send_data[0], 4, COMM_UINT, 1, 2, comm_a);
         yac_comm_bsend(send_data[1], 4, COMM_UINT, 1, 4, comm_a);
         yac_comm_bsend(send_data[0], 4, COMM_UINT, 0, 1, comm_b);
         yac_comm_bsend(send_data[1], 4, COMM_UINT, 0, 3, comm_b);

         yac_comm_wait(comm_a, 1), yac_comm_wait(comm_b, 1);
         yac_comm_wait(comm_a, 2), yac_comm_wait(comm_b, 2);
         yac_comm_wait(comm_a, 3), yac_comm_wait(comm_b, 3);
         yac_comm_wait(comm_a, 4), yac_comm_wait(comm_b, 4);

         if (num_received_messages != 4)
            PUT_ERR("error in comm_recv_callback\n")

         num_received_messages = 0;
      }

      { // test packing and unpacking

         double   send_data_dble[2][2] = {{0,1},{2,3}};
         int      send_data_int [2][2] = {{4,5},{6,7}};
         unsigned send_data_uint[2][2] = {{8,9},{10,11}};

         if (yac_comm_pack_size(2, COMM_DBLE, comm_a) != 2 * sizeof(double))
            PUT_ERR("error in comm_pack_size\n");

         if (yac_comm_pack_size(2, COMM_INT,  comm_a) != 2 * sizeof(int))
            PUT_ERR("error in comm_pack_size\n");

         if (yac_comm_pack_size(2, COMM_UINT, comm_a) != 2 * sizeof(unsigned))
            PUT_ERR("error in comm_pack_size\n");

         char send_buffer[2][2*sizeof(double)+2*sizeof(int)+2*sizeof(unsigned)];
         struct communicator * comm[2] = {comm_a, comm_b};

         for (int i = 0; i < 2; ++i) {

            int position = 0;

            yac_comm_pack(send_data_dble[i], 2, COMM_DBLE, send_buffer[i],
                          2*sizeof(double)+2*sizeof(int)+2*sizeof(unsigned),
                          &position, comm[i]);
            yac_comm_pack(send_data_int[i], 2, COMM_INT, send_buffer[i],
                          2*sizeof(double)+2*sizeof(int)+2*sizeof(unsigned),
                          &position, comm[i]);
            yac_comm_pack(send_data_uint[i], 2, COMM_UINT, send_buffer[i],
                          2*sizeof(double)+2*sizeof(int)+2*sizeof(unsigned),
                          &position, comm[i]);
         }

         unsigned count = 2*sizeof(double)+2*sizeof(int)+2*sizeof(unsigned);

         for (int i = 0; i < 2; ++i)
            yac_comm_bsend(send_buffer[i], count, COMM_PACKED, i^1, 0, comm[i]);

         char recv_buffer[2][count];

         for (int i = 0; i < 2; ++i)
            yac_comm_recv(recv_buffer[i], &count, COMM_PACKED, i^1, 0, comm[i]);

         double   recv_data_dble[2];
         int      recv_data_int [2];
         unsigned recv_data_uint[2];

         for (int i = 0; i < 2; ++i) {

            int position = 0;

            yac_comm_unpack(recv_buffer[i], count, &position, recv_data_dble, 2, COMM_DBLE,
                            comm[i]);
            yac_comm_unpack(recv_buffer[i], count, &position, recv_data_int, 2, COMM_INT,
                            comm[i]);
            yac_comm_unpack(recv_buffer[i], count, &position, recv_data_uint, 2, COMM_UINT,
                            comm[i]);

            for (int j = 0; j < 2; ++j) {

               if (recv_data_dble[j] != send_data_dble[i^1][j])
                  PUT_ERR("error in packing/unpacking (dble)\n");
               if (recv_data_int[j] != send_data_int[i^1][j])
                  PUT_ERR("error in packing/unpacking (int)\n");
               if (recv_data_uint[j] != send_data_uint[i^1][j])
                  PUT_ERR("error in packing/unpacking (uint)\n");
            }
         }
      }

      { // test translation of ranks

         unsigned rank[2][2] = {{-1,-1},{-1,-1}};

         yac_comm_translate_ranks(comm_a, child_comm_a, rank[0]);
         yac_comm_translate_ranks(comm_b, child_comm_b, rank[1]);

         for (int i = 0; i < 4; ++i)
            if (rank[i&1][i>>1] != i>>1)
               PUT_ERR("error in comm_translate_ranks\n");
      }

      { // test broadcast with callback

        struct communicator * comms[2] = {comm_a, comm_b};

        unsigned data[5] = {0,1,2,3,4};
        unsigned tag = 1;

        // for all possible root's
        for (unsigned j = 0; j < 2; ++j) {

          unsigned root = j;
          // contents {root, data size, ranks of proc that received the bcast}
          unsigned user_data[4] = {root, sizeof(data)/sizeof(data[0]), -1, -1};

          // for all processes
          for (unsigned k = 0; k < 2; ++k)
            yac_comm_bcast_callback((void*)data, sizeof(data)/sizeof(data[0]),
                                    COMM_UINT, root, tag, comms[k], user_data,
                                    bcast_callback, MSG_PRIO_HIGH);
          for (unsigned k = 0; k < 2; ++k)
            yac_comm_wait(comms[k], tag);

          for (unsigned i = 0; i < yac_get_comm_size(comms[j]); ++i)
            if (user_data[2 + i] != i)
              PUT_ERR("error in yac_comm_bcast_callback\n");
        }
      }

      { // test gatherv with callback

        struct communicator * comms[2] = {comm_a, comm_b};

        unsigned data[2][2] = {{0}, {1, 2}};
        unsigned tag = 1;
        unsigned counts[2] = {1, 2};

        // for all possible root's
        for (unsigned i = 0; i < 2; ++i) {

          unsigned root = i;

          unsigned user_data[2] = {root, 0};

          // for all ranks
          for (unsigned j = 0; j < 2; ++j)
            yac_comm_gatherv_callback(data[j], counts, COMM_UINT, root, tag,
                                      comms[j], user_data, gatherv_callback,
                                      MSG_PRIO_HIGH);
          for (unsigned j = 0; j < 2; ++j)
            yac_comm_wait(comms[j], tag);

          if (user_data[1] != 2)
            PUT_ERR("error in yac_comm_gatherv_callback\n");
        }
      }

      { // test allgather with callback

        struct communicator * comms[2] = {comm_a, comm_b};

        unsigned data[2][2] = {{0, 1}, {2, 3}};
        unsigned tag = 1;
        unsigned count = 2;
        unsigned user_data[2] = {count, 0};

        // for all ranks
        for (unsigned i = 0; i < 2; ++i)
          yac_comm_allgather_callback(data[i], count, COMM_UINT, tag, comms[i],
                                      user_data, allgather_callback,
                                      MSG_PRIO_HIGH);

        for (unsigned i = 0; i < 2; ++i)
          yac_comm_wait(comms[i], tag);

        if (user_data[1] != yac_get_comm_size(comm_a))
          PUT_ERR("error in yac_comm_allgather_callback\n");
      }

      { // test allgatherv with callback

        struct communicator * comms[2] = {comm_a, comm_b};

        unsigned data[2][2] = {{0, -1}, {1, 2}};
        unsigned tag = 1;
        unsigned counts[2] = {1, 2};
        unsigned user_data[1] = {0};

        // for all ranks
        for (unsigned i = 0; i < 2; ++i)
          yac_comm_allgatherv_callback(data[i], counts, COMM_UINT, tag, comms[i],
                                       user_data, allgatherv_callback,
                                       MSG_PRIO_HIGH);

        for (unsigned i = 0; i < 2; ++i)
          yac_comm_wait(comms[i], tag);

        if (user_data[0] != yac_get_comm_size(comm_a))
          PUT_ERR("error in yac_comm_allgatherv_callback\n");
      }

      { // test scatter with callback

        struct communicator * comms[2] = {comm_a, comm_b};

        unsigned data[4] = {0, 1, 2, 3};
        unsigned tag = 1;
        unsigned count = 2;

        // for all possible root's
        for (unsigned i = 0; i < 2; ++i) {

          unsigned root = i;

          unsigned user_data[3] = {root, count, 0};

          // for all ranks
          for (unsigned j = 0; j < 2; ++j)
            yac_comm_scatter_callback((j == root)?data:NULL, count, COMM_UINT,
                                      root, tag, comms[j], user_data,
                                      scatter_callback, MSG_PRIO_HIGH);

          for (unsigned j = 0; j < 2; ++j)
            yac_comm_wait(comms[j], tag);

          if (user_data[2] != 2)
            PUT_ERR("error in yac_comm_scatter_callback\n");
        }


      }

      { // test scatterv with callback

        struct communicator * comms[2] = {comm_a, comm_b};

        unsigned data[3] = {0, 1, 2};
        unsigned tag = 1;
        unsigned counts[2] = {1, 2};

        // for all possible root's
        for (unsigned i = 0; i < 2; ++i) {

          unsigned root = i;

          unsigned user_data[2] = {root, 0};

          // for all ranks
          for (unsigned j = 0; j < 2; ++j)
            yac_comm_scatterv_callback((j == root)?data:NULL, counts, COMM_UINT,
                                       root, tag, comms[j], user_data,
                                       scatterv_callback, MSG_PRIO_HIGH);

          for (unsigned j = 0; j < 2; ++j)
            yac_comm_wait(comms[j], tag);

          if (user_data[1] != 2)
            PUT_ERR("error in yac_comm_scatterv_callback\n");
        }


      }

      { // test alltoallv with callback

        struct communicator * comms[2] = {comm_a, comm_b};

        unsigned data[2][4] = {{0, 0},{1,2, 1,2}};
        unsigned tag = 1;
        unsigned counts[2][2] = {{1,1},{2,2}};
        unsigned user_data[1] = {0};

        for (unsigned i = 0; i < 2; ++i)
          yac_comm_alltoallv_callback(data[i], counts[i], COMM_UINT, tag,
                                      comms[i], user_data, alltoallv_callback,
                                      MSG_PRIO_HIGH);

        for (unsigned i = 0; i < 2; ++i)
          yac_comm_wait(comms[i], tag);

        if (user_data[0] != 2)
          PUT_ERR("error in yac_comm_alltoallv_callback\n");
      }

      { // test reduce with callback

        struct communicator * comms[2] = {comm_a, comm_b};

        unsigned data[2][3] = {{0, 1, 2}, {2, 1, 0}};
        unsigned tag = 1;
        unsigned count = 3;

        for (unsigned root = 0; root < 2; ++ root) {

          unsigned user_data[7] = {0, 3, COMM_UINT, root, 2, 2, 2};

          for (unsigned i = 0; i < 2; ++i)
            yac_comm_reduce_callback(data[i], count, COMM_UINT, COMM_OP_SUM,
                                     root, tag, comms[i], user_data,
                                     reduce_callback, MSG_PRIO_HIGH);

          for (unsigned i = 0; i < 2; ++i)
            yac_comm_wait(comms[i], tag);

          if (user_data[0] != 2)
            PUT_ERR("error in yac_comm_reduce_callback\n");
        }
      }

      { // test allreduce with callback

        struct communicator * comms[2] = {comm_a, comm_b};

        unsigned data[2][3] = {{0, 1, 2}, {2, 1, 0}};
        unsigned tag = 1;
        unsigned count = 3;
        unsigned user_data[6] = {0, 3, COMM_UINT, 2, 2, 2};

        for (unsigned i = 0; i < 2; ++i)
          yac_comm_allreduce_callback(data[i], count, COMM_UINT, COMM_OP_SUM,
                                      tag, comms[i], user_data,
                                      allreduce_callback, MSG_PRIO_HIGH);

        for (unsigned i = 0; i < 2; ++i)
          yac_comm_wait(comms[i], tag);

        if (user_data[0] != 2)
          PUT_ERR("error in yac_comm_allreduce_callback\n");
      }

      { // test barrier with callback

        struct communicator * comms[2] = {comm_a, comm_b};

        unsigned tag = 1;
        unsigned user_data = 0;

        for (unsigned i = 0; i < 2; ++i)
          yac_comm_barrier_callback(tag, comms[i], &user_data, barrier_callback,
                                    MSG_PRIO_HIGH);
        for (unsigned i = 0; i < 2; ++i)
          yac_comm_wait(comms[i], tag);

        if (user_data != 2)
          PUT_ERR("error in yac_comm_barrier_callback\n");
      }

      yac_free_communicator(child_comm_a);
      yac_free_communicator(child_comm_b);
      yac_free_communicator(comm_a);
      yac_free_communicator(comm_b);
   }

   { // tests with a single local rank
      struct communicator * comm;

      comm = yac_local_comm_new();

      { // test the cancelation of asynchronous receives

         unsigned send_data[2][4] = {{0,1,2,3}, {3,2,1,0}};

         yac_comm_recv_callback(4, COMM_UINT, 0, 1, comm, send_data[0],
                                recv_callback, recv_callback_cancel,
                                MSG_PRIO_HIGH);
         yac_comm_recv_callback(4, COMM_UINT, 0, 2, comm, send_data[0],
                                recv_callback, recv_callback_cancel,
                                MSG_PRIO_HIGH);
         yac_comm_recv_callback(4, COMM_UINT, 0, 3, comm, send_data[1],
                                recv_callback, recv_callback_cancel,
                                MSG_PRIO_HIGH);
         yac_comm_recv_callback(4, COMM_UINT, 0, 4, comm, send_data[1],
                                recv_callback, recv_callback_cancel,
                                MSG_PRIO_HIGH);

         yac_comm_cancel_recv(comm, 1);
         yac_comm_bsend(send_data[0], 4, COMM_UINT, 0, 2, comm);

         if (num_received_messages != 1)
            PUT_ERR("error in comm_recv_callback\n")
         if (num_canceled_messages != 1)
            PUT_ERR("error in comm_recv_callback\n")

         num_received_messages = 0;
         num_canceled_messages = 0;

         yac_comm_bsend(send_data[1], 4, COMM_UINT, 0, 3, comm);
         yac_comm_cancel_recv(comm, 4);

         if (num_received_messages != 1)
            PUT_ERR("error in comm_recv_callback\n")
         if (num_canceled_messages != 1)
            PUT_ERR("error in comm_recv_callback\n")

         num_received_messages = 0;
         num_canceled_messages = 0;
      }
      
      yac_free_communicator(comm);
   }

  { // tests with multiple different number of ranks

    // for a communicator size of 1 to 16
    for (unsigned comm_size = 1; comm_size <= 16; ++comm_size) {

      struct communicator ** comms = malloc(comm_size * sizeof(*comms));

      for (unsigned i = 0; i < comm_size; ++i)
        comms[i] = yac_local_comm_new();

      // func == 0 => no callback routine is provided
      // func == 1 => callback routine is provided
      for (unsigned func = 0; func < 2; ++func) {

        { // test broadcast with callback

          unsigned data[5] = {0,1,2,3,4};
          unsigned tag = 1;

          // for all possible root's
          for (unsigned j = 0; j < comm_size; ++j) {

            unsigned root = j;
            // contents {root, data size, ranks of proc that received the bcast}
            unsigned * user_data = malloc((comm_size + 2) * sizeof(*user_data));
            user_data[0] = root;
            user_data[1] = sizeof(data)/sizeof(data[0]);
            for (unsigned i = 0; i < comm_size; ++i) user_data[i+2] = -1;

            // for all processes
            for (unsigned i = 0; i < comm_size; ++i)
              yac_comm_bcast_callback((void*)data, sizeof(data)/sizeof(data[0]),
                                      COMM_UINT, root, tag, comms[i], user_data,
                                      (func)?bcast_callback:NULL, MSG_PRIO_HIGH);
            for (unsigned i = 0; i < comm_size; ++i) yac_comm_wait(comms[i], tag);

            if (func)
              for (unsigned i = 0; i < comm_size; ++i)
                if (user_data[2 + i] != i)
                  PUT_ERR("error in yac_comm_bcast_callback\n");

            free(user_data);
          }
        }

        { // test gatherv with callback

          unsigned total_size = (comm_size * (comm_size + 1)) / 2;
          unsigned * data = malloc(total_size * sizeof(*data));
          for (unsigned i = 0; i < total_size; ++i) data[i] = i;

          unsigned * counts = malloc(comm_size * sizeof(*counts));
          for (unsigned i = 0; i < comm_size; ++i) counts[i] = i + 1;

          unsigned tag = 1;

          // for all possible root's
          for (unsigned root = 0; root < comm_size; ++root) {

            unsigned user_data[2] = {root, 0};

            // for all ranks
            for (unsigned i = 0; i < comm_size; ++i)
              yac_comm_gatherv_callback(data + (i * (i + 1))/2, counts, COMM_UINT,
                                        root, tag, comms[i], user_data,
                                        (func)?gatherv_callback:NULL, MSG_PRIO_HIGH);
            for (unsigned i = 0; i < comm_size; ++i)
              yac_comm_wait(comms[i], tag);

            if (func)
              if (user_data[1] != comm_size)
                PUT_ERR("error in yac_comm_gatherv_callback\n");
          }

          free(counts);
          free(data);
        }

        { // test allgather with callback

          unsigned * data = malloc(2 * comm_size * sizeof(*data));
          for (unsigned i = 0; i < 2 * comm_size; ++i) data[i] = i;
          unsigned tag = 1;
          unsigned count = 2;
          unsigned user_data[2] = {count, 0};

          // for all ranks
          for (unsigned i = 0; i < comm_size; ++i)
            yac_comm_allgather_callback(data + i * 2, count, COMM_UINT, tag,
                                        comms[i], user_data,
                                        (func)?allgather_callback:NULL,
                                        MSG_PRIO_HIGH);

          for (unsigned i = 0; i < comm_size; ++i)
            yac_comm_wait(comms[i], tag);

          if (func)
            if (user_data[1] != comm_size)
              PUT_ERR("error in yac_comm_allgather_callback\n");

          free(data);
        }

        { // test allgatherv with callback

          unsigned total_size = (comm_size * (comm_size + 1)) / 2;
          unsigned * data = malloc(total_size * sizeof(*data));
          for (unsigned i = 0; i < total_size; ++i) data[i] = i;

          unsigned * counts = malloc(comm_size * sizeof(*counts));
          for (unsigned i = 0; i < comm_size; ++i) counts[i] = i + 1;

          unsigned tag = 1;
          unsigned user_data[1] = {0};

          // for all ranks
          for (unsigned i = 0; i < comm_size; ++i)
            yac_comm_allgatherv_callback(data + (i * (i + 1)) / 2, counts,
                                         COMM_UINT, tag, comms[i], user_data,
                                         (func)?allgatherv_callback:NULL,
                                         MSG_PRIO_HIGH);

          for (unsigned i = 0; i < comm_size; ++i)
            yac_comm_wait(comms[i], tag);

          if (func)
            if (user_data[0] != comm_size)
              PUT_ERR("error in yac_comm_allgatherv_callback\n");

          free(counts);
          free(data);
        }

        { // test scatter with callback

          unsigned count = 3;
          unsigned total_size = count * comm_size;
          unsigned * data = malloc(total_size * sizeof(*data));
          for (unsigned i = 0; i < total_size; ++i) data[i] = i;

          unsigned tag = 1;

          // for all possible root's
          for (unsigned root = 0; root < comm_size; ++root) {

            unsigned user_data[3] = {root, count, 0};

            // for all ranks
            for (unsigned i = 0; i < comm_size; ++i)
              yac_comm_scatter_callback((i == root)?data:NULL, count,
                                        COMM_UINT, root, tag, comms[i],
                                        user_data,
                                        (func)?scatter_callback:NULL,
                                        MSG_PRIO_HIGH);
            for (unsigned i = 0; i < comm_size; ++i)
              yac_comm_wait(comms[i], tag);

            if (func)
              if (user_data[2] != comm_size)
                PUT_ERR("error in yac_comm_scatter_callback\n");
          }

          free(data);
        }

        { // test scatterv with callback

          unsigned total_size = (comm_size * (comm_size + 1)) / 2;
          unsigned * data = malloc(total_size * sizeof(*data));
          for (unsigned i = 0; i < total_size; ++i) data[i] = i;

          unsigned * counts = malloc(comm_size * sizeof(*counts));
          counts[0] = 1;
          for (unsigned i = 1; i < comm_size; ++i) counts[i] = i + 1;

          unsigned tag = 1;

          // for all possible root's
          for (unsigned root = 0; root < comm_size; ++root) {

            unsigned user_data[2] = {root, 0};

            // for all ranks
            for (unsigned i = 0; i < comm_size; ++i)
              yac_comm_scatterv_callback((i == root)?data:NULL, counts,
                                         COMM_UINT, root, tag, comms[i],
                                         user_data,
                                         (func)?scatterv_callback:NULL,
                                         MSG_PRIO_HIGH);
            for (unsigned i = 0; i < comm_size; ++i)
              yac_comm_wait(comms[i], tag);

            if (func)
              if (user_data[1] != comm_size)
                PUT_ERR("error in yac_comm_scatterv_callback\n");
          }

          free(counts);
          free(data);
        }

        { // test alltoallv with callback

          unsigned ** data = malloc(comm_size * sizeof(*data));
          for (unsigned i = 0; i < comm_size; ++i) {
            data[i] = malloc(comm_size * (i + 1) * sizeof(**data));
            for (unsigned j = 0, offset = 0; j < comm_size; ++j)
              for (unsigned k = 0; k <= i; ++k, ++offset)
                data[i][offset] = (i * (i + 1)) / 2 + k;
          }
          unsigned tag = 1;
          unsigned * counts = malloc(comm_size * comm_size * sizeof(*counts));
          for (unsigned i = 0; i < comm_size; ++i)
            for (unsigned j = 0; j < comm_size; ++j)
              counts[comm_size * i + j] = i + 1;
          unsigned user_data[1] = {0};

          for (unsigned i = 0; i < comm_size; ++i)
            yac_comm_alltoallv_callback(data[i], counts + i * comm_size,
                                        COMM_UINT, tag, comms[i], user_data,
                                        (func)?alltoallv_callback:NULL,
                                        MSG_PRIO_HIGH);

          for (unsigned i = 0; i < comm_size; ++i)
            yac_comm_wait(comms[i], tag);

          if (func)
            if (user_data[0] != comm_size)
              PUT_ERR("error in yac_comm_alltoallv_callback\n");

          free(counts);
          for (unsigned i = 0; i < comm_size; ++i) free(data[i]);
          free(data);
        }

        { // test reduce with callback (unsigned)

          unsigned count = 3;
          unsigned  * data = malloc(comm_size * count * sizeof(*data));
          unsigned user_data[7] = {0, count, COMM_UINT, -1, 0, 0, 0};
          for (unsigned i = 0; i < comm_size; ++i) {
            for (unsigned j = 0; j < count; ++j) {
              data[i * 3 + j] = i * count + j;
              user_data[4 + j] += i * count + j;
            }
          }
          unsigned tag = 1;

          for (unsigned root = 0; root < comm_size; ++root) {

            user_data[0] = 0;
            user_data[3] = root;

            for (unsigned i = 0; i < comm_size; ++i)
              yac_comm_reduce_callback(data + i * count, count, COMM_UINT,
                                       COMM_OP_SUM, root, tag, comms[i],
                                       user_data, (func)?reduce_callback:NULL,
                                       MSG_PRIO_HIGH);

            for (unsigned i = 0; i < comm_size; ++i)
              yac_comm_wait(comms[i], tag);

            if (func)
              if (user_data[0] != comm_size)
                PUT_ERR("error in yac_comm_reduce_callback\n");
          }

          free(data);
        }

        { // test allreduce with callback (unsigned)

          unsigned count = 3;
          unsigned  * data = malloc(comm_size * count * sizeof(*data));
          unsigned user_data[6] = {0, count, COMM_UINT, 0, 0, 0};
          for (unsigned i = 0; i < comm_size; ++i) {
            for (unsigned j = 0; j < count; ++j) {
              data[i * 3 + j] = i * count + j;
              user_data[3 + j] += i * count + j;
            }
          }
          unsigned tag = 1;

          for (unsigned i = 0; i < comm_size; ++i)
            yac_comm_allreduce_callback(data + i * count, count, COMM_UINT,
                                        COMM_OP_SUM, tag, comms[i], user_data,
                                        (func)?allreduce_callback:NULL,
                                        MSG_PRIO_HIGH);

          for (unsigned i = 0; i < comm_size; ++i)
            yac_comm_wait(comms[i], tag);

          if (func)
            if (user_data[0] != comm_size)
              PUT_ERR("error in yac_comm_allreduce_callback\n");

          free(data);
        }

        { // test reduce with callback (int)

          unsigned count = 3;
          int  * data = malloc(comm_size * count * sizeof(*data));
          int user_data[7] = {0, count, COMM_INT, -1, 0, 0, 0};
          for (unsigned i = 0; i < comm_size; ++i) {
            for (unsigned j = 0; j < count; ++j) {
              data[i * 3 + j] = i * count + j;
              user_data[4 + j] += i * count + j;
            }
          }
          unsigned tag = 1;

          for (unsigned root = 0; root < comm_size; ++root) {

            user_data[0] = 0;
            user_data[3] = root;

            for (unsigned i = 0; i < comm_size; ++i)
              yac_comm_reduce_callback(data + i * count, count, COMM_INT,
                                       COMM_OP_SUM, root, tag, comms[i],
                                       user_data, (func)?reduce_callback:NULL,
                                       MSG_PRIO_HIGH);

            for (unsigned i = 0; i < comm_size; ++i)
              yac_comm_wait(comms[i], tag);

            if (func)
              if (user_data[0] != comm_size)
                PUT_ERR("error in yac_comm_reduce_callback\n");
          }

          free(data);
        }

        { // test allreduce with callback (int)

          unsigned count = 3;
          int  * data = malloc(comm_size * count * sizeof(*data));
          int user_data[6] = {0, count, COMM_INT, 0, 0, 0};
          for (unsigned i = 0; i < comm_size; ++i) {
            for (unsigned j = 0; j < count; ++j) {
              data[i * 3 + j] = i * count + j;
              user_data[3 + j] += i * count + j;
            }
          }
          unsigned tag = 1;

          for (unsigned i = 0; i < comm_size; ++i)
            yac_comm_allreduce_callback(data + i * count, count, COMM_INT,
                                        COMM_OP_SUM, tag, comms[i], user_data,
                                        (func)?allreduce_callback:NULL,
                                        MSG_PRIO_HIGH);

          for (unsigned i = 0; i < comm_size; ++i)
            yac_comm_wait(comms[i], tag);

          if (func)
            if (user_data[0] != comm_size)
              PUT_ERR("error in yac_comm_allreduce_callback\n");

          free(data);
        }

        { // test reduce with callback (double)

          unsigned count = 3;
          double  * data = malloc(comm_size * count * sizeof(*data));
          double user_data[7] = {0, count, COMM_DBLE, -1, 0, 0, 0};
          for (unsigned i = 0; i < comm_size; ++i) {
            for (unsigned j = 0; j < count; ++j) {
              data[i * 3 + j] = i * count + j;
              user_data[4 + j] += i * count + j;
            }
          }
          unsigned tag = 1;

          for (unsigned root = 0; root < comm_size; ++root) {

            user_data[0] = 0;
            user_data[3] = root;

            for (unsigned i = 0; i < comm_size; ++i)
              yac_comm_reduce_callback(data + i * count, count, COMM_DBLE,
                                       COMM_OP_SUM, root, tag, comms[i],
                                       user_data, (func)?reduce_callback:NULL,
                                       MSG_PRIO_HIGH);

            for (unsigned i = 0; i < comm_size; ++i)
              yac_comm_wait(comms[i], tag);

            if (func)
              if (user_data[0] != comm_size)
                PUT_ERR("error in yac_comm_reduce_callback\n");
          }

          free(data);
        }

        { // test allreduce with callback (double)

          unsigned count = 3;
          double  * data = malloc(comm_size * count * sizeof(*data));
          double user_data[6] = {0, count, COMM_DBLE, 0, 0, 0};
          for (unsigned i = 0; i < comm_size; ++i) {
            for (unsigned j = 0; j < count; ++j) {
              data[i * 3 + j] = i * count + j;
              user_data[3 + j] += i * count + j;
            }
          }
          unsigned tag = 1;

          for (unsigned i = 0; i < comm_size; ++i)
            yac_comm_allreduce_callback(data + i * count, count, COMM_DBLE,
                                        COMM_OP_SUM, tag, comms[i], user_data,
                                        (func)?allreduce_callback:NULL,
                                        MSG_PRIO_HIGH);

          for (unsigned i = 0; i < comm_size; ++i)
            yac_comm_wait(comms[i], tag);

          if (func)
            if (user_data[0] != comm_size)
              PUT_ERR("error in yac_comm_allreduce_callback\n");

          free(data);
        }

        { // test reduce with callback (char)

          unsigned count = 3;
          char  * data = malloc(comm_size * count * sizeof(*data));
          char user_data[7] = {0, count, COMM_BYTE, -1, 0, 0, 0};
          for (unsigned i = 0; i < comm_size; ++i) {
            for (unsigned j = 0; j < count; ++j) {
              data[i * 3 + j] = i * count + j;
              user_data[4 + j] += i * count + j;
            }
          }
          unsigned tag = 1;

          for (unsigned root = 0; root < comm_size; ++root) {

            user_data[0] = 0;
            user_data[3] = root;

            for (unsigned i = 0; i < comm_size; ++i)
              yac_comm_reduce_callback(data + i * count, count, COMM_BYTE,
                                       COMM_OP_SUM, root, tag, comms[i],
                                       user_data, (func)?reduce_callback:NULL,
                                       MSG_PRIO_HIGH);

            for (unsigned i = 0; i < comm_size; ++i)
              yac_comm_wait(comms[i], tag);

            if (func)
              if (user_data[0] != comm_size)
                PUT_ERR("error in yac_comm_reduce_callback\n");
          }

          free(data);
        }

        { // test allreduce with callback (char)

          unsigned count = 3;
          char  * data = malloc(comm_size * count * sizeof(*data));
          char user_data[6] = {0, count, COMM_BYTE, 0, 0, 0};
          for (unsigned i = 0; i < comm_size; ++i) {
            for (unsigned j = 0; j < count; ++j) {
              data[i * 3 + j] = i * count + j;
              user_data[3 + j] += i * count + j;
            }
          }
          unsigned tag = 1;

          for (unsigned i = 0; i < comm_size; ++i)
            yac_comm_allreduce_callback(data + i * count, count, COMM_BYTE,
                                        COMM_OP_SUM, tag, comms[i], user_data,
                                        (func)?allreduce_callback:NULL,
                                        MSG_PRIO_HIGH);

          for (unsigned i = 0; i < comm_size; ++i)
            yac_comm_wait(comms[i], tag);

          if (func)
            if (user_data[0] != comm_size)
              PUT_ERR("error in yac_comm_allreduce_callback\n");

          free(data);
        }

        { // test barrier with callback

          unsigned tag = 1;
          unsigned user_data = 0;

          for (unsigned i = 0; i < comm_size; ++i)
            yac_comm_barrier_callback(tag, comms[i], &user_data,
                                      (func)?barrier_callback:NULL,
                                      MSG_PRIO_HIGH);

          for (unsigned i = 0; i < comm_size; ++i)
            yac_comm_wait(comms[i], tag);

          if (func)
            if (user_data != comm_size)
              PUT_ERR("error in yac_comm_barrier_callback\n");
        }
      }

      { // test irecv and waitall

        unsigned * recv_buffer = malloc(comm_size * sizeof(*recv_buffer));
        struct comm_request ** requests = malloc(comm_size * sizeof(*requests));

        unsigned tag = 1;

        for (unsigned i = 0; i < 2; ++i) {
          for (unsigned rank = 0; rank < comm_size; ++rank) {

            if (i == 0) yac_comm_bsend(&rank, 1, COMM_UINT, (rank+1)%comm_size,
                                       tag, comms[rank]);

            yac_comm_irecv(recv_buffer + rank, 1, COMM_UINT,
                           (rank-1+comm_size)%comm_size, tag, comms[rank],
                           requests + rank);

            if (i == 1) yac_comm_bsend(&rank, 1, COMM_UINT, (rank+1)%comm_size,
                                       tag, comms[rank]);
          }

          for (unsigned rank = 0; rank < comm_size; ++rank)
            yac_comm_waitall(1, requests + rank);

          for (unsigned rank = 0; rank < comm_size; ++rank)
            if (recv_buffer[rank] != (rank-1+comm_size)%comm_size)
            PUT_ERR("error in yac_comm_irecv/yac_comm_waitall\n");
        }

        free(requests);
        free(recv_buffer);
      }

      { // test irecv, isend and waitall

        unsigned * recv_buffer = malloc(comm_size * sizeof(*recv_buffer));
        struct comm_request ** requests = malloc(2 * comm_size * sizeof(*requests));

        unsigned tag = 1;

        for (unsigned i = 0; i < 2; ++i) {
          for (unsigned rank = 0; rank < comm_size; ++rank) {

            if (i == 0) yac_comm_isend(&rank, 1, COMM_UINT, (rank+1)%comm_size,
                                       tag, comms[rank], requests + 2*rank+1);

            yac_comm_irecv(recv_buffer + rank, 1, COMM_UINT,
                           (rank-1+comm_size)%comm_size, tag, comms[rank],
                           requests + 2*rank+0);

            if (i == 1) yac_comm_isend(&rank, 1, COMM_UINT, (rank+1)%comm_size,
                                       tag, comms[rank], requests + 2*rank+1);
          }

          for (unsigned i = 0; i < 2 * comm_size; ++i)
            yac_comm_waitall(1, requests + i);

          for (unsigned rank = 0; rank < comm_size; ++rank)
            if (recv_buffer[rank] != (rank-1+comm_size)%comm_size)
            PUT_ERR("error in yac_comm_irecv/yac_comm_waitall\n");
        }

        free(requests);
        free(recv_buffer);
      }

      for (unsigned i = 0; i < comm_size; ++i)
        yac_free_communicator(comms[i]);
      free(comms);
    }
  }

   yac_comm_finalize();

   return TEST_EXIT_CODE;
}

void recv_callback(struct communicator * comm, struct comm_message message,
                   void * orig_data) {

   if (message.count != 4)
      PUT_ERR("error in message.count\n")

   if (message.dst_rank != yac_get_comm_rank(comm))
      PUT_ERR("error in message.dst_rank\n")

   unsigned * send_data = orig_data;
   unsigned i;

   for (i = 0; i < 4; ++i)
      if (((unsigned *)(message.buffer))[i] != send_data[i])
         PUT_ERR("error in message.buffer\n")

   num_received_messages++;
}

void recv_callback_cancel(struct communicator * comm, void * orig_data) {

   num_canceled_messages++;
}

void bcast_callback(struct communicator * comm, struct comm_message message,
                    void * orig_data) {

  unsigned root = ((unsigned *)orig_data)[0];
  unsigned ref_count = ((unsigned *)orig_data)[1];
  unsigned * data = (unsigned*)(message.buffer);
  unsigned rank = yac_get_comm_rank(comm);

  if (message.type != COMM_UINT)
    PUT_ERR("error in message.type\n");

  if (message.dst_rank != rank)
    PUT_ERR("error in message.dst_rank\n");

  if (message.src_rank != root)
    PUT_ERR("error in message.src_rank\n");

  for (unsigned i = 0; i < ref_count; ++i)
    if (data[i] != i) PUT_ERR("error in message.buffer\n");

  ((unsigned *)orig_data + 2)[rank] = rank;
}

void gatherv_callback(struct communicator * comm, struct comm_message message,
                     void * orig_data) {

  unsigned root = ((unsigned *)orig_data)[0];
  unsigned comm_size = yac_get_comm_size(comm);
  unsigned comm_rank = yac_get_comm_rank(comm);
  unsigned count = (comm_rank == root)?(comm_size * (comm_size + 1) / 2):(0);
  unsigned * buffer = (unsigned*)(message.buffer);

  if (count != message.count)
    PUT_ERR("error in message.count\n");

  if (message.type != COMM_UINT)
    PUT_ERR("error in message.type\n");

  for (unsigned i = 0; i < count; ++i)
    if (buffer[i] != i)
      PUT_ERR("error in message.buffer\n");

  if ((count == 0) && (buffer != NULL))
    PUT_ERR("error in message.buffer\n");

  ++((unsigned *)orig_data)[1];
}

void allgather_callback(struct communicator * comm, struct comm_message message,
                        void * orig_data) {

  unsigned comm_size = yac_get_comm_size(comm);
  unsigned comm_rank = yac_get_comm_rank(comm);
  unsigned count = *(unsigned*)orig_data;
  unsigned * buffer = (unsigned*)(message.buffer);

  if (count * comm_size != message.count)
    PUT_ERR("error in message.count\n");

  if (message.type != COMM_UINT)
    PUT_ERR("error in message.type\n");

  for (unsigned i = 0; i < count * comm_size; ++i)
    if (buffer[i] != i)
      PUT_ERR("error in message.buffer\n");

  if (message.src_rank != comm_rank)
    PUT_ERR("error in message.src_rank\n");

  ((unsigned*)orig_data)[1]++;
}

void allgatherv_callback(struct communicator * comm,
                         struct comm_message message, void * orig_data) {

  unsigned comm_size = yac_get_comm_size(comm);
  unsigned comm_rank = yac_get_comm_rank(comm);
  unsigned count = comm_size * (comm_size + 1) / 2;
  unsigned * buffer = (unsigned*)(message.buffer);

  if (count != message.count)
    PUT_ERR("error in message.count\n");

  if (message.type != COMM_UINT)
    PUT_ERR("error in message.type\n");

  for (unsigned i = 0; i < count; ++i)
    if (buffer[i] != i)
      PUT_ERR("error in message.buffer\n");

  if (message.src_rank != comm_rank)
    PUT_ERR("error in message.src_rank\n");

  ++*((unsigned*)orig_data);
}

void scatter_callback(struct communicator * comm,
                      struct comm_message message, void * user_data) {

  unsigned root = ((unsigned *)user_data)[0];
  unsigned comm_rank = yac_get_comm_rank(comm);;
  unsigned count = ((unsigned *)user_data)[1];
  unsigned * buffer = (unsigned*)(message.buffer);

  if (count != message.count)
    PUT_ERR("error in message.count\n");

  if (message.type != COMM_UINT)
    PUT_ERR("error in message.type\n");

  for (unsigned i = 0; i < count; ++i)
    if (buffer[i] != comm_rank * count + i)
      PUT_ERR("error in message.buffer\n");

  if (message.src_rank != root)
    PUT_ERR("error in message.src_rank\n");

  ((unsigned*)user_data)[2]++;
}

void scatterv_callback(struct communicator * comm,
                       struct comm_message message, void * orig_data) {

  unsigned root = ((unsigned *)orig_data)[0];
  unsigned comm_rank = yac_get_comm_rank(comm);
  unsigned n1 = comm_rank * (comm_rank + 1) / 2;
  unsigned count = comm_rank + 1;
  unsigned * buffer = (unsigned*)(message.buffer);

  if (count != message.count)
    PUT_ERR("error in message.count\n");

  if (message.type != COMM_UINT)
    PUT_ERR("error in message.type\n");

  for (unsigned i = 0; i < count; ++i)
    if (buffer[i] != i + n1)
      PUT_ERR("error in message.buffer\n");

  if (message.src_rank != root)
    PUT_ERR("error in message.src_rank\n");

  ((unsigned*)orig_data)[1]++;
}

void alltoallv_callback(struct communicator * comm,
                        struct comm_message message, void * orig_data) {

  unsigned comm_size = yac_get_comm_size(comm);
  unsigned comm_rank = yac_get_comm_rank(comm);
  unsigned count = (comm_size * (comm_size + 1)) / 2;
  unsigned * buffer = (unsigned*)(message.buffer);

  if (count != message.count)
    PUT_ERR("error in message.count\n");

  if (message.type != COMM_UINT)
    PUT_ERR("error in message.type\n");

  for (unsigned i = 0; i < count; ++i)
    if (buffer[i] != i)
      PUT_ERR("error in message.buffer\n");

  if (message.src_rank != comm_rank)
    PUT_ERR("error in message.src_rank\n");

  ((unsigned*)orig_data)[0]++;
}

void reduce_callback(struct communicator * comm,
                     struct comm_message message, void * orig_data) {

  unsigned comm_rank = yac_get_comm_rank(comm);

  unsigned count, root;
  enum yac_comm_data_type type;

  switch (message.type) {
    case (COMM_UINT):
      ((unsigned*)orig_data)[0]++;
      count = ((unsigned*)(orig_data))[1];
      type = ((unsigned*)(orig_data))[2];
      root = ((unsigned*)(orig_data))[3];
      break;
    case (COMM_INT):
      ((int*)orig_data)[0]++;
      count = ((int*)(orig_data))[1];
      type = ((int*)(orig_data))[2];
      root = ((int*)(orig_data))[3];
      break;
    case (COMM_DBLE):
      ((double*)orig_data)[0] += 1;
      count = ((double*)(orig_data))[1];
      type = ((double*)(orig_data))[2];
      root = ((double*)(orig_data))[3];
      break;
    case (COMM_BYTE):
      ((char*)orig_data)[0]++;
      count = ((char*)(orig_data))[1];
      type = ((char*)(orig_data))[2];
      root = ((char*)(orig_data))[3];
      break;
  };

  if (message.src_rank != comm_rank)
    PUT_ERR("error in message.src_rank\n");

  if (comm_rank == root) {
    if (count != message.count)
      PUT_ERR("error in message.count\n");

    if (message.type != type)
      PUT_ERR("error in message.type\n");

    unsigned * ref_buffer_uint = (unsigned*)(orig_data) + 4;
    int * ref_buffer_int = (int*)(orig_data) + 4;
    double * ref_buffer_dble = (double*)(orig_data) + 4;
    char * ref_buffer_byte = (char*)(orig_data) + 4;
    unsigned * buffer_uint = (unsigned*)(message.buffer);
    int * buffer_int = (int*)(message.buffer);
    double * buffer_dble = (double*)(message.buffer);
    char * buffer_byte = (char*)(message.buffer);

    for (unsigned i = 0; i < count; ++i)
      switch (type) {
        case (COMM_UINT):
          if (buffer_uint[i] != ref_buffer_uint[i])
            PUT_ERR("error in message.buffer\n");
          break;
        case (COMM_INT):
          if (buffer_int[i] != ref_buffer_int[i])
            PUT_ERR("error in message.buffer\n");
          break;
        case (COMM_DBLE):
          if (buffer_dble[i] != ref_buffer_dble[i])
            PUT_ERR("error in message.buffer\n");
          break;
        case (COMM_BYTE):
          if (buffer_byte[i] != ref_buffer_byte[i])
            PUT_ERR("error in message.buffer\n");
          break;
      };
  }
}

void allreduce_callback(struct communicator * comm,
                        struct comm_message message, void * orig_data) {

  unsigned comm_rank = yac_get_comm_rank(comm);
  unsigned count;
  enum yac_comm_data_type type;
  unsigned * ref_buffer_uint = (unsigned*)(orig_data) + 3;
  unsigned * buffer_uint = (unsigned*)(message.buffer);
  int * ref_buffer_int = (int*)(orig_data) + 3;
  int * buffer_int = (int*)(message.buffer);
  double * ref_buffer_dble = (double*)(orig_data) + 3;
  double * buffer_dble = (double*)(message.buffer);
  char * ref_buffer_byte = (char*)(orig_data) + 3;
  char * buffer_byte = (char*)(message.buffer);

  switch (message.type) {
    case (COMM_UINT):
      ((unsigned*)orig_data)[0]++;
      count = ((unsigned*)(orig_data))[1];
      type = ((unsigned*)(orig_data))[2];
      break;
    case (COMM_INT):
      ((int*)orig_data)[0]++;
      count = ((int*)(orig_data))[1];
      type = ((int*)(orig_data))[2];
      break;
    case (COMM_DBLE):
      ((double*)orig_data)[0] += 1;
      count = ((double*)(orig_data))[1];
      type = ((double*)(orig_data))[2];
      break;
    case (COMM_BYTE):
      ((char*)orig_data)[0]++;
      count = ((char*)(orig_data))[1];
      type = ((char*)(orig_data))[2];
      break;
  };

  if (count != message.count)
    PUT_ERR("error in message.count\n");

  if (message.type != type)
    PUT_ERR("error in message.type\n");

  for (unsigned i = 0; i < count; ++i)
    switch (type) {
      case (COMM_UINT):
        if (buffer_uint[i] != ref_buffer_uint[i])
          PUT_ERR("error in message.buffer\n");
        break;
      case (COMM_INT):
        if (buffer_int[i] != ref_buffer_int[i])
          PUT_ERR("error in message.buffer\n");
        break;
      case (COMM_DBLE):
        if (buffer_dble[i] != ref_buffer_dble[i])
          PUT_ERR("error in message.buffer\n");
        break;
      case (COMM_BYTE):
        if (buffer_byte[i] != ref_buffer_byte[i])
          PUT_ERR("error in message.buffer\n");
        break;
    };

  if (message.src_rank != comm_rank)
    PUT_ERR("error in message.src_rank\n");
}


void barrier_callback(struct communicator * comm, struct comm_message message,
                      void * orig_data) {

  if (0 != message.count)
    PUT_ERR("error in message.count\n");

  if (message.src_rank != yac_get_comm_rank(comm))
    PUT_ERR("error in message.src_rank\n");

  ++*((unsigned*)orig_data);
}