#define VERBOSE

/**
 * @file test_communicator_mpi.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "tests.h"

#ifdef USE_MPI

#include "communicator.h"
#include "communicator_mpi.h"
#include "yacpar.h"

unsigned num_errors = 0;
unsigned num_received_messages = 0;
unsigned num_canceled_messages = 0;

// callback function passed to comm_recv_callback
void recv_callback(struct communicator * comm, struct comm_message message, void * orig_data);
void recv_callback_cancel(struct communicator * comm, void * orig_data);
void recv_callback_cancel_2(struct communicator * comm, void * orig_data);

// callback function to test message priorities
void prio_recv_callback(struct communicator * comm, struct comm_message message,
                        void * user_data);

// callback function passed to yac_comm_bcast_callback
void bcast_callback(struct communicator * comm, struct comm_message message,
                    void * orig_data);

// callback function passed to yac_comm_gatherv_callback
void gatherv_callback(struct communicator * comm, struct comm_message message,
                      void * orig_data);

// callback function passed to yac_comm_allgather_callback
void allgather_callback(struct communicator * comm, struct comm_message message,
                        void * orig_data);

// callback function passed to yac_comm_allgatherv_callback
void allgatherv_callback(struct communicator * comm, struct comm_message message,
                         void * orig_data);

// callback function passed to yac_comm_scatter_callback
void scatter_callback(struct communicator * comm, struct comm_message message,
                      void * orig_data);

// callback function passed to yac_comm_scatterv_callback
void scatterv_callback(struct communicator * comm, struct comm_message message,
                       void * orig_data);

// callback function passed to yac_comm_alltoallv_callback
void alltoallv_callback(struct communicator * comm, struct comm_message message,
                        void * orig_data);

// callback function passed to yac_comm_alltoallv_callback (in random test)
void alltoallv_rand_callback(struct communicator * comm,
                             struct comm_message message, void * orig_data);

// callback function passed to yac_comm_alltoallv_callback (using character data)
void alltoallv_char_callback(struct communicator * comm,
                             struct comm_message message, void * orig_data);

// callback function passed to yac_comm_reduce_callback
void reduce_callback(struct communicator * comm,
                     struct comm_message message, void * orig_data);

// callback function passed to yac_comm_allreduce_callback
void allreduce_callback(struct communicator * comm,
                        struct comm_message message, void * orig_data);

// callback function passed to yac_comm_barrier_callback
void barrier_callback(struct communicator * comm,
                      struct comm_message message, void * orig_data);

int main (void) {

    yac_world_comm = yac_mpi_comm_new();

   { // tests with a single process in comm (if the program was started with more, then
     //                                      the global communicator split until each
     //                                      communicator contains only a single process)

      //---------------
      // setup
      //---------------
      struct communicator * comm;

      comm = yac_mpi_comm_new();

      // split if there is more than one process
      if (yac_get_comm_size(comm) > 1) {

         struct communicator * temp_comm;

         temp_comm = yac_comm_split(comm, yac_get_comm_rank(comm), 0);

         yac_free_communicator(comm);

         comm = temp_comm;
      }

      //---------------
      // testing
      //---------------

      { // test communicator size
         if (yac_get_comm_size(comm) != 1)
            PUT_ERR("error in get_comm_size\n")
      }

      { // test generation of unique ids
         char const * tag_strings[] = {"corner", "cell", "edge"};
         unsigned ids[3];

         yac_get_unique_comm_tags(
          comm, "generator string", tag_strings, 3, ids);

         if (ids[0] == ids[1] || ids[0] == ids[2] || ids[1] == ids[2])
            PUT_ERR("error in get_unique_comm_tags\n")

         unsigned tag;

         yac_get_unique_comm_tags(
          comm, "generator string", tag_strings, 1, &tag);

         if (tag != ids[0])
            PUT_ERR("error in get_unique_comm_tags\n")
      }

      { // test sending of data (first send then recv)

         unsigned send_data[4] = {0,1,2,3};

         num_received_messages = 0;

         yac_comm_bsend(send_data, 4, COMM_UINT, 0, 0, comm);

         yac_comm_recv_callback(4, COMM_UINT, 0, 0, comm, send_data,
                                recv_callback, NULL, MSG_PRIO_HIGH);

         while (yac_comm_wait(comm, 0));

         if (num_received_messages != 1)
            PUT_ERR("error in comm_recv_callback\n")
      }

      { // test sending of data (first recv then send)

         unsigned send_data[4] = {0,1,2,3};

         num_received_messages = 0;

         yac_comm_recv_callback(4, COMM_UINT, 0, 0, comm, send_data,
                                recv_callback, NULL, MSG_PRIO_HIGH);

         yac_comm_bsend(send_data, 4, COMM_UINT, 0, 0, comm);

         while (yac_comm_wait(comm, 0));

         if (num_received_messages != 1)
            PUT_ERR("error in comm_recv_callback\n")
      }

      { // test sending of two data packages (first send then recv)

         unsigned send_data[2][4] = {{0,1,2,3}, {3,2,1,0}};

         num_received_messages = 0;

         yac_comm_bsend(send_data[0], 4, COMM_UINT, 0, 0, comm);
         yac_comm_bsend(send_data[1], 4, COMM_UINT, 0, 0, comm);

         yac_comm_recv_callback(4, COMM_UINT, 0, 0, comm, send_data[0],
                                recv_callback, NULL, MSG_PRIO_HIGH);
         yac_comm_recv_callback(4, COMM_UINT, 0, 0, comm, send_data[1],
                                recv_callback, NULL, MSG_PRIO_HIGH);

         while (yac_comm_wait(comm, 0));

         if (num_received_messages != 2)
            PUT_ERR("error in comm_recv_callback\n")
      }

      { // test sending of two data packages (first recv then send)

         unsigned send_data[2][4] = {{0,1,2,3}, {3,2,1,0}};

         num_received_messages = 0;

         yac_comm_recv_callback(4, COMM_UINT, 0, 0, comm, send_data[0],
                                recv_callback, NULL, MSG_PRIO_HIGH);
         yac_comm_recv_callback(4, COMM_UINT, 0, 0, comm, send_data[1],
                                recv_callback, NULL, MSG_PRIO_HIGH);

         yac_comm_bsend(send_data[0], 4, COMM_UINT, 0, 0, comm);
         yac_comm_bsend(send_data[1], 4, COMM_UINT, 0, 0, comm);

         while (yac_comm_wait(comm, 0));

         if (num_received_messages != 2)
            PUT_ERR("error in comm_recv_callback\n")
      }

      { // test sending of two data packages with different tags

         unsigned send_data[2][4] = {{0,1,2,3}, {3,2,1,0}};

         num_received_messages = 0;

         yac_comm_bsend(send_data[0], 4, COMM_UINT, 0, 0, comm);
         yac_comm_bsend(send_data[1], 4, COMM_UINT, 0, 1, comm);

         yac_comm_recv_callback(4, COMM_UINT, 0, 0, comm, send_data[0],
                                recv_callback, NULL, MSG_PRIO_HIGH);
         yac_comm_recv_callback(4, COMM_UINT, 0, 1, comm, send_data[1],
                                recv_callback, NULL, MSG_PRIO_HIGH);

         while (yac_comm_wait(comm, 0) ||
                yac_comm_wait(comm, 1));

         if (num_received_messages != 2)
            PUT_ERR("error in comm_recv_callback\n")
      }

      { // test sending of two data packages with different tags

         unsigned send_data[2][4] = {{0,1,2,3}, {3,2,1,0}};

         num_received_messages = 0;

         yac_comm_bsend(send_data[0], 4, COMM_UINT, 0, 0, comm);
         yac_comm_bsend(send_data[1], 4, COMM_UINT, 0, 1, comm);

         yac_comm_recv_callback(4, COMM_UINT, 0, 1, comm, send_data[1],
                                recv_callback, NULL, MSG_PRIO_HIGH);
         yac_comm_recv_callback(4, COMM_UINT, 0, 0, comm, send_data[0],
                                recv_callback, NULL, MSG_PRIO_HIGH);

         while (yac_comm_wait(comm, 0) ||
                yac_comm_wait(comm, 1));

         if (num_received_messages != 2)
            PUT_ERR("error in comm_recv_callback\n")
      }

      { // test sending of two data packages with different tags using blocking recv

         unsigned send_data[2][4] = {{0,1,2,3}, {3,2,1,0}};
         unsigned recv_data[2][4];

         num_received_messages = 0;

         yac_comm_bsend(send_data[0], 4, COMM_UINT, 0, 0, comm);
         yac_comm_bsend(send_data[1], 4, COMM_UINT, 0, 1, comm);

         unsigned recv_count[2] = {4,4};

         yac_comm_recv(recv_data[0], recv_count+0, COMM_UINT, 0, 0, comm);
         yac_comm_recv(recv_data[1], recv_count+1, COMM_UINT, 0, 1, comm);

         if (recv_count[0] != 4 || recv_count[1] != 4)
            PUT_ERR("error in comm_recv count\n")

         unsigned i;

         for (i = 0; i < 4; ++i)
            if (recv_data[0][i] != send_data[0][i] ||
                recv_data[1][i] != send_data[1][i])
               PUT_ERR("error in comm_recv recv_data\n")
      }

      { // test sending of two data packages with different tags using blocking recv

         unsigned send_data[2][4] = {{0,1,2,3}, {3,2,1,0}};
         unsigned recv_data[2][4];

         num_received_messages = 0;

         yac_comm_bsend(send_data[0], 4, COMM_UINT, 0, 0, comm);
         yac_comm_bsend(send_data[1], 4, COMM_UINT, 0, 1, comm);

         unsigned recv_count[2] = {4,4};

         yac_comm_recv(recv_data[1], recv_count+1, COMM_UINT, 0, 1, comm);
         yac_comm_recv(recv_data[0], recv_count+0, COMM_UINT, 0, 0, comm);

         if (recv_count[0] != 4 || recv_count[1] != 4)
            PUT_ERR("error in comm_recv count\n")

         unsigned i;

         for (i = 0; i < 4; ++i)
            if (recv_data[0][i] != send_data[0][i] ||
                recv_data[1][i] != send_data[1][i])
               PUT_ERR("error in comm_recv recv_data\n")
      }

      { // test generation and freeing of NULL communicator

        struct communicator * null_comm;

        null_comm = yac_comm_split(comm, -1, -1);
        yac_free_communicator(null_comm);
      }

      { // a packing/unpacking example

        unsigned * uint_data = calloc(2, sizeof(*uint_data));
        double * dble_data = calloc(24, sizeof(*dble_data));

        unsigned pack_buffer_size = yac_comm_pack_size(2, COMM_UINT, comm) +
                                    yac_comm_pack_size(24, COMM_DBLE, comm);
        void * pack_buffer = malloc(pack_buffer_size);

        int position = 0;
        yac_comm_pack(uint_data, 2, COMM_UINT, pack_buffer, pack_buffer_size,
                      &position, comm);
        yac_comm_pack(dble_data, 24, COMM_DBLE, pack_buffer, pack_buffer_size,
                      &position, comm);

        free(pack_buffer);
        free(dble_data);
        free(uint_data);
      }

      { // test broadcast with callback

        unsigned data[3] = {0,1,2};
        unsigned tag = 1;
        unsigned root = 0;
        // contents {root, data size, 1 if the message was received}
        unsigned user_data[3] = {root, sizeof(data)/sizeof(data[0]), -1};

        yac_comm_bcast_callback((void*)data, sizeof(data)/sizeof(data[0]),
                                COMM_UINT, root, tag, comm, user_data,
                                bcast_callback, MSG_PRIO_HIGH);

        while (yac_comm_wait(comm, tag));

        if (user_data[2] != 1) PUT_ERR("error in yac_comm_bcast_callback\n");
      }

      { // test gatherv with callback

        unsigned rank = yac_get_comm_rank(comm);
        unsigned size = yac_get_comm_size(comm);
        unsigned * data = malloc((rank + 1) * sizeof(data));
        for (unsigned i = 0; i <= rank; ++i)
          data[i] = rank * (rank + 1) / 2 + i;
        unsigned tag = 1;
        unsigned * counts = malloc(size * sizeof(*counts));
        for (unsigned i = 0; i < size; ++i) counts[i] = i + 1;
        unsigned root = 0;
        unsigned user_data[2] = {root, -1};

        yac_comm_gatherv_callback(data, counts, COMM_UINT, root, tag, comm,
                                  user_data, gatherv_callback,
                                  MSG_PRIO_HIGH);
        free(counts);
        free(data);
        while (yac_comm_wait(comm, tag));

        if (user_data[1] != 1) PUT_ERR("error in yac_comm_gatherv_callback\n");
      }

      { // test allgather with callback

        unsigned rank = yac_get_comm_rank(comm);
        unsigned data[2] = {rank * 2 + 0, rank * 2 + 1};
        unsigned tag = 1;
        unsigned count = 2;
        unsigned user_data[2] = {count, -1};

        yac_comm_allgather_callback(data, count, COMM_UINT, tag, comm,
                                    user_data, allgather_callback,
                                    MSG_PRIO_HIGH);
        while (yac_comm_wait(comm, tag));

        if (user_data[1] != 1) PUT_ERR("error in yac_comm_allgather_callback\n");
      }

      { // test allgatherv with callback

        unsigned rank = yac_get_comm_rank(comm);
        unsigned size = yac_get_comm_size(comm);
        unsigned * data = malloc((rank + 1) * sizeof(data));
        for (unsigned i = 0; i <= rank; ++i)
          data[i] = rank * (rank + 1) / 2 + i;
        unsigned tag = 1;
        unsigned * counts = malloc(size * sizeof(*counts));
        for (unsigned i = 0; i < size; ++i) counts[i] = i + 1;
        unsigned user_data[1] = {-1};

        yac_comm_allgatherv_callback(data, counts, COMM_UINT, tag, comm,
                                     user_data, allgatherv_callback,
                                     MSG_PRIO_HIGH);
        free(counts);
        free(data);
        while (yac_comm_wait(comm, tag));

        if (user_data[0] != 1) PUT_ERR("error in yac_comm_allgatherv_callback\n");
      }

      { // test scatter with callback

        unsigned size = yac_get_comm_size(comm);
        unsigned count = 4;
        unsigned * data = malloc(size * count * sizeof(data));
        for (unsigned i = 0; i < size * count; ++i) data[i] = i;
        unsigned tag = 1;
        unsigned root = 0;
        unsigned user_data[3] = {root, count, -1};

        yac_comm_scatter_callback(data, count, COMM_UINT, root, tag, comm,
                                  user_data, scatter_callback, MSG_PRIO_HIGH);

        while (yac_comm_wait(comm, tag));

        if (user_data[2] != 1)
          PUT_ERR("error in yac_comm_scatter_callback\n");

        free(data);
      }

      { // test scatterv with callback

        unsigned size = yac_get_comm_size(comm);
        unsigned * data = malloc(size * (size + 1) / 2 * sizeof(data));
        for (unsigned i = 0; i < size * (size + 1) / 2; ++i) data[i] = i;
        unsigned tag = 1;
        unsigned * counts = malloc(size * sizeof(*counts));
        for (unsigned i = 0; i < size; ++i) counts[i] = i + 1;
        unsigned root = 0;
        unsigned user_data[2] = {root, -1};

        yac_comm_scatterv_callback(data, counts, COMM_UINT, root, tag, comm,
                                   user_data, scatterv_callback, MSG_PRIO_HIGH);

        while (yac_comm_wait(comm, tag));

        if (user_data[1] != 1)
          PUT_ERR("error in yac_comm_scatterv_callback\n");

        free(counts);
        free(data);
      }

      { // test alltoallv with callback

        unsigned size = yac_get_comm_size(comm);
        unsigned rank = yac_get_comm_rank(comm);
        unsigned * data = malloc(size * (rank + 1) * sizeof(*data));
        for (unsigned i = 0, offset = 0; i < size; ++i)
          for (unsigned j = 0; j <= rank; ++j, ++offset)
            data[offset] = (rank * (rank + 1)) / 2 + j;
        unsigned tag = 1;
        unsigned * counts = malloc(size * sizeof(*counts));
        for (unsigned i = 0; i < size; ++i)
            counts[i] = rank + 1;
        unsigned user_data = 0;

        yac_comm_alltoallv_callback(data, counts, COMM_UINT, tag, comm,
                                    &user_data, alltoallv_callback,
                                    MSG_PRIO_HIGH);

        while (yac_comm_wait(comm, tag));

        if (user_data != 1)
          PUT_ERR("error in yac_comm_alltoallv_callback\n");

        free(counts);
        free(data);
      }

      { // test reduce with callback

        unsigned size = yac_get_comm_size(comm);
        unsigned rank = yac_get_comm_rank(comm);
        unsigned count = 3;
        unsigned * user_data = calloc(3 + count, sizeof(*user_data));
        user_data[1] = count;
        for (unsigned i = 0; i < size; ++i) {
          for (unsigned j = 0; j < count; ++j) {
            user_data[3 + j] += i * count + j;
          }
        }
        unsigned * data = malloc(count * sizeof(*data));
        for (unsigned j = 0; j < count; ++j) data[j] = rank * count + j;
        unsigned tag = 1;

        for (unsigned root = 0; root < size; ++root) {

          user_data[0] = 0;
          user_data[2] = root;

          yac_comm_reduce_callback(data, count, COMM_UINT, COMM_OP_SUM, root,
                                   tag, comm, user_data, reduce_callback,
                                   MSG_PRIO_HIGH);

          while (yac_comm_wait(comm, tag));

          if (user_data[0] != 1)
            PUT_ERR("error in yac_comm_reduce_callback\n");
        }

        free(user_data);
        free(data);
      }

      { // test allreduce with callback

        unsigned size = yac_get_comm_size(comm);
        unsigned rank = yac_get_comm_rank(comm);
        unsigned count = 3;
        unsigned * user_data = calloc(2 + count, sizeof(*user_data));
        user_data[1] = count;
        for (unsigned i = 0; i < size; ++i) {
          for (unsigned j = 0; j < count; ++j) {
            user_data[2 + j] += i * count + j;
          }
        }
        unsigned * data = malloc(count * sizeof(*data));
        for (unsigned j = 0; j < count; ++j) data[j] = rank * count + j;
        unsigned tag = 1;

        yac_comm_allreduce_callback(data, count, COMM_UINT, COMM_OP_SUM, tag,
                                    comm, user_data, allreduce_callback,
                                    MSG_PRIO_HIGH);

        while (yac_comm_wait(comm, tag));

        if (user_data[0] != 1)
          PUT_ERR("error in yac_comm_allreduce_callback\n");

        free(user_data);
        free(data);
      }

      { // test barrier with callback
        unsigned user_data = 0;
        unsigned tag = 1;

        yac_comm_barrier_callback(tag, comm, &user_data, barrier_callback,
                                  MSG_PRIO_HIGH);

        while (yac_comm_wait(comm, tag));

        if (user_data != 1)
          PUT_ERR("error in yac_comm_barrier_callback\n");
      }

      //---------------
      // cleanup
      //---------------
      yac_free_communicator(comm);
   }

   { // tests with two ranks in comm (if the program was started with more, then
     //                               the global communicator split until each
     //                               communicator contains only a two process)

      //---------------
      // setup
      //---------------
      struct communicator * comm;

      comm = yac_mpi_comm_new();

      // split if there are more than two process
      if (yac_get_comm_size(comm) > 2) {

         struct communicator * temp_comm;

         temp_comm = yac_comm_split(comm, yac_get_comm_rank(comm)>>1, 0);

         yac_free_communicator(comm);

         comm = temp_comm;
      }

      //---------------
      // testing
      //---------------

      // test size of communicator
      if (yac_get_comm_size(comm) > 2 || yac_get_comm_size(comm) < 1)
         PUT_ERR("error in get_comm_size\n")

      if (yac_get_comm_size(comm) == 2) {

         unsigned rank;

         rank = yac_get_comm_rank(comm);

         { // test sending of two data packages with different tags

            unsigned send_data[2][4] = {{0,1,2,3}, {3,2,1,0}};

            num_received_messages = 0;

            yac_comm_bsend(send_data[0], 4, COMM_UINT, rank^1, 0, comm);
            yac_comm_bsend(send_data[1], 4, COMM_UINT, rank^1, 1, comm);

            yac_comm_recv_callback(4, COMM_UINT, rank^1, 1, comm, send_data[1],
                                   recv_callback, NULL, MSG_PRIO_HIGH);
            yac_comm_recv_callback(4, COMM_UINT, rank^1, 0, comm, send_data[0],
                                   recv_callback, NULL, MSG_PRIO_HIGH);

            while (yac_comm_wait(comm, 0) ||
                   yac_comm_wait(comm, 1));

            if (num_received_messages != 2)
               PUT_ERR("error in comm_recv_callback\n")

            num_received_messages = 0;
         }

         { // test sending of two data packages with different tags

            unsigned send_data[2][4] = {{0,1,2,3}, {3,2,1,0}};

            num_received_messages = 0;

            yac_comm_bsend(send_data[0], 4, COMM_UINT, rank, 0, comm);
            yac_comm_bsend(send_data[1], 4, COMM_UINT, rank, 1, comm);

            yac_comm_recv_callback(4, COMM_UINT, rank, 1, comm, send_data[1],
                               recv_callback, NULL, MSG_PRIO_HIGH);
            yac_comm_recv_callback(4, COMM_UINT, rank, 0, comm, send_data[0],
                               recv_callback, NULL, MSG_PRIO_HIGH);

            while (yac_comm_wait(comm, 0) ||
                   yac_comm_wait(comm, 1));

            if (num_received_messages != 2)
               PUT_ERR("error in comm_recv_callback\n")

            num_received_messages = 0;
         }

         { // test exchange of two data packages with different tags between processes

            unsigned send_data[2][4] = {{0,1,2,3}, {3,2,1,0}};

            num_received_messages = 0;

            yac_comm_recv_callback(4, COMM_UINT, rank^1, 1+rank, comm, send_data[0],
                                   recv_callback, NULL, MSG_PRIO_HIGH);
            yac_comm_recv_callback(4, COMM_UINT, rank^1, 3+rank, comm, send_data[1],
                                   recv_callback, NULL, MSG_PRIO_HIGH);

            yac_comm_bsend(send_data[0], 4, COMM_UINT, rank^1, 1+(rank^1), comm);
            yac_comm_bsend(send_data[1], 4, COMM_UINT, rank^1, 3+(rank^1), comm);

            while (yac_comm_wait(comm, 1) ||
                   yac_comm_wait(comm, 2) ||
                   yac_comm_wait(comm, 3) ||
                   yac_comm_wait(comm, 4));

            if (num_received_messages != 2)
               PUT_ERR("error in comm_recv_callback\n")

            num_received_messages = 0;
         }

         { // test test packing and unpacking

            double   send_data_dble[2][2] = {{0,1},{2,3}};
            int      send_data_int [2][2] = {{4,5},{6,7}};
            unsigned send_data_uint[2][2] = {{8,9},{10,11}};

            unsigned buffer_size;
            
            buffer_size = yac_comm_pack_size(2, COMM_DBLE, comm) +
                          yac_comm_pack_size(2, COMM_INT,  comm) +
                          yac_comm_pack_size(2, COMM_UINT, comm);

            char send_buffer[buffer_size];
            int position = 0;

            yac_comm_pack(send_data_dble[rank], 2, COMM_DBLE, send_buffer, buffer_size,
                          &position, comm);
            yac_comm_pack(send_data_int[rank], 2, COMM_INT, send_buffer, buffer_size,
                          &position, comm);
            yac_comm_pack(send_data_uint[rank], 2, COMM_UINT, send_buffer, buffer_size,
                          &position, comm);

            yac_comm_bsend(send_buffer, buffer_size, COMM_PACKED, rank^1, 0, comm);

            char recv_buffer[buffer_size];

            yac_comm_recv(recv_buffer, &buffer_size, COMM_PACKED, rank^1, 0, comm);

            double   recv_data_dble[2];
            int      recv_data_int [2];
            unsigned recv_data_uint[2];

            position = 0;

            yac_comm_unpack(recv_buffer, buffer_size, &position, recv_data_dble, 2,
                            COMM_DBLE, comm);
            yac_comm_unpack(recv_buffer, buffer_size, &position, recv_data_int, 2,
                            COMM_INT, comm);
            yac_comm_unpack(recv_buffer, buffer_size, &position, recv_data_uint, 2,
                            COMM_UINT, comm);

            for (int j = 0; j < 2; ++j) {

               if (recv_data_dble[j] != send_data_dble[rank^1][j])
                  PUT_ERR("error in packing/unpacking (dble)\n");
               if (recv_data_int[j] != send_data_int[rank^1][j])
                  PUT_ERR("error in packing/unpacking (int)\n");
               if (recv_data_uint[j] != send_data_uint[rank^1][j])
                  PUT_ERR("error in packing/unpacking (uint)\n");
            }
         }

        {
          if (rank == 0) {

            unsigned prio_buffer[3] = {-1, -1, -1};
            unsigned * prio_buffer_p = prio_buffer;

            yac_comm_recv_callback(1, COMM_UINT, 1, MSG_PRIO_HIGH, comm,
                                   &prio_buffer_p, prio_recv_callback,
                                   NULL, MSG_PRIO_HIGH);
            yac_comm_recv_callback(1, COMM_UINT, 1, MSG_PRIO_MIDDLE, comm,
                                   &prio_buffer_p, prio_recv_callback,
                                   NULL, MSG_PRIO_MIDDLE);
            yac_comm_recv_callback(1, COMM_UINT, 1, MSG_PRIO_LOW, comm,
                                   &prio_buffer_p, prio_recv_callback,
                                   NULL, MSG_PRIO_LOW);

            while (yac_comm_wait(comm, MSG_PRIO_LOW) ||
                   yac_comm_wait(comm, MSG_PRIO_MIDDLE) ||
                   yac_comm_wait(comm, MSG_PRIO_HIGH));

            unsigned ref_prio_buffer[3] =
              {MSG_PRIO_MIDDLE, MSG_PRIO_HIGH, MSG_PRIO_LOW};

            for (unsigned i = 0; i < 3; ++i)
              if (prio_buffer[i] != ref_prio_buffer[i])
                PUT_ERR("error in message priority handling\n");

          } else { // rank == 1

            unsigned dummy_buffer = -1;
            unsigned data_size = 1;

            // send initial messge
            unsigned msg_prio = MSG_PRIO_MIDDLE;
            yac_comm_bsend(&msg_prio, 1, COMM_UINT, 0, MSG_PRIO_MIDDLE, comm);

            // wait for first answer from middle prio
            yac_comm_recv(&dummy_buffer, &data_size, COMM_UINT, 0,
                          MSG_PRIO_MIDDLE, comm);

            // other rank is now in prio_recv_callback with middle prio waiting
            // for data

            // send low priority message (this should not interrupt middle prio
            // prio_recv_callback call)
            msg_prio = MSG_PRIO_LOW;
            yac_comm_bsend(&msg_prio, 1, COMM_UINT, 0,
                           MSG_PRIO_LOW, comm);

            // send first data set for remote callbacks
            yac_comm_bsend(&dummy_buffer, 1, COMM_UINT, 0, MSG_PRIO_HIGH+10,
                           comm);
            yac_comm_bsend(&dummy_buffer, 1, COMM_UINT, 0, MSG_PRIO_MIDDLE+10,
                           comm);
            yac_comm_bsend(&dummy_buffer, 1, COMM_UINT, 0, MSG_PRIO_LOW+10,
                           comm);

            // wait for second answer from middle prio
            yac_comm_recv(&dummy_buffer, &data_size, COMM_UINT, 0,
                          MSG_PRIO_MIDDLE, comm);

            // send high priority message (this should interrupt middle prio
            // prio_recv_callback call)
            msg_prio = MSG_PRIO_HIGH;
            yac_comm_bsend(&msg_prio, 1, COMM_UINT, MSG_PRIO_HIGH,
                           MSG_PRIO_HIGH, comm);

            // send second data set for remote callbacks
            yac_comm_bsend(&dummy_buffer, 1, COMM_UINT, 0, MSG_PRIO_HIGH+10,
                           comm);
            yac_comm_bsend(&dummy_buffer, 1, COMM_UINT, 0, MSG_PRIO_MIDDLE+10,
                           comm);
            yac_comm_bsend(&dummy_buffer, 1, COMM_UINT, 0, MSG_PRIO_LOW+10,
                           comm);

            // wait for remaining answers
            yac_comm_recv(&dummy_buffer, &data_size, COMM_UINT, 0, MSG_PRIO_HIGH,
                          comm);
            yac_comm_recv(&dummy_buffer, &data_size, COMM_UINT, 0, MSG_PRIO_LOW,
                          comm);
            yac_comm_recv(&dummy_buffer, &data_size, COMM_UINT, 0, MSG_PRIO_HIGH,
                          comm);
            yac_comm_recv(&dummy_buffer, &data_size, COMM_UINT, 0, MSG_PRIO_LOW,
                          comm);
          }
        }

        { // test broadcast with callback

          unsigned comm_size = yac_get_comm_size(comm);

          // for all possible root's
          for (unsigned i = 0; i < comm_size; ++i) {

            unsigned data[3] = {0,1,2};
            unsigned tag = 1;
            unsigned root = i;
            // contents {root, data size, 1 if the message was received}
            unsigned user_data[3] = {root, sizeof(data)/sizeof(data[0]), -1};

            yac_comm_bcast_callback((void*)data, sizeof(data)/sizeof(data[0]),
                                    COMM_UINT, root, tag, comm, user_data,
                                    bcast_callback, MSG_PRIO_HIGH);
            while (yac_comm_wait(comm, tag));

            if (user_data[2] != 1) PUT_ERR("error in yac_comm_bcast_callback\n");
          }
        }

        { // test gatherv with callback

          unsigned rank = yac_get_comm_rank(comm);
          unsigned size = yac_get_comm_size(comm);
          unsigned * data = malloc((rank + 1) * sizeof(data));
          for (unsigned i = 0; i <= rank; ++i)
            data[i] = rank * (rank + 1) / 2 + i;
          unsigned tag = 1;
          unsigned * counts = malloc(size * sizeof(*counts));
          for (unsigned i = 0; i < size; ++i) counts[i] = i + 1;

          // for all possible root's
          for (unsigned i = 0; i < size; ++i) {

            unsigned root = i;

            unsigned user_data[2] = {root, -1};

            yac_comm_gatherv_callback(data, counts, COMM_UINT, root, tag, comm,
                                      user_data, gatherv_callback,
                                      MSG_PRIO_HIGH);
            while (yac_comm_wait(comm, tag));

            if (user_data[1] != 1) PUT_ERR("error in yac_comm_gatherv_callback\n");
          }

          free(counts);
          free(data);
        }

        { // test allgather with callback

          unsigned rank = yac_get_comm_rank(comm);
          unsigned data[2] = {rank * 2 + 0, rank * 2 + 1};
          unsigned tag = 1;
          unsigned count = 2;
          unsigned user_data[2] = {count, -1};

          yac_comm_allgather_callback(data, count, COMM_UINT, tag, comm,
                                      user_data, allgather_callback,
                                      MSG_PRIO_HIGH);
          while (yac_comm_wait(comm, tag));

          if (user_data[1] != 1) PUT_ERR("error in yac_comm_allgather_callback\n");
        }

        { // test allgatherv with callback

          unsigned rank = yac_get_comm_rank(comm);
          unsigned size = yac_get_comm_size(comm);
          unsigned * data = malloc((rank + 1) * sizeof(data));
          for (unsigned i = 0; i <= rank; ++i)
            data[i] = rank * (rank + 1) / 2 + i;
          unsigned tag = 1;
          unsigned * counts = malloc(size * sizeof(*counts));
          for (unsigned i = 0; i < size; ++i) counts[i] = i + 1;
          unsigned user_data[1] = {-1};

          yac_comm_allgatherv_callback(data, counts, COMM_UINT, tag, comm,
                                       user_data, allgatherv_callback,
                                       MSG_PRIO_HIGH);
          free(counts);
          free(data);
          while (yac_comm_wait(comm, tag));

          if (user_data[0] != 1) PUT_ERR("error in yac_comm_allgatherv_callback\n");
        }

        { // test scatter with callback

          unsigned size = yac_get_comm_size(comm);
          unsigned count = 3;
          unsigned * data = malloc(size * count * sizeof(data));
          for (unsigned i = 0; i < size * count; ++i) data[i] = i;
          unsigned tag = 1;

          // for all possible root's
          for (unsigned i = 0; i < size; ++i) {

            unsigned root = i;

            unsigned user_data[3] = {root, count, -1};

            yac_comm_scatter_callback(data, count, COMM_UINT, root, tag, comm,
                                      user_data, scatter_callback,
                                      MSG_PRIO_HIGH);

            while (yac_comm_wait(comm, tag));

            if (user_data[2] != 1)
              PUT_ERR("error in yac_comm_scatter_callback\n");
          }
          free(data);
        }

        { // test scatterv with callback

          unsigned size = yac_get_comm_size(comm);
          unsigned * data = malloc(size * (size + 1) / 2 * sizeof(data));
          for (unsigned i = 0; i < size * (size + 1) / 2; ++i) data[i] = i;
          unsigned tag = 1;
          unsigned * counts = malloc(size * sizeof(*counts));
          for (unsigned i = 0; i < size; ++i) counts[i] = i + 1;

          // for all possible root's
          for (unsigned i = 0; i < size; ++i) {

            unsigned root = i;

            unsigned user_data[2] = {root, -1};

            yac_comm_scatterv_callback(data, counts, COMM_UINT, root, tag, comm,
                                       user_data, scatterv_callback,
                                       MSG_PRIO_HIGH);

            while (yac_comm_wait(comm, tag));

            if (user_data[1] != 1)
              PUT_ERR("error in yac_comm_scatterv_callback\n");
          }

          free(data);
          free(counts);
        }

        { // test alltoallv with callback

          unsigned size = yac_get_comm_size(comm);
          unsigned rank = yac_get_comm_rank(comm);
          unsigned * data = malloc(size * (rank + 1) * sizeof(*data));
          for (unsigned i = 0, offset = 0; i < size; ++i)
            for (unsigned j = 0; j <= rank; ++j, ++offset)
              data[offset] = (rank * (rank + 1)) / 2 + j;
          unsigned tag = 1;
          unsigned * counts = malloc(size * sizeof(*counts));
          for (unsigned i = 0; i < size; ++i)
              counts[i] = rank + 1;
          unsigned user_data = 0;

          yac_comm_alltoallv_callback(data, counts, COMM_UINT, tag, comm,
                                      &user_data, alltoallv_callback,
                                      MSG_PRIO_HIGH);

          while (yac_comm_wait(comm, tag));

          if (user_data != 1)
            PUT_ERR("error in yac_comm_alltoallv_callback\n");

          free(counts);
          free(data);
        }

        { // test reduce with callback

          unsigned size = yac_get_comm_size(comm);
          unsigned rank = yac_get_comm_rank(comm);
          unsigned count = 3;
          unsigned * user_data = calloc(3 + count, sizeof(*user_data));
          user_data[1] = count;
          for (unsigned i = 0; i < size; ++i) {
            for (unsigned j = 0; j < count; ++j) {
              user_data[3 + j] += i * count + j;
            }
          }
          unsigned * data = malloc(count * sizeof(*data));
          for (unsigned j = 0; j < count; ++j) data[j] = rank * count + j;
          unsigned tag = 1;

          for (unsigned root = 0; root < size; ++root) {

            user_data[0] = 0;
            user_data[2] = root;

            yac_comm_reduce_callback(data, count, COMM_UINT, COMM_OP_SUM, root,
                                     tag, comm, user_data, reduce_callback,
                                     MSG_PRIO_HIGH);

            while (yac_comm_wait(comm, tag));

            if (user_data[0] != 1)
              PUT_ERR("error in yac_comm_reduce_callback\n");
          }

          free(user_data);
          free(data);
        }

        { // test allreduce with callback

          unsigned size = yac_get_comm_size(comm);
          unsigned rank = yac_get_comm_rank(comm);
          unsigned count = 3;
          unsigned * user_data = calloc(2 + count, sizeof(*user_data));
          user_data[1] = count;
          for (unsigned i = 0; i < size; ++i) {
            for (unsigned j = 0; j < count; ++j) {
              user_data[2 + j] += i * count + j;
            }
          }
          unsigned * data = malloc(count * sizeof(*data));
          for (unsigned j = 0; j < count; ++j) data[j] = rank * count + j;
          unsigned tag = 1;

          yac_comm_allreduce_callback(data, count, COMM_UINT, COMM_OP_SUM, tag,
                                      comm, user_data, allreduce_callback,
                                      MSG_PRIO_HIGH);

          while (yac_comm_wait(comm, tag));

          if (user_data[0] != 1)
            PUT_ERR("error in yac_comm_allreduce_callback\n");

          free(user_data);
          free(data);
        }

        { // test barrier with callback
          unsigned user_data = 0;
          unsigned tag = 1;

          yac_comm_barrier_callback(tag, comm, &user_data, barrier_callback,
                                    MSG_PRIO_HIGH);

          while (yac_comm_wait(comm, tag));

          if (user_data != 1)
            PUT_ERR("error in yac_comm_barrier_callback\n");
        }
      }

      //---------------
      // cleanup
      //---------------
      yac_free_communicator(comm);
   }

   { // tests with a single process in comm (if the program was started with more, then
     //                                      the global communicator split until each
     //                                      communicator contains only a single process)

      //---------------
      // setup
      //---------------

      struct communicator * comm;

      comm = yac_mpi_comm_new();

      // split if there is more than one process
      if (yac_get_comm_size(comm) > 1) {

         struct communicator * temp_comm;

         temp_comm = yac_comm_split(comm, yac_get_comm_rank(comm), 0);

         yac_free_communicator(comm);

         comm = temp_comm;
      }

      //---------------
      // testing
      //---------------

      { // test the cancelation of asynchronous receives

         unsigned send_data[2][4] = {{0,1,2,3}, {3,2,1,0}};

         yac_comm_recv_callback(4, COMM_UINT, 0, 1, comm, send_data[0],
                                recv_callback, recv_callback_cancel,
                                MSG_PRIO_HIGH);
         yac_comm_recv_callback(4, COMM_UINT, 0, 2, comm, send_data[0],
                                recv_callback, recv_callback_cancel,
                                MSG_PRIO_HIGH);
         yac_comm_recv_callback(4, COMM_UINT, 0, 3, comm, send_data[1],
                                recv_callback, recv_callback_cancel,
                                MSG_PRIO_HIGH);
         yac_comm_recv_callback(4, COMM_UINT, 0, 4, comm, send_data[1],
                                recv_callback, recv_callback_cancel,
                                MSG_PRIO_HIGH);

         yac_comm_cancel_recv(comm, 1);
         yac_comm_bsend(send_data[0], 4, COMM_UINT, 0, 2, comm);

         while (yac_comm_wait(comm, 1) ||
                yac_comm_wait(comm, 2));

         if (num_received_messages != 1)
            PUT_ERR("error in comm_recv_callback\n")
         if (num_canceled_messages != 1)
            PUT_ERR("error in comm_recv_callback\n")

         num_received_messages = 0;
         num_canceled_messages = 0;

         yac_comm_bsend(send_data[1], 4, COMM_UINT, 0, 3, comm);
         yac_comm_cancel_recv(comm, 4);

         while (yac_comm_wait(comm, 3) ||
                yac_comm_wait(comm, 4));

         if (num_received_messages != 1)
            PUT_ERR("error in comm_recv_callback\n")
         if (num_canceled_messages != 1)
            PUT_ERR("error in comm_recv_callback\n")

         num_received_messages = 0;
         num_canceled_messages = 0;
      }

      { // test the cancelation of asynchronous receives when freeing a
        // communicator

         struct communicator * dup_comm = yac_comm_dup(comm);

         yac_comm_recv_callback(1, COMM_UINT, 0, 0, dup_comm, NULL, NULL,
                                recv_callback_cancel, MSG_PRIO_HIGH);
         yac_comm_recv_callback(1, COMM_UINT, 0, 0, dup_comm, NULL, NULL,
                                recv_callback_cancel_2, MSG_PRIO_HIGH);
         yac_comm_recv_callback(1, COMM_UINT, 0, 0, dup_comm, NULL, NULL,
                                recv_callback_cancel, MSG_PRIO_HIGH);
         yac_comm_recv_callback(1, COMM_UINT, 0, 0, dup_comm, NULL, NULL,
                                recv_callback_cancel, MSG_PRIO_HIGH);

         yac_free_communicator(dup_comm);

         if (num_canceled_messages != 4)
            PUT_ERR("error in yac_free_communicator\n")

         num_canceled_messages = 0;
      }

      //---------------
      // cleanup
      //---------------

      yac_free_communicator(comm);
   }

   { // test with four processes in comm

      //---------------
      // setup
      //---------------
      struct communicator * comm;

      comm = yac_mpi_comm_new();

      // split if there are more than four process
      if (yac_get_comm_size(comm) > 4) {

         struct communicator * temp_comm;

         temp_comm = yac_comm_split(comm, yac_get_comm_rank(comm) >> 2, 0);

         yac_free_communicator(comm);

         comm = temp_comm;
      }

      int rank = yac_get_comm_rank(comm);

      // split if there is more than one process
      if (yac_get_comm_size(comm) == 4) {

         { // testing comm_split_intercomm
            struct communicator * intercomm;

            // generate intercommunicator with process 0 in one group and 1,2,3 in the other
            intercomm = yac_comm_split_intercomm(comm, rank?1:0, NULL, NULL);

            int comm_size;

            comm_size = yac_get_comm_size(intercomm);

            if (((rank == 0) && (comm_size != 1)) || ((rank != 0) && (comm_size != 3)))
               PUT_ERR("error in get_comm_size intercomm\n");

            comm_size = yac_get_comm_remote_size(intercomm);

            if (((rank == 0) && (comm_size != 3)) || ((rank != 0) && (comm_size != 1)))
               PUT_ERR("error in get_comm_remote_size intercomm\n");

            if ((rank == 0) || (rank == 1)) {

               int recv_buffer;
               unsigned count;

               count = 1;

               yac_comm_bsend(&rank, count, COMM_INT, 0, 0, intercomm);
               yac_comm_recv(&recv_buffer, &count, COMM_INT, 0, 0, intercomm);

               if (rank != (recv_buffer^1))
                  PUT_ERR("error in comm_recv intercomm\n");
            }

            yac_free_communicator(intercomm);
         }

         { // test comm_translate_ranks

            struct communicator * temp_comm;

            temp_comm = yac_comm_split(comm, rank&1, rank>>1);

            unsigned ref_ranks[2][2] = {{0,2},{1,3}};
            unsigned ranks[2];

            yac_comm_translate_ranks(comm, temp_comm, ranks);

            if ((ranks[0] != ref_ranks[rank&1][0]) ||
                (ranks[1] != ref_ranks[rank&1][1]))
               PUT_ERR("error in comm_translate_ranks\n");

            yac_free_communicator(temp_comm);
         }

        { // test broadcast with callback

          unsigned comm_size = yac_get_comm_size(comm);

          // for all possible root's
          for (unsigned i = 0; i < comm_size; ++i) {

            unsigned data[3] = {0,1,2};
            unsigned tag = 1;
            unsigned root = i;
            // contents {root, data size, 1 if the message was received}
            unsigned user_data[3] = {root, sizeof(data)/sizeof(data[0]), -1};

            yac_comm_bcast_callback((void*)data, sizeof(data)/sizeof(data[0]),
                                    COMM_UINT, root, tag, comm, user_data,
                                    bcast_callback, MSG_PRIO_HIGH);
            while (yac_comm_wait(comm, tag));

            if (user_data[2] != 1) PUT_ERR("error in yac_comm_bcast_callback\n");
          }
        }

        { // test gatherv with callback

          unsigned rank = yac_get_comm_rank(comm);
          unsigned size = yac_get_comm_size(comm);
          unsigned * data = malloc((rank + 1) * sizeof(data));
          for (unsigned i = 0; i <= rank; ++i)
            data[i] = rank * (rank + 1) / 2 + i;
          unsigned tag = 1;
          unsigned * counts = malloc(size * sizeof(*counts));
          for (unsigned i = 0; i < size; ++i) counts[i] = i + 1;

          // for all possible root's
          for (unsigned i = 0; i < size; ++i) {

            unsigned root = i;

            unsigned user_data[2] = {root, -1};

            yac_comm_gatherv_callback(data, counts, COMM_UINT, root, tag, comm,
                                      user_data, gatherv_callback,
                                      MSG_PRIO_HIGH);
            while (yac_comm_wait(comm, tag));

            if (user_data[1] != 1) PUT_ERR("error in yac_comm_gatherv_callback\n");
          }

          free(counts);
          free(data);
        }

        { // test allgather with callback

          unsigned rank = yac_get_comm_rank(comm);
          unsigned data[2] = {rank * 2 + 0, rank * 2 + 1};
          unsigned tag = 1;
          unsigned count = 2;
          // contents {number of elements per process, 1 if the message was
          //           received}
          unsigned user_data[2] = {count, -1};

          yac_comm_allgather_callback(data, count, COMM_UINT, tag, comm,
                                      user_data, allgather_callback,
                                      MSG_PRIO_HIGH);
          while (yac_comm_wait(comm, tag));

          if (user_data[1] != 1) PUT_ERR("error in yac_comm_allgather_callback\n");
        }

        { // test allgatherv with callback

          unsigned rank = yac_get_comm_rank(comm);
          unsigned size = yac_get_comm_size(comm);
          unsigned * data = malloc((rank + 1) * sizeof(data));
          for (unsigned i = 0; i <= rank; ++i)
            data[i] = rank * (rank + 1) / 2 + i;
          unsigned tag = 1;
          unsigned * counts = malloc(size * sizeof(*counts));
          for (unsigned i = 0; i < size; ++i) counts[i] = i + 1;
          unsigned user_data[1] = {-1};

          yac_comm_allgatherv_callback(data, counts, COMM_UINT, tag, comm,
                                       user_data, allgatherv_callback,
                                       MSG_PRIO_HIGH);
          free(counts);
          free(data);
          while (yac_comm_wait(comm, tag));

          if (user_data[0] != 1) PUT_ERR("error in yac_comm_allgatherv_callback\n");
        }

        { // test scatter with callback

          unsigned size = yac_get_comm_size(comm);
          unsigned count = 6;
          unsigned * data = malloc(size * count * sizeof(data));
          for (unsigned i = 0; i < size * count; ++i) data[i] = i;
          unsigned tag = 1;

          // for all possible root's
          for (unsigned i = 0; i < size; ++i) {

            unsigned root = i;

            unsigned user_data[3] = {root, count, -1};

            yac_comm_scatter_callback(data, count, COMM_UINT, root, tag, comm,
                                      user_data, scatter_callback,
                                      MSG_PRIO_HIGH);

            while (yac_comm_wait(comm, tag));

            if (user_data[2] != 1)
              PUT_ERR("error in yac_comm_scatter_callback\n");
          }

          free(data);
        }

        { // test scatterv with callback

          unsigned size = yac_get_comm_size(comm);
          unsigned * data = malloc(size * (size + 1) / 2 * sizeof(data));
          for (unsigned i = 0; i < size * (size + 1) / 2; ++i) data[i] = i;
          unsigned tag = 1;
          unsigned * counts = malloc(size * sizeof(*counts));
          for (unsigned i = 0; i < size; ++i) counts[i] = i + 1;

          // for all possible root's
          for (unsigned i = 0; i < size; ++i) {

            unsigned root = i;

            unsigned user_data[2] = {root, -1};

            yac_comm_scatterv_callback(data, counts, COMM_UINT, root, tag, comm,
                                       user_data, scatterv_callback,
                                       MSG_PRIO_HIGH);

            while (yac_comm_wait(comm, tag));

            if (user_data[1] != 1)
              PUT_ERR("error in yac_comm_scatterv_callback\n");
          }

          free(counts);
          free(data);
        }

        { // test reduce with callback

          unsigned size = yac_get_comm_size(comm);
          unsigned rank = yac_get_comm_rank(comm);
          unsigned count = 3;
          unsigned * user_data = calloc(3 + count, sizeof(*user_data));
          user_data[1] = count;
          for (unsigned i = 0; i < size; ++i) {
            for (unsigned j = 0; j < count; ++j) {
              user_data[3 + j] += i * count + j;
            }
          }
          unsigned * data = malloc(count * sizeof(*data));
          for (unsigned j = 0; j < count; ++j) data[j] = rank * count + j;
          unsigned tag = 1;

          for (unsigned root = 0; root < size; ++root) {

            user_data[0] = 0;
            user_data[2] = root;

            yac_comm_reduce_callback(data, count, COMM_UINT, COMM_OP_SUM, root,
                                     tag, comm, user_data, reduce_callback,
                                     MSG_PRIO_HIGH);

            while (yac_comm_wait(comm, tag));

            if (user_data[0] != 1)
              PUT_ERR("error in yac_comm_reduce_callback\n");
          }

          free(user_data);
          free(data);
        }

        { // test allreduce with callback

          unsigned size = yac_get_comm_size(comm);
          unsigned rank = yac_get_comm_rank(comm);
          unsigned count = 3;
          unsigned * user_data = calloc(2 + count, sizeof(*user_data));
          user_data[1] = count;
          for (unsigned i = 0; i < size; ++i) {
            for (unsigned j = 0; j < count; ++j) {
              user_data[2 + j] += i * count + j;
            }
          }
          unsigned * data = malloc(count * sizeof(*data));
          for (unsigned j = 0; j < count; ++j) data[j] = rank * count + j;
          unsigned tag = 1;

          yac_comm_allreduce_callback(data, count, COMM_UINT, COMM_OP_SUM, tag,
                                      comm, user_data, allreduce_callback,
                                      MSG_PRIO_HIGH);

          while (yac_comm_wait(comm, tag));

          if (user_data[0] != 1)
            PUT_ERR("error in yac_comm_allreduce_callback\n");

          free(user_data);
          free(data);
        }

        { // test barrier with callback
          unsigned user_data = 0;
          unsigned tag = 1;

          yac_comm_barrier_callback(tag, comm, &user_data, barrier_callback,
                                    MSG_PRIO_HIGH);

          while (yac_comm_wait(comm, tag));

          if (user_data != 1)
            PUT_ERR("error in yac_comm_barrier_callback\n");
        }

        { // test irecv/waitall
          unsigned tag = 1;
          unsigned size = yac_get_comm_size(comm);
          unsigned rank = yac_get_comm_rank(comm);

          for (unsigned i = 0; i < 2; ++i) {

            unsigned recv_rank = -1;
            struct comm_request * requests[1];

            if (i == 0)
              yac_comm_bsend(&rank, 1, COMM_UINT, (rank+1)%size, tag, comm);
            yac_comm_irecv(&recv_rank, 1, COMM_UINT, (rank-1+size)%size, tag,
                           comm, requests);
            if (i == 1)
              yac_comm_bsend(&rank, 1, COMM_UINT, (rank+1)%size, tag, comm);
            yac_comm_waitall(1, requests);

            if (recv_rank != (rank-1+size)%size)
              PUT_ERR("error in irecv/waitall\n");
          }
        }

        { // test isend/irecv/waitall
          unsigned tag = 1;
          unsigned size = yac_get_comm_size(comm);
          unsigned rank = yac_get_comm_rank(comm);

          for (unsigned i = 0; i < 2; ++i) {

            unsigned recv_rank = -1;
            struct comm_request * requests[2];

            if (i == 0)
              yac_comm_isend(&rank, 1, COMM_UINT, (rank+1)%size, tag, comm,
                             &(requests[0]));
            yac_comm_irecv(&recv_rank, 1, COMM_UINT, (rank-1+size)%size, tag,
                           comm, &(requests[1]));
            if (i == 1)
              yac_comm_isend(&rank, 1, COMM_UINT, (rank+1)%size, tag, comm,
                             &(requests[0]));
            yac_comm_waitall(2, requests);

            if (recv_rank != (rank-1+size)%size)
              PUT_ERR("error in isend/irecv/waitall\n");
          }
        }
      }

      //---------------
      // cleanup
      //---------------

      yac_free_communicator(comm);
   }

   { // test with one to comm_size processes

      //---------------
      // setup
      //---------------
      struct communicator * comm;

      comm = yac_mpi_comm_new();

      unsigned comm_size = yac_get_comm_size(comm);

      for (unsigned i = 3; i <= comm_size; ++i) {

        struct communicator * split_comm =
          yac_comm_split(comm, yac_get_comm_rank(comm) / i, 0);

        if (yac_get_comm_size(split_comm) != i) continue;

        // func == 0 => no callback routine is provided
        // func == 1 => callback routine is provided
        for (unsigned func = 0; func < 2; ++func) {

          { // test broadcast with callback

            unsigned comm_size = yac_get_comm_size(split_comm);

            // for all possible root's
            for (unsigned i = 0; i < comm_size; ++i) {

              unsigned data[3] = {0,1,2};
              unsigned tag = 1;
              unsigned root = i;
              // contents {root, data size, 1 if the message was received}
              unsigned user_data[3] = {root, sizeof(data)/sizeof(data[0]), -1};

              yac_comm_bcast_callback((void*)data, sizeof(data)/sizeof(data[0]),
                                      COMM_UINT, root, tag, split_comm,
                                      user_data, (func)?bcast_callback:NULL,
                                      MSG_PRIO_HIGH);
              while (yac_comm_wait(split_comm, tag));

              if (func)
                if (user_data[2] != 1)
                  PUT_ERR("error in yac_comm_bcast_callback\n");
            }
          }

          { // test gatherv with callback

            unsigned rank = yac_get_comm_rank(split_comm);
            unsigned size = yac_get_comm_size(split_comm);
            unsigned * data = malloc((rank + 1) * sizeof(data));
            for (unsigned i = 0; i <= rank; ++i)
              data[i] = rank * (rank + 1) / 2 + i;
            unsigned tag = 1;
            unsigned * counts = malloc(size * sizeof(*counts));
            for (unsigned i = 0; i < size; ++i) counts[i] = i + 1;

            // for all possible root's
            for (unsigned i = 0; i < size; ++i) {

              unsigned root = i;

              unsigned user_data[2] = {root, -1};

              yac_comm_gatherv_callback(data, counts, COMM_UINT, root, tag,
                                        split_comm, user_data,
                                        (func)?gatherv_callback:NULL,
                                        MSG_PRIO_HIGH);
              while (yac_comm_wait(split_comm, tag));

              if (func)
                if (user_data[1] != 1)
                  PUT_ERR("error in yac_comm_gatherv_callback\n");
            }

            free(counts);
            free(data);
          }

          { // test allgather with callback

            unsigned rank = yac_get_comm_rank(split_comm);
            unsigned data[2] = {rank * 2 + 0, rank * 2 + 1};
            unsigned tag = 1;
            unsigned count = 2;
            // contents {number of elements per process, 1 if the message was
            //           received}
            unsigned user_data[2] = {count, -1};

            yac_comm_allgather_callback(data, count, COMM_UINT, tag, split_comm,
                                        user_data,
                                        (func)?allgather_callback:NULL,
                                        MSG_PRIO_HIGH);
            while (yac_comm_wait(split_comm, tag));

            if (func)
              if (user_data[1] != 1)
                PUT_ERR("error in yac_comm_allgather_callback\n");
          }

          { // test allgatherv with callback

            unsigned rank = yac_get_comm_rank(split_comm);
            unsigned size = yac_get_comm_size(split_comm);
            unsigned * data = malloc((rank + 1) * sizeof(data));
            for (unsigned i = 0; i <= rank; ++i)
              data[i] = rank * (rank + 1) / 2 + i;
            unsigned tag = 1;
            unsigned * counts = malloc(size * sizeof(*counts));
            for (unsigned i = 0; i < size; ++i) counts[i] = i + 1;
            unsigned user_data[1] = {-1};

            yac_comm_allgatherv_callback(data, counts, COMM_UINT, tag, split_comm,
                                         user_data,
                                         (func)?allgatherv_callback:NULL,
                                         MSG_PRIO_HIGH);
            free(counts);
            free(data);
            while (yac_comm_wait(split_comm, tag));

            if (func)
              if (user_data[0] != 1)
                PUT_ERR("error in yac_comm_allgatherv_callback\n");
          }

          { // test scatter with callback

            unsigned count = 5;
            unsigned size = yac_get_comm_size(split_comm);
            unsigned * data = malloc(size * count * sizeof(data));
            for (unsigned i = 0; i < size * count; ++i) data[i] = i;
            unsigned tag = 1;

            // for all possible root's
            for (unsigned i = 0; i < size; ++i) {

              unsigned root = i;

              unsigned user_data[3] = {root, count, -1};

              yac_comm_scatter_callback(data, count, COMM_UINT, root, tag,
                                        split_comm, user_data,
                                        (func)?scatter_callback:NULL,
                                        MSG_PRIO_HIGH);

              while (yac_comm_wait(split_comm, tag));

              if (func)
                if (user_data[2] != 1)
                  PUT_ERR("error in yac_comm_scatter_callback\n");
            }

            free(data);
          }

          { // test scatterv with callback

            unsigned size = yac_get_comm_size(split_comm);
            unsigned * data = malloc(size * (size + 1) / 2 * sizeof(data));
            for (unsigned i = 0; i < size * (size + 1) / 2; ++i) data[i] = i;
            unsigned tag = 1;
            unsigned * counts = malloc(size * sizeof(*counts));
            for (unsigned i = 0; i < size; ++i) counts[i] = i + 1;

            // for all possible root's
            for (unsigned i = 0; i < size; ++i) {

              unsigned root = i;

              unsigned user_data[2] = {root, -1};

              yac_comm_scatterv_callback(data, counts, COMM_UINT, root, tag, split_comm,
                                         user_data,
                                         (func)?scatterv_callback:NULL,
                                         MSG_PRIO_HIGH);

              while (yac_comm_wait(split_comm, tag));

              if (func)
                if (user_data[1] != 1)
                  PUT_ERR("error in yac_comm_scatterv_callback\n");
            }

            free(counts);
            free(data);
          }

          { // test alltoallv with callback

            unsigned size = yac_get_comm_size(split_comm);
            unsigned rank = yac_get_comm_rank(split_comm);
            unsigned * data = malloc(size * (rank + 1) * sizeof(*data));
            for (unsigned i = 0, offset = 0; i < size; ++i)
              for (unsigned j = 0; j <= rank; ++j, ++offset)
                data[offset] = (rank * (rank + 1)) / 2 + j;
            unsigned tag = 1;
            unsigned * counts = malloc(size * sizeof(*counts));
            for (unsigned i = 0; i < size; ++i)
                counts[i] = rank + 1;
            unsigned user_data = 0;

            yac_comm_alltoallv_callback(data, counts, COMM_UINT, tag, split_comm,
                                        &user_data,
                                        (func)?alltoallv_callback:NULL,
                                        MSG_PRIO_HIGH);

            while (yac_comm_wait(split_comm, tag));

            if (func)
              if (user_data != 1)
                PUT_ERR("error in yac_comm_alltoallv_callback\n");

            free(counts);
            free(data);
          }

          { // test alltoallv with callback

            unsigned size = yac_get_comm_size(split_comm);
            unsigned rank = yac_get_comm_rank(split_comm);
            char * data = malloc(size * sizeof(*data));
            for (unsigned i = 0; i < size; ++i)
              data[i] = rank;
            unsigned tag = 1;
            unsigned * counts = malloc(size * sizeof(*counts));
            for (unsigned i = 0; i < size; ++i)
                counts[i] = 1;
            unsigned user_data = 0;

            yac_comm_alltoallv_callback(data, counts, COMM_BYTE, tag, split_comm,
                                        &user_data,
                                        (func)?alltoallv_char_callback:NULL,
                                        MSG_PRIO_HIGH);

            while (yac_comm_wait(split_comm, tag));

            if (func)
              if (user_data != 1)
                PUT_ERR("error in yac_comm_alltoallv_callback\n");

            free(counts);
            free(data);
          }

          { // test alltoallv with callback (with pseudo random input)
#define NUM_RAND_TESTS 100

            unsigned size = yac_get_comm_size(split_comm);
            unsigned rank = yac_get_comm_rank(split_comm);
            unsigned tag = 1;

            unsigned * counts = malloc(size * size * sizeof(*counts));
            unsigned * local_counts = counts + size * rank;

            srand(23094875);

            for (unsigned t = 0; t < NUM_RAND_TESTS; ++t) {

              for (unsigned i = 0; i < size * size; ++i)
                counts[i] = rand()%10;

              unsigned data_size = 0;
              for (unsigned i = 0; i < size * size; ++i)
                data_size += counts[i];

              unsigned * data = malloc(data_size * sizeof(*data));
              for (unsigned i = 0; i < data_size; ++i)
                data[i] = rand();

              unsigned * local_data = data;
              for (unsigned i = 0; i < rank; ++i)
                for (unsigned j = 0; j < size; ++j)
                  local_data += counts[i * size + j];

              unsigned ref_data_size = 0;
              for (unsigned i = 0; i < size; ++i)
                ref_data_size += counts[i * size + rank];

              unsigned * ref_data =
                malloc((2 + ref_data_size) * sizeof(*ref_data));
              ref_data[0] = 0;
              ref_data[1] = ref_data_size;
              {
                unsigned * curr_data = data;
                unsigned * curr_ref_data = ref_data + 2;

                for (unsigned i = 0; i < size; ++i) {
                  for (unsigned j = 0; j < size; ++j) {

                    if (j == rank) {

                      memcpy(curr_ref_data, curr_data,
                             counts[i * size + j] * sizeof(*curr_ref_data));
                      curr_ref_data += counts[i * size + j];
                    }
                    curr_data += counts[i * size + j];
                  }
                }
              }

              yac_comm_alltoallv_callback(local_data, local_counts, COMM_UINT,
                                          tag, split_comm, ref_data,
                                          (func)?alltoallv_rand_callback:NULL,
                                          MSG_PRIO_HIGH);

              while (yac_comm_wait(split_comm, tag));

              if (func)
                if (ref_data[0] != 1)
                  PUT_ERR("error in yac_comm_alltoallv_callback\n");

              free(ref_data);
              free(data);
            }
            free(counts);
#undef NUM_RAND_TESTS
          }

          { // test reduce with callback

            unsigned size = yac_get_comm_size(split_comm);
            unsigned rank = yac_get_comm_rank(split_comm);
            unsigned count = 3;
            unsigned * user_data = calloc(3 + count, sizeof(*user_data));
            user_data[1] = count;
            for (unsigned i = 0; i < size; ++i) {
              for (unsigned j = 0; j < count; ++j) {
                user_data[3 + j] += i * count + j;
              }
            }
            unsigned * data = malloc(count * sizeof(*data));
            for (unsigned j = 0; j < count; ++j) data[j] = rank * count + j;
            unsigned tag = 1;

            for (unsigned root = 0; root < size; ++root) {

              user_data[0] = 0;
              user_data[2] = root;

              yac_comm_reduce_callback(data, count, COMM_UINT, COMM_OP_SUM,
                                       root, tag, split_comm, user_data,
                                       (func)?reduce_callback:NULL,
                                       MSG_PRIO_HIGH);

              while (yac_comm_wait(split_comm, tag));

              if (func)
                if (user_data[0] != 1)
                  PUT_ERR("error in yac_comm_reduce_callback\n");
            }

            free(user_data);
            free(data);
          }

          { // test allreduce with callback

            unsigned size = yac_get_comm_size(split_comm);
            unsigned rank = yac_get_comm_rank(split_comm);
            unsigned count = 3;
            unsigned * user_data = calloc(2 + count, sizeof(*user_data));
            user_data[1] = count;
            for (unsigned i = 0; i < size; ++i) {
              for (unsigned j = 0; j < count; ++j) {
                user_data[2 + j] += i * count + j;
              }
            }
            unsigned * data = malloc(count * sizeof(*data));
            for (unsigned j = 0; j < count; ++j) data[j] = rank * count + j;
            unsigned tag = 1;

            yac_comm_allreduce_callback(data, count, COMM_UINT, COMM_OP_SUM,
                                        tag, split_comm, user_data,
                                        (func)?allreduce_callback:NULL,
                                        MSG_PRIO_HIGH);

            while (yac_comm_wait(split_comm, tag));

            if (func)
              if (user_data[0] != 1)
                PUT_ERR("error in yac_comm_allreduce_callback\n");

            free(user_data);
            free(data);
          }

          { // test barrier with callback
            unsigned user_data = 0;
            unsigned tag = 1;

            yac_comm_barrier_callback(tag, split_comm, &user_data,
                                     (func)?barrier_callback:NULL,
                                      MSG_PRIO_HIGH);

            while (yac_comm_wait(split_comm, tag));

            if (func)
              if (user_data != 1)
                PUT_ERR("error in yac_comm_barrier_callback\n");
          }
        }

        yac_free_communicator(split_comm);
      }

      //---------------
      // cleanup
      //---------------

      yac_free_communicator(comm);
   }

   if (TEST_EXIT_CODE)
      yac_comm_abort(yac_world_comm);

   yac_free_communicator(yac_world_comm);

   yac_comm_finalize();

   return TEST_EXIT_CODE;
}

void recv_callback(struct communicator * comm, struct comm_message message,
                   void * orig_data) {

   if (message.count != 4)
      PUT_ERR("error in message.count\n")

   if (message.dst_rank != yac_get_comm_rank(comm))
      PUT_ERR("error in message.dst_rank\n")

   unsigned * send_data = orig_data;
   unsigned i;

   for (i = 0; i < 4; ++i)
      if (((unsigned *)(message.buffer))[i] != send_data[i])
         PUT_ERR("error in message.buffer\n")

   num_received_messages++;
}

void recv_callback_cancel_2(struct communicator * comm, void * orig_data) {

   // just to be mean we set up a new recv_callback here  :-)
   // (done to ensure, that yac_free_communicator really cancels all messages)
   yac_comm_recv_callback(1, COMM_UINT, 0, 0, comm, NULL, NULL,
                          recv_callback_cancel, MSG_PRIO_HIGH);
}

void recv_callback_cancel(struct communicator * comm, void * orig_data) {

   num_canceled_messages++;
}

void prio_recv_callback(struct communicator * comm, struct comm_message msg,
                        void * user_data) {

  unsigned ** prio_buffer = (unsigned**)user_data;
  unsigned msg_prio = *((unsigned*)(msg.buffer));

  **prio_buffer = msg_prio;
  ++*prio_buffer;

  unsigned dummy_buffer = -1;
  unsigned data_size = 1;
  yac_comm_bsend(&dummy_buffer, 1, COMM_UINT, 1, msg_prio, comm);
  yac_comm_recv(&dummy_buffer, &data_size, COMM_UINT, 1, msg_prio+10, comm);
  yac_comm_bsend(&dummy_buffer, 1, COMM_UINT, 1, msg_prio, comm);
  yac_comm_recv(&dummy_buffer, &data_size, COMM_UINT, 1, msg_prio+10, comm);
}

void bcast_callback(struct communicator * comm, struct comm_message message,
                    void * orig_data) {

  unsigned root = ((unsigned *)orig_data)[0];
  unsigned ref_count = ((unsigned *)orig_data)[1];
  unsigned * data = (unsigned*)(message.buffer);
  unsigned rank = yac_get_comm_rank(comm);

  if (message.type != COMM_UINT)
    PUT_ERR("error in message.type\n");

  if (message.dst_rank != rank)
    PUT_ERR("error in message.dst_rank\n");

  if (message.src_rank != root)
    PUT_ERR("error in message.src_rank\n");

  for (unsigned i = 0; i < ref_count; ++i)
    if (data[i] != i) PUT_ERR("error in message.buffer\n");

  ((unsigned *)orig_data)[2] = 1;
}

void gatherv_callback(struct communicator * comm, struct comm_message message,
                     void * orig_data) {

  unsigned root = ((unsigned *)orig_data)[0];
  unsigned comm_size = yac_get_comm_size(comm);
  unsigned comm_rank = yac_get_comm_rank(comm);
  unsigned count = (comm_rank == root)?(comm_size * (comm_size + 1) / 2):(0);
  unsigned * buffer = (unsigned*)(message.buffer);

  if (count != message.count)
    PUT_ERR("error in message.count\n");

  if (message.type != COMM_UINT)
    PUT_ERR("error in message.type\n");

  for (unsigned i = 0; i < count; ++i)
    if (buffer[i] != i)
      PUT_ERR("error in message.buffer\n");

  if ((count == 0) && (buffer != NULL))
    PUT_ERR("error in message.buffer\n");

  ((unsigned *)orig_data)[1] = 1;
}

void allgather_callback(struct communicator * comm, struct comm_message message,
                        void * orig_data) {

  unsigned comm_size = yac_get_comm_size(comm);
  unsigned comm_rank = yac_get_comm_rank(comm);
  unsigned count = *(unsigned*)orig_data;
  unsigned * buffer = (unsigned*)(message.buffer);

  if (count * comm_size != message.count)
    PUT_ERR("error in message.count\n");

  if (message.type != COMM_UINT)
    PUT_ERR("error in message.type\n");

  for (unsigned i = 0; i < count * comm_size; ++i)
    if (buffer[i] != i)
      PUT_ERR("error in message.buffer\n");

  if (message.src_rank != comm_rank)
    PUT_ERR("error in message.src_rank\n");

  ((unsigned *)orig_data)[1] = 1;
}

void allgatherv_callback(struct communicator * comm, struct comm_message message,
                        void * orig_data) {

  unsigned comm_size = yac_get_comm_size(comm);
  unsigned comm_rank = yac_get_comm_rank(comm);
  unsigned count = comm_size * (comm_size + 1) / 2;
  unsigned * buffer = (unsigned*)(message.buffer);

  if (count != message.count)
    PUT_ERR("error in message.count\n");

  if (message.type != COMM_UINT)
    PUT_ERR("error in message.type\n");

  for (unsigned i = 0; i < count; ++i)
    if (buffer[i] != i)
      PUT_ERR("error in message.buffer\n");

  if (message.src_rank != comm_rank)
    PUT_ERR("error in message.src_rank\n");

  ((unsigned *)orig_data)[0] = 1;
}

void scatter_callback(struct communicator * comm, struct comm_message message,
                      void * orig_data) {

  unsigned root = ((unsigned *)orig_data)[0];
  unsigned count = ((unsigned *)orig_data)[1];
  unsigned comm_rank = yac_get_comm_rank(comm);
  unsigned * buffer = (unsigned*)(message.buffer);

  if (count != message.count)
    PUT_ERR("error in message.count\n");

  if (message.type != COMM_UINT)
    PUT_ERR("error in message.type\n");

  for (unsigned i = 0; i < count; ++i)
    if (buffer[i] != i + comm_rank * count)
      PUT_ERR("error in message.buffer\n");

  if (message.src_rank != root)
    PUT_ERR("error in message.src_rank\n");

  ((unsigned *)orig_data)[2] = 1;
}

void scatterv_callback(struct communicator * comm, struct comm_message message,
                       void * orig_data) {

  unsigned root = ((unsigned *)orig_data)[0];
  unsigned comm_rank = yac_get_comm_rank(comm);
  unsigned n1 = comm_rank * (comm_rank + 1) / 2;
  unsigned count = comm_rank + 1;
  unsigned * buffer = (unsigned*)(message.buffer);

  if (count != message.count)
    PUT_ERR("error in message.count\n");

  if (message.type != COMM_UINT)
    PUT_ERR("error in message.type\n");

  for (unsigned i = 0; i < count; ++i)
    if (buffer[i] != i + n1)
      PUT_ERR("error in message.buffer\n");

  if (message.src_rank != root)
    PUT_ERR("error in message.src_rank\n");

  ((unsigned *)orig_data)[1] = 1;
}

void alltoallv_callback(struct communicator * comm,
                        struct comm_message message, void * orig_data) {

  unsigned comm_size = yac_get_comm_size(comm);
  unsigned comm_rank = yac_get_comm_rank(comm);
  unsigned count = (comm_size * (comm_size + 1)) / 2;
  unsigned * buffer = (unsigned*)(message.buffer);

  if (count != message.count)
    PUT_ERR("error in message.count\n");

  if (message.type != COMM_UINT)
    PUT_ERR("error in message.type\n");

  for (unsigned i = 0; i < count; ++i)
    if (buffer[i] != i)
      PUT_ERR("error in message.buffer\n");

  if (message.src_rank != comm_rank)
    PUT_ERR("error in message.src_rank\n");

  ((unsigned*)orig_data)[0]++;
}

void alltoallv_rand_callback(struct communicator * comm,
                             struct comm_message message, void * orig_data) {

  unsigned ref_count = ((unsigned*)orig_data)[1];
  unsigned * ref_data = (unsigned*)orig_data + 2;

  unsigned comm_rank = yac_get_comm_rank(comm);
  unsigned * buffer = (unsigned*)(message.buffer);

  if (ref_count != message.count)
    PUT_ERR("error in message.count\n");

  if (message.type != COMM_UINT)
    PUT_ERR("error in message.type\n");

  for (unsigned i = 0; i < ref_count; ++i)
    if (buffer[i] != ref_data[i])
      PUT_ERR("error in message.buffer\n");

  if (message.src_rank != comm_rank)
    PUT_ERR("error in message.src_rank\n");

  ((unsigned*)orig_data)[0]++;
}

void alltoallv_char_callback(struct communicator * comm,
                             struct comm_message message, void * orig_data) {

  unsigned comm_size = yac_get_comm_size(comm);
  unsigned comm_rank = yac_get_comm_rank(comm);
  char * buffer = (char*)(message.buffer);

  if (comm_size != message.count)
    PUT_ERR("error in message.count\n");

  if (message.type != COMM_BYTE)
    PUT_ERR("error in message.type\n");

  for (unsigned i = 0; i < comm_size; ++i)
    if (buffer[i] != (i%256))
      PUT_ERR("error in message.buffer\n");

  if (message.src_rank != comm_rank)
    PUT_ERR("error in message.src_rank\n");

  ((unsigned*)orig_data)[0]++;
}

void reduce_callback(struct communicator * comm,
                     struct comm_message message, void * orig_data) {

  unsigned comm_rank = yac_get_comm_rank(comm);
  unsigned count = ((unsigned*)(orig_data))[1];
  unsigned root = ((unsigned*)(orig_data))[2];
  unsigned * ref_buffer = (unsigned*)(orig_data) + 3;
  unsigned * buffer = (unsigned*)(message.buffer);

  if (comm_rank == root) {

    if (count != message.count)
      PUT_ERR("error in message.count\n");

    if (message.type != COMM_UINT)
      PUT_ERR("error in message.type\n");

    for (unsigned i = 0; i < count; ++i)
      if (buffer[i] != ref_buffer[i])
        PUT_ERR("error in message.buffer\n");
  }

  if (message.src_rank != comm_rank)
    PUT_ERR("error in message.src_rank\n");

  ((unsigned*)orig_data)[0]++;
}

void allreduce_callback(struct communicator * comm,
                        struct comm_message message, void * orig_data) {

  unsigned comm_rank = yac_get_comm_rank(comm);
  unsigned count = ((unsigned*)(orig_data))[1];
  unsigned * ref_buffer = (unsigned*)(orig_data) + 2;
  unsigned * buffer = (unsigned*)(message.buffer);

  if (count != message.count)
    PUT_ERR("error in message.count\n");

  if (message.type != COMM_UINT)
    PUT_ERR("error in message.type\n");

  for (unsigned i = 0; i < count; ++i)
    if (buffer[i] != ref_buffer[i])
      PUT_ERR("error in message.buffer\n");

  if (message.src_rank != comm_rank)
    PUT_ERR("error in message.src_rank\n");

  ((unsigned*)orig_data)[0]++;
}

void barrier_callback(struct communicator * comm, struct comm_message message,
                      void * orig_data) {

  if (0 != message.count)
    PUT_ERR("error in message.count\n");

  if (message.src_rank != yac_get_comm_rank(comm))
    PUT_ERR("error in message.src_rank\n");

  ++*((unsigned*)orig_data);
}

#else
int main() {

   // no mpi
   return EXIT_SKIP_TEST;
}
#endif
