/**
 * @file test_compute_overlap_area.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include "tests.h"
#include "clipping.h"
#include "geometry.h"

//#define VERBOSE

int main (void) {

  enum yac_edge_type edge_types[5] = {GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE,
                                      GREAT_CIRCLE, GREAT_CIRCLE};

  struct grid_cell Pentagon = {.coordinates_x = NULL,
                               .coordinates_y = NULL,
                               .coordinates_xyz = (double[]){-1,-1,-1,-1,-1,
                                                             -1,-1,-1,-1,-1,
                                                             -1,-1,-1,-1,-1},
                               .edge_type = edge_types,
                               .num_corners = 5};

  struct grid_cell Triangle = {.coordinates_x = NULL,
                               .coordinates_y = NULL,
                               .coordinates_xyz = (double[]){-1,-1,-1,
                                                             -1,-1,-1,
                                                             -1,-1,-1},
                               .edge_type = edge_types,
                               .num_corners = 3};

  double Pentagon_node_x, Pentagon_node_y;
  double Triangle_node_x, Triangle_node_y;

  double ref_area, area;

  static double const area_tolerance = 1.0e-10;

#ifdef VERBOSE
  puts ("----- calculating area -----\n");
#endif

  {
    /* set the test cells concave pentagon with a triangle


           10.0 (lon)   11.0 (lon)
           32.0 (lat)   32.0 (lat)
                  x--------x
                 /          \
                /    10.5    \  
               /     31.0     \  
   9.5 (lon)  /       x        \   11.5
  30.0 (lat) x                  x  30.0 


   with


                    10.5 (lon)
                    31.5 (lat)
                      x
                     / \
                    /   \
           10.0    /     \   11.0
           29.5   x-------x  29.5


    */

    {
      double coordinates_x[5] = {10.0*rad, 11.0*rad, 11.5*rad, 10.5*rad,  9.5*rad};
      double coordinates_y[5] = {32.0*rad, 32.0*rad, 30.0*rad, 31.0*rad, 30.0*rad};

       Pentagon.coordinates_x = coordinates_x;
       Pentagon.coordinates_y = coordinates_y;

       for (int i = 0; i < Pentagon.num_corners; ++i)
        LLtoXYZ(Pentagon.coordinates_x[i], Pentagon.coordinates_y[i],
                Pentagon.coordinates_xyz+i*3);
    }

    Pentagon_node_x = 10.5*rad;
    Pentagon_node_y = 31.5*rad;

    {
      double coordinates_x[3] = {10.5*rad, 11.0*rad, 10.0*rad};
      double coordinates_y[3] = {31.5*rad, 29.5*rad, 29.5*rad};

       Triangle.coordinates_x = coordinates_x;
       Triangle.coordinates_y = coordinates_y;

       for (int i = 0; i < Triangle.num_corners; ++i)
        LLtoXYZ(Triangle.coordinates_x[i], Triangle.coordinates_y[i],
                Triangle.coordinates_xyz+i*3);
    }

    Triangle_node_x = 10.5*rad;
    Triangle_node_y = 30.0*rad;

    unsigned N = 1;

    // the reference solution for the area
    yac_compute_overlap_areas( N, &Pentagon, Triangle, &ref_area );

    // this should give a bit identical results as we do not perform any triangulation
    yac_compute_concave_overlap_areas ( N, &Pentagon, Triangle, &Triangle_node_x, &Triangle_node_y, &area );
    if ( ref_area != area )
      PUT_ERR("error compute_concave_overlap_areas, bit identical results expected.\n");

    // with triangulation of the pentagon the result should not differ that much
    yac_compute_concave_overlap_areas ( N, &Triangle, Pentagon, &Pentagon_node_x, &Pentagon_node_y, &area );
    if ( fabs(ref_area-area) > area_tolerance )
      PUT_ERR("error compute_concave_overlap_areas, area difference is too large.\n");

  }
  return TEST_EXIT_CODE;
}
