/**
 * @file test_dep_list.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include "tests.h"
#include "dep_list.h"

int main (void) {

   //---------------
   // setup
   //---------------

   struct dep_list list;

   unsigned num_elements = 4;
   unsigned num_deps_per_element[4] = {1,2,2,4};
   unsigned dependencies[9] = {0, 0,1, 0,2, 0,1,2,3};

   unsigned i, j;

   // test initialisation of dependency list
   yac_init_dep_list(&list);

   yac_set_dependencies(&list, num_elements, num_deps_per_element, dependencies);

   //---------------
   // testing
   //---------------

   // test inversion of dependency list
   {
      struct dep_list inv_list;

      // reference data
      unsigned ref_num_elements = 4;
      unsigned ref_num_deps_per_element[4] = {4,2,2,1};
      unsigned ref_dependencies[9] = {0,1,2,3, 1,3, 2,3, 3};

      yac_invert_dep_list(list, &inv_list);

      if (inv_list.num_elements != ref_num_elements)
         PUT_ERR("wrong num_elements after inversion\n")

      for (i = 0; i < 4; ++i)
         if (inv_list.num_deps_per_element[i] != ref_num_deps_per_element[i])
            PUT_ERR("wrong num_deps_per_element after inversion\n")

      for (i = 0; i < 9; ++i)
         if (inv_list.dependencies[i] != ref_dependencies[i])
            PUT_ERR("wrong dependencies after inversion\n")

      // test generation of prescan of inverted list
      {
         unsigned ref_prescan[4] = {0, 4, 6, 8};

         for (i = 0; i < 4; ++i)
            if (inv_list.prescan[i] != ref_prescan[i])
               PUT_ERR("error in prescan generation\n")
      }

      yac_free_dep_list(&inv_list);
   }

   // test dependency index <--> index + element index conversion
   {

      unsigned ref_idx_element_pair[9][2] = {{0,0}, {1,0},{1,1}, {2,0},{2,2},
                                             {3,0},{3,1},{3,2},{3,3}};
      unsigned index, element;

      for (i = 0; i < 9; ++i) {

         if (yac_get_dependency_index(list, ref_idx_element_pair[i][0], ref_idx_element_pair[i][1])
             != i) PUT_ERR("error get_dependency_index\n")

         yac_get_dependency (list, i, &index, &element);

         if (index != ref_idx_element_pair[i][0] || element != ref_idx_element_pair[i][1])
            PUT_ERR("error get_dependency\n")
      }
   }

   { // test remove_dependencies_of_elements
      struct dep_list temp_list;
      unsigned element_to_remove = 2;

      yac_init_dep_list(&temp_list);
      yac_copy_dep_list(list, &temp_list);

      yac_remove_dependencies_of_elements(&temp_list, &element_to_remove, 1);

      unsigned ref_num_deps_per_element[4] = {1,2,0,4};
      unsigned ref_dependencies[7] = {0, 0,1, 0,1,2,3};

      if (temp_list.num_elements != 4)
         PUT_ERR("wrong num_elements after removing element\n")

      for (i = 0; i < 4; ++i)
         if (temp_list.num_deps_per_element[i] != ref_num_deps_per_element[i])
            PUT_ERR("wrong num_deps_per_element after removing element\n")

      for (i = 0; i < 7; ++i)
         if (temp_list.dependencies[i] != ref_dependencies[i])
            PUT_ERR("wrong dependencies after removing element\n")

      yac_free_dep_list(&temp_list);
   }

   { // test remove_dependencies
      struct dep_list temp_list;
      unsigned index_to_remove = 0;

      yac_init_dep_list(&temp_list);
      yac_copy_dep_list(list, &temp_list);

      yac_remove_dependencies(&temp_list, &index_to_remove, 1);

      unsigned ref_num_deps_per_element[4] = {0,1,1,3};
      unsigned ref_dependencies[5] = {1, 2, 1,2,3};

      if (temp_list.num_elements != 4)
         PUT_ERR("wrong num_elements after removing indices\n")

      for (i = 0; i < 4; ++i)
         if (temp_list.num_deps_per_element[i] != ref_num_deps_per_element[i])
            PUT_ERR("wrong num_deps_per_element after removing indices\n")

      for (i = 0; i < 5; ++i)
         if (temp_list.dependencies[i] != ref_dependencies[i])
            PUT_ERR("wrong dependencies after removing indices\n")

      yac_free_dep_list(&temp_list);
   }

   { // test add_dependencies
      struct dep_list temp_list;

      yac_init_dep_list(&temp_list);
      yac_copy_dep_list(list, &temp_list);

      unsigned new_dependencies[2] = {2,3};

      yac_add_dependencies(&temp_list, 1, 2, new_dependencies);

      unsigned ref_num_deps_per_element[4] = {1,4,2,4};
      unsigned ref_dependencies[11] = {0, 0,1,2,3, 0,2, 0,1,2,3};

      if (temp_list.num_elements != 4)
         PUT_ERR("wrong num_elements after removing adding dependencies\n")

      if (yac_get_total_num_dependencies(temp_list) != 11)
         PUT_ERR("wrong total number of dependencies after adding dependencies\n")

      for (i = 0; i < 4; ++i)
         if (temp_list.num_deps_per_element[i] != ref_num_deps_per_element[i])
            PUT_ERR("wrong num_deps_per_element after adding dependencies\n")

      for (i = 0; i < 11; ++i)
         if (temp_list.dependencies[i] != ref_dependencies[i])
            PUT_ERR("wrong dependencies after removing adding dependencies\n")

      yac_free_dep_list(&temp_list);
   }

   { // test add_dependencies
      struct dep_list temp_list;

      yac_init_dep_list(&temp_list);
      yac_copy_dep_list(list, &temp_list);

      unsigned new_dependencies[2][2] = {{1,2},{4,5}};

      yac_add_dependencies(&temp_list, 0, 2, new_dependencies[0]);
      yac_add_dependencies(&temp_list, 3, 2, new_dependencies[1]);

      unsigned ref_num_deps_per_element[4] = {3,2,2,6};
      unsigned ref_dependencies[13] = {0,1,2, 0,1, 0,2, 0,1,2,3,4,5};

      if (temp_list.num_elements != 4)
         PUT_ERR("wrong num_elements after removing adding dependencies\n")

      if (yac_get_total_num_dependencies(temp_list) != 13)
         PUT_ERR("wrong total number of dependencies after adding dependencies\n")

      for (i = 0; i < 4; ++i)
         if (temp_list.num_deps_per_element[i] != ref_num_deps_per_element[i])
            PUT_ERR("wrong num_deps_per_element after adding dependencies\n")

      for (i = 0; i < 1; ++i)
         if (temp_list.dependencies[i] != ref_dependencies[i])
            PUT_ERR("wrong dependencies after removing adding dependencies\n")

      yac_free_dep_list(&temp_list);
   }

   { // test packing and unpacking of dep_list
      struct dep_list unpacked_list[2];

      unsigned * buf, data_size, buf_size;

      buf = NULL, buf_size = 0, data_size = 0;

      for (i = 0; i < 2; ++i) {
         yac_pack_dep_list(list, &buf, data_size, &data_size, &buf_size);

         if (data_size != 14)
            PUT_ERR("wrong data_size returned by pack_dep_list\n")
      }

      data_size = 0;

      for (i = 0; i < 2; ++i) {
         yac_unpack_dep_list(unpacked_list+i, buf+data_size, &data_size);

         if (data_size != 14)
            PUT_ERR("wrong data_size returned by pack_dep_list\n")

         if (unpacked_list[i].num_elements != num_elements)
            PUT_ERR("wrong num_elements after unpacking\n")

         for (j = 0; j < 4; ++j)
            if (unpacked_list[i].num_deps_per_element[j] != num_deps_per_element[j])
               PUT_ERR("wrong num_deps_per_element after unpacking\n")

         for (j = 0; j < 9; ++j)
            if (unpacked_list[i].dependencies[j] != dependencies[j])
               PUT_ERR("wrong dependencies after unpacking\n")
      }

      //---------------
      // cleanup
      //---------------

      yac_free_dep_list(unpacked_list+0);
      yac_free_dep_list(unpacked_list+1);
      free(buf);
   }

   { // test packing and unpacking of an empty dep_list

      struct dep_list empty_list;

      yac_init_dep_list(&empty_list);

      struct dep_list unpacked_list[2];

      unsigned * buf, data_size, buf_size;

      buf = NULL, buf_size = 0, data_size = 0;

      for (i = 0; i < 2; ++i) {
         yac_pack_dep_list(empty_list, &buf, data_size, &data_size, &buf_size);

         if (data_size != 1)
            PUT_ERR("wrong data_size returned by pack_dep_list\n")
      }

      data_size = 0;

      for (i = 0; i < 2; ++i) {
         yac_unpack_dep_list(unpacked_list+i, buf+data_size, &data_size);

         if (data_size != 1)
            PUT_ERR("wrong data_size returned by pack_dep_list\n")

         if (unpacked_list[i].num_elements != 0)
            PUT_ERR("wrong num_elements after unpacking\n")

         if (unpacked_list[i].num_deps_per_element != NULL)
            PUT_ERR("wrong num_deps_per_element after unpacking\n")

         if (unpacked_list[i].dependencies != NULL)
            PUT_ERR("wrong dependencies after unpacking\n")
      }

      //---------------
      // cleanup
      //---------------

      yac_free_dep_list(unpacked_list+0);
      yac_free_dep_list(unpacked_list+1);
      free(buf);
   }

   return TEST_EXIT_CODE;
}
