/**
 * @file test_domain.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include <stdlib.h>
#include <stdio.h>

#include "tests.h"

#include "grid.h"
#include "grid_reg2d.h"
#include "communicator.h"
#include "communicator_local.h"
#include "domain.h"

int main (void) {

   { // tests with a single "process"

      struct communicator * comm;

      comm = yac_local_comm_new();

      { // tests setting up of a domain with only a single subdomain

         //---------------
         // setup
         //---------------
         struct grid * grid;
         unsigned num_cells[2] = {3,3};
         unsigned cyclic[2] = {0,0};
         double coordinates_x[4] = {0*rad,1*rad,2*rad,3*rad};
         double coordinates_y[4] = {0*rad,1*rad,2*rad,3*rad};

         grid = yac_reg2d_grid_new(coordinates_x, coordinates_y, num_cells,
                                   cyclic);

         struct subdomain subdomain;

         yac_init_subdomain_unstruct(&subdomain, grid, comm, NULL, NULL,
                                 NULL, NULL, NULL, NULL);

         struct domain domain;

         struct subdomain * subdomains[1] = {&subdomain};

         //---------------
         // testing
         //---------------
         yac_init_domain(&domain, comm, subdomains, 1, "dummy");
         yac_enddef_domain(&domain);

         //---------------
         // cleanup
         //---------------
         yac_free_domain(&domain);
         yac_delete_grid(grid);
      }

      { // tests a domain (4x3) that is distributed among two subdomain on the
        // same process:
        // corner, cell and edge ids for a 4 x 3 grid
        // 15--27--16--28--17--29--18--30--19
        //  |       |       |       |       |
        // 19  08  21  09  23  10  25  11  26
        //  |       |       |       |       |
        // 10--18--11--20--12--22--13--24--14
        //  |       |       |       |       |
        // 10  04  12  05  14  06  16  07  17
        //  |       |       |       |       |
        // 05--09--06--11--07--13--08--15--09
        //  |       |       |       |       |
        // 01  00  03  01  05  02  07  03  08
        //  |       |       |       |       |
        // 00--00--01--02--02--04--03--06--04
        // the domain is distributed among the subdomains as follows:
        // (index == subdomain id)
        //
        // 0--0--0--0--0--1--1--1--1
        // |     |     |     |     |
        // 0  0  0  0  0  1  1  1  1
        // |     |     |     |     |
        // 0--0--0--0--0--1--1--1--1
        // |     |     |     |     |
        // 0  0  0  0  0  1  1  1  1
        // |     |     |     |     |
        // 0--0--0--0--0--1--1--1--1
        // |     |     |     |     |
        // 0  0  0  0  0  1  1  1  1
        // |     |     |     |     |
        // 0--0--0--0--0--1--1--1--1

         //---------------
         // setup
         //---------------
         struct grid * grid_a, * grid_b;
         unsigned num_cells[2] = {3,3};
         unsigned cyclic[2] = {0,0};
         double coordinates_x[2][4] = {{0*rad,1*rad,2*rad,3*rad},
                                       {1*rad,2*rad,3*rad,4*rad}};
         double coordinates_y[4] = {0*rad,1*rad,2*rad,3*rad};

         grid_a = yac_reg2d_grid_new(coordinates_x[0], coordinates_y, num_cells,
                                     cyclic);
         grid_b = yac_reg2d_grid_new(coordinates_x[1], coordinates_y, num_cells,
                                     cyclic);

         struct subdomain subdomain_a;

         int global_corner_ids_a[16] = {0,1,2,3, 5,6,7,8, 10,11,12,13, 15,16,17,18};
         unsigned corner_is_on_rank_a[16] = {-1,-1,-1,0, -1,-1,-1,0, -1,-1,-1,0, -1,-1,-1,0};
         int global_cell_ids_a[9] = {0,1,2, 4,5,6, 8,9,10};
         unsigned cell_is_on_rank_a[9] = {-1,-1,0, -1,-1,0, -1,-1,0};
         int global_edge_ids_a[24] = {0,1,2,3,4,5,7,9,10,11,12,13,14,16,18,19,20,21,22,23,25,27,28,29};
         unsigned edge_is_on_rank_a[24] = {-1,-1,-1,-1,0,-1,0,-1,-1,-1,-1,0,-1,0,-1,-1,-1,-1,0,-1,0,-1,-1,0};

         yac_init_subdomain_unstruct(&subdomain_a, grid_a, comm,
                                     global_corner_ids_a, corner_is_on_rank_a,
                                     global_cell_ids_a, cell_is_on_rank_a,
                                     global_edge_ids_a, edge_is_on_rank_a);

         struct subdomain subdomain_b;

         int global_corner_ids_b[16] = {1,2,3,4, 6,7,8,9, 11,12,13,14, 16,17,18,19};
         unsigned corner_is_on_rank_b[16] = {0,0,-1,-1, 0,0,-1,-1, 0,0,-1,-1, 0,0,-1,-1};
         int global_cell_ids_b[9] = {1,2,3, 5,6,7, 9,10,11};
         unsigned cell_is_on_rank_b[9] = {0,-1,-1, 0,-1,-1, 0,-1,-1};
         int global_edge_ids_b[24] = {2,3,4,5,6,7,8,11,12,13,14,15,16,17,20,21,22,23,24,25,26,28,29,30};
         unsigned edge_is_on_rank_b[24] = {0,0,-1,0,-1,-1,-1,0,0,-1,0,-1,-1,-1,0,0,-1,0,-1,-1,-1,0,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_b, grid_b, comm,
                                     global_corner_ids_b, corner_is_on_rank_b,
                                     global_cell_ids_b, cell_is_on_rank_b,
                                     global_edge_ids_b, edge_is_on_rank_b);

         struct subdomain * subdomains[2];

         subdomains[0] = &subdomain_a;
         subdomains[1] = &subdomain_b;

         struct domain domain;

         yac_init_domain(&domain, comm, subdomains, 2, "dummy");
         yac_enddef_domain(&domain);

         //---------------
         // testing
         //---------------

         // is_halo_point test
         {
            { // corner points
               unsigned i;

               for (i = 0; i < 16; ++i) {

                  unsigned a, b;

                  a = (yac_is_halo_value(&domain, 0, CORNER, i, NULL))?1:0;
                  b = (corner_is_on_rank_a[i] == (unsigned)-1)?0:1;

                  if (a^b) PUT_ERR("error in is_halo_value corner\n");

                  a = (yac_is_halo_value(&domain, 1, CORNER, i, NULL))?1:0;
                  b = (corner_is_on_rank_b[i] == (unsigned)-1)?0:1;

                  if (a^b) PUT_ERR("error in is_halo_value corner\n");
               }
            }

            { // cell points
               unsigned i;

               for (i = 0; i < 9; ++i) {

                  unsigned a, b;

                  a = (yac_is_halo_value(&domain, 0, CELL, i, NULL))?1:0;
                  b = (cell_is_on_rank_a[i] == (unsigned)-1)?0:1;

                  if (a^b) PUT_ERR("error in is_halo_value cell\n");

                  a = (yac_is_halo_value(&domain, 1, CELL, i, NULL))?1:0;
                  b = (cell_is_on_rank_b[i] == (unsigned)-1)?0:1;

                  if (a^b) PUT_ERR("error in is_halo_value cell\n");
               }
            }

            { // edge points
               unsigned i;

               for (i = 0; i < 16; ++i) {

                  unsigned a, b;

                  a = (yac_is_halo_value(&domain, 0, EDGE, i, NULL))?1:0;
                  b = (edge_is_on_rank_a[i] == (unsigned)-1)?0:1;

                  if (a^b) PUT_ERR("error in is_halo_value edge\n");

                  a = (yac_is_halo_value(&domain, 1, EDGE, i, NULL))?1:0;
                  b = (edge_is_on_rank_b[i] == (unsigned)-1)?0:1;

                  if (a^b) PUT_ERR("error in is_halo_value edge\n");
               }
            }
         }

         { // yac_get_halo_points and yac_get_non_halo_points test

            { // corner
               int tmp_global_ids[2][16];
               unsigned tmp_local_ids[16];
               unsigned num_points;

               yac_get_halo_points(&domain, 0, CORNER, tmp_global_ids[0],
                                   tmp_local_ids, &num_points, 0);
               yac_get_non_halo_points(&domain, 0, CORNER, tmp_global_ids[1],
                                       NULL);

               for (unsigned i = 0, j = 0, k = 0; i < 16; ++i) {

                  if (corner_is_on_rank_a[i] != -1) {

                     if (tmp_global_ids[0][j] != global_corner_ids_a[i])
                        PUT_ERR("error in get_halo_points global corner a\n");
                     if (tmp_local_ids[j++] != i)
                        PUT_ERR("error in get_halo_points local corner a\n");
                  } else if (tmp_global_ids[1][k++] != global_corner_ids_a[i])
                      PUT_ERR("error in get_non_halo_points global corner a\n");
               }

               yac_get_halo_points(&domain, 1, CORNER, tmp_global_ids[0],
                                   tmp_local_ids, &num_points, 0);
               yac_get_non_halo_points(&domain, 1, CORNER, tmp_global_ids[1],
                                       NULL);

               for (unsigned i = 0, j = 0, k = 0; i < 16; ++i) {

                  if (corner_is_on_rank_b[i] != -1) {

                     if (tmp_global_ids[0][j] != global_corner_ids_b[i])
                        PUT_ERR("error in get_halo_points global corner b\n");
                     if (tmp_local_ids[j++] != i)
                        PUT_ERR("error in get_halo_points local corner b\n");
                  } else if (tmp_global_ids[1][k++] != global_corner_ids_b[i])
                      PUT_ERR("error in get_non_halo_points global corner b\n");
               }
            }

            { // cell
               int tmp_global_ids[2][9];
               unsigned tmp_local_ids[9];
               unsigned num_points;

               yac_get_halo_points(&domain, 0, CELL, tmp_global_ids[0],
                                   tmp_local_ids, &num_points, 0);
               yac_get_non_halo_points(&domain, 0, CELL, tmp_global_ids[1],
                                       NULL);

               for (unsigned i = 0, j = 0, k = 0; i < 9; ++i) {

                  if (cell_is_on_rank_a[i] != -1) {

                     if (tmp_global_ids[0][j] != global_cell_ids_a[i])
                        PUT_ERR("error in get_halo_points global cell a\n");
                     if (tmp_local_ids[j++] != i)
                        PUT_ERR("error in get_halo_points local cell a\n");
                  } else if (tmp_global_ids[1][k++] != global_cell_ids_a[i])
                      PUT_ERR("error in get_non_halo_points global cell a\n");
               }

               yac_get_halo_points(&domain, 1, CELL, tmp_global_ids[0],
                                   tmp_local_ids, &num_points, 0);
               yac_get_non_halo_points(&domain, 1, CELL, tmp_global_ids[1],
                                       NULL);

               for (unsigned i = 0, j = 0, k = 0; i < 9; ++i) {

                  if (cell_is_on_rank_b[i] != -1) {

                     if (tmp_global_ids[0][j] != global_cell_ids_b[i])
                        PUT_ERR("error in get_halo_points global cell b\n");
                     if (tmp_local_ids[j++] != i)
                        PUT_ERR("error in get_halo_points local cell b\n");
                  } else if (tmp_global_ids[1][k++] != global_cell_ids_b[i])
                      PUT_ERR("error in get_non_halo_points global cell b\n");
               }
            }

            { // edge
               int tmp_global_ids[2][24];
               unsigned tmp_local_ids[24];
               unsigned num_points;

               yac_get_halo_points(&domain, 0, EDGE, tmp_global_ids[0],
                                   tmp_local_ids, &num_points, 0);
               yac_get_non_halo_points(&domain, 0, EDGE, tmp_global_ids[1],
                                       NULL);

               for (unsigned i = 0, j = 0, k = 0; i < 24; ++i) {

                  if (edge_is_on_rank_a[i] != -1) {

                     if (tmp_global_ids[0][j] != global_edge_ids_a[i])
                        PUT_ERR("error in get_halo_points global edge a\n");
                     if (tmp_local_ids[j++] != i)
                        PUT_ERR("error in get_halo_points local edge a\n");
                  } else if (tmp_global_ids[1][k++] != global_edge_ids_a[i])
                      PUT_ERR("error in get_non_halo_points global edge a\n");
               }

               yac_get_halo_points(&domain, 1, EDGE, tmp_global_ids[0],
                                   tmp_local_ids, &num_points, 0);
               yac_get_non_halo_points(&domain, 1, EDGE, tmp_global_ids[1],
                                       NULL);

               for (unsigned i = 0, j = 0, k = 0; i < 24; ++i) {

                  if (edge_is_on_rank_b[i] != -1) {

                     if (tmp_global_ids[0][j] != global_edge_ids_b[i])
                        PUT_ERR("error in get_halo_points global edge b\n");
                     if (tmp_local_ids[j++] != i)
                        PUT_ERR("error in get_halo_points local edge b\n");
                  } else if (tmp_global_ids[1][k++] != global_edge_ids_b[i])
                      PUT_ERR("error in get_non_halo_points global edge b\n");
               }
            }
         }

         { // yac_get_domain_neighbours

            { // corner
               unsigned * ranks, num_ranks;

               yac_get_domain_neighbours(domain, CORNER, &ranks, &num_ranks);

               if (num_ranks != 0)
                  PUT_ERR("error in yac_get_domain_neighbours num_ranks corner\n");
               if (ranks != NULL)
                  PUT_ERR("error in yac_get_domain_neighbours ranks corner\n");
            }

            { // cell
               unsigned * ranks, num_ranks;

               yac_get_domain_neighbours(domain, CELL, &ranks, &num_ranks);

               if (num_ranks != 0)
                  PUT_ERR("error in yac_get_domain_neighbours num_ranks cell\n");
               if (ranks != NULL)
                  PUT_ERR("error in yac_get_domain_neighbours ranks cell\n");
            }

            { // edge
               unsigned * ranks, num_ranks;

               yac_get_domain_neighbours(domain, EDGE, &ranks, &num_ranks);

               if (num_ranks != 0)
                  PUT_ERR("error in yac_get_domain_neighbours num_ranks edge\n");
               if (ranks != NULL)
                  PUT_ERR("error in yac_get_domain_neighbours ranks edge\n");
            }
         }

         //---------------
         // cleanup
         //---------------

         yac_free_domain(&domain);
         yac_delete_grid(grid_a);
         yac_delete_grid(grid_b);
      }
      
      yac_free_communicator(comm);
   }

   { // tests with a four "process"

      // the global domain is a 7x7 grid (4 processes):
      //
      // 56-----57-----58-----59-----60-----61-----62-----63
      //  |      |      |      |      |      |      |      |
      //  |  42  |  43  |  44  |  45  |  46  |  47  |  48  |
      //  |      |      |      |      |      |      |      |
      // 48-----49-----50-----51-----52-----53-----54-----55
      //  |      |      |      |      |      |      |      |
      //  |  35  |  36  |  37  |  38  |  39  |  40  |  41  |
      //  |      |      |      |      |      |      |      |
      // 40-----41-----42-----43-----44-----45-----46-----47
      //  |      |      |      |      |      |      |      |
      //  |  28  |  29  |  30  |  31  |  32  |  33  |  34  |
      //  |      |      |      |      |      |      |      |
      // 32-----33-----34-----35-----36-----37-----38-----39
      //  |      |      |      |      |      |      |      |
      //  |  21  |  22  |  23  |  24  |  25  |  26  |  27  |
      //  |      |      |      |      |      |      |      |
      // 24-----25-----26-----27-----28-----29-----30-----31
      //  |      |      |      |      |      |      |      |
      //  |  14  |  15  |  16  |  17  |  18  |  19  |  20  |
      //  |      |      |      |      |      |      |      |
      // 16-----17-----18-----19-----20-----21-----22-----23
      //  |      |      |      |      |      |      |      |
      //  |  07  |  08  |  09  |  10  |  11  |  12  |  13  |
      //  |      |      |      |      |      |      |      |
      // 08-----09-----10-----11-----12-----13-----14-----15
      //  |      |      |      |      |      |      |      |
      //  |  00  |  01  |  02  |  03  |  04  |  05  |  06  |
      //  |      |      |      |      |      |      |      |
      // 00-----01-----02-----03-----04-----05-----06-----07
      //
      // the domain is distributed among the processes as follows:
      // (index == process)
      //
      // 3---3---3---3---3---3---3---3
      // | 3 | 3 | 3 | 3 | 3 | 3 | 3 |
      // 3---3---3---3---3---3---3---3
      // | 3 | 3 | 3 | 3 | 3 | 3 | 3 |
      // 3---3---3---1---2---2---3---3
      // | 1 | 1 | 1 | 2 | 2 | 2 | 2 |
      // 1---1---1---2---2---2---2---2
      // | 1 | 1 | 1 | 2 | 2 | 2 | 2 |
      // 1---1---1---1---2---2---2---2
      // | 1 | 1 | 1 | 2 | 2 | 2 | 2 |
      // 1---1---1---0---0---0---2---2
      // | 0 | 0 | 0 | 0 | 0 | 0 | 0 |
      // 0---0---0---0---0---0---0---0
      // | 0 | 0 | 0 | 0 | 0 | 0 | 0 |
      // 0---0---0---0---0---0---0---0

#define NPROC 4
      struct communicator * comm[NPROC];

      for (int i = 0; i < NPROC; ++i)
         comm[i] = yac_local_comm_new();

      //---------------
      // setup
      //---------------

      struct grid * grid[NPROC];
      unsigned num_cells[NPROC][2] = {{7,3},{4,5},{5,5},{7,3}};
      unsigned cyclic[2] = {0,0};
      double coordinates_x[NPROC][8] = {{0,1,2,3,4,5,6,7},{0,1,2,3,4},
                                        {2,3,4,5,6,7},{0,1,2,3,4,5,6,7}};
      double coordinates_y[NPROC][8] = {{0,1,2,3},{1,2,3,4,5,6},
                                        {1,2,3,4,5,6},{4,5,6,7}};

      for ( unsigned j = 0; j < NPROC; ++j ) {
         for ( unsigned i = 0; i < 8; ++i ) {
            coordinates_x[j][i] *= rad;
            coordinates_y[j][i] *= rad;
         }
      }

      for (int i = 0; i < NPROC; ++i)
         grid[i] = yac_reg2d_grid_new(coordinates_x[i],
                                      coordinates_y[i],
                                      num_cells[i], cyclic);

      struct subdomain subdomain[NPROC];

      int global_corner_ids[NPROC][36] = {{ 0, 1, 2, 3, 4, 5, 6, 7,
                                            8, 9,10,11,12,13,14,15,
                                           16,17,18,19,20,21,22,23,
                                           24,25,26,27,28,29,30,31},
                                          { 8, 9,10,11,12,
                                           16,17,18,19,20,
                                           24,25,26,27,28,
                                           32,33,34,35,36,
                                           40,41,42,43,44,
                                           48,49,50,51,52},
                                          {10,11,12,13,14,15,
                                           18,19,20,21,22,23,
                                           26,27,28,29,30,31,
                                           34,35,36,37,38,39,
                                           42,43,44,45,46,47,
                                           50,51,52,53,54,55},
                                          {32,33,34,35,36,37,38,39,
                                           40,41,42,43,44,45,46,47,
                                           48,49,50,51,52,53,54,55,
                                           56,57,58,59,60,61,62,63}};
      unsigned corner_is_on_rank[NPROC][36] = {{-1,-1,-1,-1,-1,-1,-1,-1,
                                                -1,-1,-1,-1,-1,-1,-1,-1,
                                                 1, 1, 1,-1,-1,-1, 2, 2},
                                               { 0, 0, 0, 0, 0,
                                                -1,-1,-1, 0, 0,
                                                -1,-1,-1,-1, 2,
                                                -1,-1,-1, 2, 2,
                                                 3, 3, 3,-1, 2,
                                                 3, 3, 3, 3, 3},
                                               { 0, 0, 0, 0, 0, 0,
                                                 1, 0, 0, 0,-1,-1,
                                                 1, 1,-1,-1,-1,-1,
                                                 1,-1,-1,-1,-1,-1,
                                                 3, 1,-1,-1, 3, 3,
                                                 3, 3, 3, 3, 3, 3},
                                               { 1, 1, 1, 2, 2, 2, 2, 2,
                                                -1,-1,-1, 1, 2, 2,-1,-1,
                                                -1,-1,-1,-1,-1,-1,-1,-1,
                                                -1,-1,-1,-1,-1,-1,-1,-1}};
      int global_cell_ids[NPROC][25] = {{ 0, 1, 2, 3, 4, 5, 6,
                                          7, 8, 9,10,11,12,13,
                                         14,15,16,17,18,19,20},
                                        { 7, 8, 9,10,
                                         14,15,16,17,
                                         21,22,23,24,
                                         28,29,30,31,
                                         35,36,37,38},
                                        { 9,10,11,12,13,
                                         16,17,18,19,20,
                                         23,24,25,26,27,
                                         30,31,32,33,34,
                                         37,38,39,40,41},
                                        {28,29,30,31,32,33,34,
                                         35,36,37,38,39,40,41,
                                         42,43,44,45,46,47,48}};
      unsigned cell_is_on_rank[NPROC][25] = {{-1,-1,-1,-1,-1,-1,-1,
                                              -1,-1,-1,-1,-1,-1,-1,
                                               1, 1, 1, 2, 2, 2, 2},
                                             { 0, 0, 0, 0,
                                              -1,-1,-1, 2,
                                              -1,-1,-1, 2,
                                              -1,-1,-1, 2,
                                               3, 3, 3, 3},
                                             { 0, 0, 0, 0, 0,
                                               1,-1,-1,-1,-1,
                                               1,-1,-1,-1,-1,
                                               1,-1,-1,-1,-1,
                                               3, 3, 3, 3, 3},
                                             { 1, 1, 1, 2, 2, 2, 2,
                                               -1, -1, -1, -1, -1, -1, -1,
                                               -1, -1, -1, -1, -1, -1, -1}};

      for (int i = 0; i < NPROC; ++i)
         yac_init_subdomain_unstruct(&subdomain[i], grid[i], comm[i],
                                     global_corner_ids[i], corner_is_on_rank[i],
                                     global_cell_ids[i], cell_is_on_rank[i],
                                     NULL, NULL);

      struct domain l_domain[NPROC];

      struct subdomain * blocks[NPROC];

      for (unsigned i = 0; i < NPROC; ++i)
         blocks[i] = &subdomain[i];

      for (int i = 0; i < NPROC; ++i)
         yac_init_domain(&l_domain[i], comm[i], &blocks[i], 1, "domain");

      // finish domain definition
      for (int i = 0; i < NPROC; ++i)
         yac_enddef_domain(&l_domain[i]);

      //-------------------
      // testing
      //-------------------

      { // testing yac_get_domain_neighbours

        for (int i = 0; i < NPROC; ++i) {

          unsigned ref_num_ranks[NPROC] = {2, 3, 3, 2};
          unsigned ref_ranks[NPROC][NPROC] = {{1,2},{0,2,3},{0,1,3},{1,2}};

          enum yac_location locations[2] = {CELL, CORNER};

          for (unsigned  j = 0; j < 2; ++j) {

            unsigned * ranks, num_ranks;

            yac_get_domain_neighbours(l_domain[i], locations[j], &ranks, &num_ranks);

            if (ref_num_ranks[i] != num_ranks)
              PUT_ERR("error in yac_get_domain_neighbours num_ranks\n");

            for (unsigned k = 0; k < num_ranks; ++k)
              if (ref_ranks[i][k] != ranks[k])
                PUT_ERR("error in yac_get_domain_neighbours ranks\n");

            free(ranks);
          }
        }
      }

      { // testing yac_get_matching_domain_neighbours

        unsigned * ranks = NULL, ranks_array_size = 0;

        { // global circle

          struct bounding_circle global_circle;
          LLtoXYZ(0.0, M_PI_2, global_circle.base_vector);
          global_circle.inc_angle = M_PI;

          unsigned num_ranks[NPROC];

          for (int i = 0, offset = 0; i < NPROC; offset += num_ranks[i++])
            yac_get_matching_domain_neighbours(l_domain[i], global_circle,
                                               &ranks, &ranks_array_size,
                                               &num_ranks[i], offset);

          for (int i = 0, offset = 0; i < NPROC; offset += num_ranks[i++]) {

            unsigned ref_num_ranks[NPROC] = {2, 3, 3, 2};
            unsigned ref_ranks[NPROC][NPROC] = {{1,2},{0,2,3},{0,1,3},{1,2}};

            if (ref_num_ranks[i] != num_ranks[i])
              PUT_ERR("error in yac_get_matching_domain_neighbours num_ranks\n");

            for (unsigned j = 0; j < num_ranks[i]; ++j)
              if (ref_ranks[i][j] != ranks[offset + j])
                PUT_ERR("error in yac_get_matching_domain_neighbours ranks\n");
          }
        }

        {
          struct bounding_circle circle;
          LLtoXYZ(0.0, 0.0, circle.base_vector);
          circle.inc_angle = 3.0 * rad;

          unsigned num_ranks;

          yac_get_matching_domain_neighbours(l_domain[0], circle, &ranks,
                                             &ranks_array_size, &num_ranks, 0);

          unsigned ref_num_ranks = 1;
          unsigned ref_ranks[] = {1};

          if (ref_num_ranks != num_ranks)
            PUT_ERR("error in yac_get_matching_domain_neighbours num_ranks\n");

          for (unsigned i = 0; i < num_ranks; ++i)
            if (ref_ranks[i] != ranks[i])
              PUT_ERR("error in yac_get_matching_domain_neighbours ranks\n");
        }

        {
          struct bounding_circle circle;
          LLtoXYZ(3.0 * rad, 0.0 * rad, circle.base_vector);
          circle.inc_angle = 4.0 * rad;

          unsigned num_ranks;

          yac_get_matching_domain_neighbours(l_domain[0], circle, &ranks,
                                             &ranks_array_size, &num_ranks, 0);

          unsigned ref_num_ranks = 2;
          unsigned ref_ranks[] = {1, 2};

          if (ref_num_ranks != num_ranks)
            PUT_ERR("error in yac_get_matching_domain_neighbours num_ranks\n");

          for (unsigned i = 0; i < num_ranks; ++i)
            if (ref_ranks[i] != ranks[i])
              PUT_ERR("error in yac_get_matching_domain_neighbours ranks\n");
        }

        {
          struct bounding_circle circle;
          LLtoXYZ(0.0, 0.0, circle.base_vector);
          circle.inc_angle = 1.0 * rad;

          unsigned num_ranks;

          yac_get_matching_domain_neighbours(l_domain[0], circle, &ranks,
                                             &ranks_array_size, &num_ranks, 0);

          unsigned ref_num_ranks = 0;
          unsigned ref_ranks[] = {-1};

          if (ref_num_ranks != num_ranks)
            PUT_ERR("error in yac_get_matching_domain_neighbours num_ranks\n");

          for (unsigned i = 0; i < num_ranks; ++i)
            if (ref_ranks[i] != ranks[i])
              PUT_ERR("error in yac_get_matching_domain_neighbours ranks\n");
        }

        {
          struct bounding_circle circle;
          LLtoXYZ(1.5 * rad, 3.5 * rad, circle.base_vector);
          circle.inc_angle = 1.0 * rad;

          unsigned num_ranks;

          yac_get_matching_domain_neighbours(l_domain[1], circle, &ranks,
                                             &ranks_array_size, &num_ranks, 0);

          unsigned ref_num_ranks = 0;
          unsigned ref_ranks[] = {-1};

          if (ref_num_ranks != num_ranks)
            PUT_ERR("error in yac_get_matching_domain_neighbours num_ranks\n");

          for (unsigned i = 0; i < num_ranks; ++i)
            if (ref_ranks[i] != ranks[i])
              PUT_ERR("error in yac_get_matching_domain_neighbours ranks\n");
        }

        {
          struct bounding_circle circle;
          LLtoXYZ(1.5 * rad, 2.5 * rad, circle.base_vector);
          circle.inc_angle = 1.0 * rad;

          unsigned num_ranks;

          yac_get_matching_domain_neighbours(l_domain[1], circle, &ranks,
                                             &ranks_array_size, &num_ranks, 0);

          unsigned ref_num_ranks = 1;
          unsigned ref_ranks[] = {0};

          if (ref_num_ranks != num_ranks)
            PUT_ERR("error in yac_get_matching_domain_neighbours num_ranks\n");

          for (unsigned i = 0; i < num_ranks; ++i)
            if (ref_ranks[i] != ranks[i])
              PUT_ERR("error in yac_get_matching_domain_neighbours ranks\n");
        }

        {
          struct bounding_circle circle;
          LLtoXYZ(3.0 * rad, 2.0 * rad, circle.base_vector);
          circle.inc_angle = 1.0 * rad;

          unsigned num_ranks;

          yac_get_matching_domain_neighbours(l_domain[1], circle, &ranks,
                                             &ranks_array_size, &num_ranks, 0);

          unsigned ref_num_ranks = 2;
          unsigned ref_ranks[] = {0,2};

          if (ref_num_ranks != num_ranks)
            PUT_ERR("error in yac_get_matching_domain_neighbours num_ranks\n");

          for (unsigned i = 0; i < num_ranks; ++i)
            if (ref_ranks[i] != ranks[i])
              PUT_ERR("error in yac_get_matching_domain_neighbours ranks\n");
        }

        {
          struct bounding_circle circle;
          LLtoXYZ(2.5 * rad, 4.5 * rad, circle.base_vector);
          circle.inc_angle = 1.0 * rad;

          unsigned num_ranks;

          yac_get_matching_domain_neighbours(l_domain[1], circle, &ranks,
                                             &ranks_array_size, &num_ranks, 0);

          unsigned ref_num_ranks = 2;
          unsigned ref_ranks[] = {2,3};

          if (ref_num_ranks != num_ranks)
            PUT_ERR("error in yac_get_matching_domain_neighbours num_ranks\n");

          for (unsigned i = 0; i < num_ranks; ++i)
            if (ref_ranks[i] != ranks[i])
              PUT_ERR("error in yac_get_matching_domain_neighbours ranks\n");
        }

        {
          struct bounding_circle circle;
          LLtoXYZ(3.5 * rad, 4.5 * rad, circle.base_vector);
          circle.inc_angle = 1.0 * rad;

          unsigned num_ranks;

          yac_get_matching_domain_neighbours(l_domain[2], circle, &ranks,
                                             &ranks_array_size, &num_ranks, 0);

          unsigned ref_num_ranks = 2;
          unsigned ref_ranks[] = {1,3};

          if (ref_num_ranks != num_ranks)
            PUT_ERR("error in yac_get_matching_domain_neighbours num_ranks\n");

          for (unsigned i = 0; i < num_ranks; ++i)
            if (ref_ranks[i] != ranks[i])
              PUT_ERR("error in yac_get_matching_domain_neighbours ranks\n");
        }

        {
          struct bounding_circle circle;
          LLtoXYZ(5.5 * rad, 5.5 * rad, circle.base_vector);
          circle.inc_angle = 1.0 * rad;

          unsigned num_ranks;

          yac_get_matching_domain_neighbours(l_domain[3], circle, &ranks,
                                             &ranks_array_size, &num_ranks, 0);

          unsigned ref_num_ranks = 1;
          unsigned ref_ranks[] = {2};

          if (ref_num_ranks != num_ranks)
            PUT_ERR("error in yac_get_matching_domain_neighbours num_ranks\n");

          for (unsigned i = 0; i < num_ranks; ++i)
            if (ref_ranks[i] != ranks[i])
              PUT_ERR("error in yac_get_matching_domain_neighbours ranks\n");
        }

        free(ranks);
      }

      //---------------------------------------
      // clean up
      //---------------------------------------

      for (int i = 0; i < NPROC; ++i)
         yac_free_domain(&l_domain[i]);

      for (int i = 0; i < NPROC; ++i)
         yac_delete_grid(grid[i]);

      for (int i = 0; i < NPROC; ++i)
         yac_free_communicator(comm[i]);

#undef NPROC
   }

   { // test packing and unpacking of domains

      struct communicator * comm;

      comm = yac_local_comm_new();

      { // tests a domain with only a single subdomain

         //---------------
         // setup
         //---------------
         struct grid * grid;
         unsigned num_cells[2] = {3,3};
         unsigned cyclic[2] = {0,0};

         double coords_x[4] = {0*rad,1*rad,2*rad,3*rad};
         double coords_y[4] = {0*rad,1*rad,2*rad,3*rad};

         grid = yac_reg2d_grid_new(coords_x, coords_y, num_cells, cyclic);

         struct subdomain subdomain;

         int global_cell_ids_a[9] = {0,1,2,3,4,5,6,7,8};
         unsigned cell_is_on_rank_a[9] = {-1,-1,-1, -1,-1,-1, -1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain, grid, comm, NULL, NULL,
                                     global_cell_ids_a, cell_is_on_rank_a, NULL, NULL);

         struct domain domain;

         struct subdomain * subdomains[1] = {&subdomain};

         yac_init_domain(&domain, comm, subdomains, 1, "dummy");
         yac_enddef_domain(&domain);

         //---------------
         // testing
         //---------------
         double * dble_buf;
         int * int_buf;
         unsigned dble_buf_size, int_buf_size;
         struct bounding_circle domain_extent;
         unsigned num_packed_cells;

         yac_get_subdomain_extents(domain, &domain_extent);

         yac_pack_domain(&domain, &domain_extent, 1, &dble_buf, &dble_buf_size,
                         &int_buf, &int_buf_size, &num_packed_cells);

         if (num_packed_cells != 9) PUT_ERR("wrong number of packed cells\n");

         yac_free_domain(&domain);

         unsigned dble_data_size, int_data_size;

         struct domain unpacked_domain;

         yac_unpack_domain(&unpacked_domain, comm, "dummy", dble_buf, &dble_data_size,
                           int_buf, &int_data_size);

         //---------------
         // cleanup
         //---------------
         free(dble_buf);
         free(int_buf);

         yac_free_domain(&unpacked_domain);
         yac_delete_grid(grid);
      }

      { // tests a domain (4x3) that is distributed among two subdomain on the
        // same process

         //---------------
         // setup
         //---------------
         struct grid * grid_a, * grid_b;
         unsigned num_cells[2] = {3,3};
         unsigned cyclic[2] = {0,0};

         double coords_x[5] = {0*rad,1*rad,2*rad,3*rad,4*rad};
         double coords_y[4] = {0*rad,1*rad,2*rad,3*rad};

         grid_a = yac_reg2d_grid_new(coords_x, coords_y, num_cells, cyclic);
         grid_b = yac_reg2d_grid_new(coords_x+1, coords_y, num_cells, cyclic);

         struct subdomain subdomain_a;

         int global_corner_ids_a[16] = {0,1,2,3, 5,6,7,8, 10,11,12,13, 15,16,17,18};
         unsigned corner_is_on_rank_a[16] = {-1,-1,-1,0, -1,-1,-1,0, -1,-1,-1,0, -1,-1,-1,0};
         int global_cell_ids_a[9] = {0,1,2, 4,5,6, 8,9,10};
         unsigned cell_is_on_rank_a[9] = {-1,-1,0, -1,-1,0, -1,-1,0};
         int global_edge_ids_a[24] = {0,1,2,3,4,5,7,9,10,11,12,13,14,16,18,19,20,21,22,23,25,27,28,29};
         unsigned edge_is_on_rank_a[24] = {-1,-1,-1,-1,0,-1,0,-1,-1,-1,-1,0,-1,0,-1,-1,-1,-1,0,-1,0,-1,-1,0};

         yac_init_subdomain_unstruct(&subdomain_a, grid_a, comm,
                                     global_corner_ids_a, corner_is_on_rank_a,
                                     global_cell_ids_a, cell_is_on_rank_a,
                                     global_edge_ids_a, edge_is_on_rank_a);

         struct subdomain subdomain_b;

         int global_corner_ids_b[16] = {1,2,3,4, 6,7,8,9, 11,12,13,14, 16,17,18,19};
         unsigned corner_is_on_rank_b[16] = {0,0,-1,-1, 0,0,-1,-1, 0,0,-1,-1, 0,0,-1,-1};
         int global_cell_ids_b[9] = {1,2,3, 5,6,7, 9,10,11};
         unsigned cell_is_on_rank_b[9] = {0,-1,-1, 0,-1,-1, 0,-1,-1};
         int global_edge_ids_b[24] = {2,3,4,5,6,7,8,11,12,13,14,15,16,17,20,21,22,23,24,25,26,28,29,30};
         unsigned edge_is_on_rank_b[24] = {0,0,-1,0,-1,-1,-1,0,0,-1,0,-1,-1,-1,0,0,-1,0,-1,-1,-1,0,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_b, grid_b, comm,
                                     global_corner_ids_b, corner_is_on_rank_b,
                                     global_cell_ids_b, cell_is_on_rank_b,
                                     global_edge_ids_b, edge_is_on_rank_b);

         struct subdomain * subdomains[2];

         subdomains[0] = &subdomain_a;
         subdomains[1] = &subdomain_b;

         struct domain domain;

         yac_init_domain(&domain, comm, subdomains, 2, "dummy");
         yac_enddef_domain(&domain);

         //---------------
         // testing
         //---------------

         { // pack whole domain
            double * dble_buf;
            int * int_buf;
            unsigned dble_buf_size, int_buf_size;
            struct bounding_circle domain_extents[2];
            unsigned num_packed_cells;

            yac_get_subdomain_extents(domain, domain_extents);

            yac_pack_domain(&domain, domain_extents, 2, &dble_buf,
                            &dble_buf_size, &int_buf, &int_buf_size,
                            &num_packed_cells);

            if (num_packed_cells != 18)
               PUT_ERR("wrong number of packed cells\n");

            unsigned dble_data_size, int_data_size;

            struct domain unpacked_domain;

            yac_unpack_domain(&unpacked_domain, comm, "dummy", dble_buf, &dble_data_size,
                              int_buf, &int_data_size);

            free(dble_buf);
            free(int_buf);

            yac_free_domain(&unpacked_domain);
         }

         { // pack only a part of the domain, which is associated to one subdomain
            double * dble_buf;
            int * int_buf;
            unsigned dble_buf_size, int_buf_size;
            struct bounding_circle extent;
            double tmp_vec[3];
            unsigned num_packed_cells;

            // the extent is supposed to contain cells with global ids {0,1,4,5}
            LLtoXYZ(1.0 * rad, 1.0 * rad, extent.base_vector);
            LLtoXYZ(0.5 * rad, 0.5 * rad, tmp_vec);
            extent.inc_angle = get_vector_angle(extent.base_vector, tmp_vec);

            yac_pack_domain(&domain, &extent, 1, &dble_buf, &dble_buf_size,
                            &int_buf, &int_buf_size, &num_packed_cells);

            // 4 from subdomain 0 and 2 from subdomain 1
            if (num_packed_cells != 4 + 2)
               PUT_ERR("wrong number of packed cells\n");

            unsigned dble_data_size, int_data_size;

            struct domain unpacked_domain;

            yac_unpack_domain(&unpacked_domain, comm, "dummy", dble_buf, &dble_data_size,
                              int_buf, &int_data_size);

            free(dble_buf);
            free(int_buf);

            yac_free_domain(&unpacked_domain);
         }

         { // pack only a part of the domain, which is associated to both subdomains
            int global_cell_ids[4] = {1,2,5,6};
            double * dble_buf;
            int * int_buf;
            unsigned dble_buf_size, int_buf_size;
            struct bounding_circle extent;
            double tmp_vec[3];
            unsigned num_packed_cells;

            // the extent is supposed to contain cells with global ids {0,1,4,5}
            LLtoXYZ(2.0 * rad, 1.0 * rad, extent.base_vector);
            LLtoXYZ(1.5 * rad, 0.5 * rad, tmp_vec);
            extent.inc_angle = get_vector_angle(extent.base_vector, tmp_vec);

            yac_pack_domain(&domain, &extent, 1, &dble_buf, &dble_buf_size,
                            &int_buf, &int_buf_size, &num_packed_cells);

            // 4 from subdomain 0 and 4 from subdomain 1
            if (num_packed_cells != 4 + 4)
               PUT_ERR("wrong number of packed cells\n");

            unsigned dble_data_size, int_data_size;

            struct domain unpacked_domain;

            yac_unpack_domain(&unpacked_domain, comm, "dummy", dble_buf, &dble_data_size,
                              int_buf, &int_data_size);

            free(dble_buf);
            free(int_buf);

            yac_free_domain(&unpacked_domain);
         }

         //---------------
         // cleanup
         //---------------
         yac_free_domain(&domain);
         yac_delete_grid(grid_a);
         yac_delete_grid(grid_b);
      }

      yac_free_communicator(comm);
   }
   
   yac_comm_finalize();

   return TEST_EXIT_CODE;
}
