!!
!! @file test_dummy_coupling.F90
!!
!! @copyright Copyright  (C)  2013 DKRZ, MPI-M
!!
!! @version 1.0
!!
!! @author Rene Redler <rene.redler@mpimet.mpg.de>
!!
!
!  Keywords:
!
!  Maintainer: Rene Redler <rene.redler@mpimet.mpg.de>
!
!  URL: https://doc.redmine.dkrz.de/YAC/html/index.html
!
!  This file is part of YAC.
!
!  YAC is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  YAC is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
!

MODULE mo_dummy_coupling

  IMPLICIT NONE

  PUBLIC

  INTEGER, PARAMETER :: pd =  12
  INTEGER, PARAMETER :: rd = 307

  INTEGER, PARAMETER :: dp = SELECTED_REAL_KIND(pd,rd) !< double precision
  INTEGER, PARAMETER :: wp = dp                        !< selected working precision

  INTEGER, PARAMETER :: no_of_fields = 10
  INTEGER, PARAMETER :: nbr_subdomain_ids = 1
  INTEGER, PARAMETER :: nbr_cells = 2
  INTEGER, PARAMETER :: nbr_vertices = 4

  INTEGER, PARAMETER :: max_char_length = 132

  CHARACTER(LEN=max_char_length) :: dummy_name
  CHARACTER(LEN=max_char_length) :: field_name(no_of_fields)
  CHARACTER(LEN=max_char_length) :: xml_filename
  CHARACTER(LEN=max_char_length) :: xsd_filename
  CHARACTER(LEN=max_char_length) :: grid_name
  CHARACTER(LEN=max_char_length) :: comp_name
  CHARACTER(LEN=max_char_length) :: timestep_string

  INTEGER :: info
  INTEGER :: comp_id
  INTEGER :: comp_ids(1)
  INTEGER :: cell_point_ids(1)
  INTEGER :: cell_mask_ids(1)
  INTEGER :: domain_id
  INTEGER :: subdomain_id
  INTEGER :: subdomain_ids(nbr_subdomain_ids)

  INTEGER :: nbr_fields
  INTEGER :: glb_index(nbr_cells)
  INTEGER :: owner_local(nbr_cells)
  INTEGER :: nbr_vertices_per_cell

  REAL(wp), ALLOCATABLE :: buffer(:,:)
  REAL(wp), ALLOCATABLE :: buffer_lon(:)
  REAL(wp), ALLOCATABLE :: buffer_lat(:)
  INTEGER, ALLOCATABLE  :: cell_to_vertex(:,:)

  INTEGER, ALLOCATABLE  :: cell_mask(:)
  INTEGER, ALLOCATABLE  :: field_id(:)

END MODULE mo_dummy_coupling

! ------------------------------------------------------------------------

#include "test_macros.inc"

! ------------------------------------------------------------------------

PROGRAM dummy_coupling

  USE utest

#ifdef USE_MPI

  USE mpi

  IMPLICIT NONE

  INTEGER :: ierror, npes, rank

  CALL start_test("dummy_coupling")

  CALL mpi_init (ierror)

  CALL mpi_comm_rank ( MPI_COMM_WORLD, rank, ierror )
  CALL mpi_comm_size ( MPI_COMM_WORLD, npes, ierror )

  SELECT CASE ( rank )

      CASE ( 0 )
          CALL dummy_atmosphere ( )

      CASE ( 1 )
          CALL dummy_ocean ( )

      CASE ( 2 )
          CALL dummy_io ( )

      CASE DEFAULT
          WRITE ( * , * ) "Too many processes have been launched"
          CALL test ( .FALSE. )
          CALL stop_test
          CALL exit_tests
          CALL mpi_abort (MPI_COMM_WORLD, 999, ierror)

  END SELECT

  CALL mpi_finalize (ierror)

#endif

  CALL stop_test
  CALL exit_tests

END PROGRAM dummy_coupling

! ------------------------------------------------------------------------

#ifdef USE_MPI

SUBROUTINE dummy_atmosphere ( )

  USE mpi
  USE mo_dummy_coupling
  USE mo_yac_finterface, ONLY: yac_finit, yac_fdef_comp,                    &
       &                       yac_fdef_subdomain, yac_fconnect_subdomains, &
       &                       yac_fdef_elements, yac_fdef_points,          &
       &                       yac_fdef_mask, yac_fdef_field, yac_fsearch,  &
       &                       yac_ffinalize, yac_fput, yac_fget,           &
       &                       yac_fget_nbr_fields, yac_fget_field_ids,     &
       &                       yac_fget_model_timestep,                     &
       &                       yac_fget_coupling_timestep,                  &
       &                       yac_fget_role_of_field,                      &
       &                       yac_fget_field_name_from_id,                 &
       &                       yac_fget_field_id_from_name,                 &
       &                       yac_redirstdout, CELL
  USE utest

  IMPLICIT NONE

  INTEGER :: id, role
  INTEGER :: i, ierror
  INTEGER :: local_comm, npes, rank

  ! Initialise the coupler
  xml_filename = "coupling_dummy.xml"
  xsd_filename = "coupling.xsd"
  CALL yac_finit ( xml_filename, xsd_filename )

  ! Inform the coupler about what we are
  comp_name = "dummy_atmosphere"
  CALL yac_fdef_comp ( comp_name, comp_id )
  comp_ids(1) = comp_id

  CALL yac_fget_localcomm ( local_comm, comp_id )

  CALL mpi_comm_rank ( local_comm, rank, ierror )
  CALL mpi_comm_size ( local_comm, npes, ierror )

  i = LEN_TRIM(comp_name)
  CALL yac_redirstdout ( TRIM(comp_name), i, 1, rank, npes, ierror )

  WRITE ( 6 , * ) TRIM(comp_name), " rank ", rank, ": local size is ", npes

  ! Announce one subdomain (patch) to the coupler
  grid_name = "dummy_atmosphere_grid"
  CALL yac_fdef_subdomain ( comp_id, grid_name, subdomain_id )

  subdomain_ids(1) = subdomain_id

  ALLOCATE(buffer_lon(nbr_vertices))
  ALLOCATE(buffer_lat(nbr_vertices))
  ALLOCATE(cell_to_vertex(3,nbr_cells))

  nbr_vertices_per_cell = 3

  ! Define vertices

  !      1 
  !     / \
  !    / o \
  !   /     \
  !  2-------3   Eq.
  !   \     /
  !    \ o /
  !     \ /
  !      4

  buffer_lon(1) =  0.0; buffer_lat(1) =  1.0
  buffer_lon(2) = -1.0; buffer_lat(2) =  0.0
  buffer_lon(3) =  1.0; buffer_lat(3) =  0.0
  buffer_lon(4) =  0.0; buffer_lat(4) = -1.0

  ! Connectivity
  cell_to_vertex(1,1) = 1; cell_to_vertex(2,1) = 2; cell_to_vertex(3,1) = 3 ! cell 1
  cell_to_vertex(1,2) = 2; cell_to_vertex(2,2) = 4; cell_to_vertex(3,2) = 3 ! cell 2

  ! Description of elements, here as unstructured grid
  CALL yac_fdef_elements (         &
       & subdomain_id,             &
       & nbr_vertices,             &
       & nbr_cells,                &
       & nbr_vertices_per_cell,    &
       & buffer_lon,               &
       & buffer_lat,               &
       & cell_to_vertex )

  ! Center points in cells (needed e.g. for patch recovery and nearest neighbour)

  buffer_lon(1) = 0.0; buffer_lat(1) =  0.5
  buffer_lon(2) = 0.0; buffer_lat(2) = -0.5

  CALL yac_fdef_points (           &
       & subdomain_id,             &
       & nbr_cells,                &
       & CELL,                     &
       & buffer_lon,               &
       & buffer_lat,               &
       & cell_point_ids(1) )

  DEALLOCATE (buffer_lon, buffer_lat, cell_to_vertex)

  ! Decomposition information

  DO i = 1, nbr_cells
     glb_index(i) = i
     owner_local(i) = -1
  ENDDO

  CALL yac_fdef_index_location (   &
       & subdomain_id,             &
       & nbr_cells,                &
       & CELL,                     &
       & glb_index,                &
       & owner_local )

  ! Connect subdomains
  CALL yac_fconnect_subdomains (   &
       & comp_id,                  &
       & nbr_subdomain_ids,        &
       & subdomain_ids,            &
       & domain_id )

  ! Mask generation
  ALLOCATE(cell_mask(nbr_cells))
  DO i = 1, nbr_cells
     cell_mask(i) = 0
  ENDDO

  CALL yac_fdef_mask (      &
       & nbr_cells,         &
       & cell_mask,         &
       & cell_point_ids(1), &
       & cell_mask_ids(1) )

  DEALLOCATE (cell_mask)

  field_name(1) =  "TAUX"   ! bundled field containing two components
  field_name(2) =  "TAUY"   ! bundled field containing two components
  field_name(3) =  "SFWFLX" ! bundled field containing three components
  field_name(4) =  "SFTEMP"
  field_name(5) =  "THFLX"  ! bundled field containing four components
  field_name(6) =  "ICEATM" ! bundled field containing four components
  field_name(7) =  "SST"
  field_name(8) =  "OCEANU"
  field_name(9) =  "OCEANV"
  field_name(10) = "ICEOCE" ! bundled field containing four components

  ALLOCATE(field_id(no_of_fields))

  DO i = 1, no_of_fields
     CALL yac_fdef_field (  &
          & field_name(i),  &
          & comp_id,        &
          & domain_id,      &
          & cell_point_ids, &
          & cell_mask_ids,  &
          & 1,              &
          & field_id(i) )
  ENDDO

  CALL yac_fsearch ( 1, comp_ids, no_of_fields, field_id, ierror )

  DEALLOCATE(field_id)

  ! Data exchange

  ALLOCATE(buffer(nbr_cells,5))
  buffer(:,:) = 0.0_wp

  !   field_id(1) represents "TAUX"   wind stress component
  !   field_id(2) represents "TAUY"   wind stress component
  !   field_id(3) represents "SFWFLX" surface fresh water flux
  !   field_id(4) represents "SFTEMP" surface temperature
  !   field_id(5) represents "THFLX"  total heat flux
  !   field_id(6) represents "ICEATM" ice temperatures and melt potential
  !
  !   field_id(7) represents "SST"    sea surface temperature
  !   field_id(8) represents "OCEANU" u component of ocean surface current
  !   field_id(9) represents "OCEANV" v component of ocean surface current
  !   field_id(10)represents "ICEOCE" ice thickness, concentration and temperatures
  !
  !   Get some info back from the xml files
  !
  CALL yac_fget_nbr_fields ( nbr_fields )
  ALLOCATE(field_id(nbr_fields))
  CALL yac_fget_field_ids ( nbr_fields, field_id )

  DO i = 1, nbr_fields
    CALL yac_fget_model_timestep ( field_id(i), &
                                   timestep_string )
    IF ( TRIM(timestep_string) .NE. "PT10M" ) THEN
       WRITE ( 6 , * ) "Wrong time step ", TRIM(field_name(i)), " ", TRIM(timestep_string)
       CALL test (.FALSE. )
    ENDIF
  ENDDO

  DO i = 1, nbr_fields
    CALL yac_fget_coupling_timestep ( field_name(i), timestep_string )
    IF ( TRIM(timestep_string) .NE. "PT01H" ) THEN
       WRITE ( 6 , * ) "Wrong coupling step ", TRIM(field_name(i)), " ", TRIM(timestep_string)
       CALL test (.FALSE. )
    ENDIF
  ENDDO

  DO i = 1, nbr_fields
    CALL yac_fget_field_id_from_name ( field_name(i), id )
    IF ( id .NE. field_id(i) ) THEN
      WRITE ( 6 , * ) "field ID does not match", __FILE__ , __LINE__
      CALL test (.FALSE. )
    ENDIF
  ENDDO

  DO i = 1, nbr_fields
    CALL yac_fget_role_of_field ( field_id(i), role )
    CALL yac_fget_field_name_from_id ( field_id(i), dummy_name )
    IF ( i < 7 .AND. role .NE. 1 ) THEN
      WRITE ( 6 , * ) "Wrong requested role for ", TRIM(dummy_name) , " is ", role
      CALL test (.FALSE. )
    ELSE IF ( i > 6 .AND. role .NE. 2 ) THEN
      WRITE ( 6 , * ) "Wrong requested role for ", TRIM(dummy_name) , " is ", role
      CALL test (.FALSE. )
    ENDIF
  ENDDO

  DO i = 1, nbr_fields
    CALL yac_fget_role_of_field ( field_name(i), role )
    CALL yac_fget_field_id_from_name ( field_name(i), id )
    IF ( i < 7 .AND. role .NE. 1 ) THEN
      WRITE ( 6 , * ) "Wrong requested role for ", TRIM(dummy_name) , " is ", role
      CALL test (.FALSE. )
    ELSE IF ( i > 6 .AND. role .NE. 2 ) THEN
      WRITE ( 6 , * ) "Wrong requested role for ", TRIM(dummy_name) , " is ", role
      CALL test (.FALSE. )
    ENDIF

  ENDDO

  !
  ! Send fields to oceanid
  ! --------------------

  ! meridional wind stress
  buffer(:,1) = 10.1_wp
  buffer(:,2) = 10.2_wp

  CALL yac_fput ( field_id(1), nbr_cells, 2, 1, 1, buffer(1:nbr_cells,1:2), info, ierror )
  IF ( info > 0 ) &
       WRITE ( 6 , * ) "atmosphere CPL TAUX 1", MINVAL(buffer(1:nbr_cells,1:1)), MAXVAL(buffer(1:nbr_cells,1:1))
  IF ( info > 0 ) WRITE ( 6 , * ) &
                       "atmosphere CPL TAUX 2", MINVAL(buffer(1:nbr_cells,2:2)), MAXVAL(buffer(1:nbr_cells,2:2))


  ! zonal  wind stress
  buffer(:,1) = 20.1_wp
  buffer(:,2) = 20.2_wp

  CALL yac_fput ( field_id(2), nbr_cells, 2, 1, 1, buffer(1:nbr_cells,1:2), info, ierror )
  IF ( info > 0 ) &
       WRITE ( 6 , * ) "atmosphere CPL TAUY 1", MINVAL(buffer(1:nbr_cells,1:1)), MAXVAL(buffer(1:nbr_cells,1:1))
  IF ( info > 0 ) &
       WRITE ( 6 , * ) "atmosphere CPL TAUY 2", MINVAL(buffer(1:nbr_cells,2:2)), MAXVAL(buffer(1:nbr_cells,2:2))

  ! surface fresh water flux
  buffer(:,1) = 30.1_wp
  buffer(:,2) = 30.2_wp
  buffer(:,3) = 30.3_wp

  CALL yac_fput ( field_id(3), nbr_cells, 3, 1, 1, buffer(1:nbr_cells,1:3), info, ierror )

  ! surface temperature
  buffer(:,1) = 40.1_wp
  CALL yac_fput ( field_id(4), nbr_cells, 1, 1, 1, buffer(1:nbr_cells,1:1), info, ierror )

  ! total heat flux
  buffer(:,1) = 50.1_wp
  buffer(:,2) = 50.2_wp
  buffer(:,3) = 50.3_wp
  buffer(:,4) = 50.4_wp

  CALL yac_fput ( field_id(5), nbr_cells, 4, 1, 1, buffer(1:nbr_cells,1:4), info, ierror )

  ! ice temperatures and melt potential
  buffer(:,1) = 60.1_wp
  buffer(:,2) = 60.2_wp
  buffer(:,3) = 60.3_wp
  buffer(:,4) = 60.4_wp

  CALL yac_fput ( field_id(6), nbr_cells, 4, 1, 1, buffer(1:nbr_cells,1:4), info, ierror )

  !
  ! Receive fields from ocean
  ! -------------------------

  ! SST
  CALL yac_fget ( field_id(7), nbr_cells, 1, 1, 1, buffer(1:nbr_cells,1:1), info, ierror )
  IF ( info > 0 .AND. buffer(1,1) /= 110.0_wp ) THEN
     WRITE ( 6 , * ) "atmosphere CPL SST", buffer(1,1)
     CALL test (.FALSE. )
  ENDIF

  ! zonal velocity
  CALL yac_fget ( field_id(8), nbr_cells, 1, 1, 1, buffer(1:nbr_cells,1:1), info, ierror )
  IF ( info > 0 .AND. buffer(1,1) /= 120.0_wp ) THEN
     WRITE ( 6 , * ) "atmosphere CPL OCEANU ", buffer(1,1)
     CALL test (.FALSE. )
  ENDIF

  ! meridional velocity
  CALL yac_fget ( field_id(9), nbr_cells, 1, 1, 1, buffer(1:nbr_cells,1:1), info, ierror )
  IF ( info > 0 .AND. buffer(1,1) /= 130.0_wp ) THEN
     WRITE ( 6 , * ) "atmosphere CPL OCEANV ", buffer(1,1)
     CALL test (.FALSE. )
  ENDIF

  ! Ice thickness, concentration, T1 and T2

  CALL yac_fget ( field_id(10), nbr_cells, 5, 1, 1, buffer(1:nbr_cells,1:5), info, ierror )

  IF ( info > 0 ) THEN
     IF ( buffer(1,1) /= 140.1_wp ) THEN
        WRITE ( 6 , * ) "atmosphere CPL ice 1 ", buffer(1,1)
        CALL test (.FALSE. )
     ENDIF

     IF ( buffer(1,2) /= 140.2_wp ) THEN
        WRITE ( 6 , * ) "atmosphere CPL ice 2 ", buffer(1,1)
        CALL test (.FALSE. )
     ENDIF
     IF ( buffer(1,3) /= 140.3_wp ) THEN
        WRITE ( 6 , * ) "atmosphere CPL ice 3 ", buffer(1,1)
        CALL test (.FALSE. )
     ENDIF
     IF ( buffer(1,4) /= 140.4_wp ) THEN
        WRITE ( 6 , * ) "atmosphere CPL ice 4 ", buffer(1,1)
        CALL test (.FALSE. )
     ENDIF
     IF ( buffer(1,5) /= 140.5_wp ) THEN
        WRITE ( 6 , * ) "atmosphere CPL ice 5 ", buffer(1,1)
        CALL test (.FALSE. )
     ENDIF
  ENDIF

  DEALLOCATE(buffer)
  DEALLOCATE(field_id)

  CALL yac_ffinalize

END SUBROUTINE dummy_atmosphere

! ------------------------------------------------------------------------

SUBROUTINE dummy_ocean ( )

  USE mpi
  USE mo_dummy_coupling

  USE mo_yac_finterface, ONLY: yac_finit, yac_fdef_comp,                    &
       &                       yac_fdef_subdomain, yac_fconnect_subdomains, &
       &                       yac_fdef_elements, yac_fdef_points,          &
       &                       yac_fdef_mask, yac_fdef_field, yac_fsearch,  &
       &                       yac_ffinalize, yac_fput, yac_fget,           &
       &                       yac_fget_nbr_fields, yac_fget_field_ids,     &
       &                       yac_redirstdout, CELL

  IMPLICIT NONE

  INTEGER :: i, ierror
  INTEGER :: local_comm, npes, rank

  ! Initialise the coupler
  xml_filename = "coupling_dummy.xml"
  xsd_filename = "coupling.xsd"
  CALL yac_finit ( xml_filename, xsd_filename )

  ! Inform the coupler about what we are
  comp_name = "dummy_ocean"
  CALL yac_fdef_comp ( comp_name, comp_id )
  comp_ids(1) = comp_id

  CALL yac_fget_localcomm ( local_comm, comp_id )

  CALL mpi_comm_rank ( local_comm, rank, ierror )
  CALL mpi_comm_size ( local_comm, npes, ierror )

  i = LEN_TRIM(comp_name)
  CALL yac_redirstdout ( TRIM(comp_name), i, 1, rank, npes, ierror )

  WRITE ( 6 , * ) TRIM(comp_name), " rank ", rank, ": local size is ", npes

  ! Announce one subdomain (patch) to the coupler
  grid_name = "dummy_ocean_grid"
  CALL yac_fdef_subdomain ( comp_id, grid_name, subdomain_id )

  subdomain_ids(1) = subdomain_id

  ALLOCATE(buffer_lon(nbr_vertices))
  ALLOCATE(buffer_lat(nbr_vertices))
  ALLOCATE(cell_to_vertex(3,nbr_cells))

  nbr_vertices_per_cell = 3

  ! Define vertices

  !      1 
  !     / \
  !    / o \
  !   /     \
  !  2-------3   Eq.
  !   \     /
  !    \ o /
  !     \ /
  !      4

  buffer_lon(1) =  0.0; buffer_lat(1) =  1.0
  buffer_lon(2) = -1.0; buffer_lat(2) =  0.0
  buffer_lon(3) =  1.0; buffer_lat(3) =  0.0
  buffer_lon(4) =  0.0; buffer_lat(4) = -1.0

  ! Connectivity
  cell_to_vertex(1,1) = 1; cell_to_vertex(2,1) = 2; cell_to_vertex(3,1) = 3 ! cell 1
  cell_to_vertex(1,2) = 2; cell_to_vertex(2,2) = 4; cell_to_vertex(3,2) = 3 ! cell 2

  ! Description of elements, here as unstructured grid
  CALL yac_fdef_elements (         &
       & subdomain_id,             &
       & nbr_vertices,             &
       & nbr_cells,                &
       & nbr_vertices_per_cell,    &
       & buffer_lon,               &
       & buffer_lat,               &
       & cell_to_vertex )

  ! Center points in cells (needed e.g. for patch recovery and nearest neighbour)

  buffer_lon(1) = 0.0; buffer_lat(1) =  0.5
  buffer_lon(2) = 0.0; buffer_lat(2) = -0.5

  CALL yac_fdef_points (           &
       & subdomain_id,             &
       & nbr_cells,                &
       & CELL,                     &
       & buffer_lon,               &
       & buffer_lat,               &
       & cell_point_ids(1) )

  DEALLOCATE (buffer_lon, buffer_lat, cell_to_vertex)

  ! Decomposition information

  DO i = 1, nbr_cells
     glb_index(i) = i
     owner_local(i) = -1
  ENDDO

  CALL yac_fdef_index_location (   &
       & subdomain_id,             &
       & nbr_cells,                &
       & CELL,                     &
       & glb_index,                &
       & owner_local )

  ! Connect subdomains
  CALL yac_fconnect_subdomains (   &
       & comp_id,                  &
       & nbr_subdomain_ids,        &
       & subdomain_ids,            &
       & domain_id )

  ! Mask generation
  ALLOCATE(cell_mask(nbr_cells))
  DO i = 1, nbr_cells
     cell_mask(i) = 0
  ENDDO

  CALL yac_fdef_mask (      &
       & nbr_cells,         &
       & cell_mask,         &
       & cell_point_ids(1), &
       & cell_mask_ids(1) )

  DEALLOCATE (cell_mask)

  field_name(1) =  "TAUX"   ! bundled field containing two components
  field_name(2) =  "TAUY"   ! bundled field containing two components
  field_name(3) =  "SFWFLX" ! bundled field containing three components
  field_name(4) =  "SFTEMP"
  field_name(5) =  "THFLX"  ! bundled field containing four components
  field_name(6) =  "ICEATM" ! bundled field containing four components
  field_name(7) =  "SST"
  field_name(8) =  "OCEANU"
  field_name(9) =  "OCEANV"
  field_name(10) = "ICEOCE" ! bundled field containing four components

  ALLOCATE(field_id(no_of_fields))

  DO i = 1, no_of_fields
     CALL yac_fdef_field (  &
          & field_name(i),  &
          & comp_id,        &
          & domain_id,      &
          & cell_point_ids, &
          & cell_mask_ids,  &
          & 1,              &
          & field_id(i) )
  ENDDO

  CALL yac_fsearch ( 1, comp_ids, no_of_fields, field_id, ierror )

  DEALLOCATE(field_id)

  ! Data exchange

  ALLOCATE(buffer(nbr_cells,5))
  buffer(:,:) = 0.0_wp

  !   field_id(1) represents "TAUX"   wind stress component
  !   field_id(2) represents "TAUY"   wind stress component
  !   field_id(3) represents "SFWFLX" surface fresh water flux
  !   field_id(4) represents "SFTEMP" surface temperature
  !   field_id(5) represents "THFLX"  total heat flux
  !   field_id(6) represents "ICEATM" ice temperatures and melt potential
  !
  !   field_id(7) represents "SST"    sea surface temperature
  !   field_id(8) represents "OCEANU" u component of ocean surface current
  !   field_id(9) represents "OCEANV" v component of ocean surface current
  !   field_id(10)represents "ICEOCE" ice thickness, concentration and temperatures
  !
  CALL yac_fget_nbr_fields ( nbr_fields )
  ALLOCATE(field_id(nbr_fields))
  CALL yac_fget_field_ids ( nbr_fields, field_id )
  !
  ! Send fields from ocean to atmosphere
  ! ------------------------------------
  !
  ! SST
  buffer(:,1) = 110.0_wp
  CALL yac_fput ( field_id(7), nbr_cells, 1, 1, 1, buffer(1:nbr_cells,1:1), info, ierror )

  ! zonal velocity
  buffer(:,1) = 120.0_wp
  CALL yac_fput ( field_id(8), nbr_cells, 1, 1, 1, buffer(1:nbr_cells,1:1), info, ierror )

  ! meridional velocity
  buffer(:,1) = 130.0_wp
  CALL yac_fput ( field_id(9), nbr_cells, 1, 1, 1, buffer(1:nbr_cells,1:1), info, ierror )

  ! Ice thickness, concentration, T1 and T2
  buffer(:,1) = 140.1_wp
  buffer(:,2) = 140.2_wp
  buffer(:,3) = 140.3_wp
  buffer(:,4) = 140.4_wp
  buffer(:,5) = 140.5_wp

  CALL yac_fput ( field_id(10), nbr_cells, 5, 1, 1, buffer(1:nbr_cells,1:5), info, ierror )

  !
  ! Receive fields from atmosphere
  ! ------------------------------

  ! zonal wind stress

  CALL yac_fget ( field_id(1), nbr_cells, 2, 1, 1, buffer, info, ierror )

  ! meridional wind stress

  CALL yac_fget ( field_id(2), nbr_cells, 2, 1, 1, buffer, info, ierror )

  ! freshwater flux

  CALL yac_fget ( field_id(3), nbr_cells, 3, 1, 1, buffer, info, ierror )

  ! surface air temperature
  CALL yac_fget ( field_id(4), nbr_cells, 1, 1, 1, buffer, info, ierror )

  ! total heat flux - 4 parts - record 5

  CALL yac_fget ( field_id(5), nbr_cells, 4, 1, 1, buffer, info, ierror )

  ! ice parameter

  CALL yac_fget ( field_id(6), nbr_cells, 4, 1, 1, buffer, info, ierror )

  DEALLOCATE(buffer)
  DEALLOCATE(field_id)

  CALL yac_ffinalize

END SUBROUTINE dummy_ocean

! ------------------------------------------------------------------------

SUBROUTINE dummy_io ( )

  USE mpi

  USE mo_yac_finterface, ONLY: yac_finit, yac_fget_localcomm, &
       &                       yac_ffinalize, yac_redirstdout

  USE mo_dummy_coupling, ONLY: xml_filename, xsd_filename, &
       &                       comp_name, comp_id

  IMPLICIT NONE

  INTEGER :: i, ierror
  INTEGER :: local_comm, npes, rank

  ! Initialise the coupler
  xml_filename = "coupling_dummy.xml"
  xsd_filename = "coupling.xsd"
  CALL yac_finit ( xml_filename, xsd_filename )

  ! Inform the coupler about what we are
  comp_name = "dummy_io"
  CALL yac_fdef_comp ( comp_name, comp_id )

  CALL yac_fget_localcomm ( local_comm, comp_id )

  CALL mpi_comm_rank ( local_comm, rank, ierror )
  CALL mpi_comm_size ( local_comm, npes, ierror )

  i = LEN_TRIM(comp_name)
  CALL yac_redirstdout ( TRIM(comp_name), i, 1, rank, npes, ierror )

  WRITE ( 6 , * ) TRIM(comp_name), " rank ", rank, ": local size is ", npes

  CALL yac_fsearch ( 1, (/comp_id/), 0, (/0/), ierror )

  CALL yac_ffinalize

END SUBROUTINE dummy_io

#endif
