/**
 * @file test_field_interpolation_parallel2.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#ifdef USE_MPI
#include <mpi.h>
#endif

#include "tests.h"

#include "grid.h"
#include "grid_reg2d.h"
#include "communicator.h"
#include "communicator_mpi.h"
#include "domain.h"
#include "global_search.h"
#include "interpolation_method_fixed.h"
#include "interpolation_method_avg.h"
#include "fields.h"
#include "field_interpolation_gen.h"
#include "field_interpolation_gen_src_tgt.h"

void check_data_request(struct interpolation_method_data_request * request,
                        struct interpolation_method_data_request* ref_request,
                        unsigned num_points_per_subdomain);

void submain_1(struct communicator * global_comm,
               struct communicator * domain_comm);

void submain_2(struct communicator * global_comm,
               struct communicator * domain_comm);

int mapping_on_source;

int main(int argc, char *argv[]) {

#ifndef USE_MPI
   return EXIT_SKIP_TEST;
#else

   if (argc != 2) {
      PUT_ERR("wrong number of arguments\n");
      return TEST_EXIT_CODE;
   }

   mapping_on_source = strcmp(argv[1], "src") == 0;
   if (!mapping_on_source && strcmp(argv[1], "tgt")) {
      PUT_ERR("invalid argument (has to be either \"src\" or \"tgt\")\n");
      return TEST_EXIT_CODE;
   }

   struct communicator * global_comm;

   global_comm = yac_mpi_comm_new();

   if (yac_get_comm_size(global_comm) == 4)
   {

      int my_global_rank;

      my_global_rank = yac_get_comm_rank(global_comm);

      struct communicator * comm;

      comm = yac_comm_split(global_comm, my_global_rank < 2, 0);

      if (my_global_rank < 2)
         submain_1(global_comm, comm);
      else
         submain_2(global_comm, comm);

      yac_free_communicator(comm);
   }

   yac_free_communicator(global_comm);

   MPI_Finalize();

   return TEST_EXIT_CODE;
#endif
}

/*
 * The domain is distributed among 2 processes, with two subdomains each.
 *
 * The global domain is a 4x4 grid:
 *
 *    20------21------22------23------24
 *    |       |       |       |       |
 *    |   12  |   13  |   14  |   15  |
 *    |       |       |       |       |
 *    15------16------17------18------19
 *    |       |       |       |       |
 *    |   08  |   09  |   10  |   11  |
 *    |       |       |       |       |
 *    10------11------12------13------14
 *    |       |       |       |       |
 *    |   04  |   05  |   06  |   07  |
 *    |       |       |       |       |
 *    05------06------07------08------09
 *    |       |       |       |       |
 *    |   00  |   01  |   02  |   03  |
 *    |       |       |       |       |
 *    00------01------02------03------04
 *
 * Subdomains are distributed as follows (xy => x = domain, y = subdomain)
 *
 *    10------10------01------01------01
 *    |       |       |       |       |
 *    |   10  |   10  |   01  |   01  | 
 *    |       |       |       |       |
 *    10------10------10------01------01
 *    |       |       |       |       |
 *    |   10  |   10  |   01  |   01  | 
 *    |       |       |       |       |
 *    10------00------01------11------01
 *    |       |       |       |       |
 *    |   00  |   00  |   11  |   11  | 
 *    |       |       |       |       |
 *    00------00------00------11------11
 *    |       |       |       |       |
 *    |   00  |   00  |   11  |   11  | 
 *    |       |       |       |       |
 *    00------00------11------11------11
 *
 * The mask looks as follows (# = masked points)
 *
 *    +-------+-------+-------+-------+
 *    |       |       |       |       |
 *    |       |       |       |       |
 *    |       |       |       |       |
 *    #-------#-------#-------+-------+
 *    |       |       |       |       |
 *    |       |       |       |       |
 *    |       |       |       |       |
 *    #-------#-------#-------+-------+
 *    |       |       |       |       |
 *    |       |       |       |       |
 *    |       |       |       |       |
 *    +-------+-------+-------+-------+
 *    |       |       |       |       |
 *    |       |       |       |       |
 *    |       |       |       |       |
 *    +-------+-------+-------+-------+
 */
void submain_2(struct communicator * global_comm,
               struct communicator * domain_comm) {

   unsigned i, j, k, my_rank;

   my_rank = yac_get_comm_rank(domain_comm);

   struct grid * grid[2];
   unsigned num_cells[2] = {2,2};
   unsigned cyclic[2] = {0,0};
   double coordinates_x[2][2][3] = {{{0,1,2}, {2,3,4}},{{0,1,2}, {2,3,4}}};
   double coordinates_y[2][2][3] = {{{0,1,2}, {2,3,4}},{{2,3,4}, {0,1,2}}};

   for (i = 0; i < 2; ++i) {
      for (j = 0; j < 2; ++j) {
         for (k = 0; k < 3; ++k) {
            coordinates_x[i][j][k] *= rad;
            coordinates_y[i][j][k] *= rad;
         }
      }
   }

   for (i = 0; i < 2; ++i)
      grid[i] = yac_reg2d_grid_new(coordinates_x[my_rank][i], coordinates_y[my_rank][i],
                                   num_cells, cyclic);

   struct subdomain subdomain[2];

   int global_corner_ids[2][2][9] =
      {{{ 0, 1, 2, 5, 6, 7,10,11,12},{12,13,14,17,18,19,22,23,24}},
       {{10,11,12,15,16,17,20,21,22},{ 2, 3, 4, 7, 8, 9,12,13,14}}};
   unsigned corner_is_on_rank[2][2][9] =
      {{{-1,-1, 1,-1,-1,-1, 1,-1, 0},{-1, 1,-1, 1,-1,-1,-1,-1,-1}},
       {{-1, 0, 0,-1,-1,-1,-1,-1, 0},{-1,-1,-1, 0,-1,-1, 0,-1, 0}}};
   int global_cell_ids[2][2][4] =
      {{{ 0, 1, 4, 5},{10,11,14,15}},{{ 8, 9,12,13},{ 2, 3, 6 ,7}}};
   unsigned cell_is_on_rank[2][2][4] =
      {{{-1,-1,-1,-1},{-1,-1,-1,-1}},{{-1,-1,-1,-1},{-1,-1,-1,-1}}};

   for (i = 0; i < 2; ++i)
      yac_init_subdomain_unstruct(subdomain+i, grid[i], domain_comm,
                                  global_corner_ids[my_rank][i],
                                  corner_is_on_rank[my_rank][i],
                                  global_cell_ids[my_rank][i],
                                  cell_is_on_rank[my_rank][i], NULL, NULL);

   struct domain domain;
   struct subdomain * subdomains[2] = {subdomain+0, subdomain+1};

   yac_init_domain(&domain, domain_comm, subdomains, 2, "domain 2");
   yac_enddef_domain(&domain);

   struct global_search * search;

   search = yac_global_search_new(&domain, "domain 1", global_comm, 1, 1,
                                  "component 2", "component 1");
   yac_enddef_global_search(search);

   struct points points[2];

   for (i = 0; i < 2; ++i)
      yac_init_points(points+i, grid[i], CORNER,
                      coordinates_x[my_rank][i], coordinates_y[my_rank][i]);

   struct mask mask[2];
   unsigned mask_values[2][2][9] = {{{0,0,0, 0,0,0, 1,1,1},
                                     {1,0,0, 1,0,0, 0,0,0}},
                                    {{1,1,1, 1,1,1, 0,0,0},
                                     {0,0,0, 0,0,0, 1,0,0}}};

   for (i = 0; i < 2; ++i)
      yac_init_mask(mask+i, mask_values[my_rank][i], points+i);

   struct coupling_field * field;
   struct points * points_per_subdomain[2][1] = {{points+0},{points+1}};
   struct points ** points_set[2] = {points_per_subdomain[0], points_per_subdomain[1]};
   struct mask * masks_per_subdomain[2][1] = {{mask+0},{mask+1}};
   struct mask ** mask_set[2] = {masks_per_subdomain[0], masks_per_subdomain[1]};

   field = yac_coupling_field_new("field 2", 0, &domain, points_set, mask_set, 1);

   { // test interpolation without using a mask

      struct interpolation_method * method_fixed, * method_avg;
      double fixed_value = 1337;

      method_fixed = yac_new_interpolation_method_fixed(fixed_value, search);
      method_avg = yac_new_interpolation_method_avg(search, &domain, "avg stack",
                                                    points_set, mask_set, 1, 0, 0, 0);

      //---------------------------
      // set up field interpolation
      //---------------------------

      struct field_interpolation_gen * interpolation_gen;
      struct interpolation_method * interpolation_method_stack[] = {method_avg,
                                                                    method_fixed,
                                                                    NULL};

      interpolation_gen =
        yac_new_field_interpolation_gen_src_tgt(points_set, 1, "field 2", search,
                                                "field 1",
                                                interpolation_method_stack, 0,
                                                NULL, 1, mapping_on_source);

      yac_enddef_field_interpolation_gen(interpolation_gen);

      struct field_interpolation * interpolation;

      interpolation = yac_get_field_interpolation(interpolation_gen, "field 2",
                                                  "field 1");

      //---------------------
      // do the interpolation
      //---------------------

      double source_field_point_data[2][2][9] =
         {{{ 0, 1,-1, 5, 6, 7,-1,11,-1},{12,-1,14,-1,18,19,22,23,24}},
          {{10,-1,-1,15,16,17,20,21,-1},{ 2, 3, 4,-1, 8, 9,-1,13,-1}}};

      double * source_field_subdomain_data[2][1] = {{source_field_point_data[my_rank][0]},
                                                    {source_field_point_data[my_rank][1]}};

      double ** source_field_subdomain_data_[2] = {source_field_subdomain_data[0],
                                                   source_field_subdomain_data[1]};

      double target_field[2][9] = {{-1,-1,-1,-1,-1,-1,-1,-1,-1},
                                   {-1,-1,-1,-1,-1,-1,-1,-1,-1}};

      double * target_field_[2] = {target_field[0], target_field[1]};

      double *** source_data[1] = {source_field_subdomain_data_};
      double ** target_data[1] = {target_field_};

      yac_put_field(interpolation, source_data, 1);
      yac_get_field(interpolation, target_data, 1);

      //----------------------------
      // check interpolation results
      //----------------------------

      // dimension: rank, subdomain idx, point idx
      double ref_target_data[2][2][9] = {{{1338,1338,-1,
                                           1338,3,4,
                                           -1,8,-1},
                                          {9,-1,11,
                                           -1,15,16,
                                           19,20,21}},
                                         {{1338,-1,-1,
                                           1338,13,14,
                                           1338,18,-1},
                                          {1338,1338,1338,
                                           -1,5,6,
                                           -1,10,-1}}};

      for (i = 0; i < 9; ++i)
         for (j = 0; j < 2; ++j)
            if (target_data[0][j][i] != ref_target_data[my_rank][j][i])
               PUT_ERR("error in interpolated data on target side\n");

      //--------
      // cleanup
      //--------

      yac_delete_field_interpolation(interpolation);
      yac_delete_field_interpolation_gen(interpolation_gen);
      yac_delete_interpolation_method(method_avg);
      yac_delete_interpolation_method(method_fixed);

      yac_reset_interpolation_method();
   }

   { // test interpolation using a mask (AVG_NO_INTERP)

      struct interpolation_method * method_fixed, * method_avg;
      double fixed_value = 1337;

      method_fixed = yac_new_interpolation_method_fixed(fixed_value, search);
      method_avg =
        yac_new_interpolation_method_avg_mask(search, &domain, "avg stack",
                                              points_set, mask_set, 1, 0, 0, 1, 0);

      //---------------------------
      // set up field interpolation
      //---------------------------

      struct field_interpolation_gen * interpolation_gen;
      struct interpolation_method * interpolation_method_stack[] = {method_avg,
                                                                    method_fixed,
                                                                    NULL};

      interpolation_gen =
        yac_new_field_interpolation_gen_src_tgt(points_set, 1, "field 2", search,
                                                "field 1",
                                                interpolation_method_stack, 0,
                                                NULL, 1, mapping_on_source);

      yac_enddef_field_interpolation_gen(interpolation_gen);

      struct field_interpolation * interpolation;

      interpolation = yac_get_field_interpolation(interpolation_gen, "field 2",
                                              "field 1");

      //---------------------
      // do the interpolation
      //---------------------

      double source_field_point_data[2][2][9] =
         {{{ 0, 1,-1, 5, 6, 7,-1,11,-1},{12,-1,14,-1,18,19,22,23,24}},
          {{10,-1,-1,15,16,17,20,21,-1},{ 2, 3, 4,-1, 8, 9,-1,13,-1}}};

      double * source_field_subdomain_data[2][1] = {{source_field_point_data[my_rank][0]},
                                                    {source_field_point_data[my_rank][1]}};

      double ** source_field_subdomain_data_[2] = {source_field_subdomain_data[0],
                                                   source_field_subdomain_data[1]};

      double target_field[2][9] = {{-1,-1,-1,-1,-1,-1,-1,-1,-1},
                                   {-1,-1,-1,-1,-1,-1,-1,-1,-1}};

      double * target_field_[2] = {target_field[0], target_field[1]};

      double *** source_data[1] = {source_field_subdomain_data_};
      double ** target_data[1] = {target_field_};

      yac_put_field(interpolation, source_data, 1);
      yac_get_field(interpolation, target_data, 1);

      //----------------------------
      // check interpolation results
      //----------------------------

      // dimension: rank, subdomain idx, point idx
      double ref_target_data[2][2][9] = {{{1338,1338,-1,
                                           1338,3,1338,
                                           -1,8,-1},
                                          {1338,-1,1338,
                                           -1,1338,1338,
                                           19,20,21}},
                                         {{1338,-1,-1,
                                           1338,13,1338,
                                           1338,18,-1},
                                          {1338,1338,1338,
                                           -1,1338,1338,
                                           -1,1338,-1}}};

      for (i = 0; i < 9; ++i)
         for (j = 0; j < 2; ++j)
            if (target_data[0][j][i] != ref_target_data[my_rank][j][i])
               PUT_ERR("error in interpolated data on target side\n");

      //--------
      // cleanup
      //--------

      yac_delete_field_interpolation_gen(interpolation_gen);
      yac_delete_field_interpolation(interpolation);
      yac_delete_interpolation_method(method_avg);
      yac_delete_interpolation_method(method_fixed);

      yac_reset_interpolation_method();
   }

   { // test interpolation using a mask (AVG_PARTIAL)

      struct interpolation_method * method_fixed, * method_avg;
      double fixed_value = 1337;

      method_fixed = yac_new_interpolation_method_fixed(fixed_value, search);
      method_avg =
        yac_new_interpolation_method_avg_mask(search, &domain, "avg stack",
                                              points_set, mask_set, 1, 0, 0, 1, 1);

      //---------------------------
      // set up field interpolation
      //---------------------------

      struct field_interpolation_gen * interpolation_gen;
      struct interpolation_method * interpolation_method_stack[] = {method_avg,
                                                                    method_fixed,
                                                                    NULL};

      interpolation_gen =
        yac_new_field_interpolation_gen_src_tgt(points_set, 1, "field 2", search,
                                                "field 1",
                                                interpolation_method_stack, 0,
                                                NULL, 1, mapping_on_source);

      yac_enddef_field_interpolation_gen(interpolation_gen);

      struct field_interpolation * interpolation;

      interpolation = yac_get_field_interpolation(interpolation_gen, "field 2",
                                              "field 1");

      //---------------------
      // do the interpolation
      //---------------------

      double source_field_point_data[2][2][9] =
         {{{ 0, 1,-1, 5, 6, 7,-1,-2,-1},{-2,-1,14,-1,18,19,22,23,24}},
          {{-2,-1,-1,-2,-2,-2,20,21,-1},{ 2, 3, 4,-1, 8, 9,-1,13,-1}}};

      double * source_field_subdomain_data[2][1] = {{source_field_point_data[my_rank][0]},
                                                    {source_field_point_data[my_rank][1]}};

      double ** source_field_subdomain_data_[2] = {source_field_subdomain_data[0],
                                                   source_field_subdomain_data[1]};

      double target_field[2][9] = {{-1,-1,-1,-1,-1,-1,-1,-1,-1},
                                   {-1,-1,-1,-1,-1,-1,-1,-1,-1}};

      double * target_field_[2] = {target_field[0], target_field[1]};

      double *** source_data[1] = {source_field_subdomain_data_};
      double ** target_data[1] = {target_field_};

      yac_put_field(interpolation, source_data, 1);
      yac_get_field(interpolation, target_data, 1);

      //----------------------------
      // check interpolation results
      //----------------------------

      // dimension: rank, subdomain idx, point idx
      double ref_target_data[2][2][9] = {{{1338,1338,-1,
                                           1338,3,3,
                                           -1,8,-1},
                                          {8.5,-1,1338,
                                           -1,17.5,18.5,
                                           19,20,21}},
                                         {{1338,-1,-1,
                                           1338,13,(11+16+17)/3.0,
                                           1338,18,-1},
                                          {1338,1338,1338,
                                           -1,2.5,3.5,
                                           -1,1338,-1}}};

      for (i = 0; i < 9; ++i)
         for (j = 0; j < 2; ++j)
            if (fabs(target_data[0][j][i] - ref_target_data[my_rank][j][i]) > 10e-5)
               PUT_ERR("error in interpolated data on target side\n");

      //--------
      // cleanup
      //--------

      yac_delete_field_interpolation(interpolation);
      yac_delete_field_interpolation_gen(interpolation_gen);
      yac_delete_interpolation_method(method_avg);
      yac_delete_interpolation_method(method_fixed);

      yac_reset_interpolation_method();
   }

   { // test interpolation using a source and target mask (AVG_PARTIAL + TGT_NO_INTERP)

      struct interpolation_method * method_fixed, * method_avg;
      double fixed_value = 1337;

      method_fixed = yac_new_interpolation_method_fixed(fixed_value, search);
      method_avg =
        yac_new_interpolation_method_avg_mask(search, &domain, "avg stack",
                                              points_set, mask_set, 1, 0, 0, 1, 1);

      //---------------------------
      // set up field interpolation
      //---------------------------

      struct field_interpolation_gen * interpolation_gen;
      struct interpolation_method * interpolation_method_stack[] = {method_avg,
                                                                    method_fixed,
                                                                    NULL};

      interpolation_gen =
        yac_new_field_interpolation_gen_src_tgt_mask(points_set, mask_set, 1, "field 2",
                                                     search, "field 1",
                                                     interpolation_method_stack, 0,
                                                     TGT_NO_INTERP, NULL, 1,
                                                     mapping_on_source);

      yac_enddef_field_interpolation_gen(interpolation_gen);

      struct field_interpolation * interpolation;

      interpolation = yac_get_field_interpolation(interpolation_gen, "field 2",
                                              "field 1");

      //---------------------
      // do the interpolation
      //---------------------

      double source_field_point_data[2][2][9] =
         {{{ 0, 1,-1, 5, 6, 7,-1,-2,-1},{-2,-1,14,-1,18,19,22,23,24}},
          {{-2,-1,-1,-2,-2,-2,20,21,-1},{ 2, 3, 4,-1, 8, 9,-1,13,-1}}};

      double * source_field_subdomain_data[2][1] = {{source_field_point_data[my_rank][0]},
                                                    {source_field_point_data[my_rank][1]}};

      double ** source_field_subdomain_data_[2] = {source_field_subdomain_data[0],
                                                   source_field_subdomain_data[1]};

      double target_field[2][9] = {{-1,-1,-1,-1,-1,-1,-1,-1,-1},
                                   {-1,-1,-1,-1,-1,-1,-1,-1,-1}};

      double * target_field_[2] = {target_field[0], target_field[1]};

      double *** source_data[1] = {source_field_subdomain_data_};
      double ** target_data[1] = {target_field_};

      yac_put_field(interpolation, source_data, 1);
      yac_get_field(interpolation, target_data, 1);

      //----------------------------
      // check interpolation results
      //----------------------------

      // dimension: rank, subdomain idx, point idx
      double ref_target_data[2][2][9] = {{{1338,1338,-1,
                                           1338,3,3,
                                           -1,-1,-1},
                                          {-1,-1,1338,
                                           -1,17.5,18.5,
                                           19,20,21}},
                                         {{-1,-1,-1,
                                           -1,-1,-1,
                                           1338,18,-1},
                                          {1338,1338,1338,
                                           -1,2.5,3.5,
                                           -1,1338,-1}}};

      for (i = 0; i < 9; ++i)
         for (j = 0; j < 2; ++j)
            if (fabs(target_data[0][j][i] - ref_target_data[my_rank][j][i]) > 10e-5)
               PUT_ERR("error in interpolated data on target side\n");

      //--------
      // cleanup
      //--------

      yac_delete_field_interpolation(interpolation);
      yac_delete_field_interpolation_gen(interpolation_gen);
      yac_delete_interpolation_method(method_avg);
      yac_delete_interpolation_method(method_fixed);

      yac_reset_interpolation_method();
   }

   //--------
   // cleanup
   //--------

   yac_coupling_field_delete(field);
   yac_global_search_delete(search);
   yac_free_domain(&domain);
   for (i = 0; i < 2; ++i)
      yac_delete_grid(grid[i]);
}

/*
 * The domain is distributed among 2 processes with two subdomains each
 *
 * The global domain is a 4x4 grid:
 *
 *    20------21------22------23------24
 *    |       |       |       |       |
 *    |   12  |   13  |   14  |   15  |
 *    |       |       |       |       |
 *    15------16------17------18------19
 *    |       |       |       |       |
 *    |   08  |   09  |   10  |   11  |
 *    |       |       |       |       |
 *    10------11------12------13------14
 *    |       |       |       |       |
 *    |   04  |   05  |   06  |   07  |
 *    |       |       |       |       |
 *    05------06------07------08------09
 *    |       |       |       |       |
 *    |   00  |   01  |   02  |   03  |
 *    |       |       |       |       |
 *    00------01------02------03------04
 *
 * Subdomains are distributed as follows (xy => x = domain, y = subdomain)
 *
 *    10------10------10------10------10
 *    |       |       |       |       |
 *    |   10  |   10  |   10  |   10  | 
 *    |       |       |       |       |
 *    10------10------10------10------10
 *    |       |       |       |       |
 *    |   00  |   00  |   00  |   00  | 
 *    |       |       |       |       |
 *    11------11------00------00------00
 *    |       |       |       |       |
 *    |   11  |   11  |   11  |   11  | 
 *    |       |       |       |       |
 *    01------01------11------11------11
 *    |       |       |       |       |
 *    |   01  |   01  |   01  |   01  | 
 *    |       |       |       |       |
 *    01------01------01------01------01
 *
 * The mask looks as follows (# = masked points)
 *
 *    +-------+-------+-------+-------+
 *    |       |       |       |       |
 *    |       |       |       |       |
 *    |       |       |       |       |
 *    +-------+-------+-------+-------+
 *    |       |       |       |       |
 *    |       |       |       |       |
 *    |       |       |       |       |
 *    +-------+-------#-------#-------#
 *    |       |       |       |       |
 *    |       |       |       |       |
 *    |       |       |       |       |
 *    +-------+-------#-------#-------#
 *    |       |       |       |       |
 *    |       |       |       |       |
 *    |       |       |       |       |
 *    +-------+-------+-------+-------+
 */
void submain_1(struct communicator * global_comm,
               struct communicator * domain_comm) {

   int my_rank, i, j, k;

   my_rank = yac_get_comm_rank(domain_comm);

   struct grid * grid[2];
   unsigned num_cells[2] = {4,1};
   unsigned cyclic[2] = {0,0};
   double coordinates_x[5] = {0.5,1.5,2.5,3.5,4.5};
   double coordinates_y[2][2][2] = {{{2.5,3.5},{0.5,1.5}},{{3.5,4.5},{1.5,2.5}}};

   for (i = 0; i < 5; ++i)
      coordinates_x[i] *= rad;

   for (i = 0; i < 2; ++i) {
      for (j = 0; j < 2; ++j) {
         for (k = 0; k < 2; ++k) {
            coordinates_y[i][j][k] *= rad;
         }
      }
   }

   for (i = 0; i < 2; ++i)
      grid[i] = yac_reg2d_grid_new(coordinates_x, coordinates_y[my_rank][i],
                                   num_cells, cyclic);

   struct subdomain subdomain[2];

   int global_corner_ids[2][2][10] =
      {{{10,11,12,13,14,15,16,17,18,19},{ 0, 1, 2, 3, 4, 5, 6, 7, 8, 9}},
       {{15,16,17,18,19,20,21,22,23,24},{ 5, 6, 7, 8, 9,10,11,12,13,14}}};
   unsigned corner_is_on_rank[2][2][10] =
      {{{ 1, 1,-1,-1,-1, 1, 1, 1, 1, 1},{-1,-1,-1,-1,-1,-1,-1, 1, 1, 1}},
       {{-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},{ 0, 0,-1,-1,-1,-1,-1, 0, 0, 0}}};
   int global_cell_ids[2][2][4] =
      {{{ 8, 9,10,11},{ 0, 1, 2, 3}},{{12,13,14,15},{ 4, 5, 6, 7}}};
   unsigned cell_is_on_rank[2][2][4] =
      {{{-1,-1,-1,-1},{-1,-1,-1,-1}},{{-1,-1,-1,-1},{-1,-1,-1,-1}}};

   for (i = 0; i < 2; ++i)
      yac_init_subdomain_unstruct(subdomain+i, grid[i], domain_comm,
                                  global_corner_ids[my_rank][i],
                                  corner_is_on_rank[my_rank][i],
                                  global_cell_ids[my_rank][i],
                                  cell_is_on_rank[my_rank][i], NULL, NULL);

   struct domain domain;
   struct subdomain * subdomains[2] = {subdomain+0,subdomain+1};

   yac_init_domain(&domain, domain_comm, subdomains, 2, "domain 1");
   yac_enddef_domain(&domain);

   struct global_search * search;

   search = yac_global_search_new(&domain, "domain 2", global_comm, 1, 1,
                                  "component 1", "component 2");
   yac_enddef_global_search(search);

   struct points points[2];

   for (i = 0; i < 2; ++i)
      yac_init_points(points+i, grid[i], CORNER,
                      coordinates_x, coordinates_y[my_rank][i]);

   struct mask mask[2];
   unsigned mask_values[2][2][10] = {{{0,0,1,1,1, 0,0,0,0,0},
                                      {0,0,0,0,0, 0,0,1,1,1}},
                                     {{0,0,0,0,0, 0,0,0,0,0},
                                      {0,0,1,1,1, 0,0,1,1,1}}};

   for (i = 0; i < 2; ++i)
      yac_init_mask(mask+i, mask_values[my_rank][i], points+i);

   struct coupling_field * field;
   struct points * points_per_subdomain[2][1] = {{points+0},{points+1}};
   struct points ** points_set[2] = {points_per_subdomain[0], points_per_subdomain[1]};
   struct mask * masks_per_subdomain[2][1] = {{mask+0},{mask+1}};
   struct mask ** mask_set[2] = {masks_per_subdomain[0], masks_per_subdomain[1]};

   field = yac_coupling_field_new("field 1", 0, &domain, points_set, mask_set, 1);

   { // test interpolation without using a mask

      struct interpolation_method * method_fixed, * method_avg;
      double fixed_value = 1337 + 1;

      method_fixed = yac_new_interpolation_method_fixed(fixed_value, search);
      method_avg =
        yac_new_interpolation_method_avg(search, &domain, "avg stack", points_set,
                                         mask_set, 1, 0, 0, 0);

      //---------------------------
      // set up field interpolation
      //---------------------------

      struct field_interpolation_gen * interpolation_gen;
      struct interpolation_method * interpolation_method_stack[] = {method_avg,
                                                                    method_fixed,
                                                                    NULL};

      interpolation_gen =
        yac_new_field_interpolation_gen_src_tgt(points_set, 1, "field 1", search,
                                                "field 2",
                                                interpolation_method_stack, 0,
                                                NULL, 1, mapping_on_source);

      yac_enddef_field_interpolation_gen(interpolation_gen);

      struct field_interpolation * interpolation;

      interpolation = yac_get_field_interpolation(interpolation_gen, "field 1",
                                              "field 2");

      //---------------------
      // do the interpolation
      //---------------------

      double source_field_point_data[2][2][10] =
         {{{-1,-1,12,13,14,-1,-1,-1,-1,-1},{ 0, 1, 2, 3, 4, 5, 6,-1,-1,-1}},
          {{15,16,17,18,19,20,21,22,23,24},{-1,-1, 7, 8, 9,10,11,-1,-1,-1}}};
      double * source_field_subdomain_data[2][1] = {{source_field_point_data[my_rank][0]},
                                                    {source_field_point_data[my_rank][1]}};

      double target_field[2][10] = {{-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                                    {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}};

      double ** source_field_subdomain_data_[2] = {source_field_subdomain_data[0],
                                                   source_field_subdomain_data[1]};

      double * target_field_[2] = {target_field[0], target_field[1]};

      double *** source_data[1] = {source_field_subdomain_data_};
      double ** target_data[1] = {target_field_};

      yac_put_field(interpolation, source_data, 1);
      yac_get_field(interpolation, target_data, 1);

      //----------------------------
      // check interpolation results
      //----------------------------

      // dimension: rank, subdomain idx, point idx
      double ref_target_data[2][2][10] = {{{-1,-1,15,16,1337,
                                            -1,-1,-1,-1,-1},
                                           {3,4,5,6,1337,
                                            8,9,-1,-1,-1}},
                                          {{18,19,20,21,1337,
                                            1337,1337,1337,1337,1337},
                                           {-1,-1,10,11,1337,
                                            13,14,-1,-1,-1}}};

      for (i = 0; i < 10; ++i)
         for (j = 0; j < 2; ++j)
            if (target_data[0][j][i] != ref_target_data[my_rank][j][i])
               PUT_ERR("error in interpolated data on target side\n");

      //---------
      // clean up
      //---------

      yac_delete_field_interpolation_gen(interpolation_gen);
      yac_delete_field_interpolation(interpolation);
      yac_delete_interpolation_method(method_avg);
      yac_delete_interpolation_method(method_fixed);

      yac_reset_interpolation_method();
   }

   { // test interpolation using a mask (AVG_NO_INTERP)

      struct interpolation_method * method_fixed, * method_avg;
      double fixed_value = 1337 + 1;

      method_fixed = yac_new_interpolation_method_fixed(fixed_value, search);
      method_avg =
        yac_new_interpolation_method_avg_mask(search, &domain, "avg stack",
                                              points_set, mask_set, 1, 0, 0, 1, 0);

      //---------------------------
      // set up field interpolation
      //---------------------------

      struct field_interpolation_gen * interpolation_gen;
      struct interpolation_method * interpolation_method_stack[] = {method_avg,
                                                                    method_fixed,
                                                                    NULL};

      interpolation_gen =
        yac_new_field_interpolation_gen_src_tgt(points_set, 1, "field 1", search,
                                                "field 2",
                                                interpolation_method_stack, 0,
                                                NULL, 1, mapping_on_source);

      yac_enddef_field_interpolation_gen(interpolation_gen);

      struct field_interpolation * interpolation;

      interpolation = yac_get_field_interpolation(interpolation_gen, "field 1",
                                              "field 2");

      //---------------------
      // do the interpolation
      //---------------------

      double source_field_point_data[2][2][10] =
         {{{-1,-1,12,13,14,-1,-1,-1,-1,-1},{ 0, 1, 2, 3, 4, 5, 6,-1,-1,-1}},
          {{15,16,17,18,19,20,21,22,23,24},{-1,-1, 7, 8, 9,10,11,-1,-1,-1}}};

      double * source_field_subdomain_data[2][1] = {{source_field_point_data[my_rank][0]},
                                                    {source_field_point_data[my_rank][1]}};

      double target_field[2][10] = {{-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                                    {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}};

      double ** source_field_subdomain_data_[2] = {source_field_subdomain_data[0],
                                                   source_field_subdomain_data[1]};

      double * target_field_[2] = {target_field[0], target_field[1]};

      double *** source_data[1] = {source_field_subdomain_data_};
      double ** target_data[1] = {target_field_};

      yac_put_field(interpolation, source_data, 1);
      yac_get_field(interpolation, target_data, 1);

      //----------------------------
      // check interpolation results
      //----------------------------

      // dimension: rank, subdomain idx, point idx
      double ref_target_data[2][2][10] = {{{-1,-1,1337,16,1337,
                                            -1,-1,-1,-1,-1},
                                           {3,4,5,6,1337,
                                            1337,1337,-1,-1,-1}},
                                          {{1337,1337,1337,21,1337,
                                            1337,1337,1337,1337,1337},
                                           {-1,-1,1337,11,1337,
                                            1337,1337,-1,-1,-1}}};

      for (i = 0; i < 10; ++i)
         for (j = 0; j < 2; ++j)
            if (target_data[0][j][i] != ref_target_data[my_rank][j][i])
               PUT_ERR("error in interpolated data on target side\n");

      //---------
      // clean up
      //---------

      yac_delete_field_interpolation(interpolation);
      yac_delete_field_interpolation_gen(interpolation_gen);
      yac_delete_interpolation_method(method_avg);
      yac_delete_interpolation_method(method_fixed);

      yac_reset_interpolation_method();
   }

   { // test interpolation using a mask (AVG_PARTIAL)

      struct interpolation_method * method_fixed, * method_avg;
      double fixed_value = 1337 + 1;

      method_fixed = yac_new_interpolation_method_fixed(fixed_value, search);
      method_avg =
        yac_new_interpolation_method_avg_mask(search, &domain, "avg stack",
                                              points_set, mask_set, 1, 0, 0, 1, 1);

      //---------------------------
      // set up field interpolation
      //---------------------------

      struct field_interpolation_gen * interpolation_gen;
      struct interpolation_method * interpolation_method_stack[] = {method_avg,
                                                                    method_fixed,
                                                                    NULL};

      interpolation_gen =
        yac_new_field_interpolation_gen_src_tgt(points_set, 1, "field 1", search,
                                                "field 2",
                                                interpolation_method_stack, 0,
                                                NULL, 1, mapping_on_source);

      yac_enddef_field_interpolation_gen(interpolation_gen);

      struct field_interpolation * interpolation;

      interpolation = yac_get_field_interpolation(interpolation_gen, "field 1",
                                              "field 2");

      //---------------------
      // do the interpolation
      //---------------------

      double source_field_point_data[2][2][10] =
         {{{-1,-1,-2,-2,-2,-1,-1,-1,-1,-1},{ 0, 1, 2, 3, 4, 5, 6,-1,-1,-1}},
          {{15,16,17,18,19,20,21,22,23,24},{-1,-1, -2, -2, -2,10,11,-1,-1,-1}}};

      double * source_field_subdomain_data[2][1] = {{source_field_point_data[my_rank][0]},
                                                    {source_field_point_data[my_rank][1]}};

      double target_field[2][10] = {{-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                                    {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}};

      double ** source_field_subdomain_data_[2] = {source_field_subdomain_data[0],
                                                   source_field_subdomain_data[1]};

      double * target_field_[2] = {target_field[0], target_field[1]};

      double *** source_data[1] = {source_field_subdomain_data_};
      double ** target_data[1] = {target_field_};

      yac_put_field(interpolation, source_data, 1);
      yac_get_field(interpolation, target_data, 1);

      //----------------------------
      // check interpolation results
      //----------------------------

      // dimension: rank, subdomain idx, point idx
      double ref_target_data[2][2][10] = {{{-1,-1,15.5,16,1337,
                                            -1,-1,-1,-1,-1},
                                           {3,4,5,6,1337,
                                            5.5,6.5,-1,-1,-1}},
                                          {{20.5,21.5,21,21,1337,
                                            1337,1337,1337,1337,1337},
                                           {-1,-1,(7+8+13)/3.0,11,1337,
                                            1337,1337,-1,-1,-1}}};

      for (i = 0; i < 10; ++i)
         for (j = 0; j < 2; ++j)
            if (fabs(target_data[0][j][i] - ref_target_data[my_rank][j][i]) > 10e-5)
               PUT_ERR("error in interpolated data on target side\n");

      //---------
      // clean up
      //---------

      yac_delete_field_interpolation_gen(interpolation_gen);
      yac_delete_field_interpolation(interpolation);
      yac_delete_interpolation_method(method_avg);
      yac_delete_interpolation_method(method_fixed);

      yac_reset_interpolation_method();
   }

   { // test interpolation using a source and target mask (AVG_PARTIAL + TGT_NO_INTERP)

      struct interpolation_method * method_fixed, * method_avg;
      double fixed_value = 1337 + 1;

      method_fixed = yac_new_interpolation_method_fixed(fixed_value, search);
      method_avg =
        yac_new_interpolation_method_avg_mask(search, &domain, "avg stack",
                                              points_set, mask_set, 1, 0, 0, 1, 1);

      //---------------------------
      // set up field interpolation
      //---------------------------

      struct field_interpolation_gen * interpolation_gen;
      struct interpolation_method * interpolation_method_stack[] = {method_avg,
                                                                    method_fixed,
                                                                    NULL};

      interpolation_gen =
        yac_new_field_interpolation_gen_src_tgt_mask(points_set, mask_set, 1, "field 1",
                                                     search, "field 2",
                                                     interpolation_method_stack, 0,
                                                     TGT_NO_INTERP, NULL, 1,
                                                     mapping_on_source);

      yac_enddef_field_interpolation_gen(interpolation_gen);

      struct field_interpolation * interpolation;

      interpolation = yac_get_field_interpolation(interpolation_gen, "field 1",
                                              "field 2");

      //---------------------
      // do the interpolation
      //---------------------

      double source_field_point_data[2][2][10] =
         {{{-1,-1,-2,-2,-2,-1,-1,-1,-1,-1},{ 0, 1, 2, 3, 4, 5, 6,-1,-1,-1}},
          {{15,16,17,18,19,20,21,22,23,24},{-1,-1, -2, -2, -2,10,11,-1,-1,-1}}};

      double * source_field_subdomain_data[2][1] = {{source_field_point_data[my_rank][0]},
                                                    {source_field_point_data[my_rank][1]}};

      double target_field[2][10] = {{-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                                    {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}};

      double ** source_field_subdomain_data_[2] = {source_field_subdomain_data[0],
                                                   source_field_subdomain_data[1]};

      double * target_field_[2] = {target_field[0], target_field[1]};

      double *** source_data[1] = {source_field_subdomain_data_};
      double ** target_data[1] = {target_field_};

      yac_put_field(interpolation, source_data, 1);
      yac_get_field(interpolation, target_data, 1);

      //----------------------------
      // check interpolation results
      //----------------------------

      // dimension: rank, subdomain idx, point idx
      double ref_target_data[2][2][10] = {{{-1,-1,-1,-1,-1,
                                            -1,-1,-1,-1,-1},
                                           {3,4,5,6,1337,
                                            5.5,6.5,-1,-1,-1}},
                                          {{20.5,21.5,21,21,1337,
                                            1337,1337,1337,1337,1337},
                                           {-1,-1,-1,-1,-1,
                                            1337,1337,-1,-1,-1}}};

      for (i = 0; i < 10; ++i)
         for (j = 0; j < 2; ++j)
            if (fabs(target_data[0][j][i] - ref_target_data[my_rank][j][i]) > 10e-5)
               PUT_ERR("error in interpolated data on target side\n");

      //---------
      // clean up
      //---------

      yac_delete_field_interpolation(interpolation);
      yac_delete_field_interpolation_gen(interpolation_gen);
      yac_delete_interpolation_method(method_avg);
      yac_delete_interpolation_method(method_fixed);

      yac_reset_interpolation_method();
   }

   //---------
   // clean up
   //---------

   yac_coupling_field_delete(field);
   yac_global_search_delete(search);
   yac_free_domain(&domain);
   for (i = 0; i < 2; ++i)
      yac_delete_grid(grid[i]);
}

void check_data_request(struct interpolation_method_data_request * request,
                        struct interpolation_method_data_request * ref_request,
                        unsigned num_points_per_subdomain) {

   unsigned i, j;

   for (i = 0; i < num_points_per_subdomain; ++i) {

      if (request[i].num_values != ref_request[i].num_values)
         PUT_ERR("wrong number of requested points\n");

      for (j = 0; j < ref_request[i].num_values; ++j) {

         if (request[i].global_ids[j] != ref_request[i].global_ids[j])
            PUT_ERR("wrong global_id requested\n");

         if (request[i].ranks[j] != ref_request[i].ranks[j])
            PUT_ERR("point from wrong rank requested\n");
      }
   }
}
