/**
 * @file test_field_interpolation_parallel5.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include <string.h>

#ifdef USE_MPI
#include <mpi.h>
#endif

#include "tests.h"

#include "grid.h"
#include "grid_reg2d.h"
#include "communicator.h"
#include "communicator_mpi.h"
#include "domain.h"
#include "global_search.h"
#include "interpolation_method_file.h"
#include "interpolation_method_avg.h"
#include "fields.h"
#include "field_interpolation_gen.h"
#include "field_interpolation_gen_src.h"
#include "field_interpolation_gen_tgt.h"
#include "weight_file_common.h"

static void
generate_ref_weights();

void submain_src(struct communicator * global_comm,
                 struct communicator * domain_comm);

void submain_tgt(struct communicator * global_comm,
                 struct communicator * domain_comm);

// link data for input weight file
unsigned const num_links_file = 48;
int src_address_file[48] = {
  9,10,16,17, 10,11,17,18, 15,16,22,23, 16,17,23,24, 17,18,24,25, 18,19,25,26,
  22,23,29,30, 23,24,30,31, 24,25,31,32, 25,26,32,33, 30,31,37,38, 31,32,38,39};
int tgt_address_file[48] = {
  8,8,8,8, 9,9,9,9, 13,13,13,13, 14,14,14,14, 15,15,15,15, 16,16,16,16,
  19,19,19,19, 20,20,20,20, 21,21,21,21, 22,22,22,22, 26,26,26,26, 27,27,27,27};
double weights_file[48] = {
  0.1,0.2,0.3,0.4, 0.1,0.2,0.3,0.4, 0.1,0.2,0.3,0.4, 0.1,0.2,0.3,0.4,
  0.1,0.2,0.3,0.4, 0.1,0.2,0.3,0.4, 0.1,0.2,0.3,0.4, 0.1,0.2,0.3,0.4,
  0.1,0.2,0.3,0.4, 0.1,0.2,0.3,0.4, 0.1,0.2,0.3,0.4, 0.1,0.2,0.3,0.4};

// reference link data for output weight file
unsigned ref_num_links = 4 * 36;

int ref_src_address[4 * 36];
int ref_tgt_address[4 * 36];
double ref_weights[4 * 36];

int const * ref_tgt_address_fixed = NULL;
unsigned ref_num_fixed_tgt = 0;
double const * ref_fixed_values = NULL;
int const * ref_num_tgt_per_fixed_value = NULL;
unsigned ref_num_fixed_values = 0;

/*
 * The domain is distributed among 4 processes, with two subdomains each.
 *
 * The global domain is a 4x4 grid:
 *
 *    42--79--43--81--44--83--45--85--46--87--47--89--48
 *    |       |       |       |       |       |       |
 *    66  30  68  31  70  32  72  33  74  34  76  35  78
 *    |       |       |       |       |       |       |
 *    35--65--36--67--37--69--38--71--39--73--40--75--41
 *    |       |       |       |       |       |       |
 *    53  24  55  25  57  26  59  27  61  28  63  29  64
 *    |       |       |       |       |       |       |
 *    28--52--29--54--30--56--31--58--32--60--33--62--34
 *    |       |       |       |       |       |       |
 *    40  18  42  19  44  20  46  21  48  22  50  23  51
 *    |       |       |       |       |       |       |
 *    21--39--22--41--23--43--24--45--25--47--26--49--27
 *    |       |       |       |       |       |       |
 *    27  12  29  13  31  14  33  15  35  16  37  17  38
 *    |       |       |       |       |       |       |
 *    14--26--15--28--16--30--17--32--18--34--19--36--20
 *    |       |       |       |       |       |       |
 *    14  06  16  07  18  08  20  09  22  10  24  11  25
 *    |       |       |       |       |       |       |
 *    07--13--08--15--09--17--10--19--11--21--12--23--13
 *    |       |       |       |       |       |       |
 *    01  00  03  01  05  02  07  03  09  04  11  05  12
 *    |       |       |       |       |       |       |
 *    00--00--01--02--02--04--03--06--04--08--05--10--06
 *
 * Subdomains are distributed as follows (xy => x = domain, y = subdomain)
 *
 *    30--30--30--30--30--30--30--21--21--21--20--20--20
 *    |       |       |       |       |       |       |
 *    30  30  30  30  30  30  30  21  21  21  20  20  20
 *    |       |       |       |       |       |       |
 *    30--30--30--30--30--30--30--21--21--21--20--20--20
 *    |       |       |       |       |       |       |
 *    31  31  31  31  31  31  21  21  21  21  20  20  20
 *    |       |       |       |       |       |       |
 *    31--31--31--31--31--31--21--21--21--21--20--20--20
 *    |       |       |       |       |       |       |
 *    31  31  31  31  31  31  21  21  21  21  20  20  20
 *    |       |       |       |       |       |       |
 *    31--31--31--31--01--01--01--21--21--21--11--11--11
 *    |       |       |       |       |       |       |
 *    00  00  00  01  01  01  01  11  11  11  11  11  11
 *    |       |       |       |       |       |       |
 *    00--00--00--01--01--01--01--11--11--11--11--11--11
 *    |       |       |       |       |       |       |
 *    00  00  00  01  01  01  11  11  11  11  11  11  11
 *    |       |       |       |       |       |       |
 *    00--00--00--01--01--01--10--10--10--10--10--10--10
 *    |       |       |       |       |       |       |
 *    00  00  00  01  11  01  10  10  10  10  10  10  10
 *    |       |       |       |       |       |       |
 *    00--00--00--01--01--01--10--10--10--10--10--10--10
 */

// grid information (is the same for source and target)
unsigned num_cells[4][2] = {{2,3}, {3,2}, {2,3}, {3,2}};
unsigned cyclic[2] = {0,0};
double coordinates_x[4][2][4] = {{{0,1,2},{1,2,3}},
                                 {{3,4,5,6},{3,4,5,6}},
                                 {{4,5,6},{3,4,5}},
                                 {{0,1,2,3},{0,1,2,3}}};
double coordinates_y[4][2][4] = {{{0,1,2,3},{0,1,2,3}},
                                 {{0,1,2},{1,2,3}},
                                 {{3,4,5,6},{3,4,5,6}},
                                 {{4,5,6},{3,4,5}}};

// domain information (is the same for source and target)
int global_corner_ids[4][2][12] =
  {{{0,1,2,7,8,9,14,15,16,21,22,23},
    {1,2,3,8,9,10,15,16,17,22,23,24}},
   {{3,4,5,6,10,11,12,13,17,18,19,20},
    {10,11,12,13,17,18,19,20,24,25,26,27}},
   {{25,26,27,32,33,34,39,40,41,46,47,48},
    {24,25,26,31,32,33,38,39,40,45,46,47}},
   {{28,29,30,31,35,36,37,38,42,43,44,45},
    {21,22,23,24,28,29,30,31,35,36,37,38}}};
unsigned corner_is_on_rank[4][2][12] =
  {{{-1,-1,0,-1,-1,0,-1,-1,0,3,3,0},
    {0,-1,1,0,-1,1,0,-1,-1,3,-1,-1}},
   {{-1,-1,-1,-1,-1,-1,-1,-1,0,1,1,1},
    {1,1,1,1,0,-1,-1,-1,0,2,-1,-1}},
   {{2,1,1,2,-1,-1,2,-1,-1,2,-1,-1},
    {0,-1,1,-1,-1,2,3,-1,2,3,-1,2}},
   {{3,3,3,2,-1,-1,-1,-1,-1,-1,-1,-1},
    {-1,-1,0,0,-1,-1,-1,2,3,3,3,3}}};
int global_cell_ids[4][2][6] =
  {{{0,1,6,7,12,13},{1,2,7,8,13,14}},
   {{3,4,5,9,10,11},{9,10,11,15,16,17}},
   {{22,23,28,29,34,35},{21,22,27,28,33,34}},
   {{24,25,26,30,31,32},{18,19,20,24,25,26}}};
unsigned cell_is_on_rank[4][2][6] =
  {{{-1,0,-1,0,-1,0},{-1,-1,-1,-1,-1,-1}},
   {{-1,-1,-1,1,1,1,},{-1,-1,-1,-1,-1,-1}},
   {{2,-1,2,-1,2,-1},{-1,-1,-1,-1,-1,-1}},
   {{3,3,3,-1,-1,-1},{-1,-1,-1,-1,-1,-1}}};

int mapping_on_source;

int main(int argc, char *argv[]) {

#if !(defined(USE_MPI) && defined(HAVE_C_NETCDF))
   return EXIT_SKIP_TEST;
#else

  if (argc != 2) {
    PUT_ERR("wrong number of arguments\n");
    return TEST_EXIT_CODE;
  }

  mapping_on_source = strcmp(argv[1], "src") == 0;
  if (!mapping_on_source && strcmp(argv[1], "tgt")) {
    PUT_ERR("invalid argument (has to be either \"src\" or \"tgt\")\n");
    return TEST_EXIT_CODE;
  }

  generate_ref_weights();

  for (unsigned i = 0; i < 4; ++i) {
    for (unsigned j = 0; j < 2; ++j) {
      for (unsigned k = 0; k < 4; ++k) {
        coordinates_y[i][j][k] *= rad;
        coordinates_x[i][j][k] *= rad;
      }
    }
  }

  struct communicator * global_comm;

  global_comm = yac_mpi_comm_new();

  if (yac_get_comm_size(global_comm) == 8) {

    int my_global_rank;

    my_global_rank = yac_get_comm_rank(global_comm);

    struct communicator * comm;

    comm = yac_comm_split(global_comm, my_global_rank < 4, 0);

    if (my_global_rank < 4)
      submain_src(global_comm, comm);
    else
      submain_tgt(global_comm, comm);

    yac_free_communicator(comm);
  } else
    PUT_ERR("invalid number of processes\n");

  yac_free_communicator(global_comm);

  MPI_Finalize();

  return TEST_EXIT_CODE;
#endif
}

void submain_tgt(struct communicator * global_comm,
                 struct communicator * domain_comm) {

   int my_rank;

   my_rank = yac_get_comm_rank(domain_comm);

   struct grid * grid[2];

   for (unsigned i = 0; i < 2; ++i)
      grid[i] = yac_reg2d_grid_new(coordinates_x[my_rank][i],
                                   coordinates_y[my_rank][i],
                                   num_cells[my_rank], cyclic);

   struct subdomain subdomain[2];

   for (unsigned i = 0; i < 2; ++i)
      yac_init_subdomain_unstruct(subdomain+i, grid[i], domain_comm,
                                  global_corner_ids[my_rank][i],
                                  corner_is_on_rank[my_rank][i],
                                  global_cell_ids[my_rank][i],
                                  cell_is_on_rank[my_rank][i], NULL, NULL);

   struct domain domain;
   struct subdomain * subdomains[2] = {subdomain+0,subdomain+1};

   yac_init_domain(&domain, domain_comm, subdomains, 2, "domain 2");
   yac_enddef_domain(&domain);

   struct global_search * search;

   search = yac_global_search_new(&domain, "domain 1", global_comm, 1, 1,
                                  "component 2", "component 1");
   yac_enddef_global_search(search);


   double cell_coordinates_x[4][2][3] = {{{0.5,1.5},{1.5,2.5}},
                                         {{3.5,4.5,5.5},{3.5,4.5,5.5}},
                                         {{4.5,5.5},{3.5,4.5}},
                                         {{0.5,1.5,2.5},{0.5,1.5,2.5}}};
   double cell_coordinates_y[4][2][3] = {{{0.5,1.5,2.5},{0.5,1.5,2.5}},
                                         {{0.5,1.5},{1.5,2.5}},
                                         {{3.5,4.5,5.5},{3.5,4.5,5.5}},
                                         {{4.5,5.5},{3.5,4.5}}};

   for (unsigned i = 0; i < 4; ++i) {
      for (unsigned j = 0; j < 2; ++j) {
         for (unsigned k = 0; k < 3; ++k) {
            cell_coordinates_x[i][j][k] *= rad;
            cell_coordinates_y[i][j][k] *= rad;
         }
      }
   }

   struct points points_cells[2];

   for (unsigned i = 0; i < 2; ++i)
      yac_init_points(points_cells+i, grid[i], CELL,
                      cell_coordinates_x[my_rank][i],
                      cell_coordinates_y[my_rank][i]);

   struct coupling_field * field;
   struct points *
       points_per_subdomain[2][3] = {{points_cells+0}, {points_cells+1}};

   struct points ** points_set[2] = {points_per_subdomain[0], points_per_subdomain[1]};

   field = yac_coupling_field_new("field 2", 0, &domain, points_set, NULL, 1);

   { // test interpolation

      //---------------------------
      // set up field interpolation
      //---------------------------

      struct field_interpolation_gen * interpolation_gen;

      interpolation_gen =
        yac_new_field_interpolation_gen_tgt_mask(points_set, NULL, 1, "field 2",
                                                 search, "field 1", 0,
                                                 TGT_NO_INTERP,
                                                 mapping_on_source);

      yac_enddef_field_interpolation_gen(interpolation_gen);

      struct field_interpolation * interpolation;

      interpolation = yac_get_field_interpolation(interpolation_gen, "field 2",
                                                  "field 1");

      //---------------------
      // do the interpolation
      //---------------------

      double target_field[2][6];

      for (unsigned i = 0; i < 2; ++i)
        for (unsigned j = 0; j < 6; ++j)
          target_field[i][j] = -1;

      double * target_field_[2] = {target_field[0], target_field[1]};

      double ** target_data[1] = {target_field_};

      yac_get_field(interpolation, target_data, 1);

      //---------------------------
      // check interpolation result
      //---------------------------

      // ensure that the weight file has been written
#ifdef USE_MPI
      MPI_Barrier(MPI_COMM_WORLD);
#endif

      double ref_target_field[2][6];

      for (unsigned i = 0; i < 2; ++i) {
        for (unsigned j = 0; j < 6; ++j) {

          if (cell_is_on_rank[my_rank][i][j] == -1) {

            ref_target_field[i][j] = 0;

            for (unsigned k = 0; k < 4; ++k)
              ref_target_field[i][j] +=
                ref_weights[4 * global_cell_ids[my_rank][i][j] + k] *
                (double)ref_src_address[4 * global_cell_ids[my_rank][i][j] + k];

          } else {
            ref_target_field[i][j] = -1;
          }
        }
      }

      for (unsigned i = 0; i < 2; ++i)
        for (unsigned j = 0; j < 6; ++j)
           if (fabs(target_data[0][i][j] - ref_target_field[i][j]) > 1e-14) {
               fprintf(stderr, "tgt %lf ref tgt %lf rank %d global cell id %d\n",
                       target_data[0][i][j], ref_target_field[i][j],
                       my_rank, global_cell_ids[my_rank][i][j]);
               PUT_ERR("wrong interpolation result\n");
           }

      //--------
      // cleanup
      //--------

      yac_delete_field_interpolation_gen(interpolation_gen);
      yac_delete_field_interpolation(interpolation);

      yac_reset_interpolation_method();
   }

   //--------
   // cleanup
   //--------

   yac_coupling_field_delete(field);
   yac_global_search_delete(search);
   yac_free_domain(&domain);
   for (unsigned i = 0; i < 2; ++i) {
      yac_free_points(points_cells + i);
      yac_delete_grid(grid[i]);
   }
}

void submain_src(struct communicator * global_comm,
                 struct communicator * domain_comm) {

   char const filename_in[] = "weight_file_in.nc";
   char const filename_out[] = "weight_file_out.nc";

   int my_rank;

   my_rank = yac_get_comm_rank(domain_comm);

    // create a weight file
   if (my_rank == 0) {
     enum yac_location src_locations[1] = {CORNER};
     enum yac_location tgt_location = CELL;
     int * tgt_id_fixed = NULL;
     unsigned num_fixed_tgt = 0;
     double * fixed_values = NULL;
     int * num_tgt_per_fixed_value = NULL;
     unsigned num_fixed_values = 0;
     write_weight_file(filename_in, src_address_file, tgt_address_file,
                       weights_file, num_links_file, src_locations,
                       1, (int*)&num_links_file, tgt_id_fixed, num_fixed_tgt,
                       fixed_values, num_tgt_per_fixed_value, num_fixed_values,
                       tgt_location, "domain 1", "domain 2");
   }

   struct grid * grid[2];

   for (unsigned i = 0; i < 2; ++i)
      grid[i] = yac_reg2d_grid_new(coordinates_x[my_rank][i],
                                   coordinates_y[my_rank][i],
                                   num_cells[my_rank], cyclic);

   struct subdomain subdomain[2];

   for (unsigned i = 0; i < 2; ++i)
      yac_init_subdomain_unstruct(subdomain+i, grid[i], domain_comm,
                                  global_corner_ids[my_rank][i],
                                  corner_is_on_rank[my_rank][i],
                                  global_cell_ids[my_rank][i],
                                  cell_is_on_rank[my_rank][i], NULL, NULL);

   struct domain domain;
   struct subdomain * subdomains[2] = {subdomain+0,subdomain+1};

   yac_init_domain(&domain, domain_comm, subdomains, 2, "domain 1");
   yac_enddef_domain(&domain);

   struct global_search * search;

   search = yac_global_search_new(&domain, "domain 2", global_comm, 1, 1,
                                  "component 1", "component 2");
   yac_enddef_global_search(search);

   struct points points_corner[2];

   for (unsigned i = 0; i < 2; ++i)
      yac_init_points(points_corner+i, grid[i], CORNER,
                      coordinates_x[my_rank][i], coordinates_y[my_rank][i]);

   struct coupling_field * field;
   struct points *
    points_per_subdomain[2][1] = {{points_corner+0}, {points_corner+1}};
   struct points **
    points_set[2] = {points_per_subdomain[0], points_per_subdomain[1]};

   field = yac_coupling_field_new("field 1", 0, &domain, points_set, NULL, 1);

   { // test interpolation using weights from file and avg

      struct interpolation_method * method_file, * method_avg;

      enum yac_location locations[1] = {CORNER};
      method_file =
        yac_new_interpolation_method_file(search, &domain, "file/avg stack",
                                          locations, 1, "weight_file_in.nc", 1);
      method_avg =
        yac_new_interpolation_method_avg_mask(search, &domain, "file/avg stack",
                                              points_set, NULL, 1, 0, 0, 0, 0);

      //---------------------------
      // set up field interpolation
      //---------------------------

      struct field_interpolation_gen * interpolation_gen;
      struct interpolation_method * interpolation_method_stack[] = {method_file,
                                                                    method_avg,
                                                                    NULL};

      interpolation_gen =
        yac_new_field_interpolation_gen_src(points_set, 1, "field 1", search,
                                            "field 2", interpolation_method_stack,
                                            filename_out, 1, mapping_on_source);

      yac_enddef_field_interpolation_gen(interpolation_gen);

      struct field_interpolation * interpolation;

      interpolation = yac_get_field_interpolation(interpolation_gen, "field 1",
                                                  "field 2");

      //---------------------
      // do the interpolation
      //---------------------

      double source_field_point_data[2][12];

      for (unsigned i = 0; i < 2; ++i)
        for (unsigned j = 0; j < 12; ++j)
          source_field_point_data[i][j] =
            (corner_is_on_rank[my_rank][i][j] == -1)?
              ((double)global_corner_ids[my_rank][i][j]):
              ((double)-1);

      double * source_field_subdomain_data[2][1] =
        {{source_field_point_data[0]}, {source_field_point_data[1]}};
      double ** source_data[2] =
        {source_field_subdomain_data[0], source_field_subdomain_data[1]};
      double *** source_data_collection[1] = {source_data};

      yac_put_field(interpolation, source_data_collection, 1);


      //------------------------------
      // check the written weight file
      //------------------------------

      // ensure that the weight file has been written
#ifdef USE_MPI
      MPI_Barrier(MPI_COMM_WORLD);
#endif

      if (my_rank == 0) {
        enum yac_location ref_src_locations[1] = {CORNER};
        enum yac_location ref_tgt_location = CELL;
        check_weight_file(filename_out, ref_src_address, ref_tgt_address,
                          ref_weights, ref_num_links, ref_src_locations,
                          1, (int*)&ref_num_links, ref_tgt_address_fixed,
                          ref_num_fixed_tgt, ref_fixed_values,
                          ref_num_tgt_per_fixed_value, ref_num_fixed_values,
                          ref_tgt_location, "domain 1", "domain 2");
      }

      //---------
      // clean up
      //---------

      yac_delete_field_interpolation_gen(interpolation_gen);
      yac_delete_field_interpolation(interpolation);
      yac_delete_interpolation_method(method_avg);
      yac_delete_interpolation_method(method_file);

      yac_reset_interpolation_method();
   }

   //---------
   // clean up
   //---------

   yac_coupling_field_delete(field);
   yac_global_search_delete(search);
   yac_free_domain(&domain);
   for (unsigned i = 0; i < 2; ++i) {
      yac_free_points(points_corner + i);
      yac_delete_grid(grid[i]);
   }

   // delete weight file
   if (my_rank == 0) {
     unlink(filename_in);
     unlink(filename_out);
   }
}

static void
generate_ref_weights() {

  for (unsigned i = 0; i < 6; ++i) {
    for (unsigned j = 0; j < 6; ++j) {
      ref_src_address[4 * (i * 6 + j) + 0] = 0 + i * 7 + j;
      ref_src_address[4 * (i * 6 + j) + 1] = 1 + i * 7 + j;
      ref_src_address[4 * (i * 6 + j) + 2] = 7 + i * 7 + j;
      ref_src_address[4 * (i * 6 + j) + 3] = 8 + i * 7 + j;
    }
  }
  for (unsigned i = 0; i < ref_num_links; ++i) ref_tgt_address[i] = i / 4;
  for (unsigned i = 0; i < ref_num_links; ++i) ref_weights[i] = 0.25;
  for (unsigned i = 0; i < num_links_file; ++i)
    ref_weights[tgt_address_file[i] * 4 + (i & 3)] = weights_file[i];
}

