#define VERBOSE
/**
 * @file test_gcxlatc.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "test_cxc.h"
#include "search.h"
#include "geometry.h"

void test_gcxlatc(double lon_a, double lat_a, double lon_b, double lat_b,
                  double lon_c, double lat_c, double lon_d, double lat_d,
                  double lon_ref_p, double lat_ref_p,
                  double lon_ref_q, double lat_ref_q, int ref_ret_val);

int main (void) {

   // simple example

   test_gcxlatc(-10, 10, // point a
                 10, 10, // point b
                -10,  0, // point c
                 10,  0, // point d
                 90,  0, // reference point p
                -90,  0, // reference point q
                0);      // reference return value

   // simple example

   test_gcxlatc(-10,  10, // point a
                 10, -10, // point b
                -10,   0, // point c
                 10,   0, // point d
                180,   0, // reference point p
                  0,   0, // reference point q
                10);      // reference return value

   // example from test program

   test_gcxlatc(  0,  90, // point a
                140,  60, // point b
                 90, -45, // point c
                 39, -45, // point d
                140, -45, // reference point p
                -40, -45, // reference point q
                0);       // reference return value

   // both circles do not intersect

   test_gcxlatc(-10, 10, // point a
                 10, 10, // point b
                -10, 70, // point c
                 10, 70, // point d
                -10, 70, // reference point p
                 10, 70, // reference point q
                -1);     // reference return value

   // the great circle only touches the circle of latitude

   test_gcxlatc(-90,  0, // point a
                  0, 45, // point b
                -10, 45, // point c
                 10, 45, // point d
                  0, 45, // reference point p
                  0, 45, // reference point q
                1 + 4);  // reference return value

   // the circle of latitude is on the pole and does not intersect the great circle

   test_gcxlatc(-90,  0, // point a
                  0, 45, // point b
                 10, 90, // point c
                 20, 90, // point d
                 10, 90, // reference point p
                 20, 90, // reference point q
                -1);     // reference return value

   // the circle of latitude is on the pole and does intersect the great circle

   test_gcxlatc(15      , 85, // point a
                180 + 15, 85, // point b
                10      , 90, // point c
                20      , 90, // point d
                      10, 90, // reference point p
                      10, 90, // reference point q
                1 + 4);       // reference return value

   // the circle of latitude is on the pole and does intersect the great circle

   test_gcxlatc(30, 90, // point a
                30, 70, // point b
                10, 90, // point c
                20, 90, // point d
                10, 90, // reference point p
                10, 90, // reference point q
                1 + 4);     // reference return value

   // the great circle edge has a length of zero and has no intersection with
   // the circle of latitude
   {
      struct point ref_p, ref_q;

      yac_loncxlatc((struct edge){.points = {{.lon = 30*rad, .lat = 50*rad},
                                             {.lon = 30*rad, .lat = 70*rad}},
                                  .edge_type = LON_CIRCLE},
                    (struct edge){.points = {{.lon = 10*rad, .lat = 60*rad},
                                             {.lon = 40*rad, .lat = 60*rad}},
                                  .edge_type = LAT_CIRCLE},
                    &ref_p, &ref_q);

      test_gcxlatc(30, 70,                       // point a
                   30, 70,                       // point b
                   10, 60,                       // point c
                   20, 60,                       // point d
                   ref_p.lon*deg, ref_p.lat*deg, // reference point p
                   ref_q.lon*deg, ref_q.lat*deg, // reference point q
                   0);                           // reference return value
   }

   // the great circle edge has a length of zero and an intersection with
   // the circle of latitude
   {
      struct point ref_p, ref_q;

      yac_loncxlatc((struct edge){.points = {{.lon = 15*rad, .lat = -30*rad},
                                             {.lon = 15*rad, .lat =  30*rad}},
                                  .edge_type = LON_CIRCLE},
                    (struct edge){.points = {{.lon = 10*rad, .lat = 10*rad},
                                             {.lon = 20*rad, .lat = 10*rad}},
                                  .edge_type = LAT_CIRCLE},
                    &ref_p, &ref_q);

      test_gcxlatc(15, 10,                       // point a
                   15, 10,                       // point b
                   10, 10,                       // point c
                   20, 10,                       // point d
                   ref_p.lon*deg, ref_p.lat*deg, // reference point p
                   ref_q.lon*deg, ref_q.lat*deg, // reference point q
                   1 + 4);                       // reference return value
   }

   // the latitude circle edge has a length of zero and has no intersection with
   // the great circle edge

   test_gcxlatc(-5, -5, // point a
                 5,  5, // point b
                10, 89, // point c
                10, 89, // point d
                -1, -1, // reference point p
                -1, -1, // reference point q
                -1);    // reference return value

   // the latitude circle edge has a length of zero and has no intersection with
   // the great circle edge

   test_gcxlatc(   -5, -5, // point a
                    5,  5, // point b
                   10,  0, // point c
                   10,  0, // point d
                    0,  0, // reference point p
                0+180,  0, // reference point q
                1);        // reference return value

   // the latitude circle edge has a length of zero and has an intersection with
   // the great circle edge

   test_gcxlatc(   -5, -5, // point a
                    5,  5, // point b
                    0,  0, // point c
                    0,  0, // point d
                    0,  0, // reference point p
                0+180,  0, // reference point q
                1 + 4);    // reference return value

   {
      // example taken from a toy...caused a crash

      double a[3] = {-0.18125814883147034, -0.5578552203226167, -0.80990310324198245};
      double b[3] = {-0.18125814875183416, -0.557855220392383, -0.80990310321175063};
      double c[3] = {-0.18673822182292288, -0.54232717509597339, -0.8191520442889918};
      double d[3] = {-0.1772448664054937, -0.5455036073850148, -0.8191520442889918};
      double p[3], q[3];

      yac_gcxlatc_vec(a, b, c, d, p, q);
   }
   return TEST_EXIT_CODE;
}

void test_gcxlatc(double lon_a, double lat_a, double lon_b, double lat_b,
                  double lon_c, double lat_c, double lon_d, double lat_d,
                  double lon_ref_p, double lat_ref_p,
                  double lon_ref_q, double lat_ref_q, int ref_ret_val) {

   test_cxc(lon_a, lat_a, lon_b, lat_b, lon_c, lat_c, lon_d, lat_d,
            GREAT_CIRCLE, LAT_CIRCLE,
            lon_ref_p, lat_ref_p, lon_ref_q, lat_ref_q, ref_ret_val,
            yac_gcxlatc, yac_gcxlatc_vec);
}
