/**
 * @file test_generate_cubed_sphere.c
 *
 * @copyright Copyright  (C)  2015 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <string.h>

#include "tests.h"
#include "generate_cubed_sphere.h"
#include "vtk_output.h"
#include "grid.h"
#include "geometry.h"

void write_grid_to_file(struct grid * grid, char * name);

int main(void) {

  unsigned n = 20;

   struct grid * cube_grid = generate_cubed_sphere_grid(n);

   if (yac_get_num_grid_cells(cube_grid) != n * n * 6)
     PUT_ERR("ERROR: wrong number of cells\n");

   if (yac_get_num_grid_corners(cube_grid) != n * n * 6 + 2)
     PUT_ERR("ERROR: wrong number of grid vertices\n")

#define WRITE_VTK_GRID_FILE
#ifdef WRITE_VTK_GRID_FILE
   write_grid_to_file(cube_grid, "cubed_sphere");
#endif // WRITE_VTK_GRID_FILE

   yac_delete_grid(cube_grid);

   return TEST_EXIT_CODE;
}

void get_grid_xyz_coordinates(struct grid * grid, double * points) {

   unsigned i, num_corners;

   num_corners = yac_get_num_grid_corners(grid);

   for (i = 0; i < num_corners; ++i)
      LLtoXYZ(yac_get_corner_x_coord(grid, i),
              yac_get_corner_y_coord(grid, i), points + 3 * i);
}

void get_grid_cell_data(struct grid * grid, unsigned * cell_data) {

   unsigned i, j, k;
   unsigned num_grid_cells;

   num_grid_cells = yac_get_num_grid_cells(grid);

   for (i = 0, k = 0; i < num_grid_cells; ++i) {

      unsigned const * curr_cell_corners;
      unsigned num_cell_corners;

      curr_cell_corners = yac_get_cell_corner_indices(grid, i);
      num_cell_corners = yac_get_num_cell_corners(grid, i);

      for (j = 0; j < num_cell_corners; ++j)
         cell_data[k++] = curr_cell_corners[j];
   }
}

void get_equator_point_data(unsigned num_points, double * points) {

   // the equator is approximated by a polygon with num_points corners

   double angle;

   angle = 2.0 * M_PI / (num_points + 1);

   unsigned i;

   for (i = 0; i < num_points; ++i) {

      points[3*i + 0] = sin(angle * i);
      points[3*i + 1] = cos(angle * i);
      points[3*i + 2] = 0;
   }
}

void write_grid_to_file(struct grid * grid, char * name) {

   unsigned num_grid_corners;
   unsigned num_grid_cells;

   num_grid_corners = yac_get_num_grid_corners(grid);
   num_grid_cells = yac_get_num_grid_cells(grid);

   unsigned const num_points_per_great_circle = 71;

   //-------------------------------------
   // generate point data for the vtk file
   // it will contain data for all corners of the grid and the points for the equator
   //-------------------------------------
   double * points;

   points = malloc(3 * (num_grid_corners + num_points_per_great_circle) * sizeof(*points));

   // get the point data of the grid
   get_grid_xyz_coordinates(grid, points);
   // get the point data of the equator
   get_equator_point_data(num_points_per_great_circle, points + 3 * num_grid_corners);

   //------------------------------------
   // generate cell data for the vtk file
   //------------------------------------
   unsigned * num_points_per_cell;
   unsigned * cell_data;

   // has entries for all cells of the grid and one for the equator
   num_points_per_cell = malloc((num_grid_cells + 1) * sizeof(*num_points_per_cell));

   unsigned i, total_num_cell_points;

   // get the number of points per grid cell
   for (i = 0; i < num_grid_cells; ++i)
      num_points_per_cell[i] = yac_get_num_cell_corners(grid, i);

   // get the number of corners for the equator
   num_points_per_cell[num_grid_cells] = num_points_per_great_circle;

   // count the total number of cell points
   total_num_cell_points = 0;
   for (i = 0; i <= num_grid_cells; ++i)
      total_num_cell_points += num_points_per_cell[i];

   cell_data = malloc(total_num_cell_points * sizeof(*cell_data));

   // get the cell data of the grid
   get_grid_cell_data(grid, cell_data);

   // write the cell data for the great circle
   for (i = 0; i < num_points_per_great_circle; ++i)
      cell_data[i + total_num_cell_points - num_points_per_great_circle] = num_grid_corners + i;

   //------------------------------------
   // generate scalar data
   // we write a cell type that differentiates grid cells and the equator
   //------------------------------------
   unsigned * cell_type;

   cell_type = malloc((num_grid_cells + 1) * sizeof(*cell_type));

   for (i = 0; i < num_grid_cells; ++i)
      cell_type[i] = 0;
   cell_type[num_grid_cells] = 1;

   //------------------------------------
   // generate the file name
   //------------------------------------

   char * filename;

   filename = calloc((strlen(name) + 5), sizeof(*filename));
   strcpy(filename, name);
   strcat(filename, ".vtk");

   //------------------------------------
   // generate the actual vtk file
   //------------------------------------
   VTK_FILE * file;

   file = vtk_open(filename, "grid data");
   vtk_write_point_data(file, points, num_grid_corners + num_points_per_great_circle);
   vtk_write_cell_data(file, cell_data, num_points_per_cell, num_grid_cells + 1);
   vtk_write_cell_scalars_uint(file, cell_type, num_grid_cells + 1, "cell_type");
   
   vtk_close(file);

   //------------------------------------
   // some final cleanup
   //------------------------------------

   free(filename);
   free(cell_type);
   free(cell_data);
   free(num_points_per_cell);
   free(points);
}
