#define VERBOSE
/**
 * @file test_global_search.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include <stdlib.h>
#include <stdio.h>

#define VERBOSE

#include "tests.h"

#include "grid.h"
#include "grid_reg2d.h"
#include "communicator.h"
#include "communicator_local.h"
#include "domain.h"
#include "global_search.h"

int check_ids(int const * ids_a, int const * ids_b, unsigned num_ids) {

   unsigned i, j, num_matches = 0;

   for (i = 0; i < num_ids; ++i)
      for (j = 0; j < num_ids; ++j)
         if (ids_a[i] == ids_b[j]) {
            num_matches++;
            break;
         }

   return num_matches == num_ids;
}

int main (void) {

   unsigned i, j, k, l;

   { // one directional global search with two domains on two processes with one subdomain each

      //---------------
      // setup
      //---------------

      // comm worlds on both processes
      struct communicator * comm_p1, * comm_p2;

      comm_p1 = yac_local_comm_new();
      comm_p2 = yac_local_comm_new();

      // named comms for the domains on both processes
      struct communicator * comm_named_p1, * comm_named_p2;

      comm_named_p1 = yac_named_local_comm_new(comm_p1, "P1");
      comm_named_p2 = yac_named_local_comm_new(comm_p2, "P2");

         // define a domain on P1

         struct grid * grid_p1;
         unsigned num_cells[2] = {2,2};
         unsigned cyclic[2] = {0,0};
         double coordinates_x[2][3] = {{-1,0,1}, {-1.5,-0.5,0.5}};
         double coordinates_y[2][3] = {{-1,0,1}, {-1.5,-0.5,0.5}};

         for ( unsigned i = 0; i < 2; ++i ) {
            for ( unsigned j = 0; j < 3; ++j ) {
               coordinates_x[i][j] *= rad;
               coordinates_y[i][j] *= rad;
            }
         }

         grid_p1 = yac_reg2d_grid_new(coordinates_x[0], coordinates_y[0],
                                      num_cells, cyclic);

         struct subdomain subdomain_p1;

         int global_corner_ids_p1[9] = {0,1,2,3,4,5,6,7,8};
         unsigned corner_is_on_rank_p1[9] = {-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_p1[4] = {0,1,2,3};
         unsigned cell_is_on_rank_p1[4] = {-1,-1,-1,-1};
         int global_edge_ids_p1[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned edge_is_on_rank_p1[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_p1, grid_p1, comm_named_p1,
                                 global_corner_ids_p1, corner_is_on_rank_p1,
                                 global_cell_ids_p1, cell_is_on_rank_p1,
                                 global_edge_ids_p1, edge_is_on_rank_p1);

         struct domain l_domain_p1;

         struct subdomain * blocks_p1[1] = {&subdomain_p1};

         yac_init_domain(&l_domain_p1, comm_named_p1, blocks_p1, 1, "domain_p1");

         // define a domain on P2, which matches the one on P1

         struct grid * grid_p2;

         grid_p2 = yac_reg2d_grid_new(coordinates_x[1], coordinates_y[1],
                                      num_cells, cyclic);

         struct subdomain subdomain_p2;

         int global_corner_ids_p2[9] = {0,1,2,3,4,5,6,7,8};
         unsigned corner_is_on_rank_p2[9] = {-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_p2[4] = {0,1,2,3};
         unsigned cell_is_on_rank_p2[4] = {-1,-1,-1,-1};
         int global_edge_ids_p2[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned edge_is_on_rank_p2[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_p2, grid_p2, comm_named_p2,
                                 global_corner_ids_p2, corner_is_on_rank_p2,
                                 global_cell_ids_p2, cell_is_on_rank_p2,
                                 global_edge_ids_p2, edge_is_on_rank_p2);

         struct domain l_domain_p2;

         struct subdomain * blocks_p2[1] = {&subdomain_p2};

         yac_init_domain(&l_domain_p2, comm_named_p2, blocks_p2, 1, "domain_p2");

         // finish domain definition
         yac_enddef_domain(&l_domain_p1);
         yac_enddef_domain(&l_domain_p2);

      //---------
      // testing
      //---------
      // init global search on both processes

      struct global_search * search[2];

      search[0] = yac_global_search_new(&l_domain_p1, "domain_p2", comm_p1, 1,0,
                                        "component_p1", "component_p2");
      search[1] = yac_global_search_new(&l_domain_p2, "domain_p1", comm_p2, 0,1,
                                        "component_p2", "component_p1");

      // finish initialisation of global search

      yac_enddef_global_search(search[0]);
      yac_enddef_global_search(search[1]);

      //---------------------
      // check source results
      //---------------------

      unsigned ref_num_src_per_tgt[4] = {1,2,2,4};
      unsigned ref_src_per_tgt[4][4] = {{0},{0,1},{0,2},{0,1,2,3}};
      int ref_global_tgt_ids[4] = {0,1,2,3};

      struct source_search_results const * src_search_results[4];

      yac_get_search_results_for_tgt_cells(search[0], ref_global_tgt_ids, 4,
                                       src_search_results);

      for (i = 0; i < 4; ++i) {

         struct source_search_results const * curr_result =
            yac_get_search_results_for_tgt_cell(search[0], ref_global_tgt_ids[i]);

         if (src_search_results[i] != curr_result)
            PUT_ERR("error in get_search_results_for_tgt_cells (onedirectional 2 PEs)\n");

         if (curr_result == NULL)
            PUT_ERR("error in get_search_results_for_tgt_cell (onedirectional 2 PEs)\n");

         if (curr_result->global_id != ref_global_tgt_ids[i])
            PUT_ERR("error in get_search_results_for_tgt_cell; wrong global id "
                    "(onedirectional 2 PEs)\n");

         if (curr_result->tgt_rank != 1)
            PUT_ERR("error in get_search_results_for_tgt_cell; wrong rank "
                    "(onedirectional 2 PEs)\n");

         if (curr_result->tgt_domain == NULL)
            PUT_ERR("error in get_search_results_for_tgt_cell; no target domain "
                    "(onedirectional 2 PEs)\n");

         if (curr_result->num_overlapping_src_cells_per_subdomain[0] !=
             ref_num_src_per_tgt[i])
            PUT_ERR("error in get_search_results_for_tgt_cell; wrong number of "
                    "source cells (onedirectional 2 PEs)\n");

         if (!check_ids((int*)(curr_result->overlapping_src_cells[0]),
             (int*)ref_src_per_tgt[i], ref_num_src_per_tgt[i]))
            PUT_ERR("error in get_search_results_for_tgt_cell; wrong "
                    "source cells (onedirectional 2 PEs)\n");

         if (curr_result->num_other_src_ranks != 0)
            PUT_ERR("error in get_search_results_for_tgt_cell; wrong number of "
                    "other source ranks\n");

         if (curr_result->other_src_ranks != NULL)
            PUT_ERR("error in get_search_results_for_tgt_cell; other source ranks\n");
      }

      if (yac_source_process_has_overlap_with(search[0], 0))
         PUT_ERR("error in routine source_process_has_overlap_with\n");

      if (!yac_source_process_has_overlap_with(search[0], 1))
         PUT_ERR("error in routine source_process_has_overlap_with\n");

      //---------------------
      // check target results
      //---------------------

      struct dep_list const * matching_local_cells;

      matching_local_cells = yac_get_matching_local_cell_ids(search[1], 0);

      if (matching_local_cells->num_elements != 1)
         PUT_ERR("error in global search results (onedirectional 2 PEs)\n");

      if (matching_local_cells->num_deps_per_element[0] != 4)
         PUT_ERR("error in global search results (onedirectional 2 PEs)\n");

      unsigned ref_matching_local_cell_ids[4] = {0,1,2,3};

      if (!check_ids((int*)yac_get_dependencies_of_element(*matching_local_cells, 0),
                     (int*)ref_matching_local_cell_ids, 4))
         PUT_ERR("error in global search results (onedirectional 2 PEs)\n");

      if (yac_target_process_has_overlap_with(search[1], 1))
         PUT_ERR("error in routine target_process_has_overlap_with\n");

      if (!yac_target_process_has_overlap_with(search[1], 0))
         PUT_ERR("error in routine target_process_has_overlap_with\n");

      //---------------
      // cleanup
      //---------------
      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);

      yac_free_domain(&l_domain_p1);
      yac_free_domain(&l_domain_p2);

      yac_delete_grid(grid_p1);
      yac_delete_grid(grid_p2);

      yac_free_communicator(comm_p1);
      yac_free_communicator(comm_p2);

      yac_free_communicator(comm_named_p1);
      yac_free_communicator(comm_named_p2);
   }

   { // one directional global search with two domains on two processes with
     // one subdomain each (the domains do not overlap)

      //---------------
      // setup
      //---------------

      // comm worlds on both processes
      struct communicator * comm_p1, * comm_p2;

      comm_p1 = yac_local_comm_new();
      comm_p2 = yac_local_comm_new();

      // named comms for the domains on both processes
      struct communicator * comm_named_p1, * comm_named_p2;

      comm_named_p1 = yac_named_local_comm_new(comm_p1, "P1");
      comm_named_p2 = yac_named_local_comm_new(comm_p2, "P2");

         // define a domain on P1

         struct grid * grid_p1;
         unsigned num_cells[2] = {2,2};
         unsigned cyclic[2] = {0,0};
         double coordinates_x[2][3] = {{-1,0,1}, {-181.5,-180.5,180.5}};
         double coordinates_y[2][3] = {{-1,0,1}, {-1.5,-0.5,0.5}};

         for ( unsigned i = 0; i < 2; ++i ) {
            for ( unsigned j = 0; j < 3; ++j ) {
               coordinates_x[i][j] *= rad;
               coordinates_y[i][j] *= rad;
            }
         }

         grid_p1 = yac_reg2d_grid_new(coordinates_x[0], coordinates_y[0],
                                      num_cells, cyclic);

         struct subdomain subdomain_p1;

         int global_corner_ids_p1[9] = {0,1,2,3,4,5,6,7,8};
         unsigned corner_is_on_rank_p1[9] = {-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_p1[4] = {0,1,2,3};
         unsigned cell_is_on_rank_p1[4] = {-1,-1,-1,-1};
         int global_edge_ids_p1[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned edge_is_on_rank_p1[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_p1, grid_p1, comm_named_p1,
                                 global_corner_ids_p1, corner_is_on_rank_p1,
                                 global_cell_ids_p1, cell_is_on_rank_p1,
                                 global_edge_ids_p1, edge_is_on_rank_p1);

         struct domain l_domain_p1;

         struct subdomain * blocks_p1[1] = {&subdomain_p1};

         yac_init_domain(&l_domain_p1, comm_named_p1, blocks_p1, 1, "domain_p1");

         // define a domain on P2, which matches the one on P1

         struct grid * grid_p2;

         grid_p2 = yac_reg2d_grid_new(coordinates_x[1], coordinates_y[1],
                                      num_cells, cyclic);

         struct subdomain subdomain_p2;

         int global_corner_ids_p2[9] = {0,1,2,3,4,5,6,7,8};
         unsigned corner_is_on_rank_p2[9] = {-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_p2[4] = {0,1,2,3};
         unsigned cell_is_on_rank_p2[4] = {-1,-1,-1,-1};
         int global_edge_ids_p2[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned edge_is_on_rank_p2[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_p2, grid_p2, comm_named_p2,
                                 global_corner_ids_p2, corner_is_on_rank_p2,
                                 global_cell_ids_p2, cell_is_on_rank_p2,
                                 global_edge_ids_p2, edge_is_on_rank_p2);

         struct domain l_domain_p2;

         struct subdomain * blocks_p2[1] = {&subdomain_p2};

         yac_init_domain(&l_domain_p2, comm_named_p2, blocks_p2, 1, "domain_p2");

         // finish domain definition
         yac_enddef_domain(&l_domain_p1);
         yac_enddef_domain(&l_domain_p2);

      //---------
      // testing
      //---------
      // init global search on both processes

      struct global_search * search[2];

      search[0] = yac_global_search_new(&l_domain_p1, "domain_p2", comm_p1, 1,0,
                                        "component_p1", "component_p2");
      search[1] = yac_global_search_new(&l_domain_p2, "domain_p1", comm_p2, 0,1,
                                        "component_p2", "component_p1");

      // finish initialisation of global search

      yac_enddef_global_search(search[0]);
      yac_enddef_global_search(search[1]);

      //---------------------
      // check source results
      //---------------------

      int ref_global_tgt_ids[4] = {0,1,2,3};

      struct source_search_results const * src_search_results[4];

      yac_get_search_results_for_tgt_cells(search[0], ref_global_tgt_ids, 4,
                                       src_search_results);

      for (i = 0; i < 4; ++i) {

         struct source_search_results const * curr_result =
            yac_get_search_results_for_tgt_cell(search[0], ref_global_tgt_ids[i]);

         if (src_search_results[i] != curr_result)
            PUT_ERR("error in get_search_results_for_tgt_cells (onedirectional 2 PEs)\n");

         if (curr_result != NULL)
            PUT_ERR("error in get_search_results_for_tgt_cell (onedirectional 2 PEs)\n");
      }

      if (yac_source_process_has_overlap_with(search[0], 0))
         PUT_ERR("error in routine source_process_has_overlap_with\n");

      if (yac_source_process_has_overlap_with(search[0], 1))
         PUT_ERR("error in routine source_process_has_overlap_with\n");

      //---------------------
      // check target results
      //---------------------

      struct dep_list const * matching_local_cells;

      matching_local_cells = yac_get_matching_local_cell_ids(search[1], 0);

      if (matching_local_cells != NULL)
         PUT_ERR("error in global search results (onedirectional 2 PEs)\n");

      if (yac_target_process_has_overlap_with(search[1], 1))
         PUT_ERR("error in routine target_process_has_overlap_with\n");

      if (yac_target_process_has_overlap_with(search[1], 0))
         PUT_ERR("error in routine target_process_has_overlap_with\n");

      //---------------
      // cleanup
      //---------------
      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);

      yac_free_domain(&l_domain_p1);
      yac_free_domain(&l_domain_p2);

      yac_delete_grid(grid_p1);
      yac_delete_grid(grid_p2);

      yac_free_communicator(comm_p1);
      yac_free_communicator(comm_p2);

      yac_free_communicator(comm_named_p1);
      yac_free_communicator(comm_named_p2);
   }

   { // one directional global search with two domains on two processes with
     // one subdomain each (the domains do not overlap, however their
     // bounding circles should)

      //---------------
      // setup
      //---------------

      // comm worlds on both processes
      struct communicator * comm_p1, * comm_p2;

      comm_p1 = yac_local_comm_new();
      comm_p2 = yac_local_comm_new();

      // named comms for the domains on both processes
      struct communicator * comm_named_p1, * comm_named_p2;

      comm_named_p1 = yac_named_local_comm_new(comm_p1, "P1");
      comm_named_p2 = yac_named_local_comm_new(comm_p2, "P2");

         // define a domain on P1

         struct grid * grid_p1;
         unsigned num_cells[2] = {2,2};
         unsigned cyclic[2] = {0,0};
         double coordinates_x[2][3] = {{0,1,2}, {0,1,2}};
         double coordinates_y[2][3] = {{0,1,2}, {2.0001, 3.0001, 4.0001}};

         for ( unsigned i = 0; i < 2; ++i ) {
            for ( unsigned j = 0; j < 3; ++j ) {
               coordinates_x[i][j] *= rad;
               coordinates_y[i][j] *= rad;
            }
         }

         grid_p1 = yac_reg2d_grid_new(coordinates_x[0], coordinates_y[0],
                                      num_cells, cyclic);

         struct subdomain subdomain_p1;

         int global_corner_ids_p1[9] = {0,1,2,3,4,5,6,7,8};
         unsigned corner_is_on_rank_p1[9] = {-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_p1[4] = {0,1,2,3};
         unsigned cell_is_on_rank_p1[4] = {-1,-1,-1,-1};
         int global_edge_ids_p1[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned edge_is_on_rank_p1[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_p1, grid_p1, comm_named_p1,
                                 global_corner_ids_p1, corner_is_on_rank_p1,
                                 global_cell_ids_p1, cell_is_on_rank_p1,
                                 global_edge_ids_p1, edge_is_on_rank_p1);

         struct domain l_domain_p1;

         struct subdomain * blocks_p1[1] = {&subdomain_p1};

         yac_init_domain(&l_domain_p1, comm_named_p1, blocks_p1, 1, "domain_p1");

         // define a domain on P2, which matches the one on P1

         struct grid * grid_p2;

         grid_p2 = yac_reg2d_grid_new(coordinates_x[1], coordinates_y[1],
                                      num_cells, cyclic);

         struct subdomain subdomain_p2;

         int global_corner_ids_p2[9] = {0,1,2,3,4,5,6,7,8};
         unsigned corner_is_on_rank_p2[9] = {-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_p2[4] = {0,1,2,3};
         unsigned cell_is_on_rank_p2[4] = {-1,-1,-1,-1};
         int global_edge_ids_p2[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned edge_is_on_rank_p2[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_p2, grid_p2, comm_named_p2,
                                 global_corner_ids_p2, corner_is_on_rank_p2,
                                 global_cell_ids_p2, cell_is_on_rank_p2,
                                 global_edge_ids_p2, edge_is_on_rank_p2);

         struct domain l_domain_p2;

         struct subdomain * blocks_p2[1] = {&subdomain_p2};

         yac_init_domain(&l_domain_p2, comm_named_p2, blocks_p2, 1, "domain_p2");

         // finish domain definition
         yac_enddef_domain(&l_domain_p1);
         yac_enddef_domain(&l_domain_p2);

      //---------
      // testing
      //---------
      // init global search on both processes

      struct global_search * search[2];

      search[0] = yac_global_search_new(&l_domain_p1, "domain_p2", comm_p1, 1,0,
                                        "component_p1", "component_p2");
      search[1] = yac_global_search_new(&l_domain_p2, "domain_p1", comm_p2, 0,1,
                                        "component_p2", "component_p1");

      // finish initialisation of global search

      yac_enddef_global_search(search[0]);
      yac_enddef_global_search(search[1]);

      //---------------------
      // check source results
      //---------------------

      int ref_global_tgt_ids[4] = {0,1,2,3};

      struct source_search_results const * src_search_results[4];

      yac_get_search_results_for_tgt_cells(search[0], ref_global_tgt_ids, 4,
                                       src_search_results);

      for (i = 0; i < 4; ++i) {

         struct source_search_results const * curr_result =
            yac_get_search_results_for_tgt_cell(search[0], ref_global_tgt_ids[i]);

         if (src_search_results[i] != curr_result)
            PUT_ERR("error in get_search_results_for_tgt_cells (onedirectional 2 PEs)\n");

         if (curr_result != NULL)
            PUT_ERR("error in get_search_results_for_tgt_cell (onedirectional 2 PEs)\n");
      }

      if (yac_source_process_has_overlap_with(search[0], 0))
         PUT_ERR("error in routine source_process_has_overlap_with\n");

      if (yac_source_process_has_overlap_with(search[0], 1))
         PUT_ERR("error in routine source_process_has_overlap_with\n");

      //---------------------
      // check target results
      //---------------------

      struct dep_list const * matching_local_cells;

      matching_local_cells = yac_get_matching_local_cell_ids(search[1], 0);

      if (matching_local_cells != NULL)
         PUT_ERR("error in global search results (onedirectional 2 PEs)\n");

      if (yac_target_process_has_overlap_with(search[1], 1))
         PUT_ERR("error in routine target_process_has_overlap_with\n");

      if (yac_target_process_has_overlap_with(search[1], 0))
         PUT_ERR("error in routine target_process_has_overlap_with\n");

      //---------------
      // cleanup
      //---------------
      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);

      yac_free_domain(&l_domain_p1);
      yac_free_domain(&l_domain_p2);

      yac_delete_grid(grid_p1);
      yac_delete_grid(grid_p2);

      yac_free_communicator(comm_p1);
      yac_free_communicator(comm_p2);

      yac_free_communicator(comm_named_p1);
      yac_free_communicator(comm_named_p2);
   }

   { // bidirectional global search with two domains on two processes with one subdomain each

      //---------------
      // setup
      //---------------

      // comm worlds on both processes
      struct communicator * comm_p1, * comm_p2;

      comm_p1 = yac_local_comm_new();
      comm_p2 = yac_local_comm_new();

      // named comms for the domains on both processes
      struct communicator * comm_named_p1, * comm_named_p2;

      comm_named_p1 = yac_named_local_comm_new(comm_p1, "P1");
      comm_named_p2 = yac_named_local_comm_new(comm_p2, "P2");

         // define a domain on P1

         struct grid * grid_p1;
         unsigned num_cells[2] = {2,2};
         unsigned cyclic[2] = {0,0};
         double coordinates_x[2][3] = {{-1,0,1}, {-1.5,-0.5,0.5}};
         double coordinates_y[2][3] = {{-1,0,1}, {-1.5,-0.5,0.5}};

         for ( unsigned i = 0; i < 2; ++i ) {
            for ( unsigned j = 0; j < 3; ++j ) {
               coordinates_x[i][j] *= rad;
               coordinates_y[i][j] *= rad;
            }
         }

         grid_p1 = yac_reg2d_grid_new(coordinates_x[0], coordinates_y[0],
                                      num_cells, cyclic);

         struct subdomain subdomain_p1;

         int global_corner_ids_p1[9] = {0,1,2,3,4,5,6,7,8};
         unsigned corner_is_on_rank_p1[9] = {-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_p1[4] = {0,1,2,3};
         unsigned cell_is_on_rank_p1[4] = {-1,-1,-1,-1};
         int global_edge_ids_p1[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned edge_is_on_rank_p1[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_p1, grid_p1, comm_named_p1,
                                 global_corner_ids_p1, corner_is_on_rank_p1,
                                 global_cell_ids_p1, cell_is_on_rank_p1,
                                 global_edge_ids_p1, edge_is_on_rank_p1);

         struct domain l_domain_p1;

         struct subdomain * blocks_p1[1] = {&subdomain_p1};

         yac_init_domain(&l_domain_p1, comm_named_p1, blocks_p1, 1, "domain_p1");

         // define a domain on P2, which matches the one on P1

         struct grid * grid_p2;

         grid_p2 = yac_reg2d_grid_new(coordinates_x[1], coordinates_y[1],
                                      num_cells, cyclic);

         struct subdomain subdomain_p2;

         int global_corner_ids_p2[9] = {0,1,2,3,4,5,6,7,8};
         unsigned corner_is_on_rank_p2[9] = {-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_p2[4] = {0,1,2,3};
         unsigned cell_is_on_rank_p2[4] = {-1,-1,-1,-1};
         int global_edge_ids_p2[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned edge_is_on_rank_p2[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_p2, grid_p2, comm_named_p2,
                                 global_corner_ids_p2, corner_is_on_rank_p2,
                                 global_cell_ids_p2, cell_is_on_rank_p2,
                                 global_edge_ids_p2, edge_is_on_rank_p2);

         struct domain l_domain_p2;

         struct subdomain * blocks_p2[1] = {&subdomain_p2};

         yac_init_domain(&l_domain_p2, comm_named_p2, blocks_p2, 1, "domain_p2");

         // finish domain definition
         yac_enddef_domain(&l_domain_p1);
         yac_enddef_domain(&l_domain_p2);

      //---------------
      // testing
      //---------------
      // init global search on both processes

      struct global_search * search[2];

      search[0] = yac_global_search_new(&l_domain_p1, "domain_p2", comm_p1, 1,1,
                                        "component_p1", "component_p2");
      search[1] = yac_global_search_new(&l_domain_p2, "domain_p1", comm_p2, 1,1,
                                        "component_p2", "component_p1");

      // finish initialisation of global search

      yac_enddef_global_search(search[0]);
      yac_enddef_global_search(search[1]);

      //---------------------
      // check source results
      //---------------------

      unsigned ref_num_src_per_tgt[2][4] = {{1,2,2,4},{4,2,2,1}};
      unsigned ref_src_per_tgt[2][4][4] = {{{0},{0,1},{0,2},{0,1,2,3}},
                                           {{0,1,2,3},{1,3},{2,3},{3}}};
      int ref_global_tgt_ids[4] = {0,1,2,3};

      for (k = 0; k < 2; ++k) {

         struct source_search_results const * src_search_results[4];

         yac_get_search_results_for_tgt_cells(search[k], ref_global_tgt_ids, 4,
                                          src_search_results);

         for (i = 0; i < 4; ++i) {

            struct source_search_results const * curr_result =
               yac_get_search_results_for_tgt_cell(search[k], ref_global_tgt_ids[i]);

            if (src_search_results[i] != curr_result)
               PUT_ERR("error in get_search_results_for_tgt_cells (bidirectional 2 PEs)\n");

            if (curr_result == NULL)
               PUT_ERR("error in get_search_results_for_tgt_cell (bidirectional 2 PEs)\n");

            if (curr_result->global_id != ref_global_tgt_ids[i])
               PUT_ERR("error in get_search_results_for_tgt_cell; wrong global id "
                       "(bidirectional 2 PEs)\n");

            if (curr_result->tgt_rank != (k^1))
               PUT_ERR("error in get_search_results_for_tgt_cell; wrong rank "
                       "(bidirectional 2 PEs)\n");

            if (curr_result->tgt_domain == NULL)
               PUT_ERR("error in get_search_results_for_tgt_cell; no target domain "
                       "(bidirectional 2 PEs)\n");

            if (curr_result->num_overlapping_src_cells_per_subdomain[0] !=
                ref_num_src_per_tgt[k][i])
               PUT_ERR("error in get_search_results_for_tgt_cell; wrong number of "
                       "source cells (bidirectional 2 PEs)\n");

            if (!check_ids((int*)(curr_result->overlapping_src_cells[0]),
                (int*)ref_src_per_tgt[k][i], ref_num_src_per_tgt[k][i]))
               PUT_ERR("error in get_search_results_for_tgt_cell; wrong "
                       "source cells (bidirectional 2 PEs)\n");

            if (curr_result->num_other_src_ranks != 0)
               PUT_ERR("error in get_search_results_for_tgt_cell; wrong number of "
                       "other source ranks\n");

            if (curr_result->other_src_ranks != NULL)
               PUT_ERR("error in get_search_results_for_tgt_cell; other source ranks\n");
         }

         if (!yac_source_process_has_overlap_with(search[k], k^1))
            PUT_ERR("error in routine source_process_has_overlap_with\n");

         if (yac_source_process_has_overlap_with(search[k], k))
            PUT_ERR("error in routine source_process_has_overlap_with\n");
      }

      //---------------------
      // check target results
      //---------------------

      struct dep_list const * matching_local_cells;

      for (k = 0; k < 2; ++k) {

         matching_local_cells = yac_get_matching_local_cell_ids(search[k], k^1);

         if (matching_local_cells->num_elements != 1)
            PUT_ERR("error in global search results (bidirectional 2 PEs)\n");

         if (matching_local_cells->num_deps_per_element[0] != 4)
            PUT_ERR("error in global search results (bidirectional 2 PEs)\n");

         unsigned ref_matching_local_cell_ids[4] = {0,1,2,3};

         if (!check_ids((int*)yac_get_dependencies_of_element(*matching_local_cells, 0),
                        (int*)ref_matching_local_cell_ids, 4))
            PUT_ERR("error in global search results (bidirectional 2 PEs)\n");

         if (!yac_target_process_has_overlap_with(search[k], k^1))
            PUT_ERR("error in routine target_process_has_overlap_with\n");

         if (yac_target_process_has_overlap_with(search[k], k))
            PUT_ERR("error in routine target_process_has_overlap_with\n");
      }

      //---------------
      // cleanunp
      //---------------

      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);

      yac_free_domain(&l_domain_p1);
      yac_free_domain(&l_domain_p2);

      yac_delete_grid(grid_p1);
      yac_delete_grid(grid_p2);

      yac_free_communicator(comm_p1);
      yac_free_communicator(comm_p2);

      yac_free_communicator(comm_named_p1);
      yac_free_communicator(comm_named_p2);
   }

   { // one directional global search with three domains (2 source and 1 target) on three processes with one subdomain each

      //---------------
      // setup
      //---------------

      // comm worlds on all processes
      struct communicator * comm_source_p1, * comm_source_p2, * comm_target;

      comm_source_p1 = yac_local_comm_new();
      comm_source_p2 = yac_local_comm_new();
      comm_target = yac_local_comm_new();

      // named comms for the domains on all processes
      struct communicator * comm_named_source_p1, * comm_named_source_p2,
                          * comm_named_target;

      comm_named_source_p1 = yac_named_local_comm_new(comm_source_p1, "source");
      comm_named_source_p2 = yac_named_local_comm_new(comm_source_p2, "source");
      comm_named_target = yac_named_local_comm_new(comm_target, "target");

         // define the two source domains

         struct grid * grid_source_p1, * grid_source_p2;
         unsigned num_cells[2] = {3,4};
         unsigned cyclic[2] = {0,0};
         double coordinates_source_x[2][4] = {{0,1,2,3}, {1,2,3,4}};
         double coordinates_source_y[2][5] = {{0,1,2,3,4}, {0,1,2,3,4}};

         for ( unsigned i = 0; i < 2; ++i )
            for ( unsigned j = 0; j < 4; ++j )
               coordinates_source_x[i][j] *= rad;

         for ( unsigned i = 0; i < 2; ++i )
            for ( unsigned j = 0; j < 5; ++j )
               coordinates_source_y[i][j] *= rad;

         grid_source_p1 = yac_reg2d_grid_new(coordinates_source_x[0],
                                             coordinates_source_y[0], num_cells, cyclic);
         grid_source_p2 = yac_reg2d_grid_new(coordinates_source_x[1],
                                             coordinates_source_y[1], num_cells, cyclic);

         struct subdomain subdomain_source_p1, subdomain_source_p2;

         int global_corner_ids_source[2][20] = {{0,1,2,3, 5,6,7,8, 10,11,12,13, 15,16,17,18, 20,21,22,23},
                                                {1,2,3,4, 6,7,8,9, 11,12,13,14, 16,17,18,19, 21,22,23,24}};
         unsigned corner_is_on_rank_source[2][20] = {{-1,-1,-1,1, -1,-1,-1,1, -1,-1,-1,1, -1,-1,-1,1, -1,-1,-1,1},
                                                    {0,-1,-1,-1, 0,-1,-1,-1, 0,-1,-1,-1, 0,-1,-1,-1, 0,-1,-1,-1}};
         int global_cell_ids_source[2][12] = {{0,1,2, 4,5,6, 8,9,10, 12,13,14},
                                              {1,2,3, 5,6,7, 9,10,11, 13,14,15}};
         unsigned cell_is_on_rank_source[2][12] = {{-1,-1,1, -1,-1,1, -1,-1,1, -1,-1,1},
                                                   {0,-1,-1, 0,-1,-1, 0,-1,-1, 0,-1,-1}};
         int *global_edge_ids_source = NULL;
         unsigned *edge_is_on_rank_source = NULL;

         yac_init_subdomain_unstruct(&subdomain_source_p1, grid_source_p1, comm_named_source_p1,
                                 global_corner_ids_source[0], corner_is_on_rank_source[0],
                                 global_cell_ids_source[0], cell_is_on_rank_source[0],
                                 global_edge_ids_source, edge_is_on_rank_source);
         yac_init_subdomain_unstruct(&subdomain_source_p2, grid_source_p2, comm_named_source_p2,
                                 global_corner_ids_source[1], corner_is_on_rank_source[1],
                                 global_cell_ids_source[1], cell_is_on_rank_source[1],
                                 global_edge_ids_source, edge_is_on_rank_source);

         struct domain l_domain_source_p1, l_domain_source_p2;

         struct subdomain * blocks_source[2][1] = {{&subdomain_source_p1},{&subdomain_source_p2}};

         yac_init_domain(&l_domain_source_p1, comm_named_source_p1, blocks_source[0], 1, "source");
         yac_init_domain(&l_domain_source_p2, comm_named_source_p2, blocks_source[1], 1, "source");

         // define the target domain

         struct grid * grid_target;
         unsigned num_cells_target[2] = {3,3};
         unsigned cyclic_target[2] = {0,0};
         double coordinates_target_x[4] = {0.5,1.5,2.5,3.5};
         double coordinates_target_y[4] = {0.5,1.5,2.5,3.5};

         for ( i = 0; i < 4; ++i ) {
            coordinates_target_x[i] *= rad;
            coordinates_target_y[i] *= rad;
         }

         grid_target = yac_reg2d_grid_new(coordinates_target_x, coordinates_target_y,
                                          num_cells_target, cyclic_target);

         struct subdomain subdomain_target;

         int global_corner_ids_target[16] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15};
         unsigned corner_is_on_rank_target[16] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_target[9] = {0,1,2,3,4,5,6,7,8};
         unsigned cell_is_on_rank_target[9] = {-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int *global_edge_ids_target = NULL;
         unsigned *edge_is_on_rank_target = NULL;

         yac_init_subdomain_unstruct(&subdomain_target, grid_target, comm_named_target,
                                 global_corner_ids_target, corner_is_on_rank_target,
                                 global_cell_ids_target, cell_is_on_rank_target,
                                 global_edge_ids_target, edge_is_on_rank_target);

         struct domain l_domain_target;

         struct subdomain * blocks_target[1] = {&subdomain_target};

         yac_init_domain(&l_domain_target, comm_named_target, blocks_target, 1, "target");

         // finish domain definition
         yac_enddef_domain(&l_domain_source_p1);
         yac_enddef_domain(&l_domain_source_p2);
         yac_enddef_domain(&l_domain_target);

      //---------------
      // testing
      //---------------
      // init global search on all processes

      struct global_search * search[3];

      search[0] = yac_global_search_new(&l_domain_source_p1, "target", comm_source_p1, 1,0,
                                        "component_source", "component_target");
      search[1] = yac_global_search_new(&l_domain_source_p2, "target", comm_source_p2, 1,0,
                                        "component_source", "component_target");
      search[2] = yac_global_search_new(&l_domain_target, "source", comm_target, 0,1,
                                        "component_target", "component_source");

      // finish initialisation of global search

      yac_enddef_global_search(search[0]);
      yac_enddef_global_search(search[1]);
      yac_enddef_global_search(search[2]);

      //---------------------
      // check source results
      //---------------------

      unsigned ref_num_src_per_tgt[2][9] = {{4,4,2, 4,4,2, 4,4,2},
                                            {2,4,4, 2,4,4, 2,4,4}};
      unsigned ref_src_per_tgt[2][9][4] = {{{0,1,3,4},{1,2,4,5},{2,5}, {3,4,6,7},{4,5,7,8},{5,8}, {6,7,9,10},{7,8,10,11},{8,11}},
                                           {{0,3},{0,1,3,4},{1,2,4,5}, {3,6},{3,4,6,7},{4,5,7,8}, {6,9},{6,7,9,10},{7,8,10,11}}};
      int ref_global_tgt_ids[2][9] = {{0,1,2,3,4,5,6,7,8},{0,1,2,3,4,5,6,7,8}};
      unsigned ref_num_other_src_ranks[2][9] = {{1,1,1,1,1,1,1,1,1},{1,1,1,1,1,1,1,1,1}};
      unsigned ref_other_src_ranks[2][9] = {{1,1,1,1,1,1,1,1,1},{0,0,0,0,0,0,0,0,0}};

      for (k = 0; k < 2; ++k) {

         struct source_search_results const * src_search_results[9];

         yac_get_search_results_for_tgt_cells(search[k], ref_global_tgt_ids[k], 9,
                                          src_search_results);

         for (i = 0; i < 9; ++i) {

            struct source_search_results const * curr_result =
               yac_get_search_results_for_tgt_cell(search[k], ref_global_tgt_ids[k][i]);

            if (src_search_results[i] != curr_result)
               PUT_ERR("error in get_search_results_for_tgt_cells (onedirectional 2 PEs)\n");

            if (curr_result == NULL)
               PUT_ERR("error in get_search_results_for_tgt_cell (onedirectional 2 PEs)\n");

            if (curr_result->global_id != ref_global_tgt_ids[k][i])
               PUT_ERR("error in get_search_results_for_tgt_cell; wrong global id "
                       "(onedirectional 2 PEs)\n");

            if (curr_result->tgt_rank != (2))
               PUT_ERR("error in get_search_results_for_tgt_cell; wrong rank "
                       "(onedirectional 2 PEs)\n");

            if (curr_result->tgt_domain == NULL)
               PUT_ERR("error in get_search_results_for_tgt_cell; no target domain "
                       "(onedirectional 2 PEs)\n");

            if (curr_result->num_overlapping_src_cells_per_subdomain[0] !=
                ref_num_src_per_tgt[k][i])
               PUT_ERR("error in get_search_results_for_tgt_cell; wrong number of "
                       "source cells (onedirectional 2 PEs)\n");

            if (!check_ids((int*)(curr_result->overlapping_src_cells[0]),
                (int*)ref_src_per_tgt[k][i], ref_num_src_per_tgt[k][i]))
               PUT_ERR("error in get_search_results_for_tgt_cell; wrong "
                       "source cells (onedirectional 2 PEs)\n");

            if (curr_result->num_other_src_ranks != ref_num_other_src_ranks[k][i])
               PUT_ERR("error in get_search_results_for_tgt_cell; "
                       "wrong number of other source ranks (onedirectional 2 PEs)\n");

            if (ref_num_other_src_ranks[k][i] == 0) {
               if (curr_result->other_src_ranks != NULL)
                  PUT_ERR("error in get_search_results_for_tgt_cell; wrong "
                          "other source ranks (onedirectional 2 PEs)\n");
            } else {
               if (curr_result->other_src_ranks[0] != ref_other_src_ranks[k][i])
                  PUT_ERR("error in get_search_results_for_tgt_cell; wrong "
                          "other source ranks (onedirectional 2 PEs)\n");
            }
         }

         if (yac_source_process_has_overlap_with(search[k], k))
            PUT_ERR("error in routine source_process_has_overlap_with\n");

         if (yac_source_process_has_overlap_with(search[k], k^1))
            PUT_ERR("error in routine source_process_has_overlap_with\n");

         if (!yac_source_process_has_overlap_with(search[k], 2))
            PUT_ERR("error in routine source_process_has_overlap_with\n");

      }

      //---------------------
      // check target results
      //---------------------

      struct dep_list const * matching_local_cells;

      for (k = 0; k < 2; ++k) {

         matching_local_cells = yac_get_matching_local_cell_ids(search[2], k);

         if (matching_local_cells->num_elements != 1)
            PUT_ERR("error in global search results (onedirectional 3 PEs)\n");

         if (matching_local_cells->num_deps_per_element[0] != 9)
            PUT_ERR("error in global search results (onedirectional 3 PEs)\n");

         unsigned ref_matching_local_cell_ids[2][9] = {{0,1,2,3,4,5,6,7,8},
                                                       {0,1,2,3,4,5,6,7,8}};

         if (!check_ids((int*)yac_get_dependencies_of_element(*matching_local_cells, 0),
                        (int*)ref_matching_local_cell_ids[k], 6))
            PUT_ERR("error in global search results (onedirectional 3 PEs)\n");

         if (!yac_target_process_has_overlap_with(search[2], k))
            PUT_ERR("error in routine target_process_has_overlap_with\n");
      }

      if (yac_target_process_has_overlap_with(search[2], 2))
         PUT_ERR("error in routine target_process_has_overlap_with\n");


      //---------------
      // cleanup
      //---------------
      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);
      yac_global_search_delete(search[2]);

      yac_free_domain(&l_domain_source_p1);
      yac_free_domain(&l_domain_source_p2);
      yac_free_domain(&l_domain_target);

      yac_delete_grid(grid_source_p1);
      yac_delete_grid(grid_source_p2);
      yac_delete_grid(grid_target);

      yac_free_communicator(comm_source_p1);
      yac_free_communicator(comm_source_p2);
      yac_free_communicator(comm_target);

      yac_free_communicator(comm_named_source_p1);
      yac_free_communicator(comm_named_source_p2);
      yac_free_communicator(comm_named_target);
   }

   { // bidirectional global search with two domains (each having two subdomains) on two processes

      //---------------
      // setup
      //---------------

      // comm worlds on both processes
      struct communicator * comm[2];

      comm[0] = yac_local_comm_new();
      comm[1] = yac_local_comm_new();

      // named comms for the domains on both processes
      struct communicator * comm_named[2];

      comm_named[0] = yac_named_local_comm_new(comm[0], "P1");
      comm_named[1] = yac_named_local_comm_new(comm[1], "P2");

         // define the domain on process 1

         struct grid * grid_p1[2];
         unsigned num_cells_p1[2] = {2,4};
         unsigned cyclic_p1[2] = {0,0};
         double coordinates_x_p1[2][3] = {{0,1,2},{3,4,5}};
         double coordinates_y_p1[2][5] = {{0,1,2,3,4},{0,1,2,3,4}};

         for ( unsigned i = 0; i < 2; ++i )
            for ( unsigned j = 0; j < 3; ++j )
               coordinates_x_p1[i][j] *= rad;

         for ( unsigned i = 0; i < 2; ++i )
            for ( unsigned j = 0; j < 5; ++j )
               coordinates_y_p1[i][j] *= rad;

         for (i = 0; i < 2; ++i)
            grid_p1[i] = yac_reg2d_grid_new(coordinates_x_p1[i], coordinates_y_p1[i],
                                            num_cells_p1, cyclic_p1);

         struct subdomain subdomain_p1[2];

         int global_corner_ids_p1[2][15] = {{0,1,2,3,4,5,6,7,8,9,10,11,12,13,14},
                                            {15,16,17,18,19,20,21,22,23,24,25,26,27,28,29}};
         unsigned corner_is_on_rank_p1[2][15] = {{-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                                                 {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}};
         int global_cell_ids_p1[2][8] = {{0,1,2,3,4,5,6,7},
                                         {8,9,10,11,12,13,14,15}};
         unsigned cell_is_on_rank_p1[2][8] = {{-1,-1,-1,-1,-1,-1,-1,-1},
                                              {-1,-1,-1,-1,-1,-1,-1,-1}};
         int *global_edge_ids_p1 = NULL;
         unsigned *edge_is_on_rank_p1 = NULL;

         for (i = 0; i < 2; ++i)
            yac_init_subdomain_unstruct(subdomain_p1+i, grid_p1[i], comm_named[0],
                                    global_corner_ids_p1[i], corner_is_on_rank_p1[i],
                                    global_cell_ids_p1[i], cell_is_on_rank_p1[i],
                                    global_edge_ids_p1, edge_is_on_rank_p1);

         struct domain domain_p1;

         struct subdomain * subdomains_p1[2] = {subdomain_p1+0,subdomain_p1+1};

         yac_init_domain(&domain_p1, comm_named[0], subdomains_p1, 2, "p1 domain");

         // define the domain on process 2

         struct grid * grid_p2[2];
         unsigned num_cells_p2[2] = {4,2};
         unsigned cyclic_p2[2] = {0,0};
         double coordinates_x_p2[2][5] = {{0.5,1.5,2.5,3.5,4.5},{0.5,1.5,2.5,3.5,4.5}};
         double coordinates_y_p2[2][3] = {{-0.5,0.5,1.5},{2.5,3.5,4.5}};

         for ( i = 0; i < 2; ++i )
            for ( j = 0; j < 5; ++j )
               coordinates_x_p2[i][j] *= rad;

         for ( i = 0; i < 2; ++i )
            for ( j = 0; j < 3; ++j )
               coordinates_y_p2[i][j] *= rad;

         for (i = 0; i < 2; ++i)
            grid_p2[i] = yac_reg2d_grid_new(coordinates_x_p2[i], coordinates_y_p2[i],
                                            num_cells_p2, cyclic_p2);

         struct subdomain subdomain_p2[2];

         int global_corner_ids_p2[2][15] = {{0,1,2,3,4,5,6,7,8,9,10,11,12,13,14},
                                            {15,16,17,18,19,20,21,22,23,24,25,26,27,28,29}};
         unsigned corner_is_on_rank_p2[2][15] = {{-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                                                 {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}};
         int global_cell_ids_p2[2][8] = {{0,1,2,3,4,5,6,7},
                                         {8,9,10,11,12,13,14,15}};
         unsigned cell_is_on_rank_p2[2][8] = {{-1,-1,-1,-1,-1,-1,-1,-1},
                                              {-1,-1,-1,-1,-1,-1,-1,-1}};
         int *global_edge_ids_p2 = NULL;
         unsigned *edge_is_on_rank_p2 = NULL;

         for (i = 0; i < 2; ++i)
            yac_init_subdomain_unstruct(subdomain_p2+i, grid_p2[i], comm_named[1],
                                    global_corner_ids_p2[i], corner_is_on_rank_p2[i],
                                    global_cell_ids_p2[i], cell_is_on_rank_p2[i],
                                    global_edge_ids_p2, edge_is_on_rank_p2);

         struct domain domain_p2;

         struct subdomain * subdomains_p2[2] = {subdomain_p2+0,subdomain_p2+1};

         yac_init_domain(&domain_p2, comm_named[1], subdomains_p2, 2, "p2 domain");

         // finish domain definition
         yac_enddef_domain(&domain_p1);
         yac_enddef_domain(&domain_p2);

      //---------------
      // testing
      //---------------
      // init global search on all processes

      struct global_search * search[2];

      search[0] = yac_global_search_new(&domain_p1, "p2 domain", comm[0], 1,1,
                                        "p1 component", "p2 component");
      search[1] = yac_global_search_new(&domain_p2, "p1 domain", comm[1], 1,1,
                                        "p2 component", "p1 component");

      // finish initialisation of global search

      yac_enddef_global_search(search[0]);
      yac_enddef_global_search(search[1]);

      //---------------------
      // check source results
      //---------------------

      unsigned ref_num_src_per_tgt[2][2][16] = {{{2,1,0,0,4,2,0,0,4,2,0,0,2,1,0,0},
                                                 {0,0,1,2,0,0,2,4,0,0,2,4,0,0,1,2}},
                                                {{2,4,1,2,0,0,0,0,4,2,2,1,0,0,0,0},
                                                 {0,0,0,0,1,2,2,4,0,0,0,0,2,1,4,2}}};
      unsigned ref_src_per_tgt[2][2][16][4] = {{{{0,1},{1},{-1},{-1},{0,1,2,3},{1,3},{-1},{-1},
                                                {4,5,6,7},{5,7},{-1},{-1},{6,7},{7},{-1},{-1}},
                                               {{-1},{-1},{0},{0,1},{-1},{-1},{0,2},{0,1,2,3},
                                                {-1},{-1},{4,6},{4,5,6,7},{-1},{-1},{6},{6,7}}},
                                              {{{0,4},{0,1,4,5},{4},{4,5},{-1},{-1},{-1},{-1},
                                                {2,3,6,7},{3,7},{6,7},{7},{-1},{-1},{-1},{-1}},
                                               {{-1},{-1},{-1},{-1},{0},{0,1},{0,4},{0,1,4,5},
                                                {-1},{-1},{-1},{-1},{2,3},{3},{2,3,6,7},{3,7}}}};
      int ref_global_tgt_ids[16] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11,
                                    12, 13, 14, 15};

      // for both processes
      for (k = 0; k < 2; ++k) {

         struct source_search_results const * src_search_results[16];

         yac_get_search_results_for_tgt_cells(search[k], ref_global_tgt_ids, 16,
                                          src_search_results);

         // for all target points
         for (j = 0; j < 16; ++j) {

            struct source_search_results const * curr_result =
               yac_get_search_results_for_tgt_cell(search[k], j);

            if (curr_result != src_search_results[j])
                  PUT_ERR("error in get_search_results_for_tgt_cells "
                          "(bidirectional 2 PEs 2 subdomains)\n");

            if ((ref_num_src_per_tgt[k][0][j] == 0) && (ref_num_src_per_tgt[k][1][j] == 0)) {

               if (curr_result != NULL)
                  PUT_ERR("error in get_search_results_for_tgt_cell "
                          "(bidirectional 2 PEs 2 subdomains)\n");

            } else {

               if (curr_result == NULL)
                  PUT_ERR("error in get_search_results_for_tgt_cell "
                          "(bidirectional 2 PEs 2 subdomains)\n");

               if (curr_result->global_id != j)
                  PUT_ERR("error in get_search_results_for_tgt_cell; wrong global id "
                          "(bidirectional 2 PEs 2 subdomains)\n");

               if (curr_result->tgt_rank != (k^1))
                  PUT_ERR("error in get_search_results_for_tgt_cell; wrong rank "
                          "(bidirectional 2 PEs 2 subdomains)\n");

               if (curr_result->tgt_domain == NULL)
                  PUT_ERR("error in get_search_results_for_tgt_cell; no target domain "
                          "(bidirectional 2 PEs 2 subdomains)\n");

               if (curr_result->num_other_src_ranks != 0)
                  PUT_ERR("error in get_search_results_for_tgt_cell; wrong number "
                          "of other source ranks(bidirectional 2 PEs 2 subdomains)\n");

               if (curr_result->other_src_ranks != NULL)
                  PUT_ERR("error in get_search_results_for_tgt_cell; wrong "
                          "other source ranks(bidirectional 2 PEs 2 subdomains)\n");

               // for all subdomains
               for (i = 0; i < 2; ++i) {

                  if (curr_result->num_overlapping_src_cells_per_subdomain[i] !=
                      ref_num_src_per_tgt[k][i][j])
                     PUT_ERR("error in get_search_results_for_tgt_cell; wrong number of "
                             "source cells (bidirectional 2 PEs 2 subdomains)\n");

                  if (!check_ids((int*)(curr_result->overlapping_src_cells[i]),
                      (int*)ref_src_per_tgt[k][i][j], ref_num_src_per_tgt[k][i][j]))
                     PUT_ERR("error in get_search_results_for_tgt_cell; wrong "
                             "source cells (bidirectional 2 PEs 2 subdomains)\n");
               }
            }
         }

         if (!yac_source_process_has_overlap_with(search[k], k^1))
            PUT_ERR("error in routine source_process_has_overlap_with\n");

         if (yac_source_process_has_overlap_with(search[k], k))
            PUT_ERR("error in routine source_process_has_overlap_with\n");
      }

      //---------------
      // cleanup
      //---------------
      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);

      yac_free_domain(&domain_p1);
      yac_free_domain(&domain_p2);

      yac_delete_grid(grid_p1[0]);
      yac_delete_grid(grid_p1[1]);
      yac_delete_grid(grid_p2[0]);
      yac_delete_grid(grid_p2[1]);

      yac_free_communicator(comm[0]);
      yac_free_communicator(comm[1]);

      yac_free_communicator(comm_named[0]);
      yac_free_communicator(comm_named[1]);
   }

   { // one directional global search with two domains on two processes with one subdomain each
     // the domains do not overlap

      //---------------
      // setup
      //---------------

      // comm worlds on both processes
      struct communicator * comm_p1, * comm_p2;

      comm_p1 = yac_local_comm_new();
      comm_p2 = yac_local_comm_new();

      // named comms for the domains on both processes
      struct communicator * comm_named_p1, * comm_named_p2;

      comm_named_p1 = yac_named_local_comm_new(comm_p1, "P1");
      comm_named_p2 = yac_named_local_comm_new(comm_p2, "P2");

         // define a domain on P1

         struct grid * grid_p1;
         unsigned num_cells[2] = {2,2};
         unsigned cyclic[2] = {0,0};
         double coordinates_x[2][3] = {{-10,-9,-8}, {8,9,10}};
         double coordinates_y[2][3] = {{-10,-9,-8}, {8,9,10}};

         for ( unsigned i = 0; i < 2; ++i ) {
            for ( unsigned j = 0; j < 3; ++j ) {
               coordinates_x[i][j] *= rad;
               coordinates_y[i][j] *= rad;
            }
         }

         grid_p1 = yac_reg2d_grid_new(coordinates_x[0], coordinates_y[0],
                                      num_cells, cyclic);

         struct subdomain subdomain_p1;

         int global_corner_ids_p1[9] = {0,1,2,3,4,5,6,7,8};
         unsigned corner_is_on_rank_p1[9] = {-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_p1[4] = {0,1,2,3};
         unsigned cell_is_on_rank_p1[4] = {-1,-1,-1,-1};
         int global_edge_ids_p1[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned edge_is_on_rank_p1[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_p1, grid_p1, comm_named_p1,
                                 global_corner_ids_p1, corner_is_on_rank_p1,
                                 global_cell_ids_p1, cell_is_on_rank_p1,
                                 global_edge_ids_p1, edge_is_on_rank_p1);

         struct domain l_domain_p1;

         struct subdomain * blocks_p1[1] = {&subdomain_p1};

         yac_init_domain(&l_domain_p1, comm_named_p1, blocks_p1, 1, "domain_p1");

         // define a domain on P2, which matches the one on P1

         struct grid * grid_p2;

         grid_p2 = yac_reg2d_grid_new(coordinates_x[1], coordinates_y[1],
                                      num_cells, cyclic);

         struct subdomain subdomain_p2;

         int global_corner_ids_p2[9] = {0,1,2,3,4,5,6,7,8};
         unsigned corner_is_on_rank_p2[9] = {-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_p2[4] = {0,1,2,3};
         unsigned cell_is_on_rank_p2[4] = {-1,-1,-1,-1};
         int global_edge_ids_p2[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned edge_is_on_rank_p2[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_p2, grid_p2, comm_named_p2,
                                 global_corner_ids_p2, corner_is_on_rank_p2,
                                 global_cell_ids_p2, cell_is_on_rank_p2,
                                 global_edge_ids_p2, edge_is_on_rank_p2);

         struct domain l_domain_p2;

         struct subdomain * blocks_p2[1] = {&subdomain_p2};

         yac_init_domain(&l_domain_p2, comm_named_p2, blocks_p2, 1, "domain_p2");

         // finish domain definition
         yac_enddef_domain(&l_domain_p1);
         yac_enddef_domain(&l_domain_p2);

      //---------
      // testing
      //---------
      // init global search on both processes

      struct global_search * search[2];

      search[0] = yac_global_search_new(&l_domain_p1, "domain_p2", comm_p1, 1,0,
                                        "component_p1", "component_p2");
      search[1] = yac_global_search_new(&l_domain_p2, "domain_p1", comm_p2, 0,1,
                                        "component_p2", "component_p1");

      // finish initialisation of global search

      yac_enddef_global_search(search[0]);
      yac_enddef_global_search(search[1]);

      //---------------------
      // check source results
      //---------------------

      if (yac_source_process_has_overlap_with(search[0], 0))
         PUT_ERR("error in routine source_process_has_overlap_with\n");

      if (yac_source_process_has_overlap_with(search[0], 1))
         PUT_ERR("error in routine source_process_has_overlap_with\n");

      // for all global target cell ids
      for (i = 0; i < 4; ++i)
         if (yac_get_search_results_for_tgt_cell(search[0], i) != NULL)
            PUT_ERR("error in routine get_search_results_for_tgt_cell "
                    "(onedirectional 2 PEs)\n");

      //---------------
      // cleanup
      //---------------
      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);

      yac_free_domain(&l_domain_p1);
      yac_free_domain(&l_domain_p2);

      yac_delete_grid(grid_p1);
      yac_delete_grid(grid_p2);

      yac_free_communicator(comm_p1);
      yac_free_communicator(comm_p2);

      yac_free_communicator(comm_named_p1);
      yac_free_communicator(comm_named_p2);
   }

   { // bidirectional global search with two domains on two processes each
     // the two domains share on process (global root)

      //---------------
      // setup
      //---------------

      // comm worlds on all processes
      struct communicator * global_comm[3];

      for (int i = 0; i < 3; ++i)
         global_comm[i] = yac_local_comm_new();

      // named comms for the domains on all processes
      struct communicator * domain_comm_a[2];
      struct communicator * domain_comm_b[2];

      domain_comm_a[0] = yac_named_local_comm_new(global_comm[0], "D1");
      domain_comm_a[1] = yac_named_local_comm_new(global_comm[1], "D1");
      domain_comm_b[0] = yac_named_local_comm_new(global_comm[0], "D2");
      domain_comm_b[1] = yac_named_local_comm_new(global_comm[2], "D2");

         // define the domain a

         struct grid * grid_d1[2];
         unsigned num_cells_d1[2] = {3,4};
         unsigned cyclic_d1[2] = {0,0};
         double coordinates_x_d1[2][4] = {{0,1,2,3},{1,2,3,4}};
         double coordinates_y_d1[2][5] = {{0,1,2,3,4},{0,1,2,3,4}};

         for ( unsigned i = 0; i < 2; ++i )
            for ( unsigned j = 0; j < 4; ++j )
               coordinates_x_d1[i][j] *= rad;

         for ( unsigned i = 0; i < 2; ++i )
            for ( unsigned j = 0; j < 5; ++j )
               coordinates_y_d1[i][j] *= rad;

         for (i = 0; i < 2; ++i)
            grid_d1[i] = yac_reg2d_grid_new(coordinates_x_d1[i], coordinates_y_d1[i],
                                            num_cells_d1, cyclic_d1);

         struct subdomain subdomain_d1[2];

         int global_corner_ids_d1[2][20] = {{0,1,2,3, 5,6,7,8, 10,11,12,13, 15,16,17,18, 20,21,22,23},
                                            {1,2,3,4, 6,7,8,9, 11,12,13,14, 16,17,18,19, 21,22,23,24}};
         unsigned corner_is_on_rank_d1[2][20] = {{-1,-1,-1,1, -1,-1,-1,1, -1,-1,-1,1, -1,-1,-1,1, -1,-1,-1,1},
                                                 {0,-1,-1,-1, 0,-1,-1,-1, 0,-1,-1,-1, 0,-1,-1,-1, 0,-1,-1,-1}};
         int global_cell_ids_d1[2][12] = {{0,1,2, 4,5,6, 8,9,10, 12,13,14},
                                          {1,2,3, 5,6,7, 9,10,11, 13,14,15}};
         unsigned cell_is_on_rank_d1[2][12] = {{-1,-1,1, -1,-1,1, -1,-1,1, -1,-1,1},
                                               {0,-1,-1, 0,-1,-1, 0,-1,-1, 0,-1,-1}};
         int *global_edge_ids_d1 = NULL;
         unsigned *edge_is_on_rank_d1 = NULL;

         for (i = 0; i < 2; ++i)
            yac_init_subdomain_unstruct(subdomain_d1+i, grid_d1[i], domain_comm_a[i],
                                    global_corner_ids_d1[i], corner_is_on_rank_d1[i],
                                    global_cell_ids_d1[i], cell_is_on_rank_d1[i],
                                    global_edge_ids_d1, edge_is_on_rank_d1);

         struct domain domain_d1[2];

         struct subdomain * subdomains_d1[2][1] = {{subdomain_d1+0},{subdomain_d1+1}};

         for (i = 0; i < 2; ++i)
            yac_init_domain(domain_d1+i, domain_comm_a[i], subdomains_d1[i], 1, "domain a");

         // define the domain b

         struct grid * grid_d2[2];
         unsigned num_cells_d2[2] = {4,3};
         unsigned cyclic_d2[2] = {0,0};
         double coordinates_x_d2[2][5] = {{0.5,1.5,2.5,3.5,4.5},{0.5,1.5,2.5,3.5,4.5}};
         double coordinates_y_d2[2][4] = {{0.5,1.5,2.5,3.5},{1.5,2.5,3.5,4.5}};

         for ( unsigned i = 0; i < 2; ++i )
            for ( unsigned j = 0; j < 5; ++j )
               coordinates_x_d2[i][j] *= rad;

         for ( unsigned i = 0; i < 2; ++i )
            for ( unsigned j = 0; j < 4; ++j )
               coordinates_y_d2[i][j] *= rad;

         for (i = 0; i < 2; ++i)
            grid_d2[i] = yac_reg2d_grid_new(coordinates_x_d2[i], coordinates_y_d2[i],
                                            num_cells_d2, cyclic_d2);

         struct subdomain subdomain_d2[2];

         int global_corner_ids_d2[2][20] = {{0,1,2,3,4, 5,6,7,8,9, 10,11,12,13,14, 15,16,17,18,19},
                                            {5,6,7,8,9, 10,11,12,13,14, 15,16,17,18,19, 20,21,22,23,24}};
         unsigned corner_is_on_rank_d2[2][20] = {{-1,-1,-1,-1,-1, -1,-1,-1,-1,-1, -1,-1,-1,-1,-1, 1,1,1,1,1},
                                                 {0,0,0,0,0, -1,-1,-1,-1,-1, -1,-1,-1,-1,-1, -1,-1,-1,-1,-1}};
         int global_cell_ids_d2[2][12] = {{0,1,2,3, 4,5,6,7, 8,9,10,11},
                                          {4,5,6,7, 8,9,10,11, 12,13,14,15}};
         unsigned cell_is_on_rank_d2[2][12] = {{-1,-1,-1,-1, -1,-1,-1,-1, 1,1,1,1},
                                               {0,0,0,0, -1,-1,-1,-1, -1,-1,-1,-1}};
         int *global_edge_ids_d2 = NULL;
         unsigned *edge_is_on_rank_d2 = NULL;

         for (i = 0; i < 2; ++i)
            yac_init_subdomain_unstruct(subdomain_d2+i, grid_d2[i], domain_comm_b[i],
                                    global_corner_ids_d2[i], corner_is_on_rank_d2[i],
                                    global_cell_ids_d2[i], cell_is_on_rank_d2[i],
                                    global_edge_ids_d2, edge_is_on_rank_d2);

         struct domain domain_d2[2];

         struct subdomain * subdomains_d2[2][1] = {{subdomain_d2+0},{subdomain_d2+1}};

         for (i = 0; i < 2; ++i)
            yac_init_domain(domain_d2+i, domain_comm_b[i], subdomains_d2[i], 1, "domain b");

         // finish domain definition
         for (i = 0; i < 2; ++i) {
            yac_enddef_domain(domain_d1+i);
            yac_enddef_domain(domain_d2+i);
         }

      //---------------
      // testing
      //---------------
      // init global search on all domain processes

      struct global_search * search[4];

      search[0] = yac_global_search_new(domain_d1+0, "domain b", global_comm[0], 1,1,
                                        "component a", "component b");
      search[1] = yac_global_search_new(domain_d1+1, "domain b", global_comm[1], 1,1,
                                        "component a", "component b");
      search[2] = yac_global_search_new(domain_d2+0, "domain a", global_comm[0], 1,1,
                                        "component b", "component a");
      search[3] = yac_global_search_new(domain_d2+1, "domain a", global_comm[2], 1,1,
                                        "component b", "component a");

      // finish initialisation of global search

      for (i = 0; i < 4; ++i)
         yac_enddef_global_search(search[i]);

      //---------------------
      // check source results
      //---------------------

      // [domain][process][point id]
      unsigned ref_num_src_per_tgt[2][2][16] = {{{4,4,2,0, 4,4,2,0, 4,4,2,0, 2,2,1,0},
                                                 {2,4,4,2, 2,4,4,2, 2,4,4,2, 1,2,2,1}},
                                                {{1,2,2,2, 2,4,4,4, 2,4,4,4, 1,2,2,2},
                                                 {0,0,0,0, 1,2,2,2, 2,4,4,4, 2,4,4,4}}};
      // [domain][process][tgt point][matching local cell ids]
      unsigned ref_src_per_tgt[2][2][16][4] = {{{{0,1,3,4},{1,2,4,5},{2,5},{-1}, {3,4,6,7},{4,5,7,8},{5,8},{-1},
                                                 {6,7,9,10},{7,8,10,11},{8,11},{-1}, {9,10},{10,11},{11},{-1}},
                                                {{0,3},{0,1,3,4},{1,2,4,5},{2,5}, {3,6},{3,4,6,7},{4,5,7,8},{5,8},
                                                 {6,9},{6,7,9,10},{7,8,10,11},{8,11}, {9},{9,10},{10,11},{11}}},
                                               {{{0},{0,1},{1,2},{2,3}, {0,4},{0,1,4,5},{1,2,5,6},{2,3,6,7},
                                                 {4,8},{4,5,8,9},{5,6,9,10},{6,7,10,11}, {8},{8,9},{9,10},{10,11}},
                                                {{-1},{-1},{-1},{-1}, {0},{0,1},{1,2},{2,3},
                                                 {0,4},{0,1,4,5},{1,2,5,6},{2,3,6,7}, {4,8},{4,5,8,9},{5,6,9,10},{6,7,10,11}}}};
      // [domain][process][tgt point]
      unsigned ref_num_other_src_ranks[2][2][16] = {{{1,1,1,-1, 1,1,1,-1, 1,1,1,-1, 1,1,1,-1,},
                                                     {1,1,1,0, 1,1,1,0, 1,1,1,0, 1,1,1,0}},
                                                    {{0,0,0,0, 1,1,1,1, 1,1,1,1, 1,1,1,1},
                                                     {-1,-1,-1,-1, 1,1,1,1, 1,1,1,1, 1,1,1,1}}};
      // [domain][process][tgt point]
      unsigned ref_other_src_ranks[2][2][16] = {{{1,1,1,-1, 1,1,1,-1, 1,1,1,-1, 1,1,1,-1},
                                                 {0,0,0,-1, 0,0,0,-1, 0,0,0,-1, 0,0,0,-1}},
                                                {{-1,-1,-1,-1, 2,2,2,2, 2,2,2,2, 2,2,2,2},
                                                 {-1,-1,-1,-1, 0,0,0,0, 0,0,0,0, 0,0,0,0}}};
      // [domain][tgt point]
      unsigned ref_tgt_rank[2][16] = {{0,0,0,0, 0,0,0,0, 2,2,2,2, 2,2,2,2},
                                      {0,0,1,1, 0,0,1,1, 0,0,1,1, 0,0,1,1}};

      // [domain][rank]
      unsigned ref_has_overlap_with[2][3] = {{1,0,1},{1,1,0}};

      int ref_global_tgt_ids[16] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
                                    11, 12, 13, 14, 15};

      // for both domains
      for (l = 0; l < 2; ++l) {

         // for both processes
         for (k = 0; k < 2; ++k) {

            struct source_search_results const * src_search_results[16];

            yac_get_search_results_for_tgt_cells(search[l*2+k], ref_global_tgt_ids,
                                             16, src_search_results);

            // for all target points
            for (j = 0; j < 16; ++j) {

               struct source_search_results const * curr_result =
                  yac_get_search_results_for_tgt_cell(search[l*2+k], j);

               if (curr_result != src_search_results[j])
                  PUT_ERR("error in get_search_results_for_tgt_cells "
                          "(bidirectional 3 PEs, 2 PEs per domain, common global root)\n");

               if (ref_num_src_per_tgt[l][k][j] == 0) {

                  if (curr_result != NULL)
                     PUT_ERR("error in get_search_results_for_tgt_cell "
                             "(bidirectional 3 PEs, 2 PEs per domain, common global root)\n");

               } else {

                  if (curr_result == NULL)
                     PUT_ERR("error in get_search_results_for_tgt_cell "
                             "(bidirectional 3 PEs, 2 PEs per domain, common global root)\n");

                  if (curr_result->global_id != j)
                     PUT_ERR("error in get_search_results_for_tgt_cell; wrong global id "
                             "(bidirectional 3 PEs, 2 PEs per domain, common global root)\n");

                  if (curr_result->tgt_rank != ref_tgt_rank[l][j])
                     PUT_ERR("error in get_search_results_for_tgt_cell; wrong rank "
                             "(bidirectional 3 PEs, 2 PEs per domain, common global root)\n");

                  if (curr_result->tgt_domain == NULL)
                     PUT_ERR("error in get_search_results_for_tgt_cell; no target domain "
                             "(bidirectional 3 PEs, 2 PEs per domain, common global root)\n");

                  if (curr_result->num_other_src_ranks != ref_num_other_src_ranks[l][k][j])
                     PUT_ERR("error in get_search_results_for_tgt_cell; wrong number "
                             "of other source ranks "
                             "(bidirectional 3 PEs, 2 PEs per domain, common global root)\n");

                  if (curr_result->num_other_src_ranks > 0 &&
                      curr_result->other_src_ranks[0] != ref_other_src_ranks[l][k][j])
                     PUT_ERR("error in get_search_results_for_tgt_cell; wrong "
                             "other source ranks "
                             "(bidirectional 3 PEs, 2 PEs per domain, common global root)\n");

                  if (curr_result->num_overlapping_src_cells_per_subdomain[0] !=
                      ref_num_src_per_tgt[l][k][j])
                     PUT_ERR("error in get_search_results_for_tgt_cell; wrong number of source cells "
                             "(bidirectional 3 PEs, 2 PEs per domain, common global root)\n");

                  if (!check_ids((int*)(curr_result->overlapping_src_cells[0]),
                      (int*)ref_src_per_tgt[l][k][j], ref_num_src_per_tgt[l][k][j]))
                     PUT_ERR("error in get_search_results_for_tgt_cell; wrong source cells "
                             "(bidirectional 3 PEs, 2 PEs per domain, common global root)\n");
               }
            }

            for (i = 0; i < 3; ++i)
               if (yac_source_process_has_overlap_with(search[l*2+k], i) !=
                   ref_has_overlap_with[l][i])
                  PUT_ERR("error in routine source_process_has_overlap_with "
                          "(bidirectional 3 PEs, 2 PEs per domain, common global root)\n");
         }
      }

      //---------------
      // cleanup
      //---------------
      for (i = 0; i < 4; ++i)
         yac_global_search_delete(search[i]);

      for (i = 0; i < 2; ++i) {

         yac_free_domain(domain_d1+i);
         yac_free_domain(domain_d2+i);

         yac_delete_grid(grid_d1[i]);
         yac_delete_grid(grid_d2[i]);
      }

      for (i = 0; i < 2; ++i) {

         yac_free_communicator(domain_comm_a[i]);
         yac_free_communicator(domain_comm_b[i]);
      }

      for (i = 0; i < 3; ++i)
         yac_free_communicator(global_comm[i]);
   }

   { // bidirectional global search with two domains on two processes each
     // the two domains share on process (which is not global root)

      //---------------
      // setup
      //---------------

      // comm worlds on all processes
      struct communicator * global_comm[3];

      for (int i = 0; i < 3; ++i)
         global_comm[i] = yac_local_comm_new();

      // named comms for the domains on all processes
      struct communicator * domain_comm_a[2];
      struct communicator * domain_comm_b[2];

      domain_comm_a[0] = yac_named_local_comm_new(global_comm[0], "D1");
      domain_comm_a[1] = yac_named_local_comm_new(global_comm[1], "D1");
      domain_comm_b[0] = yac_named_local_comm_new(global_comm[1], "D2");
      domain_comm_b[1] = yac_named_local_comm_new(global_comm[2], "D2");

         // define the domain a

         struct grid * grid_d1[2];
         unsigned num_cells_d1[2] = {3,4};
         unsigned cyclic_d1[2] = {0,0};
         double coordinates_x_d1[2][4] = {{0,1,2,3},{1,2,3,4}};
         double coordinates_y_d1[2][5] = {{0,1,2,3,4},{0,1,2,3,4}};

         for ( unsigned i = 0; i < 2; ++i )
            for ( unsigned j = 0; j < 4; ++j )
               coordinates_x_d1[i][j] *= rad;

         for ( unsigned i = 0; i < 2; ++i )
            for ( unsigned j = 0; j < 5; ++j )
               coordinates_y_d1[i][j] *= rad;

         for (i = 0; i < 2; ++i)
            grid_d1[i] = yac_reg2d_grid_new(coordinates_x_d1[i], coordinates_y_d1[i],
                                            num_cells_d1, cyclic_d1);

         struct subdomain subdomain_d1[2];

         int global_corner_ids_d1[2][20] = {{0,1,2,3, 5,6,7,8, 10,11,12,13, 15,16,17,18, 20,21,22,23},
                                            {1,2,3,4, 6,7,8,9, 11,12,13,14, 16,17,18,19, 21,22,23,24}};
         unsigned corner_is_on_rank_d1[2][20] = {{-1,-1,-1,1, -1,-1,-1,1, -1,-1,-1,1, -1,-1,-1,1, -1,-1,-1,1},
                                                 {0,-1,-1,-1, 0,-1,-1,-1, 0,-1,-1,-1, 0,-1,-1,-1, 0,-1,-1,-1}};
         int global_cell_ids_d1[2][12] = {{0,1,2, 4,5,6, 8,9,10, 12,13,14},
                                          {1,2,3, 5,6,7, 9,10,11, 13,14,15}};
         unsigned cell_is_on_rank_d1[2][12] = {{-1,-1,1, -1,-1,1, -1,-1,1, -1,-1,1},
                                               {0,-1,-1, 0,-1,-1, 0,-1,-1, 0,-1,-1}};
         int *global_edge_ids_d1 = NULL;
         unsigned *edge_is_on_rank_d1 = NULL;

         for (i = 0; i < 2; ++i)
            yac_init_subdomain_unstruct(subdomain_d1+i, grid_d1[i], domain_comm_a[i],
                                    global_corner_ids_d1[i], corner_is_on_rank_d1[i],
                                    global_cell_ids_d1[i], cell_is_on_rank_d1[i],
                                    global_edge_ids_d1, edge_is_on_rank_d1);

         struct domain domain_d1[2];

         struct subdomain * subdomains_d1[2][1] = {{subdomain_d1+0},{subdomain_d1+1}};

         for (i = 0; i < 2; ++i)
            yac_init_domain(domain_d1+i, domain_comm_a[i], subdomains_d1[i], 1, "domain a");

         // define the domain b

         struct grid * grid_d2[2];
         unsigned num_cells_d2[2] = {4,3};
         unsigned cyclic_d2[2] = {0,0};
         double coordinates_x_d2[2][5] = {{0.5,1.5,2.5,3.5,4.5},{0.5,1.5,2.5,3.5,4.5}};
         double coordinates_y_d2[2][4] = {{0.5,1.5,2.5,3.5},{1.5,2.5,3.5,4.5}};

         for ( unsigned i = 0; i < 2; ++i )
            for ( unsigned j = 0; j < 5; ++j )
               coordinates_x_d2[i][j] *= rad;

         for ( unsigned i = 0; i < 2; ++i )
            for ( unsigned j = 0; j < 4; ++j )
               coordinates_y_d2[i][j] *= rad;

         for (i = 0; i < 2; ++i)
            grid_d2[i] = yac_reg2d_grid_new(coordinates_x_d2[i], coordinates_y_d2[i],
                                            num_cells_d2, cyclic_d2);

         struct subdomain subdomain_d2[2];

         int global_corner_ids_d2[2][20] = {{0,1,2,3,4, 5,6,7,8,9, 10,11,12,13,14, 15,16,17,18,19},
                                            {5,6,7,8,9, 10,11,12,13,14, 15,16,17,18,19, 20,21,22,23,24}};
         unsigned corner_is_on_rank_d2[2][20] = {{-1,-1,-1,-1,-1, -1,-1,-1,-1,-1, -1,-1,-1,-1,-1, 1,1,1,1,1},
                                                 {0,0,0,0,0, -1,-1,-1,-1,-1, -1,-1,-1,-1,-1, -1,-1,-1,-1,-1}};
         int global_cell_ids_d2[2][12] = {{0,1,2,3, 4,5,6,7, 8,9,10,11},
                                          {4,5,6,7, 8,9,10,11, 12,13,14,15}};
         unsigned cell_is_on_rank_d2[2][12] = {{-1,-1,-1,-1, -1,-1,-1,-1, 1,1,1,1},
                                               {0,0,0,0, -1,-1,-1,-1, -1,-1,-1,-1}};
         int *global_edge_ids_d2 = NULL;
         unsigned *edge_is_on_rank_d2 = NULL;

         for (i = 0; i < 2; ++i)
            yac_init_subdomain_unstruct(subdomain_d2+i, grid_d2[i], domain_comm_b[i],
                                    global_corner_ids_d2[i], corner_is_on_rank_d2[i],
                                    global_cell_ids_d2[i], cell_is_on_rank_d2[i],
                                    global_edge_ids_d2, edge_is_on_rank_d2);

         struct domain domain_d2[2];

         struct subdomain * subdomains_d2[2][1] = {{subdomain_d2+0},{subdomain_d2+1}};

         for (i = 0; i < 2; ++i)
            yac_init_domain(domain_d2+i, domain_comm_b[i], subdomains_d2[i], 1, "domain b");

         // finish domain definition
         for (i = 0; i < 2; ++i) {
            yac_enddef_domain(domain_d1+i);
            yac_enddef_domain(domain_d2+i);
         }

      //---------------
      // testing
      //---------------
      // init global search on all domain processes

      struct global_search * search[4];

      search[0] = yac_global_search_new(domain_d1+0, "domain b", global_comm[0], 1,1,
                                        "component a", "component b");
      search[1] = yac_global_search_new(domain_d1+1, "domain b", global_comm[1], 1,1,
                                        "component a", "component b");
      search[2] = yac_global_search_new(domain_d2+0, "domain a", global_comm[1], 1,1,
                                        "component b", "component a");
      search[3] = yac_global_search_new(domain_d2+1, "domain a", global_comm[2], 1,1,
                                        "component b", "component a");

      // finish initialisation of global search

      for (i = 0; i < 4; ++i)
         yac_enddef_global_search(search[i]);

      //---------------------
      // check source results
      //---------------------

      // [domain][process][point id]
      // [domain][process][point id]
      unsigned ref_num_src_per_tgt[2][2][16] = {{{4,4,2,0, 4,4,2,0, 4,4,2,0, 2,2,1,0},
                                                 {2,4,4,2, 2,4,4,2, 2,4,4,2, 1,2,2,1}},
                                                {{1,2,2,2, 2,4,4,4, 2,4,4,4, 1,2,2,2},
                                                 {0,0,0,0, 1,2,2,2, 2,4,4,4, 2,4,4,4}}};
      // [domain][process][tgt point][matching local cell ids]
      unsigned ref_src_per_tgt[2][2][16][4] = {{{{0,1,3,4},{1,2,4,5},{2,5},{-1}, {3,4,6,7},{4,5,7,8},{5,8},{-1},
                                                 {6,7,9,10},{7,8,10,11},{8,11},{-1}, {9,10},{10,11},{11},{-1}},
                                                {{0,3},{0,1,3,4},{1,2,4,5},{2,5}, {3,6},{3,4,6,7},{4,5,7,8},{5,8},
                                                 {6,9},{6,7,9,10},{7,8,10,11},{8,11}, {9},{9,10},{10,11},{11}}},
                                               {{{0},{0,1},{1,2},{2,3}, {0,4},{0,1,4,5},{1,2,5,6},{2,3,6,7},
                                                 {4,8},{4,5,8,9},{5,6,9,10},{6,7,10,11}, {8},{8,9},{9,10},{10,11}},
                                                {{-1},{-1},{-1},{-1}, {0},{0,1},{1,2},{2,3},
                                                 {0,4},{0,1,4,5},{1,2,5,6},{2,3,6,7}, {4,8},{4,5,8,9},{5,6,9,10},{6,7,10,11}}}};
      // [domain][process][tgt point]
      unsigned ref_num_other_src_ranks[2][2][16] = {{{1,1,1,-1, 1,1,1,-1, 1,1,1,-1, 1,1,1,-1,},
                                                     {1,1,1,0, 1,1,1,0, 1,1,1,0, 1,1,1,0}},
                                                    {{0,0,0,0, 1,1,1,1, 1,1,1,1, 1,1,1,1},
                                                     {-1,-1,-1,-1, 1,1,1,1, 1,1,1,1, 1,1,1,1}}};
      // [domain][process][tgt point]
      unsigned ref_other_src_ranks[2][2][16] = {{{1,1,1,-1, 1,1,1,-1, 1,1,1,-1, 1,1,1,-1},
                                                 {0,0,0,-1, 0,0,0,-1, 0,0,0,-1, 0,0,0,-1}},
                                                {{-1,-1,-1,-1, 2,2,2,2, 2,2,2,2, 2,2,2,2},
                                                 {-1,-1,-1,-1, 1,1,1,1, 1,1,1,1, 1,1,1,1}}};
      // [domain][tgt point]
      unsigned ref_tgt_rank[2][16] = {{1,1,1,1, 1,1,1,1, 2,2,2,2, 2,2,2,2},
                                      {0,0,1,1, 0,0,1,1, 0,0,1,1, 0,0,1,1}};

      // [domain][rank]
      unsigned ref_has_overlap_with[2][3] = {{0,1,1},{1,1,0}};

      int ref_global_tgt_ids[16] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
                                    11, 12, 13, 14, 15};

      // for both domains
      for (l = 0; l < 2; ++l) {

         // for both processes
         for (k = 0; k < 2; ++k) {

            struct source_search_results const * src_search_results[16];

            yac_get_search_results_for_tgt_cells(search[l*2+k], ref_global_tgt_ids,
                                             16, src_search_results);

            // for all target points
            for (j = 0; j < 16; ++j) {

               struct source_search_results const * curr_result =
                  yac_get_search_results_for_tgt_cell(search[l*2+k], j);

               if (curr_result != src_search_results[j])
                  PUT_ERR("error in get_search_results_for_tgt_cells "
                          "(bidirectional 3 PEs, 2 PEs per domain, common global root)\n");

               if (ref_num_src_per_tgt[l][k][j] == 0) {

                  if (curr_result != NULL)
                     PUT_ERR("error in get_search_results_for_tgt_cell "
                             "(bidirectional 3 PEs, 2 PEs per domain, common global root)\n");

               } else {

                  if (curr_result == NULL)
                     PUT_ERR("error in get_search_results_for_tgt_cell "
                             "(bidirectional 3 PEs, 2 PEs per domain, common global root)\n");

                  if (curr_result->global_id != j)
                     PUT_ERR("error in get_search_results_for_tgt_cell; wrong global id "
                             "(bidirectional 3 PEs, 2 PEs per domain, common global root)\n");

                  if (curr_result->tgt_rank != ref_tgt_rank[l][j])
                     PUT_ERR("error in get_search_results_for_tgt_cell; wrong rank "
                             "(bidirectional 3 PEs, 2 PEs per domain, common global root)\n");

                  if (curr_result->tgt_domain == NULL)
                     PUT_ERR("error in get_search_results_for_tgt_cell; no target domain "
                             "(bidirectional 3 PEs, 2 PEs per domain, common global root)\n");

                  if (curr_result->num_other_src_ranks != ref_num_other_src_ranks[l][k][j])
                     PUT_ERR("error in get_search_results_for_tgt_cell; wrong number "
                             "of other source ranks "
                             "(bidirectional 3 PEs, 2 PEs per domain, common global root)\n");

                  if (curr_result->num_other_src_ranks > 0 &&
                      curr_result->other_src_ranks[0] != ref_other_src_ranks[l][k][j])
                     PUT_ERR("error in get_search_results_for_tgt_cell; wrong "
                             "other source ranks "
                             "(bidirectional 3 PEs, 2 PEs per domain, common global root)\n");

                  if (curr_result->num_overlapping_src_cells_per_subdomain[0] !=
                      ref_num_src_per_tgt[l][k][j])
                     PUT_ERR("error in get_search_results_for_tgt_cell; wrong number of source cells "
                             "(bidirectional 3 PEs, 2 PEs per domain, common global root)\n");

                  if (!check_ids((int*)(curr_result->overlapping_src_cells[0]),
                      (int*)ref_src_per_tgt[l][k][j], ref_num_src_per_tgt[l][k][j]))
                     PUT_ERR("error in get_search_results_for_tgt_cell; wrong source cells "
                             "(bidirectional 3 PEs, 2 PEs per domain, common global root)\n");
               }
            }

            for (i = 0; i < 3; ++i)
               if (yac_source_process_has_overlap_with(search[l*2+k], i) !=
                   ref_has_overlap_with[l][i])
                  PUT_ERR("error in routine source_process_has_overlap_with "
                          "(bidirectional 3 PEs, 2 PEs per domain, common global root)\n");
         }
      }

      //---------------
      // cleanup
      //---------------
      for (i = 0; i < 4; ++i)
         yac_global_search_delete(search[i]);

      for (i = 0; i < 2; ++i) {

         yac_free_domain(domain_d1+i);
         yac_free_domain(domain_d2+i);

         yac_delete_grid(grid_d1[i]);
         yac_delete_grid(grid_d2[i]);
      }

      for (i = 0; i < 2; ++i) {

         yac_free_communicator(domain_comm_a[i]);
         yac_free_communicator(domain_comm_b[i]);
      }

      for (i = 0; i < 3; ++i)
         yac_free_communicator(global_comm[i]);
   }

   {
      // the global source domain is a 4x4 grid (3 processes):
      //
      //       20--36--21--37--22--38--23--39--24
      //       |       |       |       |       |
      //       28  12  30  13  32  14  34  15  35
      //       |       |       |       |       |
      //       15--27--16--29--17--31--18--33--19
      //       |       |       |       |       |
      //       19  08  21  09  23  10  25  11  26
      //       |       |       |       |       |
      //       10--18--11--20--12--22--13--24--14
      //       |       |       |       |       |
      //       10  04  12  05  14  06  16  07  17
      //       |       |       |       |       |
      //       05--09--06--11--07--13--08--15--09
      //       |       |       |       |       |
      //       01  00  03  01  05  02  07  03  08
      //       |       |       |       |       |
      //       00--00--01--02--02--04--03--06--04
      //
      // the global target domain is a 3x3 grid (2 processes):
      //
      //       12--21--13--22--14--23--15
      //       |       |       |       |
      //       15  06  17  07  19  08  20
      //       |       |       |       |
      //       08--14--09--16--10--18--11
      //       |       |       |       |
      //       08  03  10  04  12  05  13
      //       |       |       |       |
      //       04--07--05--09--06--11--07
      //       |       |       |       |
      //       01  00  03  01  05  02  06
      //       |       |       |       |
      //       00--00--01--02--02--04--03
      //
      //---------------
      // setup
      //---------------

      // comm worlds on both processes

      #define NUM_SRC 3
      #define NUM_TGT 2
      struct communicator * comm_src[NUM_SRC], * comm_tgt[NUM_TGT];

      for (int i = 0; i < NUM_SRC; ++i)
         comm_src[i] = yac_local_comm_new();
      for (int i = 0; i < NUM_TGT; ++i)
         comm_tgt[i] = yac_local_comm_new();

      // named comms for the domains on both processes
      struct communicator * comm_named_src[NUM_SRC], * comm_named_tgt[NUM_TGT];

      for (int i = 0; i < NUM_SRC; ++i)
         comm_named_src[i] = yac_named_local_comm_new(comm_src[i], "src");
      for (int i = 0; i < NUM_TGT; ++i)
         comm_named_tgt[i] = yac_named_local_comm_new(comm_tgt[i], "tgt");

      // define domains on source

      struct grid * grid_src[NUM_SRC];
      unsigned num_cells_src[NUM_SRC][2] = {{4,2}, {4,4}, {4,2}};
      unsigned cyclic_src[2] = {0,0};
      double coordinates_x_src[5] = {0.0,1.0,2.0,3.0,4.0};
      double coordinates_y_src[NUM_SRC][5] = {{0.0,1.0,2.0},
                                              {0.0,1.0,2.0,3.0,4.0},
                                              {2.0,3.0,4.0}};

      for ( unsigned i = 0; i < 5; ++i )
         coordinates_x_src[i] *= rad;

      for ( unsigned j = 0; j < NUM_SRC; ++j )
         for ( unsigned i = 0; i < 4; ++i )
         coordinates_y_src[j][i] *= rad;

      for (int i = 0; i < NUM_SRC; ++i)
         grid_src[i] = yac_reg2d_grid_new(coordinates_x_src, coordinates_y_src[i],
                                          num_cells_src[i], cyclic_src);

      struct subdomain subdomain_src[NUM_SRC];

      int global_corner_ids_src[NUM_SRC][25] = {{0,1,2,3,4, 5,6,7,8,9, 10,11,12,13,14},
                                                {0,1,2,3,4, 5,6,7,8,9, 10,11,12,13,14, 15,16,17,18,19, 20,21,22,23,24},
                                                {10,11,12,13,14, 15,16,17,18,19, 20,21,22,23,24}};
      unsigned corner_is_on_rank_src[NUM_SRC][25] = {{-1,-1,-1,-1, -1,-1,-1,-1,-1, 1,1,1,1,1},
                                                     {0,0,0,0,0, 0,0,0,0,0, -1,-1,-1,-1,-1, 2,2,2,2,2, 2,2,2,2,2},
                                                     {1,1,1,1,1, -1,-1,-1,-1,-1, -1,-1,-1,-1,-1}};
      int global_cell_ids_src[NUM_SRC][16] = {{0,1,2,3, 4,5,6,7},
                                              {0,1,2,3, 4,5,6,7, 8,9,10,11, 12,13,14,15},
                                              {8,9,10,11, 12,13,14,15}};
      unsigned cell_is_on_rank_src[NUM_SRC][16] = {{-1,-1,-1,-1, 1,1,1,1},
                                                   {0,0,0,0, -1,-1,-1,-1, -1,-1,-1,-1, 2,2,2,2},
                                                   {1,1,1,1, -1,-1,-1,-1}};

      for (int i = 0; i < NUM_SRC; ++i)
         yac_init_subdomain_unstruct(&subdomain_src[i], grid_src[i], comm_named_src[i],
                                 global_corner_ids_src[i], corner_is_on_rank_src[i],
                                 global_cell_ids_src[i], cell_is_on_rank_src[i],
                                 NULL, NULL);

      struct domain l_domain_src[NUM_SRC];

      struct subdomain * blocks_src[NUM_SRC] = {&subdomain_src[0],
                                                &subdomain_src[1],
                                                &subdomain_src[2]};

      for (int i = 0; i < NUM_SRC; ++i)
         yac_init_domain(&l_domain_src[i], comm_named_src[i], &blocks_src[i], 1, "domain_src");

      // define a domain on the target

      struct grid * grid_tgt[NUM_TGT];
      unsigned num_cells_tgt[2] = {3,2};
      unsigned cyclic_tgt[2] = {0,0};
      double coordinates_x_tgt[4] = {0.5,1.5,2.5,3.5};
      double coordinates_y_tgt[2][3] = {{0.5,1.5,2.5}, {1.5,2.5,3.5}};

      for ( unsigned i = 0; i < 4; ++i )
         coordinates_x_tgt[i] *= rad;

      for ( unsigned i = 0; i < 3; ++i )
         for ( unsigned j = 0; j < NUM_TGT; ++j )
            coordinates_y_tgt[j][i] *= rad;

      for (int i = 0; i < NUM_TGT; ++i)
         grid_tgt[i] = yac_reg2d_grid_new(coordinates_x_tgt,
                                          coordinates_y_tgt[i],
                                          num_cells_tgt, cyclic_tgt);

      struct subdomain subdomain_tgt[NUM_TGT];

      int global_corner_ids_tgt[NUM_TGT][12] = {{0,1,2,3,4,5,6,7,8,9,10,11},
                                                {4,5,6,7,8,9,10,11,12,13,14,15}};
      unsigned corner_is_on_rank_tgt[NUM_TGT][12] = {{-1,-1,-1,-1, -1,-1,-1,-1, 1,1,1,1},
                                                     {0,0,0,0, -1,-1,-1,-1, -1,-1,-1,-1}};
      int global_cell_ids_tgt[NUM_TGT][6] = {{0,1,2,3,4,5},
                                             {3,4,5,6,7,8}};
      unsigned cell_is_on_rank_tgt[NUM_TGT][6] = {{-1,-1,-1, -1,1,1},
                                                  {0,-1,-1, -1,-1,-1}};

      for (int i = 0; i < NUM_TGT; ++i)
         yac_init_subdomain_unstruct(&subdomain_tgt[i], grid_tgt[i], comm_named_tgt[i],
                                 global_corner_ids_tgt[i], corner_is_on_rank_tgt[i],
                                 global_cell_ids_tgt[i], cell_is_on_rank_tgt[i],
                                 NULL, NULL);

      struct domain l_domain_tgt[NUM_TGT];

      struct subdomain * blocks_tgt[NUM_TGT] = {&subdomain_tgt[0],
                                                &subdomain_tgt[1]};

      for (int i = 0; i < NUM_TGT; ++i)
         yac_init_domain(&l_domain_tgt[i], comm_named_tgt[i], &blocks_tgt[i], 1, "domain_tgt");

      // finish domain definition
      for (int i = 0; i < NUM_SRC; ++i)
         yac_enddef_domain(&l_domain_src[i]);
      for (int i = 0; i < NUM_TGT; ++i)
         yac_enddef_domain(&l_domain_tgt[i]);

      //-------------------------------------
      // init global search on both processes
      //-------------------------------------

      struct global_search * search[NUM_SRC + NUM_TGT];

      for (int i = 0; i < NUM_SRC; ++i)
         search[i] =
            yac_global_search_new(&l_domain_src[i], "domain_tgt", comm_src[i],
                                  1,0, "component_src", "component_tgt");
      for (int i = 0; i < NUM_TGT; ++i)
         search[NUM_SRC + i] =
            yac_global_search_new(&l_domain_tgt[i], "domain_src", comm_tgt[i],
                                  0,1, "component_tgt", "component_src");

      //---------------------------------------
      // finish initialisation of global search
      //---------------------------------------

      for (int i = 0; i < NUM_SRC + NUM_TGT; ++i)
         yac_enddef_global_search(search[i]);

      //---------------------
      // check source results
      //---------------------

      // [process]
      unsigned ref_num_tgt_cells[3] = {6,9,6};

      // [process][tgt_point]
      int ref_global_tgt_ids[3][9] = {{0,1,2, 3,4,5},
                                      {0,1,2, 3,4,5, 6,7,8},
                                      {3,4,5, 6,7,8}};

      // [process][point id]
      unsigned ref_num_src_per_tgt[3][9] = {{4,4,4, 2,2,2},
                                            {4,4,4, 4,4,4, 4,4,4},
                                            {2,2,2, 4,4,4}};
      // [process][tgt point][matching local cell ids]
      unsigned ref_src_per_tgt[3][9][4] = {{{0,1,4,5},{1,2,5,6},{2,3,6,7},{4,5},{5,6},{6,7},{-1},{-1},{-1}},
                                           {{0,1,4,5},{1,2,5,6},{2,3,6,7},{4,5,8,9},{5,6,9,10},{6,7,10,11},{8,9,12,13},{9,10,13,14},{10,11,14,15}},
                                           {{0,1},{1,2},{2,3},{0,1,4,5},{1,2,5,6},{2,3,6,7},{-1},{-1},{-1}}};
      // [process][tgt point]
      unsigned ref_num_other_src_ranks[3][9] = {{1,1,1, 2,2,2, -1,-1,-1},
                                                {1,1,1, 2,2,2, 1,1,1},
                                                {2,2,2, 1,1,1, -1,-1,-1}};
      // [process][tgt point]
      unsigned ref_other_src_ranks[3][9][2] = {{{1},{1},{1},{1,2},{1,2},{1,2},{-1},{-1},{-1}},
                                               {{0},{0},{0},{0,2},{0,2},{0,2},{2},{2},{2}},
                                               {{0,1},{0,1},{0,1},{1},{1},{1},{-1},{-1},{-1}}};
      // [tgt point]
      unsigned ref_tgt_rank[3][9] = {{3,3,3, 3,4,4, -4,-4,-4},
                                     {3,3,3, 3,4,4, 4,4,4},
                                     {3,4,4, 4,4,4, -4,-4,-4}};

      // for all src processes
      for (k = 0; k < 3; ++k) {

         struct source_search_results const *
            src_search_results[ref_num_tgt_cells[k]];

         yac_get_search_results_for_tgt_cells(search[k], ref_global_tgt_ids[k],
                                          ref_num_tgt_cells[k],
                                          src_search_results);

         // for all overlapping target points
         for (int j = 0; j < ref_num_tgt_cells[k]; ++j) {

            struct source_search_results const * curr_result =
               yac_get_search_results_for_tgt_cell(search[k],
                                               ref_global_tgt_ids[k][j]);

            if (curr_result != src_search_results[j])
               PUT_ERR("error in get_search_results_for_tgt_cells\n");

            if (ref_num_src_per_tgt[k][j] == 0) {

               if (curr_result != NULL)
                  PUT_ERR("error in get_search_results_for_tgt_cell\n");

            } else {

               if (curr_result == NULL)
                  PUT_ERR("error in get_search_results_for_tgt_cell\n");

               if (curr_result->global_id != ref_global_tgt_ids[k][j])
                  PUT_ERR("error in get_search_results_for_tgt_cell; wrong global id\n");

               if (curr_result->tgt_rank != ref_tgt_rank[k][j])
                  PUT_ERR("error in get_search_results_for_tgt_cell; wrong rank\n");

               if (curr_result->tgt_domain == NULL)
                  PUT_ERR("error in get_search_results_for_tgt_cell; no target domain\n");

               if (curr_result->num_other_src_ranks != ref_num_other_src_ranks[k][j])
                  PUT_ERR("error in get_search_results_for_tgt_cell; wrong number "
                          "of other source ranks\n");

               if ((curr_result->num_other_src_ranks == 2 &&
                    !((curr_result->other_src_ranks[0] == ref_other_src_ranks[k][j][0] &&
                       curr_result->other_src_ranks[1] == ref_other_src_ranks[k][j][1]) ||
                      (curr_result->other_src_ranks[0] == ref_other_src_ranks[k][j][1] &&
                       curr_result->other_src_ranks[1] == ref_other_src_ranks[k][j][0]))) ||
                   (curr_result->num_other_src_ranks == 1 &&
                    curr_result->other_src_ranks[0] != ref_other_src_ranks[k][j][0]))
                  PUT_ERR("error in get_search_results_for_tgt_cell; wrong "
                          "other source ranks\n");

               if (curr_result->num_overlapping_src_cells_per_subdomain[0] !=
                   ref_num_src_per_tgt[k][j])
                  PUT_ERR("error in get_search_results_for_tgt_cell; wrong number of source cells\n");

               if (!check_ids((int*)(curr_result->overlapping_src_cells[0]),
                   (int*)ref_src_per_tgt[k][j], ref_num_src_per_tgt[k][j]))
                  PUT_ERR("error in get_search_results_for_tgt_cell; wrong source cells\n");
            }
         }

         for (i = 3; i < 4; ++i)
            if (!yac_source_process_has_overlap_with(search[k], i))
               PUT_ERR("error in routine source_process_has_overlap_with\n");
      }

      //---------------------------------------
      // clean up
      //---------------------------------------

      for (int i = 0; i < NUM_SRC + NUM_TGT; ++i)
         yac_global_search_delete(search[i]);

      for (int i = 0; i < NUM_SRC; ++i)
         yac_free_domain(&l_domain_src[i]);
      for (int i = 0; i < NUM_TGT; ++i)
         yac_free_domain(&l_domain_tgt[i]);

      for (int i = 0; i < NUM_SRC; ++i)
         yac_delete_grid(grid_src[i]);
      for (int i = 0; i < NUM_TGT; ++i)
         yac_delete_grid(grid_tgt[i]);

      for (int i = 0; i < NUM_SRC; ++i)
         yac_free_communicator(comm_src[i]);
      for (int i = 0; i < NUM_TGT; ++i)
         yac_free_communicator(comm_tgt[i]);

      for (int i = 0; i < NUM_SRC; ++i)
         yac_free_communicator(comm_named_src[i]);
      for (int i = 0; i < NUM_TGT; ++i)
         yac_free_communicator(comm_named_tgt[i]);
   }

   {
      // the global source domain is a 7x7 grid (4 processes):
      //
      // 56-----57-----58-----59-----60-----61-----62-----63
      //  |      |      |      |      |      |      |      |
      //  |  42  |  43  |  44  |  45  |  46  |  47  |  48  |
      //  |      |      |      |      |      |      |      |
      // 48-----49-----50-----51-----52-----53-----54-----55
      //  |      |      |      |      |      |      |      |
      //  |  35  |  36  |  37  |  38  |  39  |  40  |  41  |
      //  |      |      |      |      |      |      |      |
      // 40-----41-----42-----43-----44-----45-----46-----47
      //  |      |      |      |      |      |      |      |
      //  |  28  |  29  |  30  |  31  |  32  |  33  |  34  |
      //  |      |      |      |      |      |      |      |
      // 32-----33-----34-----35-----36-----37-----38-----39
      //  |      |      |      |      |      |      |      |
      //  |  21  |  22  |  23  |  24  |  25  |  26  |  27  |
      //  |      |      |      |      |      |      |      |
      // 24-----25-----26-----27-----28-----29-----30-----31
      //  |      |      |      |      |      |      |      |
      //  |  14  |  15  |  16  |  17  |  18  |  19  |  20  |
      //  |      |      |      |      |      |      |      |
      // 16-----17-----18-----19-----20-----21-----22-----23
      //  |      |      |      |      |      |      |      |
      //  |  07  |  08  |  09  |  10  |  11  |  12  |  13  |
      //  |      |      |      |      |      |      |      |
      // 08-----09-----10-----11-----12-----13-----14-----15
      //  |      |      |      |      |      |      |      |
      //  |  00  |  01  |  02  |  03  |  04  |  05  |  06  |
      //  |      |      |      |      |      |      |      |
      // 00-----01-----02-----03-----04-----05-----06-----07
      //
      // the domain is distributed among the processes as follows:
      // (index == process)
      //
      // 3---3---3---3---3---3---3---3
      // | 3 | 3 | 3 | 3 | 3 | 3 | 3 |
      // 3---3---3---3---3---3---3---3
      // | 3 | 3 | 3 | 3 | 3 | 3 | 3 |
      // 3---3---3---1---2---2---3---3
      // | 1 | 1 | 1 | 2 | 2 | 2 | 2 |
      // 1---1---1---2---2---2---2---2
      // | 1 | 1 | 1 | 2 | 2 | 2 | 2 |
      // 1---1---1---1---2---2---2---2
      // | 1 | 1 | 1 | 2 | 2 | 2 | 2 |
      // 1---1---1---0---0---0---2---2
      // | 0 | 0 | 0 | 0 | 0 | 0 | 0 |
      // 0---0---0---0---0---0---0---0
      // | 0 | 0 | 0 | 0 | 0 | 0 | 0 |
      // 0---0---0---0---0---0---0---0
      //
      // the global target domain is also a 7x7 grid (1 process)
      //
      //---------------
      // setup
      //---------------

      // comm worlds on both processes

#undef NUM_SRC
#undef NUM_TGT
#define NUM_SRC 4
#define NUM_TGT 1

      struct communicator * comm_src[NUM_SRC], * comm_tgt[NUM_TGT];

      for (int i = 0; i < NUM_SRC; ++i)
         comm_src[i] = yac_local_comm_new();
      for (int i = 0; i < NUM_TGT; ++i)
         comm_tgt[i] = yac_local_comm_new();

      // named comms for the domains on both processes
      struct communicator * comm_named_src[NUM_SRC], * comm_named_tgt[NUM_TGT];

      for (int i = 0; i < NUM_SRC; ++i)
         comm_named_src[i] = yac_named_local_comm_new(comm_src[i], "src");
      for (int i = 0; i < NUM_TGT; ++i)
         comm_named_tgt[i] = yac_named_local_comm_new(comm_tgt[i], "tgt");

      // define domains on source

      struct grid * grid_src[NUM_SRC];
      unsigned num_cells_src[NUM_SRC][2] = {{7,3},{4,5},{5,5},{7,3}};
      unsigned cyclic_src[2] = {0,0};
      double coordinates_x_src[NUM_SRC][8] = {{0,1,2,3,4,5,6,7},{0,1,2,3,4},
                                              {2,3,4,5,6,7},{0,1,2,3,4,5,6,7}};
      double coordinates_y_src[NUM_SRC][8] = {{0,1,2,3},{1,2,3,4,5,6},
                                              {1,2,3,4,5,6},{4,5,6,7}};

      for ( unsigned j = 0; j < NUM_SRC; ++j ) {
         for ( unsigned i = 0; i < 8; ++i ) {
            coordinates_x_src[j][i] *= rad;
            coordinates_y_src[j][i] *= rad;
         }
      }

      for (int i = 0; i < NUM_SRC; ++i)
         grid_src[i] = yac_reg2d_grid_new(coordinates_x_src[i],
                                          coordinates_y_src[i],
                                          num_cells_src[i], cyclic_src);

      struct subdomain subdomain_src[NUM_SRC];

      int global_corner_ids_src[NUM_SRC][36] = {{ 0, 1, 2, 3, 4, 5, 6, 7,
                                                  8, 9,10,11,12,13,14,15,
                                                 16,17,18,19,20,21,22,23,
                                                 24,25,26,27,28,29,30,31},
                                                { 8, 9,10,11,12,
                                                 16,17,18,19,20,
                                                 24,25,26,27,28,
                                                 32,33,34,35,36,
                                                 40,41,42,43,44,
                                                 48,49,50,51,52},
                                                {10,11,12,13,14,15,
                                                 18,19,20,21,22,23,
                                                 26,27,28,29,30,31,
                                                 34,35,36,37,38,39,
                                                 42,43,44,45,46,47,
                                                 50,51,52,53,54,55},
                                                {32,33,34,35,36,37,38,39,
                                                 40,41,42,43,44,45,46,47,
                                                 48,49,50,51,52,53,54,55,
                                                 56,57,58,59,60,61,62,63}};
      unsigned corner_is_on_rank_src[NUM_SRC][36] = {{-1,-1,-1,-1,-1,-1,-1,-1,
                                                      -1,-1,-1,-1,-1,-1,-1,-1,
                                                       1, 1, 1,-1,-1,-1, 2, 2},
                                                     { 0, 0, 0, 0, 0,
                                                      -1,-1,-1, 0, 0,
                                                      -1,-1,-1,-1, 2,
                                                      -1,-1,-1, 2, 2,
                                                       3, 3, 3,-1, 2,
                                                       3, 3, 3, 3, 3},
                                                     { 0, 0, 0, 0, 0, 0,
                                                       1, 0, 0, 0,-1,-1,
                                                       1, 1,-1,-1,-1,-1,
                                                       1,-1,-1,-1,-1,-1,
                                                       3, 1,-1,-1, 3, 3,
                                                       3, 3, 3, 3, 3, 3},
                                                     { 1, 1, 1, 2, 2, 2, 2, 2,
                                                      -1,-1,-1, 1, 2, 2,-1,-1,
                                                      -1,-1,-1,-1,-1,-1,-1,-1,
                                                      -1,-1,-1,-1,-1,-1,-1,-1}};
      int global_cell_ids_src[NUM_SRC][25] = {{ 0, 1, 2, 3, 4, 5, 6,
                                                7, 8, 9,10,11,12,13,
                                               14,15,16,17,18,19,20},
                                              { 7, 8, 9,10,
                                               14,15,16,17,
                                               21,22,23,24,
                                               28,29,30,31,
                                               35,36,37,38},
                                              { 9,10,11,12,13,
                                               16,17,18,19,20,
                                               23,24,25,26,27,
                                               30,31,32,33,34,
                                               37,38,39,40,41},
                                              {28,29,30,31,32,33,34,
                                               35,36,37,38,39,40,41,
                                               42,43,44,45,46,47,48}};
      unsigned cell_is_on_rank_src[NUM_SRC][25] = {{-1,-1,-1,-1,-1,-1,-1,
                                                    -1,-1,-1,-1,-1,-1,-1,
                                                     1, 1, 1, 2, 2, 2, 2},
                                                   { 0, 0, 0, 0,
                                                    -1,-1,-1, 2,
                                                    -1,-1,-1, 2,
                                                    -1,-1,-1, 2,
                                                     3, 3, 3, 3},
                                                   { 0, 0, 0, 0, 0,
                                                     1,-1,-1,-1,-1,
                                                     1,-1,-1,-1,-1,
                                                     1,-1,-1,-1,-1,
                                                     3, 3, 3, 3, 3},
                                                   { 1, 1, 1, 2, 2, 2, 2,
                                                     3, 3, 3, 3, 3, 3, 3,
                                                     3, 3, 3, 3, 3, 3, 3}};

      for (int i = 0; i < NUM_SRC; ++i)
         yac_init_subdomain_unstruct(&subdomain_src[i], grid_src[i], comm_named_src[i],
                                 global_corner_ids_src[i], corner_is_on_rank_src[i],
                                 global_cell_ids_src[i], cell_is_on_rank_src[i],
                                 NULL, NULL);

      struct domain l_domain_src[NUM_SRC];

      struct subdomain * blocks_src[NUM_SRC];

      for (unsigned i = 0; i < NUM_SRC; ++i)
         blocks_src[i] = &subdomain_src[i];

      for (int i = 0; i < NUM_SRC; ++i)
         yac_init_domain(&l_domain_src[i], comm_named_src[i], &blocks_src[i], 1, "domain_src");

      // define a domain on the target

      struct grid * grid_tgt[NUM_TGT];
      unsigned num_cells_tgt[NUM_TGT][2] = {{7,7}};
      unsigned cyclic_tgt[2] = {0,0};
      double coordinates_x_tgt[NUM_TGT][8] = {{0,1,2,3,4,5,6,7}};
      double coordinates_y_tgt[NUM_TGT][8] = {{0,1,2,3,4,5,6,7}};

      for ( unsigned i = 0; i < NUM_TGT; ++i ) {
         for ( unsigned j = 0; j < 8; ++j ) {
            coordinates_x_tgt[i][j] *= rad;
            coordinates_y_tgt[i][j] *= rad;
         }
      }

      for (int i = 0; i < NUM_TGT; ++i)
         grid_tgt[i] = yac_reg2d_grid_new(coordinates_x_tgt[i],
                                          coordinates_y_tgt[i],
                                          num_cells_tgt[i], cyclic_tgt);

      struct subdomain subdomain_tgt[NUM_TGT];

      int global_corner_ids_tgt[NUM_TGT][64] = {{ 0, 1, 2, 3, 4, 5, 6, 7,
                                                  8, 9,10,11,12,13,14,15,
                                                 16,17,18,19,20,21,22,23,
                                                 24,25,26,27,28,29,30,31,
                                                 32,33,34,35,36,37,38,39,
                                                 40,41,42,43,44,45,46,47,
                                                 48,49,50,51,52,53,54,55,
                                                 56,57,58,59,60,61,62,63}};
      unsigned corner_is_on_rank_tgt[NUM_TGT][64] = {{-1,-1,-1,-1,-1,-1,-1,-1,
                                                      -1,-1,-1,-1,-1,-1,-1,-1,
                                                      -1,-1,-1,-1,-1,-1,-1,-1,
                                                      -1,-1,-1,-1,-1,-1,-1,-1,
                                                      -1,-1,-1,-1,-1,-1,-1,-1,
                                                      -1,-1,-1,-1,-1,-1,-1,-1,
                                                      -1,-1,-1,-1,-1,-1,-1,-1,
                                                      -1,-1,-1,-1,-1,-1,-1,-1}};
      int global_cell_ids_tgt[NUM_TGT][49] = {{ 0, 1, 2, 3, 4, 5, 6,
                                                7, 8, 9,10,11,12,13,
                                               14,15,16,17,18,19,20,
                                               21,22,23,24,25,26,27,
                                               28,29,30,31,32,33,34,
                                               35,36,37,38,39,40,41,
                                               42,43,44,45,46,47,48}};
      unsigned cell_is_on_rank_tgt[NUM_TGT][49] = {{-1,-1,-1,-1,-1,-1,-1,
                                                    -1,-1,-1,-1,-1,-1,-1,
                                                    -1,-1,-1,-1,-1,-1,-1,
                                                    -1,-1,-1,-1,-1,-1,-1,
                                                    -1,-1,-1,-1,-1,-1,-1,
                                                    -1,-1,-1,-1,-1,-1,-1,
                                                    -1,-1,-1,-1,-1,-1,-1}};

      for (int i = 0; i < NUM_TGT; ++i)
         yac_init_subdomain_unstruct(&subdomain_tgt[i], grid_tgt[i], comm_named_tgt[i],
                                 global_corner_ids_tgt[i], corner_is_on_rank_tgt[i],
                                 global_cell_ids_tgt[i], cell_is_on_rank_tgt[i],
                                 NULL, NULL);

      struct domain l_domain_tgt[NUM_TGT];

      struct subdomain * blocks_tgt[NUM_TGT];

      for (unsigned i = 0; i < NUM_TGT; ++i)
         blocks_tgt[i] = &subdomain_tgt[i];

      for (int i = 0; i < NUM_TGT; ++i)
         yac_init_domain(&l_domain_tgt[i], comm_named_tgt[i], &blocks_tgt[i], 1, "domain_tgt");

      // finish domain definition
      for (int i = 0; i < NUM_SRC; ++i)
         yac_enddef_domain(&l_domain_src[i]);
      for (int i = 0; i < NUM_TGT; ++i)
         yac_enddef_domain(&l_domain_tgt[i]);

      //-------------------------------------
      // init global search on both processes
      //-------------------------------------

      struct global_search * search[NUM_SRC + NUM_TGT];

      for (int i = 0; i < NUM_SRC; ++i)
         search[i] =
            yac_global_search_new(&l_domain_src[i], "domain_tgt", comm_src[i],
                                  1,0, "component_src", "component_tgt");
      for (int i = 0; i < NUM_TGT; ++i)
         search[NUM_SRC + i] =
            yac_global_search_new(&l_domain_tgt[i], "domain_src", comm_tgt[i],
                                  0,1, "component_tgt", "component_src");

      //---------------------------------------
      // finish initialisation of global search
      //---------------------------------------

      for (int i = 0; i < NUM_SRC + NUM_TGT; ++i)
         yac_enddef_global_search(search[i]);

      //---------------------
      // check source results
      //---------------------

      // for all src processes
      for (k = 0; k < NUM_SRC; ++k) {

         for (i = 0; i < NUM_TGT; ++i)
            if (yac_source_process_has_overlap_with(search[k], i))
               PUT_ERR("error in routine source_process_has_overlap_with\n");
      }

      //---------------------------------------
      // clean up
      //---------------------------------------

      for (int i = 0; i < NUM_SRC + NUM_TGT; ++i)
         yac_global_search_delete(search[i]);

      for (int i = 0; i < NUM_SRC; ++i)
         yac_free_domain(&l_domain_src[i]);
      for (int i = 0; i < NUM_TGT; ++i)
         yac_free_domain(&l_domain_tgt[i]);

      for (int i = 0; i < NUM_SRC; ++i)
         yac_delete_grid(grid_src[i]);
      for (int i = 0; i < NUM_TGT; ++i)
         yac_delete_grid(grid_tgt[i]);

      for (int i = 0; i < NUM_SRC; ++i)
         yac_free_communicator(comm_src[i]);
      for (int i = 0; i < NUM_TGT; ++i)
         yac_free_communicator(comm_tgt[i]);

      for (int i = 0; i < NUM_SRC; ++i)
         yac_free_communicator(comm_named_src[i]);
      for (int i = 0; i < NUM_TGT; ++i)
         yac_free_communicator(comm_named_tgt[i]);
   }

   {
      // the global source domain is a 6x5 grid (4 processes):
      //
      // 35-----36-----37-----38-----39-----40-----41
      //  |      |      |      |      |      |      |
      //  |  24  |  25  |  26  |  27  |  28  |  29  |
      //  |      |      |      |      |      |      |
      // 28-----29-----30-----31-----32-----33-----34
      //  |      |      |      |      |      |      |
      //  |  18  |  19  |  20  |  21  |  22  |  23  |
      //  |      |      |      |      |      |      |
      // 21-----22-----23-----24-----25-----26-----27
      //  |      |      |      |      |      |      |
      //  |  12  |  13  |  14  |  15  |  16  |  17  |
      //  |      |      |      |      |      |      |
      // 14-----15-----16-----17-----18-----19-----20
      //  |      |      |      |      |      |      |
      //  |  06  |  07  |  08  |  09  |  10  |  11  |
      //  |      |      |      |      |      |      |
      // 07-----08-----09-----10-----11-----12-----13
      //  |      |      |      |      |      |      |
      //  |  00  |  01  |  02  |  03  |  04  |  05  |
      //  |      |      |      |      |      |      |
      // 00-----01-----02-----03-----04-----05-----06
      //
      // the domain is distributed among the processes as follows:
      // (index == process)
      //
      // +---+---+---+---+---+---+
      // | 2 | 2 | 2 | 3 | 3 | 3 |
      // +---+---+---+---+---+---+
      // | 2 | 2 | 2 | 3 | 3 | 3 |
      // +---+---+---+---+---+---+
      // | 2 | 2 | 2 | 3 | 3 | 3 |
      // +---+---+---+---+---+---+
      // | 0 | 0 | 0 | 1 | 1 | 1 |
      // +---+---+---+---+---+---+
      // | 0 | 0 | 0 | 1 | 1 | 1 |
      // +---+---+---+---+---+---+
      //
      // the global target domain is also a 5x6 grid (4 process)
      //
      // 36-----37-----38-----39-----40-----41
      //  |      |      |      |      |      |
      //  |  25  |  26  |  27  |  28  |  29  |
      //  |      |      |      |      |      |
      // 30-----31-----32-----33-----34-----35
      //  |      |      |      |      |      |
      //  |  20  |  21  |  22  |  23  |  24  |
      //  |      |      |      |      |      |
      // 24-----25-----26-----27-----28-----29
      //  |      |      |      |      |      |
      //  |  15  |  16  |  17  |  18  |  19  |
      //  |      |      |      |      |      |
      // 18-----19-----20-----21-----22-----23
      //  |      |      |      |      |      |
      //  |  10  |  11  |  12  |  13  |  14  |
      //  |      |      |      |      |      |
      // 12-----13-----14-----15-----16-----17
      //  |      |      |      |      |      |
      //  |  05  |  06  |  07  |  08  |  09  |
      //  |      |      |      |      |      |
      // 06-----07-----08-----09-----10-----11
      //  |      |      |      |      |      |
      //  |  00  |  01  |  02  |  03  |  04  |
      //  |      |      |      |      |      |
      // 00-----01-----02-----03-----04-----05
      //
      // the domain is distributed among the processes as follows:
      // (index == process)
      //
      // +---+---+---+---+---+
      // | 2 | 2 | 3 | 3 | 3 |
      // +---+---+---+---+---+
      // | 2 | 2 | 3 | 3 | 3 |
      // +---+---+---+---+---+
      // | 2 | 2 | 3 | 3 | 3 |
      // +---+---+---+---+---+
      // | 0 | 0 | 0 | 1 | 1 |
      // +---+---+---+---+---+
      // | 0 | 0 | 0 | 1 | 1 |
      // +---+---+---+---+---+
      // | 0 | 0 | 0 | 1 | 1 |
      // +---+---+---+---+---+
      //
      //---------------
      // setup
      //---------------

      // comm worlds on both processes

#undef NUM_SRC
#undef NUM_TGT
#define NUM_SRC 4
#define NUM_TGT 4

      struct communicator * comm_src[NUM_SRC], * comm_tgt[NUM_TGT];

      for (int i = 0; i < NUM_TGT; ++i)
         comm_tgt[i] = yac_local_comm_new();
      for (int i = 0; i < NUM_SRC; ++i)
         comm_src[i] = yac_local_comm_new();

      // named comms for the domains on both processes
      struct communicator * comm_named_src[NUM_SRC], * comm_named_tgt[NUM_TGT];

      for (int i = 0; i < NUM_SRC; ++i)
         comm_named_src[i] = yac_named_local_comm_new(comm_src[i], "src");
      for (int i = 0; i < NUM_TGT; ++i)
         comm_named_tgt[i] = yac_named_local_comm_new(comm_tgt[i], "tgt");

      // define domains on source

      struct grid * grid_src[NUM_SRC];
      unsigned num_cells_src[NUM_SRC][2] = {{4,3}, {4,3}, {4,4}, {4,4}};
      unsigned cyclic_src[2] = {0,0};
      double coordinates_x_src[NUM_SRC][5] = {{-3,-2,-1,0,1},{-1,0,1,2,3},
                                              {-3,-2,-1,0,1},{-1,0,1,2,3}};
      double coordinates_y_src[NUM_SRC][5] = {{-3,-2,-1,0},{-3,-2,-1,0},
                                              {-2,-1,0,1,2},{-2,-1,0,1,2}};

      for ( unsigned j = 0; j < NUM_SRC; ++j ) {
         for ( unsigned i = 0; i < 5; ++i ) {
            coordinates_x_src[j][i] *= rad;
            coordinates_y_src[j][i] *= rad;
         }
      }

      for (int i = 0; i < NUM_SRC; ++i)
         grid_src[i] = yac_reg2d_grid_new(coordinates_x_src[i],
                                          coordinates_y_src[i],
                                          num_cells_src[i], cyclic_src);

      struct subdomain subdomain_src[NUM_SRC];

      int global_corner_ids_src[NUM_SRC][25] =
        {{0,1,2,3,4, 7,8,9,10,11, 14,15,16,17,18, 21,22,23,24,25},
         {2,3,4,5,6, 9,10,11,12,13, 16,17,18,19,20, 23,24,25,26,27},
         {7,8,9,10,11, 14,15,16,17,18, 21,22,23,24,25, 28,29,30,31,32, 35,36,37,38,39},
         {9,10,11,12,13, 16,17,18,19,20, 23,24,25,26,27, 30,31,32,33,34, 37,38,39,40,41}};
      unsigned corner_is_on_rank_src[NUM_SRC][25] = 
        {{-1,-1,-1,-1,1, -1,-1,-1,-1,1, -1,-1,-1,-1,1, 2,2,2,2,3},
         {0,-1,-1,-1,-1, 0,-1,-1,-1,-1, 0,-1,-1,-1,-1, 2,3,3,3,3},
         {0,0,0,0,1, -1,-1,-1,-1,3, -1,-1,-1,-1,3, -1,-1,-1,-1,3, -1,-1,-1,-1,3},
         {0,1,1,1,1, 2,-1,-1,-1,-1, 2,-1,-1,-1,-1, 2,-1,-1,-1,-1, 2,-1,-1,-1,-1}};
      int global_cell_ids_src[NUM_SRC][16] = 
        {{0,1,2,3, 6,7,8,9, 12,13,14,15},
         {2,3,4,5, 8,9,10,11, 14,15,16,17},
         {6,7,8,9, 12,13,14,15, 18,19,20,21, 24,25,26,27},
         {8,9,10,11, 14,15,16,17, 20,21,22,23, 26,27,28,29}};
      unsigned cell_is_on_rank_src[NUM_SRC][16] =
        {{-1,-1,-1,1, -1,-1,-1,1, 2,2,2,3},
         {0,-1,-1,-1, 0,-1,-1,-1, 2,3,3,3},
         {0,0,0,1, -1,-1,-1,3, -1,-1,-1,3, -1,-1,-1,3},
         {0,1,1,1, 2,-1,-1,-1, 2,-1,-1,-1, 2,-1,-1,-1}};

      for (int i = 0; i < NUM_SRC; ++i)
         yac_init_subdomain_unstruct(&subdomain_src[i], grid_src[i], comm_named_src[i],
                                 global_corner_ids_src[i], corner_is_on_rank_src[i],
                                 global_cell_ids_src[i], cell_is_on_rank_src[i],
                                 NULL, NULL);

      struct domain l_domain_src[NUM_SRC];

      struct subdomain * blocks_src[NUM_SRC];

      for (unsigned i = 0; i < NUM_SRC; ++i)
         blocks_src[i] = &subdomain_src[i];

      for (int i = 0; i < NUM_SRC; ++i)
         yac_init_domain(&l_domain_src[i], comm_named_src[i], &blocks_src[i], 1, "domain_src");

      // define a domain on the target

      struct grid * grid_tgt[NUM_TGT];
      unsigned num_cells_tgt[NUM_TGT][2] = {{4,4},{3,4},{3,4},{4,4}};
      unsigned cyclic_tgt[2] = {0,0};
      double coordinates_x_tgt[NUM_TGT][6] = {{-2.5,-1.5,-0.5,0.5,1.5},
                                              {-0.5,0.5,1.5,2.5},
                                              {-2.5,-1.5,-0.5,0.5},
                                              {-1.5,-0.5,0.5,1.5,2.5}};
      double coordinates_y_tgt[NUM_TGT][6] = {{-2.5,-1.5,-0.5,0.5,1.5},
                                              {-2.5,-1.5,-0.5,0.5,1.5},
                                              {-0.5,0.5,1.5,2.5,3.5},
                                              {-0.5,0.5,1.5,2.5,3.5}};

      for ( unsigned i = 0; i < NUM_TGT; ++i ) {
         for ( unsigned j = 0; j < 6; ++j ) {
            coordinates_x_tgt[i][j] *= rad;
            coordinates_y_tgt[i][j] *= rad;
         }
      }

      for (int i = 0; i < NUM_TGT; ++i)
         grid_tgt[i] = yac_reg2d_grid_new(coordinates_x_tgt[i],
                                          coordinates_y_tgt[i],
                                          num_cells_tgt[i], cyclic_tgt);

      struct subdomain subdomain_tgt[NUM_TGT];

      int global_corner_ids_tgt[NUM_TGT][25] =
        {{0,1,2,3,4, 6,7,8,9,10, 12,13,14,15,16, 18,19,20,21,22, 24,25,26,27,28},
         {2,3,4,5, 8,9,10,11, 14,15,16,17, 20,21,22,23, 26,27,28,29},
         {12,13,14,15, 18,19,20,21, 24,25,26,27, 30,31,32,33, 36,37,38,39},
         {13,14,15,16,17, 19,20,21,22,23, 15,26,27,28,29, 31,32,33,34,35, 37,38,39,40,41}};
      unsigned corner_is_on_rank_tgt[NUM_TGT][25] =
        {{-1,-1,-1,-1,1, -1,-1,-1,-1,1, -1,-1,-1,-1,1, -1,-1,-1,-1,1, 2,2,2,3,3},
         {0,-1,-1,-1, 0,-1,-1,-1, 0,-1,-1,-1, 0,-1,-1,-1, 3,3,3,3},
         {0,0,0,0, -1,-1,-1,3, -1,-1,-1,3, -1,-1,-1,3, -1,-1,-1,3},
         {0,0,0,1,1, 2,-1,-1,-1,-1, 2,-1,-1,-1,-1, 2,-1,-1,-1,-1, 2,-1,-1,-1,-1}};
      int global_cell_ids_tgt[NUM_TGT][16] =
        {{0,1,2,3, 5,6,7,8, 10,11,12,13, 15,16,17,18},
         {2,3,4, 7,8,9, 12,13,14, 17,18,19},
         {10,11,12, 15,16,17, 20,21,22, 25,26,27},
         {11,12,13,14, 16,17,18,19, 21,22,23,24, 26,27,28,29}};
      unsigned cell_is_on_rank_tgt[NUM_TGT][16] =
        {{-1,-1,-1,1, -1,-1,-1,1, -1,-1,-1,1, 2,2,3,3},
         {0,-1,-1, 0,-1,-1, 0,-1,-1, 3,3,3},
         {0,0,0, -1,-1,3, -1,-1,3, -1,-1,3},
         {0,0,1,1, 2,-1,-1,-1, 2,-1,-1,-1, 2,-1,-1,-1}};

      for (int i = 0; i < NUM_TGT; ++i)
         yac_init_subdomain_unstruct(&subdomain_tgt[i], grid_tgt[i], comm_named_tgt[i],
                                 global_corner_ids_tgt[i], corner_is_on_rank_tgt[i],
                                 global_cell_ids_tgt[i], cell_is_on_rank_tgt[i],
                                 NULL, NULL);

      struct domain l_domain_tgt[NUM_TGT];

      struct subdomain * blocks_tgt[NUM_TGT];

      for (unsigned i = 0; i < NUM_TGT; ++i)
         blocks_tgt[i] = &subdomain_tgt[i];

      for (int i = 0; i < NUM_TGT; ++i)
         yac_init_domain(&l_domain_tgt[i], comm_named_tgt[i], &blocks_tgt[i], 1, "domain_tgt");

      // finish domain definition
      for (int i = 0; i < NUM_SRC; ++i)
         yac_enddef_domain(&l_domain_src[i]);
      for (int i = 0; i < NUM_TGT; ++i)
         yac_enddef_domain(&l_domain_tgt[i]);

      //-------------------------------------
      // init global search on both processes
      //-------------------------------------

      struct global_search * search[NUM_SRC + NUM_TGT];

      for (int i = 0; i < NUM_SRC; ++i)
         search[i] =
            yac_global_search_new(&l_domain_src[i], "domain_tgt", comm_src[i],
                                  1,0, "component_src", "component_tgt");
      for (int i = 0; i < NUM_TGT; ++i)
         search[NUM_SRC + i] =
            yac_global_search_new(&l_domain_tgt[i], "domain_src", comm_tgt[i],
                                  0,1, "component_tgt", "component_src");

      //---------------------------------------
      // finish initialisation of global search
      //---------------------------------------

      for (int i = 0; i < NUM_SRC + NUM_TGT; ++i)
         yac_enddef_global_search(search[i]);

      //---------------------
      // check source results
      //---------------------

      // number of overlapping source cells per target
      unsigned ref_num_src_per_tgt[30] = {4,4,4,4,4,
                                          4,4,4,4,4,
                                          4,4,4,4,4,
                                          4,4,4,4,4,
                                          2,2,2,2,2,
                                          0,0,0,0,0};

      int ref_src_per_tgt[30][4] = {{0,1,6,7},{1,2,7,8},{2,3,8,9},{3,4,9,10},{4,5,10,11},
                                    {6,7,12,13},{7,8,13,14},{8,9,14,15},{9,10,15,16},{10,11,16,17},
                                    {12,13,18,19},{13,14,19,20},{14,15,20,21},{15,16,21,22},{16,17,22,23},
                                    {18,19,24,25},{19,20,25,26},{20,21,26,27},{21,22,27,28},{22,23,28,29},
                                    {24,25},{25,26},{26,27},{27,28},{28,29},
                                    {-1},{-1},{-1},{-1},{-1}};

      unsigned tgt_cell_owner[74] = {0,0,0,1,1, 0,0,0,1,1, 0,0,0,1,1,
                                     2,2,3,3,3, 2,2,3,3,3, 2,2,3,3,3};

      int global_tgt_ids[30];
      for (unsigned i = 0; i < 30; ++i) global_tgt_ids[i] = i;

      struct source_search_results const * src_search_results[4][30];

      // for all src processes
      for (i = 0; i < NUM_SRC; ++i)
        yac_get_search_results_for_tgt_cells(search[i], global_tgt_ids, 30,
                                             src_search_results[i]);

      // for all target cells
      for (unsigned i = 0; i < 30; ++i) {

        unsigned src_proc_mask = 0;
        for (unsigned j = 0; j < ref_num_src_per_tgt[i]; ++j)
          for (unsigned k = 0; k < NUM_SRC; ++k)
            for (unsigned l = 0; l < num_cells_src[k][0] * num_cells_src[k][1]; ++l)
              src_proc_mask |=
                (ref_src_per_tgt[i][j] == global_cell_ids_src[k][l]) << k;

        // for all src processes
        for (j = 0; j < NUM_SRC; ++j) {

          unsigned src_overlaps_with_tgt_cell = src_proc_mask & (1 << j);

          struct source_search_results const * curr_result =
            yac_get_search_results_for_tgt_cell(search[j], global_tgt_ids[i]);

          if (curr_result != src_search_results[j][i])
            PUT_ERR("error in get_search_results_for_tgt_cells\n");

          if (curr_result == NULL) continue;

          if (curr_result->global_id != i)
            PUT_ERR("error in get_search_results_for_tgt_cell; wrong global id\n");

          if (curr_result->tgt_rank != tgt_cell_owner[i])
            PUT_ERR("error in get_search_results_for_tgt_cell; wrong rank\n");

          if (curr_result->tgt_domain == NULL)
            PUT_ERR("error in get_search_results_for_tgt_cell; no target domain\n");

          if (!src_overlaps_with_tgt_cell) {
            if (curr_result != NULL)
              PUT_ERR("error in get_search_results_for_tgt_cell\n");
            continue;
          }

          unsigned ref_num_other_src_ranks = 0;
          unsigned ref_other_src_ranks[NUM_SRC - 1];

          for (unsigned k = 0; k < NUM_SRC; ++k) {
            if ((k == j) || !(src_proc_mask & (1 << k))) continue;
            ref_other_src_ranks[ref_num_other_src_ranks] = k + 4;
            ref_num_other_src_ranks++;
          }

          if (curr_result->num_other_src_ranks != ref_num_other_src_ranks)
            PUT_ERR("error in get_search_results_for_tgt_cell; wrong number "
                    "of other source ranks\n");

          if (!check_ids((int*)(curr_result->other_src_ranks),
                         (int*)(ref_other_src_ranks), ref_num_other_src_ranks))
            PUT_ERR("error in get_search_results_for_tgt_cell; wrong "
                    "other source ranks\n");

          unsigned curr_ref_num_src_per_tgt = 0;
          unsigned curr_ref_src_per_tgt[4];

          for (unsigned k = 0; k < ref_num_src_per_tgt[i]; ++k) {
            for (unsigned l = 0; l < num_cells_src[j][0] * num_cells_src[j][1]; ++l)
              if (global_cell_ids_src[j][l] == ref_src_per_tgt[i][k]) {
              curr_ref_src_per_tgt[curr_ref_num_src_per_tgt] = l;
              curr_ref_num_src_per_tgt++;
            }
          }

          if (curr_result->num_overlapping_src_cells_per_subdomain[0] !=
              curr_ref_num_src_per_tgt)
            PUT_ERR("error in get_search_results_for_tgt_cell; wrong number of source cells\n");

          if (!check_ids((int*)(curr_result->overlapping_src_cells[0]),
              (int*)curr_ref_src_per_tgt, curr_ref_num_src_per_tgt))
            PUT_ERR("error in get_search_results_for_tgt_cell; wrong source cells\n");
        }

      }

      //---------------------------------------
      // clean up
      //---------------------------------------

      for (int i = 0; i < NUM_SRC + NUM_TGT; ++i)
         yac_global_search_delete(search[i]);

      for (int i = 0; i < NUM_SRC; ++i)
         yac_free_domain(&l_domain_src[i]);
      for (int i = 0; i < NUM_TGT; ++i)
         yac_free_domain(&l_domain_tgt[i]);

      for (int i = 0; i < NUM_SRC; ++i)
         yac_delete_grid(grid_src[i]);
      for (int i = 0; i < NUM_TGT; ++i)
         yac_delete_grid(grid_tgt[i]);

      for (int i = 0; i < NUM_SRC; ++i)
         yac_free_communicator(comm_src[i]);
      for (int i = 0; i < NUM_TGT; ++i)
         yac_free_communicator(comm_tgt[i]);

      for (int i = 0; i < NUM_SRC; ++i)
         yac_free_communicator(comm_named_src[i]);
      for (int i = 0; i < NUM_TGT; ++i)
         yac_free_communicator(comm_named_tgt[i]);
   }

   yac_comm_finalize();

   return TEST_EXIT_CODE;
}
