/**
 * @file test_global_search_parallel.c
 *
 * @copyright Copyright  (C)  2015 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#ifdef USE_MPI
#include <mpi.h>
#endif

#include "tests.h"

#include "grid.h"
#include "grid_reg2d.h"
#include "communicator.h"
#include "communicator_mpi.h"
#include "domain.h"
#include "global_search.h"

void submain_1(struct communicator * global_comm,
               struct communicator * domain_comm);

void submain_2(struct communicator * global_comm,
               struct communicator * domain_comm);

int check_ids(int const * ids_a, int const * ids_b, unsigned num_ids) {

   unsigned i, j, num_matches = 0;

   for (i = 0; i < num_ids; ++i)
      for (j = 0; j < num_ids; ++j)
         if (ids_a[i] == ids_b[j]) {
            num_matches++;
            break;
         }

   return num_matches == num_ids;
}

int main (void) {

#ifndef USE_MPI
   return EXIT_SKIP_TEST;
#else

   struct communicator * global_comm;

   global_comm = yac_mpi_comm_new();

   if (yac_get_comm_size(global_comm) == 4)
   {

      int my_global_rank;

      my_global_rank = yac_get_comm_rank(global_comm);

      struct communicator * comm;

      comm = yac_comm_split(global_comm, my_global_rank < 2, 0);

      if (my_global_rank < 2)
         submain_1(global_comm, comm);
      else
         submain_2(global_comm, comm);

      yac_free_communicator(comm);
   }

   yac_free_communicator(global_comm);

   MPI_Finalize();

   return TEST_EXIT_CODE;
#endif
}

/*
 * The domain is distributed among 2 processes, with two subdomains each.
 *
 * The global domain is a 4x4 grid:
 *
 *    20------21------22------23------24
 *    |       |       |       |       |
 *    |   12  |   13  |   14  |   15  |
 *    |       |       |       |       |
 *    15------16------17------18------19
 *    |       |       |       |       |
 *    |   08  |   09  |   10  |   11  |
 *    |       |       |       |       |
 *    10------11------12------13------14
 *    |       |       |       |       |
 *    |   04  |   05  |   06  |   07  |
 *    |       |       |       |       |
 *    05------06------07------08------09
 *    |       |       |       |       |
 *    |   00  |   01  |   02  |   03  |
 *    |       |       |       |       |
 *    00------01------02------03------04
 *
 * Subdomains are distributed as follows (xy => x = domain, y = subdomain)
 *
 *    10------10------01------01------01
 *    |       |       |       |       |
 *    |   10  |   10  |   01  |   01  | 
 *    |       |       |       |       |
 *    10------10------10------01------01
 *    |       |       |       |       |
 *    |   10  |   10  |   01  |   01  | 
 *    |       |       |       |       |
 *    10------00------01------11------01
 *    |       |       |       |       |
 *    |   00  |   00  |   11  |   11  | 
 *    |       |       |       |       |
 *    00------00------00------11------11
 *    |       |       |       |       |
 *    |   00  |   00  |   11  |   11  | 
 *    |       |       |       |       |
 *    00------00------11------11------11
 */
void submain_2(struct communicator * global_comm,
               struct communicator * domain_comm) {

   unsigned i, j, k, my_rank;

   my_rank = yac_get_comm_rank(domain_comm);

   struct grid * grid[2];
   unsigned num_cells[2] = {2,2};
   unsigned cyclic[2] = {0,0};
   double coordinates_x[2][2][3] = {{{0,1,2}, {2,3,4}},{{0,1,2}, {2,3,4}}};
   double coordinates_y[2][2][3] = {{{0,1,2}, {2,3,4}},{{2,3,4}, {0,1,2}}};

   for (i = 0; i < 2; ++i) {
      for (j = 0; j < 2; ++j) {
         for (k = 0; k < 3; ++k) {
            coordinates_x[i][j][k] *= rad;
            coordinates_y[i][j][k] *= rad;
         }
      }
   }

   for (i = 0; i < 2; ++i)
      grid[i] = yac_reg2d_grid_new(coordinates_x[my_rank][i], coordinates_y[my_rank][i],
                                   num_cells, cyclic);

   struct subdomain subdomain[2];

   int global_corner_ids[2][2][9] =
      {{{ 0, 1, 2, 5, 6, 7,10,11,12},{12,13,14,17,18,19,22,23,24}},
       {{10,11,12,15,16,17,20,21,22},{ 2, 3, 4, 7, 8, 9,12,13,14}}};
   unsigned corner_is_on_rank[2][2][9] =
      {{{-1,-1, 1,-1,-1,-1, 1,-1, 0},{-1, 1,-1, 1,-1,-1,-1,-1,-1}},
       {{-1, 0, 0,-1,-1,-1,-1,-1, 0},{-1,-1,-1, 0,-1,-1, 0,-1, 0}}};
   int global_cell_ids[2][2][4] =
      {{{ 0, 1, 4, 5},{10,11,14,15}},{{ 8, 9,12,13},{ 2, 3, 6 ,7}}};
   unsigned cell_is_on_rank[2][2][4] =
      {{{-1,-1,-1,-1},{-1,-1,-1,-1}},{{-1,-1,-1,-1},{-1,-1,-1,-1}}};

   for (i = 0; i < 2; ++i)
      yac_init_subdomain_unstruct(subdomain+i, grid[i], domain_comm,
                                  global_corner_ids[my_rank][i],
                                  corner_is_on_rank[my_rank][i],
                                  global_cell_ids[my_rank][i],
                                  cell_is_on_rank[my_rank][i], NULL, NULL);

   struct domain domain;
   struct subdomain * subdomains[2] = {subdomain+0, subdomain+1};

   yac_init_domain(&domain, domain_comm, subdomains, 2, "domain 2");
   yac_enddef_domain(&domain);

   struct global_search * search;

   search = yac_global_search_new(&domain, "domain 1", global_comm, 1, 1,
                                  "component 2", "componente 1");
   yac_enddef_global_search(search);

   //---------------------
   // check source results
   //---------------------

   // dimensions [process][subdomain][tgt cell]
   unsigned ref_num_src_per_tgt[2][2][16] =
      {{{4,2,0,0, 2,1,0,0, 0,0,0,0, 0,0,0,0},
        {0,0,0,0, 0,1,2,1, 0,2,4,2, 0,1,2,1}},
       {{0,0,0,0, 2,1,0,0, 4,2,0,0, 2,1,0,0},
        {0,2,4,2, 0,1,2,1, 0,0,0,0, 0,0,0,0}}};
   // dimensions [process][subdomain][tgt cell][src cell]
   unsigned ref_src_per_tgt[2][2][16][4] =
      {{{{0,1,2,3},{1,3},{-1},{-1}, {2,3},{3},{-1},{-1},
         {-1},{-1},{-1},{-1}, {-1},{-1},{-1},{-1}},
        {{-1},{-1},{-1},{-1}, {-1},{0},{0,1},{1},
         {-1},{0,2},{0,1,2,3},{1,3}, {-1},{2},{2,3},{3}}},
       {{{-1},{-1},{-1},{-1}, {0,1},{1},{-1},{-1},
         {0,1,2,3},{1,3},{-1},{-1}, {2,3},{3},{-1},{-1}},
        {{-1},{0,2},{0,1,2,3},{1,3}, {-1},{2},{2,3},{3},
         {-1},{-1},{-1},{-1}, {-1},{-1},{-1},{-1}}}};
   // dimensions [process][tgt_cell]
   unsigned ref_num_other_src_ranks[2][16];
   for (unsigned i = 0; i < 2; ++i)
     for (unsigned j = 0; j < 16; ++j)
       ref_num_other_src_ranks[i][j] = ref_num_src_per_tgt[i^1][0][j] ||
                                       ref_num_src_per_tgt[i^1][1][j];

   int ref_global_tgt_ids[16] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15};
   unsigned ref_tgt_rank[16] = {0,0,0,0, 1,1,1,1, 0,0,0,0, 1,1,1,1};

   struct source_search_results const * src_search_results[16];

   yac_get_search_results_for_tgt_cells(search, ref_global_tgt_ids, 16,
                                        src_search_results);

   for (i = 0; i < 16; ++i) {

      struct source_search_results const * curr_result =
         yac_get_search_results_for_tgt_cell(search, ref_global_tgt_ids[i]);

      if (src_search_results[i] != curr_result)
         PUT_ERR("error in get_search_results_for_tgt_cells\n");

      if ((ref_num_src_per_tgt[my_rank][0][i] ||
           ref_num_src_per_tgt[my_rank][1][i]) ^ (curr_result != NULL))
        PUT_ERR("error in get_search_results_for_tgt_cells\n");

      if (curr_result == NULL) continue;

      if (curr_result->global_id != ref_global_tgt_ids[i])
         PUT_ERR("error in get_search_results_for_tgt_cell; wrong global id\n");

      if (curr_result->tgt_rank != ref_tgt_rank[i])
         PUT_ERR("error in get_search_results_for_tgt_cell; wrong rank\n");

      if (curr_result->tgt_domain == NULL)
         PUT_ERR("error in get_search_results_for_tgt_cell; no target domain\n");

      for (unsigned subdomain_idx = 0; subdomain_idx < 2; ++subdomain_idx) {

         if (curr_result->num_overlapping_src_cells_per_subdomain[subdomain_idx] !=
             ref_num_src_per_tgt[my_rank][subdomain_idx][i])
            PUT_ERR("error in get_search_results_for_tgt_cell; wrong number of\n");

         if (!check_ids((int*)(curr_result->overlapping_src_cells[subdomain_idx]),
                               (int*)ref_src_per_tgt[my_rank][subdomain_idx][i],
                               ref_num_src_per_tgt[my_rank][subdomain_idx][i]))
            PUT_ERR("error in get_search_results_for_tgt_cell; wrong source cells\n");
      }


      if (curr_result->num_other_src_ranks != ref_num_other_src_ranks[my_rank][i])
         PUT_ERR("error in get_search_results_for_tgt_cell; wrong number of "
                 "other source ranks\n");

      if (curr_result->num_other_src_ranks != 0) {
        if ((curr_result->other_src_ranks == NULL) ||
            (curr_result->other_src_ranks[0] != ((my_rank^1)+2)))
           PUT_ERR("error in get_search_results_for_tgt_cell; other source ranks\n");
      }
   }

   unsigned ref_overlaps[4] = {1,1,0,0};

   for (unsigned i = 0; i < 4; ++i)
      if (yac_source_process_has_overlap_with(search, i) != ref_overlaps[i])
         PUT_ERR("error in routine source_process_has_overlap_with\n");

   //---------------------
   // check target results
   //---------------------

   // dimensions [rank][tgt rank][subdomain]
   unsigned ref_num_overlaping_cells[2][2][2] = {{{4,4},{2,4}},{{4,4},{4,2}}};
   // dimensions [rank][tgt rank][subdomain][src cells]
   unsigned ref_matching_local_cell_ids[2][2][2][4] =
      {{{{0,1,2,3},{0,1,2,3}},{{2,3},{0,1,2,3}}},
       {{{0,1,2,3},{0,1,2,3}},{{0,1,2,3},{2,3}}}};

   for (unsigned i = 0; i < 2; ++i) {

      struct dep_list const * matching_local_cells =
         yac_get_matching_local_cell_ids(search, i);

      if (matching_local_cells == NULL) {

        PUT_ERR("error in global search results\n");
        continue;
      }

      if (matching_local_cells->num_elements != 2)
         PUT_ERR("error in global search results\n");

      for (unsigned subdomain_idx = 0; subdomain_idx < 2; ++subdomain_idx) {
         if (matching_local_cells->num_deps_per_element[subdomain_idx] !=
             ref_num_overlaping_cells[my_rank][i][subdomain_idx])
            PUT_ERR("error in global search results\n");


         if (!check_ids((int*)yac_get_dependencies_of_element(
                        *matching_local_cells, subdomain_idx),
                        (int*)ref_matching_local_cell_ids[my_rank][i][subdomain_idx],
                        ref_num_overlaping_cells[my_rank][i][subdomain_idx]))
            PUT_ERR("error in global search results\n");
      }

      if (!yac_target_process_has_overlap_with(search, i))
         PUT_ERR("error in routine target_process_has_overlap_with\n");
   }

   //--------
   // cleanup
   //--------

   yac_global_search_delete(search);
   yac_free_domain(&domain);
   for (i = 0; i < 2; ++i)
      yac_delete_grid(grid[i]);
}

/*
 * The domain is distributed among 2 processes with two subdomains each
 *
 * The global domain is a 4x4 grid:
 *
 *    20------21------22------23------24
 *    |       |       |       |       |
 *    |   12  |   13  |   14  |   15  |
 *    |       |       |       |       |
 *    15------16------17------18------19
 *    |       |       |       |       |
 *    |   08  |   09  |   10  |   11  |
 *    |       |       |       |       |
 *    10------11------12------13------14
 *    |       |       |       |       |
 *    |   04  |   05  |   06  |   07  |
 *    |       |       |       |       |
 *    05------06------07------08------09
 *    |       |       |       |       |
 *    |   00  |   01  |   02  |   03  |
 *    |       |       |       |       |
 *    00------01------02------03------04
 *
 * Subdomains are distributed as follows (xy => x = domain, y = subdomain)
 *
 *    10------10------10------10------10
 *    |       |       |       |       |
 *    |   10  |   10  |   10  |   10  | 
 *    |       |       |       |       |
 *    10------10------10------10------10
 *    |       |       |       |       |
 *    |   00  |   00  |   00  |   00  | 
 *    |       |       |       |       |
 *    11------11------00------00------00
 *    |       |       |       |       |
 *    |   11  |   11  |   11  |   11  | 
 *    |       |       |       |       |
 *    01------01------11------11------11
 *    |       |       |       |       |
 *    |   01  |   01  |   01  |   01  | 
 *    |      4 |       |       |       |
 *    01------01------01------01------01
 */
void submain_1(struct communicator * global_comm,
               struct communicator * domain_comm) {

   int my_rank, i, j, k;

   my_rank = yac_get_comm_rank(domain_comm);

   struct grid * grid[2];
   unsigned num_cells[2] = {4,1};
   unsigned cyclic[2] = {0,0};
   double coordinates_x[5] = {0.5,1.5,2.5,3.5,4.5};
   double coordinates_y[2][2][2] = {{{2.5,3.5},{0.5,1.5}},{{3.5,4.5},{1.5,2.5}}};

   for (i = 0; i < 5; ++i)
      coordinates_x[i] *= rad;

   for (i = 0; i < 2; ++i) {
      for (j = 0; j < 2; ++j) {
         for (k = 0; k < 2; ++k) {
            coordinates_y[i][j][k] *= rad;
         }
      }
   }

   for (i = 0; i < 2; ++i)
      grid[i] = yac_reg2d_grid_new(coordinates_x, coordinates_y[my_rank][i],
                                   num_cells, cyclic);

   struct subdomain subdomain[2];

   int global_corner_ids[2][2][10] =
      {{{10,11,12,13,14,15,16,17,18,19},{ 0, 1, 2, 3, 4, 5, 6, 7, 8, 9}},
       {{15,16,17,18,19,20,21,22,23,24},{ 5, 6, 7, 8, 9,10,11,12,13,14}}};
   unsigned corner_is_on_rank[2][2][10] =
      {{{ 1, 1,-1,-1,-1, 1, 1, 1, 1, 1},{-1,-1,-1,-1,-1,-1,-1, 1, 1, 1}},
       {{-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},{ 0, 0,-1,-1,-1,-1,-1, 0, 0, 0}}};
   int global_cell_ids[2][2][4] =
      {{{ 8, 9,10,11},{ 0, 1, 2, 3}},{{12,13,14,15},{ 4, 5, 6, 7}}};
   unsigned cell_is_on_rank[2][2][4] =
      {{{-1,-1,-1,-1},{-1,-1,-1,-1}},{{-1,-1,-1,-1},{-1,-1,-1,-1}}};

   for (i = 0; i < 2; ++i)
      yac_init_subdomain_unstruct(subdomain+i, grid[i], domain_comm,
                                  global_corner_ids[my_rank][i],
                                  corner_is_on_rank[my_rank][i],
                                  global_cell_ids[my_rank][i],
                                  cell_is_on_rank[my_rank][i], NULL, NULL);

   struct domain domain;
   struct subdomain * subdomains[2] = {subdomain+0,subdomain+1};

   yac_init_domain(&domain, domain_comm, subdomains, 2, "domain 1");
   yac_enddef_domain(&domain);

   struct global_search * search;

   search = yac_global_search_new(&domain, "domain 2", global_comm, 1, 1,
                                  "componente 1", "component 2");
   yac_enddef_global_search(search);

   //---------------------
   // check source results
   //---------------------

   // dimensions [process][subdomain][tgt cell]
   unsigned ref_num_src_per_tgt[2][2][16] =
      {{{0,0,0,0, 0,0,0,0, 1,2,2,2, 1,2,2,2},
        {1,2,2,2, 1,2,2,2, 0,0,0,0, 0,0,0,0}},
       {{0,0,0,0, 0,0,0,0, 0,0,0,0, 1,2,2,2},
        {0,0,0,0, 1,2,2,2, 1,2,2,2, 0,0,0,0}}};
   // dimensions [process][subdomain][tgt cell][src cell]
   unsigned ref_src_per_tgt[2][2][16][4] =
      {{{{-1},{-1},{-1},{-1}, {-1},{-1},{-1},{-1},
         {0},{0,1},{1,2},{2,3}, {0},{0,1},{1,2},{2,3}},
        {{0},{0,1},{1,2},{2,3}, {0},{0,1},{1,2},{2,3},
         {-1},{-1},{-1},{-1}, {-1},{-1},{-1},{-1}}},
       {{{-1},{-1},{-1},{-1}, {-1},{-1},{-1},{-1},
         {-1},{-1},{-1},{-1}, {0},{0,1},{1,2},{2,3}},
        {{-1},{-1},{-1},{-1}, {0},{0,1},{1,2},{2,3},
         {0},{0,1},{1,2},{2,3}, {-1},{-1},{-1},{-1}}}};
   // dimensions [process][tgt_cell]
   unsigned ref_num_other_src_ranks[2][16];
   for (unsigned i = 0; i < 2; ++i)
     for (unsigned j = 0; j < 16; ++j)
       ref_num_other_src_ranks[i][j] = ref_num_src_per_tgt[i^1][0][j] ||
                                       ref_num_src_per_tgt[i^1][1][j];

   int ref_global_tgt_ids[16] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15};
   unsigned ref_tgt_rank[16] = {2,2,3,3, 2,2,3,3, 3,3,2,2, 3,3,2,2};

   struct source_search_results const * src_search_results[16];

   yac_get_search_results_for_tgt_cells(search, ref_global_tgt_ids, 16,
                                        src_search_results);

   for (i = 0; i < 16; ++i) {

      struct source_search_results const * curr_result =
         yac_get_search_results_for_tgt_cell(search, ref_global_tgt_ids[i]);

      if (src_search_results[i] != curr_result)
         PUT_ERR("error in get_search_results_for_tgt_cells\n");

      if ((ref_num_src_per_tgt[my_rank][0][i] ||
           ref_num_src_per_tgt[my_rank][1][i]) ^ (curr_result != NULL))
        PUT_ERR("error in get_search_results_for_tgt_cells\n");

      if (curr_result == NULL) continue;

      if (curr_result->global_id != ref_global_tgt_ids[i])
         PUT_ERR("error in get_search_results_for_tgt_cell; wrong global id\n");

      if (curr_result->tgt_rank != ref_tgt_rank[i])
         PUT_ERR("error in get_search_results_for_tgt_cell; wrong rank\n");

      if (curr_result->tgt_domain == NULL)
         PUT_ERR("error in get_search_results_for_tgt_cell; no target domain\n");

      for (unsigned subdomain_idx = 0; subdomain_idx < 2; ++subdomain_idx) {

         if (curr_result->num_overlapping_src_cells_per_subdomain[subdomain_idx] !=
             ref_num_src_per_tgt[my_rank][subdomain_idx][i])
            PUT_ERR("error in get_search_results_for_tgt_cell; wrong number of\n");

         if (!check_ids((int*)(curr_result->overlapping_src_cells[subdomain_idx]),
                               (int*)ref_src_per_tgt[my_rank][subdomain_idx][i],
                               ref_num_src_per_tgt[my_rank][subdomain_idx][i]))
            PUT_ERR("error in get_search_results_for_tgt_cell; wrong source cells\n");
      }


      if (curr_result->num_other_src_ranks != ref_num_other_src_ranks[my_rank][i])
         PUT_ERR("error in get_search_results_for_tgt_cell; wrong number of "
                 "other source ranks\n");

      if (curr_result->num_other_src_ranks != 0) {
        if ((curr_result->other_src_ranks == NULL) ||
            (curr_result->other_src_ranks[0] != (my_rank^1)))
           PUT_ERR("error in get_search_results_for_tgt_cell; other source ranks\n");
      }
   }

   unsigned ref_overlaps[4] = {0,0,1,1};

   for (unsigned i = 0; i < 4; ++i)
      if (yac_source_process_has_overlap_with(search, i) != ref_overlaps[i])
         PUT_ERR("error in routine source_process_has_overlap_with\n");

   //---------------------
   // check target results
   //---------------------

   // dimensions [rank][tgt rank][subdomain]
   unsigned ref_num_overlaping_cells[2][2][2] = {{{3,2},{2,3}},{{3,4},{2,4}}};
   // dimensions [rank][tgt rank][subdomain][src cells]
   unsigned ref_matching_local_cell_ids[2][2][2][4] =
      {{{{1,2,3},{0,1}},{{0,1},{1,2,3}}},
       {{{1,2,3},{0,1,2,3}},{{0,1},{0,1,2,3}}}};

   for (unsigned i = 2; i < 4; ++i) {

      struct dep_list const * matching_local_cells =
         yac_get_matching_local_cell_ids(search, i);

      if (matching_local_cells->num_elements != 2)
         PUT_ERR("error in global search results\n");

      for (unsigned subdomain_idx = 0; subdomain_idx < 2; ++subdomain_idx) {
         if (matching_local_cells->num_deps_per_element[subdomain_idx] !=
             ref_num_overlaping_cells[my_rank][i^2][subdomain_idx])
            PUT_ERR("error in global search results\n");

         if (!check_ids((int*)yac_get_dependencies_of_element(
                        *matching_local_cells, subdomain_idx),
                        (int*)ref_matching_local_cell_ids[my_rank][i^2][subdomain_idx],
                        ref_num_overlaping_cells[my_rank][i^2][subdomain_idx]))
            PUT_ERR("error in global search results\n");
      }

      if (!yac_target_process_has_overlap_with(search, i))
         PUT_ERR("error in routine target_process_has_overlap_with\n");
   }

   //---------
   // clean up
   //---------

   yac_global_search_delete(search);
   yac_free_domain(&domain);
   for (i = 0; i < 2; ++i)
      yac_delete_grid(grid[i]);
}
