/**
 * @file test_global_search_parallel2.c
 *
 * @copyright Copyright  (C)  2015 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#ifdef USE_MPI
#include <mpi.h>
#endif

#include "tests.h"

#include "grid.h"
#include "grid_reg2d.h"
#include "communicator.h"
#include "communicator_mpi.h"
#include "domain.h"
#include "global_search.h"

void target_main(struct communicator * global_comm,
                 struct communicator * target_comm);

void source_main(struct communicator * global_comm,
                 struct communicator * source_comm);

int main (void) {

#ifndef USE_MPI
   return EXIT_SKIP_TEST;
#else

   struct communicator * global_comm;

   global_comm = yac_mpi_comm_new();

   if (yac_get_comm_size(global_comm) == 5)
   { 
      int my_global_rank = yac_get_comm_rank(global_comm);

      // split processes into source and target

      unsigned target = my_global_rank < 1;

      struct communicator * comm;

      comm = yac_comm_split(global_comm, target, 0);

      if (target)
         target_main(global_comm, comm);
      else
         source_main(global_comm, comm);

      yac_free_communicator(comm);
   }

   yac_free_communicator(global_comm);

   MPI_Finalize();

   return TEST_EXIT_CODE;
#endif
}

void source_main(struct communicator * global_comm,
                 struct communicator * source_comm) {

   unsigned i, j;

   // corner and cell ids for a 7 x 7 grid

   // 56-----57-----58-----59-----60-----61-----62-----63
   //  |      |      |      |      |      |      |      |
   //  |  42  |  43  |  44  |  45  |  46  |  47  |  48  |
   //  |      |      |      |      |      |      |      |
   // 48-----49-----50-----51-----52-----53-----54-----55
   //  |      |      |      |      |      |      |      |
   //  |  35  |  36  |  37  |  38  |  39  |  40  |  41  |
   //  |      |      |      |      |      |      |      |
   // 40-----41-----42-----43-----44-----45-----46-----47
   //  |      |      |      |      |      |      |      |
   //  |  28  |  29  |  30  |  31  |  32  |  33  |  34  |
   //  |      |      |      |      |      |      |      |
   // 32-----33-----34-----35-----36-----37-----38-----39
   //  |      |      |      |      |      |      |      |
   //  |  21  |  22  |  23  |  24  |  25  |  26  |  27  |
   //  |      |      |      |      |      |      |      |
   // 24-----25-----26-----27-----28-----29-----30-----31
   //  |      |      |      |      |      |      |      |
   //  |  14  |  15  |  16  |  17  |  18  |  19  |  20  |
   //  |      |      |      |      |      |      |      |
   // 16-----17-----18-----19-----20-----21-----22-----23
   //  |      |      |      |      |      |      |      |
   //  |  07  |  08  |  09  |  10  |  11  |  12  |  13  |
   //  |      |      |      |      |      |      |      |
   // 08-----09-----10-----11-----12-----13-----14-----15
   //  |      |      |      |      |      |      |      |
   //  |  00  |  01  |  02  |  03  |  04  |  05  |  06  |
   //  |      |      |      |      |      |      |      |
   // 00-----01-----02-----03-----04-----05-----06-----07
   //
   // the domain is distributed among the processes as follows:
   // (index == process)
   //
   // 3---3---3---3---3---3---3---3
   // | 3 | 3 | 3 | 3 | 3 | 3 | 3 |
   // 3---3---3---3---3---3---3---3
   // | 3 | 3 | 3 | 3 | 3 | 3 | 3 |
   // 3---3---3---1---2---2---3---3
   // | 1 | 1 | 1 | 2 | 2 | 2 | 2 |
   // 1---1---1---2---2---2---2---2
   // | 1 | 1 | 1 | 2 | 2 | 2 | 2 |
   // 1---1---1---1---2---2---2---2
   // | 1 | 1 | 1 | 2 | 2 | 2 | 2 |
   // 1---1---1---0---0---0---2---2
   // | 0 | 0 | 0 | 0 | 0 | 0 | 0 |
   // 0---0---0---0---0---0---0---0
   // | 0 | 0 | 0 | 0 | 0 | 0 | 0 |
   // 0---0---0---0---0---0---0---0
   //
   //--------------------
   // set up global search
   //--------------------

   int my_source_rank = yac_get_comm_rank(source_comm);

   // define a domain on P1

   struct grid * grid;
   unsigned num_cells[4][2] = {{7,3},{4,5},{5,5},{7,3}};
   unsigned cyclic[2] = {0,0};
   double coordinates_x[4][8] = {{0,1,2,3,4,5,6,7},{0,1,2,3,4},
                                 {2,3,4,5,6,7},{0,1,2,3,4,5,6,7}};
   double coordinates_y[4][8] = {{0,1,2,3},{1,2,3,4,5,6},
                                 {1,2,3,4,5,6},{4,5,6,7}};

   for ( i = 0; i < 4; ++i ) {
      for ( j = 0; j < 8; ++j ) {
         coordinates_x[i][j] *= rad;
         coordinates_y[i][j] *= rad;
      }
   }

   grid = yac_reg2d_grid_new(coordinates_x[my_source_rank], coordinates_y[my_source_rank],
                         num_cells[my_source_rank], cyclic);

   struct subdomain subdomain;

   int global_corner_ids[4][36] = {{ 0, 1, 2, 3, 4, 5, 6, 7,
                                     8, 9,10,11,12,13,14,15,
                                    16,17,18,19,20,21,22,23,
                                    24,25,26,27,28,29,30,31},
                                   { 8, 9,10,11,12,
                                    16,17,18,19,20,
                                    24,25,26,27,28,
                                    32,33,34,35,36,
                                    40,41,42,43,44,
                                    48,49,50,51,52},
                                   {10,11,12,13,14,15,
                                    18,19,20,21,22,23,
                                    26,27,28,29,30,31,
                                    34,35,36,37,38,39,
                                    42,43,44,45,46,47,
                                    50,51,52,53,54,55},
                                   {32,33,34,35,36,37,38,39,
                                    40,41,42,43,44,45,46,47,
                                    48,49,50,51,52,53,54,55,
                                    56,57,58,59,60,61,62,63}};
   unsigned corner_is_on_rank[4][36] = {{-1,-1,-1,-1,-1,-1,-1,-1,
                                         -1,-1,-1,-1,-1,-1,-1,-1,
                                          1, 1, 1,-1,-1,-1, 2, 2},
                                        { 0, 0, 0, 0, 0,
                                         -1,-1,-1, 0, 0,
                                         -1,-1,-1,-1, 2,
                                         -1,-1,-1, 2, 2,
                                          3, 3, 3,-1, 2,
                                          3, 3, 3, 3, 3},
                                        { 0, 0, 0, 0, 0, 0,
                                          1, 0, 0, 0,-1,-1,
                                          1, 1,-1,-1,-1,-1,
                                          1,-1,-1,-1,-1,-1,
                                          3, 1,-1,-1, 3, 3,
                                          3, 3, 3, 3, 3, 3},
                                        { 1, 1, 1, 2, 2, 2, 2, 2,
                                         -1,-1,-1, 1, 2, 2,-1,-1,
                                         -1,-1,-1,-1,-1,-1,-1,-1,
                                         -1,-1,-1,-1,-1,-1,-1,-1}};
   int global_cell_ids[4][25] = {{ 0, 1, 2, 3, 4, 5, 6,
                                   7, 8, 9,10,11,12,13,
                                  14,15,16,17,18,19,20},
                                 { 7, 8, 9,10,
                                  14,15,16,17,
                                  21,22,23,24,
                                  28,29,30,31,
                                  35,36,37,38},
                                 { 9,10,11,12,13,
                                  16,17,18,19,20,
                                  23,24,25,26,27,
                                  30,31,32,33,34,
                                  37,38,39,40,41},
                                 {28,29,30,31,32,33,34,
                                  35,36,37,38,39,40,41,
                                  42,43,44,45,46,47,48}};
   unsigned cell_is_on_rank[4][25] = {{-1,-1,-1,-1,-1,-1,-1,
                                       -1,-1,-1,-1,-1,-1,-1,
                                        1, 1, 1, 2, 2, 2, 2},
                                      { 0, 0, 0, 0,
                                       -1,-1,-1, 2,
                                       -1,-1,-1, 2,
                                       -1,-1,-1, 2,
                                        3, 3, 3, 3},
                                      { 0, 0, 0, 0, 0,
                                        1,-1,-1,-1,-1,
                                        1,-1,-1,-1,-1,
                                        1,-1,-1,-1,-1,
                                        3, 3, 3, 3, 3},
                                      { 1, 1, 1, 2, 2, 2, 2,
                                        3, 3, 3, 3, 3, 3, 3,
                                        3, 3, 3, 3, 3, 3, 3}};

   yac_init_subdomain_unstruct(&subdomain, grid, source_comm,
                               global_corner_ids[my_source_rank],
                               corner_is_on_rank[my_source_rank],
                               global_cell_ids[my_source_rank],
                               cell_is_on_rank[my_source_rank], NULL, NULL);

   struct domain domain;

   struct subdomain * blocks[1] = {&subdomain};

   yac_init_domain(&domain, source_comm, blocks, 1, "source_domain");

   // finish domain definition
   yac_enddef_domain(&domain);

   // init global search on both processes

   struct global_search * search;

   search = yac_global_search_new(&domain, "target_domain", global_comm, 1,0,
                                  "source_component", "target_component");

   // finish initialisation of global search

   yac_enddef_global_search(search);

   //---------------
   // cleanup
   //---------------

   yac_global_search_delete(search);
   yac_free_domain(&domain);
   yac_delete_grid(grid);
}

void target_main(struct communicator * global_comm,
                 struct communicator * target_comm) {

   struct grid * grid;
   unsigned num_cells[2] = {7,7};
   unsigned cyclic[2] = {0,0};
   double coordinates_x[8] = {0,1,2,3,4,5,6,7};
   double coordinates_y[8] = {0,1,2,3,4,5,6,7};

   for ( unsigned i = 0; i < 8; ++i ) {
      coordinates_x[i] *= rad;
      coordinates_y[i] *= rad;
   }

   grid = yac_reg2d_grid_new(coordinates_x, coordinates_y, num_cells, cyclic);

   struct subdomain subdomain;

   int global_corner_ids[64] =
      { 0, 1, 2, 3, 4, 5, 6, 7,
        8, 9,10,11,12,13,14,15,
       16,17,18,19,20,21,22,23,
       24,25,26,27,28,29,30,31,
       32,33,34,35,36,37,38,39,
       40,41,42,43,44,45,46,47,
       48,49,50,51,52,53,54,55,
       56,57,58,59,60,61,62,63};
   unsigned corner_is_on_rank[64] =
      {-1,-1,-1,-1,-1,-1,-1,-1,
       -1,-1,-1,-1,-1,-1,-1,-1,
       -1,-1,-1,-1,-1,-1,-1,-1,
       -1,-1,-1,-1,-1,-1,-1,-1,
       -1,-1,-1,-1,-1,-1,-1,-1,
       -1,-1,-1,-1,-1,-1,-1,-1,
       -1,-1,-1,-1,-1,-1,-1,-1,
       -1,-1,-1,-1,-1,-1,-1,-1};
   int global_cell_ids[49] =
      { 0, 1, 2, 3, 4, 5, 6,
        7, 8, 9,10,11,12,13,
       14,15,16,17,18,19,20,
       21,22,23,24,25,26,27,
       28,29,30,31,32,33,34,
       35,36,37,38,39,40,41,
       42,43,44,45,46,47,48};
   unsigned cell_is_on_rank[49] =
      {-1,-1,-1,-1,-1,-1,-1,
       -1,-1,-1,-1,-1,-1,-1,
       -1,-1,-1,-1,-1,-1,-1,
       -1,-1,-1,-1,-1,-1,-1,
       -1,-1,-1,-1,-1,-1,-1,
       -1,-1,-1,-1,-1,-1,-1,
       -1,-1,-1,-1,-1,-1,-1};

   yac_init_subdomain_unstruct(&subdomain, grid, target_comm,
                               global_corner_ids, corner_is_on_rank,
                               global_cell_ids, cell_is_on_rank, NULL, NULL);

   struct domain domain;
   struct subdomain * subdomains[1] = {&subdomain};

   yac_init_domain(&domain, target_comm, subdomains, 1, "target_domain");
   yac_enddef_domain(&domain);

   struct global_search * search;

   search = yac_global_search_new(&domain, "source_domain", global_comm, 0, 1,
                                  "target_component", "source_component");
   yac_enddef_global_search(search);

   yac_global_search_delete(search);
   yac_free_domain(&domain);
   yac_delete_grid(grid);
}
