/**
 * @file test_global_search_parallel3.c
 *
 * @copyright Copyright  (C)  2017 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#ifdef USE_MPI
#include <mpi.h>
#endif

#include "tests.h"

#include "grid.h"
#include "grid_reg2d.h"
#include "communicator.h"
#include "communicator_mpi.h"
#include "domain.h"
#include "global_search.h"

int main (void) {

#ifndef USE_MPI
  return EXIT_SKIP_TEST;
#else

  struct communicator * global_comm = yac_mpi_comm_new();

  if (yac_get_comm_size(global_comm) == 7) {

    unsigned global_rank = yac_get_comm_rank(global_comm);

    unsigned define_comp[2] = {global_rank < 5, global_rank > 1};
    unsigned order = global_rank & 1;

    struct communicator * domain_comms[2] =
      {yac_comm_split(global_comm, define_comp[0], 0),
       yac_comm_split(global_comm, define_comp[1], 0)};

    for (unsigned i = 1; i < 4; ++i) {

      unsigned domain_is_source[2] = {i & 1, i & 2};

      struct grid * grids[2];
      struct subdomain subdomains[2];
      struct subdomain * blocks[2][1] = {{&(subdomains[0])},{&(subdomains[1])}};
      char domain_names[2][16] = {"domain 1", "domain 2"};
      char component_names[2][16] = {"component 1", "component 2"};
      struct domain domains[2];

      for (unsigned j = 0; j < 2; ++j) {

        unsigned domain_idx = j ^ order;

        if (!define_comp[domain_idx]) continue;

        // corner and cell ids for a 7 x 7 grid

        // 56-----57-----58-----59-----60-----61-----62-----63
        //  |      |      |      |      |      |      |      |
        //  |  42  |  43  |  44  |  45  |  46  |  47  |  48  |
        //  |      |      |      |      |      |      |      |
        // 48-----49-----50-----51-----52-----53-----54-----55
        //  |      |      |      |      |      |      |      |
        //  |  35  |  36  |  37  |  38  |  39  |  40  |  41  |
        //  |      |      |      |      |      |      |      |
        // 40-----41-----42-----43-----44-----45-----46-----47
        //  |      |      |      |      |      |      |      |
        //  |  28  |  29  |  30  |  31  |  32  |  33  |  34  |
        //  |      |      |      |      |      |      |      |
        // 32-----33-----34-----35-----36-----37-----38-----39
        //  |      |      |      |      |      |      |      |
        //  |  21  |  22  |  23  |  24  |  25  |  26  |  27  |
        //  |      |      |      |      |      |      |      |
        // 24-----25-----26-----27-----28-----29-----30-----31
        //  |      |      |      |      |      |      |      |
        //  |  14  |  15  |  16  |  17  |  18  |  19  |  20  |
        //  |      |      |      |      |      |      |      |
        // 16-----17-----18-----19-----20-----21-----22-----23
        //  |      |      |      |      |      |      |      |
        //  |  07  |  08  |  09  |  10  |  11  |  12  |  13  |
        //  |      |      |      |      |      |      |      |
        // 08-----09-----10-----11-----12-----13-----14-----15
        //  |      |      |      |      |      |      |      |
        //  |  00  |  01  |  02  |  03  |  04  |  05  |  06  |
        //  |      |      |      |      |      |      |      |
        // 00-----01-----02-----03-----04-----05-----06-----07
        //
        // the domain is distributed among the processes as follows:
        // (index == process)
        //
        // 4---4---4---4---4---4---4---4
        // | 4 | 4 | 4 | 4 | 4 | 4 | 4 |
        // 4---4---4---4---4---4---4---4
        // | 4 | 4 | 4 | 4 | 4 | 4 | 4 |
        // 3---3---3---3---4---4---4---4
        // | 3 | 3 | 3 | 3 | 3 | 3 | 3 |
        // 2---2---2---2---3---3---3---3
        // | 2 | 2 | 2 | 2 | 2 | 2 | 2 |
        // 1---1---1---1---2---2---2---2
        // | 1 | 1 | 1 | 1 | 1 | 1 | 1 |
        // 0---0---0---0---1---1---1---1
        // | 0 | 0 | 0 | 0 | 0 | 0 | 0 |
        // 0---0---0---0---0---0---0---0
        // | 0 | 0 | 0 | 0 | 0 | 0 | 0 |
        // 0---0---0---0---0---0---0---0
        //
        // 0---0---1---2---3---4---4---4
        // | 0 | 0 | 1 | 2 | 3 | 4 | 4 |
        // 0---0---1---2---3---4---4---4
        // | 0 | 0 | 1 | 2 | 3 | 4 | 4 |
        // 0---0---1---2---3---4---4---4
        // | 0 | 0 | 1 | 2 | 3 | 4 | 4 |
        // 0---0---1---2---3---4---4---4
        // | 0 | 0 | 1 | 2 | 3 | 4 | 4 |
        // 0---0---0---1---2---3---4---4
        // | 0 | 0 | 1 | 2 | 3 | 4 | 4 |
        // 0---0---0---1---2---3---4---4
        // | 0 | 0 | 1 | 2 | 3 | 4 | 4 |
        // 0---0---0---1---2---3---4---4
        // | 0 | 0 | 1 | 2 | 3 | 4 | 4 |
        // 0---0---0---1---2---3---4---4

        unsigned domain_rank = yac_get_comm_rank(domain_comms[domain_idx]);

        // define domain
        unsigned num_cells[2][2] = {{7,3}, {3,7}};
        unsigned cyclic[2] = {0,0};
        double coordinates_x[2][5][8] =
          {{{0,1,2,3,4,5,6,7},
            {0,1,2,3,4,5,6,7},
            {0,1,2,3,4,5,6,7},
            {0,1,2,3,4,5,6,7},
            {0,1,2,3,4,5,6,7}},
           {{0,1,2,3},
            {1,2,3,4},
            {2,3,4,5},
            {3,4,5,6},
            {4,5,6,7}}};
        double coordinates_y[2][5][8] =
          {{{0,1,2,3},
            {1,2,3,4},
            {2,3,4,5},
            {3,4,5,6},
            {4,5,6,7}},
           {{0,1,2,3,4,5,6,7},
            {0,1,2,3,4,5,6,7},
            {0,1,2,3,4,5,6,7},
            {0,1,2,3,4,5,6,7},
            {0,1,2,3,4,5,6,7}}};

        for (unsigned i = 0; i < 2; ++i) {
          for (unsigned j = 0; j < 5; ++j) {
            for (unsigned k = 0; k < 8; ++k) {
              coordinates_x[i][j][k] *= rad;
              coordinates_y[i][j][k] *= rad;
            }
          }
        }

        grids[domain_idx] =
          yac_reg2d_grid_new(coordinates_x[domain_idx][domain_rank],
                             coordinates_y[domain_idx][domain_rank],
                             num_cells[domain_idx], cyclic);

        int global_corner_ids[2][5][32] =
          {{{0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,
             25,26,27,28,29,30,31},
            {8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,
             31,32,33,34,35,36,37,38,39},
            {16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,
             38,39,40,41,42,43,44,45,46,47},
            {24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,
             46,47,48,49,50,51,52,53,54,55},
            {32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49,50,51,52,53,
             54,55,56,57,58,59,60,61,62,63}},
           {{0,1,2,3,8,9,10,11,16,17,18,19,24,25,26,27,32,33,34,35,40,41,42,43,
             48,49,50,51,56,57,58,59},
            {1,2,3,4,9,10,11,12,17,18,19,20,25,26,27,28,33,34,35,36,41,42,43,44,
             49,50,51,52,57,58,59,60},
            {2,3,4,5,10,11,12,13,18,19,20,21,26,27,28,29,34,35,36,37,42,43,44,
             45,50,51,52,53,58,59,60,61},
            {3,4,5,6,11,12,13,14,19,20,21,22,27,28,29,30,35,36,37,38,43,44,45,
             46,51,52,53,54,59,60,61,62},
            {4,5,6,7,12,13,14,15,20,21,22,23,28,29,30,31,36,37,38,39,44,45,46,
             47,52,53,54,55,60,61,62,63}}};
        unsigned corner_is_on_rank[2][5][32] =
          {{{-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,1,1,1,
             1,1,1,1,1,2,2,2,2},
            {0,0,0,0,0,0,0,0,0,0,0,0,-1,-1,-1,-1,-1,-1,-1,-1,2,2,2,2,2,2,2,2,2,
             2,2,2},
            {2,2,2,2,2,2,2,2,2,2,2,2,-1,-1,-1,-1,-1,-1,-1,-1,3,3,3,3,3,3,3,3,3,
             3,3,3},
            {3,3,3,3,3,3,3,3,3,3,3,3,-1,-1,-1,-1,-1,-1,-1,-1,4,4,4,4,4,4,4,4,4,
             4,4,4},
            {4,4,4,4,4,4,4,4,4,4,4,4,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
             -1,-1,-1,-1,-1,-1}},
           {{-1,-1,-1,1,-1,-1,-1,1,-1,-1,-1,1,-1,-1,-1,1,-1,-1,1,2,-1,-1,1,2,-1,
             -1,1,2,-1,-1,1,2},
            {0,0,-1,2,0,0,-1,2,0,0,-1,2,0,0,-1,2,0,-1,2,3,0,-1,2,3,0,-1,2,3,0,
             -1,2,3},
            {0,1,-1,3,0,1,-1,3,0,1,-1,3,0,1,-1,3,1,-1,3,4,1,-1,3,4,1,-1,3,4,1,
             -1,3,4},
            {1,2,-1,4,1,2,-1,4,1,2,-1,4,1,2,-1,4,2,-1,4,4,2,-1,4,4,2,-1,4,4,2,
             -1,4,4},
            {2,3,-1,-1,2,3,-1,-1,2,3,-1,-1,2,3,-1,-1,3,-1,-1,-1,3,-1,-1,-1,3,-1,
             -1,-1,3,-1,-1,-1}}};
        int global_cell_ids[2][5][21] =
          {{{0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20},
            {7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27},
            {14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34},
            {21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,41},
            {28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48}},
           {{0,1,2,7,8,9,14,15,16,21,22,23,28,29,30,35,36,37,42,43,44},
            {1,2,3,8,9,10,15,16,17,22,23,24,29,30,31,36,37,38,43,44,45},
            {2,3,4,9,10,11,16,17,18,23,24,25,30,31,32,37,38,39,44,45,46},
            {3,4,5,10,11,12,17,18,19,24,25,26,31,32,33,38,39,40,45,46,47},
            {4,5,6,11,12,13,18,19,20,25,26,27,32,33,34,39,40,41,46,47,48}}};
        unsigned cell_is_on_rank[2][5][21] =
          {{{-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,1,1,1,1,1,1,1},
            {0,0,0,0,0,0,0,-1,-1,-1,-1,-1,-1,-1,2,2,2,2,2,2,2},
            {1,1,1,1,1,1,1,-1,-1,-1,-1,-1,-1,-1,3,3,3,3,3,3,3},
            {2,2,2,2,2,2,2,-1,-1,-1,-1,-1,-1,-1,4,4,4,4,4,4,4},
            {3,3,3,3,3,3,3,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}},
           {{-1,-1,1,-1,-1,1,-1,-1,1,-1,-1,1,-1,-1,1,-1,-1,1,-1,-1,1},
            {0,-1,2,0,-1,2,0,-1,2,0,-1,2,0,-1,2,0,-1,2,0,-1,2},
            {1,-1,3,1,-1,3,1,-1,3,1,-1,3,1,-1,3,1,-1,3,1,-1,3},
            {2,-1,4,2,-1,4,2,-1,4,2,-1,4,2,-1,4,2,-1,4,2,-1,4},
            {3,-1,-1,3,-1,-1,3,-1,-1,3,-1,-1,3,-1,-1,3,-1,-1,3,-1,-1}}};

        yac_init_subdomain_unstruct(&(subdomains[domain_idx]), grids[domain_idx],
                                    domain_comms[domain_idx],
                                    global_corner_ids[domain_idx][domain_rank],
                                    corner_is_on_rank[domain_idx][domain_rank],
                                    global_cell_ids[domain_idx][domain_rank],
                                    cell_is_on_rank[domain_idx][domain_rank],
                                    NULL, NULL);
        yac_init_domain(&(domains[domain_idx]), domain_comms[domain_idx],
                        blocks[domain_idx], 1, domain_names[domain_idx]);
      }

      for (unsigned j = 0; j < 2; ++j) {

        unsigned domain_idx = j ^ order;
        // finish domain definition
        if (define_comp[domain_idx]) yac_enddef_domain(&(domains[domain_idx]));
      }

      if (define_comp[0] && define_comp[1]) {

        struct global_search * searches[2];

        struct domain * domains_[2] = {&(domains[order]), &(domains[order^1])};
        unsigned domain_is_source_[2] = {domain_is_source[order],
                                         domain_is_source[order^1]};
        char const * component_names_[2] = {component_names[order],
                                            component_names[order^1]};

        yac_global_search_all_local_new(
          domains_, global_comm, domain_is_source_, component_names_, searches);

        yac_enddef_global_search(searches[0]);
        yac_enddef_global_search(searches[1]);

        yac_global_search_delete(searches[0]);
        yac_global_search_delete(searches[1]);
        yac_free_domain(&(domains[0]));
        yac_free_domain(&(domains[1]));
        yac_delete_grid(grids[0]);
        yac_delete_grid(grids[1]);

      } else {

        unsigned domain_idx = (define_comp[0])?0:1;

        struct global_search * search =
          yac_global_search_new(&(domains[domain_idx]),
                                domain_names[domain_idx^1], global_comm,
                                domain_is_source[domain_idx],
                                domain_is_source[domain_idx^1],
                                component_names[domain_idx],
                                component_names[domain_idx^1]);

        yac_enddef_global_search(search);

        yac_global_search_delete(search);
        yac_free_domain(&(domains[domain_idx]));
        yac_delete_grid(grids[domain_idx]);
      }
    }

    yac_free_communicator(domain_comms[0]);
    yac_free_communicator(domain_comms[1]);

  } else
    PUT_ERR("Wrong number of processes.");

  yac_free_communicator(global_comm);

  MPI_Finalize();

  return TEST_EXIT_CODE;
#endif
}