/**
 * @file test_grid_bnd_circle.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "tests.h"
#include "geometry.h"
#include "grid.h"
#include "grid_reg2d.h"
#include "grid_unstruct.h"

// tests whether all corners of grid are withing the bouding circle
void test_circle(struct grid * grid, struct bounding_circle circle);
// tests whether a point is within the bounding circle
unsigned point_in_circle(double lon, double lat, struct bounding_circle circle);

int main (void) {

   { // test simple case (2x2 regular grid)
      //---------------
      // setup
      //---------------
      double coord_x[3] = {0,1,2};
      double coord_y[3] = {0,1,2};
      unsigned num_cells[2] = {2,2};
      unsigned cyclic[2] = {0,0};
      struct grid * grid;

      for (unsigned i = 0; i < 3; ++i) coord_x[i] *= rad;
      for (unsigned i = 0; i < 3; ++i) coord_y[i] *= rad;

      grid = yac_reg2d_grid_new(coord_x, coord_y, num_cells, cyclic);

      //---------------
      // testing
      //---------------
      struct bounding_circle circle;

      yac_get_grid_bounding_circle(grid, &circle);

      test_circle(grid, circle);

      //---------------
      // cleanup
      //---------------
      yac_delete_grid(grid);
   }

   { // test pole example (2x2 unstructured grid that contains the north pole)
      //---------------
      // setup
      //---------------
      double coord_x[9] = {-45,0,45, 270,0,90, 225,180,135};
      double coord_y[9] = {83,85,83, 85,90,85, 83,85,83};
      unsigned num_corners = 9;
      unsigned num_cells = 4;
      unsigned num_corners_per_cell[4] = {4,4,4,4};
      unsigned cell_to_corner_dep[16] = {0,1,4,3, 1,2,5,4, 3,4,7,6, 4,5,8,7};
      struct dep_list cell_to_vertex;
      struct grid * grid;

      for (unsigned i = 0; i < 9; ++i) coord_x[i] *= rad;
      for (unsigned i = 0; i < 9; ++i) coord_y[i] *= rad;

      yac_init_dep_list(&cell_to_vertex);
      yac_set_dependencies(&cell_to_vertex, num_cells, num_corners_per_cell,
                           cell_to_corner_dep);

      grid = yac_unstruct_grid_new(coord_x, coord_y, num_corners, cell_to_vertex);

      //---------------
      // testing
      //---------------
      struct bounding_circle circle;

      yac_get_grid_bounding_circle(grid, &circle);

      test_circle(grid, circle);

      //---------------
      // cleanup
      //---------------
      yac_delete_grid(grid);
   }

   { // test ring around the globe on the equator (1x8 unstructured grid around the equator)
      //---------------
      // setup
      //---------------
      double coord_x[16] = {0,45,90,135,180,225,270,315,0,45,90,135,180,225,270,315};
      double coord_y[16] = {-5,-5,-5,-5,-5,-5,-5,-5,5,5,5,5,5,5,5,5};
      unsigned num_corners = 16;
      unsigned num_cells = 8;
      unsigned num_corners_per_cell[8] = {4,4,4,4,4,4,4,4};
      unsigned cell_to_corner_dep[4*8] = {0,1,9,8, 1,2,10,9, 2,3,11,10, 3,4,12,11,
                                          4,5,13,12, 5,6,14,13, 6,7,15,14, 7,0,8,15};
      struct dep_list cell_to_vertex;
      struct grid * grid;

      for (unsigned i = 0; i < 16; ++i) coord_x[i] *= rad;
      for (unsigned i = 0; i < 16; ++i) coord_y[i] *= rad;

      yac_init_dep_list(&cell_to_vertex);
      yac_set_dependencies(&cell_to_vertex, num_cells, num_corners_per_cell,
                           cell_to_corner_dep);

      grid = yac_unstruct_grid_new(coord_x, coord_y, num_corners, cell_to_vertex);

      //---------------
      // testing
      //---------------
      struct bounding_circle circle;

      yac_get_grid_bounding_circle(grid, &circle);

      test_circle(grid, circle);

      //---------------
      // cleanup
      //---------------
      yac_delete_grid(grid);
   }

   return TEST_EXIT_CODE;
}

void test_circle(struct grid * grid, struct bounding_circle circle) {

   unsigned i, num_corners;

   num_corners = yac_get_num_grid_corners(grid);

   for (i = 0; i < num_corners; ++i) {

      if (!point_in_circle(yac_get_corner_x_coord(grid, i),
                           yac_get_corner_y_coord(grid, i),
                           circle)) {

         PUT_ERR("point is not in bounding circle\n");
      }
   }
}

unsigned point_in_circle(double lon, double lat, struct bounding_circle circle) {

   static double const tol = 1.0e-08;
   double angle;
   double xyz[3];

   if (circle.inc_angle + tol >= M_PI) return 1 == 1;

   LLtoXYZ(lon, lat, xyz);

   angle = get_vector_angle(xyz, circle.base_vector);

   return angle <= circle.inc_angle + tol;
}
