#define VERBOSE
/**
 * @file test_interpolation_method.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */
#include <stdlib.h>
#include <stdio.h>

#include "tests.h"

#include "grid.h"
#include "grid_reg2d.h"
#include "communicator.h"
#include "communicator_local.h"
#include "domain.h"
#include "global_search.h"
#include "interpolation_method_fixed.h"
#include "interpolation_method_avg.h"
#include "interpolation_method_nnn.h"
#include "fields.h"
#include "utils.h"

void check_data_request(struct interpolation_method_data_request * request,
                        struct interpolation_method_data_request* ref_request,
                        unsigned num_points_per_subdomain);

void
delete_interpolation_method_data_request_stack(
   struct interpolation_method_data_request * request,
   unsigned num_points_per_subdomain);

int main (void) {

   unsigned i, j;

   unsigned component_id = yac_pointer_to_unique_id(NULL);

   { // simple fixed value "interpolation" on a single process

      //--------------------
      // set up global search
      //--------------------

         // one directional global search with two domains on two processes with one subdomain each

         // comm worlds on both processes
         struct communicator * comm_p1, * comm_p2;

         comm_p1 = yac_local_comm_new();
         comm_p2 = yac_local_comm_new();

         // named comms for the domains on both processes
         struct communicator * comm_named_p1, * comm_named_p2;

         comm_named_p1 = yac_named_local_comm_new(comm_p1, "P1");
         comm_named_p2 = yac_named_local_comm_new(comm_p2, "P2");

         // define a domain on P1

         struct grid * grid_p1;
         unsigned num_cells[2] = {2,2};
         unsigned cyclic[2] = {0,0};
         double coordinates_x[2][3] = {{-1,0,1}, {-1.5,-0.5,0.5}};
         double coordinates_y[2][3] = {{-1,0,1}, {-1.5,-0.5,0.5}};

         for ( i = 0; i < 2; ++i ) {
            for ( j = 0; j < 3; ++j ) {
               coordinates_x[i][j] *= rad;
               coordinates_y[i][j] *= rad;
            }
         }

         grid_p1 = yac_reg2d_grid_new(coordinates_x[0], coordinates_y[0],
                                      num_cells, cyclic);

         struct subdomain subdomain_p1;

         int global_corner_ids_p1[9] = {0,1,2,3,4,5,6,7,8};
         unsigned corner_is_on_rank_p1[9] = {-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_p1[4] = {0,1,2,3};
         unsigned cell_is_on_rank_p1[4] = {-1,-1,-1,-1};
         int global_edge_ids_p1[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned edge_is_on_rank_p1[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_p1, grid_p1, comm_named_p1,
                                     global_corner_ids_p1, corner_is_on_rank_p1,
                                     global_cell_ids_p1, cell_is_on_rank_p1,
                                     global_edge_ids_p1, edge_is_on_rank_p1);

         struct domain l_domain_p1;

         struct subdomain * blocks_p1[1] = {&subdomain_p1};

         yac_init_domain(&l_domain_p1, comm_named_p1, blocks_p1, 1, "domain_p1");

         // define a domain on P2, which matches the one on P1

         struct grid * grid_p2;

         grid_p2 = yac_reg2d_grid_new(coordinates_x[1], coordinates_y[1],
                                      num_cells, cyclic);

         struct subdomain subdomain_p2;

         int global_corner_ids_p2[9] = {0,1,2,3,4,5,6,7,8};
         unsigned corner_is_on_rank_p2[9] = {-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_p2[4] = {0,1,2,3};
         unsigned cell_is_on_rank_p2[4] = {-1,-1,-1,-1};
         int global_edge_ids_p2[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned edge_is_on_rank_p2[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_p2, grid_p2, comm_named_p2,
                                     global_corner_ids_p2, corner_is_on_rank_p2,
                                     global_cell_ids_p2, cell_is_on_rank_p2,
                                     global_edge_ids_p2, edge_is_on_rank_p2);

         struct domain l_domain_p2;

         struct subdomain * blocks_p2[1] = {&subdomain_p2};

         yac_init_domain(&l_domain_p2, comm_named_p2, blocks_p2, 1, "domain_p2");

         // finish domain definition
         yac_enddef_domain(&l_domain_p1);
         yac_enddef_domain(&l_domain_p2);

         // init global search on both processes

         struct global_search * search[2];

         search[0] = yac_global_search_new(&l_domain_p1, "domain_p2", comm_p1,
                                           1,0, "component_p1", "component_p2");
         search[1] = yac_global_search_new(&l_domain_p2, "domain_p1", comm_p2,
                                           0,1, "component_p2", "component_p1");

         // finish initialisation of global search

         yac_enddef_global_search(search[0]);
         yac_enddef_global_search(search[1]);

      //----------------------
      // set up coupling field
      //----------------------

      struct coupling_field * field;

      field = yac_coupling_field_new("field name", component_id, &l_domain_p1, NULL, NULL, 0);

      //----------------------------------------
      // test generation of interpolation method
      //----------------------------------------

      struct interpolation_method * method_fixed;
      double fixed_value = 1337;

      method_fixed = yac_new_interpolation_method_fixed(fixed_value, search[0]);

      //----------
      // do search
      //----------

      struct interpolation_method * interpolation_method_stack[] = {method_fixed,
                                                                    NULL};
      struct interpolation_plan_element * interpolation_plan_element_stack[] = {NULL, NULL};
      int global_ids[] = {0,1,2,3,4,5,6,7,8};
      unsigned num_points = sizeof(global_ids) / sizeof(global_ids[0]);
      unsigned num_interpolated_points;

      yac_do_search_interpolation_method(interpolation_method_stack,
         interpolation_plan_element_stack, global_ids, NULL, NULL, CORNER, num_points,
         &num_interpolated_points, 0);
      yac_enddef_interpolation_method(interpolation_method_stack);

      //---------------------
      // do the interpolation
      //---------------------

      double send_buffer[9];

      yac_execute_interpolation_plan_element(
        interpolation_plan_element_stack, NULL, send_buffer);

      struct interpolation_method_data_request data_request_;
      struct interpolation_method_data_request * data_request = &data_request_;
      yac_get_source_exchange_data_interpolation_method(
        interpolation_method_stack, &data_request, 1);

      //---------------------------------
      // testing results of interpolation
      //---------------------------------

      for (i = 0; i < 9; ++i)
         if (send_buffer[i] != fixed_value)
            PUT_ERR("wrong result buffer a\n");

      //---------------
      // cleanup
      //---------------

      delete_interpolation_method_data_request_stack(data_request, 1);
      yac_delete_interpolation_method(method_fixed);

      for (i = 0; i < sizeof(interpolation_plan_element_stack) /
                      sizeof(interpolation_plan_element_stack[0]); ++i) {
         yac_free_interpolation_plan_element(interpolation_plan_element_stack[i]);
         free(interpolation_plan_element_stack[i]);
      }

      yac_coupling_field_delete(field);

      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);

      yac_free_domain(&l_domain_p1);
      yac_free_domain(&l_domain_p2);

      yac_delete_grid(grid_p1);
      yac_delete_grid(grid_p2);

      yac_free_communicator(comm_p1);
      yac_free_communicator(comm_p2);

      yac_free_communicator(comm_named_p1);
      yac_free_communicator(comm_named_p2);

      yac_reset_interpolation_method();
   }

   { // linear point interpolation with one source process that has two domains and
     // a target process

      //--------
      // setup
      //--------

      // communicators

      struct communicator * comm_src, * comm_tgt,
                          * comm_named_src, * comm_named_tgt;

      comm_src = yac_local_comm_new();
      comm_tgt = yac_local_comm_new();
      comm_named_src = yac_named_local_comm_new(comm_src, "source");
      comm_named_tgt = yac_named_local_comm_new(comm_tgt, "target");

      // grid

      struct grid * grid_src[2], * grid_tgt;
      unsigned num_cells_src[2][2] = {{2,4},{2,4}};
      unsigned num_cells_tgt[2] = {2,3};
      unsigned cyclic[2] = {0,0};
      double coordinates_x_src[2][3] = {{0,1,2},{3,4,5}};
      double coordinates_y_src[2][5] = {{0,1,2,3,4},{0,1,2,3,4}};
      double coordinates_x_tgt[3] = {0.5,1.5,2.5};
      double coordinates_y_tgt[4] = {0.5,1.5,2.5,3.5};

      for ( i = 0; i < 2; ++i )
         for ( j = 0; j < 3; ++j )
            coordinates_x_src[i][j] *= rad;

      for ( i = 0; i < 2; ++i )
         for ( j = 0; j < 5; ++j )
            coordinates_y_src[i][j] *= rad;

      for ( i = 0; i < 3; ++i )
         coordinates_x_tgt[i] *= rad;

      for ( i = 0; i < 4; ++i )
         coordinates_y_tgt[i] *= rad;

      grid_src[0] = yac_reg2d_grid_new(coordinates_x_src[0], coordinates_y_src[0],
                                       num_cells_src[0], cyclic);
      grid_src[1] = yac_reg2d_grid_new(coordinates_x_src[1], coordinates_y_src[1],
                                       num_cells_src[1], cyclic);
      grid_tgt = yac_reg2d_grid_new(coordinates_x_tgt, coordinates_y_tgt,
                                    num_cells_tgt, cyclic);

      // subdomain

      struct subdomain subdomain_src[2], subdomain_tgt;
      int global_corner_ids_src[2][15] = {{0,1,2,3,4,5,6,7,8,9,10,11,12,13,14},
                                          {15,16,17,18,19,20,21,22,23,24,25,26,27,28,29}};
      unsigned corner_is_on_rank_src[2][15] = {{-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
                                               {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}};
      int global_cell_ids_src[2][8] = {{0,1,2,3,4,5,6,7},{8,9,10,11,12,13,14,15}};
      unsigned cell_is_on_rank_src[2][8] = {{-1,-1,-1,-1,-1,-1,-1,-1},{-1,-1,-1,-1,-1,-1,-1,-1}};

      int global_corner_ids_tgt[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
      unsigned corner_is_on_rank_tgt[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
      int global_cell_ids_tgt[6] = {0,1,2,3,4,5};
      unsigned cell_is_on_rank_tgt[6] = {-1,-1,-1,-1,-1,-1};

      yac_init_subdomain_unstruct(subdomain_src+0, grid_src[0], comm_named_src,
                                  global_corner_ids_src[0], corner_is_on_rank_src[0],
                                  global_cell_ids_src[0], cell_is_on_rank_src[0],
                                  NULL, NULL);
      yac_init_subdomain_unstruct(subdomain_src+1, grid_src[1], comm_named_src,
                                  global_corner_ids_src[1], corner_is_on_rank_src[1],
                                  global_cell_ids_src[1], cell_is_on_rank_src[1],
                                  NULL, NULL);
      yac_init_subdomain_unstruct(&subdomain_tgt, grid_tgt, comm_named_tgt,
                                  global_corner_ids_tgt, corner_is_on_rank_tgt,
                                  global_cell_ids_tgt, cell_is_on_rank_tgt,
                                  NULL, NULL);

      // domain

      struct domain domain_src, domain_tgt;
      struct subdomain * block_src[2] = {subdomain_src+0,subdomain_src+1};
      struct subdomain * block_tgt[1] = {&subdomain_tgt};

      yac_init_domain(&domain_src, comm_named_src, block_src, 2, "source domain");
      yac_init_domain(&domain_tgt, comm_named_tgt, block_tgt, 1, "target domain");

      yac_enddef_domain(&domain_src);
      yac_enddef_domain(&domain_tgt);

      // global search

      struct global_search * search_src, * search_tgt;

      search_src = yac_global_search_new(&domain_src, "target domain", comm_src,
                                         1, 0, "source component", "target component");
      search_tgt = yac_global_search_new(&domain_tgt, "source domain", comm_tgt,
                                         0, 1, "target component", "source component");

      yac_enddef_global_search(search_src);
      yac_enddef_global_search(search_tgt);

      // points

      struct points points_src[2];

      yac_init_points(points_src+0, grid_src[0], CORNER,
                      coordinates_x_src[0], coordinates_y_src[0]);
      yac_init_points(points_src+1, grid_src[1], CORNER,
                      coordinates_x_src[1], coordinates_y_src[1]);

      // point sets

      struct points * points_per_subdomain[2][1] = {{points_src+0},{points_src+1}};
      struct points ** points_set_src[2] = {points_per_subdomain[0],
                                            points_per_subdomain[1]};

      //-----------------------------------------
      // test generation of interpolation method
      //-----------------------------------------

      struct interpolation_method * method_fixed, * method_avg;
      double fixed_value = 1337;

      method_fixed = yac_new_interpolation_method_fixed(fixed_value, search_src);
      method_avg =
        yac_new_interpolation_method_avg(search_src, &domain_src, "avg stack",
                                     points_set_src, NULL, 1, 0, 0, 0);

      //----------------
      // do the search
      //----------------

      struct interpolation_method * method_stack[] = {method_avg, method_fixed, NULL};
      struct interpolation_plan_element * plan_stack[] = {NULL,NULL,NULL};

      int global_ids[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
      unsigned num_points = 12;
      unsigned num_interpolated_points[3] = {0,0,0};
      double coordinates_x[12] = {1.5,2.5,3.5,1.5,2.5,3.5,1.5,2.5,3.5,1.5,2.5,3.5};
      double coordinates_y[12] = {0.5,0.5,0.5,1.5,1.5,1.5,2.5,2.5,2.5,3.5,3.5,3.5};

      for ( i = 0; i < 12; ++i ) {
         coordinates_x[i] *= rad;
         coordinates_y[i] *= rad;
      }

      yac_do_search_interpolation_method(method_stack, plan_stack, global_ids,
         coordinates_x, coordinates_y, CORNER, num_points, num_interpolated_points, 1);
      yac_enddef_interpolation_method(method_stack);

      //----------------------
      // do the interpolation
      //----------------------

      double in_data_raw[30] = {0,0,0,1,1,1,2,2,2,3,3,3,4,4,4,
                                5,5,5,6,6,6,7,7,7,8,8,8,9,9,9};
      double * subdomain_in_data[2][1] = {{in_data_raw+0}, {in_data_raw+15}};
      double ** in_data[3] = {NULL, subdomain_in_data[0], subdomain_in_data[1]};
      double out_data[12];

      yac_execute_interpolation_plan_element(plan_stack, in_data, out_data);

      //---------------------------------
      // testing results of interpolation
      //---------------------------------

      struct interpolation_method_data_request data_request_[2];
      struct interpolation_method_data_request * data_request = data_request_;

      yac_get_source_exchange_data_interpolation_method(
         method_stack, &data_request, 1);

      double ref_out_data[12] = {0.5,5.5,1.5,6.5,2.5,7.5,3.5,8.5,
                                 fixed_value,fixed_value,
                                 fixed_value,fixed_value};

      int ref_global_ids[12] = {0,2,3,5,6,8,9,11,1,4,10,7};

      for (i = 0; i < 12; ++i)
         if (ref_global_ids[i] != global_ids[i])
            PUT_ERR("wrong global id order");

      for (i = 0; i < 12; ++i)
         if (out_data[i] != ref_out_data[i])
            PUT_ERR("wrong result buffer\n");

      //---------------
      // cleanup
      //---------------

      for (i = 0; i < sizeof(method_stack) /
                      sizeof(method_stack[0]); ++i)
         if (method_stack[i] != NULL)
            yac_delete_interpolation_method(method_stack[i]);
      for (i = 0; i < sizeof(plan_stack) /
                      sizeof(plan_stack[0]); ++i)
         if (plan_stack[i] != NULL) {
            yac_free_interpolation_plan_element(plan_stack[i]);
            free(plan_stack[i]);
         }
      delete_interpolation_method_data_request_stack(data_request, 1);

      yac_free_points(points_src+0);
      yac_free_points(points_src+1);

      yac_global_search_delete(search_src);
      yac_global_search_delete(search_tgt);

      yac_free_domain(&domain_src);
      yac_free_domain(&domain_tgt);

      yac_delete_grid(grid_src[0]);
      yac_delete_grid(grid_src[1]);
      yac_delete_grid(grid_tgt);

      yac_free_communicator(comm_src);
      yac_free_communicator(comm_tgt);

      yac_free_communicator(comm_named_src);
      yac_free_communicator(comm_named_tgt);

      yac_reset_interpolation_method();
   }

   { // fixed value point interpolation with two source processes and a single target process

      // the global domain is a 4x2 grid:
      //       10--18--11--19--12--20--13--21--14
      //       |       |       |       |       |
      //       10  04  12  05  14  06  16  07  17
      //       |       |       |       |       |
      //       05--09--06--11--07--13--08--15--09
      //       |       |       |       |       |
      //       01  00  03  01  05  02  07  03  08
      //       |       |       |       |       |
      //       00--00--01--02--02--04--03--06--04
      //
      // domain_a is distributed among two processes with one subdomains each
      // the partitioning is as follows:
      //
      //       0--0--0--0--0--1--1--1--1
      //       |     |     |     |     |
      //       0  0  0  0  0  1  1  1  1
      //       |     |     |     |     |
      //       0--0--0--0--1--1--1--1--1
      //       |     |     |     |     |
      //       0  0  0  0  1  1  1  1  1
      //       |     |     |     |     |
      //       0--0--0--0--1--1--1--1--1
      //
      //---------------
      // setup
      //---------------

      // comm worlds on both processes
      struct communicator * comm_p1_src, * comm_p2_src, * comm_p1_tgt;

      comm_p1_src = yac_local_comm_new();
      comm_p2_src = yac_local_comm_new();
      comm_p1_tgt = yac_local_comm_new();

      // named comms for the domains on both processes
      struct communicator * comm_named_p1_src, * comm_named_p2_src,
                          * comm_named_p1_tgt;

      comm_named_p1_src = yac_named_local_comm_new(comm_p1_src, "src");
      comm_named_p2_src = yac_named_local_comm_new(comm_p2_src, "src");
      comm_named_p1_tgt = yac_named_local_comm_new(comm_p1_tgt, "tgt");

         // define domains on source

         struct grid * grid_p1_src, * grid_p2_src;
         unsigned num_cells_src[2] = {3,2};
         unsigned cyclic_src[2] = {0,0};
         double coordinates_x_src[2][4] = {{0,1,2,3}, {1,2,3,4}};
         double coordinates_y_src[2][3] = {{0,1,2},   {0,1,2}};

         for ( i = 0; i < 2; ++i )
            for ( j = 0; j < 4; ++j )
               coordinates_x_src[i][j] *= rad;

         for ( i = 0; i < 2; ++i )
            for ( j = 0; j < 3; ++j )
               coordinates_y_src[i][j] *= rad;

         grid_p1_src = yac_reg2d_grid_new(coordinates_x_src[0], coordinates_y_src[0],
                                          num_cells_src, cyclic_src);
         grid_p2_src = yac_reg2d_grid_new(coordinates_x_src[1], coordinates_y_src[1],
                                          num_cells_src, cyclic_src);

         struct subdomain subdomain_p1_src, subdomain_p2_src;

         int global_corner_ids_src[2][12] = {{0,1,2,3, 5,6,7,8, 10,11,12,13},
                                             {1,2,3,4, 6,7,8,9, 11,12,13,14}};
         unsigned corner_is_on_rank_src[2][12] = {{-1,-1,1,1, -1,-1,1,1, -1,-1,-1,1},
                                                  {0,-1,-1,-1, 0,-1,-1,-1, 0,0,-1,-1}};
         int global_cell_ids_src[2][6] = {{0,1,2, 4,5,6},
                                          {1,2,3, 5,6,7}};
         unsigned cell_is_on_rank_src[2][6] = {{-1,-1,1, -1,-1,1},
                                               {0,-1,-1, 0,-1,-1}};

         yac_init_subdomain_unstruct(&subdomain_p1_src, grid_p1_src, comm_named_p1_src,
                                     global_corner_ids_src[0], corner_is_on_rank_src[0],
                                     global_cell_ids_src[0], cell_is_on_rank_src[0],
                                     NULL, NULL);

         yac_init_subdomain_unstruct(&subdomain_p2_src, grid_p2_src, comm_named_p2_src,
                                     global_corner_ids_src[1], corner_is_on_rank_src[1],
                                     global_cell_ids_src[1], cell_is_on_rank_src[1],
                                     NULL, NULL);

         struct domain l_domain_p1_src, l_domain_p2_src;

         struct subdomain * blocks_src[2][1] = {{&subdomain_p1_src},
                                                {&subdomain_p2_src}};

         yac_init_domain(&l_domain_p1_src, comm_named_p1_src, blocks_src[0], 1, "domain_src");
         yac_init_domain(&l_domain_p2_src, comm_named_p2_src, blocks_src[1], 1, "domain_src");

         // define a domain on the target

         struct grid * grid_tgt;
         unsigned num_cells_tgt[2] = {4,2};
         unsigned cyclic_tgt[2] = {0,0};
         double coordinates_x_tgt[5] = {0.5,1.5,2.5,3.5,4.5};
         double coordinates_y_tgt[3] = {0.5,1.5,2.5};

         for ( i = 0; i < 5; ++i )
            coordinates_x_tgt[i] *= rad;
         for ( i = 0; i < 3; ++i )
            coordinates_y_tgt[i] *= rad;

         grid_tgt = yac_reg2d_grid_new(coordinates_x_tgt, coordinates_y_tgt,
                                       num_cells_tgt, cyclic_tgt);

         struct subdomain subdomain_tgt;

         int global_corner_ids_tgt[15] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14};
         unsigned corner_is_on_rank_tgt[15] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_tgt[8] = {0,1,2,3,4,5,6,7};
         unsigned cell_is_on_rank_tgt[8] = {-1,-1,-1,-1,-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_tgt, grid_tgt, comm_named_p1_tgt,
                                     global_corner_ids_tgt, corner_is_on_rank_tgt,
                                     global_cell_ids_tgt, cell_is_on_rank_tgt,
                                     NULL, NULL);

         struct domain l_domain_tgt;

         struct subdomain * blocks_tgt[1] = {&subdomain_tgt};

         yac_init_domain(&l_domain_tgt, comm_named_p1_tgt, blocks_tgt, 1, "domain_tgt");

         // finish domain definition
         yac_enddef_domain(&l_domain_p1_src);
         yac_enddef_domain(&l_domain_p2_src);
         yac_enddef_domain(&l_domain_tgt);

         //-------------------------------------
         // init global search on both processes
         //-------------------------------------

         struct global_search * search[3];

         search[0] = yac_global_search_new(&l_domain_p1_src, "domain_tgt", comm_p1_src,
                                           1,0, "component_src", "component_tgt");
         search[1] = yac_global_search_new(&l_domain_p2_src, "domain_tgt", comm_p2_src,
                                           1,0, "component_src", "component_tgt");
         search[2] = yac_global_search_new(&l_domain_tgt, "domain_src", comm_p1_tgt,
                                           0,1, "component_tgt", "component_src");

         //---------------------------------------
         // finish initialisation of global search
         //---------------------------------------

         yac_enddef_global_search(search[0]);
         yac_enddef_global_search(search[1]);
         yac_enddef_global_search(search[2]);

      //---------------------------------
      // initialise points on source side
      //---------------------------------

      struct points points_src[2];

      yac_init_points(points_src+0, grid_p1_src, CORNER,
                      coordinates_x_src[0], coordinates_y_src[0]);
      yac_init_points(points_src+1, grid_p2_src, CORNER,
                      coordinates_x_src[1], coordinates_y_src[1]);

      //-------------------------------------
      // set up coupling field on source side
      //-------------------------------------

      struct coupling_field * field_src[2];
      struct points * points_per_subdomain[2][1] = {{points_src+0},{points_src+1}};
      struct points ** point_sets[2][1] = {{points_per_subdomain[0]},
                                           {points_per_subdomain[1]}};

      field_src[0] = yac_coupling_field_new("field name", component_id, &l_domain_p1_src, point_sets[0], NULL, 1);
      field_src[1] = yac_coupling_field_new("field name", component_id, &l_domain_p2_src, point_sets[1], NULL, 1);

      //----------------------------------------
      // test generation of interpolation method
      //----------------------------------------

      struct interpolation_method * method_fixed[2];
      double fixed_value = 1337;

      method_fixed[0] = yac_new_interpolation_method_fixed(fixed_value, search[0]);
      method_fixed[1] = yac_new_interpolation_method_fixed(2*fixed_value, search[1]);
      //----------
      // do search
      //----------

      struct interpolation_method * method_stack[2][2] = {{method_fixed[0],
                                                           NULL},
                                                          {method_fixed[1],
                                                           NULL}};
      struct interpolation_plan_element * interpolation_plan_element_stack[2][2] = {{NULL, NULL},
                                                                    {NULL, NULL}};
      int global_ids[2][8] = {{0,1,2,3,4,5,6,7}, {0,1,2,3,4,5,6,7}};
      unsigned num_points = 8;
      unsigned num_interpolated_points[2];
      double coordinates_x_tgt_cells[2][8] = {{1,2,1,2,1,2,1,2},{1,2,1,2,1,2,1,2}};
      double coordinates_y_tgt_cells[2][8] = {{1,1,2,2,3,3,4,4},{1,1,2,2,3,3,4,4}};

      for ( i = 0; i < 2; ++i ) {
         for ( j = 0; j < 8; ++j ) {
            coordinates_x_tgt_cells[i][j] *= rad;
            coordinates_y_tgt_cells[i][j] *= rad;
         }
      }

      yac_do_search_interpolation_method(method_stack[0], interpolation_plan_element_stack[0],
         global_ids[0], coordinates_x_tgt_cells[0], coordinates_y_tgt_cells[0], CORNER,
         num_points, num_interpolated_points+0, 1);
      yac_do_search_interpolation_method(method_stack[1], interpolation_plan_element_stack[1],
         global_ids[1], coordinates_x_tgt_cells[1], coordinates_y_tgt_cells[1], CORNER,
         num_points, num_interpolated_points+1, 1);
      yac_enddef_interpolation_method(method_stack[0]);
      yac_enddef_interpolation_method(method_stack[1]);

      //---------------------
      // do the interpolation
      //---------------------

      double send_buffer[2][8];

      yac_execute_interpolation_plan_element(interpolation_plan_element_stack[0], NULL, send_buffer[0]);
      yac_execute_interpolation_plan_element(interpolation_plan_element_stack[1], NULL, send_buffer[1]);

      //---------------------------------
      // testing results of interpolation
      //---------------------------------

      struct interpolation_method_data_request data_request_[2];
      struct interpolation_method_data_request * data_request[2][1] = {{data_request_+0},
                                                                       {data_request_+1}};

      yac_get_source_exchange_data_interpolation_method(method_stack[0], data_request[0], 1);
      yac_get_source_exchange_data_interpolation_method(method_stack[1], data_request[1], 1);

      unsigned ref_num_interpolated_points[2] = {0,8};
      double ref_out_data[8] = {2*fixed_value,2*fixed_value,
                                2*fixed_value,2*fixed_value,
                                2*fixed_value,2*fixed_value,
                                2*fixed_value,2*fixed_value};
      int ref_global_ids[8] = {0,1,2,3,4,5,6,7};

      for (i = 0; i < 2; ++i)
         if (ref_num_interpolated_points[i] != num_interpolated_points[i])
            PUT_ERR("wrong number of interpolated points\n");

      for (i = 0; i < 8; ++i)
         if (ref_global_ids[i] != global_ids[1][i])
            PUT_ERR("wrong global id order\n");

      for (i = 0; i < 8; ++i)
         if (send_buffer[1][i] != ref_out_data[i])
            PUT_ERR("wrong result buffer c\n");

      //---------------
      // cleanup
      //---------------

      for (j = 0; j < 2; ++j) {
         for (i = 0; i < sizeof(method_stack[j]) /
                         sizeof(method_stack[j][0]); ++i)
            if (method_stack[j][i] != NULL)
               yac_delete_interpolation_method(method_stack[j][i]);
         for (i = 0; i < sizeof(interpolation_plan_element_stack[j]) /
                         sizeof(interpolation_plan_element_stack[j][0]); ++i) {
            yac_free_interpolation_plan_element(interpolation_plan_element_stack[j][i]);
            free(interpolation_plan_element_stack[j][i]);
         }
         delete_interpolation_method_data_request_stack(data_request[j][0], 1);
      }

      yac_coupling_field_delete(field_src[0]);
      yac_coupling_field_delete(field_src[1]);

      yac_free_points(points_src+0);
      yac_free_points(points_src+1);

      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);
      yac_global_search_delete(search[2]);

      yac_free_domain(&l_domain_p1_src);
      yac_free_domain(&l_domain_p2_src);
      yac_free_domain(&l_domain_tgt);

      yac_delete_grid(grid_p1_src);
      yac_delete_grid(grid_p2_src);
      yac_delete_grid(grid_tgt);

      yac_free_communicator(comm_p1_src);
      yac_free_communicator(comm_p2_src);
      yac_free_communicator(comm_p1_tgt);

      yac_free_communicator(comm_named_p1_src);
      yac_free_communicator(comm_named_p2_src);
      yac_free_communicator(comm_named_p1_tgt);

      yac_reset_interpolation_method();
   }

   { // linear point interpolation with two source processes and a single target process

      // the global domain is a 4x2 grid:
      //       08--14--09--15--10--16--11
      //       |       |       |       |
      //       08  03  10  04  12  05  13
      //       |       |       |       |
      //       04--07--05--09--06--11--07
      //       |       |       |       |
      //       01  00  03  01  05  02  06
      //       |       |       |       |
      //       00--00--01--02--02--04--03
      //
      // domain_a is distributed among two processes with one subdomain each
      // the partitioning is as follows:
      //
      //       0--0--0--0--0--1--1
      //       |     |     |     |
      //       0  0  0  0  1  1  1
      //       |     |     |     |
      //       0--0--0--1--1--1--1
      //       |     |     |     |
      //       0  0  0  1  1  1  1
      //       |     |     |     |
      //       0--0--0--1--1--1--1
      //
      //---------------
      // setup
      //---------------

      // comm worlds on both processes
      struct communicator * comm_p1_src, * comm_p2_src, * comm_p1_tgt;

      comm_p1_src = yac_local_comm_new();
      comm_p2_src = yac_local_comm_new();
      comm_p1_tgt = yac_local_comm_new();

      // named comms for the domains on both processes
      struct communicator * comm_named_p1_src, * comm_named_p2_src, * comm_named_p1_tgt;

      comm_named_p1_src = yac_named_local_comm_new(comm_p1_src, "src");
      comm_named_p2_src = yac_named_local_comm_new(comm_p2_src, "src");
      comm_named_p1_tgt = yac_named_local_comm_new(comm_p1_tgt, "tgt");

         // define domains on source

         struct grid * grid_p1_src, * grid_p2_src;
         unsigned num_cells_src[2] = {2,2};
         unsigned cyclic_src[2] = {0,0};
         double coordinates_x_src[2][3] = {{0,1,2}, {1,2,3}};
         double coordinates_y_src[2][3] = {{0,1,2}, {0,1,2}};

         for ( i = 0; i < 2; ++i ) {
            for ( j = 0; j < 3; ++j ) {
               coordinates_x_src[i][j] *= rad;
               coordinates_y_src[i][j] *= rad;
            }
         }

         grid_p1_src = yac_reg2d_grid_new(coordinates_x_src[0], coordinates_y_src[0],
                                          num_cells_src, cyclic_src);
         grid_p2_src = yac_reg2d_grid_new(coordinates_x_src[1], coordinates_y_src[1],
                                          num_cells_src, cyclic_src);

         struct subdomain subdomain_p1_src, subdomain_p2_src;

         int global_corner_ids_src[2][9] = {{0,1,2, 4,5,6, 8,9,10},
                                            {1,2,3, 5,6,7, 9,10,11}};
         unsigned corner_is_on_rank_src[2][9] = {{-1,-1,1, -1,-1,1, -1,-1,-1,},
                                                 {0,-1,-1, 0,-1,-1, 0,0,-1}};
         int global_cell_ids_src[2][4] = {{0,1,3,4},
                                          {1,2,4,5}};
         unsigned cell_is_on_rank_src[2][4] = {{-1,1,-1,-1},
                                               {-1,-1,0,-1}};

         yac_init_subdomain_unstruct(&subdomain_p1_src, grid_p1_src, comm_named_p1_src,
                                     global_corner_ids_src[0], corner_is_on_rank_src[0],
                                     global_cell_ids_src[0], cell_is_on_rank_src[0],
                                     NULL, NULL);

         yac_init_subdomain_unstruct(&subdomain_p2_src, grid_p2_src, comm_named_p2_src,
                                     global_corner_ids_src[1], corner_is_on_rank_src[1],
                                     global_cell_ids_src[1], cell_is_on_rank_src[1],
                                     NULL, NULL);

         struct domain l_domain_p1_src, l_domain_p2_src;

         struct subdomain * blocks_src[2][1] = {{&subdomain_p1_src},
                                                {&subdomain_p2_src}};

         yac_init_domain(&l_domain_p1_src, comm_named_p1_src, blocks_src[0], 1, "domain_src");
         yac_init_domain(&l_domain_p2_src, comm_named_p2_src, blocks_src[1], 1, "domain_src");

         // define a domain on the target

         struct grid * grid_tgt;
         unsigned num_cells_tgt[2] = {2,2};
         unsigned cyclic_tgt[2] = {0,0};
         double coordinates_x_tgt[3] = {0.5*rad,1.5*rad,2.5*rad};
         double coordinates_y_tgt[3] = {0.5*rad,1.5*rad,2.5*rad};

         grid_tgt = yac_reg2d_grid_new(coordinates_x_tgt, coordinates_y_tgt,
                                       num_cells_tgt, cyclic_tgt);

         struct subdomain subdomain_tgt;

         int global_corner_ids_tgt[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned corner_is_on_rank_tgt[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_tgt[4] = {0,1,2,3};
         unsigned cell_is_on_rank_tgt[4] = {-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_tgt, grid_tgt, comm_named_p1_tgt,
                                     global_corner_ids_tgt, corner_is_on_rank_tgt,
                                     global_cell_ids_tgt, cell_is_on_rank_tgt,
                                     NULL, NULL);

         struct domain l_domain_tgt;

         struct subdomain * blocks_tgt[1] = {&subdomain_tgt};

         yac_init_domain(&l_domain_tgt, comm_named_p1_tgt, blocks_tgt, 1, "domain_tgt");

         // finish domain definition
         yac_enddef_domain(&l_domain_p1_src);
         yac_enddef_domain(&l_domain_p2_src);
         yac_enddef_domain(&l_domain_tgt);

         //-------------------------------------
         // init global search on both processes
         //-------------------------------------

         struct global_search * search[3];

         search[0] = yac_global_search_new(&l_domain_p1_src, "domain_tgt", comm_p1_src,
                                           1,0, "component_src", "component_tgt");
         search[1] = yac_global_search_new(&l_domain_p2_src, "domain_tgt", comm_p2_src,
                                           1,0, "component_src", "component_tgt");
         search[2] = yac_global_search_new(&l_domain_tgt, "domain_src", comm_p1_tgt,
                                           0,1, "component_tgt", "component_src");

         //---------------------------------------
         // finish initialisation of global search
         //---------------------------------------

         yac_enddef_global_search(search[0]);
         yac_enddef_global_search(search[1]);
         yac_enddef_global_search(search[2]);

      //---------------------------------
      // initialise points on source side
      //---------------------------------

      struct points points_src[2];

      yac_init_points(points_src+0, grid_p1_src, CORNER,
                      coordinates_x_src[0], coordinates_y_src[0]);
      yac_init_points(points_src+1, grid_p2_src, CORNER,
                      coordinates_x_src[1], coordinates_y_src[1]);

      //-------------------------------------
      // set up point sets
      //-------------------------------------

      struct points * points_per_subdomain[2][1] = {{points_src+0},{points_src+1}};
      struct points ** point_sets[2][1] = {{points_per_subdomain[0]},
                                           {points_per_subdomain[1]}};

      //----------------------------------------
      // test generation of interpolation method
      //----------------------------------------

      struct interpolation_method * method_avg[2];

      method_avg[0] = yac_new_interpolation_method_avg(search[0], &l_domain_p1_src,
                                                   "avg stack", point_sets[0],
                                                   NULL, 1, 0, 0, 0);
      method_avg[1] = yac_new_interpolation_method_avg(search[1], &l_domain_p2_src,
                                                   "avg stack", point_sets[1],
                                                   NULL, 1, 0, 0, 0);

      //----------
      // do search
      //----------

      struct interpolation_method * method_stack[2][2] = {{method_avg[0],
                                                           NULL},
                                                          {method_avg[1],
                                                           NULL}};
      struct interpolation_plan_element * interpolation_plan_element_stack[2][2] = {{NULL, NULL},
                                                                    {NULL, NULL}};
      int global_ids[2][9] = {{0,1,2,3,4,5,6,7,8}, {0,1,2,3,4,5,6,7,8}};
      unsigned num_points = 9;
      unsigned num_interpolated_points[2];
      double coordinates_x_tgt_cells[2][9] = {{0.5,1.5,2.5, 0.5,1.5,2.5, 0.5,1.5,2.5},
                                              {0.5,1.5,2.5, 0.5,1.5,2.5, 0.5,1.5,2.5}};
      double coordinates_y_tgt_cells[2][9] = {{0.5,0.5,0.5, 1.5,1.5,1.5, 2.5,2.5,2.5},
                                              {0.5,0.5,0.5, 1.5,1.5,1.5, 2.5,2.5,2.5}};

      for ( i = 0; i < 2; ++i ) {
         for ( j = 0; j < 9; ++j ) {
            coordinates_x_tgt_cells[i][j] *= rad;
            coordinates_y_tgt_cells[i][j] *= rad;
         }
      }

      yac_do_search_interpolation_method(method_stack[0], interpolation_plan_element_stack[0],
         global_ids[0], coordinates_x_tgt_cells[0], coordinates_y_tgt_cells[0], CORNER,
         num_points, num_interpolated_points+0, 1);
      yac_do_search_interpolation_method(method_stack[1], interpolation_plan_element_stack[1],
         global_ids[1], coordinates_x_tgt_cells[1], coordinates_y_tgt_cells[1], CORNER,
         num_points, num_interpolated_points+1, 1);
      yac_enddef_interpolation_method(method_stack[0]);
      yac_enddef_interpolation_method(method_stack[1]);

      //---------------------
      // do the interpolation
      //---------------------

      double in_data_raw[2][2][9] = {{{-1,-1,-1,-1,-1,-1,-1,-1,-1},
                                      {0,1,2,4,5,6,8,9,10}},
                                     {{1,5,9,10,-1,-1,-1,-1,-1},
                                      {1,2,3,5,6,7,9,10,11}}};
      double * subdomain_in_data[2][1] = {{in_data_raw[0][1]}, {in_data_raw[1][1]}};
      double * halo_data[2][1] = {{in_data_raw[0][0]}, {in_data_raw[1][0]}};
      double ** in_data[2][2] = {{halo_data[0], subdomain_in_data[0]},
                                 {halo_data[1], subdomain_in_data[1]}};
      double send_buffer[2][9];

      yac_execute_interpolation_plan_element(interpolation_plan_element_stack[0], in_data[0], send_buffer[0]);
      yac_execute_interpolation_plan_element(interpolation_plan_element_stack[1], in_data[1], send_buffer[1]);

      //---------------------------------
      // testing results of interpolation
      //---------------------------------

      struct interpolation_method_data_request data_request_[2];
      struct interpolation_method_data_request * data_request[2][1] = {{data_request_+0},
                                                                       {data_request_+1}};

      yac_get_source_exchange_data_interpolation_method(method_stack[0], data_request[0], 1);
      yac_get_source_exchange_data_interpolation_method(method_stack[1], data_request[1], 1);

      struct interpolation_method_data_request ref_data_request[2] =
         {{0, NULL, NULL}, {4, (int[]){1,5,9,10}, (int[]){0,0,0,0}}};

      check_data_request(data_request[0][0], ref_data_request+0, 1);
      check_data_request(data_request[1][0], ref_data_request+1, 1);

      unsigned ref_num_interpolated_points[2] = {2,4};
      double ref_out_data[2][4] = {{2.5,6.5},{3.5,4.5,7.5,8.5}};

      int ref_global_ids[2][4] = {{0,3},{1,2,4,5}};

      for (i = 0; i < 2; ++i)
         if (ref_num_interpolated_points[i] != num_interpolated_points[i])
            PUT_ERR("wrong number of interpolated points\n");

      for (i = 0; i < 2; ++i)
         for (j = 0; j < ref_num_interpolated_points[i]; ++j) {

            if (ref_global_ids[i][j] != global_ids[i][j])
               PUT_ERR("wrong global id order\n");

            if (send_buffer[i][j] != ref_out_data[i][j])
               PUT_ERR("wrong result buffer d\n");
         }

      //---------------
      // cleanup
      //---------------

      for (j = 0; j < 2; ++j) {
         for (i = 0; i < sizeof(method_stack[j]) /
                         sizeof(method_stack[j][0]); ++i)
            if (method_stack[j][i] != NULL)
               yac_delete_interpolation_method(method_stack[j][i]);
         for (i = 0; i < sizeof(interpolation_plan_element_stack[j]) /
                         sizeof(interpolation_plan_element_stack[j][0]); ++i) {
            yac_free_interpolation_plan_element(interpolation_plan_element_stack[j][i]);
            free(interpolation_plan_element_stack[j][i]);
         }
         delete_interpolation_method_data_request_stack(data_request[j][0], 1);
      }

      yac_free_points(points_src+0);
      yac_free_points(points_src+1);

      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);
      yac_global_search_delete(search[2]);

      yac_free_domain(&l_domain_p1_src);
      yac_free_domain(&l_domain_p2_src);
      yac_free_domain(&l_domain_tgt);

      yac_delete_grid(grid_p1_src);
      yac_delete_grid(grid_p2_src);
      yac_delete_grid(grid_tgt);

      yac_free_communicator(comm_p1_src);
      yac_free_communicator(comm_p2_src);
      yac_free_communicator(comm_p1_tgt);

      yac_free_communicator(comm_named_p1_src);
      yac_free_communicator(comm_named_p2_src);
      yac_free_communicator(comm_named_p1_tgt);

      yac_reset_interpolation_method();
   }

   { // linear point interpolation with two source processes and a single target process
     // the source fields have two struct points per subdomain

      // the global domain is a 4x2 grid:
      //       08--14--09--15--10--16--11
      //       |       |       |       |
      //       08  03  10  04  12  05  13
      //       |       |       |       |
      //       04--07--05--09--06--11--07
      //       |       |       |       |
      //       01  00  03  01  05  02  06
      //       |       |       |       |
      //       00--00--01--02--02--04--03
      //
      // domain_a is distributed among two processes with one subdomain each
      // the partitioning is as follows:
      //
      //       0--0--0--0--0--1--1
      //       |     |     |     |
      //       0  0  0  0  1  1  1
      //       |     |     |     |
      //       0--0--0--1--1--1--1
      //       |     |     |     |
      //       0  0  0  1  1  1  1
      //       |     |     |     |
      //       0--0--0--1--1--1--1
      //
      //---------------
      // setup
      //---------------

      // comm worlds on both processes
      struct communicator * comm_p1_src, * comm_p2_src, * comm_p1_tgt;

      comm_p1_src = yac_local_comm_new();
      comm_p2_src = yac_local_comm_new();
      comm_p1_tgt = yac_local_comm_new();

      // named comms for the domains on both processes
      struct communicator * comm_named_p1_src, * comm_named_p2_src, * comm_named_p1_tgt;

      comm_named_p1_src = yac_named_local_comm_new(comm_p1_src, "src");
      comm_named_p2_src = yac_named_local_comm_new(comm_p2_src, "src");
      comm_named_p1_tgt = yac_named_local_comm_new(comm_p1_tgt, "tgt");

         // define domains on source

         struct grid * grid_p1_src, * grid_p2_src;
         unsigned num_cells_src[2] = {2,2};
         unsigned cyclic_src[2] = {0,0};
         double coordinates_x_src[2][3] = {{0,1,2}, {1,2,3}};
         double coordinates_y_src[2][3] = {{0,1,2}, {0,1,2}};

         for ( i = 0; i < 2; ++i ) {
            for ( j = 0; j < 3; ++j ) {
               coordinates_x_src[i][j] *= rad;
               coordinates_y_src[i][j] *= rad;
            }
         }

         grid_p1_src = yac_reg2d_grid_new(coordinates_x_src[0], coordinates_y_src[0],
                                          num_cells_src, cyclic_src);
         grid_p2_src = yac_reg2d_grid_new(coordinates_x_src[1], coordinates_y_src[1],
                                          num_cells_src, cyclic_src);

         struct subdomain subdomain_p1_src, subdomain_p2_src;

         int global_corner_ids_src[2][9] = {{0,1,2, 4,5,6, 8,9,10},
                                            {1,2,3, 5,6,7, 9,10,11}};
         unsigned corner_is_on_rank_src[2][9] = {{-1,-1,1, -1,-1,1, -1,-1,-1,},
                                                 {0,-1,-1, 0,-1,-1, 0,0,-1}};
         int global_cell_ids_src[2][4] = {{0,1,3,4},
                                          {1,2,4,5}};
         unsigned cell_is_on_rank_src[2][4] = {{-1,1,-1,-1},
                                               {-1,-1,0,-1}};

         yac_init_subdomain_unstruct(&subdomain_p1_src, grid_p1_src, comm_named_p1_src,
                                     global_corner_ids_src[0], corner_is_on_rank_src[0],
                                     global_cell_ids_src[0], cell_is_on_rank_src[0],
                                     NULL, NULL);

         yac_init_subdomain_unstruct(&subdomain_p2_src, grid_p2_src, comm_named_p2_src,
                                     global_corner_ids_src[1], corner_is_on_rank_src[1],
                                     global_cell_ids_src[1], cell_is_on_rank_src[1],
                                     NULL, NULL);

         struct domain l_domain_p1_src, l_domain_p2_src;

         struct subdomain * blocks_src[2][1] = {{&subdomain_p1_src},
                                                {&subdomain_p2_src}};

         yac_init_domain(&l_domain_p1_src, comm_named_p1_src, blocks_src[0], 1, "domain_src");
         yac_init_domain(&l_domain_p2_src, comm_named_p2_src, blocks_src[1], 1, "domain_src");

         // define a domain on the target

         struct grid * grid_tgt;
         unsigned num_cells_tgt[2] = {2,2};
         unsigned cyclic_tgt[2] = {0,0};
         double coordinates_x_tgt[3] = {0.5*rad,1.5*rad,2.5*rad};
         double coordinates_y_tgt[3] = {0.5*rad,1.5*rad,2.5*rad};

         grid_tgt = yac_reg2d_grid_new(coordinates_x_tgt, coordinates_y_tgt,
                                       num_cells_tgt, cyclic_tgt);

         struct subdomain subdomain_tgt;

         int global_corner_ids_tgt[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned corner_is_on_rank_tgt[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_tgt[4] = {0,1,2,3};
         unsigned cell_is_on_rank_tgt[4] = {-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_tgt, grid_tgt, comm_named_p1_tgt,
                                     global_corner_ids_tgt, corner_is_on_rank_tgt,
                                     global_cell_ids_tgt, cell_is_on_rank_tgt,
                                     NULL, NULL);

         struct domain l_domain_tgt;

         struct subdomain * blocks_tgt[1] = {&subdomain_tgt};

         yac_init_domain(&l_domain_tgt, comm_named_p1_tgt, blocks_tgt, 1, "domain_tgt");

         // finish domain definition
         yac_enddef_domain(&l_domain_p1_src);
         yac_enddef_domain(&l_domain_p2_src);
         yac_enddef_domain(&l_domain_tgt);

         //-------------------------------------
         // init global search on both processes
         //-------------------------------------

         struct global_search * search[3];

         search[0] = yac_global_search_new(&l_domain_p1_src, "domain_tgt", comm_p1_src,
                                           1,0, "component_src", "component_tgt");
         search[1] = yac_global_search_new(&l_domain_p2_src, "domain_tgt", comm_p2_src,
                                           1,0, "component_src", "component_tgt");
         search[2] = yac_global_search_new(&l_domain_tgt, "domain_src", comm_p1_tgt,
                                           0,1, "component_tgt", "component_src");

         //---------------------------------------
         // finish initialisation of global search
         //---------------------------------------

         yac_enddef_global_search(search[0]);
         yac_enddef_global_search(search[1]);
         yac_enddef_global_search(search[2]);

      //---------------------------------
      // initialise points on source side
      //---------------------------------

      struct points corner_points_src[2];
      struct points cell_points_src[2];

      yac_init_points(corner_points_src+0, grid_p1_src, CORNER,
                      coordinates_x_src[0], coordinates_y_src[0]);
      yac_init_points(corner_points_src+1, grid_p2_src, CORNER,
                      coordinates_x_src[1], coordinates_y_src[1]);

      double cell_coordinates_x_src[2][2] = {{0.5,1.5},{1.5,2.5}};
      double cell_coordinates_y_src[2][2] = {{0.5,1.5},{0.5,1.5}};

      for ( i = 0; i < 2; ++i ) {
         for ( j = 0; j < 2; ++j ) {
            cell_coordinates_x_src[i][j] *= rad;
            cell_coordinates_y_src[i][j] *= rad;
         }
      }

      yac_init_points(cell_points_src+0, grid_p1_src, CELL,
                      cell_coordinates_x_src[0], cell_coordinates_y_src[0]);
      yac_init_points(cell_points_src+1, grid_p2_src, CELL,
                      cell_coordinates_x_src[1], cell_coordinates_y_src[1]);

      //-------------------------------------
      // set up coupling field on source side
      //-------------------------------------

      struct points * points_per_subdomain[2][2] = {{corner_points_src+0, cell_points_src+0},
                                                    {corner_points_src+1, cell_points_src+1}};
      struct points ** point_sets[2][1] = {{points_per_subdomain[0]},
                                           {points_per_subdomain[1]}};

      //----------------------------------------
      // test generation of interpolation method
      //----------------------------------------

      struct interpolation_method * method_avg[2];

      method_avg[0] = yac_new_interpolation_method_avg(search[0], &l_domain_p1_src,
                                                   "avg stack", point_sets[0],
                                                   NULL, 2, 0, 0, 0);
      method_avg[1] = yac_new_interpolation_method_avg(search[1], &l_domain_p2_src,
                                                   "avg stack", point_sets[1],
                                                   NULL, 2, 0, 0, 0);

      //----------
      // do search
      //----------

      struct interpolation_method * method_stack[2][2] = {{method_avg[0],
                                                           NULL},
                                                          {method_avg[1],
                                                           NULL}};
      struct interpolation_plan_element * interpolation_plan_element_stack[2][2] = {{NULL, NULL},
                                                                    {NULL, NULL}};
      int global_ids[2][9] = {{0,1,2,3,4,5,6,7,8}, {0,1,2,3,4,5,6,7,8}};
      unsigned num_points = 9;
      unsigned num_interpolated_points[2];
      double coordinates_x_tgt_cells[2][9] = {{0.5,1.5,2.5, 0.5,1.5,2.5, 0.5,1.5,2.5},
                                              {0.5,1.5,2.5, 0.5,1.5,2.5, 0.5,1.5,2.5}};
      double coordinates_y_tgt_cells[2][9] = {{0.5,0.5,0.5, 1.5,1.5,1.5, 2.5,2.5,2.5},
                                              {0.5,0.5,0.5, 1.5,1.5,1.5, 2.5,2.5,2.5}};

      for ( i = 0; i < 2; ++i ) {
         for ( j = 0; j < 9; ++j ) {
            coordinates_x_tgt_cells[i][j] *= rad;
            coordinates_y_tgt_cells[i][j] *= rad;
         }
      }

      yac_do_search_interpolation_method(method_stack[0], interpolation_plan_element_stack[0],
         global_ids[0], coordinates_x_tgt_cells[0], coordinates_y_tgt_cells[0], CORNER,
         num_points, num_interpolated_points+0, 2);
      yac_do_search_interpolation_method(method_stack[1], interpolation_plan_element_stack[1],
         global_ids[1], coordinates_x_tgt_cells[1], coordinates_y_tgt_cells[1], CORNER,
         num_points, num_interpolated_points+1, 2);
      yac_enddef_interpolation_method(method_stack[0]);
      yac_enddef_interpolation_method(method_stack[1]);

      //---------------------
      // do the interpolation
      //---------------------

      double in_data_raw[2][2][9] = {{{-1,-1,-1,-1,-1,-1,-1,-1,-1},
                                      {0,1,2,4,5,6,8,9,10}},
                                     {{1,5,9,10,-1,-1,-1,-1,-1},
                                      {1,2,3,5,6,7,9,10,11}}};
      double * subdomain_in_data[2][2] = {{in_data_raw[0][1], NULL}, {in_data_raw[1][1], NULL}};
      double * halo_in_data[2][2] = {{in_data_raw[0][0], NULL}, {in_data_raw[1][0], NULL}};
      double ** in_data[2][2] = {{halo_in_data[0], subdomain_in_data[0]},
                                 {halo_in_data[1], subdomain_in_data[1]}};
      double send_buffer[2][9];

      yac_execute_interpolation_plan_element(interpolation_plan_element_stack[0], in_data[0], send_buffer[0]);
      yac_execute_interpolation_plan_element(interpolation_plan_element_stack[1], in_data[1], send_buffer[1]);

      //---------------------------------
      // testing results of interpolation
      //---------------------------------

      struct interpolation_method_data_request data_request_[2][2];
      struct interpolation_method_data_request * data_request[2] = {data_request_[0],
                                                                    data_request_[1]};

      yac_get_source_exchange_data_interpolation_method(method_stack[0], data_request + 0, 2);
      yac_get_source_exchange_data_interpolation_method(method_stack[1], data_request + 1, 2);

      struct interpolation_method_data_request ref_data_request[2][2] =
         {{{0, NULL, NULL}, {0, NULL, NULL}},
          {{4, (int[]){1,5,9,10}, (int[]){0,0,0,0}}, {0, NULL, NULL}}};

      check_data_request(data_request[0], ref_data_request[0], 2);
      check_data_request(data_request[1], ref_data_request[1], 2);

      unsigned ref_num_interpolated_points[2] = {2,4};
      double ref_out_data[2][4] = {{2.5,6.5},{3.5,4.5,7.5,8.5}};

      int ref_global_ids[2][4] = {{0,3},{1,2,4,5}};

      for (i = 0; i < 2; ++i)
         if (ref_num_interpolated_points[i] != num_interpolated_points[i])
            PUT_ERR("wrong number of interpolated points\n");

      for (i = 0; i < 2; ++i)
         for (j = 0; j < ref_num_interpolated_points[i]; ++j) {

            if (ref_global_ids[i][j] != global_ids[i][j])
               PUT_ERR("wrong global id order\n");

            if (send_buffer[i][j] != ref_out_data[i][j])
               PUT_ERR("wrong result buffer\n");
         }

      //---------------
      // cleanup
      //---------------

      for (j = 0; j < 2; ++j) {
         for (i = 0; i < sizeof(method_stack[j]) /
                         sizeof(method_stack[j][0]); ++i)
            if (method_stack[j][i] != NULL)
               yac_delete_interpolation_method(method_stack[j][i]);
         for (i = 0; i < sizeof(interpolation_plan_element_stack[j]) /
                         sizeof(interpolation_plan_element_stack[j][0]); ++i) {
            yac_free_interpolation_plan_element(interpolation_plan_element_stack[j][i]);
            free(interpolation_plan_element_stack[j][i]);
         }
         delete_interpolation_method_data_request_stack(data_request[j], 2);
      }

      yac_free_points(corner_points_src+0);
      yac_free_points(corner_points_src+1);
      yac_free_points(cell_points_src+0);
      yac_free_points(cell_points_src+1);

      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);
      yac_global_search_delete(search[2]);

      yac_free_domain(&l_domain_p1_src);
      yac_free_domain(&l_domain_p2_src);
      yac_free_domain(&l_domain_tgt);

      yac_delete_grid(grid_p1_src);
      yac_delete_grid(grid_p2_src);
      yac_delete_grid(grid_tgt);

      yac_free_communicator(comm_p1_src);
      yac_free_communicator(comm_p2_src);
      yac_free_communicator(comm_p1_tgt);

      yac_free_communicator(comm_named_p1_src);
      yac_free_communicator(comm_named_p2_src);
      yac_free_communicator(comm_named_p1_tgt);

      yac_reset_interpolation_method();
   }

   { // simple fixed value "interpolation" on four source processes and a
     // grid with only a single cell

      //--------------------
      // set up global search
      //--------------------

         // one directional global search with two domains (source and target)
         // on one target process and four source processes with one subdomain
         // each

         // comm worlds all processes
         struct communicator * comm_src[4], * comm_tgt;

         for (i = 0; i < 4; ++i)
            comm_src[i] = yac_local_comm_new();
         comm_tgt = yac_local_comm_new();

         // named comms for the domains
         struct communicator * comm_named_src[4], * comm_named_tgt;

         for (i = 0; i < 4; ++i)
            comm_named_src[i] = yac_named_local_comm_new(comm_src[i], "source");
         comm_named_tgt = yac_named_local_comm_new(comm_tgt, "target");

         // define a domain on the source processes

         struct grid * grid_src[4];
         unsigned num_cells[2] = {1,1};
         unsigned cyclic[2] = {0,0};
         double coordinates_x[2] = {0,1*rad};
         double coordinates_y[2] = {0,1*rad};

         for (i = 0; i < 4; ++i)
            grid_src[i] = yac_reg2d_grid_new(coordinates_x, coordinates_y,
                                             num_cells, cyclic);

         struct subdomain subdomain_src[4];

         int global_corner_ids_src[4] = {0,1,2,3};
         unsigned corner_is_on_rank_src[4][4] = {{-1, 1, 2, 3},
                                                 { 0,-1, 2, 3},
                                                 { 0, 1,-1, 3},
                                                 { 0, 1, 2,-1}};
         int global_cell_ids_src[1] = {0};
         unsigned cell_is_on_rank_src[4][1] = {{-1},{0},{0},{0}};

         for (i = 0; i < 4; ++i)
            yac_init_subdomain_unstruct(subdomain_src+i, grid_src[i], comm_named_src[i],
                                        global_corner_ids_src, corner_is_on_rank_src[i],
                                        global_cell_ids_src, cell_is_on_rank_src[i],
                                        NULL, NULL);

         struct domain domain_src[4];

         struct subdomain * blocks_src[4][1] = {{subdomain_src+0},
                                                {subdomain_src+1},
                                                {subdomain_src+2},
                                                {subdomain_src+3}};

         for (i = 0; i < 4; ++i)
            yac_init_domain(domain_src+i, comm_named_src[i], blocks_src[i], 1,
                        "source domain");

         // define the target domain

         struct grid * grid_tgt;

         grid_tgt = yac_reg2d_grid_new(coordinates_x, coordinates_y,
                                       num_cells, cyclic);

         struct subdomain subdomain_tgt;

         int global_corner_ids_tgt[4] = {0,1,2,3};
         unsigned corner_is_on_rank_tgt[4] = {-1,-1,-1,-1};
         int global_cell_ids_tgt[1] = {0};
         unsigned cell_is_on_rank_tgt[1] = {-1};

         yac_init_subdomain_unstruct(&subdomain_tgt, grid_tgt, comm_named_tgt,
                                     global_corner_ids_tgt, corner_is_on_rank_tgt,
                                     global_cell_ids_tgt, cell_is_on_rank_tgt,
                                     NULL, NULL);

         struct domain domain_tgt;

         struct subdomain * blocks_tgt[1] = {&subdomain_tgt};

         yac_init_domain(&domain_tgt, comm_named_tgt, blocks_tgt, 1, "target domain");

         // finish domain definition

         for (i = 0; i < 4; ++i)
            yac_enddef_domain(domain_src+i);
         yac_enddef_domain(&domain_tgt);

         // init global search on both processes

         struct global_search * search_src[4];
         struct global_search * search_tgt;

         for (i = 0; i < 4; ++i)
            search_src[i] =
               yac_global_search_new(domain_src+i, "target domain", comm_src[i],
                                     1,0, "source component", "target component");
         search_tgt =
            yac_global_search_new(&domain_tgt, "source domain", comm_tgt, 0,1,
                                  "target component", "source component");

         // finish initialisation of global search

         for (i = 0; i < 4; ++i)
            yac_enddef_global_search(search_src[i]);
         yac_enddef_global_search(search_tgt);

      //----------------------
      // set up coupling field
      //----------------------

      struct coupling_field * field[4];

      for (i = 0; i < 4; ++i)
         field[i] = yac_coupling_field_new("source field", component_id, domain_src+i, NULL, NULL, 0);

      //----------------------------------------
      // test generation of interpolation method
      //----------------------------------------

      struct interpolation_method * method_fixed[4];
      double fixed_value = 1337;

      for (i = 0; i < 4; ++i)
         method_fixed[i] = yac_new_interpolation_method_fixed(fixed_value, search_src[i]);

      //----------
      // do search
      //----------

      struct interpolation_method * interpolation_method_stack[4][2] =
         {{method_fixed[0], NULL}, {method_fixed[1], NULL},
          {method_fixed[2], NULL}, {method_fixed[3], NULL}};
      struct interpolation_plan_element * interpolation_plan_element_stack[4][2] =
         {{NULL, NULL}, {NULL, NULL}, {NULL, NULL}, {NULL, NULL}};
      int global_ids[4][1] = {{0},{0},{0},{0}};
      double coordinates_x_search[1] = {0.5*rad};
      double coordinates_y_search[1] = {0.5*rad};
      unsigned num_points = 1;
      unsigned num_interpolated_points[4];

      for (i = 0; i < 4; ++i)
         yac_do_search_interpolation_method(interpolation_method_stack[i],
            interpolation_plan_element_stack[i], global_ids[i], coordinates_x_search,
            coordinates_y_search, CORNER, num_points, num_interpolated_points+i, 0);
      for (i = 0; i < 4; ++i)
         yac_enddef_interpolation_method(interpolation_method_stack[i]);

      //---------------------
      // do the interpolation
      //---------------------

      double send_buffer[4][1] = {{-1},{-1},{-1},{-1}};

      for (i = 0; i < 4; ++i)
         yac_execute_interpolation_plan_element(
            interpolation_plan_element_stack[i], NULL, send_buffer[i]);

      //---------------------------------
      // testing results of interpolation
      //---------------------------------

      struct interpolation_method_data_request data_request_[4];
      struct interpolation_method_data_request * data_request[4][1] =
        {{data_request_+0}, {data_request_+1}, {data_request_+2}, {data_request_+3}};

      for (i = 0; i < 4; ++i)
         yac_get_source_exchange_data_interpolation_method(
            interpolation_method_stack[i], data_request[i], 1);

      unsigned num_no_interp = 0, num_fixed = 0;
      for (i = 0; i < 4; ++i)
         if (send_buffer[i][0] == fixed_value)
            num_fixed++;
         else if (send_buffer[i][0] == -1)
            num_no_interp++;
         else
            PUT_ERR("wrong result buffer f\n");

      if (num_fixed != 1 || num_no_interp != 3)
         PUT_ERR("wrong result buffer g\n");

      //---------------
      // cleanup
      //---------------

      for (i = 0; i < 4; ++i)
         yac_delete_interpolation_method(method_fixed[i]);

      for (j = 0; j < 4; ++j) {
         for (i = 0; i < sizeof(interpolation_plan_element_stack[j]) /
                         sizeof(interpolation_plan_element_stack[j][0]); ++i) {
            yac_free_interpolation_plan_element(interpolation_plan_element_stack[j][i]);
            free(interpolation_plan_element_stack[j][i]);
         }
         delete_interpolation_method_data_request_stack(data_request[j][0], 1);
      }

      for (i = 0; i < 4; ++i)
         yac_coupling_field_delete(field[i]);

      for (i = 0; i < 4; ++i)
         yac_global_search_delete(search_src[i]);
      yac_global_search_delete(search_tgt);

      for (i = 0; i < 4; ++i)
         yac_free_domain(domain_src+i);
      yac_free_domain(&domain_tgt);

      for (i = 0; i < 4; ++i)
         yac_delete_grid(grid_src[i]);
      yac_delete_grid(grid_tgt);

      for (i = 0; i < 4; ++i)
         yac_free_communicator(comm_src[i]);
      yac_free_communicator(comm_tgt);

      for (i = 0; i < 4; ++i)
         yac_free_communicator(comm_named_src[i]);
      yac_free_communicator(comm_named_tgt);

      yac_reset_interpolation_method();
   }

   { // interpolation with two source processes and a single target process
     // linear and nearest-neighbour interpolation is used (on the second
     // source process both interpolations need data from remote processes)

      // the global domain is a 4x2 grid:
      //       08--14--09--15--10--16--11
      //       |       |       |       |
      //       08  03  10  04  12  05  13
      //       |       |       |       |
      //       04--07--05--09--06--11--07
      //       |       |       |       |
      //       01  00  03  01  05  02  06
      //       |       |       |       |
      //       00--00--01--02--02--04--03
      //
      // domain_a is distributed among two processes with one subdomain each
      // the partitioning is as follows:
      //
      //       0--0--0--0--0--1--1
      //       |     |     |     |
      //       0  0  0  0  1  1  1
      //       |     |     |     |
      //       0--0--0--1--1--1--1
      //       |     |     |     |
      //       0  0  0  1  1  1  1
      //       |     |     |     |
      //       0--0--0--1--1--1--1
      //
      //---------------
      // setup
      //---------------

      // comm worlds on both processes
      struct communicator * comm_p1_src, * comm_p2_src, * comm_p1_tgt;

      comm_p1_src = yac_local_comm_new();
      comm_p2_src = yac_local_comm_new();
      comm_p1_tgt = yac_local_comm_new();

      // named comms for the domains on both processes
      struct communicator * comm_named_p1_src, * comm_named_p2_src, * comm_named_p1_tgt;

      comm_named_p1_src = yac_named_local_comm_new(comm_p1_src, "src");
      comm_named_p2_src = yac_named_local_comm_new(comm_p2_src, "src");
      comm_named_p1_tgt = yac_named_local_comm_new(comm_p1_tgt, "tgt");

         // define domains on source

         struct grid * grid_p1_src, * grid_p2_src;
         unsigned num_cells_src[2] = {2,2};
         unsigned cyclic_src[2] = {0,0};
         double coordinates_x_src[2][3] = {{0,1,2}, {1,2,3}};
         double coordinates_y_src[2][3] = {{0,1,2}, {0,1,2}};

         for ( i = 0; i < 2; ++i ) {
            for ( j = 0; j < 3; ++j ) {
               coordinates_x_src[i][j] *= rad;
               coordinates_y_src[i][j] *= rad;
            }
         }

         grid_p1_src = yac_reg2d_grid_new(coordinates_x_src[0], coordinates_y_src[0],
                                          num_cells_src, cyclic_src);
         grid_p2_src = yac_reg2d_grid_new(coordinates_x_src[1], coordinates_y_src[1],
                                          num_cells_src, cyclic_src);

         struct subdomain subdomain_p1_src, subdomain_p2_src;

         int global_corner_ids_src[2][9] = {{0,1,2, 4,5,6, 8,9,10},
                                            {1,2,3, 5,6,7, 9,10,11}};
         unsigned corner_is_on_rank_src[2][9] = {{-1,-1,1, -1,-1,1, -1,-1,-1,},
                                                 {0,-1,-1, 0,-1,-1, 0,0,-1}};
         int global_cell_ids_src[2][4] = {{0,1,3,4},
                                          {1,2,4,5}};
         unsigned cell_is_on_rank_src[2][4] = {{-1,1,-1,-1},
                                               {-1,-1,0,-1}};

         yac_init_subdomain_unstruct(&subdomain_p1_src, grid_p1_src, comm_named_p1_src,
                                     global_corner_ids_src[0], corner_is_on_rank_src[0],
                                     global_cell_ids_src[0], cell_is_on_rank_src[0],
                                     NULL, NULL);

         yac_init_subdomain_unstruct(&subdomain_p2_src, grid_p2_src, comm_named_p2_src,
                                     global_corner_ids_src[1], corner_is_on_rank_src[1],
                                     global_cell_ids_src[1], cell_is_on_rank_src[1],
                                     NULL, NULL);

         struct domain l_domain_p1_src, l_domain_p2_src;

         struct subdomain * blocks_src[2][1] = {{&subdomain_p1_src},
                                                {&subdomain_p2_src}};

         yac_init_domain(&l_domain_p1_src, comm_named_p1_src, blocks_src[0], 1, "domain_src");
         yac_init_domain(&l_domain_p2_src, comm_named_p2_src, blocks_src[1], 1, "domain_src");

         // define a domain on the target

         struct grid * grid_tgt;
         unsigned num_cells_tgt[2] = {2,2};
         unsigned cyclic_tgt[2] = {0,0};
         double coordinates_x_tgt[3] = {0.5*rad,1.5*rad,2.5*rad};
         double coordinates_y_tgt[3] = {0.5*rad,1.5*rad,2.5*rad};

         grid_tgt = yac_reg2d_grid_new(coordinates_x_tgt, coordinates_y_tgt,
                                       num_cells_tgt, cyclic_tgt);

         struct subdomain subdomain_tgt;

         int global_corner_ids_tgt[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned corner_is_on_rank_tgt[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_tgt[4] = {0,1,2,3};
         unsigned cell_is_on_rank_tgt[4] = {-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_tgt, grid_tgt, comm_named_p1_tgt,
                                     global_corner_ids_tgt, corner_is_on_rank_tgt,
                                     global_cell_ids_tgt, cell_is_on_rank_tgt,
                                     NULL, NULL);

         struct domain l_domain_tgt;

         struct subdomain * blocks_tgt[1] = {&subdomain_tgt};

         yac_init_domain(&l_domain_tgt, comm_named_p1_tgt, blocks_tgt, 1, "domain_tgt");

         // finish domain definition
         yac_enddef_domain(&l_domain_p1_src);
         yac_enddef_domain(&l_domain_p2_src);
         yac_enddef_domain(&l_domain_tgt);

         //-------------------------------------
         // init global search on both processes
         //-------------------------------------

         struct global_search * search[3];

         search[0] = yac_global_search_new(&l_domain_p1_src, "domain_tgt", comm_p1_src,
                                           1,0, "component_src", "component_tgt");
         search[1] = yac_global_search_new(&l_domain_p2_src, "domain_tgt", comm_p2_src,
                                           1,0, "component_src", "component_tgt");
         search[2] = yac_global_search_new(&l_domain_tgt, "domain_src", comm_p1_tgt,
                                           0,1, "component_tgt", "component_src");

         //---------------------------------------
         // finish initialisation of global search
         //---------------------------------------

         yac_enddef_global_search(search[0]);
         yac_enddef_global_search(search[1]);
         yac_enddef_global_search(search[2]);

      //---------------------------------
      // initialise points on source side
      //---------------------------------

      struct points points_src[2];

      yac_init_points(points_src+0, grid_p1_src, CORNER,
                      coordinates_x_src[0], coordinates_y_src[0]);
      yac_init_points(points_src+1, grid_p2_src, CORNER,
                      coordinates_x_src[1], coordinates_y_src[1]);

      //-------------------------------------
      // set up point sets
      //-------------------------------------

      struct points * points_per_subdomain[2][1] = {{points_src+0},{points_src+1}};
      struct points ** point_sets[2][1] = {{points_per_subdomain[0]},
                                           {points_per_subdomain[1]}};

      //----------------------------------------
      // test generation of interpolation method
      //----------------------------------------

      struct interpolation_method * method_avg[2];
      struct interpolation_method * method_nnn[2];

      method_avg[0] =
        yac_new_interpolation_method_avg(search[0], &l_domain_p1_src,
                                         "avg/nnn stack", point_sets[0],
                                                   NULL, 1, 0, 0, 0);
      method_avg[1] =
        yac_new_interpolation_method_avg(search[1], &l_domain_p2_src,
                                         "avg/nnn stack", point_sets[1],
                                                   NULL, 1, 0, 0, 0);
      method_nnn[0] =
        yac_new_interpolation_method_nnn(search[0], &l_domain_p1_src,
                                         "avg/nnn stack", point_sets[0], NULL,
                                         1, 0, 4, 0);
      method_nnn[1] =
        yac_new_interpolation_method_nnn(search[1], &l_domain_p2_src,
                                         "avg/nnn stack", point_sets[1], NULL,
                                         1, 0, 4, 0);

      //----------
      // do search
      //----------

      struct interpolation_method * method_stack[2][3] = {{method_avg[0],
                                                           method_nnn[0],
                                                           NULL},
                                                          {method_avg[1],
                                                           method_nnn[1],
                                                           NULL}};
      struct interpolation_plan_element *
        interpolation_plan_element_stack[2][3] = {{NULL, NULL, NULL},
                                                  {NULL, NULL, NULL}};
      int global_ids[2][8] = {{0,1,2,3,4,5,6,7}, {0,1,2,3,4,5,6,7}};
      unsigned num_points = 8;
      unsigned num_interpolated_points[2][2];
      double coordinates_x_tgt_cells[2][8] = {{0.5,1.5,2.5,3.1, 0.5,1.5,2.5,3.1},
                                              {0.5,1.5,2.5,3.1, 0.5,1.5,2.5,3.1}};
      double coordinates_y_tgt_cells[2][8] = {{0.5,0.5,0.5,0.5, 1.5,1.5,1.5,1.5},
                                              {0.5,0.5,0.5,0.5, 1.5,1.5,1.5,1.5}};

      for ( i = 0; i < 2; ++i ) {
         for ( j = 0; j < 8; ++j ) {
            coordinates_x_tgt_cells[i][j] *= rad;
            coordinates_y_tgt_cells[i][j] *= rad;
         }
      }

      yac_do_search_interpolation_method(method_stack[0], interpolation_plan_element_stack[0],
         global_ids[0], coordinates_x_tgt_cells[0], coordinates_y_tgt_cells[0], CORNER,
         num_points, num_interpolated_points[0], 1);
      yac_do_search_interpolation_method(method_stack[1], interpolation_plan_element_stack[1],
         global_ids[1], coordinates_x_tgt_cells[1], coordinates_y_tgt_cells[1], CORNER,
         num_points, num_interpolated_points[1], 1);
      yac_enddef_interpolation_method(method_stack[0]);
      yac_enddef_interpolation_method(method_stack[1]);

      //---------------------
      // do the interpolation
      //---------------------

      double in_data_raw[2][2][9] = {{{-1,-1,-1,-1,-1,-1,-1,-1,-1},
                                      {0,1,2,4,5,6,8,9,10}},
                                     {{1,5,9,10,10,-1,-1,-1,-1},
                                      {1,2,3,5,6,7,9,10,11}}};
      double * subdomain_in_data[2][1] = {{in_data_raw[0][1]}, {in_data_raw[1][1]}};
      double * halo_data[2][1] = {{in_data_raw[0][0]}, {in_data_raw[1][0]}};
      double ** in_data[2][2] = {{halo_data[0], subdomain_in_data[0]},
                                 {halo_data[1], subdomain_in_data[1]}};
      double send_buffer[2][9];

      yac_execute_interpolation_plan_element(interpolation_plan_element_stack[0], in_data[0], send_buffer[0]);
      yac_execute_interpolation_plan_element(interpolation_plan_element_stack[1], in_data[1], send_buffer[1]);

      //---------------------------------
      // testing results of interpolation
      //---------------------------------

      struct interpolation_method_data_request data_request_[2];
      struct interpolation_method_data_request * data_request[2][1] = {{data_request_+0},
                                                                       {data_request_+1}};

      yac_get_source_exchange_data_interpolation_method(method_stack[0], data_request[0], 1);
      yac_get_source_exchange_data_interpolation_method(method_stack[1], data_request[1], 1);

      struct interpolation_method_data_request ref_data_request[2] =
         {{0, NULL, NULL}, {5, (int[]){1,5,9,10,10}, (int[]){0,0,0,0,0}}};

      check_data_request(data_request[0][0], ref_data_request+0, 1);
      check_data_request(data_request[1][0], ref_data_request+1, 1);

      unsigned ref_num_interpolated_points[2][2] = {{2,0},{4,2}};
      double ref_out_data[2][6] = {{2.5,6.5},{3.5,4.5,7.5,8.5,4.5,8.5}};

      int ref_global_ids[2][6] = {{0,4},{1,2,5,6,3,7}};

      for (unsigned i = 0; i < 2; ++i)
        for (unsigned j = 0; j < 2; ++j)
          if (ref_num_interpolated_points[i][j] !=
              num_interpolated_points[i][j])
            PUT_ERR("wrong number of interpolated points\n");

      for (i = 0; i < 2; ++i)
         for (j = 0; j < ref_num_interpolated_points[i][0] +
                         ref_num_interpolated_points[i][1]; ++j) {

            if (ref_global_ids[i][j] != global_ids[i][j])
               PUT_ERR("wrong global id order\n");

            if (send_buffer[i][j] != ref_out_data[i][j])
               PUT_ERR("wrong result buffer d\n");
         }

      //---------------
      // cleanup
      //---------------

      for (j = 0; j < 2; ++j) {
         for (i = 0; i < sizeof(method_stack[j]) /
                         sizeof(method_stack[j][0]); ++i)
            if (method_stack[j][i] != NULL)
               yac_delete_interpolation_method(method_stack[j][i]);
         for (i = 0; i < sizeof(interpolation_plan_element_stack[j]) /
                         sizeof(interpolation_plan_element_stack[j][0]); ++i) {
            yac_free_interpolation_plan_element(interpolation_plan_element_stack[j][i]);
            free(interpolation_plan_element_stack[j][i]);
         }
         delete_interpolation_method_data_request_stack(data_request[j][0], 1);
      }

      yac_free_points(points_src+0);
      yac_free_points(points_src+1);

      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);
      yac_global_search_delete(search[2]);

      yac_free_domain(&l_domain_p1_src);
      yac_free_domain(&l_domain_p2_src);
      yac_free_domain(&l_domain_tgt);

      yac_delete_grid(grid_p1_src);
      yac_delete_grid(grid_p2_src);
      yac_delete_grid(grid_tgt);

      yac_free_communicator(comm_p1_src);
      yac_free_communicator(comm_p2_src);
      yac_free_communicator(comm_p1_tgt);

      yac_free_communicator(comm_named_p1_src);
      yac_free_communicator(comm_named_p2_src);
      yac_free_communicator(comm_named_p1_tgt);

      yac_reset_interpolation_method();
   }

   return TEST_EXIT_CODE;
}

void check_data_request(struct interpolation_method_data_request * request,
                        struct interpolation_method_data_request * ref_request,
                        unsigned num_points_per_subdomain) {

   unsigned i, j;

   for (i = 0; i < num_points_per_subdomain; ++i) {

      if (request[i].num_values != ref_request[i].num_values)
         PUT_ERR("wrong number of requested points\n");

      for (j = 0; j < ref_request[i].num_values; ++j) {

         if (request[i].global_ids[j] != ref_request[i].global_ids[j])
            PUT_ERR("wrong global_id requested\n");

         if (request[i].ranks[j] != ref_request[i].ranks[j])
            PUT_ERR("point from wrong rank requested\n");
      }
   }
}

void
delete_interpolation_method_data_request_stack(
   struct interpolation_method_data_request * request,
   unsigned num_points_per_subdomain) {

   if (request != NULL) {

      unsigned i;

      for (i = 0; i < num_points_per_subdomain; ++i) {
         free(request[i].global_ids);
         free(request[i].ranks);
      }
   }
}
