/**
 * @file test_interpolation_method_nnn.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "tests.h"

#include "grid.h"
#include "grid_reg2d.h"
#include "communicator.h"
#include "communicator_local.h"
#include "domain.h"
#include "global_search.h"
#include "interpolation_method_nnn.h"
#include "interpolation_method_fixed.h"
#include "utils.h"

double const tol = 1e-7;

void check_data_request(struct interpolation_method_data_request * request,
                        struct interpolation_method_data_request* ref_request,
                        unsigned num_points_per_subdomain);

void
delete_interpolation_method_data_request_stack(
   struct interpolation_method_data_request * request,
   unsigned num_points_per_subdomain);

double get_weighted_nnn(struct points * src_points, unsigned * src_idx,
                        double * src_values, unsigned num_src_idx,
                        double tgt_coord_x, double tgt_coord_y);

int main (void) {

   unsigned i, j, k;

   { // interpolation on a single process (1 src and 1 tgt)
     // 1, 3 and 4 nearest neighbours per target point

      //--------------------
      // set up global search
      //--------------------

         // one directional global search with two domains on two processes with one subdomain each

         // comm worlds on both processes
         struct communicator * comm_p1, * comm_p2;

         comm_p1 = yac_local_comm_new();
         comm_p2 = yac_local_comm_new();
         
         // named comms for the domains on both processes
         struct communicator * comm_named_p1, * comm_named_p2;

         comm_named_p1 = yac_named_local_comm_new(comm_p1, "P1");
         comm_named_p2 = yac_named_local_comm_new(comm_p2, "P2");

         // define a domain on P1

         struct grid * grid_p1;
         unsigned num_cells[2] = {2,2};
         unsigned cyclic[2] = {0,0};
         double coordinates_x[3] = {-1*rad,0,1*rad};
         double coordinates_y[3] = {-1*rad,0,1*rad};

         grid_p1 = yac_reg2d_grid_new(coordinates_x, coordinates_y,
                                      num_cells, cyclic);

         struct subdomain subdomain_p1;

         int global_corner_ids_p1[9] = {0,1,2,3,4,5,6,7,8};
         unsigned corner_is_on_rank_p1[9] = {-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_p1[4] = {0,1,2,3};
         unsigned cell_is_on_rank_p1[4] = {-1,-1,-1,-1};
         int global_edge_ids_p1[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned edge_is_on_rank_p1[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_p1, grid_p1, comm_named_p1,
                                     global_corner_ids_p1, corner_is_on_rank_p1,
                                     global_cell_ids_p1, cell_is_on_rank_p1,
                                     global_edge_ids_p1, edge_is_on_rank_p1);

         struct domain l_domain_p1;

         struct subdomain * blocks_p1[1] = {&subdomain_p1};

         yac_init_domain(&l_domain_p1, comm_named_p1, blocks_p1, 1, "domain_p1");

         // define a domain on P2, which matches the one on P1

         struct grid * grid_p2;

         grid_p2 = yac_reg2d_grid_new(coordinates_x, coordinates_y,
                                      num_cells, cyclic);

         struct subdomain subdomain_p2;

         int global_corner_ids_p2[9] = {0,1,2,3,4,5,6,7,8};
         unsigned corner_is_on_rank_p2[9] = {-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_p2[4] = {0,1,2,3};
         unsigned cell_is_on_rank_p2[4] = {-1,-1,-1,-1};
         int global_edge_ids_p2[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned edge_is_on_rank_p2[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_p2, grid_p2, comm_named_p2,
                                     global_corner_ids_p2, corner_is_on_rank_p2,
                                     global_cell_ids_p2, cell_is_on_rank_p2,
                                     global_edge_ids_p2, edge_is_on_rank_p2);

         struct domain l_domain_p2;

         struct subdomain * blocks_p2[1] = {&subdomain_p2};

         yac_init_domain(&l_domain_p2, comm_named_p2, blocks_p2, 1, "domain_p2");

         // finish domain definition
         yac_enddef_domain(&l_domain_p1);
         yac_enddef_domain(&l_domain_p2);

         // init global search on both processes

         struct global_search * search[2];

         search[0] = yac_global_search_new(&l_domain_p1, "domain_p2", comm_p1,
                                           1,0, "component_p1", "component_p2");
         search[1] = yac_global_search_new(&l_domain_p2, "domain_p1", comm_p2,
                                           0,1, "component_p2", "component_p1");

         // finish initialisation of global search

         yac_enddef_global_search(search[0]);
         yac_enddef_global_search(search[1]);

         // points

         struct points points_p1;

         yac_init_points(&points_p1, grid_p1, CORNER, coordinates_x, coordinates_y);

         // point set

         struct points * points_per_subdomain[1] = {&points_p1};
         struct points ** points_set[1] = {&(points_per_subdomain[0])};

      //----------------------------------------
      // generation of interpolation method
      //----------------------------------------

      struct interpolation_method * method_1nn, * method_3nn, * method_4nn;

      method_1nn =
        yac_new_interpolation_method_nnn(search[0], &l_domain_p1, "nnn stack",
                                     points_set, NULL, 1, 0, 1, 0);
      method_3nn =
        yac_new_interpolation_method_nnn(search[0], &l_domain_p1, "nnn stack",
                                     points_set, NULL, 1, 0, 3, 0);
      method_4nn =
        yac_new_interpolation_method_nnn(search[0], &l_domain_p1, "nnn stack",
                                     points_set, NULL, 1, 0, 4, 0);

      //----------
      // do search
      //----------

      struct interpolation_method * interpolation_method_stack[3][2] =
         {{method_1nn, NULL}, {method_3nn, NULL}, {method_4nn, NULL}};
      struct interpolation_plan_element * interpolation_plan_element_stack[3][2] =
         {{NULL, NULL}, {NULL, NULL}, {NULL, NULL}};
      int global_ids[3][4] = {{0,1,2,3}, {0,1,2,3}, {0,1,2,3}};
      unsigned num_points = 4;
      unsigned num_interpolated_points[3];
      double coordinates_x_tgt[3][4] = {{-0.75, -0.75,  0.75, 0.75},
                                        {-0.75, -0.75,  0.75, 0.75},
                                        {-0.75, -0.75,  0.75, 0.75}};
      double coordinates_y_tgt[3][4] = {{-0.75,  0.75, -0.75, 0.75},
                                        {-0.75,  0.75, -0.75, 0.75},
                                        {-0.75,  0.75, -0.75, 0.75}};

      for ( i = 0; i < 3; ++i ) {
         for ( j = 0; j < 4; ++j ) {
            coordinates_x_tgt[i][j] *= rad;
            coordinates_y_tgt[i][j] *= rad;
         }
      }

      for (i = 0; i < 3; ++i)
         yac_do_search_interpolation_method(interpolation_method_stack[i],
            interpolation_plan_element_stack[i], global_ids[i], coordinates_x_tgt[i],
            coordinates_y_tgt[i], CELL, num_points, num_interpolated_points+i, 0);
      for (i = 0; i < 3; ++i)
         yac_enddef_interpolation_method(interpolation_method_stack[i]);

      //---------------------
      // do the interpolation
      //---------------------

      double input_per_subdomain_raw[1][9] = {{0,1,2,3,4,5,6,7,8}};
      double * input_per_subdomain[1][1] = {{input_per_subdomain_raw[0]}};
      double ** input[1 + 1] = {NULL, input_per_subdomain[0]};
      double output[3][4];

      for (i = 0; i < 3; ++i)
         yac_execute_interpolation_plan_element(interpolation_plan_element_stack[i],
                                                input, output[i]);

      //---------------------------------
      // testing results of interpolation
      //---------------------------------

      struct interpolation_method_data_request data_request_[3][1];
      struct interpolation_method_data_request * data_request[3][1] =
         {{data_request_[0]}, {data_request_[1]}, {data_request_[2]}};

      for (i = 0; i < 3; ++i)
         yac_get_source_exchange_data_interpolation_method(interpolation_method_stack[i],
                                                       data_request[i], 1);

      double ref_output[3][4] = {{0, 6, 2, 8},
                                 {(0+1+3)/3.0, (3+6+7)/3.0, (1+2+5)/3.0, (5+7+8)/3.0},
                                 {(0+1+3+4)/4.0, (3+4+6+7)/4.0,
                                  (1+2+4+5)/4.0, (4+5+7+8)/4.0}};

      for (i = 0; i < 3; ++i)
         for (j = 0; j < 4; ++j)
            if (fabs(output[i][j] - ref_output[i][j]) > tol)
               PUT_ERR("wrong result output buffer a\n");

      //---------------
      // cleanup
      //---------------

      yac_delete_interpolation_method(method_1nn);
      yac_delete_interpolation_method(method_3nn);
      yac_delete_interpolation_method(method_4nn);

      for (j = 0; j < 3; ++j) {
         for (i = 0; i < sizeof(interpolation_plan_element_stack[0]) /
                         sizeof(interpolation_plan_element_stack[0][0]); ++i) {
            yac_free_interpolation_plan_element(interpolation_plan_element_stack[j][i]);
            free(interpolation_plan_element_stack[j][i]);
            delete_interpolation_method_data_request_stack(data_request_[j], 1);
         }
      }

      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);

      yac_free_domain(&l_domain_p1);
      yac_free_domain(&l_domain_p2);

      yac_free_points(&points_p1);

      yac_delete_grid(grid_p1);
      yac_delete_grid(grid_p2);

      yac_free_communicator(comm_p1);
      yac_free_communicator(comm_p2);

      yac_free_communicator(comm_named_p1);
      yac_free_communicator(comm_named_p2);
      
      yac_reset_interpolation_method();
   }

   { // interpolation on a single process (1 src and 1 tgt)
     // identical grids for source and target
     // distance weighted nnn
     // 1, 3 and 4 nearest neighbours per target point

      //--------------------
      // set up global search
      //--------------------

         // one directional global search with two domains on two processes with one subdomain each

         // comm worlds on both processes
         struct communicator * comm_p1, * comm_p2;

         comm_p1 = yac_local_comm_new();
         comm_p2 = yac_local_comm_new();

         // named comms for the domains on both processes
         struct communicator * comm_named_p1, * comm_named_p2;

         comm_named_p1 = yac_named_local_comm_new(comm_p1, "P1");
         comm_named_p2 = yac_named_local_comm_new(comm_p2, "P2");

         // define a domain on P1

         struct grid * grid_p1;
         unsigned num_cells[2] = {2,2};
         unsigned cyclic[2] = {0,0};
         double coordinates_x[3] = {-1*rad,0,1*rad};
         double coordinates_y[3] = {-1*rad,0,1*rad};

         grid_p1 = yac_reg2d_grid_new(coordinates_x, coordinates_y,
                                      num_cells, cyclic);

         struct subdomain subdomain_p1;

         int global_corner_ids_p1[9] = {0,1,2,3,4,5,6,7,8};
         unsigned corner_is_on_rank_p1[9] = {-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_p1[4] = {0,1,2,3};
         unsigned cell_is_on_rank_p1[4] = {-1,-1,-1,-1};
         int global_edge_ids_p1[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned edge_is_on_rank_p1[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_p1, grid_p1, comm_named_p1,
                                     global_corner_ids_p1, corner_is_on_rank_p1,
                                     global_cell_ids_p1, cell_is_on_rank_p1,
                                     global_edge_ids_p1, edge_is_on_rank_p1);

         struct domain l_domain_p1;

         struct subdomain * blocks_p1[1] = {&subdomain_p1};

         yac_init_domain(&l_domain_p1, comm_named_p1, blocks_p1, 1, "domain_p1");

         // define a domain on P2, which matches the one on P1

         struct grid * grid_p2;

         grid_p2 = yac_reg2d_grid_new(coordinates_x, coordinates_y,
                                      num_cells, cyclic);

         struct subdomain subdomain_p2;

         int global_corner_ids_p2[9] = {0,1,2,3,4,5,6,7,8};
         unsigned corner_is_on_rank_p2[9] = {-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_p2[4] = {0,1,2,3};
         unsigned cell_is_on_rank_p2[4] = {-1,-1,-1,-1};
         int global_edge_ids_p2[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned edge_is_on_rank_p2[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_p2, grid_p2, comm_named_p2,
                                     global_corner_ids_p2, corner_is_on_rank_p2,
                                     global_cell_ids_p2, cell_is_on_rank_p2,
                                     global_edge_ids_p2, edge_is_on_rank_p2);

         struct domain l_domain_p2;

         struct subdomain * blocks_p2[1] = {&subdomain_p2};

         yac_init_domain(&l_domain_p2, comm_named_p2, blocks_p2, 1, "domain_p2");

         // finish domain definition
         yac_enddef_domain(&l_domain_p1);
         yac_enddef_domain(&l_domain_p2);

         // init global search on both processes

         struct global_search * search[2];

         search[0] = yac_global_search_new(&l_domain_p1, "domain_p2", comm_p1,
                                           1,0, "component_p1", "component_p2");
         search[1] = yac_global_search_new(&l_domain_p2, "domain_p1", comm_p2,
                                           0,1, "component_p2", "component_p1");

         // finish initialisation of global search

         yac_enddef_global_search(search[0]);
         yac_enddef_global_search(search[1]);

         // points

         struct points points_p1;

         double cell_coordinates_x[2] = {-0.5, 0.5};
         double cell_coordinates_y[2] = {-0.5, 0.5};

         for ( j = 0; j < 2; ++j ) {
            cell_coordinates_x[j] *= rad;
            cell_coordinates_y[j] *= rad;
         }

         yac_init_points(&points_p1, grid_p1, CELL,
                         cell_coordinates_x, cell_coordinates_y);

         // point set

         struct points * points_per_subdomain[1] = {&points_p1};
         struct points ** points_set[1] = {&(points_per_subdomain[0])};

      //----------------------------------------
      // generation of interpolation method
      //----------------------------------------

      struct interpolation_method * method_1nn, * method_3nn, * method_4nn;

      method_1nn =
        yac_new_interpolation_method_nnn(search[0], &l_domain_p1, "nnn stack",
                                     points_set, NULL, 1, 1, 1, 0);
      method_3nn =
        yac_new_interpolation_method_nnn(search[0], &l_domain_p1, "nnn stack",
                                     points_set, NULL, 1, 1, 2, 0);
      method_4nn =
        yac_new_interpolation_method_nnn(search[0], &l_domain_p1, "nnn stack",
                                     points_set, NULL, 1, 1, 4, 0);

      //----------
      // do search
      //----------

      struct interpolation_method * interpolation_method_stack[3][2] =
         {{method_1nn, NULL}, {method_3nn, NULL}, {method_4nn, NULL}};
      struct interpolation_plan_element * interpolation_plan_element_stack[3][2] =
         {{NULL, NULL}, {NULL, NULL}, {NULL, NULL}};
      int global_ids[3][4] = {{0,1,2,3}, {0,1,2,3}, {0,1,2,3}};
      unsigned num_points = 4;
      unsigned num_interpolated_points[3];
      double coordinates_x_tgt[3][4] = {{-0.5,  0.5, -0.5, 0.5},
                                        {-0.5,  0.5, -0.5, 0.5},
                                        {-0.5,  0.5, -0.5, 0.5}};
      double coordinates_y_tgt[3][4] = {{-0.5, -0.5,  0.5, 0.5},
                                        {-0.5, -0.5,  0.5, 0.5},
                                        {-0.5, -0.5,  0.5, 0.5}};

      for ( i = 0; i < 3; ++i ) {
         for ( j = 0; j < 4; ++j ) {
            coordinates_x_tgt[i][j] *= rad;
            coordinates_y_tgt[i][j] *= rad;
         }
      }

      for (i = 0; i < 3; ++i)
         yac_do_search_interpolation_method(interpolation_method_stack[i],
            interpolation_plan_element_stack[i], global_ids[i], coordinates_x_tgt[i],
            coordinates_y_tgt[i], CELL, num_points, num_interpolated_points+i, 0);
      for (i = 0; i < 3; ++i)
         yac_enddef_interpolation_method(interpolation_method_stack[i]);

      //---------------------
      // do the interpolation
      //---------------------

      double input_per_subdomain_raw[1][4] = {{0,1,2,3}};
      double * input_per_subdomain[1][1] = {{input_per_subdomain_raw[0]}};
      double ** input[1 + 1] = {NULL, input_per_subdomain[0]};
      double output[3][4];

      for (i = 0; i < 3; ++i)
         yac_execute_interpolation_plan_element(interpolation_plan_element_stack[i],
                                                input, output[i]);

      //---------------------------------
      // testing results of interpolation
      //---------------------------------

      struct interpolation_method_data_request data_request_[3][1];
      struct interpolation_method_data_request * data_request[3][1] =
         {{data_request_[0]}, {data_request_[1]}, {data_request_[2]}};

      for (i = 0; i < 3; ++i)
         yac_get_source_exchange_data_interpolation_method(interpolation_method_stack[i],
                                                       data_request[i], 1);

      double ref_output[4] = {0, 1, 2, 3};

      for (i = 0; i < 3; ++i)
         for (j = 0; j < 4; ++j)
            if (fabs(output[i][j] - ref_output[j]) > tol)
               PUT_ERR("wrong result output buffer a\n");

      //---------------
      // cleanup
      //---------------

      yac_delete_interpolation_method(method_1nn);
      yac_delete_interpolation_method(method_3nn);
      yac_delete_interpolation_method(method_4nn);

      for (j = 0; j < 3; ++j) {
         for (i = 0; i < sizeof(interpolation_plan_element_stack[0]) /
                         sizeof(interpolation_plan_element_stack[0][0]); ++i) {
            yac_free_interpolation_plan_element(interpolation_plan_element_stack[j][i]);
            free(interpolation_plan_element_stack[j][i]);
            delete_interpolation_method_data_request_stack(data_request_[j], 1);
         }
      }

      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);

      yac_free_domain(&l_domain_p1);
      yac_free_domain(&l_domain_p2);

      yac_free_points(&points_p1);

      yac_delete_grid(grid_p1);
      yac_delete_grid(grid_p2);

      yac_free_communicator(comm_p1);
      yac_free_communicator(comm_p2);

      yac_free_communicator(comm_named_p1);
      yac_free_communicator(comm_named_p2);

      yac_reset_interpolation_method();
   }

   { // interpolation on a single process (1 src and 1 tgt)
     // single target cell is completely covered by masked source cells

      //--------------------
      // set up global search
      //--------------------

         // one directional global search with two domains on two processes with one subdomain each

         // comm worlds on both processes
         struct communicator * comm_src, * comm_tgt;

         comm_src = yac_local_comm_new();
         comm_tgt = yac_local_comm_new();
         
         // named comms for the domains on both processes
         struct communicator * comm_named_src, * comm_named_tgt;

         comm_named_src = yac_named_local_comm_new(comm_src, "SRC");
         comm_named_tgt = yac_named_local_comm_new(comm_tgt, "TGT");

         // define the src domain

         struct grid * grid_src;
         unsigned num_cells_src[2] = {4,2};
         unsigned cyclic[2] = {0,0};
         double coordinates_x_src[5] = {-3*rad,-2*rad,-1*rad,0,1*rad};
         double coordinates_y_src[3] = {-1*rad,0,1*rad};

         grid_src = yac_reg2d_grid_new(coordinates_x_src, coordinates_y_src,
                                       num_cells_src, cyclic);

         struct subdomain subdomain_src;

         int global_corner_ids_src[15] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14};
         unsigned corner_is_on_rank_src[15] =
          {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_src[8] = {0,1,2,3,4,5,6,7};
         unsigned cell_is_on_rank_src[8] = {-1,-1,-1,-1,-1,-1,-1,-1};
         int global_edge_ids_src[22] =
          {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21};
         unsigned edge_is_on_rank_src[22] =
          {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_src, grid_src, comm_named_src,
                                     global_corner_ids_src, corner_is_on_rank_src,
                                     global_cell_ids_src, cell_is_on_rank_src,
                                     global_edge_ids_src, edge_is_on_rank_src);

         struct domain l_domain_src;

         struct subdomain * blocks_src[1] = {&subdomain_src};

         yac_init_domain(&l_domain_src, comm_named_src, blocks_src, 1, "domain_src");

         // define the tgt domain

         struct grid * grid_tgt;
         unsigned num_cells_tgt[2] = {1,1};
         double coordinates_x_tgt[3] = {-0.5*rad,0.5*rad};
         double coordinates_y_tgt[3] = {-0.5*rad,0.5*rad};

         grid_tgt = yac_reg2d_grid_new(coordinates_x_tgt, coordinates_y_tgt,
                                       num_cells_tgt, cyclic);

         struct subdomain subdomain_tgt;

         int global_corner_ids_tgt[4] = {0,1,2,3};
         unsigned corner_is_on_rank_tgt[4] = {-1,-1,-1,-1};
         int global_cell_ids_tgt[1] = {0};
         unsigned cell_is_on_rank_tgt[1] = {-1};
         int global_edge_ids_tgt[4] = {0,1,2,3};
         unsigned edge_is_on_rank_tgt[4] = {-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_tgt, grid_tgt, comm_named_tgt,
                                     global_corner_ids_tgt, corner_is_on_rank_tgt,
                                     global_cell_ids_tgt, cell_is_on_rank_tgt,
                                     global_edge_ids_tgt, edge_is_on_rank_tgt);

         struct domain l_domain_tgt;

         struct subdomain * blocks_tgt[1] = {&subdomain_tgt};

         yac_init_domain(&l_domain_tgt, comm_named_tgt, blocks_tgt, 1, "domain_tgt");

         // finish domain definition
         yac_enddef_domain(&l_domain_src);
         yac_enddef_domain(&l_domain_tgt);

         // init global search on both processes

         struct global_search * search[2];

         search[0] = yac_global_search_new(&l_domain_src, "domain_tgt", comm_src,
                                           1,0, "component_src", "component_tgt");
         search[1] = yac_global_search_new(&l_domain_tgt, "domain_src", comm_tgt,
                                           0,1, "component_tgt", "component_src");

         // finish initialisation of global search

         yac_enddef_global_search(search[0]);
         yac_enddef_global_search(search[1]);

         // points

         struct points points_src;
         double cell_coordinates_x_src[8] = {-2.5*rad,-1.5*rad,-0.5*rad,0.5*rad,
                                             -2.5*rad,-1.5*rad,-0.5*rad,0.5*rad};
         double cell_coordinates_y_src[8] = {-0.5*rad,-0.5*rad,-0.5*rad,-0.5*rad,
                                              0.5*rad, 0.5*rad, 0.5*rad, 0.5*rad};

         yac_init_points(&points_src, grid_src, CELL,
                     cell_coordinates_x_src, cell_coordinates_y_src);

         // point set

         struct points * points_per_subdomain[1] = {&points_src};
         struct points ** points_set[1] = {&(points_per_subdomain[0])};

         // mask

         struct mask mask_src;
         unsigned mask_values[8] = {0,1,1,1,1,1,1,1};

         yac_init_mask(&mask_src, mask_values, &points_src);

         // mask set

         struct mask * mask_per_subdomain[1] = {&mask_src};
         struct mask ** mask_set[1] = {&(mask_per_subdomain[0])};

      //----------------------------------------
      // generation of interpolation method
      //----------------------------------------

      struct interpolation_method * method_1nn;

      method_1nn =
        yac_new_interpolation_method_nnn_mask(search[0], &l_domain_src, "nnn stack",
                                          points_set, mask_set, 1, 0, 1, 0, 1);

      //----------
      // do search
      //----------

      struct interpolation_method * interpolation_method_stack[2] =
         {method_1nn, NULL};
      struct interpolation_plan_element * interpolation_plan_element_stack[2] =
         {NULL, NULL};
      int global_ids[1] = {0};
      unsigned num_points = 1;
      unsigned num_interpolated_points[1];
      double cell_coordinates_x_tgt[1] = {0.0};
      double cell_coordinates_y_tgt[1] = {0.0};

      yac_do_search_interpolation_method(interpolation_method_stack,
        interpolation_plan_element_stack, global_ids, cell_coordinates_x_tgt,
        cell_coordinates_y_tgt, CELL, num_points, num_interpolated_points, 0);
      yac_enddef_interpolation_method(interpolation_method_stack);

      //---------------------
      // do the interpolation
      //---------------------

      double input_per_subdomain_raw[8] = {0,1,2,3,4,5,6,7};
      double * input_per_subdomain[1] = {input_per_subdomain_raw};
      double ** input[1 + 1] = {NULL, input_per_subdomain};
      double output = -1;

      yac_execute_interpolation_plan_element(interpolation_plan_element_stack,
                                             input, &output);

      //---------------------------------
      // testing results of interpolation
      //---------------------------------

      struct interpolation_method_data_request data_request_[1];
      struct interpolation_method_data_request * data_request[1] =
        {data_request_};

      yac_get_source_exchange_data_interpolation_method(
        interpolation_method_stack, data_request, 1);

      double ref_output = 0;

      if (num_interpolated_points[0] != 1) {
        PUT_ERR("wrong number of result in output buffer\n");
      } else if (output != ref_output)
        PUT_ERR("wrong result output buffer\n");

      //---------------
      // cleanup
      //---------------

      yac_delete_interpolation_method(method_1nn);

      for (i = 0; i < sizeof(interpolation_plan_element_stack) /
                      sizeof(interpolation_plan_element_stack[0]); ++i) {
         yac_free_interpolation_plan_element(interpolation_plan_element_stack[i]);
         free(interpolation_plan_element_stack[i]);
         delete_interpolation_method_data_request_stack(data_request_, 1);
      }

      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);

      yac_free_domain(&l_domain_src);
      yac_free_domain(&l_domain_tgt);

      yac_free_points(&points_src);
      yac_free_mask(&mask_src);

      yac_delete_grid(grid_src);
      yac_delete_grid(grid_tgt);

      yac_free_communicator(comm_src);
      yac_free_communicator(comm_tgt);

      yac_free_communicator(comm_named_src);
      yac_free_communicator(comm_named_tgt);
      
      yac_reset_interpolation_method();
   }

   { // interpolation on a two processes (2 src and 1 tgt), each source processes
     // has one subdomain
     // 1, 3 and 4 nearest neighbours per target point
     // corner, cell and edge ids for a 4 x 3 grid
     // 15--27--16--28--17--29--18--30--19
     //  |       |       |       |       |
     // 19  08  21  09  23  10  25  11  26
     //  |       |       |       |       |
     // 10--18--11--20--12--22--13--24--14
     //  |       |       |       |       |
     // 10  04  12  05  14  06  16  07  17
     //  |       |       |       |       |
     // 05--09--06--11--07--13--08--15--09
     //  |       |       |       |       |
     // 01  00  03  01  05  02  07  03  08
     //  |       |       |       |       |
     // 00--00--01--02--02--04--03--06--04
     // the domain is distributed among the processes as follows:
     // (index == process)
     //
     // 0--0--0--0--0--1--1--1--1
     // |     |     |     |     |
     // 0  0  0  0  0  1  1  1  1
     // |     |     |     |     |
     // 0--0--0--0--0--1--1--1--1
     // |     |     |     |     |
     // 0  0  0  0  0  1  1  1  1
     // |     |     |     |     |
     // 0--0--0--0--1--1--1--1--1
     // |     |     |     |     |
     // 0  0  0  0  1  1  1  1  1
     // |     |     |     |     |
     // 0--0--0--0--1--1--1--1--1

      //--------------------
      // set up global search
      //--------------------

         // one directional global search with two domains on three processes with one subdomain each

         // comm worlds on both processes
         struct communicator * comm_src[2], * comm_tgt;

         comm_src[0] = yac_local_comm_new();
         comm_src[1] = yac_local_comm_new();
         comm_tgt = yac_local_comm_new();
         
         // named comms for the domains on both processes
         struct communicator * comm_named_src[2], * comm_named_tgt;

         comm_named_src[0] = yac_named_local_comm_new(comm_src[0], "src");
         comm_named_src[1] = yac_named_local_comm_new(comm_src[1], "src");
         comm_named_tgt = yac_named_local_comm_new(comm_tgt, "tgt");

         // define a domain on source side

         struct grid * grid_src[2];
         unsigned num_cells_src[2] = {3,3};
         unsigned cyclic_src[2] = {0,0};
         double coordinates_src_x[2][4] = {{0,1,2,3},{1,2,3,4}};
         double coordinates_src_y[2][4] = {{0,1,2,3},{0,1,2,3}};

         for ( i = 0; i < 2; ++i ) {
            for ( j = 0; j < 4; ++j ) {
               coordinates_src_x[i][j] *= rad;
               coordinates_src_y[i][j] *= rad;
            }
         }

         grid_src[0] = yac_reg2d_grid_new(coordinates_src_x[0], coordinates_src_y[0], num_cells_src, cyclic_src);
         grid_src[1] = yac_reg2d_grid_new(coordinates_src_x[1], coordinates_src_y[1], num_cells_src, cyclic_src);

         struct subdomain subdomain_src[2];

         int global_corner_ids_src[2][16] = {{0,1,2,3, 5,6,7,8, 10,11,12,13, 15,16,17,18},
                                             {1,2,3,4, 6,7,8,9, 11,12,13,14, 16,17,18,19}};
         unsigned corner_is_on_rank_src[2][16] = {{-1,-1,1,1, -1,-1,1,1, -1,-1,-1,1, -1,-1,-1,1},
                                                  {0,-1,-1,-1, 0,-1,-1,-1, 0,0,-1,-1, 0,0,-1,-1}};
         int global_cell_ids_src[2][9] = {{0,1,2, 4,5,6, 8,9,10},{1,2,3, 5,6,7, 9,10,11}};
         unsigned cell_is_on_rank_src[2][9] = {{-1,-1,1, -1,-1,1, -1,-1,1},
                                               {0,-1,-1, 0,-1,-1, 0,-1,-1}};
         int global_edge_ids_src[2][24] = {{0,1,2,3,4,5,7, 9,10,11,12,13,14,16, 18,19,20,21,22,23,25, 27,28,29},
                                           {2,3,4,5,6,7,8, 11,12,13,14,15,16,17, 20,21,22,23,24,25,26, 28,29,30}};
         unsigned edge_is_on_rank_src[2][24] = {{-1,-1,-1,-1,1,1,1, -1,-1,-1,-1,1,-1,1,
                                                 -1,-1,-1,-1,1,-1,1, -1,-1,1},
                                                {0,0,-1,-1,-1,-1,-1, 0,0,-1,0,-1,-1,-1,
                                                 0,0,-1,0,-1,-1,-1, 0,-1,-1}};

         for (i = 0; i < 2; ++i)
            yac_init_subdomain_unstruct(subdomain_src+i, grid_src[i], comm_named_src[i],
                                        global_corner_ids_src[i], corner_is_on_rank_src[i],
                                        global_cell_ids_src[i], cell_is_on_rank_src[i],
                                        global_edge_ids_src[i], edge_is_on_rank_src[i]);

         struct domain domain_src[2];

         struct subdomain * blocks_src[2][1] = {{subdomain_src+0},{subdomain_src+1}};

         yac_init_domain(domain_src+0, comm_named_src[0], blocks_src[0], 1, "domain_src");
         yac_init_domain(domain_src+1, comm_named_src[1], blocks_src[1], 1, "domain_src");

         // define a domain on the target, which matches the one on the source processes

         struct grid * grid_tgt;
         unsigned num_cells_tgt[2] = {4,3};
         unsigned cyclic_tgt[2] = {0,0};
         double coordinates_tgt_x[5] = {0,1*rad,2*rad,3*rad,4*rad};
         double coordinates_tgt_y[4] = {0,1*rad,2*rad,3*rad};

         grid_tgt = yac_reg2d_grid_new(coordinates_tgt_x, coordinates_tgt_y,
                                       num_cells_tgt, cyclic_tgt);

         struct subdomain subdomain_tgt;

         int global_corner_ids_tgt[20] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19};
         unsigned corner_is_on_rank_tgt[20] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
                                               -1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_tgt[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned cell_is_on_rank_tgt[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_edge_ids_tgt[31] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,
                                        17,18,19,20,21,22,23,24,25,26,27,28,29,30};
         unsigned edge_is_on_rank_tgt[31] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
                                             -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_tgt, grid_tgt, comm_named_tgt,
                                     global_corner_ids_tgt, corner_is_on_rank_tgt,
                                     global_cell_ids_tgt, cell_is_on_rank_tgt,
                                     global_edge_ids_tgt, edge_is_on_rank_tgt);

         struct domain domain_tgt;

         struct subdomain * blocks_tgt[1] = {&subdomain_tgt};

         yac_init_domain(&domain_tgt, comm_named_tgt, blocks_tgt, 1, "domain_tgt");

         // finish domain definition
         yac_enddef_domain(domain_src+0);
         yac_enddef_domain(domain_src+1);
         yac_enddef_domain(&domain_tgt);

         // init global search on both processes

         struct global_search * search[3];

         search[0] = yac_global_search_new(domain_src+0, "domain_tgt", comm_src[0],
                                           1,0, "component_src", "component_tgt");
         search[1] = yac_global_search_new(domain_src+1, "domain_tgt", comm_src[1],
                                           1,0, "component_src", "component_tgt");
         search[2] = yac_global_search_new(&domain_tgt, "domain_src", comm_tgt,
                                           0,1, "component_tgt", "component_src");

         // finish initialisation of global search

         yac_enddef_global_search(search[0]);
         yac_enddef_global_search(search[1]);
         yac_enddef_global_search(search[2]);

         // points

         struct points points_src[2];

         yac_init_points(points_src+0, grid_src[0], CORNER, coordinates_src_x[0], coordinates_src_y[0]);
         yac_init_points(points_src+1, grid_src[1], CORNER, coordinates_src_x[1], coordinates_src_y[1]);

         // point set

         struct points * points_per_subdomain[2][1] = {{points_src+0},{points_src+1}};
         struct points ** points_set[2][1] = {{points_per_subdomain[0]},
                                              {points_per_subdomain[1]}};

      //----------------------------------------
      // generation of interpolation method
      //----------------------------------------

      struct interpolation_method * method_1nn[2], * method_3nn[2], * method_4nn[2];

      for (i = 0; i < 2; ++i) {
         method_1nn[i] =
          yac_new_interpolation_method_nnn(search[i], domain_src+i, "nnn stack",
                                       points_set[i], NULL, 1, 0, 1, 0);
         method_3nn[i] =
          yac_new_interpolation_method_nnn(search[i], domain_src+i, "nnn stack",
                                       points_set[i], NULL, 1, 0, 3, 0);
         method_4nn[i] =
          yac_new_interpolation_method_nnn(search[i], domain_src+i, "nnn stack",
                                       points_set[i], NULL, 1, 0, 4, 0);
      }

      //----------
      // do search
      //----------

      struct interpolation_method * interpolation_method_stack[2][3][2] =
         {{{method_1nn[0], NULL}, {method_3nn[0], NULL}, {method_4nn[0], NULL}},
          {{method_1nn[1], NULL}, {method_3nn[1], NULL}, {method_4nn[1], NULL}}};
      struct interpolation_plan_element * interpolation_plan_element_stack[2][3][2] =
         {{{NULL, NULL}, {NULL, NULL}, {NULL, NULL}},
          {{NULL, NULL}, {NULL, NULL}, {NULL, NULL}}};
      int global_ids[2][3][12] = {{{0,1,2,3,4,5,6,7,8,9,10,11},
                                   {0,1,2,3,4,5,6,7,8,9,10,11},
                                   {0,1,2,3,4,5,6,7,8,9,10,11}},
                                  {{0,1,2,3,4,5,6,7,8,9,10,11},
                                   {0,1,2,3,4,5,6,7,8,9,10,11},
                                   {0,1,2,3,4,5,6,7,8,9,10,11}}};
      unsigned num_points = 12;
      unsigned num_interpolated_points[2][3];
      double coordinates_x_tgt[2][3][12] = {{{0.75, 1.75, 2.75, 3.75,
                                              0.25, 1.25, 2.25, 3.25,
                                              0.75, 1.75, 2.75, 3.75},
                                             {0.75, 1.75, 2.75, 3.75,
                                              0.25, 1.25, 2.25, 3.25,
                                              0.75, 1.75, 2.75, 3.75},
                                             {0.75, 1.75, 2.75, 3.75,
                                              0.25, 1.25, 2.25, 3.25,
                                              0.75, 1.75, 2.75, 3.75}},
                                            {{0.75, 1.75, 2.75, 3.75,
                                              0.25, 1.25, 2.25, 3.25,
                                              0.75, 1.75, 2.75, 3.75},
                                             {0.75, 1.75, 2.75, 3.75,
                                              0.25, 1.25, 2.25, 3.25,
                                              0.75, 1.75, 2.75, 3.75},
                                             {0.75, 1.75, 2.75, 3.75,
                                              0.25, 1.25, 2.25, 3.25,
                                              0.75, 1.75, 2.75, 3.75}}};
      double coordinates_y_tgt[2][3][12] = {{{0.25, 0.25, 0.25, 0.25,
                                              1.25, 1.25, 1.25, 1.25,
                                              2.25, 2.25, 2.25, 2.25},
                                             {0.25, 0.25, 0.25, 0.25,
                                              1.25, 1.25, 1.25, 1.25,
                                              2.25, 2.25, 2.25, 2.25},
                                             {0.25, 0.25, 0.25, 0.25,
                                              1.25, 1.25, 1.25, 1.25,
                                              2.25, 2.25, 2.25, 2.25}},
                                            {{0.25, 0.25, 0.25, 0.25,
                                              1.25, 1.25, 1.25, 1.25,
                                              2.25, 2.25, 2.25, 2.25},
                                             {0.25, 0.25, 0.25, 0.25,
                                              1.25, 1.25, 1.25, 1.25,
                                              2.25, 2.25, 2.25, 2.25},
                                             {0.25, 0.25, 0.25, 0.25,
                                              1.25, 1.25, 1.25, 1.25,
                                              2.25, 2.25, 2.25, 2.25}}};

      for ( i = 0; i < 2; ++i ) {
         for ( j = 0; j < 3; ++j ) {
            for ( k = 0; k < 12; ++k ) {
               coordinates_x_tgt[i][j][k] *= rad;
               coordinates_y_tgt[i][j][k] *= rad;
            }
         }
      }
      
      for (j = 0; j < 3; ++j)
         for (i = 0; i < 2; ++i)
            yac_do_search_interpolation_method(interpolation_method_stack[i][j],
               interpolation_plan_element_stack[i][j], global_ids[i][j],
               coordinates_x_tgt[i][j], coordinates_y_tgt[i][j], CELL, num_points,
               num_interpolated_points[i]+j, 0);
      for (j = 0; j < 3; ++j)
         for (i = 0; i < 2; ++i)
            yac_enddef_interpolation_method(interpolation_method_stack[i][j]);

      //------------------------------------
      // get information on requested points
      //------------------------------------

      struct interpolation_method_data_request data_request_[2][3];
      struct interpolation_method_data_request * data_request[2][3] =
         {{data_request_[0]+0, data_request_[0]+1, data_request_[0]+2},
          {data_request_[1]+0, data_request_[1]+1, data_request_[1]+2}};

      for (i = 0; i < 2; ++i)
         for (j = 0; j < 3; ++j)
            yac_get_source_exchange_data_interpolation_method(interpolation_method_stack[i][j],
                                                          data_request[i]+j, 1);

      //---------------------
      // do the interpolation
      //---------------------

      double input_per_subdomain_raw[2][16] = {{0,1,-1,-1,5,6,-1,-1,10,11,12,-1,15,16,17,-1},
                                               {-1,2,3,4,-1,7,8,9,-1,-1,13,14,-1,-1,18,19}};
      double * input_per_subdomain[2] = {input_per_subdomain_raw[0],
                                         input_per_subdomain_raw[1]};
      double input_remote_data_raw[2][3][16];
      double * input_remote_data[2][3] = {{input_remote_data_raw[0][0],
                                           input_remote_data_raw[0][1],
                                           input_remote_data_raw[0][2]},
                                          {input_remote_data_raw[1][0],
                                           input_remote_data_raw[1][1],
                                           input_remote_data_raw[1][2]}};
      double ** input[2][3][1 + 1] = {{{input_remote_data[0]+0, input_per_subdomain+0},
                                       {input_remote_data[0]+1, input_per_subdomain+0},
                                       {input_remote_data[0]+2, input_per_subdomain+0}},
                                      {{input_remote_data[1]+0, input_per_subdomain+1},
                                       {input_remote_data[1]+1, input_per_subdomain+1},
                                       {input_remote_data[1]+2, input_per_subdomain+1}}};
      double output[2][3][9];

      // set requested remote data
      for (i = 0; i < 2; ++i)
         for (j = 0; j < 3; ++j)
            for (k = 0; k < data_request_[i][j].num_values; ++k)
               input_remote_data[i][j][k] = data_request_[i][j].global_ids[k];

      for (i = 0; i < 2; ++i)
         for (j = 0; j < 3; ++j)
            yac_execute_interpolation_plan_element(interpolation_plan_element_stack[i][j],
                                                   input[i][j], output[i][j]);

      //---------------------------------
      // testing results of interpolation
      //---------------------------------

      unsigned ref_num_interpolated_points[2][3] = {{6,6,6},{6,6,6}};
      int ref_interpolated_global_ids[2][3][6] = {{{0,1,4,5,8,9},
                                                   {0,1,4,5,8,9},
                                                   {0,1,4,5,8,9}},
                                                  {{2,3,6,7,10,11},
                                                   {2,3,6,7,10,11},
                                                   {2,3,6,7,10,11}}};
      double ref_output[2][3][6] = {{{1,2,5,6,11,12},
                                     {(0+1+6)/3.0,(1+2+7)/3.0,
                                      (10+5+6)/3.0,(11+6+7)/3.0,
                                      (10+11+16)/3.0,(11+12+17)/3.0},
                                     {(0+1+5+6)/4.0,(1+2+6+7)/4.0,
                                      (5+6+10+11)/4.0,(6+7+11+12)/4.0,
                                      (10+11+15+16)/4.0,(11+12+16+17)/4.0}},
                                    {{3,4,7,8,13,14},
                                     {(2+3+8)/3.0,(3+4+9)/3.0,
                                      (12+7+8)/3.0,(13+8+9)/3.0,
                                      (12+13+18)/3.0,(13+14+19)/3.0},
                                     {(2+3+7+8)/4.0,(3+4+8+9)/4.0,
                                      (7+8+12+13)/4.0,(8+9+13+14)/4.0,
                                      (12+13+17+18)/4.0,(13+14+18+19)/4.0}}};

      for (i = 0; i < 2; ++i)
         for (j = 0; j < 3; ++j)
            if (ref_num_interpolated_points[i][j] != num_interpolated_points[i][j])
               PUT_ERR("wrong num of interpolated points\n");

      for (i = 0; i < 2; ++i)
         for (j = 0; j < 3; ++j)
            for (k = 0; k < num_interpolated_points[i][j]; ++k)
               if (ref_interpolated_global_ids[i][j][k] != global_ids[i][j][k])
                  PUT_ERR("wrong global ids interpolated by a process\n");

      for (i = 0; i < 2; ++i)
         for (j = 0; j < 3; ++j)
            for (k = 0; k < num_interpolated_points[i][j]; ++k)
               if (fabs(ref_output[i][j][k] - output[i][j][k]) > tol)
                  PUT_ERR("wrong result output buffer b\n");

      //---------------
      // cleanup
      //---------------

      for (i = 0; i < 2; ++i) {
         yac_delete_interpolation_method(method_1nn[i]);
         yac_delete_interpolation_method(method_3nn[i]);
         yac_delete_interpolation_method(method_4nn[i]);
      }

      for (k = 0; k < 2; ++k) {
         for (j = 0; j < 3; ++j) {
            for (i = 0; i < sizeof(interpolation_plan_element_stack[0][0]) /
                            sizeof(interpolation_plan_element_stack[0][0][0]); ++i) {
               yac_free_interpolation_plan_element(interpolation_plan_element_stack[k][j][i]);
               free(interpolation_plan_element_stack[k][j][i]);
            }
            delete_interpolation_method_data_request_stack(data_request[k][j], 1);
         }
      }

      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);
      yac_global_search_delete(search[2]);

      yac_free_domain(domain_src+0);
      yac_free_domain(domain_src+1);
      yac_free_domain(&domain_tgt);

      yac_free_points(points_src+0);
      yac_free_points(points_src+1);

      yac_delete_grid(grid_src[0]);
      yac_delete_grid(grid_src[1]);
      yac_delete_grid(grid_tgt);

      yac_free_communicator(comm_src[0]);
      yac_free_communicator(comm_src[1]);
      yac_free_communicator(comm_tgt);

      yac_free_communicator(comm_named_src[0]);
      yac_free_communicator(comm_named_src[1]);
      yac_free_communicator(comm_named_tgt);

      yac_reset_interpolation_method();
   }

   { // interpolation on a two processes (2 src and 1 tgt), each source processes
     // has one subdomain
     // 1, 3 and 4 nearest neighbours per target point
     // corner, cell and edge ids for a 4 x 3 grid
     // 15--27--16--28--17--29--18--30--19
     //  |       |       |       |       |
     // 19  08  21  09  23  10  25  11  26
     //  |       |       |       |       |
     // 10--18--11--20--12--22--13--24--14
     //  |       |       |       |       |
     // 10  04  12  05  14  06  16  07  17
     //  |       |       |       |       |
     // 05--09--06--11--07--13--08--15--09
     //  |       |       |       |       |
     // 01  00  03  01  05  02  07  03  08
     //  |       |       |       |       |
     // 00--00--01--02--02--04--03--06--04
     // the domain is distributed among the processes as follows:
     // (index == process)
     //
     // 0--0--0--0--0--1--1--1--1
     // |     |     |     |     |
     // 0  0  0  0  0  1  1  1  1
     // |     |     |     |     |
     // 0--0--0--0--0--1--1--1--1
     // |     |     |     |     |
     // 0  0  0  0  0  1  1  1  1
     // |     |     |     |     |
     // 0--0--0--0--1--1--1--1--1
     // |     |     |     |     |
     // 0  0  0  0  1  1  1  1  1
     // |     |     |     |     |
     // 0--0--0--0--1--1--1--1--1
     //
     // target mask (X == masked):
     //
     // +-----+-----+-----+-----+
     // |     |     |     |     |
     // |  O  |  X  |  X  |  X  |
     // |     |     |     |     |
     // +-----+-----+-----+-----+
     // |     |     |     |     |
     // |  O  |  X  |  X  |  X  |
     // |     |     |     |     |
     // +-----+-----+-----+-----+
     // |     |     |     |     |
     // |  O  |  X  |  X  |  X  |
     // |     |     |     |     |
     // +-----+-----+-----+-----+

      //--------------------
      // set up global search
      //--------------------

         // one directional global search with two domains on three processes with one subdomain each

         // comm worlds on both processes
         struct communicator * comm_src[2], * comm_tgt;

         comm_src[0] = yac_local_comm_new();
         comm_src[1] = yac_local_comm_new();
         comm_tgt = yac_local_comm_new();
         
         // named comms for the domains on both processes
         struct communicator * comm_named_src[2], * comm_named_tgt;

         comm_named_src[0] = yac_named_local_comm_new(comm_src[0], "src");
         comm_named_src[1] = yac_named_local_comm_new(comm_src[1], "src");
         comm_named_tgt = yac_named_local_comm_new(comm_tgt, "tgt");

         // define a domain on source side

         struct grid * grid_src[2];
         unsigned num_cells_src[2] = {3,3};
         unsigned cyclic_src[2] = {0,0};
         double coordinates_src_x[2][4] = {{0,1,2,3},{1,2,3,4}};
         double coordinates_src_y[2][4] = {{0,1,2,3},{0,1,2,3}};

         for ( i = 0; i < 2; ++i ) {
            for ( j = 0; j < 4; ++j ) {
               coordinates_src_x[i][j] *= rad;
               coordinates_src_y[i][j] *= rad;
            }
         }

         grid_src[0] = yac_reg2d_grid_new(coordinates_src_x[0], coordinates_src_y[0], num_cells_src, cyclic_src);
         grid_src[1] = yac_reg2d_grid_new(coordinates_src_x[1], coordinates_src_y[1], num_cells_src, cyclic_src);

         struct subdomain subdomain_src[2];

         int global_corner_ids_src[2][16] = {{0,1,2,3, 5,6,7,8, 10,11,12,13, 15,16,17,18},
                                             {1,2,3,4, 6,7,8,9, 11,12,13,14, 16,17,18,19}};
         unsigned corner_is_on_rank_src[2][16] = {{-1,-1,1,1, -1,-1,1,1, -1,-1,-1,1, -1,-1,-1,1},
                                                  {0,-1,-1,-1, 0,-1,-1,-1, 0,0,-1,-1, 0,0,-1,-1}};
         int global_cell_ids_src[2][9] = {{0,1,2, 4,5,6, 8,9,10},{1,2,3, 5,6,7, 9,10,11}};
         unsigned cell_is_on_rank_src[2][9] = {{-1,-1,1, -1,-1,1, -1,-1,1},
                                               {0,-1,-1, 0,-1,-1, 0,-1,-1}};
         int global_edge_ids_src[2][24] = {{0,1,2,3,4,5,7, 9,10,11,12,13,14,16, 18,19,20,21,22,23,25, 27,28,29},
                                           {2,3,4,5,6,7,8, 11,12,13,14,15,16,17, 20,21,22,23,24,25,26, 28,29,30}};
         unsigned edge_is_on_rank_src[2][24] = {{-1,-1,-1,-1,1,1,1, -1,-1,-1,-1,1,-1,1,
                                                 -1,-1,-1,-1,1,-1,1, -1,-1,1},
                                                {0,0,-1,-1,-1,-1,-1, 0,0,-1,0,-1,-1,-1,
                                                 0,0,-1,0,-1,-1,-1, 0,-1,-1}};

         for (i = 0; i < 2; ++i)
            yac_init_subdomain_unstruct(subdomain_src+i, grid_src[i], comm_named_src[i],
                                        global_corner_ids_src[i], corner_is_on_rank_src[i],
                                        global_cell_ids_src[i], cell_is_on_rank_src[i],
                                        global_edge_ids_src[i], edge_is_on_rank_src[i]);

         struct domain domain_src[2];

         struct subdomain * blocks_src[2][1] = {{subdomain_src+0},{subdomain_src+1}};

         yac_init_domain(domain_src+0, comm_named_src[0], blocks_src[0], 1, "domain_src");
         yac_init_domain(domain_src+1, comm_named_src[1], blocks_src[1], 1, "domain_src");

         // define a domain on the target, which matches the one on the source processes

         struct grid * grid_tgt;
         unsigned num_cells_tgt[2] = {4,3};
         unsigned cyclic_tgt[2] = {0,0};
         double coordinates_tgt_x[5] = {0,1*rad,2*rad,3*rad,4*rad};
         double coordinates_tgt_y[4] = {0,1*rad,2*rad,3*rad};

         grid_tgt = yac_reg2d_grid_new(coordinates_tgt_x, coordinates_tgt_y,
                                       num_cells_tgt, cyclic_tgt);

         struct subdomain subdomain_tgt;

         int global_corner_ids_tgt[20] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19};
         unsigned corner_is_on_rank_tgt[20] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
                                               -1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_tgt[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned cell_is_on_rank_tgt[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_edge_ids_tgt[31] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,
                                        17,18,19,20,21,22,23,24,25,26,27,28,29,30};
         unsigned edge_is_on_rank_tgt[31] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
                                             -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_tgt, grid_tgt, comm_named_tgt,
                                     global_corner_ids_tgt, corner_is_on_rank_tgt,
                                     global_cell_ids_tgt, cell_is_on_rank_tgt,
                                     global_edge_ids_tgt, edge_is_on_rank_tgt);

         struct domain domain_tgt;

         struct subdomain * blocks_tgt[1] = {&subdomain_tgt};

         yac_init_domain(&domain_tgt, comm_named_tgt, blocks_tgt, 1, "domain_tgt");

         // finish domain definition
         yac_enddef_domain(domain_src+0);
         yac_enddef_domain(domain_src+1);
         yac_enddef_domain(&domain_tgt);

         // init global search on both processes

         struct global_search * search[3];

         search[0] = yac_global_search_new(domain_src+0, "domain_tgt", comm_src[0],
                                           1,0, "component_src", "component_tgt");
         search[1] = yac_global_search_new(domain_src+1, "domain_tgt", comm_src[1],
                                           1,0, "component_src", "component_tgt");
         search[2] = yac_global_search_new(&domain_tgt, "domain_src", comm_tgt,
                                           0,1, "component_tgt", "component_src");

         // finish initialisation of global search

         yac_enddef_global_search(search[0]);
         yac_enddef_global_search(search[1]);
         yac_enddef_global_search(search[2]);

         // points

         struct points points_src[2];

         yac_init_points(points_src+0, grid_src[0], CORNER, coordinates_src_x[0], coordinates_src_y[0]);
         yac_init_points(points_src+1, grid_src[1], CORNER, coordinates_src_x[1], coordinates_src_y[1]);

         // point sets

         struct points * points_per_subdomain[2][1] = {{points_src+0},{points_src+1}};
         struct points ** points_set[2][1] = {{points_per_subdomain[0]},
                                              {points_per_subdomain[1]}};

      //----------------------------------------
      // generation of interpolation method
      //----------------------------------------

      struct interpolation_method * method_1nn[2], * method_3nn[2], * method_4nn[2];

      for (i = 0; i < 2; ++i) {
         method_1nn[i] =
          yac_new_interpolation_method_nnn(search[i], domain_src+i, "nnn stack",
                                       points_set[i], NULL, 1, 0, 1, 0);
         method_3nn[i] =
          yac_new_interpolation_method_nnn(search[i], domain_src+i, "nnn stack",
                                       points_set[i], NULL, 1, 0, 3, 0);
         method_4nn[i] =
          yac_new_interpolation_method_nnn(search[i], domain_src+i, "nnn stack",
                                       points_set[i], NULL, 1, 0, 4, 0);
      }

      //----------
      // do search
      //----------

      struct interpolation_method * interpolation_method_stack[2][3][2] =
         {{{method_1nn[0], NULL}, {method_3nn[0], NULL}, {method_4nn[0], NULL}},
          {{method_1nn[1], NULL}, {method_3nn[1], NULL}, {method_4nn[1], NULL}}};
      struct interpolation_plan_element * interpolation_plan_element_stack[2][3][2] =
         {{{NULL, NULL}, {NULL, NULL}, {NULL, NULL}},
          {{NULL, NULL}, {NULL, NULL}, {NULL, NULL}}};
      int global_ids[2][3][12] = {{{0,4,8}, {0,4,8}, {0,4,8}},
                                  {{-1}, {-1}, {-1}}};
      unsigned num_points[2] = {3,0};
      unsigned num_interpolated_points[2][3];
      double coordinates_x_tgt[2][3][3] = {{{0.75, 0.25, 0.75},
                                            {0.75, 0.25, 0.75},
                                            {0.75, 0.25, 0.75}},
                                           {{-1}, {-1}, {-1}}};
      double coordinates_y_tgt[2][3][3] = {{{0.25, 1.25, 2.25},
                                            {0.25, 1.25, 2.25},
                                            {0.25, 1.25, 2.25}},
                                           {{-1}, {-1}, {-1}}};

      for ( i = 0; i < 2; ++i ) {
         for ( j = 0; j < 3; ++j ) {
            for ( k = 0; k < 3; ++k ) {
               coordinates_x_tgt[i][j][k] *= rad;
               coordinates_y_tgt[i][j][k] *= rad;
            }
         }
      }
      
      for (j = 0; j < 3; ++j)
         for (i = 0; i < 2; ++i)
            yac_do_search_interpolation_method(interpolation_method_stack[i][j],
               interpolation_plan_element_stack[i][j], global_ids[i][j],
               (num_points[i]>0)?coordinates_x_tgt[i][j]:NULL,
               (num_points[i]>0)?coordinates_y_tgt[i][j]:NULL, CELL,
               num_points[i], num_interpolated_points[i]+j, 0);
      for (j = 0; j < 3; ++j)
         for (i = 0; i < 2; ++i)
            yac_enddef_interpolation_method(interpolation_method_stack[i][j]);

      //------------------------------------
      // get information on requested points
      //------------------------------------

      struct interpolation_method_data_request data_request_[2][3];
      struct interpolation_method_data_request * data_request[2][3] =
         {{data_request_[0]+0, data_request_[0]+1, data_request_[0]+2},
          {data_request_[1]+0, data_request_[1]+1, data_request_[1]+2}};

      for (i = 0; i < 2; ++i)
         for (j = 0; j < 3; ++j)
            yac_get_source_exchange_data_interpolation_method(interpolation_method_stack[i][j],
                                                          data_request[i]+j, 1);

      //---------------------
      // do the interpolation
      //---------------------

      double input_per_subdomain_raw[2][16] = {{0,1,-1,-1,5,6,-1,-1,10,11,12,-1,15,16,17,-1},
                                               {-1,2,3,4,-1,7,8,9,-1,-1,13,14,-1,-1,18,19}};
      double * input_per_subdomain[2] = {input_per_subdomain_raw[0],
                                         input_per_subdomain_raw[1]};
      double input_remote_data_raw[2][3][16];
      double * input_remote_data[2][3] = {{input_remote_data_raw[0][0],
                                           input_remote_data_raw[0][1],
                                           input_remote_data_raw[0][2]},
                                          {input_remote_data_raw[0][0],
                                           input_remote_data_raw[0][1],
                                           input_remote_data_raw[0][2]}};
      double ** input[2][3][1 + 1] = {{{input_remote_data[0]+0, input_per_subdomain+0},
                                       {input_remote_data[0]+1, input_per_subdomain+0},
                                       {input_remote_data[0]+2, input_per_subdomain+0}},
                                      {{input_remote_data[1]+0, input_per_subdomain+1},
                                       {input_remote_data[1]+1, input_per_subdomain+1},
                                       {input_remote_data[1]+2, input_per_subdomain+1}}};
      double output[2][3][3];

      for (i = 0; i < 2; ++i)
        for (j = 0; j < 3; ++j)
          for (k = 0; k < 3; ++k)
            output[i][j][k] = -1;

      // set requested remote data
      for (i = 0; i < 2; ++i)
         for (j = 0; j < 3; ++j)
            for (k = 0; k < data_request_[i][j].num_values; ++k)
               input_remote_data[i][j][k] = data_request_[i][j].global_ids[k];

      for (i = 0; i < 2; ++i)
         for (j = 0; j < 3; ++j)
            yac_execute_interpolation_plan_element(interpolation_plan_element_stack[i][j],
                                                   input[i][j], output[i][j]);

      //---------------------------------
      // testing results of interpolation
      //---------------------------------

      unsigned ref_num_interpolated_points[2][3] = {{3,3,3},{0,0,0}};
      int ref_interpolated_global_ids[2][3][3] = {{{0,4,8},
                                                   {0,4,8},
                                                   {0,4,8}},
                                                  {{-1},
                                                   {-1},
                                                   {-1}}};
      double ref_output[2][3][3] = {{{1,5,11},
                                     {(0+1+6)/3.0,(10+5+6)/3.0,(10+11+16)/3.0},
                                     {(0+1+5+6)/4.0,(5+6+10+11)/4.0,(10+11+15+16)/4.0}},
                                    {{-1}, {-1}, {-1}}};

      for (i = 0; i < 2; ++i)
         for (j = 0; j < 3; ++j)
            if (ref_num_interpolated_points[i][j] != num_interpolated_points[i][j])
               PUT_ERR("wrong num of interpolated points\n");

      for (i = 0; i < 2; ++i)
         for (j = 0; j < 3; ++j)
            for (k = 0; k < num_interpolated_points[i][j]; ++k)
               if (ref_interpolated_global_ids[i][j][k] != global_ids[i][j][k])
                  PUT_ERR("wrong global ids interpolated by a process\n");

      for (i = 0; i < 2; ++i)
         for (j = 0; j < 3; ++j)
            for (k = 0; k < num_interpolated_points[i][j]; ++k)
               if (fabs(ref_output[i][j][k] - output[i][j][k]) > tol)
                  PUT_ERR("wrong result output buffer b\n");

      //---------------
      // cleanup
      //---------------

      for (i = 0; i < 2; ++i) {
         yac_delete_interpolation_method(method_1nn[i]);
         yac_delete_interpolation_method(method_3nn[i]);
         yac_delete_interpolation_method(method_4nn[i]);
      }

      for (k = 0; k < 2; ++k) {
         for (j = 0; j < 3; ++j) {
            for (i = 0; i < sizeof(interpolation_plan_element_stack[0][0]) /
                            sizeof(interpolation_plan_element_stack[0][0][0]); ++i) {
               yac_free_interpolation_plan_element(interpolation_plan_element_stack[k][j][i]);
               free(interpolation_plan_element_stack[k][j][i]);
            }
            delete_interpolation_method_data_request_stack(data_request[k][j], 1);
         }
      }

      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);
      yac_global_search_delete(search[2]);

      yac_free_domain(domain_src+0);
      yac_free_domain(domain_src+1);
      yac_free_domain(&domain_tgt);

      yac_free_points(points_src+0);
      yac_free_points(points_src+1);

      yac_delete_grid(grid_src[0]);
      yac_delete_grid(grid_src[1]);
      yac_delete_grid(grid_tgt);

      yac_free_communicator(comm_src[0]);
      yac_free_communicator(comm_src[1]);
      yac_free_communicator(comm_tgt);

      yac_free_communicator(comm_named_src[0]);
      yac_free_communicator(comm_named_src[1]);
      yac_free_communicator(comm_named_tgt);

      yac_reset_interpolation_method();
   }

   { // interpolation on a two processes (2 src and 1 tgt), one source process has
     // one subdomain and the other two
     // 1, 3 and 4 nearest neighbours per target point (some points lie outside
     // the source domain)
     // corner, cell and edge ids for a 4 x 3 grid
     // 15--27--16--28--17--29--18--30--19
     //  |       |       |       |       |
     // 19  08  21  09  23  10  25  11  26
     //  |       |       |       |       |
     // 10--18--11--20--12--22--13--24--14
     //  |       |       |       |       |
     // 10  04  12  05  14  06  16  07  17
     //  |       |       |       |       |
     // 05--09--06--11--07--13--08--15--09
     //  |       |       |       |       |
     // 01  00  03  01  05  02  07  03  08
     //  |       |       |       |       |
     // 00--00--01--02--02--04--03--06--04
     // the domain is distributed among the processes as follows:
     // (index xy => x == subdomain
     //              y == process
     //
     // 00--00--00--00--00--01--01--01--01
     // |        |       |       |       |
     // 00  00  00  00  00  01  01  01  01
     // |        |       |       |       |
     // 00--00--00--00--00--01--01--01--01
     // |        |       |       |       |
     // 00  00  00  00  00  11  11  11  11
     // |        |       |       |       |
     // 00--00--00--00--11--11--11--11--11
     // |        |       |       |       |
     // 00  00  00  00  11  11  11  11  11
     // |        |       |       |       |
     // 00--00--00--00--11--11--11--11--11


      //--------------------
      // set up global search
      //--------------------

         // one directional global search with two domains on three processes with one subdomain each

         // comm worlds on both processes
         struct communicator * comm_src[2], * comm_tgt;

         comm_src[0] = yac_local_comm_new();
         comm_src[1] = yac_local_comm_new();
         comm_tgt = yac_local_comm_new();
         
         // named comms for the domains on both processes
         struct communicator * comm_named_src[2], * comm_named_tgt;

         comm_named_src[0] = yac_named_local_comm_new(comm_src[0], "src");
         comm_named_src[1] = yac_named_local_comm_new(comm_src[1], "src");
         comm_named_tgt = yac_named_local_comm_new(comm_tgt, "tgt");

         // define a domain on source side

         struct grid * grid_src_p0;
         struct grid * grid_src_p1[2];
         unsigned num_cells_src_p0[2] = {3,3};
         unsigned num_cells_src_p1_0[2] = {3,3};
         unsigned num_cells_src_p1_1[2] = {3,2};
         unsigned cyclic_src[2] = {0,0};
         double coordinates_src_x_p0[4] = {0,1*rad,2*rad,3*rad};
         double coordinates_src_y_p0[4] = {0,1*rad,2*rad,3*rad};
         double coordinates_src_x_p1_0[4] = {1*rad,2*rad,3*rad,4*rad};
         double coordinates_src_y_p1_0[4] = {0,1*rad,2*rad,3*rad};
         double coordinates_src_x_p1_1[4] = {1*rad,2*rad,3*rad,4*rad};
         double coordinates_src_y_p1_1[3] = {1*rad,2*rad,3*rad};

         grid_src_p0 = yac_reg2d_grid_new(coordinates_src_x_p0, coordinates_src_y_p0, num_cells_src_p0, cyclic_src);
         grid_src_p1[0] = yac_reg2d_grid_new(coordinates_src_x_p1_0, coordinates_src_y_p1_0, num_cells_src_p1_0, cyclic_src);
         grid_src_p1[1] = yac_reg2d_grid_new(coordinates_src_x_p1_1, coordinates_src_y_p1_1, num_cells_src_p1_1, cyclic_src);

         struct subdomain subdomain_src_p0;
         struct subdomain subdomain_src_p1[2];

         int global_corner_ids_src_p0[16] = {0,1,2,3, 5,6,7,8, 10,11,12,13, 15,16,17,18};
         int global_corner_ids_src_p1_0[16] = {1,2,3,4, 6,7,8,9, 11,12,13,14, 16,17,18,19};
         int global_corner_ids_src_p1_1[12] = {6,7,8,9, 11,12,13,14, 16,17,18,19};
         unsigned corner_is_on_rank_src_p0[16] = {-1,-1,1,1, -1,-1,1,1, -1,-1,-1,1, -1,-1,-1,1};
         unsigned corner_is_on_rank_src_p1_0[16] = {0,-1,-1,-1, 0,-1,-1,-1, 0,0,1,1, 0,0,1,1};
         unsigned corner_is_on_rank_src_p1_1[12] = {0,1,1,1, 0,0,-1,-1, 0,0,-1,-1};
         int global_cell_ids_src_p0[9] = {0,1,2, 4,5,6, 8,9,10};
         int global_cell_ids_src_p1_0[9] = {1,2,3, 5,6,7, 9,10,11};
         int global_cell_ids_src_p1_1[6] = {5,6,7, 9,10,11};
         unsigned cell_is_on_rank_src_p0[9] = {-1,-1,1, -1,-1,1, -1,-1,1};
         unsigned cell_is_on_rank_src_p1_0[9] = {0,-1,-1, 0,-1,-1, 0,1,1};
         unsigned cell_is_on_rank_src_p1_1[6] = {0,1,1, 0,-1,-1};
         int global_edge_ids_src_p0[24] = {0,1,2,3,4,5,7, 9,10,11,12,13,14,16, 18,19,20,21,22,23,25, 27,28,29};
         int global_edge_ids_src_p1_0[24] = {2,3,4,5,6,7,8, 11,12,13,14,15,16,17, 20,21,22,23,24,25,26, 28,29,30};
         int global_edge_ids_src_p1_1[17] = {11,12,13,14,15,16,17, 20,21,22,23,24,25,26, 28,29,30};
         unsigned edge_is_on_rank_src_p0[24] = {-1,-1,-1,-1,1,1,1, -1,-1,-1,-1,1,-1,1, -1,-1,-1,-1,1,-1,1, -1,-1,1};
         unsigned edge_is_on_rank_src_p1_0[24] = {0,0,-1,-1,-1,-1,-1, 0,0,-1,0,-1,-1,-1, 0,0,1,0,1,1,1, 0,1,1};
         unsigned edge_is_on_rank_src_p1_1[17] = {0,0,1,0,1,1,1, 0,0,-1,0,-1,-1,-1, 0,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_src_p0, grid_src_p0, comm_named_src[0],
                                     global_corner_ids_src_p0, corner_is_on_rank_src_p0,
                                     global_cell_ids_src_p0, cell_is_on_rank_src_p0,
                                     global_edge_ids_src_p0, edge_is_on_rank_src_p0);
         yac_init_subdomain_unstruct(subdomain_src_p1+0, grid_src_p1[0], comm_named_src[1],
                                     global_corner_ids_src_p1_0, corner_is_on_rank_src_p1_0,
                                     global_cell_ids_src_p1_0, cell_is_on_rank_src_p1_0,
                                     global_edge_ids_src_p1_0, edge_is_on_rank_src_p1_0);
         yac_init_subdomain_unstruct(subdomain_src_p1+1, grid_src_p1[1], comm_named_src[1],
                                     global_corner_ids_src_p1_1, corner_is_on_rank_src_p1_1,
                                     global_cell_ids_src_p1_1, cell_is_on_rank_src_p1_1,
                                     global_edge_ids_src_p1_1, edge_is_on_rank_src_p1_1);

         struct domain domain_src[2];

         struct subdomain * blocks_src_p0[1] = {&subdomain_src_p0};
         struct subdomain * blocks_src_p1[2] = {subdomain_src_p1+0, subdomain_src_p1+1};

         yac_init_domain(domain_src+0, comm_named_src[0], blocks_src_p0, 1, "domain_src");
         yac_init_domain(domain_src+1, comm_named_src[1], blocks_src_p1, 2, "domain_src");

         // define a domain on the target, which matches the one on the source processes

         struct grid * grid_tgt;
         unsigned num_cells_tgt[2] = {4,3};
         unsigned cyclic_tgt[2] = {0,0};
         double coordinates_tgt_x[5] = {0,1*rad,2*rad,3*rad,4*rad};
         double coordinates_tgt_y[4] = {0,1*rad,2*rad,3*rad};

         grid_tgt = yac_reg2d_grid_new(coordinates_tgt_x, coordinates_tgt_y,
                                       num_cells_tgt, cyclic_tgt);

         struct subdomain subdomain_tgt;

         int global_corner_ids_tgt[20] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19};
         unsigned corner_is_on_rank_tgt[20] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
                                               -1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_tgt[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned cell_is_on_rank_tgt[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_edge_ids_tgt[31] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,
                                        17,18,19,20,21,22,23,24,25,26,27,28,29,30};
         unsigned edge_is_on_rank_tgt[31] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
                                             -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_tgt, grid_tgt, comm_named_tgt,
                                     global_corner_ids_tgt, corner_is_on_rank_tgt,
                                     global_cell_ids_tgt, cell_is_on_rank_tgt,
                                     global_edge_ids_tgt, edge_is_on_rank_tgt);

         struct domain domain_tgt;

         struct subdomain * blocks_tgt[1] = {&subdomain_tgt};

         yac_init_domain(&domain_tgt, comm_named_tgt, blocks_tgt, 1, "domain_tgt");

         // finish domain definition
         yac_enddef_domain(domain_src+0);
         yac_enddef_domain(domain_src+1);
         yac_enddef_domain(&domain_tgt);

         // init global search on both processes

         struct global_search * search[3];

         search[0] = yac_global_search_new(domain_src+0, "domain_tgt", comm_src[0],
                                           1,0, "component_src", "component_tgt");
         search[1] = yac_global_search_new(domain_src+1, "domain_tgt", comm_src[1],
                                           1,0, "component_src", "component_tgt");
         search[2] = yac_global_search_new(&domain_tgt, "domain_src", comm_tgt,
                                           0,1, "component_tgt", "component_src");

         // finish initialisation of global search

         yac_enddef_global_search(search[0]);
         yac_enddef_global_search(search[1]);
         yac_enddef_global_search(search[2]);

         // points

         struct points points_src_p0;
         struct points points_src_p1[2];

         yac_init_points(&points_src_p0, grid_src_p0, CORNER, coordinates_src_x_p0, coordinates_src_y_p0);
         yac_init_points(points_src_p1+0, grid_src_p1[0], CORNER, coordinates_src_x_p1_0, coordinates_src_y_p1_0);
         yac_init_points(points_src_p1+1, grid_src_p1[1], CORNER, coordinates_src_x_p1_1, coordinates_src_y_p1_1);

         // point sets

         struct points * points_per_subdomain_p0[1] = {&points_src_p0};
         struct points * points_per_subdomain_p1[2][1] = {{points_src_p1+0},
                                                          {points_src_p1+1}};
         struct points ** points_set[2][2] = {{points_per_subdomain_p0, NULL},
                                              {points_per_subdomain_p1[0],
                                               points_per_subdomain_p1[1]}};

      //----------------------------------------
      // generation of interpolation method
      //----------------------------------------

      struct interpolation_method * method_1nn[2], * method_3nn[2], * method_21nn[2];
      struct interpolation_method * method_fixed[3][2];

      for (int i = 0; i < 3; ++i) {
        for (int j = 0; j < 2; ++j) {
          method_fixed[i][j] =
            yac_new_interpolation_method_fixed(1337, search[j]);
        }
      }

      for (i = 0; i < 2; ++i) {
         method_1nn[i] =
          yac_new_interpolation_method_nnn(search[i], domain_src+i, "nnn stack",
                                       points_set[i], NULL, 1, 0, 1, 0);
         method_3nn[i] =
          yac_new_interpolation_method_nnn(search[i], domain_src+i, "nnn stack",
                                       points_set[i], NULL, 1, 0, 3, 0);
         method_21nn[i] =
          yac_new_interpolation_method_nnn(search[i], domain_src+i, "nnn stack",
                                       points_set[i], NULL, 1, 0, 21, 0);
      }

      //----------
      // do search
      //----------

      struct interpolation_method * interpolation_method_stack[2][3][3] =
         {{{method_1nn[0], method_fixed[0][0], NULL},
           {method_3nn[0], method_fixed[1][0], NULL},
           {method_21nn[0], method_fixed[2][0], NULL}},
          {{method_1nn[1], method_fixed[0][1], NULL},
           {method_3nn[1], method_fixed[1][1], NULL},
           {method_21nn[1], method_fixed[2][1], NULL}}};
      struct interpolation_plan_element * interpolation_plan_element_stack[2][3][3] =
         {{{NULL, NULL, NULL}, {NULL, NULL, NULL}, {NULL, NULL, NULL}},
          {{NULL, NULL, NULL}, {NULL, NULL, NULL}, {NULL, NULL, NULL}}};
      int global_ids[2][3][11] = {{{0,1,2,3,4,5,6,7,8,9,10},
                                   {0,1,2,3,4,5,6,7,8,9,10},
                                   {0,1,2,3,4,5,6,7,8,9,10}},
                                  {{0,1,2,3,4,5,6,7,8,9,10},
                                   {0,1,2,3,4,5,6,7,8,9,10},
                                   {0,1,2,3,4,5,6,7,8,9,10}}};
      unsigned num_points = 11;
      unsigned num_interpolated_points[2][3][2];
      double coordinates_x_tgt[2][3][11] = {{{1.753, 2.253, 1.753, 4.253, 2.253,
                                              3.753, 4.253, 2.753, 1.753, 2.253, 2.753},
                                             {1.753, 2.253, 1.753, 4.253, 2.253,
                                              3.753, 4.253, 2.753, 1.753, 2.253, 2.753},
                                             {1.753, 2.253, 1.753, 4.253, 2.253,
                                              3.753, 4.253, 2.753, 1.753, 2.253, 2.753}},
                                            {{1.753, 2.253, 1.753, 4.253, 2.253,
                                              3.753, 4.253, 2.753, 1.753, 2.253, 2.753},
                                             {1.753, 2.253, 1.753, 4.253, 2.253,
                                              3.753, 4.253, 2.753, 1.753, 2.253, 2.753},
                                             {1.753, 2.253, 1.753, 4.253, 2.253,
                                              3.753, 4.253, 2.753, 1.753, 2.253, 2.753}}};
      double coordinates_y_tgt[2][3][11] = {{{-0.25, -0.25, 1.25, 1.25, 1.75, 1.75,
                                              1.75, 2.25, 3.25, 3.25, 3.25},
                                             {-0.25, -0.25, 1.25, 1.25, 1.75, 1.75,
                                              1.75, 2.25, 3.25, 3.25, 3.25},
                                             {-0.25, -0.25, 1.25, 1.25, 1.75, 1.75,
                                              1.75, 2.25, 3.25, 3.25, 3.25}},
                                            {{-0.25, -0.25, 1.25, 1.25, 1.75, 1.75,
                                              1.75, 2.25, 3.25, 3.25, 3.25},
                                             {-0.25, -0.25, 1.25, 1.25, 1.75, 1.75,
                                              1.75, 2.25, 3.25, 3.25, 3.25},
                                             {-0.25, -0.25, 1.25, 1.25, 1.75, 1.75,
                                              1.75, 2.25, 3.25, 3.25, 3.25}}};

      for ( i = 0; i < 2; ++i ) {
         for ( j = 0; j < 3; ++j ) {
            for ( k = 0; k < 11; ++k ) {
               coordinates_x_tgt[i][j][k] *= rad;
               coordinates_y_tgt[i][j][k] *= rad;
            }
         }
      }
      
      for (j = 0; j < 3; ++j)
         for (i = 0; i < 2; ++i)
            yac_do_search_interpolation_method(interpolation_method_stack[i][j],
               interpolation_plan_element_stack[i][j], global_ids[i][j],
               coordinates_x_tgt[i][j], coordinates_y_tgt[i][j], CELL, num_points,
               num_interpolated_points[i][j], 0);
      for (j = 0; j < 3; ++j)
         for (i = 0; i < 2; ++i)
            yac_enddef_interpolation_method(interpolation_method_stack[i][j]);

      //------------------------------------
      // get information on requested points
      //------------------------------------

      struct interpolation_method_data_request data_request_[2][3];
      struct interpolation_method_data_request * data_request[2][3] =
         {{data_request_[0]+0, data_request_[0]+1, data_request_[0]+2},
          {data_request_[1]+0, data_request_[1]+1, data_request_[1]+2}};

      for (i = 0; i < 2; ++i)
         for (j = 0; j < 3; ++j)
            yac_get_source_exchange_data_interpolation_method(interpolation_method_stack[i][j],
                                                          data_request[i]+j, 1);

      //---------------------
      // do the interpolation
      //---------------------

      double input_per_subdomain_raw_p0[16] = {0,1,-1,-1,5,6,-1,-1,10,11,12,-1,15,16,17,-1};
      double input_per_subdomain_raw_p1_0[16] = {-1,2,3,4,-1,7,8,9,-1,-1,-1,-1,-1,-1,-1,-1};
      double input_per_subdomain_raw_p1_1[12] = {-1,-1,-1,-1,-1,-1,13,14,-1,-1,18,19};
      double * input_per_subdomain_p0[1] = {input_per_subdomain_raw_p0};
      double * input_per_subdomain_p1_0[1] = {input_per_subdomain_raw_p1_0};
      double * input_per_subdomain_p1_1[1] = {input_per_subdomain_raw_p1_1};
      double input_remote_data_raw[2][3][16];
      double * input_remote_data[2][3] = {{input_remote_data_raw[0][0],
                                           input_remote_data_raw[0][1],
                                           input_remote_data_raw[0][2]},
                                          {input_remote_data_raw[1][0],
                                           input_remote_data_raw[1][1],
                                           input_remote_data_raw[1][2]}};
      double ** input_p0[3][1 + 1] = {{input_remote_data[0]+0, input_per_subdomain_p0},
                                      {input_remote_data[0]+1, input_per_subdomain_p0},
                                      {input_remote_data[0]+2, input_per_subdomain_p0}};
      double ** input_p1[3][1 + 2] = {{input_remote_data[1]+0, input_per_subdomain_p1_0, input_per_subdomain_p1_1},
                                      {input_remote_data[1]+1, input_per_subdomain_p1_0, input_per_subdomain_p1_1},
                                      {input_remote_data[1]+2, input_per_subdomain_p1_0, input_per_subdomain_p1_1}};
      double output[2][3][11];

      // set requested remote data
      for (i = 0; i < 2; ++i)
         for (j = 0; j < 3; ++j)
            for (k = 0; k < data_request_[i][j].num_values; ++k)
               input_remote_data[i][j][k] = data_request_[i][j].global_ids[k];

      for (i = 0; i < 2; ++i)
         for (j = 0; j < 3; ++j)
            for (k = 0; k < 11; ++k)
               output[i][j][k] = -1;

      for (i = 0; i < 2; ++i) {
         for (j = 0; j < 3; ++j)
            yac_execute_interpolation_plan_element(interpolation_plan_element_stack[0][j],
                                                   input_p0[j], output[0][j]);
         for (j = 0; j < 3; ++j)
            yac_execute_interpolation_plan_element(interpolation_plan_element_stack[1][j],
                                                   input_p1[j], output[1][j]);
      }

      //---------------------------------
      // testing results of interpolation
      //---------------------------------

      unsigned ref_num_interpolated_points[2][3][2] = {{{1, 0},
                                                        {1, 0},
                                                        {0, 0}},
                                                       {{10, 0},
                                                        {10, 0},
                                                        {0, 11}}};
      int ref_interpolated_global_ids[2][3][11] = {{{2}, {2}, {-1}},
                                                   {{0,1,3,4,5,6,7,8,9,10},
                                                    {0,1,3,4,5,6,7,8,9,10},
                                                    {0,1,2,3,4,5,6,7,8,9,10}}};
      double ref_output[2][3][11] = {{{7,-1},
                                      {(6+7+12)/3.0,-1},
                                      {-1}},
                                    {{2,2,9,12,14,14,13,17,17,18},
                                     {(1+2+3)/3.0,(2+3+7)/3.0,
                                      (4+9+14)/3.0,(7+12+13)/3.0,
                                      (9+13+14)/3.0,(9+14+19)/3.0,
                                      (12+13+18)/3.0,(16+17+18)/3.0,
                                      (12+17+18)/3.0,(17+18+19)/3.0},
                                     {1337, 1337, 1337, 1337, 1337, 1337, 1337,
                                      1337, 1337, 1337, 1337}}};

      for (i = 0; i < 2; ++i)
         for (j = 0; j < 3; ++j)
            for (k = 0; k < 2; ++k)
               if (ref_num_interpolated_points[i][j][k] !=
                   num_interpolated_points[i][j][k])
                  PUT_ERR("wrong num of interpolated points\n");

      for (i = 0; i < 2; ++i)
         for (j = 0; j < 3; ++j)
            for (k = 0; k < num_interpolated_points[i][j][0] +
                 num_interpolated_points[i][j][1]; ++k)
               if (ref_interpolated_global_ids[i][j][k] != global_ids[i][j][k])
                  PUT_ERR("wrong global ids interpolated by a process\n");

      for (i = 0; i < 2; ++i)
         for (j = 0; j < 3; ++j)
            for (k = 0; k < num_interpolated_points[i][j][0] +
                 num_interpolated_points[i][j][1]; ++k)
               if (fabs(ref_output[i][j][k] - output[i][j][k]) > tol)
                  PUT_ERR("wrong result output buffer\n");

      //---------------
      // cleanup
      //---------------

      for (i = 0; i < 2; ++i) {
         yac_delete_interpolation_method(method_1nn[i]);
         yac_delete_interpolation_method(method_3nn[i]);
         yac_delete_interpolation_method(method_21nn[i]);
         for (int j = 0; j < 3; ++j)
          yac_delete_interpolation_method(method_fixed[j][i]);
      }

      for (k = 0; k < 2; ++k) {
         for (j = 0; j < 3; ++j) {
            for (i = 0; i < sizeof(interpolation_plan_element_stack[0][0]) /
                            sizeof(interpolation_plan_element_stack[0][0][0]); ++i) {
               yac_free_interpolation_plan_element(interpolation_plan_element_stack[k][j][i]);
               free(interpolation_plan_element_stack[k][j][i]);
            }
            delete_interpolation_method_data_request_stack(data_request[k][j], 1);
         }
      }

      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);
      yac_global_search_delete(search[2]);

      yac_free_domain(domain_src+0);
      yac_free_domain(domain_src+1);
      yac_free_domain(&domain_tgt);

      yac_free_points(&points_src_p0);
      yac_free_points(points_src_p1+0);
      yac_free_points(points_src_p1+1);

      yac_delete_grid(grid_src_p0);
      yac_delete_grid(grid_src_p1[0]);
      yac_delete_grid(grid_src_p1[1]);
      yac_delete_grid(grid_tgt);

      yac_free_communicator(comm_src[0]);
      yac_free_communicator(comm_src[1]);
      yac_free_communicator(comm_tgt);

      yac_free_communicator(comm_named_src[0]);
      yac_free_communicator(comm_named_src[1]);
      yac_free_communicator(comm_named_tgt);

      yac_reset_interpolation_method();
   }

   { // interpolation on a two processes (2 src and 1 tgt), each source processes
     // has one subdomain
     // 1, 3 and 4 nearest neighbours per target point
     // weighting by distance is used
     // corner, cell and edge ids for a 4 x 3 grid
     // 15--27--16--28--17--29--18--30--19
     //  |       |       |       |       |
     // 19  08  21  09  23  10  25  11  26
     //  |       |       |       |       |
     // 10--18--11--20--12--22--13--24--14
     //  |       |       |       |       |
     // 10  04  12  05  14  06  16  07  17
     //  |       |       |       |       |
     // 05--09--06--11--07--13--08--15--09
     //  |       |       |       |       |
     // 01  00  03  01  05  02  07  03  08
     //  |       |       |       |       |
     // 00--00--01--02--02--04--03--06--04
     // the domain is distributed among the processes as follows:
     // (index == process)
     //
     // 0--0--0--0--0--1--1--1--1
     // |     |     |     |     |
     // 0  0  0  0  0  1  1  1  1
     // |     |     |     |     |
     // 0--0--0--0--0--1--1--1--1
     // |     |     |     |     |
     // 0  0  0  0  0  1  1  1  1
     // |     |     |     |     |
     // 0--0--0--0--1--1--1--1--1
     // |     |     |     |     |
     // 0  0  0  0  1  1  1  1  1
     // |     |     |     |     |
     // 0--0--0--0--1--1--1--1--1

      //--------------------
      // set up global search
      //--------------------

         // one directional global search with two domains on three processes with one subdomain each

         // comm worlds on both processes
         struct communicator * comm_src[2], * comm_tgt;

         comm_src[0] = yac_local_comm_new();
         comm_src[1] = yac_local_comm_new();
         comm_tgt = yac_local_comm_new();
         
         // named comms for the domains on both processes
         struct communicator * comm_named_src[2], * comm_named_tgt;

         comm_named_src[0] = yac_named_local_comm_new(comm_src[0], "src");
         comm_named_src[1] = yac_named_local_comm_new(comm_src[1], "src");
         comm_named_tgt = yac_named_local_comm_new(comm_tgt, "tgt");

         // define a domain on source side

         struct grid * grid_src[2];
         unsigned num_cells_src[2] = {3,3};
         unsigned cyclic_src[2] = {0,0};
         double coordinates_src_x[2][4] = {{0,1,2,3},{1,2,3,4}};
         double coordinates_src_y[2][4] = {{0,1,2,3},{0,1,2,3}};

         for ( i = 0; i < 2; ++i ) {
            for ( j = 0; j < 4; ++j ) {
               coordinates_src_x[i][j] *= rad;
               coordinates_src_y[i][j] *= rad;
            }
         }

         grid_src[0] = yac_reg2d_grid_new(coordinates_src_x[0], coordinates_src_y[0], num_cells_src, cyclic_src);
         grid_src[1] = yac_reg2d_grid_new(coordinates_src_x[1], coordinates_src_y[1], num_cells_src, cyclic_src);

         struct subdomain subdomain_src[2];

         int global_corner_ids_src[2][16] = {{0,1,2,3, 5,6,7,8, 10,11,12,13, 15,16,17,18},
                                             {1,2,3,4, 6,7,8,9, 11,12,13,14, 16,17,18,19}};
         unsigned corner_is_on_rank_src[2][16] = {{-1,-1,1,1, -1,-1,1,1, -1,-1,-1,1, -1,-1,-1,1},
                                                  {0,-1,-1,-1, 0,-1,-1,-1, 0,0,-1,-1, 0,0,-1,-1}};
         int global_cell_ids_src[2][9] = {{0,1,2, 4,5,6, 8,9,10},{1,2,3, 5,6,7, 9,10,11}};
         unsigned cell_is_on_rank_src[2][9] = {{-1,-1,1, -1,-1,1, -1,-1,1},
                                               {0,-1,-1, 0,-1,-1, 0,-1,-1}};
         int global_edge_ids_src[2][24] = {{0,1,2,3,4,5,7, 9,10,11,12,13,14,16, 18,19,20,21,22,23,25, 27,28,29},
                                           {2,3,4,5,6,7,8, 11,12,13,14,15,16,17, 20,21,22,23,24,25,26, 28,29,30}};
         unsigned edge_is_on_rank_src[2][24] = {{-1,-1,-1,-1,1,1,1, -1,-1,-1,-1,1,-1,1,
                                                 -1,-1,-1,-1,1,-1,1, -1,-1,1},
                                                {0,0,-1,-1,-1,-1,-1, 0,0,-1,0,-1,-1,-1,
                                                 0,0,-1,0,-1,-1,-1, 0,-1,-1}};

         for (i = 0; i < 2; ++i)
            yac_init_subdomain_unstruct(subdomain_src+i, grid_src[i], comm_named_src[i],
                                        global_corner_ids_src[i], corner_is_on_rank_src[i],
                                        global_cell_ids_src[i], cell_is_on_rank_src[i],
                                        global_edge_ids_src[i], edge_is_on_rank_src[i]);

         struct domain domain_src[2];

         struct subdomain * blocks_src[2][1] = {{subdomain_src+0},{subdomain_src+1}};

         yac_init_domain(domain_src+0, comm_named_src[0], blocks_src[0], 1, "domain_src");
         yac_init_domain(domain_src+1, comm_named_src[1], blocks_src[1], 1, "domain_src");

         // define a domain on the target, which matches the one on the source processes

         struct grid * grid_tgt;
         unsigned num_cells_tgt[2] = {4,3};
         unsigned cyclic_tgt[2] = {0,0};
         double coordinates_tgt_x[5] = {0,1*rad,2*rad,3*rad,4*rad};
         double coordinates_tgt_y[4] = {0,1*rad,2*rad,3*rad};

         grid_tgt = yac_reg2d_grid_new(coordinates_tgt_x, coordinates_tgt_y,
                                       num_cells_tgt, cyclic_tgt);

         struct subdomain subdomain_tgt;

         int global_corner_ids_tgt[20] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19};
         unsigned corner_is_on_rank_tgt[20] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
                                               -1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_cell_ids_tgt[12] = {0,1,2,3,4,5,6,7,8,9,10,11};
         unsigned cell_is_on_rank_tgt[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
         int global_edge_ids_tgt[31] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,
                                        17,18,19,20,21,22,23,24,25,26,27,28,29,30};
         unsigned edge_is_on_rank_tgt[31] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
                                             -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};

         yac_init_subdomain_unstruct(&subdomain_tgt, grid_tgt, comm_named_tgt,
                                     global_corner_ids_tgt, corner_is_on_rank_tgt,
                                     global_cell_ids_tgt, cell_is_on_rank_tgt,
                                     global_edge_ids_tgt, edge_is_on_rank_tgt);

         struct domain domain_tgt;

         struct subdomain * blocks_tgt[1] = {&subdomain_tgt};

         yac_init_domain(&domain_tgt, comm_named_tgt, blocks_tgt, 1, "domain_tgt");

         // finish domain definition
         yac_enddef_domain(domain_src+0);
         yac_enddef_domain(domain_src+1);
         yac_enddef_domain(&domain_tgt);

         // init global search on both processes

         struct global_search * search[3];

         search[0] = yac_global_search_new(domain_src+0, "domain_tgt", comm_src[0],
                                           1,0, "component_src", "component_tgt");
         search[1] = yac_global_search_new(domain_src+1, "domain_tgt", comm_src[1],
                                           1,0, "component_src", "component_tgt");
         search[2] = yac_global_search_new(&domain_tgt, "domain_src", comm_tgt,
                                           0,1, "component_tgt", "component_src");

         // finish initialisation of global search

         yac_enddef_global_search(search[0]);
         yac_enddef_global_search(search[1]);
         yac_enddef_global_search(search[2]);

         // points

         struct points points_src[2];

         yac_init_points(points_src+0, grid_src[0], CORNER, coordinates_src_x[0], coordinates_src_y[0]);
         yac_init_points(points_src+1, grid_src[1], CORNER, coordinates_src_x[1], coordinates_src_y[1]);

         // point set

         struct points * points_per_subdomain[2][1] = {{points_src+0},{points_src+1}};
         struct points ** points_set[2][1] = {{points_per_subdomain[0]},
                                              {points_per_subdomain[1]}};

      //----------------------------------------
      // generation of interpolation method
      //----------------------------------------

      struct interpolation_method * method_1nn[2], * method_3nn[2], * method_4nn[2];

      for (i = 0; i < 2; ++i) {
         method_1nn[i] =
          yac_new_interpolation_method_nnn(search[i], domain_src+i, "nnn stack",
                                       points_set[i], NULL, 1, 1, 1, 0);
         method_3nn[i] =
          yac_new_interpolation_method_nnn(search[i], domain_src+i, "nnn stack",
                                       points_set[i], NULL, 1, 1, 3, 0);
         method_4nn[i] =
          yac_new_interpolation_method_nnn(search[i], domain_src+i, "nnn stack",
                                       points_set[i], NULL, 1, 1, 4, 0);
      }

      //----------
      // do search
      //----------

      struct interpolation_method * interpolation_method_stack[2][3][2] =
         {{{method_1nn[0], NULL}, {method_3nn[0], NULL}, {method_4nn[0], NULL}},
          {{method_1nn[1], NULL}, {method_3nn[1], NULL}, {method_4nn[1], NULL}}};
      struct interpolation_plan_element * interpolation_plan_element_stack[2][3][2] =
         {{{NULL, NULL}, {NULL, NULL}, {NULL, NULL}},
          {{NULL, NULL}, {NULL, NULL}, {NULL, NULL}}};
      int global_ids[2][3][12] = {{{0,1,2,3,4,5,6,7,8,9,10,11},
                                   {0,1,2,3,4,5,6,7,8,9,10,11},
                                   {0,1,2,3,4,5,6,7,8,9,10,11}},
                                  {{0,1,2,3,4,5,6,7,8,9,10,11},
                                   {0,1,2,3,4,5,6,7,8,9,10,11},
                                   {0,1,2,3,4,5,6,7,8,9,10,11}}};
      unsigned num_points = 12;
      unsigned num_interpolated_points[2][3];
      double coordinates_x_tgt[2][3][12] = {{{0.75, 1.75, 2.75, 3.75,
                                              0.25, 1.25, 2.25, 3.25,
                                              0.75, 1.75, 2.75, 3.75},
                                             {0.75, 1.75, 2.75, 3.75,
                                              0.25, 1.25, 2.25, 3.25,
                                              0.75, 1.75, 2.75, 3.75},
                                             {0.75, 1.75, 2.75, 3.75,
                                              0.25, 1.25, 2.25, 3.25,
                                              0.75, 1.75, 2.75, 3.75}},
                                            {{0.75, 1.75, 2.75, 3.75,
                                              0.25, 1.25, 2.25, 3.25,
                                              0.75, 1.75, 2.75, 3.75},
                                             {0.75, 1.75, 2.75, 3.75,
                                              0.25, 1.25, 2.25, 3.25,
                                              0.75, 1.75, 2.75, 3.75},
                                             {0.75, 1.75, 2.75, 3.75,
                                              0.25, 1.25, 2.25, 3.25,
                                              0.75, 1.75, 2.75, 3.75}}};
      double coordinates_y_tgt[2][3][12] = {{{0.25, 0.25, 0.25, 0.25,
                                              1.25, 1.25, 1.25, 1.25,
                                              2.25, 2.25, 2.25, 2.25},
                                             {0.25, 0.25, 0.25, 0.25,
                                              1.25, 1.25, 1.25, 1.25,
                                              2.25, 2.25, 2.25, 2.25},
                                             {0.25, 0.25, 0.25, 0.25,
                                              1.25, 1.25, 1.25, 1.25,
                                              2.25, 2.25, 2.25, 2.25}},
                                            {{0.25, 0.25, 0.25, 0.25,
                                              1.25, 1.25, 1.25, 1.25,
                                              2.25, 2.25, 2.25, 2.25},
                                             {0.25, 0.25, 0.25, 0.25,
                                              1.25, 1.25, 1.25, 1.25,
                                              2.25, 2.25, 2.25, 2.25},
                                             {0.25, 0.25, 0.25, 0.25,
                                              1.25, 1.25, 1.25, 1.25,
                                              2.25, 2.25, 2.25, 2.25}}};

      for ( i = 0; i < 2; ++i ) {
         for ( j = 0; j < 3; ++j ) {
            for ( k = 0; k < 12; ++k ) {
               coordinates_x_tgt[i][j][k] *= rad;
               coordinates_y_tgt[i][j][k] *= rad;
            }
         }
      }
      
      for (j = 0; j < 3; ++j)
         for (i = 0; i < 2; ++i)
            yac_do_search_interpolation_method(interpolation_method_stack[i][j],
               interpolation_plan_element_stack[i][j], global_ids[i][j],
               coordinates_x_tgt[i][j], coordinates_y_tgt[i][j], CELL, num_points,
               num_interpolated_points[i]+j, 0);
      for (j = 0; j < 3; ++j)
         for (i = 0; i < 2; ++i)
            yac_enddef_interpolation_method(interpolation_method_stack[i][j]);

      //------------------------------------
      // get information on requested points
      //------------------------------------

      struct interpolation_method_data_request data_request_[2][3];
      struct interpolation_method_data_request * data_request[2][3] =
         {{data_request_[0]+0, data_request_[0]+1, data_request_[0]+2},
          {data_request_[1]+0, data_request_[1]+1, data_request_[1]+2}};

      for (i = 0; i < 2; ++i)
         for (j = 0; j < 3; ++j)
            yac_get_source_exchange_data_interpolation_method(interpolation_method_stack[i][j],
                                                          data_request[i]+j, 1);

      //---------------------
      // do the interpolation
      //---------------------

      double input_per_subdomain_raw[2][16] = {{0,1,-1,-1,5,6,-1,-1,10,11,12,-1,15,16,17,-1},
                                               {-1,2,3,4,-1,7,8,9,-1,-1,13,14,-1,-1,18,19}};
      double * input_per_subdomain[2] = {input_per_subdomain_raw[0],
                                         input_per_subdomain_raw[1]};
      double input_remote_data_raw[2][3][16];
      double * input_remote_data[2][3] = {{input_remote_data_raw[0][0],
                                           input_remote_data_raw[0][1],
                                           input_remote_data_raw[0][2]},
                                          {input_remote_data_raw[1][0],
                                           input_remote_data_raw[1][1],
                                           input_remote_data_raw[1][2]}};
      double ** input[2][3][1 + 1] = {{{input_remote_data[0]+0, input_per_subdomain+0},
                                       {input_remote_data[0]+1, input_per_subdomain+0},
                                       {input_remote_data[0]+2, input_per_subdomain+0}},
                                      {{input_remote_data[1]+0, input_per_subdomain+1},
                                       {input_remote_data[1]+1, input_per_subdomain+1},
                                       {input_remote_data[1]+2, input_per_subdomain+1}}};
      double output[2][3][9];

      // set requested remote data
      for (i = 0; i < 2; ++i)
         for (j = 0; j < 3; ++j)
            for (k = 0; k < data_request_[i][j].num_values; ++k)
               input_remote_data[i][j][k] = data_request_[i][j].global_ids[k];

      for (i = 0; i < 2; ++i)
         for (j = 0; j < 3; ++j)
            yac_execute_interpolation_plan_element(interpolation_plan_element_stack[i][j],
                                                   input[i][j], output[i][j]);

      //---------------------------------
      // testing results of interpolation
      //---------------------------------

      unsigned ref_num_interpolated_points[2][3] = {{6,6,6},{6,6,6}};
      int ref_interpolated_global_ids[2][3][6] = {{{0,1,4,5,8,9},
                                                   {0,1,4,5,8,9},
                                                   {0,1,4,5,8,9}},
                                                  {{2,3,6,7,10,11},
                                                   {2,3,6,7,10,11},
                                                   {2,3,6,7,10,11}}};
      double ref_output[2][3][6] =
        {{{1,2,5,6,11,12},
          {get_weighted_nnn(points_src+0, (unsigned[]){0,1,5}, (double[]){0,1,6}, 3,
                            coordinates_x_tgt[0][1][0], coordinates_y_tgt[0][1][0]),
           get_weighted_nnn(points_src+0, (unsigned[]){1,2,6}, (double[]){1,2,7}, 3,
                            coordinates_x_tgt[0][1][1], coordinates_y_tgt[0][1][1]),
           get_weighted_nnn(points_src+0, (unsigned[]){8,4,5}, (double[]){10,5,6}, 3,
                            coordinates_x_tgt[0][1][2], coordinates_y_tgt[0][1][2]),
           get_weighted_nnn(points_src+0, (unsigned[]){9,5,6}, (double[]){11,6,7}, 3,
                            coordinates_x_tgt[0][1][3], coordinates_y_tgt[0][1][3]),
           get_weighted_nnn(points_src+0, (unsigned[]){8,9,13}, (double[]){10,11,16}, 3,
                            coordinates_x_tgt[0][1][4], coordinates_y_tgt[0][1][4]),
           get_weighted_nnn(points_src+0, (unsigned[]){9,10,14}, (double[]){11,12,17}, 3,
                            coordinates_x_tgt[0][1][5], coordinates_y_tgt[0][1][5])},
          {get_weighted_nnn(points_src+0, (unsigned[]){0,1,4,5}, (double[]){0,1,5,6}, 4,
                            coordinates_x_tgt[0][2][0], coordinates_y_tgt[0][2][0]),
           get_weighted_nnn(points_src+0, (unsigned[]){1,2,5,6}, (double[]){1,2,6,7}, 4,
                            coordinates_x_tgt[0][2][1], coordinates_y_tgt[0][2][1]),
           get_weighted_nnn(points_src+0, (unsigned[]){4,5,8,9}, (double[]){5,6,10,11}, 4,
                            coordinates_x_tgt[0][2][2], coordinates_y_tgt[0][2][2]),
           get_weighted_nnn(points_src+0, (unsigned[]){5,6,9,10}, (double[]){6,7,11,12}, 4,
                            coordinates_x_tgt[0][2][3], coordinates_y_tgt[0][2][3]),
           get_weighted_nnn(points_src+0, (unsigned[]){8,9,12,13}, (double[]){10,11,15,16}, 4,
                            coordinates_x_tgt[0][2][4], coordinates_y_tgt[0][2][4]),
           get_weighted_nnn(points_src+0, (unsigned[]){9,10,13,14}, (double[]){11,12,16,17}, 4,
                            coordinates_x_tgt[0][2][5], coordinates_y_tgt[0][2][5])}},
         {{3,4,7,8,13,14},
          {get_weighted_nnn(points_src+1, (unsigned[]){1,2,6}, (double[]){2,3,8}, 3,
                            coordinates_x_tgt[1][1][0], coordinates_y_tgt[1][1][0]),
           get_weighted_nnn(points_src+1, (unsigned[]){2,3,7}, (double[]){3,4,9}, 3,
                            coordinates_x_tgt[1][1][1], coordinates_y_tgt[1][1][1]),
           get_weighted_nnn(points_src+1, (unsigned[]){9,5,6}, (double[]){12,7,8}, 3,
                            coordinates_x_tgt[1][1][2], coordinates_y_tgt[1][1][2]),
           get_weighted_nnn(points_src+1, (unsigned[]){10,6,7}, (double[]){13,8,9}, 3,
                            coordinates_x_tgt[1][1][3], coordinates_y_tgt[1][1][3]),
           get_weighted_nnn(points_src+1, (unsigned[]){9,10,14}, (double[]){12,13,18}, 3,
                            coordinates_x_tgt[1][1][4], coordinates_y_tgt[1][1][4]),
           get_weighted_nnn(points_src+1, (unsigned[]){10,11,15}, (double[]){13,14,19}, 3,
                            coordinates_x_tgt[1][1][5], coordinates_y_tgt[1][1][5])},
          {get_weighted_nnn(points_src+1, (unsigned[]){1,2,5,6}, (double[]){2,3,7,8}, 4,
                            coordinates_x_tgt[1][2][0], coordinates_y_tgt[1][2][0]),
           get_weighted_nnn(points_src+1, (unsigned[]){2,3,6,7}, (double[]){3,4,8,9}, 4,
                            coordinates_x_tgt[1][2][1], coordinates_y_tgt[1][2][1]),
           get_weighted_nnn(points_src+1, (unsigned[]){5,6,9,10}, (double[]){7,8,12,13}, 4,
                            coordinates_x_tgt[1][2][2], coordinates_y_tgt[1][2][2]),
           get_weighted_nnn(points_src+1, (unsigned[]){6,7,10,11}, (double[]){8,9,13,14}, 4,
                            coordinates_x_tgt[1][2][3], coordinates_y_tgt[1][2][3]),
           get_weighted_nnn(points_src+1, (unsigned[]){9,10,13,14}, (double[]){12,13,17,18}, 4,
                            coordinates_x_tgt[1][2][4], coordinates_y_tgt[1][2][4]),
           get_weighted_nnn(points_src+1, (unsigned[]){10,11,14,15}, (double[]){13,14,18,19}, 4,
                            coordinates_x_tgt[1][2][5], coordinates_y_tgt[1][2][5])}}};

      for (i = 0; i < 2; ++i)
         for (j = 0; j < 3; ++j)
            if (ref_num_interpolated_points[i][j] != num_interpolated_points[i][j])
               PUT_ERR("wrong num of interpolated points\n");

      for (i = 0; i < 2; ++i)
         for (j = 0; j < 3; ++j)
            for (k = 0; k < num_interpolated_points[i][j]; ++k)
               if (ref_interpolated_global_ids[i][j][k] != global_ids[i][j][k])
                  PUT_ERR("wrong global ids interpolated by a process\n");

      for (i = 0; i < 2; ++i)
         for (j = 0; j < 3; ++j)
            for (k = 0; k < num_interpolated_points[i][j]; ++k)
               if (fabs(ref_output[i][j][k] - output[i][j][k]) > tol)
                  PUT_ERR("wrong result output buffer b\n");

      //---------------
      // cleanup
      //---------------

      for (i = 0; i < 2; ++i) {
         yac_delete_interpolation_method(method_1nn[i]);
         yac_delete_interpolation_method(method_3nn[i]);
         yac_delete_interpolation_method(method_4nn[i]);
      }

      for (k = 0; k < 2; ++k) {
         for (j = 0; j < 3; ++j) {
            for (i = 0; i < sizeof(interpolation_plan_element_stack[0][0]) /
                            sizeof(interpolation_plan_element_stack[0][0][0]); ++i) {
               yac_free_interpolation_plan_element(interpolation_plan_element_stack[k][j][i]);
               free(interpolation_plan_element_stack[k][j][i]);
            }
            delete_interpolation_method_data_request_stack(data_request[k][j], 1);
         }
      }

      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);
      yac_global_search_delete(search[2]);

      yac_free_domain(domain_src+0);
      yac_free_domain(domain_src+1);
      yac_free_domain(&domain_tgt);

      yac_free_points(points_src+0);
      yac_free_points(points_src+1);

      yac_delete_grid(grid_src[0]);
      yac_delete_grid(grid_src[1]);
      yac_delete_grid(grid_tgt);

      yac_free_communicator(comm_src[0]);
      yac_free_communicator(comm_src[1]);
      yac_free_communicator(comm_tgt);

      yac_free_communicator(comm_named_src[0]);
      yac_free_communicator(comm_named_src[1]);
      yac_free_communicator(comm_named_tgt);

      yac_reset_interpolation_method();
   }

   { // interpolation on two processes (1 src and 1 tgt)
     // src and tgt processes both have global regular grids

      //--------------------
      // set up global search
      //--------------------

         // comm worlds on both processes
         struct communicator * comm_src, * comm_tgt;

         comm_src = yac_local_comm_new();
         comm_tgt = yac_local_comm_new();

         // named comms for the domains on both processes
         struct communicator * comm_named_src, * comm_named_tgt;

         comm_named_src = yac_named_local_comm_new(comm_src, "src");
         comm_named_tgt = yac_named_local_comm_new(comm_tgt, "tgt");

         // define a domain on source side

         struct grid * grid_src;
         unsigned num_cells[2] = {20,20};
         unsigned cyclic[2] = {1,0};
         double coordinates_x[20];
         double coordinates_y[21];

         for (unsigned i = 0; i < 20; ++i)
            coordinates_x[i] = (0.0 + (double)i * 18.0) * rad;

         for (unsigned i = 0; i < 21; ++i)
            coordinates_y[i] = (-90.0 + (double)i * 9.0) * rad;

         grid_src = yac_reg2d_grid_new(coordinates_x, coordinates_y,
                                       num_cells, cyclic);

         struct subdomain subdomain_src;

         int global_corner_ids[20*21];
         unsigned corner_is_on_rank[20*21];
         int global_cell_ids[20*20];
         unsigned cell_is_on_rank[20*20];

         for (unsigned i = 0; i < 20*21; ++i) global_corner_ids[i] = i;
         for (unsigned i = 0; i < 20*21; ++i) corner_is_on_rank[i] = -1;
         for (unsigned i = 0; i < 20*20; ++i) global_cell_ids[i] = i;
         for (unsigned i = 0; i < 20*20; ++i) cell_is_on_rank[i] = -1;

         yac_init_subdomain_unstruct(
            &subdomain_src, grid_src, comm_named_src, global_corner_ids,
            corner_is_on_rank, global_cell_ids, cell_is_on_rank, NULL, NULL);

         struct domain domain_src;

         struct subdomain * blocks_src = &subdomain_src;

         yac_init_domain(&domain_src, comm_named_src, &blocks_src, 1,
                         "domain_src");

         // define a domain on the target, which matches the one on the source processes

         struct grid * grid_tgt;

         grid_tgt = yac_reg2d_grid_new(coordinates_x, coordinates_y,
                                       num_cells, cyclic);

         struct subdomain subdomain_tgt;

         yac_init_subdomain_unstruct(
            &subdomain_tgt, grid_tgt, comm_named_tgt, global_corner_ids,
            corner_is_on_rank, global_cell_ids, cell_is_on_rank, NULL, NULL);

         struct domain domain_tgt;

         struct subdomain * blocks_tgt = &subdomain_tgt;

         yac_init_domain(&domain_tgt, comm_named_tgt, &blocks_tgt, 1,
                         "domain_tgt");

         // finish domain definition
         yac_enddef_domain(&domain_src);
         yac_enddef_domain(&domain_tgt);

         // init global search on both processes

         struct global_search * search[2];

         search[0] = yac_global_search_new(&domain_src, "domain_tgt", comm_src,
                                           1,0, "component_src", "component_tgt");
         search[1] = yac_global_search_new(&domain_tgt, "domain_src", comm_tgt,
                                           0,1, "component_tgt", "component_src");

         // finish initialisation of global search

         yac_enddef_global_search(search[0]);
         yac_enddef_global_search(search[1]);

         // points

         double coordinates_cell_x[20];
         double coordinates_cell_y[20];

         for (unsigned i = 0; i < 20; ++i)
            coordinates_cell_x[i] = (9.0 + (double)i * 18.0) * rad;

         for (unsigned i = 0; i < 20; ++i)
            coordinates_cell_y[i] = (-85.5 + (double)i * 9.0) * rad;

         struct points points_src;
         yac_init_points(&points_src, grid_src, CELL, coordinates_cell_x,
                         coordinates_cell_y);

         // point set

         struct points * points_per_subdomain = &points_src;
         struct points ** points_set = &points_per_subdomain;

      //----------------------------------------
      // generation of interpolation method
      //----------------------------------------

      struct interpolation_method * method_1nn;

      method_1nn =
         yac_new_interpolation_method_nnn(search[0], &domain_src, "nnn stack",
                                          &points_set, NULL, 1, 1, 1, 0);

      //----------
      // do search
      //----------

      struct interpolation_method * interpolation_method_stack[2] =
         {method_1nn, NULL};
      struct interpolation_plan_element * interpolation_plan_element_stack[2] =
         {NULL, NULL};
      int global_ids[20*20];
      double coordinates_cell_x_tgt[20*20];
      double coordinates_cell_y_tgt[20*20];
      unsigned num_points = 20*20;
      unsigned num_interpolated_points;

      for (unsigned i = 0; i < 20*20; ++i) {
         global_ids[i] = i;
         coordinates_cell_x_tgt[i] = coordinates_cell_x[i%20];
         coordinates_cell_y_tgt[i] = coordinates_cell_y[i/20];
      }

      yac_do_search_interpolation_method(interpolation_method_stack,
         interpolation_plan_element_stack, global_ids,
         coordinates_cell_x_tgt, coordinates_cell_y_tgt, CELL, num_points,
         &num_interpolated_points, 0);
      yac_enddef_interpolation_method(interpolation_method_stack);

      //------------------------------------
      // get information on requested points
      //------------------------------------

      struct interpolation_method_data_request data_request_;
      struct interpolation_method_data_request * data_request = & data_request_;

      yac_get_source_exchange_data_interpolation_method(
         interpolation_method_stack, &data_request, 1);

      //---------------------
      // do the interpolation
      //---------------------

      double input_per_subdomain_raw[20*20];
      double * input_per_subdomain = input_per_subdomain_raw;
      double ** input[2] = {NULL, &input_per_subdomain};
      double output[20*20];

      for (unsigned i = 0; i < 20*20; ++i)
         input_per_subdomain_raw[i] = (double)i;

      yac_execute_interpolation_plan_element(interpolation_plan_element_stack,
                                             input, output);

      //---------------------------------
      // testing results of interpolation
      //---------------------------------
      if (20*20 != num_interpolated_points)
         PUT_ERR("wrong num of interpolated points\n");

      for (k = 0; k < num_interpolated_points; ++k)
         if (k != global_ids[k])
            PUT_ERR("wrong global ids interpolated by a process\n");

      for (k = 0; k < num_interpolated_points; ++k)
         if (fabs((double)k - output[k]) > tol)
            PUT_ERR("wrong result output buffer b\n");

      //---------------
      // cleanup
      //---------------

      yac_delete_interpolation_method(method_1nn);

      yac_free_interpolation_plan_element(interpolation_plan_element_stack[0]);
      free(interpolation_plan_element_stack[0]);
      delete_interpolation_method_data_request_stack(data_request, 1);

      yac_global_search_delete(search[0]);
      yac_global_search_delete(search[1]);

      yac_free_domain(&domain_src);
      yac_free_domain(&domain_tgt);

      yac_free_points(&points_src);;

      yac_delete_grid(grid_src);
      yac_delete_grid(grid_tgt);

      yac_free_communicator(comm_src);
      yac_free_communicator(comm_tgt);

      yac_free_communicator(comm_named_src);
      yac_free_communicator(comm_named_tgt);

      yac_reset_interpolation_method();
   }

   return TEST_EXIT_CODE;
}

double get_weighted_nnn(struct points * src_points, unsigned * src_idx,
                        double * src_values, unsigned num_src_idx,
                        double tgt_coord_x, double tgt_coord_y) {

  double tgt_coord_3d[3];

  LLtoXYZ(tgt_coord_x, tgt_coord_y, tgt_coord_3d);

  double distances[num_src_idx];
  double distances_sum = 0;

  for (unsigned i = 0; i < num_src_idx; ++i) {

    double src_coord_2d[2];
    double src_coord_3d[3];
    yac_get_point_coordinates(src_points, src_idx[i], src_coord_2d);
    LLtoXYZ(src_coord_2d[0], src_coord_2d[1], src_coord_3d);

    distances[i] = 1.0 / get_vector_angle(tgt_coord_3d, src_coord_3d);
    distances_sum += distances[i];
  }

  double result = 0;

  for (unsigned i = 0; i < num_src_idx; ++i)
    result += (distances[i] / distances_sum) * src_values[i];

  return result;
}

void check_data_request(struct interpolation_method_data_request * request,
                        struct interpolation_method_data_request * ref_request,
                        unsigned num_points_per_subdomain) {

   unsigned i, j;

   for (i = 0; i < num_points_per_subdomain; ++i) {

      if (request[i].num_values != ref_request[i].num_values)
         PUT_ERR("wrong number of requested points\n");

      for (j = 0; j < ref_request[i].num_values; ++j) {

         if (request[i].global_ids[j] != ref_request[i].global_ids[j])
            PUT_ERR("wrong global_id requested\n");

         if (request[i].ranks[j] != ref_request[i].ranks[j])
            PUT_ERR("point from wrong rank requested\n");
      }
   }
}

void
delete_interpolation_method_data_request_stack(
   struct interpolation_method_data_request * request,
   unsigned num_points_per_subdomain) {

   if (request != NULL) {

      unsigned i;

      for (i = 0; i < num_points_per_subdomain; ++i) {
         free(request[i].global_ids);
         free(request[i].ranks);
      }
   }
}
