/**
 * @file test_interpolation_method_nnn_parallel.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#ifdef USE_MPI
#include <mpi.h>
#endif

#include "tests.h"

#include "grid.h"
#include "grid_reg2d.h"
#include "communicator.h"
#include "communicator_mpi.h"
#include "domain.h"
#include "global_search.h"
#include "interpolation_method_nnn.h"

double const tol = 1e-7;

void check_data_request(struct interpolation_method_data_request * request,
                        struct interpolation_method_data_request* ref_request,
                        unsigned num_points_per_subdomain);

void target_main(struct communicator * global_comm,
                 struct communicator * target_comm);

void source_main(struct communicator * global_comm,
                 struct communicator * source_comm);

void
delete_interpolation_method_data_request_stack(
   struct interpolation_method_data_request * request,
   unsigned num_points_per_subdomain);

int main (void) {

#ifndef USE_MPI
   return EXIT_SKIP_TEST;
#else

   struct communicator * global_comm;

   global_comm = yac_mpi_comm_new();

   if (yac_get_comm_size(global_comm) == 5)
   {
      int my_global_rank = yac_get_comm_rank(global_comm);

      // split processes into source and target

      unsigned target = my_global_rank < 1;

      struct communicator * comm;

      comm = yac_comm_split(global_comm, target, 0);

      if (target)
         target_main(global_comm, comm);
      else
         source_main(global_comm, comm);

      yac_free_communicator(comm);
   }

   yac_free_communicator(global_comm);

   MPI_Finalize();

   return TEST_EXIT_CODE;
#endif
}

void source_main(struct communicator * global_comm,
                 struct communicator * source_comm) {

   unsigned i, j;

   // 1 and 5 nearest neighbours per target point with fixed value fallback

   // corner and cell ids for a 7 x 7 grid (x == target point position)

   // 56-----57-----58-----59-----60-----61-----62-----63
   //  |     x|     x|     x|     x|     x|     x|     x|
   //  |  42  |  43  |  44  |  45  |  46  |  47  |  48  |
   //  |      |      |      |      |      |      |      |
   // 48-----49-----50-----51-----52-----53-----54-----55
   //  |     x|     x|     x|     x|     x|     x|     x|
   //  |  35  |  36  |  37  |  38  |  39  |  40  |  41  |
   //  |      |      |      |      |      |      |      |
   // 40-----41-----42-----43-----44-----45-----46-----47
   //  |     x|     x|     x|     x|     x|     x|     x|
   //  |  28  |  29  |  30  |  31  |  32  |  33  |  34  |
   //  |      |      |      |      |      |      |      |
   // 32-----33-----34-----35-----36-----37-----38-----39
   //  |     x|     x|     x|     x|     x|     x|     x|
   //  |  21  |  22  |  23  |  24  |  25  |  26  |  27  |
   //  |      |      |      |      |      |      |      |
   // 24-----25-----26-----27-----28-----29-----30-----31
   //  |     x|     x|     x|     x|     x|     x|     x|
   //  |  14  |  15  |  16  |  17  |  18  |  19  |  20  |
   //  |      |      |      |      |      |      |      |
   // 16-----17-----18-----19-----20-----21-----22-----23
   //  |     x|     x|     x|     x|     x|     x|     x|
   //  |  07  |  08  |  09  |  10  |  11  |  12  |  13  |
   //  |      |      |      |      |      |      |      |
   // 08-----09-----10-----11-----12-----13-----14-----15
   //  |     x|     x|     x|     x|     x|     x|     x|
   //  |  00  |  01  |  02  |  03  |  04  |  05  |  06  |
   //  |      |      |      |      |      |      |      |
   // 00-----01-----02-----03-----04-----05-----06-----07
   //
   // the domain is distributed among the processes as follows:
   // (index == process)
   //
   // 3---3---3---3---3---3---3---3
   // | 3 | 3 | 3 | 3 | 3 | 3 | 3 |
   // 3---3---3---3---3---3---3---3
   // | 3 | 3 | 3 | 3 | 3 | 3 | 3 |
   // 3---3---3---1---2---2---3---3
   // | 1 | 1 | 1 | 2 | 2 | 2 | 2 |
   // 1---1---1---2---2---2---2---2
   // | 1 | 1 | 1 | 2 | 2 | 2 | 2 |
   // 1---1---1---1---2---2---2---2
   // | 1 | 1 | 1 | 2 | 2 | 2 | 2 |
   // 1---1---1---0---0---0---2---2
   // | 0 | 0 | 0 | 0 | 0 | 0 | 0 |
   // 0---0---0---0---0---0---0---0
   // | 0 | 0 | 0 | 0 | 0 | 0 | 0 |
   // 0---0---0---0---0---0---0---0
   //
   // the mask looks as follows (# == masked point)
   //
   // +---+---+---+---+---+---+---#
   // |   |   |   |   |   |   |   |
   // +---+---+---+---+---+---#---+
   // |   |   |   |   |   |   |   |
   // +---+---+---+---+---#---+---+
   // |   |   |   |   |   |   |   |
   // +---+---+---+---#---+---+---+
   // |   |   |   |   |   |   |   |
   // +---+---+---#---+---+---+---+
   // |   |   |   |   |   |   |   |
   // +---+---#---+---+---+---+---+
   // |   |   |   |   |   |   |   |
   // +---#---+---+---+---+---+---+
   // |   |   |   |   |   |   |   |
   // #---+---+---+---+---+---+---+
   //--------------------
   // set up global search
   //--------------------

   int my_source_rank = yac_get_comm_rank(source_comm);

   // define a domain on P1

   struct grid * grid;
   unsigned num_cells[4][2] = {{7,3},{4,5},{5,5},{7,3}};
   unsigned cyclic[2] = {0,0};
   double coordinates_x[4][8] = {{0,1,2,3,4,5,6,7},{0,1,2,3,4},
                                 {2,3,4,5,6,7},{0,1,2,3,4,5,6,7}};
   double coordinates_y[4][8] = {{0,1,2,3},{1,2,3,4,5,6},
                                 {1,2,3,4,5,6},{4,5,6,7}};

   for ( i = 0; i < 4; ++i ) {
      for ( j = 0; j < 8; ++j ) {
         coordinates_x[i][j] *= rad;
         coordinates_y[i][j] *= rad;
      }
   }

   grid = yac_reg2d_grid_new(coordinates_x[my_source_rank], coordinates_y[my_source_rank],
                         num_cells[my_source_rank], cyclic);

   struct subdomain subdomain;

   int global_corner_ids[4][36] = {{ 0, 1, 2, 3, 4, 5, 6, 7,
                                     8, 9,10,11,12,13,14,15,
                                    16,17,18,19,20,21,22,23,
                                    24,25,26,27,28,29,30,31},
                                   { 8, 9,10,11,12,
                                    16,17,18,19,20,
                                    24,25,26,27,28,
                                    32,33,34,35,36,
                                    40,41,42,43,44,
                                    48,49,50,51,52},
                                   {10,11,12,13,14,15,
                                    18,19,20,21,22,23,
                                    26,27,28,29,30,31,
                                    34,35,36,37,38,39,
                                    42,43,44,45,46,47,
                                    50,51,52,53,54,55},
                                   {32,33,34,35,36,37,38,39,
                                    40,41,42,43,44,45,46,47,
                                    48,49,50,51,52,53,54,55,
                                    56,57,58,59,60,61,62,63}};
   unsigned corner_is_on_rank[4][36] = {{-1,-1,-1,-1,-1,-1,-1,-1,
                                         -1,-1,-1,-1,-1,-1,-1,-1,
                                          1, 1, 1,-1,-1,-1, 2, 2,
                                          1, 1, 1, 1, 2, 2, 2, 2},
                                        { 0, 0, 0, 0, 0,
                                         -1,-1,-1, 0, 0,
                                         -1,-1,-1,-1, 2,
                                         -1,-1,-1, 2, 2,
                                          3, 3, 3,-1, 2,
                                          3, 3, 3, 3, 3},
                                        { 0, 0, 0, 0, 0, 0,
                                          1, 0, 0, 0,-1,-1,
                                          1, 1,-1,-1,-1,-1,
                                          1,-1,-1,-1,-1,-1,
                                          3, 1,-1,-1, 3, 3,
                                          3, 3, 3, 3, 3, 3},
                                        { 1, 1, 1, 2, 2, 2, 2, 2,
                                         -1,-1,-1, 1, 2, 2,-1,-1,
                                         -1,-1,-1,-1,-1,-1,-1,-1,
                                         -1,-1,-1,-1,-1,-1,-1,-1}};
   int global_cell_ids[4][25] = {{ 0, 1, 2, 3, 4, 5, 6,
                                   7, 8, 9,10,11,12,13,
                                  14,15,16,17,18,19,20},
                                 { 7, 8, 9,10,
                                  14,15,16,17,
                                  21,22,23,24,
                                  28,29,30,31,
                                  35,36,37,38},
                                 { 9,10,11,12,13,
                                  16,17,18,19,20,
                                  23,24,25,26,27,
                                  30,31,32,33,34,
                                  37,38,39,40,41},
                                 {28,29,30,31,32,33,34,
                                  35,36,37,38,39,40,41,
                                  42,43,44,45,46,47,48}};
   unsigned cell_is_on_rank[4][25] = {{-1,-1,-1,-1,-1,-1,-1,
                                       -1,-1,-1,-1,-1,-1,-1,
                                        1, 1, 1, 2, 2, 2, 2},
                                      { 0, 0, 0, 0,
                                       -1,-1,-1, 2,
                                       -1,-1,-1, 2,
                                       -1,-1,-1, 2,
                                        3, 3, 3, 3},
                                      { 0, 0, 0, 0, 0,
                                        1,-1,-1,-1,-1,
                                        1,-1,-1,-1,-1,
                                        1,-1,-1,-1,-1,
                                        3, 3, 3, 3, 3},
                                      { 1, 1, 1, 2, 2, 2, 2,
                                        3, 3, 3, 3, 3, 3, 3,
                                        3, 3, 3, 3, 3, 3, 3}};

   yac_init_subdomain_unstruct(&subdomain, grid, source_comm,
                               global_corner_ids[my_source_rank],
                               corner_is_on_rank[my_source_rank],
                               global_cell_ids[my_source_rank],
                               cell_is_on_rank[my_source_rank], NULL, NULL);

   struct domain domain;

   struct subdomain * blocks[1] = {&subdomain};

   yac_init_domain(&domain, source_comm, blocks, 1, "source_domain");

   // finish domain definition
   yac_enddef_domain(&domain);

   // init global search on both processes

   struct global_search * search;

   search = yac_global_search_new(&domain, "target_domain", global_comm, 1,0,
                                  "source_component", "target_component");

   // finish initialisation of global search

   yac_enddef_global_search(search);

   // points

   struct points points;

   yac_init_points(&points, grid, CORNER, coordinates_x[my_source_rank],
                   coordinates_y[my_source_rank]);

   // masks

   struct mask mask;

   unsigned mask_values[4][36] = {{1,0,0,0,0,0,0,0,
                                   0,1,0,0,0,0,0,0,
                                   0,0,1,0,0,0,0,0,
                                   0,0,0,1,0,0,0,0},
                                  {0,1,0,0,0,
                                   0,0,1,0,0,
                                   0,0,0,1,0,
                                   0,0,0,0,1,
                                   0,0,0,0,0,
                                   0,0,0,0,0},
                                  {0,0,0,0,0,0,
                                   1,0,0,0,0,0,
                                   0,1,0,0,0,0,
                                   0,0,1,0,0,0,
                                   0,0,0,1,0,0,
                                   0,0,0,0,1,0},
                                  {0,0,0,0,1,0,0,0,
                                   0,0,0,0,0,1,0,0,
                                   0,0,0,0,0,0,1,0,
                                   0,0,0,0,0,0,0,1}};

   yac_init_mask(&mask, mask_values[my_source_rank], &points);

   // point sets

   struct points * points_per_subdomain[1] = {&points};
   struct points ** points_set[1] = {&(points_per_subdomain[0])};
   struct mask * masks_per_subdomain[1] = {&mask};
   struct mask ** masks_set[1] = {&(masks_per_subdomain[0])};

   //----------------------------------------
   // generation of interpolation method
   //----------------------------------------

   struct interpolation_method * method_1nn, * method_5nn;

   method_1nn =
    yac_new_interpolation_method_nnn_mask(search, &domain, "nnn stack", points_set,
                                      masks_set, 1, 0, 1, 0, 1);
   method_5nn =
    yac_new_interpolation_method_nnn_mask(search, &domain, "nnn stack", points_set,
                                      masks_set, 1, 0, 5, 0, 1);

   //----------
   // do search
   //----------

   struct interpolation_method * interpolation_method_stack[2][2] =
      {{method_1nn, NULL}, {method_5nn, NULL}};
   struct interpolation_plan_element * interpolation_plan_element_stack[2][2] =
      {{NULL, NULL}, {NULL, NULL}};
   int global_ids[2][49] = {{ 0, 1, 2, 3, 4, 5, 6,
                              7, 8, 9,10,11,12,13,
                             14,15,16,17,18,19,20,
                             21,22,23,24,25,26,27,
                             28,29,30,31,32,33,34,
                             35,36,37,38,39,40,41,
                             42,43,44,45,46,47,48},
                            { 0, 1, 2, 3, 4, 5, 6,
                              7, 8, 9,10,11,12,13,
                             14,15,16,17,18,19,20,
                             21,22,23,24,25,26,27,
                             28,29,30,31,32,33,34,
                             35,36,37,38,39,40,41,
                             42,43,44,45,46,47,48}};
   unsigned num_points = 49;
   unsigned num_interpolated_points[2];
   double coordinates_x_tgt[2][49] = {{0.75,1.75,2.75,3.75,4.75,5.75,6.75,
                                       0.75,1.75,2.75,3.75,4.75,5.75,6.75,
                                       0.75,1.75,2.75,3.75,4.75,5.75,6.75,
                                       0.75,1.75,2.75,3.75,4.75,5.75,6.75,
                                       0.75,1.75,2.75,3.75,4.75,5.75,6.75,
                                       0.75,1.75,2.75,3.75,4.75,5.75,6.75,
                                       0.75,1.75,2.75,3.75,4.75,5.75,6.75},
                                      {0.75,1.75,2.75,3.75,4.75,5.75,6.75,
                                       0.75,1.75,2.75,3.75,4.75,5.75,6.75,
                                       0.75,1.75,2.75,3.75,4.75,5.75,6.75,
                                       0.75,1.75,2.75,3.75,4.75,5.75,6.75,
                                       0.75,1.75,2.75,3.75,4.75,5.75,6.75,
                                       0.75,1.75,2.75,3.75,4.75,5.75,6.75,
                                       0.75,1.75,2.75,3.75,4.75,5.75,6.75}};
   double coordinates_y_tgt[2][49] = {{0.75,0.75,0.75,0.75,0.75,0.75,0.75,
                                       1.75,1.75,1.75,1.75,1.75,1.75,1.75,
                                       2.75,2.75,2.75,2.75,2.75,2.75,2.75,
                                       3.75,3.75,3.75,3.75,3.75,3.75,3.75,
                                       4.75,4.75,4.75,4.75,4.75,4.75,4.75,
                                       5.75,5.75,5.75,5.75,5.75,5.75,5.75,
                                       6.75,6.75,6.75,6.75,6.75,6.75,6.75},
                                      {0.75,0.75,0.75,0.75,0.75,0.75,0.75,
                                       1.75,1.75,1.75,1.75,1.75,1.75,1.75,
                                       2.75,2.75,2.75,2.75,2.75,2.75,2.75,
                                       3.75,3.75,3.75,3.75,3.75,3.75,3.75,
                                       4.75,4.75,4.75,4.75,4.75,4.75,4.75,
                                       5.75,5.75,5.75,5.75,5.75,5.75,5.75,
                                       6.75,6.75,6.75,6.75,6.75,6.75,6.75}};

   for ( i = 0; i < 2; ++i ) {
      for ( j = 0; j < 49; ++j ) {
         coordinates_x_tgt[i][j] *= rad;
         coordinates_y_tgt[i][j] *= rad;
      }
   }

   for (i = 0; i < 2; ++i)
      yac_do_search_interpolation_method(interpolation_method_stack[i],
         interpolation_plan_element_stack[i], global_ids[i], coordinates_x_tgt[i],
         coordinates_y_tgt[i], CORNER, num_points, num_interpolated_points+i, 0);
   for (i = 0; i < 2; ++i)
      yac_enddef_interpolation_method(interpolation_method_stack[i]);

   //------------------------------------
   // get information on requested points
   //------------------------------------

   struct interpolation_method_data_request data_request_[2];
   struct interpolation_method_data_request * data_request[2][1] =
      {{data_request_+0},{data_request_+1}};

   for (i = 0; i < 2; ++i)
      yac_get_source_exchange_data_interpolation_method(interpolation_method_stack[i],
                                                    data_request[i], 1);

   //---------------------
   // do the interpolation
   //---------------------

   double input_per_subdomain_raw[4][36] = {{ 0, 1, 2, 3, 4, 5, 6, 7,
                                              8, 9,10,11,12,13,14,15,
                                             -1,-1,-1,19,20,21,-1,-1,
                                             -1,-1,-1,-1,-1,-1,-1,-1},
                                            {-1,-1,-1,-1,-1,
                                             16,17,18,-1,-1,
                                             24,25,26,27,-1,
                                             32,33,34,-1,-1,
                                             -1,-1,-1,43,-1,
                                             -1,-1,-1,-1,-1},
                                            {-1,-1,-1,-1,-1,-1,
                                             -1,-1,-1,-1,22,23,
                                             -1,-1,28,29,30,31,
                                             -1,35,36,37,38,39,
                                             -1,-1,44,45,-1,-1,
                                             -1,-1,-1,-1,-1,-1,},
                                            {-1,-1,-1,-1,-1,-1,-1,-1,
                                             40,41,42,-1,-1,-1,46,47,
                                             48,49,50,51,52,53,54,55,
                                             56,57,58,59,60,61,62,63}};
   double * input_per_subdomain[1] = {input_per_subdomain_raw[my_source_rank]};
   double input_remote_data_raw[2][21];
   double * input_remote_data[2] = {input_remote_data_raw[0],input_remote_data_raw[1]};
   double ** input[2][1 + 1] = {{input_remote_data+0, input_per_subdomain},
                                {input_remote_data+1, input_per_subdomain}};
   double output[2][21];

   // set requested remote data
   for (i = 0; i < 2; ++i)
      for (j = 0; j < data_request_[i].num_values; ++j)
         input_remote_data[i][j] = data_request_[i].global_ids[j];

   for (i = 0; i < 2; ++i)
      for (j = 0; j < 21; ++j)
         output[i][j] = -1;

   for (i = 0; i < 2; ++i)
      yac_execute_interpolation_plan_element(interpolation_plan_element_stack[i],
                                             input[i], output[i]);

   //---------------------------------
   // testing results of interpolation
   //---------------------------------

      unsigned ref_num_interpolated_points[4] = {14,9,12,14};
      int ref_interpolated_global_ids[4][14] = {{0,1,2,3,4,5,6,7,8,9,10,11,12,13},
                                                {14,15,16,21,22,23,28,29,30},
                                                {17,18,19,20,24,25,26,27,31,32,33,34},
                                                {35,36,37,38,39,40,41,42,43,44,45,
                                                 46,47,48}};
      double ref_output[4][2][14] = {{{8,10,11,12,13,14,15,17,17,19,20,21,22,23},
                                      {(1+8+10+16+17)/5.0,(1+2+10+11+17)/5.0,
                                       (2+3+10+11+12)/5.0,(3+4+11+12+13)/5.0,
                                       (4+5+12+13+14)/5.0,(5+6+13+14+15)/5.0,
                                       (6+7+14+15+23)/5.0,(8+16+17+24+25)/5.0,
                                       (10+17+19+25+26)/5.0,(10+11+19+20+26)/5.0,
                                       (11+12+19+20+21)/5.0,(12+13+20+21+22)/5.0,
                                       (13+14+21+22+23)/5.0,(14+15+22+23+31)/5.0}},
                                     {{25,26,26,33,34,35,41,42,43},
                                      {(16+17+24+25+26)/5.0,(17+19+25+26+34)/5.0,
                                       (19+20+26+28+35)/5.0,(24+25+32+33+34)/5.0,
                                       (25+26+33+34+35)/5.0,(26+28+34+35+43)/5.0,
                                       (32+33+40+41+42)/5.0,(33+34+41+42+43)/5.0,
                                       (34+35+42+43+44)/5.0}},
                                     {{28,29,30,31,35,37,38,39,44,44,46,47},
                                      {(19+20+21+28+29)/5.0,(20+21+28+29+30)/5.0,
                                       (21+22+29+30+31)/5.0,(22+23+30+31+39)/5.0,
                                       (28+29+35+37+44)/5.0,(28+29+30+37+38)/5.0,
                                       (29+30+37+38+39)/5.0,(30+31+38+39+47)/5.0,
                                       (35+37+43+44+52)/5.0,(37+38+44+46+53)/5.0,
                                       (37+38+39+46+47)/5.0,(38+39+46+47+55)/5.0}},
                                     {{49,50,51,52,53,53,55,57,58,59,60,61,62,62},
                                      {(40+41+48+49+50)/5.0,(41+42+49+50+51)/5.0,
                                       (42+43+50+51+52)/5.0,(43+44+51+52+53)/5.0,
                                       (44+46+52+53+61)/5.0,(46+47+53+55+62)/5.0,
                                       (46+47+53+55+62)/5.0,(48+49+56+57+58)/5.0,
                                       (49+50+57+58+59)/5.0,(50+51+58+59+60)/5.0,
                                       (51+52+59+60+61)/5.0,(52+53+60+61+62)/5.0,
                                       (53+55+60+61+62)/5.0,(47+53+55+61+62)/5.0}}};

      for (i = 0; i < 2; ++i)
            if (ref_num_interpolated_points[my_source_rank] != num_interpolated_points[i])
               PUT_ERR("wrong num of interpolated points\n");

      for (i = 0; i < 2; ++i)
         for (j = 0; j < ref_num_interpolated_points[my_source_rank]; ++j)
            if (ref_interpolated_global_ids[my_source_rank][j] != global_ids[i][j])
               PUT_ERR("wrong global ids interpolated by a process\n");

      // for (i = 0; i < 2; ++i)
         // for (j = 0; j < ref_num_interpolated_points[my_source_rank]; ++j)
            // if (fabs(ref_output[my_source_rank][i][j] - output[i][j]) > tol)
               // PUT_ERR("wrong result output buffer\n");

   //---------------
   // cleanup
   //---------------

   yac_delete_interpolation_method(method_1nn);
   yac_delete_interpolation_method(method_5nn);

   for (j = 0; j < 2; ++j) {
      for (i = 0; i < sizeof(interpolation_plan_element_stack[0]) /
                      sizeof(interpolation_plan_element_stack[0][0]); ++i) {
         yac_free_interpolation_plan_element(interpolation_plan_element_stack[j][i]);
         free(interpolation_plan_element_stack[j][i]);
      }
      delete_interpolation_method_data_request_stack(data_request_+j, 1);
   }

   yac_global_search_delete(search);
   yac_free_domain(&domain);
   yac_free_mask(&mask);
   yac_free_points(&points);
   yac_delete_grid(grid);

   yac_reset_interpolation_method();
}

void target_main(struct communicator * global_comm,
                 struct communicator * target_comm) {

   struct grid * grid;
   unsigned num_cells[2] = {7,7};
   unsigned cyclic[2] = {0,0};
   double coordinates_x[8] = {0,1,2,3,4,5,6,7};
   double coordinates_y[8] = {0,1,2,3,4,5,6,7};

   for ( unsigned i = 0; i < 8; ++i ) {
      coordinates_x[i] *= rad;
      coordinates_y[i] *= rad;
   }

   grid = yac_reg2d_grid_new(coordinates_x, coordinates_y, num_cells, cyclic);

   struct subdomain subdomain;

   int global_corner_ids[64] =
      { 0, 1, 2, 3, 4, 5, 6, 7,
        8, 9,10,11,12,13,14,15,
       16,17,18,19,20,21,22,23,
       24,25,26,27,28,29,30,31,
       32,33,34,35,36,37,38,39,
       40,41,42,43,44,45,46,47,
       48,49,50,51,52,53,54,55,
       56,57,58,59,60,61,62,63};
   unsigned corner_is_on_rank[64] =
      {-1,-1,-1,-1,-1,-1,-1,-1,
       -1,-1,-1,-1,-1,-1,-1,-1,
       -1,-1,-1,-1,-1,-1,-1,-1,
       -1,-1,-1,-1,-1,-1,-1,-1,
       -1,-1,-1,-1,-1,-1,-1,-1,
       -1,-1,-1,-1,-1,-1,-1,-1,
       -1,-1,-1,-1,-1,-1,-1,-1,
       -1,-1,-1,-1,-1,-1,-1,-1};
   int global_cell_ids[49] =
      { 0, 1, 2, 3, 4, 5, 6,
        7, 8, 9,10,11,12,13,
       14,15,16,17,18,19,20,
       21,22,23,24,25,26,27,
       28,29,30,31,32,33,34,
       35,36,37,38,39,40,41,
       42,43,44,45,46,47,48};
   unsigned cell_is_on_rank[49] =
      {-1,-1,-1,-1,-1,-1,-1,
       -1,-1,-1,-1,-1,-1,-1,
       -1,-1,-1,-1,-1,-1,-1,
       -1,-1,-1,-1,-1,-1,-1,
       -1,-1,-1,-1,-1,-1,-1,
       -1,-1,-1,-1,-1,-1,-1,
       -1,-1,-1,-1,-1,-1,-1};

   yac_init_subdomain_unstruct(&subdomain, grid, target_comm,
                               global_corner_ids, corner_is_on_rank,
                               global_cell_ids, cell_is_on_rank, NULL, NULL);

   struct domain domain;
   struct subdomain * subdomains[1] = {&subdomain};

   yac_init_domain(&domain, target_comm, subdomains, 1, "target_domain");
   yac_enddef_domain(&domain);

   struct global_search * search;

   search = yac_global_search_new(&domain, "source_domain", global_comm, 0, 1,
                                  "target_component", "source_component");
   yac_enddef_global_search(search);

   yac_global_search_delete(search);
   yac_free_domain(&domain);
   yac_delete_grid(grid);

   yac_reset_interpolation_method();
}

void check_data_request(struct interpolation_method_data_request * request,
                        struct interpolation_method_data_request * ref_request,
                        unsigned num_points_per_subdomain) {

   unsigned i, j;

   for (i = 0; i < num_points_per_subdomain; ++i) {

      if (request[i].num_values != ref_request[i].num_values)
         PUT_ERR("wrong number of requested points\n");

      for (j = 0; j < ref_request[i].num_values; ++j) {

         if (request[i].global_ids[j] != ref_request[i].global_ids[j])
            PUT_ERR("wrong global_id requested\n");

         if (request[i].ranks[j] != ref_request[i].ranks[j])
            PUT_ERR("point from wrong rank requested\n");
      }
   }
}

void
delete_interpolation_method_data_request_stack(
   struct interpolation_method_data_request * request,
   unsigned num_points_per_subdomain) {

   if (request != NULL) {

      unsigned i;

      for (i = 0; i < num_points_per_subdomain; ++i) {
         free(request[i].global_ids);
         free(request[i].ranks);
      }
   }
}
