/**
 * @file test_interpolation_method_spmap_parallel.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "tests.h"

#include "grid.h"
#include "grid_reg2d.h"
#include "communicator.h"
#include "communicator_mpi.h"
#include "domain.h"
#include "global_search.h"
#include "interpolation_method_spmap.h"
#include "utils.h"

void check_data_request(struct interpolation_method_data_request * request,
                        struct interpolation_method_data_request* ref_request,
                        unsigned num_points_per_subdomain);

void
delete_interpolation_method_data_request_stack(
   struct interpolation_method_data_request * request,
   unsigned num_points_per_subdomain);

void target_main(struct communicator * global_comm,
                 struct communicator * target_comm);

void source_main(struct communicator * global_comm,
                 struct communicator * source_comm);

int main (void) {

#ifndef USE_MPI
   return EXIT_SKIP_TEST;
#else

   struct communicator * global_comm;

   global_comm = yac_mpi_comm_new();

   if (yac_get_comm_size(global_comm) == 8)
   { 
      int my_global_rank = yac_get_comm_rank(global_comm);

      // split processes into source and target

      unsigned target = my_global_rank < 4;

      struct communicator * comm;

      comm = yac_comm_split(global_comm, target, 0);

      if (target)
         target_main(global_comm, comm);
      else
         source_main(global_comm, comm);

      yac_free_communicator(comm);
   } else {
      PUT_ERR("ERROR: wrong number of processes\n");
   }

   yac_free_communicator(global_comm);

   MPI_Finalize();

   return TEST_EXIT_CODE;
#endif
}

void source_main(struct communicator * global_comm,
                 struct communicator * source_comm) {

   unsigned i;


   // corner and cell ids for a 6 x 5 grid (x == target point position)
   //
   // 35-----36-----37-----38-----39-----40-----41
   //  |      |      |      |      |      |      |
   //  |  24  |  25  |  26  |  27  |  28  |  29  |
   //  |      |      |      |      |      |      |
   // 28-----29-----30-----31-----32-----33-----34
   //  |      |      |      |      |      |      |
   //  |  18  |  19  |  20  |  21  |  22  |  23  |
   //  |      |      |      |      |      |      |
   // 21-----22-----23-----24-----25-----26-----27
   //  |      |      |      |      |      |      |
   //  |  12  |  13  |  14  |  15  |  16  |  17  |
   //  |      |      |      |      |      |      |
   // 14-----15-----16-----17-----18-----19-----20
   //  |      |      |      |      |      |      |
   //  |  06  |  07  |  08  |  09  |  10  |  11  |
   //  |      |      |      |      |      |      |
   // 07-----08-----09-----10-----11-----12-----13
   //  |      |      |      |      |      |      |
   //  |  00  |  01  |  02  |  03  |  04  |  05  |
   //  |      |      |      |      |      |      |
   // 00-----01-----02-----03-----04-----05-----06
   //
   // the domain is distributed among the processes as follows:
   // (index == process)
   //
   // +---+---+---+---+---+---+
   // | 2 | 2 | 2 | 3 | 3 | 3 |
   // +---+---+---+---+---+---+
   // | 2 | 2 | 2 | 3 | 3 | 3 |
   // +---+---+---+---+---+---+
   // | 2 | 2 | 2 | 3 | 3 | 3 |
   // +---+---+---+---+---+---+
   // | 0 | 0 | 0 | 1 | 1 | 1 |
   // +---+---+---+---+---+---+
   // | 0 | 0 | 0 | 1 | 1 | 1 |
   // +---+---+---+---+---+---+
   //
   //--------------------
   // set up global search
   //--------------------

   int my_source_rank = yac_get_comm_rank(source_comm);

   // define domains on source

   struct grid * grid;
   unsigned num_cells[4][2] = {{4,3}, {4,3}, {4,4}, {4,4}};
   unsigned cyclic[2] = {0,0};
   double coordinates_x[4][5] = {{-3,-2,-1,0,1},{-1,0,1,2,3},
                                 {-3,-2,-1,0,1},{-1,0,1,2,3}};
   double coordinates_y[4][5] = {{-3,-2,-1,0},{-3,-2,-1,0},
                                 {-2,-1,0,1,2},{-2,-1,0,1,2}};

   for ( unsigned i = 0; i < 4; ++i ) {
      for ( unsigned j = 0; j < 5; ++j ) {
         coordinates_x[i][j] *= rad;
         coordinates_y[i][j] *= rad;
      }
   }

   double coordinates_x_src_points[4][4] = {{-2.5,-1.5,-0.5,0.5},{-0.5,0.5,1.5,2.5},
                                            {-2.5,-1.5,-0.5,0.5},{-0.5,0.5,1.5,2.5}};
   double coordinates_y_src_points[4][4] = {{-2.5,-1.5,-0.5},{-2.5,-1.5,-0.5},
                                            {-1.5,-0.5,0.5,1.5},{-1.5,-0.5,0.5,1.5}};

   for ( unsigned i = 0; i < 4; ++i ) {
      for ( unsigned j = 0; j < 4; ++j ) {
         coordinates_x_src_points[i][j] *= rad;
         coordinates_y_src_points[i][j] *= rad;
      }
   }

   grid = yac_reg2d_grid_new(coordinates_x[my_source_rank],
                             coordinates_y[my_source_rank],
                             num_cells[my_source_rank], cyclic);

   struct subdomain subdomain;

   int global_corner_ids[4][25] = {{0,1,2,3,4, 7,8,9,10,11, 14,15,16,17,18, 21,22,23,24,25},
                                   {2,3,4,5,6, 9,10,11,12,13, 16,17,18,19,20, 23,24,25,26,27},
                                   {7,8,9,10,11, 14,15,16,17,18, 21,22,23,24,25, 28,29,30,31,32, 35,36,37,38,39},
                                   {9,10,11,12,13, 16,17,18,19,20, 23,24,25,26,27, 30,31,32,33,34, 37,38,39,40,41}};
   unsigned corner_is_on_rank[4][25] = {{-1,-1,-1,-1,1, -1,-1,-1,-1,1, -1,-1,-1,-1,1, 2,2,2,2,3},
                                        {0,-1,-1,-1,-1, 0,-1,-1,-1,-1, 0,-1,-1,-1,-1, 2,3,3,3,3},
                                        {0,0,0,0,1, -1,-1,-1,-1,3, -1,-1,-1,-1,3, -1,-1,-1,-1,3, -1,-1,-1,-1,3},
                                        {0,1,1,1,1, 2,-1,-1,-1,-1, 2,-1,-1,-1,-1, 2,-1,-1,-1,-1, 2,-1,-1,-1,-1}};
   int global_cell_ids[4][16] = {{0,1,2,3, 6,7,8,9, 12,13,14,15},
                                 {2,3,4,5, 8,9,10,11, 14,15,16,17},
                                 {6,7,8,9, 12,13,14,15, 18,19,20,21, 24,25,26,27},
                                 {8,9,10,11, 14,15,16,17, 20,21,22,23, 26,27,28,29}};
   unsigned cell_is_on_rank[4][16] = {{-1,-1,-1,1, -1,-1,-1,1, 2,2,2,3},
                                      {0,-1,-1,-1, 0,-1,-1,-1, 2,3,3,3},
                                      {0,0,0,1, -1,-1,-1,3, -1,-1,-1,3, -1,-1,-1,3},
                                      {0,1,1,1, 2,-1,-1,-1, 2,-1,-1,-1, 2,-1,-1,-1}};

   yac_init_subdomain_unstruct(&subdomain, grid, source_comm,
                               global_corner_ids[my_source_rank],
                               corner_is_on_rank[my_source_rank],
                               global_cell_ids[my_source_rank],
                               cell_is_on_rank[my_source_rank],
                               NULL, NULL);

   struct domain l_domain;
   struct subdomain * blocks = &subdomain;

   yac_init_domain(&l_domain, source_comm, &blocks, 1, "domain_src");

   yac_enddef_domain(&l_domain);

   //-------------------------------------
   // do the global search
   //-------------------------------------

   struct global_search * search;

   search = yac_global_search_new(&l_domain, "domain_tgt", global_comm,
                                  1,0, "component_p1", "component_p2");
   yac_enddef_global_search(search);

   //---------------------------------
   // initialise points on source side
   //---------------------------------

   struct points cell_points;

   yac_init_points(&cell_points, grid, CELL,
                   coordinates_x_src_points[my_source_rank],
                   coordinates_y_src_points[my_source_rank]);

   //------------------------------------------
   // set up coupling point sets on source side
   //------------------------------------------

   struct points * points_per_subdomain = &cell_points;
   struct points ** point_sets = &points_per_subdomain;

   //----------------------------------------
   // test generation of interpolation method
   //----------------------------------------

   struct interpolation_method * method_spmap;

   method_spmap =
    yac_new_interpolation_method_spmap(search, &l_domain, "spmap stack",
                                       &point_sets, 1, 0);

   //----------
   // do search
   //----------

   struct interpolation_method * method_stack[2] = {method_spmap, NULL};
   struct interpolation_plan_element * interpolation_plan_element_stack[2] = {NULL, NULL};
   int global_ids[4][20] = {{0,1,2,3,5,6,7,8,10,11,12,13},
                            {1,2,3,4,6,7,8,9,11,12,13,14},
                            {0,1,2,3,5,6,7,8,10,11,12,13,15,16,17,18,20,21,22,23},
                            {1,2,3,4,6,7,8,9,11,12,13,14,16,17,18,19,21,22,23,24}};
   unsigned num_points[4] = {12,12,20,20};
   // define target points (for target domain see below in main_target)

   double coordinates_x_tgt_cell_point[5] = {-2.0,-1.0,0.0,1.0,2.0};
   double coordinates_y_tgt_cell_point[6] = {-2.0,-1.0,0.0,1.0,2.0,3.0};
   double x_coordinates[20];
   double y_coordinates[20];
   unsigned num_interpolated_points[1];

   for (unsigned i = 0; i < num_points[my_source_rank]; ++i) {
      x_coordinates[i] = (coordinates_x_tgt_cell_point[
        global_ids[my_source_rank][i]%5] + 0.1) * rad;
      y_coordinates[i] = (coordinates_y_tgt_cell_point[
        global_ids[my_source_rank][i]/5] + 0.1) * rad;
   }

   yac_do_search_interpolation_method(
    method_stack, interpolation_plan_element_stack, global_ids[my_source_rank],
    x_coordinates, y_coordinates, CELL, num_points[my_source_rank],
    num_interpolated_points, 1);

   yac_enddef_interpolation_method(method_stack);

   //---------------------
   // do the interpolation
   //---------------------

   double in_data_raw[4][2][16] =
    {{{-1},{1,2,3,-1, 7,8,9,-1, -1,-1,-1,-1}},
     {{-1},{-1,4,5,6, -1,10,11,12, -1,-1,-1,-1}},
     {{-1},{-1,-1,-1,-1, 13,14,15,-1, 19,20,21,-1, 25,26,27,-1}},
     {{-1},{-1,-1,-1,-1, -1,16,17,18, -1,22,23,24, -1,28,29,30}}};
   double * remote_data = in_data_raw[my_source_rank][0];
   double * subdomain_in_data = in_data_raw[my_source_rank][1];
   double ** in_data[2] = {&remote_data, &subdomain_in_data};
   double send_buffer[12] = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};

   yac_execute_interpolation_plan_element(interpolation_plan_element_stack,
                                          in_data, send_buffer);

   //---------------------------------
   // testing results of interpolation
   //---------------------------------

   struct interpolation_method_data_request data_request_;
   struct interpolation_method_data_request * data_request = &data_request_;

   yac_get_source_exchange_data_interpolation_method(method_stack, &data_request, 1);

   struct interpolation_method_data_request ref_data_request[4] =
      {{0, NULL, NULL} ,{0, NULL, NULL} ,{0, NULL, NULL} ,{0, NULL, NULL}};

   unsigned ref_num_interpolated_points[4] = {2,3,6,9};
   unsigned ref_global_ids[4][9] = {{0,1}, {2,3,4}, {5,6,10,11,15,16},
                                    {7,8,9,12,13,14,17,18,19}};

   double ref_out_data[4][9] = {{1+2+7+8,3+9}, {4+10,5+11,6+12},
                                {13+14,15,19+20,21,25+26,27},
                                {16,17,18,22,23,24,28,29,30}};

   check_data_request(data_request, &ref_data_request[my_source_rank], 1);

   if (ref_num_interpolated_points[my_source_rank] != num_interpolated_points[0]) {
     PUT_ERR("wrong number of interpolated points\n");
     printf ( "wrong number of points [%d] : %d %d \n", my_source_rank,
              ref_num_interpolated_points[my_source_rank], num_interpolated_points[0]);
   }

   for (i = 0; i < ref_num_interpolated_points[my_source_rank]; ++i) {
     if (ref_global_ids[my_source_rank][i] != global_ids[my_source_rank][i]) {
       PUT_ERR("wrong global id order\n");
       printf ("wrong id order         [%d] : %d %d %d \n", my_source_rank,
               i, ref_global_ids[my_source_rank][i], global_ids[my_source_rank][i]);
     }

     // compare reference data with results obtained via spmap

     // in this test results should be identical, however on some systems it differs
     // in the last bits
     if (send_buffer[i] != ref_out_data[my_source_rank][i]) {
       PUT_ERR("wrong result buffer\n");
       printf ("wrong results global ID    %d %d:\n",
                      ref_global_ids[my_source_rank][i],
                      global_ids[my_source_rank][i]);
       printf ("wrong results           [%d] %d : %.15f %.15f \n",
               my_source_rank, i, send_buffer[i], ref_out_data[my_source_rank][i] );
     }
   }

   //---------------
   // cleanup
   //---------------

   for (i = 0; i < sizeof(method_stack)/sizeof(method_stack[0]); ++i)
      if (method_stack[i] != NULL)
         yac_delete_interpolation_method(method_stack[i]);
   for (i = 0; i < sizeof(interpolation_plan_element_stack) /
        sizeof(interpolation_plan_element_stack[0]); ++i) {
      yac_free_interpolation_plan_element(interpolation_plan_element_stack[i]);
      free(interpolation_plan_element_stack[i]);
   }
   delete_interpolation_method_data_request_stack(data_request, 1);

   yac_free_points(&cell_points);
   yac_global_search_delete(search);
   yac_free_domain(&l_domain);
   yac_delete_grid(grid);

   yac_reset_interpolation_method();
}

void target_main(struct communicator * global_comm,
                 struct communicator * target_comm) {

   // corner and cell ids for a 5 x 6 grid (x == target point position)
   //
   // 36-----37-----38-----39-----40-----41
   //  |      |      |      |      |      |
   //  |  25  |  26  |  27  |  28  |  29  |
   //  |      |      |      |      |      |
   // 30-----31-----32-----33-----34-----35
   //  |      |      |      |      |      |
   //  |  20  |  21  |  22  |  23  |  24  |
   //  |      |      |      |      |      |
   // 24-----25-----26-----27-----28-----29
   //  |      |      |      |      |      |
   //  |  15  |  16  |  17  |  18  |  19  |
   //  |      |      |      |      |      |
   // 18-----19-----20-----21-----22-----23
   //  |      |      |      |      |      |
   //  |  10  |  11  |  12  |  13  |  14  |
   //  |      |      |      |      |      |
   // 12-----13-----14-----15-----16-----17
   //  |      |      |      |      |      |
   //  |  05  |  06  |  07  |  08  |  09  |
   //  |      |      |      |      |      |
   // 06-----07-----08-----09-----10-----11
   //  |      |      |      |      |      |
   //  |  00  |  01  |  02  |  03  |  04  |
   //  |      |      |      |      |      |
   // 00-----01-----02-----03-----04-----05
   //
   // the domain is distributed among the processes as follows:
   // (index == process)
   //
   // +---+---+---+---+---+
   // | 2 | 2 | 3 | 3 | 3 |
   // +---+---+---+---+---+
   // | 2 | 2 | 3 | 3 | 3 |
   // +---+---+---+---+---+
   // | 2 | 2 | 3 | 3 | 3 |
   // +---+---+---+---+---+
   // | 0 | 0 | 0 | 1 | 1 |
   // +---+---+---+---+---+
   // | 0 | 0 | 0 | 1 | 1 |
   // +---+---+---+---+---+
   // | 0 | 0 | 0 | 1 | 1 |
   // +---+---+---+---+---+
   //
   //--------------------
   // set up global search
   //--------------------

   int my_target_rank = yac_get_comm_rank(target_comm);

   // define a domain on the target

   struct grid * grid;
   unsigned num_cells[4][2] = {{4,4},{3,4},{3,4},{4,4}};
   unsigned cyclic[2] = {0,0};
   double coordinates_x[4][5] = {{-2.5,-1.5,-0.5,0.5,1.5},
                                 {-0.5,0.5,1.5,2.5},
                                 {-2.5,-1.5,-0.5,0.5},
                                 {-1.5,-0.5,0.5,1.5,2.5}};
   double coordinates_y[4][5] = {{-2.5,-1.5,-0.5,0.5,1.5},
                                 {-2.5,-1.5,-0.5,0.5,1.5},
                                 {-0.5,0.5,1.5,2.5,3.5},
                                 {-0.5,0.5,1.5,2.5,3.5}};

   for ( unsigned i = 0; i < 4; ++i ) {
      for ( unsigned j = 0; j < 5; ++j ) {
         coordinates_x[i][j] *= rad;
         coordinates_y[i][j] *= rad;
      }
   }

   grid = yac_reg2d_grid_new(coordinates_x[my_target_rank], coordinates_y[my_target_rank],
                             num_cells[my_target_rank], cyclic);

   struct subdomain subdomain;

   int global_corner_ids[4][25] = {{0,1,2,3,4, 6,7,8,9,10, 12,13,14,15,16, 18,19,20,21,22, 24,25,26,27,28},
                                   {2,3,4,5, 8,9,10,11, 14,15,16,17, 20,21,22,23, 26,27,28,29},
                                   {12,13,14,15, 18,19,20,21, 24,25,26,27, 30,31,32,33, 36,37,38,39},
                                   {13,14,15,16,17, 19,20,21,22,23, 15,26,27,28,29, 31,32,33,34,35, 37,38,39,40,41}};
   unsigned corner_is_on_rank[4][25] = {{-1,-1,-1,-1,1, -1,-1,-1,-1,1, -1,-1,-1,-1,1, -1,-1,-1,-1,1, 2,2,2,3,3},
                                        {0,-1,-1,-1, 0,-1,-1,-1, 0,-1,-1,-1, 0,-1,-1,-1, 3,3,3,3},
                                        {0,0,0,0, -1,-1,-1,3, -1,-1,-1,3, -1,-1,-1,3, -1,-1,-1,3},
                                        {0,0,0,1,1, 2,-1,-1,-1,-1, 2,-1,-1,-1,-1, 2,-1,-1,-1,-1, 2,-1,-1,-1,-1}};
   int global_cell_ids[4][16] = {{0,1,2,3, 5,6,7,8, 10,11,12,13, 15,16,17,18},
                                 {2,3,4, 7,8,9, 12,13,14, 17,18,19},
                                 {10,11,12, 15,16,17, 20,21,22, 25,26,27},
                                 {11,12,13,14, 16,17,18,19, 21,22,23,24, 26,27,28,29}};
   unsigned cell_is_on_rank[4][16] = {{-1,-1,-1,1, -1,-1,-1,1, -1,-1,-1,1, 2,2,3,3},
                                      {0,-1,-1, 0,-1,-1, 0,-1,-1, 3,3,3},
                                      {0,0,0, -1,-1,3, -1,-1,3, -1,-1,3},
                                      {0,0,1,1, 2,-1,-1,-1, 2,-1,-1,-1, 2,-1,-1,-1}};

   yac_init_subdomain_unstruct(&subdomain, grid, target_comm,
                               global_corner_ids[my_target_rank],
                               corner_is_on_rank[my_target_rank],
                               global_cell_ids[my_target_rank],
                               cell_is_on_rank[my_target_rank],
                               NULL, NULL);

   struct domain l_domain;

   struct subdomain * blocks = &subdomain;

   yac_init_domain(&l_domain, target_comm, &blocks, 1, "domain_tgt");

   // finish domain definition
   yac_enddef_domain(&l_domain);

   //-------------------------------------
   // init global search on both processes
   //-------------------------------------

   struct global_search * search;

   search = yac_global_search_new(&l_domain, "domain_src", global_comm,
                                  0,1, "component_p2", "component_p1");
   yac_enddef_global_search(search);

   //---------------
   // cleanup
   //---------------

   yac_global_search_delete(search);
   yac_free_domain(&l_domain);
   yac_delete_grid(grid);

   yac_reset_interpolation_method();
}

void
check_data_request(struct interpolation_method_data_request * request,
                   struct interpolation_method_data_request * ref_request,
                   unsigned num_points_per_subdomain) {

   unsigned i, j;

   for (i = 0; i < num_points_per_subdomain; ++i) {

      if (request[i].num_values != ref_request[i].num_values)  {
         PUT_ERR("wrong number of requested points\n");
         printf ("check_data_request subdomain %d : num_values %d != %d\n", i, request[i].num_values, ref_request[i].num_values);
      }

      for (j = 0; j < ref_request[i].num_values; ++j) {

         if (request[i].global_ids[j] != ref_request[i].global_ids[j]) {
            PUT_ERR("wrong global_id requested\n");
            printf ("check_data_request subdomain %d global_id %d : %d %d\n", i, j, request[i].global_ids[j], ref_request[i].global_ids[j]);
         }

         if (request[i].ranks[j] != ref_request[i].ranks[j]) {
            PUT_ERR("point from wrong rank requested\n");
            printf ("check_data_request subdomain %d rank %d : %d %d\n", i, j, request[i].ranks[j], ref_request[i].ranks[j]);
         }
      }
   }
}

void
delete_interpolation_method_data_request_stack(
   struct interpolation_method_data_request * request,
   unsigned num_points_per_subdomain) {

   if (request != NULL) {

      unsigned i;

      for (i = 0; i < num_points_per_subdomain; ++i) {
         free(request[i].global_ids);
         free(request[i].ranks);
      }
   }
}
