//#define VERBOSE

/**
 * @file test_partial_weights.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "grid.h"
#include "clipping.h"
#include "utils.h"
#include "area.h"
#include "tests.h"
#include "geometry.h"

int main (void) {

  /* For the test at the Equator we cannot get better than 1.0e-11, while
     for the Pole test we are somewhat better.

    tolerance test for the deviation from 1 of  the sum of weights: 1.0e-11

  */

  double area[1024];
  double tgt_area, ref_area;
  double weight[100];
  double weight_sum;

  double const epsilon = 1.0e-10; // relative precision

  struct grid_cell *SourceCell;
  struct grid_cell  TargetCell;

  unsigned nSourceCells = 4;

  unsigned ncs  = 4;
  unsigned nct  = 4;

  unsigned n;

  enum yac_edge_type quad_type[] = {GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE};

  SourceCell = malloc ( nSourceCells  * sizeof(*SourceCell) );

  for ( n = 0; n <  nSourceCells; n++ ) {
    SourceCell[n].num_corners   = ncs;
    SourceCell[n].edge_type     = quad_type;
    SourceCell[n].coordinates_x = malloc ( ncs * sizeof(SourceCell[n].coordinates_x[0]) );
    SourceCell[n].coordinates_y = malloc ( ncs * sizeof(SourceCell[n].coordinates_y[0]) );
    SourceCell[n].coordinates_xyz = malloc ( 3 * ncs * sizeof(SourceCell[n].coordinates_xyz[0]) );
  }

  TargetCell.num_corners   = nct;
  TargetCell.edge_type     = quad_type;

  TargetCell.coordinates_x = malloc ( nct * sizeof(*TargetCell.coordinates_x) );
  TargetCell.coordinates_y = malloc ( nct * sizeof(*TargetCell.coordinates_y) );
  TargetCell.coordinates_xyz = malloc ( 3 * nct * sizeof(*TargetCell.coordinates_xyz) );

#ifdef VERBOSE
  printf (" ---------- test 1.) ----------- \n");
#endif

  /* 1.) 4 source elements overlapping a target cell

    Target Cell

              0.0 (lon)  [1]
              0.7 (lat)
              / \
             /   \
   -0.5 [2] /     \  0.5 [0]
    0.0     \     /  0.0
             \   /
              \ /
              0.0  [3]
             -0.5
  */

  TargetCell.coordinates_x[0] =  0.5*rad;
  TargetCell.coordinates_y[0] =  0.0*rad;
  TargetCell.coordinates_x[1] =  0.0*rad;
  TargetCell.coordinates_y[1] =  0.7*rad;
  TargetCell.coordinates_x[2] = -0.5*rad;
  TargetCell.coordinates_y[2] =  0.0*rad;
  TargetCell.coordinates_x[3] =  0.0*rad;
  TargetCell.coordinates_y[3] = -0.5*rad;

  /* source cell test data

         0.0          1.0 (lon)
         1.8          1.8 (lat)
   ------------------
           |        |
       1   |    0   |
           |        |
           |        |  1.0
   ------------------  0.6
         0.0        |
         0.6        |
                    |
       2   |    3   |
           |        |
   ------------------  1.0
                      -0.6
  */

  SourceCell[0].coordinates_x[0] =  1.0*rad;
  SourceCell[0].coordinates_y[0] =  1.8*rad;
  SourceCell[0].coordinates_x[1] =  0.0*rad;
  SourceCell[0].coordinates_y[1] =  1.8*rad;
  SourceCell[0].coordinates_x[2] =  0.0*rad;
  SourceCell[0].coordinates_y[2] =  0.6*rad;
  SourceCell[0].coordinates_x[3] =  1.0*rad;
  SourceCell[0].coordinates_y[3] =  0.6*rad;

  SourceCell[1].coordinates_x[0] =  SourceCell[0].coordinates_x[1];
  SourceCell[1].coordinates_y[0] =  SourceCell[0].coordinates_y[1];
  SourceCell[1].coordinates_x[1] = -1.0*rad;
  SourceCell[1].coordinates_y[1] =  1.8*rad;
  SourceCell[1].coordinates_x[2] = -1.0*rad;
  SourceCell[1].coordinates_y[2] =  0.6*rad;
  SourceCell[1].coordinates_x[3] =  SourceCell[0].coordinates_x[2];
  SourceCell[1].coordinates_y[3] =  SourceCell[0].coordinates_y[2];

  SourceCell[2].coordinates_x[0] =  SourceCell[0].coordinates_x[2];
  SourceCell[2].coordinates_y[0] =  SourceCell[0].coordinates_y[2];
  SourceCell[2].coordinates_x[1] = -1.0*rad;
  SourceCell[2].coordinates_y[1] =  0.6*rad;
  SourceCell[2].coordinates_x[2] = -1.0*rad;
  SourceCell[2].coordinates_y[2] = -0.6*rad;
  SourceCell[2].coordinates_x[3] =  0.0*rad;
  SourceCell[2].coordinates_y[3] = -0.6*rad;

  SourceCell[3].coordinates_x[0] =  SourceCell[0].coordinates_x[3];
  SourceCell[3].coordinates_y[0] =  SourceCell[0].coordinates_y[3];
  SourceCell[3].coordinates_x[1] =  SourceCell[0].coordinates_x[2];
  SourceCell[3].coordinates_y[1] =  SourceCell[0].coordinates_y[2];
  SourceCell[3].coordinates_x[2] =  0.0*rad;
  SourceCell[3].coordinates_y[2] = -0.6*rad;
  SourceCell[3].coordinates_x[3] =  1.0*rad;
  SourceCell[3].coordinates_y[3] = -0.6*rad;

  for (int i = 0; i < nSourceCells; ++i)
    for (int j = 0; j < SourceCell[i].num_corners; ++j)
      LLtoXYZ(SourceCell[i].coordinates_x[j], SourceCell[i].coordinates_y[j],
              SourceCell[i].coordinates_xyz + j * 3);
  for (int j = 0; j < TargetCell.num_corners; ++j)
    LLtoXYZ(TargetCell.coordinates_x[j], TargetCell.coordinates_y[j],
            TargetCell.coordinates_xyz + j * 3);

  yac_compute_overlap_areas ( nSourceCells, SourceCell, TargetCell, area );

  tgt_area = yac_huiliers_area(TargetCell);
  for (n = 0; n < nSourceCells; ++n)
    weight[n] = area[n] / tgt_area;

  weight_sum = 0.0;
  for ( n = 0; n < nSourceCells; ++n ) weight_sum += weight[n];

  if ( fabs(weight_sum-1.0) > epsilon ) {
    printf ("test 1.) weight %lf\n", weight_sum );
    PUT_ERR("test 1.) weight deviates from expected value of 1.0!\n");
  }

  yac_correct_weights ( nSourceCells, weight );

  weight_sum = 0.0;
  for ( n = 0; n < nSourceCells; ++n ) weight_sum += weight[n];

  if ( fabs(weight_sum-1.0) > epsilon ) {
    printf ("test 1.) weight %lf\n", weight_sum );
    PUT_ERR("test 1.) weight deviates from expected value of 1.0!\n");
  }

#ifdef VERBOSE
  printf (" ---------- test 2.) ----------- \n");
#endif

  /* 2.) Next test is over the pole

     Target cell located over the pole o with
     vertices x and source cells 0 to 3 arranged as


 135.0         90.0       45.0 (lon)
               89.0       88.5 (lat)
         ------------------
        |        |        |
        |        |        |
        |       1x        |
        |        |    0   |
-180.0   ---x----o----x----  0.0        pole @ -180.0; 90.0
        |   2    |        |
        |        |        |
        |       3x        |
        |        |        |
        |        |        |
-135.0   ------------------ -45.0
               -90.0         -0.6

  */

  TargetCell.coordinates_x[0] =    0.0*rad;
  TargetCell.coordinates_y[0] =   89.5*rad;
  TargetCell.coordinates_x[1] =   90.0*rad;
  TargetCell.coordinates_y[1] =   89.5*rad;
  TargetCell.coordinates_x[2] = -180.0*rad;
  TargetCell.coordinates_y[2] =   89.5*rad;
  TargetCell.coordinates_x[3] =  -90.0*rad;
  TargetCell.coordinates_y[3] =   89.5*rad;

  SourceCell[0].coordinates_x[0] =   45.0*rad;
  SourceCell[0].coordinates_y[0] =   88.5*rad;
  SourceCell[0].coordinates_x[1] =   90.0*rad;
  SourceCell[0].coordinates_y[1] =   89.0*rad;
  SourceCell[0].coordinates_x[2] = -180.0*rad;
  SourceCell[0].coordinates_y[2] =   90.0*rad;
  SourceCell[0].coordinates_x[3] =    0.0*rad;
  SourceCell[0].coordinates_y[3] =   89.0*rad;

  SourceCell[1].coordinates_x[0] =  SourceCell[0].coordinates_x[1];
  SourceCell[1].coordinates_y[0] =  SourceCell[0].coordinates_y[1];
  SourceCell[1].coordinates_x[1] =  135.0*rad;
  SourceCell[1].coordinates_y[1] =   88.5*rad;
  SourceCell[1].coordinates_x[2] = -180.0*rad;
  SourceCell[1].coordinates_y[2] =   89.0*rad;
  SourceCell[1].coordinates_x[3] =  SourceCell[0].coordinates_x[2];
  SourceCell[1].coordinates_y[3] =  SourceCell[0].coordinates_y[2];

  SourceCell[2].coordinates_x[0] =  SourceCell[0].coordinates_x[2];
  SourceCell[2].coordinates_y[0] =  SourceCell[0].coordinates_y[2];
  SourceCell[2].coordinates_x[1] = -180.0*rad;
  SourceCell[2].coordinates_y[1] =   89.0*rad;
  SourceCell[2].coordinates_x[2] = -135.0*rad;
  SourceCell[2].coordinates_y[2] =   88.5*rad;
  SourceCell[2].coordinates_x[3] =  -90.0*rad;
  SourceCell[2].coordinates_y[3] =   89.0*rad;

  SourceCell[3].coordinates_x[0] =  SourceCell[0].coordinates_x[3];
  SourceCell[3].coordinates_y[0] =  SourceCell[0].coordinates_y[3];
  SourceCell[3].coordinates_x[1] =  SourceCell[0].coordinates_x[2];
  SourceCell[3].coordinates_y[1] =  SourceCell[0].coordinates_y[2];
  SourceCell[3].coordinates_x[2] =  -90.0*rad;
  SourceCell[3].coordinates_y[2] =   89.0*rad;
  SourceCell[3].coordinates_x[3] =  -45.0*rad;
  SourceCell[3].coordinates_y[3] =   88.5*rad;

  for (int i = 0; i < nSourceCells; ++i)
    for (int j = 0; j < SourceCell[i].num_corners; ++j)
      LLtoXYZ(SourceCell[i].coordinates_x[j], SourceCell[i].coordinates_y[j],
              SourceCell[i].coordinates_xyz + j * 3);
  for (int j = 0; j < TargetCell.num_corners; ++j)
    LLtoXYZ(TargetCell.coordinates_x[j], TargetCell.coordinates_y[j],
            TargetCell.coordinates_xyz + j * 3);

  yac_compute_overlap_areas ( nSourceCells, SourceCell, TargetCell, area );

  tgt_area = yac_huiliers_area(TargetCell);
  for (n = 0; n < nSourceCells; ++n)
    weight[n] = area[n] / tgt_area;

  weight_sum = 0.0;
  for ( n = 0; n < nSourceCells; ++n ) weight_sum += weight[n];

  if ( fabs(weight_sum-1.0) > epsilon ) {
    printf ("test 2.) weight %lf\n", weight_sum );
    PUT_ERR("test 2.) weight deviates from expected value of 1.0!\n");
  }

  yac_correct_weights ( nSourceCells, weight );

  weight_sum = 0.0;
  for ( n = 0; n < nSourceCells; ++n ) weight_sum += weight[n];

  if ( fabs(weight_sum-1.0) > epsilon ) {
    printf ("test 2.) weight %lf\n", weight_sum );
    PUT_ERR("test 2.) weight deviates from expected value of 1.0!\n");
  }
  /* Free memory allocated in main */

  for ( n = 0; n < nSourceCells; n++ ) {
    free(SourceCell[n].coordinates_x);
    free(SourceCell[n].coordinates_y);
    free(SourceCell[n].coordinates_xyz);
  }
  free (SourceCell);

#ifdef VERBOSE
  printf (" ---------- test 3.) ----------- \n");
#endif

  nSourceCells = 1;

  SourceCell = malloc ( nSourceCells  * sizeof(*SourceCell) );

  for ( n = 0; n < nSourceCells; n++ ) {
    SourceCell[n].num_corners   = ncs;
    SourceCell[n].edge_type     = quad_type;
    SourceCell[n].coordinates_x = malloc ( ncs * sizeof(SourceCell[n].coordinates_x[0]) );
    SourceCell[n].coordinates_y = malloc ( ncs * sizeof(SourceCell[n].coordinates_y[0]) );
    SourceCell[n].coordinates_xyz = malloc ( 3 * ncs * sizeof(SourceCell[n].coordinates_xyz[0]) );
  }

  TargetCell.coordinates_x[0] =  -180.0*rad;
  TargetCell.coordinates_y[0] =   -90.0*rad;
  TargetCell.coordinates_x[1] =  -150.0*rad;
  TargetCell.coordinates_y[1] =   -90.0*rad;
  TargetCell.coordinates_x[2] =  -150.0*rad;
  TargetCell.coordinates_y[2] =   -60.0*rad;
  TargetCell.coordinates_x[3] =  -180.0*rad;
  TargetCell.coordinates_y[3] =   -60.0*rad;

  SourceCell[0].coordinates_x[0] = -180.0*rad;
  SourceCell[0].coordinates_y[0] =  -90.0*rad;
  SourceCell[0].coordinates_x[1] = -150.0*rad;
  SourceCell[0].coordinates_y[1] =  -90.0*rad;
  SourceCell[0].coordinates_x[2] = -150.0*rad;
  SourceCell[0].coordinates_y[2] =  -60.0*rad;
  SourceCell[0].coordinates_x[3] = -180.0*rad;
  SourceCell[0].coordinates_y[3] =  -60.0*rad;

  for (int i = 0; i < nSourceCells; ++i)
    for (int j = 0; j < SourceCell[i].num_corners; ++j)
      LLtoXYZ(SourceCell[i].coordinates_x[j], SourceCell[i].coordinates_y[j],
              SourceCell[i].coordinates_xyz + j * 3);
  for (int j = 0; j < TargetCell.num_corners; ++j)
    LLtoXYZ(TargetCell.coordinates_x[j], TargetCell.coordinates_y[j],
            TargetCell.coordinates_xyz + j * 3);

  yac_compute_overlap_areas ( nSourceCells, SourceCell, TargetCell, area );

  tgt_area = yac_huiliers_area(TargetCell);
  for (n = 0; n < nSourceCells; ++n)
    weight[n] = area[n] / tgt_area;

  weight_sum = 0.0;
  for ( n = 0; n < nSourceCells; ++n ) weight_sum += weight[n];

  if ( fabs(weight_sum-1.0) > epsilon ) {
    printf ("test 3.) weight %lf\n", weight_sum );
    PUT_ERR("test 3.) weight deviates from expected value of 1.0!\n");
  }

  yac_correct_weights ( nSourceCells, weight );

  weight_sum = 0.0;
  for ( n = 0; n < nSourceCells; ++n ) weight_sum += weight[n];

  if ( fabs(weight_sum-1.0) > epsilon ) {
    printf ("test 3.) weight %lf\n", weight_sum );
    PUT_ERR("test 3.) weight deviates from expected value of 1.0!\n");
  }

#ifdef VERBOSE
  printf (" ---------- test 4.) ----------- \n");
#endif

  TargetCell.coordinates_x[0] =  -180.0*rad;
  TargetCell.coordinates_y[0] =    60.0*rad;
  TargetCell.coordinates_x[1] =  -150.0*rad;
  TargetCell.coordinates_y[1] =    60.0*rad;
  TargetCell.coordinates_x[2] =  -150.0*rad;
  TargetCell.coordinates_y[2] =    90.0*rad;
  TargetCell.coordinates_x[3] =  -180.0*rad;
  TargetCell.coordinates_y[3] =    90.0*rad;

  SourceCell[0].coordinates_x[0] = -180.0*rad;
  SourceCell[0].coordinates_y[0] =   60.0*rad;
  SourceCell[0].coordinates_x[1] = -150.0*rad;
  SourceCell[0].coordinates_y[1] =   60.0*rad;
  SourceCell[0].coordinates_x[2] = -150.0*rad;
  SourceCell[0].coordinates_y[2] =   90.0*rad;
  SourceCell[0].coordinates_x[3] = -180.0*rad;
  SourceCell[0].coordinates_y[3] =   90.0*rad;

  for (int i = 0; i < nSourceCells; ++i)
    for (int j = 0; j < SourceCell[i].num_corners; ++j)
      LLtoXYZ(SourceCell[i].coordinates_x[j], SourceCell[i].coordinates_y[j],
              SourceCell[i].coordinates_xyz + j * 3);
  for (int j = 0; j < TargetCell.num_corners; ++j)
    LLtoXYZ(TargetCell.coordinates_x[j], TargetCell.coordinates_y[j],
            TargetCell.coordinates_xyz + j * 3);

  yac_compute_overlap_areas ( nSourceCells, SourceCell, TargetCell, area );

  tgt_area = yac_huiliers_area(TargetCell);
  for (n = 0; n < nSourceCells; ++n)
    weight[n] = area[n] / tgt_area;

  weight_sum = 0.0;
  for ( n = 0; n < nSourceCells; ++n ) weight_sum += weight[n];

  if ( fabs(weight_sum-1.0) > epsilon ) {
    printf ("test 4.) weight %lf\n", weight_sum );
    PUT_ERR("test 4.) weight deviates from expected value of 1.0!\n");
  }

  yac_correct_weights ( nSourceCells, weight );

  weight_sum = 0.0;
  for ( n = 0; n < nSourceCells; ++n ) weight_sum += weight[n];

  if ( fabs(weight_sum-1.0) > epsilon ) {
    printf ("test 4.) weight %lf\n", weight_sum );
    PUT_ERR("test 4.) weight deviates from expected value of 1.0!\n");
  }

  for ( n = 0; n < nSourceCells; n++ ) {
    free(SourceCell[n].coordinates_x);
    free(SourceCell[n].coordinates_y);
    free(SourceCell[n].coordinates_xyz);
  }

  free (SourceCell);

#ifdef VERBOSE
  printf (" ---------- test 5.) ----------- \n");
#endif

  nSourceCells = 3;

  SourceCell = malloc ( nSourceCells  * sizeof(*SourceCell) );

  for ( n = 0; n < nSourceCells; n++ ) {
    SourceCell[n].num_corners   = ncs;
    SourceCell[n].edge_type     = quad_type;
    SourceCell[n].coordinates_x = malloc ( ncs * sizeof(SourceCell[n].coordinates_x[0]) );
    SourceCell[n].coordinates_y = malloc ( ncs * sizeof(SourceCell[n].coordinates_y[0]) );
    SourceCell[n].coordinates_xyz = malloc ( 3 * ncs * sizeof(SourceCell[n].coordinates_xyz[0]) );
  }

  TargetCell.coordinates_x[0] =    25.0*rad;
  TargetCell.coordinates_y[0] =    75.0*rad;
  TargetCell.coordinates_x[1] =    30.0*rad;
  TargetCell.coordinates_y[1] =    75.0*rad;
  TargetCell.coordinates_x[2] =    30.0*rad;
  TargetCell.coordinates_y[2] =    80.0*rad;
  TargetCell.coordinates_x[3] =    25.0*rad;
  TargetCell.coordinates_y[3] =    80.0*rad;

  SourceCell[0].coordinates_x[0] =   30.0*rad;
  SourceCell[0].coordinates_y[0] =   75.0*rad;
  SourceCell[0].coordinates_x[1] =   35.0*rad;
  SourceCell[0].coordinates_y[1] =   75.0*rad;
  SourceCell[0].coordinates_x[2] =   35.0*rad;
  SourceCell[0].coordinates_y[2] =   80.0*rad;
  SourceCell[0].coordinates_x[3] =   30.0*rad;
  SourceCell[0].coordinates_y[3] =   80.0*rad;

  SourceCell[1].coordinates_x[0] =   25.0*rad;
  SourceCell[1].coordinates_y[0] =   75.0*rad;
  SourceCell[1].coordinates_x[1] =   30.0*rad;
  SourceCell[1].coordinates_y[1] =   75.0*rad;
  SourceCell[1].coordinates_x[2] =   30.0*rad;
  SourceCell[1].coordinates_y[2] =   80.0*rad;
  SourceCell[1].coordinates_x[3] =   25.0*rad;
  SourceCell[1].coordinates_y[3] =   80.0*rad;

  SourceCell[2].coordinates_x[0] =   20.0*rad;
  SourceCell[2].coordinates_y[0] =   75.0*rad;
  SourceCell[2].coordinates_x[1] =   25.0*rad;
  SourceCell[2].coordinates_y[1] =   75.0*rad;
  SourceCell[2].coordinates_x[2] =   25.0*rad;
  SourceCell[2].coordinates_y[2] =   80.0*rad;
  SourceCell[2].coordinates_x[3] =   20.0*rad;
  SourceCell[2].coordinates_y[3] =   80.0*rad;

  for (int i = 0; i < nSourceCells; ++i)
    for (int j = 0; j < SourceCell[i].num_corners; ++j)
      LLtoXYZ(SourceCell[i].coordinates_x[j], SourceCell[i].coordinates_y[j],
              SourceCell[i].coordinates_xyz + j * 3);
  for (int j = 0; j < TargetCell.num_corners; ++j)
    LLtoXYZ(TargetCell.coordinates_x[j], TargetCell.coordinates_y[j],
            TargetCell.coordinates_xyz + j * 3);

  yac_compute_overlap_areas ( nSourceCells, SourceCell, TargetCell, area );

  tgt_area = yac_huiliers_area(TargetCell);
  for (n = 0; n < nSourceCells; ++n)
    weight[n] = area[n] / tgt_area;

  weight_sum = 0.0;
  for ( n = 0; n < nSourceCells; ++n ) weight_sum += weight[n];

  if ( fabs(weight_sum-1.0) > epsilon ) {
    printf ("test 5.) weight %lf\n", weight_sum );
    PUT_ERR("test 5.) weight deviates from expected value of 1.0!\n");
  }

  yac_correct_weights ( nSourceCells, weight );

  weight_sum = 0.0;
  for ( n = 0; n < nSourceCells; ++n ) weight_sum += weight[n];

  if ( fabs(weight_sum-1.0) > epsilon ) {
    printf ("test 5.) weight %lf\n", weight_sum );
    PUT_ERR("test 5.) weight deviates from expected value of 1.0!\n");
  }

  for ( n = 0; n < nSourceCells; n++ ) {
    free(SourceCell[n].coordinates_x);
    free(SourceCell[n].coordinates_y);
    free(SourceCell[n].coordinates_xyz);
  }

  free (SourceCell);

#ifdef VERBOSE
  printf (" ---------- test 6.) ----------- \n");
#endif

  nSourceCells = 1;

  SourceCell = malloc ( nSourceCells  * sizeof(*SourceCell) );

  for ( n = 0; n < nSourceCells; n++ ) {
    SourceCell[n].num_corners   = 3;
    SourceCell[n].edge_type     = quad_type;
    SourceCell[n].coordinates_x = malloc ( ncs * sizeof(SourceCell[n].coordinates_x[0]) );
    SourceCell[n].coordinates_y = malloc ( ncs * sizeof(SourceCell[n].coordinates_y[0]) );
    SourceCell[n].coordinates_xyz = malloc ( 3 * ncs * sizeof(SourceCell[n].coordinates_xyz[0]) );
  }

  TargetCell.coordinates_x[0] = 0.098707396293595859;
  TargetCell.coordinates_y[0] = 0.33303337284067219;
  TargetCell.coordinates_x[1] = 0.11782629535552268;
  TargetCell.coordinates_y[1] = 0.34548477527681493;
  TargetCell.coordinates_x[2] = 0.098174496441371703;
  TargetCell.coordinates_y[2] = 0.35333560925858365;
  TargetCell.num_corners      = 3;

  SourceCell[0].coordinates_x[0] = 0.073631077818510776;
  SourceCell[0].coordinates_y[0] = 0.34361169648638362;
  SourceCell[0].coordinates_x[1] = 0.098174770424681035;
  SourceCell[0].coordinates_y[1] = 0.34361169648638362;
  SourceCell[0].coordinates_x[2] = 0.098174770424681035;
  SourceCell[0].coordinates_y[2] = 0.36815538909255391;
  SourceCell[0].coordinates_x[3] = 0.073631077818510776;
  SourceCell[0].coordinates_y[3] = 0.36815538909255391;

  for (int i = 0; i < nSourceCells; ++i)
    for (int j = 0; j < SourceCell[i].num_corners; ++j)
      LLtoXYZ(SourceCell[i].coordinates_x[j], SourceCell[i].coordinates_y[j],
              SourceCell[i].coordinates_xyz + j * 3);
  for (int j = 0; j < TargetCell.num_corners; ++j)
    LLtoXYZ(TargetCell.coordinates_x[j], TargetCell.coordinates_y[j],
            TargetCell.coordinates_xyz + j * 3);

  yac_compute_overlap_areas ( nSourceCells, SourceCell, TargetCell, area );

  tgt_area = yac_huiliers_area(TargetCell);
  for (n = 0; n < nSourceCells; ++n)
    weight[n] = area[n] / tgt_area;

  if ( weight[0] < 0 ) {
    printf ("test 6.) weight is smaller than zero (%.2e)\n", weight[0] );
    PUT_ERR("test 6.) weight deviates from expected value > 0.0!\n");
  }

  for ( n = 0; n < nSourceCells; n++ ) {
    free(SourceCell[n].coordinates_x);
    free(SourceCell[n].coordinates_y);
    free(SourceCell[n].coordinates_xyz);
  }

  free (SourceCell);

#ifdef VERBOSE
  printf (" ---------- test 7.) ----------- \n");
#endif
  {
    enum yac_edge_type edge_types[4] = {LAT_CIRCLE, LON_CIRCLE,
                                        LAT_CIRCLE, LON_CIRCLE};
    nSourceCells = 4;

    SourceCell = malloc ( nSourceCells  * sizeof(*SourceCell) );

    // TargetCell.edge_type = quad_type;
    TargetCell.edge_type = edge_types;

    for ( n = 0; n < nSourceCells; n++ ) {
      SourceCell[n].num_corners   = 4;
      // SourceCell[n].edge_type     = quad_type;
      SourceCell[n].edge_type     = edge_types;
      SourceCell[n].coordinates_x = malloc ( ncs * sizeof(SourceCell[n].coordinates_x[0]) );
      SourceCell[n].coordinates_y = malloc ( ncs * sizeof(SourceCell[n].coordinates_y[0]) );
      SourceCell[n].coordinates_xyz = malloc ( 3 * ncs * sizeof(SourceCell[n].coordinates_xyz[0]) );
    }

    TargetCell.coordinates_x[0] = -0.5*rad;
    TargetCell.coordinates_y[0] = -0.5*rad;
    TargetCell.coordinates_x[1] =  0.5*rad;
    TargetCell.coordinates_y[1] = -0.5*rad;
    TargetCell.coordinates_x[2] =  0.5*rad;
    TargetCell.coordinates_y[2] =  0.5*rad;
    TargetCell.coordinates_x[3] = -0.5*rad;
    TargetCell.coordinates_y[3] =  0.5*rad;
    TargetCell.num_corners      = 4;

    SourceCell[0].coordinates_x[0] = -1*rad;
    SourceCell[0].coordinates_y[0] = -1*rad;
    SourceCell[0].coordinates_x[1] =  0*rad;
    SourceCell[0].coordinates_y[1] = -1*rad;
    SourceCell[0].coordinates_x[2] =  0*rad;
    SourceCell[0].coordinates_y[2] =  0*rad;
    SourceCell[0].coordinates_x[3] = -1*rad;
    SourceCell[0].coordinates_y[3] =  0*rad;
    SourceCell[1].coordinates_x[0] =  0*rad;
    SourceCell[1].coordinates_y[0] = -1*rad;
    SourceCell[1].coordinates_x[1] =  1*rad;
    SourceCell[1].coordinates_y[1] = -1*rad;
    SourceCell[1].coordinates_x[2] =  1*rad;
    SourceCell[1].coordinates_y[2] =  0*rad;
    SourceCell[1].coordinates_x[3] =  0*rad;
    SourceCell[1].coordinates_y[3] =  0*rad;
    SourceCell[2].coordinates_x[0] = -1*rad;
    SourceCell[2].coordinates_y[0] =  0*rad;
    SourceCell[2].coordinates_x[1] =  0*rad;
    SourceCell[2].coordinates_y[1] =  0*rad;
    SourceCell[2].coordinates_x[2] =  0*rad;
    SourceCell[2].coordinates_y[2] =  1*rad;
    SourceCell[2].coordinates_x[3] = -1*rad;
    SourceCell[2].coordinates_y[3] =  1*rad;
    SourceCell[3].coordinates_x[0] =  0*rad;
    SourceCell[3].coordinates_y[0] =  0*rad;
    SourceCell[3].coordinates_x[1] =  1*rad;
    SourceCell[3].coordinates_y[1] =  0*rad;
    SourceCell[3].coordinates_x[2] =  1*rad;
    SourceCell[3].coordinates_y[2] =  1*rad;
    SourceCell[3].coordinates_x[3] =  0*rad;
    SourceCell[3].coordinates_y[3] =  1*rad;

    for (int i = 0; i < nSourceCells; ++i)
      for (int j = 0; j < SourceCell[i].num_corners; ++j)
        LLtoXYZ(SourceCell[i].coordinates_x[j], SourceCell[i].coordinates_y[j],
                SourceCell[i].coordinates_xyz + j * 3);
    for (int j = 0; j < TargetCell.num_corners; ++j)
      LLtoXYZ(TargetCell.coordinates_x[j], TargetCell.coordinates_y[j],
              TargetCell.coordinates_xyz + j * 3);

    yac_compute_overlap_areas ( nSourceCells, SourceCell, TargetCell, area );

    tgt_area = yac_pole_area(TargetCell);
    for (n = 0; n < nSourceCells; ++n)
      tgt_area -= area[n];

    if ( fabs(tgt_area) > yac_area_tol() ) {
      PUT_ERR("test 7.) error in partial area computation\n");
    }

    for ( n = 0; n < nSourceCells; n++ ) {
      free(SourceCell[n].coordinates_x);
      free(SourceCell[n].coordinates_y);
      free(SourceCell[n].coordinates_xyz);
    }

    free (SourceCell);

    TargetCell.edge_type = quad_type;
  }

#ifdef VERBOSE
  printf (" ---------- test 8.) ----------- \n");
#endif
  {
    enum yac_edge_type edge_types[4] = {LAT_CIRCLE, LON_CIRCLE,
                                        LAT_CIRCLE, LON_CIRCLE};
    nSourceCells = 1;

    SourceCell = malloc ( nSourceCells  * sizeof(*SourceCell) );

    TargetCell.edge_type = quad_type;

    for ( n = 0; n < nSourceCells; n++ ) {
      SourceCell[n].num_corners   = 4;
      SourceCell[n].edge_type     = edge_types;
      SourceCell[n].coordinates_x = malloc ( ncs * sizeof(SourceCell[n].coordinates_x[0]) );
      SourceCell[n].coordinates_y = malloc ( ncs * sizeof(SourceCell[n].coordinates_y[0]) );
      SourceCell[n].coordinates_xyz = malloc ( 3 * ncs * sizeof(SourceCell[n].coordinates_xyz[0]) );
    }

    TargetCell.coordinates_x[0] = -0.5*rad;
    TargetCell.coordinates_y[0] = -0.5*rad;
    TargetCell.coordinates_x[1] =  0.5*rad;
    TargetCell.coordinates_y[1] = -0.5*rad;
    TargetCell.coordinates_x[2] =  0.0*rad;
    TargetCell.coordinates_y[2] =  0.5*rad;
    TargetCell.num_corners      = 3;

    SourceCell[0].coordinates_x[0] = -1*rad;
    SourceCell[0].coordinates_y[0] = -1*rad;
    SourceCell[0].coordinates_x[1] =  1*rad;
    SourceCell[0].coordinates_y[1] = -1*rad;
    SourceCell[0].coordinates_x[2] =  1*rad;
    SourceCell[0].coordinates_y[2] =  1*rad;
    SourceCell[0].coordinates_x[3] = -1*rad;
    SourceCell[0].coordinates_y[3] =  1*rad;

    for (int i = 0; i < nSourceCells; ++i)
      for (int j = 0; j < SourceCell[i].num_corners; ++j)
        LLtoXYZ(SourceCell[i].coordinates_x[j], SourceCell[i].coordinates_y[j],
                SourceCell[i].coordinates_xyz + j * 3);
    for (int j = 0; j < TargetCell.num_corners; ++j)
      LLtoXYZ(TargetCell.coordinates_x[j], TargetCell.coordinates_y[j],
              TargetCell.coordinates_xyz + j * 3);

    yac_compute_overlap_areas ( nSourceCells, SourceCell, TargetCell, area );

    tgt_area = yac_pole_area(TargetCell);
    for (n = 0; n < nSourceCells; ++n)
      tgt_area -= area[n];

    if ( fabs(tgt_area) > yac_area_tol() ) {
      PUT_ERR("test 7.) error in partial area computation\n");
    }

    for ( n = 0; n < nSourceCells; n++ ) {
      free(SourceCell[n].coordinates_x);
      free(SourceCell[n].coordinates_y);
      free(SourceCell[n].coordinates_xyz);
    }

    free (SourceCell);

    TargetCell.edge_type = quad_type;
  }

#ifdef VERBOSE
  printf (" ---------- test 9.) ----------- \n");
#endif
  {
    enum yac_edge_type edge_types[4] = {LAT_CIRCLE, LON_CIRCLE,
                                        LAT_CIRCLE, LON_CIRCLE};
    nSourceCells = 1;

    SourceCell = malloc ( nSourceCells  * sizeof(*SourceCell) );

    TargetCell.edge_type = edge_types;

    for ( n = 0; n < nSourceCells; n++ ) {
      SourceCell[n].num_corners   = 3;
      SourceCell[n].edge_type     = quad_type;
      SourceCell[n].coordinates_x = malloc ( ncs * sizeof(SourceCell[n].coordinates_x[0]) );
      SourceCell[n].coordinates_y = malloc ( ncs * sizeof(SourceCell[n].coordinates_y[0]) );
      SourceCell[n].coordinates_xyz = malloc ( 3 * ncs * sizeof(SourceCell[n].coordinates_xyz[0]) );
    }

    TargetCell.coordinates_x[0] = -1*rad;
    TargetCell.coordinates_y[0] = -1*rad;
    TargetCell.coordinates_x[1] =  1*rad;
    TargetCell.coordinates_y[1] = -1*rad;
    TargetCell.coordinates_x[2] =  1*rad;
    TargetCell.coordinates_y[2] =  1*rad;
    TargetCell.coordinates_x[3] = -1*rad;
    TargetCell.coordinates_y[3] =  1*rad;
    TargetCell.num_corners      = 4;

    SourceCell[0].coordinates_x[0] = -0.5*rad;
    SourceCell[0].coordinates_y[0] = -0.5*rad;
    SourceCell[0].coordinates_x[1] =  0.5*rad;
    SourceCell[0].coordinates_y[1] = -0.5*rad;
    SourceCell[0].coordinates_x[2] =  0.0*rad;
    SourceCell[0].coordinates_y[2] =  0.5*rad;

    for (int i = 0; i < nSourceCells; ++i)
      for (int j = 0; j < SourceCell[i].num_corners; ++j)
        LLtoXYZ(SourceCell[i].coordinates_x[j], SourceCell[i].coordinates_y[j],
                SourceCell[i].coordinates_xyz + j * 3);
    for (int j = 0; j < TargetCell.num_corners; ++j)
      LLtoXYZ(TargetCell.coordinates_x[j], TargetCell.coordinates_y[j],
              TargetCell.coordinates_xyz + j * 3);

    yac_compute_overlap_areas ( nSourceCells, SourceCell, TargetCell, area );

    ref_area = yac_pole_area(SourceCell[0]);
    for (n = 0; n < nSourceCells; ++n)
      ref_area -= area[n];

    if ( fabs(ref_area) > yac_area_tol() ) {
      PUT_ERR("test 7.) error in partial area computation\n");
    }

    for ( n = 0; n < nSourceCells; n++ ) {
      free(SourceCell[n].coordinates_x);
      free(SourceCell[n].coordinates_y);
      free(SourceCell[n].coordinates_xyz);
    }

    free (SourceCell);

    TargetCell.edge_type = quad_type;
  }

#ifdef VERBOSE
  printf (" ---------- test 10.) ----------- \n");
#endif
  {
    enum yac_edge_type edge_types[4] = {LAT_CIRCLE, LON_CIRCLE,
                                        LAT_CIRCLE, LON_CIRCLE};
    nSourceCells = 4;

    SourceCell = malloc ( nSourceCells  * sizeof(*SourceCell) );

    TargetCell.edge_type = edge_types;

    for ( n = 0; n < nSourceCells; n++ ) {
      SourceCell[n].num_corners   = 3;
      SourceCell[n].edge_type     = quad_type;
      SourceCell[n].coordinates_x = malloc ( ncs * sizeof(SourceCell[n].coordinates_x[0]) );
      SourceCell[n].coordinates_y = malloc ( ncs * sizeof(SourceCell[n].coordinates_y[0]) );
      SourceCell[n].coordinates_xyz = malloc ( 3 * ncs * sizeof(SourceCell[n].coordinates_xyz[0]) );
    }

    TargetCell.coordinates_x[0] = 2.5034566458293663;
    TargetCell.coordinates_x[1] = 2.5280003384355365;
    TargetCell.coordinates_x[2] = 2.5280003384355365;
    TargetCell.coordinates_x[3] = 2.5034566458293663;
    TargetCell.coordinates_y[0] = 0.5890486225480862;
    TargetCell.coordinates_y[1] = 0.5890486225480862;
    TargetCell.coordinates_y[2] = 0.6135923151542565;
    TargetCell.coordinates_y[3] = 0.6135923151542565;
    TargetCell.num_corners      = 4;

    SourceCell[0].coordinates_x[0] = 2.5132750564642796;
    SourceCell[0].coordinates_x[1] = 2.5132751216401110;
    SourceCell[0].coordinates_x[2] = 2.5530451220055963;
    SourceCell[0].coordinates_y[0] = 0.6227096831849960;
    SourceCell[0].coordinates_y[1] = 0.5888365165503746;
    SourceCell[0].coordinates_y[2] = 0.6044728875052012;

    SourceCell[1].coordinates_x[0] = 2.5530451220055963;
    SourceCell[1].coordinates_x[1] = 2.5132751216401110;
    SourceCell[1].coordinates_x[2] = 2.5514531161521616;
    SourceCell[1].coordinates_y[0] = 0.6044728875052012;
    SourceCell[1].coordinates_y[1] = 0.5888365165503746;
    SourceCell[1].coordinates_y[2] = 0.5706926293017953;

    SourceCell[2].coordinates_x[0] = 2.4735050741253155;
    SourceCell[2].coordinates_x[1] = 2.5132751216401110;
    SourceCell[2].coordinates_x[2] = 2.5132750564642796;
    SourceCell[2].coordinates_y[0] = 0.6044725603601205;
    SourceCell[2].coordinates_y[1] = 0.5888365165503746;
    SourceCell[2].coordinates_y[2] = 0.6227096831849960;

    SourceCell[3].coordinates_x[0] = 2.4750971948235070;
    SourceCell[3].coordinates_x[1] = 2.5132751216401110;
    SourceCell[3].coordinates_x[2] = 2.4735050741253155;
    SourceCell[3].coordinates_y[0] = 0.5706923461529829;
    SourceCell[3].coordinates_y[1] = 0.5888365165503746;
    SourceCell[3].coordinates_y[2] = 0.6044725603601205;

    for (int i = 0; i < nSourceCells; ++i)
      for (int j = 0; j < SourceCell[i].num_corners; ++j)
        LLtoXYZ(SourceCell[i].coordinates_x[j], SourceCell[i].coordinates_y[j],
                SourceCell[i].coordinates_xyz + j * 3);
    for (int j = 0; j < TargetCell.num_corners; ++j)
      LLtoXYZ(TargetCell.coordinates_x[j], TargetCell.coordinates_y[j],
              TargetCell.coordinates_xyz + j * 3);

    yac_compute_overlap_areas ( nSourceCells, SourceCell, TargetCell, area );

    ref_area = yac_pole_area(TargetCell);
    for (n = 0; n < nSourceCells; ++n)
      ref_area -= area[n];

    if ( fabs(ref_area) > yac_area_tol() ) {
      PUT_ERR("test 7.) error in partial area computation\n");
    }

    for ( n = 0; n < nSourceCells; n++ ) {
      free(SourceCell[n].coordinates_x);
      free(SourceCell[n].coordinates_y);
      free(SourceCell[n].coordinates_xyz);
    }

    free (SourceCell);

    TargetCell.edge_type = quad_type;
  }

#ifdef VERBOSE
  printf (" ---------- test 11.) ----------- \n");
#endif
  {
    enum yac_edge_type edge_types[4] = {LAT_CIRCLE, LON_CIRCLE,
                                        LAT_CIRCLE, LON_CIRCLE};
    nSourceCells = 3;

    SourceCell = malloc ( nSourceCells  * sizeof(*SourceCell) );

    TargetCell.edge_type = edge_types;

    for ( n = 0; n < nSourceCells; n++ ) {
      SourceCell[n].num_corners   = 3;
      SourceCell[n].edge_type     = quad_type;
      SourceCell[n].coordinates_x = malloc ( ncs * sizeof(SourceCell[n].coordinates_x[0]) );
      SourceCell[n].coordinates_y = malloc ( ncs * sizeof(SourceCell[n].coordinates_y[0]) );
      SourceCell[n].coordinates_xyz = malloc ( 3 * ncs * sizeof(SourceCell[n].coordinates_xyz[0]) );
    }

    TargetCell.coordinates_x[0] = 0.6626797003665970;
    TargetCell.coordinates_x[1] = 0.6872233929727672;
    TargetCell.coordinates_x[2] = 0.6872233929727672;
    TargetCell.coordinates_x[3] = 0.6626797003665970;
    TargetCell.coordinates_y[0] = 1.2271846303085130;
    TargetCell.coordinates_y[1] = 1.2271846303085130;
    TargetCell.coordinates_y[2] = 1.2517283229146832;
    TargetCell.coordinates_y[3] = 1.2517283229146832;
    TargetCell.num_corners      = 4;

    SourceCell[0].coordinates_x[0] = 0.6910166804158336;
    SourceCell[0].coordinates_x[1] = 0.6283648515564006;
    SourceCell[0].coordinates_x[2] = 0.7428780447168533;
    SourceCell[0].coordinates_y[0] = 1.2515918327919406;
    SourceCell[0].coordinates_y[1] = 1.2207207913058662;
    SourceCell[0].coordinates_y[2] = 1.2190404189581385;

    SourceCell[1].coordinates_x[0] = 0.5657152198049494;
    SourceCell[1].coordinates_x[1] = 0.6283648515564006;
    SourceCell[1].coordinates_x[2] = 0.6910166804158336;
    SourceCell[1].coordinates_y[0] = 1.2515932566235379;
    SourceCell[1].coordinates_y[1] = 1.2207207913058662;
    SourceCell[1].coordinates_y[2] = 1.2515918327919406;

    SourceCell[2].coordinates_x[0] = 0.6283674880262698;
    SourceCell[2].coordinates_x[1] = 0.5657152198049494;
    SourceCell[2].coordinates_x[2] = 0.6910166804158336;
    SourceCell[2].coordinates_y[0] = 1.2831450892755811;
    SourceCell[2].coordinates_y[1] = 1.2515932566235379;
    SourceCell[2].coordinates_y[2] = 1.2515918327919406;

    for (int i = 0; i < nSourceCells; ++i)
      for (int j = 0; j < SourceCell[i].num_corners; ++j)
        LLtoXYZ(SourceCell[i].coordinates_x[j], SourceCell[i].coordinates_y[j],
                SourceCell[i].coordinates_xyz + j * 3);
    for (int j = 0; j < TargetCell.num_corners; ++j)
      LLtoXYZ(TargetCell.coordinates_x[j], TargetCell.coordinates_y[j],
              TargetCell.coordinates_xyz + j * 3);

    yac_compute_overlap_areas ( nSourceCells, SourceCell, TargetCell, area );

    ref_area = yac_pole_area(TargetCell);
    for (n = 0; n < nSourceCells; ++n)
      ref_area -= area[n];

    if ( fabs(ref_area) > yac_area_tol() ) {
      PUT_ERR("test 7.) error in partial area computation\n");
    }

    for ( n = 0; n < nSourceCells; n++ ) {
      free(SourceCell[n].coordinates_x);
      free(SourceCell[n].coordinates_y);
      free(SourceCell[n].coordinates_xyz);
    }

    free (SourceCell);

    TargetCell.edge_type = quad_type;
  }

#ifdef VERBOSE
  printf (" ---------- test 12.) ----------- \n");
#endif
  {
    enum yac_edge_type edge_types[4] = {LAT_CIRCLE, LON_CIRCLE,
                                        LAT_CIRCLE, LON_CIRCLE};
    nSourceCells = 7;

    SourceCell = malloc ( nSourceCells  * sizeof(*SourceCell) );

    TargetCell.edge_type = quad_type;

    for ( n = 0; n < nSourceCells; n++ ) {
      SourceCell[n].num_corners   = 4;
      SourceCell[n].edge_type     = edge_types;
      SourceCell[n].coordinates_x = malloc ( ncs * sizeof(SourceCell[n].coordinates_x[0]) );
      SourceCell[n].coordinates_y = malloc ( ncs * sizeof(SourceCell[n].coordinates_y[0]) );
      SourceCell[n].coordinates_xyz = malloc ( 3 * ncs * sizeof(SourceCell[n].coordinates_xyz[0]) );
    }

    TargetCell.coordinates_x[0] = -2.1798479726998332;
    TargetCell.coordinates_x[1] = -2.1991148570983681;
    TargetCell.coordinates_x[2] = -2.2190472361409284;
    TargetCell.coordinates_y[0] = -0.0311624518327736;
    TargetCell.coordinates_y[1] =  0.0000000043633749;
    TargetCell.coordinates_y[2] = -0.0338654325037920;
    TargetCell.num_corners      = 3;

    SourceCell[0].coordinates_x[0] = 4.0987966652304335;
    SourceCell[0].coordinates_x[1] = 4.1233403578366037;
    SourceCell[0].coordinates_x[2] = 4.1233403578366037;
    SourceCell[0].coordinates_x[3] = 4.0987966652304335;
    SourceCell[0].coordinates_y[0] = -0.0490873852123405;
    SourceCell[0].coordinates_y[1] = -0.0490873852123405;
    SourceCell[0].coordinates_y[2] = -0.0245436926061703;
    SourceCell[0].coordinates_y[3] = -0.0245436926061703;

    SourceCell[1].coordinates_x[0] = 4.0742529726242633;
    SourceCell[1].coordinates_x[1] = 4.0987966652304335;
    SourceCell[1].coordinates_x[2] = 4.0987966652304335;
    SourceCell[1].coordinates_x[3] = 4.0742529726242633;
    SourceCell[1].coordinates_y[0] = -0.0490873852123405;
    SourceCell[1].coordinates_y[1] = -0.0490873852123405;
    SourceCell[1].coordinates_y[2] = -0.0245436926061703;
    SourceCell[1].coordinates_y[3] = -0.0245436926061703;

    SourceCell[2].coordinates_x[0] = 4.0742529726242633;
    SourceCell[2].coordinates_x[1] = 4.0987966652304335;
    SourceCell[2].coordinates_x[2] = 4.0987966652304335;
    SourceCell[2].coordinates_x[3] = 4.0742529726242633;
    SourceCell[2].coordinates_y[0] = 0.0000000000000000;
    SourceCell[2].coordinates_y[1] = 0.0000000000000000;
    SourceCell[2].coordinates_y[2] = 0.0245436926061703;
    SourceCell[2].coordinates_y[3] = 0.0245436926061703;

    SourceCell[3].coordinates_x[0] = 4.0987966652304335;
    SourceCell[3].coordinates_x[1] = 4.1233403578366037;
    SourceCell[3].coordinates_x[2] = 4.1233403578366037;
    SourceCell[3].coordinates_x[3] = 4.0987966652304335;
    SourceCell[3].coordinates_y[0] = -0.0245436926061703;
    SourceCell[3].coordinates_y[1] = -0.0245436926061703;
    SourceCell[3].coordinates_y[2] =  0.0000000000000000;
    SourceCell[3].coordinates_y[3] =  0.0000000000000000;

    SourceCell[4].coordinates_x[0] = 4.0742529726242633;
    SourceCell[4].coordinates_x[1] = 4.0987966652304335;
    SourceCell[4].coordinates_x[2] = 4.0987966652304335;
    SourceCell[4].coordinates_x[3] = 4.0742529726242633;
    SourceCell[4].coordinates_y[0] = -0.0245436926061703;
    SourceCell[4].coordinates_y[1] = -0.0245436926061703;
    SourceCell[4].coordinates_y[2] =  0.0000000000000000;
    SourceCell[4].coordinates_y[3] =  0.0000000000000000;

    SourceCell[5].coordinates_x[0] = 4.0497092800180932;
    SourceCell[5].coordinates_x[1] = 4.0742529726242633;
    SourceCell[5].coordinates_x[2] = 4.0742529726242633;
    SourceCell[5].coordinates_x[3] = 4.0497092800180932;
    SourceCell[5].coordinates_y[0] = -0.0245436926061703;
    SourceCell[5].coordinates_y[1] = -0.0245436926061703;
    SourceCell[5].coordinates_y[2] =  0.0000000000000000;
    SourceCell[5].coordinates_y[3] =  0.0000000000000000;

    SourceCell[6].coordinates_x[0] = 4.0497092800180932;
    SourceCell[6].coordinates_x[1] = 4.0742529726242633;
    SourceCell[6].coordinates_x[2] = 4.0742529726242633;
    SourceCell[6].coordinates_x[3] = 4.0497092800180932;
    SourceCell[6].coordinates_y[0] = -0.0490873852123405;
    SourceCell[6].coordinates_y[1] = -0.0490873852123405;
    SourceCell[6].coordinates_y[2] = -0.0245436926061703;
    SourceCell[6].coordinates_y[3] = -0.0245436926061703;

    for (int i = 0; i < nSourceCells; ++i)
      for (int j = 0; j < SourceCell[i].num_corners; ++j)
        LLtoXYZ(SourceCell[i].coordinates_x[j], SourceCell[i].coordinates_y[j],
                SourceCell[i].coordinates_xyz + j * 3);
    for (int j = 0; j < TargetCell.num_corners; ++j)
      LLtoXYZ(TargetCell.coordinates_x[j], TargetCell.coordinates_y[j],
              TargetCell.coordinates_xyz + j * 3);

    yac_compute_overlap_areas ( nSourceCells, SourceCell, TargetCell, area );

    ref_area = yac_pole_area(TargetCell);
    for (n = 0; n < nSourceCells; ++n)
      ref_area -= area[n];

    if ( fabs(ref_area) > yac_area_tol() ) {
      PUT_ERR("test 12.) error in partial area computation\n");
    }

    for ( n = 0; n < nSourceCells; n++ ) {
      free(SourceCell[n].coordinates_x);
      free(SourceCell[n].coordinates_y);
      free(SourceCell[n].coordinates_xyz);
    }

    free (SourceCell);

    TargetCell.edge_type = quad_type;
    TargetCell.num_corners      = 4;
  }

#ifdef VERBOSE
  printf (" ---------- test 13.) ----------- \n");
#endif
  {
    enum yac_edge_type edge_types[4] = {LAT_CIRCLE, LON_CIRCLE,
                                        LAT_CIRCLE, LON_CIRCLE};
    nSourceCells = 4;

    SourceCell = malloc ( nSourceCells  * sizeof(*SourceCell) );

    TargetCell.edge_type = edge_types;

    for ( n = 0; n < nSourceCells; n++ ) {
      SourceCell[n].num_corners   = 3;
      SourceCell[n].edge_type     = quad_type;
      SourceCell[n].coordinates_x = malloc ( ncs * sizeof(SourceCell[n].coordinates_x[0]) );
      SourceCell[n].coordinates_y = malloc ( ncs * sizeof(SourceCell[n].coordinates_y[0]) );
      SourceCell[n].coordinates_xyz = malloc ( 3 * ncs * sizeof(SourceCell[n].coordinates_xyz[0]) );
    }


    TargetCell.coordinates_x[0] =  3.4361169648638361;
    TargetCell.coordinates_x[1] =  3.4606606574700067;
    TargetCell.coordinates_x[2] =  3.4606606574700067;
    TargetCell.coordinates_x[3] =  3.4361169648638361;
    TargetCell.coordinates_y[0] = -0.0245436926061703;
    TargetCell.coordinates_y[1] = -0.0245436926061703;
    TargetCell.coordinates_y[2] =  0.0000000000000000;
    TargetCell.coordinates_y[3] =  0.0000000000000000;
    TargetCell.num_corners      = 4;

    SourceCell[0].coordinates_x[0] = -2.8075010104343816;
    SourceCell[0].coordinates_x[1] = -2.8274333887952685;
    SourceCell[0].coordinates_x[2] = -2.8467002736213392;
    SourceCell[0].coordinates_y[0] = -0.0338654341705874;
    SourceCell[0].coordinates_y[1] =  0.0000000011622542;
    SourceCell[0].coordinates_y[2] = -0.0311624540855508;

    SourceCell[1].coordinates_x[0] = -2.8081665040027728;
    SourceCell[1].coordinates_x[1] = -2.8274333887952685;
    SourceCell[1].coordinates_x[2] = -2.7882305020025413;
    SourceCell[1].coordinates_y[0] =  0.0311624566805337;
    SourceCell[1].coordinates_y[1] =  0.0000000011622542;
    SourceCell[1].coordinates_y[2] = -0.0026744729717723;

    SourceCell[2].coordinates_x[0] = -2.7882305020025413;
    SourceCell[2].coordinates_x[1] = -2.8274333887952685;
    SourceCell[2].coordinates_x[2] = -2.8075010104343816;
    SourceCell[2].coordinates_y[0] = -0.0026744729717723;
    SourceCell[2].coordinates_y[1] =  0.0000000011622542;
    SourceCell[2].coordinates_y[2] = -0.0338654341705874;

    SourceCell[3].coordinates_x[0] = -2.8467002736213392;
    SourceCell[3].coordinates_x[1] = -2.8274333887952685;
    SourceCell[3].coordinates_x[2] = -2.8666362753717674;
    SourceCell[3].coordinates_y[0] = -0.0311624540855508;
    SourceCell[3].coordinates_y[1] =  0.0000000011622542;
    SourceCell[3].coordinates_y[2] =  0.0026744751289656;

    for (int i = 0; i < nSourceCells; ++i)
      for (int j = 0; j < SourceCell[i].num_corners; ++j)
        LLtoXYZ(SourceCell[i].coordinates_x[j], SourceCell[i].coordinates_y[j],
                SourceCell[i].coordinates_xyz + j * 3);
    for (int j = 0; j < TargetCell.num_corners; ++j)
      LLtoXYZ(TargetCell.coordinates_x[j], TargetCell.coordinates_y[j],
              TargetCell.coordinates_xyz + j * 3);

    yac_compute_overlap_areas ( nSourceCells, SourceCell, TargetCell, area );

    ref_area = yac_pole_area(TargetCell);
    for (n = 0; n < nSourceCells; ++n)
      ref_area -= area[n];

    if ( fabs(ref_area) > yac_area_tol() ) {
      PUT_ERR("test 13.) error in partial area computation\n");
    }

    for ( n = 0; n < nSourceCells; n++ ) {
      free(SourceCell[n].coordinates_x);
      free(SourceCell[n].coordinates_y);
      free(SourceCell[n].coordinates_xyz);
    }

    free (SourceCell);

    TargetCell.edge_type = quad_type;
    TargetCell.num_corners      = 4;
  }

#ifdef VERBOSE
  printf (" ---------- test 14.) ----------- \n");
#endif
  {
    enum yac_edge_type edge_types[4] = {LAT_CIRCLE, LON_CIRCLE,
                                        LAT_CIRCLE, LON_CIRCLE};
    nSourceCells = 3;

    SourceCell = malloc ( nSourceCells  * sizeof(*SourceCell) );

    TargetCell.edge_type = edge_types;

    for ( n = 0; n < nSourceCells; n++ ) {
      SourceCell[n].num_corners   = 3;
      SourceCell[n].edge_type     = quad_type;
      SourceCell[n].coordinates_x = malloc ( ncs * sizeof(SourceCell[n].coordinates_x[0]) );
      SourceCell[n].coordinates_y = malloc ( ncs * sizeof(SourceCell[n].coordinates_y[0]) );
      SourceCell[n].coordinates_xyz = malloc ( 3 * ncs * sizeof(SourceCell[n].coordinates_xyz[0]) );
    }

    TargetCell.coordinates_x[0] =  3.1170489609836229;
    TargetCell.coordinates_x[1] =  3.1415926535897931;
    TargetCell.coordinates_x[2] =  3.1415926535897931;
    TargetCell.coordinates_x[3] =  3.1170489609836229;
    TargetCell.coordinates_y[0] = -1.0062913968529805;
    TargetCell.coordinates_y[1] = -1.0062913968529805;
    TargetCell.coordinates_y[2] = -0.9817477042468103;
    TargetCell.coordinates_y[3] = -0.9817477042468103;
    TargetCell.num_corners      = 4;

    SourceCell[0].coordinates_x[0] =  3.1415926521557882;
    SourceCell[0].coordinates_x[1] =  3.1415926520689506;
    SourceCell[0].coordinates_x[2] = -3.0774466652769172;
    SourceCell[0].coordinates_y[0] = -0.9805860393425995;
    SourceCell[0].coordinates_y[1] = -1.0172219678978514;
    SourceCell[0].coordinates_y[2] = -0.9979427097227050;

    SourceCell[1].coordinates_x[0] =  3.0774466624490269;
    SourceCell[1].coordinates_x[1] =  3.1415926520689506;
    SourceCell[1].coordinates_x[2] =  3.1415926521557882;
    SourceCell[1].coordinates_y[0] = -0.9979427096430752;
    SourceCell[1].coordinates_y[1] = -1.0172219678978514;
    SourceCell[1].coordinates_y[2] = -0.9805860393425995;

    SourceCell[2].coordinates_x[0] =  3.0808931330453300;
    SourceCell[2].coordinates_x[1] =  3.0774466624490269;
    SourceCell[2].coordinates_x[2] =  3.1415926521557882;
    SourceCell[2].coordinates_y[0] = -0.9613498550843829;
    SourceCell[2].coordinates_y[1] = -0.9979427096430752;
    SourceCell[2].coordinates_y[2] = -0.9805860393425995;

    for (int i = 0; i < nSourceCells; ++i)
      for (int j = 0; j < SourceCell[i].num_corners; ++j)
        LLtoXYZ(SourceCell[i].coordinates_x[j], SourceCell[i].coordinates_y[j],
                SourceCell[i].coordinates_xyz + j * 3);
    for (int j = 0; j < TargetCell.num_corners; ++j)
      LLtoXYZ(TargetCell.coordinates_x[j], TargetCell.coordinates_y[j],
              TargetCell.coordinates_xyz + j * 3);

    yac_compute_overlap_areas ( nSourceCells, SourceCell, TargetCell, area );

    ref_area = yac_pole_area(TargetCell);
    for (n = 0; n < nSourceCells; ++n)
      ref_area -= area[n];

    if ( fabs(ref_area) > yac_area_tol() ) {
      PUT_ERR("test 14.) error in partial area computation\n");
    }

    for ( n = 0; n < nSourceCells; n++ ) {
      free(SourceCell[n].coordinates_x);
      free(SourceCell[n].coordinates_y);
      free(SourceCell[n].coordinates_xyz);
    }

    free (SourceCell);

    TargetCell.edge_type = quad_type;
    TargetCell.num_corners      = 4;
  }

#ifdef VERBOSE
  printf (" ---------- test 15.) ----------- \n");
#endif
  {
    enum yac_edge_type edge_types[4] = {LAT_CIRCLE, LON_CIRCLE,
                                        LAT_CIRCLE, LON_CIRCLE};
    nSourceCells = 198;

    SourceCell = malloc ( nSourceCells  * sizeof(*SourceCell) );

    TargetCell.edge_type = quad_type;

    for ( n = 0; n < nSourceCells; n++ ) {
      SourceCell[n].num_corners   = 4;
      SourceCell[n].edge_type     = edge_types;
      SourceCell[n].coordinates_x = malloc ( ncs * sizeof(SourceCell[n].coordinates_x[0]) );
      SourceCell[n].coordinates_y = malloc ( ncs * sizeof(SourceCell[n].coordinates_y[0]) );
      SourceCell[n].coordinates_xyz = malloc ( 3 * ncs * sizeof(SourceCell[n].coordinates_xyz[0]) );
    }

    TargetCell.num_corners      = 3;
    TargetCell.coordinates_x[0] = 22.8313*rad;
    TargetCell.coordinates_y[0] = 89.7418*rad;
    TargetCell.coordinates_x[1] = 49.1687*rad;
    TargetCell.coordinates_y[1] = 89.7418*rad;
    TargetCell.coordinates_x[2] = 36*rad;
    TargetCell.coordinates_y[2] = 89.8335*rad;

    double dx = ((49.2 - 22.8) /((double)nSourceCells))*rad;

    for (int i = 0; i < nSourceCells; ++i) {
      SourceCell[i].coordinates_x[0] = 22.8*rad + dx*((double)i);
      SourceCell[i].coordinates_y[0] = 89.7333*rad;
      SourceCell[i].coordinates_x[1] = 22.8*rad + dx*((double)(i+1));
      SourceCell[i].coordinates_y[1] = 89.7333*rad;
      SourceCell[i].coordinates_x[2] = 22.8*rad + dx*((double)(i+1));
      SourceCell[i].coordinates_y[2] = 89.8667*rad;
      SourceCell[i].coordinates_x[3] = 22.8*rad + dx*((double)i);
      SourceCell[i].coordinates_y[3] = 89.8667*rad;
    }

    for (int i = 0; i < nSourceCells; ++i)
      for (int j = 0; j < SourceCell[i].num_corners; ++j)
        LLtoXYZ(SourceCell[i].coordinates_x[j], SourceCell[i].coordinates_y[j],
                SourceCell[i].coordinates_xyz + j * 3);
    for (int j = 0; j < TargetCell.num_corners; ++j)
      LLtoXYZ(TargetCell.coordinates_x[j], TargetCell.coordinates_y[j],
              TargetCell.coordinates_xyz + j * 3);

    yac_compute_overlap_areas ( nSourceCells, SourceCell, TargetCell, area );

    ref_area = yac_pole_area(TargetCell);
    for (n = 0; n < nSourceCells; ++n)
      ref_area -= area[n];

    if ( fabs(ref_area) > yac_area_tol() ) {
      PUT_ERR("test 15.) error in partial area computation\n");
    }

    for ( n = 0; n < nSourceCells; n++ ) {
      free(SourceCell[n].coordinates_x);
      free(SourceCell[n].coordinates_y);
      free(SourceCell[n].coordinates_xyz);
    }

    free (SourceCell);

    TargetCell.edge_type = quad_type;
    TargetCell.num_corners      = 4;
  }

#ifdef VERBOSE
  printf (" ---------- test 16.) ----------- \n");
#endif
  {
    struct grid_cell SourceCell[] =
      {{.coordinates_x = (double[]){4.9999997254688875,
                                    4.9999997254688875,
                                    5.0078539201557488,
                                    5.0078539201557488},
        .coordinates_y = (double[]){-1.2878426515089207,
                                    -1.2946756570757916,
                                    -1.2946797849754812,
                                    -1.2878469125666649},
        .coordinates_xyz = (double[4*3]){-1},
        .edge_type = quad_type,
        .num_corners = 4},
       {.coordinates_x = (double[]){5.0078539201557488,
                                    5.0078539201557488,
                                    5.0157075822103927,
                                    5.0157075822103927},
        .coordinates_y = (double[]){-1.2946797849754812,
                                    -1.3015127905423520,
                                    -1.3015170516000960,
                                    -1.2946840460332254},
        .coordinates_xyz = (double[4*3]){-1},
        .edge_type = quad_type,
        .num_corners = 4},
       {.coordinates_x = (double[]){5.0078539201557488,
                                    5.0078539201557488,
                                    5.0157075822103927,
                                    5.0157075822103927},
        .coordinates_y = (double[]){-1.2878469125666649,
                                    -1.2946797849754812,
                                    -1.2946840460332254,
                                    -1.2878513067824635},
        .coordinates_xyz = (double[4*3]){-1},
        .edge_type = quad_type,
        .num_corners = 4},
       {.coordinates_x = (double[]){4.9999997254688875,
                                    4.9999997254688875,
                                    5.0078539201557488,
                                    5.0078539201557488},
        .coordinates_y = (double[]){-1.2946756570757916,
                                    -1.3015085294846078,
                                    -1.3015127905423520,
                                    -1.2946797849754812},
        .coordinates_xyz = (double[4*3]){-1},
        .edge_type = quad_type,
        .num_corners = 4}};

    struct grid_cell TargetCell =
      {.coordinates_x = (double[]){5.0069132916587327,
                                   5.0130492148102759,
                                   5.0130492148102759,
                                   5.0069132916587327},
       .coordinates_y = (double[]){-1.2946797849754812,
                                   -1.2946797849754812,
                                   -1.2885438618239387,
                                   -1.2885438618239387},
       .coordinates_xyz = (double[4*3]){-1},
       .edge_type = (enum yac_edge_type[]){LAT_CIRCLE, LON_CIRCLE,
                                           LAT_CIRCLE, LON_CIRCLE},
       .num_corners = 4};

    unsigned nSourceCells = sizeof(SourceCell)/sizeof(SourceCell[0]);

    double area[nSourceCells];

    for (int i = 0; i < nSourceCells; ++i)
      for (int j = 0; j < SourceCell[i].num_corners; ++j)
        LLtoXYZ(SourceCell[i].coordinates_x[j], SourceCell[i].coordinates_y[j],
                SourceCell[i].coordinates_xyz + j * 3);
    for (int j = 0; j < TargetCell.num_corners; ++j)
      LLtoXYZ(TargetCell.coordinates_x[j], TargetCell.coordinates_y[j],
              TargetCell.coordinates_xyz + j * 3);

    yac_compute_overlap_areas ( nSourceCells, SourceCell, TargetCell, area );

    double ref_area = yac_pole_area(TargetCell);
    for (unsigned n = 0; n < nSourceCells; ++n)
      ref_area -= area[n];

    if ( fabs(ref_area) > yac_area_tol() )
      PUT_ERR("test 16.) error in partial area computation\n");
  }

#ifdef VERBOSE
  printf (" ---------- test 17.) ----------- \n");
#endif
  {
    struct grid_cell SourceCell[] =
      {{.coordinates_x = (double[]){-2.8655528147976832,
                                    -2.9038195735214831,
                                    -2.8254194450162888},
        .coordinates_y = (double[]){-0.0615856127966505,
                                    -0.1228534567110898,
                                    -0.1292806814915037},
        .coordinates_xyz = (double[3*3]){-1},
        .edge_type = quad_type,
        .num_corners = 3},
       {.coordinates_x = (double[]){-2.9440830021466091,
                                    -2.9038195735214831,
                                    -2.8655528147976832},
        .coordinates_y = (double[]){-0.0558979744839555,
                                    -0.1228534567110898,
                                    -0.0615856127966505},
        .coordinates_xyz = (double[3*3]){-1},
        .edge_type = quad_type,
        .num_corners = 3}};

    struct grid_cell TargetCell =
      {.coordinates_x = (double[]){-2.8460897445189417,
                                   -2.8659852121839959,
                                   -2.8853055611946927},
       .coordinates_y = (double[]){-0.0961602003314924,
                                   -0.0622820530745658,
                                   -0.0933150814318732},
       .coordinates_xyz = (double[3*3]){-1},
       .edge_type = quad_type,
       .num_corners = 3};

    unsigned nSourceCells = sizeof(SourceCell)/sizeof(SourceCell[0]);

    double area[nSourceCells];

    for (int i = 0; i < nSourceCells; ++i)
      for (int j = 0; j < SourceCell[i].num_corners; ++j)
        LLtoXYZ(SourceCell[i].coordinates_x[j], SourceCell[i].coordinates_y[j],
                SourceCell[i].coordinates_xyz + j * 3);
    for (int j = 0; j < TargetCell.num_corners; ++j)
      LLtoXYZ(TargetCell.coordinates_x[j], TargetCell.coordinates_y[j],
              TargetCell.coordinates_xyz + j * 3);

    yac_compute_overlap_areas(nSourceCells, SourceCell, TargetCell, area );

    double ref_area = yac_pole_area(TargetCell);
    for (unsigned n = 0; n < nSourceCells; ++n)
      ref_area -= area[n];

    if ( fabs(ref_area) > yac_area_tol() )
      PUT_ERR("test 17.) error in partial area computation\n");
  }

#ifdef VERBOSE
  printf (" ---------- test 18.) ----------- \n");
#endif
  {
    struct grid_cell SourceCell[] =
      {{.coordinates_x = (double[]){2.9198084624304621,
                                    2.9720969737232750,
                                    3.0078074923914948},
        .coordinates_y = (double[]){0.6123031304013200,
                                    0.5471403770103002,
                                    0.6192726969308253},
        .coordinates_xyz = (double[3*3]){-1},
        .edge_type = quad_type,
        .num_corners = 3},
       {.coordinates_x = (double[]){2.8889398582596519,
                                    2.9720969737232750,
                                    2.9198084624304621},
        .coordinates_y = (double[]){0.5392518296263771,
                                    0.5471403770103002,
                                    0.6123031304013200},
        .coordinates_xyz = (double[3*3]){-1},
        .edge_type = quad_type,
        .num_corners = 3},
       {.coordinates_x = (double[]){3.0187962726049782,
                                    2.9720969737232750,
                                    2.9378118571213374},
        .coordinates_y = (double[]){0.4805223860987507,
                                    0.5471403770103002,
                                    0.4745836829538209},
        .coordinates_xyz = (double[3*3]){-1},
        .edge_type = quad_type,
        .num_corners = 3},
       {.coordinates_x = (double[]){2.9378118571213374,
                                    2.9720969737232750,
                                    2.8889398582596519},
        .coordinates_y = (double[]){0.4745836829538209,
                                    0.5471403770103002,
                                    0.5392518296263771},
        .coordinates_xyz = (double[3*3]){-1},
        .edge_type = quad_type,
        .num_corners = 3},
       {.coordinates_x = (double[]){3.0565195473916882,
                                    2.9720969737232750,
                                    3.0187962726049782},
        .coordinates_y = (double[]){0.5519553785110132,
                                    0.5471403770103002,
                                    0.4805223860987507},
        .coordinates_xyz = (double[3*3]){-1},
        .edge_type = quad_type,
        .num_corners = 3},
       {.coordinates_x = (double[]){3.0078074923914948,
                                    2.9720969737232750,
                                    3.0565195473916882},
        .coordinates_y = (double[]){0.6192726969308253,
                                    0.5471403770103002,
                                    0.5519553785110132},
        .coordinates_xyz = (double[3*3]){-1},
        .edge_type = quad_type,
        .num_corners = 3}};

    struct grid_cell TargetCell =
      {.coordinates_x = (double[]){3.0127596408133406,
                                   2.9702182441873108,
                                   2.9943996418061438},
       .coordinates_y = (double[]){0.5498596133493898,
                                   0.5469967117356207,
                                   0.5140000458840724},
       .coordinates_xyz = (double[3*3]){-1},
       .edge_type = quad_type,
       .num_corners = 3};

    unsigned nSourceCells = sizeof(SourceCell)/sizeof(SourceCell[0]);

    double area[nSourceCells];

    for (int i = 0; i < nSourceCells; ++i)
      for (int j = 0; j < SourceCell[i].num_corners; ++j)
        LLtoXYZ(SourceCell[i].coordinates_x[j], SourceCell[i].coordinates_y[j],
                SourceCell[i].coordinates_xyz + j * 3);
    for (int j = 0; j < TargetCell.num_corners; ++j)
      LLtoXYZ(TargetCell.coordinates_x[j], TargetCell.coordinates_y[j],
              TargetCell.coordinates_xyz + j * 3);

    yac_compute_overlap_areas(nSourceCells, SourceCell, TargetCell, area );

    double ref_area = yac_pole_area(TargetCell);
    for (unsigned n = 0; n < nSourceCells; ++n)
      ref_area -= area[n];

    if ( fabs(ref_area) > yac_area_tol() )
      PUT_ERR("test 18.) error in partial area computation\n");
  }

  free (TargetCell.coordinates_x);
  free (TargetCell.coordinates_y);
  free (TargetCell.coordinates_xyz);

  return TEST_EXIT_CODE;
}
