/**
 * @file test_point_sphere_part.c
 *
 * @copyright Copyright  (C)  2016 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include <stdlib.h>
#include <stdio.h>

#include "tests.h"
#include "grid_search.h"
#include "read_icon_grid.h"
#include "grid.h"
#include "sphere_part.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

void random_test();
void single_point_test();
void empty_branch_test();

int main(void) {

  random_test();
  single_point_test();
  empty_branch_test();

  return TEST_EXIT_CODE;
}

struct distance_index {
  double distance;
  unsigned index;
};

static int compare_distance_index(const void * a, const void * b) {

  int ret = (((struct distance_index*)a)->distance >
             ((struct distance_index*)b)->distance) -
            (((struct distance_index*)a)->distance <
             ((struct distance_index*)b)->distance);
  if (ret == 0)
    ret = (((struct distance_index*)a)->index >
           ((struct distance_index*)b)->index) -
          (((struct distance_index*)a)->index <
           ((struct distance_index*)b)->index);

  return ret;
}

void random_test() {

  unsigned const num_tests = 100;
  unsigned const num_points = 100;

  double * x_coordinates[2];
  double * y_coordinates[2];
  double * xyz_coordinates[2];

  x_coordinates[0] = malloc(4 * num_points * sizeof(*x_coordinates[0])) ; 
  x_coordinates[1] = x_coordinates[0] + 1 * num_points;
  y_coordinates[0] = x_coordinates[0] + 2 * num_points;
  y_coordinates[1] = x_coordinates[0] + 3 * num_points;
  xyz_coordinates[0] = malloc(2 * 3 * num_points * sizeof(*xyz_coordinates[0]));
  xyz_coordinates[1] = xyz_coordinates[0] + 3 * num_points;

  unsigned * local_point_ids = NULL;
  unsigned local_point_ids_array_size = 0;
  unsigned * num_local_point_ids =
    malloc(num_points * sizeof(*num_local_point_ids));
  struct distance_index * distances =
    malloc(num_points * num_points * sizeof(*distances));

  srand(1365441);

  for (unsigned i = 0; i < num_tests; ++i) {

    for (unsigned j = 0; j < 2; ++j) {
      for (unsigned k = 0; k < num_points; ++k) {

        x_coordinates[j][k] =
          2.0 * M_PI * (((double)rand()) / ((double)RAND_MAX));
        y_coordinates[j][k] =
          M_PI * (((double)rand()) / ((double)RAND_MAX)) - M_PI_2;

        LLtoXYZ(x_coordinates[j][k], y_coordinates[j][k],
                xyz_coordinates[j] + 3 * k);
      }
    }

    for (unsigned j = 0; j < num_points; ++j) {
      for (unsigned k = 0; k < num_points; ++k) {

        distances[j * num_points + k].distance =
        get_vector_angle(xyz_coordinates[1] + j * 3,
                         xyz_coordinates[0] + k * 3);
        distances[j * num_points + k].index = k;
      }
      qsort(distances + j * num_points, num_points, sizeof(*distances),
            compare_distance_index);
    }

    struct point_sphere_part_search * search =
      yac_point_sphere_part_search_new(num_points, x_coordinates[0],
                                       y_coordinates[0]);

    for (unsigned n = 1; n < 16; ++n) {

      yac_point_sphere_part_search_NNN(search, num_points, x_coordinates[1],
                                       y_coordinates[1], n, NULL, NULL,
                                       &local_point_ids,
                                       &local_point_ids_array_size,
                                       num_local_point_ids);


      // check results
      for (unsigned j = 0, offset = 0; j < num_points; ++j) {

        struct distance_index * curr_distances = distances + j * num_points;

        unsigned ref_num_local_points = 1, m = 0;
        for (; (ref_num_local_points < num_points); ++ref_num_local_points)
          if (curr_distances[ref_num_local_points-1].distance !=
              curr_distances[ref_num_local_points].distance)
            if (++m == n) break;

        if (num_local_point_ids[j] != ref_num_local_points) {
          PUT_ERR("wrong number of local points found\n");
          continue;
        }

        unsigned num_matching_points = 0;
        for (unsigned k = 0; k < ref_num_local_points; ++k)
          for (unsigned l = 0; l < ref_num_local_points; ++l)
            if (curr_distances[k].index == local_point_ids[offset+l])
              ++num_matching_points;

        if (num_matching_points != ref_num_local_points)
          PUT_ERR("wrong results\n");

        offset += ref_num_local_points;
      }
    }
    yac_delete_point_sphere_part_search(search);
  }

  free(distances);
  free(local_point_ids);
  free(num_local_point_ids);
  free(xyz_coordinates[0]);
  free(x_coordinates[0]);
}

void single_point_test() {

  struct point_sphere_part_search * search;
  {
    double x_coordinate = 0, y_coordinate = 0;
    search = yac_point_sphere_part_search_new(1, &x_coordinate, &y_coordinate);
  }

  double x_coordinates[] =
    {0.0*rad, 90.0*rad, 180.0*rad, -90.0 * rad,  0.0*rad,   0.0*rad};
  double y_coordinates[] =
    {0.0*rad,  0.0*rad,   0.0*rad,   0.0 * rad, 90.0*rad, -90.0*rad};
  unsigned num_points = sizeof(x_coordinates) / sizeof(x_coordinates[0]);

  unsigned * local_point_ids = NULL;
  unsigned local_point_ids_array_size = 0;
  unsigned * num_local_point_ids =
    malloc(num_points * sizeof(*num_local_point_ids));

  yac_point_sphere_part_search_NN(search, num_points, x_coordinates,
                                  y_coordinates, NULL, &local_point_ids,
                                  &local_point_ids_array_size,
                                  num_local_point_ids);

  for (unsigned i = 0; i < num_points; ++i)
    if ((num_local_point_ids[i] != 1) || (local_point_ids[i] != 0))
      PUT_ERR("single_point_test: wrong result\n");

  free(local_point_ids);
  free(num_local_point_ids);
  yac_delete_point_sphere_part_search(search);
}

void empty_branch_test() {

  struct point_sphere_part_search * search;

  double x_coordinates_a[] = {0,0,0,0};
  double y_coordinates_a[] = {1.0*rad, 2.0*rad, 3.0*rad, 4.0*rad};
  unsigned num_points_a = sizeof(x_coordinates_a) / sizeof(x_coordinates_a[0]);
  search = yac_point_sphere_part_search_new(num_points_a, x_coordinates_a,
                                            y_coordinates_a);


  double x_coordinates_b[] =
    {-1.0*rad, -1.0*rad,  1.0*rad, 1.0*rad};
  double y_coordinates_b[] =
    {-1.0*rad,  1.0*rad, -1.0*rad, 1.0*rad};
  unsigned num_points_b = sizeof(x_coordinates_a) / sizeof(x_coordinates_a[0]);

  unsigned * local_point_ids = NULL;
  unsigned local_point_ids_array_size = 0;
  unsigned * num_local_point_ids =
    malloc(num_points_b * sizeof(*num_local_point_ids));

  struct distance_index distances[num_points_a * num_points_b];

  for (unsigned i = 0; i < num_points_b; ++i) {

    double xyz_coordinates_b[3];
    LLtoXYZ(x_coordinates_b[i], y_coordinates_b[i], xyz_coordinates_b);

    for (unsigned j = 0; j < num_points_a; ++j) {

      double xyz_coordinates_a[3];
      LLtoXYZ(x_coordinates_a[j], y_coordinates_a[j], xyz_coordinates_a);

      distances[i * num_points_a + j].distance =
        get_vector_angle(xyz_coordinates_a, xyz_coordinates_b);
      distances[i * num_points_a + j].index = j;
    }
    qsort(distances + i * num_points_a, num_points_a, sizeof(*distances),
          compare_distance_index);
  }

  yac_point_sphere_part_search_NN(search, num_points_b, x_coordinates_b,
                                  y_coordinates_b, NULL, &local_point_ids,
                                  &local_point_ids_array_size,
                                  num_local_point_ids);

    // check results
  for (unsigned j = 0, offset = 0; j < num_points_b; ++j) {

    struct distance_index * curr_distances = distances + j * num_points_a;

    unsigned ref_num_local_points;
    for (ref_num_local_points = 1;
         (ref_num_local_points < num_points_a); ++ref_num_local_points)
      if (curr_distances[ref_num_local_points-1].distance !=
          curr_distances[ref_num_local_points].distance) break;

    if (num_local_point_ids[j] != ref_num_local_points) {
      PUT_ERR("wrong number of local points found\n");
      continue;
    }

    unsigned num_matching_points = 0;
    for (unsigned k = 0; k < ref_num_local_points; ++k)
      for (unsigned l = 0; l < ref_num_local_points; ++l)
        if (curr_distances[k].index == local_point_ids[offset+l])
          ++num_matching_points;

    if (num_matching_points != ref_num_local_points)
      PUT_ERR("wrong results\n");

    offset += ref_num_local_points;
  }

  free(local_point_ids);
  free(num_local_point_ids);
  yac_delete_point_sphere_part_search(search);
}
