/**
 * @file test_read_fesom.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_C_NETCDF
#include <netcdf.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <unistd.h>

#include "tests.h"
#include "read_fesom_grid.h"
#include "vtk_output.h"
#include "grid.h"
#include "geometry.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

void write_grid_to_file(struct grid * grid, char * name);
int file_exists(const char * filename);

#ifdef HAVE_C_NETCDF

void write_dummy_grid_file(char * name);
void check_grid(struct grid * grid);

int main(void) {

   write_dummy_grid_file("fesom_grid.nc");

   struct grid * fesom_grid;

   fesom_grid = read_fesom_grid("fesom_grid.nc");

   unlink("fesom_grid.nc");

   check_grid(fesom_grid);

#define WRITE_VTK_GRID_FILE
#ifdef WRITE_VTK_GRID_FILE
   write_grid_to_file(fesom_grid, "fesom");
#endif // WRITE_VTK_GRID_FILE

   yac_delete_grid(fesom_grid);

   return TEST_EXIT_CODE;
}

void get_grid_xyz_coordinates(struct grid * grid, double * points) {

   unsigned i, num_corners;
   double lon, lat;

   double const rad = M_PI / 180.0;

   num_corners = yac_get_num_grid_corners(grid);

   for (i = 0; i < num_corners; ++i) {
      lon = yac_get_corner_x_coord(grid, i) * rad;
      lat = yac_get_corner_y_coord(grid, i) * rad;
      LLtoXYZ(lon, lat, points + 3 * i);
   }
}

void get_grid_cell_data(struct grid * grid, unsigned * cell_data) {

   unsigned i, j, k;
   unsigned num_grid_cells;

   num_grid_cells = yac_get_num_grid_cells(grid);

   for (i = 0, k = 0; i < num_grid_cells; ++i) {

      unsigned const * curr_cell_corners;
      unsigned num_cell_corners;

      curr_cell_corners = yac_get_cell_corner_indices(grid, i);
      num_cell_corners = yac_get_num_cell_corners(grid, i);

      for (j = 0; j < num_cell_corners; ++j)
         cell_data[k++] = curr_cell_corners[j];
   }
}

void get_equator_point_data(unsigned num_points, double * points) {

   // the equator is approximated by a polygon with num_points corners

   double angle;

   angle = 2.0 * M_PI / (num_points + 1);

   unsigned i;

   for (i = 0; i < num_points; ++i) {

      points[3*i + 0] = sin(angle * i);
      points[3*i + 1] = cos(angle * i);
      points[3*i + 2] = 0;
   }
}

void write_grid_to_file(struct grid * grid, char * name) {

   unsigned num_grid_corners;
   unsigned num_grid_cells;

   num_grid_corners = yac_get_num_grid_corners(grid);
   num_grid_cells = yac_get_num_grid_cells(grid);

   unsigned const num_points_per_great_circle = 71;

   //-------------------------------------
   // generate point data for the vtk file
   // it will contain data for all corners of the grid and the points for the equator
   //-------------------------------------
   double * points;

   points = malloc(3 * (num_grid_corners + num_points_per_great_circle) * sizeof(*points));

   // get the point data of the grid
   get_grid_xyz_coordinates(grid, points);
   // get the point data of the equator
   get_equator_point_data(num_points_per_great_circle, points + 3 * num_grid_corners);

   //------------------------------------
   // generate cell data for the vtk file
   //------------------------------------
   unsigned * num_points_per_cell;
   unsigned * cell_data;

   // has entries for all cells of the grid and one for the equator
   num_points_per_cell = malloc((num_grid_cells + 1) * sizeof(*num_points_per_cell));

   unsigned i, total_num_cell_points;

   // get the number of points per grid cell
   for (i = 0; i < num_grid_cells; ++i)
      num_points_per_cell[i] = yac_get_num_cell_corners(grid, i);

   // get the number of corners for the equator
   num_points_per_cell[num_grid_cells] = num_points_per_great_circle;

   // count the total number of cell points
   total_num_cell_points = 0;
   for (i = 0; i <= num_grid_cells; ++i)
      total_num_cell_points += num_points_per_cell[i];

   cell_data = malloc(total_num_cell_points * sizeof(*cell_data));

   // get the cell data of the grid
   get_grid_cell_data(grid, cell_data);

   // write the cell data for the great circle
   for (i = 0; i < num_points_per_great_circle; ++i)
      cell_data[i + total_num_cell_points - num_points_per_great_circle] = num_grid_corners + i;

   //------------------------------------
   // generate scalar data
   // we write a cell type that differentiates grid cells and the equator
   //------------------------------------
   unsigned * cell_type;

   cell_type = malloc((num_grid_cells + 1) * sizeof(*cell_type));

   for (i = 0; i < num_grid_cells; ++i)
      cell_type[i] = 0;
   cell_type[num_grid_cells] = 1;

   //------------------------------------
   // generate the file name
   //------------------------------------

   char * filename;

   filename = calloc((strlen(name) + 5), sizeof(*filename));
   strcpy(filename, name);
   strcat(filename, ".vtk");

   //------------------------------------
   // generate the actual vtk file
   //------------------------------------
   VTK_FILE * file;

   file = vtk_open(filename, "grid data");
   vtk_write_point_data(file, points, num_grid_corners + num_points_per_great_circle);
   vtk_write_cell_data(file, cell_data, num_points_per_cell, num_grid_cells + 1);
   vtk_write_cell_scalars_uint(file, cell_type, num_grid_cells + 1, "cell_type");
   
   vtk_close(file);

   //------------------------------------
   // some final cleanup
   //------------------------------------

   free(filename);
   free(cell_type);
   free(cell_data);
   free(num_points_per_cell);
   free(points);
}

int file_exists(const char * filename) {

   FILE * file = fopen(filename, "r");

   if (file != NULL) {
      fclose(file);
      return 1;
   }
   return 0;
}

static void handle_error(int status) {
  const char *err_string;
  err_string = nc_strerror(status);
  printf ("%s \n", err_string);
  if (status != NC_NOERR) {
    exit(EXIT_FAILURE);
  }
}

void write_dummy_grid_file(char * file_name) {

  int ncid, status;

  // create file
  status = nc_create(file_name, NC_CLOBBER, &ncid);
  if (status != NC_NOERR) handle_error(status);

  int dim_ncells_id;
  int dim_nv_id;
  int dim_ids[2];

  // define dimensions
  status = nc_def_dim(ncid, "ncells", 8, &dim_ncells_id);
  if (status != NC_NOERR) handle_error(status);
  status = nc_def_dim(ncid, "nv", 9, &dim_nv_id);
  if (status != NC_NOERR) handle_error(status);

  dim_ids[0] = dim_ncells_id;
  dim_ids[1] = dim_nv_id;

  int var_lon_id, var_lon_v_id, var_lat_id, var_lat_v_id;

  // define grid arrays
  status = nc_def_var(ncid, "lon", NC_DOUBLE, 1, &dim_ncells_id, &var_lon_id);
  if (status != NC_NOERR) handle_error(status);
  status = nc_def_var(ncid, "lon_vertices", NC_DOUBLE, 2, dim_ids, &var_lon_v_id);
  status = nc_def_var(ncid, "lat", NC_DOUBLE, 1, &dim_ncells_id, &var_lat_id);
  if (status != NC_NOERR) handle_error(status);
  status = nc_def_var(ncid, "lat_vertices", NC_DOUBLE, 2, dim_ids, &var_lat_v_id);
  if (status != NC_NOERR) handle_error(status);

  // end definition
  status = nc_enddef(ncid);
  if (status != NC_NOERR) handle_error(status);

  // write grid data

  double lon[8] = {1,2,3,4,3,3,5.5,5.5};
  double lon_vertices[8*9] = {0,1,2,2,1,0,0,0,0,
                              1,3,2,2,2,2,2,2,2,
                              3,4,2,2,2,2,2,2,2,
                              3,5,4,4,4,4,4,4,4,
                              2,4,4,2,2,2,2,2,2,
                              1,2,4,5,4,2,2,2,2,
                              4,6,7,6,5,4,4,4,4,
                              5,6,4,4,4,4,4,4,4};
  double lat[8] = {1.5,0.5,0.5,0.5,1.5,3,1.5,0.5};
  double lat_vertices[8*9] = {1,0,1,2,3,2,2,2,2,
                              0,0,1,1,1,1,1,1,1,
                              0,1,1,1,1,1,1,1,1,
                              0,0,1,1,1,1,1,1,1,
                              1,1,2,2,2,2,2,2,2,
                              3,2,2,3,4,4,4,4,4,
                              1,1,2,3,3,2,2,2,2,
                              0,1,1,1,1,1,1,1,1};

  status = nc_put_var_double(ncid, var_lon_id, lon);
  if (status != NC_NOERR) handle_error(status);
  status = nc_put_var_double(ncid, var_lon_v_id, lon_vertices);
  if (status != NC_NOERR) handle_error(status);
  status = nc_put_var_double(ncid, var_lat_id, lat);
  if (status != NC_NOERR) handle_error(status);
  status = nc_put_var_double(ncid, var_lat_v_id, lat_vertices);
  if (status != NC_NOERR) handle_error(status);

  status = nc_close(ncid);
  if (status != NC_NOERR) handle_error(status);
}

void check_grid(struct grid * grid) {

  unsigned ref_num_cells = 8;

  if (yac_get_num_grid_cells(grid) != ref_num_cells)
    PUT_ERR("wrong number of grid cells");

  unsigned ref_num_corners_per_cell[8] = {6,3,3,3,4,6,6,3};

  for (unsigned i = 0; i < ref_num_cells; ++i)
    if (yac_get_num_cell_corners(grid, i) != ref_num_corners_per_cell[i])
      PUT_ERR("wrong number of corners per CELL");
}
#else

int main(void) {
   return EXIT_SKIP_TEST;
}
#endif

