/**
 * @file test_read_icon.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "tests.h"
#include "read_icon_grid.h"
#include "vtk_output.h"
#include "grid.h"
#include "geometry.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

static void write_grid_to_file(struct grid * grid, char * name);
static int file_exists(const char * filename);

#ifdef HAVE_C_NETCDF
int main(void) {

   struct grid * icon_grid;

   if (!file_exists("iconR2B00-grid.nc"))
      return EXIT_SKIP_TEST;

   icon_grid = read_icon_grid("iconR2B00-grid.nc");

   if (yac_get_num_grid_cells(icon_grid) != 80)
      PUT_ERR("wrong number of grid cells");

#define WRITE_VTK_GRID_FILE
#ifdef WRITE_VTK_GRID_FILE
   write_grid_to_file(icon_grid, "icon");
#endif // WRITE_VTK_GRID_FILE

   yac_delete_grid(icon_grid);

   return TEST_EXIT_CODE;
}

static void
get_grid_xyz_coordinates(struct grid * grid, double * points) {

   unsigned i, num_corners;
   double lon, lat;

   num_corners = yac_get_num_grid_corners(grid);

   for (i = 0; i < num_corners; ++i) {
      lon = yac_get_corner_x_coord(grid, i);
      lat = yac_get_corner_y_coord(grid, i);
      LLtoXYZ(lon, lat, points + 3 * i);
   }
}

static void
get_grid_cell_data(struct grid * grid, unsigned * cell_data) {

   unsigned i, j, k;
   unsigned num_grid_cells;

   num_grid_cells = yac_get_num_grid_cells(grid);

   for (i = 0, k = 0; i < num_grid_cells; ++i) {

      unsigned const * curr_cell_corners;
      unsigned num_cell_corners;

      curr_cell_corners = yac_get_cell_corner_indices(grid, i);
      num_cell_corners = yac_get_num_cell_corners(grid, i);

      for (j = 0; j < num_cell_corners; ++j)
         cell_data[k++] = curr_cell_corners[j];
   }
}

static void
rotate_vector(double * v, double * axis, double angle, double * out) {

  double t = 1 - cos(angle);

  double R[3][3] = {{axis[0] * axis[0] * t + cos(angle),
                     axis[0] * axis[1] * t - axis[2] * sin(angle),
                     axis[0] * axis[2] * t + axis[1] * sin(angle)},
                    {axis[1] * axis[0] * t + axis[2] * sin(angle),
                     axis[1] * axis[1] * t + cos(angle),
                     axis[1] * axis[2] * t - axis[0] * sin(angle)},
                    {axis[2] * axis[0] * t - axis[1] * sin(angle),
                     axis[2] * axis[1] * t + axis[0] * sin(angle),
                     axis[2] * axis[2] * t + cos(angle)}};

  for (unsigned i = 0; i < 3; ++i) {
    out[i] = 0;
    for (unsigned j = 0; j < 3; ++j) {
      out[i] += R[i][j] * v[j];
    }
  }
}

static void
get_bounding_circle_point_data(struct bounding_circle bnd_circle,
                               unsigned num_points, double * points) {

  double n_x[3] = {1, 0, 0};
  double n_y[3] = {0, 1, 0};
  double * n;

  n = (get_vector_angle(n_x, bnd_circle.base_vector) >
       get_vector_angle(n_y, bnd_circle.base_vector))?n_x:n_y;

  double temp_vector[3];
  double start_vector[3];

  crossproduct_d(bnd_circle.base_vector, n, temp_vector);
  rotate_vector(bnd_circle.base_vector, temp_vector, bnd_circle.inc_angle,
                start_vector);

  double angle = 2.0 * M_PI / (num_points + 1);

  for (unsigned i = 0; i < num_points; ++i)
    rotate_vector(start_vector, bnd_circle.base_vector, angle * (double)i,
                  points + 3 * i);
}

static void
get_equator_point_data(unsigned num_points, double * points) {

   // the equator is approximated by a polygon with num_points corners

   struct bounding_circle equator;
   equator.base_vector[0] = 0;
   equator.base_vector[1] = 0;
   equator.base_vector[2] = 1;
   equator.inc_angle = M_PI_2;

   get_bounding_circle_point_data(equator, num_points, points);
}

static void
write_grid_to_file(struct grid * grid, char * name) {

   unsigned num_grid_corners;
   unsigned num_grid_cells;

   num_grid_corners = yac_get_num_grid_corners(grid);
   num_grid_cells = yac_get_num_grid_cells(grid);

   unsigned const num_points_per_great_circle = 71;

   //-------------------------------------
   // generate point data for the vtk file
   // it will contain data for all corners of the grid and the points for the equator
   //-------------------------------------
   double * points;

   points = malloc(3 * (num_grid_corners + num_points_per_great_circle) * sizeof(*points));

   // get the point data of the grid
   get_grid_xyz_coordinates(grid, points);
   // get the point data of the equator
   get_equator_point_data(num_points_per_great_circle, points + 3 * num_grid_corners);

   //------------------------------------
   // generate cell data for the vtk file
   //------------------------------------
   unsigned * num_points_per_cell;
   unsigned * cell_data;

   // has entries for all cells of the grid and one for the equator
   num_points_per_cell = malloc((num_grid_cells + 1) * sizeof(*num_points_per_cell));

   unsigned i, total_num_cell_points;

   // get the number of points per grid cell
   for (i = 0; i < num_grid_cells; ++i)
      num_points_per_cell[i] = yac_get_num_cell_corners(grid, i);

   // get the number of corners for the equator
   num_points_per_cell[num_grid_cells] = num_points_per_great_circle;

   // count the total number of cell points
   total_num_cell_points = 0;
   for (i = 0; i <= num_grid_cells; ++i)
      total_num_cell_points += num_points_per_cell[i];

   cell_data = malloc(total_num_cell_points * sizeof(*cell_data));

   // get the cell data of the grid
   get_grid_cell_data(grid, cell_data);

   // write the cell data for the great circle
   for (i = 0; i < num_points_per_great_circle; ++i)
      cell_data[i + total_num_cell_points - num_points_per_great_circle] = num_grid_corners + i;

   //------------------------------------
   // generate scalar data
   // we write a cell type that differentiates grid cells and the equator
   //------------------------------------
   unsigned * cell_type;

   cell_type = malloc((num_grid_cells + 1) * sizeof(*cell_type));

   for (i = 0; i < num_grid_cells; ++i)
      cell_type[i] = 0;
   cell_type[num_grid_cells] = 1;

   //------------------------------------
   // generate the file name
   //------------------------------------

   char * filename;

   filename = calloc((strlen(name) + 5), sizeof(*filename));
   strcpy(filename, name);
   strcat(filename, ".vtk");

   //------------------------------------
   // generate the actual vtk file
   //------------------------------------
   VTK_FILE * file;

   file = vtk_open(filename, "grid data");
   vtk_write_point_data(file, points, num_grid_corners + num_points_per_great_circle);
   vtk_write_cell_data(file, cell_data, num_points_per_cell, num_grid_cells + 1);
   vtk_write_cell_scalars_uint(file, cell_type, num_grid_cells + 1, "cell_type");
   
   vtk_close(file);

   //------------------------------------
   // some final cleanup
   //------------------------------------

   free(filename);
   free(cell_type);
   free(cell_data);
   free(num_points_per_cell);
   free(points);
}

static int
file_exists(const char * filename) {

   FILE * file = fopen(filename, "r");

   if (file != NULL) {
      fclose(file);
      return 1;
   }
   return 0;
}

#else
int main(void) {
    return EXIT_SKIP_TEST;
}
#endif

