/**
 * @file weight_file_common.c
 *
 * @copyright Copyright  (C)  2016 DKRZ, MPI-M
 *
 * @version 1.0
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 *
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *
 * URL: https://doc.redmine.dkrz.de/YAC/html/index.html
 *
 * This file is part of YAC.
 *
 * YAC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAC.  If not, see <http://www.gnu.org/licenses/gpl.txt>.
 *
 */
 
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>

#include "tests.h"
#include "weight_file_common.h"
#include "points.h"
#include "interpolation_method_file.h"

#ifdef HAVE_C_NETCDF

#include <netcdf.h>

static void handle_error(int status) {
  const char *err_string;
  err_string = nc_strerror(status);
  printf ("%s \n", err_string);
  if (status != NC_NOERR) {
    exit(EXIT_FAILURE);
  }
}

void write_weight_file(char const * file_name, int const * src_id,
                       int const * tgt_id, double const * weights,
                       unsigned num_links,
                       enum yac_location const * src_locations,
                       unsigned num_src_fields,
                       int const * num_links_per_src_field,
                       int * tgt_id_fixed, unsigned num_fixed_tgt,
                       double * fixed_values,
                       int * num_tgt_per_fixed_value,
                       unsigned num_fixed_values,
                       enum yac_location tgt_location,
                       char const * src_grid_name,
                       char const * tgt_grid_name) {

  int ncid, status;

  // create file
  status = nc_create(file_name, NC_CLOBBER, &ncid);
  if (status != NC_NOERR) handle_error(status);

  int dim_weight_id[6];

  // define dimensions
  if (num_links > 0) {
    status = nc_def_dim(ncid, "num_links", num_links, &dim_weight_id[0]);
    if (status != NC_NOERR) handle_error(status);
    status = nc_def_dim(ncid, "num_wgts", 1, &dim_weight_id[1]);
    if (status != NC_NOERR) handle_error(status);
  }
  status = nc_def_dim(ncid, "num_src_fields", num_src_fields, &dim_weight_id[2]);
  if (status != NC_NOERR) handle_error(status);
  status = nc_def_dim(ncid, "max_loc_str_len", YAC_MAX_LOC_STR_LEN, &dim_weight_id[3]);
  if (status != NC_NOERR) handle_error(status);
  if (num_fixed_values > 0) {
    status = nc_def_dim(ncid, "num_fixed_values", num_fixed_values,
                        &dim_weight_id[4]);
    status = nc_def_dim(ncid, "num_fixed_dst", num_fixed_tgt, &dim_weight_id[5]);
    if (status != NC_NOERR) handle_error(status);
  }

  int var_src_add_id, var_dst_add_id, var_weight_id, var_num_links_id,
      src_var_locs_id, tgt_var_loc_id, var_fixed_values_id,
      var_num_dst_per_fixed_value_id, var_dst_add_fixed_id;

  // define variables
  if (num_links > 0) {
    status = nc_def_var(ncid, "src_address", NC_INT, 1, dim_weight_id, &var_src_add_id);
    if (status != NC_NOERR) handle_error(status);
    status = nc_def_var(ncid, "dst_address", NC_INT, 1, dim_weight_id, &var_dst_add_id);
    if (status != NC_NOERR) handle_error(status);
    status = nc_def_var(ncid, "remap_matrix", NC_DOUBLE, 2, dim_weight_id, &var_weight_id);
    if (status != NC_NOERR) handle_error(status);
    status = nc_def_var(ncid, "num_links_per_src_field", NC_INT, 1, dim_weight_id + 2, &var_num_links_id);
    if (status != NC_NOERR) handle_error(status);
  }
  status = nc_def_var(ncid, "src_locations", NC_CHAR, 2, &dim_weight_id[2],
                      &src_var_locs_id);
  if (status != NC_NOERR) handle_error(status);
  status = nc_def_var(ncid, "dst_location", NC_CHAR, 1, &dim_weight_id[3],
                      &tgt_var_loc_id);
  if (status != NC_NOERR) handle_error(status);
  if (num_fixed_values > 0) {
    status = nc_def_var(ncid, "fixed_values", NC_DOUBLE, 1, &dim_weight_id[4],
                        &var_fixed_values_id);
    if (status != NC_NOERR) handle_error(status);
    status = nc_def_var(ncid, "num_dst_per_fixed_value", NC_INT, 1, &dim_weight_id[4],
                        &var_num_dst_per_fixed_value_id);
    if (status != NC_NOERR) handle_error(status);
    status = nc_def_var(ncid, "dst_address_fixed", NC_INT, 1, &dim_weight_id[5],
                        &var_dst_add_fixed_id);
    if (status != NC_NOERR) handle_error(status);
  }

  // put attributes
  status = nc_put_att_text(ncid, NC_GLOBAL, "version",
                           strlen(YAC_WEIGHT_FILE_VERSION_STRING),
                           YAC_WEIGHT_FILE_VERSION_STRING);
  if (status != NC_NOERR) handle_error(status);
  status = nc_put_att_text(ncid, NC_GLOBAL, "src_grid_name",
                           strlen(src_grid_name), src_grid_name);
  if (status != NC_NOERR) handle_error(status);
  status = nc_put_att_text(ncid, NC_GLOBAL, "dst_grid_name",
                           strlen(tgt_grid_name), tgt_grid_name);
  if (status != NC_NOERR) handle_error(status);
  {
    char const * str_logical[2] = {"FALSE", "TRUE"};
    status = nc_put_att_text(ncid, NC_GLOBAL, "contains_fixed_dst",
                             strlen(str_logical[num_fixed_values > 0]),
                             str_logical[num_fixed_values > 0]);
    if (status != NC_NOERR) handle_error(status);
    status = nc_put_att_text(ncid, NC_GLOBAL, "contains_links",
                             strlen(str_logical[num_links > 0]),
                             str_logical[num_links > 0]);
    if (status != NC_NOERR) handle_error(status);
  }

  // end definition
  status = nc_enddef(ncid);
  if (status != NC_NOERR) handle_error(status);

  // write mapping data

  if (num_links > 0) {
    int * src_address = malloc(num_links * sizeof(*src_address));
    int * tgt_address = malloc(num_links * sizeof(*tgt_address));

    for (unsigned i = 0; i < num_links; i++) {
      src_address[i] = src_id[i] + 1;
      tgt_address[i] = tgt_id[i] + 1;
    }

    status = nc_put_var_int(ncid, var_src_add_id, src_address);
    if (status != NC_NOERR) handle_error(status);
    status = nc_put_var_int(ncid, var_dst_add_id, tgt_address);
    if (status != NC_NOERR) handle_error(status);

    free(src_address);
    free(tgt_address);

    status = nc_put_var_double(ncid, var_weight_id, weights);
    if (status != NC_NOERR) handle_error(status);

    status = nc_put_var_int(ncid, var_num_links_id, num_links_per_src_field);
    if (status != NC_NOERR) handle_error(status);
  }

  for (unsigned i = 0; i < num_src_fields; ++i) {
    char const * loc_str = yac_loc2str(src_locations[i]);
    size_t str_start[2] = {i, 0};
    size_t str_count[2] = {1, strlen(loc_str)};
    status = nc_put_vara_text(ncid, src_var_locs_id, str_start, str_count,
                              loc_str);
    if (status != NC_NOERR) handle_error(status);
  }

  {
    char const * loc_str = yac_loc2str(tgt_location);
    size_t str_start[1] = {0};
    size_t str_count[1] = {strlen(loc_str)};
    status = nc_put_vara_text(ncid, tgt_var_loc_id, str_start, str_count,
                              loc_str);
  }
  if (status != NC_NOERR) handle_error(status);

  if (num_fixed_values > 0) {

    int * tgt_address_fixed =
      malloc(num_fixed_tgt * sizeof(*tgt_address_fixed));
    for (unsigned i = 0; i < num_fixed_tgt; i++)
      tgt_address_fixed[i] = tgt_id_fixed[i] + 1;

    status = nc_put_var_int(ncid, var_dst_add_fixed_id, tgt_address_fixed);
    if (status != NC_NOERR) handle_error(status);
    status = nc_put_var_double(ncid, var_fixed_values_id, fixed_values);
    if (status != NC_NOERR) handle_error(status);
    status = nc_put_var_int(ncid, var_num_dst_per_fixed_value_id,
                            num_tgt_per_fixed_value);
    if (status != NC_NOERR) handle_error(status);

    free(tgt_address_fixed);
  }

  status = nc_close(ncid);
  if (status != NC_NOERR) handle_error(status);
}

void check_weight_file(char const * file_name, int const * ref_src_address,
                       int const * ref_tgt_address, double const * ref_weights,
                       unsigned ref_num_links,
                       enum yac_location const * ref_src_locations,
                       unsigned ref_num_src_fields,
                       int const * ref_num_links_per_src_field,
                       int const * ref_tgt_address_fixed,
                       unsigned ref_num_fixed_tgt,
                       double const * ref_fixed_values,
                       int const * ref_num_tgt_per_fixed_value,
                       unsigned ref_num_fixed_values,
                       enum yac_location ref_tgt_location,
                       char const * ref_src_grid_name,
                       char const * ref_tgt_grid_name) {

  int ncid, status;

  // open file
  status = nc_open(file_name, NC_NOWRITE, &ncid);
  if (status != NC_NOERR) handle_error(status);

  int dimid, var_id;
  size_t num_wgts;
  status = nc_inq_dimid(ncid, "num_wgts", &dimid);
  if (status != NC_NOERR) handle_error(status);
  status = nc_inq_dimlen(ncid, dimid, &num_wgts);
  if (status != NC_NOERR) handle_error(status);
  if (num_wgts != 1)
      yac_internal_abort_message("ERROR(check_weight_file): test only supports "
                                 "num_wgts == 1", __FILE__, __LINE__);

  // global attributes
  size_t str_version_len;
  char * str_version;
  var_id = NC_GLOBAL;
  status = nc_inq_attlen(ncid, var_id, "version", &str_version_len);
  if (status != NC_NOERR) handle_error(status);
  str_version = malloc(str_version_len + 1);
  str_version[str_version_len] = '\0';
  status = nc_get_att_text(ncid, var_id, "version", str_version);
  if (status != NC_NOERR) handle_error(status);

  size_t str_fixed_len;
  char * str_fixed;
  status = nc_inq_attlen(ncid, var_id, "contains_fixed_dst", &str_fixed_len);
  if (status != NC_NOERR) handle_error(status);
  str_fixed = malloc(str_fixed_len + 1);
  str_fixed[str_fixed_len] = '\0';
  status = nc_get_att_text(ncid, var_id, "contains_fixed_dst", str_fixed);
  if (status != NC_NOERR) handle_error(status);

  unsigned contains_fixed_dst;
  if ((strlen("FALSE") == str_fixed_len) &&
      !strncmp("FALSE", str_fixed, str_fixed_len))
      contains_fixed_dst = 0;
  else if ((strlen("TRUE") == str_fixed_len) &&
      !strncmp("TRUE", str_fixed, str_fixed_len))
      contains_fixed_dst = 1;
  else
    yac_internal_abort_message("ERROR(check_weight_file): invalid global "
                               "attribute contains_fixed_dst", __FILE__,
                               __LINE__);

  char * src_grid_name, * tgt_grid_name;
  size_t src_grid_name_len = 0, tgt_grid_name_len = 0;
  var_id = NC_GLOBAL;
  status = nc_inq_attlen(ncid, var_id, "src_grid_name", &src_grid_name_len);
  if (status != NC_NOERR) handle_error(status);
  src_grid_name = malloc(src_grid_name_len + 1);
  src_grid_name[src_grid_name_len] = '\0';
  status = nc_get_att_text(ncid, var_id, "src_grid_name", src_grid_name);
  if (status != NC_NOERR) handle_error(status);

  status = nc_inq_attlen(ncid, var_id, "dst_grid_name", &tgt_grid_name_len);
  if (status != NC_NOERR) handle_error(status);
  tgt_grid_name = malloc(tgt_grid_name_len + 1);
  tgt_grid_name[tgt_grid_name_len] = '\0';
  status = nc_get_att_text(ncid, var_id, "dst_grid_name", tgt_grid_name);
  if (status != NC_NOERR) handle_error(status);

  size_t str_link_len;
  char * str_link;
  var_id = NC_GLOBAL;
  status = nc_inq_attlen(ncid, var_id, "contains_links", &str_link_len);
  if (status != NC_NOERR) handle_error(status);
  str_link = malloc(str_link_len + 1);
  str_link[str_link_len] = '\0';
  status = nc_get_att_text(ncid, var_id, "contains_links", str_link);
  if (status != NC_NOERR) handle_error(status);

  unsigned contains_links;
  if ((strlen("FALSE") == str_link_len) &&
      !strncmp("FALSE", str_link, str_link_len))
      contains_links = 0;
  else if ((strlen("TRUE") == str_link_len) &&
      !strncmp("TRUE", str_link, str_link_len))
      contains_links = 1;
  else
    yac_internal_abort_message("ERROR(check_weight_file): invalid global "
                               "attribute contains_links", __FILE__, __LINE__);
  free(str_link);
    

  // get number of links from file
  size_t num_links = 0;
  if (contains_links) {
    status = nc_inq_dimid(ncid, "num_links", &dimid);
    if (status != NC_NOERR) handle_error(status);
    status = nc_inq_dimlen(ncid, dimid, &num_links);
    if (status != NC_NOERR) handle_error(status);
    if (num_links == 0)
      yac_internal_abort_message("ERROR(check_weight_file): no links defined",
                                 __FILE__, __LINE__);
  }
  if (ref_num_links != num_links)
    PUT_ERR("wrong number of links in weight file\n");

  // get number of source fields from file
  size_t num_src_fields = 0;
  if (contains_links) {
    status = nc_inq_dimid(ncid, "num_src_fields", &dimid);
    if (status != NC_NOERR) handle_error(status);
    status = nc_inq_dimlen(ncid, dimid, &num_src_fields);
    if (status != NC_NOERR) handle_error(status);
    if (num_src_fields == 0)
        yac_internal_abort_message("ERROR(check_weight_file): no source fields",
                                   __FILE__, __LINE__);
    if (ref_num_src_fields != num_src_fields)
      PUT_ERR("wrong number of source fields in weight file\n");
  }

  // get max location string length from file
  size_t max_loc_str_len;
  status = nc_inq_dimid(ncid, "max_loc_str_len", &dimid);
  if (status != NC_NOERR) handle_error(status);
  status = nc_inq_dimlen(ncid, dimid, &max_loc_str_len);
  if (status != NC_NOERR) handle_error(status);
  if (max_loc_str_len != YAC_MAX_LOC_STR_LEN)
      yac_internal_abort_message("ERROR(check_weight_file): wrong max location "
                                 "string length", __FILE__, __LINE__);

  size_t num_fixed_values = 0;
  size_t num_fixed_tgt = 0;
  if (contains_fixed_dst) {

    // get number of fixed values
    status = nc_inq_dimid(ncid, "num_fixed_values", &dimid);
    if (status != NC_NOERR) handle_error(status);
    status = nc_inq_dimlen(ncid, dimid, &num_fixed_values);
    if (status != NC_NOERR) handle_error(status);

    // get number for fixed target points
    status = nc_inq_dimid(ncid, "num_fixed_dst", &dimid);
    if (status != NC_NOERR) handle_error(status);
    status = nc_inq_dimlen(ncid, dimid, &num_fixed_tgt);
    if (status != NC_NOERR) handle_error(status);
  }

  int * src_address = malloc(num_links * sizeof(*src_address));
  int * tgt_address = malloc(num_links * sizeof(*tgt_address));
  double * weights = malloc(num_links * sizeof(*weights));
  int * num_links_per_src_field =
    malloc(num_src_fields * sizeof(*num_links_per_src_field));
  enum yac_location * src_locations =
    malloc(num_src_fields * sizeof(*src_locations));
  enum yac_location tgt_location;
  double * fixed_values = malloc(num_fixed_values * sizeof(*fixed_values));
  int * num_tgt_per_fixed_value =
    malloc(num_fixed_values * sizeof(*num_tgt_per_fixed_value));
  int * tgt_address_fixed = malloc(num_fixed_tgt * sizeof(*tgt_address_fixed));

  // read data
  if (contains_links) {
    status = nc_inq_varid(ncid, "src_address", &var_id);
    if (status != NC_NOERR) handle_error(status);
    status = nc_get_var_int(ncid, var_id, src_address);
    if (status != NC_NOERR) handle_error(status);

    status = nc_inq_varid(ncid, "dst_address", &var_id);
    if (status != NC_NOERR) handle_error(status);
    status = nc_get_var_int(ncid, var_id, tgt_address);
    if (status != NC_NOERR) handle_error(status);

    status = nc_inq_varid(ncid, "remap_matrix", &var_id);
    if (status != NC_NOERR) handle_error(status);
    status = nc_get_var_double(ncid, var_id, weights);
    if (status != NC_NOERR) handle_error(status);

    status = nc_inq_varid(ncid, "num_links_per_src_field", &var_id);
    if (status != NC_NOERR) handle_error(status);
    status = nc_get_var_int(ncid, var_id, num_links_per_src_field);
    if (status != NC_NOERR) handle_error(status);
  }

  status = nc_inq_varid(ncid, "src_locations", &var_id);
  if (status != NC_NOERR) handle_error(status);
  for (unsigned i = 0; i < num_src_fields; ++i) {
    char loc_str[YAC_MAX_LOC_STR_LEN];
    size_t str_start[2] = {i, 0};
    size_t str_count[2] = {1, YAC_MAX_LOC_STR_LEN};
    status = nc_get_vara_text(ncid, var_id, str_start, str_count, loc_str);
    if (status != NC_NOERR) handle_error(status);
    src_locations[i] = yac_str2loc(loc_str);
  }

  status = nc_inq_varid(ncid, "dst_location", &var_id);
  if (status != NC_NOERR) handle_error(status);
  {
    char loc_str[YAC_MAX_LOC_STR_LEN];
    status = nc_get_var_text(ncid, var_id, loc_str);
    if (status != NC_NOERR) handle_error(status);
    tgt_location = yac_str2loc(loc_str);
  }

  if (contains_fixed_dst) {

    status = nc_inq_varid(ncid, "fixed_values", &var_id);
    if (status != NC_NOERR) handle_error(status);
    status = nc_get_var_double(ncid, var_id, fixed_values);
    if (status != NC_NOERR) handle_error(status);

    status = nc_inq_varid(ncid, "num_dst_per_fixed_value", &var_id);
    if (status != NC_NOERR) handle_error(status);
    status = nc_get_var_int(ncid, var_id, num_tgt_per_fixed_value);
    if (status != NC_NOERR) handle_error(status);

    status = nc_inq_varid(ncid, "dst_address_fixed", &var_id);
    if (status != NC_NOERR) handle_error(status);
    status = nc_get_var_int(ncid, var_id, tgt_address_fixed);
    if (status != NC_NOERR) handle_error(status);
  }

  // check data
  if ((strlen(YAC_WEIGHT_FILE_VERSION_STRING) != str_version_len) ||
      strncmp(YAC_WEIGHT_FILE_VERSION_STRING, str_version, str_version_len))
    PUT_ERR("wrong version string\n");
  if ((strlen(ref_src_grid_name) != src_grid_name_len) ||
      strncmp(ref_src_grid_name, src_grid_name, src_grid_name_len))
    PUT_ERR("wrong src_grid_name\n");
  if ((strlen(ref_tgt_grid_name) != tgt_grid_name_len) ||
      strncmp(ref_tgt_grid_name, tgt_grid_name, tgt_grid_name_len))
    PUT_ERR("wrong tgt_grid_name\n");

  if (contains_links != (ref_num_links > 0)) {
    PUT_ERR("file contains links, but reference data does not\n");
  } else if (contains_links) {
    for (unsigned i = 0; i < MIN(num_links, ref_num_links); ++i)
      if (ref_src_address[i] != src_address[i] - 1)
        PUT_ERR("wrong src_address\n");
    if ((strlen(ref_src_grid_name) != src_grid_name_len) ||
        strncmp(ref_src_grid_name, src_grid_name, src_grid_name_len))
      PUT_ERR("wrong src_grid_name\n");
    for (unsigned i = 0; i < MIN(num_links, ref_num_links); ++i)
      if (ref_tgt_address[i] != tgt_address[i] - 1)
        PUT_ERR("wrong tgt_address\n");
    if ((strlen(ref_tgt_grid_name) != tgt_grid_name_len) ||
        strncmp(ref_tgt_grid_name, tgt_grid_name, tgt_grid_name_len))
      PUT_ERR("wrong src_grid_name\n");
    for (unsigned i = 0; i < MIN(num_links, ref_num_links); ++i)
      if (fabs(ref_weights[i] - weights[i]) > 1e-10)
        PUT_ERR("wrong weight\n");
    for (unsigned i = 0; i < MIN(num_src_fields, ref_num_src_fields); ++i)
      if (ref_num_links_per_src_field[i] != num_links_per_src_field[i])
        PUT_ERR("wrong number of links per source field\n");
  }
  for (unsigned i = 0; i < MIN(num_src_fields, ref_num_src_fields); ++i)
    if (ref_src_locations[i] != src_locations[i])
      PUT_ERR("wrong source location\n");
  if (ref_tgt_location != tgt_location) PUT_ERR("wrong target location\n");

  if (contains_fixed_dst != (ref_num_fixed_values > 0)) {
    PUT_ERR("file contains fixed target data, but reference data does not\n");
  } else if (contains_fixed_dst) {
    if (ref_num_fixed_values != num_fixed_values) {
      PUT_ERR("wrong number of fixed values\n");
    } else {
      unsigned * used_flag = calloc(num_fixed_values, sizeof(*used_flag));
      unsigned match_count = 0;
      for (unsigned i = 0, ref_offset = 0; i < ref_num_fixed_values;
           ref_offset += ref_num_tgt_per_fixed_value[i++]) {
        for (unsigned j = 0, offset = 0; j < num_fixed_values;
             offset += num_tgt_per_fixed_value[j++]) {
          if ((!used_flag[j]) && (ref_fixed_values[i] == fixed_values[j]) &&
              (ref_num_tgt_per_fixed_value[i] == num_tgt_per_fixed_value[j])) {

            for (unsigned k = 0; k < num_tgt_per_fixed_value[j]; ++k) {
              if (ref_tgt_address_fixed[ref_offset + k] !=
                  tgt_address_fixed[offset + k] - 1)
                PUT_ERR("wrong fixed target address\n");
            }

            used_flag[j] = 1;
            match_count++;
            break;
          }
        }
      }
      if (match_count != ref_num_fixed_values)
        PUT_ERR("wrong fixed values data");
      free(used_flag);
    }
  }

  // clean up
  free(str_version);
  free(fixed_values);
  free(num_tgt_per_fixed_value);
  free(tgt_address_fixed);
  free(src_grid_name);
  free(tgt_grid_name);
  free(src_address);
  free(tgt_address);
  free(weights);
  free(num_links_per_src_field);
  free(src_locations);
  free(str_fixed);

  // close file
  status = nc_close(ncid);
  if (status != NC_NOERR) handle_error(status);
}

#else // HAVE_C_NETCDF

void write_weight_file(char const * file_name, int const * src_id,
                       int const * tgt_id, double const * weights,
                       unsigned num_links,
                       enum yac_location const * src_locations,
                       unsigned num_src_fields,
                       int const * num_links_per_src_field,
                       int * tgt_id_fixed, unsigned num_fixed_tgt,
                       double * fixed_values,
                       int * num_tgt_per_fixed_value,
                       unsigned num_fixed_values,
                       enum yac_location tgt_location,
                       char const * src_grid_name,
                       char const * tgt_grid_name) {

  yac_internal_abort_message("ERROR(write_weight_file): YAC compiled with "
                             "netcdf support", __FILE__, __LINE__);
}

void check_weight_file(char const * file_name, int const * ref_src_address,
                       int const * ref_tgt_address, double const * ref_weights,
                       unsigned ref_num_links,
                       enum yac_location const * ref_src_locations,
                       unsigned ref_num_src_fields,
                       int const * ref_num_links_per_src_field,
                       int const * ref_tgt_id_fixed, unsigned ref_num_fixed_tgt,
                       double const * ref_fixed_values,
                       int const * ref_num_tgt_per_fixed_value,
                       unsigned ref_num_fixed_values,
                       enum yac_location ref_tgt_location,
                       char const * ref_src_grid_name,
                       char const * ref_tgt_grid_name) {

  yac_internal_abort_message("ERROR(write_weight_file): YAC compiled with "
                             "netcdf support", __FILE__, __LINE__);
}

#endif
